
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "global.h"
#include "mbuffer.h"
#include "adaptive_loop_filter.h"
#ifndef QC_FILTER_H
#define QC_FILTER_H

#ifdef ADAPTIVE_LOOP_FILTER

#ifdef _WIN32
#define NO_OUTPUT_FILE        ":NUL"
#else
#define NO_OUTPUT_FILE        "/dev/null"
#endif

#define TMP_YUV_FILE          "tmp.yuv"
#define NUM_BITS              9
#define NO_TEST_FILT           3       // Filter supports (5/7/9)
#define NO_PREV_FILTERS       16
#define PREDICT_COEFF          1
#define NO_VAR_BINS           16
#define NO_FILTERS            16
#define FILTER_LENGTH          9
#define SQUARE(X)             ((X) * (X))
#define VAR_SIZE               3
#define MAX_SQR_FILT_LENGTH   (SQUARE(FILTER_LENGTH) / 2 + 2)
#define SQR_FILT_LENGTH_9SYM  (SQUARE(9) / 4 + 2) 
#define SQR_FILT_LENGTH_7SYM  (SQUARE(7) / 4 + 2) 
#define SQR_FILT_LENGTH_5SYM  (SQUARE(5) / 4 + 2) 
#define MAX_SCAN_VAL    11
#define MAX_EXP_GOLOMB  16

#define int8 char

#ifndef max
#define max(a, b)	(((a) > (b))? (a): (b))
#endif

#ifndef min
#define min(a, b)	(((a) < (b))? (a): (b))
#endif

#define ROUND(a)  (((a) < 0)? (int)((a) - 0.5) : (int)((a) + 0.5))

#define FATAL_ERROR_0(MESSAGE, EXITCODE)                                \
          {                                                             \
          printf(MESSAGE);                                            \
          exit(EXITCODE);                                             \
          }

#define FATAL_ERROR_1(MESSAGE, PARAM_1, EXITCODE)                       \
          {                                                             \
          printf(MESSAGE, PARAM_1);                                   \
          exit(EXITCODE);                                             \
          }

#define FATAL_ERROR_2(MESSAGE, PARAM_1, PARAM_2, EXITCODE)              \
          {                                                             \
          printf(MESSAGE, PARAM_1, PARAM_2);                          \
          exit(EXITCODE);                                             \
          }


int    InitPFGlobalBuffers(alf_parameter_set_rbsp_t *alfps);
void   FreePFGlobalBuffers(alf_parameter_set_rbsp_t *alfps);
void   SetPFParameters(pfStruct *p , pfStruct *pf_in);



int golombDecode(int k, Bitstream *bitstream, char *tracestring);


void decodeAuxInfo(pfStruct *pf, int frNo, Bitstream *bitstream);
void readPredFlags(pfStruct *pf, Bitstream *bitstream);
void readFilterCodingParams(pfStruct *pf, Bitstream *bitstream);
void readFilterCoeffs(pfStruct *pf,Bitstream *bitstream);


typedef struct
{
  imgpel **imgY;
  imgpel ***imgUV;
  int img_height;
  int img_width;
} imgStruct;




//
//  memalloc.c
//
void initMatrix_int(int ***m2D, int d1, int d2);
void initMatrix_double(double ***m2D, int d1, int d2);
void initMatrix3D_int(int ****m3D, int d1, int d2, int d3);
void initMatrix3D_double(double ****m3D, int d1, int d2, int d3);
void initMatrix4D_int(int *****m4D, int d1, int d2, int d3, int d4);
void initMatrix4D_double(double *****m4D, int d1, int d2, int d3, int d4);
void destroyMatrix_int(int **m2D);
void destroyMatrix_double(double **m2D);
void destroyMatrix3D_int(int ***m3D, int d1);
void destroyMatrix3D_double(double ***m3D, int d1);
void destroyMatrix4D_int(int ****m4D, int d1, int d2);
void destroyMatrix4D_double(double ****m4D, int d1, int d2);

void initMatrix_imgpel(imgpel ***m2D, int d1, int d2);
void initMatrix3D_imgpel(imgpel ****m3D, int d1, int d2, int d3);
void initMatrix4D_imgpel(imgpel *****m4D, int d1, int d2, int d3, int d4);
void destroyMatrix_imgpel(imgpel **m2D);
void destroyMatrix3D_imgpel(imgpel ***m3D, int d1);
void destroyMatrix4D_imgpel(imgpel ****m4D, int d1, int d2);
// 
//  misc.c
//
int Clip3Fun_post(int low, int high, int val);
imgpel Clip_post(int high, int val);
void int2Byte(imgStruct *imgIn, int width, int height);
void initImg(imgStruct *imgIn, int height, int width);
void padImage(imgpel **imgY,  imgpel **imgY_pad, int fl, int img_height, int img_width);
void calcVar(imgpel **imgY_var, imgpel **imgY_pad, int pad_size, int fl, int img_height, int img_width);

void calcVar_block(imgpel **imgY_var, imgpel **imgY_pad, int pad_size, int fl, int img_height, int img_width,
                   int img_height_start, int img_height_end, int img_width_start,int img_width_end,int blk_addr,StorablePicture *p,int num_blk);
void init_filterCoeff();
void delete_filterCoeff();
void reconstructFilterCoeffs(pfStruct *pf,int **pfilterCoeffSym, int bit_depth);
// 
//  decVarPostprocessing.c
//

double decFilterVarPred(imgpel **imgY_rec, imgpel **imgY_org, imgpel **varImg, int filtNo_best, double **filterCoeffSym, 
                        int varIndTab[NO_VAR_BINS], int usePrevFiltBest[NO_VAR_BINS], int prevFiltInd, int codedFlag);



int read_ALFLumaParam_qc(alf_parameter_set_rbsp_t *alfps, Bitstream *s,int inloop);
void setQCfilterImgSize(int height,int width);
void readPFbitstream (Bitstream *bitstream,int frNo, int img_height, int img_width);
void FilteringProcess_qc( imgpel **ImgDec, imgpel **ImgRest,pfStruct *pf);
void FilteringProcessBlock_qc_consecutive(int blk_addr, imgpel** ImgDec, imgpel** ImgRest, int blk_size, int width_in_alf_blks,StorablePicture *p,int num_blk);
void DecFilter_qc(imgpel** ImgDec,pfStruct *pf,int inloop);
void setFilterStream_qc(pfStruct *pf,int inloop);
void resetDecFilterFrNo();
void resetFilterFrNo();
#endif
imgpel   **imgY_pad;
int **imgY_temp;

int mult_fact_int;
int var_max;
#endif 

