#include "alf_filter.h"
#include "adaptive_loop_filter.h"
#include "vlc.h"
#include "memalloc.h"


#ifdef ADAPTIVE_LOOP_FILTER

static int first = 1;
static imgpel **imgY_var;
int img_height,img_width;
static int **filterCoeffSym;
static int *filterCoeffDefault, ***filterCoeffPrev, **filterCoeffPrevSelected;
static int availableFilters;
static int g_frNo=0,g_frNoDec=0;

extern StorablePicture *dec_picture;

int pattern9x9Sym[] = 
{
  0,
  1,  2,  3,
  4,  5,  6,  7,  8,
  9, 10, 11, 12, 13, 14, 15,
  16, 17, 18, 19, 20, 19, 18, 17, 16,
  15, 14, 13, 12, 11, 10,  9, 
  8,  7,  6,  5,  4,
  3,  2,  1,
  0
};

int weights9x9Sym[] = 
{
  2,
  2,  2,  2,   
  2,  2,  2,  2,  2, 
  2, 2,  2,  2,  2,  2,  2,  
  2,  2, 2,  2,  1,  1
};

int pattern9x9Sym_Quart[] = 
{
  0,  0,  0,  0,  1,  0,  0,  0,  0,
  0,  0,  0,  2,  3,  4,  0,  0,  0,
  0,  0,  5,  6,  7,  8,  9,  0,  0,  
  0, 10, 11, 12, 13, 14, 15, 16,  0,
  17, 18, 19, 20, 21, 22
};

int pattern7x7Sym[] = 
{
  0,
  1,  2,  3,
  4,  5,  6,  7,  8,
  9, 10, 11, 12, 11, 10, 9,
  8,  7,  6,  5,  4,
  3,  2,  1,
  0
};

int weights7x7Sym[] = 
{
  2,  
  2,  2,  2,   
  2,  2,  2,  2,  2,    
  2,  2,  2,  1,  1
};

int pattern7x7Sym_Quart[] = 
{
  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  1,  0,  0,  0,  0,
  0,  0,  0,  2,  3,  4,  0,  0,  0,
  0,  0,  5,  6,  7,  8,  9,  0,  0,  
  0, 10, 11, 12, 13, 14,  
};

int pattern5x5Sym[] = 
{
  0,
  1,  2,  3,
  4,  5,  6,  5,  4,
  3,  2,  1,
  0
};

int weights5x5Sym[] = 
{
  2, 
  2, 2, 2,
  2, 2, 1, 1
};

int pattern5x5Sym_Quart[] = 
{
  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  1,  0,  0,  0,  0,
  0,  0,  0,  2,  3,  4,  0,  0,  0,
  0,  0,  5,  6,  7,  8,  0,  0,  0,  
};

int pattern9x9Sym_9[] = 
{
  4,  
  12, 13, 14,  
  20, 21, 22, 23, 24, 
  28, 29, 30, 31, 32, 33, 34,      
  36, 37, 38, 39, 40, 39, 38, 37, 36, 
  34, 33, 32, 31, 30, 29, 28,  
  24, 23, 22, 21, 20, 
  14, 13, 12, 
  4,  
};
int pattern9x9Sym_7[] = 
{    
  13,   
  21, 22, 23,  
  29, 30, 31, 32, 33,       
  37, 38, 39, 40, 39, 38, 37,  
  33, 32, 31, 30, 29,   
  23, 22, 21,  
  13  

};

int pattern9x9Sym_5[] = 
{
  22, 
  30, 31, 32,    
  38, 39, 40, 39, 38, 
  32, 31, 30, 
  22,  
};
static int *patternTab_filt[NO_TEST_FILT]={pattern9x9Sym_9, pattern9x9Sym_7, pattern9x9Sym_5}; 
static int flTab[]={9/2, 7/2, 5/2};


int InitPFGlobalBuffers(alf_parameter_set_rbsp_t *alfps)
{
  // allocate PF parameter set
  if ((alfps->active_ilf=calloc (sizeof (pfStruct), 1)) == NULL)
    no_mem_exit ("InitPFGlobalBuffers: active_ilf");
  get_mem1Dint(&(alfps->active_ilf->kMinTab), MAX_SQR_FILT_LENGTH);
  get_mem1Dint(&(alfps->active_ilf->codedVarBins), NO_VAR_BINS);
  get_mem1Dint(&(alfps->active_ilf->varIndTab), NO_VAR_BINS);
  get_mem1Dint(&(alfps->active_ilf->usePrevFiltBest), NO_VAR_BINS);
  initMatrix_int(&(alfps->active_ilf->diffQFilterCoeffIntPP), NO_VAR_BINS, MAX_SQR_FILT_LENGTH);
  initMatrix_imgpel(&imgY_var, img_height, img_width); 
  initMatrix_imgpel(&imgY_pad, img_height+2*(FILTER_LENGTH/2), img_width+2*(FILTER_LENGTH/2));


  initMatrix_int(&imgY_temp, img_height+2*VAR_SIZE+3, img_width+2*VAR_SIZE+3);

  init_filterCoeff();

  alfps->active_ilf->filtNo_best = 0;
  alfps->active_ilf->prevFiltInd = 0;


  return 0;
}

void FreePFGlobalBuffers(alf_parameter_set_rbsp_t *alfps)
{
  pfStruct *p = alfps->active_ilf;
  assert(alfps != NULL);


  destroyMatrix_imgpel(imgY_var); 
  destroyMatrix_imgpel(imgY_pad);
  destroyMatrix_int(imgY_temp);
  delete_filterCoeff();

  if(p->kMinTab != NULL)
  {
    free_mem1Dint (p->kMinTab);
    p->kMinTab=NULL;
  }
  if(p->codedVarBins != NULL)
  {
    free_mem1Dint (p->codedVarBins);
    p->codedVarBins=NULL;
  }
  if(p->varIndTab != NULL)
  {
    free_mem1Dint (p->varIndTab);
    p->varIndTab=NULL;
  }
  if(p->usePrevFiltBest != NULL)
  {
    free_mem1Dint (p->usePrevFiltBest);
    p->usePrevFiltBest=NULL;
  }
  destroyMatrix_int(p->diffQFilterCoeffIntPP);
  free (p);
  p=NULL;

}



void SetPFParameters(pfStruct *p, pfStruct *pf_in)
{
  int i;
  p->filtNo_best = pf_in->filtNo_best;
  p->prevFiltInd = pf_in->prevFiltInd;
  p->sqrFiltLength = pf_in->sqrFiltLength;
  p->nonZeroCoeffs = pf_in->nonZeroCoeffs;
  p->filters_per_group = pf_in->filters_per_group;
  p->force0 = pf_in->force0;
  p->predMethod = pf_in->predMethod;
  p->minKStart = pf_in->minKStart;
  p->savedFilters = pf_in->savedFilters;
  memcpy(p->kMinTab, pf_in->kMinTab, sizeof(int)*MAX_SQR_FILT_LENGTH);
  memcpy(p->codedVarBins, pf_in->codedVarBins, sizeof(int)*NO_VAR_BINS);
  memcpy(p->varIndTab, pf_in->varIndTab, sizeof(int)*NO_VAR_BINS);
  memcpy(p->usePrevFiltBest, pf_in->usePrevFiltBest, sizeof(int)*NO_VAR_BINS);

  for(i = 0; i < NO_VAR_BINS; i++) 
    memcpy((p->diffQFilterCoeffIntPP[i]),(pf_in->diffQFilterCoeffIntPP[i]),sizeof(int)*MAX_SQR_FILT_LENGTH);
}

void getCurrentFilter(int **filterCoeffSym,pfStruct *pf)
{ 
  int i,  k, varInd;
  int *patternMap, *patternMapTab[NO_TEST_FILT]={pattern9x9Sym_Quart, pattern7x7Sym_Quart, pattern5x5Sym_Quart};
  static int first=1;
#if PREDICT_COEFF==1
  if (g_frNoDec!=0)
  {
    for(varInd=0; varInd<NO_VAR_BINS; ++varInd)
    {		
      if (pf->usePrevFiltBest[varInd]==1)
      {
        memcpy(filterCoeffPrevSelected[varInd],filterCoeffPrev[pf->prevFiltInd][varInd],sizeof(int)*MAX_SQR_FILT_LENGTH);
      }
      else
      {
        memcpy(filterCoeffPrevSelected[varInd],filterCoeffDefault,sizeof(int)*MAX_SQR_FILT_LENGTH);
      }
    }
    patternMap=patternMapTab[pf->filtNo_best];
    for(varInd=0; varInd<NO_VAR_BINS; ++varInd)
    {		
      k=0;
      for(i = 0; i < MAX_SQR_FILT_LENGTH; i++)
      {
        if (patternMap[i]>0)
        {
          filterCoeffPrevSelected[varInd][i]+=filterCoeffSym[pf->varIndTab[varInd]][k];
          k++;
        }
        else
        {
          filterCoeffPrevSelected[varInd][i]=0;
        }
      }
    }
  }
  else
#endif
  {
    availableFilters=0;
    for(varInd=0; varInd<NO_VAR_BINS; ++varInd)
    {		
      memset(filterCoeffPrevSelected[varInd],0,sizeof(int)*MAX_SQR_FILT_LENGTH);
    }
    patternMap=patternMapTab[pf->filtNo_best];
    for(varInd=0; varInd<NO_VAR_BINS; ++varInd)
    {		
      k=0;
      for(i = 0; i < MAX_SQR_FILT_LENGTH; i++)
      {
        if (patternMap[i]>0)
        {
          filterCoeffPrevSelected[varInd][i]=filterCoeffSym[pf->varIndTab[varInd]][k];
          k++;
        }
        else
        {
          filterCoeffPrevSelected[varInd][i]=0;
        }
      }
    }
  }
}


void resetDecFilterFrNo()
{
  g_frNoDec =0;
}

void resetFilterFrNo()
{
  g_frNo =0;
}

void storeDecodeFilter(int codedFlag)
{
  int varInd;
  //  int i;

  if (dec_picture->slice_type==I_SLICE)
  {
    g_frNoDec =0;
    availableFilters =0;
  }

  if(codedFlag)
  {
    for(varInd=0; varInd<NO_VAR_BINS; ++varInd)
    {
      memcpy(filterCoeffPrev[g_frNoDec%NO_PREV_FILTERS][varInd],filterCoeffPrevSelected[varInd],sizeof(int)*MAX_SQR_FILT_LENGTH);
    }
    availableFilters++; g_frNoDec++;
    availableFilters=min(NO_PREV_FILTERS-1,availableFilters);
  }
}


double filterFrame(imgpel **imgY_rec_post, imgpel **imgY_rec, imgpel **varImg,int filtNo)
{
  int i, j, ii, jj, pixelInt,m=0;
  imgpel *p_imgY_var,*p_imgY_pad;
  int max_val=(1 << (img->bitdepth_luma))-1;

  int fl=FILTER_LENGTH/2;
  int *pattern=pattern9x9Sym;
  int fl_temp;
  int last_coef= MAX_SQR_FILT_LENGTH-1;
  imgpel *im1,*im2;
  int *coef;
  int num_bits_minus_1= NUM_BITS-1;
  int offset = (1<<(NUM_BITS-2));
  int *pattern_fix=patternTab_filt[filtNo];
  fl_temp=flTab[filtNo];


  // Filter
  for (i = fl; i < img_height+fl; i++)
  {
    p_imgY_var = imgY_var[i-fl];
    p_imgY_pad = imgY_pad[i];
    for (j = fl; j < img_width+fl; j++)
    {
      coef = filterCoeffPrevSelected[*(p_imgY_var++)];
      pixelInt=coef[last_coef];
      pattern=pattern_fix;

      for (ii=-fl_temp, m = 0; ii<0; ii++,m++)
      {
        im1= &(imgY_pad[i+ii][j-m]);
        im2= &(imgY_pad[i-ii][j+m]);
        for (jj=-m; jj<=m; jj++)
          pixelInt+=((*(im1++)+ *(im2--))*coef[*(pattern++)]);
      }

      im1= &(p_imgY_pad[j-fl_temp]);
      im2= &(p_imgY_pad[j+fl_temp]);	
      for (jj=0; jj<fl_temp; jj++)
        pixelInt+=((*(im1++)+ *(im2--))*coef[*(pattern++)]);

      pixelInt+=(p_imgY_pad[j]*coef[*(pattern++)]);
      pixelInt=(int)((pixelInt+offset) >> (num_bits_minus_1));
      imgY_rec_post[i-fl][j-fl]=max(0, min(pixelInt,max_val));

    }
  }
  return 0;
}



void setQCfilterImgSize(int height,int width)
{
  img_height=height;
  img_width=width;
}


void DecFilter_qc(imgpel** imgY_rec,pfStruct *pf,int inloop)
{
  int i;
  int numBits = NUM_BITS; 
  int fl=FILTER_LENGTH/2;
  static int first =1;
  int **pfilterCoeffSym;
  pfilterCoeffSym= filterCoeffSym;


  if(pf->sqrFiltLength == SQR_FILT_LENGTH_9SYM)
    pf->filtNo_best = 0;
  else if(pf->sqrFiltLength == SQR_FILT_LENGTH_7SYM)
    pf->filtNo_best = 1;
  else
    pf->filtNo_best = 2;

  if(pf->nonZeroCoeffs)
  {
    //// Reconstruct filter coefficients
    reconstructFilterCoeffs( pf, pfilterCoeffSym, numBits);
  }
  else
  {

    for(i = 0; i < NO_VAR_BINS; i++)
    {
      pf->varIndTab[i] = 0;
      memset(pfilterCoeffSym[i],0,sizeof(int)*MAX_SQR_FILT_LENGTH);
    }
  }
  getCurrentFilter(pfilterCoeffSym,pf);
  padImage(imgY_rec, imgY_pad, fl, img_height, img_width);
  memset(imgY_temp[0],0,sizeof(int)*(img_height+2*VAR_SIZE)*(img_width+2*VAR_SIZE));


  if (first)
  {
    int mult_fact_int_tab[4]= {1,114,41,21};
    mult_fact_int = mult_fact_int_tab[VAR_SIZE];
    var_max= NO_VAR_BINS-1;
    first=0;
  }
  calcVar(imgY_var, imgY_pad, fl, VAR_SIZE, img_height, img_width);
}


void setFilterStream_qc(pfStruct *pf,int inloop)
{
  storeDecodeFilter(pf->nonZeroCoeffs);

}


/*!
*****************************************************************************************
* \brief
*    filtering process per block
*****************************************************************************************
*/

int read_ALFLumaParam_qc(alf_parameter_set_rbsp_t *alfps, Bitstream *bitstream,int inloop)
{
  int i;
  int numBits = NUM_BITS; 
  pfStruct *pf;
  int **pfilterCoeffSym;
  int *pfrNo;
  if(first)
  {
    initMatrix_int(&filterCoeffSym, NO_VAR_BINS, MAX_SQR_FILT_LENGTH);
    initMatrix3D_int(&filterCoeffPrev, NO_PREV_FILTERS, NO_VAR_BINS, MAX_SQR_FILT_LENGTH); 
    initMatrix_int(&filterCoeffPrevSelected, NO_VAR_BINS, MAX_SQR_FILT_LENGTH); 
    filterCoeffDefault = (int *) calloc(MAX_SQR_FILT_LENGTH, sizeof(int));
    filterCoeffDefault[MAX_SQR_FILT_LENGTH - 2] = (1 << (NUM_BITS - 1));
    first = 0;
  }

  {
    pf = alfps->active_ilf;
    pfilterCoeffSym = filterCoeffSym;
#if PREDICT_COEFF==1
    if(img->type==I_SLICE)
    {
      g_frNo=0;
    }
#endif
    pfrNo = &g_frNo;
  }
  //// zero filterCoeffSym - debugging purposes
  for(i = 0; i < NO_VAR_BINS; i++)
  {
    memset(pfilterCoeffSym[i],0,sizeof(int)*MAX_SQR_FILT_LENGTH);
    memset(pf->diffQFilterCoeffIntPP[i],0,sizeof(int)*MAX_SQR_FILT_LENGTH);
  }
  // Read bitstream
  decodeAuxInfo(pf,  *pfrNo, bitstream);

#if PREDICT_COEFF==1
  (*(pfrNo))++;
#endif

  if(pf->nonZeroCoeffs)
  {
    if(pf->filters_per_group > 1)
    {
      readPredFlags(pf, bitstream);
    }
    else
    {
      pf->force0 = 0;
      pf->predMethod = 0;
    }
    readFilterCodingParams(pf, bitstream);
    if(pf->force0)
    {
      pf->savedFilters = 0;
      for(i = 0; i < pf->filters_per_group; ++i)
        pf->savedFilters += pf->codedVarBins[i];
    }
    else
    {
      pf->savedFilters = pf->filters_per_group;
    }
    readFilterCoeffs(pf, bitstream);
  }
  if(!inloop)
  {
    // padding
    if (bitstream->frame_bitoffset%8 != 0 )
    {
      while(bitstream->frame_bitoffset%8 != 0)
      {
        u_1("ALFPS: padding bits", bitstream);
      }
    }
  }

  return 0;
}

void FilteringProcessBlock_qc_consecutive(int blk_addr, imgpel** ImgDec, imgpel** ImgRest, int blk_size, int width_in_alf_blks,StorablePicture *p,int num_blk)
{
  int i, j, ii, jj, pixelInt,m;
  imgpel *p_imgY_var,*p_imgY_pad;

  int max_val=(1 << (img->bitdepth_luma))-1;

  int *coef;
  int num_bits_minus_1= NUM_BITS-1;
  int offset=(1<<(NUM_BITS - 2));
  imgpel *im1,*im2;

  int fl_temp = flTab[p->active_ilf->filtNo_best];
  int *pattern=patternTab_filt[p->active_ilf->filtNo_best];
  int *pattern_fix=pattern;

  int fl=FILTER_LENGTH/2;
  int last_coef= MAX_SQR_FILT_LENGTH-1;
  int blk_x, blk_y, max_blk_x, max_blk_y;

  GetSamplePosForBALF(blk_addr, &blk_x, &blk_y, blk_size, width_in_alf_blks);

  max_blk_x = blk_x + blk_size*num_blk;
  max_blk_y = blk_y + blk_size;

  if(max_blk_x > img_width)  max_blk_x = img_width;
  if(max_blk_y > img_height) max_blk_y = img_height;


  calcVar_block(imgY_var,imgY_pad,fl, VAR_SIZE, img_height, img_width, blk_y, max_blk_y, blk_x,max_blk_x,blk_addr,p,num_blk);

  for (i = blk_y+fl; i < max_blk_y+fl; i++)
  {
    p_imgY_var= &(imgY_var[i-fl][blk_x]);
    p_imgY_pad= imgY_pad[i];
    for (j = blk_x+fl; j < max_blk_x+fl; j++)
    {
      coef = filterCoeffPrevSelected[*(p_imgY_var++)];
      pixelInt=coef[last_coef];
      pattern=pattern_fix;  
      for (ii=-fl_temp, m = 0; ii<0; ii++,m++)
      {
        im1= &(imgY_pad[i+ii][j-m]);
        im2= &(imgY_pad[i-ii][j+m]);
        for (jj=-m; jj<=m; jj++)
          pixelInt+=((*(im1++)+ *(im2--))*coef[*(pattern++)]);
      }
      im1= &(p_imgY_pad[j-fl_temp]); 
      im2= &(p_imgY_pad[j+fl_temp]);	
      for (jj=0; jj<fl_temp; jj++)
        pixelInt+=((*(im1++)+ *(im2--))*coef[*(pattern++)]);

      pixelInt+=(p_imgY_pad[j]*coef[*(pattern++)]);
      pixelInt=(int)((pixelInt+offset) >> (num_bits_minus_1));

      ImgRest[i-fl][j-fl]=max(0, min(pixelInt,max_val));
    }
  }

}



void FilteringProcess_qc( imgpel **ImgDec, imgpel **ImgRest,pfStruct *pf)
{
  int inloop =1;

  DecFilter_qc(ImgDec,pf,inloop);

  calcVar(imgY_var, imgY_pad,FILTER_LENGTH/2, VAR_SIZE, img_height, img_width);

  filterFrame(ImgRest, ImgDec, imgY_var,pf->filtNo_best);

#if PREDICT_COEFF==1 
  setFilterStream_qc(pf,inloop);
#endif
}


#endif