#ifndef _QC_FILTER_COMMON_H_
#define _QC_FILTER_COMMON_H_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "global.h"
#include "vlc.h"
#include "mbuffer.h"
#include "switched_filters.h"
 
#ifdef ADAPTIVE_LOOP_FILTER
#define MIN_BLOCK_IND          2

#define FORCE_ZERO             1
#define PREDICT_COEFF          1

#define NUM_BITS               9
#define NUM_BITS_POST          10

#define NO_TEST_FILT           3       // Filter supports (5/7/9)

#define NO_VAR_BINS           16 
#define NO_PREV_FILTERS       16
#define NO_FILTERS            16
#define VAR_SIZE               3
#define FILTER_LENGTH          9
#define SQUARE(X)             ((X) * (X))

#define MAX_SQR_FILT_LENGTH   (SQUARE(FILTER_LENGTH) / 2 + 2)
#define SQR_FILT_LENGTH_9SYM  (SQUARE(9) / 4 + 2) 
#define SQR_FILT_LENGTH_7SYM  (SQUARE(7) / 4 + 2) 
#define SQR_FILT_LENGTH_5SYM  (SQUARE(5) / 4 + 2) 
#define MAX_SCAN_VAL    11
#define MAX_EXP_GOLOMB  16

#define ROUND(a)  (((a) < 0)? (int)((a) - 0.5) : (int)((a) + 0.5))


#define min(a, b) (((a) < (b)) ? (a) : (b))
#define FATAL_ERROR_0(MESSAGE, EXITCODE)                                \
          {                                                             \
            printf(MESSAGE);                                            \
            exit(EXITCODE);                                             \
          }

#define FATAL_ERROR_1(MESSAGE, PARAM_1, EXITCODE)                       \
          {                                                             \
            printf(MESSAGE, PARAM_1);                                   \
            exit(EXITCODE);                                             \
          }

#define FATAL_ERROR_2(MESSAGE, PARAM_1, PARAM_2, EXITCODE)              \
          {                                                             \
            printf(MESSAGE, PARAM_1, PARAM_2);                          \
            exit(EXITCODE);                                             \
          }



void initMatrix_int(int ***m2D, int d1, int d2);
void initMatrix_double(double ***m2D, int d1, int d2);
void initMatrix_ushort(unsigned short ***m2D, int d1, int d2);
void initMatrix3D_int(int ****m3D, int d1, int d2, int d3);
void initMatrix3D_double(double ****m3D, int d1, int d2, int d3);
void initMatrix3D_ushort(unsigned short ****m3D, int d1, int d2, int d3);
void initMatrix4D_int(int *****m4D, int d1, int d2, int d3, int d4);
void initMatrix4D_double(double *****m4D, int d1, int d2, int d3, int d4);
void initMatrix4D_ushort(unsigned short *****m4D, int d1, int d2, int d3, int d4);


void destroyMatrix_int(int **m2D);
void destroyMatrix_double(double **m2D);
void destroyMatrix_ushort(unsigned short **m2D);
void destroyMatrix3D_int(int ***m3D, int d1);
void destroyMatrix3D_double(double ***m3D, int d1);
void destroyMatrix3D_ushort(unsigned short ***m3D, int d1);
void destroyMatrix4D_int(int ****m4D, int d1, int d2);
void destroyMatrix4D_double(double ****m4D, int d1, int d2);
void destroyMatrix4D_ushort(unsigned short ****m4D, int d1, int d2);



// 
//  cholesky.c
//

double calculateErrorCoeffProvidedInt(double **A, double *b, int *c, int size,int *pattern);
double calculateErrorCoeffProvided(double **A, double *b, double *c, int size);



int gnsSolveByChol(double **LHS, double *rhs, double *x, int noEq);
double calculateError(double **A, double *b, int size);
double calculateErrorAbs(double **A, double *b, double y, int size);



double mergeFiltersGreedy(double **yGlobalSeq, double ***EGlobalSeq, double *pixAccGlobalSeq, 
                          int intervalBest[NO_VAR_BINS][2], int sqrFiltLength, int noIntervals);
double findFilterGroupingError(double ***EGlobalSeq, double **yGlobalSeq, double *pixAccGlobalSeq, 
                               int intervalBest[NO_VAR_BINS][2], int sqrFiltLength, int filters_per_fr);


double findFilterCoeff(double ***EGlobalSeq, double **yGlobalSeq, double *pixAccGlobalSeq, int **filterCoeffSeq, 
                       int **filterCoeffQuantSeq, int intervalBest[NO_VAR_BINS][2], int varIndTab[NO_VAR_BINS], 
                       int sqrFiltLength, int filters_per_fr, int *weights, int bit_depth, 
                       double errorTabForce0Coeff[NO_VAR_BINS][2]);

// 
//  calcBits.c
//

int codeAuxInfo(int filtNo, int noFilters, int varIndTab[NO_VAR_BINS], int usePrevFiltBest[NO_VAR_BINS], 
                int prevFiltInd, int frNo, Bitstream *bitstream, int createBitstream,int realfiltNo);

int sendAllFiltersPPPred(int **FilterCoeffQuant, int fl, int sqrFiltLength,  int filters_per_group, 
                         Bitstream *bitstream, int createBistream);
int sendAllFiltersPPPredForce0(int **FilterCoeffQuant, int fl, int sqrFiltLength, int filters_per_group, 
                               int codedVarBins[NO_VAR_BINS], Bitstream *bitstream, int createBistream);

double findBestCoeffCodMethod(Bitstream *bitstream, int codedVarBins[NO_VAR_BINS], int *forceCoeff0, 
                              int **filterCoeffSymQuant, int fl, int sqrFiltLength, 
                              int filters_per_fr, double errorForce0CoeffTab[NO_VAR_BINS][2], 
                              double *errorQuant, double lambda);





void padImage(imgpel **imgY,  imgpel **imgY_pad, int fl, int img_height, int img_width);
void  calcVar(imgpel **imgY_var, imgpel **imgY_rec, int fl, int img_height, int img_width);
Bitstream * allocBitstream(int size);
int freeBitstream(Bitstream * bitstream);


#endif
#endif
