//
//
//

#include "alf_filter_common.h"
#include "vlc.h"
#include "cabac.h" // POST_FILTER - bitstream
#ifdef ADAPTIVE_LOOP_FILTER
int depthInt9x9Sym[] = 
{
              5, 
           5, 6, 5, 
        5, 6, 7, 6, 5,
     5, 6, 7, 8, 7, 6, 5,
  5, 6, 7, 8, 9, 9 
};

int depthInt7x7Sym[] = 
{
           4, 
        4, 5, 4, 
     4, 5, 6, 5, 4, 
  4, 5, 6, 7, 7 
};

int depthInt5x5Sym[] = 
{
        3,   
     3, 4, 3,
  3, 4, 5, 5  
};

int *pDepthIntTab[3]={
  depthInt5x5Sym, 
  depthInt7x7Sym, 
  depthInt9x9Sym, 
};



// POST_FILTER - bitstream
extern Bitstream *varFiltStream;
extern int useVarFilter;

//
// from http://en.wikipedia.org/wiki/Golomb_coding
// M=2^k
//
int golombEncode(int coeff, int k, Bitstream *bitstream, char *tracestring)
{
  int q, i, m;
  int len = 0;
  int symbol = abs(coeff);

  m = (int)pow(2.0, k);
  q = symbol / m;

  for (i = 0; i < q; i++)
    len += u_1(tracestring, 1, bitstream);    // write q ones
  len += u_1(tracestring, 0, bitstream);      // write one zero

  for(i = 0; i < k; i++)
  {
    len += u_1(tracestring, symbol & 0x01, bitstream); 
    symbol >>= 1;
  }

  if(coeff != 0)
  {
    int sign = (coeff > 0)? 1: 0;
    len += u_1(tracestring, sign, bitstream); 
  }
  return(len);
}

//
//
//
int lengthGolomb(int coeffVal, int k)
{
  int m = 2 << (k - 1);
  int q = coeffVal / m;
  if(coeffVal != 0)
    return(q + 2 + k);
  else
    return(q + 1 + k);
}


//
//  force0: 
//    0 - normal
//    1 - force0 flag followed by codedVarBins[0...filters_per_group-1]
//
//  predMethod:
//    0 - filter coefficients non predicted
//    1 - filter coefficients predicted
//
int writePredFlags(int force0, int predMethod, int codedVarBins[NO_VAR_BINS], 
                   int filters_per_group, Bitstream *bitstream, int createBitstream)
{
  int bit_cnt = 0;
  int ind;

  if(createBitstream)
  {
    bit_cnt += u_1("force0", force0, bitstream);
    if(force0)
      for(ind = 0; ind < filters_per_group; ++ind)
        bit_cnt += u_1("codedVarBins[i]", codedVarBins[ind], bitstream);
    bit_cnt += u_1("predMethod", predMethod, bitstream);
  }
  else
  {
    if(force0)
      bit_cnt = 2 + filters_per_group;
    else
      bit_cnt = 2;
  }
  return bit_cnt;
}


int writeFilterCodingParams(int minKStart, int maxScanVal, int kMinTab[], Bitstream *bitstream, int createBistream)
{
  int scanPos;
  int golombIndexBit;
  int bit_cnt = 0;
  int kMin;

  if(createBistream)
  {
    // Golomb parameters
    bit_cnt += u_v (3, "expGolomb index", minKStart - 1, bitstream);

    kMin = minKStart; 
    for(scanPos = 0; scanPos < maxScanVal; scanPos++)
    {
      golombIndexBit = (kMinTab[scanPos] != kMin)? 1: 0;

      assert(kMinTab[scanPos] <= kMin + 1);

      bit_cnt += u_1("expGolomb index", golombIndexBit, bitstream);
      kMin = kMinTab[scanPos];
    }    
  }
  else
  {
    bit_cnt += 3;                               // minKStart
    bit_cnt += maxScanVal;                      // golombIndexBit[]
  }

  return bit_cnt;
}


int writeFilterCoeffs(int sqrFiltLength, int filters_per_group, int pDepthInt[], 
                      int **FilterCoeff, int kMinTab[], Bitstream *bitstream, int createBitstream)
{
  int ind, scanPos, i;
  int bit_cnt = 0;

  for(ind = 0; ind < filters_per_group; ++ind)
  {
    for(i = 0; i < sqrFiltLength; i++)
    {	
      scanPos = pDepthInt[i] - 1;
      if (createBitstream)
        bit_cnt += golombEncode(FilterCoeff[ind][i], kMinTab[scanPos], bitstream, "full-pel filter");
      else
        bit_cnt += lengthGolomb(abs(FilterCoeff[ind][i]), kMinTab[scanPos]);
    }
  }

  return bit_cnt;
}


int codeAuxInfo(int filtNo, int noFilters, int varIndTab[NO_VAR_BINS], int usePrevFiltBest[NO_VAR_BINS], 
                int prevFiltInd, int frNo, Bitstream *bitstream, int createBitstream,int realfiltNo)
{
  int i, predPattern, filterPattern[NO_VAR_BINS], startSecondFilter=0, bitCt=0, codePrediction;
  SyntaxElement symbol, *sym=&symbol;

  int lentab_pred[3]={1, 2, 2};
  int codtab_pred[3]={1, 1, 0};
  int lentab_noFilters[3]={1, 2, 2};
  int codtab_noFilters[3]={1, 1, 0};
  int lentab_filtNo[3]={2, 2, 1};
  int codtab_filtNo[3]={0, 1, 1};

  codePrediction = 0;
  //printf("pf->sqrFiltLength=%d\n",realfiltNo);
  if(frNo > 0)
  {
    codePrediction = 1;

    if(createBitstream == 1)
    {
      if(filtNo >= 0)
        bitCt += u_1("Nonzero coefficients", 1, bitstream);
      else
        bitCt += u_1("Nonzero coefficients", 0, bitstream);
    }
    else
    {
      bitCt += 1;
    }
  }


  // Filter order, number of filters & variance for which they are used
  
    sym->len=lentab_filtNo[realfiltNo];
    if (createBitstream==1){
      sym->inf=codtab_filtNo[realfiltNo];
      symbol2vlc(sym);
      writeUVLC2buffer(sym, bitstream);
    }
    bitCt+=sym->len;
  if(filtNo >= 0)
  {
    if(noFilters > 1)
    {
      memset(filterPattern, 0, NO_VAR_BINS * sizeof(int)); 
      for(i = 1; i < NO_VAR_BINS; ++i)
        if(varIndTab[i] != varIndTab[i-1])
        {
          filterPattern[i] = 1;
          startSecondFilter = i;
        }
    }

    sym->len=lentab_noFilters[min(noFilters-1, 2)];
    if(createBitstream == 1)
    {
      sym->inf=codtab_noFilters[min(noFilters-1, 2)];
      symbol2vlc(sym);
      writeUVLC2buffer(sym, bitstream);
    }
    bitCt+=sym->len;

    if(noFilters == 2)
    {
      if(createBitstream == 1)
        u_v (4, "startSecondFilter", startSecondFilter, bitstream);
      bitCt+=4;
    }
    else if(noFilters > 2)
    {
      if(createBitstream == 1)
        for(i = 1; i < NO_VAR_BINS; i++)
          u_1("filterPattern[i]", filterPattern[i], bitstream);
      bitCt+=15;  // Was 16
    }
  }

  if(codePrediction == 1)
  {
    if (createBitstream == 1)
      u_v (4, "prevFiltInd", prevFiltInd, bitstream);
    bitCt+=4;

    // prediction pattern
    predPattern = (usePrevFiltBest[0] == 1)? 0: 1;
    for(i = 1; i < NO_VAR_BINS; i++)
    {
      if(usePrevFiltBest[i] == 0)
      {
        predPattern=2;
        break;
      }
    }

    sym->len = lentab_pred[predPattern];
    if(createBitstream == 1)
    {
      sym->inf = codtab_pred[predPattern];
      symbol2vlc(sym);
      writeUVLC2buffer(sym, bitstream);
    }
    bitCt += sym->len;

    if(predPattern == 2)
    {
      if(createBitstream == 1)
        for(i = 0; i < NO_VAR_BINS; i++)
          u_1("usePrevFiltBest[i]", usePrevFiltBest[i], bitstream);
      bitCt += 16;
    }
  }
  return(bitCt);
}


//
//
//
int codeFilterCoeff(int **pDiffQFilterCoeffIntPP, Bitstream *bitstream, int fl, int sqrFiltLength, 
                    int filters_per_group, int createBitstream)
{
  int i, k, kMin, kStart, minBits, ind, scanPos, maxScanVal, coeffVal, len = 0,
    *pDepthInt=NULL, kMinTab[MAX_SQR_FILT_LENGTH], bitsCoeffScan[MAX_SCAN_VAL][MAX_EXP_GOLOMB],
    minKStart, minBitsKStart, bitsKStart;

  pDepthInt = pDepthIntTab[fl-2];

  maxScanVal = 0;
  for(i = 0; i < sqrFiltLength; i++)
    maxScanVal = max(maxScanVal, pDepthInt[i]);

  // vlc for all
  memset(bitsCoeffScan, 0, MAX_SCAN_VAL * MAX_EXP_GOLOMB * sizeof(int));
  for(ind=0; ind<filters_per_group; ++ind){	
    for(i = 0; i < sqrFiltLength; i++){	     
      scanPos=pDepthInt[i]-1;
      coeffVal=abs(pDiffQFilterCoeffIntPP[ind][i]);
      for (k=1; k<15; k++){
        bitsCoeffScan[scanPos][k]+=lengthGolomb(coeffVal, k);
      }
    }
  }

  minBitsKStart = 0;
  minKStart = -1;
  for(k = 1; k < 8; k++)
  { 
    bitsKStart = 0; 
    kStart = k;
    for(scanPos = 0; scanPos < maxScanVal; scanPos++)
    {
      kMin = kStart; 
      minBits = bitsCoeffScan[scanPos][kMin];

      if(bitsCoeffScan[scanPos][kStart+1] < minBits)
      {
        kMin = kStart + 1; 
        minBits = bitsCoeffScan[scanPos][kMin];
      }
      kStart = kMin;
      bitsKStart += minBits;
    }
    if((bitsKStart < minBitsKStart) || (k == 1))
    {
      minBitsKStart = bitsKStart;
      minKStart = k;
    }
  }

  kStart = minKStart; 
  for(scanPos = 0; scanPos < maxScanVal; scanPos++)
  {
    kMin = kStart; 
    minBits = bitsCoeffScan[scanPos][kMin];

    if(bitsCoeffScan[scanPos][kStart+1] < minBits)
    {
      kMin = kStart + 1; 
      minBits = bitsCoeffScan[scanPos][kMin];
    }

    kMinTab[scanPos] = kMin;
    kStart = kMin;
  }

  // Coding parameters
  len += writeFilterCodingParams(minKStart, maxScanVal, kMinTab, bitstream, createBitstream);

  // Filter coefficients
  len += writeFilterCoeffs(sqrFiltLength, filters_per_group, pDepthInt, pDiffQFilterCoeffIntPP, 
    kMinTab, bitstream, createBitstream);

  return len;
}

//
//
//
void collectStatCodeFilterCoeffForce0(int **pDiffQFilterCoeffIntPP, int fl, int sqrFiltLength, 
                                     int filters_per_group, int bitsVarBin[])
{
  int i, k, kMin, kStart, minBits, ind, scanPos, maxScanVal, coeffVal, 
    *pDepthInt=NULL, kMinTab[MAX_SQR_FILT_LENGTH], bitsCoeffScan[MAX_SCAN_VAL][MAX_EXP_GOLOMB],
    minKStart, minBitsKStart, bitsKStart;

  pDepthInt=pDepthIntTab[fl-2];

  maxScanVal=0;
  for (i=0; i<sqrFiltLength; i++)
  {
    maxScanVal=max(maxScanVal, pDepthInt[i]);
  }

  // vlc for all
  memset(bitsCoeffScan, 0, MAX_SCAN_VAL * MAX_EXP_GOLOMB * sizeof(int));
  for(ind=0; ind<filters_per_group; ++ind){	
    for(i = 0; i < sqrFiltLength; i++){	     
      scanPos=pDepthInt[i]-1;
      coeffVal=abs(pDiffQFilterCoeffIntPP[ind][i]);
      for (k=1; k<15; k++){
        bitsCoeffScan[scanPos][k] += lengthGolomb(coeffVal, k);
      }
    }
  }

  minBitsKStart=0;
  minKStart = -1;
  for (k=1; k<8; k++){ 
    bitsKStart=0; kStart=k;
    for (scanPos=0; scanPos<maxScanVal; scanPos++){
      kMin=kStart; minBits=bitsCoeffScan[scanPos][kMin];

      if (bitsCoeffScan[scanPos][kStart+1]<minBits){
        kMin=kStart+1; minBits=bitsCoeffScan[scanPos][kMin];
      }
      kStart=kMin;
      bitsKStart+=minBits;
    }
    if (bitsKStart<minBitsKStart || k==1){
      minBitsKStart=bitsKStart;
      minKStart=k;
    }
  }

  kStart = minKStart; 
  for (scanPos=0; scanPos<maxScanVal; scanPos++)
  {
    kMin=kStart; minBits=bitsCoeffScan[scanPos][kMin];

    if (bitsCoeffScan[scanPos][kStart+1]<minBits)
    {
      kMin = kStart+1; 
      minBits = bitsCoeffScan[scanPos][kMin];
    }
   
    kMinTab[scanPos] = kMin;
    kStart = kMin;
  }

  for(ind=0; ind<filters_per_group; ++ind){
    bitsVarBin[ind]=0;
    for(i = 0; i < sqrFiltLength; i++){	
      scanPos=pDepthInt[i]-1;
      bitsVarBin[ind] += lengthGolomb(abs(pDiffQFilterCoeffIntPP[ind][i]), kMinTab[scanPos]);
    }
  }
}


//
//
//
int sendAllFiltersPPPred(int **FilterCoeffQuant, int fl, int sqrFiltLength, 
                         int filters_per_group, Bitstream *bitstream, int createBistream)
{
  int ind, bit_ct = 0, bit_ct0 = 0, i;
  int predMethod = 0;
  int force0 = 0;
  static int **diffFilterCoeffQuant;
  static int first = 1;
  
  if(first == 1)
  {
    initMatrix_int(&diffFilterCoeffQuant, NO_VAR_BINS, MAX_SQR_FILT_LENGTH);
    first = 0;
  }

  bit_ct0 = codeFilterCoeff(FilterCoeffQuant, bitstream, fl, sqrFiltLength, filters_per_group, 0);

  for(ind = 0; ind < filters_per_group; ++ind)
  {		
    if(ind == 0)
    {
      for(i = 0; i < sqrFiltLength; i++)
        diffFilterCoeffQuant[ind][i] = FilterCoeffQuant[ind][i];
    }
    else
    {
      for(i = 0; i < sqrFiltLength; i++)
        diffFilterCoeffQuant[ind][i] = FilterCoeffQuant[ind][i] - FilterCoeffQuant[ind-1][i];
    }
  }

  if(codeFilterCoeff(diffFilterCoeffQuant, bitstream, fl, sqrFiltLength, filters_per_group, 0) >= bit_ct0)
  {
    predMethod = 0;  
    if(filters_per_group > 1)
      bit_ct += writePredFlags(force0, predMethod, NULL, 0, bitstream, createBistream);
    bit_ct += codeFilterCoeff(FilterCoeffQuant, bitstream, fl, sqrFiltLength, filters_per_group, createBistream);
  }
  else
  {
    predMethod = 1;
    if(filters_per_group > 1)
      bit_ct += writePredFlags(force0, predMethod, NULL, 0, bitstream, createBistream);
    bit_ct += codeFilterCoeff(diffFilterCoeffQuant, bitstream, fl, sqrFiltLength, filters_per_group, createBistream);
  }

  return(bit_ct);
}


//
//
//
int sendAllFiltersPPPredForce0(int **FilterCoeffQuant, int fl, int sqrFiltLength, int filters_per_group, 
                               int codedVarBins[NO_VAR_BINS], Bitstream *bitstream, int createBistream)
{
  int ind, bit_ct=0, bit_ct0, i, j;
  int filters_per_group_temp, filters_per_group_diff;
  int chosenPred = 0;
  static int **diffFilterCoeffQuant, **FilterCoeffQuantTemp, first = 1;
  int force0 = 1;

  if(first == 1)
  {
    initMatrix_int(&diffFilterCoeffQuant, NO_VAR_BINS, MAX_SQR_FILT_LENGTH);
    initMatrix_int(&FilterCoeffQuantTemp, NO_VAR_BINS, MAX_SQR_FILT_LENGTH);
    first = 0;
  }

  i = 0;
  for(ind = 0; ind < filters_per_group; ind++)
  {		
    if(codedVarBins[ind] == 1)
    {
      for(j = 0; j < sqrFiltLength; j++)
        FilterCoeffQuantTemp[i][j]=FilterCoeffQuant[ind][j];
      i++;
    }
  }
  filters_per_group_diff = filters_per_group_temp = i;

  for(ind = 0; ind < filters_per_group; ++ind)
  {		
    if(ind == 0)
    {
      for(i = 0; i < sqrFiltLength; i++)
        diffFilterCoeffQuant[ind][i] = FilterCoeffQuantTemp[ind][i];
    }
    else
    {
      for(i = 0; i < sqrFiltLength; i++)
        diffFilterCoeffQuant[ind][i] = FilterCoeffQuantTemp[ind][i] - FilterCoeffQuantTemp[ind-1][i];
    }
  }

  if(!((filters_per_group_temp == 0) && (filters_per_group == 1)))
  {
    bit_ct0 = codeFilterCoeff(FilterCoeffQuantTemp, bitstream, fl, sqrFiltLength, filters_per_group_temp, 0);

    if(codeFilterCoeff(diffFilterCoeffQuant, bitstream, fl, sqrFiltLength, filters_per_group_diff, 0) >= bit_ct0)
    {
      chosenPred = 0;
      bit_ct += writePredFlags(force0, chosenPred, codedVarBins, filters_per_group, bitstream, createBistream);
      bit_ct += codeFilterCoeff(FilterCoeffQuantTemp, bitstream, fl, sqrFiltLength, filters_per_group_temp, createBistream);
    }
    else
    {
      chosenPred = 1;
      bit_ct += writePredFlags(force0, chosenPred, codedVarBins, filters_per_group, bitstream, createBistream);
      bit_ct += codeFilterCoeff(diffFilterCoeffQuant, bitstream, fl, sqrFiltLength, filters_per_group_temp, createBistream);
    }
  }

  return(bit_ct);
}


//
//  Decide Lagrangian, best coding method 
//
void decideCoeffForce0(int codedVarBins[NO_VAR_BINS], double errorForce0Coeff[], double errorForce0CoeffTab[NO_VAR_BINS][2], 
                       int bitsVarBin[NO_VAR_BINS], double lambda, int filters_per_fr)
{
  int filtNo;
  double lagrangianDiff;

  errorForce0Coeff[0]=errorForce0Coeff[1]=0;
  for(filtNo=0; filtNo<filters_per_fr; filtNo++)
  {
    // No coeffcient prediction bits used
    lagrangianDiff=errorForce0CoeffTab[filtNo][0]-(errorForce0CoeffTab[filtNo][1]+lambda*bitsVarBin[filtNo]);
    codedVarBins[filtNo]=(lagrangianDiff>0)? 1 : 0;
    errorForce0Coeff[0]+=errorForce0CoeffTab[filtNo][codedVarBins[filtNo]];
    errorForce0Coeff[1]+=errorForce0CoeffTab[filtNo][1];
  }
    
}

//
// errorForce0CoeffTab[filtNo][0] - error when all coeff zero, 
// errorForce0CoeffTab[filtNo][1] - error with quant coeff
//
double findBestCoeffCodMethod(Bitstream *bitstream, int codedVarBins[NO_VAR_BINS], int *forceCoeff0, 
                              int **filterCoeffSymQuant, int fl, int sqrFiltLength, 
                              int filters_per_fr, double errorForce0CoeffTab[NO_VAR_BINS][2], 
                              double *errorQuant, double lambda)
{
  int bitsVarBin[NO_VAR_BINS], createBistream, coeffBits, coeffBitsForce0;
  double errorForce0Coeff[2], lagrangianForce0, lagrangian;


  collectStatCodeFilterCoeffForce0(filterCoeffSymQuant, fl, sqrFiltLength,  
    filters_per_fr, bitsVarBin);

  decideCoeffForce0(codedVarBins, errorForce0Coeff, errorForce0CoeffTab, bitsVarBin, lambda, filters_per_fr);
  
  coeffBitsForce0 = sendAllFiltersPPPredForce0(filterCoeffSymQuant, fl, sqrFiltLength, 
    filters_per_fr, codedVarBins, bitstream, createBistream=0);

  coeffBits = sendAllFiltersPPPred(filterCoeffSymQuant, fl, sqrFiltLength, filters_per_fr, 
    bitstream, createBistream=0);

  lagrangianForce0=errorForce0Coeff[0]+lambda*coeffBitsForce0;
  lagrangian=errorForce0Coeff[1]+lambda*coeffBits;
#if FORCE_ZERO==1
  if (lagrangianForce0<lagrangian)
  {
    *errorQuant=errorForce0Coeff[0];
    *forceCoeff0=1;
    return(lagrangianForce0);
  }
  else
  {
    *errorQuant=errorForce0Coeff[1];
    *forceCoeff0=0;
    return(lagrangian);
  }
#else
  *errorQuant=errorForce0Coeff[1];
  *forceCoeff0=0;
  return(lagrangian);
#endif
}

#endif


