
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <limits.h>
#include <float.h>
#include "global.h"
#include "memalloc.h"
#include "image.h"
#include "alf_filter.h"
#include "adaptive_loop_filter.h"
#ifdef ADAPTIVE_LOOP_FILTER

int pattern9x9Sym[] = 
{
                   0,
	           1,  2,  3,
		   4,  5,  6,  7,  8,
       9, 10, 11, 12, 13, 14, 15,
  16, 17, 18, 19, 20, 19, 18, 17, 16,
      15, 14, 13, 12, 11, 10,  9, 
           8,  7,  6,  5,  4,
		       3,  2,  1,
			       0
};
 
int weights9x9Sym[] = 
{
                  2,
              2,  2,  2,   
          2,  2,  2,  2,  2, 
	   2, 2,  2,  2,  2,  2,  2,  
   2,  2, 2,  2,  1,  1
};

int pattern9x9Sym_Quart[] = 
{
   0,  0,  0,  0,  1,  0,  0,  0,  0,
   0,  0,  0,  2,  3,  4,  0,  0,  0,
   0,  0,  5,  6,  7,  8,  9,  0,  0,  
   0, 10, 11, 12, 13, 14, 15, 16,  0,
  17, 18, 19, 20, 21, 22
};

int pattern7x7Sym[] = 
{
                   0,
	           1,  2,  3,
		   4,  5,  6,  7,  8,
       9, 10, 11, 12, 11, 10, 9,
           8,  7,  6,  5,  4,
		       3,  2,  1,
			       0
};

int weights7x7Sym[] = 
{
                  2,  
              2,  2,  2,   
	      2,  2,  2,  2,  2,    
      2,  2,  2,  1,  1
};

int pattern7x7Sym_Quart[] = 
{
   0,  0,  0,  0,  0,  0,  0,  0,  0,
   0,  0,  0,  0,  1,  0,  0,  0,  0,
   0,  0,  0,  2,  3,  4,  0,  0,  0,
   0,  0,  5,  6,  7,  8,  9,  0,  0,  
   0, 10, 11, 12, 13, 14,  
};

int pattern5x5Sym[] = 
{
                   0,
	           1,  2,  3,
		   4,  5,  6,  5,  4,
               3,  2,  1,
			       0
};

int weights5x5Sym[] = 
{
           2, 
        2, 2, 2,
     2, 2, 1, 1
};

int pattern5x5Sym_Quart[] = 
{
   0,  0,  0,  0,  0,  0,  0,  0,  0,
   0,  0,  0,  0,  0,  0,  0,  0,  0,
   0,  0,  0,  0,  1,  0,  0,  0,  0,
   0,  0,  0,  2,  3,  4,  0,  0,  0,
   0,  0,  5,  6,  7,  8,  0,  0,  0,  
};


int pattern9x9Sym_9[] = 
{
                   4,  
              12, 13, 14,  
          20, 21, 22, 23, 24, 
      28, 29, 30, 31, 32, 33, 34,      
  36, 37, 38, 39, 40, 39, 38, 37, 36, 
      34, 33, 32, 31, 30, 29, 28,  
          24, 23, 22, 21, 20, 
              14, 13, 12, 
                   4,  
};
int pattern9x9Sym_7[] = 
{    
               13,   
           21, 22, 23,  
       29, 30, 31, 32, 33,       
   37, 38, 39, 40, 39, 38, 37,  
       33, 32, 31, 30, 29,   
           23, 22, 21,  
               13  
                     
};

int pattern9x9Sym_5[] = 
{
          22, 
      30, 31, 32,    
  38, 39, 40, 39, 38, 
      32, 31, 30, 
          22,  
 };


static int *patternTab_filt[NO_TEST_FILT]={pattern9x9Sym_9, pattern9x9Sym_7, pattern9x9Sym_5}; 
static int *patternTab[NO_TEST_FILT]={pattern9x9Sym, pattern7x7Sym, pattern5x5Sym}; 
static int *patternMapTab[NO_TEST_FILT]={pattern9x9Sym_Quart, pattern7x7Sym_Quart, pattern5x5Sym_Quart};
static int *weightsTab[NO_TEST_FILT]={weights9x9Sym, weights7x7Sym, weights5x5Sym};
static int sqrFiltLengthTab[]={SQR_FILT_LENGTH_9SYM, SQR_FILT_LENGTH_7SYM, SQR_FILT_LENGTH_5SYM};
static int flTab[]={9/2, 7/2, 5/2};

static int  varIndTab[NO_VAR_BINS];
static double ***yGlobalSym, ****EGlobalSym,  *pixAcc;
static int **g_filterCoeffSymQuant;

static int **g_filterCoeffSym,**g_filterCoeffPrevSelected,**g_filterCoeffPrevSelected_best,**g_filterCoeffPrevSelected_fullframe;
static int ***g_filterCoeffPrev;
static imgpel **varImg,**maskImg;
static int im_height,im_width;
static Bitstream *gBitstream,*gBitstream_fullframe,*gBitstream_best;
static int g_availableFilters=0, g_frNoCoded=0,g_coded=0,g_coded_Best=0,g_coded_fullframe=0;



void init_qc_filter(int height,int width)
{
	im_height=height;
	im_width=width;
	initMatrix4D_double(&EGlobalSym, NO_TEST_FILT,  NO_VAR_BINS, MAX_SQR_FILT_LENGTH, MAX_SQR_FILT_LENGTH);
	initMatrix3D_double(&yGlobalSym, NO_TEST_FILT, NO_VAR_BINS, MAX_SQR_FILT_LENGTH); 

    initMatrix3D_int(&g_filterCoeffPrev, NO_PREV_FILTERS, NO_VAR_BINS, MAX_SQR_FILT_LENGTH); 
	initMatrix_int(&g_filterCoeffSym, NO_VAR_BINS, MAX_SQR_FILT_LENGTH); 
	initMatrix_int(&g_filterCoeffPrevSelected, NO_VAR_BINS, MAX_SQR_FILT_LENGTH); 
	initMatrix_int(&g_filterCoeffPrevSelected_best, NO_VAR_BINS, MAX_SQR_FILT_LENGTH); 
	initMatrix_int(&g_filterCoeffPrevSelected_fullframe, NO_VAR_BINS, MAX_SQR_FILT_LENGTH); 
	initMatrix_int(&g_filterCoeffSymQuant, NO_VAR_BINS, MAX_SQR_FILT_LENGTH); 

	pixAcc = (double *) calloc(NO_VAR_BINS, sizeof(double));
	get_mem2Dpel(&varImg, im_height, im_width);
	get_mem2Dpel(&maskImg, im_height, im_width);
	gBitstream=allocBitstream(1000);
	gBitstream_best=allocBitstream(1000);
	gBitstream_fullframe=allocBitstream(1000);
	
}

void delete_qc_filter()
{
	destroyMatrix4D_double(EGlobalSym, NO_TEST_FILT,  NO_VAR_BINS);
	destroyMatrix3D_double(yGlobalSym, NO_TEST_FILT);
	destroyMatrix_int(g_filterCoeffSymQuant);

	destroyMatrix3D_int(g_filterCoeffPrev, NO_PREV_FILTERS); 
	destroyMatrix_int(g_filterCoeffSym);
	destroyMatrix_int(g_filterCoeffPrevSelected);
	destroyMatrix_int(g_filterCoeffPrevSelected_best);
	destroyMatrix_int(g_filterCoeffPrevSelected_fullframe);

	free(pixAcc);
	free_mem2Dpel(varImg);
	free_mem2Dpel(maskImg);
	freeBitstream(gBitstream);
	freeBitstream(gBitstream_best);
	freeBitstream(gBitstream_fullframe);
	
}

void   CalValImage( imgpel **ImgDec)
{
	calcVar(varImg, ImgDec, VAR_SIZE, im_height, im_width);
}




int storeInBlockMartix_update(double *****blk_corr,double ***blk_pixAcc,alf_parameter_set_rbsp_t* temp_alfps,int prev_idx)
{
  int i,k,varInd;
  int x, y, bn, bx, by, max_bx, max_by ,bbx,bby;
  int sqrFiltLength=SQR_FILT_LENGTH_5SYM;
  double *p_blk_corr,*p_blk_corr_sum;
  int num_alf_block;
  int WidthInAlfBlks;

  int matrix_size = sqrFiltLength*(sqrFiltLength+3)/2;

  int max_HeightinPrevAlfBlks;
  int max_WidthinPrevAlfBlks;
  int prev_blk_size =1<<(prev_idx);
  num_alf_block = temp_alfps->num_alf_block;
  WidthInAlfBlks = temp_alfps->WidthInAlfBlks;
  bn   = 1<<(temp_alfps->alf_block_size_idx - prev_idx);
  assert(bn>0);

  max_HeightinPrevAlfBlks = input->img_height / prev_blk_size;
  max_WidthinPrevAlfBlks  = input->img_width  / prev_blk_size;
  if( (input->img_width  % prev_blk_size)!=0 ) max_WidthinPrevAlfBlks++;
  if( (input->img_height % prev_blk_size)!=0 ) max_HeightinPrevAlfBlks++;

  for (i=0; i<num_alf_block; i++)
  {
		bby = (i / WidthInAlfBlks);
		bbx = (i % WidthInAlfBlks);
        by = bby * bn;
        bx = bbx * bn;
        max_by = by + bn;
        max_bx = bx + bn;
		if(max_by>max_HeightinPrevAlfBlks) max_by = max_HeightinPrevAlfBlks;
        if(max_bx>max_WidthinPrevAlfBlks ) max_bx = max_WidthinPrevAlfBlks;
		
		for (varInd=0;varInd<NO_VAR_BINS;varInd++)
		{
			p_blk_corr_sum = blk_corr[varInd][bby][bbx][0];	
			p_blk_corr = blk_corr[varInd][by][bx][0];
			
			memcpy(p_blk_corr_sum, p_blk_corr,sizeof(double)*matrix_size);

			blk_pixAcc[varInd][bby][bbx] = blk_pixAcc[varInd][by][bx];	

			for(x=bx+1; x<max_bx; x++)
			{		  
			  p_blk_corr_sum = blk_corr[varInd][bby][bbx][0];
			  p_blk_corr = blk_corr[varInd][by][x][0];
			  
			  for (k=0;k<matrix_size;k++)
				*(p_blk_corr_sum++) += *(p_blk_corr++);
			  
			  blk_pixAcc[varInd][bby][bbx]+= blk_pixAcc[varInd][by][x];
			} 
			for(y=by+1; y<max_by; y++)
			{
			  for(x=bx; x<max_bx; x++)
			  {		  
				  p_blk_corr_sum = blk_corr[varInd][bby][bbx][0];
				  p_blk_corr = blk_corr[varInd][y][x][0];

				  for (k=0;k<matrix_size;k++)
					*(p_blk_corr_sum++) += *(p_blk_corr++);

				  blk_pixAcc[varInd][bby][bbx]+= blk_pixAcc[varInd][y][x];
			  } 
			}

		}
  }

  return 1;
}



int fillInMartixfromBlockMatrix_update(double *****blk_corr,double ***blk_pixAcc,alf_parameter_set_rbsp_t* temp_alfps,int getfromImg)
{
  int i,k,l,varInd;
  int bx, by;
  int sqrFiltLength=SQR_FILT_LENGTH_5SYM;
  double ** p_blk_corr;
  int num_alf_block;
  int WidthInAlfBlks;
  int block_ct=0;
  double **E,*yy;
  int filtNo =2; 
  int bn  = temp_alfps->alf_block_size;
  int max_by,y;
  int max_bx,x;
  memset( pixAcc, 0,sizeof(double)*NO_VAR_BINS);
  for (varInd=0; varInd<NO_VAR_BINS; varInd++)
  {
	memset(yGlobalSym[filtNo][varInd],0,sizeof(double)*MAX_SQR_FILT_LENGTH);
    for (k=0; k<sqrFiltLength; k++)
	{
	  memset(EGlobalSym[filtNo][varInd][k],0,sizeof(double)*MAX_SQR_FILT_LENGTH);
    }
  }
  for (i=0; i<im_height; i++)
    memset(maskImg[i],0,sizeof(imgpel)*im_width); 

  num_alf_block = temp_alfps->num_alf_block;
  WidthInAlfBlks = temp_alfps->WidthInAlfBlks;

  for (i=0; i<num_alf_block; i++)
  {
    if(getfromImg || temp_alfps->alf_blk_data[i].alf_blk_flag)
    {
        by = (i / WidthInAlfBlks) ;
        bx = (i % WidthInAlfBlks) ;
		for (varInd=0;varInd<NO_VAR_BINS;varInd++)
		{
            E= EGlobalSym[filtNo][varInd];
            yy= yGlobalSym[filtNo][varInd];
			p_blk_corr = blk_corr[varInd][by][bx];
			for (k=0; k<sqrFiltLength; k++)
			{
				double *pE = E[k];
				for (l=k; l<sqrFiltLength; l++)
					pE[l] += (double)p_blk_corr[k][l-k];

				yy[k]+=(double)p_blk_corr[k][sqrFiltLength-k];
			}
			pixAcc[varInd]+=blk_pixAcc[varInd][by][bx];
		}

		by = by * bn;
		bx = bx * bn;
		max_by = by + bn;
		max_bx = bx + bn;
		if(max_by>im_height) max_by = im_height;
		if(max_bx>im_width ) max_bx = im_width;    
		for(y=by; y<max_by; y++)
		{
		  for(x=bx; x<max_bx; x++)
			  maskImg[y][x] =1;
		}
		block_ct++;
	}
  }

  // Matrix EGlobalSeq is symmetric, only part of it is calculated
  for (varInd=0; varInd<NO_VAR_BINS; varInd++)
  {
	double **pE = EGlobalSym[filtNo][varInd];
    for (k=1; k<sqrFiltLength; k++)
	{
      for (l=0; l<k; l++)
	  {
        pE[k][l]=pE[l][k];
      }
    }
  }
  return block_ct;
}


void storeInBlockMartix(double *****blk_corr,double ***blk_pixAcc,imgpel** ImgOrg, imgpel** ImgDec)
{
  int i,j,k,l,varInd,ii,jj;
  int fl=MIN_NUM_TAP/2, sqrFiltLength=SQR_FILT_LENGTH_5SYM;
  int fl2=9/2; //extended size at each side of the frame
  int ELocal[MAX_SQR_FILT_LENGTH];
  int yLocal;
  int blk_y,blk_x;
  double** corr;
  int filtNo =2;
  double **E,*y;

  memset( pixAcc, 0,sizeof(double)*NO_VAR_BINS);
  for (varInd=0;varInd<NO_VAR_BINS;varInd++)
  {	
	  memset(yGlobalSym[filtNo][varInd],0,sizeof(double)*MAX_SQR_FILT_LENGTH);
      for (k=0; k<sqrFiltLength; k++)
	      memset(EGlobalSym[filtNo][varInd][k],0,sizeof(double)*MAX_SQR_FILT_LENGTH);
      for (i=0; i<(img->max_HeightInAlfBlks>>MIN_BLOCK_IND); i++)
	  {
		  for (j=0; j<(img->max_WidthInAlfBlks>>MIN_BLOCK_IND); j++)
		  {
			  memset(blk_corr[varInd][i][j][0],0,sizeof(double)*sqrFiltLength*(sqrFiltLength+1+2)/2);
		  }
		  memset(blk_pixAcc[varInd][i],0,sizeof(double)*(img->max_WidthInAlfBlks>>MIN_BLOCK_IND));
	  }
  }
  for (i = fl2; i < im_height+fl2; i++)
  {
	for (j = fl2; j < im_width+fl2; j++)
	{
	  int pix;
      varInd=min(varImg[i-fl2][j-fl2], NO_VAR_BINS-1);
      k=0; 
	  memset(ELocal, 0, sqrFiltLength*sizeof(int));
	  maskImg[i-fl2][j-fl2] =1;
	  blk_y=(i-fl2)/(img->min_alf_block_size<<MIN_BLOCK_IND);
	  blk_x=(j-fl2)/(img->min_alf_block_size<<MIN_BLOCK_IND);
      for (ii=-fl; ii<0; ii++)
	  {
        for (jj=-fl-ii; jj<=fl+ii; jj++)
		{  
          ELocal[pattern5x5Sym[k++]]+=(ImgDec[i+ii][j+jj]+ImgDec[i-ii][j-jj]);
        }
      }
      for (jj=-fl; jj<0; jj++)
          ELocal[pattern5x5Sym[k++]]+=(ImgDec[i][j+jj]+ImgDec[i][j-jj]);
	  ELocal[pattern5x5Sym[k++]]+=ImgDec[i][j];
      ELocal[sqrFiltLength-1]=1;
      yLocal=imgY_org[i-fl2][j-fl2];

 	  corr = blk_corr[varInd][blk_y][blk_x];
      E= EGlobalSym[filtNo][varInd];
	  y= yGlobalSym[filtNo][varInd];

	  for (k=0; k<sqrFiltLength; k++)
	  {
		int E_temp = (ELocal[k]*ELocal[k]);
		double *pE = E[k];
  	    corr[k][0]+=E_temp;
		pE[k] += E_temp;
		
		for (l=k+1; l<sqrFiltLength; l++)
		{
			E_temp = (ELocal[k]*ELocal[l]);
			corr[k][l-k]+=E_temp;
			pE[l]+=(E_temp);
		}     
		E_temp = (ELocal[k]*yLocal);
		corr[k][sqrFiltLength-k]+=(E_temp);
		y[k]+=(E_temp);
      }
	  pix = yLocal*yLocal;
	  blk_pixAcc[varInd][blk_y][blk_x] += pix;
	  pixAcc[varInd]+= pix;
    }
  } 

  // Matrix EGlobalSeq is symmetric, only part of it is calculated
  for (varInd=0; varInd<NO_VAR_BINS; varInd++)
  {
    for (k=1; k<sqrFiltLength; k++)
	{
      for (l=0; l<k; l++)
	  {
        EGlobalSym[filtNo][varInd][k][l]=EGlobalSym[filtNo][varInd][l][k];
      }
    }
  }
}


int fillInMartixfromBlockMatrix(double *****blk_corr,double ***blk_pixAcc,alf_parameter_set_rbsp_t* temp_alfps,int getfromImg ,imgpel** ImgOrg, imgpel** ImgDec)
{
  int i,k,l,varInd;
  int x, y, bn, bx, by, max_bx, max_by;
  int sqrFiltLength=SQR_FILT_LENGTH_5SYM;
  double ** p_blk_corr;
  int num_alf_block;
  int WidthInAlfBlks;
  int block_ct=0;
  double **E,*yy;
  int filtNo =2; 
 
  int max_HeightinPrevAlfBlks;
  int max_WidthinPrevAlfBlks;
  int prev_blk_size =(img->min_alf_block_size<<MIN_BLOCK_IND);
 
  int bn2;

  max_HeightinPrevAlfBlks = input->img_height / prev_blk_size;
  max_WidthinPrevAlfBlks  = input->img_width  / prev_blk_size;
  if( (input->img_width  % prev_blk_size)!=0 ) max_WidthinPrevAlfBlks++;
  if( (input->img_height % prev_blk_size)!=0 ) max_HeightinPrevAlfBlks++;

  memset( pixAcc, 0,sizeof(double)*NO_VAR_BINS);
  for (varInd=0; varInd<NO_VAR_BINS; varInd++)
  {
	memset(yGlobalSym[filtNo][varInd],0,sizeof(double)*MAX_SQR_FILT_LENGTH);
    for (k=0; k<sqrFiltLength; k++)
	{
	  memset(EGlobalSym[filtNo][varInd][k],0,sizeof(double)*MAX_SQR_FILT_LENGTH);
    }
  }
  for (i=0; i<im_height; i++)
    memset(maskImg[i],0,sizeof(imgpel)*im_width);

  num_alf_block = temp_alfps->num_alf_block;
  WidthInAlfBlks = temp_alfps->WidthInAlfBlks;
  bn   = (temp_alfps->alf_block_size_idx-img->min_alf_block_size_idx-MIN_BLOCK_IND);
  assert(bn>=0);

  bn2  = temp_alfps->alf_block_size;

  for (i=0; i<num_alf_block; i++)
  {
    if(getfromImg || temp_alfps->alf_blk_data[i].alf_blk_flag)
    {
        by = ((i / WidthInAlfBlks)<<bn);
        bx = ((i % WidthInAlfBlks)<<bn);
        max_by = by + (1<<bn);
        max_bx = bx + (1<<bn);
        if(max_by>max_HeightinPrevAlfBlks) max_by = max_HeightinPrevAlfBlks;
        if(max_bx>max_WidthinPrevAlfBlks ) max_bx = max_WidthinPrevAlfBlks;
		
		for (varInd=0;varInd<NO_VAR_BINS;varInd++)
		{
            E= EGlobalSym[filtNo][varInd];
            yy= yGlobalSym[filtNo][varInd];
			for(y=by; y<max_by; y++)
			{
			  for(x=bx; x<max_bx; x++)
			  {
				  p_blk_corr = blk_corr[varInd][y][x];
				  for (k=0; k<sqrFiltLength; k++)
				  {
					double *pE = E[k];
					for (l=k; l<sqrFiltLength; l++)
						pE[l] += p_blk_corr[k][l-k];
					
					yy[k]+=p_blk_corr[k][sqrFiltLength-k];
				  }
				  pixAcc[varInd]+=blk_pixAcc[varInd][y][x];
			  } 
			}
		}
		by = (i / WidthInAlfBlks) * bn2;
		bx = (i % WidthInAlfBlks) * bn2;
		max_by = by + bn2;
		max_bx = bx + bn2;
		if(max_by>im_height) max_by = im_height;
		if(max_bx>im_width ) max_bx = im_width;    
		for(y=by; y<max_by; y++)
		{
		  for(x=bx; x<max_bx; x++)
			  maskImg[y][x] =1;
		}
		block_ct++;
	}
  }

  // Matrix EGlobalSeq is symmetric, only part of it is calculated
  for (varInd=0; varInd<NO_VAR_BINS; varInd++)
  {
	double **pE = EGlobalSym[filtNo][varInd];
    for (k=1; k<sqrFiltLength; k++)
	{
      for (l=0; l<k; l++)
	  {
        pE[k][l]=pE[l][k];
      }
    }
  }
  return block_ct;
}


int fillInMartix_direct(alf_parameter_set_rbsp_t* temp_alfps,imgpel** ImgOrg, imgpel** ImgDec,int getfromImg,int tap)
{
  int i,k,l,varInd,ii,jj;
  int x, y, bn, bx, by, max_bx, max_by;
  int num_alf_block;
  int WidthInAlfBlks;

  int fl =tap/2;
  int sqrFiltLength=((SQUARE(tap)/4 + 1) + 1);
  int fl2=9/2; //extended size at each side of the frame
  int ELocal[MAX_SQR_FILT_LENGTH];
  int yLocal;
  int *p_pattern;
  int block_ct=0;
  int filtNo =2; 
  double **E,*yy;
  if (tap==9)
		filtNo =0;
  else if (tap==7)
		filtNo =1;

  p_pattern= patternTab[filtNo];

  memset( pixAcc, 0,sizeof(double)*NO_VAR_BINS);
  for (varInd=0; varInd<NO_VAR_BINS; varInd++)
  {
	memset(yGlobalSym[filtNo][varInd],0,sizeof(double)*MAX_SQR_FILT_LENGTH);
    for (k=0; k<sqrFiltLength; k++)
	{
	  memset(EGlobalSym[filtNo][varInd][k],0,sizeof(double)*MAX_SQR_FILT_LENGTH);
    }
  }
  for (i=0; i<im_height; i++)
    memset(maskImg[i],0,sizeof(imgpel)*im_width);

  if (getfromImg)
  {
    int j;
	for (i=0,y=fl2; i<im_height; i++,y++)
	{
		for (j=0,x=fl2; j<im_width; j++,x++)
		{
			maskImg[i][j] =1;
			varInd=min(varImg[i][j], NO_VAR_BINS-1);
			k=0; 
			memset(ELocal, 0, sqrFiltLength*sizeof(int));
		  for (ii=-fl; ii<0; ii++)
		  {
			for (jj=-fl-ii; jj<=fl+ii; jj++)
			{  
			  ELocal[p_pattern[k++]]+=(ImgDec[y+ii][x+jj]+ImgDec[y-ii][x-jj]);
			}
		  }
		  for (jj=-fl; jj<0; jj++)
			  ELocal[p_pattern[k++]]+=(ImgDec[y][x+jj]+ImgDec[y][x-jj]);
		  ELocal[p_pattern[k++]]+=ImgDec[y][x];
			ELocal[sqrFiltLength-1]=1;
			yLocal=imgY_org[i][j];

			pixAcc[varInd]+=(yLocal*yLocal);
            E= EGlobalSym[filtNo][varInd];
            yy= yGlobalSym[filtNo][varInd];

			for (k=0; k<sqrFiltLength; k++)
			{
				for (l=k; l<sqrFiltLength; l++)
				  E[k][l]+=(double)(ELocal[k]*ELocal[l]);
				yy[k]+=(double)(ELocal[k]*yLocal);
			}
		}//j
	  }//i
  }
  else
  {
	num_alf_block = temp_alfps->num_alf_block;
	WidthInAlfBlks = temp_alfps->WidthInAlfBlks;
	bn   = temp_alfps->alf_block_size / img->min_alf_block_size;
    for (i=0; i<num_alf_block; i++)
    {
		if(temp_alfps->alf_blk_data[i].alf_blk_flag)
		{
			by = (i / WidthInAlfBlks) * bn;
			bx = (i % WidthInAlfBlks) * bn;
			max_by = by + bn;
			max_bx = bx + bn;
			by = by*img->min_alf_block_size;
			bx = bx*img->min_alf_block_size;
			max_by = max_by*img->min_alf_block_size;
			max_bx = max_bx*img->min_alf_block_size;
			if(max_by>im_height) max_by = im_height;
			if(max_bx>im_width ) max_bx = im_width;    
			for(y=by+fl2; y<max_by+fl2; y++)
			{
			  for(x=bx+fl2; x<max_bx+fl2; x++)
			  {
				  maskImg[y-fl2][x-fl2] =1;
				  varInd=min(varImg[y-fl2][x-fl2], NO_VAR_BINS-1);
				  k=0; 
				  memset(ELocal, 0, sqrFiltLength*sizeof(int));
				  for (ii=-fl; ii<0; ii++)
				  {
					for (jj=-fl-ii; jj<=fl+ii; jj++)
					{  
					  ELocal[p_pattern[k++]]+=(ImgDec[y+ii][x+jj]+ImgDec[y-ii][x-jj]);
					}
				  }
				  for (jj=-fl; jj<0; jj++)
					  ELocal[p_pattern[k++]]+=(ImgDec[y][x+jj]+ImgDec[y][x-jj]);
				  ELocal[p_pattern[k++]]+=ImgDec[y][x];

				  ELocal[sqrFiltLength-1]=1;
				  yLocal=imgY_org[y-fl2][x-fl2];

				  pixAcc[varInd]+=(yLocal*yLocal);
				  E= EGlobalSym[filtNo][varInd];
				  yy= yGlobalSym[filtNo][varInd];
				  for (k=0; k<sqrFiltLength; k++)
				  {
					for (l=k; l<sqrFiltLength; l++)
					{
					  E[k][l]+=(double)(ELocal[k]*ELocal[l]);
					}
					yy[k]+=(double)(ELocal[k]*yLocal);
				  }
				}
			  }
			block_ct++;
		}
	  }
  }
  // Matrix EGlobalSeq is symmetric, only part of it is calculated
  for (varInd=0; varInd<NO_VAR_BINS; varInd++)
  {
	double **pE = EGlobalSym[filtNo][varInd];
    for (k=1; k<sqrFiltLength; k++)
	{
      for (l=0; l<k; l++)
	  {
        pE[k][l]=pE[l][k];
      }
    }
  }
  return block_ct;
}



void storeCodedFilter()
{

    int varInd;

    for(varInd=0; varInd<NO_VAR_BINS; ++varInd)
	  memcpy(g_filterCoeffPrevSelected_best[varInd],g_filterCoeffPrevSelected[varInd],sizeof(int)*MAX_SQR_FILT_LENGTH);
}

void storeCodedFilter_fullframe()
{
    int varInd;
    for(varInd=0; varInd<NO_VAR_BINS; ++varInd)
		memcpy(g_filterCoeffPrevSelected_fullframe[varInd],g_filterCoeffPrevSelected[varInd],sizeof(int)*MAX_SQR_FILT_LENGTH);
}

void LoadCodedFilter_fullframe()
{
    int varInd;
    for(varInd=0; varInd<NO_VAR_BINS; ++varInd)
	  memcpy(g_filterCoeffPrevSelected[varInd],g_filterCoeffPrevSelected_fullframe[varInd],sizeof(int)*MAX_SQR_FILT_LENGTH);
}

void setCodedFilter()
{

    int varInd;
	if(g_coded_Best)
	{
		for(varInd=0; varInd<NO_VAR_BINS; ++varInd)
		  memcpy(g_filterCoeffPrev[g_frNoCoded%NO_PREV_FILTERS][varInd],g_filterCoeffPrevSelected_best[varInd],sizeof(int)*MAX_SQR_FILT_LENGTH);
		g_availableFilters++; 
		g_frNoCoded++;
		g_availableFilters=min(NO_PREV_FILTERS-1,g_availableFilters);
	}

}



int send_ALFLumaParam_qc(Bitstream * bitstream)
{
    bitstream->bits_to_go= gBitstream->bits_to_go;
  	bitstream->byte_buf= gBitstream->byte_buf;
	memcpy(bitstream->streamBuffer+bitstream->byte_pos,gBitstream->streamBuffer,sizeof(byte)*gBitstream->byte_pos);
	bitstream->byte_pos+=gBitstream->byte_pos;
	return (gBitstream->byte_pos+1)*8-gBitstream->bits_to_go;
}


void saveBestBitstream()
{
//	  printf("gBitstream->byte_pos=%d,%d\n",gBitstream->byte_pos,gBitstream->bits_to_go);

	  storeCodedFilter();
      g_coded_Best=g_coded; 
      gBitstream_best->bits_to_go=gBitstream->bits_to_go;
      gBitstream_best->byte_pos= gBitstream->byte_pos;
  	  gBitstream_best->byte_buf= gBitstream->byte_buf;
	  memcpy(gBitstream_best->streamBuffer,gBitstream->streamBuffer,sizeof(byte)*gBitstream->byte_pos);
}

void LoadBestBitstream()
{
//	  printf("gBitstream_best->byte_pos=%d,%d\n",gBitstream_best->byte_pos,gBitstream_best->bits_to_go);
      g_coded=g_coded_Best; 
	  gBitstream->bits_to_go=gBitstream_best->bits_to_go;
      gBitstream->byte_pos= gBitstream_best->byte_pos;
	  gBitstream->byte_buf=gBitstream_best->byte_buf;

	  memcpy(gBitstream->streamBuffer,gBitstream_best->streamBuffer,sizeof(byte)*gBitstream_best->byte_pos);
}



void saveBitstream_fullframe()
{
	  storeCodedFilter_fullframe();
	  g_coded_fullframe=g_coded;
      gBitstream_fullframe->bits_to_go=gBitstream->bits_to_go;
      gBitstream_fullframe->byte_pos= gBitstream->byte_pos;
  	  gBitstream_fullframe->byte_buf= gBitstream->byte_buf;
	  memcpy(gBitstream_fullframe->streamBuffer,gBitstream->streamBuffer,sizeof(byte)*gBitstream->byte_pos);
}

void LoadBitstream_fullframe()
{
	  int i,j;
      for (i=0; i<im_height; i++)
	  {
        for (j=0; j<im_width; j++)
			maskImg[i][j]=1;
	  }
	  LoadCodedFilter_fullframe();
	  g_coded= g_coded_fullframe;
      gBitstream->bits_to_go=gBitstream_fullframe->bits_to_go;
      gBitstream->byte_pos= gBitstream_fullframe->byte_pos;
	  gBitstream->byte_buf=gBitstream_fullframe->byte_buf;
	  memcpy(gBitstream->streamBuffer,gBitstream_fullframe->streamBuffer,sizeof(byte)*gBitstream_fullframe->byte_pos);
}



//
//  Prediction  
//

double testPrevFilterFast(double ***A, double **b, double *pixAcc, int **filterCoeffSym, int usePrevFilt[],int filtNo)
{
  int i;
  double seFilt= 0, error=0;
  int sqrFiltLength = sqrFiltLengthTab[filtNo];
  for(i = 0; i < NO_VAR_BINS; i++)
  {
	seFilt = calculateErrorCoeffProvidedInt(A[i], b[i], filterCoeffSym[i], sqrFiltLength, patternTab_filt[filtNo]);
    if(seFilt < 0)
    {
      usePrevFilt[i] = 1;
      error += (seFilt+pixAcc[i]);
    }
    else
    {
      usePrevFilt[i] = 0;
      error += pixAcc[i];
    } 
  }
  return(error);
}


double testPrevFilter(imgpel **imgY_rec, imgpel **imgY_org, imgpel **imgY_pad, imgpel **varImg, imgpel **maskImg,
                      int **filterCoeffSym, int usePrevFilt[], int img_height, int img_width,int mask_value,int filtNo)
{
  int i, j, ii, jj, pixelInt;
  double error;
  int fl = FILTER_LENGTH/2;

  int fl_filter = flTab[filtNo];
  int *pattern = patternTab_filt[filtNo];
  double se[NO_VAR_BINS] = {0};


  int offset = (1<<(NUM_BITS-2));
  for(i = fl; i < img_height+fl; i++)
  {
    for(j = fl; j < img_width+fl; j++)
	{
      if (maskImg[i - fl][j - fl]==mask_value)
	  {
	   int varInd = varImg[i - fl][j - fl];
       imgpel *im1,*im2;
	   int *coef= filterCoeffSym[varInd];
        pixelInt = filterCoeffSym[varInd][MAX_SQR_FILT_LENGTH - 1]; 

		pattern=patternTab_filt[filtNo];
		  for (ii=-fl_filter; ii<0; ii++)
		  {
			im1= &(imgY_pad[i+ii][j-fl_filter-ii]);
			im2= &(imgY_pad[i-ii][j+fl_filter+ii]);
			for (jj=-fl_filter-ii; jj<=fl_filter+ii; jj++,im1++,im2--)
				pixelInt+=((*im1+ *im2)*coef[*(pattern++)]);
		  }
	    im1= &(imgY_pad[i][j-fl_filter]);
		im2= &(imgY_pad[i][j+fl_filter]);
		for (jj=-fl_filter; jj<0; jj++,im1++,im2--)
		  pixelInt+=((*im1+ *im2)*coef[*(pattern++)]);
		pixelInt+=(imgY_pad[i][j]*coef[*(pattern++)]);
        pixelInt=(int)((pixelInt+offset) >> (NUM_BITS - 1));
        pixelInt = Clip3Fun(0, (1 << input->BitDepthLuma)-1, pixelInt);
		im1 = &imgY_rec[i - fl][j - fl];
		im2 = &imgY_org[i - fl][j - fl];
        se[varInd] += (pixelInt-*im1)*(pixelInt+*im1- *im2- *im2);
      }
    }
  }

  error = 0;

  for(i = 0; i < NO_VAR_BINS; i++)
  {
    if(se[i] < 0)
      usePrevFilt[i] = 1;
    else
      usePrevFilt[i] = 0;
  }

  return(error);
}



void preFilterFr(double **imgY_preFilter, imgpel **imgY_rec, imgpel **imgY_org, imgpel **imgY_pad, imgpel **varImg,  imgpel **maskImg,
                 int **filterCoeffSym, int usePrevFilt[], int img_height, int img_width,int mask_value,int filtNo)
{
  int i, j, ii, jj;
  int fl = FILTER_LENGTH/2;

  double factor = 1.0/(double)(1<<(NUM_BITS-1));

  int pixelInt;

  int fl_filter = flTab[filtNo];
  int *pattern = patternTab_filt[filtNo];


  for(i = fl; i < img_height + fl; i++)
  {
    for(j = fl; j < img_width+fl; j++)
	{
      if (maskImg[i - fl][j - fl]==mask_value)
	  {
        //int varInd = min(varImg[i - fl][j - fl], NO_VAR_BINS - 1);
        int varInd = varImg[i - fl][j - fl];
        if(usePrevFilt[varInd] == 1)
        {
		   imgpel *im1,*im2;
			int *coef=filterCoeffSym[varInd];
			pixelInt = filterCoeffSym[varInd][MAX_SQR_FILT_LENGTH - 1];

			pattern= patternTab_filt[filtNo];

			  for (ii=-fl_filter; ii<0; ii++)
			  {
				im1= &(imgY_pad[i+ii][j-fl_filter-ii]);
				im2= &(imgY_pad[i-ii][j+fl_filter+ii]);
				for (jj=-fl_filter-ii; jj<=fl_filter+ii; jj++,im1++,im2--)
					pixelInt+=((*im1+ *im2)*coef[*(pattern++)]);
			  }
			im1= &(imgY_pad[i][j-fl_filter]);
			im2= &(imgY_pad[i][j+fl_filter]);	
			for (jj=-fl_filter; jj<0; jj++,im1++,im2--)
			  pixelInt+=((*im1+ *im2)*coef[*(pattern++)]);
			pixelInt+=(imgY_pad[i][j]*coef[*(pattern++)]);

			imgY_preFilter[i - fl][j - fl] = (double)(pixelInt)*factor;
        }
        else
        {
          imgY_preFilter[i - fl][j - fl] = (double)imgY_rec[i - fl][j - fl];
        }
      }
	  else
	  {
		 imgY_preFilter[i - fl][j - fl] = (double)imgY_rec[i - fl][j - fl];
	  }
    }
  }
}


void filterFrame(imgpel** ImgOrg, imgpel** ImgDec, imgpel** ImgRest,int filtNo)
{
  int i,j,ii,jj;
  int  *pattern; 
  int fl, fl_temp, sqrFiltLength;
  int pixelInt;
  int offset = (1<<(NUM_BITS - 2));

  pattern=patternTab_filt[filtNo];
  fl_temp=flTab[filtNo];
  sqrFiltLength=MAX_SQR_FILT_LENGTH;  fl=FILTER_LENGTH/2;

  for (i = fl; i < im_height+fl; i++)
  {
    for (j = fl; j < im_width+fl; j++)
	{
		  int varInd=varImg[i-fl][j-fl];
          imgpel *im1,*im2;
		  int *coef = g_filterCoeffPrevSelected[varInd];

		  pattern=patternTab_filt[filtNo];
          pixelInt=g_filterCoeffPrevSelected[varInd][sqrFiltLength-1]; 
		  for (ii=-fl_temp; ii<0; ii++)
		  {
			im1= &(ImgDec[i+ii][j-fl_temp-ii]);
			im2= &(ImgDec[i-ii][j+fl_temp+ii]);
			for (jj=-fl_temp-ii; jj<=fl_temp+ii; jj++,im1++,im2--)
				pixelInt+=((*im1+ *im2)*coef[*(pattern++)]);
		  }
		  im1= &(ImgDec[i][j-fl_temp]);
		  im2= &(ImgDec[i][j+fl_temp]);	
		  for (jj=-fl_temp; jj<0; jj++,im1++,im2--)
			  pixelInt+=((*im1+ *im2)*coef[*(pattern++)]);
		  pixelInt+=(ImgDec[i][j]*coef[*(pattern++)]);

		  pixelInt=(int)((pixelInt+offset) >> (NUM_BITS - 1));
          ImgRest[i-fl][j-fl] = Clip3Fun(0, (1 << input->BitDepthLuma)-1, pixelInt);
    }
  }
}


void calcPredFilterCoeff(int **filterCoeffPrevSelected, int **filterCoeffSym, int ***filterCoeffPrev, int filtNo, int varIndTab[], int usePrevFiltBest[], 
                         int frNo, int prevFiltInd)
{
  int *patternMap, varInd, i, k;
  static int *filterCoeffDefault;
  static int first=0;

  if (first==0)
  {
    filterCoeffDefault = (int *) calloc(MAX_SQR_FILT_LENGTH, sizeof(int));
	filterCoeffDefault[MAX_SQR_FILT_LENGTH - 2] =(1 << (NUM_BITS - 1));    
    first=1;
  }
  if(img->type==I_SLICE)   //initialize all static variables
  {
    g_availableFilters=0;
    g_frNoCoded=0;
    memset(filterCoeffDefault,0,MAX_SQR_FILT_LENGTH* sizeof(int));
	filterCoeffDefault[MAX_SQR_FILT_LENGTH - 2] = (1 << (NUM_BITS - 1));
  }
  patternMap=patternMapTab[filtNo];

  if (g_frNoCoded>0)
  {
    for(varInd=0; varInd<NO_VAR_BINS; ++varInd)
	{		
      if (usePrevFiltBest[varInd]==1)
	  {
        for(i = 0; i < MAX_SQR_FILT_LENGTH; i++)
		{
          filterCoeffPrevSelected[varInd][i]=filterCoeffPrev[prevFiltInd][varInd][i];
        }
      }
      else
	  {
        for(i = 0; i < MAX_SQR_FILT_LENGTH; i++)
		{
          filterCoeffPrevSelected[varInd][i]=filterCoeffDefault[i];
        }
      }
    }

    for(varInd=0; varInd<NO_VAR_BINS; ++varInd)
	{		
      k=0;
      for(i = 0; i < MAX_SQR_FILT_LENGTH; i++)
	  {
        if (patternMap[i]>0)
		{
          filterCoeffPrevSelected[varInd][i]+=filterCoeffSym[varIndTab[varInd]][k];
          k++;
        }
		else
		{
		  filterCoeffPrevSelected[varInd][i]=0;
		}
      }
    }

  }
  else
  {
    for(varInd=0; varInd<NO_VAR_BINS; ++varInd)
	{		
      k=0;
      for(i = 0; i < MAX_SQR_FILT_LENGTH; i++)
	  {
        if (patternMap[i]>0)
		{
          filterCoeffPrevSelected[varInd][i]=filterCoeffSym[varIndTab[varInd]][k];
          k++;
        }
        else
		{
          filterCoeffPrevSelected[varInd][i]=0;
        }
      }
    }
  }
}


int codeFiltCoeff(Bitstream *bitstream, int **filterCoeffSymQuant, int filtNo, int varIndTab[], int usePrevFiltBest[], int prevFiltInd,
                 int filters_per_fr_best, int frNo)
{
  int varInd, i, forceCoeff0, codedVarBins[NO_VAR_BINS], coeffBits, createBistream,   sqrFiltLength=sqrFiltLengthTab[filtNo], 
    fl=flTab[filtNo], coded;

  for(varInd=0; varInd<filters_per_fr_best; varInd++)
  {
    codedVarBins[varInd]=0;
    for(i = 0; i < MAX_SQR_FILT_LENGTH; i++)
	{
      if (filterCoeffSymQuant[varInd][i] != 0)
	  {
        codedVarBins[varInd]=1;
        break;
      }
    }
  }

  forceCoeff0=0;
  for(varInd=0; varInd<filters_per_fr_best; varInd++)
  {
    if (codedVarBins[varInd] == 0)
	{
      forceCoeff0=1;
      break;
    }
  }

  coded=1;
  if (forceCoeff0==1 && filters_per_fr_best==1)
  {
    coded=0;
    coeffBits = codeAuxInfo(-1, filters_per_fr_best, varIndTab, usePrevFiltBest, prevFiltInd, 
      frNo, bitstream, createBistream=1,filtNo);
  }
  else
  {
    coeffBits = codeAuxInfo(filtNo, filters_per_fr_best, varIndTab, usePrevFiltBest, prevFiltInd, 
      frNo, bitstream, createBistream=1,filtNo);
  }

  if (forceCoeff0==0) 
  {
    coeffBits += sendAllFiltersPPPred(filterCoeffSymQuant, fl, sqrFiltLength, 
      filters_per_fr_best, bitstream, createBistream=1);
  }
  else if (filters_per_fr_best>1)
  {
    coeffBits += sendAllFiltersPPPredForce0(filterCoeffSymQuant, fl, sqrFiltLength, 
      filters_per_fr_best, codedVarBins, bitstream, createBistream=1);
  }
  return(coded);
}





int findBestFilterVarPred(Bitstream *bitstream, double **ySym, double ***ESym, double *pixAcc, int **filterCoeffSym, int **filterCoeffSymQuant, 
                             int ***filterCoeffPrev, int filtNo, int availableFilters, int *filters_per_fr_best, int varIndTab[], int usePrevFiltBest[], 
                             int  *prevFiltInd, imgpel **imgY_rec, imgpel **varImg, imgpel **maskImg, imgpel **imgY_pad, int frNo, int qp)
{
  int filters_per_fr, firstFilt, coded, forceCoeff0,
    interval[NO_VAR_BINS][2], intervalBest[NO_VAR_BINS][2];
  int i, j, ii, jj, k, varInd, /*vectorInd=0,*/ ELocal[MAX_SQR_FILT_LENGTH];
  static double ***E_temp, **y_temp, *pixAcc_temp;
  double  error, errorMin, lambda, lagrangian, lagrangianMin,  yLocal;
  static imgpel  **imgY_filt;
  static double **imgY_preFilter;
  int fl, flMax=flTab[0], sqrFiltLength;
  int *pattern, *patternMap, *weights;
  int numBits, coeffBits;
  double errorForce0CoeffTab[NO_VAR_BINS][2];
  int  codedVarBins[NO_VAR_BINS], createBistream /*, forceCoeff0 */;
  int  usePrevFilt[NO_VAR_BINS], usePrevFiltDefault[NO_VAR_BINS];
  static int first=0;


  for (i = 0; i < NO_VAR_BINS; i++)
    usePrevFiltDefault[i]=usePrevFilt[i]=1;
  lambda = 0.68* pow (2, (qp - SHIFT_QP) / 3.0);
  lambda = lambda * (1<<(2*img->BitDepthIncrease));
  sqrFiltLength=MAX_SQR_FILT_LENGTH;  fl=FILTER_LENGTH/2;
  if (first==0)
  {
    initMatrix_ushort(&imgY_filt, im_height, im_width);
    initMatrix_double(&imgY_preFilter, im_height, im_width);
    initMatrix3D_double(&E_temp, NO_VAR_BINS, MAX_SQR_FILT_LENGTH, MAX_SQR_FILT_LENGTH);
    initMatrix_double(&y_temp, NO_VAR_BINS, MAX_SQR_FILT_LENGTH); 
    pixAcc_temp = (double *) calloc(NO_VAR_BINS, sizeof(double));
    first=1;
  }
#if PREDICT_COEFF==1
  if(img->type==I_SLICE)   //initialize all static variables
#endif
  {
    g_availableFilters=0;
    g_frNoCoded=0;
  }

  sqrFiltLength=sqrFiltLengthTab[filtNo];   fl=flTab[filtNo];
  weights=weightsTab[filtNo];               patternMap=patternMapTab[filtNo];  pattern=patternTab[filtNo];

  memcpy(pixAcc_temp,pixAcc,sizeof(double)*NO_VAR_BINS);
  for (varInd=0; varInd<NO_VAR_BINS; varInd++)
  {
     memcpy(y_temp[varInd],ySym[varInd],sizeof(double)*sqrFiltLength);
     for (k=0; k<sqrFiltLength; k++)
		memcpy(E_temp[varInd][k],ESym[varInd][k],sizeof(double)*sqrFiltLength);
  }
 
  // zero all variables 
  (*prevFiltInd) = 0;
  memset(varIndTab,0,sizeof(int)*NO_VAR_BINS);
  memset(usePrevFiltBest,0,sizeof(int)*NO_VAR_BINS);

  for(i = 0; i < NO_VAR_BINS; i++)
	memset(filterCoeffSym[i],0,sizeof(int)*MAX_SQR_FILT_LENGTH);

  if (g_frNoCoded>0)
  {
    errorMin=0;

    for (i=0; i<availableFilters; i++)
	{
      error = testPrevFilterFast(ESym, ySym, pixAcc, g_filterCoeffPrev[i], usePrevFilt,filtNo);
      if (error<errorMin || i==0)
	  {
        errorMin=error;
        (*prevFiltInd)=i;
      }
    }
    error=testPrevFilter(imgY_rec, imgY_org, imgY_pad, varImg, maskImg, g_filterCoeffPrev[(*prevFiltInd)], usePrevFiltBest, 
      im_height, im_width,1,filtNo);
    preFilterFr(imgY_preFilter, imgY_rec, imgY_org, imgY_pad, varImg, maskImg, g_filterCoeffPrev[(*prevFiltInd)], 
      usePrevFiltBest, im_height, im_width,1,filtNo);

  }


  // If frNo>0 pixAcc and ySym have to be calculated again
  if (g_frNoCoded>0)
  {
    memset(pixAcc_temp,0,sizeof(double)*NO_VAR_BINS);
	for (varInd=0; varInd<NO_VAR_BINS; varInd++)
		memset(y_temp[varInd],0,sizeof(double)*sqrFiltLength);

    for (i = flMax; i < im_height+flMax; i++)
	{
      for (j = flMax; j < im_width+flMax; j++)
	  {
        if (maskImg[i - flMax][j - flMax]==1)
		{
          varInd=min(varImg[i-flMax][j-flMax], NO_VAR_BINS-1);
          k=0; memset(ELocal, 0, sqrFiltLength*sizeof(int));
		  for (ii=-fl; ii<0; ii++)
		  {
			for (jj=-fl-ii; jj<=fl+ii; jj++)
			{  
			  ELocal[pattern[k++]]+=(imgY_pad[i+ii][j+jj]+imgY_pad[i-ii][j-jj]);
			}
		  }
		  for (jj=-fl; jj<0; jj++)
			  ELocal[pattern[k++]]+=(imgY_pad[i][j+jj]+imgY_pad[i][j-jj]);
		  ELocal[pattern[k++]]+=imgY_pad[i][j];
          ELocal[sqrFiltLength-1]=1;
          yLocal=imgY_org[i-flMax][j-flMax]-imgY_preFilter[i-flMax][j-flMax];
          pixAcc_temp[varInd]+=(yLocal*yLocal);
          for (k=0; k<sqrFiltLength; k++)
		  {
            y_temp[varInd][k]+=(double)(ELocal[k]*yLocal);
          }
        }
      }
    }
  }
  firstFilt=1;  lagrangianMin=0;
  filters_per_fr=NO_FILTERS;

  while(filters_per_fr>=1)
  {
    findFilterGroupingError(E_temp, y_temp, pixAcc_temp, interval, sqrFiltLength, filters_per_fr);
    findFilterCoeff(E_temp, y_temp, pixAcc_temp, filterCoeffSym, filterCoeffSymQuant, interval,
      varIndTab, sqrFiltLength, filters_per_fr, weights, numBits=NUM_BITS,  errorForce0CoeffTab);

    lagrangian=findBestCoeffCodMethod(bitstream, codedVarBins, &forceCoeff0, filterCoeffSymQuant, fl, 
      sqrFiltLength, filters_per_fr, errorForce0CoeffTab, &error, lambda);

    if (lagrangian<lagrangianMin || firstFilt==1){
      firstFilt=0;
      lagrangianMin=lagrangian;

      (*filters_per_fr_best)=filters_per_fr;
      memcpy(intervalBest, interval, NO_VAR_BINS*2*sizeof(int));
    }
    filters_per_fr--;
	
  }


  findFilterCoeff(E_temp, y_temp, pixAcc_temp, filterCoeffSym, filterCoeffSymQuant, intervalBest,
    varIndTab, sqrFiltLength, (*filters_per_fr_best), weights, numBits=NUM_BITS, errorForce0CoeffTab);
  findBestCoeffCodMethod(bitstream, codedVarBins, &forceCoeff0, filterCoeffSymQuant, fl, sqrFiltLength, 
    (*filters_per_fr_best), errorForce0CoeffTab, &error, lambda);

  coded=1;
  if (forceCoeff0==1 && (*filters_per_fr_best)==1)
  {
    coded=0;
    coeffBits = codeAuxInfo(-1, (*filters_per_fr_best), varIndTab, usePrevFiltBest, (*prevFiltInd), 
      g_frNoCoded, bitstream, createBistream=0,filtNo);
  }
  else{
    coeffBits = codeAuxInfo(filtNo, (*filters_per_fr_best), varIndTab, usePrevFiltBest, (*prevFiltInd), 
      g_frNoCoded, bitstream, createBistream=0,filtNo);
  }
  if (forceCoeff0==0){
    coeffBits += sendAllFiltersPPPred(filterCoeffSymQuant, fl, sqrFiltLength, 
      (*filters_per_fr_best), bitstream, createBistream=0);
  }
  else
  {
    if ((*filters_per_fr_best)==1)
	{
      for(varInd=0; varInd<(*filters_per_fr_best); varInd++)
	  {
        memset(filterCoeffSym[varInd],0,sizeof(int)*MAX_SQR_FILT_LENGTH);
	    memset(filterCoeffSymQuant[varInd],0,sizeof(int)*MAX_SQR_FILT_LENGTH);
      }
    }
    else{
      coeffBits += sendAllFiltersPPPredForce0(filterCoeffSymQuant, fl, sqrFiltLength, 
        (*filters_per_fr_best), codedVarBins, bitstream, createBistream=0);

      for(varInd=0; varInd<(*filters_per_fr_best); varInd++)
	  {
        if (codedVarBins[varInd]==0)
		{
			memset(filterCoeffSym[varInd],0,sizeof(int)*MAX_SQR_FILT_LENGTH);
			memset(filterCoeffSymQuant[varInd],0,sizeof(int)*MAX_SQR_FILT_LENGTH);
        }
      }
    }
  } 
  return(coeffBits);
}



void FilteringFrameLuma_qc(imgpel** ImgOrg, imgpel** imgY_pad, imgpel** ImgFilt,int tap,alf_parameter_set_rbsp_t* temp_alfps)
{
	static int first_local=0;
	int  filtNo,filters_per_fr, prevFiltInd,usePrevFilt[NO_VAR_BINS];

	static double **ySym, ***ESym;
	static Bitstream  *varFiltStream;
	imgpel   **imgY_rec=enc_picture->imgY;
	if (first_local==0)
	{
		varFiltStream = allocBitstream(10000);
		first_local=1;
	}

	if (tap==9)
		filtNo =0;
	else if (tap==7)
		filtNo =1;
	else
		filtNo=2;
   
	ESym=EGlobalSym[filtNo];  
	ySym=yGlobalSym[filtNo];
	findBestFilterVarPred(varFiltStream, ySym, ESym, pixAcc, g_filterCoeffSym, g_filterCoeffSymQuant,
						g_filterCoeffPrev, filtNo, g_availableFilters,&filters_per_fr, varIndTab,  
						usePrevFilt, &prevFiltInd, imgY_rec, varImg, maskImg, imgY_pad, (img->framepoc),img->qp);

	calcPredFilterCoeff(g_filterCoeffPrevSelected, g_filterCoeffSym, g_filterCoeffPrev, filtNo, varIndTab,
					  usePrevFilt,img->framepoc, prevFiltInd);

	//filter the frame

	filterFrame(ImgOrg, imgY_pad, ImgFilt,filtNo);

	g_coded=codeFiltCoeff(varFiltStream, g_filterCoeffSymQuant, filtNo, varIndTab, usePrevFilt, prevFiltInd,
			filters_per_fr,g_frNoCoded);

	gBitstream->bits_to_go= varFiltStream->bits_to_go;
	gBitstream->byte_pos= varFiltStream->byte_pos;
	gBitstream->byte_buf= varFiltStream->byte_buf;
	memcpy(gBitstream->streamBuffer,varFiltStream->streamBuffer,sizeof(byte)*varFiltStream->byte_pos);
	//printf("filter bits=%d\n",varFiltStream->byte_pos*8-varFiltStream->bits_to_go+8);
	varFiltStream->byte_pos =0;
	varFiltStream->bits_to_go =8;
}


#endif

