
#include "alf_filter_common.h"
#ifdef ADAPTIVE_LOOP_FILTER
Bitstream * allocBitstream(int size)
{
  Bitstream *tmp;

  tmp =  (Bitstream *) calloc(1, sizeof(Bitstream));
  tmp->streamBuffer = (byte *) calloc(size, sizeof(byte));
  tmp->bits_to_go = 8;
  tmp->byte_buf = 0;
  tmp->byte_pos = 0;

  return tmp;
}
int freeBitstream(Bitstream *bitstream)
{
  int tmp = 0;
  tmp += bitstream->byte_pos * 8;
  tmp += 8 - bitstream->bits_to_go;

  free(bitstream->streamBuffer);
  bitstream->streamBuffer = NULL;
  free(bitstream);

  return tmp;
}

imgpel Clip_post(int high, int val)
{
  return (imgpel)(((val > high)? high: val));
}

void padImage(imgpel **imgY,  imgpel **imgY_pad, int fl, int img_height, int img_width)
{
  int x, y;

  for(y = 0; y < img_height; ++y)
     memcpy(imgY_pad[y+fl]+fl,imgY[y],sizeof(imgpel)* img_width);
  
  for(x = fl; x < img_width + fl; ++x) 
  {
    for(y = 0; y < fl; ++y)
      imgY_pad[y][x] = imgY_pad[fl][x];

    for(y = img_height + fl; y < img_height + 2 * fl; ++y)
      imgY_pad[y][x] = imgY_pad[img_height + fl - 1][x];
  }

  for(y = 0; y < img_height + 2 * fl; ++y)
  {
    for(x = 0; x < fl; ++x)
      imgY_pad[y][x] = imgY_pad[y][fl];

    for(x = img_width + fl; x < img_width + 2 * fl; ++x)
      imgY_pad[y][x] = imgY_pad[y][img_width + fl - 1];
  }
}


void  calcVar(imgpel **imgY_var, imgpel **imgY_rec, int fl, int img_height, int img_width)
{
  int i, j, ii, jj;
  int sum;
  static imgpel **imgY_pad;
  static int **imgY_temp, first = 0;
  int var_max= NO_VAR_BINS-1;

  int mult_fact_int;
  int shift= (11+img->BitDepthIncrease);
  int mult_fact_int_tab[4]= {1,114,41,21};
  

  if(first == 0)
  {
    initMatrix_ushort(&imgY_pad, img_height+2*fl+2, img_width+2*fl+2);
    initMatrix_int(&imgY_temp, img_height+2*fl, img_width+2*fl);
    first = 1;
  }

  padImage(imgY_rec, imgY_pad, fl+1, img_height, img_width);

  mult_fact_int = mult_fact_int_tab[fl];
  
if (fl==0)
{
  shift=img->BitDepthIncrease;
  for(i = 1; i < img_height + 1; ++i)
    for(j = 1; j < img_width + 1; ++j)	
	{
      imgY_temp[i-1][j-1] = abs(2 * imgY_pad[i][j] - imgY_pad[i][j+1] - imgY_pad[i][j-1])+
                            abs(2 * imgY_pad[i][j] - imgY_pad[i-1][j] - imgY_pad[i+1][j]);
	  imgY_var[i-1][j-1]=(imgpel) Clip_post(var_max, (int) ((imgY_temp[i-1][j-1])>>shift));
	}
}
else
{
  shift= (11+img->BitDepthIncrease);

  for(i = 1; i < img_height + 2 * fl + 1; ++i)
    for(j = 1; j < img_width + 2 * fl + 1; ++j)	
      imgY_temp[i-1][j-1] = abs(2 * imgY_pad[i][j] - imgY_pad[i][j+1] - imgY_pad[i][j-1])+
                            abs(2 * imgY_pad[i][j] - imgY_pad[i-1][j] - imgY_pad[i+1][j]);
  for(i = fl; i < img_height + fl; ++i)
  {
    for(j = fl; j < img_width + fl; ++j)
    {	
      sum = 0;
      for(ii = i - fl; ii <= i + fl; ii++)
        for(jj = j - fl; jj <= j + fl; jj++)
          sum+=imgY_temp[ii][jj];

	  imgY_var[i - fl][j - fl] =(imgpel) Clip_post(var_max, (int) ((sum * mult_fact_int)>>shift));
    }
  }
}
}


#endif



