#include <stdlib.h>
#include <assert.h>
#include <memory.h>
#include "global.h"

#include "elements.h"
#include "vlc.h"
#include "image.h"
#include "mb_access.h"
#include "cabac.h"
#include "header.h"
#include <math.h>

#ifdef MV_COMPETITION
#include "mv_competition.h"
extern MV_Competition mv_comp;
extern int ******send_index_mv_prediction;        // int predModeMV2[mode][10][list][4][4][MAX_MODEPREDMV+1]; 
#endif
#ifdef MB32X32
extern int no32x32, no64x64;
extern Macroblock  MB32;
extern Macroblock  MB64;//MB64X64
extern int   cbp_32[1<<(2*(MAX_MB_EXT_LEVEL-1))], cbp_32_RDCost[1<<(2*(MAX_MB_EXT_LEVEL-1))];//MB64X64

extern int   cbp16[1<<(2*MAX_MB_EXT_LEVEL)];
extern int64 cbp_blk16[1<<(2*MAX_MB_EXT_LEVEL)];
extern int   ****cofAC32[1<<(2*MAX_MB_EXT_LEVEL)];     // [8x8block][4x4block][level/run][scan_pos]
extern int   ***cofDC32[1<<(2*MAX_MB_EXT_LEVEL)];                       // [yuv][level/run][scan_pos]
extern int   luma_transform_size_8x8_flag32[1<<(2*MAX_MB_EXT_LEVEL)];
extern int    **cof32AC16x16[1<<(2*MAX_MB_EXT_LEVEL)];

extern int bestmode32;
extern int bestmode64;
//int mb64_stat[2][5];//MB64X64
extern int mb32type[1<<(2*(MAX_MB_EXT_LEVEL-1))];//MB64X64

extern int mb32_stat[2][5];
extern int mb32_stat_currentpass[2][3][5];
#ifdef MB64X64
extern int mb64_stat[2][5];
extern int mb64_stat_currentpass[2][3][5];
#endif
extern int eos_bit_32;
extern int mb32_mvd[2][BLOCK_MULTIPLE<<MAX_MB_EXT_LEVEL][BLOCK_MULTIPLE<<MAX_MB_EXT_LEVEL][2];          //!< indices correspond to [forw,backw][block_y][block_x][x,y]
void unary_bin_encode(EncodingEnvironmentPtr eep_frame, unsigned int symbol, BiContextTypePtr ctx, int ctx_offset);
void unary_exp_golomb_mv_encode(EncodingEnvironmentPtr eep_dp, unsigned int symbol, BiContextTypePtr ctx, unsigned int max_bin);
int MBType2Value (Macroblock* currMB);
int writeIntraModes(void);
int ZeroRef (Macroblock* currMB);
int BType2CtxRef (int btype);

int writeDeltaQP(int rdopt, int cbp);
int writeCBP_Big(int rdopt, int cbp);
extern int writeCBP_Transform_size (int rdopt, int mb32);
int writeLumaCoeff (int rdopt, int mb32);

extern char ****geo_mask[3];//[mb_ext_level][rho][theta][y][x]
extern int curr_mb32i, curr_mbi;
extern int rho_bits[3];//[mb_ext_level]
extern int geo_stat[2][3];
extern int best8x8pdir_geo[GEO_PART_NUM][2];

int MBType2Value_geo (Macroblock* currMB);
int writeMotionInfo2NAL_geo (int geo_mode, int mb_ext_level);
void writeMB_geoInfo_CABAC (SyntaxElement *se, EncodingEnvironmentPtr eep_dp);

/*!
 ************************************************************************
 * \brief
 *    level           : 1-32x32; 2-64x64
 *    CurrentMbAddr   : 16x16MB index in the raster scan order starting from 0
 *    CurrentMbAddr_x : 16x16MB x-axis index
 *    CurrentMbAddr_y : 16x16MB y_axis index
 *    extend_x        : the horizonal next 16x16MB x-axis index in the current cluster
 *    extend_y        : the vertical  next 16x16MB y-axis index in the current cluster
 *    num_MB          : number of 16x16 MB in the cluster
 *
 ************************************************************************
 */
void find_Mb_cluster(int level, int CurrentMbAddr, int *CurrentMbAddr_x, int *CurrentMbAddr_y, int *extend_x, int *extend_y, int *num_MB, int *nextMbAddr)
{
  int extend = 1<<level;
  *CurrentMbAddr_x = xPicPos[CurrentMbAddr];
  *CurrentMbAddr_y = yPicPos[CurrentMbAddr];
  *extend_x = min(*CurrentMbAddr_x+extend-1, (int)(img->PicWidthInMbs-1));
  *extend_y = min(*CurrentMbAddr_y+extend-1, (int)(img->FrameHeightInMbs-1));

  *num_MB = (*extend_x - *CurrentMbAddr_x + 1) * (*extend_y - *CurrentMbAddr_y + 1);

  if(*extend_x+1 <= (int)(img->PicWidthInMbs-1))
    *nextMbAddr = CurrentMbAddr + extend;
  else if(*extend_y+1 <= (int)(img->FrameHeightInMbs-1))
    *nextMbAddr = (*extend_y+1) * img->PicWidthInMbs;
  else
    *nextMbAddr = -1;

}


void SetFrameOffset32(int CurrentMbAddr)
{
    int i, j;
//    int initialMbAddr = CurrentMbAddr;
    int encodeMbAddr;
    int CurrentMbAddr_x, CurrentMbAddr_y, extend_x, extend_y, num_MB, nextMbAddr;//, mbcount=0;
#ifdef MB64X64
    find_Mb_cluster(2, CurrentMbAddr, &CurrentMbAddr_x, &CurrentMbAddr_y, &extend_x, &extend_y, &num_MB, &nextMbAddr);
#else
    find_Mb_cluster(1, CurrentMbAddr, &CurrentMbAddr_x, &CurrentMbAddr_y, &extend_x, &extend_y, &num_MB, &nextMbAddr);
#endif
    for(i=CurrentMbAddr_y; i<=extend_y; i++)
    {
      for(j=CurrentMbAddr_x; j<=extend_x; j++)
      {
        encodeMbAddr = i*img->PicWidthInMbs + j;
        set_MB_parameters (encodeMbAddr);
        SetFrameOffset();
      }
    }
}
void Setnz_coeff(int CurrentMbAddr, int mb_ext_level)//tm_vlc
{
  int i, j, m, n;
  //    int initialMbAddr = CurrentMbAddr;
  int encodeMbAddr;
  int CurrentMbAddr_x, CurrentMbAddr_y, extend_x, extend_y, num_MB, nextMbAddr;//, mbcount=0;

  find_Mb_cluster(mb_ext_level, CurrentMbAddr, &CurrentMbAddr_x, &CurrentMbAddr_y, &extend_x, &extend_y, &num_MB, &nextMbAddr);

  for(i=CurrentMbAddr_y; i<=extend_y; i++)
  {
    for(j=CurrentMbAddr_x; j<=extend_x; j++)
    {
      encodeMbAddr = i*img->PicWidthInMbs + j;

      for (m=0; m < (4 + img->num_blk8x8_uv); m++)
        for (n=0; n < 4; n++)
          img->nz_coeff[encodeMbAddr][n][m]=0;        
    }
  }
}

void prepare_data_for_write(int encodeMbAddr, Macroblock *MB, int mbcount, int mode)
{
  int ****i4p, ***i3p;
  Macroblock  *currMB  = &img->mb_data[encodeMbAddr];
  int **i2p;

  currMB->cbp = cbp16[mbcount];
  currMB->cbp_blk = cbp_blk16[mbcount];
  currMB->mb_type = mode;
  currMB->luma_transform_size_8x8_flag = luma_transform_size_8x8_flag32[mbcount];

  i4p=cofAC32[mbcount]; cofAC32[mbcount]=img->cofAC; img->cofAC=i4p;
  i3p=cofDC32[mbcount]; cofDC32[mbcount]=img->cofDC; img->cofDC=i3p;
  i2p=cof32AC16x16[mbcount]; cof32AC16x16[mbcount]=img->cofAC16x16; img->cofAC16x16=i2p;

  memcpy(currMB->b8mode,MB->b8mode, BLOCK_MULTIPLE * sizeof(int));
  memcpy(currMB->b8pdir,MB->b8pdir, BLOCK_MULTIPLE * sizeof(int));    
  currMB->bi_pred_me = MB->bi_pred_me;
}

void propagate_qp_info(int CurrentMbAddr, int CurrentMbAddr_x, int CurrentMbAddr_y, int extend_x, int extend_y)      
{
  int i, j, encodeMbAddr;
  int qp,prev_qp,prev_delta_qp;

  qp            = img->mb_data[CurrentMbAddr].qp;
  prev_qp       = img->mb_data[CurrentMbAddr].prev_qp;
  prev_delta_qp = img->mb_data[CurrentMbAddr].prev_delta_qp;

  for(i=CurrentMbAddr_y; i<=extend_y; i++)
  {
    for(j=CurrentMbAddr_x; j<=extend_x; j++)
    {
      encodeMbAddr = i*img->PicWidthInMbs + j;
      img->mb_data[encodeMbAddr].qp             = qp;
      img->mb_data[encodeMbAddr].prev_qp        = prev_qp;
      img->mb_data[encodeMbAddr].prev_delta_qp  = prev_delta_qp;
      set_chroma_qp(&img->mb_data[encodeMbAddr]);
      img->mb_data[encodeMbAddr].delta_qp = img->mb_data[encodeMbAddr].qp - img->mb_data[encodeMbAddr].prev_qp;
    }
  }
}

void propagate_qp_infoFinal(int CurrentMbAddr, int CurrentMbAddr_x, int CurrentMbAddr_y, int extend_x, int extend_y)      
{
  int i, j, ii, jj, encodeMbAddr;
  int qp,prev_qp,prev_delta_qp;
  int mb32_count=0,mb16_count=0;

  for(i=CurrentMbAddr_y; i<=extend_y; i+=2)  
  {      
    for(j=CurrentMbAddr_x; j<=extend_x; j+=2)   
    {             
      for(ii=0; ii<2; ii++)
      {
        for(jj=0; jj<2; jj++)
        {
          encodeMbAddr = (i+ii)*img->PicWidthInMbs + (j+jj);
          assert(saveMode.MBsize64==P8x8);
            if(saveMode.MBsize32[mb32_count]==P8x8)
            {
              qp            = saveMode.MBsize16_qp[mb16_count];
              prev_qp       = saveMode.MBsize16_prev_qp[mb16_count];
              prev_delta_qp = saveMode.MBsize16_prev_delta_qp[mb16_count];
            }
            else
            {
              qp            = saveMode.MBsize32_qp[mb32_count];
              prev_qp       = saveMode.MBsize32_prev_qp[mb32_count];
              prev_delta_qp = saveMode.MBsize32_prev_delta_qp[mb32_count];
            }
          img->mb_data[encodeMbAddr].qp             = qp;
          img->mb_data[encodeMbAddr].prev_qp        = prev_qp;
          img->mb_data[encodeMbAddr].prev_delta_qp  = prev_delta_qp;
          set_chroma_qp(&img->mb_data[encodeMbAddr]);
          img->mb_data[encodeMbAddr].delta_qp = img->mb_data[encodeMbAddr].qp - img->mb_data[encodeMbAddr].prev_qp;

          mb16_count++;
        }
      }
      mb32_count++;
    }
  }
}


void propagate_mb_info(int mb_ext_level, Macroblock* currMB32, int CurrentMbAddr, int CurrentMbAddr_x, int CurrentMbAddr_y, int extend_x, int extend_y)      
{
  int i, j, encodeMbAddr;
  int k;
  //set 16x16 block type 
  for(i=CurrentMbAddr_y; i<=extend_y; i++)
  {
    for(j=CurrentMbAddr_x; j<=extend_x; j++)
    {
      encodeMbAddr = i*img->PicWidthInMbs + j;

#if !P_DIRECT
      if(img->type==P_SLICE && currMB32->mb_type == 0)
        assert(currMB32->cbp==0);
#endif

      if(currMB32->mb_type == 0 && currMB32->cbp==0)
        img->mb_data[encodeMbAddr].skip_flag = 1;
      else
        img->mb_data[encodeMbAddr].skip_flag = 0;

      if(img->mb_data[CurrentMbAddr].mb_type == 0)
        img->mb_data[encodeMbAddr].mb_type = 0;
      else
      {
          img->mb_data[encodeMbAddr].mb_type = 1;
      }
      for(k=0; k<4; k++)
        img->mb_data[encodeMbAddr].b8mode[k] = img->mb_data[encodeMbAddr].mb_type;

      if(currMB32->geo_mode != -1)
      {
        int m,n;
        int rho = currMB32->geo_mode>>5, theta_idx = currMB32->geo_mode & 31;  
        int start_y = (i - CurrentMbAddr_y)*16;
        int start_x = (j - CurrentMbAddr_x)*16;
        for(m=0; m<16; m++)
          for(n=0; n<16; n++)
            img->mb_data[encodeMbAddr].geo_mask[m][n] = geo_mask[mb_ext_level][rho][theta_idx][start_y+m][start_x+n];
      }
      else
      {
        img->mb_data[encodeMbAddr].geo_mask[0][0] = -1;
      }
    }//j
  }//i
}

#ifdef MB32X32
void writeRefFrame_CABAC32(SyntaxElement *se, EncodingEnvironmentPtr eep_dp)
{
  MotionInfoContexts  *ctx    = img->currentSlice->mot_ctx;
//  Macroblock          *currMB = &img->mb_data[img->current_mb_nr];
  int                 addctx  = 0;

  int   a, b;
  int   act_ctx;
  int   act_sym;
  char** refframe_array = enc_picture->ref_idx[se->value2];

  int bslice = (img->type==B_SLICE);

  int   b8a, b8b;

  PixelPos block_a, block_b;

  int block_a_is_direct=0, block_b_is_direct=0, block_a_b8mode_is_direct=0, block_b_b8mode_is_direct=0;
  int mb_nr = img->current_mb_nr, subblock_x=0, subblock_y=0; 
  int mb_ext_level = se->mb_ext_level;

  if(img->subblock_x==0 && img->subblock_y==0 )
  {
    subblock_x = 0;
    subblock_y = 0;
    mb_nr      = img->current_mb_nr;
  }
  else if(img->subblock_x==0 && img->subblock_y==(2<<mb_ext_level) )
  {
    subblock_x = 0;
    subblock_y = 0;
    mb_nr      = img->current_mb_nr + (img->PicWidthInMbs << (mb_ext_level-1));
  }
  else if(img->subblock_x==(2<<mb_ext_level) && img->subblock_y==0 )
  {
    subblock_x = 0;
    subblock_y = 0;
    mb_nr      = img->current_mb_nr + (1<<(mb_ext_level-1));
  }
  else
  {
    error("subblock_x and subblock_y are out of range\n", -1);
  }

  getLuma4x4Neighbour(mb_nr, subblock_x, subblock_y, -1,  0, &block_a);
  getLuma4x4Neighbour(mb_nr, subblock_x, subblock_y,  0, -1, &block_b);

  b8a=((block_a.x >> 1) & 0x01)+2*((block_a.y >> 1) & 0x01);
  b8b=((block_b.x >> 1) & 0x01)+2*((block_b.y >> 1) & 0x01);
  

// ???????????????????????????????????????????????????????????????????????????????????? peisong
  if(block_b.available && (mb_nr == img->current_mb_nr || mb_nr == img->current_mb_nr+(1<<(mb_ext_level-1))) )//block_b should be out of the current 32x32 block
  {
    block_b_is_direct = IS_DIRECT(&img->mb_data[block_b.mb_addr]);
    block_b_b8mode_is_direct = img->mb_data[block_b.mb_addr].b8mode[b8b]==0 && bslice;
  }
  if(block_a.available && (mb_nr == img->current_mb_nr || mb_nr == img->current_mb_nr + (img->PicWidthInMbs << (mb_ext_level-1))) )//block_a should be out of the current 32x32 block
  {
    block_a_is_direct = IS_DIRECT(&img->mb_data[block_a.mb_addr]);
    block_a_b8mode_is_direct = img->mb_data[block_a.mb_addr].b8mode[b8a]==0 && bslice;
  }


  if (!block_b.available)
    b=0;
  else if (( block_b_is_direct && !giRDOpt_B8OnlyFlag) || block_b_b8mode_is_direct )
    b=0;
  else
  {
      b = (refframe_array[block_b.pos_y][block_b.pos_x] > 0 ? 1 : 0);
  }

  if (!block_a.available)
    a=0;
  else if (( block_a_is_direct && !giRDOpt_B8OnlyFlag) || block_a_b8mode_is_direct)
    a=0;
  else 
  {
      a = (refframe_array[block_a.pos_y][block_a.pos_x] > 0 ? 1 : 0);
  }

  act_ctx     = a + 2*b; 
  se->context = act_ctx; // store context
  act_sym     = se->value1;

  if (act_sym==0)
  {
    biari_encode_symbol(eep_dp, 0, ctx->ref_no_contexts[addctx] + act_ctx );
  }
  else
  {
    biari_encode_symbol(eep_dp, 1, ctx->ref_no_contexts[addctx] + act_ctx);
    act_sym--;
    act_ctx=4;
    unary_bin_encode(eep_dp, act_sym,ctx->ref_no_contexts[addctx]+act_ctx,1);
  }
}


void writeMVRES_4_8_CABAC32(SyntaxElement *se, EncodingEnvironmentPtr eep_dp)
{

  MotionInfoContexts  *ctx    = img->currentSlice->mot_ctx;
  PixelPos block_a, block_b;
  int mb_nr=img->current_mb_nr, subblock_x=0, subblock_y=0; 
  int mb_ext_level = se->mb_ext_level;
  int a_x,a_y, b_x,b_y;
  int act_ctx;
  int  mv_res = se->value1;
  int b_res,a_res;

  if(img->subblock_x==0 && img->subblock_y==0 )
  {
    subblock_x = 0;
    subblock_y = 0;
    mb_nr      = img->current_mb_nr;
  }
  else if(img->subblock_x==0 && img->subblock_y==(2<<mb_ext_level) )
  {
    subblock_x = 0;
    subblock_y = 0;
    mb_nr      = img->current_mb_nr + (img->PicWidthInMbs << (mb_ext_level-1));
  }
  else if(img->subblock_x==(2<<mb_ext_level) && img->subblock_y==0 )
  {
    subblock_x = 0;
    subblock_y = 0;
    mb_nr      = img->current_mb_nr + (1<<(mb_ext_level- 1));
  }
  else
  {
    error("subblock_x and subblock_y are out of the range\n", -1);
  }

  getLuma4x4Neighbour(mb_nr, subblock_x, subblock_y, -1,  0, &block_a);
  getLuma4x4Neighbour(mb_nr, subblock_x, subblock_y,  0, -1, &block_b);

  if (block_b.available)
  {
	b_x = absm(enc_picture->mv[0][block_b.pos_y][block_b.pos_x][0]);
	b_y = absm(enc_picture->mv[0][block_b.pos_y][block_b.pos_x][1]);
  }
  else
  {
	  b_x=2;  
	  b_y=2;
  }          
  if (block_a.available)
  {
    a_x = absm(enc_picture->mv[0][block_a.pos_y][block_a.pos_x][0]);
	a_y = absm(enc_picture->mv[0][block_a.pos_y][block_a.pos_x][1]);
  }
  else
  {
	  a_x=2;  
	  a_y=2;
  }

  b_res =0;
  if (((b_x&1)==0)&& ((b_y&1)==0))
	b_res=1;
  a_res =0;
  if (((a_x&1)==0)&& ((a_y&1)==0))
	a_res=1;

  act_ctx = a_res*2 + b_res;

  biari_encode_symbol(eep_dp, mv_res, &ctx->adp_mv_res_contexts[0][act_ctx] );

}



void writeMVD_CABAC32(SyntaxElement *se, EncodingEnvironmentPtr eep_dp)
{

  int a, b;
  int act_ctx;
  int act_sym;
  int mv_pred_res;
  int mv_local_err;
  int mv_sign;

  int mv_res = se->value2 & 0x01;
  int list_idx = (se->value2>>1) & 0x01;
  int k = (se->value2>>2); // MVD component


  PixelPos block_a, block_b;

  MotionInfoContexts  *ctx    = img->currentSlice->mot_ctx;
//  Macroblock          *currMB = &img->mb_data[img->current_mb_nr];

  int mb_nr=img->current_mb_nr, subblock_x=0, subblock_y=0; 
  int mb_ext_level = se->mb_ext_level;

  if(img->subblock_x==0 && img->subblock_y==0 )
  {
    subblock_x = 0;
    subblock_y = 0;
    mb_nr      = img->current_mb_nr;
  }
  else if(img->subblock_x==0 && img->subblock_y==(2<<mb_ext_level) )
  {
    subblock_x = 0;
    subblock_y = 0;
    mb_nr      = img->current_mb_nr + (img->PicWidthInMbs << (mb_ext_level-1));
  }
  else if(img->subblock_x==(2<<mb_ext_level) && img->subblock_y==0 )
  {
    subblock_x = 0;
    subblock_y = 0;
    mb_nr      = img->current_mb_nr + (1<<(mb_ext_level- 1));
  }
  else
  {
    error("subblock_x and subblock_y are out of the range\n", -1);
  }

  getLuma4x4Neighbour(mb_nr, subblock_x, subblock_y, -1,  0, &block_a);
  getLuma4x4Neighbour(mb_nr, subblock_x, subblock_y,  0, -1, &block_b);

  if (block_b.available)
  {
    b = absm(img->mb_data[block_b.mb_addr].mvd[list_idx][block_b.y][block_b.x][k]);

	b= b >> mv_res;

  }
  else
    b=0;
          
  if (block_a.available)
  {
    a = absm(img->mb_data[block_a.mb_addr].mvd[list_idx][block_a.y][block_a.x][k]);

	a= a >> mv_res;

  }
  else
    a = 0;

  if ((mv_local_err=a+b)<3)
    act_ctx = 5*k;
  else
  {
    if (mv_local_err>32)
      act_ctx=5*k+3;
    else
      act_ctx=5*k+2;
  }

  mv_pred_res = se->value1;
  se->context = act_ctx;

  act_sym = absm(mv_pred_res);

  if (act_sym == 0)
    biari_encode_symbol(eep_dp, 0, &ctx->mv_res_contexts[0][act_ctx] );
  else
  {
    biari_encode_symbol(eep_dp, 1, &ctx->mv_res_contexts[0][act_ctx] );
    act_sym--;
    act_ctx=5*k;
    unary_exp_golomb_mv_encode(eep_dp,act_sym,ctx->mv_res_contexts[1]+act_ctx,3);
    mv_sign = (mv_pred_res<0) ? 1: 0;
    biari_encode_symbol_eq_prob(eep_dp, (signed short) mv_sign);
  }
}

void writeCBP_CABAC32(SyntaxElement *se, EncodingEnvironmentPtr eep_dp)
{
//  TextureInfoContexts *ctx = img->currentSlice->tex_ctx;
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];

  int cbp = se->value1; // symbol to encode
  int curr_cbp_idx;
  
  curr_cbp_idx = 1;
  writeCBP_BIT_CABAC (0, cbp, cbp, currMB, curr_cbp_idx, eep_dp);
}

void writeCBP_CABAC_luma1bit(SyntaxElement *se, EncodingEnvironmentPtr eep_dp)
{
//  TextureInfoContexts *ctx = img->currentSlice->tex_ctx;
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];

  int cbp = se->value1; // symbol to encode
  int curr_cbp_idx;

  curr_cbp_idx = 1;

  if((cbp&15)!=0)
    writeCBP_BIT_CABAC (0, 1, 1, currMB, curr_cbp_idx, eep_dp);
  else
    writeCBP_BIT_CABAC (0, 0, 0, currMB, curr_cbp_idx, eep_dp);


}

#endif


void SetMotionVectorPredictor32 (short  pmv[2],
                               char   **refPic,
                               short  ***tmp_mv,
                               short  ref_frame,
                               int    list,
                               int    block_x,
                               int    block_y,
                               int    blockshape_x,
                               int    blockshape_y,
                               int    mb_ext_level);

void writeMVD_CABAC32(SyntaxElement *se, EncodingEnvironmentPtr eep_dp);
/*!
************************************************************************
* \brief
*    Writes motion vectors of an 16x16 block
************************************************************************
*/
int write32MotionVector8x8 (int  i0,
                          int  j0,
                          int  i1,
                          int  j1,
                          int  refframe,
                          int  list_idx,
                          int  mv_mode,
                          int  mb_ext_level)
{
  int            i, j, k, l, m;
  int            curr_mvd;
  DataPartition* dataPart;
  
  int            rate       = 0;
  int            step_h     = input->part_size[mv_mode][0] << mb_ext_level;
  int            step_v     = input->part_size[mv_mode][1] << mb_ext_level;
  Macroblock*    currMB     = &img->mb_data[img->current_mb_nr];
  SyntaxElement* currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*         currSlice  = img->currentSlice;
  int*           bitCount   = currMB->bitcounter;
  const int*     partMap    = assignSE2partition[input->partition_mode];
  int            refindex   = refframe;
  
  short******    all_mv     = img->all_mv;
  short******    pred_mv    = img->pred_mv;
  
  if (currMB->bi_pred_me && currMB->b8pdir[0]==2 && mv_mode == 1 && refindex == 0)
    all_mv = currMB->bi_pred_me == 1? img->bipred_mv1 : img->bipred_mv2 ;  

  for (j=j0; j<j1; j+=step_v)
  {
    for (i=i0; i<i1; i+=step_h)
    {

      int mv_shift =0; //default to 1/8 pel
      int currMvVec;
      int predMvVec;


#ifdef MV_COMPETITION
#ifdef RDO_Q
      if((input->Transform8x8Mode)&&(pass_with_writing == 1) && (input->mv_competition > 0) &&(!input->UseRDO_Q))
#endif
      {
        short best_prediction[2];
        SetMotionVectorPredictor_Competition32 (best_prediction, enc_picture->ref_idx[list_idx], enc_picture->mv[list_idx], (short)refindex, list_idx, i, j,(input->blc_size[mv_mode][0]<<mb_ext_level),(input->blc_size[mv_mode][1]<<mb_ext_level),mb_ext_level);
        GetBestPredVector(best_prediction, all_mv[j][i][list_idx][refindex][mv_mode][0],all_mv[j][i][list_idx][refindex][mv_mode][1],img->lambda_mf[img->type][img->qp],refindex,list_idx,mv_mode,i*4,j*4); 
        pred_mv[j][i][list_idx][refindex][mv_mode][0]=best_prediction[0];
        pred_mv[j][i][list_idx][refindex][mv_mode][1]=best_prediction[1];
      }
#endif
      

#ifdef RDO_Q
      if(input->UseRDO_Q)
      {
#ifdef MV_COMPETITION
        if( (input->mv_competition > 0))    
        {
          short best_prediction[2];
          SetMotionVectorPredictor_Competition32 (best_prediction, enc_picture->ref_idx[list_idx], enc_picture->mv[list_idx], (short)refindex, list_idx, i, j,(input->blc_size[mv_mode][0]<<mb_ext_level),(input->blc_size[mv_mode][1]<<mb_ext_level),mb_ext_level);
          GetBestPredVector(best_prediction, all_mv[j][i][list_idx][refindex][mv_mode][0],all_mv[j][i][list_idx][refindex][mv_mode][1],img->lambda_mf[img->type][img->qp],refindex,list_idx,mv_mode,i*4,j*4); 
          pred_mv[j][i][list_idx][refindex][mv_mode][0]=best_prediction[0];
          pred_mv[j][i][list_idx][refindex][mv_mode][1]=best_prediction[1];
        }
        else
        {
#endif
          short pred_mv_stored[2];
          SetMotionVectorPredictor32 (pred_mv_stored, enc_picture->ref_idx[list_idx], enc_picture->mv[list_idx], (short)refindex, list_idx, i, j, (input->blc_size[mv_mode][0]<<mb_ext_level),(input->blc_size[mv_mode][1]<<mb_ext_level), mb_ext_level);
          pred_mv[j][i][list_idx][refindex][mv_mode][0]=pred_mv_stored[0];
          pred_mv[j][i][list_idx][refindex][mv_mode][1]=pred_mv_stored[1];
#ifdef MV_COMPETITION
        }  
#endif
      }
#endif   //RDO_Q


	  if (img->mv_res)
	  { 
		  if (((all_mv[j][i][list_idx][refindex][mv_mode][0]& 1) ==0)&&((all_mv[j][i][list_idx][refindex][mv_mode][1]& 1))==0)
			  mv_shift =1; //switch to 1/4 pel
		  if(input->symbol_mode == CABAC)
		  {
			  img->subblock_x = i; // position used for context determination
			  img->subblock_y = j; // position used for context determination
			  currSE->value1 = mv_shift;
  			  currSE->type   = SE_MVD; 
			  currSE->mb_ext_level  = mb_ext_level;
			  currSE->value2 = 0;
			  currSE->writing  = writeMVRES_4_8_CABAC32;
			  dataPart = &(currSlice->partArr[partMap[SE_MVD]]);
			  dataPart->writeSyntaxElement (currSE, dataPart);
#if TRACE	
			  snprintf(currSE->tracestring, TRACESTRING_SIZE, "mv_res = %3d  ", mv_shift);
#endif
			  bitCount[BITS_INTER_MB] += currSE->len;
			  rate                    += currSE->len;
			  currSE++;  
			  currMB->currSEnr++;

		  }
	  }
	  else
	  {
		  mv_shift=0;
	  }


      for (k=0; k<2; k++) 
      {        

		if (img->mv_res && mv_shift )
		{
			    int scaled_mvd ;
				currMvVec=roundInt(all_mv[j][i][list_idx][refindex][mv_mode][k],mv_shift);
				predMvVec=roundInt(pred_mv[j][i][list_idx][refindex][mv_mode][k],mv_shift);
				curr_mvd = currMvVec - predMvVec;
		        scaled_mvd = curr_mvd*(1<< mv_shift);
				//--- store (oversampled) mvd ---
				//only save mvd for the upper left 16x16 block for the context of the 2nd partition of 32x16 and 16x32        
				if(j==0 && i==0)
				{
				  for (l=0; l < 4; l++)
				  for (m=0; m < 4; m++)
				  {
					currMB->mvd[list_idx][j+l][i+m][k] = scaled_mvd; //mvd[2][BLOCK_MULTIPLE][BLOCK_MULTIPLE][2]
				  }
				}
				//the mvd for the whole 32x32 is saved in mb32_mvd and will be copied to each 16x16 block's mvd in write_macroblock_cluster
				for (l=0; l < step_v; l++)
				  for (m=0; m < step_h; m++)
				  {
					mb32_mvd[list_idx][j+l][i+m][k] = scaled_mvd;
				  }
			
	    }
		else

		{
			curr_mvd = all_mv[j][i][list_idx][refindex][mv_mode][k] - pred_mv[j][i][list_idx][refindex][mv_mode][k];
	        
			//--- store (oversampled) mvd ---
			//only save mvd for the upper left 16x16 block for the context of the 2nd partition of 32x16 and 16x32        
			if(j==0 && i==0)
			{
			  for (l=0; l < 4; l++)
			  for (m=0; m < 4; m++)
			  {
				currMB->mvd[list_idx][j+l][i+m][k] = curr_mvd; //mvd[2][BLOCK_MULTIPLE][BLOCK_MULTIPLE][2]
			  }
			}
			//the mvd for the whole 32x32 is saved in mb32_mvd and will be copied to each 16x16 block's mvd in write_macroblock_cluster
			for (l=0; l < step_v; l++)
			  for (m=0; m < step_h; m++)
			  {
				mb32_mvd[list_idx][j+l][i+m][k] = curr_mvd;
			  }
		}
          currSE->value1 = curr_mvd;
          currSE->value2 = 0;
          currSE->type   = SE_MVD;
          if (input->symbol_mode == UVLC)
          {
            currSE->mapping = se_linfo;
          }
          else
          {
            img->subblock_x = i; // position used for context determination
            img->subblock_y = j; // position used for context determination

		    currSE->value2  = (2*k+list_idx)*2+mv_shift;

            currSE->mb_ext_level  = mb_ext_level;
            currSE->writing = writeMVD_CABAC32;
          }  
          dataPart = &(currSlice->partArr[partMap[SE_MVD]]);
          dataPart->writeSyntaxElement (currSE, dataPart);
          bitCount[BITS_INTER_MB] += currSE->len;
          rate                    += currSE->len;
          currSE++;  
          currMB->currSEnr++;
      }
#ifdef MV_COMPETITION
      if (input->mv_competition > 0)
      { 
        int maxmode=0;
        if (img->type == P_SLICE)
          maxmode = mv_comp.nb_mode_for_mvp;
        else if (img->type == B_SLICE)
          maxmode = mv_comp.nb_mode_for_mvb;
        //Rahul ----- reverse logic....0 means "send it", 1 means "dont send"
        if (send_index_mv_prediction[mv_mode][refindex][list_idx][j][i][maxmode] == 0)
          rate+=writeIndexForMotionVectorPrediction (refframe, list_idx, mv_mode, i, j);
      }
#endif
    }
  }
  return rate;
}


void writeRefFrame_CABAC32(SyntaxElement *se, EncodingEnvironmentPtr eep_dp);

/*!
************************************************************************
* \brief
*    Codes the reference frame
************************************************************************
*/
int writeReferenceFrame32 (int mode, int i, int j, int fwd_flag, int  ref, int  mb_ext_level)
{
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  int*            bitCount  = currMB->bitcounter;
  const int*      partMap   = assignSE2partition[input->partition_mode];
  int             rate      = 0;
  DataPartition*  dataPart;
  int             num_ref   = ( fwd_flag ? listXsize[LIST_0+currMB->list_offset]: listXsize[LIST_1+currMB->list_offset]);
  int             flag_mode = 0;
  
  if( num_ref == 1 )
  {
    return 0;
  }
  
  if ( num_ref == 2 )
  {
    flag_mode = 1;
  }
  
  currSE->value1 = ref;
  currSE->value2  = 0;
  currSE->type   = SE_REFFRAME;
  
  dataPart = &(currSlice->partArr[partMap[currSE->type]]);
  if (input->symbol_mode == UVLC)
  {

    if (img->mv_res)
    {
      if(ref == 0 && mode <= 4)
        currSE->value2 = ref_idx_ctx(mode, i, j, !fwd_flag, ref, mb_ext_level);
      write_ref_idx(mode, currSE, dataPart);
    }
    else

    {
      if( flag_mode )
      {
        currSE->bitpattern = 1 - currSE->value1;
        currSE->len = 1;
        writeSyntaxElement2Buf_Fixed(currSE, dataPart->bitstream);
      }
      else
      {
        currSE->mapping = ue_linfo;     
        dataPart->writeSyntaxElement (currSE, dataPart);
      }
    }
  }
  else
  {
    currSE->context = BType2CtxRef (mode);
    img->subblock_x = i; // position used for context determination
    img->subblock_y = j; // position used for context determination
    currSE->mb_ext_level = mb_ext_level;
    currSE->writing = writeRefFrame_CABAC32;
    currSE->value2 = (fwd_flag)? LIST_0:LIST_1;
    dataPart->writeSyntaxElement (currSE, dataPart);
  }
  bitCount[BITS_INTER_MB] += currSE->len;
  rate                    += currSE->len;
  currSE++;
  currMB->currSEnr++;
  
  return rate;
}


/*!
************************************************************************
* \brief
*    Writes motion info
************************************************************************
*/
int writeMotionInfo2NAL32 (int mb_ext_level)
{
  int k, j0, i0, refframe;
  int jj;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  int             no_bits   = 0;
  int   bframe          = (img->type==B_SLICE);
  int   step_h0         = ((input->blc_size[IS_P8x8(currMB) ? 4 : currMB->mb_type][0]<<mb_ext_level) >> 2);
  int   step_v0         = ((input->blc_size[IS_P8x8(currMB) ? 4 : currMB->mb_type][1]<<mb_ext_level) >> 2);

  //=== If multiple ref. frames, write reference frame for the MB ===
  if (IS_INTERMV (currMB))
  {
#ifdef GROUP_REF
    int group_flag=0;
    if(input->symbol_mode == UVLC)
    {
    if(currMB->mb_type>1 && listXsize[LIST_0]>1)
    {
      const int*      partMap   = assignSE2partition[input->partition_mode];
      group_flag = group_ref (currMB, mb_ext_level);
      no_bits += u_1 ("group_flag", group_flag, (img->currentSlice->partArr[partMap[SE_REFFRAME]]).bitstream);
    }
    }
#endif
    // if UVLC is turned on, a 8x8 macroblock with all ref=0 in a P-frame is signalled in macroblock mode
#ifdef GROUP_REF
    if((input->symbol_mode==UVLC && group_flag==0) || input->symbol_mode==CABAC)
#else
    if (!IS_P8x8 (currMB) || !ZeroRef (currMB) || input->symbol_mode==CABAC || bframe)
#endif
    {
      for (j0=0; j0<(4<<mb_ext_level); j0+=step_v0)
      {
        jj = img->block_y+j0;
        for (i0=0; i0<(4<<mb_ext_level); i0+=step_h0)
        {
          k=(j0>>mb_ext_level)+(i0 >> (1+mb_ext_level));
          if ((currMB->b8pdir[k]==0 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)//has forward vector
          {
            no_bits += writeReferenceFrame32 (currMB->b8mode[k], i0, j0, 1, enc_picture->ref_idx[LIST_0][jj][img->block_x+i0], mb_ext_level);
          }
//#ifdef GROUP_REF
//          if(group_flag)
//            break;
//#endif
        }
//#ifdef GROUP_REF
//        if(group_flag)
//          break;
//#endif
      }

      for (j0=0; j0<(4<<mb_ext_level); j0+=step_v0)
      {
        jj = img->block_y+j0;
        for (i0=0; i0<(4<<mb_ext_level); i0+=step_h0)
        {
          k=(j0>>mb_ext_level)+(i0 >> (1+mb_ext_level));
          if ((currMB->b8pdir[k]==1 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)//has backward vector
          {
            no_bits += writeReferenceFrame32 (currMB->b8mode[k], i0, j0, 0, enc_picture->ref_idx[LIST_1][jj][img->block_x+i0], mb_ext_level);
          }
        }
      }
    }
  }
  
  //===== write forward motion vectors =====
  if (IS_INTERMV (currMB))
  {
    for (j0=0; j0<(4<<mb_ext_level); j0+=step_v0)
    {
      jj = img->block_y+j0;
      for (i0=0; i0<(4<<mb_ext_level); i0+=step_h0)
      {
        k=(j0>>mb_ext_level)+(i0 >> (1+mb_ext_level));
        if ((currMB->b8pdir[k]==0 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)//has forward vector
        {
          refframe  = enc_picture->ref_idx[LIST_0][jj][img->block_x+i0];
          no_bits  += write32MotionVector8x8 (i0, j0, i0+step_h0, j0+step_v0, refframe, LIST_0, currMB->b8mode[k], mb_ext_level);
        }
      }
    }
  }
  
  
  //===== write backward motion vectors =====
  if (IS_INTERMV (currMB) && bframe)
  {
    for (j0=0; j0<(4<<mb_ext_level); j0+=step_v0)
    {
      jj = img->block_y+j0;
      for (i0=0; i0<(4<<mb_ext_level); i0+=step_h0)
      {
        k=(j0>>mb_ext_level)+(i0 >> (1+mb_ext_level));
        if ((currMB->b8pdir[k]==1 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)//has backward vector
        {
          refframe  = enc_picture->ref_idx[LIST_1][jj][img->block_x+i0];
          no_bits  += write32MotionVector8x8 (i0, j0, i0+step_h0, j0+step_v0, refframe, LIST_1, currMB->b8mode[k], mb_ext_level);
        }
      }
    }
  }
  return no_bits;
}


void writeMB_typeInfo_CABAC32(SyntaxElement *se, EncodingEnvironmentPtr eep_dp)
{
  int a, b;
  int act_ctx = 0;
  int act_sym;
#if !B_MBTYPE
  signed short csym;
#endif
  int bframe   = (img->type==B_SLICE);
  int mode_sym = 0;
  int mode16x16;


  MotionInfoContexts *ctx         = (img->currentSlice)->mot_ctx;
  Macroblock         *currMB      = &img->mb_data[img->current_mb_nr];
  int                curr_mb_type = se->value1;

#if B_MBTYPE
  int dir1 = se->dir_partition1;
  int dir2 = se->dir_partition2;
  if(bframe)
    assert(curr_mb_type<=3 || curr_mb_type>=22);
#endif

  if(img->type == I_SLICE)  // INTRA-frame
  {
    if (currMB->mb_available_up == NULL)
      b = 0;
    else 
      b = ((currMB->mb_available_up->mb_type != I4MB &&  currMB->mb_available_up->mb_type != I8MB) ? 1 : 0 );

    if (currMB->mb_available_left == NULL)
      a = 0;
    else 
      a = ((currMB->mb_available_left->mb_type != I4MB &&  currMB->mb_available_left->mb_type != I8MB) ? 1 : 0 );
    
    act_ctx     = a + b;
    act_sym     = curr_mb_type;
    se->context = act_ctx; // store context

    if (act_sym==0) // 4x4 Intra
    {
      biari_encode_symbol(eep_dp, 0, ctx->mb_type_contexts[0] + act_ctx );
    }
    else // 16x16 Intra
    {
      biari_encode_symbol(eep_dp, 1, ctx->mb_type_contexts[0] + act_ctx );

      biari_encode_symbol_final(eep_dp, 0);

      mode_sym = act_sym-1; // Values in the range of 0...23
      act_ctx  = 4;
      act_sym  = mode_sym/12;
      biari_encode_symbol(eep_dp, (signed short) act_sym, ctx->mb_type_contexts[0] + act_ctx ); // coding of AC/no AC
      mode_sym = mode_sym % 12;
      act_sym  = mode_sym / 4; // coding of cbp: 0,1,2
      act_ctx  = 5;
      if (act_sym==0)
      {
        biari_encode_symbol(eep_dp, 0, ctx->mb_type_contexts[0] + act_ctx );
      }
      else
      {
        biari_encode_symbol(eep_dp, 1, ctx->mb_type_contexts[0] + act_ctx );
        act_ctx=6;
        biari_encode_symbol(eep_dp, (signed short) (act_sym!=1), ctx->mb_type_contexts[0] + act_ctx );
      }
      mode_sym = mode_sym & 0x03; // coding of I pred-mode: 0,1,2,3
      act_sym  = mode_sym >> 1;
      act_ctx  = 7;
      biari_encode_symbol(eep_dp, (signed short) act_sym, ctx->mb_type_contexts[0] + act_ctx );
      act_ctx  = 8;
      act_sym  = mode_sym & 0x01;
      biari_encode_symbol(eep_dp, (signed short) act_sym, ctx->mb_type_contexts[0] + act_ctx );
    }
  }
  else // INTER
  {
    
    if (bframe)
    {
      if (currMB->mb_available_up == NULL)
        b = 0;
      else
        b = ((currMB->mb_available_up->mb_type != 0) ? 1 : 0 );

      if (currMB->mb_available_left == NULL)
        a = 0;
      else
        a = ((currMB->mb_available_left->mb_type != 0) ? 1 : 0 );
      act_ctx = a + b;
      se->context = act_ctx; // store context
    }
    act_sym = curr_mb_type;

    if (act_sym>=(mode16x16=(bframe?24:7)))
    {
      mode_sym = act_sym-mode16x16;
      act_sym  = mode16x16; // 16x16 mode info
    }

    if (!bframe)
    {
      switch (act_sym)
      {
#if P_DIRECT    
      case 0:       //0001    ...this will only happen when mode=0 and cbp!=0 i.e Pdirect
        //biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[1][4]);  commented by peisong
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[1][5]);
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[1][6]);
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[1][6]);
        break;
      case 1:       //0000    ...need to modify mbtype of P16x16 so as to accomodate Pdirect
        //biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[1][4]);  commented by peisong
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[1][5]);
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[1][6]);
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[1][6]);
        break;
#else
      case 0:
        break;
      case 1:
        //biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[1][4]);  commented by peisong
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[1][5]);
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[1][6]);
        break;
#endif
      case 2:
//        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[1][4]); commented by peisong
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[1][5]);
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[1][7]);
        break;
      case 3:
//        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[1][4]); commented by peisong
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[1][5]);
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[1][7]);
        break;
      case 4:
      case 5:
//        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[1][4]); commented by peisong
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[1][5]);
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[1][6]);
        break;
      //case 6:   this indicates Intra and it doesnt happen for BigBlocks
      //  biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[1][4]);
      //  biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[1][7]);
      //  break;
      //case 7:   this indicates Intra and it doesnt happen for BigBlocks
      //  biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[1][4]);
      //  biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[1][7]);
      //  break;
      default:
        printf ("Unsupported MB-MODE in writeMB_typeInfo_CABAC32!\n");
        exit (1);
      }
    }
    else //===== B-FRAMES =====
    {
      if (act_sym==0)
      {
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[2][act_ctx]);
      }
#if B_MBTYPE
      else if (act_sym==1)  //16x16           101
      {
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][act_ctx]); 
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[2][4]);
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][5]);
        //biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][6]);
        writeB_pDirInfo_CABAC (dir1 , eep_dp, &ctx->mb_type_contexts[2][6]);
      }
      else if (act_sym==2)  //16x8            110
      {
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][act_ctx]); 
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][4]);
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[2][5]);
        //biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[2][6]);
        writeB_pDirInfo_CABAC (dir1 , eep_dp, &ctx->mb_type_contexts[2][6]);
        writeB_pDirInfo_CABAC (dir2 , eep_dp, &ctx->mb_type_contexts[2][6]);
      }
      else if (act_sym==3)  //8x16            111
      {
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][act_ctx]); 
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][4]);
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][5]);
        //biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][6]);
        writeB_pDirInfo_CABAC (dir1 , eep_dp, &ctx->mb_type_contexts[2][6]);
        writeB_pDirInfo_CABAC (dir2 , eep_dp, &ctx->mb_type_contexts[2][6]);
      }
      else if (act_sym==22)  //8x8            100
      {
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][act_ctx]); 
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[2][4]);
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[2][5]);
        //biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][6]);
      }
      else if (act_sym>22)  //Intra modes in B slice      1101 
      {
          //Intra in BigBlocks
          //this will never happen...
      }
#else
      else if (act_sym<=2)
      {
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][act_ctx]);
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[2][4]);
        csym = (act_sym-1 != 0);
        biari_encode_symbol (eep_dp, csym, &ctx->mb_type_contexts[2][6]);
      }
      else if (act_sym<=10)
      {
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][act_ctx]);
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][4]);
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[2][5]);
        csym=(((act_sym-3)>>2)&0x01) != 0;
        biari_encode_symbol (eep_dp, csym, &ctx->mb_type_contexts[2][6]);
        csym=(((act_sym-3)>>1)&0x01) != 0;
        biari_encode_symbol (eep_dp, csym, &ctx->mb_type_contexts[2][6]);
        csym=((act_sym-3)&0x01) != 0;
        biari_encode_symbol (eep_dp, csym, &ctx->mb_type_contexts[2][6]);
      }
      else if (act_sym==11 || act_sym==22)
      {
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][act_ctx]);
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][4]);
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][5]);
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][6]);
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][6]);
        csym = (act_sym != 11);
        biari_encode_symbol (eep_dp, csym, &ctx->mb_type_contexts[2][6]);
      }
      else
      {
        if (act_sym > 22) act_sym--;
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][act_ctx]);
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][4]);
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][5]);
        csym=(((act_sym-12)>>3)&0x01) != 0;
        biari_encode_symbol (eep_dp, csym, &ctx->mb_type_contexts[2][6]);
        csym=(((act_sym-12)>>2)&0x01) != 0;
        biari_encode_symbol (eep_dp, csym, &ctx->mb_type_contexts[2][6]);
        csym=(((act_sym-12)>>1)&0x01) != 0;
        biari_encode_symbol (eep_dp, csym, &ctx->mb_type_contexts[2][6]);
        csym=((act_sym-12)&0x01) != 0;
        biari_encode_symbol (eep_dp, csym, &ctx->mb_type_contexts[2][6]);         
        if (act_sym >=22) act_sym++;
      }
#endif
    }

    if(act_sym==mode16x16) // additional info for 16x16 Intra-mode
    {
      if( mode_sym==25 )
      {
        biari_encode_symbol_final(eep_dp, 1 );
        return;
      }
      biari_encode_symbol_final(eep_dp, 0 );

      act_ctx = 8;
      act_sym = mode_sym/12;
      biari_encode_symbol(eep_dp, (signed short) act_sym, ctx->mb_type_contexts[1] + act_ctx ); // coding of AC/no AC
      mode_sym = mode_sym % 12;

      act_sym = mode_sym / 4; // coding of cbp: 0,1,2
      act_ctx = 9;
      if (act_sym==0)
      {
        biari_encode_symbol(eep_dp, 0, ctx->mb_type_contexts[1] + act_ctx );
      }
      else
      {
        biari_encode_symbol(eep_dp, 1, ctx->mb_type_contexts[1] + act_ctx );
        biari_encode_symbol(eep_dp, (signed short) (act_sym!=1), ctx->mb_type_contexts[1] + act_ctx );
      }

      mode_sym = mode_sym % 4; // coding of I pred-mode: 0,1,2,3
      act_ctx  = 10;
      act_sym  = mode_sym/2;
      biari_encode_symbol(eep_dp, (signed short) act_sym, ctx->mb_type_contexts[1] + act_ctx );
      act_sym  = mode_sym%2;
      biari_encode_symbol(eep_dp, (signed short) act_sym, ctx->mb_type_contexts[1] + act_ctx );
    }
  }
}


#ifdef MB32X32
int writeCBP32(int rdopt, int cbp, int dquant)
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  int*            bitCount  = currMB->bitcounter;
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[input->partition_mode];
  //int             cbp       = currMB->cbp;
  DataPartition*  dataPart;

  
//  int   b8, b4;
//  int*  DCLevel = img->cofDC[0][0];
//  int*  DCRun   = img->cofDC[0][1];
//  int*  ACLevel;
//  int*  ACRun;
  
#ifdef MB32X32
  if (1)//(!IS_NEWINTRA (&MB32))
#else
  if (!IS_NEWINTRA (currMB))
#endif
  {
    //=====   C B P   =====
    //---------------------
    currSE->value1 = cbp;

    if (input->symbol_mode == UVLC)  currSE->mapping = cbp_linfo_inter;
    currSE->type = SE_CBP_INTER;
    if (input->symbol_mode == CABAC)   currSE->writing = writeCBP_CABAC32;
    
    // choose the appropriate data partition
    dataPart = &(currSlice->partArr[partMap[currSE->type]]);
    
    dataPart->writeSyntaxElement(currSE, dataPart);
    bitCount[BITS_CBP_MB] += currSE->len;
    rate                  += currSE->len;
    // proceed to next SE
    currSE++;  
    currMB->currSEnr++;        
  }
  

  if(cbp != 0 && dquant)
  {
        currSE->value1 = currMB->delta_qp;
      
      if (input->symbol_mode==UVLC)   currSE->mapping = se_linfo;
      else                            currSE->writing = writeDquant_CABAC;
      
      if (IS_INTER (currMB))  currSE->type = SE_DELTA_QUANT_INTER;
      else                    currSE->type = SE_DELTA_QUANT_INTRA;
      
      
      // choose the appropriate data partition
      dataPart = &(img->currentSlice->partArr[partMap[currSE->type]]);
      dataPart->writeSyntaxElement(  currSE, dataPart);
      bitCount[BITS_DELTA_QUANT_MB] += currSE->len;
      rate                          += currSE->len;
      // proceed to next SE
      currSE++;
      currMB->currSEnr++;
  }

  return rate;
}


#endif
int mbtype_mapping[]={1,2,3,0};
void re_mapping(int *mbtype_value, int mb_ext_level)
{
  if(mb_ext_level>0)
    *mbtype_value = mbtype_mapping[*mbtype_value];
}

int write_mb_type_cavlc(Macroblock *MB, int mb_ext_level)
{
  int             mb_nr      = img->current_mb_nr;
  Macroblock*     currMB     = &img->mb_data[mb_nr];
  SyntaxElement  *currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
  int*            bitCount   = currMB->bitcounter;
  Slice*          currSlice  = img->currentSlice;
  DataPartition*  dataPart;
  const int*      partMap    = assignSE2partition[input->partition_mode];
  int             no_bits    = 0;
  // choose the appropriate data partition
  dataPart = &(currSlice->partArr[partMap[SE_MBTYPE]]);


  if (MB->mb_type != 0 || ((img->type == B_SLICE) && MB->cbp != 0))//non-skip
  {
    //===== Run Length Coding: Non-Skipped macroblock =====
    if((MAX_MB_EXT_LEVEL==2 && (mb_ext_level==2 || (no64x64 && mb_ext_level==1) || (no32x32 && mb_ext_level==0)))
      ||(MAX_MB_EXT_LEVEL==1 && (mb_ext_level==1 || (no32x32 && mb_ext_level==0))) )//tm_vlc
    {
      currSE->value1  = img->cod_counter;
      currSE->value2  = 0;
      currSE->mapping = ue_linfo;
      currSE->type    = SE_MBTYPE;
      dataPart->writeSyntaxElement( currSE, dataPart);
    }
    else 
    {
      currSE->len = u_1 ("mb_skip_flag", 0, dataPart->bitstream);    
    }

    bitCount[BITS_MB_MODE] += currSE->len;
    no_bits                += currSE->len;
    currSE++;
    currMB->currSEnr++;
    
#ifdef MV_COMPETITION
    if ((input->mv_competition > 0)&& (rdopt==0))
      if ((img->type==P_SLICE) && (mv_comp.nb_mode_for_skip > 1))  
        no_bits  += write_predictor_index_for_skip_mode();
#endif
      
      // Reset cod counter
    //if((MAX_MB_EXT_LEVEL==2 && (mb_ext_level==2 || (no64x64 && mb_ext_level==1) || (no32x32 && mb_ext_level==0)))
    //  ||(MAX_MB_EXT_LEVEL==1 && (mb_ext_level==1 || (no32x32 && mb_ext_level==0))) )//tm_vlc
      img->cod_counter = 0;

      // Put out mb mode
      currSE->value1  = MBType2Value (MB);
      
      if (img->type != B_SLICE)
      {
        currSE->value1--;

        if(currSE->value1 == 4 && mb_ext_level > 0)
          currSE->value1 = 3;


        if(currSE->value1 == 4 && mb_ext_level == 0)
        {
          if(!ZeroRef_mvres (currMB))
            currSE->value1 = 3;
          else
            currSE->value1 = 4;

        }
      }

      //re_mapping(&(currSE->value1), mb_ext_level);

      if(mb_ext_level>0)
      {
        currSE->type    = SE_MBTYPE;
        currSE->value2  = 0;
        writeSyntaxElement_bigblk_type(currSE, dataPart);
      }
      else
      {
      currSE->mapping = ue_linfo;

      currSE->type    = SE_MBTYPE;
      currSE->value2  = 0;

      dataPart->writeSyntaxElement( currSE, dataPart);
      }
      bitCount[BITS_MB_MODE] += currSE->len;
      no_bits                += currSE->len;
      currSE++;
      currMB->currSEnr++;


      if(MB->mb_type == 3 || MB->geo_mode != -1)
      {
        currSE->value1  = MB->geo_mode;
        currSE->value2  = rho_bits[mb_ext_level]+THETA_BITS;
        currSE->len     = writeMB_geoInfo_VLC(currSE, dataPart);
        bitCount[BITS_MB_MODE] += currSE->len;
        no_bits                += currSE->len;
        currSE++;
        currMB->currSEnr++;
      }
  }
  else //skip
  {
    //Run Length Coding: Skipped macroblock
    if((MAX_MB_EXT_LEVEL==2 && (mb_ext_level==2 || (no64x64 && mb_ext_level==1) || (no32x32 && mb_ext_level==0)))      
      ||(MAX_MB_EXT_LEVEL==1 && (mb_ext_level==1 || (no32x32 && mb_ext_level==0))) )//tm_vlc
      img->cod_counter++;
    else
    {
      currSE->len = u_1 ("mb_skip_flag", 1, dataPart->bitstream);   
      bitCount[BITS_MB_MODE] += currSE->len;
      no_bits                += currSE->len;
      currSE++;
      currMB->currSEnr++;
    }
    
    currMB->skip_flag = 1;

    if(input->symbol_mode==UVLC)
      Setnz_coeff(img->current_mb_nr, mb_ext_level);

#ifdef MV_COMPETITION
      if (input->mv_competition > 0)
        currMB->send_index_predictor_skip = send_index_for_skip_prediction(currMB->best_predictor_for_skip);
#endif      
      
      //
      //if((MAX_MB_EXT_LEVEL==2 && mb_ext_level==2) ||(MAX_MB_EXT_LEVEL==1 && mb_ext_level==1))//tm_vlc
      {
        
        int mb_nr = img->current_mb_nr + img->PicWidthInMbs*((1<<mb_ext_level)-1) + (1<<mb_ext_level);
 
        //if(FmoGetNextMBNr(img->current_mb_nr) == -1 && img->cod_counter>0)
        if(mb_nr == (int)img->PicSizeInMbs && img->cod_counter>0)//tm_vlc
        {
          // Put out run
          currSE->value1  = img->cod_counter;
          currSE->value2  = 0;
          currSE->mapping = ue_linfo;
          currSE->type    = SE_MBTYPE;

          dataPart->writeSyntaxElement( currSE, dataPart);
          bitCount[BITS_MB_MODE] += currSE->len;
          no_bits                += currSE->len;
          currSE++;
          currMB->currSEnr++;

#ifdef MV_COMPETITION
          if (input->mv_competition > 0)
          {
            if(rdopt==0)
            {
              if ((img->type==P_SLICE) && (mv_comp.nb_mode_for_skip > 1))
                no_bits  += write_predictor_index_for_skip_mode();
            }
          }
#endif     

          // Reset cod counter
          img->cod_counter = 0;

        }
      }
  }

  return no_bits;
}

int write_mb_type(Macroblock *MB, int mb_ext_level)
{
  int             mb_nr      = img->current_mb_nr;
  Macroblock*     currMB     = &img->mb_data[mb_nr];
  int             mb_type;  
  SyntaxElement  *currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
  int*            bitCount   = currMB->bitcounter;
  Slice*          currSlice  = img->currentSlice;
  DataPartition*  dataPart;
  const int*      partMap    = assignSE2partition[input->partition_mode];
  int             no_bits    = 0;

  // choose the appropriate data partition
  dataPart = &(currSlice->partArr[partMap[SE_MBTYPE]]);

    //========= write mb_skip_flag (CABAC) =========
    if(MB->geo_mode != -1)
      mb_type         = MBType2Value_geo (MB);
    else
    mb_type         = MBType2Value (MB); //MB64 set in write_macroblock_cluster

      currSE->value1  = mb_type;
    currSE->value2  = MB->cbp; //MB32.cbp set in write_macroblock_cluster
    currSE->type    = SE_MBTYPE;
    currSE->writing = writeMB_skip_flagInfo_CABAC;
    dataPart->writeSyntaxElement( currSE, dataPart);
    bitCount[BITS_MB_MODE] += currSE->len;
    no_bits                += currSE->len;
    currSE++;
    currMB->currSEnr++;


    CheckAvailabilityOfNeighborsCABAC();

    //========= write mb_type (CABAC) =========
#if P_DIRECT
    if (MB->mb_type != 0 || MB->cbp != 0)
#else
    if (MB->mb_type != 0 || ((img->type == B_SLICE) && MB->cbp != 0))//non-skip
#endif
    {
      currSE->value1  = mb_type;
      currSE->value2  = 0;
      currSE->type    = SE_MBTYPE;
#if B_MBTYPE//_BIG    //to see what value to set in dir1 and dir2, check input parameter in function MBType2Value()
      if(MB->geo_mode != -1)
      {
        //mbtype = 3;//use B_8x16
        currSE->dir_partition1  = best8x8pdir_geo[MB->geo_mode][0];
        currSE->dir_partition2  = best8x8pdir_geo[MB->geo_mode][1];
      }
      else
      {
      currSE->dir_partition1 = MB->b8pdir[0];//this is for Partition 1
      currSE->dir_partition2 = MB->b8pdir[3];//this is for Partition 2
      }
#endif
      if(mb_ext_level > 0)
        currSE->writing = writeMB_typeInfo_CABAC32;
      else
        currSE->writing = writeMB_typeInfo_CABAC;

      dataPart->writeSyntaxElement( currSE, dataPart);
#if TRACE
      if (img->type == B_SLICE) 
        snprintf(currSE->tracestring, TRACESTRING_SIZE, "mb_type (B_SLICE) (%2d,%2d) = %3d",img->mb_x, img->mb_y, currMB->mb_type);
      else                      
        snprintf(currSE->tracestring, TRACESTRING_SIZE, "mb_type (P_SLICE) (%2d,%2d) = %3d",img->mb_x, img->mb_y, currMB->mb_type);
#endif
      bitCount[BITS_MB_MODE] += currSE->len;
      no_bits                += currSE->len;
      currSE++;
      currMB->currSEnr++;

      if(MB->mb_type == 3 || MB->geo_mode != -1)
      {
        currSE->value1  = MB->geo_mode;
        currSE->value2  = rho_bits[mb_ext_level]+THETA_BITS;
        currSE->writing = writeMB_geoInfo_CABAC;
        dataPart->writeSyntaxElement( currSE, dataPart);
#if TRACE
        snprintf(currSE->tracestring, TRACESTRING_SIZE, "mb_skip_flag");
#endif
        bitCount[BITS_MB_MODE] += currSE->len;
        no_bits                += currSE->len;
        currSE++;
        currMB->currSEnr++;
      }
    }

  return no_bits;
}

#ifdef MB64X64
/*!
************************************************************************
* \brief
*    Passes the chosen syntax elements to the NAL for 16x16 block inside of 64x64 block
* \param coeff_rate
*    bitrate of Luma and Chroma coeff
* \param (mb32i,mb16i)
*    mb32i: index of 32x32 block inside 64x64
*    mb16i: index of 16x16 block inside 32x32
************************************************************************
*/
int write_one_macroblock64 (int eos_bit, int mb32i, int mb16i)
{
  Macroblock* currMB   = &img->mb_data[img->current_mb_nr];
  int*        bitCount = currMB->bitcounter;
  int i;
  int no_bits;
  
  extern int cabac_encoding;

  //===== init and update number of intra macroblocks =====
  if (img->current_mb_nr==0)
    intras=0;
  
  if (IS_INTRA(currMB))
    intras++;
  
  //--- write non-slice termination symbol if the macroblock is not the first one in its slice ---
  if (input->symbol_mode==CABAC && img->current_mb_nr!=img->currentSlice->start_mb_nr && eos_bit && mb32i==0 && mb16i==0)
  {
    write_terminating_bit (0);
  }
  
  cabac_encoding = 1;
  
  //--- write macroblock ---
  no_bits = writeMBLayer_Big (0, &i, 2, mb32i, mb16i);  // i is temporary
  if (!((currMB->mb_type !=0 ) || ((img->type==B_SLICE) && currMB->cbp != 0) ))
  { 
    for (i=0; i < 4; i++)
      memset(img->nz_coeff [img->current_mb_nr][i], 0, (4 + img->num_blk8x8_uv) * sizeof(int));  // CAVLC
  }
  
  //set_last_dquant(); //this function can be removed! -peisong
  
  //--- set total bit-counter ---
#ifdef MV_COMPETITION
  if (input->mv_competition > 0)
  {
    bitCount[BITS_TOTAL_MB] = bitCount[BITS_MB_MODE]  + bitCount[BITS_COEFF_Y_MB]     
      + bitCount[BITS_INTER_MB] + bitCount[BITS_CBP_MB]  
      + bitCount[BITS_DELTA_QUANT_MB] + bitCount[BITS_COEFF_UV_MB]
      + bitCount[BITS_SKIP_PRED]
      + bitCount[BITS_MV_PRED];
  }
  else
#endif
    
  {
    bitCount[BITS_TOTAL_MB] = bitCount[BITS_MB_MODE]  + bitCount[BITS_COEFF_Y_MB]     
      + bitCount[BITS_INTER_MB] + bitCount[BITS_CBP_MB]  
      + bitCount[BITS_DELTA_QUANT_MB] + bitCount[BITS_COEFF_UV_MB];
  }  
  
#if DUMP_MBTYPE 
  statsRDPass[img->curr_rd_pass].bit_use_mb_type[img->type]      += bitCount[BITS_MB_MODE];
#else
  stats->bit_use_mb_type[img->type]      += bitCount[BITS_MB_MODE];
#endif
  stats->bit_use_coeffY[img->type]       += bitCount[BITS_COEFF_Y_MB] ;
  stats->tmp_bit_use_cbp[img->type]      += bitCount[BITS_CBP_MB];
  stats->bit_use_coeffC[img->type]       += bitCount[BITS_COEFF_UV_MB];
  stats->bit_use_delta_quant[img->type]  += bitCount[BITS_DELTA_QUANT_MB];
  stats->bit_use_mode[img->type][currMB->mb_type]+= bitCount[BITS_INTER_MB];

  stats->bit_slice += bitCount[BITS_TOTAL_MB];
  
  cabac_encoding = 0;
  return no_bits;
}

#endif  ///endif of MB64x64


/*!
************************************************************************
* \brief
*    Codes macroblock header
* \param rdopt
*    true for calls during RD-optimization
* \param coeff_rate
*    bitrate of Luma and Chroma coeff
* \param (bigger_block_type,mb16i)
*        (1,0) - 32x32, 32x16, 16x32 mode info, mv info, coeff of upper left 16x16, dequant
*        (1,1~)-                                         coeff of other      16x16
*        (0,0) - 16x16 mode, upper left 16x16 partition mode, mv, coeff, dequant
*        (0,1~)-                        16x16 partition mode, mv, coeff
************************************************************************
*/
int writeMBLayer_Big (int rdopt, int *coeff_rate, int bigger_block_type, int mb32i, int mb16i)
{
  int             i;
  int             mb_nr      = img->current_mb_nr;
//  int             prev_mb_nr = FmoGetPreviousMBNr(img->current_mb_nr); //ignore -peisong
  Macroblock*     currMB     = &img->mb_data[mb_nr];
//  Macroblock*     prevMB     = mb_nr ? (&img->mb_data[prev_mb_nr]) : NULL; //ignore -peisong
  SyntaxElement  *currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
  int*            bitCount   = currMB->bitcounter;
  Slice*          currSlice  = img->currentSlice;
  DataPartition*  dataPart;
  const int*      partMap    = assignSE2partition[input->partition_mode];
  int             no_bits    = 0;
//  int             skip       = currMB->mb_type ? 0:((img->type == B_SLICE) ? !currMB->cbp:1);//ignore -peisong
  //int             mb_type=-1;  
//  int             prevMbSkipped = 0; //ignore -peisong
//  Macroblock      *topMB = NULL;
  

  currMB->IntraChromaPredModeFlag = IS_INTRA(currMB);
  

  // choose the appropriate data partition
  dataPart = &(currSlice->partArr[partMap[SE_MBTYPE]]);
  
  if(img->type == I_SLICE)
  {
    // should never come to this place
    // No big block in I slice
    assert(0);
    //========= write mb_type (I_SLICE) =========
    currSE->value1  = MBType2Value (currMB);

      currSE->value2  = 0;
    currSE->type    = SE_MBTYPE;
    
    if (input->symbol_mode == UVLC)  
      currSE->mapping = ue_linfo;
    else
      currSE->writing = writeMB_typeInfo_CABAC;
    
    dataPart->writeSyntaxElement( currSE, dataPart);
#if TRACE
    snprintf(currSE->tracestring, TRACESTRING_SIZE,   "mb_type (I_SLICE) (%2d,%2d) = %3d",img->mb_x, img->mb_y, currMB->mb_type);
#endif
    bitCount[BITS_MB_MODE] += currSE->len;
    no_bits                += currSE->len;
    currSE++;
    currMB->currSEnr++;
    
  }
  // not I_SLICE, CABAC
  else
  {

#if defined MB32X32 && !defined TEST_ONLY_Z_SCAN 

#ifdef MB64X64
    if(bigger_block_type == 2)
    {
      if(mb32i==0 && mb16i==0)
      {
        if(input->symbol_mode == CABAC)
          no_bits += write_mb_type(&MB64, 2);
        else
          no_bits += write_mb_type_cavlc(&MB64, 2);//tm_vlc

        currSE     = &img->MB_SyntaxElements[currMB->currSEnr];

      }
    }
    else if(MB64.write_mb64_mbtype && mb16i==0)
    {
      Macroblock mb32x32;
      mb32x32.mb_type = P8x8;
      mb32x32.geo_mode = -1;
      if(input->symbol_mode == CABAC)
        no_bits += write_mb_type(&mb32x32, 2);
      else
        no_bits += write_mb_type_cavlc(&mb32x32, 2);//tm_vlc

      currSE     = &img->MB_SyntaxElements[currMB->currSEnr];

    }
#endif
    /********************************************
     *MB type for mode 32x32, 32x16, 16x32, 16x16
     *******************************************/
    if(bigger_block_type <= 1)
    {
      if(mb16i==0)
      {
        if(input->symbol_mode == CABAC)
          no_bits += write_mb_type(&MB32, 1);
        else
          no_bits += write_mb_type_cavlc(&MB32, 1);//tm_vlc

        currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
      }
    }
#endif

    if(bigger_block_type ==0)
    {
      if(input->symbol_mode == CABAC)
        no_bits += write_mb_type(currMB, 0);
      else
        no_bits += write_mb_type_cavlc(currMB, 0);//tm_vlc
      currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
    }

    if (input->symbol_mode == CABAC)
     CheckAvailabilityOfNeighborsCABAC();
  }
  
#ifdef MV_COMPETITION
  if (input->mv_competition > 0)
  {
    if((rdopt==0) && (input->symbol_mode == CABAC) && (img->type==P_SLICE) && (mv_comp.nb_mode_for_skip > 1))  
    {  
      //make sure that this function is called only once...whether it is big blocks or 16x16
      if(bigger_block_type==2)
      {
        if(mb32i==0 && mb16i==0 && MB64.mb_type==0)
          no_bits  += write_predictor_index_for_skip_mode64();
      }
      if(bigger_block_type==1)
      { 
        //32x32 P-Skip
        if(mb16i==0 && MB32.mb_type==0)
          no_bits  += write_predictor_index_for_skip_mode32();
      }
      else if(bigger_block_type==0)
      {
        //16x16 P-Skip
        if(currMB->mb_type==0)
          no_bits  += write_predictor_index_for_skip_mode  ();
      }
    }
  }
#endif
 
  if (bigger_block_type == 0)
  {
    //init NoMbPartLessThan8x8Flag
    currMB->NoMbPartLessThan8x8Flag = (IS_DIRECT(currMB) && !(active_sps->direct_8x8_inference_flag))? 0: 1;
    //===== BITS FOR 8x8 SUB-PARTITION MODES =====
    if(IS_P8x8 (currMB))     //Rahul---this will happen for 16x16 having 8x8 partition
    {
      dataPart = &(currSlice->partArr[partMap[SE_MBTYPE]]);

      for (i=0; i<4; i++)
      {
        if (input->symbol_mode==UVLC)   
          currSE->mapping = ue_linfo;
        else
        {
#if B_MBTYPE//_BIG
          currSE->dir_partition1 = currMB->b8pdir[i];    //what value to set here-> see the input parameter in B8Mode2Value()
#endif
          currSE->writing = writeB8_typeInfo_CABAC;
        }

        currSE->value1  = B8Mode2Value (currMB->b8mode[i], currMB->b8pdir[i]);
        currSE->value2  = 0;
        currSE->type    = SE_MBTYPE;
        dataPart->writeSyntaxElement (currSE, dataPart);
        bitCount[BITS_MB_MODE]+= currSE->len;
        no_bits               += currSE->len;
        currSE++;
        currMB->currSEnr++;

        //set NoMbPartLessThan8x8Flag for P8x8 mode
        currMB->NoMbPartLessThan8x8Flag &= (currMB->b8mode[i]==0 && active_sps->direct_8x8_inference_flag) || 
          (currMB->b8mode[i]==4);
      }

#ifdef MV_COMPETITION
      if (input->mv_competition > 0)
        pass_with_writing = !rdopt;
#endif
      no_bits += writeMotionInfo2NAL  ();
      currSE   = &img->MB_SyntaxElements[currMB->currSEnr];
    }
  }
  //============= Transform size flag for INTRA MBs =============
  //-------------------------------------------------------------
  //transform size flag for INTRA_4x4 and INTRA_8x8 modes
  if ((currMB->mb_type == I8MB || currMB->mb_type == I4MB) && input->Transform8x8Mode)
  {
    currSE->value1 = currMB->luma_transform_size_8x8_flag;
    currSE->type   = SE_HEADER;
    
    if( input->symbol_mode==UVLC)
    {
      currSE->mapping = ue_linfo;
      currSE->type    = SE_MBTYPE;
      currSE->bitpattern = currMB->luma_transform_size_8x8_flag;
      currSE->len = 1;
      writeSyntaxElement2Buf_Fixed(currSE, dataPart->bitstream);
    }
    else
    {
      currSE->writing = writeMB_transform_size_CABAC;
      dataPart->writeSyntaxElement(currSE, dataPart);
    }
#if TRACE
    snprintf(currSE->tracestring, TRACESTRING_SIZE, "transform_size_8x8_flag = %3d", currMB->luma_transform_size_8x8_flag);
#endif
    
    bitCount[BITS_MB_MODE] += currSE->len;
    no_bits                += currSE->len;
    currSE++;
    currMB->currSEnr++;
  }
  
  
  //===== BITS FOR INTRA PREDICTION MODES ====
  no_bits += writeIntraModes();
  //===== BITS FOR CHROMA INTRA PREDICTION MODE ====
  if (currMB->IntraChromaPredModeFlag && img->yuv_format != YUV400)
    no_bits += writeChromaIntraPredMode();
  else if(!rdopt) //GB CHROMA !!!!!
    currMB->c_ipred_mode = DC_PRED_8; //setting c_ipred_mode to default is not the right place here
  //resetting in rdopt.c (but where ??)
  //with cabac and bframes maybe it could crash without this default
  //since cabac needs the right neighborhood for the later MBs
  
  //----- motion information -----
  if(bigger_block_type==2)
  {
    if(mb32i==0 && mb16i==0)
    {
      if(MB64.mb_type != 0)
      {
#ifdef MV_COMPETITION
        if (input->mv_competition > 0)
          pass_with_writing = !rdopt;
#endif

        if(MB64.geo_mode != -1)
          no_bits  += writeMotionInfo2NAL_geo  (MB64.geo_mode, 2);
        else
          no_bits  += writeMotionInfo2NAL32  (2);
      }
    }
  }
  if(bigger_block_type==1)
  {
    if(mb16i==0)
    {
      if(MB32.mb_type != 0)
      {
#ifdef MV_COMPETITION
        if (input->mv_competition > 0)
          pass_with_writing = !rdopt;
#endif
        if(MB32.geo_mode != -1)
          no_bits  += writeMotionInfo2NAL_geo  (MB32.geo_mode, 1);
        else
          no_bits  += writeMotionInfo2NAL32  (1);
      }
    }
  }
  else if(bigger_block_type==0 && currMB->mb_type !=0 && currMB->mb_type !=P8x8)
  {
#ifdef MV_COMPETITION
    if (input->mv_competition > 0)
      pass_with_writing = !rdopt;
#endif

    if(currMB->geo_mode != -1)
      no_bits  += writeMotionInfo2NAL_geo  (currMB->geo_mode, 0);
    else
    no_bits  += writeMotionInfo2NAL  ();
  }

  if(bigger_block_type == 2)
  {
    // this function only takes care of 64x64 skip, 64x64, 64x32, 32x64....Rahul
    // so there will be only Delta_QP at 64x64 level

#if P_DIRECT
    if(MB64.mb_type!=0 || MB64.cbp!=0)
#else
    if(MB64.mb_type!=0 || (img->type==B_SLICE && MB64.cbp!=0))
#endif
    {
      if(mb32i==0 && mb16i==0)    //64x64 cbp
      {
        //      no_bits += writeCBP32(rdopt, MB64.cbp, 1);//with dquant
        no_bits += writeCBP_Big (rdopt, MB64.cbp);
        no_bits += writeDeltaQP (rdopt, MB64.cbp);
      }

      if(MB64.cbp!=0)           
      {
        if(mb16i == 0)            //32x32 cbp 
        {
          //        no_bits += writeCBP32(rdopt, MB32.cbp, 0);// dquant is already written in 64x64, so no need to write at 32x32 level.
          no_bits += writeCBP_Big  (rdopt, MB32.cbp);
        }

        if(MB32.cbp != 0)         //16x16 cbp, delta_qp, luma coeff, chroma coeff
        {
          //*coeff_rate = writeCBPandLumaCoeff (rdopt, 1, mb16i);
          no_bits +=    writeCBP_Transform_size (rdopt, 1);
          *coeff_rate = writeLumaCoeff          (rdopt, 1);

          if (img->yuv_format != YUV400)
            *coeff_rate  += writeChromaCoeff ();

          no_bits  += *coeff_rate;           
        }
        else
        {
          if(input->symbol_mode==UVLC)//tm_vlc
            Setnz_coeff(img->current_mb_nr, 1);
        }
      }
      else
      {         
        if(input->symbol_mode==UVLC)//tm_vlc
          Setnz_coeff(img->current_mb_nr, 2);
      }

    }

  }
  else
  {
#if defined MB32X32 && !defined TEST_ONLY_Z_SCAN    
#if P_DIRECT
    if(  (bigger_block_type==1 && (MB32.mb_type!=0 || MB32.cbp!=0))
      || (bigger_block_type==0 && /*MB32.cbp!=0 && */(currMB->mb_type!=0 || currMB->cbp!=0))  )
#else
    if(  (bigger_block_type==1 && (MB32.mb_type!=0 || (img->type==B_SLICE && MB32.cbp!=0)))
      || (bigger_block_type==0 && /*MB32.cbp!=0 && */(currMB->mb_type!=0 || (img->type==B_SLICE && currMB->cbp!=0)))  )
#endif
#else
    if ((currMB->mb_type!=0) || (img->type==B_SLICE && (currMB->cbp!=0)))
#endif
    {
      if(bigger_block_type==1 && mb16i==0)
      {
        //      no_bits += writeCBP32(rdopt, MB32.cbp, !delta_qp_sent);//with dquant
        no_bits += writeCBP_Big    (rdopt, MB32.cbp);
        no_bits += writeDeltaQP    (rdopt, MB32.cbp);
      }
      if( bigger_block_type==0 || (bigger_block_type==1 && MB32.cbp!=0) )
      {
        //*coeff_rate = writeCBPandLumaCoeff (rdopt, bigger_block_type, mb16i);
        no_bits +=    writeCBP_Transform_size (rdopt, bigger_block_type);
        if(bigger_block_type==0)
          no_bits +=    writeDeltaQP            (rdopt, img->mb_data[img->current_mb_nr].cbp);
        *coeff_rate = writeLumaCoeff          (rdopt, bigger_block_type);
        if (img->yuv_format != YUV400)
          *coeff_rate  += writeChromaCoeff ();

        no_bits  += *coeff_rate;  
      }

      if(input->symbol_mode==UVLC && bigger_block_type==1 && MB32.cbp==0)//tm_vlc
        Setnz_coeff(img->current_mb_nr, 1);
    }
  }

  return no_bits;
}

/*!
************************************************************************
* \brief
*    Passes the chosen syntax elements to the NAL for 16x16 block inside of 32x32 block
* \param coeff_rate
*    bitrate of Luma and Chroma coeff
* \param (mb32,mbi)
*        (1,0) - 32x32, 32x16, 16x32 mode info, mv info, coeff of upper left 16x16, dequant
*        (1,1~)-                                         coeff of other      16x16
*        (0,0) - 16x16 mode, upper left 16x16 partition mode, mv, coeff, dequant
*        (0,1~)-                        16x16 partition mode, mv, coeff
************************************************************************
*/
int write_one_macroblock32 (int eos_bit, int mb32, int mbi)
{
  Macroblock* currMB   = &img->mb_data[img->current_mb_nr];
  int*        bitCount = currMB->bitcounter;
  int i;
  int no_bits;
  
  extern int cabac_encoding;
  //===== init and update number of intra macroblocks =====
  if (img->current_mb_nr==0)
    intras=0;
  
  if (IS_INTRA(currMB))
    intras++;
  
  //--- write non-slice termination symbol if the macroblock is not the first one in its slice ---
  if (input->symbol_mode==CABAC && img->current_mb_nr!=img->currentSlice->start_mb_nr && eos_bit && mbi==0 && eos_bit_32)
  {
    write_terminating_bit (0);
  }
  
  cabac_encoding = 1;
  
  //--- write macroblock ---
  no_bits = writeMBLayer_Big (0, &i, mb32, 0, mbi);  // i is temporary
  if (!((currMB->mb_type !=0 ) || ((img->type==B_SLICE) && currMB->cbp != 0) ))
  { 
    for (i=0; i < 4; i++)
      memset(img->nz_coeff [img->current_mb_nr][i], 0, (4 + img->num_blk8x8_uv) * sizeof(int));  // CAVLC
  }
  
  //set_last_dquant(); //this function can be removed! -peisong
  
  //--- set total bit-counter ---
#ifdef MV_COMPETITION
  if (input->mv_competition > 0)
  {
    bitCount[BITS_TOTAL_MB] = bitCount[BITS_MB_MODE]  + bitCount[BITS_COEFF_Y_MB]     
      + bitCount[BITS_INTER_MB] + bitCount[BITS_CBP_MB]  
      + bitCount[BITS_DELTA_QUANT_MB] + bitCount[BITS_COEFF_UV_MB]
      + bitCount[BITS_SKIP_PRED]
      + bitCount[BITS_MV_PRED];
  }
  else
#endif
  {
    bitCount[BITS_TOTAL_MB] = bitCount[BITS_MB_MODE]  + bitCount[BITS_COEFF_Y_MB]     
      + bitCount[BITS_INTER_MB] + bitCount[BITS_CBP_MB]  
      + bitCount[BITS_DELTA_QUANT_MB] + bitCount[BITS_COEFF_UV_MB];
  }  
  stats->bit_slice += bitCount[BITS_TOTAL_MB];

  if(mb32)
  {
#if DUMP_MBTYPE 
    statsRDPass[img->curr_rd_pass].bit_use_mb_type[img->type]      += bitCount[BITS_MB_MODE];
#else
    stats->bit_use_mb_type[img->type]      += bitCount[BITS_MB_MODE];
#endif
    stats->bit_use_coeffY[img->type]       += bitCount[BITS_COEFF_Y_MB] ;
    stats->tmp_bit_use_cbp[img->type]      += bitCount[BITS_CBP_MB];
    stats->bit_use_coeffC[img->type]       += bitCount[BITS_COEFF_UV_MB];
    stats->bit_use_delta_quant[img->type]  += bitCount[BITS_DELTA_QUANT_MB];
    stats->bit_use_mode[img->type][currMB->mb_type]+= bitCount[BITS_INTER_MB];
  }


  cabac_encoding = 0;
  return no_bits;
}

int writeLumaCoeff16x16P_CABAC ()
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[input->partition_mode];
  int*            bitCount  = currMB->bitcounter;
  DataPartition*  dataPart;
  
  int   level, run;
  int   k;
  int*  ACLevel = img->cofAC16x16[0];
  int*  ACRun   = img->cofAC16x16[1];
  
  img->subblock_x = 0;//img->pix_x;  // horiz. position for coeff_count context
  img->subblock_y = 0;//img->pix_y;  // vert.  position for coeff_count context

  level=1; // get inside loop
  for(k=0; k<=256 && level !=0; k++)
  {
    level = currSE->value1 = ACLevel[k]; // level
    run   = currSE->value2 = ACRun  [k]; // run
    
    currSE->writing = writeRunLevel16x16P_CABAC;
    
    currSE->context     = LUMA_16x16P; 
    currSE->type        = SE_LUM_AC_INTER;
    img->is_intra_block = 0;    // only applicable to inter blocks
    
    // choose the appropriate data partition
    dataPart = &(currSlice->partArr[partMap[img->type != B_SLICE ? currSE->type : SE_BFRAME]]);
    dataPart->writeSyntaxElement (currSE, dataPart);
    bitCount[BITS_COEFF_Y_MB] += currSE->len;
    rate                      += currSE->len;
    /* proceed to next SE */
    currSE++;  
    currMB->currSEnr++;
    
  }
  
  return rate;
}

int writeLumaCoeff16x16P ()
{
  int  rate = 0;
  
  if( input->symbol_mode == UVLC) // allow here if 4x4 or UVLC
  {
    assert(0);
  }
  else 
    rate += writeLumaCoeff16x16P_CABAC ();
  
  return rate;
}

int writeLumaCoeff16x8_CABAC (int blk)
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[input->partition_mode];
  int*            bitCount  = currMB->bitcounter;
  DataPartition*  dataPart;
  
  int   level, run;
  int   k;
  int*  ACLevel = img->cofAC16x8[blk][0];
  int*  ACRun   = img->cofAC16x8[blk][1];
  
  img->subblock_x = 0;               // horiz. position for coeff_count context
  img->subblock_y = (blk == 0)?0:2;  // vert.  position for coeff_count context
  
  level=1; // get inside loop
  for(k=0; k<=128 && level !=0; k++)
  {
    level = currSE->value1 = ACLevel[k]; // level
    run   = currSE->value2 = ACRun  [k]; // run
    
    currSE->writing = writeRunLevel16x8_CABAC;
    
    currSE->context     = LUMA_16x8P;

    currSE->type        = SE_LUM_AC_INTER;
    img->is_intra_block = 0;    // only applicable to inter blocks
    
    // choose the appropriate data partition
    dataPart = &(currSlice->partArr[partMap[img->type != B_SLICE ? currSE->type : SE_BFRAME]]);
    dataPart->writeSyntaxElement (currSE, dataPart);
    bitCount[BITS_COEFF_Y_MB] += currSE->len;
    rate                      += currSE->len;
#if TRACE
    snprintf(currSE->tracestring, TRACESTRING_SIZE, "Luma8x8 sng(%2d) level =%3d run =%2d", k, level,run);
#endif
    /* proceed to next SE */
    currSE++;  
    currMB->currSEnr++;
    
  }
  
  return rate;
}

int writeLumaCoeff16x8 ( int blk16x8 )
{
  int  rate = 0;
  
  if( input->symbol_mode == UVLC) // allow here if 4x4 or UVLC
  {
    assert(0);
  }
  else 
    rate += writeLumaCoeff16x8_CABAC (blk16x8);
  
  return rate;
}

int writeLumaCoeff8x16_CABAC (int blk)
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[input->partition_mode];
  int*            bitCount  = currMB->bitcounter;
  DataPartition*  dataPart;
  
  int   level, run;
  int   k;
  int*  ACLevel = img->cofAC8x16[blk][0];
  int*  ACRun   = img->cofAC8x16[blk][1];
  
  img->subblock_x = (blk == 0)?0:2;     // horiz. position for coeff_count context
  img->subblock_y = 0;                  // vert.  position for coeff_count context
  
  level=1; // get inside loop
  for(k=0; k<=128 && level !=0; k++)
  {
    level = currSE->value1 = ACLevel[k]; // level
    run   = currSE->value2 = ACRun  [k]; // run
    
    currSE->writing = writeRunLevel8x16_CABAC;
    
    currSE->context     = LUMA_8x16P;

    currSE->type        = SE_LUM_AC_INTER;
    img->is_intra_block = 0;    // only applicable to inter blocks
    
    // choose the appropriate data partition
    dataPart = &(currSlice->partArr[partMap[img->type != B_SLICE ? currSE->type : SE_BFRAME]]);
    dataPart->writeSyntaxElement (currSE, dataPart);
    bitCount[BITS_COEFF_Y_MB] += currSE->len;
    rate                      += currSE->len;
#if TRACE
    snprintf(currSE->tracestring, TRACESTRING_SIZE, "Luma8x8 sng(%2d) level =%3d run =%2d", k, level,run);
#endif
    /* proceed to next SE */
    currSE++;  
    currMB->currSEnr++;
    
  }
  
  return rate;
}

int writeLumaCoeff8x16 ( int blk8x16 )
{
  int  rate = 0;
  
  if( input->symbol_mode == UVLC) 
  {
    assert(0);
  }
  else 
    rate += writeLumaCoeff8x16_CABAC (blk8x16);
  
  return rate;
}

#ifdef MB32X32
int writeLumaCoefBigBlock(Macroblock *currMB, int mb32, int rdopt)
#else
int writeLumaCoefBigBlock(Macroblock *currMB, int rdopt)
#endif
{
  int rate = 0;

  assert(currMB->mb_type >= 0 && currMB->mb_type <= 3);
  assert(currMB->luma_transform_size_8x8_flag);
#ifdef MB32X32
  if((currMB->mb_type < 2 || mb32 || currMB->geo_mode!=-1)&& currMB->luma_transform_size_8x8_flag == 2)  // 16x16 partition
#else
  if(currMB->mb_type < 2 && currMB->luma_transform_size_8x8_flag == 2)  // 16x16 partition
#endif
  {
    if(currMB->cbp & 0xf)
      rate += writeLumaCoeff16x16P(); 
  }
  else if(currMB->mb_type == 2 && currMB->luma_transform_size_8x8_flag == 2)  // 16x8 partition
  {
    if(currMB->cbp & 0x3)
      rate += writeLumaCoeff16x8(0);
    if(currMB->cbp & 0xc)
      rate += writeLumaCoeff16x8(1);
  }
  else if(currMB->mb_type == 3 && currMB->luma_transform_size_8x8_flag == 2)  // 8x16 partition
  {
    if(currMB->cbp & 0x5)
      rate += writeLumaCoeff8x16(0);
    if(currMB->cbp & 0xa)
      rate += writeLumaCoeff8x16(1);
  }

  return rate;
}


// write 1 bit indicating luma cbp
int writelumaCBPTransform_size(int BigBlock)
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  int*            bitCount  = currMB->bitcounter;
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[input->partition_mode];
  int             cbp       = currMB->cbp;
  DataPartition*  dataPart;
  int             need_transform_size_flag;   //ADD-VG-24062004
  

    //=====   C B P   =====
    //---------------------
    currSE->value1 = cbp;

    if (IS_OLDINTRA (currMB) || currMB->mb_type == SI4MB ||  currMB->mb_type == I8MB)
    {
      if (input->symbol_mode == UVLC)  currSE->mapping = cbp_linfo_intra;
      currSE->type = SE_CBP_INTRA;
    }
    else
    {
      if (input->symbol_mode == UVLC)  currSE->mapping = cbp_linfo_inter;
      currSE->type = SE_CBP_INTER;
    }

    // choose the appropriate data partition
    dataPart = &(currSlice->partArr[partMap[currSE->type]]);

    if (input->symbol_mode == CABAC)   
    {
      currSE->writing = writeCBP_CABAC_luma1bit;
      dataPart->writeSyntaxElement(currSE, dataPart);
	  bitCount[BITS_CBP_MB] += currSE->len;
      rate                  += currSE->len;
      // proceed to next SE
	  currSE++;  
	  currMB->currSEnr++;
    }
	else
    {
      currSE->len = u_1 ("luma1bit", (cbp&15)!=0?1:0, dataPart->bitstream);
      bitCount[BITS_CBP_MB] += currSE->len;
      rate                  += currSE->len;
      // proceed to next SE
	  currSE++;  
	  currMB->currSEnr++;
    }

    //============= Transform Size Flag for INTER MBs =============
    //-------------------------------------------------------------

 
#if P_DIRECT
    need_transform_size_flag = (((currMB->mb_type >= 0 && currMB->mb_type <= 3)||
      (IS_DIRECT(currMB) && active_sps->direct_8x8_inference_flag) ||
      (currMB->NoMbPartLessThan8x8Flag))
      && currMB->mb_type != I8MB && currMB->mb_type != I4MB
      && (currMB->cbp&15)
      && input->Transform8x8Mode);
#else
    need_transform_size_flag = (((currMB->mb_type >= 1 && currMB->mb_type <= 3)||
      (IS_DIRECT(currMB) && active_sps->direct_8x8_inference_flag) ||
      (currMB->NoMbPartLessThan8x8Flag))
      && currMB->mb_type != I8MB && currMB->mb_type != I4MB
      && (currMB->cbp&15)
      && input->Transform8x8Mode);
#endif


    if (need_transform_size_flag)
    {
      if (input->DisableSmallTransInLargeMB && BigBlock)
        return rate;

      currSE->value1 = currMB->luma_transform_size_8x8_flag;
      currSE->type   = SE_HEADER;
      
      if( input->symbol_mode==UVLC)
      {
        currSE->len = u_1 ("currMB->luma_transform_size_8x8_flag>0", (currMB->luma_transform_size_8x8_flag>0)?1:0, dataPart->bitstream);
        if(currMB->luma_transform_size_8x8_flag>0 && currMB->mb_type < 4)
          currSE->len += u_1 ("currMB->luma_transform_size_8x8_flag=2", (currMB->luma_transform_size_8x8_flag>1)?1:0, dataPart->bitstream);        
      }
      else
      {
        currSE->writing = writeMB_transform_size_CABAC;
        dataPart->writeSyntaxElement(currSE, dataPart);
      }
      bitCount[BITS_MB_MODE] += currSE->len;
      rate                   += currSE->len;
      currSE++;
      currMB->currSEnr++;
    }

    return rate;
}
void writeCBP_CABAC_chroma(SyntaxElement *se, EncodingEnvironmentPtr eep_dp)
{
  TextureInfoContexts *ctx = img->currentSlice->tex_ctx;
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];
  
  int a, b;
  int curr_cbp_ctx;
  int cbp = se->value1; // symbol to encode
  int cbp_bit;
  
  if (img->yuv_format != YUV400)
  {
    // coding of chroma part
    b = 0;
    if (currMB->mb_available_up != NULL)
      b = ((currMB->mb_available_up)->cbp > 15) ? 1 : 0;

    a = 0;
    if (currMB->mb_available_left != NULL)
      a = ((currMB->mb_available_left)->cbp > 15) ? 1 : 0;
    
    curr_cbp_ctx = a+2*b;
    cbp_bit = (cbp > 15 ) ? 1 : 0;
    biari_encode_symbol(eep_dp, (signed short) cbp_bit, ctx->cbp_contexts[1] + curr_cbp_ctx );
    
    if (cbp > 15)
    {
      b = 0;
      if (currMB->mb_available_up != NULL)
      {
        if ((currMB->mb_available_up)->cbp > 15)
          b = (( ((currMB->mb_available_up)->cbp >> 4) == 2) ? 1 : 0);
      }

      a = 0;
      if (currMB->mb_available_left != NULL)
      {
        if ((currMB->mb_available_left)->cbp > 15)
          a = (( ((currMB->mb_available_left)->cbp >> 4) == 2) ? 1 : 0);
      }
      
      curr_cbp_ctx = a+2*b;
      cbp_bit = ((cbp>>4) == 2) ? 1 : 0;
      biari_encode_symbol(eep_dp, (signed short) cbp_bit, ctx->cbp_contexts[2] + curr_cbp_ctx );
    }
  }
}


int write_chroma_cbp()
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  int*            bitCount  = currMB->bitcounter;
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[input->partition_mode];
  int             cbp       = currMB->cbp;
  DataPartition*  dataPart;

  

    //=====   C B P   =====
    //---------------------
    currSE->value1 = cbp;
 
    if (IS_OLDINTRA (currMB) || currMB->mb_type == SI4MB ||  currMB->mb_type == I8MB)
    {
      if (input->symbol_mode == UVLC)  currSE->mapping = cbp_linfo_intra;
      currSE->type = SE_CBP_INTRA;
    }
    else
    {
      if (input->symbol_mode == UVLC)  currSE->mapping = cbp_linfo_inter;
      currSE->type = SE_CBP_INTER;
    }

    // choose the appropriate data partition
    dataPart = &(currSlice->partArr[partMap[currSE->type]]);

    if (input->symbol_mode == CABAC)
    {
      currSE->writing = writeCBP_CABAC_chroma;        
      dataPart->writeSyntaxElement(currSE, dataPart);
    }
    else
    {
      currSE->len = u_1 ("chroma cbp > 0", (cbp>15)?1:0, dataPart->bitstream);
      if(cbp>15)
      {
        currSE->len += u_1 ("chroma cbp = 2", ((cbp>>4)==2)?1:0, dataPart->bitstream);
      }
    }

    bitCount[BITS_CBP_MB] += currSE->len;
    rate                  += currSE->len;
    // proceed to next SE
    currSE++;  
    currMB->currSEnr++;

    return rate;
}

#endif
