
/*!
***************************************************************************
* \file rdopt.c
*
* \brief
*    Rate-Distortion optimized mode decision
*
* \author
*    - Heiko Schwarz              <hschwarz@hhi.de>
*    - Valeri George              <george@hhi.de>
*    - Lowell Winger              <lwinger@lsil.com>
*    - Alexis Michael Tourapis    <alexismt@ieee.org>
* \date
*    12. April 2001
**************************************************************************
*/

#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <limits.h>
#include <memory.h>
#include <string.h>

#include "global.h"

#include "rdopt_coding_state.h"
#include "memalloc.h"
#include "mb_access.h"
#include "elements.h"
#include "intrarefresh.h"
#include "image.h"
#include "transform8x8.h"
#include "cabac.h"   
#include "vlc.h"
#include "fast_me.h"
//#include "ratectl.h"            // head file for rate control
#include "mode_decision.h"
#include "fmo.h"


#define KS_MV
void LumaResidualCoding_geo (int mbi_x, int mbi_y, int geo_mode, int mode_idx, int mb_ext_level);
void ChromaResidualCoding_geo (int mbi_x, int mbi_y, int geo_mode, int mode_idx, int mb_ext_level);

#ifdef MV_COMPETITION
#include "mv_competition.h"
extern MV_Competition mv_comp;
extern int skip_mode;
#endif

#ifdef RDO_Q
int saved_cbp, saved_I16cbp, saved_I8cbp;
int saved_i16mode;
int saved_cofAC[4][4][2][65];
int saved_I16cofAC[4][4][2][65], saved_I8cofAC[4][4][2][65];
int saved_cofDC[3][2][65];
imgpel saved_enc_picture_imgY[16][16], saved_I16enc_picture_imgY[16][16], saved_I8enc_picture_imgY[16][16];
imgpel saved_img_mpr[16][16], saved_I8img_mpr[16][16];
#endif
#ifdef USE_INTRA_MDDT
int   bestCofAC16[2][16*17];
#endif 

int   bestCofAC16x16 [2][257];
int   bestCofAC16x8  [2][2][129];
int   bestCofAC8x16  [2][2][129];

imgpel pred[16][16];

#define FASTMODE 1
//#define RESET_STATE

extern const int LEVELMVLIMIT[17][6];
extern int   QP2QUANT[40];

extern short OffsetList4x4[15][16];
extern short OffsetList8x8[5][64];
extern const int OffsetBits;

extern int writeDeltaQP(int rdopt, int cbp);
extern int writeCBP_Transform_size (int rdopt, int mb32);
extern int writeLumaCoeff (int rdopt, int mb32);

#ifdef MB32X32
imgpel   rec_mbY[16<<MAX_MB_EXT_LEVEL][16<<MAX_MB_EXT_LEVEL], rec_mbU[16<<MAX_MB_EXT_LEVEL][16<<MAX_MB_EXT_LEVEL], rec_mbV[16<<MAX_MB_EXT_LEVEL][16<<MAX_MB_EXT_LEVEL];    // reconstruction values
#else
imgpel   rec_mbY[16][16], rec_mbU[16][16], rec_mbV[16][16];    // reconstruction values
#endif

//RD_8x8DATA tr4x4, tr8x8;
RD_8x8DATA tr4x4[NumberOfQP], tr8x8[NumberOfQP];      // for 5 QP..

int   ****cofAC=NULL;   //this is used in store_macroblock_parameters() 
int   ***cofDC=NULL;                       // [yuv][level/run][scan_pos]

int   ****cofAC8x8=NULL;        // [8x8block][4x4block][level/run][scan_pos]
int   **cofAC4x4=NULL, ****cofAC4x4intern=NULL; // [level/run][scan_pos]
int   cbp, cbp8x8, cnt_nonz_8x8;
int64 cbp_blk;
int   cbp_blk8x8;
char  frefframe[4][4], brefframe[4][4];
int   b8mode[4], b8pdir[4];
//short best8x8mode [4];                // [block]
short best8x8pdir  [MAXMODE][4];       // [mode][block]
short best8x8fwref [MAXMODE][4];       // [mode][block]
short best8x8bwref [MAXMODE][4];       // [mode][block]


CSptr cs_mb=NULL, cs_b8=NULL, cs_cm=NULL, cs_imb=NULL, cs_ib8=NULL, cs_ib4=NULL, cs_pc=NULL;
#ifdef USE_INTRA_MDDT
CSptr cs_i16 = NULL;
#endif 
CSptr cs_smb=NULL, cs_eachQP = NULL;
int   best_c_imode;
int   best_i16offset;
short best_mode;
short  bi_pred_me;

//mixed transform sizes definitions
int   luma_transform_size_8x8_flag;

short all_mv8x8[2][2][4][4][2];       //[8x8_data/temp_data][LIST][block_x][block_y][MVx/MVy]
short pred_mv8x8[2][2][4][4][2];

//int   ****cofAC_8x8ts = NULL;        // [8x8block][4x4block][level/run][scan_pos]
//int64    cbp_blk8_8x8ts;
//int      cbp8_8x8ts;
//int      cost8_8x8ts;
//int      cnt_nonz8_8x8ts;
#ifdef RDO_Q
imgpel rec_mbY_intra[4][4][16][16], rec_mbU_intra[4][4][16][16], rec_mbV_intra[4][4][16][16];    // reconstruction values
int   i16offset_intra[4][4];
int   cbp_intra[4][4];
int64 cbp_blk_intra[4][4];
//double rdcost_intra[4][4];
int   cofAC_intra[4][4][12][4][2][65];
int   cofDC_intra[4][4][3][2][65];                       // [yuv][level/run][scan_pos]
char   b4_ipredmode_intra[4][4][16];
char   b8_ipredmode8x8_intra[4][4][4][4];
char   b4_ipredmode_mb_intra[4][4][16];
char   b8_ipredmode8x8_mb_intra[4][4][16];
#define IS_INTRA_MODE(mode)  ((mode==I4MB)||(mode==I8MB)||(mode==I16MB))
#define IS_INTER_SLICE(type) ((type) == B_SLICE || (type) == P_SLICE)
#endif

void StoreMV8x8(int dir, short bestQP);
void RestoreMV8x8(int dir, short bestQP);
// end of mixed transform sizes definitions

#ifdef MB32X32
extern int   ****cofAC32[1<<(2*MAX_MB_EXT_LEVEL)], ****cofAC32_RDCost[1<<(2*MAX_MB_EXT_LEVEL)];// ****cofAC8x8=NULL;        // [8x8block][4x4block][level/run][scan_pos]
extern int   ***cofDC32[1<<(2*MAX_MB_EXT_LEVEL)], ***cofDC32_RDCost[1<<(2*MAX_MB_EXT_LEVEL)];                       // [yuv][level/run][scan_pos]
extern int    **cof32AC16x16[1<<(2*MAX_MB_EXT_LEVEL)], **cof32AC16x16_RDCost[1<<(2*MAX_MB_EXT_LEVEL)];//[2][257];
extern int    ***cof32AC16x8[1<<(2*MAX_MB_EXT_LEVEL)], ***cof32AC16x8_RDCost[1<<(2*MAX_MB_EXT_LEVEL)];//[2][2][129];
extern int    ***cof32AC8x16[1<<(2*MAX_MB_EXT_LEVEL)], ***cof32AC8x16_RDCost[1<<(2*MAX_MB_EXT_LEVEL)];//[2][2][129];

#endif

// Residue Color Transform
char  b4_ipredmode[16], b4_intra_pred_modes[16];

/*!
************************************************************************
* \brief
*    delete structure for RD-optimized mode decision
************************************************************************
*/
void clear_rdopt ()
{
  int i;
#ifdef MB32X32
  for(i=0; i<(1<<(2*MAX_MB_EXT_LEVEL)); i++)
  {
    free_mem_DCcoeff (cofDC32[i]);
    free_mem_ACcoeff (cofAC32[i]);
    free_mem_DCcoeff (cofDC32_RDCost[i]);
    free_mem_ACcoeff (cofAC32_RDCost[i]);
    free_mem2Dint(cof32AC16x16[i]);
    free_mem3Dint(cof32AC16x8[i], 2);
    free_mem3Dint(cof32AC8x16[i], 2);
    free_mem2Dint(cof32AC16x16_RDCost[i]);
    free_mem3Dint(cof32AC16x8_RDCost[i], 2);
    free_mem3Dint(cof32AC8x16_RDCost[i], 2);
  }
#endif
  free_mem_DCcoeff (cofDC);
  free_mem_ACcoeff (cofAC);
  for(i=0;i<NumberOfQP;i++)
  {
    free_mem_ACcoeff (tr4x4[i].cofAC8x8);
  }
  free_mem_ACcoeff (cofAC8x8);

  free_mem_ACcoeff (cofAC4x4intern);
  
  if (input->Transform8x8Mode)
  {
  for(i=0;i<NumberOfQP;i++)
  {
    free_mem_ACcoeff (tr8x8[i].cofAC8x8);
  }
  }
  
  // structure for saving the coding state
  delete_coding_state (cs_mb);
  delete_coding_state (cs_b8);
  delete_coding_state (cs_cm);
  delete_coding_state (cs_imb);
  delete_coding_state (cs_ib8);
  delete_coding_state (cs_ib4);
  delete_coding_state (cs_pc);
#ifdef USE_INTRA_MDDT
  delete_coding_state (cs_i16);
#endif 
  delete_coding_state (cs_smb);
  delete_coding_state (cs_eachQP);
}


/*!
************************************************************************
* \brief
*    create structure for RD-optimized mode decision
************************************************************************
*/
void init_rdopt ()
{
  int i;
#ifdef MB32X32
  for(i=0; i<(1<<(2*MAX_MB_EXT_LEVEL)); i++)
  {
    get_mem_DCcoeff (&cofDC32[i]);
    get_mem_ACcoeff (&cofAC32[i]);
    get_mem_DCcoeff (&cofDC32_RDCost[i]);
    get_mem_ACcoeff (&cofAC32_RDCost[i]);

    get_mem2Dint(&cof32AC16x16[i],2,257);
    get_mem3Dint(&cof32AC16x8[i], 2, 2, 129);
    get_mem3Dint(&cof32AC8x16[i], 2, 2, 129);
    get_mem2Dint(&cof32AC16x16_RDCost[i],2,257);
    get_mem3Dint(&cof32AC16x8_RDCost[i], 2, 2, 129);
    get_mem3Dint(&cof32AC8x16_RDCost[i], 2, 2, 129);
  }
#endif
  rdopt = NULL;
  
  get_mem_DCcoeff (&cofDC);
  get_mem_ACcoeff (&cofAC);
//this is used in Intra 8x8 mode decision
  for(i=0;i<NumberOfQP;i++)
  {
    get_mem_ACcoeff (&(tr4x4[i].cofAC8x8)); 
  }
    get_mem_ACcoeff (&cofAC8x8);

  get_mem_ACcoeff (&cofAC4x4intern);
  cofAC4x4 = cofAC4x4intern[0][0];
  
  if (input->Transform8x8Mode)
  {
  for(i=0;i<NumberOfQP;i++)
  {
    get_mem_ACcoeff (&(tr8x8[i].cofAC8x8));
  }
  }
  
  // structure for saving the coding state
  cs_mb  = create_coding_state ();
  cs_b8  = create_coding_state ();
  cs_cm  = create_coding_state ();
  cs_imb = create_coding_state ();
  cs_ib8 = create_coding_state ();
  cs_ib4 = create_coding_state ();
  cs_pc  = create_coding_state ();
#ifdef USE_INTRA_MDDT
  cs_i16  = create_coding_state ();
#endif 
  cs_smb      = create_coding_state ();
  cs_eachQP   = create_coding_state ();
}



/*!
*************************************************************************************
* \brief
*    Updates the pixel map that shows, which reference frames are reliable for
*    each MB-area of the picture.
*
* \note
*    The new values of the pixel_map are taken from the temporary buffer refresh_map
*
*************************************************************************************
*/
void UpdatePixelMap()
{
  int mx,my,y,x,i,j;
  if (img->type==I_SLICE)
  {
    for (y=0; y<img->height; y++)
      for (x=0; x<img->width; x++)
      {
        pixel_map[y][x]=1;
      }
  }
  else
  {
    for (my=0; my<img->height >> 3; my++)
      for (mx=0; mx<img->width >> 3;  mx++)
      {
        j = my*8 + 8;
        i = mx*8 + 8;
        if (refresh_map[my][mx])
        {
          for (y=my*8; y<j; y++)
            for (x=mx*8; x<i; x++)  
              pixel_map[y][x] = 1;
        }
        else
        {
          for (y=my*8; y<j; y++)
            for (x=mx*8; x<i; x++)  
              pixel_map[y][x] = min(pixel_map[y][x] + 1, input->num_ref_frames+1);
        }
      }
  }
}

/*!
*************************************************************************************
* \brief
*    Checks if a given reference frame is reliable for the current
*    macroblock, given the motion vectors that the motion search has
*    returned.
*
* \return
*    If the return value is 1, the reference frame is reliable. If it
*    is 0, then it is not reliable.
*
* \note
*    A specific area in each reference frame is assumed to be unreliable
*    if the same area has been intra-refreshed in a subsequent frame.
*    The information about intra-refreshed areas is kept in the pixel_map.
*
*************************************************************************************
*/
int CheckReliabilityOfRef (int block, int list_idx, int ref, int mode)
{
  int y,x, block_y, block_x, dy, dx, y_pos, x_pos, yy, xx, pres_x, pres_y;
  int maxold_x  = img->width-1;
  int maxold_y  = img->height-1;
  int ref_frame = ref+1;
  
  int by0 = (mode>=4?2*(block >> 1):mode==2?2*block:0);
  int by1 = by0 + (mode>=4||mode==2?2:4);
  int bx0 = (mode>=4?2*(block & 0x01):mode==3?2*block:0);
  int bx1 = bx0 + (mode>=4||mode==3?2:4);
  
  for (block_y=by0; block_y<by1; block_y++)
  {
    for (block_x=bx0; block_x<bx1; block_x++)
    {
      y_pos  = img->all_mv[block_y][block_x][list_idx][ref][mode][1];
      y_pos += (img->block_y + block_y) * BLOCK_SIZE * 4;
      x_pos  = img->all_mv[block_y][block_x][list_idx][ref][mode][0];
      x_pos += (img->block_x + block_x) * BLOCK_SIZE * 4;
      
      /* Here we specify which pixels of the reference frame influence
      the reference values and check their reliability. This is
      based on the function Get_Reference_Pixel */
      
      dy = y_pos & 3;
      dx = x_pos & 3;
      
      y_pos = (y_pos-dy) >> 2;
      x_pos = (x_pos-dx) >> 2;
      
      if (dy==0 && dx==0) //full-pel
      {
        for (y=y_pos ; y < y_pos + BLOCK_SIZE ; y++)
          for (x=x_pos ; x < x_pos + BLOCK_SIZE ; x++)
            if (pixel_map[max(0,min(maxold_y,y))][max(0,min(maxold_x,x))] < ref_frame)
              return 0;
      }
      else  /* other positions */
      {
        if (dy == 0)
        {
          for (y = y_pos ; y < y_pos + BLOCK_SIZE ; y++)
          {
            pres_y = max(0,min(maxold_y,y));
            for (x = x_pos ; x < x_pos + BLOCK_SIZE ; x++)
            {
              for(xx = -2 ; xx < 4 ; xx++) {
                pres_x = max(0, min( maxold_x, x + xx));
                if (pixel_map[pres_y][pres_x] < ref_frame)
                  return 0;
              }
            }
          }
        }        
        else if (dx == 0)
        {
          for (y = y_pos ; y < y_pos + BLOCK_SIZE ; y++)
            for (x=x_pos ; x < x_pos + BLOCK_SIZE ; x++)
            {
              pres_x = max(0,min(maxold_x,x));
              for(yy=-2;yy<4;yy++) {
                pres_y = max(0,min(maxold_y, yy + y));
                if (pixel_map[pres_y][pres_x] < ref_frame)
                  return 0;
              }
            }
        }
        else if (dx == 2)
        {
          for (y = y_pos ; y < y_pos + BLOCK_SIZE ; y++)
            for (x = x_pos ; x < x_pos + BLOCK_SIZE ; x++)
            {
              for(yy=-2;yy<4;yy++) {
                pres_y = max(0,min(maxold_y, yy + y));
                for(xx=-2;xx<4;xx++) {
                  pres_x = max(0,min(maxold_x, xx + x));
                  if (pixel_map[pres_y][pres_x] < ref_frame)
                    return 0;
                }
              }
            }
        }
        else if (dy == 2)
        {
          for (y = y_pos ; y < y_pos + BLOCK_SIZE ; y++)
            for (x = x_pos ; x < x_pos + BLOCK_SIZE ; x++)
            {
              for(xx=-2;xx<4;xx++) {
                pres_x = max(0,min(maxold_x, xx + x));
                for(yy=-2;yy<4;yy++) {
                  pres_y = max(0,min(maxold_y, yy + y));
                  if (pixel_map[pres_y][pres_x] < ref_frame)
                    return 0;
                }
              }
            }
        }
        else
        {
          for (y = y_pos ; y < y_pos + BLOCK_SIZE ; y++)
          {
            for (x = x_pos ; x < x_pos + BLOCK_SIZE ; x++)
            {
              pres_y = dy == 1 ? y : y + 1;
              pres_y = max(0,min(maxold_y,pres_y));
              
              for(xx=-2;xx<4;xx++) 
              {
                pres_x = max(0,min(maxold_x,xx + x));
                if (pixel_map[pres_y][pres_x] < ref_frame)
                  return 0;
              }
              
              pres_x = dx == 1 ? x : x + 1;
              pres_x = max(0,min(maxold_x,pres_x));
              
              for(yy=-2;yy<4;yy++) 
              {
                pres_y = max(0,min(maxold_y, yy + y));
                if (pixel_map[pres_y][pres_x] < ref_frame)
                  return 0;
              }
            }
          }
        }        
      }
    }
  }
  return 1;
}



/*!
*************************************************************************************
* \brief
*    R-D Cost for an 4x4 Intra block
*************************************************************************************
*/
double RDCost_for_4x4IntraBlocks (int*    nonzero,
                                  int     b8,
                                  int     b4,
                                  int     ipmode,
                                  double  lambda,
                                  double  min_rdcost,
                                  int mostProbableMode)
{
  double  rdcost;
  int     dummy, rate;
  int64   distortion  = 0;
  int     block_x     = 8*(b8 & 0x01)+4*(b4 & 0x01);
  int     block_y     = 8*(b8 >> 1)+4*(b4 >> 1);
  int     pic_pix_x   = img->pix_x+block_x;
  int     pic_pix_y   = img->pix_y+block_y;
  int     pic_opix_y  = img->opix_y+block_y;
  imgpel  **imgY      = enc_picture->imgY;
  
  Slice          *currSlice    =  img->currentSlice;
  Macroblock     *currMB       = &img->mb_data[img->current_mb_nr];
  SyntaxElement  *currSE       = &img->MB_SyntaxElements[currMB->currSEnr];
  const int      *partMap      = assignSE2partition[input->partition_mode];
  DataPartition  *dataPart;
  
  //===== perform DCT, Q, IQ, IDCT, Reconstruction =====
  dummy = 0;
#ifdef USE_INTRA_MDDT
    if(input->UseIntraMDDT)
    {
      *nonzero = (ipmode == 2) ? 
        dct_luma (block_x, block_y, &dummy, 1, ipmode): 
      klt_luma_sep (block_x, block_y, &dummy, ipmode);
    }
    else 
      *nonzero = dct_luma (block_x, block_y, &dummy, 1, ipmode);
#else
    *nonzero = dct_luma (block_x, block_y, &dummy, 1);
#endif
    
    //===== get distortion (SSD) of 4x4 block =====
  distortion += compute_SSE(imgY_org, imgY , pic_opix_y, pic_pix_y, pic_pix_x, pic_pix_x, 4, 4);
      if(distortion > min_rdcost)
      {
        return (double)distortion;
      }

    //===== RATE for INTRA PREDICTION MODE  (SYMBOL MODE MUST BE SET TO UVLC) =====
    currSE->value1 = (mostProbableMode == ipmode) ? -1 : ipmode < mostProbableMode ? ipmode : ipmode-1;
    
    //--- set position and type ---
    currSE->context = 4*b8 + b4;
    currSE->type    = SE_INTRAPREDMODE;
    
    //--- choose data partition ---
    dataPart = &(currSlice->partArr[partMap[SE_INTRAPREDMODE]]);
    //--- encode and update rate ---
    if (input->symbol_mode == UVLC)    
      writeSyntaxElement_Intra4x4PredictionMode(currSE, dataPart);
    else
    {
      currSE->writing = writeIntraPredMode_CABAC;
      dataPart->writeSyntaxElement (currSE, dataPart);
    }
    
    rate = currSE->len;
    currSE++;
    currMB->currSEnr++;
    
    //===== RATE for LUMINANCE COEFFICIENTS =====
    if (input->symbol_mode == UVLC)
    {
      rate  += writeCoeff4x4_CAVLC (LUMA, b8, b4, 0);
    }
    else
    {
      rate  += writeLumaCoeff4x4_CABAC (b8, b4, 1);
    }
    //reset_coding_state (cs_cm);
    rdcost = (double)distortion + lambda*(double)rate;

      return rdcost;
}
                                  
/*!
*************************************************************************************
* \brief
*    Mode Decision for an 4x4 Intra block
*************************************************************************************
*/
int Mode_Decision_for_4x4IntraBlocks (int  b8,  int  b4,  double  lambda,  int*  min_cost)
{
  int     ipmode, best_ipmode = 0, i, j, k, y, cost, dummy;
  int     c_nz, nonzero = 0, diff[16];
  imgpel  rec4x4[4][4];
  double  rdcost;
  int     block_x     = 8*(b8 & 0x01)+4*(b4 & 0x01);
  int     block_y     = 8*(b8 >> 1)+4*(b4 >> 1);
  int     block_x4    = block_x >> 2;
  int     block_y4    = block_y >> 2;
  int     pic_pix_x   = img->pix_x+block_x;
  int     pic_pix_y   = img->pix_y+block_y;
  int     pic_opix_x   = img->opix_x+block_x;
  int     pic_opix_y   = img->opix_y+block_y;
  int     pic_block_x = pic_pix_x >> 2;
  int     pic_block_y = pic_pix_y >> 2;
  double  min_rdcost  = 1e30;
  
  int left_available, up_available, all_available;
  
  char   upMode;
  char   leftMode;
  int     mostProbableMode;
  
  PixelPos left_block;
  PixelPos top_block;
  
  int fixedcost = (int) floor(4 * lambda );

#ifdef BEST_NZ_COEFF
  int best_nz_coeff = 0;
  int best_coded_block_flag = 0;
  int bit_pos = 1 + ((((b8>>1)<<1)+(b4>>1))<<2) + (((b8&1)<<1)+(b4&1));
  static int64 cbp_bits;

  if (b8==0 && b4==0)
      cbp_bits = 0;

  
#endif
  
  getLuma4x4Neighbour(img->current_mb_nr, block_x4, block_y4, -1,  0, &left_block);
  getLuma4x4Neighbour(img->current_mb_nr, block_x4, block_y4,  0, -1, &top_block);
  
  // constrained intra pred
  if (input->UseConstrainedIntraPred)
  {
    left_block.available = left_block.available ? img->intra_block[left_block.mb_addr] : 0;
    top_block.available  = top_block.available  ? img->intra_block[top_block.mb_addr]  : 0;
  }
  
  upMode            =  top_block.available ? img->ipredmode[top_block.pos_y ][top_block.pos_x ] : -1;
  leftMode          = left_block.available ? img->ipredmode[left_block.pos_y][left_block.pos_x] : -1;
  
  mostProbableMode  = (upMode < 0 || leftMode < 0) ? DC_PRED : upMode < leftMode ? upMode : leftMode;
  
  *min_cost = INT_MAX;
  
  //===== INTRA PREDICTION FOR 4x4 BLOCK =====
  intrapred_luma (pic_pix_x, pic_pix_y, &left_available, &up_available, &all_available);
  
  //===== LOOP OVER ALL 4x4 INTRA PREDICTION MODES =====
  for (ipmode=0; ipmode<NO_INTRA_PMODE; ipmode++)
  {
    int available_mode =  (ipmode==DC_PRED) ||
      ((ipmode==VERT_PRED||ipmode==VERT_LEFT_PRED||ipmode==DIAG_DOWN_LEFT_PRED) && up_available ) ||
      ((ipmode==HOR_PRED||ipmode==HOR_UP_PRED) && left_available ) ||(all_available);
    
    if (input->IntraDisableInterOnly==0 || img->type != I_SLICE)
    {
      if (input->Intra4x4ParDisable && (ipmode==VERT_PRED||ipmode==HOR_PRED))
        continue;      
      
      if (input->Intra4x4DiagDisable && (ipmode==DIAG_DOWN_LEFT_PRED||ipmode==DIAG_DOWN_RIGHT_PRED))
        continue;      
      
      if (input->Intra4x4DirDisable && ipmode>=VERT_RIGHT_PRED)
        continue;
    }
    
    if( available_mode)
    {
      if (!input->rdopt)
      {
        for (k=j=0; j<4; j++)
        {
          int jj = pic_opix_y+j;
          for (i=0; i<4; i++, k++)
          {
            diff[k] = imgY_org[jj][pic_opix_x+i] - img->mprr[ipmode][j][i];
          }
        }
        //cost  = (ipmode == mostProbableMode) ? 0 : (int)floor(4 * lambda );
        cost  = (ipmode == mostProbableMode) ? 0 : fixedcost;
        cost += SATD (diff, input->hadamard);
        if (cost < *min_cost)
        {
          best_ipmode = ipmode;
          *min_cost   = cost;
        }
      }
      else
      {
          // get prediction and prediction error
          for (j=0; j<4; j++)
          {
            memcpy(&img->mpr[block_y+j][block_x], img->mprr[ipmode][j], BLOCK_SIZE * sizeof(imgpel));
            for (i=0; i<4; i++)
            {
              img->m7[j][i] = (int) (imgY_org[pic_opix_y+j][pic_opix_x+i] - img->mprr[ipmode][j][i]);
            }
          }
          
          //===== store the coding state =====
          //store_coding_state (cs_cm);
          // get and check rate-distortion cost
#ifdef BEST_NZ_COEFF
          img->mb_data[img->current_mb_nr].cbp_bits = cbp_bits;
#endif
          if ((rdcost = RDCost_for_4x4IntraBlocks (&c_nz, b8, b4, ipmode, lambda, min_rdcost, mostProbableMode)) < min_rdcost)
          {
            //--- set coefficients ---
            memcpy(cofAC4x4[0],img->cofAC[b8][b4][0], 18 * sizeof(int));
            memcpy(cofAC4x4[1],img->cofAC[b8][b4][1], 18 * sizeof(int));
            
            //--- set reconstruction ---
            for (y=0; y<4; y++)
            {
              memcpy(rec4x4[y],&enc_picture->imgY[pic_pix_y+y][pic_pix_x], BLOCK_SIZE * sizeof(imgpel));
            } 
            //--- flag if dct-coefficients must be coded ---
            nonzero = c_nz;
            
            //--- set best mode update minimum cost ---
            min_rdcost    = rdcost;
            best_ipmode   = ipmode;
#ifdef BEST_NZ_COEFF
            best_nz_coeff = img->nz_coeff [img->current_mb_nr][block_x4][block_y4];
            best_coded_block_flag = (int)((img->mb_data[img->current_mb_nr].cbp_bits>>bit_pos)&(int64)(1));
#endif            
            //store_coding_state (cs_ib4);
          }
          
#ifndef RESET_STATE
          reset_coding_state (cs_cm);
#endif
      }
    }
  }


#ifdef BEST_NZ_COEFF
  img->nz_coeff [img->current_mb_nr][block_x4][block_y4] = best_nz_coeff;
  cbp_bits &= (~(int64)(1<<bit_pos));
  cbp_bits |= (int64)(best_coded_block_flag<<bit_pos);
#endif
   
    //===== set intra mode prediction =====
    img->ipredmode[pic_block_y][pic_block_x] = best_ipmode;
    img->mb_data[img->current_mb_nr].intra_pred_modes[4*b8+b4] = mostProbableMode == best_ipmode ? -1 : best_ipmode < mostProbableMode ? best_ipmode : best_ipmode-1;
    
  if (!input->rdopt)
    {
        // get prediction and prediction error
        for (j=0; j<4; j++)
        {
          int jj = pic_opix_y+j;
          for (i=0; i<4; i++)
          {
            img->mpr[block_y+j][block_x+i]  = img->mprr[best_ipmode][j][i];
            img->m7[j][i]                   = imgY_org[jj][pic_opix_x+i] - img->mprr[best_ipmode][j][i];
          }
        }
#ifdef USE_INTRA_MDDT
        if(input->UseIntraMDDT)
        {
          nonzero = (best_ipmode == 2) ? 
            dct_luma     (block_x, block_y, &dummy, 1, best_ipmode) : 
          klt_luma_sep (block_x, block_y, &dummy, best_ipmode);
        }
        else 
          nonzero = dct_luma     (block_x, block_y, &dummy, 1, best_ipmode);
#else
        nonzero = dct_luma (block_x, block_y, &dummy, 1);
#endif
  }
  else
  {
    //===== restore coefficients =====
    for (j=0; j<2; j++)
    {
      memcpy (img->cofAC[b8][b4][j],cofAC4x4[j], 18 * sizeof(int));
    } 
   
    //===== restore reconstruction and prediction (needed if single coeffs are removed) =====
    for (y=0; y<BLOCK_SIZE; y++)
    {
      memcpy (&enc_picture->imgY[pic_pix_y+y][pic_pix_x],rec4x4[y],    BLOCK_SIZE * sizeof(imgpel));
      memcpy (&img->mpr[block_y+y][block_x],img->mprr[best_ipmode][y], BLOCK_SIZE * sizeof(imgpel));
    }
  }  
  return nonzero;
}


/*!
*************************************************************************************
* \brief
*    Mode Decision for an 8x8 Intra block
*************************************************************************************
*/
int Mode_Decision_for_8x8IntraBlocks(int b8,double lambda,int *cost)
{
  int  nonzero=0, b4;
  int  cost4x4;
  
  *cost = (int)floor(6.0 * lambda + 0.4999);
  
  for (b4=0; b4<4; b4++)
  {
    if (Mode_Decision_for_4x4IntraBlocks (b8, b4, lambda, &cost4x4))
    {
      nonzero        = 1;
    }
    *cost += cost4x4;
  }
#ifdef RESET_STATE
  reset_coding_state (cs_cm);
#endif
  
  return nonzero;
}

/*!
*************************************************************************************
* \brief
*    4x4 Intra mode decision for an macroblock
*************************************************************************************
*/
int Mode_Decision_for_Intra4x4Macroblock (double lambda,  int* cost)
{
  int  cbp=0, b8, cost8x8;
  
  for (*cost=0, b8=0; b8<4; b8++)
  {
    if (Mode_Decision_for_8x8IntraBlocks (b8, lambda, &cost8x8))
    {
      cbp |= (1<<b8);
    }
    *cost += cost8x8;
  }
  
  return cbp;
}


/*!
*************************************************************************************
* \brief
*    R-D Cost for an 8x8 Partition
*************************************************************************************
*/
double RDCost_for_8x8blocks (int*    cnt_nonz,   // --> number of nonzero coefficients
                             int64*  cbp_blk,    // --> cbp blk
                             double  lambda,     // <-- lagrange multiplier
                             int     block,      // <-- 8x8 block number
                             int     mode,       // <-- partitioning mode
                             short   pdir,       // <-- prediction direction
                             short   ref,        // <-- reference frame
                             short   bwd_ref
                             ,double* min_rdcost
                             )    // <-- abp type
{
  int  rate=0;
  int64 distortion=0;
  int  dummy = 0, mrate;
  int  fw_mode, bw_mode;
  int  cbp     = 0;
  int  pax     = 8*(block & 0x01);
  int  pay     = 8*(block >> 1);
  int  i0      = pax >> 2;
  int  j0      = pay >> 2;
  int  bframe  = (img->type==B_SLICE);
  int  direct  = (bframe && mode==0);
  int  b8value = B8Mode2Value (mode, pdir);
  
  Macroblock    *currMB    = &img->mb_data[img->current_mb_nr];
  SyntaxElement *currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice         *currSlice = img->currentSlice;
  DataPartition *dataPart;
  const int     *partMap   = assignSE2partition[input->partition_mode];
  
  EncodingEnvironmentPtr eep_dp;
  
  if (input->mv_competition > 0)// skip_mode is for P_SKIP..and here P_SKIP will never happen coz this is 8x8
    skip_mode = FALSE;
#if P_DIRECT    //for sub mb (8x8 and below), there is no SKIP...PSkip or Bskip...and also there is no P_Direct for 8x8...
  img->NoResidueDirect = 0;
#endif


  //=====
  //=====  GET COEFFICIENTS, RECONSTRUCTIONS, CBP
  //=====
  currMB->bi_pred_me=0;
  
  if (direct)
  {
    if (direct_pdir[img->block_y+j0][img->block_x+i0]<0) // mode not allowed
      return (1e20);
    else
      *cnt_nonz = LumaResidualCoding8x8 (0, 0, mode, -1, &cbp, cbp_blk, block, direct_pdir[img->block_y+j0][img->block_x+i0], 0, 0, 
      (short)max(0,direct_ref_idx[LIST_0][img->block_y+j0][img->block_x+i0]), direct_ref_idx[LIST_1][img->block_y+j0][img->block_x+i0]);
  }
  else
  {
    fw_mode   = (pdir==0||pdir==2 ? mode : 0);
    bw_mode   = (pdir==1||pdir==2 ? mode : 0);
    *cnt_nonz = LumaResidualCoding8x8 (0, 0, mode, -1, &cbp, cbp_blk, block, pdir, fw_mode, bw_mode, ref, bwd_ref);
  }
  
  //=====
  //=====   GET DISTORTION
  //=====
  distortion += compute_SSE(imgY_org, enc_picture->imgY, img->opix_y+pay, img->pix_y+pay, img->pix_x+pax, img->pix_x+pax, 8, 8);

   if(distortion > *min_rdcost)
     return (double)distortion;

  //=====
  //=====   GET RATE
  //=====
  //----- block 8x8 mode -----
  if (input->symbol_mode == UVLC)
  {
    ue_linfo (b8value, dummy, &mrate, &dummy);
    rate += mrate;
  }
  else
  {
    currSE->value1  = b8value;
#if B_MBTYPE
    currSE->dir_partition1 = pdir;    //what value to set here-> see the input parameter in B8Mode2Value()
#endif
    currSE->writing = writeB8_typeInfo_CABAC;

    currSE->type    = SE_MBTYPE;
    dataPart = &(currSlice->partArr[partMap[currSE->type]]);
    dataPart->writeSyntaxElement (currSE, dataPart);
    rate += currSE->len;
    currSE++;
    currMB->currSEnr++;
  }
  
  //----- motion information -----
  if (!direct)
  {
    if ((img->num_ref_idx_l0_active > 1 ) && (pdir==0 || pdir==2))
      rate  += writeReferenceFrame (mode, i0, j0, 1, ref);
    if(img->num_ref_idx_l1_active > 1 && img->type== B_SLICE)
    {
      if (pdir==1 || pdir==2)
      {
        rate  += writeReferenceFrame (mode, i0, j0, 0, bwd_ref);
      }
    }

#ifdef MV_COMPETITION
    if (input->mv_competition)
      pass_with_writing = 0;
#endif

    if (pdir==0 || pdir==2)
    {
      rate  += writeMotionVector8x8 (i0, j0, i0+2, j0+2, ref,LIST_0, mode);
    }
    if (pdir==1 || pdir==2)
    {
      rate  += writeMotionVector8x8 (i0, j0, i0+2, j0+2, bwd_ref, LIST_1, mode);
    }

#ifdef MV_COMPETITION
    if (input->mv_competition)
      pass_with_writing = 1;
#endif
  }
  
  //----- coded block pattern (for CABAC only) -----
  if (input->symbol_mode == CABAC)
  {
    dataPart = &(currSlice->partArr[partMap[SE_CBP_INTER]]);
    eep_dp   = &(dataPart->ee_cabac);
    mrate    = arienco_bits_written (eep_dp);
    writeCBP_BIT_CABAC (block, ((*cnt_nonz>0)?1:0), cbp8x8, currMB, 1, eep_dp);
    mrate    = arienco_bits_written (eep_dp) - mrate;
    rate    += mrate;
  }
  
  //----- luminance coefficients -----
  if (*cnt_nonz)
  {
    rate += writeLumaCoeff8x8 (block, mode, currMB->luma_transform_size_8x8_flag);
  }
  
  return (double)distortion + lambda * (double)rate;
}


/*!
*************************************************************************************
* \brief
*    Gets mode offset for intra16x16 mode
*************************************************************************************
*/
int I16Offset (int cbp, int i16mode)
{
  return (cbp&15?13:1) + i16mode + ((cbp&0x30)>>2);
}


/*!
*************************************************************************************
* \brief
*    Sets modes and reference frames for a macroblock
*************************************************************************************
*/
void SetModesAndRefframeForBlocks (int mode)
{
  int i,j,k,l;
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];
  int  bframe  = (img->type==B_SLICE);
  int  block_x, block_y;
  int  cur_ref[2];  
  
  //--- macroblock type ---
  currMB->mb_type = mode;    
  currMB->bi_pred_me= (mode == 1 ? img->bi_pred_me[mode] : 0);  
  
  //--- block 8x8 mode and prediction direction ---
  switch (mode)
  {
  case 0:
    for(i=0;i<4;i++)
    {
      currMB->b8mode[i] = 0;
      currMB->b8pdir[i] = (bframe ? direct_pdir[img->block_y + (i >> 1)*2][img->block_x + (i & 0x01)*2] : 0);
    }
    break;
  case 1:
  case 2:
  case 3:
    for(i=0;i<4;i++)
    {
      currMB->b8mode[i] = mode;
      currMB->b8pdir[i] = best8x8pdir[mode][i];
    }
    break;
  case P8x8:              //NOT NECESSARY...DUMMY VALUES
    //for(i=0;i<4;i++)    //is taken care inside SetCoeffAndReconstruction8x8()
    //{
    //  currMB->b8mode[i]   = best8x8mode[i];
    //  currMB->b8pdir[i]   = best8x8pdir[mode][i];
    //}
    break;
  case I4MB:
    for(i=0;i<4;i++)
    {
      currMB->b8mode[i] = IBLOCK; 
      currMB->b8pdir[i] = -1;
    }
    break;
  case I16MB:
    for(i=0;i<4;i++)
    {
      currMB->b8mode[i] =  0;
      currMB->b8pdir[i] = -1;
    }
    break;
  case I8MB:
    for(i=0;i<4;i++)
    {
      currMB->b8mode[i] = I8MB;
      currMB->b8pdir[i] = -1;
    }
    //switch to 8x8 transform
    currMB->luma_transform_size_8x8_flag = 1;
    break;
  default:
    printf ("Unsupported mode in SetModesAndRefframeForBlocks!\n");
    exit (1);
  }
  
//#define IS_FW ((best8x8pdir[mode][k]==0 || best8x8pdir[mode][k]==2) && (mode!=P8x8 || best8x8mode[k]!=0 || !bframe))
//#define IS_BW ((best8x8pdir[mode][k]==1 || best8x8pdir[mode][k]==2) && (mode!=P8x8 || best8x8mode[k]!=0))
#define IS_FW ((best8x8pdir[mode][k]==0 || best8x8pdir[mode][k]==2) && (mode!=P8x8 || !bframe))
#define IS_BW ((best8x8pdir[mode][k]==1 || best8x8pdir[mode][k]==2) && (mode!=P8x8))

  //--- reference frame arrays ---
  if (mode==0 || mode==I4MB || mode==I16MB || mode==I8MB)
  {
    if (bframe)
    {
      if (!mode)
      {
          memcpy(&enc_picture->ref_idx[LIST_0][img->block_y    ][img->block_x],&direct_ref_idx[LIST_0][img->block_y    ][img->block_x], 4 * sizeof(char));
          memcpy(&enc_picture->ref_idx[LIST_0][img->block_y + 1][img->block_x],&direct_ref_idx[LIST_0][img->block_y + 1][img->block_x], 4 * sizeof(char));
          memcpy(&enc_picture->ref_idx[LIST_0][img->block_y + 2][img->block_x],&direct_ref_idx[LIST_0][img->block_y + 2][img->block_x], 4 * sizeof(char));
          memcpy(&enc_picture->ref_idx[LIST_0][img->block_y + 3][img->block_x],&direct_ref_idx[LIST_0][img->block_y + 3][img->block_x], 4 * sizeof(char));

          memcpy(&enc_picture->ref_idx[LIST_1][img->block_y    ][img->block_x],&direct_ref_idx[LIST_1][img->block_y    ][img->block_x], 4 * sizeof(char));
          memcpy(&enc_picture->ref_idx[LIST_1][img->block_y + 1][img->block_x],&direct_ref_idx[LIST_1][img->block_y + 1][img->block_x], 4 * sizeof(char));
          memcpy(&enc_picture->ref_idx[LIST_1][img->block_y + 2][img->block_x],&direct_ref_idx[LIST_1][img->block_y + 2][img->block_x], 4 * sizeof(char));
          memcpy(&enc_picture->ref_idx[LIST_1][img->block_y + 3][img->block_x],&direct_ref_idx[LIST_1][img->block_y + 3][img->block_x], 4 * sizeof(char));
      }
      else
      {
          memset(&enc_picture->ref_idx[LIST_0][img->block_y    ][img->block_x],-1, 4 * sizeof(char));
          memset(&enc_picture->ref_idx[LIST_0][img->block_y + 1][img->block_x],-1, 4 * sizeof(char));
          memset(&enc_picture->ref_idx[LIST_0][img->block_y + 2][img->block_x],-1, 4 * sizeof(char));
          memset(&enc_picture->ref_idx[LIST_0][img->block_y + 3][img->block_x],-1, 4 * sizeof(char));

          memset(&enc_picture->ref_idx[LIST_1][img->block_y    ][img->block_x],-1, 4 * sizeof(char));
          memset(&enc_picture->ref_idx[LIST_1][img->block_y + 1][img->block_x],-1, 4 * sizeof(char));
          memset(&enc_picture->ref_idx[LIST_1][img->block_y + 2][img->block_x],-1, 4 * sizeof(char));
          memset(&enc_picture->ref_idx[LIST_1][img->block_y + 3][img->block_x],-1, 4 * sizeof(char));
      }
    }
    else
    {
      if (!mode)
      {
          memset(&enc_picture->ref_idx[LIST_0][img->block_y    ][img->block_x],0, 4 * sizeof(char));
          memset(&enc_picture->ref_idx[LIST_0][img->block_y + 1][img->block_x],0, 4 * sizeof(char));
          memset(&enc_picture->ref_idx[LIST_0][img->block_y + 2][img->block_x],0, 4 * sizeof(char));
          memset(&enc_picture->ref_idx[LIST_0][img->block_y + 3][img->block_x],0, 4 * sizeof(char));
      }
      else
      {
          memset(&enc_picture->ref_idx[LIST_0][img->block_y    ][img->block_x],-1, 4 * sizeof(char));
          memset(&enc_picture->ref_idx[LIST_0][img->block_y + 1][img->block_x],-1, 4 * sizeof(char));
          memset(&enc_picture->ref_idx[LIST_0][img->block_y + 2][img->block_x],-1, 4 * sizeof(char));
          memset(&enc_picture->ref_idx[LIST_0][img->block_y + 3][img->block_x],-1, 4 * sizeof(char));
      }
    }
  }
  else
  {
    if (bframe)
    {
      for (j=0;j<4;j++)
      {
        block_y = img->block_y + j;
        for (i=0;i<4;i++)
        {
          block_x = img->block_x + i;
          k = 2*(j >> 1) + (i >> 1);
          l = 2*(j & 0x01) + (i & 0x01);
          
            enc_picture->ref_idx[LIST_0][block_y][block_x] = -1;
            enc_picture->ref_idx[LIST_1][block_y][block_x] = -1;

          //if(mode == P8x8 && best8x8mode[k]==0)
          //{
          //  enc_picture->ref_idx[LIST_0][block_y][block_x] = direct_ref_idx[LIST_0][block_y][block_x];
          //  enc_picture->ref_idx[LIST_1][block_y][block_x] = direct_ref_idx[LIST_1][block_y][block_x];
          //}
          /*else*/ if (mode ==1 && currMB->bi_pred_me && IS_FW && IS_BW)
          {
            enc_picture->ref_idx[LIST_0][block_y][block_x] = 0;
            enc_picture->ref_idx[LIST_1][block_y][block_x] = 0;
          }
          else
          {
            enc_picture->ref_idx[LIST_0][block_y][block_x] = (IS_FW ? best8x8fwref[mode][k] : -1);
            enc_picture->ref_idx[LIST_1][block_y][block_x] = (IS_BW ? best8x8bwref[mode][k] : -1);
          }
        }
      }
    }
    else
    {
      for (j=0;j<4;j++)
      {
        block_y = img->block_y + j;
        for (i=0;i<4;i++)
        {
          block_x = img->block_x + i;
          k = 2*(j >> 1) + (i >> 1);
          l = 2*(j & 0x01) + (i & 0x01);
          enc_picture->ref_idx[LIST_0][block_y][block_x] = (IS_FW ? best8x8fwref[mode][k] : -1);
        }
      }
    }
  }
  
  if (bframe)
  {
    
    for (j = img->block_y; j < img->block_y + 4; j++)
      for (i = img->block_x; i < img->block_x + 4;i++)
      {
        cur_ref[LIST_0] = (int) enc_picture->ref_idx[LIST_0][j][i];
        cur_ref[LIST_1] = (int) enc_picture->ref_idx[LIST_1][j][i];
        
        enc_picture->ref_pic_id [LIST_0][j][i] = (cur_ref[LIST_0]>=0 
          ? enc_picture->ref_pic_num[LIST_0 + currMB->list_offset][cur_ref[LIST_0]]
          : -1);
        enc_picture->ref_pic_id [LIST_1][j][i] = (cur_ref[LIST_1]>=0 
          ? enc_picture->ref_pic_num[LIST_1 + currMB->list_offset][cur_ref[LIST_1]]
          : -1);
      }
  }
  else
  {  
    for (j = img->block_y; j < img->block_y + 4; j++)
      for (i = img->block_x; i < img->block_x + 4;i++)
      {
        cur_ref[LIST_0] = (int) enc_picture->ref_idx[LIST_0][j][i];
        enc_picture->ref_pic_id [LIST_0][j][i] = (cur_ref[LIST_0]>=0 
          ? enc_picture->ref_pic_num[LIST_0 + currMB->list_offset][cur_ref[LIST_0]]
          : -1);
      }
  }
  
#undef IS_FW
#undef IS_BW


  if(mode != P8x8)
  //==== motion vectors =====
  SetMotionVectorsMB (currMB, bframe);


}


#ifdef USE_INTRA_MDDT
/*! 
*************************************************************************************
* \brief
*   RD-optimized mode decision for INTRA16x16 mode
*
* \para Intra16x16_Mode_Decision_RDopt()
*    <paragraph>
*
* \author
*    - Yan Ye                      <yye@qualcomm.com>
*************************************************************************************
*/
void Intra16x16_Mode_Decision_RDopt (Macroblock* currMB, int* i16mode, double lambda)
{
  extern int MBType2Value (Macroblock* currMB);
  int i, j;
  
  int mode16, best_mode16;
  int max_mode = 4;
  double rdcost, min_rdcost = 1e30;
  int distortion, rate; 
  imgpel **imgY = enc_picture->imgY; 
  int y;
  extern CSptr cs_i16;
  SyntaxElement  *currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
  DataPartition*  dataPart;
  const int*      partMap    = assignSE2partition[input->partition_mode];
  Slice*          currSlice  = img->currentSlice;
  int  bestCofAC16[2][16*17];
  int  best_cbp=0;
  imgpel bestRec[MB_BLOCK_SIZE][MB_BLOCK_SIZE];
  
  intrapred_luma_16x16 ();   /* make intra pred for all 4 new modes */
  
  {
    PixelPos up;          //!< pixel position p(0,-1)
    PixelPos left[17];    //!< pixel positions p(-1, -1..15)
    
    int up_avail, left_avail, left_up_avail;
    
    for (i=0;i<17;i++)
    {
      getNeighbour(img->current_mb_nr, -1 ,  i-1 , 1, &left[i]);
    }
    
    getNeighbour(img->current_mb_nr, 0     ,  -1 , 1, &up);
    
    if (!(input->UseConstrainedIntraPred))
    {
      up_avail   = up.available;
      left_avail = left[1].available;
      left_up_avail = left[0].available;
    }
    else
    {
      up_avail      = up.available ? img->intra_block[up.mb_addr] : 0;
      for (i=1, left_avail=1; i<17;i++)
        left_avail  &= left[i].available ? img->intra_block[left[i].mb_addr]: 0;
      left_up_avail = left[0].available ? img->intra_block[left[0].mb_addr]: 0;
    }
    
    
    best_mode16 = DC_PRED_16;
    for(mode16 = 0; mode16 < max_mode; mode16++)
    {
      if (input->IntraDisableInterOnly == 0 || img->type != I_SLICE)
      {
        if (input->Intra16x16ParDisable && (mode16==VERT_PRED_16||mode16==HOR_PRED_16))
          continue;
        
        if (input->Intra16x16PlaneDisable && mode16==PLANE_16)
          continue;
      }
      
      //check if there are neighbours to predict from
      if ((mode16==VERT_PRED_16 && !up_avail  ) || 
        (mode16==HOR_PRED_16  && !left_avail) || 
        (mode16==PLANE_16     && (!left_avail || !up_avail || !left_up_avail)))
      {
        continue; // edge, do nothing
      }
      
      if(input->symbol_mode == UVLC)
      {    
        currMB->cbp = dct_luma_16x16 (mode16);
      }   // if UVLC coding
      else
      {
        currMB->cbp = klt_luma16x16_sep_fast (mode16); 
        //===== get distortion (SSD) of 4x4 block =====
        distortion = 0;
  distortion += (int)compute_SSE(imgY_org, imgY, img->opix_y, img->pix_y, img->opix_x, img->pix_x, 16, 16);
        if(distortion > min_rdcost)
        {
          continue;
        }

        store_coding_state (cs_i16);
        
        // rate for MB_TYPE 
        img->i16offset = I16Offset  (currMB->cbp, mode16);
        currMB->mb_type = I16MB; 
        currSE->value1  = MBType2Value (currMB);
        currSE->value2  = 0;
        currSE->type    = SE_MBTYPE;
        
        if (input->symbol_mode == UVLC)  
          currSE->mapping = ue_linfo;
        else
          currSE->writing = writeMB_typeInfo_CABAC;
        dataPart = &(currSlice->partArr[partMap[SE_MBTYPE]]);
        dataPart->writeSyntaxElement( currSE, dataPart);
        rate                = currSE->len;
        currSE++;
        currMB->currSEnr++;
        
        //===== RATE for LUMINANCE COEFFICIENTS =====
#ifdef MB32X32
        //rate += writeCBPandLumaCoeff(1,0,0);
        rate += writeCBP_Transform_size (1, 0);
        rate += writeDeltaQP            (1, img->mb_data[img->current_mb_nr].cbp);
        rate += writeLumaCoeff          (1, 0);
#else
//        rate += writeCBPandLumaCoeff(1);
        rate += writeCBP_Transform_size (1, 0);
        rate += writeDeltaQP            (1, img->mb_data[img->current_mb_nr].cbp);
        rate += writeLumaCoeff          (1, 0);
#endif
        reset_coding_state (cs_i16);
        
        rdcost = (double)distortion + lambda*(double)rate;
        if(rdcost < min_rdcost)
        {
          best_mode16 = mode16;
          min_rdcost = rdcost;
          best_cbp = currMB->cbp;
          memcpy(bestCofAC16[0], img->cofAC16[0], sizeof(int)*16*17);
          memcpy(bestCofAC16[1], img->cofAC16[1], sizeof(int)*16*17);
          for(j = 0; j < MB_BLOCK_SIZE; j++)
            memcpy(bestRec[j], &imgY[img->pix_y+j][img->pix_x], sizeof(imgpel)*MB_BLOCK_SIZE);
        }
      }
    }
    
    *i16mode = best_mode16;
  }
  
  currMB->cbp = best_cbp;
  if(currMB->cbp)
    currMB->cbp = 15;
  memcpy(img->cofAC16[0], bestCofAC16[0], sizeof(int)*16*17);
  memcpy(img->cofAC16[1], bestCofAC16[1], sizeof(int)*16*17);
  for(y = 0; y < MB_BLOCK_SIZE; y++)
  {
    memcpy(&imgY[img->pix_y+y][img->pix_x], &bestRec[y][0], MB_BLOCK_SIZE*sizeof(imgpel));
  }
}
#endif

/*!
*************************************************************************************
* \brief
*    Intra 16x16 mode decision
*************************************************************************************
*/
void Intra16x16_Mode_Decision (Macroblock* currMB, int* i16mode)
{
  intrapred_luma_16x16 ();   /* make intra pred for all 4 new modes */
  find_sad_16x16 (i16mode);   /* get best new intra mode */
  currMB->cbp = dct_luma_16x16 (*i16mode);
}


/*!
*************************************************************************************
* \brief
*    Sets Coefficients and reconstruction for an 8x8 block
*************************************************************************************
*/
void SetCoeffAndReconstruction8x8 (Macroblock* currMB, short QPcount)
{
  int block, k, j, i;
  int cur_ref[2];
  
  //============= MIXED TRANSFORM SIZES FOR 8x8 PARTITION ==============
  //--------------------------------------------------------------------
  int l;
  int bframe = img->type==B_SLICE; 
  
  if (currMB->luma_transform_size_8x8_flag)
  {
    
    //============= set mode and ref. frames ==============
      currMB->b8mode[0]   = tr8x8[QPcount].part8x8mode[0];
      currMB->b8mode[1]   = tr8x8[QPcount].part8x8mode[1];
      currMB->b8mode[2]   = tr8x8[QPcount].part8x8mode[2];
      currMB->b8mode[3]   = tr8x8[QPcount].part8x8mode[3];

      currMB->b8pdir[0]   = tr8x8[QPcount].part8x8pdir[0];
      currMB->b8pdir[1]   = tr8x8[QPcount].part8x8pdir[1];
      currMB->b8pdir[2]   = tr8x8[QPcount].part8x8pdir[2];
      currMB->b8pdir[3]   = tr8x8[QPcount].part8x8pdir[3];
    
    if (bframe)
    {
      for (j = 0;j<4;j++)
        for (i = 0;i<4;i++)
        {
          k = 2*(j >> 1)+(i >> 1);
          l = 2*(j & 0x01)+(i & 0x01);
          enc_picture->ref_idx[LIST_0][img->block_y+j][img->block_x+i] = ((currMB->b8pdir[k] & 0x01) == 0) ? tr8x8[QPcount].part8x8fwref[k] : - 1;
          enc_picture->ref_idx[LIST_1][img->block_y+j][img->block_x+i] = (currMB->b8pdir[k] > 0) ? tr8x8[QPcount].part8x8bwref[k] : - 1;
        }
    }
    else
    {
      for (j = 0;j<4;j++)
        for (i = 0;i<4;i++)
        {
          k = 2*(j >> 1)+(i >> 1);
          l = 2*(j & 0x01)+(i & 0x01);
          enc_picture->ref_idx[LIST_0][img->block_y+j][img->block_x+i] = tr8x8[QPcount].part8x8fwref[k];
        }
    }
    
    
    for (j = img->block_y;j<img->block_y + BLOCK_MULTIPLE;j++)
    {
      for (i = img->block_x;i<img->block_x + BLOCK_MULTIPLE;i++)
      {
        cur_ref[LIST_0] = (int) enc_picture->ref_idx[LIST_0][j][i];
        
        enc_picture->ref_pic_id [LIST_0][j][i] =(cur_ref[LIST_0]>=0 
          ? enc_picture->ref_pic_num[LIST_0 + currMB->list_offset][cur_ref[LIST_0]]
          : -1);
      }
    }
    
    if (bframe)
    {
      for (j = img->block_y; j < img->block_y + BLOCK_MULTIPLE; j++)
      {
        for (i = img->block_x;i<img->block_x + BLOCK_MULTIPLE;i++)
        {
          cur_ref[LIST_1] = (int) enc_picture->ref_idx[LIST_1][j][i];
          
          enc_picture->ref_pic_id [LIST_1][j][i] = (cur_ref[LIST_1]>=0 
            ? enc_picture->ref_pic_num[LIST_1 + currMB->list_offset][cur_ref[LIST_1]]
            : -1);
        }
        
      }
    }
    
    //====== set the mv's for 8x8 partition with transform size 8x8 ======
    //set new mv data for 8x8 transform
    RestoreMV8x8(0, QPcount);

  //==== motion vectors =====
  SetMotionVectorsMB (currMB, bframe);  

    //============= get pre-calculated data ==============
    //restore coefficients from 8x8 transform
    
    for (block = 0; block<4; block++)
    {
      for (k = 0; k<4; k++)
        for (j = 0; j<2; j++)
          memcpy (img->cofAC[block][k][j],tr8x8[QPcount].cofAC8x8[block][k][j], 65 * sizeof(int));            
    }     
    //restore reconstruction 
    if (tr8x8[QPcount].cnt_nonz_8x8 <= _LUMA_8x8_COEFF_COST_ && 
      ((img->qp + img->bitdepth_luma_qp_scale)!=0 || img->lossless_qpprime_flag==0))
    {
      currMB->cbp     = 0;
      currMB->cbp_blk = 0;
      for (j = 0; j < MB_BLOCK_SIZE; j++)
        {
          memcpy(&enc_picture->imgY[img->pix_y+j][img->pix_x], tr8x8[QPcount].mpr8x8[j], MB_BLOCK_SIZE * sizeof(imgpel));
        }
    }
    else
    {
      currMB->cbp     = tr8x8[QPcount].cbp8x8;     //      cbp8_8x8ts;
      currMB->cbp_blk = tr8x8[QPcount].cbp_blk8x8; //      cbp_blk8_8x8ts;
     
        for (j = 0; j < MB_BLOCK_SIZE; j++)
        {
          memcpy (&enc_picture->imgY[img->pix_y+j][img->pix_x],tr8x8[QPcount].rec_mbY8x8[j], MB_BLOCK_SIZE * sizeof(imgpel));   
        }
    }
  }
  else
  {
      currMB->b8mode[0]   = tr4x4[QPcount].part8x8mode[0];
      currMB->b8mode[1]   = tr4x4[QPcount].part8x8mode[1];
      currMB->b8mode[2]   = tr4x4[QPcount].part8x8mode[2];
      currMB->b8mode[3]   = tr4x4[QPcount].part8x8mode[3];

      currMB->b8pdir[0]   = tr4x4[QPcount].part8x8pdir[0];
      currMB->b8pdir[1]   = tr4x4[QPcount].part8x8pdir[1];
      currMB->b8pdir[2]   = tr4x4[QPcount].part8x8pdir[2];
      currMB->b8pdir[3]   = tr4x4[QPcount].part8x8pdir[3];

    if (bframe)
    {
      for (j = 0;j<4;j++)
        for (i = 0;i<4;i++)
        {
          k = 2*(j >> 1)+(i >> 1);
          l = 2*(j & 0x01)+(i & 0x01);
          enc_picture->ref_idx[LIST_0][img->block_y+j][img->block_x+i] = ((currMB->b8pdir[k] & 0x01) == 0) ? tr4x4[QPcount].part8x8fwref[k] : - 1;
          enc_picture->ref_idx[LIST_1][img->block_y+j][img->block_x+i] = (currMB->b8pdir[k] > 0) ? tr4x4[QPcount].part8x8bwref[k] : - 1;
        }
    }
    else
    {
      for (j = 0;j<4;j++)
        for (i = 0;i<4;i++)
        {
          k = 2*(j >> 1)+(i >> 1);
          l = 2*(j & 0x01)+(i & 0x01);
          enc_picture->ref_idx[LIST_0][img->block_y+j][img->block_x+i] = tr4x4[QPcount].part8x8fwref[k];
        }
    }

    for (j = img->block_y;j<img->block_y + BLOCK_MULTIPLE;j++)
    {
      for (i = img->block_x;i<img->block_x + BLOCK_MULTIPLE;i++)
      {
        cur_ref[LIST_0] = (int) enc_picture->ref_idx[LIST_0][j][i];
        
        enc_picture->ref_pic_id [LIST_0][j][i] =(cur_ref[LIST_0]>=0 
          ? enc_picture->ref_pic_num[LIST_0 + currMB->list_offset][cur_ref[LIST_0]]
          : -1);
      }
    }
    
    if (bframe)
    {
      for (j = img->block_y; j < img->block_y + BLOCK_MULTIPLE; j++)
      {
        for (i = img->block_x;i<img->block_x + BLOCK_MULTIPLE;i++)
        {
          cur_ref[LIST_1] = (int) enc_picture->ref_idx[LIST_1][j][i];
          
          enc_picture->ref_pic_id [LIST_1][j][i] = (cur_ref[LIST_1]>=0 
            ? enc_picture->ref_pic_num[LIST_1 + currMB->list_offset][cur_ref[LIST_1]]
            : -1);
        }
      }
    }

    //====== set the mv's for 8x8 partition with transform size 4x4 ======
    //set new mv data for 4x4 transform
    RestoreMV8x8(1, QPcount);

  //==== motion vectors =====
  SetMotionVectorsMB (currMB, bframe);
    //============= get pre-calculated data ==============
    //---------------------------------------------------
    //--- restore coefficients ---
    for (block = 0; block<4+img->num_blk8x8_uv; block++)
    {
      for (k = 0; k<4; k++)
        for (j = 0; j<2; j++)
          memcpy (img->cofAC[block][k][j],tr4x4[QPcount].cofAC8x8[block][k][j], 65 * sizeof(int));            
    }
    
    if (tr4x4[QPcount].cnt_nonz_8x8<=5 && ((img->qp + img->bitdepth_luma_qp_scale)!=0 || img->lossless_qpprime_flag==0))
    {
      currMB->cbp     = 0;
      currMB->cbp_blk = 0;
      for (j = 0; j < MB_BLOCK_SIZE; j++)
        {
          memcpy (&enc_picture->imgY[img->pix_y+j][img->pix_x],tr4x4[QPcount].mpr8x8[j], MB_BLOCK_SIZE * sizeof(imgpel));            
        }
    }
    else
    {
      currMB->cbp     = tr4x4[QPcount].cbp8x8;     //     cbp8x8;
      currMB->cbp_blk = tr4x4[QPcount].cbp_blk8x8; //     cbp_blk8x8;
          for (j = 0; j < MB_BLOCK_SIZE; j++)
          {
            memcpy (&enc_picture->imgY[img->pix_y+j][img->pix_x],tr4x4[QPcount].rec_mbY8x8[j], MB_BLOCK_SIZE * sizeof(imgpel));            
          }
    }
  }
}


/*!
*************************************************************************************
* \brief
*    Sets motion vectors for a macroblock
*************************************************************************************
*/
void SetMotionVectorsMB (Macroblock* currMB, int bframe)
{
  int i, j, k, l, mode8, pdir8, ref, by, bx;
  short ******all_mv  = img->all_mv;
  int  bw_ref;
  int jdiv, jmod;
  
  if (!bframe)
  {
    for (j = 0; j<4; j++)
    {
      jmod = j & 0x01;
      jdiv = j >>   1;
      by    = img->block_y+j;
      for (i = 0; i<4; i++)
      {
        mode8 = currMB->b8mode[k=2*jdiv+(i>>1)];
        l     = 2*jmod + (i & 0x01);
        
        bx   = img->block_x+i;
        
        pdir8 = currMB->b8pdir[k];
        ref    = enc_picture->ref_idx[LIST_0][by][bx];
        
        if (pdir8>=0) 
        {
          enc_picture->mv[LIST_0][by][bx][0] = all_mv [j][i][LIST_0][ ref][mode8][0];
          enc_picture->mv[LIST_0][by][bx][1] = all_mv [j][i][LIST_0][ ref][mode8][1];
        }
        else
        {
          enc_picture->mv[LIST_0][by][bx][0] = 0;
          enc_picture->mv[LIST_0][by][bx][1] = 0;
        }
      }
    }
  }
  else
  {
    for (j = 0; j<4; j++)
    {
      jmod = j & 0x01;
      jdiv = j >>   1;
      by    = img->block_y+j;
      for (i = 0; i<4; i++)
      {
        mode8 = currMB->b8mode[k=2*jdiv+(i>>1)];
        l     = 2*jmod + (i & 0x01);
        
        bx    = img->block_x+i;
        
        pdir8 = currMB->b8pdir[k];
        ref    = enc_picture->ref_idx[LIST_0][by][bx];
        bw_ref = enc_picture->ref_idx[LIST_1][by][bx];
        
        if (currMB->bi_pred_me && (pdir8 == 2) && currMB->mb_type==1)
        {
          all_mv  = currMB->bi_pred_me == 1 ? img->bipred_mv1 : img->bipred_mv2;
          ref = 0;
          bw_ref = 0;
        }
        
        if (pdir8==-1) // intra
        {
          enc_picture->mv[LIST_0][by][bx][0] = 0;
          enc_picture->mv[LIST_0][by][bx][1] = 0;
          enc_picture->mv[LIST_1][by][bx][0] = 0;
          enc_picture->mv[LIST_1][by][bx][1] = 0;
        }
        else if (pdir8==0) // list 0
        {
          enc_picture->mv[LIST_0][by][bx][0] = all_mv [j][i][LIST_0][ ref][mode8][0];
          enc_picture->mv[LIST_0][by][bx][1] = all_mv [j][i][LIST_0][ ref][mode8][1];
          enc_picture->mv[LIST_1][by][bx][0] = 0;
          enc_picture->mv[LIST_1][by][bx][1] = 0;
          enc_picture->ref_idx[LIST_1][by][bx] = -1;
        }
        else if (pdir8==1) // list 1
        {
          enc_picture->mv[LIST_0][by][bx][0] = 0;
          enc_picture->mv[LIST_0][by][bx][1] = 0;          
          enc_picture->ref_idx[LIST_0][by][bx] = -1;
          enc_picture->mv[LIST_1][by][bx][0] = all_mv [j][i][LIST_1][bw_ref][mode8][0];
          enc_picture->mv[LIST_1][by][bx][1] = all_mv [j][i][LIST_1][bw_ref][mode8][1];
        }
        else if (pdir8==2) // bipredictive
        {
          enc_picture->mv[LIST_0][by][bx][0] = all_mv [j][i][LIST_0][ ref][mode8][0];
          enc_picture->mv[LIST_0][by][bx][1] = all_mv [j][i][LIST_0][ ref][mode8][1];                   
          enc_picture->mv[LIST_1][by][bx][0] = all_mv [j][i][LIST_1][bw_ref][mode8][0];
          enc_picture->mv[LIST_1][by][bx][1] = all_mv [j][i][LIST_1][bw_ref][mode8][1];
        }
        else
        {
          error("invalid direction mode", 255);
        }
        
      }
    }
  }
}



/*!
*************************************************************************************
* \brief
*    R-D Cost for a macroblock
*************************************************************************************
*/
int RDCost_for_macroblocks (double   lambda,       // <-- lagrange multiplier
                            int      mode,         // <-- modus (0-COPY/DIRECT, 1-16x16, 2-16x8, 3-8x16, 4-8x8(+), 5-Intra4x4, 6-Intra16x16)
                            int      *mode_idx,
                            double*  min_rdcost,   // <-> minimum rate-distortion cost
                            double*  min_rate,     // --> bitrate of mode which has minimum rate-distortion cost. 
                            int i16mode,
                            short QPcount)
{
  
  int         i, j; 
  int         rate = 0, coeff_rate = 0;
  int64       distortion = 0;
  double      rdcost;
  Macroblock  *currMB   = &img->mb_data[img->current_mb_nr];
//  Macroblock  *prevMB   = img->current_mb_nr ? &img->mb_data[img->current_mb_nr-1] : NULL;
  int         bframe    = (img->type==B_SLICE);
  int         use_of_cc =  (img->type!=I_SLICE &&  input->symbol_mode!=CABAC);
  int         dummy;
  int         cr_cbp = 0;
  int      return_val = 0;
  int tmp_cc;
  int cc_rate;
 
#ifdef MV_COMPETITION
  short iter, nbiter;
  short ******all_mv = img->all_mv;
#endif

  //=====
  //=====  SET REFERENCE FRAMES AND BLOCK MODES
  //=====
  if(currMB->geo_mode >= 0)
    currMB->mb_type = mode;
  else
  SetModesAndRefframeForBlocks (mode);
  
#ifdef MV_COMPETITION
  if ((input->mv_competition > 0) && (mode == 0) && (img->type==P_SLICE))
    nbiter = mv_comp.nb_mode_for_skip + 1;
  else
    nbiter = 1;
  
  // Let's iterate on each possible mode for the skip
  for (iter= 0; iter < nbiter; iter++)
  {
    if ((input->mv_competition > 0) && (mode == 0) && (img->type==P_SLICE))  
    {
      distortion = 0;
      coeff_rate = 0;
      rate = 0;
      cr_cbp = 0;
    }
    
#endif

    //=====
    //=====  GET COEFFICIENTS, RECONSTRUCTIONS, CBP
    //=====
    if (bframe && mode==0)
    {
      int block_x=img->pix_x>>2;
      int block_y=img->pix_y>>2;    
      for (j = block_y;j< block_y + 4;j++)
        for (i = block_x;i<block_x + 4;i++)
          if (direct_pdir[j][i] < 0)
            return 0;
    }
    
#ifdef MV_COMPETITION
    if ((input->mv_competition > 0) && (mode == 0) && (img->type==P_SLICE))
    {
      // The last iteration consist of re-launching with the bestmode,
      // to set needed variables for the JM
      if (iter == (nbiter - 1))
        mv_comp.current_predictor_for_skip = currMB->best_predictor_for_skip;
      else
      {
        mv_comp.current_predictor_for_skip = iter;
        if ((mv_comp.mv_pred_skip[mv_comp.current_predictor_for_skip][0] == NOT_AVAILABLE) || (mv_comp.mv_pred_skip[mv_comp.current_predictor_for_skip][1] == NOT_AVAILABLE))
        {
          continue;
        }
      }
      
      LumaResidualCoding (0, 0, mode, *mode_idx);
      if (img->yuv_format!=YUV400)
        ChromaResidualCoding (0, 0, mode, *mode_idx, &dummy);
    }
    else //if((input->mv_competition > 0) && (mode == 0) && (img->type==P_SLICE))
    {
#endif
      if(currMB->geo_mode != -1)
      {
        LumaResidualCoding_geo (0, 0, currMB->geo_mode, *mode_idx, 0);

      }
      else
      if (mode<P8x8)
      {
        LumaResidualCoding (0, 0, mode, *mode_idx);
#if !P_DIRECT        
        if(mode==0 && currMB->cbp!=0 && (img->type != B_SLICE || img->NoResidueDirect==1)) 
          return 0;
#endif
      }
      else if (mode==P8x8)
      {
		    SetCoeffAndReconstruction8x8 (currMB, QPcount);
      }
      else if (mode==I4MB)
      {
#if defined RDO_Q
        if(input->UseRDO_Q) 
        {
//          if(isfirstQP)
//          {
            if (currMB->c_ipred_mode == DC_PRED_8)    //only store during 1st chroma Pred, for other chroma Pred restore it.
              currMB->cbp = Mode_Decision_for_Intra4x4Macroblock (lambda, &dummy);
            set_I4MB_parameters();
          //}
          //else
          //{
          //  restore_macroblock_parameters_intra(mode); 
          //}
        }
        else
#endif
          currMB->cbp = Mode_Decision_for_Intra4x4Macroblock (lambda, &dummy);
      }
      else if (mode == I16MB)
      {
      
#if defined RDO_Q
        if(input->UseRDO_Q)
        {
          //if(isfirstQP)
          //{
              //only store during 1st chroma Pred, for other chroma Pred restore it.
              if (currMB->c_ipred_mode == DC_PRED_8)
              {
#ifdef USE_INTRA_MDDT         
                if(input->symbol_mode == CABAC && input->UseIntraMDDT)
                  Intra16x16_Mode_Decision_RDopt(currMB, &i16mode, lambda);
                else
#endif
                  Intra16x16_Mode_Decision(currMB, &i16mode);
              }
              set_I16MB_parameters(&i16mode);
          //}
          //else
          //{
          //  restore_macroblock_parameters_intra(mode);
          //}
        }
        else
#endif
        {
#ifdef USE_INTRA_MDDT        
                if(input->symbol_mode == CABAC && input->UseIntraMDDT)
                  Intra16x16_Mode_Decision_RDopt(currMB, &i16mode, lambda);
                else
#endif
                  Intra16x16_Mode_Decision(currMB, &i16mode);
        }
      }
      else if (mode==I8MB)
      {
#if defined RDO_Q
        if(input->UseRDO_Q) //speed up intra coding in RDOQ for MB size 16x16, not for 32x32 or higher
        {
          //if(isfirstQP)
          //{
            //only store during 1st chroma Pred, for other chroma Pred restore it.
            if (currMB->c_ipred_mode == DC_PRED_8)
              currMB->cbp = Mode_Decision_for_new_Intra8x8Macroblock(lambda, &dummy);    
            set_I8MB_parameters();
          //}
          //else
          //{
          //  restore_macroblock_parameters_intra(mode); 
          //}
        }
        else
#endif
          currMB->cbp = Mode_Decision_for_new_Intra8x8Macroblock(lambda, &dummy);
      }
 
#if defined RDO_Q
      if(input->UseRDO_Q)
      {
        if(1)
//        if (img->type == I_SLICE  || !IS_INTRA_MODE(mode) || (IS_INTRA_MODE(mode) && IS_INTER_SLICE(img->type)))
        { 
          img->i16offset = 0;
          dummy = 0;
          if (img->yuv_format!=YUV400)
          {
            if(currMB->geo_mode != -1)
            {
              ChromaResidualCoding_geo (0, 0, currMB->geo_mode, *mode_idx, 0);
            }
            else
              ChromaResidualCoding (0, 0 , mode, *mode_idx, &dummy);
          }

          if (mode==I16MB)     
            img->i16offset = I16Offset  (currMB->cbp, i16mode);
          
          if(IS_INTER_SLICE(img->type) && IS_INTRA_MODE(mode))
            store_macroblock_parameters_intra(mode); 
        }
      }
      else
#endif
      {
      img->i16offset = 0;
      dummy = 0;
      if (img->yuv_format!=YUV400)
      {
        if(currMB->geo_mode != -1)
          ChromaResidualCoding_geo (0, 0 , currMB->geo_mode, *mode_idx, 0);
        else
          ChromaResidualCoding (0, 0 , mode, *mode_idx, &dummy);
      }
      if (mode==I16MB)     
        img->i16offset = I16Offset  (currMB->cbp, i16mode);
      }


#ifdef MV_COMPETITION
    }  //else end //if((input->mv_competition > 0) && (mode == 0) && (img->type==P_SLICE))
#endif

      {    
      //=====   GET DISTORTION
  distortion += compute_SSE(imgY_org, enc_picture->imgY, img->opix_y, img->pix_y, img->opix_x, img->opix_x, MB_BLOCK_SIZE, MB_BLOCK_SIZE);
  distortion += compute_SSE(imgUV_org[0], enc_picture->imgUV[0], img->opix_c_y, img->pix_c_y, img->opix_c_x, img->opix_c_x, img->mb_cr_size_y, img->mb_cr_size_x);
  distortion += compute_SSE(imgUV_org[1], enc_picture->imgUV[1], img->opix_c_y, img->pix_c_y, img->opix_c_x, img->opix_c_x, img->mb_cr_size_y, img->mb_cr_size_x);

    if (distortion > *min_rdcost)
      rdcost = (double)distortion;
    else
    {
      //=====   S T O R E   C O D I N G   S T A T E   =====
      //---------------------------------------------------
      store_coding_state (cs_cm);
      
      //=====
      //=====   GET RATE
      //=====
      //----- macroblock header -----
      if (use_of_cc)  //use_of_cc =  (img->type!=I_SLICE &&  input->symbol_mode!=CABAC);
      {
        if (currMB->mb_type!=0 || (bframe && currMB->cbp!=0))
        {
          // cod counter and macroblock mode are written ==> do not consider code counter
          tmp_cc = img->cod_counter;
          rate   = writeMBLayer (1, &coeff_rate);
          ue_linfo (tmp_cc, dummy, &cc_rate, &dummy);
          rate  -= cc_rate;
          img->cod_counter = tmp_cc;
        }
        else
        {
          // cod counter is just increased  ==> get additional rate
          ue_linfo (img->cod_counter+1, dummy, &rate,    &dummy);
          ue_linfo (img->cod_counter,   dummy, &cc_rate, &dummy);
          rate -= cc_rate;
        }
      }
      else
      {

        rate = writeMBLayer (1, &coeff_rate);
      }
      
      //=====   R E S T O R E   C O D I N G   S T A T E   =====
      //-------------------------------------------------------
      reset_coding_state (cs_cm);

      rdcost = (double)distortion + lambda * max(0.5,(double)rate);
    }

  }  

    if (rdcost < *min_rdcost)
    {                
        //=====   U P D A T E   M I N I M U M   C O S T   =====
        //-----------------------------------------------------
        *min_rdcost = rdcost;
        *min_rate = lambda * (double)coeff_rate;

#ifdef MV_COMPETITION      
          if ((input->mv_competition > 0) && (mode == 0) && (img->type==P_SLICE))
          {
            int by, bx;
            
            currMB->best_predictor_for_skip = mv_comp.current_predictor_for_skip;
            
            for (by = 0;by < 4;by++)
              for (bx = 0;bx < 4;bx++)
              {
                all_mv [by][bx][0][0][0][0] = mv_comp.mv_pred_skip[currMB->best_predictor_for_skip][0];
                all_mv [by][bx][0][0][0][1] = mv_comp.mv_pred_skip[currMB->best_predictor_for_skip][1];
              }
          }
#endif
          return_val = 1;
    }
    else 
    {
#ifdef MV_COMPETITION
      if (!((input->mv_competition > 0) && (mode == 0) && (img->type==P_SLICE)))
      {
#endif
#if FASTMODE
        // Reordering RDCost comparison order of mode 0 and mode 1 in P_SLICE
        // if RDcost of mode 0 and mode 1 is same, we choose best_mode is 0
        // This might not always be good since mode 0 is more biased towards rate than quality.
        if((img->type!=P_SLICE || mode != 0 || rdcost != *min_rdcost) || input->ProfileIDC>=FREXT_HP)
#endif
          return_val = 0;
#ifdef MV_COMPETITION
      }
#endif
    }
  
#ifdef MV_COMPETITION
  }  //for (iter= 0; iter < nbiter; iter++)
#endif 

  return (return_val);
}

#ifdef RDO_Q
/*!
****************************************************************************
* \brief
*    save encoding information for intra MB
****************************************************************************
*/
void store_macroblock_parameters_intra (int mode)
{
  int        i, j, k;
  int        num_blk8x8 = 4 + img->num_blk8x8_uv;
  Macroblock *currMB  = &img->mb_data[img->current_mb_nr];
  //																		  I4 I16	    I8  PCM
  int        intra_mode_mapping_table[] = {-1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 1, -1, -1, 2, 3};
  int        luma_mode = intra_mode_mapping_table[mode];
  int        chroma_mode = currMB->c_ipred_mode; 
  
  assert(luma_mode != -1);
  
  if(currMB->mb_type == I8MB)
  {
    for(j = 0; j < BLOCK_MULTIPLE; j++)
    {
      //memcpy(b8_ipredmode8x8_intra[luma_mode][chroma_mode][j], &img->ipredmode[img->block_y+j][img->block_x],BLOCK_MULTIPLE * sizeof(char));
      memcpy(b8_ipredmode8x8_intra[luma_mode][chroma_mode][j], &img->ipredmode8x8[img->block_y+j][img->block_x], BLOCK_MULTIPLE * sizeof(char));
    }
    memcpy(b8_ipredmode8x8_mb_intra[luma_mode][chroma_mode], currMB->intra_pred_modes8x8, BLOCK_MULTIPLE*BLOCK_MULTIPLE*sizeof(char));
  }
  else if (currMB->mb_type == I4MB)
  {
    for(j = 0; j < BLOCK_MULTIPLE; j++)
      memcpy(&b4_ipredmode_intra[luma_mode][chroma_mode][BLOCK_MULTIPLE * j], &img->ipredmode[img->block_y + j][img->block_x],BLOCK_MULTIPLE * sizeof(char));
    memcpy(b4_ipredmode_mb_intra[luma_mode][chroma_mode], currMB->intra_pred_modes, BLOCK_MULTIPLE*BLOCK_MULTIPLE*sizeof(char));
  }
  
  //--- reconstructed blocks ----
  for (j = 0; j < MB_BLOCK_SIZE; j++)
  {
    memcpy(rec_mbY_intra[luma_mode][chroma_mode][j],&enc_picture->imgY[img->pix_y+j][img->pix_x], MB_BLOCK_SIZE * sizeof(imgpel));
  }
  if (img->yuv_format != YUV400)
  {
    for (j = 0; j<img->mb_cr_size_y; j++)
    {
      memcpy(rec_mbU_intra[luma_mode][chroma_mode][j],&enc_picture->imgUV[0][img->pix_c_y+j][img->pix_c_x], img->mb_cr_size_x * sizeof(imgpel));
      memcpy(rec_mbV_intra[luma_mode][chroma_mode][j],&enc_picture->imgUV[1][img->pix_c_y+j][img->pix_c_x], img->mb_cr_size_x * sizeof(imgpel));
    }
  }
  
  
  //--- coeff, cbp, kac ---
  for(i = 0; i < num_blk8x8; i++)
    for(j = 0; j < 4; j++)
      for(k = 0; k < 2; k++)
        memcpy(cofAC_intra[luma_mode][chroma_mode][i][j][k], img->cofAC[i][j][k], sizeof(int)*65);

      for(i = 0; i < 3; i++)
        for(j = 0; j < 2; j++)
          memcpy(cofDC_intra[luma_mode][chroma_mode][i][j], img->cofDC[i][j], sizeof(int)*65);

        cbp_intra[luma_mode][chroma_mode]     = currMB->cbp;
        cbp_blk_intra[luma_mode][chroma_mode] = currMB->cbp_blk;
        i16offset_intra[luma_mode][chroma_mode] = img->i16offset;
}

/*!
****************************************************************************
* \brief
*    assign saved encoding information for intra MB
****************************************************************************
*/
void restore_macroblock_parameters_intra (int mode)
{
  int  i, j, k;
  int num_blk8x8 = 4 + img->num_blk8x8_uv;
  Macroblock  *currMB  = &img->mb_data[img->current_mb_nr];
  int         intra_mode_mapping_table[] = 
  {-1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 1, -1, -1, 2, 3};
  int         luma_mode = intra_mode_mapping_table[mode];
  int        chroma_mode = currMB->c_ipred_mode; 
  int         bframe   = (img->type==B_SLICE);  
  int        block_x, block_y;  
  
  assert(luma_mode != -1);
  
  //===== reconstruction values =====
  for (j = 0; j < MB_BLOCK_SIZE; j++)
  {
    memcpy(&enc_picture->imgY[img->pix_y+j][img->pix_x],rec_mbY_intra[luma_mode][chroma_mode][j], MB_BLOCK_SIZE * sizeof(imgpel));
  }
  if (img->yuv_format != YUV400)
  {
    for (j = 0; j<img->mb_cr_size_y; j++)
    {
      memcpy(&enc_picture->imgUV[0][img->pix_c_y+j][img->pix_c_x],rec_mbU_intra[luma_mode][chroma_mode][j], img->mb_cr_size_x * sizeof(imgpel));
      memcpy(&enc_picture->imgUV[1][img->pix_c_y+j][img->pix_c_x],rec_mbV_intra[luma_mode][chroma_mode][j], img->mb_cr_size_x * sizeof(imgpel));
    }
  }
  
  //===== coefficients and cbp =====
  for(i = 0; i < num_blk8x8; i++)
    for(j = 0; j < 4; j++)
      for(k = 0; k < 2; k++)
        memcpy(img->cofAC[i][j][k], cofAC_intra[luma_mode][chroma_mode][i][j][k], sizeof(int)*65);

      for(i = 0; i < 3; i++)
        for(j = 0; j < 2; j++)
          memcpy(img->cofDC[i][j], cofDC_intra[luma_mode][chroma_mode][i][j], sizeof(int)*65);

        currMB->cbp      = cbp_intra[luma_mode][chroma_mode];
        currMB->cbp_blk  = cbp_blk_intra[luma_mode][chroma_mode];
        //==== macroblock type ====
        currMB->mb_type = mode;
        img->i16offset = i16offset_intra[luma_mode][chroma_mode];
        
        //==== transform size flag ====
        currMB->luma_transform_size_8x8_flag = (currMB->mb_type == I8MB);
        
        rdopt->luma_transform_size_8x8_flag  = currMB->luma_transform_size_8x8_flag;
        
     
        if (bframe)
        {
          for (j=0; j<4; j++)
          {
            block_y = img->block_y + j;
            for (i=0; i<4; i++)
            {          
              block_x = img->block_x + i;
              k = 2*(j >> 1)+(i >> 1);
              
              // forward
              enc_picture->ref_idx    [LIST_1][block_y][block_x]    = -1;
              enc_picture->ref_pic_id [LIST_1][block_y][block_x]    = -1;
              enc_picture->mv         [LIST_1][block_y][block_x][0] = 0;
              enc_picture->mv         [LIST_1][block_y][block_x][1] = 0;
            }
          }
        }
        
        if(currMB->mb_type == I8MB)
        {
          for(j = 0; j < BLOCK_MULTIPLE; j++)
          {
            memcpy(&img->ipredmode[img->block_y+j][img->block_x],b8_ipredmode8x8_intra[luma_mode][chroma_mode][j], BLOCK_MULTIPLE * sizeof(char));
            memcpy(&img->ipredmode8x8[img->block_y+j][img->block_x], b8_ipredmode8x8_intra[luma_mode][chroma_mode][j], BLOCK_MULTIPLE * sizeof(char));
          }
          memcpy(currMB->intra_pred_modes8x8, b8_ipredmode8x8_mb_intra[luma_mode][chroma_mode], BLOCK_MULTIPLE*BLOCK_MULTIPLE*sizeof(char));
        }
        else if (currMB->mb_type == I4MB)
        {
          for(j = 0; j < BLOCK_MULTIPLE; j++)
            memcpy(&img->ipredmode[img->block_y + j][img->block_x],&b4_ipredmode_intra[luma_mode][chroma_mode][BLOCK_MULTIPLE * j], BLOCK_MULTIPLE * sizeof(char));
          memcpy(currMB->intra_pred_modes, b4_ipredmode_mb_intra[luma_mode][chroma_mode], BLOCK_MULTIPLE*BLOCK_MULTIPLE*sizeof(char));
        }
}

/*!
****************************************************************************
* \brief
*    save/assign encoding information for Intra4x4 MB
****************************************************************************
*/
void set_I4MB_parameters()
{
  Macroblock  *currMB   = &img->mb_data[img->current_mb_nr];
  int *src1, *dest1;
  imgpel *src2, *dest2, *src3, *dest3;
  int b8, b4,j, y;
  
  for (b8=0; b8<4; b8++)
  {
    for (b4=0; b4<4; b4++)
    {
      for (j=0; j<2; j++)    
      {
        if (currMB->c_ipred_mode == DC_PRED_8)
        {
          src1 = img->cofAC[b8][b4][j];
          dest1 = saved_cofAC[b8][b4][j];
        }
        else
        {
          dest1 = img->cofAC[b8][b4][j];
          src1 = saved_cofAC[b8][b4][j];
        }
        memcpy (dest1, src1, 18 * sizeof(int));
      }
    }
  }
  
  for (y=0; y<MB_BLOCK_SIZE; y++)
  {                
    if (currMB->c_ipred_mode == DC_PRED_8)
    {
      src2 = &enc_picture->imgY[img->pix_y+y][img->pix_x];
      dest2 = &saved_enc_picture_imgY[y][0];
      src3 = &img->mpr[y][0];
      dest3 = &saved_img_mpr[y][0];
    }
    else
    {
      dest2 = &enc_picture->imgY[img->pix_y+y][img->pix_x];
      src2 = &saved_enc_picture_imgY[y][0];
      dest3 = &img->mpr[y][0];
      src3 = &saved_img_mpr[y][0];
    }
    
    memcpy (dest2, src2, MB_BLOCK_SIZE * sizeof(imgpel));                
    memcpy (dest3, src3, MB_BLOCK_SIZE * sizeof(imgpel)); 
  }
  
  if(currMB->c_ipred_mode == DC_PRED_8)
  {
    saved_cbp = currMB->cbp;
  }
  else
  {
    currMB->cbp = saved_cbp;
  }
}

/*!
****************************************************************************
* \brief
*    save/assign encoding information for Intra16x16 MB
****************************************************************************
*/
void set_I16MB_parameters(int *i16mode)
{
  Macroblock  *currMB   = &img->mb_data[img->current_mb_nr];
  int *src1, *dest1;
  imgpel *src2, *dest2;
  int b8, b4,j, y;
  
  for (b8=0; b8<4; b8++)
  {
    for (b4=0; b4<4; b4++)
    {
      for (j=0; j<2; j++)    
      {
        if (currMB->c_ipred_mode == DC_PRED_8){
          src1 = img->cofAC[b8][b4][j];
          dest1 = saved_I16cofAC[b8][b4][j];
        }
        else
        {
          dest1 = img->cofAC[b8][b4][j];
          src1 = saved_I16cofAC[b8][b4][j];
        }
        memcpy (dest1, src1, 18 * sizeof(int));
      }
    }
  }
  
  for(j=0; j<2; j++)
  {
    if (currMB->c_ipred_mode == DC_PRED_8)
    {
      src1 = img->cofDC[0][j];
      dest1 = saved_cofDC[0][j];
    }
    else
    {
      dest1 = img->cofDC[0][j];
      src1 = saved_cofDC[0][j];
    }
    memcpy(dest1, src1, 18 * sizeof(int));
  }
  
  for (y=0; y<MB_BLOCK_SIZE; y++){                
    if (currMB->c_ipred_mode == DC_PRED_8){
      src2 = &enc_picture->imgY[img->pix_y+y][img->pix_x];
      dest2 = &saved_I16enc_picture_imgY[y][0];
    }
    else
    {
      dest2 = &enc_picture->imgY[img->pix_y+y][img->pix_x];
      src2 = &saved_I16enc_picture_imgY[y][0];
    }
    memcpy (dest2, src2, MB_BLOCK_SIZE * sizeof(imgpel));                
  }
  
  if(currMB->c_ipred_mode == DC_PRED_8){
    saved_I16cbp = currMB->cbp;
    saved_i16mode = *i16mode;
  }
  else
  {
    currMB->cbp = saved_I16cbp;
    *i16mode = saved_i16mode;
  }
}

/*!
****************************************************************************
* \brief
*    save/assign encoding information for Intra8x8 MB
****************************************************************************
*/
void set_I8MB_parameters()
{
  Macroblock  *currMB   = &img->mb_data[img->current_mb_nr];
  int *src1, *dest1;
  imgpel *src2, *dest2, *src3, *dest3;
  int b8, b4,j, y;
  
  for (b8=0; b8<4; b8++)
  {
    for (b4=0; b4<4; b4++)
    {
      for (j=0; j<2; j++)    
      {
        if (currMB->c_ipred_mode == DC_PRED_8){
          src1 = img->cofAC[b8][b4][j];
          dest1 = saved_I8cofAC[b8][b4][j];
        }    
        else
        {
          dest1 = img->cofAC[b8][b4][j];
          src1 = saved_I8cofAC[b8][b4][j];
        }
        memcpy (dest1, src1, 65 * sizeof(int));
      }
    }
  }
  
  for (y=0; y<MB_BLOCK_SIZE; y++){                
    if (currMB->c_ipred_mode == DC_PRED_8){
      src2 = &enc_picture->imgY[img->pix_y+y][img->pix_x];
      dest2 = &saved_I8enc_picture_imgY[y][0];
      src3 = &img->mpr[y][0];
      dest3 = &saved_I8img_mpr[y][0];
    }
    else
    {
      dest2 = &enc_picture->imgY[img->pix_y+y][img->pix_x];
      src2 = &saved_I8enc_picture_imgY[y][0];
      dest3 = &img->mpr[y][0];
      src3 = &saved_I8img_mpr[y][0];
    }
    memcpy (dest2, src2, MB_BLOCK_SIZE * sizeof(imgpel));                
    memcpy (dest3, src3, MB_BLOCK_SIZE * sizeof(imgpel)); 
  }
  
  if(currMB->c_ipred_mode == DC_PRED_8){
    
    saved_I8cbp = currMB->cbp;
  }
  else
  {
    currMB->cbp = saved_I8cbp;
  }
}

#endif

/*!
*************************************************************************************
* \brief
*    Store macroblock parameters
*************************************************************************************
*/
void store_macroblock_parameters (int mode)
{
  int  j, ****i4p, ***i3p;
  Macroblock *currMB  = &img->mb_data[img->current_mb_nr];
  int        bframe   = (img->type==B_SLICE);
  
  //--- store best mode ---
  best_mode = mode;
  best_c_imode = currMB->c_ipred_mode;
  best_i16offset = img->i16offset;
  
  // If condition is not really necessary.
  bi_pred_me = (mode == 1) ? currMB->bi_pred_me : 0;  
  
  memcpy(b8mode, currMB->b8mode, BLOCK_MULTIPLE * sizeof(int));
  memcpy(b8pdir, currMB->b8pdir, BLOCK_MULTIPLE * sizeof(int));
  
  // Residue Color Transform
  //for (k = 0, j=img->block_y; j<img->block_y+BLOCK_MULTIPLE; j++, k+=BLOCK_MULTIPLE)
  memcpy(b4_intra_pred_modes,currMB->intra_pred_modes, MB_BLOCK_PARTITIONS * sizeof(char));
  memcpy(b8_intra_pred_modes8x8,currMB->intra_pred_modes8x8, MB_BLOCK_PARTITIONS * sizeof(char));
  for (j = 0 ; j<BLOCK_MULTIPLE; j++)
  {
    memcpy(&b4_ipredmode[j * BLOCK_MULTIPLE],&img->ipredmode[img->block_y + j][img->block_x],BLOCK_MULTIPLE * sizeof(char));
    memcpy(b8_ipredmode8x8[j],&img->ipredmode8x8[img->block_y + j][img->block_x],BLOCK_MULTIPLE * sizeof(char));
  }  
  //--- reconstructed blocks ----
  for (j = 0; j < MB_BLOCK_SIZE; j++)
  {
    memcpy(rec_mbY[j],&enc_picture->imgY[img->pix_y+j][img->pix_x], MB_BLOCK_SIZE * sizeof(imgpel));
  }
 
  if (img->yuv_format != YUV400)
  {
    for (j = 0; j<img->mb_cr_size_y; j++)
    {
      memcpy(rec_mbU[j],&enc_picture->imgUV[0][img->pix_c_y+j][img->pix_c_x], img->mb_cr_size_x * sizeof(imgpel));
      memcpy(rec_mbV[j],&enc_picture->imgUV[1][img->pix_c_y+j][img->pix_c_x], img->mb_cr_size_x * sizeof(imgpel));
    }
  }

 
  //--- coeff, cbp, kac ---
#if P_DIRECT    //lets not make the cbp=0 for mode=0 in Pslice..coz it can be Pdirect also...
  if(1)
#else
  if (mode || bframe)
#endif
  {
    i4p=cofAC; cofAC=img->cofAC; img->cofAC=i4p;
    i3p=cofDC; cofDC=img->cofDC; img->cofDC=i3p;
    cbp     = currMB->cbp;
    cbp_blk = currMB->cbp_blk;
  }
  else
  {
    cbp_blk = cbp = 0;
  }
  
  //--- store transform size ---
  luma_transform_size_8x8_flag = currMB->luma_transform_size_8x8_flag;
  
  for (j = 0; j<4; j++)
    memcpy(frefframe[j],&enc_picture->ref_idx[LIST_0][img->block_y+j][img->block_x], BLOCK_MULTIPLE * sizeof(char));
  
  if (bframe)
  {
    for (j = 0; j<4; j++)
      memcpy(brefframe[j],&enc_picture->ref_idx[LIST_1][img->block_y+j][img->block_x], BLOCK_MULTIPLE * sizeof(char));
  }  
  
 
#ifdef USE_INTRA_MDDT
  if(input->UseIntraMDDT)
  {
    memcpy(bestCofAC16[0], img->cofAC16[0], sizeof(int)*16*17);
    memcpy(bestCofAC16[1], img->cofAC16[1], sizeof(int)*16*17);
  }
#endif 

  memcpy(bestCofAC16x16[0],    img->cofAC16x16[0],    sizeof(int)*257);
  memcpy(bestCofAC16x16[1],    img->cofAC16x16[1],    sizeof(int)*257);
  memcpy(bestCofAC16x8 [0][0], img->cofAC16x8 [0][0], sizeof(int)*129);
  memcpy(bestCofAC16x8 [0][1], img->cofAC16x8 [0][1], sizeof(int)*129);
  memcpy(bestCofAC16x8 [1][0], img->cofAC16x8 [1][0], sizeof(int)*129);
  memcpy(bestCofAC16x8 [1][1], img->cofAC16x8 [1][1], sizeof(int)*129);
  memcpy(bestCofAC8x16 [0][0], img->cofAC8x16 [0][0], sizeof(int)*129);
  memcpy(bestCofAC8x16 [0][1], img->cofAC8x16 [0][1], sizeof(int)*129);
  memcpy(bestCofAC8x16 [1][0], img->cofAC8x16 [1][0], sizeof(int)*129);
  memcpy(bestCofAC8x16 [1][1], img->cofAC8x16 [1][1], sizeof(int)*129);

}


/*!
*************************************************************************************
* \brief
*    Set stored macroblock parameters
*************************************************************************************
*/
void set_stored_macroblock_parameters (short bestQPcount)
{
  int  i, j, k, ****i4p, ***i3p;
  Macroblock  *currMB  = &img->mb_data[img->current_mb_nr];
  int         mode     = best_mode;
  int         bframe   = (img->type==B_SLICE);
//  char    **ipredmodes = img->ipredmode;
  
  imgpel        **imgY  = enc_picture->imgY;
  imgpel       ***imgUV = enc_picture->imgUV;
  int        block_x, block_y;  
  short   *cur_mv;
  
  //===== reconstruction values =====
  for (j = 0; j < MB_BLOCK_SIZE; j++)
  {
    memcpy(&imgY[img->pix_y+j][img->pix_x],rec_mbY[j], MB_BLOCK_SIZE * sizeof(imgpel));
  }
  
  if (img->yuv_format != YUV400)
  {
    for (j = 0; j<img->mb_cr_size_y; j++)
    {
      memcpy(&imgUV[0][img->pix_c_y+j][img->pix_c_x],rec_mbU[j], img->mb_cr_size_x * sizeof(imgpel));
      memcpy(&imgUV[1][img->pix_c_y+j][img->pix_c_x],rec_mbV[j], img->mb_cr_size_x * sizeof(imgpel));
    }
  }
  
  //===== coefficients and cbp =====
  i4p=cofAC; cofAC=img->cofAC; img->cofAC=i4p;
  i3p=cofDC; cofDC=img->cofDC; img->cofDC=i3p;
  currMB->cbp      = cbp;
  currMB->cbp_blk = cbp_blk;
  //==== macroblock type ====
  currMB->mb_type = mode;
  
  memcpy(currMB->b8mode,b8mode, BLOCK_MULTIPLE * sizeof(int));
  memcpy(currMB->b8pdir,b8pdir, BLOCK_MULTIPLE * sizeof(int));
  
  currMB->bi_pred_me = currMB->mb_type == 1 ? bi_pred_me : 0;  

  if (mode == P8x8 && input->Transform8x8Mode)
  {
    RestoreMV8x8(!luma_transform_size_8x8_flag , bestQPcount);       //1  = restore 4x4 mvs and 0 =  restore 8x8 mvs
  }
  //==== transform size flag ====
  if (((currMB->cbp & 15) == 0) && !(currMB->mb_type == I4MB || currMB->mb_type == I8MB))
    currMB->luma_transform_size_8x8_flag = 0;
  else
    currMB->luma_transform_size_8x8_flag = luma_transform_size_8x8_flag;
  
  rdopt->luma_transform_size_8x8_flag  = currMB->luma_transform_size_8x8_flag;
  
  //==== reference frames =====
  for (j = 0; j < 4; j++)
  {
    block_y = img->block_y + j;
    for (i = 0; i < 4; i++)
    {
      block_x = img->block_x + i;
      k = 2*(j >> 1)+(i >> 1);
      
      // backward prediction or intra
      if ((currMB->b8pdir[k] == 1) || IS_INTRA(currMB))
      {
        enc_picture->ref_idx    [LIST_0][block_y][block_x]    = -1;
        enc_picture->ref_pic_id [LIST_0][block_y][block_x]    = -1;          
        enc_picture->mv         [LIST_0][block_y][block_x][0] = 0;
        enc_picture->mv         [LIST_0][block_y][block_x][1] = 0;
      }
      else
      {
        if (currMB->bi_pred_me && (currMB->b8pdir[k] == 2) && currMB->mb_type==1)
        {
          cur_mv = currMB->bi_pred_me == 1 
            ? img->bipred_mv1[j][i][LIST_0][0][currMB->b8mode[k]] 
            : img->bipred_mv2[j][i][LIST_0][0][currMB->b8mode[k]];
          
          enc_picture->ref_idx    [LIST_0][block_y][block_x] = 0;                         
          enc_picture->ref_pic_id [LIST_0][block_y][block_x] = enc_picture->ref_pic_num[LIST_0 + currMB->list_offset][0];
          enc_picture->mv         [LIST_0][block_y][block_x][0] = cur_mv[0];
          enc_picture->mv         [LIST_0][block_y][block_x][1] = cur_mv[1];
        }
        else
        {
          cur_mv = img->all_mv[j][i][LIST_0][(short)frefframe[j][i]][currMB->b8mode[k]];
          
          enc_picture->ref_idx    [LIST_0][block_y][block_x]    = frefframe[j][i];
          enc_picture->ref_pic_id [LIST_0][block_y][block_x]    = enc_picture->ref_pic_num[LIST_0 + currMB->list_offset][(short)frefframe[j][i]];
          enc_picture->mv         [LIST_0][block_y][block_x][0] = cur_mv[0];
          enc_picture->mv         [LIST_0][block_y][block_x][1] = cur_mv[1];
        }
      }
      
      // forward prediction or intra
      if ((currMB->b8pdir[k] == 0) || IS_INTRA(currMB))
      {
        enc_picture->ref_idx    [LIST_1][block_y][block_x]    = -1;
        enc_picture->ref_pic_id [LIST_1][block_y][block_x]    = -1;
        enc_picture->mv         [LIST_1][block_y][block_x][0] = 0;
        enc_picture->mv         [LIST_1][block_y][block_x][1] = 0;
      }
    }
  }
  
  if (bframe)
  {
    for (j=0; j<4; j++)
    {
      block_y = img->block_y + j;
      for (i=0; i<4; i++)
      {          
        block_x = img->block_x + i;
        k = 2*(j >> 1)+(i >> 1);
        
        // forward
        if (IS_INTRA(currMB)||(currMB->b8pdir[k] == 0))
        {
          enc_picture->ref_idx    [LIST_1][block_y][block_x]    = -1;
          enc_picture->ref_pic_id [LIST_1][block_y][block_x]    = -1;
          enc_picture->mv         [LIST_1][block_y][block_x][0] = 0;
          enc_picture->mv         [LIST_1][block_y][block_x][1] = 0;
        }
        else
        {
          if (currMB->bi_pred_me && (currMB->b8pdir[k] == 2) && currMB->mb_type==1)
          {
            cur_mv = currMB->bi_pred_me == 1 
              ? img->bipred_mv1[j][i][LIST_1][0][currMB->b8mode[k]] 
              : img->bipred_mv2[j][i][LIST_1][0][currMB->b8mode[k]];
            
            enc_picture->ref_idx    [LIST_1][block_y][block_x] = 0; 
            enc_picture->ref_pic_id [LIST_1][block_y][block_x] = enc_picture->ref_pic_num[LIST_1 + currMB->list_offset][0];
            enc_picture->mv         [LIST_1][block_y][block_x][0] = cur_mv[0];
            enc_picture->mv         [LIST_1][block_y][block_x][1] = cur_mv[1];        
          }
          else
          {
            cur_mv = img->all_mv[j][i][LIST_1][(short)brefframe[j][i]][currMB->b8mode[k]];
            
            enc_picture->ref_idx    [LIST_1][block_y][block_x] = brefframe[j][i];
            enc_picture->ref_pic_id [LIST_1][block_y][block_x] = enc_picture->ref_pic_num[LIST_1 + currMB->list_offset][(short)brefframe[j][i]];
            enc_picture->mv         [LIST_1][block_y][block_x][0] = cur_mv[0];
            enc_picture->mv         [LIST_1][block_y][block_x][1] = cur_mv[1];
          }
        }
      }
    }
  }
  
  //==== intra prediction modes ====
  currMB->c_ipred_mode = best_c_imode;
  img->i16offset = best_i16offset;
  
  if(currMB->mb_type == I8MB)
  {
    memcpy(currMB->intra_pred_modes8x8,b8_intra_pred_modes8x8, MB_BLOCK_PARTITIONS * sizeof(char));            
    memcpy(currMB->intra_pred_modes   ,b8_intra_pred_modes8x8, MB_BLOCK_PARTITIONS * sizeof(char));            
    for(j = 0; j < BLOCK_MULTIPLE; j++)
    {
      memcpy(&img->ipredmode   [img->block_y+j][img->block_x], b8_ipredmode8x8[j], BLOCK_MULTIPLE * sizeof(char));
      memcpy(&img->ipredmode8x8[img->block_y+j][img->block_x], b8_ipredmode8x8[j], BLOCK_MULTIPLE * sizeof(char));
    }
  }
  else if (mode!=I4MB && mode!=I8MB)
  {
    memset(currMB->intra_pred_modes,DC_PRED, MB_BLOCK_PARTITIONS * sizeof(char));
    for(j = img->block_y; j < img->block_y + BLOCK_MULTIPLE; j++)
      memset(&img->ipredmode[j][img->block_x], DC_PRED, BLOCK_MULTIPLE * sizeof(char));
  }
  // Residue Color Transform
  else if (mode == I4MB)
  {
    memcpy(currMB->intra_pred_modes,b4_intra_pred_modes, MB_BLOCK_PARTITIONS * sizeof(char));
    for(j = 0; j < BLOCK_MULTIPLE; j++)
      memcpy(&img->ipredmode[img->block_y + j][img->block_x],&b4_ipredmode[BLOCK_MULTIPLE * j], BLOCK_MULTIPLE * sizeof(char));
  }
  
  //==== motion vectors =====
  //SetMotionVectorsMB (currMB, bframe);

#ifdef USE_INTRA_MDDT
  if(input->UseIntraMDDT)
  {
    memcpy(img->cofAC16[0], bestCofAC16[0], sizeof(int)*16*17);
    memcpy(img->cofAC16[1], bestCofAC16[1], sizeof(int)*16*17);
  }
#endif 

  memcpy(img->cofAC16x16[0],    bestCofAC16x16[0],    sizeof(int)*257);
  memcpy(img->cofAC16x16[1],    bestCofAC16x16[1],    sizeof(int)*257);
  memcpy(img->cofAC16x8 [0][0], bestCofAC16x8 [0][0], sizeof(int)*129);
  memcpy(img->cofAC16x8 [0][1], bestCofAC16x8 [0][1], sizeof(int)*129);
  memcpy(img->cofAC16x8 [1][0], bestCofAC16x8 [1][0], sizeof(int)*129);
  memcpy(img->cofAC16x8 [1][1], bestCofAC16x8 [1][1], sizeof(int)*129);
  memcpy(img->cofAC8x16 [0][0], bestCofAC8x16 [0][0], sizeof(int)*129);
  memcpy(img->cofAC8x16 [0][1], bestCofAC8x16 [0][1], sizeof(int)*129);
  memcpy(img->cofAC8x16 [1][0], bestCofAC8x16 [1][0], sizeof(int)*129);
  memcpy(img->cofAC8x16 [1][1], bestCofAC8x16 [1][1], sizeof(int)*129);

}



/*!
*************************************************************************************
* \brief
*    Set reference frames and motion vectors
*************************************************************************************
*/
void SetRefAndMotionVectors (int block, int mode, int pdir, int fwref, int bwref)
{
  int     i, j=0;
  int     bslice  = (img->type==B_SLICE);
  int     pmode   = (mode==1||mode==2||mode==3?mode:4);
  int     j0      = ((block >> 1)<<1);
  int     i0      = ((block & 0x01)<<1);
  int     j1      = j0 + (input->part_size[pmode][1]);
  int     i1      = i0 + (input->part_size[pmode][0]);
  int     block_x, block_y;
  short   *cur_mv;
  Macroblock  *currMB  = &img->mb_data[img->current_mb_nr];
  
  if (pdir<0)
  {
    for (j = img->block_y + j0; j < img->block_y + j1; j++)
    {
      for (i=img->block_x + i0; i<img->block_x +i1; i++)
      {
        enc_picture->ref_pic_id[LIST_0][j][i] = -1;
        enc_picture->ref_pic_id[LIST_1][j][i] = -1;
      }
      memset(&enc_picture->ref_idx[LIST_0][j][img->block_x + i0], -1, (input->part_size[pmode][0]) * sizeof(char));
      memset(&enc_picture->ref_idx[LIST_1][j][img->block_x + i0], -1, (input->part_size[pmode][0]) * sizeof(char));
      memset(enc_picture->mv[LIST_0][j][img->block_x + i0], 0, 2*(input->part_size[pmode][0]) * sizeof(short));
      memset(enc_picture->mv[LIST_1][j][img->block_x + i0], 0, 2*(input->part_size[pmode][0]) * sizeof(short));
    }
    return;
  }
  
  if (!bslice)
  {
    for (j=j0; j<j1; j++)
    {
      block_y = img->block_y + j;
      memset(&enc_picture->ref_idx   [LIST_0][block_y][img->block_x + i0], fwref, (input->part_size[pmode][0]) * sizeof(char));
      for (i=i0; i<i1; i++)
      {
        block_x = img->block_x + i;        
        cur_mv = img->all_mv[j][i][LIST_0][fwref][mode];                
        enc_picture->mv        [LIST_0][block_y][block_x][0] = cur_mv[0];
        enc_picture->mv        [LIST_0][block_y][block_x][1] = cur_mv[1];
        enc_picture->ref_pic_id[LIST_0][block_y][block_x] = enc_picture->ref_pic_num[LIST_0+currMB->list_offset][fwref];
      }
    }
    return;
  }
  else
  {
    for (j=j0; j<j1; j++)
    {
      block_y = img->block_y + j;
      for (i=i0; i<i1; i++)
      {
        block_x = img->block_x + i;
        if (mode==0)
        {
          pdir  = direct_pdir[block_y][block_x];
          fwref = direct_ref_idx[LIST_0][block_y][block_x];
          bwref = direct_ref_idx[LIST_1][block_y][block_x];
        }
        
        if ((pdir==0 || pdir==2))
        {
          if (currMB->bi_pred_me && (pdir == 2) && mode == 1)
          {
            cur_mv = currMB->bi_pred_me == 1 
              ? img->bipred_mv1[j][i][LIST_0][0][mode]
              : img->bipred_mv2[j][i][LIST_0][0][mode];
            
            enc_picture->mv        [LIST_0][block_y][block_x][0] = cur_mv[0];
            enc_picture->mv        [LIST_0][block_y][block_x][1] = cur_mv[1];
            enc_picture->ref_idx   [LIST_0][block_y][block_x]    = 0;            
            enc_picture->ref_pic_id[LIST_0][block_y][block_x]    = enc_picture->ref_pic_num[LIST_0+currMB->list_offset][0];
          }
          else
          {
            cur_mv = img->all_mv[j][i][LIST_0][fwref][mode];
            
            enc_picture->mv        [LIST_0][block_y][block_x][0] = cur_mv[0];
            enc_picture->mv        [LIST_0][block_y][block_x][1] = cur_mv[1];
            enc_picture->ref_idx   [LIST_0][block_y][block_x] = fwref;
            enc_picture->ref_pic_id[LIST_0][block_y][block_x] = 
              enc_picture->ref_pic_num[LIST_0+currMB->list_offset][(short)enc_picture->ref_idx[LIST_0][block_y][block_x]];
          }
        }
        else
        {
          enc_picture->mv        [LIST_0][block_y][block_x][0] = 0;
          enc_picture->mv        [LIST_0][block_y][block_x][1] = 0;
          enc_picture->ref_idx   [LIST_0][block_y][block_x]    = -1;
          enc_picture->ref_pic_id[LIST_0][block_y][block_x]    = -1;
        }
        
        if ((pdir==1 || pdir==2))
        {
          if (currMB->bi_pred_me && (pdir == 2) && mode == 1)
          {
            cur_mv = currMB->bi_pred_me == 1 
              ? img->bipred_mv1[j][i][LIST_1][0][mode]
              : img->bipred_mv2[j][i][LIST_1][0][mode];
            
            enc_picture->mv        [LIST_1][block_y][block_x][0] = cur_mv[0];
            enc_picture->mv        [LIST_1][block_y][block_x][1] = cur_mv[1];
            enc_picture->ref_idx   [LIST_1][block_y][block_x]    = 0;            
            enc_picture->ref_pic_id[LIST_1][block_y][block_x]    = enc_picture->ref_pic_num[LIST_1+currMB->list_offset][0];
          }
          else
          {
            cur_mv = img->all_mv[j][i][LIST_1][bwref][mode];
            
            enc_picture->mv        [LIST_1][block_y][block_x][0] = cur_mv[0];
            enc_picture->mv        [LIST_1][block_y][block_x][1] = cur_mv[1];
            enc_picture->ref_idx   [LIST_1][block_y][block_x] = bwref;
            enc_picture->ref_pic_id[LIST_1][block_y][block_x] = 
              enc_picture->ref_pic_num[LIST_1+currMB->list_offset][(short)enc_picture->ref_idx[LIST_1][block_y][block_x]];
          }
        }
        else
        {
          enc_picture->mv        [LIST_1][block_y][block_x][0] = 0;
          enc_picture->mv        [LIST_1][block_y][block_x][1] = 0;
          enc_picture->ref_idx   [LIST_1][block_y][block_x]    = -1;
          enc_picture->ref_pic_id[LIST_1][block_y][block_x]    = -1;
        }
      }
    }
  }
}

/*!
*************************************************************************************
* \brief
*    Store motion vectors for 8x8 partition
*************************************************************************************
*/

void StoreMVBlock8x8(int dir, int block8x8, int mode, int ref, int bw_ref, int pdir8, int bframe)
{
  int i, j, i0, j0, ii, jj;
  short ******all_mv  = img->all_mv;
  short ******pred_mv = img->pred_mv;
  
  
  i0 = (block8x8 & 0x01) << 1;
  j0 = (block8x8 >> 1) << 1;
  ii = i0+2;
  jj = j0+2;
  
  if (!bframe)
  {
    if (pdir8>=0) //(mode8!=IBLOCK)&&(mode8!=I16MB))  // && ref != -1)
    {
      for (j=j0; j<jj; j++)
        for (i=i0; i<ii; i++)
        {
          all_mv8x8 [dir][LIST_0][j][i][0] = all_mv [j][i][LIST_0][ref][4][0];
          all_mv8x8 [dir][LIST_0][j][i][1] = all_mv [j][i][LIST_0][ref][4][1];
          pred_mv8x8[dir][LIST_0][j][i][0] = pred_mv[j][i][LIST_0][ref][4][0];
          pred_mv8x8[dir][LIST_0][j][i][1] = pred_mv[j][i][LIST_0][ref][4][1];
        }
    }
  }
  else
  {
    if (pdir8==0) // forward
    {
      for (j=j0; j<jj; j++)
        for (i=i0; i<ii; i++)
        {
          all_mv8x8 [dir][LIST_0][j][i][0] = all_mv [j][i][LIST_0][ref][mode][0];
          all_mv8x8 [dir][LIST_0][j][i][1] = all_mv [j][i][LIST_0][ref][mode][1];
          pred_mv8x8[dir][LIST_0][j][i][0] = pred_mv[j][i][LIST_0][ref][mode][0];
          pred_mv8x8[dir][LIST_0][j][i][1] = pred_mv[j][i][LIST_0][ref][mode][1];
        }
    }
    else if (pdir8==1) // backward
    {
      for (j=j0; j<jj; j++)
        for (i=i0; i<ii; i++)
        {
          all_mv8x8 [dir][LIST_1][j][i][0] = all_mv [j][i][LIST_1][bw_ref][mode][0];
          all_mv8x8 [dir][LIST_1][j][i][1] = all_mv [j][i][LIST_1][bw_ref][mode][1];
          pred_mv8x8[dir][LIST_1][j][i][0] = pred_mv[j][i][LIST_1][bw_ref][mode][0];
          pred_mv8x8[dir][LIST_1][j][i][1] = pred_mv[j][i][LIST_1][bw_ref][mode][1];
        }
    }
    else if (pdir8==2) // bidir
    {
      for (j=j0; j<jj; j++)
        for (i=i0; i<ii; i++)
        {
          all_mv8x8 [dir][LIST_0][j][i][0] = all_mv [j][i][LIST_0][ref][mode][0];
          all_mv8x8 [dir][LIST_0][j][i][1] = all_mv [j][i][LIST_0][ref][mode][1];
          pred_mv8x8[dir][LIST_0][j][i][0] = pred_mv[j][i][LIST_0][ref][mode][0];
          pred_mv8x8[dir][LIST_0][j][i][1] = pred_mv[j][i][LIST_0][ref][mode][1];
          
          all_mv8x8 [dir][LIST_1][j][i][0] = all_mv [j][i][LIST_1][bw_ref][mode][0];
          all_mv8x8 [dir][LIST_1][j][i][1] = all_mv [j][i][LIST_1][bw_ref][mode][1];
          pred_mv8x8[dir][LIST_1][j][i][0] = pred_mv[j][i][LIST_1][bw_ref][mode][0];
          pred_mv8x8[dir][LIST_1][j][i][1] = pred_mv[j][i][LIST_1][bw_ref][mode][1];
        }
    }
    else
    {
      error("invalid direction mode", 255);
    }
  }
}



/*!
*************************************************************************************
* \brief
*    Store motion vectors of 8x8 partitions of one macroblock
*************************************************************************************
*/
void StoreMV8x8(int dir, short QPcount)
{
  int block8x8;
  
  int bframe = (img->type == B_SLICE);
  
  for (block8x8=0; block8x8<4; block8x8++)
    StoreMVBlock8x8(dir, block8x8, tr8x8[QPcount].part8x8mode[block8x8], tr8x8[QPcount].part8x8fwref[block8x8], 
    tr8x8[QPcount].part8x8bwref[block8x8], tr8x8[QPcount].part8x8pdir[block8x8], bframe);
}

/*!
*************************************************************************************
* \brief
*    Restore motion vectors for 8x8 partition
*************************************************************************************
*/
void RestoreMVBlock8x8(int dir, int block8x8, RD_8x8DATA tr, int bframe)
{
  int i, j, i0, j0, ii, jj;
  short ******all_mv  = img->all_mv;
  short ******pred_mv = img->pred_mv;
  short pdir8  = tr.part8x8pdir [block8x8];
  short mode   = tr.part8x8mode [block8x8];
  short ref    = tr.part8x8fwref[block8x8];
  short bw_ref = tr.part8x8bwref[block8x8];
  
  i0 = (block8x8 & 0x01) << 1;
  j0 = (block8x8 >> 1) << 1;
  ii = i0+2;
  jj = j0+2;
  
  if (!bframe)
  {
    if (pdir8>=0) //(mode8!=IBLOCK)&&(mode8!=I16MB))  // && ref != -1)
    {
      for (j=j0; j<jj; j++)
        for (i=i0; i<ii; i++)
        {
          all_mv [j][i][LIST_0][ref][4][0] = all_mv8x8 [dir][LIST_0][j][i][0] ;
          all_mv [j][i][LIST_0][ref][4][1] = all_mv8x8 [dir][LIST_0][j][i][1] ;
          pred_mv[j][i][LIST_0][ref][4][0] = pred_mv8x8[dir][LIST_0][j][i][0];
          pred_mv[j][i][LIST_0][ref][4][1] = pred_mv8x8[dir][LIST_0][j][i][1];
        }
    }
  }
  else
  {
    if (pdir8==0) // forward
    {
      for (j=j0; j<jj; j++)
        for (i=i0; i<ii; i++)
        {
          all_mv [j][i][LIST_0][ref][mode][0] = all_mv8x8 [dir][LIST_0][j][i][0] ;
          all_mv [j][i][LIST_0][ref][mode][1] = all_mv8x8 [dir][LIST_0][j][i][1] ;
          pred_mv[j][i][LIST_0][ref][mode][0] = pred_mv8x8[dir][LIST_0][j][i][0];
          pred_mv[j][i][LIST_0][ref][mode][1] = pred_mv8x8[dir][LIST_0][j][i][1];
        }
    }
    else if (pdir8==1) // backward
    {
      for (j=j0; j<jj; j++)
        for (i=i0; i<ii; i++)
        {
          all_mv [j][i][LIST_1][bw_ref][mode][0] = all_mv8x8 [dir][LIST_1][j][i][0] ;
          all_mv [j][i][LIST_1][bw_ref][mode][1] = all_mv8x8 [dir][LIST_1][j][i][1] ;
          pred_mv[j][i][LIST_1][bw_ref][mode][0] = pred_mv8x8[dir][LIST_1][j][i][0];
          pred_mv[j][i][LIST_1][bw_ref][mode][1] = pred_mv8x8[dir][LIST_1][j][i][1];
        }
    }
    else if (pdir8==2) // bidir
    {
      for (j=j0; j<jj; j++)
        for (i=i0; i<ii; i++)
        {
          all_mv [j][i][LIST_0][ref][mode][0] = all_mv8x8 [dir][LIST_0][j][i][0] ;
          all_mv [j][i][LIST_0][ref][mode][1] = all_mv8x8 [dir][LIST_0][j][i][1] ;
          pred_mv[j][i][LIST_0][ref][mode][0] = pred_mv8x8[dir][LIST_0][j][i][0];
          pred_mv[j][i][LIST_0][ref][mode][1] = pred_mv8x8[dir][LIST_0][j][i][1];
          
          all_mv [j][i][LIST_1][bw_ref][mode][0] = all_mv8x8 [dir][LIST_1][j][i][0] ;
          all_mv [j][i][LIST_1][bw_ref][mode][1] = all_mv8x8 [dir][LIST_1][j][i][1] ;
          pred_mv[j][i][LIST_1][bw_ref][mode][0] = pred_mv8x8[dir][LIST_1][j][i][0];
          pred_mv[j][i][LIST_1][bw_ref][mode][1] = pred_mv8x8[dir][LIST_1][j][i][1];
        }
    }
    else
    {
      error("invalid direction mode", 255);
    }
  }
}

/*!
*************************************************************************************
* \brief
*    Restore motion vectors of 8x8 partitions of one macroblock
*************************************************************************************
*/
void RestoreMV8x8(int dir, short QPcount)
{
  int block8x8;
  
  int bframe = (img->type == B_SLICE);
  
  for (block8x8=0; block8x8<4; block8x8++)
    RestoreMVBlock8x8(dir, block8x8, tr8x8[QPcount], bframe);   
}


/*!
*************************************************************************************
* \brief
*    Store predictors for 8x8 partition
*************************************************************************************
*/

void StoreNewMotionVectorsBlock8x8(int dir, int block8x8, int mode, int fw_ref, int bw_ref, int pdir8, int bframe)
{
  int i, j, i0, j0, ii, jj;
  short ******all_mv  = img->all_mv;
  short ******pred_mv = img->pred_mv;
  
  
  i0 = (block8x8 & 0x01) << 1;
  j0 = (block8x8 >> 1) << 1;
  ii = i0+2;
  jj = j0+2;
  
  if (pdir8<0)
  {
    for (j=j0; j<jj; j++)
    {
      memset(&all_mv8x8[dir][LIST_0][j][i0], 0, 2 * 2 * sizeof(short));
      memset(&all_mv8x8[dir][LIST_1][j][i0], 0, 2 * 2 * sizeof(short));
    }
    return;
  }
  
  if (!bframe)
  {
    for (j=j0; j<jj; j++)
    {
      for (i=i0; i<ii; i++)
      {
        all_mv8x8 [dir][LIST_0][j][i][0] = all_mv [j][i][LIST_0][fw_ref][4][0];
        all_mv8x8 [dir][LIST_0][j][i][1] = all_mv [j][i][LIST_0][fw_ref][4][1];
        pred_mv8x8[dir][LIST_0][j][i][0] = pred_mv[j][i][LIST_0][fw_ref][4][0];
        pred_mv8x8[dir][LIST_0][j][i][1] = pred_mv[j][i][LIST_0][fw_ref][4][1];
      }
    }
    return;
  }
  else
  {
    if ((pdir8==0 || pdir8==2))
    {
      for (j=j0; j<jj; j++)
        for (i=i0; i<ii; i++)
        {
          all_mv8x8 [dir][LIST_0][j][i][0] = all_mv [j][i][LIST_0][fw_ref][mode][0];
          all_mv8x8 [dir][LIST_0][j][i][1] = all_mv [j][i][LIST_0][fw_ref][mode][1];
          pred_mv8x8[dir][LIST_0][j][i][0] = pred_mv[j][i][LIST_0][fw_ref][mode][0];
          pred_mv8x8[dir][LIST_0][j][i][1] = pred_mv[j][i][LIST_0][fw_ref][mode][1];
        }
    }
    else
    {
      for (j=j0; j<jj; j++)
        memset(&all_mv8x8[dir][LIST_0][j][i0], 0, 2 * 2 * sizeof(short));
    }
    
    if ((pdir8==1 || pdir8==2))
    {
      for (j=j0; j<jj; j++)
        for (i=i0; i<ii; i++)
        {
          all_mv8x8 [dir][LIST_1][j][i][0] = all_mv [j][i][LIST_1][bw_ref][mode][0];
          all_mv8x8 [dir][LIST_1][j][i][1] = all_mv [j][i][LIST_1][bw_ref][mode][1];
          pred_mv8x8[dir][LIST_1][j][i][0] = pred_mv[j][i][LIST_1][bw_ref][mode][0];
          pred_mv8x8[dir][LIST_1][j][i][1] = pred_mv[j][i][LIST_1][bw_ref][mode][1];
        }
    }
    else
    {
      for (j=j0; j<jj; j++)
        memset(&all_mv8x8[dir][LIST_1][j][i0], 0, 2 * 2 * sizeof(short));
    }
  }
}

/*!
************************************************************************
* \brief
*    store coding state (for rd-optimized mode decision), used for 8x8 transformation
************************************************************************
*/
void store_coding_state_cs_cm()
{
  store_coding_state(cs_cm);
}

/*!
************************************************************************
* \brief
*    restore coding state (for rd-optimized mode decision), used for 8x8 transformation
************************************************************************
*/
void reset_coding_state_cs_cm()
{
  reset_coding_state(cs_cm);
}

void assign_enc_picture_params(int mode, int best_pdir, int block, int list_offset, int best_fw_ref, int best_bw_ref, int bframe)
{
  int i,j;
  int block_x, block_y;
  short *cur_mv;
  
  if (mode==1)
  {
    if (best_pdir==1)
    {
      for (j=img->block_y+(block&2); j<img->block_y+(block&2) + BLOCK_MULTIPLE; j++)
      {
        block_x = img->block_x+(block&1)*2;
        
        memset(&enc_picture->ref_idx[LIST_0][j][block_x], -1 ,     BLOCK_MULTIPLE * sizeof(char));
        memset(enc_picture->mv      [LIST_0][j][block_x],  0 , 2 * BLOCK_MULTIPLE * sizeof(short));
        for (i=block_x; i<block_x + BLOCK_MULTIPLE; i++)
        {
          enc_picture->ref_pic_id [LIST_0][j][i]    = -1;
        }
      }
    }
    else if (img->bi_pred_me[mode])
    {
      for (j=0; j<BLOCK_MULTIPLE; j++)
      {
        block_y = img->block_y+(block&2)+j;
        block_x = img->block_x+(block&1)*2;
        memset(&enc_picture->ref_idx[LIST_0][block_y][block_x], 0, BLOCK_MULTIPLE * sizeof(char));
        for (i=0; i<BLOCK_MULTIPLE; i++)
        {            
          cur_mv = img->bi_pred_me[mode] == 1 
            ? img->bipred_mv1[i][j][LIST_0][0][mode] 
            : img->bipred_mv2[i][j][LIST_0][0][mode];
          
          enc_picture->ref_pic_id [LIST_0][block_y][block_x + i]    = enc_picture->ref_pic_num[LIST_0 + list_offset][0];  
          enc_picture->mv         [LIST_0][block_y][block_x + i][0] = cur_mv[0];
          enc_picture->mv         [LIST_0][block_y][block_x + i][1] = cur_mv[1];
        }
      }
    }
    else 
    {
      for (j=0; j<BLOCK_MULTIPLE; j++)
      {
        block_y = img->block_y+(block&2)+j;
        block_x = img->block_x+(block&1)*2;
        memset(&enc_picture->ref_idx[LIST_0][block_y][block_x], best_fw_ref , BLOCK_MULTIPLE * sizeof(char));
        for (i=0; i<BLOCK_MULTIPLE; i++)
        {                                
          cur_mv = img->all_mv[j][i][LIST_0][best_fw_ref][mode];
          
          enc_picture->ref_pic_id [LIST_0][block_y][block_x + i]    = enc_picture->ref_pic_num[LIST_0 + list_offset][best_fw_ref];  
          enc_picture->mv         [LIST_0][block_y][block_x + i][0] = cur_mv[0];
          enc_picture->mv         [LIST_0][block_y][block_x + i][1] = cur_mv[1];
        }          
      }
    }
    
    if (bframe)
    {
      if (best_pdir==0)
      {
        for (j=img->block_y+(block&2); j<img->block_y+(block&2) + BLOCK_MULTIPLE; j++)
        {
          block_x = img->block_x+(block&1)*2;
          memset(&enc_picture->ref_idx[LIST_1][j][block_x], -1 , BLOCK_MULTIPLE * sizeof(char));
          memset(enc_picture->mv[LIST_1][j][block_x], 0 , 2 * BLOCK_MULTIPLE * sizeof(short));
          for (i=block_x; i<block_x + BLOCK_MULTIPLE; i++)
          {
            enc_picture->ref_pic_id [LIST_1][j][i] = -1;
          }
        }
      }
      else
      {
        if (img->bi_pred_me[mode])
        {
          for (j=0; j<BLOCK_MULTIPLE; j++)
          {
            block_y = img->block_y+(block&2)+j;
            block_x = img->block_x+(block&1)*2;
            memset(&enc_picture->ref_idx[LIST_1][block_y][block_x], 0, BLOCK_MULTIPLE * sizeof(char));
            for (i=0; i<BLOCK_MULTIPLE; i++)
            {                     
              cur_mv = img->bi_pred_me[mode] == 1 
                ? img->bipred_mv1[i][j][LIST_1][0][mode] 
                : img->bipred_mv2[i][j][LIST_1][0][mode];
              
              enc_picture->ref_pic_id [LIST_1][block_y][block_x + i] = 
                enc_picture->ref_pic_num[LIST_1 + list_offset][0];
              enc_picture->mv         [LIST_1][block_y][block_x + i][0] = cur_mv[0];
              enc_picture->mv         [LIST_1][block_y][block_x + i][1] = cur_mv[1];
            }
          }
        }
        else 
        {
          for (j=0; j<BLOCK_MULTIPLE; j++)
          {
            block_y = img->block_y+(block&2)+j;
            block_x = img->block_x+(block&1)*2;
            memset(&enc_picture->ref_idx[LIST_1][block_y][block_x], best_bw_ref, BLOCK_MULTIPLE * sizeof(char));
            for (i=0; i<BLOCK_MULTIPLE; i++)
            {                     
              
              enc_picture->ref_pic_id [LIST_1][block_y][block_x + i] = 
                enc_picture->ref_pic_num[LIST_1 + list_offset][best_bw_ref];
              if(best_bw_ref>=0)
              {
                cur_mv = img->all_mv[j][i][LIST_1][best_bw_ref][mode];
                enc_picture->mv[LIST_1][block_y][block_x + i][0] = cur_mv[0];
                enc_picture->mv[LIST_1][block_y][block_x + i][1] = cur_mv[1];
              }
            }            
          }
        }
      }
    }
  }
  else if (mode==2)
  {
    for (j=0; j<2; j++)
    {
      block_y = img->block_y + block * 2 + j;
      for (i=0; i<BLOCK_MULTIPLE; i++)
      {
        block_x = img->block_x + i;
        if (best_pdir==1)
        {
          enc_picture->ref_idx    [LIST_0][block_y][block_x]    = -1;
          enc_picture->ref_pic_id [LIST_0][block_y][block_x]    = -1;
          enc_picture->mv         [LIST_0][block_y][block_x][0] = 0;
          enc_picture->mv         [LIST_0][block_y][block_x][1] = 0;
        }
        else
        {                     
          cur_mv = img->all_mv[j+block*2][i][LIST_0][best_fw_ref][mode];
          
          enc_picture->ref_idx    [LIST_0][block_y][block_x]    = best_fw_ref;
          enc_picture->ref_pic_id [LIST_0][block_y][block_x]    = 
            enc_picture->ref_pic_num[LIST_0 + list_offset][best_fw_ref];
          enc_picture->mv         [LIST_0][block_y][block_x][0] = cur_mv[0];
          enc_picture->mv         [LIST_0][block_y][block_x][1] = cur_mv[1];
        }
        
        if (bframe)
        {
          if (best_pdir==0)
          {
            enc_picture->ref_idx    [LIST_1][block_y][block_x]    = -1;
            enc_picture->ref_pic_id [LIST_1][block_y][block_x]    = -1;
            enc_picture->mv         [LIST_1][block_y][block_x][0] = 0;
            enc_picture->mv         [LIST_1][block_y][block_x][1] = 0;
          }
          else
          {
            enc_picture->ref_idx[LIST_1][block_y][block_x] = best_bw_ref;
            if(best_bw_ref>=0)
            {
              cur_mv = img->all_mv[j+ block*2][i][LIST_1][best_bw_ref][mode];
              
              enc_picture->ref_pic_id [LIST_1][block_y][block_x] = 
                enc_picture->ref_pic_num[LIST_1 + list_offset][best_bw_ref];
              enc_picture->mv[LIST_1][block_y][block_x][0] = cur_mv[0];
              enc_picture->mv[LIST_1][block_y][block_x][1] = cur_mv[1];
            }                       
          }
        }
      }
    }
  }
  else
  {
    for (j=0; j<BLOCK_MULTIPLE; j++)
    {
      block_y = img->block_y+j;
      for (i=0; i<2; i++)
      {
        block_x = img->block_x + block*2 + i;
        if (best_pdir==1)
        {
          enc_picture->ref_idx    [LIST_0][block_y][block_x]    = -1;
          enc_picture->ref_pic_id [LIST_0][block_y][block_x]    = -1;
          enc_picture->mv         [LIST_0][block_y][block_x][0] = 0;
          enc_picture->mv         [LIST_0][block_y][block_x][1] = 0;
        }
        else
        {
          cur_mv = img->all_mv[j][block*2+i][LIST_0][best_fw_ref][mode];
          
          enc_picture->ref_idx    [LIST_0][block_y][block_x] = best_fw_ref;
          enc_picture->ref_pic_id [LIST_0][block_y][block_x] = 
            enc_picture->ref_pic_num[LIST_0 + list_offset][best_fw_ref];          
          enc_picture->mv[LIST_0][block_y][block_x][0] = cur_mv[0];
          enc_picture->mv[LIST_0][block_y][block_x][1] = cur_mv[1];
        }
        
        if (bframe)
        {
          if (best_pdir==0)
          {
            enc_picture->ref_idx    [LIST_1][block_y][block_x]    = -1;
            enc_picture->ref_pic_id [LIST_1][block_y][block_x]    = -1;
            enc_picture->mv         [LIST_1][block_y][block_x][0] = 0;
            enc_picture->mv         [LIST_1][block_y][block_x][1] = 0;
          }
          else
          {
            enc_picture->ref_idx[LIST_1][block_y][block_x] = best_bw_ref;
            if(best_bw_ref>=0)
            {
              cur_mv = img->all_mv[j][block*2+i][LIST_1][best_bw_ref][mode];
              enc_picture->ref_pic_id [LIST_1][block_y][block_x] = 
                enc_picture->ref_pic_num[LIST_1 + list_offset][best_bw_ref];
              
              enc_picture->mv[LIST_1][block_y][block_x][0] = cur_mv[0];
              enc_picture->mv[LIST_1][block_y][block_x][1] = cur_mv[1];
            }
          }
        }
      }
    }
  }
}

void update_refresh_map(int intra, int intra1, Macroblock *currMB)
{
  if (input->RestrictRef==1)
  {
    // Modified for Fast Mode Decision. Inchoon Choi, SungKyunKwan Univ.
    if (input->rdopt<2)
    {
      refresh_map[2*img->mb_y  ][2*img->mb_x  ] = (intra ? 1 : 0);
      refresh_map[2*img->mb_y  ][2*img->mb_x+1] = (intra ? 1 : 0);
      refresh_map[2*img->mb_y+1][2*img->mb_x  ] = (intra ? 1 : 0);
      refresh_map[2*img->mb_y+1][2*img->mb_x+1] = (intra ? 1 : 0);
    }
  }
  else if (input->RestrictRef==2)
  {
    refresh_map[2*img->mb_y  ][2*img->mb_x  ] = (currMB->mb_type==I16MB || currMB->mb_type==I4MB ? 1 : 0);
    refresh_map[2*img->mb_y  ][2*img->mb_x+1] = (currMB->mb_type==I16MB || currMB->mb_type==I4MB ? 1 : 0);
    refresh_map[2*img->mb_y+1][2*img->mb_x  ] = (currMB->mb_type==I16MB || currMB->mb_type==I4MB ? 1 : 0);
    refresh_map[2*img->mb_y+1][2*img->mb_x+1] = (currMB->mb_type==I16MB || currMB->mb_type==I4MB ? 1 : 0);
  }
}  

void set_stored_macroblock_parameters1 ()
{
  Macroblock  *currMB  = &img->mb_data[img->current_mb_nr];
  memcpy(currMB->b8mode,b8mode, BLOCK_MULTIPLE * sizeof(int));
  memcpy(currMB->b8pdir,b8pdir, BLOCK_MULTIPLE * sizeof(int));
}
