
/*!
*************************************************************************************
* \file weighted_prediction.c
*
* \brief
*    Estimate weights for WP
*
* \author
*    Main contributors (see contributors.h for copyright, address and affiliation details)
*     - Alexis Michael Tourapis         <alexismt@ieee.org>
*************************************************************************************
*/
#include <stdlib.h>
#include "contributors.h"

#include "global.h"
#include <memory.h>
#include "image.h"

/*!
************************************************************************
* \brief
*    Tests P slice weighting factors to perform or not WP RD decision
************************************************************************
*/
int test_wp_P_slice(int select_offset)
{
  int i, j, n;
  
  double dc_org = 0.0;
  int index;
  double dc_ref[MAX_REFERENCE_PICTURES];
  
  pel_t*  ref_pic;   
  pel_t*  ref_pic_w;   
  int default_weight;
  int default_weight_chroma;
  int list_offset   = 0;
  int weight[2][MAX_REFERENCE_PICTURES][3]; 
  int offset[2][MAX_REFERENCE_PICTURES][3];       
  int clist;
  int perform_wp = 0;      
  
  luma_log_weight_denom = 5;
  chroma_log_weight_denom = 5;
  wp_luma_round = 1 << (luma_log_weight_denom - 1);
  wp_chroma_round = 1 << (chroma_log_weight_denom - 1);
  default_weight = 1<<luma_log_weight_denom;
  default_weight_chroma = 1<<chroma_log_weight_denom;
  
  /* set all values to defaults */
  for (i = 0; i < 2 + list_offset; i++)
  {
    for (j = 0; j < listXsize[i]; j++)
    {
      for (n = 0; n < 3; n++)			//0=Luma  1,2=Chroma
      {
        weight[i][j][n] = default_weight;
        wp_weight[i][j][n] = default_weight;
        wp_offset[i][j][n] = 0;
        offset[i][j][n] = 0;
      }
    }
  }

  dc_org = img->dc_org;

  for (clist=0; clist<2 + list_offset; clist++)
  {
    for (n = 0; n < listXsize[clist]; n++)
    {
      dc_ref[n] = 0.0;
      
      ref_pic       = listX[clist][n]->imgY_11;
      ref_pic_w     = listX[clist][n]->imgY_11_w;
      
      // Y
      for (i = 0; i < img->height * img->width; i++)
      {
        dc_ref[n] += (double) ref_pic[i];
      }
      
      if (select_offset==0)
      {
        if (dc_ref[n] != 0.0)
          weight[clist][n][0] = (int) (default_weight * dc_org / dc_ref[n] + 0.5);
        else
          weight[clist][n][0] = default_weight;  // only used when reference picture is black
        if (weight[clist][n][0] < -64 || weight[clist][n][0] >127)
          weight[clist][n][0] = 32;
      }
      else
      {        
        offset[clist][n][0] = (int) ((dc_org-dc_ref[n])/(img->height*img->width)+0.5);
#ifdef USE_NEW_OFFSET
        if(input->UseNewOffset)
          offset[clist][n][0] = img->frameOffset[clist][n]; 
#endif 
        offset[clist][n][0] = (offset[clist][n][0]+((img->bitdepth_luma-8)>>1))>>(img->bitdepth_luma-8);
        offset[clist][n][0] = (offset[clist][n][0]<-128) ? -128: (offset[clist][n][0]>127) ? 127:offset[clist][n][0];
        offset[clist][n][0] = offset[clist][n][0]<<(img->bitdepth_luma-8);
        weight[clist][n][0] = default_weight;
      }
      
      /* for now always use default weight for chroma weight */
      weight[clist][n][1] = default_weight_chroma;
      weight[clist][n][2] = default_weight_chroma;
                  
    }
  }
  
  for (clist=0; clist<2 + list_offset; clist++)
  {
    for (index = 0; index < listXsize[clist]; index++)
    {
        if ((weight[clist][index][0] != default_weight) ||  (offset[clist][index][0] != 0))		//check for Luma only
        {
          perform_wp = 1;
          break;
        }
    }
    if (perform_wp == 1) break;
  }


  return perform_wp;


}

void estimate_weighting_factor_P_slice(int select_offset)
{
  int i, j, n;
  
  double dc_org = 0.0;
  int index;
  int comp;
  double dc_ref[MAX_REFERENCE_PICTURES];

  pel_t*  ref_pic;   
  pel_t*  ref_pic_w;   
  int default_weight[3];

  int list_offset   = 0;
  int weight[2][MAX_REFERENCE_PICTURES][3]; 
  int offset[2][MAX_REFERENCE_PICTURES][3];       
  int clist;
 
  
  luma_log_weight_denom = 5;
  chroma_log_weight_denom = 5;
  wp_luma_round = 1 << (luma_log_weight_denom - 1);
  wp_chroma_round = 1 << (chroma_log_weight_denom - 1);
  default_weight[0] = 1<<luma_log_weight_denom;
  default_weight[1] = default_weight[2] = 1<<chroma_log_weight_denom;
  
  /* set all values to defaults */
  for (i = 0; i < 2 + list_offset; i++)
  {
    for (j = 0; j < listXsize[i]; j++)
    {
      for (n = 0; n < 3; n++)
      {
        weight[i][j][n] = default_weight[n];
        wp_weight[i][j][n] = default_weight[n];
        wp_offset[i][j][n] = 0;
        offset[i][j][n] = 0;
      }
    }
  }
  
  dc_org = img->dc_org;

  for (clist=0; clist<2 + list_offset; clist++)
  {
    for (n = 0; n < listXsize[clist]; n++)
    {
      dc_ref[n] = 0.0;
      
      ref_pic       = listX[clist][n]->imgY_11;
      ref_pic_w     = listX[clist][n]->imgY_11_w;
      
      // Y
      for (i = 0; i < img->height * img->width; i++)
      {
        dc_ref[n] += (double) ref_pic[i];
      }
      
      if (select_offset==0.0)
      {
        if (dc_ref[n] != 0)
          weight[clist][n][0] = (int) (default_weight[0] * dc_org / dc_ref[n] + 0.5);
        else
          weight[clist][n][0] = default_weight[0];  // only used when reference picture is black
        if (weight[clist][n][0] < -64 || weight[clist][n][0] >127)
          weight[clist][n][0] = default_weight[0];
      }
      else
      {        
        offset[clist][n][0] = (int) ((dc_org-dc_ref[n])/(img->height*img->width)+0.5);
#ifdef USE_NEW_OFFSET
        if(input->UseNewOffset)
          offset[clist][n][0] = img->frameOffset[clist][n];
#endif
        offset[clist][n][0] = (offset[clist][n][0]+((img->bitdepth_luma-8)>>1))>>(img->bitdepth_luma-8);
        offset[clist][n][0] = (offset[clist][n][0]<-128) ? -128: (offset[clist][n][0]>127) ? 127:offset[clist][n][0];
        offset[clist][n][0] = offset[clist][n][0]<<(img->bitdepth_luma-8);
        weight[clist][n][0] = default_weight[0];
      }
      
      
      /* for now always use default weight for chroma weight */
      weight[clist][n][1] = default_weight[1];
      weight[clist][n][2] = default_weight[2];
      
      
      
      /* store weighted reference pic for motion estimation */
      for (i = 0; i < img->height * img->width; i++)
      {          
        ref_pic_w[i] = Clip3 (0, img->max_imgpel_value , (((int) ref_pic[i] * weight[clist][n][0] + wp_luma_round) >> luma_log_weight_denom) + offset[clist][n][0]);
      }
      for (i = 0; i < 4*(img->height + 2*IMG_PAD_SIZE) ; i++)
      {
        for (j = 0; j< 4*(img->width + 2*IMG_PAD_SIZE); j++)
        {
          listX[LIST_0][n]->imgY_ups_w[i][j] =   Clip3 (0, img->max_imgpel_value, (((int) listX[LIST_0 ][n]->imgY_ups[i][j] * weight[clist][n][0] + wp_luma_round) >> luma_log_weight_denom) + offset[clist][n][0]);
        }
      }
    }
  }
  
  for (clist=0; clist<2 + list_offset; clist++)
  {
    for (index = 0; index < listXsize[clist]; index++)
    {
      for (comp=0; comp < 3; comp ++)
      {
        wp_weight[clist][index][comp] = weight[clist][index][comp];
        wp_offset[clist][index][comp] = offset[clist][index][comp];
        explicit_weight[clist][index][comp] = weight[clist][index][comp];
        // printf("index %d component %d weight %d offset %d\n",index,comp,weight[0][index][comp],offset[0][index][comp]);
      }
    }
  }
  
}

/*!
************************************************************************
* \brief
*    test_wp_B_slice:
*    Tests B slice weighting prediction
* select_method = 0  means Weights
* select_method = 1  means Offsets
************************************************************************
*/
int test_wp_B_slice(int select_method)		
{

  int i, j, n;
  int index;
  
  int default_weight[3];
  // this needs to be fixed. 
  int list_offset   = 0;
  int weight[6][MAX_REFERENCE_PICTURES][3]; 
  int offset[6][MAX_REFERENCE_PICTURES][3];       
  int clist;
  int wf_weight=64;
  int perform_wp = 0;      
  int active_refs[2];

  int bipred_rounding_control, testedWeight, temp;
  imgpel recPic;
  double errorMin, error;//, errorLuma[128];

  luma_log_weight_denom = 6;
  chroma_log_weight_denom = 6;

  wp_luma_round = 1 << (luma_log_weight_denom - 1);
  wp_chroma_round = 1 << (chroma_log_weight_denom - 1);
  default_weight[0] = 1<<luma_log_weight_denom;             // 1<<6 = 64
  default_weight[1] = 1<<chroma_log_weight_denom;
  default_weight[2] = 1<<chroma_log_weight_denom;
  
// set all Weights to defaults
// set all Offsets to defaults if select_method=0 
  for (clist=0; clist<2 + list_offset; clist++)
  {
    for (n = 0; n < listXsize[clist]; n++)
    {
      weight[clist][n][0] = default_weight[0];
      weight[clist][n][1] = default_weight[1];
      weight[clist][n][2] = default_weight[2];
      offset[clist][n][0] = 0;
      offset[clist][n][1] = 0;
      offset[clist][n][2] = 0;
#ifdef USE_NEW_OFFSET
      if(input->UseNewOffset && select_method==1)
      {
        offset[clist][n][0] = img->frameOffset[clist][n];
        offset[clist][n][0] = (offset[clist][n][0]+((img->bitdepth_luma-8)>>1))>>(img->bitdepth_luma-8);
        offset[clist][n][0] = (offset[clist][n][0]<-128) ? -128: (offset[clist][n][0]>127) ? 127:offset[clist][n][0];
        offset[clist][n][0] = offset[clist][n][0]<<(img->bitdepth_luma-8);
      }
#endif
    }
  }

  if (select_method == 0)
  {
    bipred_rounding_control = (img->nal_reference_idc!=0);
    errorMin=0; 
    for (testedWeight=0; testedWeight<128; testedWeight++)
    {
      error=0;
      for (i = 0; i < img->height; i++)
      {
        for (j = 0; j < img->width; j++)
        {
          recPic=clip1a((testedWeight*listX[0][0]->imgY[i][j]+(128-testedWeight)*listX[1][0]->imgY[i][j]+ 2*wp_luma_round*(1-bipred_rounding_control))>>(luma_log_weight_denom + 1));
          temp = imgY_org[i][j] - recPic;
          error += temp*temp;
        }
      }
      //errorLuma[testedWeight]=error;
      if (error<errorMin || testedWeight==0)
      {
        errorMin=error; 
        wf_weight=testedWeight;
      }
    }


    clist=0;
    for (n = 0; n < listXsize[clist]; n++)
    {
      weight[clist][n][0] = wf_weight;
      weight[clist][n][1] = wf_weight;
      weight[clist][n][2] = wf_weight;
    }
    clist=1;
    for (n = 0; n < listXsize[clist]; n++)
    {
      weight[clist][n][0] = (128-wf_weight);
      weight[clist][n][1] = (128-wf_weight);
      weight[clist][n][2] = (128-wf_weight);
    }
  }

  for (clist=0; clist<2; clist++)
  {
    for (index = 0; index < listXsize[clist]; index++)
    {
        wp_offset[clist][index][0] = offset[clist][index][0];
        wp_offset[clist][index][1] = offset[clist][index][1];
        wp_offset[clist][index][2] = offset[clist][index][2];
    }
  }

  active_refs[0]=input->B_List0_refs == 0 ? listXsize[0] : min(input->B_List0_refs,listXsize[0]);
  active_refs[1]=input->B_List1_refs == 0 ? listXsize[1] : min(input->B_List1_refs,listXsize[1]);


  if(select_method==1)
  {
    for (clist=0; clist<2; clist++)
    {
      for (index = 0; index < active_refs[clist]; index++)
      {
          if (wp_offset[clist][index][0] != 0)
          {
            perform_wp = 1;
            break;
          }
        if (perform_wp == 1) break;
      }
      if (perform_wp == 1) break;
    }
    return (perform_wp);
  }
  else
    return (wf_weight);

}




/*!
************************************************************************
* \brief
*    Estimates reference picture weighting factors
************************************************************************
*/
void estimate_weighting_factor_B_slice()
{
  int i, j, k, n;
  
  int index;
  int comp;
  int log_weight_denom;
  
  pel_t*  ref_pic;   
  pel_t*  ref_pic_w;   
  pel_t**  ref_qpic;   
  pel_t**  ref_qpic_w;   

  int default_weight[3];
  int list_offset   = 0;
  int weight[6][MAX_REFERENCE_PICTURES][3]; 
  int offset[6][MAX_REFERENCE_PICTURES][3];       
  int clist;
  int wf_weight=64;
  
  int bipred_rounding_control, testedWeight, temp;
  imgpel recPic;
  double errorMin, error;//, errorLuma[128];



  luma_log_weight_denom = 6;
  chroma_log_weight_denom = 6;
  wp_luma_round = 1 << (luma_log_weight_denom - 1);
  wp_chroma_round = 1 << (chroma_log_weight_denom - 1);
  default_weight[0] = 1<<luma_log_weight_denom;
  default_weight[1] = 1<<chroma_log_weight_denom;
  default_weight[2] = 1<<chroma_log_weight_denom;
  
  /* set all values to defaults */
  for (i = 0; i < 2 + list_offset; i++)
  {
    for (j = 0; j < listXsize[i]; j++)
    {
      for (n = 0; n < 3; n++)
      {
        wp_weight[i][j][n] = default_weight[n];
        weight   [i][j][n] = default_weight[n];
        wp_offset[i][j][n] = 0;
        offset   [i][j][n] = 0;
      }
    }
  }
  
  if (active_pps->weighted_bipred_idc == 2) //! implicit mode
  {
    for (i = 0; i < listXsize[LIST_0]; i++)
    {
      for (j = 0; j < listXsize[LIST_1]; j++)
      {
        for (comp = 0; comp < 3; comp++)
        {
          log_weight_denom = (comp == 0) ? luma_log_weight_denom : chroma_log_weight_denom;         
          wbp_weight[1][i][j][comp] = default_weight[comp];//im_weight[1][i][j][comp] ;
          wbp_weight[0][i][j][comp] = default_weight[comp];//im_weight[0][i][j][comp];
        }
      }
    }
    
    for (clist=0; clist<2 + list_offset; clist++)
    {
      for (index = 0; index < listXsize[clist]; index++)
      {
        wp_weight[clist][index][0] = default_weight[0];
        wp_weight[clist][index][1] = default_weight[1];
        wp_weight[clist][index][2] = default_weight[2];
        wp_offset[clist][index][0] = 0;
        wp_offset[clist][index][1] = 0;
        wp_offset[clist][index][2] = 0;

        explicit_weight[clist][index][0] = default_weight[0];
        explicit_weight[clist][index][1] = default_weight[1];
        explicit_weight[clist][index][2] = default_weight[2];

      }
    }
        
    for (k= 0; k<2; k++) 
    {
      for (i = 0; i < listXsize[k]; i++)
      {
        memcpy(listX[k][i]->imgY_11_w, listX[k][i]->imgY_11, img->height * img->width * sizeof(imgpel));
        memcpy(&listX[k][i]->imgY_ups_w[0][0], &listX[k][i]->imgY_ups[0][0], 
          16 * (img->height + 2*IMG_PAD_SIZE) * (img->width + 2*IMG_PAD_SIZE) * sizeof(imgpel));
      }
    }
  }
  else			//explicit
  {
    for (clist=0; clist<2 + list_offset; clist++)
    {
      for (n = 0; n < listXsize[clist]; n++)
      {
        weight[clist][n][0] = default_weight[0];
        weight[clist][n][1] = default_weight[1];
        weight[clist][n][2] = default_weight[2];
        offset[clist][n][0] = 0;
        offset[clist][n][1] = 0;
        offset[clist][n][2] = 0;
        
#ifdef USE_NEW_OFFSET
        if(input->UseNewOffset && img->rdPassWeights==0)
        {
          offset[clist][n][0] = img->frameOffset[clist][n];
          offset[clist][n][0] = (offset[clist][n][0]+((img->bitdepth_luma-8)>>1))>>(img->bitdepth_luma-8);
          offset[clist][n][0] = (offset[clist][n][0]<-128) ? -128: (offset[clist][n][0]>127) ? 127:offset[clist][n][0];
          offset[clist][n][0] = offset[clist][n][0]<<(img->bitdepth_luma-8);
        }
#endif
      }
    }

    if(img->rdPassWeights == 1)
    {
      bipred_rounding_control = (img->nal_reference_idc!=0);
      errorMin=0; 
      for (testedWeight=0; testedWeight<128; testedWeight++)
      {
        error=0;
        for (i = 0; i < img->height; i++)
        {
          for (j = 0; j < img->width; j++)
          {
            recPic=clip1a((testedWeight*listX[0][0]->imgY[i][j]+(128-testedWeight)*listX[1][0]->imgY[i][j]+ 2*wp_luma_round*(1-bipred_rounding_control))>>(luma_log_weight_denom + 1));
            temp = imgY_org[i][j] - recPic;
            error += temp*temp;
          }
        }
        //errorLuma[testedWeight]=error;
        if (error<errorMin || testedWeight==0)
        {
          errorMin=error; 
          wf_weight=testedWeight;
        }
      }

      clist=0;
      for (n = 0; n < listXsize[clist]; n++)
      {
        weight[clist][n][0] = wf_weight;
        weight[clist][n][1] = wf_weight;
        weight[clist][n][2] = wf_weight;
      }

      clist=1;
      for (n = 0; n < listXsize[clist]; n++)
      {
        weight[clist][n][0] = (128-wf_weight);
        weight[clist][n][1] = (128-wf_weight);
        weight[clist][n][2] = (128-wf_weight);
      }
    }


      for (clist=0; clist<2 + list_offset; clist++)
      {
        for (index = 0; index < listXsize[clist]; index++)
        {
            wp_weight[clist][index][0] = weight[clist][index][0];
            wp_weight[clist][index][1] = weight[clist][index][1];
            wp_weight[clist][index][2] = weight[clist][index][2];

            explicit_weight[clist][index][0] = weight[clist][index][0];
            explicit_weight[clist][index][1] = weight[clist][index][1];
            explicit_weight[clist][index][2] = weight[clist][index][2];

            wp_offset[clist][index][0] = offset[clist][index][0];
            wp_offset[clist][index][1] = offset[clist][index][1];
            wp_offset[clist][index][2] = offset[clist][index][2];
            //printf("%d %d\n",wp_weight[clist][index][comp],wp_offset[clist][index][comp]);
        }
      }

//Rahul--wp_weight is used to populate weights for B frames (wbp_weight) and then wp_weights are set to default
    for (i = 0; i < listXsize[LIST_0]; i++)
    {
      for (j = 0; j < listXsize[LIST_1]; j++)
      {
        for (comp = 0; comp < 3; comp++)
        {
          log_weight_denom = (comp == 0) ? luma_log_weight_denom : chroma_log_weight_denom;
          wbp_weight[0][i][j][comp] = wp_weight[0][i][comp];
          wbp_weight[1][i][j][comp] = wp_weight[1][j][comp];
        }
      }
    }    

//   In B Frames
//1.       Use Explicit weights for  Bi-Directional Prediction (wbp_weights)
//2.       Use Default  weights for Uni-Directional Prediction (wp_weights)

    for (clist=0; clist<2 + list_offset; clist++)
    {
      for (index = 0; index < listXsize[clist]; index++)
      {
        wp_weight[clist][index][0] = default_weight[0];
        wp_weight[clist][index][1] = default_weight[1];
        wp_weight[clist][index][2] = default_weight[2];
      }
    }

    for (clist=0; clist<=1;  clist++)
    {
      for (n = 0; n < listXsize[clist]; n++)
      {
        ref_pic       = listX[clist][n]->imgY_11;
        ref_pic_w     = listX[clist][n]->imgY_11_w;
        ref_qpic      = listX[clist][n]->imgY_ups;
        ref_qpic_w    = listX[clist][n]->imgY_ups_w;

        for (i = 0; i < img->height * img->width; i++)
        {
            ref_pic_w[i]      =   Clip3 (0, img->max_imgpel_value, (((int) ref_pic[i]     * wp_weight[clist][n][0] + wp_luma_round) >> luma_log_weight_denom) + wp_offset[clist][n][0]);
        }
        for (i = 0; i < 4*(img->height + 2*IMG_PAD_SIZE) ; i++)
        {
          for (j = 0; j< 4*(img->width + 2*IMG_PAD_SIZE); j++)
          {
            ref_qpic_w[i][j]  =   Clip3 (0, img->max_imgpel_value, (((int) ref_qpic[i][j] * wp_weight[clist][n][0] + wp_luma_round) >> luma_log_weight_denom) + wp_offset[clist][n][0]);        
          }
        }

      } //number of frames in list
    }   //number of lists
  }

}
   

double Calc_dc_org()
{
  double dc_org = 0;
  int i,j;
  for (i = 0; i < img->height; i++)
  {
    for (j = 0; j < img->width; j++)
    {
      dc_org += (double) imgY_org[i][j];
    }
  }
  return dc_org;
}

