/* 
  ====================================================================================================================

	The copyright in this software is being made available under the License included below. 
	This software may be subject to other third party and 	contributor rights, including patent rights, and no such 
	rights are granted under this license.
 
	Copyright (c) 2010, QUALCOMM INC.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification, are permitted only for 
	the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
	promoting such standards. The following conditions are required to be met:

		* Redistributions of source code must retain the above copyright notice, this list of conditions and 
		  the following disclaimer.
		* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
		  the following disclaimer in the documentation and/or other materials provided with the distribution.
		* The name of QUALCOMM INC.  may not be used to endorse or promote products derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
	INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  ====================================================================================================================
*/

#ifndef _ADAPTIVE_LOOP_FILTER_H_
#define _ADAPTIVE_LOOP_FILTER_H_

#include "global.h"
#include "mbuffer.h"
#ifdef ADAPTIVE_LOOP_FILTER
#include "alf_filter.h"
//------- defines
#define MAX_NUM_TAP    9
#define MAX_NUM_TAP_C  5
#define NUM_BIT_SHIFT  8
#define MAX_NUM_COEF   42
#define MAX_NUM_COEF_C 14
#define MAX_ALF_PARAM_SIZE 50000
#define MAX_NUM_BLOCK  60000
#define PYRAMIDAL_MEM_SIZE 341
typedef struct
{
  Boolean      valid_flag[PYRAMIDAL_MEM_SIZE];
  Boolean      part_flag[PYRAMIDAL_MEM_SIZE];
  Boolean      alf_blk_flag[PYRAMIDAL_MEM_SIZE];
} AlfQtBlock;
//------- end


//------- prototype
alf_parameter_set_rbsp_t *AllocALFPS (int *memory_size);
void   FreeALFPS (alf_parameter_set_rbsp_t *alfps);
int    InitALFGlobalBuffers(void);
void   FreeALFGlobalBurrers(void);
void   SetALFParameters(StorablePicture *p, alf_parameter_set_rbsp_t *alfps);
void   MakeExtendFrame( imgpel** ImgDec, imgpel** ImgExt, int width, int height, int offset );
void   SetRestFrame(imgpel** ImgRest, imgpel** ImgDec, int width, int height);
void   SetALFImgSize(ImageParameters *img, seq_parameter_set_rbsp_t *p);


int    GetALFBaseType(int slice_type, int ref);
void   InitFilterBase();
void   SetALFCoefChroma(StorablePicture *p);


//------- filtering for luma
void   AdaptiveLoopFilterProcess(StorablePicture *p);
void   FilteringFrame(StorablePicture *p, imgpel** ImgDec, imgpel** ImgRest);

//------- filtering for chroma
void   AdaptiveLoopFilterProcessChroma(StorablePicture *p);
void   FilteringFrameChroma(StorablePicture *p, imgpel ***ImgUVDec, imgpel ***ImgUVRest);
void   FilteringProcessChroma(imgpel ***ImgUVDec, imgpel ***ImgUVRest, int *qh, int tap, int chroma_id);

//------- filtering based block
void   BlockAdaptiveFilteringProcess(StorablePicture *p, imgpel** ImgDec, imgpel** ImgRest);
void   FilteringProcessBlock(int blk_addr, imgpel** ImgDec, imgpel** ImgRest, int *qh, int tap, int blk_size, int width_in_alf_blks);
void   CopyDecToRestBlock(int blk_addr, imgpel** ImgDec, imgpel** ImgRest, int blk_size, int width_in_alf_blks);
void   SetBlockSizeParameter(alf_parameter_set_rbsp_t *alfps, int size_idx);
int    CalcNumOfAlfBlocks(int blk_size, int width, int height);
void   GetBlockPosForBALF(int blk_addr, int *x, int *y, int width_in_blks);
void   GetSamplePosForBALF(int blk_addr, int *x, int *y, int blk_size, int width_in_blks);
int    ALFBlockIsAvailable(int blk_addr);
void   CheckAvailabilityOfNeighborsAlfBlock();
void   GetNeighbourAlfBlock(int curr_alfb_nr, int xN, int yN, PixelPos *pix);
void   InitBlockPartition(int pblk_addr, alf_parameter_set_rbsp_t *alfps);
void   TransQbpToFbp(int pblk_addr, alf_parameter_set_rbsp_t *alfps);
void   SetParentBlockSizeParam(alf_parameter_set_rbsp_t *alfps, int block_size);

//------- read ALF parameters
int    read_ALFParameterSet(alf_parameter_set_rbsp_t *alfps, DataPartition *p);
int    read_ALFChromaParam(alf_parameter_set_rbsp_t *alfps, Bitstream *s);
int    read_ALFBlockControlParam(alf_parameter_set_rbsp_t *alfps,  DataPartition *p);
int    read_BlockFlagRunLength(alf_parameter_set_rbsp_t *alfps, Bitstream *s);
int    read_BlockFlagCABAC(alf_parameter_set_rbsp_t *alfps, DataPartition *p);
int    read_QuadTreeBlockPartitioningParam(alf_parameter_set_rbsp_t *alfps, Bitstream *s);
void   read_ALFHeader();


#endif // ADAPTIVE_LOOP_FILTER

#endif // _ADAPTIVE_LOOP_FILTER_H_
