/* 
  ====================================================================================================================

	The copyright in this software is being made available under the License included below. 
	This software may be subject to other third party and 	contributor rights, including patent rights, and no such 
	rights are granted under this license.
 
	Copyright (c) 2010, QUALCOMM INC.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification, are permitted only for 
	the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
	promoting such standards. The following conditions are required to be met:

		* Redistributions of source code must retain the above copyright notice, this list of conditions and 
		  the following disclaimer.
		* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
		  the following disclaimer in the documentation and/or other materials provided with the distribution.
		* The name of QUALCOMM INC.  may not be used to endorse or promote products derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
	INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  ====================================================================================================================
*/
#include "alf_filter.h"
#include "vlc.h"

#ifdef ADAPTIVE_LOOP_FILTER

static int **filterCoeffTmp;
static int **filterCoeffSymTmp;

void initMatrix_int(int ***m2D, int d1, int d2)
{
  int i;

  if(!(*m2D = (int **) calloc(d1, sizeof(int *))))
    FATAL_ERROR_0("initMatrix_int: memory allocation problem\n", -1);
  if(!((*m2D)[0] = (int *) calloc(d1 * d2, sizeof(int))))
    FATAL_ERROR_0("initMatrix_int: memory allocation problem\n", -1);

  for(i = 1; i < d1; i++)
    (*m2D)[i] = (*m2D)[i-1] + d2;
}


void initMatrix_double(double ***m2D, int d1, int d2)
{
  int i;

  if(!(*m2D = (double **) calloc(d1, sizeof(double *))))
    FATAL_ERROR_0("initMatrix_double: memory allocation problem\n", -1);
  if(!((*m2D)[0] = (double *) calloc(d1 * d2, sizeof(double))))
    FATAL_ERROR_0("initMatrix_double: memory allocation problem\n", -1);

  for(i = 1; i < d1; i++)
    (*m2D)[i] = (*m2D)[i-1] + d2;
}


void initMatrix_imgpel(imgpel ***m2D, int d1, int d2)
{
  int i;

  if(!(*m2D = (imgpel **) calloc(d1, sizeof(imgpel *))))
    FATAL_ERROR_0("initMatrix_imgpel: memory allocation problem\n", -1);
  if(!((*m2D)[0] = (imgpel *) calloc(d1 * d2, sizeof(imgpel))))
    FATAL_ERROR_0("initMatrix_imgpel: memory allocation problem\n", -1);

  for(i = 1; i < d1; i++)
    (*m2D)[i] = (*m2D)[i-1] + d2;
}


void initMatrix3D_int(int ****m3D, int d1, int d2, int d3)
{
  int  j;

  if(!((*m3D) = (int ***) calloc(d1, sizeof(int **))))
    FATAL_ERROR_0("initMatrix3D_int: memory allocation problem\n", -1);

  for(j = 0; j < d1; j++)
    initMatrix_int((*m3D) + j, d2, d3);
}

void initMatrix3D_double(double ****m3D, int d1, int d2, int d3)
{
  int  j;

  if(!((*m3D) = (double ***) calloc(d1, sizeof(double **))))
    FATAL_ERROR_0("initMatrix3D_double: memory allocation problem\n", -1);

  for(j = 0; j < d1; j++)
    initMatrix_double((*m3D) + j, d2, d3);
}


void initMatrix3D_imgpel(imgpel ****m3D, int d1, int d2, int d3)
{
  int  j;

  if(!((*m3D) = (imgpel ***) calloc(d1, sizeof(imgpel **))))
    FATAL_ERROR_0("initMatrix3D_ushort: memory allocation problem\n", -1);

  for(j = 0; j < d1; j++)
    initMatrix_imgpel((*m3D) + j, d2, d3);
}


void initMatrix4D_int(int *****m4D, int d1, int d2, int d3, int d4)
{
  int  j;

  if(!((*m4D) = (int ****) calloc(d1, sizeof(int ***))))
    FATAL_ERROR_0("initMatrix4D_int: memory allocation problem\n", -1);

  for(j = 0; j < d1; j++)
    initMatrix3D_int((*m4D) + j, d2, d3, d4);
}

void initMatrix4D_double(double *****m4D, int d1, int d2, int d3, int d4)
{
  int  j;

  if(!((*m4D) = (double ****) calloc(d1, sizeof(double ***))))
    FATAL_ERROR_0("initMatrix4D_double: memory allocation problem\n", -1);

  for(j = 0; j < d1; j++)
    initMatrix3D_double((*m4D) + j, d2, d3, d4);
}

void initMatrix4D_imgpel(imgpel *****m4D, int d1, int d2, int d3, int d4)
{
  int  j;

  if(!((*m4D) = (imgpel ****) calloc(d1, sizeof(imgpel ***))))
    FATAL_ERROR_0("initMatrix4D_ushort: memory allocation problem\n", -1);

  for(j = 0; j < d1; j++)
    initMatrix3D_imgpel((*m4D) + j, d2, d3, d4);
}


void destroyMatrix_int(int **m2D)
{
  if(m2D)
  {
    if(m2D[0])
      free(m2D[0]);
    else
      FATAL_ERROR_0("destroyMatrix_int: memory free problem\n", -1);
    free(m2D);
  } 
  else
  {
    FATAL_ERROR_0("destroyMatrix_int: memory free problem\n", -1);
  }
}

void destroyMatrix_double(double **m2D)
{
  if(m2D)
  {
    if(m2D[0])
      free(m2D[0]);
    else
      FATAL_ERROR_0("destroyMatrix_double: memory free problem\n", -1);
    free(m2D);
  } 
  else
  {
    FATAL_ERROR_0("destroyMatrix_double: memory free problem\n", -1);
  }
}


void destroyMatrix_imgpel(imgpel **m2D)
{
  if(m2D)
  {
    if(m2D[0])
      free(m2D[0]);
    else
      FATAL_ERROR_0("destroyMatrix_imgpel: memory free problem\n", -1);
    free(m2D);
  } 
  else
  {
    FATAL_ERROR_0("destroyMatrix_imgpel: memory free problem\n", -1);
  }
}


void destroyMatrix3D_int(int ***m3D, int d1)
{
  int i;

  if(m3D)
  {
    for(i = 0; i < d1; i++)
      destroyMatrix_int(m3D[i]);
    free(m3D);
  } 
  else
  {
    FATAL_ERROR_0("destroyMatrix3D_int: memory free problem\n", -1);
  }
}

void destroyMatrix3D_double(double ***m3D, int d1)
{
  int i;

  if(m3D)
  {
    for(i = 0; i < d1; i++)
      destroyMatrix_double(m3D[i]);
    free(m3D);
  } 
  else
  {
    FATAL_ERROR_0("destroyMatrix3D_double: memory free problem\n", -1);
  }
}


void destroyMatrix3D_imgpel(imgpel ***m3D, int d1)
{
  int i;

  if(m3D)
  {
    for(i = 0; i < d1; i++)
      destroyMatrix_imgpel(m3D[i]);
    free(m3D);
  } 
  else
  {
    FATAL_ERROR_0("destroyMatrix3D_imgpel: memory free problem\n", -1);
  }
}


void destroyMatrix4D_int(int ****m4D, int d1, int d2)
{
  int  j;

  if(m4D)
  {
    for(j = 0; j < d1; j++)
      destroyMatrix3D_int(m4D[j], d2);
    free(m4D);
  } 
  else
  {
    FATAL_ERROR_0("destroyMatrix4D_int: memory free problem\n", -1);
  }
}


void destroyMatrix4D_double(double ****m4D, int d1, int d2)
{
  int  j;

  if(m4D)
  {
    for(j = 0; j < d1; j++)
      destroyMatrix3D_double(m4D[j], d2);
    free(m4D);
  } 
  else
  {
    FATAL_ERROR_0("destroyMatrix4D_double: memory free problem\n", -1);
  }
}

void destroyMatrix4D_imgpel(imgpel ****m4D, int d1, int d2)
{
  int  j;

  if(m4D)
  {
    for(j = 0; j < d1; j++)
      destroyMatrix3D_imgpel(m4D[j], d2);
    free(m4D);
  } 
  else
  {
    FATAL_ERROR_0("destroyMatrix4D_ushort: memory free problem\n", -1);
  }
}



int Clip3Fun_post(int low, int high, int val)
{
  return ((val < low)? low: (val > high)? high: val);
}
imgpel Clip_post(int high, int val)
{
  return (imgpel)(((val > high)? high: val));
}

void initImg(imgStruct *imgIn, int height, int width)
{
  initMatrix_imgpel(&(imgIn->imgY), height, width);
  initMatrix3D_imgpel(&(imgIn->imgUV), 2, height/2, width/2);
  imgIn->img_height = height;
  imgIn->img_width = width;
}



void padImage(imgpel **imgY,  imgpel **imgY_pad, int fl, int img_height, int img_width)
{
  int x, y;
  int x_size = img_width + fl;
  int x_size2= img_width + 2 * fl;
  int y_size = img_height + fl;
  int y_size2= img_height + 2 * fl;
  for(y = 0; y < img_height; ++y)
  {
    memcpy(imgY_pad[y+fl]+fl,imgY[y],sizeof(imgpel)* img_width);
  }
  for(x = fl; x < img_width + fl; ++x) 
  {
    imgpel temp=imgY_pad[fl][x];
    for(y = 0; y < fl; ++y)
    {
      imgY_pad[y][x] = temp;
      //imgY_pad[y][x] = imgY_pad[fl][x];
    }
    temp=imgY_pad[y_size - 1][x];
    for(y = y_size; y < y_size2; ++y)
      imgY_pad[y][x] = temp;
    //imgY_pad[y][x] = imgY_pad[img_height + fl - 1][x];
  }

  for(y = 0; y < img_height + 2 * fl; ++y)
  {
    imgpel temp=imgY_pad[y][fl];
    imgpel *p_img= imgY_pad[y];
    for(x = 0; x < fl; ++x)
      *(p_img++) = temp;
    temp=imgY_pad[y][x_size - 1];
    p_img= &(imgY_pad[y][x_size]);
    for(x = x_size; x < x_size2; ++x)
      *(p_img++)  = temp;
  }
}


void  calcVar(imgpel **imgY_var, imgpel **imgY_pad,int pad_size, int fl, int img_height, int img_width)
{
  int i, j, ii, jj;
  int sum;
  int *p_imgY_temp;
  int shift= (11+img->BitDepthIncrease);
  int fl2plusOne= (VAR_SIZE<<1)+1;
  int pad_offset = pad_size-fl-1;

  if (VAR_SIZE ==0)
  {
    imgpel *p_imgY_var;
    shift = img->BitDepthIncrease;
    //current
    for(i = 1; i < img_height + fl2plusOne; i++)
    {
      imgpel *p_imgY_pad = &imgY_pad[pad_offset+i][pad_offset];
      imgpel *p_imgY_pad_up   = &imgY_pad[pad_offset+i+1][pad_offset];
      imgpel *p_imgY_pad_down = &imgY_pad[pad_offset+i-1][pad_offset];
      p_imgY_temp = &imgY_temp[i-1][0];
      p_imgY_var  = &imgY_var [i-1][0];
      for(j = 1; j < img_width +fl2plusOne; j++)	
      {
        *(p_imgY_temp) = abs((p_imgY_pad[j]<<1) - p_imgY_pad[j+1] - p_imgY_pad[j-1])+
          abs((p_imgY_pad[j]<<1) - p_imgY_pad_down[j] - p_imgY_pad_up[j]);
        *(p_imgY_var++) =(imgpel) Clip_post(var_max, (int) ((*(p_imgY_temp++))>>shift));
      }
    }
    return;
  }


  for(i = 1; i < img_height + fl2plusOne; i++)
  {
    imgpel *p_imgY_pad = &imgY_pad[pad_offset+i][pad_offset];
    imgpel *p_imgY_pad_up   = &imgY_pad[pad_offset+i+1][pad_offset];
    imgpel *p_imgY_pad_down = &imgY_pad[pad_offset+i-1][pad_offset];
    p_imgY_temp = &imgY_temp[i-1][0];
    for(j = 1; j < img_width +fl2plusOne; j++)	
    {
      *(p_imgY_temp++) = abs((p_imgY_pad[j]<<1) - p_imgY_pad[j+1] - p_imgY_pad[j-1])+
        abs((p_imgY_pad[j]<<1) - p_imgY_pad_down[j] - p_imgY_pad_up[j]);
    }
  }
  {
    //int temp;
    int sum_0=0;
    int y=img_height+((VAR_SIZE+1)<<1);

    int *p_imgY_temp_sum;
    i = fl;
    j = fl;

    memset(imgY_temp[y],0,sizeof(int)*(img_width+((VAR_SIZE+1)<<1)));
    //--------------------------------------------------------------------------------------------

    for(ii = i - fl; ii <= i + fl; ii++)
    {
      p_imgY_temp= &imgY_temp[ii][j - fl];
      p_imgY_temp_sum=&imgY_temp[y][j - fl];
      for(jj = j - fl; jj <= j + fl; jj++)
      {
        *(p_imgY_temp_sum++) += *(p_imgY_temp++);
      }
    }
    p_imgY_temp_sum=&imgY_temp[y][j - fl];
    for(jj = j - fl; jj <= j + fl; jj++)
      sum_0+=*(p_imgY_temp_sum++);
    imgY_var[i - fl][j - fl] = (imgpel) Clip_post(var_max, (int) ((sum_0 * mult_fact_int)>>shift));
    //--------------------------------------------------------------------------------------------
    sum = sum_0;
    for(j = fl+1; j < img_width + fl; ++j)
    {	
      int k=j+fl;
      for(ii = i - fl; ii <= i + fl; ii++)
        imgY_temp[y][k] += (imgY_temp[ii][k]);

      sum += (imgY_temp[y][k]-imgY_temp[y][j - fl-1]);
      imgY_var[i - fl][j - fl] = (imgpel) Clip_post(var_max, (int) ((sum * mult_fact_int)>>shift));

    }
    //--------------------------------------------------------------------------------------------

    for(i = fl+1; i < img_height + fl; ++i)
    {
      imgpel  *pimgY_var= &imgY_var[i-fl][0];
      int *p_imgY_temp1;
      int *p_imgY_temp2;
      sum = sum_0;	
      j= fl;
      p_imgY_temp1= &imgY_temp[i+fl  ][j - fl];
      p_imgY_temp2= &imgY_temp[i-fl-1][j - fl];
      p_imgY_temp_sum=&imgY_temp[y][j - fl];
      for(jj = j - fl; jj <= j + fl; jj++)
      {
        int diff = *(p_imgY_temp1++)-*(p_imgY_temp2++);
        *(p_imgY_temp_sum++) += diff;
        sum += diff;
      }
      sum_0=sum;

      *(pimgY_var++) = (imgpel) Clip_post(var_max, (int) ((sum * mult_fact_int)>>shift));
      //--------------------------------------------------------------------------------------------
      p_imgY_temp_sum=imgY_temp[y];
      for(j = fl+1; j < img_width + fl; ++j)
      {	
        int k = j+fl;
        p_imgY_temp_sum[k] += (imgY_temp[i + fl][k]-imgY_temp[i-fl-1][k]);
        sum += (p_imgY_temp_sum[k]-p_imgY_temp_sum[j-fl-1]);
        *(pimgY_var++) = (imgpel) Clip_post(var_max, (int) ((sum * mult_fact_int)>>shift));
      }
    }
  }

}


void calcVar_block(imgpel **imgY_var, imgpel **imgY_pad,int pad_size, int fl, int img_height, int img_width
                   , int h_start, int h_end, int w_start,int w_end,int blk_addr,StorablePicture *p,int num_blk)
{
  static int pre_w_end=0;
  int i, j, ii, jj;
  int sum;
  int *p_imgY_temp;
  int fl2=fl<<1;
  int shift= (11+img->BitDepthIncrease);
  int fl2plusOne= (VAR_SIZE<<1)+1;
  int pad_offset = pad_size-fl-1;


  if (fl ==0)
  {
    imgpel *p_imgY_var;
    shift = img->BitDepthIncrease;

    if (!p->alf_blk_data[blk_addr].alfbAvailD || p->alf_blk_data[p->alf_blk_data[blk_addr].alfbAddrD].alf_blk_flag==0) 
    {
      for(i = h_start+1; i < h_start+1; i++)
      {
        imgpel *p_imgY_pad = &imgY_pad[pad_offset+i][pad_offset];
        imgpel *p_imgY_pad_up   = &imgY_pad[pad_offset+i+1][pad_offset];
        imgpel *p_imgY_pad_down = &imgY_pad[pad_offset+i-1][pad_offset];

        p_imgY_temp = &imgY_temp[i-1][w_start];
        p_imgY_var  = &imgY_var[i-1][w_start];
        for(j = w_start+1; j < w_start+1; j++)	
        {
          *(p_imgY_temp) = abs((p_imgY_pad[j]<<1) - p_imgY_pad[j+1] - p_imgY_pad[j-1])+
            abs((p_imgY_pad[j]<<1) - p_imgY_pad_down[j] - p_imgY_pad_up[j]);
          *(p_imgY_var++) =(imgpel) Clip_post(var_max, (int) ((*(p_imgY_temp++))>>shift));

        }
      }
    }
    if (w_end!=img_width)
    {
      for (ii = w_start,jj=0;ii<w_end;ii+= p->alf_block_size,jj++)
      {
        //upper
        if (!p->alf_blk_data[blk_addr+jj].alfbAvailB || p->alf_blk_data[p->alf_blk_data[blk_addr+jj].alfbAddrB].alf_blk_flag==0) 
        {
          int iii = ii+ 1;
          for(i = h_start+1; i < h_start + 1; ++i)
          {
            imgpel *p_imgY_pad = &imgY_pad[pad_offset+i][pad_offset];
            imgpel *p_imgY_pad_up   = &imgY_pad[pad_offset+i+1][pad_offset];
            imgpel *p_imgY_pad_down = &imgY_pad[pad_offset+i-1][pad_offset];

            p_imgY_temp = &imgY_temp[i-1][ii];
            p_imgY_var  = &imgY_var[i-1][ii];
            for(j = iii; j < iii+ p->alf_block_size; j++)	
            {
              *(p_imgY_temp) = abs((p_imgY_pad[j]<<1) - p_imgY_pad[j+1] - p_imgY_pad[j-1])+
                abs((p_imgY_pad[j]<<1) - p_imgY_pad_down[j] - p_imgY_pad_up[j]);
              *(p_imgY_var++) =(imgpel) Clip_post(var_max, (int) ((*(p_imgY_temp++))>>shift));
            }
          }
        }
      }
    }
    else
    {
      for (ii = w_start,jj=0;jj<num_blk-1;ii+= p->alf_block_size,jj++)
      {
        //upper
        if (!p->alf_blk_data[blk_addr+jj].alfbAvailB || p->alf_blk_data[p->alf_blk_data[blk_addr+jj].alfbAddrB].alf_blk_flag==0) 
        {
          int iii = ii+ 1;
          for(i = h_start+1; i < h_start +1; ++i)
          {
            imgpel *p_imgY_pad = &imgY_pad[pad_offset+i][pad_offset];
            imgpel *p_imgY_pad_up   = &imgY_pad[pad_offset+i+1][pad_offset];
            imgpel *p_imgY_pad_down = &imgY_pad[pad_offset+i-1][pad_offset];

            p_imgY_temp = &imgY_temp[i-1][ii];
            p_imgY_var  = &imgY_var[i-1][ii];
            for(j = iii; j < iii+p->alf_block_size; j++)	
            {
              *(p_imgY_temp) = abs((p_imgY_pad[j]<<1) - p_imgY_pad[j+1] - p_imgY_pad[j-1])+
                abs((p_imgY_pad[j]<<1) - p_imgY_pad_down[j] - p_imgY_pad_up[j]);
              *(p_imgY_var++) =(imgpel) Clip_post(var_max, (int) ((*(p_imgY_temp++))>>shift));
            }
          }
        }
      }
      jj= num_blk-1;
      ii = w_start+jj*p->alf_block_size;
      {
        //upper
        if (!p->alf_blk_data[blk_addr+jj].alfbAvailB || p->alf_blk_data[p->alf_blk_data[blk_addr+jj].alfbAddrB].alf_blk_flag==0) 
        {
          for(i = h_start+1; i < h_start + 1; ++i)
          {
            imgpel *p_imgY_pad = &imgY_pad[pad_offset+i][pad_offset];
            imgpel *p_imgY_pad_up   = &imgY_pad[pad_offset+i+1][pad_offset];
            imgpel *p_imgY_pad_down = &imgY_pad[pad_offset+i-1][pad_offset];

            p_imgY_temp = &imgY_temp[i-1][ii];
            p_imgY_var  = &imgY_var[i-1][ii];
            for(j = ii+ 1; j < w_end+1; j++)	
            {
              *(p_imgY_temp) = abs((p_imgY_pad[j]<<1) - p_imgY_pad[j+1] - p_imgY_pad[j-1])+
                abs((p_imgY_pad[j]<<1) - p_imgY_pad_down[j] - p_imgY_pad_up[j]);
              *(p_imgY_var++) =(imgpel) Clip_post(var_max, (int) ((*(p_imgY_temp++))>>shift));
            }
          }
        }
      }
    }
    //left
    if (!p->alf_blk_data[blk_addr].alfbAvailA || p->alf_blk_data[p->alf_blk_data[blk_addr].alfbAddrA].alf_blk_flag==0) 
    {
      for(i = h_start+1; i < h_end +1; i++)
      {
        imgpel *p_imgY_pad = &imgY_pad[pad_offset+i][pad_offset];
        imgpel *p_imgY_pad_up   = &imgY_pad[pad_offset+i+1][pad_offset];
        imgpel *p_imgY_pad_down = &imgY_pad[pad_offset+i-1][pad_offset];

        p_imgY_temp = &imgY_temp[i-1][w_start];
        p_imgY_var  = &imgY_var[i-1][w_start];
        for(j = w_start+1; j < w_start +1; j++)	
        {
          *(p_imgY_temp) = abs((p_imgY_pad[j]<<1) - p_imgY_pad[j+1] - p_imgY_pad[j-1])+
            abs((p_imgY_pad[j]<<1) - p_imgY_pad_down[j] - p_imgY_pad_up[j]);
          *(p_imgY_var++) =(imgpel) Clip_post(var_max, (int) ((*(p_imgY_temp++))>>shift));
        }
      }
    }

    for(i = h_start+1; i < h_end + 1; i++)
    {
      imgpel *p_imgY_pad = &imgY_pad[pad_offset+i][pad_offset];
      imgpel *p_imgY_pad_up   = &imgY_pad[pad_offset+i+1][pad_offset];
      imgpel *p_imgY_pad_down = &imgY_pad[pad_offset+i-1][pad_offset];
      p_imgY_temp = &imgY_temp[i-1][w_start];
      p_imgY_var = &imgY_var[i-1][w_start];
      for(j = w_start+ 1; j < w_end +1; j++)	
      {
        *(p_imgY_temp) = abs((p_imgY_pad[j]<<1) - p_imgY_pad[j+1] - p_imgY_pad[j-1])+
          abs((p_imgY_pad[j]<<1) - p_imgY_pad_down[j] - p_imgY_pad_up[j]);
        *(p_imgY_var++) =(imgpel) Clip_post(var_max, (int) ((*(p_imgY_temp++))>>shift));
      }
    }
    return;
  }


  if (!p->alf_blk_data[blk_addr].alfbAvailD || p->alf_blk_data[p->alf_blk_data[blk_addr].alfbAddrD].alf_blk_flag==0) 
  {
    for(i = h_start+1; i < h_start+fl2plusOne; i++)
    {
      imgpel *p_imgY_pad = &imgY_pad[pad_offset+i][pad_offset];
      imgpel *p_imgY_pad_up   = &imgY_pad[pad_offset+i+1][pad_offset];
      imgpel *p_imgY_pad_down = &imgY_pad[pad_offset+i-1][pad_offset];

      p_imgY_temp = &imgY_temp[i-1][w_start];
      for(j = w_start+1; j < w_start+fl2plusOne; j++)	
      {
        *(p_imgY_temp++) = abs((p_imgY_pad[j]<<1) - p_imgY_pad[j+1] - p_imgY_pad[j-1])+
          abs((p_imgY_pad[j]<<1) - p_imgY_pad_down[j] - p_imgY_pad_up[j]);
      }
    }
  }
  if (w_end!=img_width)
  {
    for (ii = w_start,jj=0;ii<w_end;ii+= p->alf_block_size,jj++)
    {
      //upper
      if (!p->alf_blk_data[blk_addr+jj].alfbAvailB || p->alf_blk_data[p->alf_blk_data[blk_addr+jj].alfbAddrB].alf_blk_flag==0) 
      {
        int iii = ii+ fl2plusOne;
        for(i = h_start+1; i < h_start + fl2plusOne; ++i)
        {
          imgpel *p_imgY_pad = &imgY_pad[pad_offset+i][pad_offset];
          imgpel *p_imgY_pad_up   = &imgY_pad[pad_offset+i+1][pad_offset];
          imgpel *p_imgY_pad_down = &imgY_pad[pad_offset+i-1][pad_offset];

          p_imgY_temp = &imgY_temp[i-1][ii+ fl2];
          for(j = iii; j < iii+ p->alf_block_size; j++)	
          {
            *(p_imgY_temp++) = abs((p_imgY_pad[j]<<1) - p_imgY_pad[j+1] - p_imgY_pad[j-1])+
              abs((p_imgY_pad[j]<<1) - p_imgY_pad_down[j] - p_imgY_pad_up[j]);
          }
        }
      }
    }
  }
  else
  {
    for (ii = w_start,jj=0;jj<num_blk-1;ii+= p->alf_block_size,jj++)
    {

      if (!p->alf_blk_data[blk_addr+jj].alfbAvailB || p->alf_blk_data[p->alf_blk_data[blk_addr+jj].alfbAddrB].alf_blk_flag==0) 
      {
        int iii = ii+ fl2plusOne;
        for(i = h_start+1; i < h_start +fl2plusOne; ++i)
        {
          imgpel *p_imgY_pad = &imgY_pad[pad_offset+i][pad_offset];
          imgpel *p_imgY_pad_up   = &imgY_pad[pad_offset+i+1][pad_offset];
          imgpel *p_imgY_pad_down = &imgY_pad[pad_offset+i-1][pad_offset];

          p_imgY_temp = &imgY_temp[i-1][ii+ fl2];
          for(j = iii; j < iii+p->alf_block_size; j++)	
          {
            *(p_imgY_temp++) = abs((p_imgY_pad[j]<<1) - p_imgY_pad[j+1] - p_imgY_pad[j-1])+
              abs((p_imgY_pad[j]<<1) - p_imgY_pad_down[j] - p_imgY_pad_up[j]);
          }
        }
      }
    }
    jj= num_blk-1;
    ii = w_start+jj*p->alf_block_size;
    {
      //upper
      if (!p->alf_blk_data[blk_addr+jj].alfbAvailB || p->alf_blk_data[p->alf_blk_data[blk_addr+jj].alfbAddrB].alf_blk_flag==0) 
      {
        for(i = h_start+1; i < h_start + fl2plusOne; ++i)
        {
          imgpel *p_imgY_pad = &imgY_pad[pad_offset+i][pad_offset];
          imgpel *p_imgY_pad_up   = &imgY_pad[pad_offset+i+1][pad_offset];
          imgpel *p_imgY_pad_down = &imgY_pad[pad_offset+i-1][pad_offset];

          p_imgY_temp = &imgY_temp[i-1][ii+ fl2];
          for(j = ii+ fl2plusOne; j < w_end+fl2plusOne; j++)	
          {
            *(p_imgY_temp++) = abs((p_imgY_pad[j]<<1) - p_imgY_pad[j+1] - p_imgY_pad[j-1])+
              abs((p_imgY_pad[j]<<1) - p_imgY_pad_down[j] - p_imgY_pad_up[j]);
          }
        }
      }
    }
  }


  if (!p->alf_blk_data[blk_addr].alfbAvailA || p->alf_blk_data[p->alf_blk_data[blk_addr].alfbAddrA].alf_blk_flag==0) 
  {
    for(i = h_start+fl2plusOne; i < h_end +fl2plusOne; i++)
    {
      imgpel *p_imgY_pad = &imgY_pad[pad_offset+i][pad_offset];
      imgpel *p_imgY_pad_up   = &imgY_pad[pad_offset+i+1][pad_offset];
      imgpel *p_imgY_pad_down = &imgY_pad[pad_offset+i-1][pad_offset];

      p_imgY_temp = &imgY_temp[i-1][w_start];
      for(j = w_start+1; j < w_start +fl2plusOne; j++)	
      {
        *(p_imgY_temp++) = abs((p_imgY_pad[j]<<1) - p_imgY_pad[j+1] - p_imgY_pad[j-1])+
          abs((p_imgY_pad[j]<<1) - p_imgY_pad_down[j] - p_imgY_pad_up[j]);
      }
    }
  }

  for(i = h_start+fl2plusOne; i < h_end + fl2plusOne; i++)
  {
    p_imgY_temp = &imgY_temp[i-1][w_start+ fl2];
    for(j = w_start+ fl2plusOne; j < w_end +fl2plusOne; j++)	
    {
      imgpel *p_imgY_pad = &imgY_pad[pad_offset+i][pad_offset];
      imgpel *p_imgY_pad_up   = &imgY_pad[pad_offset+i+1][pad_offset];
      imgpel *p_imgY_pad_down = &imgY_pad[pad_offset+i-1][pad_offset];

      *(p_imgY_temp++) = abs((p_imgY_pad[j]<<1) - p_imgY_pad[j+1] - p_imgY_pad[j-1])+
        abs((p_imgY_pad[j]<<1) - p_imgY_pad_down[j] - p_imgY_pad_up[j]);
    }
  }
  {
    int sum_0=0;
    int y=img_height+((VAR_SIZE+1)<<1);	
    int *p_imgY_temp_sum;
    i = h_start+fl;
    j = w_start+fl;

    //--------------------------------------------------------------------------------------------
    if ((p->alf_blk_data[blk_addr].alfbAvailB && p->alf_blk_data[p->alf_blk_data[blk_addr].alfbAddrB].alf_blk_flag==1)&&
      (p->alf_block_size>4||(w_start - pre_w_end<=0)||(w_start - pre_w_end > fl+1))) 
    {
      int *p_imgY_temp1= &imgY_temp[i+fl  ][j - fl];
      int *p_imgY_temp2= &imgY_temp[i-fl-1][j - fl];

      p_imgY_temp_sum=&imgY_temp[y][j - fl];
      for(jj = j - fl; jj <= j + fl; jj++)
      {
        *(p_imgY_temp_sum++) += *(p_imgY_temp1++)-*(p_imgY_temp2++);
      }
      p_imgY_temp_sum=&imgY_temp[y][j - fl];
      for(jj = j - fl; jj <= j + fl; jj++)
        sum_0+=*(p_imgY_temp_sum++);
    }
    else
    {
      memset(&imgY_temp[y][w_start],0,sizeof(int)*((fl<<1)+1));

      for(ii = i - fl; ii <= i + fl; ii++)
      {
        p_imgY_temp= &imgY_temp[ii][j - fl];
        p_imgY_temp_sum=&imgY_temp[y][j - fl];
        for(jj = j - fl; jj <= j + fl; jj++)
        {
          *(p_imgY_temp_sum++) += *(p_imgY_temp++);
        }
      }	
      p_imgY_temp_sum=&imgY_temp[y][j - fl];
      for(jj = j - fl; jj <= j + fl; jj++)
        sum_0+=*(p_imgY_temp_sum++);
    }

    imgY_var[i - fl][j - fl] = (imgpel) Clip_post(var_max, (int) ((sum_0 * mult_fact_int)>>shift));
    //--------------------------------------------------------------------------------------------
    sum = sum_0;
    for(j = w_start+fl+1; j < w_end + fl; ++j)
    {	
      int k=j+fl;
      imgY_temp[y][k]=0;
      for(ii = i - fl; ii <= i + fl; ii++)
        imgY_temp[y][k] += (imgY_temp[ii][k]);
      sum += (imgY_temp[y][k]-imgY_temp[y][j - fl-1]);

      imgY_var[i - fl][j - fl] = (imgpel) Clip_post(var_max, (int) ((sum * mult_fact_int)>>shift));

    }
    //--------------------------------------------------------------------------------------------

    for(i = h_start+fl+1; i < h_end + fl; ++i)
    {
      imgpel  *pimgY_var= &imgY_var[i-fl][w_start];
      int *p_imgY_temp1;
      int *p_imgY_temp2;
      sum = sum_0;	
      j= w_start+fl;
      p_imgY_temp1= &imgY_temp[i+fl  ][j - fl];
      p_imgY_temp2= &imgY_temp[i-fl-1][j - fl];
      p_imgY_temp_sum=&imgY_temp[y][j - fl];
      for(jj = j - fl; jj <= j + fl; jj++)
      {
        int diff = *(p_imgY_temp1++)-*(p_imgY_temp2++);
        *(p_imgY_temp_sum++) += diff;
        sum += diff;
      }
      sum_0=sum;

      *(pimgY_var++) = (imgpel) Clip_post(var_max, (int) ((sum * mult_fact_int)>>shift));
      //--------------------------------------------------------------------------------------------
      p_imgY_temp_sum=imgY_temp[y];
      for(j = w_start+fl+1; j < w_end + fl; ++j)
      {	
        int k = j+fl;
        p_imgY_temp_sum[k] += (imgY_temp[i + fl][k]-imgY_temp[i-fl-1][k]);
        sum += (p_imgY_temp_sum[k]-p_imgY_temp_sum[j-fl-1]);

        *(pimgY_var++) = (imgpel) Clip_post(var_max, (int) ((sum * mult_fact_int)>>shift));
      }
    }
  }
  pre_w_end = w_end;
}


int depthInt9x9Sym[] = 
{
  5, 
  5, 6, 5, 
  5, 6, 7, 6, 5,
  5, 6, 7, 8, 7, 6, 5,
  5, 6, 7, 8, 9, 9 
};

int depthInt7x7Sym[] = 
{
  4, 
  4, 5, 4, 
  4, 5, 6, 5, 4, 
  4, 5, 6, 7, 7 
};

int depthInt5x5Sym[] = 
{
  3,   
  3, 4, 3,
  3, 4, 5, 5  
};

int *pDepthIntTab[3]={
  depthInt5x5Sym, 
  depthInt7x7Sym, 
  depthInt9x9Sym, 
};

// from http://en.wikipedia.org/wiki/Golomb_coding
// M=2^k
//
int golombDecode(int k, Bitstream *bitstream, char *tracestring)
{
  int q = 0;
  int nr = 0;
  int m = (int)pow(2.0, k);
  int a;

  while(u_1("full-pel filter", bitstream)) 
    q++;
  for(a = 0; a < k; ++a)          // read out the sequential log2(M) bits
    if(u_1("full-pel filter", bitstream))
      nr += 1 << a;
  nr += q * m;                    // add the bits and the multiple of M
  if(nr != 0)
    nr = (u_1("full-pel filter", bitstream))? nr: -nr;
  return nr; 
}

void readPredFlags(pfStruct *pf, Bitstream *bitstream)
{
  int ind;

  pf->force0 = u_1("force0", bitstream);
  if(pf->force0)
    for(ind = 0; ind < pf->filters_per_group; ++ind)
      pf->codedVarBins[ind] = u_1("codedVarBins[i]", bitstream);
  pf->predMethod = u_1("predMethod", bitstream);
}



void readFilterCodingParams(pfStruct *pf, Bitstream *bitstream)
{
  int ind, scanPos;
  int golombIndexBit;
  int kMin;
  int maxScanVal;
  int *pDepthInt;
  int fl;

  // Determine fl
  if(pf->sqrFiltLength == SQR_FILT_LENGTH_9SYM)
    fl = 4;
  else if(pf->sqrFiltLength == SQR_FILT_LENGTH_7SYM)
    fl = 3;
  else
    fl = 2;

  // Determine maxScanVal
  maxScanVal = 0;
  pDepthInt = pDepthIntTab[fl - 2];
  for(ind = 0; ind < pf->sqrFiltLength; ind++)
    maxScanVal = max(maxScanVal, pDepthInt[ind]);

  // Golomb parameters
  pf->minKStart = 1 + u_v(3, "expGolomb index", bitstream);

  kMin = pf->minKStart;
  for(scanPos = 0; scanPos < maxScanVal; scanPos++)
  {
    golombIndexBit = u_1("expGolomb index", bitstream);
    if(golombIndexBit)
      pf->kMinTab[scanPos] = kMin + 1;
    else
      pf->kMinTab[scanPos] = kMin;
    kMin = pf->kMinTab[scanPos];
  }
}


void readFilterCoeffs(pfStruct *pf,Bitstream *bitstream)
{
  int ind, scanPos, i;
  int *pDepthInt;
  int fl;

  if(pf->sqrFiltLength == SQR_FILT_LENGTH_9SYM)
    fl = 4;
  else if(pf->sqrFiltLength == SQR_FILT_LENGTH_7SYM)
    fl = 3;
  else
    fl = 2;

  pDepthInt = pDepthIntTab[fl - 2];

  for(ind = 0; ind < pf->savedFilters; ++ind)
  {
    for(i = 0; i < pf->sqrFiltLength; i++)
    {	
      scanPos = pDepthInt[i] - 1;
      pf->diffQFilterCoeffIntPP[ind][i] = golombDecode(pf->kMinTab[scanPos], bitstream, "full-pel filter");
    }
  }
}

void decodeAuxInfo(pfStruct *pf, int frNo, Bitstream *bitstream)
{
  int i;
  int filterPattern[NO_VAR_BINS] = {0};
  int predPattern;
  int tmp;
  int sqrFiltLengthTab[] = {SQR_FILT_LENGTH_9SYM, SQR_FILT_LENGTH_7SYM, SQR_FILT_LENGTH_5SYM};

  pf->nonZeroCoeffs = 1;

  if(frNo > 0)
    pf->nonZeroCoeffs = u_1("Nonzero coefficients", bitstream);


  {

    tmp = u_1("codtab_filtNo", bitstream);
    if(tmp)
    {
      pf->sqrFiltLength = sqrFiltLengthTab[2];
    }
    else
    {
      tmp += u_1("codtab_filtNo", bitstream);
      pf->sqrFiltLength = sqrFiltLengthTab[tmp];
    }
  }
  if(pf->nonZeroCoeffs)
  {


    tmp = u_1("codtab_noFilters", bitstream);
    if(tmp)
    {
      pf->filters_per_group = 1;
    }
    else
    {
      tmp = u_1("codtab_noFilters", bitstream);
      if(tmp)
        pf->filters_per_group = 2;
      else
        pf->filters_per_group = 3;    // More than two filters
    }

    if(pf->filters_per_group == 2)
    {
      int startSecondFilter = u_v (4, "startSecondFilter", bitstream);
      filterPattern[startSecondFilter] = 1;
    }
    else if(pf->filters_per_group > 2)
    {
      pf->filters_per_group = 1;
      for(i = 1; i < NO_VAR_BINS; ++i)
      {
        filterPattern[i] = u_1("filterPattern[i]", bitstream);
        pf->filters_per_group += filterPattern[i];
      }
    }
    else
    {
      memset(filterPattern, 0, NO_VAR_BINS * sizeof(int));
    }
  }

  // Reconstruct varIndTab[]
  memset(pf->varIndTab, 0, NO_VAR_BINS * sizeof(int));
  if(pf->nonZeroCoeffs)
  {
    for(i = 1; i < NO_VAR_BINS; ++i)
      if(filterPattern[i])
        pf->varIndTab[i] = pf->varIndTab[i-1] + 1;
      else
        pf->varIndTab[i] = pf->varIndTab[i-1];
  }

  if(frNo > 0)
  {
    pf->prevFiltInd = u_v(4, "prevFiltInd", bitstream);
    tmp = u_1("codtab_pred", bitstream);
    if(tmp)
    {
      predPattern = 0;
    }
    else
    {
      tmp = u_1("codtab_pred", bitstream);
      if(tmp)
        predPattern = 1;
      else
        predPattern = 2;
    }

    // Reconstruct usePrevFiltBest[]
    if(predPattern == 0)
    {
      for(i = 0; i < NO_VAR_BINS; ++i)
        pf->usePrevFiltBest[i] = 1;
    } 
    else if(predPattern == 1)
    {
      pf->usePrevFiltBest[0] = 0;
      for(i = 1; i < NO_VAR_BINS; ++i)
        pf->usePrevFiltBest[i] = 1;
    }
    else
    {
      for(i = 0; i < NO_VAR_BINS; ++i)
        pf->usePrevFiltBest[i] = u_1("usePrevFiltBest[i]", bitstream);
    }
  }
}

void init_filterCoeff()
{
  initMatrix_int(&filterCoeffTmp, NO_VAR_BINS, MAX_SQR_FILT_LENGTH);      // initialized and set to zero
  initMatrix_int(&filterCoeffSymTmp, NO_VAR_BINS, MAX_SQR_FILT_LENGTH);   // initialized and set to zero
}
void delete_filterCoeff()
{
  destroyMatrix_int(filterCoeffTmp); 
  destroyMatrix_int(filterCoeffSymTmp); 
}

void reconstructFilterCoeffs(pfStruct *pf,int **pfilterCoeffSym, int bit_depth)
{
  int i, src, ind;
  int factor = (1 << (bit_depth - 1));

  // Copy non zero filters in filterCoeffTmp
  for(ind = 0; ind < pf->savedFilters; ++ind)
  {
    for(i = 0; i < pf->sqrFiltLength; i++)
      filterCoeffTmp[ind][i] = pf->diffQFilterCoeffIntPP[ind][i];
  }
  // Undo prediction
  for(ind = 0; ind < pf->savedFilters; ++ind)
  {
    if((!pf->predMethod) || (ind == 0)) 
    {
      memcpy(filterCoeffSymTmp[ind],filterCoeffTmp[ind],sizeof(int)*pf->sqrFiltLength);
    }
    else
    {
      // Prediction
      for(i = 0; i < pf->sqrFiltLength; ++i)
        filterCoeffSymTmp[ind][i] = (int)(filterCoeffTmp[ind][i] + filterCoeffSymTmp[ind - 1][i]);
    }
  }

  // Inverse quantization
  // Add filters forced to zero
  if(pf->force0)
  {
    assert(pf->savedFilters < pf->filters_per_group);
    src = 0;
    for(ind = 0; ind < pf->filters_per_group; ++ind)
    {
      if(pf->codedVarBins[ind])
      {
        memcpy(pfilterCoeffSym[ind],filterCoeffSymTmp[src],sizeof(int)*pf->sqrFiltLength);
        ++src;
      }
      else
      {
        memset(pfilterCoeffSym[ind],0,sizeof(int)*pf->sqrFiltLength);
      }
    }
    assert(src == pf->savedFilters);
  }
  else
  {
    assert(pf->savedFilters == pf->filters_per_group);
    for(ind = 0; ind < pf->filters_per_group; ++ind)
      memcpy(pfilterCoeffSym[ind],filterCoeffSymTmp[ind],sizeof(int)*pf->sqrFiltLength);
  }
}
#endif