/* 
  ====================================================================================================================

	The copyright in this software is being made available under the License included below. 
	This software may be subject to other third party and 	contributor rights, including patent rights, and no such 
	rights are granted under this license.
 
	Copyright (c) 2010, QUALCOMM INC.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification, are permitted only for 
	the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
	promoting such standards. The following conditions are required to be met:

		* Redistributions of source code must retain the above copyright notice, this list of conditions and 
		  the following disclaimer.
		* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
		  the following disclaimer in the documentation and/or other materials provided with the distribution.
		* The name of QUALCOMM INC.  may not be used to endorse or promote products derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
	INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  ====================================================================================================================
*/
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "global.h"
#include "image.h"
#include "mb_access.h"
#include "loopfilter.h"

byte mixedModeEdgeFlag, fieldModeFilteringFlag;

/*********************************************************************************************************/

#define  IClip( Min, Max, Val) (((Val)<(Min))? (Min):(((Val)>(Max))? (Max):(Val)))

// NOTE: In principle, the alpha and beta tables are calculated with the formulas below
//       Alpha( qp ) = 0.8 * (2^(qp/6)  -  1)
//       Beta ( qp ) = 0.5 * qp  -  7

// The tables actually used have been "hand optimized" though (by Anthony Joch). So, the
// table values might be a little different to formula-generated values. Also, the first
// few values of both tables is set to zero to force the filter off at low qps 

byte ALPHA_TABLE[52]  = {0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,4,4,5,6,  7,8,9,10,12,13,15,17,  20,22,25,28,32,36,40,45,  50,56,63,71,80,90,101,113,  127,144,162,182,203,226,255,255} ;
byte  BETA_TABLE[52]  = {0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,2,2,2,3,  3,3,3, 4, 4, 4, 6, 6,   7, 7, 8, 8, 9, 9,10,10,  11,11,12,12,13,13, 14, 14,   15, 15, 16, 16, 17, 17, 18, 18} ;
byte CLIP_TAB[52][5]  =
{
  { 0, 0, 0, 0, 0},{ 0, 0, 0, 0, 0},{ 0, 0, 0, 0, 0},{ 0, 0, 0, 0, 0},{ 0, 0, 0, 0, 0},{ 0, 0, 0, 0, 0},{ 0, 0, 0, 0, 0},{ 0, 0, 0, 0, 0},
  { 0, 0, 0, 0, 0},{ 0, 0, 0, 0, 0},{ 0, 0, 0, 0, 0},{ 0, 0, 0, 0, 0},{ 0, 0, 0, 0, 0},{ 0, 0, 0, 0, 0},{ 0, 0, 0, 0, 0},{ 0, 0, 0, 0, 0},
  { 0, 0, 0, 0, 0},{ 0, 0, 0, 1, 1},{ 0, 0, 0, 1, 1},{ 0, 0, 0, 1, 1},{ 0, 0, 0, 1, 1},{ 0, 0, 1, 1, 1},{ 0, 0, 1, 1, 1},{ 0, 1, 1, 1, 1},
  { 0, 1, 1, 1, 1},{ 0, 1, 1, 1, 1},{ 0, 1, 1, 1, 1},{ 0, 1, 1, 2, 2},{ 0, 1, 1, 2, 2},{ 0, 1, 1, 2, 2},{ 0, 1, 1, 2, 2},{ 0, 1, 2, 3, 3},
  { 0, 1, 2, 3, 3},{ 0, 2, 2, 3, 3},{ 0, 2, 2, 4, 4},{ 0, 2, 3, 4, 4},{ 0, 2, 3, 4, 4},{ 0, 3, 3, 5, 5},{ 0, 3, 4, 6, 6},{ 0, 3, 4, 6, 6},
  { 0, 4, 5, 7, 7},{ 0, 4, 5, 8, 8},{ 0, 4, 6, 9, 9},{ 0, 5, 7,10,10},{ 0, 6, 8,11,11},{ 0, 6, 8,13,13},{ 0, 7,10,14,14},{ 0, 8,11,16,16},
  { 0, 9,12,18,18},{ 0,10,13,20,20},{ 0,11,15,23,23},{ 0,13,17,25,25}
} ;

char chroma_edge[2][4][4] = //[dir][edge][yuv_format]
{ { {-1, 0, 0, 0},
    {-1,-1,-1, 1},
    {-1, 1, 1, 2},
    {-1,-1,-1, 3}},

  { {-1, 0, 0, 0},
    {-1,-1, 1, 1},
    {-1, 1, 2, 2},
    {-1,-1, 3, 3}}};

#ifdef EIGHTH_PEL
void GetStrength(byte Strength[16],struct img_par *img,int MbQAddr,int dir,int edge, int mvlimit_x, int mvlimit_y, StorablePicture *p);
#else
void GetStrength(byte Strength[16],struct img_par *img,int MbQAddr,int dir,int edge, int mvlimit,StorablePicture *p);
#endif
void EdgeLoop(imgpel** Img, byte Strength[16],struct img_par *img, int MbQAddr, int AlphaC0Offset, int BetaOffset, int dir, int edge, int width, int yuv, int uv, StorablePicture *p);
void DeblockMb(ImageParameters *img, StorablePicture *p, int MbQAddr) ;

/*!
 *****************************************************************************************
 * \brief
 *    Filter all macroblocks in order of increasing macroblock address.
 *****************************************************************************************
 */
void DeblockPicture(ImageParameters *img, StorablePicture *p)
{
  unsigned i;

  for (i=0; i<p->PicSizeInMbs; i++)
  {
    DeblockMb( img, p, i ) ;
  }
} 


/*!
 *****************************************************************************************
 * \brief
 *    Deblocking filter for one macroblock.
 *****************************************************************************************
 */

void DeblockMb(ImageParameters *img, StorablePicture *p, int MbQAddr)
{
  int           EdgeCondition;
  int           dir,edge;
  byte          Strength[16];
  int           mb_x, mb_y;
  int           filterNon8x8LumaEdgesFlag[4] = {1,1,1,1};
  int           filterLeftMbEdgeFlag;
  int           filterTopMbEdgeFlag;
  //int           fieldModeMbFlag;
#ifdef EIGHTH_PEL
  int           mvlimit_x=4 << img->mv_res;
  int           mvlimit_y=4 << img->mv_res;
#else
  int           mvlimit=4;
#endif
  int           i, StrengthSum;
  Macroblock    *MbQ;
  imgpel **imgY   = p->imgY;
  imgpel ***imgUV = p->imgUV;

  int           edge_cr;
  
  
  img->DeblockCall = 1;
  get_mb_pos (MbQAddr, &mb_x, &mb_y);
  filterLeftMbEdgeFlag  = (mb_x != 0);
  filterTopMbEdgeFlag   = (mb_y != 0);

  MbQ  = &(img->mb_data[MbQAddr]) ; // current Mb



  if (MbQ->mb_type == I8MB)
    assert(MbQ->luma_transform_size_8x8_flag);

  filterNon8x8LumaEdgesFlag[1] = 
  filterNon8x8LumaEdgesFlag[3] = !(MbQ->luma_transform_size_8x8_flag);

  // return, if filter is disabled
  if (MbQ->LFDisableIdc==1) {
    img->DeblockCall = 0;
    return;
  }

  if (MbQ->LFDisableIdc==2)
  {
    // don't filter at slice boundaries
    filterLeftMbEdgeFlag = MbQ->mbAvailA;
    // if this the bottom of a frame macroblock pair then always filter the top edge
    filterTopMbEdgeFlag  = MbQ->mbAvailB;
  }

  img->current_mb_nr = MbQAddr;
  CheckAvailabilityOfNeighbors();

  for( dir=0 ; dir<2 ; dir++ )                                             // vertical edges, than horicontal edges
  {
    EdgeCondition = (dir && filterTopMbEdgeFlag) || (!dir && filterLeftMbEdgeFlag); // can not filter beyond picture boundaries
    for( edge=0 ; edge<4 ; edge++ )                                            // first 4 vertical strips of 16 pel
    {                                                                                         // then  4 horicontal
      if( edge || EdgeCondition )
      {
        edge_cr = chroma_edge[dir][edge][p->chroma_format_idc];

#ifdef EIGHTH_PEL
        GetStrength(Strength,img,MbQAddr,dir,edge, mvlimit_x, mvlimit_y, p); // Strength for 4 blks in 1 stripe
#else
        GetStrength(Strength,img,MbQAddr,dir,edge, mvlimit, p); // Strength for 4 blks in 1 stripe
#endif   
  StrengthSum = Strength[0];
        for (i = 1; i < 16; i++) StrengthSum += Strength[i];


        if(MbQ->mb_type >= 1 && MbQ->mb_type <= 3 && MbQ->luma_transform_size_8x8_flag == 2)
        {
#ifdef GEOM
          if(MbQ->mb_type == 1 || MbQ->geo_mode!=-1)
#else
          if(MbQ->mb_type == 1)
#endif
          {
            filterNon8x8LumaEdgesFlag[2] = 0;
          }
          else if(MbQ->mb_type == 2 && !dir)
          {
            filterNon8x8LumaEdgesFlag[2] = 0;
          }
          else if(MbQ->mb_type == 3 && dir)
          {
            filterNon8x8LumaEdgesFlag[2] = 0;
          }
        }

        if( StrengthSum )                      // only if one of the 16 Strength bytes is != 0
        {
          if (filterNon8x8LumaEdgesFlag[edge])
            EdgeLoop( imgY, Strength, img, MbQAddr, MbQ->LFAlphaC0Offset, MbQ->LFBetaOffset, dir, edge, p->size_x, 0, 0, p) ; 
          if( (imgUV != NULL) && (edge_cr >= 0))
          {
            EdgeLoop( imgUV[0], Strength, img, MbQAddr, MbQ->LFAlphaC0Offset, MbQ->LFBetaOffset, dir, edge_cr, p->size_x_cr, 1, 0, p) ; 
            EdgeLoop( imgUV[1], Strength, img, MbQAddr, MbQ->LFAlphaC0Offset, MbQ->LFBetaOffset, dir, edge_cr, p->size_x_cr, 1, 1, p) ; 
          }
        }

        if (dir && !edge && mixedModeEdgeFlag) {
          // this is the extra horizontal edge between a frame macroblock pair and a field above it
          img->DeblockCall = 2;
#ifdef EIGHTH_PEL
          GetStrength(Strength,img,MbQAddr,dir,4, mvlimit_x, mvlimit_y, p); // Strength for 4 blks in 1 stripe
#else
          GetStrength(Strength,img,MbQAddr,dir,4, mvlimit, p); // Strength for 4 blks in 1 stripe
#endif
          if( *((int*)Strength) )                      // only if one of the 4 Strength bytes is != 0
          {
            if (filterNon8x8LumaEdgesFlag[edge])
              EdgeLoop( imgY, Strength, img, MbQAddr, MbQ->LFAlphaC0Offset, MbQ->LFBetaOffset, dir, 4, p->size_x, 0, 0, p) ; 
            if( (imgUV != NULL) && (edge_cr >= 0))
            {
              EdgeLoop( imgUV[0], Strength, img, MbQAddr, MbQ->LFAlphaC0Offset, MbQ->LFBetaOffset, dir, 4, p->size_x_cr, 1, 0, p) ; 
              EdgeLoop( imgUV[1], Strength, img, MbQAddr, MbQ->LFAlphaC0Offset, MbQ->LFBetaOffset, dir, 4, p->size_x_cr, 1, 1, p) ; 
            }
          }
          img->DeblockCall = 1;
        }
      }
    }//end edge
  }//end loop dir
  img->DeblockCall = 0;
}

  /*!
 *********************************************************************************************
 * \brief
 *    returns a buffer of 16 Strength values for one stripe in a mb (for different Frame types)
 *********************************************************************************************
 */

int  ININT_STRENGTH[4] = {0x04040404, 0x03030303, 0x03030303, 0x03030303} ; 
byte BLK_NUM[2][4][4]  = {{{0,4,8,12},{1,5,9,13},{2,6,10,14},{3,7,11,15}},{{0,1,2,3},{4,5,6,7},{8,9,10,11},{12,13,14,15}}} ;
byte BLK_4_TO_8[16]    = {0,0,1,1,0,0,1,1,2,2,3,3,2,2,3,3} ;
#define ANY_INTRA (MbP->mb_type==I4MB||MbP->mb_type==I16MB||MbP->mb_type==IPCM||MbQ->mb_type==I4MB||MbQ->mb_type==I16MB||MbQ->mb_type==IPCM)


#ifdef EIGHTH_PEL
  void GetStrength(byte Strength[16],struct img_par *img,int MbQAddr,int dir,int edge, int mvlimit_x, int mvlimit_y, StorablePicture *p)
#else
  void GetStrength(byte Strength[16],struct img_par *img,int MbQAddr,int dir,int edge, int mvlimit, StorablePicture *p)
#endif
{
  byte     StrValue;
  int    xQ, yQ;
  Macroblock    *MbQ;
  Macroblock    *MbP;
  PixelPos pixP, pixMB;
 
  MbQ = &(img->mb_data[MbQAddr]);
  
  edge = edge << 2;
  xQ = dir ? 0 : edge - 1;
  yQ = dir ? (edge < 16 ? edge - 1: 0) : 0;
  getNeighbour(MbQAddr, xQ, yQ, 1, &pixMB);
  pixP = pixMB;
  MbP = &(img->mb_data[pixP.mb_addr]);
                                                               
  if(  !(MbP->mb_type==I4MB || MbP->mb_type==I16MB || MbP->mb_type==IPCM || MbP->mb_type==I8MB)
    && !(MbQ->mb_type==I4MB || MbQ->mb_type==I16MB || MbQ->mb_type==IPCM || MbQ->mb_type==I8MB) )
  {

    int64    ref_p0,ref_p1,ref_q0,ref_q1;
    int      blkP, blkQ, idx;
    int      blk_x, blk_x2, blk_y, blk_y2 ;
    short  ***list0_mv = p->mv[LIST_0];
    short  ***list1_mv = p->mv[LIST_1];
    char   **list0_refIdxArr = p->ref_idx[LIST_0];
    char   **list1_refIdxArr = p->ref_idx[LIST_1];
    int64  **list0_refPicIdArr = p->ref_pic_id[LIST_0];
    int64  **list1_refPicIdArr = p->ref_pic_id[LIST_1];
    int    mb_x, mb_y;
    
    get_mb_block_pos (MbQAddr, &mb_x, &mb_y);
    mb_x <<= 2;
    mb_y <<= 2;
    yQ += dir;
    xQ += (1 - dir);

    for( idx = 0 ; idx < 16 ; idx += BLOCK_SIZE )
    {
      if (dir)
      {
        xQ = idx;
        pixP.x = (short) (pixMB.x + idx);
        pixP.pos_x =  (short) (pixMB.pos_x + idx);
      }
      else
      {
        yQ = idx;
        pixP.y =  (short) (pixMB.y + idx);
        pixP.pos_y =  (short) (pixMB.pos_y + idx);
      }

      blkQ = (yQ & 0xFFFC) + (xQ >> 2);
      blkP = (pixP.y & 0xFFFC) + (pixP.x >> 2);

      if( ((MbQ->cbp_blk &  ((int64)1 << blkQ )) != 0) || ((MbP->cbp_blk &  ((int64)1 << blkP)) != 0) )
        StrValue = 2 ;
      else
      {       
        // if no coefs, but vector difference >= 1 set Strength=1 
        blk_y  = mb_y + (blkQ >> 2);
        blk_x  = mb_x + (blkQ  & 3);
        blk_y2 = pixP.pos_y >> 2;
        blk_x2 = pixP.pos_x >> 2;

        {
          ref_p0 = list0_refIdxArr[blk_y][blk_x]<0   ? INT64_MIN : list0_refPicIdArr[blk_y][blk_x];
          ref_q0 = list0_refIdxArr[blk_y2][blk_x2]<0 ? INT64_MIN : list0_refPicIdArr[blk_y2][blk_x2];
          ref_p1 = list1_refIdxArr[blk_y][blk_x]<0   ? INT64_MIN : list1_refPicIdArr[blk_y][blk_x];
          ref_q1 = list1_refIdxArr[blk_y2][blk_x2]<0 ? INT64_MIN : list1_refPicIdArr[blk_y2][blk_x2];
          if ( ((ref_p0==ref_q0) && (ref_p1==ref_q1)) || ((ref_p0==ref_q1) && (ref_p1==ref_q0))) 
          {
            // L0 and L1 reference pictures of p0 are different; q0 as well
            if (ref_p0 != ref_p1) 
            { 
              // compare MV for the same reference picture
              if (ref_p0==ref_q0) 
              {
#ifdef EIGHTH_PEL
                StrValue =  (byte)  
                  ( (abs( list0_mv[blk_y][blk_x][0] - list0_mv[blk_y2][blk_x2][0]) >= mvlimit_x)
                  | (abs( list0_mv[blk_y][blk_x][1] - list0_mv[blk_y2][blk_x2][1]) >= mvlimit_y)
                  | (abs( list1_mv[blk_y][blk_x][0] - list1_mv[blk_y2][blk_x2][0]) >= mvlimit_x)
                  | (abs( list1_mv[blk_y][blk_x][1] - list1_mv[blk_y2][blk_x2][1]) >= mvlimit_y));
#else  
                StrValue =  (byte)  
                  ( (abs( list0_mv[blk_y][blk_x][0] - list0_mv[blk_y2][blk_x2][0]) >= 4)
                  | (abs( list0_mv[blk_y][blk_x][1] - list0_mv[blk_y2][blk_x2][1]) >= mvlimit)
                  | (abs( list1_mv[blk_y][blk_x][0] - list1_mv[blk_y2][blk_x2][0]) >= 4)
                  | (abs( list1_mv[blk_y][blk_x][1] - list1_mv[blk_y2][blk_x2][1]) >= mvlimit));
#endif
              }
              else 
              {
#ifdef EIGHTH_PEL
                StrValue =  (byte)  
                  ( (abs( list0_mv[blk_y][blk_x][0] - list1_mv[blk_y2][blk_x2][0]) >= mvlimit_x)
                  | (abs( list0_mv[blk_y][blk_x][1] - list1_mv[blk_y2][blk_x2][1]) >= mvlimit_y)
                  | (abs( list1_mv[blk_y][blk_x][0] - list0_mv[blk_y2][blk_x2][0]) >= mvlimit_x)
                  | (abs( list1_mv[blk_y][blk_x][1] - list0_mv[blk_y2][blk_x2][1]) >= mvlimit_y));
#else
                StrValue =  (byte)  
                  ( (abs( list0_mv[blk_y][blk_x][0] - list1_mv[blk_y2][blk_x2][0]) >= 4)
                  | (abs( list0_mv[blk_y][blk_x][1] - list1_mv[blk_y2][blk_x2][1]) >= mvlimit)
                  | (abs( list1_mv[blk_y][blk_x][0] - list0_mv[blk_y2][blk_x2][0]) >= 4)
                  | (abs( list1_mv[blk_y][blk_x][1] - list0_mv[blk_y2][blk_x2][1]) >= mvlimit));
#endif
              } 
            }
            else 
            { // L0 and L1 reference pictures of p0 are the same; q0 as well

#ifdef EIGHTH_PEL
              StrValue =  (byte)  
                ( (abs( list0_mv[blk_y][blk_x][0] - list0_mv[blk_y2][blk_x2][0]) >= mvlimit_x)
                | (abs( list0_mv[blk_y][blk_x][1] - list0_mv[blk_y2][blk_x2][1]) >= mvlimit_y)
                | (abs( list1_mv[blk_y][blk_x][0] - list1_mv[blk_y2][blk_x2][0]) >= mvlimit_x)
                | (abs( list1_mv[blk_y][blk_x][1] - list1_mv[blk_y2][blk_x2][1]) >= mvlimit_y))
                &&
                ( (abs( list0_mv[blk_y][blk_x][0] - list1_mv[blk_y2][blk_x2][0]) >= mvlimit_x)
                | (abs( list0_mv[blk_y][blk_x][1] - list1_mv[blk_y2][blk_x2][1]) >= mvlimit_y)
                | (abs( list1_mv[blk_y][blk_x][0] - list0_mv[blk_y2][blk_x2][0]) >= mvlimit_x)
                | (abs( list1_mv[blk_y][blk_x][1] - list0_mv[blk_y2][blk_x2][1]) >= mvlimit_y));
#else
              StrValue =  (byte)  
                ( (abs( list0_mv[blk_y][blk_x][0] - list0_mv[blk_y2][blk_x2][0]) >= 4)
                | (abs( list0_mv[blk_y][blk_x][1] - list0_mv[blk_y2][blk_x2][1]) >= mvlimit)
                | (abs( list1_mv[blk_y][blk_x][0] - list1_mv[blk_y2][blk_x2][0]) >= 4)
                | (abs( list1_mv[blk_y][blk_x][1] - list1_mv[blk_y2][blk_x2][1]) >= mvlimit))
                &&
                ( (abs( list0_mv[blk_y][blk_x][0] - list1_mv[blk_y2][blk_x2][0]) >= 4)
                | (abs( list0_mv[blk_y][blk_x][1] - list1_mv[blk_y2][blk_x2][1]) >= mvlimit)
                | (abs( list1_mv[blk_y][blk_x][0] - list0_mv[blk_y2][blk_x2][0]) >= 4)
                | (abs( list1_mv[blk_y][blk_x][1] - list0_mv[blk_y2][blk_x2][1]) >= mvlimit));
#endif
            }       
          }
          else 
          {
            StrValue =  1;        
          } 
        }
      }

      memset(&Strength[idx], (byte) StrValue, BLOCK_SIZE * sizeof(byte));
    }
  }
  else
  {
    // Start with Strength=3. or Strength=4 for Mb-edge
    StrValue = (edge == 0) ? 4 : 3;
    memset(&Strength[0], (byte) StrValue, MB_BLOCK_SIZE * sizeof(byte));
  }
}



/*!
 *****************************************************************************************
 * \brief
 *    Filters one edge of 16 (luma) or 8 (chroma) pel
 *****************************************************************************************
 */
void EdgeLoop(imgpel** Img, byte Strength[16],struct img_par *img, int MbQAddr, int AlphaC0Offset, int BetaOffset,
              int dir, int edge, int width, int yuv, int uv, StorablePicture *p)
{
  int      pel, ap = 0, aq = 0, Strng ;
  int      incP, incQ;
  int      C0, c0, Delta, dif, AbsDelta ;
  int      L2 = 0, L1, L0, R0, R1, R2 = 0, RL0, L3, R3 ;
  int      Alpha = 0, Beta = 0 ;
  byte*    ClipTab = NULL;   
  int      small_gap;
  int      indexA, indexB;
  int      PelNum;
  int      StrengthIdx;
  imgpel   *SrcPtrP, *SrcPtrQ;
  int      QP;
  int      xQ, yQ;

  Macroblock *MbQ, *MbP;
  PixelPos pixP, pixQ, pixMB1, pixMB2;
  int      bitdepth_scale; 
  int      pelnum_cr[2][4] =  {{0,8,16,16}, {0,8, 8,16}};  //[dir:0=vert, 1=hor.][yuv_format]
  
  if (!yuv)
    bitdepth_scale = 1<<(img->bitdepth_luma - 8);
  else
    bitdepth_scale = 1<<(img->bitdepth_chroma - 8);
  
  PelNum = yuv ? pelnum_cr[dir][p->chroma_format_idc] : 16 ;
 
  MbQ = &(img->mb_data[MbQAddr]);
  
  edge = edge << 2;
  xQ = dir ? 0 : edge - 1;
  yQ = dir ? (edge < 16 ? edge - 1: 0) : 0;
  getNeighbour(MbQAddr, xQ, yQ, 1-yuv, &pixMB1);

  if (pixMB1.available || (MbQ->LFDisableIdc== 0))
  {
    int dirM1 = (dir - 1);
    pixP = pixMB1;

    MbP = &(img->mb_data[pixP.mb_addr]);
    fieldModeFilteringFlag = 0;

    // Average QP of the two blocks
    QP = yuv ? (MbP->qpc[uv] + MbQ->qpc[uv] + 1) >> 1 : (MbP->qp + MbQ->qp + 1) >> 1;

    indexA = IClip(0, MAX_QP, QP + AlphaC0Offset);
    indexB = IClip(0, MAX_QP, QP + BetaOffset);

    Alpha  =ALPHA_TABLE[indexA] * bitdepth_scale;
    Beta   =BETA_TABLE[indexB]  * bitdepth_scale;

    getNeighbour(MbQAddr, (xQ -= dirM1), (yQ += dir), 1-yuv, &pixMB2);

    if (Alpha != 0 && Beta != 0)
    {
      ClipTab=CLIP_TAB[indexA];
      pixQ = pixMB2;

      for (pel=0 ; pel<PelNum ; pel++ )
      {
        StrengthIdx = (yuv&&(PelNum==8)) ? (((pel>>1)<<2)+(pel%2)) : pel ;

        incQ = dir ? (width) : 1;
        incP = dir ? (width) : 1;
        SrcPtrQ = &(Img[pixQ.pos_y][pixQ.pos_x]);
        SrcPtrP = &(Img[pixP.pos_y][pixP.pos_x]);

        L0  = SrcPtrP[0] ;
        R0  = SrcPtrQ[0] ;
        L1  = SrcPtrP[-incP] ;
        R1  = SrcPtrQ[ incQ] ;
        L2  = SrcPtrP[-incP*2] ;
        R2  = SrcPtrQ[ incQ*2] ;
        L3  = SrcPtrP[-incP*3] ;
        R3  = SrcPtrQ[ incQ*3] ;

        if ( (Strng = Strength[StrengthIdx]) )
        {
          AbsDelta  = abs( Delta = R0 - L0 )  ;

          if ( AbsDelta < Alpha )
          {
            C0  = ClipTab[ Strng ] * bitdepth_scale;
            if( ((abs( R0 - R1) - Beta )  & (abs(L0 - L1) - Beta )) < 0  ) 
            {
              if( !yuv)
              {
                aq  = (abs( R0 - R2) - Beta ) < 0  ;
                ap  = (abs( L0 - L2) - Beta ) < 0  ;
              }

              RL0             = L0 + R0 ;

              if(Strng == 4 )    // INTRA strong filtering
              {
                if( yuv)  // Chroma
                {
                  SrcPtrQ[0] = ((R1 << 1) + R0 + L1 + 2) >> 2; 
                  SrcPtrP[0] = ((L1 << 1) + L0 + R1 + 2) >> 2;                                           
                }
                else  // Luma
                {
                  small_gap = (AbsDelta < ((Alpha >> 2) + 2));

                  aq &= small_gap;
                  ap &= small_gap;

                  SrcPtrQ[0]   = aq ? ( L1 + ((R1 + RL0) << 1) +  R2 + 4) >> 3 : ((R1 << 1) + R0 + L1 + 2) >> 2 ;
                  SrcPtrP[0]   = ap ? ( R1 + ((L1 + RL0) << 1) +  L2 + 4) >> 3 : ((L1 << 1) + L0 + R1 + 2) >> 2 ;

                  SrcPtrQ[ incQ] =   aq  ? ( R2 + R0 + R1 + L0 + 2) >> 2 : R1;
                  SrcPtrP[-incP] =   ap  ? ( L2 + L1 + L0 + R0 + 2) >> 2 : L1;

                  SrcPtrQ[ incQ*2] = aq ? (((R3 + R2) <<1) + R2 + R1 + RL0 + 4) >> 3 : R2;
                  SrcPtrP[-incP*2] = ap ? (((L3 + L2) <<1) + L2 + L1 + RL0 + 4) >> 3 : L2;
                }
              }
              else                                                     // normal filtering
              {
                c0               = yuv? (C0+1):(C0 + ap + aq) ;
                dif              = IClip( -c0, c0, ( (Delta << 2) + (L1 - R1) + 4) >> 3 ) ;
                if(!yuv)
                {
                  SrcPtrP[0]  = IClip(0, img->max_imgpel_value, L0 + dif) ;
                  SrcPtrQ[0]  = IClip(0, img->max_imgpel_value, R0 - dif) ;
                } 
                else 
                {
                  SrcPtrP[0]  = IClip(0, img->max_imgpel_value_uv, L0 + dif) ;
                  SrcPtrQ[0]  = IClip(0, img->max_imgpel_value_uv, R0 - dif) ;
                }

                if( !yuv )
                {
                  if( ap )
                    SrcPtrP[-incP] += IClip( -C0,  C0, ( L2 + ((RL0 + 1) >> 1) - (L1<<1)) >> 1 ) ;
                  if( aq  )
                    SrcPtrQ[ incQ] += IClip( -C0,  C0, ( R2 + ((RL0 + 1) >> 1) - (R1<<1)) >> 1 ) ;
                } ;
              } ;
            } ; 
          } ;
        } ; // strng != 0
        pixP.pos_x = (short) (pixP.pos_x + dir);
        pixP.pos_y = (short) (pixP.pos_y - dirM1);
        pixQ.pos_x = (short) (pixQ.pos_x + dir);
        pixQ.pos_y = (short) (pixQ.pos_y - dirM1);
      } ;
    }
  }
}

