/* 
  ====================================================================================================================

	The copyright in this software is being made available under the License included below. 
	This software may be subject to other third party and 	contributor rights, including patent rights, and no such 
	rights are granted under this license.
 
	Copyright (c) 2010, QUALCOMM INC.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification, are permitted only for 
	the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
	promoting such standards. The following conditions are required to be met:

		* Redistributions of source code must retain the above copyright notice, this list of conditions and 
		  the following disclaimer.
		* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
		  the following disclaimer in the documentation and/or other materials provided with the distribution.
		* The name of QUALCOMM INC.  may not be used to endorse or promote products derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
	INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  ====================================================================================================================
*/
#include <math.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>

#include "global.h"
#include "elements.h"
#include "defines.h"
#include "fmo.h"
#include "vlc.h"
#include "mbuffer.h"
#include "header.h"
#include "switched_filters.h"
#include "image.h"


//
//  Clip a value between low and high
//
int Clip3Fun(int low, int high, int val)
{
  return ((val < low)? low: (val > high)? high: val);
}

double STANDARD_2D_FILTER[15][36] = {
  {   0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
  1.0/64.0  ,   -5.0/64.0  ,   52.0/64.0  ,   20.0/64.0  ,   -5.0/64.0  ,   1.0/64.0  ,
  0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0
  }, // a_pos
  {   0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
  1.0/32.0  ,   -5.0/32.0  ,   20.0/32.0  ,   20.0/32.0  ,   -5.0/32.0  ,   1.0/32.0  ,
  0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0
  }, // b_pos
  {   0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
  1.0/64.0  ,   -5.0/64.0  ,   20.0/64.0  ,   52.0/64.0  ,   -5.0/64.0  ,   1.0/64.0  ,
  0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0
  }, // c_pos
  {   0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,   52.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,   20.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0
  }, // d_pos
  {   0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
  1.0/64.0  ,   -5.0/64.0  ,   40.0/64.0  ,   20.0/64.0  ,   -5.0/64.0  ,   1.0/64.0  ,
  0.0       ,    0.0       ,   20.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0
  }, // e_pos
  {   1.0/2048.0,   -5.0/2048.0,   20.0/2048.0,   20.0/2048.0,   -5.0/2048.0,   1.0/2048.0,
  -5.0/2048.0,   25.0/2048.0, -100.0/2048.0, -100.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
  52.0/2048.0, -260.0/2048.0, 1040.0/2048.0, 1040.0/2048.0, -260.0/2048.0,  52.0/2048.0,
  20.0/2048.0, -100.0/2048.0,  400.0/2048.0,  400.0/2048.0, -100.0/2048.0,  20.0/2048.0,
  -5.0/2048.0,   25.0/2048.0, -100.0/2048.0, -100.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
  1.0/2048.0,   -5.0/2048.0,   20.0/2048.0,   20.0/2048.0,   -5.0/2048.0,   1.0/2048.0
  }, // f_pos
  {   0.0       ,    0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,   0.0       ,
  1.0/64.0  ,   -5.0/64.0  ,   20.0/64.0  ,   40.0/64.0  ,   -5.0/64.0  ,   1.0/64.0  ,
  0.0       ,    0.0       ,    0.0       ,   20.0/64.0  ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,   0.0
  }, // g_pos
  {   0.0       ,    0.0       ,    1.0/32.0  ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,   -5.0/32.0  ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,   20.0/32.0  ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,   20.0/32.0  ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,   -5.0/32.0  ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,    1.0/32.0  ,    0.0       ,    0.0       ,   0.0
  }, // h_pos
  {   1.0/2048.0,   -5.0/2048.0,   52.0/2048.0,   20.0/2048.0,   -5.0/2048.0,   1.0/2048.0,
  -5.0/2048.0,   25.0/2048.0, -260.0/2048.0, -100.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
  20.0/2048.0, -100.0/2048.0, 1040.0/2048.0,  400.0/2048.0, -100.0/2048.0,  20.0/2048.0,
  20.0/2048.0, -100.0/2048.0, 1040.0/2048.0,  400.0/2048.0, -100.0/2048.0,  20.0/2048.0,
  -5.0/2048.0,   25.0/2048.0, -260.0/2048.0, -100.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
  1.0/2048.0,   -5.0/2048.0,   52.0/2048.0,   20.0/2048.0,   -5.0/2048.0,   1.0/2048.0
  }, // i_pos
  {   1.0/1024.0,   -5.0/1024.0,   20.0/1024.0,   20.0/1024.0,   -5.0/1024.0,   1.0/1024.0,
  -5.0/1024.0,   25.0/1024.0, -100.0/1024.0, -100.0/1024.0,   25.0/1024.0,  -5.0/1024.0,
  20.0/1024.0, -100.0/1024.0,  400.0/1024.0,  400.0/1024.0, -100.0/1024.0,  20.0/1024.0,
  20.0/1024.0, -100.0/1024.0,  400.0/1024.0,  400.0/1024.0, -100.0/1024.0,  20.0/1024.0,
  -5.0/1024.0,   25.0/1024.0, -100.0/1024.0, -100.0/1024.0,   25.0/1024.0,  -5.0/1024.0,
  1.0/1024.0,   -5.0/1024.0,   20.0/1024.0,   20.0/1024.0,   -5.0/1024.0,   1.0/1024.0
  }, // j_pos
  {   1.0/2048.0,   -5.0/2048.0,   20.0/2048.0,   52.0/2048.0,   -5.0/2048.0,   1.0/2048.0,
  -5.0/2048.0,   25.0/2048.0, -100.0/2048.0, -260.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
  20.0/2048.0, -100.0/2048.0,  400.0/2048.0, 1040.0/2048.0, -100.0/2048.0,  20.0/2048.0,
  20.0/2048.0, -100.0/2048.0,  400.0/2048.0, 1040.0/2048.0, -100.0/2048.0,  20.0/2048.0,
  -5.0/2048.0,   25.0/2048.0, -100.0/2048.0, -260.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
  1.0/2048.0,   -5.0/2048.0,   20.0/2048.0,   52.0/2048.0,   -5.0/2048.0,   1.0/2048.0
  }, // i_pos
  {   0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,   20.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,   52.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0
  }, // l_pos
  {   0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,   20.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
  1.0/64.0  ,   -5.0/64.0  ,   40.0/64.0  ,   20.0/64.0  ,   -5.0/64.0  ,   1.0/64.0  ,
  0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
  0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0
  }, // m_pos
  {   1.0/2048.0,   -5.0/2048.0,   20.0/2048.0,   20.0/2048.0,   -5.0/2048.0,   1.0/2048.0,
  -5.0/2048.0,   25.0/2048.0, -100.0/2048.0, -100.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
  20.0/2048.0, -100.0/2048.0,  400.0/2048.0,  400.0/2048.0, -100.0/2048.0,  20.0/2048.0,
  52.0/2048.0, -260.0/2048.0, 1040.0/2048.0, 1040.0/2048.0, -260.0/2048.0,  52.0/2048.0,
  -5.0/2048.0,   25.0/2048.0, -100.0/2048.0, -100.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
  1.0/2048.0,   -5.0/2048.0,   20.0/2048.0,   20.0/2048.0,   -5.0/2048.0,   1.0/2048.0
  }, // n_pos
  {   0.0         ,    0.0         ,  0.0     ,    1.0/64.0  ,    0.0       ,   0.0       ,
  0.0         ,    0.0         ,  0.0     ,   -5.0/64.0  ,    0.0       ,   0.0       ,
  0.0         ,    0.0         ,  0.0     ,   20.0/64.0  ,    0.0       ,   0.0       ,
  1.0/64.0    ,   -5.0/64.0    , 20.0/64.0,   40.0/64.0  ,   -5.0/64.0  ,   1.0/64.0  ,
  0.0         ,    0.0         ,  0.0     ,   -5.0/64.0  ,    0.0       ,   0.0       ,
  0.0         ,    0.0         ,  0.0     ,    1.0/64.0  ,    0.0       ,   0.0
  } // o_pos
};

//
//
int SYMMETRIC_1[15][SQR_FILTER] = {
  {
    0,   0,   0,   0,   0,   0,  
      0,   0,   0,   0,   0,   0,  
      -1,  -2,  97,  42, -11,   3, 
      0,   0,   0,   0,   0,   0,  
      0,   0,   0,   0,   0,   0,  
      0,   0,   0,   0,   0,   0  
  },  // a_pos
  {
    0,   0,   0,   0,   0,   0,  
      0,   0,   0,   0,   0,   0,  
      3, -16,  77,  77, -16,   3, 
      0,   0,   0,   0,   0,   0,  
      0,   0,   0,   0,   0,   0,  
      0,   0,   0,   0,   0,   0  
    },  // b_pos
    {
      0,   0,   0,   0,   0,   0,  
        0,   0,   0,   0,   0,   0,  
        3, -11,  42,  97,  -2,  -1, 
        0,   0,   0,   0,   0,   0,  
        0,   0,   0,   0,   0,   0,  
        0,   0,   0,   0,   0,   0  
    },  // c_pos
    {
      0,   0,  -1,   0,   0,   0,  
        0,   0,  -2,   0,   0,   0,  
        0,   0,  97,   0,   0,   0,  
        0,   0,  42,   0,   0,   0,  
        0,   0, -11,   0,   0,   0,  
        0,   0,   3,   0,   0,   0
      },  // d_pos
      {
        0,   0,   0,   0,   0,   0,  
          0,  -7,   4,  -2,  -3,   0,  
          0,   4,  66,  36,   1,   0,  
          0,  -2,  36,   7,  -6,   0,  
          0,  -3,   1,  -6,   2,   0,  
          0,   0,   0,   0,   0,   0
      },  // e_pos
      {
        0,   0,   0,   0,   0,   0,  
          0,  -4,  -1,  -1,  -4,   0,  
          0,  -2,  56,  56,  -2,   0,  
          0,  -7,  25,  25,  -7,   0,  
          0,   1,  -4,  -4,   1,   0,  
          0,   0,   0,   0,   0,   0
        },  // f_pos
        {
          0,   0,   0,   0,   0,   0,  
            0,  -3,  -2,   4,  -7,   0,  
            0,   1,  36,  66,   4,   0,  
            0,  -6,   7,  36,  -2,   0,  
            0,   2,  -6,   1,  -3,   0,  
            0,   0,   0,   0,   0,   0
        },  // g_pos
        {
          0,   0,   3,   0,   0,   0, 
            0,   0, -16,   0,   0,   0,  
            0,   0,  77,   0,   0,   0,  
            0,   0,  77,   0,   0,   0,  
            0,   0, -16,   0,   0,   0,  
            0,   0,   3,   0,   0,   0
          },  // h_pos
          {
            0,   0,   0,   0,   0,   0,  
              0,  -4,  -2,  -7,   1,   0,  
              0,  -1,  56,  25,  -4,   0,  
              0,  -1,  56,  25,  -4,   0,  
              0,  -4,  -2,  -7,   1,   0,  
              0,   0,   0,   0,   0,   0
          },  // i_pos
          {
            0,   0,   0,   0,   0,   0,  
              0,   0,  -6,  -6,   0,   0,  
              0,  -6,  44,  44,  -6,   0,  
              0,  -6,  44,  44,  -6,   0,  
              0,   0,  -6,  -6,   0,   0,  
              0,   0,   0,   0,   0,   0
            },  // j_pos
            {
              0,   0,   0,   0,   0,   0,  
                0,   1,  -7,  -2,  -4,   0,  
                0,  -4,  25,  56,  -1,   0,  
                0,  -4,  25,  56,  -1,   0,  
                0,   1,  -7,  -2,  -4,   0,  
                0,   0,   0,   0,   0,   0
            },  // k_pos
            {
              0,   0,   3,   0,   0,   0, 
                0,   0, -11,   0,   0,   0,  
                0,   0,  42,   0,   0,   0,  
                0,   0,  97,   0,   0,   0,  
                0,   0,  -2,   0,   0,   0,  
                0,   0,  -1,   0,   0,   0
              },  // l_pos
              {
                0,   0,   0,   0,   0,   0,  
                  0,  -3,   1,  -6,   2,   0,  
                  0,  -2,  36,   7,  -6,   0,  
                  0,   4,  66,  36,   1,   0,  
                  0,  -7,   4,  -2,  -3,   0,  
                  0,   0,   0,   0,   0,   0
              },  // m_pos
              {
                0,   0,   0,   0,   0,   0,  
                  0,   1,  -4,  -4,   1,   0,  
                  0,  -7,  25,  25,  -7,   0,  
                  0,  -2,  56,  56,  -2,   0,  
                  0,  -4,  -1,  -1,  -4,   0,  
                  0,   0,   0,   0,   0,   0
                },  // n_pos
                {
                  0,   0,   0,   0,   0,   0,  
                    0,   2,  -6,   1,  -3,   0,  
                    0,  -6,   7,  36,  -2,   0,  
                    0,   1,  36,  66,   4,   0,  
                    0,  -3,  -2,   4,  -7,   0,  
                    0,   0,   0,   0,   0,   0
                }  // o_pos
};


//
//
//
int SYMMETRIC_2[15][SQR_FILTER] = {
  {
    0,   0,   0,   0,   0,   0,  
      0,   0,   0,   0,   0,   0,  
      4, -14, 108,  38, -12,   4, 
      0,   0,   0,   0,   0,   0,  
      0,   0,   0,   0,   0,   0,  
      0,   0,   0,   0,   0,   0  
  },  // a_pos
  {
    0,   0,   0,   0,   0,   0,  
      0,   0,   0,   0,   0,   0,  
      6, -23,  81,  81, -23,   6, 
      0,   0,   0,   0,   0,   0,  
      0,   0,   0,   0,   0,   0,  
      0,   0,   0,   0,   0,   0  
    },  // b_pos
    {
      0,   0,   0,   0,   0,   0,  
        0,   0,   0,   0,   0,   0,  
        4, -12,  38, 108, -14,   4, 
        0,   0,   0,   0,   0,   0,  
        0,   0,   0,   0,   0,   0,  
        0,   0,   0,   0,   0,   0  
    },  // c_pos
    {
      0,   0,   4,   0,   0,   0,  
        0,   0, -14,   0,   0,   0,  
        0,   0, 108,   0,   0,   0,  
        0,   0,  38,   0,   0,   0,  
        0,   0, -12,   0,   0,   0,  
        0,   0,   4,   0,   0,   0
      },  // d_pos
      {
        0,   0,   0,   0,   0,   0,  
          0,  -6,   2,  -4,  -3,   0,  
          0,   2,  71,  38,   0,   0,  
          0,  -4,  38,   7,  -7,   0,  
          0,  -3,   0,  -7,   4,   0,  
          0,   0,   0,   0,   0,   0
      },  // e_pos
      {
        0,   0,   0,   0,   0,   0,  
          0,  -3,  -2,  -2,  -3,   0,  
          0,  -3,  58,  58,  -3,   0,  
          0,  -9,  26,  26,  -9,   0,  
          0,   2,  -5,  -5,   2,   0,  
          0,   0,   0,   0,   0,   0
        },  // f_pos
        {
          0,   0,   0,   0,   0,   0,  
            0,  -3,  -4,   2,  -6,   0,  
            0,   0,  38,  71,   2,   0,  
            0,  -7,   7,  38,  -4,   0,  
            0,   4,  -7,   0,  -3,   0,  
            0,   0,   0,   0,   0,   0
        },  // g_pos
        {
          0,   0,   6,   0,   0,   0,  
            0,   0, -23,   0,   0,   0,  
            0,   0,  81,   0,   0,   0,  
            0,   0,  81,   0,   0,   0,  
            0,   0, -23,   0,   0,   0,  
            0,   0,   6,   0,   0,   0
          },  // h_pos
          {
            0,   0,   0,   0,   0,   0,  
              0,  -3,  -3,  -9,   2,   0,  
              0,  -2,  58,  26,  -5,   0,  
              0,  -2,  58,  26,  -5,   0,  
              0,  -3,  -3,  -9,   2,   0,  
              0,   0,   0,   0,   0,   0
          },  // i_pos
          {
            0,   0,   0,   0,   0,   0,  
              0,   1,  -7,  -7,   1,   0,  
              0,  -7,  45,  45,  -7,   0,  
              0,  -7,  45,  45,  -7,   0,  
              0,   1,  -7,  -7,   1,   0,  
              0,   0,   0,   0,   0,   0
            },  // j_pos
            {
              0,   0,   0,   0,   0,   0,  
                0,   2,  -9,  -3,  -3,   0,  
                0,  -5,  26,  58,  -2,   0,  
                0,  -5,  26,  58,  -2,   0,  
                0,   2,  -9,  -3,  -3,   0,  
                0,   0,   0,   0,   0,   0
            },  // k_pos
            {
              0,   0,   4,   0,   0,   0,  
                0,   0, -12,   0,   0,   0,  
                0,   0,  38,   0,   0,   0,  
                0,   0, 108,   0,   0,   0,  
                0,   0, -14,   0,   0,   0,  
                0,   0,   4,   0,   0,   0
              },  // l_pos
              {
                0,   0,   0,   0,   0,   0,  
                  0,  -3,   0,  -7,   4,   0,  
                  0,  -4,  38,   7,  -7,   0,  
                  0,   2,  71,  38,   0,   0,  
                  0,  -6,   2,  -4,  -3,   0,  
                  0,   0,   0,   0,   0,   0
              },  // m_pos
              {
                0,   0,   0,   0,   0,   0,  
                  0,   2,  -5,  -5,   2,   0,  
                  0,  -9,  26,  26,  -9,   0,  
                  0,  -3,  58,  58,  -3,   0,  
                  0,  -3,  -2,  -2,  -3,   0,  
                  0,   0,   0,   0,   0,   0
                },  // n_pos
                {
                  0,   0,   0,   0,   0,   0,  
                    0,   4,  -7,   0,  -3,   0,  
                    0,  -7,   7,  38,  -4,   0,  
                    0,   0,  38,  71,   2,   0,  
                    0,  -3,  -4,   2,  -6,   0,  
                    0,   0,   0,   0,   0,   0
                }   // o_pos
};



int getblock2DFilt(int ref_idx, StorablePicture **list, int x_pos, int y_pos,
                   int mvx, int mvy, int img_width, int img_height, int block[16][16], int size_x,int size_y)
{
#ifdef EIGHTH_PEL
  if(img->mv_res && (((mvx&1)!=0)||((mvy&1)!=0))) 
  {
    int mv_mul = 4 << img->mv_res;
    get_block( ref_idx, list, x_pos*mv_mul+mvx, y_pos*mv_mul+mvy, img, block,size_x,size_y );
    return 1;
  }
  else  
#endif
  {
    int x_sub, y_sub, sub_pos;
    int i,j,offset=0;
#ifdef EIGHTH_PEL  
    mvx >>= img->mv_res; 
    mvy >>= img->mv_res;
#endif
    x_sub = (mvx < 0)? (4 - (-mvx%4)) % 4 : (mvx % 4);
    y_sub = (mvy < 0)? (4 - (-mvy%4)) % 4 : (mvy % 4);
    sub_pos = x_sub + (y_sub<<2);



    switch(img->filter[sub_pos])
    {
    case 0:
      if(sub_pos==7)  //subpel "g"//strong filter
        getblock2DFilt_quarter_pel       (ref_idx, list, x_pos, y_pos, mvx, mvy, img_width,img_height, block,size_x,size_y);
      else
        getblockH264HPFFilt_quarter_pel  (ref_idx, list, x_pos, y_pos, mvx, mvy, img_width,img_height, block,size_x,size_y);
      break;
    case 1:
    case 2:
      getblock2DFilt_quarter_pel       (ref_idx, list, x_pos, y_pos, mvx, mvy, img_width,img_height, block,size_x,size_y);
      break;

    case 3:
      getblockOneEighthFilt_quarter_pel(ref_idx, list, x_pos, y_pos, mvx, mvy, img_width,img_height, block,size_x,size_y);
      break;

    default:
      printf("sub_pos is out of range!\n");
      exit(1);
    }
    //add offset if its present 
    if(img->filterParam != SIFO_FILTERS_ONLY) 
    {
      if(ref_idx == 0)  
      {
        if(listX[LIST_0] == list)
          offset = img->subpelOffset[LIST_0][sub_pos];
        else 
          offset = img->subpelOffset[LIST_1][sub_pos];
      }
      else
      {
        if(list == listX[LIST_0])
        {
          if(listX[LIST_0][ref_idx] == listX[LIST_1][0])
            offset = img->subpelOffset[LIST_1][sub_pos];
          else
            offset = img->imgOffset[LIST_0][ref_idx];
        }
        else 
        {
          if (listX[LIST_1][ref_idx] == listX[LIST_0][0])
            offset = img->subpelOffset[LIST_0][sub_pos];
          else
            offset = img->imgOffset[LIST_1][ref_idx];
        }
      }
      if(offset)
      {
        for(i = 0; i < size_y; ++i) //y
          for(j = 0; j < size_x; ++j) //x
            block[j][i] = Clip3(0, img->max_imgpel_value, block[j][i] + offset);
      }
    }

    return sub_pos;
  }
}

const int ONE_FOURTH_TAP_HPF[3][2] =
{
  {20,20},
  {-5,-4},
  { 1, 0},
};

int getblockH264HPFFilt_quarter_pel(int ref_frame, StorablePicture **list, int x_pos, int y_pos, int mvx, int mvy, int img_width, int img_height, int block[16][16], int size_x, int size_y)
{
  int x_sub, y_sub, comp_x, comp_y, pos_y1, pos_y2, pos_x1, pos_x2;
  int i, j, ii, jj, ci;
  int maxold_x,maxold_y;
  int sub_pos, comp_sub_pos;

  int comp_block[2][16][16];
  int comp_loc[][2][2] =
  { 
    {{2,0},{0,0}},  //a {half, integer}
    {{2,0},{2,0}},  //b 
    {{2,0},{4,0}},  //c {half, integer}
    {{0,2},{0,0}},  //d {half, integer}
    {{2,0},{0,2}},  //e {half, half}
    {{2,0},{2,2}},  //f {half, "j"}
    {{2,0},{4,2}},  //g {half, half} 
    {{0,2},{0,2}},  //h
    {{0,2},{2,2}},  //i {half, "j"}
    {{2,2},{2,2}},  //j 
    {{4,2},{2,2}},  //k {half, "j"}
    {{0,2},{0,4}},  //l {half, integer}
    {{2,4},{0,2}},  //m {half, half}
    {{2,4},{2,2}},  //n {half, "j"}
    {{2,4},{4,2}},  //o {half, half}
  };
  int comp_num[] = {2,1,2,2,2,2,2,1,2,1,2,2,2,2,2}; // number of contributing components
  int AF[3], is;
  int temp_2D[6+16][16];
  imgpel **imgY = list[ref_frame]->imgY;

  x_sub = (mvx < 0)? (4 - (-mvx%4)) % 4 : (mvx % 4);
  y_sub = (mvy < 0)? (4 - (-mvy%4)) % 4 : (mvy % 4);
  sub_pos = x_sub + (y_sub<<2);

  //x_pos and y_pos are in full pel positions, so convert them to 1/4th pel positions
  x_pos = (x_pos<<2) + mvx;
  y_pos = (y_pos<<2) + mvy;


  // the upper left integer pel position
  if(x_pos < 0 && x_sub != 0)
    x_pos = (x_pos-4)/4;
  else
    x_pos /= 4;

  if(y_pos < 0 && y_sub != 0)
    y_pos = (y_pos-4)/4;
  else
    y_pos /= 4;

  maxold_x = img_width-1;
  maxold_y = img_height-1;

  if (sub_pos==0) 
  {  /* fullpel position */
    // Filter tap = 6, so filter offset = (6/2)-1 = 2 and block size is 4, so (max-2-4)=(max-6)
    if ( (y_pos >= 2) && (y_pos < maxold_y - 18) && (x_pos >= 2) && (x_pos < maxold_x - 18))
    {//Image Non-Border safe positions...so dont need Clipping
      for (i = 0; i < size_y; i++)
        for (j = 0; j < size_x; j++)      
          block[j][i] = imgY[y_pos+i][x_pos+j]; //block[x][y]
    }
    else
    {//Image Border positions...so needs Clipping
      for (i = 0; i < size_y; i++)
        for (j = 0; j < size_x; j++)      
          block[j][i] = imgY[Clip3Fun(0,maxold_y,y_pos+i)][Clip3Fun(0,maxold_x,x_pos+j)]; //block[x][y]
    }
  }
  else 
  {
    if ( (y_pos >= 2) && (y_pos < maxold_y - 18) && (x_pos >= 2) && (x_pos < maxold_x - 18))
    {
      //Image Non-Border safe positions...so dont need Clipping
      int inter_loc_x, inter_loc_y; // the location of to-be-interpolated subpel

      for(ci=0; ci<comp_num[sub_pos-1]; ci++) // contributing components
      {
        comp_x = comp_loc[sub_pos-1][ci][0];
        comp_y = comp_loc[sub_pos-1][ci][1];

        comp_sub_pos = comp_x%4 + ((comp_y%4)<<2);          // pos 1..15 in a 4x4 block

        if(comp_sub_pos==0) // integer pel
        {
          for (i = 0; i < size_y; i++)
            for (j = 0; j < size_x; j++)  
            {
              comp_block[ci][j][i] = imgY[y_pos+comp_y/4+i][x_pos+comp_x/4+j]; //comp_block[ci][x][y]
            }        
        }
        else if(comp_sub_pos ==10 ) // "j"
        {
          for(j=0; j<size_x; j++)
          {
            for(ii=-2; ii<=3+size_y; ii++)// y
            {
              is=0;
              pos_y1 = y_pos+ii;
              for(jj = 0; jj < 3; jj++) //x
              {
                inter_loc_x = ((x_pos+j)<<2)+2;
                pos_x1 = (inter_loc_x>>2)-2+jj;//(inter_loc_x/4)-2+jj;
                pos_x2 = pos_x1-jj + (5-jj);   //(inter_loc_x/4)-2+(5-jj);
                is += ONE_FOURTH_TAP_HPF[2-jj][0]*(imgY[pos_y1][pos_x1]+imgY[pos_y1][pos_x2]);
              }
              temp_2D[ii+2][j]=is;
            }
          }

          for(i=0; i<size_y; i++)
            for(j=0; j<size_x; j++)
            {
              //AF[0]=A+F; AF[1]=B+E; AF[2]=C+D;
              for(ii = 0; ii < 3; ii++)
              {
                AF[ii] = temp_2D[ii+i][j]+temp_2D[5-ii+i][j];
              }

              AF[0] -= AF[1]; //(A+F)-(B+E)
              AF[0] >>= 2;
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
              if (img->BitDepthIncrease==0)
              {
                AF[2] = max(-2371, min(18640, AF[2]));
              }
#else
              AF[2] = max(-2371, min(18640, AF[2]));
#endif
              AF[1] = AF[2] - AF[1];   // (C+D)-(B+E)
              AF[0] += AF[1];
              AF[0] >>= 2;
              AF[0] += AF[2];

              comp_block[ci][j][i] = AF[0];
            }
        }
        else //1-D half pel
        {
          char H_half_pel_positions = ((comp_sub_pos>>2)==0)? 1 : 0;
          char V_half_pel_positions = ((comp_sub_pos%4 )==0)? 1 : 0;

          for(i=0; i<size_y; i++)
            for(j=0; j<size_x; j++)
            {
              is = 0;

              if(H_half_pel_positions)    // horizontal 1/2 position 
              {
                inter_loc_y = ((i+y_pos)<<2)+comp_y;
                pos_y1 = inter_loc_y>>2;//inter_loc_y/4;
                for(jj = 0; jj < 3; jj++)
                {
                  inter_loc_x = ((j+x_pos)<<2)+comp_x;
                  pos_x1 = (inter_loc_x>>2)-2+jj;//inter_loc_x/4-2+jj;
                  pos_x2 = pos_x1-jj + (5-jj);   //inter_loc_x/4-2+(5-jj);
                  is += ONE_FOURTH_TAP_HPF[2-jj][0]*(imgY[pos_y1][pos_x1]+imgY[pos_y1][pos_x2]);
                }
                comp_block[ci][j][i] = is;
              }
              else if(V_half_pel_positions) // vertical 1/2 position
              {
                inter_loc_x = ((j+x_pos)<<2)+comp_x; 
                pos_x1 = inter_loc_x>>2;//inter_loc_x/4;
                for(ii = 0; ii < 3; ii++)
                {
                  inter_loc_y = ((i+y_pos)<<2)+comp_y;
                  pos_y1 = (inter_loc_y>>2)-2+ii;//inter_loc_y/4-2+ii;
                  pos_y2 = pos_y1-ii + (5-ii);   //inter_loc_y/4-2+(5-ii);
                  is += ONE_FOURTH_TAP_HPF[2-ii][0]*(imgY[pos_y1][pos_x1]+imgY[pos_y2][pos_x1]);
                }
                comp_block[ci][j][i] = is;
              } 
            }               
        }         
      }

    }
    else    //Image Border positions...so needs Clipping
    {
      int inter_loc_x, inter_loc_y; // the location of to-be-interpolated subpel

      for(ci=0; ci<comp_num[sub_pos-1]; ci++) // contributing components
      {
        comp_x = comp_loc[sub_pos-1][ci][0];
        comp_y = comp_loc[sub_pos-1][ci][1];

        comp_sub_pos = comp_x%4 + ((comp_y%4)<<2);          // pos 1..15 in a 4x4 block

        if(comp_sub_pos==0) // integer pel
        {
          for(i=0; i<size_y; i++)
            for(j=0; j<size_x; j++)
            {
              comp_block[ci][j][i] = imgY[Clip3Fun(0,maxold_y,y_pos+comp_y/4+i)][Clip3Fun(0,maxold_x,x_pos+comp_x/4+j)]; //comp_block[ci][x][y]
            }        
        }
        else if(comp_sub_pos ==10 ) // "j"
        {
          for(j=0; j<size_x; j++)
          {
            for(ii=-2; ii<=3+size_y; ii++)// y
            {
              is=0;
              inter_loc_y = y_pos+ii;
              pos_y1 = Clip3Fun(0, maxold_y, inter_loc_y);
              if(inter_loc_y<0 && pos_y1 > 0)
                pos_y1 -=1;

              for(jj = 0; jj < 3; jj++) //x
              {
                inter_loc_x = ((x_pos+j)<<2)+2;
                pos_x1 = Clip3Fun(0, maxold_x, inter_loc_x/4-2+jj);
                if(inter_loc_x<0 && pos_x1 > 0)
                  pos_x1 -=1;
                pos_x2 = Clip3Fun(0, maxold_x, inter_loc_x/4-2+(5-jj));
                if(inter_loc_x<0 && pos_x2 > 0)
                  pos_x2 -=1;
                is += ONE_FOURTH_TAP_HPF[2-jj][0]*(imgY[pos_y1][pos_x1]+imgY[pos_y1][pos_x2]);
              }
              temp_2D[ii+2][j]=is;
            }
          }

          for(i=0; i<size_y; i++)
            for(j=0; j<size_x; j++)
            {
              //AF[0]=A+F; AF[1]=B+E; AF[2]=C+D;
              for(ii = 0; ii < 3; ii++)
              {
                AF[ii] = temp_2D[ii+i][j]+temp_2D[5-ii+i][j];
              }

              AF[0] -= AF[1]; //(A+F)-(B+E)
              AF[0] >>= 2;
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
              if (img->BitDepthIncrease==0)
              {
                AF[2] = max(-2371, min(18640, AF[2]));
              }
#else
              AF[2] = max(-2371, min(18640, AF[2]));
#endif
              AF[1] = AF[2] - AF[1];   // (C+D)-(B+E)
              AF[0] += AF[1];
              AF[0] >>= 2;
              AF[0] += AF[2];

              comp_block[ci][j][i] = AF[0];
            }
        }
        else //1-D half pel
        {
          char H_half_pel_positions = ((comp_sub_pos>>2)==0)? 1 : 0;
          char V_half_pel_positions = ((comp_sub_pos%4 )==0)? 1 : 0;

          for(i=0; i<size_y; i++)
            for(j=0; j<size_x; j++)
            {
              is = 0;
              if(H_half_pel_positions)    // horizontal 1/2 position 
              {
                inter_loc_y = ((i+y_pos)<<2)+comp_y;
                pos_y1 = Clip3Fun(0, maxold_y, inter_loc_y/4);
                if(inter_loc_y<0 && pos_y1 > 0)// to be removed
                  pos_y1 -=1;

                for(jj = 0; jj < 3; jj++)
                {
                  inter_loc_x = ((j+x_pos)<<2)+comp_x;
                  pos_x1 = Clip3Fun(0, maxold_x, inter_loc_x/4-2+jj);
                  if(inter_loc_x<0 && pos_x1 > 0)
                    pos_x1 -=1;

                  inter_loc_x = ((j+x_pos)<<2)+comp_x;
                  pos_x2 = Clip3Fun(0, maxold_x, inter_loc_x/4-2+(5-jj));
                  if(inter_loc_x<0 && pos_x2 > 0)
                    pos_x2 -=1;

                  is += ONE_FOURTH_TAP_HPF[2-jj][0]*(imgY[pos_y1][pos_x1]+imgY[pos_y1][pos_x2]);
                }
                comp_block[ci][j][i] = is;
              }
              else if(V_half_pel_positions) // vertical 1/2 position
              {
                inter_loc_x = ((j+x_pos)<<2)+comp_x; 
                pos_x1 = Clip3Fun(0, maxold_x, inter_loc_x/4);
                if(inter_loc_x<0 && pos_x1 > 0)
                  pos_x1 -=1;

                for(ii = 0; ii < 3; ii++)
                {
                  inter_loc_y = ((i+y_pos)<<2)+comp_y;
                  pos_y1 = Clip3Fun(0, maxold_y, inter_loc_y/4-2+ii);
                  if(inter_loc_y<0 && pos_y1 > 0)
                    pos_y1 -=1;

                  inter_loc_y = ((i+y_pos)<<2)+comp_y;
                  pos_y2 = Clip3Fun(0, maxold_y, inter_loc_y/4-2+(5-ii));
                  if(inter_loc_y<0 && pos_y2 > 0)
                    pos_y2 -=1;

                  is += ONE_FOURTH_TAP_HPF[2-ii][0]*(imgY[pos_y1][pos_x1]+imgY[pos_y2][pos_x1]);
                }
                comp_block[ci][j][i] = is;
              } 
            }               
        }         

      }
    }

    //  F a b c     00 01 02 03
    //  d e f g     04 05 06 07
    //  h i j k     08 09 10 11
    //  l m n o     12 13 14 15

    // interpolate quarter-pel and shift down half-pel at the same time
    switch(sub_pos)
    {
    case 1:
    case 3:
    case 4:
    case 12: // a, c, d, l    //1/4th pel
      for(i=0; i<size_x; i++)    
        for(j=0; j<size_y; j++)
        {
          block[i][j] = (comp_block[0][i][j] + (comp_block[1][i][j]<<5) + 32) >> 6;
        }                
        break;

    case 2:
    case 8:  // b, h,       //half pel
      for(i=0; i<size_x; i++)    
        for(j=0; j<size_y; j++)
        {
          block[i][j] = (comp_block[0][i][j] + 16)>>5;
        }
        break;

    case 10: // j
      for(i=0; i<size_x; i++)    
        for(j=0; j<size_y; j++)
        {
          block[i][j] = (comp_block[0][i][j] + 32)>>6;
        }
        break;

    case 5:
    case 7:
    case 13:
    case 15:      // e, g, m, o
      for(i=0; i<size_x; i++)    
        for(j=0; j<size_y; j++)
        {
          block[i][j] = (comp_block[0][i][j] + comp_block[1][i][j] + 32) >> 6;
        }                
        break;

    case 6:
    case 9:
    case 11:
    case 14:  // i, k,f, n            
      for(i=0; i<size_x; i++)    
        for(j=0; j<size_y; j++)
        {
          block[i][j] = (comp_block[0][i][j] + (comp_block[1][i][j]>>1) + 32) >> 6;        
        }        
        break;
    default:
      printf("sub_pos is out of range!\n");
      exit(1);
    }

    for(i=0; i<size_x; i++)    
      for(j=0; j<size_y; j++)
      {
        block[i][j] = Clip3Fun(0, img->max_imgpel_value, block[i][j]);
      }
  }

  return sub_pos;
}

extern void get_eighthpel_block_direct(int ref_frame, StorablePicture **list, int x_pos, int y_pos, struct img_par *img, int block[16][16], int size_x, int size_y);
int getblockOneEighthFilt_quarter_pel(int ref_idx, StorablePicture **list, int x_pos, int y_pos,
                                      int mvx, int mvy, int img_width, int img_height, 
                                      int block[16][16], int size_x, int size_y)
{
  int sub_pos, x_sub, y_sub;
  int x8,y8;
  x_sub = (mvx < 0)? (4 - (-mvx%4)) % 4 : (mvx % 4);
  y_sub = (mvy < 0)? (4 - (-mvy%4)) % 4 : (mvy % 4);
  sub_pos = x_sub + (y_sub<<2);

  //get 1/8th position
  x8 = ((0 + x_pos) * 4 + mvx)<<1;
  y8 = ((0 + y_pos) * 4 + mvy)<<1;
  get_eighthpel_block_direct( ref_idx, list, x8, y8, img, block, size_x, size_y);
  return sub_pos;
}


int getblock2DFilt_quarter_pel(int ref_idx, StorablePicture **list, int x_pos, int y_pos,
                               int mvx, int mvy, int img_width, int img_height, 
                               int block[16][16], int size_x, int size_y)
{
  int i,j, ii, jj;
  int x4, y4, sub_pos, pos_x, pos_y, x_sub, y_sub;
  int is;
  int *FILTER = NULL;
  int FILTER_SIZE = 6;
  int FILTER_OFFSET = (FILTER_SIZE) / 2 - 1;
  imgpel **imgY = list[ref_idx]->imgY;
  int xFullpel,yFullpel,xFullpel_j,yFullpel_i;

  int *filtP;

  x_sub = (mvx < 0)? (4 - (-mvx%4)) % 4 : (mvx % 4);
  y_sub = (mvy < 0)? (4 - (-mvy%4)) % 4 : (mvy % 4);
  sub_pos = x_sub + (y_sub<<2);     // pos 1..15 in a 4x4 block

  xFullpel = ((x_pos<<2) + mvx)>>2;
  yFullpel = ((y_pos<<2) + mvy)>>2;

  //if(0)
  if((xFullpel>=2) && (xFullpel< img_width-2-size_x) && (yFullpel>=2) && (yFullpel< img_height-2-size_y))
  {
    for(i = 0; i < size_y; ++i) //y
      for(j = 0; j < size_x; ++j) //x
      {
        xFullpel_j = xFullpel + j;
        yFullpel_i = yFullpel + i;

        if(sub_pos != 0)
        {
          if(img->filter[sub_pos] == 0)
          {
            assert(sub_pos == 7);
            // No filter is needed in this case. Filter f0[4][4] is used below. 
          }
          else if(img->filter[sub_pos] == 1)
            FILTER = SYMMETRIC_1[sub_pos-1];
          else if (img->filter[sub_pos] == 2)
            FILTER = SYMMETRIC_2[sub_pos-1];
          else
            error("img->filterSequence should be less than 3", -1);

          if((sub_pos == 7)  && (img->filter[sub_pos] == 0))		// g
          {	
            // Strong filter position
            static const int f0[4][4] = {{  0,  5,  5,  0},
            {  5, 22, 22,  5},
            {  5, 22, 22,  5},
            {  0,  5,  5,  0}};
            int sfp = 0;

            sfp += 5 * (imgY[yFullpel_i][xFullpel_j-1] + imgY[yFullpel_i+1][xFullpel_j-1] +
              imgY[yFullpel_i-1][xFullpel_j] + imgY[yFullpel_i+2][xFullpel_j] +
              imgY[yFullpel_i-1][xFullpel_j+1] + imgY[yFullpel_i+2][xFullpel_j+1] +
              imgY[yFullpel_i][xFullpel_j+2] + imgY[yFullpel_i+1][xFullpel_j+2]);

            sfp += 22 * (imgY[yFullpel_i][xFullpel_j] + imgY[yFullpel_i+1][xFullpel_j] +
              imgY[yFullpel_i][xFullpel_j+1] + imgY[yFullpel_i+1][xFullpel_j+1]);

            block[j][i] = Clip3Fun(0, img->max_imgpel_value, (sfp + 64) >> 7);
          }
          else
          {
            //  F a b c     00 01 02 03
            //  d e f g     04 05 06 07
            //  h i j k     08 09 10 11
            //  l m n o     12 13 14 15

            int ii_start, ii_end, jj_start, jj_end;

            switch (sub_pos)
            {
            case 1: case 3:
              is = 0;
              filtP = &FILTER[12];
              for (jj=-2; jj <= 3; jj++)
              {
                is += (*filtP++) * imgY[yFullpel_i][xFullpel_j+jj];
              }
              break;
            case 4: case 12:
              is = FILTER[2] * imgY[yFullpel_i-2][xFullpel_j] + FILTER[8] * imgY[yFullpel_i-1][xFullpel_j] +
                FILTER[14] * imgY[yFullpel_i][xFullpel_j] + FILTER[20] * imgY[yFullpel_i+1][xFullpel_j] +
                FILTER[26] * imgY[yFullpel_i+2][xFullpel_j] + FILTER[32] * imgY[yFullpel_i+3][xFullpel_j];
              break;
            case 2:
              is = FILTER[12] * (imgY[yFullpel_i][xFullpel_j-2] + imgY[yFullpel_i][xFullpel_j+3]) +
                FILTER[13] * (imgY[yFullpel_i][xFullpel_j-1] + imgY[yFullpel_i][xFullpel_j+2]) +
                FILTER[14] * (imgY[yFullpel_i][xFullpel_j] + imgY[yFullpel_i][xFullpel_j+1]);
              break;
            case 8:
              is = FILTER[2] * (imgY[yFullpel_i-2][xFullpel_j] + imgY[yFullpel_i+3][xFullpel_j]) +
                FILTER[8] * (imgY[yFullpel_i-1][xFullpel_j] + imgY[yFullpel_i+2][xFullpel_j]) +
                FILTER[14] * (imgY[yFullpel_i][xFullpel_j] + imgY[yFullpel_i+1][xFullpel_j]);
              break;
            case 6: case 14:
              is = 0;
              for (ii=1; ii <= 4; ii++)
              {
                filtP = &FILTER[ii*FILTER_SIZE + 1];
                is += (*filtP++) * (imgY[yFullpel_i-2+ii][xFullpel_j-1] + imgY[yFullpel_i-2+ii][xFullpel_j+2]);
                is += (*filtP++) * (imgY[yFullpel_i-2+ii][xFullpel_j] + imgY[yFullpel_i-2+ii][xFullpel_j+1]);
              }
              break;
            case 9: case 11:
              is = 0;
              filtP = &FILTER[7];
              for (jj=-1; jj <= 2; jj++)
              {
                is += (*filtP++) * (imgY[yFullpel_i-1][xFullpel_j+jj] + imgY[yFullpel_i+2][xFullpel_j+jj]);
              }
              filtP = &FILTER[13];
              for (jj=-1; jj <= 2; jj++)
              {
                is += (*filtP++) * (imgY[yFullpel_i][xFullpel_j+jj] + imgY[yFullpel_i+1][xFullpel_j+jj]);
              }
              break;
            case 5: case 15:
              is = FILTER[7] * imgY[yFullpel_i-1][xFullpel_j-1] + FILTER[28] * imgY[yFullpel_i+2][xFullpel_j+2] +
                FILTER[8] * (imgY[yFullpel_i-1][xFullpel_j] + imgY[yFullpel_i][xFullpel_j-1]) +
                FILTER[9] * (imgY[yFullpel_i-1][xFullpel_j+1] + imgY[yFullpel_i+1][xFullpel_j-1]) +
                FILTER[14] * imgY[yFullpel_i][xFullpel_j] + FILTER[21] * (imgY[yFullpel_i+1][xFullpel_j+1]) +
                FILTER[16] * (imgY[yFullpel_i][xFullpel_j+2] + imgY[yFullpel_i+2][xFullpel_j]) +
                FILTER[22] * (imgY[yFullpel_i+1][xFullpel_j+2] + imgY[yFullpel_i+2][xFullpel_j+1]) +
                FILTER[10] * (imgY[yFullpel_i-1][xFullpel_j+2] + imgY[yFullpel_i+2][xFullpel_j-1]) +
                FILTER[15] * (imgY[yFullpel_i][xFullpel_j+1] + imgY[yFullpel_i+1][xFullpel_j]);
              break;
            case 10:
              is = 0;
              is += FILTER[8] *  (imgY[yFullpel_i-1][xFullpel_j] + imgY[yFullpel_i-1][xFullpel_j+1] +
                imgY[yFullpel_i][xFullpel_j-1] + imgY[yFullpel_i][xFullpel_j+2] +
                imgY[yFullpel_i+1][xFullpel_j-1] + imgY[yFullpel_i+1][xFullpel_j+2] +
                imgY[yFullpel_i+2][xFullpel_j] + imgY[yFullpel_i+2][xFullpel_j+1]);
              is += FILTER[14] * (imgY[yFullpel_i][xFullpel_j] + imgY[yFullpel_i][xFullpel_j+1] +
                imgY[yFullpel_i+1][xFullpel_j] + imgY[yFullpel_i+1][xFullpel_j+1]);
              if (img->filter[sub_pos] == 2)
              {
                is += FILTER[7] * (imgY[yFullpel_i-1][xFullpel_j-1] + imgY[yFullpel_i-1][xFullpel_j+2] +
                  imgY[yFullpel_i+2][xFullpel_j-1] + imgY[yFullpel_i+2][xFullpel_j+2]);
              }
              break;
            case 7: case 13:
              is = FILTER[10] * imgY[yFullpel_i-1][xFullpel_j+2] + FILTER[25] * imgY[yFullpel_i+2][xFullpel_j-1] +
                FILTER[9] * (imgY[yFullpel_i-1][xFullpel_j+1] + imgY[yFullpel_i][xFullpel_j+2]) +
                FILTER[8] * (imgY[yFullpel_i-1][xFullpel_j] + imgY[yFullpel_i+1][xFullpel_j+2]) +
                FILTER[15] * imgY[yFullpel_i][xFullpel_j+1] + FILTER[20] * (imgY[yFullpel_i+1][xFullpel_j]) +
                FILTER[13] * (imgY[yFullpel_i][xFullpel_j-1] + imgY[yFullpel_i+2][xFullpel_j+1]) +
                FILTER[19] * (imgY[yFullpel_i+1][xFullpel_j-1] + imgY[yFullpel_i+2][xFullpel_j]) +
                FILTER[7] * (imgY[yFullpel_i-1][xFullpel_j-1] + imgY[yFullpel_i+2][xFullpel_j+2]) +
                FILTER[14] * (imgY[yFullpel_i][xFullpel_j] + imgY[yFullpel_i+1][xFullpel_j+1]);
              break;
            default:
              ii_start=1;
              jj_start=1;
              ii_end=FILTER_SIZE-1;   //4x4 filter
              jj_end=FILTER_SIZE-1; 
              if(sub_pos==1 || sub_pos==2 || sub_pos==3)  //a,b,c horizontal filter
              {
                ii_start=2;
                ii_end=3;
                jj_start=0;
                jj_end=FILTER_SIZE; 
              }
              if(sub_pos==4 || sub_pos==8 || sub_pos==12) //d,h,l vertical filter
              {
                ii_start=0;
                ii_end=FILTER_SIZE;
                jj_start=2;
                jj_end=3; 
              }
              is = 0;

              for(ii = ii_start; ii < ii_end; ++ii)
              {
                filtP = FILTER + (FILTER_SIZE * ii) + jj_start;
                for(jj = jj_start; jj < jj_end; ++jj)
                {
                  pos_y = (yFullpel_i) - FILTER_OFFSET + ii;
                  pos_x = (xFullpel_j) - FILTER_OFFSET + jj;
                  is += ((*filtP++) * imgY[pos_y][pos_x]);
                }
              }
              break;

            }
            block[j][i] = Clip3Fun(0, img->max_imgpel_value, (is + 64) >> 7);
          }
        }
        else
        {
          block[j][i] = imgY[yFullpel_i][xFullpel_j];
        }
      }

  }  
  else
  {
    for(i = 0; i < size_y; ++i) //y
      for(j = 0; j < size_x; ++j) //x
      {
        x4 = ((j + x_pos) <<2) + mvx;
        y4 = ((i + y_pos) <<2) + mvy;

        if(sub_pos != 0)
        {
          if(img->filter[sub_pos] == 0)
            assert(sub_pos == 7);
          // No filter is needed in this case. Filter f0[4][4] is used below. 
          else if(img->filter[sub_pos] == 1)
            FILTER = SYMMETRIC_1[sub_pos-1];
          else if (img->filter[sub_pos] == 2)
            FILTER = SYMMETRIC_2[sub_pos-1];
          else
            error("img->filterSequence should be less than 3", -1);

          if((sub_pos == 7)  && (img->filter[sub_pos] == 0))		// g
          {	
            // Strong filter position
            static const int f0[4][4] = {{  0,  5,  5,  0},
            {  5, 22, 22,  5},
            {  5, 22, 22,  5},
            {  0,  5,  5,  0}};
            int sfp = 0;
            int l, k;
            for(l = 0; l < 4; ++l)
            {
              for(k = 0; k < 4; ++k)
              {
                pos_y = Clip3Fun(0, img_height - 1, y4 / 4 - 1 + l);
                if((y4 < 0) && (pos_y > 0))
                  pos_y -= 1;
                pos_x = Clip3Fun(0, img_width - 1, x4 / 4 - 1 + k);
                if((x4 < 0) && (pos_x > 0))
                  pos_x -=1;
                sfp += f0[l][k] * imgY[pos_y][pos_x];
              }
            }
            block[j][i] = Clip3Fun(0, (1 << img->bitdepth_luma) - 1, (sfp + 64) >> 7);
          }
          else
          {
            //  F a b c     00 01 02 03
            //  d e f g     04 05 06 07
            //  h i j k     08 09 10 11
            //  l m n o     12 13 14 15
            int ii_start, ii_end, jj_start, jj_end;
            ii_start=1;
            jj_start=1;
            ii_end=FILTER_SIZE-1;   //4x4 filter
            jj_end=FILTER_SIZE-1; 

            if(sub_pos==1 || sub_pos==2 || sub_pos==3)  //a,b,c horizontal filter
            {
              ii_start=2;
              ii_end=3;
              jj_start=0;
              jj_end=FILTER_SIZE; 
            }
            if(sub_pos==4 || sub_pos==8 || sub_pos==12) //d,h,l vertical filter
            {
              ii_start=0;
              ii_end=FILTER_SIZE;
              jj_start=2;
              jj_end=3; 
            }
            is = 0;

            for(ii = ii_start; ii < ii_end; ++ii)
            {
              filtP = FILTER + (FILTER_SIZE * ii) + jj_start;
              for(jj = jj_start; jj < jj_end; ++jj)
              {
                pos_y = Clip3(0, img_height - 1, y4 / 4 - FILTER_OFFSET + ii);
                if((y4 < 0) && (pos_y > 0))
                  pos_y -= 1;
                pos_x = Clip3(0, img_width - 1, x4 / 4 - FILTER_OFFSET + jj);
                if((x4 < 0) && (pos_x > 0))
                  pos_x -= 1;
                is += ((*filtP++) * imgY[pos_y][pos_x]);
              }
            }
            block[j][i] = Clip3Fun(0, img->max_imgpel_value, (is + 64) >> 7);
          }
        }
        else
        {
          block[j][i] = imgY[Clip3(0, img_height - 1, y4 / 4)][Clip3(0, img_width - 1, x4 / 4)];
        }
      }
  }
  return sub_pos;
}


void readSIFOHeader()
{
  Slice *currSlice = img->currentSlice;
  int dP_nr = assignSE2partition[currSlice->dp_mode][SE_HEADER];
  DataPartition *partition = &(currSlice->partArr[dP_nr]);
  Bitstream *currStream = partition->bitstream;



  if(img->type != I_SLICE)
  {

    img->filterParam = u_v(1, "SH: use offsets", currStream); // 1 bit only
    {
      // Read filters with bestFilter option
      int bestFilter;
      int sub_pos;

      memset(img->filter, 0, 16 * sizeof(int));

      bestFilter = u_v (3, "SH: bestFilter", currStream);

      if(bestFilter == 0)
      {
        for(sub_pos = 1; sub_pos < 16; ++sub_pos)
        {
          img->filter[sub_pos] = u_v (2, "SH: filter", currStream);
        }
      }
      else
      {
        for(sub_pos = 1; sub_pos < 16; ++sub_pos)
          img->filter[sub_pos] = bestFilter - 1;
      }
    }

    if(img->filterParam == SIFO_USE_OFFSETS)
    {
      // Read offsets
      int sub_pos, frame, nonzero, list;
      int listNo = (img->type == B_SLICE)? 2: 1;

      for(list = 0; list < listNo; ++list) 
      {    
        memset(img->subpelOffset[list], 0, 16 * sizeof(int));
        memset(img->imgOffset[list], 0, 16 * sizeof(int));

        nonzero = u_1("SH: nonzero offset", currStream);
        if(nonzero)
        {
          for(frame = 0; frame < listXsize[list]; ++frame)
          {       
            if(frame == 0)
            {     
              for(sub_pos = 0; sub_pos < 16; ++sub_pos)   
              {
                img->subpelOffset[list][sub_pos] = se_v("SH: subpelOffset", currStream);
                img->subpelOffset[list][sub_pos] *= (1<<img->BitDepthIncrease);
              }
            }
            else
            { 
              img->imgOffset[list][frame] = se_v("SH: imgOffset", currStream);
              img->imgOffset[list][frame] *= (1<<img->BitDepthIncrease);
            }
          }
        }
      }
    }
  }
}
