/* 
  ====================================================================================================================

	The copyright in this software is being made available under the License included below. 
	This software may be subject to other third party and 	contributor rights, including patent rights, and no such 
	rights are granted under this license.
 
	Copyright (c) 2010, QUALCOMM INC.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification, are permitted only for 
	the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
	promoting such standards. The following conditions are required to be met:

		* Redistributions of source code must retain the above copyright notice, this list of conditions and 
		  the following disclaimer.
		* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
		  the following disclaimer in the documentation and/or other materials provided with the distribution.
		* The name of QUALCOMM INC.  may not be used to endorse or promote products derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
	INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  ====================================================================================================================
*/

#ifndef _ADAPTIVE_LOOP_FILTER_H_
#define _ADAPTIVE_LOOP_FILTER_H_
#include "global.h"

//------- defines
#define MIN_NUM_TAP    5
#define MAX_NUM_TAP    9
#define MAX_NUM_TAP_C  5
#define NUM_BIT_SHIFT  8

#define MAX_NUM_COEF   42
#define MIN_NUM_COEF   8
#define MAX_NUM_COEF_C 14

#define MAX_ALF_PARAM_SIZE 50000
#define NUM_OF_REDESIGN 3
#define MIN_NUM_BLOCK  10
#define MAX_NUM_BLOCK  60000
#define PYRAMIDAL_MEM_SIZE 341
#define MIN_NUM_PARENT_BLOCK 7
#ifdef ADAPTIVE_LOOP_FILTER
typedef struct
{
  Boolean      valid_flag[PYRAMIDAL_MEM_SIZE];
  Boolean      part_flag[PYRAMIDAL_MEM_SIZE];
  Boolean      alf_blk_flag[PYRAMIDAL_MEM_SIZE];
} AlfQtBlock;
//------- end

//------- prototype
alf_parameter_set_rbsp_t *AllocALFPS(int *memory_size);
void   FreeALFPS (alf_parameter_set_rbsp_t *alfps);
int    InitALFGlobalBuffers(void);
void   FreeALFGlobalBurrers(void);
void   SetALFParameters(alf_parameter_set_rbsp_t *alfps, alf_parameter_set_rbsp_t *p);
void   MakeExtendFrame( imgpel** ImgDec, imgpel** ImgExt, int width, int height, int offset );
void   SetRestFrame(imgpel** ImgRest, imgpel** ImgDec, int width, int height);

//------- prediction of coeff
int    GetALFBaseType(int slice_type, int ref);
void   InitFilterBase();
void   SetALFDiffCoefChroma(alf_parameter_set_rbsp_t *alfps);
void   SetALFBaseCoefChroma(alf_parameter_set_rbsp_t *alfps);

void   CalcALFLambda(int qp);
int64  FindDistortionBlock(int blk_addr, imgpel** ImgOrg, imgpel** ImgCmp, int blk_size, int width_in_alf_blks);
void   CalcRDCost(imgpel** ImgOrg, imgpel** ImgCmp, alf_parameter_set_rbsp_t *alfps, int64* rate, int64* dist, double* cost);
void   CalcRDCostChroma(imgpel*** ImgUVOrg, imgpel*** ImgUVCmp, alf_parameter_set_rbsp_t *alfps, int64* rate, int64* dist, double* cost);
void   CalcRDCostBlock(imgpel** ImgOrg, imgpel** ImgCmp, imgpel** ImgDec, alf_parameter_set_rbsp_t *alfps, int64* rate, int64* dist, double* cost);
int    Gauss(double **a, int N);
void   FilterCoefQuickSort( double *coef_data, int *coef_num, int upper, int lower );
void   QuantFilterCoef(double* h, int* qh, int tap, int bitdepth);
void   ClearFilterCoefInt(int* qh, int N);

//------- filtering for luma
void   AdaptiveLoopFilterProcess(Picture *pic);
void   FirstFilteringFrameLuma(imgpel** ImgOrg, imgpel** ImgDec, imgpel** ImgRest);
void   FilterTapDecision(imgpel** ImgOrg, imgpel** ImgDec, imgpel** ImgRest);

//------- filtering for chroma
void   AdaptiveLoopFilterProcessChroma();
void   FilteringFrameChroma(imgpel ***ImgUVOrg, imgpel ***ImgUVDec, imgpel ***ImgUVRest);
void   FilteringProcessChroma(imgpel ***ImgUVDec, imgpel ***ImgUVRest, int *qh, int tap, int chroma_id);

//------- filtering based block
void   BlockAdaptiveFilterControl(imgpel** ImgOrg, imgpel** ImgDec, imgpel** ImgRest);
void   SetBlockSizeParameter(alf_parameter_set_rbsp_t *alfps, int size_idx);
int    CalcNumOfAlfBlocks(int blk_size, int width, int height);
int    SetALFBlockFlag(alf_parameter_set_rbsp_t *alfps, imgpel** ImgOrg, imgpel** ImgDec, imgpel** ImgRest);
void   CopyDecToRestBlock(alf_parameter_set_rbsp_t *alfps, imgpel** ImgDec, imgpel** ImgRest);
void   GetBlockPosForBALF(int blk_addr, int *x, int *y, int width_in_blks);
void   GetSamplePosForBALF(int blk_addr, int *x, int *y, int blk_size, int width_in_blks);
int    get_mem_blk_corr2( double ******blk_corr, int dim,int height, int width, int rows, int columns);
void   free_mem_blk_corr2( double *****blk_corr, int dim, int height, int width );
int    get_mem_blk_corr0(  double ****blk_corr, int dim,int height, int width);
void   free_mem_blk_corr0( double ***blk_corr, int dim);

int    ALFBlockIsAvailable(int blk_addr);
void   CheckAvailabilityOfNeighborsAlf();
void   GetNeighbourAlfBlock(int curr_alfb_nr, int xN, int yN, PixelPos *pix);
void   SetBlockPartitioningFlag(int pblk_addr, alf_parameter_set_rbsp_t *alfps, imgpel **ImgOrg, imgpel **ImgDec, imgpel **ImgRest);
void   TransQbpToFbp(int pblk_addr, alf_parameter_set_rbsp_t *alfps);
void   SetParentBlockSizeParam(alf_parameter_set_rbsp_t *alfps, int block_size);

//------- send ALF parameters
int    send_ALFParameterSet(alf_parameter_set_rbsp_t *alfps, Bitstream *bitstream, imgpel **ImgOrg, imgpel **ImgDec, imgpel **ImgRest);
int    send_ALFChromaParam(alf_parameter_set_rbsp_t *alfps, Bitstream *bitstream);
int    send_ALFBlockControlParam(alf_parameter_set_rbsp_t *alfps, Bitstream *bitstream, imgpel **ImgOrg, imgpel **ImgDec, imgpel **ImgRest);
int    send_BlockFlagCABAC(alf_parameter_set_rbsp_t *alfps, Bitstream *bitstream);
int    send_QuadTreeBlockPartitioningParam(alf_parameter_set_rbsp_t *alfps, Bitstream *bitstream, imgpel **ImgOrg, imgpel **ImgDec, imgpel **ImgRest);
int    TerminateSliceForALF(Picture* pic);

#endif // ADAPTIVE_LOOP_FILTER

#endif // _ADAPTIVE_LOOP_FILTER_H_
