/* 
  ====================================================================================================================

	The copyright in this software is being made available under the License included below. 
	This software may be subject to other third party and 	contributor rights, including patent rights, and no such 
	rights are granted under this license.
 
	Copyright (c) 2010, QUALCOMM INC.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification, are permitted only for 
	the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
	promoting such standards. The following conditions are required to be met:

		* Redistributions of source code must retain the above copyright notice, this list of conditions and 
		  the following disclaimer.
		* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
		  the following disclaimer in the documentation and/or other materials provided with the distribution.
		* The name of QUALCOMM INC.  may not be used to endorse or promote products derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
	INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  ====================================================================================================================
*/

#ifndef _GLOBAL_H_
#define _GLOBAL_H_

#include <stdio.h>

#include "defines.h"
#include "parset.h"
#include "q_matrix_offset.h"


//    defines min and max macros for non WIN32 environments
#if !defined(WIN32) || defined(__GNUC__)
#define max(a, b) (((a) > (b)) ? (a) : (b))
#define min(a, b) (((a) < (b)) ? (a) : (b))
#endif


// Tool defines (AIF, 1/8pel, MVC) is moved to define.h by AT 2006/11/01
#ifdef MV_COMPETITION
#define MAX_MV_PREDICTOR  8
#endif

#ifdef WIN32
#define  snprintf _snprintf
#endif

#if defined(WIN32) && !defined(__GNUC__)
typedef __int64   int64;
# define FORMAT_OFF_T "I64d"
#ifndef INT64_MIN
# define INT64_MIN        (-9223372036854775807i64 - 1i64)
#endif
#else
typedef long long int64;
# define FORMAT_OFF_T "lld"
#ifndef INT64_MIN
# define INT64_MIN        (-9223372036854775807LL - 1LL)
#endif
#endif

#ifdef WIN32
#define  snprintf _snprintf
#define  open     _open
#define  close    _close
#define  read     _read
#define  write    _write
#define  lseek    _lseeki64
#define  fsync    _commit
#define  OPENFLAGS_WRITE _O_WRONLY|_O_CREAT|_O_BINARY|_O_TRUNC
#define  OPEN_PERMISSIONS _S_IREAD | _S_IWRITE
#define  OPENFLAGS_READ  _O_RDONLY|_O_BINARY
#else
#define  OPENFLAGS_WRITE O_WRONLY|O_CREAT|O_TRUNC
#define  OPENFLAGS_READ  O_RDONLY
#define  OPEN_PERMISSIONS S_IRUSR | S_IWUSR
#endif


/***********************************************************************
* T y p e    d e f i n i t i o n s    f o r    T M L
***********************************************************************
*/

//#define pel_t byte

#define imgpel unsigned short
#define distpel int
//#define imgpel byte
//#define distpel unsigned short

#define pel_t imgpel

//! Data Partitioning Modes
typedef enum
{
  PAR_DP_1,   //!< no data partitioning is supported
    PAR_DP_3    //!< data partitioning with 3 partitions
} PAR_DP_TYPE;


//! Output File Types
typedef enum
{
  PAR_OF_ANNEXB,    //!< Annex B byte stream format
    PAR_OF_RTP       //!< RTP packets in outfile
} PAR_OF_TYPE;

//! Field Coding Types
typedef enum 
{
  FRAME_CODING
} CodingType;

//! definition of H.264 syntax elements
typedef enum 
{
  SE_HEADER,
    SE_PTYPE,
    SE_MBTYPE,
    SE_REFFRAME,
    SE_INTRAPREDMODE,
    SE_MVD,
#ifdef MV_COMPETITION
    SE_MV_PREDICTOR,
#endif
    SE_CBP_INTRA,
    SE_LUM_DC_INTRA,
    SE_CHR_DC_INTRA,
    SE_LUM_AC_INTRA,
    SE_CHR_AC_INTRA,
    SE_CBP_INTER,
    SE_LUM_DC_INTER,
    SE_CHR_DC_INTER,
    SE_LUM_AC_INTER,
    SE_CHR_AC_INTER,
    SE_DELTA_QUANT_INTER,
    SE_DELTA_QUANT_INTRA,
    SE_BFRAME,
    SE_EOS,
    SE_MAX_ELEMENTS  //!< number of maximum syntax elements
} SE_type;         // substituting the definitions in elements.h


typedef enum 
{
  INTER_MB,
    INTRA_MB_4x4,
    INTRA_MB_16x16
} IntraInterDecision;


typedef enum 
{
  BITS_HEADER,
    BITS_TOTAL_MB,
    BITS_MB_MODE,
    BITS_INTER_MB,
    BITS_CBP_MB,
    BITS_COEFF_Y_MB,
    BITS_COEFF_UV_MB,
    BITS_DELTA_QUANT_MB,
#ifdef MV_COMPETITION
    BITS_SKIP_PRED,
    BITS_MV_PRED,
#endif
    MAX_BITCOUNTER_MB
} BitCountType;


typedef enum 
{
  NO_SLICES,
    FIXED_MB,
    FIXED_RATE,
    CALLBACK,
    FMO
} SliceMode;


typedef enum 
{
  UVLC,
    CABAC
} SymbolMode;


typedef enum 
{
  FRAME
} PictureStructure;           //!< New enum for field processing

typedef enum 
{
  P_SLICE = 0,
    B_SLICE,
    I_SLICE,
    SP_SLICE,
    SI_SLICE
} SliceType;

/***********************************************************************
* D a t a    t y p e s   f o r  C A B A C
***********************************************************************
*/

//! struct to characterize the state of the arithmetic coding engine
typedef struct
{
  unsigned int  Elow, Erange;
  unsigned int  Ebuffer;
  unsigned int  Ebits_to_go;
  unsigned int  Ebits_to_follow;
  byte          *Ecodestrm;
  int           *Ecodestrm_len;
  int           C;
  int           E;
  
  // storage in case of recode MB
  unsigned int  ElowS, ErangeS;
  unsigned int  EbufferS;
  unsigned int  Ebits_to_goS;
  unsigned int  Ebits_to_followS;
  byte          *EcodestrmS;
  int           *Ecodestrm_lenS;
  int           CS;
  int           ES;
} EncodingEnvironment;

typedef EncodingEnvironment *EncodingEnvironmentPtr;

//! struct for context management
typedef struct
{
  unsigned short state;         // index into state-table CP  
  unsigned char  MPS;           // Least Probable Symbol 0/1 CP
  
  unsigned long  count;
  
} BiContextType;

typedef BiContextType *BiContextTypePtr;


/**********************************************************************
* C O N T E X T S   F O R   T M L   S Y N T A X   E L E M E N T S
**********************************************************************
*/


#define NUM_MB_TYPE_CTX  11
#define NUM_B8_TYPE_CTX  9
#define NUM_MV_RES_CTX   10
#define NUM_REF_NO_CTX   6
#define NUM_DELTA_QP_CTX 4
#define NUM_MB_AFF_CTX 4

#define NUM_TRANSFORM_SIZE_CTX 3
#ifdef ADAPTIVE_LOOP_FILTER
#define NUM_ALF_BLK_CTX  3
#endif

#ifdef ADAPTIVE_MV_RES
#define NUM_ADP_MV_RES_CTX 4
#endif

typedef struct
{
  BiContextType mb_type_contexts [3][NUM_MB_TYPE_CTX];
  BiContextType b8_type_contexts [2][NUM_B8_TYPE_CTX];
  BiContextType mv_res_contexts  [2][NUM_MV_RES_CTX];
#ifdef MV_COMPETITION
  BiContextType mv_predictor_skip_contexts  [MAX_MV_PREDICTOR];
  BiContextType mv_predictor_mvp_contexts   [MAX_MV_PREDICTOR];
  BiContextType mv_predictor_mvb_contexts   [MAX_MV_PREDICTOR];
#endif
  BiContextType ref_no_contexts  [2][NUM_REF_NO_CTX];
  BiContextType delta_qp_contexts   [NUM_DELTA_QP_CTX];
  BiContextType mb_aff_contexts     [NUM_MB_AFF_CTX];
  BiContextType transform_size_contexts   [NUM_TRANSFORM_SIZE_CTX];

#ifdef ADAPTIVE_LOOP_FILTER
  BiContextType alf_blk_flag_contexts     [NUM_ALF_BLK_CTX];
#endif

#ifdef ADAPTIVE_MV_RES
  BiContextType adp_mv_res_contexts[2][NUM_ADP_MV_RES_CTX];
#endif


} MotionInfoContexts;


#define NUM_IPR_CTX    2
#define NUM_CIPR_CTX   4
#define NUM_CBP_CTX    4
#define NUM_BCBP_CTX   4
#define NUM_MAP_CTX   15
#define NUM_LAST_CTX  15
#define NUM_ONE_CTX    5
#define NUM_ABS_CTX    5


typedef struct
{
  BiContextType  ipr_contexts [NUM_IPR_CTX]; 
  BiContextType  cipr_contexts[NUM_CIPR_CTX]; 
  BiContextType  cbp_contexts [3][NUM_CBP_CTX];
  BiContextType  bcbp_contexts[NUM_BLOCK_TYPES][NUM_BCBP_CTX];
  BiContextType  map_contexts [NUM_BLOCK_TYPES][NUM_MAP_CTX];
  BiContextType  last_contexts[NUM_BLOCK_TYPES][NUM_LAST_CTX];
  BiContextType  one_contexts [NUM_BLOCK_TYPES][NUM_ONE_CTX];
  BiContextType  abs_contexts [NUM_BLOCK_TYPES][NUM_ABS_CTX];
  BiContextType  fld_map_contexts [NUM_BLOCK_TYPES][NUM_MAP_CTX];
  BiContextType  fld_last_contexts[NUM_BLOCK_TYPES][NUM_LAST_CTX];
} TextureInfoContexts;

//*********************** end of data type definition for CABAC *******************

typedef struct pix_pos
{
  int available;
  int mb_addr;
  int x;
  int y;
  int pos_x;
  int pos_y;
} PixelPos;

/*! Buffer structure for RMPNI commands */
typedef struct RMPNIbuffer_s
{
  int RMPNI;
  int Data;
  struct RMPNIbuffer_s *Next;
} RMPNIbuffer_t;

/*! Buffer structure for decoded referenc picture marking commands */
typedef struct DecRefPicMarking_s
{
  int memory_management_control_operation;
  int difference_of_pic_nums_minus1;
  int long_term_pic_num;
  int long_term_frame_idx;
  int max_long_term_frame_idx_plus1;
  struct DecRefPicMarking_s *Next;
} DecRefPicMarking_t;

//! Syntaxelement
typedef struct syntaxelement
{
  int                 type;           //!< type of syntax element for data part.
  int                 value1;         //!< numerical value of syntax element
  int                 value2;         //!< for blocked symbols, e.g. run/level
  int                 dir_partition1;         //direction (0,1,2) for partition1
  int                 dir_partition2;         //direction (0,1,2) for partition2
  int                 mb_ext_level;
  int                 len;            //!< length of code
  int                 inf;            //!< info part of UVLC code
  unsigned int        bitpattern;     //!< UVLC bitpattern
  int                 context;        //!< CABAC context
  int                 k;              //!< CABAC context for coeff_count,uv
  
#if TRACE
#define             TRACESTRING_SIZE 100            //!< size of trace string
  char                tracestring[TRACESTRING_SIZE];  //!< trace string
#endif
  
  //!< for mapping of syntaxElement to UVLC
  void    (*mapping)(int value1, int value2, int* len_ptr, int* info_ptr);
  //!< used for CABAC: refers to actual coding method of each individual syntax element type
  void    (*writing)(struct syntaxelement *, EncodingEnvironmentPtr);
} SyntaxElement;

//! Macroblock
typedef struct macroblock
{
  int                 currSEnr;                   //!< number of current syntax element
  int                 slice_nr;
  int                 delta_qp;
  int                 qp;                         //!< QP luma
  int                 qpc[2];                     //!< QP chroma
  int                 bitcounter[MAX_BITCOUNTER_MB];
  
  struct macroblock   *mb_available_up;   //!< pointer to neighboring MB (CABAC)
  struct macroblock   *mb_available_left; //!< pointer to neighboring MB (CABAC)
  
  int                 mb_type;
#ifdef GEOM
  int                 geo_mode;
  char                geo_mask[16][16];
#endif

  int                 mvd[2][BLOCK_MULTIPLE][BLOCK_MULTIPLE][2];          //!< indices correspond to [forw,backw][block_y][block_x][x,y]
  char                intra_pred_modes[MB_BLOCK_PARTITIONS];
  char                intra_pred_modes8x8[MB_BLOCK_PARTITIONS];                             //!< four 8x8 blocks in a macroblock
  int                 cbp ;
  int64               cbp_blk ;    //!< 1 bit set for every 4x4 block with coefs (not implemented for INTRA)
  int                 b8mode[4];
  int                 b8pdir[4];
  int64               cbp_bits;
  
  int                 c_ipred_mode;      //!< chroma intra prediction mode
  int                 IntraChromaPredModeFlag;
  
  int                 list_offset;
  
  int mbAddrA, mbAddrB, mbAddrC, mbAddrD;
  int mbAvailA, mbAvailB, mbAvailC, mbAvailD;
  
  int                 all_blk_8x8;
  int                 luma_transform_size_8x8_flag;
  int                 NoMbPartLessThan8x8Flag;
  short    bi_pred_me;
  
  int                 prev_qp;
  int                 prev_delta_qp;
  int                 prev_cbp;
  
  int                 LFDisableIdc;
  int                 LFAlphaC0Offset;
  int                 LFBetaOffset;
  
  int                 skip_flag;

#ifdef MV_COMPETITION
  int                 best_predictor_for_skip;  // Index of the best predictor for the skip mode
  short               send_index_predictor_skip; 
  short               send_index_predictor_mv; 
#endif


  int write_mb32_mbtype;  // this is used to indicate if a full MB32x32 is available or not. If not, 
                          // no need to send mb_type in function writeMBLayer_Big(). 
  int write_mb64_mbtype;  // this is used to indicate if a full MB64x64 is available or not. If not, 
                          // no need to send mb_type in function writeMBLayer_Big().
} Macroblock;


//! Bitstream
typedef struct
{
  int             byte_pos;           //!< current position in bitstream;
  int             bits_to_go;         //!< current bitcounter
  byte            byte_buf;           //!< current buffer for last written byte
  int             stored_byte_pos;    //!< storage for position in bitstream;
  int             stored_bits_to_go;  //!< storage for bitcounter
  byte            stored_byte_buf;    //!< storage for buffer of last written byte
  
  byte            byte_buf_skip;      //!< current buffer for last written byte
  int             byte_pos_skip;      //!< storage for position in bitstream;
  int             bits_to_go_skip;    //!< storage for bitcounter
  
  byte            *streamBuffer;      //!< actual buffer for written bytes
  int             write_flag;         //!< Bitstream contains data and needs to be written
  
} Bitstream;

//! DataPartition
typedef struct datapartition
{
  
  Bitstream           *bitstream;
  EncodingEnvironment ee_cabac;
  
  int                 (*writeSyntaxElement)(SyntaxElement *, struct datapartition *);
  /*!< virtual function;
  actual method depends on chosen data partition and
  entropy coding method  */
} DataPartition;

//! Slice
typedef struct
{
  int                 picture_id;
  int                 qp;
  int                 picture_type; //!< picture type
  int                 start_mb_nr;
  int                 max_part_nr;  //!< number of different partitions
  int                 num_mb;       //!< number of MBs in the slice
  DataPartition       *partArr;     //!< array of partitions
  MotionInfoContexts  *mot_ctx;     //!< pointer to struct of context models for use in CABAC
  TextureInfoContexts *tex_ctx;     //!< pointer to struct of context models for use in CABAC
  
  RMPNIbuffer_t        *rmpni_buffer; //!< stores the slice temporary buffer remapping commands
  
  int                 ref_pic_list_reordering_flag_l0;
  int                 *reordering_of_pic_nums_idc_l0;
  int                 *abs_diff_pic_num_minus1_l0;
  int                 *long_term_pic_idx_l0;
  int                 ref_pic_list_reordering_flag_l1;
  int                 *reordering_of_pic_nums_idc_l1;
  int                 *abs_diff_pic_num_minus1_l1;
  int                 *long_term_pic_idx_l1;
  
  Boolean             (*slice_too_big)(int bits_slice); //!< for use of callback functions
#ifdef ADAPTIVE_LOOP_FILTER
  int                 alf_flag_bit_to_go;
  byte                *alf_flag_stream_pointer;
  int                 sh_byte_pos;
  int                 model_number;
#endif
} Slice;


#define MAXSLICEPERPICTURE 100
typedef struct 
{
  int   no_slices;
  int   idr_flag;
  Slice *slices[MAXSLICEPERPICTURE];
  int bits_per_picture;
  float distortion_y;
  float distortion_u;
  float distortion_v;
} Picture;

Picture *frame_pic;
Picture *frame_pic_1;
Picture *frame_pic_2;
Picture *frame_pic_3;
#ifdef ADAPTIVE_LOOP_FILTER
typedef struct
{
  int          tap;
  //int          num_coeff;
  //int          *coeff;
  int          tap_chroma;
  int          num_coeff_chroma;
  int          *coeff_chroma;
} ALFParam;

typedef struct alfblock
{
  Boolean           alf_blk_flag;
  int               currSEnr;             //!< number of current syntax element
  struct alfblock   *alfb_available_up;   //!< pointer to neighboring Block (CABAC)
  struct alfblock   *alfb_available_left; //!< pointer to neighboring Block (CABAC)
  int               alfbAddrA, alfbAddrB, alfbAddrC, alfbAddrD;
  int               alfbAvailA, alfbAvailB, alfbAvailC, alfbAvailD;
} AlfBlock;

typedef struct
{
  int          alf_flag;
  int          chroma_idc;
  ALFParam     alf;
  int          block_control_flag;
  int          alf_block_size_idx;
  int          alf_block_size;
  int          WidthInAlfBlks;
  int          HeightInAlfBlks;
  int          num_alf_block;
  AlfBlock     *alf_blk_data;
  int          qt_partition_flag;
  int          max_layer_level;
  int          parent_block_size;
  int          WidthInParentBlks;
  int          HeightInParentBlks;
  int          num_parent_block;
} alf_parameter_set_rbsp_t;

alf_parameter_set_rbsp_t *active_alfps;
#endif // ADAPTIVE_LOOP_FILTER

// global picture format dependend buffers, mem allocation in image.c
imgpel **imgY_org;           //!< Reference luma image
imgpel ***imgUV_org;         //!< Reference chroma image
int    **img4Y_tmp;          //!< for quarter pel interpolation

unsigned int *yPicPos;
unsigned int *xPicPos;
unsigned int log2_max_frame_num_minus4;
unsigned int log2_max_pic_order_cnt_lsb_minus4;

int  me_tot_time,me_time;
pic_parameter_set_rbsp_t *active_pps;
seq_parameter_set_rbsp_t *active_sps;


int dsr_new_search_range; //!<Dynamic Search Range.
//////////////////////////////////////////////////////////////////////////
// B pictures
// motion vector : forward, backward, direct


//Weighted prediction
int ***wp_weight;  // weight in [list][index][component] order
int ***wp_offset;  // offset in [list][index][component] order
int ****wbp_weight;  // weight in [list][fwd_index][bwd_idx][component] order

int luma_log_weight_denom;
int chroma_log_weight_denom;
int wp_luma_round;
int wp_chroma_round;

// global picture format dependend buffers, mem allocation in image.c (field picture)
imgpel   **imgY_org_frm;
imgpel  ***imgUV_org_frm;

imgpel   **imgY_com;               //!< Encoded luma images
imgpel  ***imgUV_com;              //!< Encoded croma images

char    ***direct_ref_idx;           //!< direct mode reference index buffer
short    **direct_pdir;              //!< direct mode reference index buffer

int intras;         //!< Counts the intra updates in each frame.

int  frame_ctr[5];
int  frame_no, nextP_tr_fld, nextP_tr_frm;
int  tot_time;

#define ET_SIZE 300      //!< size of error text buffer
char errortext[ET_SIZE]; //!< buffer for error message for exit with error()

// Residue Color Transform
int dc_level[2][4][4], dc_level_temp[2][4][4];
char b8_ipredmode8x8[4][4], b8_intra_pred_modes8x8[16];

typedef struct mbdata
{
  int mb_type;
  int qp;
  int prev_qp;
  int prev_delta_qp;
  int level_of_child;
  short mv[2][MAXMODE][4][2];
  char ref_idx[2][MAXMODE][4];
  char pdir[MAXMODE][4];
  short bipred_mv[2][2][4][2];
  short bi_pred_me;
  struct mbdata *NextMB[4];
#ifdef GEOM//GEOM_RESTURCTURE
  int geo_mode;
  int best8x8fwref_geo[2];
  int best8x8pdir_geo[2];
  int best8x8bwref_geo[2];
  short geo_mv[2][2][2];
#endif
}MBdata;

typedef struct
{
  int  **resY;               //!< Residue of Luminance
  imgpel ***decY;            //!< Decoded values at the simulated decoders
  imgpel ****decref;         //!< Reference frames of the simulated decoders
  imgpel ***decY_best;       //!< Decoded frames for the best mode for all decoders
  imgpel **RefBlock;
  byte **status_map;
  byte **dec_mb_mode;
} Decoders;
extern Decoders *decs;

//! SNRParameters
typedef struct
{
  float snr_y;               //!< current Y SNR
  float snr_u;               //!< current U SNR
  float snr_v;               //!< current V SNR
  float snr_y1;              //!< SNR Y(dB) first frame
  float snr_u1;              //!< SNR U(dB) first frame
  float snr_v1;              //!< SNR V(dB) first frame
  float snr_yt[5];           //!< SNR Y(dB) based on frame type
  float snr_ut[5];           //!< SNR U(dB) based on frame type
  float snr_vt[5];           //!< SNR V(dB) based on frame type
  float snr_ya;              //!< Average SNR Y(dB) remaining frames
  float snr_ua;              //!< Average SNR U(dB) remaining frames
  float snr_va;              //!< Average SNR V(dB) remaining frames
  float sse_y;               //!< SSE Y
  float sse_u;               //!< SSE U
  float sse_v;               //!< SSE V
  float msse_y;              //!< Average SSE Y
  float msse_u;              //!< Average SSE U
  float msse_v;              //!< Average SSE V
  int   frame_ctr;           //!< number of coded frames
} SNRParameters;

#define FILE_NAME_SIZE 200
//! all input parameters
typedef struct
{
  int HierP_Level;
  int HierP_QPoffset;
  int ProfileIDC;               //!< profile idc
  int LevelIDC;                 //!< level idc

  int no_frames;                //!< number of frames to be encoded
  int qp0;                      //!< QP of first frame
  int qpN;                      //!< QP of remaining frames
  int jumpd;                    //!< number of frames to skip in input sequence (e.g 2 takes frame 0,3,6,9...)
  int mv_res;                   //!< motion vector resolution: 0 = 1/4-pel, 1 = 1/8-pel
#ifdef MV_COMPETITION
  int mv_competition;
  char predictors_skip[1024];		//contains maximum 12 chars...check in encoder.cfg
  char predictors_mvp[1024];
  char predictors_mvb[1024];
#endif
  int DisableSmallTransInLargeMB;
  int hadamard;                 /*!< 0: 'normal' SAD in sub pixel search.  1: use 4x4 Hadamard transform and '
                                Sum of absolute transform difference' in sub pixel search                   */
  int DisableSubpelME;          //!< Disable Subpixel Motion Estimation
  int search_range;             /*!< search range - integer pel search and 16x16 blocks.  The search window is
                                generally around the predicted vector. Max vector is 2xmcrange.  For 8x8
                                and 4x4 block sizes the search range is 1/2 of that for 16x16 blocks.       */
  int num_ref_frames;           //!< number of reference frames to be used
  int P_List0_refs;
  int B_List1_refs;
  int B_List0_refs;
  int Log2MaxFNumMinus4;
  int Log2MaxPOCLsbMinus4;

  int img_width;                //!< image width  (must be a multiple of 16 pels)
  int img_height;               //!< image height (must be a multiple of 16 pels)
  int yuv_format;               //!< YUV format (0=4:0:0, 1=4:2:0, 2=4:2:2, 3=4:4:4)
  int intra_upd;                /*!< For error robustness. 0: no special action. 1: One GOB/frame is intra coded
                                as regular 'update'. 2: One GOB every 2 frames is intra coded etc.
                                In connection with this intra update, restrictions is put on motion vectors
                                to prevent errors to propagate from the past                                */
  int blc_size[8][2];           //!< array for different block sizes
  int part_size[8][2];          //!< array for different partition sizes
  int slice_mode;               //!< Indicate what algorithm to use for setting slices
  int slice_argument;           //!< Argument to the specified slice algorithm
  int UseConstrainedIntraPred;  //!< 0: Inter MB pixels are allowed for intra prediction 1: Not allowed
  int  infile_header;           //!< If input file has a header set this to the length of the header
  char infile[FILE_NAME_SIZE];             //!< YUV 4:2:0 input format
  char outfile[FILE_NAME_SIZE];            //!< H.264 compressed output bitstream
  char ReconFile[FILE_NAME_SIZE];          //!< Reconstructed Pictures
  char TraceFile[FILE_NAME_SIZE];          //!< Trace Outputs
  int intra_period;             //!< Random Access period though intra
  int EnableOpenGOP;            //!< support for open gops.

  int idr_enable;         //!< Encode intra slices as IDR
  int start_frame;        //!< Encode sequence starting from Frame start_frame

  // B pictures
  int successive_Bframe;        //!< number of B frames that will be used
  int qpB;                      //!< QP for non-reference B slice coded pictures
  int qpBRSOffset;                     //!< QP for reference B slice coded pictures
  int direct_spatial_mv_pred_flag;              //!< Direct Mode type to be used (0: Temporal, 1: Spatial)
  int directInferenceFlag;      //!< Direct Inference Flag

  int BiPredMotionEstimation;
  int BiPredMERefinements;
  int BiPredMESearchRange;
  int BiPredMESubPel;

  int SkipIntraInInterSlices;    //!< Skip intra type checking in inter slices if best_mode is skip/direct
  int BRefPictures;              //!< B coded reference pictures replace P pictures (0: not used, 1: used)
  int HierarchicalCoding;
  int HierarchyLevelQPEnable;
  char ExplicitHierarchyFormat[1024];  //!< Explicit GOP format (HierarchicalCoding==3). 
  int ReferenceReorder;          //!< Reordering based on Poc distances
  int PocMemoryManagement;       //!< Memory management based on Poc distances for hierarchical coding

  int symbol_mode;              //!< Specifies the mode the symbols are mapped on bits
  int of_mode;                  //!< Specifies the mode of the output file
  int InterSearchSkip;
  int InterSearch16x16;
  int InterSearch16x8;
  int InterSearch8x16;
  int InterSearch8x8;
  int InterSearch8x4;
  int InterSearch4x8;
  int InterSearch4x4;

  int IntraDisableInterOnly;
  int Intra4x4ParDisable;
  int Intra4x4DiagDisable;
  int Intra4x4DirDisable;
  int Intra16x16ParDisable;
  int Intra16x16PlaneDisable;
  int ChromaIntraDisable;

  double FrameRate;

  int EPZSPattern;
  int EPZSDual;
  int EPZSFixed;
  int EPZSTemporal;
  int EPZSSpatialMem;
  int EPZSMinThresScale;
  int EPZSMaxThresScale;
  int EPZSMedThresScale;
  int EPZSSubPelME;
  int EPZSSubPelThresScale;

  int chroma_qp_index_offset;
#ifdef _FULL_SEARCH_RANGE_
  int full_search;
#endif
#ifdef _ADAPT_LAST_GROUP_
  int last_frame;
#endif
#ifdef _CHANGE_QP_
  int qpN2, qpB2, qp2start;
  int qp02, qpBRS2Offset;
#endif
  int rdopt;
  int disthres;
  int nobskip;
#ifdef _LEAKYBUCKET_
  int NumberLeakyBuckets;
  char LeakyBucketRateFile[FILE_NAME_SIZE];
  char LeakyBucketParamFile[FILE_NAME_SIZE];
#endif
  int PicInterlace;           //!< picture adaptive frame/field
  int MbInterlace;            //!< macroblock adaptive frame/field

  int IntraBottom;            //!< Force Intra Bottom at GOP periods.

  int LossRateA;              //!< assumed loss probablility of partition A (or full slice), in per cent, used for loss-aware R/D optimization
  int LossRateB;              //!< assumed loss probablility of partition B, in per cent, used for loss-aware R/D 
  int LossRateC;              //!< assumed loss probablility of partition C, in per cent, used for loss-aware R/D 
  int NoOfDecoders;
  int NumFramesInELSubSeq;
  int NumFrameIn2ndIGOP;


  int LFSendParameters;
  int LFDisableIdc;
  int LFAlphaC0Offset;
  int LFBetaOffset;

  int SparePictureOption;
  int SPDetectionThreshold;
  int SPPercentageThreshold;

  // FMO
  char SliceGroupConfigFileName[FILE_NAME_SIZE];    //!< Filename for config info fot type 0, 2, 6
  int num_slice_groups_minus1;           //!< "FmoNumSliceGroups" in encoder.cfg, same as FmoNumSliceGroups, which should be erased later
  int slice_group_map_type; 

  int *top_left;                         //!< top_left and bottom_right store values indicating foregrounds
  int *bottom_right; 
  byte *slice_group_id;                   //!< slice_group_id is for slice group type being 6  
  int *run_length_minus1;                //!< run_length_minus1 is for slice group type being 0

  int slice_group_change_direction_flag;
  int slice_group_change_rate_minus1;
  int slice_group_change_cycle;

  int redundant_pic_flag;   //! encoding of redundant pictures
  int pic_order_cnt_type;   //! POC type

  int context_init_method;
  int model_number;
  int Transform8x8Mode;
  int ReportFrameStats;
  int DisplayEncParams;
  int Verbose;

  int RCEnable;    
  int bit_rate;
  int SeinitialQP;
  int basicunit;
  int channel_type;

  // FastME enable
  int FMEnable;
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
  int InputBitDepth;
#endif
#ifdef USE_INTRA_MDDT
  int UseIntraMDDT; 
#endif
#ifdef RDO_Q
  int UseRDO_Q;
#endif
#ifdef USE_HP_FILTER
  int UseHPFilter;
#endif
  int DynamicSearchRange;//!< Dynamic Search Range
  int FMEScale;
  //////////////////////////////////////////////////////////////////////////
  // Fidelity Range Extensions
  int BitDepthLuma;
  int BitDepthChroma;
  int img_height_cr;
  int img_width_cr;
  int rgb_input_flag;
  int cb_qp_index_offset;
  int cr_qp_index_offset;

  // Lossless Coding
  int lossless_qpprime_y_zero_flag;

  //Residue Color Transform
  int residue_transform_flag;

  // Lambda Params
  int UseExplicitLambdaParams;
  double LambdaWeight[6];

  // Fast Mode Decision
  int EarlySkipEnable;
  int SelectiveIntraEnable;
  int DisposableP;
  int DispPQPOffset;

  //Redundant picture
  int NumRedundantHierarchy;   //!< number of entries to allocate redundant pictures in a GOP
  int PrimaryGOPLength;        //!< GOP length of primary pictures  
  int NumRefPrimary;           //!< number of reference frames for primary picture

  int HierB_AdaptiveCadence;

#ifdef ADAPTIVE_LOOP_FILTER
  int UseAdaptiveLoopFilter;
#endif

} InputParameters;

//! ImageParameters
typedef struct
{
  int number;                  //!< current image number to be encoded
  int pn;                      //!< picture number
  int LevelIndex;              //!< mapped level idc
  int current_mb_nr;
  int total_number_mb;
  int current_slice_nr;
  int type;
  int structure;               //!< picture structure
  int num_ref_frames;          //!< number of reference frames to be used
  int max_num_references;      //!< maximum number of reference pictures that may occur

  int qp;                      //!< quant for the current frame
  float framerate;
  int width;                   //!< Number of pels
  int width_cr;                //!< Number of pels chroma
  int height;                  //!< Number of lines
  int height_cr;               //!< Number of lines  chroma
  int height_cr_frame;         //!< Number of lines  chroma frame
  int subblock_x;              //!< current subblock horizontal
  int subblock_y;              //!< current subblock vertical
  int is_intra_block;
  int is_v_block;
  int mb_y_upd;
  int mb_y_intra;              //!< which GOB to intra code
  int block_c_x;               //!< current block chroma vertical
  char **ipredmode;             //!< intra prediction mode
  char **ipredmode8x8;          //!< help storage for 8x8 modes, inserted by YV
  int mv_res;                  //!< motion vector resolution: 0 = 1/4-pel, 1 = 1/8-pel
  int cod_counter;             //!< Current count of number of skipped macroblocks in a row
  int ***nz_coeff;             //!< number of coefficients per block (CAVLC)
  
  int mb_x;                    //!< current MB horizontal
  int mb_y;                    //!< current MB vertical
  int block_x;                 //!< current block horizontal
  int block_y;                 //!< current block vertical
  int pix_x;                   //!< current pixel horizontal
  int pix_y;                   //!< current pixel vertical
  int pix_c_x;                 //!< current pixel chroma horizontal
  int pix_c_y;                 //!< current pixel chroma vertical
  
  int opix_x;                   //!< current original picture pixel horizontal
  int opix_y;                   //!< current original picture pixel vertical
  int opix_c_x;                 //!< current original picture pixel chroma horizontal
  int opix_c_y;                 //!< current original picture pixel chroma vertical
  
  
  // some temporal buffers
  imgpel mprr[9][16][16];      //!< all 9 prediction modes? // enlarged from 4 to 16 for ABT (is that neccessary?)
  
  imgpel mprr_2[5][16][16];    //!< all 4 new intra prediction modes
  imgpel mprr_3[9][8][8];      //!< all 9 prediction modes for 8x8 transformation
  imgpel mprr_c[2][4][16][16]; //!< chroma intra prediction modes
  imgpel mpr[16<<MAX_MB_EXT_LEVEL][16<<MAX_MB_EXT_LEVEL];          //!< current best prediction mode
  int m7[16][16];              //!< the diff pixel values between orginal image and prediction
  
  int ****cofAC;               //!< AC coefficients [8x8block][4x4block][level/run][scan_pos]
  int ***cofDC;                //!< DC coefficients [yuv][level/run][scan_pos]

  Picture     *currentPicture; //!< The coded picture currently in the works (typically frame_pic, top_pic, or bottom_pic)
  Slice       *currentSlice;                                //!< pointer to current Slice data struct
  Macroblock    *mb_data;                                   //!< array containing all MBs of a whole frame
  SyntaxElement   MB_SyntaxElements[MAX_SYMBOLS_PER_MB];    //!< temporal storage for all chosen syntax elements of one MB
  
  int *quad;               //!< Array containing square values,used for snr computation  */                                         /* Values are limited to 5000 for pixel differences over 70 (sqr(5000)).
  int *intra_block;
  
  int tr;
  unsigned int rd_pass;
  int direct_intraP_ref[4][4];
  int pstruct_next_P;
  int imgtr_next_P_frm;
  int imgtr_last_P_frm;
  int imgtr_next_P_fld;
  int imgtr_last_P_fld;
  
  // B pictures
  double b_interval;
  int p_interval;
  int b_frame_to_code;
  int fw_mb_mode;
  int bw_mb_mode;
  
  short****** pred_mv;                 //!< motion vector predictors for all block types and all reference frames
  short****** all_mv;                  //!< replaces local all_mv
  short****** all_mv32;                  //!< replaces local all_mv
#ifdef RDO_Q
#ifdef GEOM
  short  ****all32_mv8_geo[5][GEO_PART_NUM][2];
  short  ****all32_mv4_geo[5][GEO_PART_NUM][2];
  short  ****pred32_mv8_geo[5][GEO_PART_NUM][2];
  short  ****pred32_mv4_geo[5][GEO_PART_NUM][2];

  short  ****all16_mv8_geo[1<<(2*MAX_MB_EXT_LEVEL)][GEO_PART_NUM][2];                  //!< replaces local all_mv
  short  ****all16_mv4_geo[1<<(2*MAX_MB_EXT_LEVEL)][GEO_PART_NUM][2];                  //!< replaces local all_mv
  short  ****pred16_mv8_geo[1<<(2*MAX_MB_EXT_LEVEL)][GEO_PART_NUM][2];                  //!< replaces local all_mv
  short  ****pred16_mv4_geo[1<<(2*MAX_MB_EXT_LEVEL)][GEO_PART_NUM][2];                  //!< replaces local all_mv

  short  ****all_mv8_geo[GEO_PART_NUM][2];                  //!< replaces local all_mv
  short  ****all_mv4_geo[GEO_PART_NUM][2];                  //!< replaces local all_mv
  short  ****pred_mv8_geo[GEO_PART_NUM][2];                  //!< replaces local all_mv
  short  ****pred_mv4_geo[GEO_PART_NUM][2];                  //!< replaces local all_mv

#endif

  short  ******all_mv8[1<<(2*MAX_MB_EXT_LEVEL)];                  //!< replaces local all_mv
  short  ******all_mv4[1<<(2*MAX_MB_EXT_LEVEL)];                  //!< replaces local all_mv
  short  ******bipred_mv1_8[1<<(2*MAX_MB_EXT_LEVEL)];
  short  ******bipred_mv2_8[1<<(2*MAX_MB_EXT_LEVEL)];
  short  ******bipred_mv1_4[1<<(2*MAX_MB_EXT_LEVEL)];
  short  ******bipred_mv2_4[1<<(2*MAX_MB_EXT_LEVEL)];

  short  ******all32_mv8[5];
  short  ******all32_mv4[5];
  short  ******bipred32_mv1_8[5];
  short  ******bipred32_mv2_8[5];
  short  ******bipred32_mv1_4[5];
  short  ******bipred32_mv2_4[5];
#endif

  short****** bipred_mv1;              //!< Biprediction MVs
  short****** bipred_mv2;              //!< Biprediction MVs
  short bi_pred_me[MAXMODE];
  
  int LFDisableIdc;
  int LFAlphaC0Offset;
  int LFBetaOffset;
  
  int direct_spatial_mv_pred_flag;              //!< Direct Mode type to be used (0: Temporal, 1: Spatial)
  
  int num_ref_idx_l0_active;
  int num_ref_idx_l1_active;
  
  int mvscale[6][MAX_REFERENCE_PICTURES];
  
#ifdef MV_COMPETITION
  int mvscale_hb[2][2][MAX_REFERENCE_PICTURES][MAX_REFERENCE_PICTURES]; 
#endif

  int buf_cycle;
  int i16offset;
  
  int layer;             //!< which layer this picture belonged to
  int old_layer;         //!< old layer number
  int NoResidueDirect;
  
  //the following should probably go in sequence parameters
  // unsigned int log2_max_frame_num_minus4;
  unsigned int pic_order_cnt_type;
  // for poc mode 0, POC200301
  // unsigned int log2_max_pic_order_cnt_lsb_minus4;  
  // for poc mode 1, POC200301
  unsigned int delta_pic_order_always_zero_flag;
  int offset_for_non_ref_pic;
  unsigned int num_ref_frames_in_pic_order_cnt_cycle;
  int offset_for_ref_frame[1];  // MAX_LENGTH_POC_CYCLE in decoder
  
  // POC200301
  //the following is for slice header syntax elements of poc
  // for poc mode 0.
  unsigned int pic_order_cnt_lsb;
  int delta_pic_order_cnt_bottom;
  // for poc mode 1.
  int delta_pic_order_cnt[2];
  
  
  // POC200301
  signed int framepoc;    
  signed int ThisPOC;     //!< current picture POC
  unsigned int frame_num;   //!< frame_num for this frame
  
  unsigned int PicWidthInMbs;
  unsigned int PicHeightInMapUnits;
  unsigned int FrameHeightInMbs;
  unsigned int PicHeightInMbs;
  unsigned int PicSizeInMbs;
  unsigned int FrameSizeInMbs;
  
  //the following should probably go in picture parameters
  unsigned int pic_order_present_flag; // ????????
  
  //the following are sent in the slice header
  //  int delta_pic_order_cnt[2];
  int nal_reference_idc;
#ifdef  USE_HP_FILTER//BROUND
  int bipred_rounding_control;
#endif  
  int adaptive_ref_pic_buffering_flag;
  int no_output_of_prior_pics_flag;
  int long_term_reference_flag;
  
  DecRefPicMarking_t *dec_ref_pic_marking_buffer;
  
  int model_number;
  int DeblockCall;
  
  int last_has_mmco_5;
  int pre_frame_num;
  
  int slice_group_change_cycle;
  
  int pic_unit_size_on_disk;
  int bitdepth_luma;
  int bitdepth_chroma;
  int bitdepth_luma_qp_scale;
  int bitdepth_chroma_qp_scale;
  int bitdepth_lambda_scale;
  // Lagrangian Parameters
  double **lambda_md;     //!< Mode decision Lambda
  double **lambda_me;     //!< Motion Estimation Lambda
  int    **lambda_mf;     //!< Integer formatted Motion Estimation Lambda
  
  unsigned int dc_pred_value_luma;   //!< luma value for DC prediction (depends on luma pel bit depth)
  unsigned int dc_pred_value_chroma; //!< chroma value for DC prediction (depends on chroma pel bit depth)
  int max_imgpel_value;              //!< max value that one picture element (pixel) can take (depends on pic_unit_bitdepth)
  int max_imgpel_value_uv;
  
  int num_blk8x8_uv;
  int num_cdc_coeff;
  int yuv_format;
  int lossless_qpprime_flag;
  int mb_cr_size_x;
  int mb_cr_size_y;
  
  int chroma_qp_offset[2];      //!< offset for qp for chroma [0-Cb, 1-Cr] 
  
  int auto_crop_right;
  int auto_crop_bottom;
  int last_valid_reference;
  int bytes_in_picture;
#ifdef RDO_Q
  int masterQP;
#endif

#ifdef  INTERNAL_BIT_DEPTH_INCREASE
  int BitDepthIncrease;
  int BitDepthIncreaseChroma;
#endif
#ifdef USE_INTRA_MDDT
  byte scanOrder4x4[9][16][2];
  byte scanOrder8x8[9][64][2];
  int  scanStats4x4[9][16];
  int  scanStats8x8[9][64];
  int  update4x4[9];
  int  update8x8[9];
  int  update4x4Count[9];
  int  update8x8Count[9];
  int  update4x4Thres[9];
  int  update8x8Thres[9];
  
  byte scanOrder16x16[4][256][2];
  int  cofAC16[2][16*17];
#endif 
  int    dct_coeff16I[16][16];
  int    dct_coeff8I [16][16];
  int    **cofAC16x16;//[2][257];
  int    ***cofAC16x8;//[2][2][129];
  int    ***cofAC8x16;//[2][2][129];

#ifdef SWITCHED_FILTERS
  int filterSequence[16];                         // Filter selection computed on sequence
  int subpelOffset[3][2][16];                        // Sub-pel offsets for Frame0
  int imgOffset[3][2][MAX_REFERENCE_PICTURES];       // Image offset for frames
  int filterParam;
  int bestFilter;

  int noOffsets[2];
  int currMEOffset[2];
  int noMEOffsets[2];
  int min_mcost_offset[2];
  int min_mcost_offset_bi[2];
  int offsetMETab[2][16];
  int search_point_qp_x_offset[15];
  int search_point_qp_y_offset[15];
#endif  // SWITCHED_FILTERS

  int curr_rd_pass;
  int KeyP;
  int Q_Level;
#ifdef GEOM
  int   geo_scheme;
#endif


#ifdef ADAPTIVE_LOOP_FILTER
  int  min_alf_block_size_idx;
  int  min_alf_block_size;
  int  max_WidthInAlfBlks;
  int  max_HeightInAlfBlks;
  int  max_num_alf_block;
  int  current_alfb_nr;
  int  alf_block_size;
  int  WidthInAlfBlks;
  int  HeightInAlfBlks;
  int  num_alf_block;
  AlfBlock *alf_blk_data;
#endif

} ImageParameters;

  short isfirstQP;
  int finalbestQP;
  int Prev_QP,Prev_Delta_QP;

#ifdef RDO_Q
int Motion_Selected;
int Intra_Selected; 
int final_mb_encoding;
int saved_best_mode ;

#define IS_INTRA_MODE(mode)  ((mode==I4MB)||(mode==I8MB)||(mode==I16MB))
#endif


#define NUM_PIC_TYPE 5
//!< statistics
typedef struct
{
  int   quant0;                 //!< quant for the first frame
  int   quant1;                 //!< average quant for the remaining frames
  float bitr;                   //!< bit rate for current frame, used only for output til terminal
  float bitrate;                //!< average bit rate for the sequence except first frame
  int64   bit_ctr;                //!< counter for bit usage
  int64   bit_ctr_n;              //!< bit usage for the current frame
  int   bit_slice;              //!< number of bits in current slice
  int   bit_ctr_emulationprevention; //!< stored bits needed to prevent start code emulation
  int   b8_mode_0_use[NUM_PIC_TYPE][2];
  int   mode_use_transform_8x8[NUM_PIC_TYPE][MAXMODE];
  int   mode_use_transform_4x4[NUM_PIC_TYPE][MAXMODE];
  int   intra_chroma_mode[4];
  
  // B pictures
  int   successive_Bframe;
  int   *mode_use_Bframe;
  int   *bit_use_mode_Bframe;
  int64   bit_ctr_I;
  int64   bit_ctr_P;
  int64   bit_ctr_B;
  float bitrate_I;
  float bitrate_P;
  float bitrate_B;
    
  int64   mode_use            [NUM_PIC_TYPE][MAXMODE]; //!< Macroblock mode usage for Intra frames
  int64   bit_use_mode        [NUM_PIC_TYPE][MAXMODE]; //!< statistics of bit usage
  int64   bit_use_stuffingBits[NUM_PIC_TYPE];
  int64   bit_use_mb_type     [NUM_PIC_TYPE];
  int64   bit_use_header      [NUM_PIC_TYPE];
  int64   tmp_bit_use_cbp     [NUM_PIC_TYPE];
  int64   bit_use_coeffY      [NUM_PIC_TYPE];
  int64   bit_use_coeffC      [NUM_PIC_TYPE];
  int64   bit_use_delta_quant [NUM_PIC_TYPE];
  
  int   em_prev_bits_frm;
  int  *em_prev_bits;
  int   bit_ctr_parametersets;
  int   bit_ctr_parametersets_n;

#ifdef MV_COMPETITION
  int64   predictor_for_skip_use    [NUM_PIC_TYPE][MAX_MV_PREDICTOR];
  int64   predictor_for_skip_trans  [NUM_PIC_TYPE];
  int64   bit_use_pred_for_skip     [NUM_PIC_TYPE];
  int64   bit_use_pred_for_mv_inter [NUM_PIC_TYPE];
#endif
} StatParameters;

//!< For MB level field/frame coding tools
//!< temporary structure to store MB data for field/frame coding
typedef struct
{
  double min_rdcost;
  
  imgpel rec_mbY[16][16];       // hold the Y component of reconstructed MB
  imgpel rec_mbU[16][16], rec_mbV[16][16]; 
  int    ****cofAC;
  int    ***cofDC;
  int    mb_type;
  short  bi_pred_me;
  
  int    b8mode[4], b8pdir[4];
  char   **ipredmode;
  char   intra_pred_modes[16];
  char   intra_pred_modes8x8[16];
  int    cbp;
  int64  cbp_blk;
  int    mode;
  short  ******pred_mv;        //!< predicted motion vectors
  short  ******all_mv;         //!< all modes motion vectors
  char   refar[2][4][4];       //!< reference frame array [list][y][x]
  int    i16offset;
  int    c_ipred_mode;
  int    luma_transform_size_8x8_flag;
  int    NoMbPartLessThan8x8Flag;
  
  int    qp;
  int    prev_qp;
  int    prev_delta_qp;
  int    delta_qp;
  int    prev_cbp;
} RD_DATA;


//!< Set Explicit GOP Parameters.
//!< Currently only supports Enhancement GOP but could be easily extended
typedef struct
{
  int slice_type;       //! Slice type
  int display_no;       //! GOP Display order
  int reference_idc;    //! Is reference?
  int slice_qp;         //! Assigned QP
  int hierarchy_layer;    //! Hierarchy layer (used with GOP Hierarchy option 2)
  int hierarchyPocDelta;  //! Currently unused
} GOP_DATA;


typedef struct
{
  int cost8x8;
  short part8x8mode[4];
  short part8x8pdir[4];
  char  part8x8fwref[4];
  char  part8x8bwref[4];
  imgpel rec_mbY8x8[16][16];    
  imgpel mpr8x8[16][16];
  double rdcost[4];
  int   ****cofAC8x8;        // [8x8block][4x4block][level/run][scan_pos]
  int   cbp_blk8x8;
  int   cbp8x8, cnt_nonz_8x8;
} RD_8x8DATA;         //array of size 5 is declared for this structure to store data for 5 QP in RDOQ 

typedef struct
{  
  double lambda_md;     //!< Mode decision Lambda
  double lambda_me;     //!< Motion Estimation Lambda
  int    lambda_mf;     //!< Integer formatted Motion Estimation Lambda
  
  short  valid[MAXMODE];
  short  list_offset[2];
  short  best_ref[2];
  int    best_mcost[2];
} RD_PARAMS;

typedef struct
{
  int64  dist;
  int    rate; 
  double rdcost;
} MYSTAT;

GOP_DATA *gop_structure;
RD_DATA *rdopt; 
RD_DATA rddata_top_frame_mb; //!< For MB level field/frame coding tools

extern InputParameters *input;
extern ImageParameters *img;
extern StatParameters  *stats;
extern SNRParameters *snr;

// files
FILE *p_stat;                    //!< status file for the last encoding session
FILE *p_log;                     //!< SNR file
FILE *p_trace;                   //!< Trace file
int  p_in;                       //!< original YUV file handle
int  p_dec;                      //!< decoded image file handle



                                 /***********************************************************************
                                 * P r o t o t y p e s   f o r    T M L
                                 ************************************************************************/

void init(void);

#ifdef USE_INTRA_MDDT
int  klt_luma_sep(int block_x,int block_y,int *coeff_cost, int ipmode);
int  klt_luma8x8_sep_fast(int b8, int *coeff_cost, int ipmode);
int  klt_luma16x16_sep_fast (int ipmode);
int  dct_luma(int pos_mb1,int pos_mb2,int *cnt_nonz, int intra,  int ipmode);
int  dct_luma8x8(int b8,int *coeff_cost, int intra, int ipmode);
void Intra16x16_Mode_Decision_RDopt (Macroblock* currMB, int* i16mode, double lambda);
#else
int  dct_luma(int pos_mb1,int pos_mb2,int *cnt_nonz, int intra);
int  dct_luma8x8(int b8,int *coeff_cost, int intra);
#endif
int  dct_luma_16x16(int);
int  dct_chroma(int uv,int i11);

int  sign(int a,int b);


void intrapred_luma(int CurrPixX,int CurrPixY, int *left_available, int *up_available, int *all_available);
void intrapred_luma_16x16(void);
void intrapred_chroma(int,int,int uv);
int  find_sad_16x16(int *intra_mode);

int Mode_Decision_for_8x8IntraBlocks(int b8,double lambda,int *cost);
int Mode_Decision_for_Intra4x4Macroblock (double lambda,  int* cost);
int Mode_Decision_for_4x4IntraBlocks (int  b8,  int  b4,  double  lambda,  int*  min_cost);
double RDCost_for_4x4IntraBlocks (int* nonzero,int b8,int b4,int ipmode,double  lambda,double  min_rdcost, int mostProbableMode);
void Intra16x16_Mode_Decision (Macroblock* currMB, int* i16mode);

int    Mode_Decision_for_new_Intra8x8Macroblock (double lambda, int *min_cost);
int    Mode_Decision_for_new_8x8IntraBlocks (int b8, double lambda, int *min_cost);
void   intrapred_luma_8x8(int img_x,int img_y, int *left_available, int *up_available, int *all_available);
double RDCost_for_8x8IntraBlocks(int *c_nz, int b8, int ipmode, double lambda, double min_rdcost, int mostProbableMode);
void   LowPassForIntra8x8Pred(int *PredPel, int block_up_left, int block_up, int block_left);

void init_poc(void);

void init_img(void);
void report(void);
void information_init(void);
int  get_picture_type(void);
int clip1a(int a);
void DeblockFrame(ImageParameters *img, imgpel **, imgpel ***) ;

int  SATD8X8(int*, int);

void LumaPrediction4x4 (int, int, int, int, int, short, short);
int  SATD (int*, int);
int  find_SATD (int c_diff[MB_PIXELS], int blocktype, int mb_ext_level);

void LumaResidualCoding (int mbi_x, int mbi_y, int mode, int mode_idx);
void ChromaResidualCoding (int mbi_x, int mbi_y, int mode, int mode_idx, int*);
void IntraChromaPrediction (int*, int*, int*);
void ChromaPrediction4x4 (int, int, int, int, int, int, short, short);

int writeMBLayer (int rdopt, int *coeff_rate);

extern int*   refbits;
extern int**** motion_cost;

void  Get_Direct_Motion_Vectors (int mb_ext_level);
void  PartitionMotionSearch     (int, int, int, int, int mb_ext_level);
int   BIDPartitionCost          (int, int, short, short, int,int mb_ext_level);
int   LumaResidualCoding8x8     (int, int, int, int, int*, int64*, int, short, int, int, short, short);
int   writeLumaCoeff8x8         (int, int, int);
#ifdef USE_INTRA_MDDT
int   writeLumaCoeff16x16       (void);
#endif
int   writeMotionVector8x8      (int  i0, int  j0, int  i1, int  j1, int  refframe, int  list_idx, int  mv_mode, int mb_ext_level);
int   writeReferenceFrame       (int, int, int, int, int, int mb_ext_level);
int   writeAbpCoeffIndex        (int, int, int, int);
int   writeChromaIntraPredMode  (void);

void poc_based_ref_management(int current_pic_num);
int  picture_coding_decision (Picture *picture1, Picture *picture2, int qp);

unsigned CeilLog2( unsigned uiVal);

int   BPredPartitionCost  (int, int, short, short, int, int, int mb_ext_level);
void  LumaPrediction4x4Bi (int, int,   int,   int, int, short, short, int);
int   SATDBI (int* , int );
int  B8Mode2Value (int b8mode, int b8pdir);
void FindSkipModeMotionVector (int mb_ext_level);

// dynamic mem allocation
int  init_global_buffers(void);
void free_global_buffers(void);
void no_mem_exit  (char *where);

int  get_mem_mv  (short*******);
void free_mem_mv (short******);
void free_img    (void);

int  get_mem_ACcoeff  (int*****);
int  get_mem_DCcoeff  (int****);
void free_mem_ACcoeff (int****);
void free_mem_DCcoeff (int***);

MBdata *init_ExtMBdata(int level);
void free_ExtMBdata(MBdata* pMBdata);
void CopyMBdata(MBdata *Dest,MBdata *Src);

void combine_field(void);

Picture *malloc_picture(void);
void     free_picture (Picture *pic);

int   encode_one_slice(int SLiceGroupId, Picture *pic, int TotalCodedMBs);   //! returns the number of MBs in the slice

void  start_macroblock(int mb_addr);
void  set_MB_parameters (int mb_addr);           //! sets up img-> according to input-> and currSlice->

int   writeMotionInfo2NAL (int mb_ext_level);

void  terminate_macroblock(Boolean *end_of_slice, Boolean *recode_macroblock);
int   slice_too_big(int rlc_bits);
void  write_one_macroblock(int eos_bit);
void  proceed2nextMacroblock(void);

void free_slice_list(Picture *currPic);

void report_stats_on_error(void);

#ifdef RDO_Q
typedef struct 
{
  int significantBits[16][2];
  int lastBits[16][2];
  int greaterOneBits[2][5][2]; // c1 and c2
  int greaterOneState[5];
  int blockCbpBits[4][2]; // c1 and c2
} estBitsCabacStruct;

typedef struct 
{
  int64 level[3];
  int pre_level;//TREL_CAVLC
  int coeff_ctr;//TREL_CAVLC
  int64 levelDouble;
  double errLevel[3];
  int noLevels;
} levelDataStruct;


typedef struct {

  // important variables of data partition array
  int                   no_part;
  EncodingEnvironment  *encenv;
  Bitstream            *bitstream;

  // contexts for binary arithmetic coding
  MotionInfoContexts   *mot_ctx;
  TextureInfoContexts  *tex_ctx;

  // syntax element number and bitcounters
  int                   currSEnr;
  int                   bitcounter[MAX_BITCOUNTER_MB];

  // elements of current macroblock
  int                   mvd[2][BLOCK_MULTIPLE][BLOCK_MULTIPLE][2];
  int64                 cbp_bits;
} CSobj;
typedef CSobj* CSptr;


void  delete_coding_state  (CSptr);  //!< delete structure
CSptr create_coding_state  (void);       //!< create structure

void  store_coding_state   (CSptr);  //!< store parameters
void  reset_coding_state   (CSptr);  //!< restore parameters

void estRunLevel_CABAC (int context);
void est_writeRunLevel_CABAC1(levelDataStruct levelData[], int levelTabMin[], int type, double lambda, int kStart, 
                              int kStop, int noCoeff, int estCBP);
void est_writeRunLevel_CABAC2(levelDataStruct levelData[], int levelTabMin[], int type, double lambda, int kStart, 
                              int kStop, int noCoeff, int estCBP);
void est_writeRunLevel_CABAC(levelDataStruct levelData[], int levelTabMin[], int type, double lambda, int kStart, 
                             int kStop, int noCoeff, int estCBP);
int est_write_and_store_CBP_block_bit(Macroblock* currMB, int type);
#endif

#if TRACE
void  trace2out(SyntaxElement *se);
#endif


void error(char *text, int code);
int  start_sequence(void);
int  terminate_sequence(void);
int  start_slice(void);
int  terminate_slice(int, Picture *pic);
int  write_PPS(int, int);


void InitMotionVectorSearchModule(void);
void writeVlcByteAlign(Bitstream* currStream);

int   writeLumaCoeff4x4_CABAC     (int, int, int);
int   writeLumaCoeff8x8_CABAC     (int, int);
#ifdef USE_INTRA_MDDT
int   writeLumaCoeff16x16_CABAC   (void);
#endif


int writeCBP_Transform_size (int rdopt, int mb32);
int writeDeltaQP            (int rdopt, int cbp);
int writeLumaCoeff          (int rdopt, int mb32);

//============= rate-distortion optimization ===================
void  clear_rdopt      (void);
void  init_rdopt       (void);
//============= restriction of reference frames based on the latest intra-refreshes==========
void process_2nd_IGOP(void);
void SetImgType(void);

extern Boolean In2ndIGOP;
extern int start_frame_no_in_this_IGOP;
extern int start_tr_in_this_IGOP;
extern int FirstFrameIn2ndIGOP;
#define IMG_NUMBER (img->number-start_frame_no_in_this_IGOP)

void AllocNalPayloadBuffer(void);
void FreeNalPayloadBuffer(void);
void SODBtoRBSP(Bitstream *currStream);
int RBSPtoEBSP(byte *streamBuffer, int begin_bytepos, int end_bytepos, int min_num_bytes);
int Bytes_After_Header;

// Fast ME enable
int BlockMotionSearch (short,int,int,int,int,int, int, int mb_ext_level);
double encode_one_macroblock (int Re_Encode);
void set_chroma_qp(Macroblock *currMB);

//context_ini.h
void  create_context_memory (void);
void  free_context_memory   (void);
void  init_contexts  (void);
void  store_contexts (void);
void  update_rd_picture_contexts  (int);
void  SetCtxModelNumber (void);
void reset_coding_state_cs_cm(void);
int giRDOpt_B8OnlyFlag;
#ifdef BEST_NZ_COEFF
int gaaiMBAFF_NZCoeff[4][12];
#endif

int do_Re_Encode;

MBdata *ExtMBdata,*tmpExtMBdata;
short index16x16; //0~3 inside 32x32
short index32x32; //0~3 inside 64x64

short final_mb_encoding_bestmode;
short reencoding_during_modedecision;
int   final_mb_encoding_geo_mode; //GEOM_RESTURCTURE
int        FrameNumberInFile;
short      last_coded_frame_was_reference;    //this will take care of proper frame numbering when 2 GOPs are side-by-side
imgpel *Pel_12to8;
#endif

#ifdef USE_INTRA_MDDT
void updateScanOrder(int first);
void normalizeScanStats();
#endif 


#ifdef ADAPTIVE_MV_RES
short roundInt(short mv,int mvres);
#endif

// GOP Hierarchy
void init_gop_structure(void); 
void interpret_gop_structure(void);
void create_hierarchy(void);
void clear_gop_structure(void);
void encode_enhancement_layer(void);
void create_adaptive_hierarchy(int number);

