/* 
  ====================================================================================================================

	The copyright in this software is being made available under the License included below. 
	This software may be subject to other third party and 	contributor rights, including patent rights, and no such 
	rights are granted under this license.
 
	Copyright (c) 2010, QUALCOMM INC.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification, are permitted only for 
	the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
	promoting such standards. The following conditions are required to be met:

		* Redistributions of source code must retain the above copyright notice, this list of conditions and 
		  the following disclaimer.
		* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
		  the following disclaimer in the documentation and/or other materials provided with the distribution.
		* The name of QUALCOMM INC.  may not be used to endorse or promote products derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
	INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  ====================================================================================================================
*/


#ifndef _PARSET_H_
#define _PARSET_H_

#include "nalucommon.h"
#include "global.h"

#define MAXIMUMPARSETRBSPSIZE   1500
#define MAXIMUMPARSETNALUSIZE   1500

#define MAXSPS  32
#define MAXPPS  256

//! Boolean Type
#ifdef FALSE
#  define Boolean int
#else
typedef enum {
  FALSE,
  TRUE
} Boolean;
#endif

#define MAXIMUMVALUEOFcpb_cnt   32
typedef struct
{
  unsigned int  cpb_cnt;                                          // ue(v)
  unsigned int  bit_rate_scale;                                   // u(4)
  unsigned int  cpb_size_scale;                                   // u(4)
  unsigned int  bit_rate_value [MAXIMUMVALUEOFcpb_cnt];         // ue(v)
  unsigned int  cpb_size_value[MAXIMUMVALUEOFcpb_cnt];          // ue(v)
  unsigned int  vbr_cbr_flag[MAXIMUMVALUEOFcpb_cnt];            // u(1)
  unsigned int  initial_cpb_removal_delay_length_minus1;          // u(5)
  unsigned int  cpb_removal_delay_length_minus1;                  // u(5)
  unsigned int  dpb_output_delay_length_minus1;                   // u(5)
  unsigned int  time_offset_length;                               // u(5)
} hrd_parameters_t;


#define MAXnum_slice_groups_minus1  8
typedef struct
{
  Boolean   Valid;                  // indicates the parameter set is valid
  unsigned int  pic_parameter_set_id;                             // ue(v)
  unsigned int  seq_parameter_set_id;                             // ue(v)
  Boolean   entropy_coding_mode_flag;                         // u(1)

  Boolean   transform_8x8_mode_flag;                          // u(1)

  // if( pic_order_cnt_type < 2 )  in the sequence parameter set
  Boolean      pic_order_present_flag;                           // u(1)
  unsigned int  num_slice_groups_minus1;                          // ue(v)
  unsigned int  slice_group_map_type;                        // ue(v)
  // if( slice_group_map_type = = 0 )
  unsigned int  run_length_minus1[MAXnum_slice_groups_minus1]; // ue(v)
  // else if( slice_group_map_type = = 2 )
  unsigned int  top_left[MAXnum_slice_groups_minus1];         // ue(v)
  unsigned int  bottom_right[MAXnum_slice_groups_minus1];     // ue(v)
  // else if( slice_group_map_type = = 3 || 4 || 5
  Boolean   slice_group_change_direction_flag;            // u(1)
  unsigned int  slice_group_change_rate_minus1;               // ue(v)
  // else if( slice_group_map_type = = 6 )
  unsigned int  pic_size_in_map_units_minus1;                  // ue(v)
  byte      *slice_group_id;                              // complete MBAmap u(v)

  int       num_ref_idx_l0_active_minus1;                     // ue(v)
  int       num_ref_idx_l1_active_minus1;                     // ue(v)
  int       pic_init_qp_minus26;                              // se(v)
  int       pic_init_qs_minus26;                              // se(v)
  int       chroma_qp_index_offset;                           // se(v)

  int       cb_qp_index_offset;                               // se(v)
  int       cr_qp_index_offset;                               // se(v)

  Boolean   deblocking_filter_control_present_flag;           // u(1)
  Boolean   constrained_intra_pred_flag;                      // u(1)
  Boolean   redundant_pic_cnt_present_flag;                   // u(1)
  Boolean   vui_pic_parameters_flag;                          // u(1)
} pic_parameter_set_rbsp_t;


#define MAXnum_ref_frames_in_pic_order_cnt_cycle  256
typedef struct
{
  Boolean   Valid;                  // indicates the parameter set is valid

  unsigned int  profile_idc;                                      // u(8)
  Boolean   constrained_set0_flag;                            // u(1)
  Boolean   constrained_set1_flag;                            // u(1)
  Boolean   constrained_set2_flag;                            // u(1)
  Boolean   constrained_set3_flag;                            // u(1)
  unsigned int  level_idc;                                        // u(8)
  unsigned int  seq_parameter_set_id;                             // ue(v)
  unsigned int  chroma_format_idc;                                // ue(v)

  unsigned int  bit_depth_luma_minus8;                            // ue(v)
  unsigned int  bit_depth_chroma_minus8;                          // ue(v)
  unsigned int  log2_max_frame_num_minus4;                        // ue(v)
  unsigned int pic_order_cnt_type;
  // if( pic_order_cnt_type == 0 ) 
  unsigned int log2_max_pic_order_cnt_lsb_minus4;                 // ue(v)
  // else if( pic_order_cnt_type == 1 )
  Boolean delta_pic_order_always_zero_flag;               // u(1)
  int     offset_for_non_ref_pic;                         // se(v)
  unsigned int  num_ref_frames_in_pic_order_cnt_cycle;          // ue(v)
  // for( i = 0; i < num_ref_frames_in_pic_order_cnt_cycle; i++ )
  int   offset_for_ref_frame[MAXnum_ref_frames_in_pic_order_cnt_cycle];   // se(v)
  unsigned int  num_ref_frames;                                   // ue(v)
  Boolean   gaps_in_frame_num_value_allowed_flag;             // u(1)
  unsigned int  pic_width_in_mbs_minus1;                          // ue(v)
  unsigned int  pic_height_in_map_units_minus1;                   // ue(v)
  Boolean   direct_8x8_inference_flag;                        // u(1)
  Boolean   frame_cropping_flag;                              // u(1)
  unsigned int  frame_cropping_rect_left_offset;                // ue(v)
  unsigned int  frame_cropping_rect_right_offset;               // ue(v)
  unsigned int  frame_cropping_rect_top_offset;                 // ue(v)
  unsigned int  frame_cropping_rect_bottom_offset;              // ue(v)
} seq_parameter_set_rbsp_t;

pic_parameter_set_rbsp_t *AllocPPS (void);
seq_parameter_set_rbsp_t *AllocSPS (void);
void FreePPS (pic_parameter_set_rbsp_t *pps);
void FreeSPS (seq_parameter_set_rbsp_t *sps);

void GenerateParameterSets (void);
void FreeParameterSets (void);

NALU_t *GenerateSeq_parameter_set_NALU (void);
NALU_t *GeneratePic_parameter_set_NALU (int);

// The following are local helpers, but may come handy in the future, hence public
void GenerateSequenceParameterSet(seq_parameter_set_rbsp_t *sps, int SPS_id);
void GeneratePictureParameterSet( pic_parameter_set_rbsp_t *pps, seq_parameter_set_rbsp_t *sps, int PPS_id, 
                                 int cb_qp_index_offset, int cr_qp_index_offset);



int GenerateSeq_parameter_set_rbsp (seq_parameter_set_rbsp_t *sps, unsigned char *buf);
int GeneratePic_parameter_set_rbsp (pic_parameter_set_rbsp_t *pps, unsigned char *buf);


#endif
