/* 
====================================================================================================================

The copyright in this software is being made available under the License included below. 
This software may be subject to other third party and 	contributor rights, including patent rights, and no such 
rights are granted under this license.

Copyright (c) 2010, QUALCOMM INC.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted only for 
the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
promoting such standards. The following conditions are required to be met:

* Redistributions of source code must retain the above copyright notice, this list of conditions and 
the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
the following disclaimer in the documentation and/or other materials provided with the distribution.
* The name of QUALCOMM INC.  may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

====================================================================================================================
*/
#include "alf_filter_common.h"
#ifdef ADAPTIVE_LOOP_FILTER
Bitstream * allocBitstream(int size)
{
  Bitstream *tmp;

  tmp =  (Bitstream *) calloc(1, sizeof(Bitstream));
  tmp->streamBuffer = (byte *) calloc(size, sizeof(byte));
  tmp->bits_to_go = 8;
  tmp->byte_buf = 0;
  tmp->byte_pos = 0;

  return tmp;
}
int freeBitstream(Bitstream *bitstream)
{
  int tmp = 0;
  tmp += bitstream->byte_pos * 8;
  tmp += 8 - bitstream->bits_to_go;

  free(bitstream->streamBuffer);
  bitstream->streamBuffer = NULL;
  free(bitstream);

  return tmp;
}

imgpel Clip_post(int high, int val)
{
  return (imgpel)(((val > high)? high: val));
}

void padImage(imgpel **imgY,  imgpel **imgY_pad, int fl, int img_height, int img_width)
{
  int x, y;

  for(y = 0; y < img_height; ++y)
    memcpy(imgY_pad[y+fl]+fl,imgY[y],sizeof(imgpel)* img_width);

  for(x = fl; x < img_width + fl; ++x) 
  {
    for(y = 0; y < fl; ++y)
      imgY_pad[y][x] = imgY_pad[fl][x];

    for(y = img_height + fl; y < img_height + 2 * fl; ++y)
      imgY_pad[y][x] = imgY_pad[img_height + fl - 1][x];
  }

  for(y = 0; y < img_height + 2 * fl; ++y)
  {
    for(x = 0; x < fl; ++x)
      imgY_pad[y][x] = imgY_pad[y][fl];

    for(x = img_width + fl; x < img_width + 2 * fl; ++x)
      imgY_pad[y][x] = imgY_pad[y][img_width + fl - 1];
  }
}


void  calcVar(imgpel **imgY_var, imgpel **imgY_rec, int fl, int img_height, int img_width)
{
  int i, j, ii, jj;
  int sum;
  static imgpel **imgY_pad;
  static int **imgY_temp, first = 0;
  int var_max= NO_VAR_BINS-1;

  int mult_fact_int;
  int shift= (11+img->BitDepthIncrease);
  int mult_fact_int_tab[4]= {1,114,41,21};


  if(first == 0)
  {
    initMatrix_ushort(&imgY_pad, img_height+2*fl+2, img_width+2*fl+2);
    initMatrix_int(&imgY_temp, img_height+2*fl, img_width+2*fl);
    first = 1;
  }

  padImage(imgY_rec, imgY_pad, fl+1, img_height, img_width);

  mult_fact_int = mult_fact_int_tab[fl];

  if (fl==0)
  {
    shift=img->BitDepthIncrease;
    for(i = 1; i < img_height + 1; ++i)
      for(j = 1; j < img_width + 1; ++j)	
      {
        imgY_temp[i-1][j-1] = abs(2 * imgY_pad[i][j] - imgY_pad[i][j+1] - imgY_pad[i][j-1])+
          abs(2 * imgY_pad[i][j] - imgY_pad[i-1][j] - imgY_pad[i+1][j]);
        imgY_var[i-1][j-1]=(imgpel) Clip_post(var_max, (int) ((imgY_temp[i-1][j-1])>>shift));
      }
  }
  else
  {
    shift= (11+img->BitDepthIncrease);

    for(i = 1; i < img_height + 2 * fl + 1; ++i)
      for(j = 1; j < img_width + 2 * fl + 1; ++j)	
        imgY_temp[i-1][j-1] = abs(2 * imgY_pad[i][j] - imgY_pad[i][j+1] - imgY_pad[i][j-1])+
        abs(2 * imgY_pad[i][j] - imgY_pad[i-1][j] - imgY_pad[i+1][j]);
    for(i = fl; i < img_height + fl; ++i)
    {
      for(j = fl; j < img_width + fl; ++j)
      {	
        sum = 0;
        for(ii = i - fl; ii <= i + fl; ii++)
          for(jj = j - fl; jj <= j + fl; jj++)
            sum+=imgY_temp[ii][jj];

        imgY_var[i - fl][j - fl] =(imgpel) Clip_post(var_max, (int) ((sum * mult_fact_int)>>shift));
      }
    }
  }
}




//alf...memory allocations
void initMatrix_int(int ***m2D, int d1, int d2)
{
  int i;

  if(!(*m2D = (int **) calloc(d1, sizeof(int *))))
    FATAL_ERROR_0("initMatrix_int: memory allocation problem\n", -1);
  if(!((*m2D)[0] = (int *) calloc(d1 * d2, sizeof(int))))
    FATAL_ERROR_0("initMatrix_int: memory allocation problem\n", -1);

  for(i = 1; i < d1; i++)
    (*m2D)[i] = (*m2D)[i-1] + d2;
}

void initMatrix_double(double ***m2D, int d1, int d2)
{
  int i;

  if(!(*m2D = (double **) calloc(d1, sizeof(double *))))
    FATAL_ERROR_0("initMatrix_double: memory allocation problem\n", -1);
  if(!((*m2D)[0] = (double *) calloc(d1 * d2, sizeof(double))))
    FATAL_ERROR_0("initMatrix_double: memory allocation problem\n", -1);

  for(i = 1; i < d1; i++)
    (*m2D)[i] = (*m2D)[i-1] + d2;
}

void initMatrix_ushort(unsigned short ***m2D, int d1, int d2)
{
  int i;

  if(!(*m2D = (unsigned short **) calloc(d1, sizeof(unsigned short *))))
    FATAL_ERROR_0("initMatrix_ushort: memory allocation problem\n", -1);
  if(!((*m2D)[0] = (unsigned short *) calloc(d1 * d2, sizeof(unsigned short))))
    FATAL_ERROR_0("initMatrix_ushort: memory allocation problem\n", -1);

  for(i = 1; i < d1; i++)
    (*m2D)[i] = (*m2D)[i-1] + d2;
}

void initMatrix3D_int(int ****m3D, int d1, int d2, int d3)
{
  int  j;

  if(!((*m3D) = (int ***) calloc(d1, sizeof(int **))))
    FATAL_ERROR_0("initMatrix3D_int: memory allocation problem\n", -1);

  for(j = 0; j < d1; j++)
    initMatrix_int((*m3D) + j, d2, d3);
}

void initMatrix3D_double(double ****m3D, int d1, int d2, int d3)
{
  int  j;

  if(!((*m3D) = (double ***) calloc(d1, sizeof(double **))))
    FATAL_ERROR_0("initMatrix3D_double: memory allocation problem\n", -1);

  for(j = 0; j < d1; j++)
    initMatrix_double((*m3D) + j, d2, d3);
}

void initMatrix4D_double(double *****m4D, int d1, int d2, int d3, int d4)
{
  int  j;

  if(!((*m4D) = (double ****) calloc(d1, sizeof(double ***))))
    FATAL_ERROR_0("initMatrix4D_double: memory allocation problem\n", -1);

  for(j = 0; j < d1; j++)
    initMatrix3D_double((*m4D) + j, d2, d3, d4);
}

void destroyMatrix_int(int **m2D)
{
  if(m2D)
  {
    if(m2D[0])
      free(m2D[0]);
    else
      FATAL_ERROR_0("destroyMatrix_int: memory free problem\n", -1);
    free(m2D);
  } 
  else
  {
    FATAL_ERROR_0("destroyMatrix_int: memory free problem\n", -1);
  }
}

void destroyMatrix_double(double **m2D)
{
  if(m2D)
  {
    if(m2D[0])
      free(m2D[0]);
    else
      FATAL_ERROR_0("destroyMatrix_double: memory free problem\n", -1);
    free(m2D);
  } 
  else
  {
    FATAL_ERROR_0("destroyMatrix_double: memory free problem\n", -1);
  }
}

void destroyMatrix3D_int(int ***m3D, int d1)
{
  int i;

  if(m3D)
  {
    for(i = 0; i < d1; i++)
      destroyMatrix_int(m3D[i]);
    free(m3D);
  } 
  else
  {
    FATAL_ERROR_0("destroyMatrix3D_int: memory free problem\n", -1);
  }
}

void destroyMatrix3D_double(double ***m3D, int d1)
{
  int i;

  if(m3D)
  {
    for(i = 0; i < d1; i++)
      destroyMatrix_double(m3D[i]);
    free(m3D);
  } 
  else
  {
    FATAL_ERROR_0("destroyMatrix3D_double: memory free problem\n", -1);
  }
}

void destroyMatrix4D_int(int ****m4D, int d1, int d2)
{
  int  j;

  if(m4D)
  {
    for(j = 0; j < d1; j++)
      destroyMatrix3D_int(m4D[j], d2);
    free(m4D);
  } 
  else
  {
    FATAL_ERROR_0("destroyMatrix4D_int: memory free problem\n", -1);
  }
}

void destroyMatrix4D_double(double ****m4D, int d1, int d2)
{
  int  j;

  if(m4D)
  {
    for(j = 0; j < d1; j++)
      destroyMatrix3D_double(m4D[j], d2);
    free(m4D);
  } 
  else
  {
    FATAL_ERROR_0("destroyMatrix4D_double: memory free problem\n", -1);
  }
}

#endif



