/* 
====================================================================================================================

The copyright in this software is being made available under the License included below. 
This software may be subject to other third party and 	contributor rights, including patent rights, and no such 
rights are granted under this license.

Copyright (c) 2010, QUALCOMM INC.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted only for 
the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
promoting such standards. The following conditions are required to be met:

* Redistributions of source code must retain the above copyright notice, this list of conditions and 
the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
the following disclaimer in the documentation and/or other materials provided with the distribution.
* The name of QUALCOMM INC.  may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

====================================================================================================================
*/

#include <math.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>

#include "global.h"

#include "elements.h"
#include "header.h"
#include "mbuffer.h"
#include "defines.h"
#include "vlc.h"
#include "parset.h"

#ifdef SWITCHED_FILTERS
#include "switched_filters.h"
#endif  // SWITCHED_FILTERS

// A little trick to avoid those horrible #if TRACE all over the source code
#if TRACE
#define SYMTRACESTRING(s) strncpy(sym->tracestring,s,TRACESTRING_SIZE)
#else
#define SYMTRACESTRING(s) // do nothing
#endif

int * assignSE2partition[2] ;
int assignSE2partition_NoDP[SE_MAX_ELEMENTS] =
{  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
int assignSE2partition_DP[SE_MAX_ELEMENTS] =
{  0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 2, 2, 2, 2, 0, 0, 0, 0 } ;

static int ref_pic_list_reordering(Bitstream *bitstream);
static int dec_ref_pic_marking    (Bitstream *bitstream);

/*!
********************************************************************************************
* \brief 
*    Write a slice header
*
* \return
*    number of bits used 
********************************************************************************************
*/
int SliceHeader()
{
  int dP_nr = assignSE2partition[0][SE_HEADER];
  Bitstream *bitstream = img->currentSlice->partArr[dP_nr].bitstream;
  Slice* currSlice = img->currentSlice;
  int len = 0;

  int num_bits_slice_group_change_cycle;
  float numtmp;  

  len  = ue_v("SH: first_mb_in_slice", img->current_mb_nr,   bitstream);

  len += ue_v("SH: slice_type",        get_picture_type (),   bitstream);

  len += ue_v("SH: pic_parameter_set_id" , active_pps->pic_parameter_set_id ,bitstream);

  len += u_v (log2_max_frame_num_minus4 + 4,"SH: frame_num", img->frame_num, bitstream);

  if (img->currentPicture->idr_flag)
  {
    // idr_pic_id
    len += ue_v ("SH: idr_pic_id", (img->number % 2), bitstream);
  }

  if (img->pic_order_cnt_type == 0)
  {
    img->pic_order_cnt_lsb = (img->framepoc & ~((((unsigned int)(-1)) << (log2_max_pic_order_cnt_lsb_minus4+4))) );
    len += u_v (log2_max_pic_order_cnt_lsb_minus4+4, "SH: pic_order_cnt_lsb", img->pic_order_cnt_lsb, bitstream);

    if (img->pic_order_present_flag)
    {
      len += se_v ("SH: delta_pic_order_cnt_bottom", img->delta_pic_order_cnt_bottom, bitstream);
    }
  }
  if (img->pic_order_cnt_type == 1 && !img->delta_pic_order_always_zero_flag)
  {
    len += se_v ("SH: delta_pic_order_cnt[0]", img->delta_pic_order_cnt[0], bitstream);

    if (img->pic_order_present_flag)
    {
      len += se_v ("SH: delta_pic_order_cnt[1]", img->delta_pic_order_cnt[1], bitstream);
    }
  }

  // Direct Mode Type selection for B pictures
  if (img->type==B_SLICE)
  {
    len +=  u_1 ("SH: direct_spatial_mv_pred_flag", img->direct_spatial_mv_pred_flag, bitstream);    
  }

  if ((img->type == P_SLICE) || (img->type == B_SLICE))
  {
    int override_flag;
    if ((img->type == P_SLICE))
    {
      override_flag = (img->num_ref_idx_l0_active != (active_pps->num_ref_idx_l0_active_minus1 +1)) ? 1 : 0;
    }
    else
    {
      override_flag = ((img->num_ref_idx_l0_active != (active_pps->num_ref_idx_l0_active_minus1 +1)) 
        || (img->num_ref_idx_l1_active != (active_pps->num_ref_idx_l1_active_minus1 +1))) ? 1 : 0;
    }

    len +=  u_1 ("SH: num_ref_idx_active_override_flag", override_flag, bitstream);

    if (override_flag) 
    {
      len += ue_v ("SH: num_ref_idx_l0_active_minus1", img->num_ref_idx_l0_active-1, bitstream);
      if (img->type==B_SLICE)
      {
        len += ue_v ("SH: num_ref_idx_l1_active_minus1", img->num_ref_idx_l1_active-1, bitstream);
      }
    }

  }
  len += ref_pic_list_reordering(bitstream);

  if (img->nal_reference_idc)
    len += dec_ref_pic_marking(bitstream);

  if(img->type!=I_SLICE)
  {
    len += ue_v("SH: cabac_init_idc", img->model_number, bitstream);
  }

  len += se_v("SH: slice_qp_delta", (currSlice->qp - 26 - active_pps->pic_init_qp_minus26), bitstream);  

  if (active_pps->deblocking_filter_control_present_flag)
  {
    len += ue_v("SH: disable_deblocking_filter_idc",img->LFDisableIdc, bitstream);  // Turn loop filter on/off on slice basis 

    if (img->LFDisableIdc!=1)
    {
      len += se_v ("SH: slice_alpha_c0_offset_div2", img->LFAlphaC0Offset / 2, bitstream);

      len += se_v ("SH: slice_beta_offset_div2", img->LFBetaOffset / 2, bitstream);
    }
  }

  len += ue_v ("SH: motion vector resolution", img->mv_res, bitstream);
  if ( active_pps->num_slice_groups_minus1>0 &&
    active_pps->slice_group_map_type>=3 && active_pps->slice_group_map_type<=5)
  {
    numtmp=img->PicHeightInMapUnits*img->PicWidthInMbs/(float)(active_pps->slice_group_change_rate_minus1+1)+1;
    num_bits_slice_group_change_cycle = (int)ceil(log(numtmp)/log(2));

    //! img->slice_group_change_cycle can be changed before calling FmoInit()
    len += u_v (num_bits_slice_group_change_cycle, "SH: slice_group_change_cycle", img->slice_group_change_cycle, bitstream);
  }

#ifdef USE_INTRA_MDDT
  len += u_v(1, "SE: use_intra_MDDT", input->UseIntraMDDT, bitstream);
#endif

#ifdef USE_HP_FILTER
#ifdef SWITCHED_FILTERS
  len += u_v(3, "SE: use_HP_filter", input->UseHPFilter, bitstream);
#else
  len += u_v(2, "SE: use_HP_filter", input->UseHPFilter, bitstream);
#endif
#endif

#ifdef SWITCHED_FILTERS
  if(img->type != I_SLICE)
  {	
    if(input->UseHPFilter == HPF_SIFO_FPO)
    {
      int sub_pos;

      if(img->filterParam == SIFO_FIRST_PASS_FPO)
      {    
        len += u_v(1, "SH: use offsets", SIFO_USE_OFFSETS, bitstream);  // 1 bit only
        len += u_v (3, "SH: bestFilter", img->bestFilter, bitstream);
      }

      if(img->filterParam == SIFO_FIRST_PASS_FPO)
      {    
        if(img->bestFilter == NO_BEST_FILTER)
        {
          for(sub_pos = 1; sub_pos < 16; sub_pos++)
          {
            len += u_v (2, "SH: filterSequence", img->filterSequence[sub_pos], bitstream);
          }
        }
      } 

      // Save offsets
      if((img->filterParam == SIFO_FIRST_PASS_FPO))
      {
        int list;
        int listNo = (img->type == B_SLICE)? 2: 1;
        int yuv = 0;
        for(list = 0; list < listNo; ++list) 
        {
          int frame;
          int sub_pos;
          int nonzero = getNonzero(yuv,list);

          len += u_1("nonzero offset", nonzero, bitstream);
          if(nonzero)
          {
            for(frame = 0; frame < listXsize[list]; ++frame)
            {
              if(frame == 0)     
              {    
                for(sub_pos = 0; sub_pos < 16; ++sub_pos)   
                {
                  len += se_v("SH: subpelOffset", img->subpelOffset[yuv][list][sub_pos]/(1<<img->BitDepthIncrease), bitstream);
                }         
              }
              else              
              {
                len += se_v("SH: imgOffset", img->imgOffset[yuv][list][frame]/(1<<img->BitDepthIncrease), bitstream);
              }
            }
          }
        }
      }
    }
  }
#endif  // SWITCHED_FILTERS

#ifdef ADAPTIVE_LOOP_FILTER
  currSlice->alf_flag_bit_to_go      = bitstream->bits_to_go;
  currSlice->alf_flag_stream_pointer = &(bitstream->streamBuffer[bitstream->byte_pos]);
  len += u_1("SH: adaptive_loop_filter_flag", 0, bitstream);

  // padding end of slice header for ALF
  if(bitstream->bits_to_go!=8)
  {
    while(bitstream->bits_to_go!=8)
    {
      len += u_1("SH: padding bits for ALF", 1, bitstream);
    }
  }
  currSlice->sh_byte_pos = bitstream->byte_pos;
#endif


  return len;
}

/*!
********************************************************************************************
* \brief 
*    writes the ref_pic_list_reordering syntax
*    based on content of according fields in img structure
*
* \return
*    number of bits used 
********************************************************************************************
*/
static int ref_pic_list_reordering(Bitstream *bitstream)
{
  Slice *currSlice = img->currentSlice;

  int i, len=0;

  if ((img->type!=I_SLICE) /*&&(img->type!=SI_IMG)*/ )
  {
    len += u_1 ("SH: ref_pic_list_reordering_flag_l0", currSlice->ref_pic_list_reordering_flag_l0, bitstream);
    if (currSlice->ref_pic_list_reordering_flag_l0)
    {
      i=-1;
      do
      {
        i++;
        len += ue_v ("SH: reordering_of_pic_num_idc", currSlice->reordering_of_pic_nums_idc_l0[i], bitstream);
        if (currSlice->reordering_of_pic_nums_idc_l0[i]==0 ||
          currSlice->reordering_of_pic_nums_idc_l0[i]==1)
        {
          len += ue_v ("SH: abs_diff_pic_num_minus1_l0", currSlice->abs_diff_pic_num_minus1_l0[i], bitstream);
        }
        else
        {
          if (currSlice->reordering_of_pic_nums_idc_l0[i]==2)
          {
            len += ue_v ("SH: long_term_pic_idx_l0", currSlice->long_term_pic_idx_l0[i], bitstream);
          }
        }

      } while (currSlice->reordering_of_pic_nums_idc_l0[i] != 3);
    }
  }

  if (img->type==B_SLICE)
  {
    len += u_1 ("SH: ref_pic_list_reordering_flag_l1", currSlice->ref_pic_list_reordering_flag_l1, bitstream);
    if (currSlice->ref_pic_list_reordering_flag_l1)
    {
      i=-1;
      do
      {
        i++;
        len += ue_v ("SH: remapping_of_pic_num_idc", currSlice->reordering_of_pic_nums_idc_l1[i], bitstream);
        if (currSlice->reordering_of_pic_nums_idc_l1[i]==0 ||
          currSlice->reordering_of_pic_nums_idc_l1[i]==1)
        {
          len += ue_v ("SH: abs_diff_pic_num_minus1_l1", currSlice->abs_diff_pic_num_minus1_l1[i], bitstream);
        }
        else
        {
          if (currSlice->reordering_of_pic_nums_idc_l1[i]==2)
          {
            len += ue_v ("SH: long_term_pic_idx_l1", currSlice->long_term_pic_idx_l1[i], bitstream);
          }
        }
      } while (currSlice->reordering_of_pic_nums_idc_l1[i] != 3);
    }
  }

  return len;
}


/*!
************************************************************************
* \brief
*    write the memory management control operations
*
* \return
*    number of bits used 
************************************************************************
*/
static int dec_ref_pic_marking(Bitstream *bitstream)
{
  DecRefPicMarking_t *tmp_drpm;

  int val, len=0;

  if (img->currentPicture->idr_flag)
  {
    len += u_1("SH: no_output_of_prior_pics_flag", img->no_output_of_prior_pics_flag, bitstream);
    len += u_1("SH: long_term_reference_flag", img->long_term_reference_flag, bitstream);
  }
  else
  {
    img->adaptive_ref_pic_buffering_flag = (img->dec_ref_pic_marking_buffer!=NULL);

    len += u_1("SH: adaptive_ref_pic_buffering_flag", img->adaptive_ref_pic_buffering_flag, bitstream);

    if (img->adaptive_ref_pic_buffering_flag)
    {
      tmp_drpm = img->dec_ref_pic_marking_buffer;
      // write Memory Management Control Operation 
      do
      {
        if (tmp_drpm==NULL) error ("Error encoding MMCO commands", 500);

        val = tmp_drpm->memory_management_control_operation;
        len += ue_v("SH: memory_management_control_operation", val, bitstream);

        if ((val==1)||(val==3)) 
        {
          len += 1 + ue_v("SH: difference_of_pic_nums_minus1", tmp_drpm->difference_of_pic_nums_minus1, bitstream);
        }
        if (val==2)
        {
          len+= ue_v("SH: long_term_pic_num", tmp_drpm->long_term_pic_num, bitstream);
        }
        if ((val==3)||(val==6))
        {
          len+= ue_v("SH: long_term_frame_idx", tmp_drpm->long_term_frame_idx, bitstream);
        }
        if (val==4)
        {
          len += ue_v("SH: max_long_term_pic_idx_plus1", tmp_drpm->max_long_term_frame_idx_plus1, bitstream);
        }

        tmp_drpm=tmp_drpm->Next;

      } while (val != 0);

    }
  }
  return len;
}


/*!
************************************************************************
* \brief
*    Selects picture type and codes it to symbol
*
* \return
*    symbol value for picture type
************************************************************************
*/
int get_picture_type()
{
  // set this value to zero for transmission without signaling 
  // that the whole picture has the same slice type
  int same_slicetype_for_whole_frame = 5;

  switch (img->type)
  {
  case I_SLICE:
    return 2 + same_slicetype_for_whole_frame;
    break;
  case P_SLICE:
    return 0 + same_slicetype_for_whole_frame;
    break;
  case B_SLICE:
    return 1 + same_slicetype_for_whole_frame;
    break;
  default:
    error("Picture Type not supported!",1);
    break;
  }

  return 0;
}

