/* 
====================================================================================================================

The copyright in this software is being made available under the License included below. 
This software may be subject to other third party and 	contributor rights, including patent rights, and no such 
rights are granted under this license.

Copyright (c) 2010, QUALCOMM INC.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted only for 
the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
promoting such standards. The following conditions are required to be met:

* Redistributions of source code must retain the above copyright notice, this list of conditions and 
the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
the following disclaimer in the documentation and/or other materials provided with the distribution.
* The name of QUALCOMM INC.  may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

====================================================================================================================
*/

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>

#include "global.h"

#include "image.h"
#include "mb_access.h"
#include "elements.h"
#include "cabac.h"
#include "vlc.h"
#include "memalloc.h"
#include "elements.h"
#include "mode_decision.h"
#include "mb_access.h"

extern int   **cofAC4x4;

// Notation for comments regarding prediction and predictors.
// The pels of the 4x4 block are labelled a..p. The predictor pels above
// are labelled A..H, from the left I..P, and from above left X, as follows:
//
//  X A B C D E F G H
//  I a b c d
//  J e f g h
//  K i j k l
//  L m n o p
//
// Notation for comments regarding prediction and predictors.
// The pels of the 4x4 block are labelled a..p. The predictor pels above
// are labelled A..H, from the left I..P, and from above left X, as follows:
//
//  X  A  B  C  D  E  F  G  H  I  J  K  L  M   N  O  P  
//  Q  a1 b1 c1 d1 e1 f1 g1 h1
//  R  a2 b2 c2 d2 e2 f2 g2 h2
//  S  a3 b3 c3 d3 e3 f3 g3 h3
//  T  a4 b4 c4 d4 e4 f4 g4 h4
//  U  a5 b5 c5 d5 e5 f5 g5 h5
//  V  a6 b6 c6 d6 e6 f6 g6 h6
//  W  a7 b7 c7 d7 e7 f7 g7 h7
//  Z  a8 b8 c8 d8 e8 f8 g8 h8
// Predictor array index definitions
#define P_X (PredPel[0])
#define P_A (PredPel[1])
#define P_B (PredPel[2])
#define P_C (PredPel[3])
#define P_D (PredPel[4])
#define P_E (PredPel[5])
#define P_F (PredPel[6])
#define P_G (PredPel[7])
#define P_H (PredPel[8])
#define P_I (PredPel[9])
#define P_J (PredPel[10])
#define P_K (PredPel[11])
#define P_L (PredPel[12])
#define P_M (PredPel[13])
#define P_N (PredPel[14])
#define P_O (PredPel[15])
#define P_P (PredPel[16])
#define P_Q (PredPel[17])
#define P_R (PredPel[18])
#define P_S (PredPel[19])
#define P_T (PredPel[20])
#define P_U (PredPel[21])
#define P_V (PredPel[22])
#define P_W (PredPel[23])
#define P_Z (PredPel[24])

/*!
************************************************************************
* \brief
*    Make intra 4x4 prediction according to all 9 prediction modes.
*    The routine uses left and upper neighbouring points from
*    previous coded blocks to do this (if available). Notice that
*    inaccessible neighbouring points are signalled with a negative
*    value in the predmode array .
*
*  \par Input:
*     Starting point of current 4x4 block image posision
*
*  \par Output:
*      none
************************************************************************
*/
void intrapred_luma(int img_x,int img_y, int *left_available, int *up_available, int *all_available)
{
  int i,j;
  int s0;
  int PredPel[13];  // array of predictor pels
  imgpel **imgY = enc_picture->imgY;  // For MB level frame/field coding tools -- set default to imgY

  int ioff = (img_x & 15);
  int joff = (img_y & 15);
  int mb_nr=img->current_mb_nr;

  PixelPos pix_a[4];
  PixelPos pix_b, pix_c, pix_d;

  int block_available_up;
  int block_available_left;
  int block_available_up_left;
  int block_available_up_right;

  for (i=0;i<4;i++)
  {
    getNeighbour(mb_nr, ioff -1 , joff +i , 1, &pix_a[i]);
  }

  getNeighbour(mb_nr, ioff    , joff -1 , 1, &pix_b);
  getNeighbour(mb_nr, ioff +4 , joff -1 , 1, &pix_c);
  getNeighbour(mb_nr, ioff -1 , joff -1 , 1, &pix_d);

  pix_c.available = pix_c.available && !((ioff==4) && ((joff==4)||(joff==12)));

  if (input->UseConstrainedIntraPred)
  {
    for (i=0, block_available_left=1; i<4;i++)
      block_available_left  &= pix_a[i].available ? img->intra_block[pix_a[i].mb_addr]: 0;
    block_available_up       = pix_b.available ? img->intra_block [pix_b.mb_addr] : 0;
    block_available_up_right = pix_c.available ? img->intra_block [pix_c.mb_addr] : 0;
    block_available_up_left  = pix_d.available ? img->intra_block [pix_d.mb_addr] : 0;
  }
  else
  {
    block_available_left     = pix_a[0].available;
    block_available_up       = pix_b.available;
    block_available_up_right = pix_c.available;
    block_available_up_left  = pix_d.available;
  }

  *left_available = block_available_left;
  *up_available   = block_available_up;
  *all_available  = block_available_up && block_available_left && block_available_up_left;

  i = (img_x & 15);
  j = (img_y & 15);

  // form predictor pels
  if (block_available_up)
  {
    P_A = imgY[pix_b.pos_y][pix_b.pos_x+0];
    P_B = imgY[pix_b.pos_y][pix_b.pos_x+1];
    P_C = imgY[pix_b.pos_y][pix_b.pos_x+2];
    P_D = imgY[pix_b.pos_y][pix_b.pos_x+3];

  }
  else
  {
    P_A = P_B = P_C = P_D = img->dc_pred_value_luma;
  }

  if (block_available_up_right)
  {
    P_E = imgY[pix_c.pos_y][pix_c.pos_x+0];
    P_F = imgY[pix_c.pos_y][pix_c.pos_x+1];
    P_G = imgY[pix_c.pos_y][pix_c.pos_x+2];
    P_H = imgY[pix_c.pos_y][pix_c.pos_x+3];
  }
  else
  {
    P_E = P_F = P_G = P_H = P_D;
  }

  if (block_available_left)
  {
    P_I = imgY[pix_a[0].pos_y][pix_a[0].pos_x];
    P_J = imgY[pix_a[1].pos_y][pix_a[1].pos_x];
    P_K = imgY[pix_a[2].pos_y][pix_a[2].pos_x];
    P_L = imgY[pix_a[3].pos_y][pix_a[3].pos_x];
  }
  else
  {
    P_I = P_J = P_K = P_L = img->dc_pred_value_luma;
  }

  if (block_available_up_left)
  {
    P_X = imgY[pix_d.pos_y][pix_d.pos_x];
  }
  else
  {
    P_X = img->dc_pred_value_luma;
  }

  for(i=0;i<9;i++)
    img->mprr[i][0][0]=-1;

  ///////////////////////////////
  // make DC prediction
  ///////////////////////////////
  s0 = 0;
  if (block_available_up && block_available_left)
  {   
    // no edge
    s0 = (P_A + P_B + P_C + P_D + P_I + P_J + P_K + P_L + 4) >> (BLOCK_SHIFT + 1);
  }
  else if (!block_available_up && block_available_left)
  {
    // upper edge
    s0 = (P_I + P_J + P_K + P_L + 2) >> BLOCK_SHIFT;;             
  }
  else if (block_available_up && !block_available_left)
  {
    // left edge
    s0 = (P_A + P_B + P_C + P_D + 2) >> BLOCK_SHIFT;             
  }
  else //if (!block_available_up && !block_available_left)
  {
    // top left corner, nothing to predict from
    s0 = img->dc_pred_value_luma;                           
  }

  // store DC prediction
  for (j=0; j < BLOCK_SIZE; j++)
  {
    for (i=0; i < BLOCK_SIZE; i++)
      img->mprr[DC_PRED][j][i] = s0;
  }

  ///////////////////////////////
  // make horiz and vert prediction
  ///////////////////////////////

  for (i=0; i < BLOCK_SIZE; i++)
  {
    img->mprr[VERT_PRED][0][i] = 
      img->mprr[VERT_PRED][1][i] = 
      img->mprr[VERT_PRED][2][i] = 
      img->mprr[VERT_PRED][3][i] = (&P_A)[i];
    img->mprr[HOR_PRED][i][0]  = 
      img->mprr[HOR_PRED][i][1]  = 
      img->mprr[HOR_PRED][i][2]  = 
      img->mprr[HOR_PRED][i][3]  = (&P_I)[i];
  }

  if(!block_available_up)
    img->mprr[VERT_PRED][0][0]=-1;
  if(!block_available_left)
    img->mprr[HOR_PRED][0][0]=-1;

  if (block_available_up) 
  {
    // Mode DIAG_DOWN_LEFT_PRED
    img->mprr[DIAG_DOWN_LEFT_PRED][0][0] = (P_A + P_C + 2*(P_B) + 2) >> 2;
    img->mprr[DIAG_DOWN_LEFT_PRED][0][1] = 
      img->mprr[DIAG_DOWN_LEFT_PRED][1][0] = (P_B + P_D + 2*(P_C) + 2) >> 2;
    img->mprr[DIAG_DOWN_LEFT_PRED][0][2] =
      img->mprr[DIAG_DOWN_LEFT_PRED][1][1] =
      img->mprr[DIAG_DOWN_LEFT_PRED][2][0] = (P_C + P_E + 2*(P_D) + 2) >> 2;
    img->mprr[DIAG_DOWN_LEFT_PRED][0][3] = 
      img->mprr[DIAG_DOWN_LEFT_PRED][1][2] = 
      img->mprr[DIAG_DOWN_LEFT_PRED][2][1] = 
      img->mprr[DIAG_DOWN_LEFT_PRED][3][0] = (P_D + P_F + 2*(P_E) + 2) >> 2;
    img->mprr[DIAG_DOWN_LEFT_PRED][1][3] = 
      img->mprr[DIAG_DOWN_LEFT_PRED][2][2] = 
      img->mprr[DIAG_DOWN_LEFT_PRED][3][1] = (P_E + P_G + 2*(P_F) + 2) >> 2;
    img->mprr[DIAG_DOWN_LEFT_PRED][2][3] = 
      img->mprr[DIAG_DOWN_LEFT_PRED][3][2] = (P_F + P_H + 2*(P_G) + 2) >> 2;
    img->mprr[DIAG_DOWN_LEFT_PRED][3][3] = (P_G + 3*(P_H) + 2) >> 2;

    // Mode VERT_LEFT_PRED
    img->mprr[VERT_LEFT_PRED][0][0] = (P_A + P_B + 1) >> 1;
    img->mprr[VERT_LEFT_PRED][0][1] = 
      img->mprr[VERT_LEFT_PRED][2][0] = (P_B + P_C + 1) >> 1;
    img->mprr[VERT_LEFT_PRED][0][2] = 
      img->mprr[VERT_LEFT_PRED][2][1] = (P_C + P_D + 1) >> 1;
    img->mprr[VERT_LEFT_PRED][0][3] = 
      img->mprr[VERT_LEFT_PRED][2][2] = (P_D + P_E + 1) >> 1;
    img->mprr[VERT_LEFT_PRED][2][3] = (P_E + P_F + 1) >> 1;
    img->mprr[VERT_LEFT_PRED][1][0] = (P_A + 2*P_B + P_C + 2) >> 2;
    img->mprr[VERT_LEFT_PRED][1][1] = 
      img->mprr[VERT_LEFT_PRED][3][0] = (P_B + 2*P_C + P_D + 2) >> 2;
    img->mprr[VERT_LEFT_PRED][1][2] = 
      img->mprr[VERT_LEFT_PRED][3][1] = (P_C + 2*P_D + P_E + 2) >> 2;
    img->mprr[VERT_LEFT_PRED][1][3] = 
      img->mprr[VERT_LEFT_PRED][3][2] = (P_D + 2*P_E + P_F + 2) >> 2;
    img->mprr[VERT_LEFT_PRED][3][3] = (P_E + 2*P_F + P_G + 2) >> 2;

  }

  /*  Prediction according to 'diagonal' modes */
  if (block_available_left) 
  {
    // Mode HOR_UP_PRED
    img->mprr[HOR_UP_PRED][0][0] = (P_I + P_J + 1) >> 1;
    img->mprr[HOR_UP_PRED][0][1] = (P_I + 2*P_J + P_K + 2) >> 2;
    img->mprr[HOR_UP_PRED][0][2] = 
      img->mprr[HOR_UP_PRED][1][0] = (P_J + P_K + 1) >> 1;
    img->mprr[HOR_UP_PRED][0][3] = 
      img->mprr[HOR_UP_PRED][1][1] = (P_J + 2*P_K + P_L + 2) >> 2;
    img->mprr[HOR_UP_PRED][1][2] = 
      img->mprr[HOR_UP_PRED][2][0] = (P_K + P_L + 1) >> 1;
    img->mprr[HOR_UP_PRED][1][3] = 
      img->mprr[HOR_UP_PRED][2][1] = (P_K + 2*P_L + P_L + 2) >> 2;
    img->mprr[HOR_UP_PRED][3][0] = 
      img->mprr[HOR_UP_PRED][2][2] = 
      img->mprr[HOR_UP_PRED][2][3] = 
      img->mprr[HOR_UP_PRED][3][1] = 
      img->mprr[HOR_UP_PRED][3][2] = 
      img->mprr[HOR_UP_PRED][3][3] = P_L;
  }

  /*  Prediction according to 'diagonal' modes */
  if (block_available_up && block_available_left && block_available_up_left) 
  {
    // Mode DIAG_DOWN_RIGHT_PRED
    img->mprr[DIAG_DOWN_RIGHT_PRED][3][0] = (P_L + 2*P_K + P_J + 2) >> 2; 
    img->mprr[DIAG_DOWN_RIGHT_PRED][2][0] =
      img->mprr[DIAG_DOWN_RIGHT_PRED][3][1] = (P_K + 2*P_J + P_I + 2) >> 2; 
    img->mprr[DIAG_DOWN_RIGHT_PRED][1][0] =
      img->mprr[DIAG_DOWN_RIGHT_PRED][2][1] = 
      img->mprr[DIAG_DOWN_RIGHT_PRED][3][2] = (P_J + 2*P_I + P_X + 2) >> 2; 
    img->mprr[DIAG_DOWN_RIGHT_PRED][0][0] =
      img->mprr[DIAG_DOWN_RIGHT_PRED][1][1] =
      img->mprr[DIAG_DOWN_RIGHT_PRED][2][2] =
      img->mprr[DIAG_DOWN_RIGHT_PRED][3][3] = (P_I + 2*P_X + P_A + 2) >> 2; 
    img->mprr[DIAG_DOWN_RIGHT_PRED][0][1] =
      img->mprr[DIAG_DOWN_RIGHT_PRED][1][2] =
      img->mprr[DIAG_DOWN_RIGHT_PRED][2][3] = (P_X + 2*P_A + P_B + 2) >> 2;
    img->mprr[DIAG_DOWN_RIGHT_PRED][0][2] =
      img->mprr[DIAG_DOWN_RIGHT_PRED][1][3] = (P_A + 2*P_B + P_C + 2) >> 2;
    img->mprr[DIAG_DOWN_RIGHT_PRED][0][3] = (P_B + 2*P_C + P_D + 2) >> 2;

    // Mode VERT_RIGHT_PRED
    img->mprr[VERT_RIGHT_PRED][0][0] = 
      img->mprr[VERT_RIGHT_PRED][2][1] = (P_X + P_A + 1) >> 1;
    img->mprr[VERT_RIGHT_PRED][0][1] = 
      img->mprr[VERT_RIGHT_PRED][2][2] = (P_A + P_B + 1) >> 1;
    img->mprr[VERT_RIGHT_PRED][0][2] = 
      img->mprr[VERT_RIGHT_PRED][2][3] = (P_B + P_C + 1) >> 1;
    img->mprr[VERT_RIGHT_PRED][0][3] = (P_C + P_D + 1) >> 1;
    img->mprr[VERT_RIGHT_PRED][1][0] = 
      img->mprr[VERT_RIGHT_PRED][3][1] = (P_I + 2*P_X + P_A + 2) >> 2;
    img->mprr[VERT_RIGHT_PRED][1][1] = 
      img->mprr[VERT_RIGHT_PRED][3][2] = (P_X + 2*P_A + P_B + 2) >> 2;
    img->mprr[VERT_RIGHT_PRED][1][2] = 
      img->mprr[VERT_RIGHT_PRED][3][3] = (P_A + 2*P_B + P_C + 2) >> 2;
    img->mprr[VERT_RIGHT_PRED][1][3] = (P_B + 2*P_C + P_D + 2) >> 2;
    img->mprr[VERT_RIGHT_PRED][2][0] = (P_X + 2*P_I + P_J + 2) >> 2;
    img->mprr[VERT_RIGHT_PRED][3][0] = (P_I + 2*P_J + P_K + 2) >> 2;

    // Mode HOR_DOWN_PRED
    img->mprr[HOR_DOWN_PRED][0][0] = 
      img->mprr[HOR_DOWN_PRED][1][2] = (P_X + P_I + 1) >> 1;
    img->mprr[HOR_DOWN_PRED][0][1] = 
      img->mprr[HOR_DOWN_PRED][1][3] = (P_I + 2*P_X + P_A + 2) >> 2;
    img->mprr[HOR_DOWN_PRED][0][2] = (P_X + 2*P_A + P_B + 2) >> 2;
    img->mprr[HOR_DOWN_PRED][0][3] = (P_A + 2*P_B + P_C + 2) >> 2;
    img->mprr[HOR_DOWN_PRED][1][0] = 
      img->mprr[HOR_DOWN_PRED][2][2] = (P_I + P_J + 1) >> 1;
    img->mprr[HOR_DOWN_PRED][1][1] = 
      img->mprr[HOR_DOWN_PRED][2][3] = (P_X + 2*P_I + P_J + 2) >> 2;
    img->mprr[HOR_DOWN_PRED][2][0] = 
      img->mprr[HOR_DOWN_PRED][3][2] = (P_J + P_K + 1) >> 1;
    img->mprr[HOR_DOWN_PRED][2][1] = 
      img->mprr[HOR_DOWN_PRED][3][3] = (P_I + 2*P_J + P_K + 2) >> 2;
    img->mprr[HOR_DOWN_PRED][3][0] = (P_K + P_L + 1) >> 1;
    img->mprr[HOR_DOWN_PRED][3][1] = (P_J + 2*P_K + P_L + 2) >> 2;
  }
}


/*!
************************************************************************
* \brief
*    Make intra 8x8 prediction according to all 9 prediction modes.
*    The routine uses left and upper neighbouring points from
*    previous coded blocks to do this (if available). Notice that
*    inaccessible neighbouring points are signalled with a negative
*    value in the predmode array .
*
*  \par Input:
*     Starting point of current 8x8 block image posision
*
*  \par Output:
*      none
************************************************************************
*/
void intrapred_luma_8x8(int img_x,int img_y, int *left_available, int *up_available, int *all_available)
{
  int i,j;
  int s0;
  int PredPel[25];  // array of predictor pels
  imgpel **imgY = enc_picture->imgY;  // For MB level frame/field coding tools -- set default to imgY

  int ioff = (img_x & 15);
  int joff = (img_y & 15);
  int mb_nr=img->current_mb_nr;

  PixelPos pix_a[8];
  PixelPos pix_b, pix_c, pix_d;

  int block_available_up;
  int block_available_left;
  int block_available_up_left;
  int block_available_up_right;

  for (i=0;i<8;i++)
  {
    getNeighbour(mb_nr, ioff -1 , joff +i , 1, &pix_a[i]);
  }

  getNeighbour(mb_nr, ioff    , joff -1 , 1, &pix_b);
  getNeighbour(mb_nr, ioff +8 , joff -1 , 1, &pix_c);
  getNeighbour(mb_nr, ioff -1 , joff -1 , 1, &pix_d);

  pix_c.available = pix_c.available &&!(ioff == 8 && joff == 8);

  if (input->UseConstrainedIntraPred)
  {
    for (i=0, block_available_left=1; i<8;i++)
      block_available_left  &= pix_a[i].available ? img->intra_block[pix_a[i].mb_addr]: 0;
    block_available_up       = pix_b.available ? img->intra_block [pix_b.mb_addr] : 0;
    block_available_up_right = pix_c.available ? img->intra_block [pix_c.mb_addr] : 0;
    block_available_up_left  = pix_d.available ? img->intra_block [pix_d.mb_addr] : 0;
  }
  else
  {
    block_available_left     = pix_a[0].available;
    block_available_up       = pix_b.available;
    block_available_up_right = pix_c.available;
    block_available_up_left  = pix_d.available;
  }

  *left_available = block_available_left;
  *up_available   = block_available_up;
  *all_available  = block_available_up && block_available_left && block_available_up_left;

  i = (img_x & 15);
  j = (img_y & 15);

  // form predictor pels
  // form predictor pels
  if (block_available_up)
  {
    P_A = imgY[pix_b.pos_y][pix_b.pos_x+0];
    P_B = imgY[pix_b.pos_y][pix_b.pos_x+1];
    P_C = imgY[pix_b.pos_y][pix_b.pos_x+2];
    P_D = imgY[pix_b.pos_y][pix_b.pos_x+3];
    P_E = imgY[pix_b.pos_y][pix_b.pos_x+4];
    P_F = imgY[pix_b.pos_y][pix_b.pos_x+5];
    P_G = imgY[pix_b.pos_y][pix_b.pos_x+6];
    P_H = imgY[pix_b.pos_y][pix_b.pos_x+7];
  }
  else
  {
    P_A = P_B = P_C = P_D = P_E = P_F = P_G = P_H = img->dc_pred_value_luma;
  }

  if (block_available_up_right)
  {
    P_I = imgY[pix_c.pos_y][pix_c.pos_x+0];
    P_J = imgY[pix_c.pos_y][pix_c.pos_x+1];
    P_K = imgY[pix_c.pos_y][pix_c.pos_x+2];
    P_L = imgY[pix_c.pos_y][pix_c.pos_x+3];
    P_M = imgY[pix_c.pos_y][pix_c.pos_x+4];
    P_N = imgY[pix_c.pos_y][pix_c.pos_x+5];
    P_O = imgY[pix_c.pos_y][pix_c.pos_x+6];
    P_P = imgY[pix_c.pos_y][pix_c.pos_x+7];

  }
  else
  {
    P_I = P_J = P_K = P_L = P_M = P_N = P_O = P_P = P_H;
  }

  if (block_available_left)
  {
    P_Q = imgY[pix_a[0].pos_y][pix_a[0].pos_x];
    P_R = imgY[pix_a[1].pos_y][pix_a[1].pos_x];
    P_S = imgY[pix_a[2].pos_y][pix_a[2].pos_x];
    P_T = imgY[pix_a[3].pos_y][pix_a[3].pos_x];
    P_U = imgY[pix_a[4].pos_y][pix_a[4].pos_x];
    P_V = imgY[pix_a[5].pos_y][pix_a[5].pos_x];
    P_W = imgY[pix_a[6].pos_y][pix_a[6].pos_x];
    P_Z = imgY[pix_a[7].pos_y][pix_a[7].pos_x];
  }
  else
  {
    P_Q = P_R = P_S = P_T = P_U = P_V = P_W = P_Z = img->dc_pred_value_luma;
  }

  if (block_available_up_left)
  {
    P_X = imgY[pix_d.pos_y][pix_d.pos_x];
  }
  else
  {
    P_X = img->dc_pred_value_luma;
  }

  for(i=0;i<9;i++)
    img->mprr_3[i][0][0]=-1;

  LowPassForIntra8x8Pred(&(P_X), block_available_up_left, block_available_up, block_available_left);

  ///////////////////////////////
  // make DC prediction
  ///////////////////////////////
  s0 = 0;
  if (block_available_up && block_available_left)
  {   
    // no edge
    s0 = (P_A + P_B + P_C + P_D + P_E + P_F + P_G + P_H + P_Q + P_R + P_S + P_T + P_U + P_V + P_W + P_Z + 8) >> 4;
  }
  else if (!block_available_up && block_available_left)
  {
    // upper edge
    s0 = (P_Q + P_R + P_S + P_T + P_U + P_V + P_W + P_Z + 4) >> 3;             
  }
  else if (block_available_up && !block_available_left)
  {
    // left edge
    s0 = (P_A + P_B + P_C + P_D + P_E + P_F + P_G + P_H + 4) >> 3;             
  }
  else //if (!block_available_up && !block_available_left)
  {
    // top left corner, nothing to predict from
    s0 = img->dc_pred_value_luma;                           
  }

  // store DC prediction
  for (j=0; j < 2*BLOCK_SIZE; j++)
  {
    for (i=0; i < 2*BLOCK_SIZE; i++)
    {
      img->mprr_3[DC_PRED][i][j] = s0;
    }
  }


  ///////////////////////////////
  // make horiz and vert prediction
  ///////////////////////////////

  for (i=0; i < 2*BLOCK_SIZE; i++)
  {
    img->mprr_3[VERT_PRED][0][i] = 
      img->mprr_3[VERT_PRED][1][i] = 
      img->mprr_3[VERT_PRED][2][i] = 
      img->mprr_3[VERT_PRED][3][i] = 
      img->mprr_3[VERT_PRED][4][i] = 
      img->mprr_3[VERT_PRED][5][i] = 
      img->mprr_3[VERT_PRED][6][i] = 
      img->mprr_3[VERT_PRED][7][i] = (&P_A)[i];
    img->mprr_3[HOR_PRED][i][0]  = 
      img->mprr_3[HOR_PRED][i][1]  = 
      img->mprr_3[HOR_PRED][i][2]  = 
      img->mprr_3[HOR_PRED][i][3]  = 
      img->mprr_3[HOR_PRED][i][4]  = 
      img->mprr_3[HOR_PRED][i][5]  = 
      img->mprr_3[HOR_PRED][i][6]  = 
      img->mprr_3[HOR_PRED][i][7]  = (&P_Q)[i];
  }

  if(!block_available_up)img->mprr_3[VERT_PRED][0][0]=-1;
  if(!block_available_left)img->mprr_3[HOR_PRED][0][0]=-1;

  ///////////////////////////////////
  // make diagonal down left prediction
  ///////////////////////////////////
  if (block_available_up) 
  {
    // Mode DIAG_DOWN_LEFT_PRED
    img->mprr_3[DIAG_DOWN_LEFT_PRED][0][0] = (P_A + P_C + 2*(P_B) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_LEFT_PRED][0][1] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][1][0] = (P_B + P_D + 2*(P_C) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_LEFT_PRED][0][2] =
      img->mprr_3[DIAG_DOWN_LEFT_PRED][1][1] =
      img->mprr_3[DIAG_DOWN_LEFT_PRED][2][0] = (P_C + P_E + 2*(P_D) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_LEFT_PRED][0][3] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][1][2] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][2][1] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][3][0] = (P_D + P_F + 2*(P_E) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_LEFT_PRED][0][4] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][1][3] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][2][2] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][3][1] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][4][0] = (P_E + P_G + 2*(P_F) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_LEFT_PRED][0][5] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][1][4] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][2][3] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][3][2] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][4][1] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][5][0] = (P_F + P_H + 2*(P_G) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_LEFT_PRED][0][6] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][1][5] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][2][4] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][3][3] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][4][2] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][5][1] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][6][0] = (P_G + P_I + 2*(P_H) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_LEFT_PRED][0][7] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][1][6] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][2][5] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][3][4] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][4][3] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][5][2] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][6][1] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][7][0] = (P_H + P_J + 2*(P_I) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_LEFT_PRED][1][7] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][2][6] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][3][5] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][4][4] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][5][3] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][6][2] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][7][1] = (P_I + P_K + 2*(P_J) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_LEFT_PRED][2][7] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][3][6] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][4][5] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][5][4] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][6][3] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][7][2] = (P_J + P_L + 2*(P_K) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_LEFT_PRED][3][7] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][4][6] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][5][5] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][6][4] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][7][3] = (P_K + P_M + 2*(P_L) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_LEFT_PRED][4][7] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][5][6] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][6][5] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][7][4] = (P_L + P_N + 2*(P_M) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_LEFT_PRED][5][7] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][6][6] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][7][5] = (P_M + P_O + 2*(P_N) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_LEFT_PRED][6][7] = 
      img->mprr_3[DIAG_DOWN_LEFT_PRED][7][6] = (P_N + P_P + 2*(P_O) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_LEFT_PRED][7][7] = (P_O + 3*(P_P) + 2) >> 2;

    ///////////////////////////////////
    // make vertical left prediction
    ///////////////////////////////////
    img->mprr_3[VERT_LEFT_PRED][0][0] = (P_A + P_B + 1) >> 1;
    img->mprr_3[VERT_LEFT_PRED][0][1] = 
      img->mprr_3[VERT_LEFT_PRED][2][0] = (P_B + P_C + 1) >> 1;
    img->mprr_3[VERT_LEFT_PRED][0][2] = 
      img->mprr_3[VERT_LEFT_PRED][2][1] = 
      img->mprr_3[VERT_LEFT_PRED][4][0] = (P_C + P_D + 1) >> 1;
    img->mprr_3[VERT_LEFT_PRED][0][3] = 
      img->mprr_3[VERT_LEFT_PRED][2][2] = 
      img->mprr_3[VERT_LEFT_PRED][4][1] = 
      img->mprr_3[VERT_LEFT_PRED][6][0] = (P_D + P_E + 1) >> 1;
    img->mprr_3[VERT_LEFT_PRED][0][4] = 
      img->mprr_3[VERT_LEFT_PRED][2][3] = 
      img->mprr_3[VERT_LEFT_PRED][4][2] = 
      img->mprr_3[VERT_LEFT_PRED][6][1] = (P_E + P_F + 1) >> 1;
    img->mprr_3[VERT_LEFT_PRED][0][5] = 
      img->mprr_3[VERT_LEFT_PRED][2][4] = 
      img->mprr_3[VERT_LEFT_PRED][4][3] = 
      img->mprr_3[VERT_LEFT_PRED][6][2] = (P_F + P_G + 1) >> 1;
    img->mprr_3[VERT_LEFT_PRED][0][6] = 
      img->mprr_3[VERT_LEFT_PRED][2][5] = 
      img->mprr_3[VERT_LEFT_PRED][4][4] = 
      img->mprr_3[VERT_LEFT_PRED][6][3] = (P_G + P_H + 1) >> 1;
    img->mprr_3[VERT_LEFT_PRED][0][7] = 
      img->mprr_3[VERT_LEFT_PRED][2][6] = 
      img->mprr_3[VERT_LEFT_PRED][4][5] = 
      img->mprr_3[VERT_LEFT_PRED][6][4] = (P_H + P_I + 1) >> 1;
    img->mprr_3[VERT_LEFT_PRED][2][7] = 
      img->mprr_3[VERT_LEFT_PRED][4][6] = 
      img->mprr_3[VERT_LEFT_PRED][6][5] = (P_I + P_J + 1) >> 1;
    img->mprr_3[VERT_LEFT_PRED][4][7] = 
      img->mprr_3[VERT_LEFT_PRED][6][6] = (P_J + P_K + 1) >> 1;
    img->mprr_3[VERT_LEFT_PRED][6][7] = (P_K + P_L + 1) >> 1;
    img->mprr_3[VERT_LEFT_PRED][1][0] = (P_A + P_C + 2*P_B + 2) >> 2;
    img->mprr_3[VERT_LEFT_PRED][1][1] = 
      img->mprr_3[VERT_LEFT_PRED][3][0] = (P_B + P_D + 2*P_C + 2) >> 2;
    img->mprr_3[VERT_LEFT_PRED][1][2] = 
      img->mprr_3[VERT_LEFT_PRED][3][1] = 
      img->mprr_3[VERT_LEFT_PRED][5][0] = (P_C + P_E + 2*P_D + 2) >> 2;
    img->mprr_3[VERT_LEFT_PRED][1][3] = 
      img->mprr_3[VERT_LEFT_PRED][3][2] = 
      img->mprr_3[VERT_LEFT_PRED][5][1] = 
      img->mprr_3[VERT_LEFT_PRED][7][0] = (P_D + P_F + 2*P_E + 2) >> 2;
    img->mprr_3[VERT_LEFT_PRED][1][4] = 
      img->mprr_3[VERT_LEFT_PRED][3][3] = 
      img->mprr_3[VERT_LEFT_PRED][5][2] = 
      img->mprr_3[VERT_LEFT_PRED][7][1] = (P_E + P_G + 2*P_F + 2) >> 2;
    img->mprr_3[VERT_LEFT_PRED][1][5] = 
      img->mprr_3[VERT_LEFT_PRED][3][4] = 
      img->mprr_3[VERT_LEFT_PRED][5][3] = 
      img->mprr_3[VERT_LEFT_PRED][7][2] = (P_F + P_H + 2*P_G + 2) >> 2;
    img->mprr_3[VERT_LEFT_PRED][1][6] = 
      img->mprr_3[VERT_LEFT_PRED][3][5] = 
      img->mprr_3[VERT_LEFT_PRED][5][4] = 
      img->mprr_3[VERT_LEFT_PRED][7][3] = (P_G + P_I + 2*P_H + 2) >> 2;
    img->mprr_3[VERT_LEFT_PRED][1][7] = 
      img->mprr_3[VERT_LEFT_PRED][3][6] = 
      img->mprr_3[VERT_LEFT_PRED][5][5] = 
      img->mprr_3[VERT_LEFT_PRED][7][4] = (P_H + P_J + 2*P_I + 2) >> 2;
    img->mprr_3[VERT_LEFT_PRED][3][7] = 
      img->mprr_3[VERT_LEFT_PRED][5][6] = 
      img->mprr_3[VERT_LEFT_PRED][7][5] = (P_I + P_K + 2*P_J + 2) >> 2;
    img->mprr_3[VERT_LEFT_PRED][5][7] = 
      img->mprr_3[VERT_LEFT_PRED][7][6] = (P_J + P_L + 2*P_K + 2) >> 2;
    img->mprr_3[VERT_LEFT_PRED][7][7] = (P_K + P_M + 2*P_L + 2) >> 2;
  }

  ///////////////////////////////////
  // make diagonal down right prediction
  ///////////////////////////////////
  if (block_available_up && block_available_left && block_available_up_left) 
  {
    // Mode DIAG_DOWN_RIGHT_PRED
    img->mprr_3[DIAG_DOWN_RIGHT_PRED][7][0] = (P_Z + P_V + 2*(P_W) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_RIGHT_PRED][6][0] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][7][1] = (P_W + P_U + 2*(P_V) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_RIGHT_PRED][5][0] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][6][1] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][7][2] = (P_V + P_T + 2*(P_U) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_RIGHT_PRED][4][0] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][5][1] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][6][2] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][7][3] = (P_U + P_S + 2*(P_T) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_RIGHT_PRED][3][0] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][4][1] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][5][2] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][6][3] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][7][4] = (P_T + P_R + 2*(P_S) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_RIGHT_PRED][2][0] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][3][1] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][4][2] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][5][3] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][6][4] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][7][5] = (P_S + P_Q + 2*(P_R) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_RIGHT_PRED][1][0] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][2][1] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][3][2] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][4][3] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][5][4] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][6][5] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][7][6] = (P_R + P_X + 2*(P_Q) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_RIGHT_PRED][0][0] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][1][1] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][2][2] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][3][3] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][4][4] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][5][5] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][6][6] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][7][7] = (P_Q + P_A + 2*(P_X) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_RIGHT_PRED][0][1] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][1][2] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][2][3] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][3][4] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][4][5] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][5][6] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][6][7] = (P_X + P_B + 2*(P_A) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_RIGHT_PRED][0][2] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][1][3] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][2][4] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][3][5] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][4][6] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][5][7] = (P_A + P_C + 2*(P_B) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_RIGHT_PRED][0][3] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][1][4] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][2][5] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][3][6] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][4][7] = (P_B + P_D + 2*(P_C) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_RIGHT_PRED][0][4] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][1][5] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][2][6] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][3][7] = (P_C + P_E + 2*(P_D) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_RIGHT_PRED][0][5] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][1][6] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][2][7] = (P_D + P_F + 2*(P_E) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_RIGHT_PRED][0][6] = 
      img->mprr_3[DIAG_DOWN_RIGHT_PRED][1][7] = (P_E + P_G + 2*(P_F) + 2) >> 2;
    img->mprr_3[DIAG_DOWN_RIGHT_PRED][0][7] = (P_F + P_H + 2*(P_G) + 2) >> 2;

    ///////////////////////////////////
    // make vertical right prediction
    ///////////////////////////////////
    img->mprr_3[VERT_RIGHT_PRED][0][0] = 
      img->mprr_3[VERT_RIGHT_PRED][2][1] = 
      img->mprr_3[VERT_RIGHT_PRED][4][2] = 
      img->mprr_3[VERT_RIGHT_PRED][6][3] = (P_X + P_A + 1) >> 1;
    img->mprr_3[VERT_RIGHT_PRED][0][1] = 
      img->mprr_3[VERT_RIGHT_PRED][2][2] = 
      img->mprr_3[VERT_RIGHT_PRED][4][3] = 
      img->mprr_3[VERT_RIGHT_PRED][6][4] = (P_A + P_B + 1) >> 1;
    img->mprr_3[VERT_RIGHT_PRED][0][2] = 
      img->mprr_3[VERT_RIGHT_PRED][2][3] = 
      img->mprr_3[VERT_RIGHT_PRED][4][4] = 
      img->mprr_3[VERT_RIGHT_PRED][6][5] = (P_B + P_C + 1) >> 1;
    img->mprr_3[VERT_RIGHT_PRED][0][3] = 
      img->mprr_3[VERT_RIGHT_PRED][2][4] = 
      img->mprr_3[VERT_RIGHT_PRED][4][5] = 
      img->mprr_3[VERT_RIGHT_PRED][6][6] = (P_C + P_D + 1) >> 1;
    img->mprr_3[VERT_RIGHT_PRED][0][4] = 
      img->mprr_3[VERT_RIGHT_PRED][2][5] = 
      img->mprr_3[VERT_RIGHT_PRED][4][6] = 
      img->mprr_3[VERT_RIGHT_PRED][6][7] = (P_D + P_E + 1) >> 1;
    img->mprr_3[VERT_RIGHT_PRED][0][5] = 
      img->mprr_3[VERT_RIGHT_PRED][2][6] = 
      img->mprr_3[VERT_RIGHT_PRED][4][7] = (P_E + P_F + 1) >> 1;
    img->mprr_3[VERT_RIGHT_PRED][0][6] = 
      img->mprr_3[VERT_RIGHT_PRED][2][7] = (P_F + P_G + 1) >> 1;
    img->mprr_3[VERT_RIGHT_PRED][0][7] = (P_G + P_H + 1) >> 1;
    img->mprr_3[VERT_RIGHT_PRED][1][0] = 
      img->mprr_3[VERT_RIGHT_PRED][3][1] = 
      img->mprr_3[VERT_RIGHT_PRED][5][2] = 
      img->mprr_3[VERT_RIGHT_PRED][7][3] = (P_Q + P_A + 2*P_X + 2) >> 2;
    img->mprr_3[VERT_RIGHT_PRED][1][1] = 
      img->mprr_3[VERT_RIGHT_PRED][3][2] = 
      img->mprr_3[VERT_RIGHT_PRED][5][3] = 
      img->mprr_3[VERT_RIGHT_PRED][7][4] = (P_X + P_B + 2*P_A + 2) >> 2;
    img->mprr_3[VERT_RIGHT_PRED][1][2] = 
      img->mprr_3[VERT_RIGHT_PRED][3][3] = 
      img->mprr_3[VERT_RIGHT_PRED][5][4] = 
      img->mprr_3[VERT_RIGHT_PRED][7][5] = (P_A + P_C + 2*P_B + 2) >> 2;
    img->mprr_3[VERT_RIGHT_PRED][1][3] = 
      img->mprr_3[VERT_RIGHT_PRED][3][4] = 
      img->mprr_3[VERT_RIGHT_PRED][5][5] = 
      img->mprr_3[VERT_RIGHT_PRED][7][6] = (P_B + P_D + 2*P_C + 2) >> 2;
    img->mprr_3[VERT_RIGHT_PRED][1][4] = 
      img->mprr_3[VERT_RIGHT_PRED][3][5] = 
      img->mprr_3[VERT_RIGHT_PRED][5][6] = 
      img->mprr_3[VERT_RIGHT_PRED][7][7] = (P_C + P_E + 2*P_D + 2) >> 2;
    img->mprr_3[VERT_RIGHT_PRED][1][5] = 
      img->mprr_3[VERT_RIGHT_PRED][3][6] = 
      img->mprr_3[VERT_RIGHT_PRED][5][7] = (P_D + P_F + 2*P_E + 2) >> 2;
    img->mprr_3[VERT_RIGHT_PRED][1][6] = 
      img->mprr_3[VERT_RIGHT_PRED][3][7] = (P_E + P_G + 2*P_F + 2) >> 2;
    img->mprr_3[VERT_RIGHT_PRED][1][7] = (P_F + P_H + 2*P_G + 2) >> 2;
    img->mprr_3[VERT_RIGHT_PRED][2][0] =
      img->mprr_3[VERT_RIGHT_PRED][4][1] =
      img->mprr_3[VERT_RIGHT_PRED][6][2] = (P_R + P_X + 2*P_Q + 2) >> 2;
    img->mprr_3[VERT_RIGHT_PRED][3][0] =
      img->mprr_3[VERT_RIGHT_PRED][5][1] =
      img->mprr_3[VERT_RIGHT_PRED][7][2] = (P_S + P_Q + 2*P_R + 2) >> 2;
    img->mprr_3[VERT_RIGHT_PRED][4][0] =
      img->mprr_3[VERT_RIGHT_PRED][6][1] = (P_T + P_R + 2*P_S + 2) >> 2;
    img->mprr_3[VERT_RIGHT_PRED][5][0] =
      img->mprr_3[VERT_RIGHT_PRED][7][1] = (P_U + P_S + 2*P_T + 2) >> 2;
    img->mprr_3[VERT_RIGHT_PRED][6][0] = (P_V + P_T + 2*P_U + 2) >> 2;
    img->mprr_3[VERT_RIGHT_PRED][7][0] = (P_W + P_U + 2*P_V + 2) >> 2;

    ///////////////////////////////////
    // make horizontal down prediction
    ///////////////////////////////////

    img->mprr_3[HOR_DOWN_PRED][0][0] = 
      img->mprr_3[HOR_DOWN_PRED][1][2] = 
      img->mprr_3[HOR_DOWN_PRED][2][4] = 
      img->mprr_3[HOR_DOWN_PRED][3][6] = (P_Q + P_X + 1) >> 1;
    img->mprr_3[HOR_DOWN_PRED][1][0] = 
      img->mprr_3[HOR_DOWN_PRED][2][2] = 
      img->mprr_3[HOR_DOWN_PRED][3][4] = 
      img->mprr_3[HOR_DOWN_PRED][4][6] = (P_R + P_Q + 1) >> 1;
    img->mprr_3[HOR_DOWN_PRED][2][0] = 
      img->mprr_3[HOR_DOWN_PRED][3][2] = 
      img->mprr_3[HOR_DOWN_PRED][4][4] = 
      img->mprr_3[HOR_DOWN_PRED][5][6] = (P_S + P_R + 1) >> 1;
    img->mprr_3[HOR_DOWN_PRED][3][0] = 
      img->mprr_3[HOR_DOWN_PRED][4][2] = 
      img->mprr_3[HOR_DOWN_PRED][5][4] = 
      img->mprr_3[HOR_DOWN_PRED][6][6] = (P_T + P_S + 1) >> 1;
    img->mprr_3[HOR_DOWN_PRED][4][0] = 
      img->mprr_3[HOR_DOWN_PRED][5][2] = 
      img->mprr_3[HOR_DOWN_PRED][6][4] = 
      img->mprr_3[HOR_DOWN_PRED][7][6] = (P_U + P_T + 1) >> 1;
    img->mprr_3[HOR_DOWN_PRED][5][0] = 
      img->mprr_3[HOR_DOWN_PRED][6][2] = 
      img->mprr_3[HOR_DOWN_PRED][7][4] = (P_V + P_U + 1) >> 1;
    img->mprr_3[HOR_DOWN_PRED][6][0] = 
      img->mprr_3[HOR_DOWN_PRED][7][2] = (P_W + P_V + 1) >> 1;
    img->mprr_3[HOR_DOWN_PRED][7][0] = (P_Z + P_W + 1) >> 1;
    img->mprr_3[HOR_DOWN_PRED][0][1] =
      img->mprr_3[HOR_DOWN_PRED][1][3] =
      img->mprr_3[HOR_DOWN_PRED][2][5] =
      img->mprr_3[HOR_DOWN_PRED][3][7] = (P_Q + P_A + 2*P_X + 2) >> 2;
    img->mprr_3[HOR_DOWN_PRED][1][1] =
      img->mprr_3[HOR_DOWN_PRED][2][3] =
      img->mprr_3[HOR_DOWN_PRED][3][5] =
      img->mprr_3[HOR_DOWN_PRED][4][7] = (P_X + P_R + 2*P_Q + 2) >> 2;
    img->mprr_3[HOR_DOWN_PRED][2][1] =
      img->mprr_3[HOR_DOWN_PRED][3][3] =
      img->mprr_3[HOR_DOWN_PRED][4][5] =
      img->mprr_3[HOR_DOWN_PRED][5][7] = (P_Q + P_S + 2*P_R + 2) >> 2;
    img->mprr_3[HOR_DOWN_PRED][3][1] =
      img->mprr_3[HOR_DOWN_PRED][4][3] =
      img->mprr_3[HOR_DOWN_PRED][5][5] =
      img->mprr_3[HOR_DOWN_PRED][6][7] = (P_R + P_T + 2*P_S + 2) >> 2;
    img->mprr_3[HOR_DOWN_PRED][4][1] =
      img->mprr_3[HOR_DOWN_PRED][5][3] =
      img->mprr_3[HOR_DOWN_PRED][6][5] =
      img->mprr_3[HOR_DOWN_PRED][7][7] = (P_S + P_U + 2*P_T + 2) >> 2;
    img->mprr_3[HOR_DOWN_PRED][5][1] =
      img->mprr_3[HOR_DOWN_PRED][6][3] =
      img->mprr_3[HOR_DOWN_PRED][7][5] = (P_T + P_V + 2*P_U + 2) >> 2;
    img->mprr_3[HOR_DOWN_PRED][6][1] =
      img->mprr_3[HOR_DOWN_PRED][7][3] = (P_U + P_W + 2*P_V + 2) >> 2;
    img->mprr_3[HOR_DOWN_PRED][7][1] = (P_V + P_Z + 2*P_W + 2) >> 2;
    img->mprr_3[HOR_DOWN_PRED][0][2] = 
      img->mprr_3[HOR_DOWN_PRED][1][4] = 
      img->mprr_3[HOR_DOWN_PRED][2][6] = (P_X + P_B + 2*P_A + 2) >> 2;
    img->mprr_3[HOR_DOWN_PRED][0][3] = 
      img->mprr_3[HOR_DOWN_PRED][1][5] = 
      img->mprr_3[HOR_DOWN_PRED][2][7] = (P_A + P_C + 2*P_B + 2) >> 2;
    img->mprr_3[HOR_DOWN_PRED][0][4] = 
      img->mprr_3[HOR_DOWN_PRED][1][6] = (P_B + P_D + 2*P_C + 2) >> 2;
    img->mprr_3[HOR_DOWN_PRED][0][5] = 
      img->mprr_3[HOR_DOWN_PRED][1][7] = (P_C + P_E + 2*P_D + 2) >> 2;
    img->mprr_3[HOR_DOWN_PRED][0][6] = (P_D + P_F + 2*P_E + 2) >> 2;
    img->mprr_3[HOR_DOWN_PRED][0][7] = (P_E + P_G + 2*P_F + 2) >> 2;
  }

  ///////////////////////////////////
  // make horizontal up prediction
  ///////////////////////////////////
  if (block_available_left)
  {
    img->mprr_3[HOR_UP_PRED][0][0] = (P_Q + P_R + 1) >> 1;
    img->mprr_3[HOR_UP_PRED][1][0] =
      img->mprr_3[HOR_UP_PRED][0][2] = (P_R + P_S + 1) >> 1;
    img->mprr_3[HOR_UP_PRED][2][0] =
      img->mprr_3[HOR_UP_PRED][1][2] =
      img->mprr_3[HOR_UP_PRED][0][4] = (P_S + P_T + 1) >> 1;
    img->mprr_3[HOR_UP_PRED][3][0] =
      img->mprr_3[HOR_UP_PRED][2][2] =
      img->mprr_3[HOR_UP_PRED][1][4] =
      img->mprr_3[HOR_UP_PRED][0][6] = (P_T + P_U + 1) >> 1;
    img->mprr_3[HOR_UP_PRED][4][0] =
      img->mprr_3[HOR_UP_PRED][3][2] =
      img->mprr_3[HOR_UP_PRED][2][4] =
      img->mprr_3[HOR_UP_PRED][1][6] = (P_U + P_V + 1) >> 1;
    img->mprr_3[HOR_UP_PRED][5][0] =
      img->mprr_3[HOR_UP_PRED][4][2] =
      img->mprr_3[HOR_UP_PRED][3][4] =
      img->mprr_3[HOR_UP_PRED][2][6] = (P_V + P_W + 1) >> 1;
    img->mprr_3[HOR_UP_PRED][6][0] =
      img->mprr_3[HOR_UP_PRED][5][2] =
      img->mprr_3[HOR_UP_PRED][4][4] =
      img->mprr_3[HOR_UP_PRED][3][6] = (P_W + P_Z + 1) >> 1;
    img->mprr_3[HOR_UP_PRED][4][6] =
      img->mprr_3[HOR_UP_PRED][4][7] =
      img->mprr_3[HOR_UP_PRED][5][4] =
      img->mprr_3[HOR_UP_PRED][5][5] =
      img->mprr_3[HOR_UP_PRED][5][6] =
      img->mprr_3[HOR_UP_PRED][5][7] =
      img->mprr_3[HOR_UP_PRED][6][2] =
      img->mprr_3[HOR_UP_PRED][6][3] =
      img->mprr_3[HOR_UP_PRED][6][4] =
      img->mprr_3[HOR_UP_PRED][6][5] =
      img->mprr_3[HOR_UP_PRED][6][6] =
      img->mprr_3[HOR_UP_PRED][6][7] =
      img->mprr_3[HOR_UP_PRED][7][0] =
      img->mprr_3[HOR_UP_PRED][7][1] =
      img->mprr_3[HOR_UP_PRED][7][2] =
      img->mprr_3[HOR_UP_PRED][7][3] =
      img->mprr_3[HOR_UP_PRED][7][4] =
      img->mprr_3[HOR_UP_PRED][7][5] =
      img->mprr_3[HOR_UP_PRED][7][6] =
      img->mprr_3[HOR_UP_PRED][7][7] = P_Z;
    img->mprr_3[HOR_UP_PRED][6][1] =
      img->mprr_3[HOR_UP_PRED][5][3] =
      img->mprr_3[HOR_UP_PRED][4][5] =
      img->mprr_3[HOR_UP_PRED][3][7] = (P_W + 3*P_Z + 2) >> 2;
    img->mprr_3[HOR_UP_PRED][5][1] =
      img->mprr_3[HOR_UP_PRED][4][3] =
      img->mprr_3[HOR_UP_PRED][3][5] =
      img->mprr_3[HOR_UP_PRED][2][7] = (P_Z + P_V + 2*P_W + 2) >> 2;
    img->mprr_3[HOR_UP_PRED][4][1] =
      img->mprr_3[HOR_UP_PRED][3][3] =
      img->mprr_3[HOR_UP_PRED][2][5] =
      img->mprr_3[HOR_UP_PRED][1][7] = (P_W + P_U + 2*P_V + 2) >> 2;
    img->mprr_3[HOR_UP_PRED][3][1] =
      img->mprr_3[HOR_UP_PRED][2][3] =
      img->mprr_3[HOR_UP_PRED][1][5] =
      img->mprr_3[HOR_UP_PRED][0][7] = (P_V + P_T + 2*P_U + 2) >> 2;
    img->mprr_3[HOR_UP_PRED][2][1] =
      img->mprr_3[HOR_UP_PRED][1][3] =
      img->mprr_3[HOR_UP_PRED][0][5] = (P_U + P_S + 2*P_T + 2) >> 2;
    img->mprr_3[HOR_UP_PRED][1][1] =
      img->mprr_3[HOR_UP_PRED][0][3] = (P_T + P_R + 2*P_S + 2) >> 2;
    img->mprr_3[HOR_UP_PRED][0][1] = (P_S + P_Q + 2*P_R + 2) >> 2;
  }
}


/*!
************************************************************************
* \brief
*    16x16 based luma prediction
*
* \par Input:
*    Image parameters
*
* \par Output:
*    none
************************************************************************
*/
void intrapred_luma_16x16()
{
  int s0=0,s1,s2;
  imgpel s[2][16];
  int i,j;

  int ih,iv;
  int ib,ic,iaa;

  imgpel   **imgY_pred = enc_picture->imgY;  // For Mb level field/frame coding tools -- default to frame pred
  int          mb_nr = img->current_mb_nr;

  PixelPos up;          //!< pixel position p(0,-1)
  PixelPos left[17];    //!< pixel positions p(-1, -1..15)

  int up_avail, left_avail, left_up_avail;

  for (i=0;i<17;i++)
  {
    getNeighbour(mb_nr, -1,  i-1, 1, &left[i]);
  }

  getNeighbour(mb_nr,    0,   -1, 1, &up);

  if (!(input->UseConstrainedIntraPred))
  {
    up_avail      = up.available;
    left_avail    = left[1].available;
    left_up_avail = left[0].available;
  }
  else
  {
    up_avail      = up.available ? img->intra_block[up.mb_addr] : 0;
    for (i=1, left_avail=1; i<17;i++)
      left_avail  &= left[i].available ? img->intra_block[left[i].mb_addr]: 0;
    left_up_avail = left[0].available ? img->intra_block[left[0].mb_addr]: 0;
  }

  s1=s2=0;
  // make DC prediction
  if (up_avail)
  {
    for (i=0; i < MB_BLOCK_SIZE; i++)
      s1 += imgY_pred[up.pos_y][up.pos_x+i];    // sum hor pix
  }

  if (left_avail)
  {
    for (i=0; i < MB_BLOCK_SIZE; i++)      
      s2 += imgY_pred[left[i+1].pos_y][left[i+1].pos_x];    // sum vert pix
  }

  if (up_avail && left_avail)
    s0=(s1+s2+16)/(2*MB_BLOCK_SIZE);             // no edge

  if (!up_avail && left_avail)
    s0=(s2+8)/MB_BLOCK_SIZE;                     // upper edge

  if (up_avail && !left_avail)
    s0=(s1+8)/MB_BLOCK_SIZE;                     // left edge

  if (!up_avail && !left_avail)
    s0=img->dc_pred_value_luma;                       // top left corner, nothing to predict from

  // vertical prediction
  if (up_avail)
    memcpy(s[0], &imgY_pred[up.pos_y][up.pos_x], MB_BLOCK_SIZE * sizeof(imgpel));

  // horizontal prediction
  if (left_avail)
  {
    for (i=0; i < MB_BLOCK_SIZE; i++)
      s[1][i]=imgY_pred[left[i+1].pos_y][left[i+1].pos_x];
  }

  for (j=0; j < MB_BLOCK_SIZE; j++)
  {
    memcpy(img->mprr_2[VERT_PRED_16][j], s[0], MB_BLOCK_SIZE * sizeof(imgpel)); // store vertical prediction
    for (i=0; i < MB_BLOCK_SIZE; i++)
    {
      img->mprr_2[HOR_PRED_16 ][j][i]=s[1][j]; // store horizontal prediction
      img->mprr_2[DC_PRED_16  ][j][i]=s0;      // store DC prediction
    }
  }
  if (!up_avail || !left_avail || !left_up_avail) // edge
    return;

  // 16 bit integer plan pred

  ih=0;
  iv=0;
  for (i=1;i<9;i++)
  {
    if (i<8)
      ih += i*(imgY_pred[up.pos_y][up.pos_x+7+i] - imgY_pred[up.pos_y][up.pos_x+7-i]);
    else
      ih += i*(imgY_pred[up.pos_y][up.pos_x+7+i] - imgY_pred[left[0].pos_y][left[0].pos_x]);

    iv += i*(imgY_pred[left[8+i].pos_y][left[8+i].pos_x] - imgY_pred[left[8-i].pos_y][left[8-i].pos_x]);
  }
  ib=(5*ih+32)>>6;
  ic=(5*iv+32)>>6;

  iaa=16*(imgY_pred[up.pos_y][up.pos_x+15]+imgY_pred[left[16].pos_y][left[16].pos_x]);

  for (j=0;j< MB_BLOCK_SIZE;j++)
  {
    for (i=0;i< MB_BLOCK_SIZE;i++)
    {
      img->mprr_2[PLANE_16][j][i]=max(0,min((int)img->max_imgpel_value,(iaa+(i-7)*ib +(j-7)*ic + 16)/32));// store plane prediction
    }
  }
}


/*!
*************************************************************************************
* \brief
*    4x4 Intra mode decision for an macroblock
*************************************************************************************
*/
int Mode_Decision_for_Intra4x4Macroblock (double lambda,  int* cost)
{
  int  cbp=0, b8, cost8x8;

  for (*cost=0, b8=0; b8<4; b8++)
  {
    if (Mode_Decision_for_8x8IntraBlocks (b8, lambda, &cost8x8))
    {
      cbp |= (1<<b8);
    }
    *cost += cost8x8;
  }

  return cbp;
}
/*!
*************************************************************************************
* \brief
*    Mode Decision for an 8x8 Intra block
*************************************************************************************
*/
int Mode_Decision_for_8x8IntraBlocks(int b8,double lambda,int *cost)
{
  int  nonzero=0, b4;
  int  cost4x4;

  *cost = (int)floor(6.0 * lambda + 0.4999);

  for (b4=0; b4<4; b4++)
  {
    if (Mode_Decision_for_4x4IntraBlocks (b8, b4, lambda, &cost4x4))
    {
      nonzero        = 1;
    }
    *cost += cost4x4;
  }
#ifdef RESET_STATE
  reset_coding_state (cs_cm);
#endif

  return nonzero;
}

/*!
*************************************************************************************
* \brief
*    Mode Decision for an 4x4 Intra block
*************************************************************************************
*/
int Mode_Decision_for_4x4IntraBlocks (int  b8,  int  b4,  double  lambda,  int*  min_cost)
{
  int     ipmode, best_ipmode = 0, i, j, y;
  int     c_nz, nonzero = 0;
  imgpel  rec4x4[4][4];
  double  rdcost;
  int     block_x     = 8*(b8 & 0x01)+4*(b4 & 0x01);
  int     block_y     = 8*(b8 >> 1)+4*(b4 >> 1);
  int     block_x4    = block_x >> 2;
  int     block_y4    = block_y >> 2;
  int     pic_pix_x   = img->pix_x+block_x;
  int     pic_pix_y   = img->pix_y+block_y;
  int     pic_opix_x   = img->opix_x+block_x;
  int     pic_opix_y   = img->opix_y+block_y;
  int     pic_block_x = pic_pix_x >> 2;
  int     pic_block_y = pic_pix_y >> 2;
  double  min_rdcost  = 1e30;

  int left_available, up_available, all_available;

  char   upMode;
  char   leftMode;
  int     mostProbableMode;

  PixelPos left_block;
  PixelPos top_block;

  getLuma4x4Neighbour(img->current_mb_nr, block_x4, block_y4, -1,  0, &left_block);
  getLuma4x4Neighbour(img->current_mb_nr, block_x4, block_y4,  0, -1, &top_block);

  // constrained intra pred
  if (input->UseConstrainedIntraPred)
  {
    left_block.available = left_block.available ? img->intra_block[left_block.mb_addr] : 0;
    top_block.available  = top_block.available  ? img->intra_block[top_block.mb_addr]  : 0;
  }

  upMode            =  top_block.available ? img->ipredmode[top_block.pos_y ][top_block.pos_x ] : -1;
  leftMode          = left_block.available ? img->ipredmode[left_block.pos_y][left_block.pos_x] : -1;

  mostProbableMode  = (upMode < 0 || leftMode < 0) ? DC_PRED : upMode < leftMode ? upMode : leftMode;

  *min_cost = INT_MAX;

  //===== INTRA PREDICTION FOR 4x4 BLOCK =====
  intrapred_luma (pic_pix_x, pic_pix_y, &left_available, &up_available, &all_available);

  //===== LOOP OVER ALL 4x4 INTRA PREDICTION MODES =====
  for (ipmode=0; ipmode<NO_INTRA_PMODE; ipmode++)
  {
    int available_mode =  (ipmode==DC_PRED) ||
      ((ipmode==VERT_PRED||ipmode==VERT_LEFT_PRED||ipmode==DIAG_DOWN_LEFT_PRED) && up_available ) ||
      ((ipmode==HOR_PRED||ipmode==HOR_UP_PRED) && left_available ) ||(all_available);

    if (input->IntraDisableInterOnly==0 || img->type != I_SLICE)
    {
      if (input->Intra4x4ParDisable && (ipmode==VERT_PRED||ipmode==HOR_PRED))
        continue;      

      if (input->Intra4x4DiagDisable && (ipmode==DIAG_DOWN_LEFT_PRED||ipmode==DIAG_DOWN_RIGHT_PRED))
        continue;      

      if (input->Intra4x4DirDisable && ipmode>=VERT_RIGHT_PRED)
        continue;
    }

    if( available_mode)
    {
      {
        // get prediction and prediction error
        for (j=0; j<4; j++)
        {
          memcpy(&img->mpr[block_y+j][block_x], img->mprr[ipmode][j], BLOCK_SIZE * sizeof(imgpel));
          for (i=0; i<4; i++)
          {
            img->m7[j][i] = (int) (imgY_org[pic_opix_y+j][pic_opix_x+i] - img->mprr[ipmode][j][i]);
          }
        }

        //===== store the coding state =====
        //store_coding_state (cs_cm);
        // get and check rate-distortion cost
#ifdef BEST_NZ_COEFF
        img->mb_data[img->current_mb_nr].cbp_bits = cbp_bits;
#endif
        if ((rdcost = RDCost_for_4x4IntraBlocks (&c_nz, b8, b4, ipmode, lambda, min_rdcost, mostProbableMode)) < min_rdcost)
        {
          //--- set coefficients ---
          memcpy(cofAC4x4[0],img->cofAC[b8][b4][0], 18 * sizeof(int));
          memcpy(cofAC4x4[1],img->cofAC[b8][b4][1], 18 * sizeof(int));

          //--- set reconstruction ---
          for (y=0; y<4; y++)
          {
            memcpy(rec4x4[y],&enc_picture->imgY[pic_pix_y+y][pic_pix_x], BLOCK_SIZE * sizeof(imgpel));
          } 
          //--- flag if dct-coefficients must be coded ---
          nonzero = c_nz;

          //--- set best mode update minimum cost ---
          min_rdcost    = rdcost;
          best_ipmode   = ipmode;
#ifdef BEST_NZ_COEFF
          best_nz_coeff = img->nz_coeff [img->current_mb_nr][block_x4][block_y4];
          best_coded_block_flag = (int)((img->mb_data[img->current_mb_nr].cbp_bits>>bit_pos)&(int64)(1));
#endif            
          //store_coding_state (cs_ib4);
        }

#ifndef RESET_STATE
        reset_coding_state (cs_cm);
#endif
      }
    }
  }


#ifdef BEST_NZ_COEFF
  img->nz_coeff [img->current_mb_nr][block_x4][block_y4] = best_nz_coeff;
  cbp_bits &= (~(int64)(1<<bit_pos));
  cbp_bits |= (int64)(best_coded_block_flag<<bit_pos);
#endif

  //===== set intra mode prediction =====
  img->ipredmode[pic_block_y][pic_block_x] = best_ipmode;
  img->mb_data[img->current_mb_nr].intra_pred_modes[4*b8+b4] = mostProbableMode == best_ipmode ? -1 : best_ipmode < mostProbableMode ? best_ipmode : best_ipmode-1;

  {
    //===== restore coefficients =====
    for (j=0; j<2; j++)
    {
      memcpy (img->cofAC[b8][b4][j],cofAC4x4[j], 18 * sizeof(int));
    } 

    //===== restore reconstruction and prediction (needed if single coeffs are removed) =====
    for (y=0; y<BLOCK_SIZE; y++)
    {
      memcpy (&enc_picture->imgY[pic_pix_y+y][pic_pix_x],rec4x4[y],    BLOCK_SIZE * sizeof(imgpel));
      memcpy (&img->mpr[block_y+y][block_x],img->mprr[best_ipmode][y], BLOCK_SIZE * sizeof(imgpel));
    }
  }  
  return nonzero;
}




/*!
*************************************************************************************
* \brief
*    R-D Cost for an 4x4 Intra block
*************************************************************************************
*/
double RDCost_for_4x4IntraBlocks (int*    nonzero,
                                  int     b8,
                                  int     b4,
                                  int     ipmode,
                                  double  lambda,
                                  double  min_rdcost,
                                  int mostProbableMode)
{
  double  rdcost;
  int     dummy, rate;
  int64   distortion  = 0;
  int     block_x     = 8*(b8 & 0x01)+4*(b4 & 0x01);
  int     block_y     = 8*(b8 >> 1)+4*(b4 >> 1);
  int     pic_pix_x   = img->pix_x+block_x;
  int     pic_pix_y   = img->pix_y+block_y;
  int     pic_opix_y  = img->opix_y+block_y;
  imgpel  **imgY      = enc_picture->imgY;

  Slice          *currSlice    =  img->currentSlice;
  Macroblock     *currMB       = &img->mb_data[img->current_mb_nr];
  SyntaxElement  *currSE       = &img->MB_SyntaxElements[currMB->currSEnr];
  const int      *partMap      = assignSE2partition[0];
  DataPartition  *dataPart;

  //===== perform DCT, Q, IQ, IDCT, Reconstruction =====
  dummy = 0;
#ifdef USE_INTRA_MDDT
  if(input->UseIntraMDDT)
  {
    *nonzero = (ipmode == 2) ? 
      dct_luma (block_x, block_y, &dummy, 1, ipmode): 
    klt_luma_sep (block_x, block_y, &dummy, ipmode);
  }
  else 
    *nonzero = dct_luma (block_x, block_y, &dummy, 1, ipmode);
#else
  *nonzero = dct_luma (block_x, block_y, &dummy, 1);
#endif

  //===== get distortion (SSD) of 4x4 block =====
  distortion += compute_SSE(imgY_org, imgY , pic_opix_y, pic_pix_y, pic_pix_x, pic_pix_x, 4, 4);
  if(distortion > min_rdcost)
  {
    return (double)distortion;
  }

  //===== RATE for INTRA PREDICTION MODE  (SYMBOL MODE MUST BE SET TO UVLC) =====
  currSE->value1 = (mostProbableMode == ipmode) ? -1 : ipmode < mostProbableMode ? ipmode : ipmode-1;

  //--- set position and type ---
  currSE->context = 4*b8 + b4;
  currSE->type    = SE_INTRAPREDMODE;

  //--- choose data partition ---
  dataPart = &(currSlice->partArr[partMap[SE_INTRAPREDMODE]]);
  //--- encode and update rate ---
  currSE->writing = writeIntraPredMode_CABAC;
  dataPart->writeSyntaxElement (currSE, dataPart);
  rate = currSE->len;
  currSE++;
  currMB->currSEnr++;

  //===== RATE for LUMINANCE COEFFICIENTS =====
  rate  += writeLumaCoeff4x4_CABAC (b8, b4, 1);
  //reset_coding_state (cs_cm);
  rdcost = (double)distortion + lambda*(double)rate;

  return rdcost;
}



/*! 
*************************************************************************************
* \brief
*    8x8 Intra mode decision for a macroblock
*************************************************************************************
*/

int Mode_Decision_for_new_Intra8x8Macroblock (double lambda, int *min_cost)
{
  int  cbp=0, b8, cost8x8;

  *min_cost = (int)floor(6.0 * lambda + 0.4999);

  for (b8=0; b8<4; b8++)
  {
    if (Mode_Decision_for_new_8x8IntraBlocks (b8, lambda, &cost8x8))
    {
      cbp |= (1<<b8);
    }
    *min_cost += cost8x8;
  }

  return cbp;
}

/*! 
*************************************************************************************
* \brief
*    8x8 Intra mode decision for a macroblock
*************************************************************************************
*/

int Mode_Decision_for_new_8x8IntraBlocks (int b8, double lambda, int *min_cost)
{
  int     ipmode, best_ipmode = 0, i, j, k, y;
  int     c_nz, nonzero = 0;
  imgpel  rec8x8[8][8];
  double  rdcost = 0.0;
  int     block_x     = 8*(b8 & 0x01);
  int     block_y     = 8*(b8 >> 1);
  int     pic_pix_x   = img->pix_x+block_x;
  int     pic_pix_y   = img->pix_y+block_y;
  int     pic_opix_x   = img->opix_x+block_x;
  int     pic_opix_y   = img->opix_y+block_y;
  int     pic_block_x = pic_pix_x/4;
  int     pic_block_y = pic_pix_y/4;
  double  min_rdcost  = 1e30;
  imgpel    **imgY_orig  = imgY_org;
  extern  int ****cofAC8x8; 
  int left_available, up_available, all_available;
  char   upMode;
  char   leftMode;
  int     mostProbableMode;  

  PixelPos left_block;
  PixelPos top_block;

  Macroblock     *currMB       = &img->mb_data[img->current_mb_nr];

  getLuma4x4Neighbour(img->current_mb_nr, block_x/4, block_y/4, -1,  0, &left_block);
  getLuma4x4Neighbour(img->current_mb_nr, block_x/4, block_y/4,  0, -1, &top_block);

  if (input->UseConstrainedIntraPred)
  {
    top_block.available  = top_block.available ? img->intra_block [top_block.mb_addr] : 0;
    left_block.available = left_block.available ? img->intra_block [left_block.mb_addr] : 0;
  }

  if(b8 >> 1)
    upMode    =  top_block.available ? img->ipredmode8x8[top_block.pos_y ][top_block.pos_x ] : -1; 
  else
    upMode    =  top_block.available ? img->ipredmode   [top_block.pos_y ][top_block.pos_x ] : -1;
  if(b8 & 0x01)
    leftMode  = left_block.available ? img->ipredmode8x8[left_block.pos_y][left_block.pos_x] : -1;
  else
    leftMode  = left_block.available ? img->ipredmode[left_block.pos_y][left_block.pos_x] : -1;

  mostProbableMode  = (upMode < 0 || leftMode < 0) ? DC_PRED : upMode < leftMode ? upMode : leftMode;

  *min_cost = INT_MAX;

  //===== INTRA PREDICTION FOR 8x8 BLOCK =====
  intrapred_luma_8x8 (pic_pix_x, pic_pix_y, &left_available, &up_available, &all_available);

  //===== LOOP OVER ALL 8x8 INTRA PREDICTION MODES =====
  for (ipmode=0; ipmode<NO_INTRA_PMODE; ipmode++)
  {
    if( (ipmode==DC_PRED) ||
      ((ipmode==VERT_PRED||ipmode==VERT_LEFT_PRED||ipmode==DIAG_DOWN_LEFT_PRED) && up_available ) ||
      ((ipmode==HOR_PRED||ipmode==HOR_UP_PRED) && left_available ) ||
      (all_available) )
    {
      {
        // get prediction and prediction error
        for (j=0; j<8; j++)
        {
          memcpy(&img->mpr[block_y+j][block_x],img->mprr_3[ipmode][j], 8 * sizeof(imgpel));
          for (i=0; i<8; i++)
          {
            img->m7[j][i] = imgY_orig[pic_opix_y+j][pic_opix_x+i] - img->mprr_3[ipmode][j][i];
          }
        }

        if ((rdcost = RDCost_for_8x8IntraBlocks (&c_nz, b8, ipmode, lambda, min_rdcost, mostProbableMode)) < min_rdcost)
        {
          //--- set coefficients ---
          for(k=0; k<4; k++) // do 4x now
          {
            for (j=0; j<2; j++)
              memcpy(cofAC8x8[b8][k][j],img->cofAC[b8][k][j], 65 * sizeof(int));
          }

          //--- set reconstruction ---
          for (y=0; y<8; y++)
          {
            memcpy(rec8x8[y],&enc_picture->imgY[pic_pix_y+y][pic_pix_x], 8 * sizeof(imgpel));
          }

          //--- flag if dct-coefficients must be coded ---
          nonzero = c_nz;

          //--- set best mode update minimum cost ---
          min_rdcost  = rdcost;
          best_ipmode = ipmode;
        }
        reset_coding_state_cs_cm();  
      }
    }
  }

  //===== set intra mode prediction =====
  img->ipredmode8x8[pic_block_y][pic_block_x] = best_ipmode;
  currMB->intra_pred_modes8x8[4*b8] = (mostProbableMode == best_ipmode) 
    ? -1 
    : (best_ipmode < mostProbableMode ? best_ipmode : best_ipmode-1);

  for(j = img->mb_y*4+(b8 >> 1)*2; j < img->mb_y*4+(b8 >> 1)*2 + 2; j++)   //loop 4x4s in the subblock for 8x8 prediction setting
    memset(&img->ipredmode8x8[j][img->mb_x*4+(b8 & 0x01)*2], best_ipmode, 2 * sizeof(char));

  {
    //===== restore coefficients =====
    for(k=0; k<4; k++) // do 4x now    
    {
      for (j=0; j<2; j++)
        memcpy(img->cofAC[b8][k][j],cofAC8x8[b8][k][j], 65 * sizeof(int));
    }

    //===== restore reconstruction and prediction (needed if single coeffs are removed) =====
    for (y=0; y<8; y++) 
    {
      memcpy(&enc_picture->imgY[pic_pix_y+y][pic_pix_x], rec8x8[y], 8 * sizeof(imgpel));
      memcpy(&  img->mpr[block_y+y][block_x], img->mprr_3[best_ipmode][y], 8 * sizeof(imgpel));
    }
  }
  //if(final_mb_encoding && b8==0)
  //{
  //  int ii, jj;
  //  for(jj=0; jj<8; jj++)
  //  {
  //    for(ii=0; ii<8; ii++)
  //      printf("%d ", enc_picture->imgY[pic_pix_y+jj][pic_pix_x+ii]);
  //    printf("\n");
  //  }
  //}
  return nonzero;
}

/*! 
*************************************************************************************
* \brief
*    Prefiltering for Intra8x8 prediction
*************************************************************************************
*/
void LowPassForIntra8x8Pred(int *PredPel, int block_up_left, int block_up, int block_left)
{
  int i;
  int LoopArray[25];


  for(i = 0; i < 25; i++)
    LoopArray[i] = PredPel[i] ;

  if(block_up)
  {
    if(block_up_left) 
    {
      LoopArray[1] = ((&P_X)[0] + ((&P_X)[1]<<1) + (&P_X)[2] + 2)>>2;
    }
    else
      LoopArray[1] = ((&P_X)[1] + ((&P_X)[1]<<1) + (&P_X)[2] + 2)>>2; 


    for(i = 2; i <16; i++)
    {
      LoopArray[i] = ((&P_X)[i-1] + ((&P_X)[i]<<1) + (&P_X)[i+1] + 2)>>2;
    }
    LoopArray[16] = (P_P + (P_P<<1) + P_O + 2)>>2;
  }

  if(block_up_left) 
  {

    if(block_up && block_left)
    {
      LoopArray[0] = (P_Q + (P_X<<1) + P_A +2)>>2;
    }
    else
    {
      if(block_up)
        LoopArray[0] = (P_X + (P_X<<1) + P_A +2)>>2;
      else
        if(block_left)
          LoopArray[0] = (P_X + (P_X<<1) + P_Q +2)>>2;
    }

  }

  if(block_left)
  {
    if(block_up_left)
      LoopArray[17] = (P_X + (P_Q<<1) + P_R + 2)>>2; 
    else
      LoopArray[17] = (P_Q + (P_Q<<1) + P_R + 2)>>2;

    for(i = 18; i <24; i++)
    {
      LoopArray[i] = ((&P_X)[i-1] + ((&P_X)[i]<<1) + (&P_X)[i+1] + 2)>>2;
    }
    LoopArray[24] = (P_W + (P_Z<<1) + P_Z + 2)>>2;
  }

  for(i = 0; i < 25; i++)
    PredPel[i] = LoopArray[i];
}






/*! 
*************************************************************************************
* \brief
*    R-D Cost for an 8x8 Intra block
*************************************************************************************
*/
double RDCost_for_8x8IntraBlocks(int *nonzero, int b8, int ipmode, double lambda, double min_rdcost, int mostProbableMode)
{
  double  rdcost = 0.0;
  int     dummy, rate;
  int64   distortion  = 0;
  int     block_x     = 8*(b8 & 0x01);
  int     block_y     = 8*(b8 >> 1);
  int     pic_pix_x   = img->pix_x+block_x;
  int     pic_pix_y   = img->pix_y+block_y;
  int     pic_opix_y  = img->opix_y+block_y;
  imgpel    **imgY_orig  = imgY_org;
  imgpel    **imgY       = enc_picture->imgY;

  Slice          *currSlice    =  img->currentSlice;
  Macroblock     *currMB       = &img->mb_data[img->current_mb_nr];
  SyntaxElement  *currSE       = &img->MB_SyntaxElements[currMB->currSEnr];
  const int      *partMap      = assignSE2partition[0];
  DataPartition  *dataPart;

  //===== perform DCT, Q, IQ, IDCT, Reconstruction =====
  dummy = 0;

#ifdef USE_INTRA_MDDT
  if(input->UseIntraMDDT)
  {
    *nonzero = (ipmode == 2) ? 
      dct_luma8x8 (b8, &dummy, 1, ipmode) : 
    klt_luma8x8_sep_fast (b8, &dummy, ipmode);
  }
  else
    *nonzero = dct_luma8x8 (b8, &dummy, 1, ipmode) ; 
#else
  *nonzero = dct_luma8x8 (b8, &dummy, 1);
#endif

  //===== get distortion (SSD) of 8x8 block =====
  distortion += compute_SSE(imgY_orig, imgY, pic_opix_y, pic_pix_y, pic_pix_x, pic_pix_x, 8, 8);
  if(distortion > min_rdcost)
  {
    return (double)distortion;
  }

  //===== RATE for INTRA PREDICTION MODE  (SYMBOL MODE MUST BE SET TO UVLC) =====
  currSE->value1 = (mostProbableMode == ipmode) ? -1 : ipmode < mostProbableMode ? ipmode : ipmode-1;

  //--- set position and type ---
  currSE->context = b8;
  currSE->type    = SE_INTRAPREDMODE;

  //--- choose data partition ---
  if (img->type!=B_SLICE)
    dataPart = &(currSlice->partArr[partMap[SE_INTRAPREDMODE]]);
  else
    dataPart = &(currSlice->partArr[partMap[SE_BFRAME]]);

  //--- encode and update rate ---
  currSE->writing = writeIntraPredMode_CABAC;
  dataPart->writeSyntaxElement (currSE, dataPart);

  rate = currSE->len;
  currSE++;
  currMB->currSEnr++;

  //===== RATE for LUMINANCE COEFFICIENTS =====
  rate  += writeLumaCoeff8x8_CABAC (b8, 1);

  rdcost = (double)distortion + lambda*(double)rate;
  return rdcost;
}











#ifdef USE_INTRA_MDDT
/*! 
*************************************************************************************
* \brief
*   RD-optimized mode decision for INTRA16x16 mode
*
* \para Intra16x16_Mode_Decision_RDopt()
*    <paragraph>
*
* \author
*    - Yan Ye                      <yye@qualcomm.com>
*************************************************************************************
*/
void Intra16x16_Mode_Decision_RDopt (Macroblock* currMB, int* i16mode, double lambda)
{
  extern int MBType2Value (Macroblock* currMB);
  int i, j;

  int mode16, best_mode16;
  int max_mode = 4;
  double rdcost, min_rdcost = 1e30;
  int distortion, rate; 
  imgpel **imgY = enc_picture->imgY; 
  int y;
  extern CSptr cs_i16;
  SyntaxElement  *currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
  DataPartition*  dataPart;
  const int*      partMap    = assignSE2partition[0];
  Slice*          currSlice  = img->currentSlice;
  int  bestCofAC16[2][16*17];
  int  best_cbp=0;
  imgpel bestRec[MB_BLOCK_SIZE][MB_BLOCK_SIZE];

  intrapred_luma_16x16 ();   /* make intra pred for all 4 new modes */

  {
    PixelPos up;          //!< pixel position p(0,-1)
    PixelPos left[17];    //!< pixel positions p(-1, -1..15)

    int up_avail, left_avail, left_up_avail;

    for (i=0;i<17;i++)
    {
      getNeighbour(img->current_mb_nr, -1 ,  i-1 , 1, &left[i]);
    }

    getNeighbour(img->current_mb_nr, 0     ,  -1 , 1, &up);

    if (!(input->UseConstrainedIntraPred))
    {
      up_avail   = up.available;
      left_avail = left[1].available;
      left_up_avail = left[0].available;
    }
    else
    {
      up_avail      = up.available ? img->intra_block[up.mb_addr] : 0;
      for (i=1, left_avail=1; i<17;i++)
        left_avail  &= left[i].available ? img->intra_block[left[i].mb_addr]: 0;
      left_up_avail = left[0].available ? img->intra_block[left[0].mb_addr]: 0;
    }


    best_mode16 = DC_PRED_16;
    for(mode16 = 0; mode16 < max_mode; mode16++)
    {
      if (input->IntraDisableInterOnly == 0 || img->type != I_SLICE)
      {
        if (input->Intra16x16ParDisable && (mode16==VERT_PRED_16||mode16==HOR_PRED_16))
          continue;

        if (input->Intra16x16PlaneDisable && mode16==PLANE_16)
          continue;
      }

      //check if there are neighbours to predict from
      if ((mode16==VERT_PRED_16 && !up_avail  ) || 
        (mode16==HOR_PRED_16  && !left_avail) || 
        (mode16==PLANE_16     && (!left_avail || !up_avail || !left_up_avail)))
      {
        continue; // edge, do nothing
      }

      {
        currMB->cbp = klt_luma16x16_sep_fast (mode16); 
        //===== get distortion (SSD) of 4x4 block =====
        distortion = 0;
        distortion += (int)compute_SSE(imgY_org, imgY, img->opix_y, img->pix_y, img->opix_x, img->pix_x, 16, 16);
        if(distortion > min_rdcost)
        {
          continue;
        }

        store_coding_state (cs_i16);

        // rate for MB_TYPE 
        img->i16offset = I16Offset  (currMB->cbp, mode16);
        currMB->mb_type = I16MB; 
        currSE->value1  = MBType2Value (currMB);
        currSE->value2  = 0;
        currSE->type    = SE_MBTYPE;

        currSE->writing = writeMB_typeInfo_CABAC;
        dataPart = &(currSlice->partArr[partMap[SE_MBTYPE]]);
        dataPart->writeSyntaxElement( currSE, dataPart);
        rate                = currSE->len;
        currSE++;
        currMB->currSEnr++;

        //===== RATE for LUMINANCE COEFFICIENTS =====

        rate += writeCBP_Transform_size (1, 0);
        rate += writeDeltaQP            (1, img->mb_data[img->current_mb_nr].cbp);
        rate += writeLumaCoeff          (1, 0);
        reset_coding_state (cs_i16);

        rdcost = (double)distortion + lambda*(double)rate;
        if(rdcost < min_rdcost)
        {
          best_mode16 = mode16;
          min_rdcost = rdcost;
          best_cbp = currMB->cbp;
          memcpy(bestCofAC16[0], img->cofAC16[0], sizeof(int)*16*17);
          memcpy(bestCofAC16[1], img->cofAC16[1], sizeof(int)*16*17);
          for(j = 0; j < MB_BLOCK_SIZE; j++)
            memcpy(bestRec[j], &imgY[img->pix_y+j][img->pix_x], sizeof(imgpel)*MB_BLOCK_SIZE);
        }
      }
    }

    *i16mode = best_mode16;
  }

  currMB->cbp = best_cbp;
  if(currMB->cbp)
    currMB->cbp = 15;
  memcpy(img->cofAC16[0], bestCofAC16[0], sizeof(int)*16*17);
  memcpy(img->cofAC16[1], bestCofAC16[1], sizeof(int)*16*17);
  for(y = 0; y < MB_BLOCK_SIZE; y++)
  {
    memcpy(&imgY[img->pix_y+y][img->pix_x], &bestRec[y][0], MB_BLOCK_SIZE*sizeof(imgpel));
  }
}
#endif

/*!
*************************************************************************************
* \brief
*    Intra 16x16 mode decision
*************************************************************************************
*/
void Intra16x16_Mode_Decision (Macroblock* currMB, int* i16mode)
{
  intrapred_luma_16x16 ();   /* make intra pred for all 4 new modes */
  find_sad_16x16 (i16mode);   /* get best new intra mode */
  currMB->cbp = dct_luma_16x16 (*i16mode);
}
