/* 
====================================================================================================================

The copyright in this software is being made available under the License included below. 
This software may be subject to other third party and 	contributor rights, including patent rights, and no such 
rights are granted under this license.

Copyright (c) 2010, QUALCOMM INC.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted only for 
the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
promoting such standards. The following conditions are required to be met:

* Redistributions of source code must retain the above copyright notice, this list of conditions and 
the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
the following disclaimer in the documentation and/or other materials provided with the distribution.
* The name of QUALCOMM INC.  may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

====================================================================================================================
*/

#include <stdlib.h>
#include <assert.h>
#include <limits.h>
#include <memory.h>

#include "global.h"

#include "elements.h"
#include "macroblock.h"
#include "refbuf.h"
#include "mb_access.h"
#include "vlc.h"
#include "image.h"
#include "mb_access.h"
#include "cabac.h"

#ifdef MV_COMPETITION
#include "mv_competition.h"
extern int skip_mode;
extern int ******send_index_mv_prediction;
extern MV_Competition mv_comp;
#endif

extern const byte QP_SCALE_CR[52] ;


//Rate control
extern int DELTA_QP,DELTA_QP2;
extern int QP,QP2;

int writeDeltaQP(int rdopt, int cbp);
int writeCBP_Transform_size (int rdopt, int mb32);
int writeLumaCoeff (int rdopt, int mb32);

void CheckAvailabilityOfNeighbors_Z(void);
int write_mb_type(Macroblock *MB, int mb_ext_level);
void writeCBP_CABAC32(SyntaxElement *se, EncodingEnvironmentPtr eep_dp);
void writeCBP_CABAC_chroma(SyntaxElement *se, EncodingEnvironmentPtr eep_dp);
void writeCBP_CABAC_luma1bit(SyntaxElement *se, EncodingEnvironmentPtr eep_dp);
extern Macroblock  MB32;
extern int mb32_mvd[2][BLOCK_MULTIPLE<<MAX_MB_EXT_LEVEL][BLOCK_MULTIPLE<<MAX_MB_EXT_LEVEL][2];          //!< indices correspond to [forw,backw][block_y][block_x][x,y]
extern int delta_qp_sent;
extern int eos_bit_32;
extern int only_16x16_transform;
extern Macroblock  MB64;
extern int   cbp_32[1<<(2*(MAX_MB_EXT_LEVEL-1))], cbp_32_RDCost[1<<(2*(MAX_MB_EXT_LEVEL-1))];

#ifdef RDO_Q
void SetMotionVectorPredictor (short  pmv[2], char   **refPic, short  ***tmp_mv, short  ref_frame, int list, int block_x, int block_y, int blockshape_x, int blockshape_y, int mb_ext_level);
#endif

#ifdef GEOM
int writeMB_geoInfo_VLC(SyntaxElement *se, struct datapartition *dataPart);
extern int coeff_cost4x4[4][4];
extern int rho_bits[3];//[mb_ext_level]
extern int geo_stat[2][3];
extern int rho_bits[3];//[mb_ext_level]
extern char ****geo_mask[3];
extern char **mask;
extern int best8x8pdir_geo[GEO_PART_NUM][2];

void writeMB_geoInfo_CABAC (SyntaxElement *se, EncodingEnvironmentPtr eep_dp);
int MBType2Value_geo (Macroblock* currMB);
int writeMotionInfo2NAL_geo (int geo_mode, int mb_ext_level);
void writeMB_geoInfo_CABAC (SyntaxElement *se, EncodingEnvironmentPtr eep_dp);
#endif

int writelumaCBPTransform_size(int BigBlock);
int write_chroma_cbp();
int writeLumaCoefBigBlock(Macroblock *currMB, int mb32, int rdopt);

#ifdef ADAPTIVE_MV_RES
short roundInt(short mv,int mvres)
{
  short roundMV=mv, sign;
  sign=mv>0? 1 : -1;
  roundMV*=sign;
  if (mvres ==1)
    roundMV=roundMV/2;
  else if (mvres ==2)
    roundMV=roundMV/4;
  roundMV*=sign;
  return(roundMV);
}
#endif
/*!
************************************************************************
* \brief
*    updates the coordinates for the next macroblock to be processed
*
* \param mb_addr
*    macroblock address in scan order
************************************************************************
*/
void set_MB_parameters (int mb_addr)
{
  img->current_mb_nr = mb_addr;

  get_mb_block_pos(mb_addr, &img->mb_x, &img->mb_y);

  img->block_x = img->mb_x << 2;
  img->block_y = img->mb_y << 2;

  img->pix_x   = img->block_x << 2;
  img->pix_y   = img->block_y << 2;

  img->opix_x   = img->pix_x;
  img->opix_y   = img->block_y << 2;
  img->mb_data[mb_addr].list_offset = 0;


  if (img->yuv_format != YUV400)
  {
    img->pix_c_x = (img->mb_cr_size_x * img->pix_x) >> 4;
    img->pix_c_y = (img->mb_cr_size_y * img->pix_y) >> 4;

    img->opix_c_x = (img->mb_cr_size_x * img->opix_x) >> 4;
    img->opix_c_y = (img->mb_cr_size_y * img->opix_y) >> 4;
  }
  //  printf ("set_MB_parameters: mb %d,  mb_x %d,  mb_y %d\n", mb_addr, img->mb_x, img->mb_y);
}


/*!
************************************************************************
* \brief
*    clip to range [0..max luma pel value]
************************************************************************
*/
int clip1a(int a)
{
  return ((a)>img->max_imgpel_value?img->max_imgpel_value:((a)<0?0:(a)));
}

/*!
************************************************************************
* \brief
*    updates the coordinates and statistics parameter for the
*    next macroblock
************************************************************************
*/
void proceed2nextMacroblock(void)
{
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];
  int*        bitCount = currMB->bitcounter;
  int i;

  if (bitCount[BITS_TOTAL_MB] > 128 + 256 * img->bitdepth_luma + 2 * img->mb_cr_size_y * img->mb_cr_size_x * img->bitdepth_chroma)
    printf("Warning!!! Number of bits (%d) of macroblock_layer() data seems to exceed defined limit.\n", bitCount[BITS_TOTAL_MB]);

  // Update the statistics
  stats->bit_use_mb_type[img->type]      += bitCount[BITS_MB_MODE];
  stats->bit_use_coeffY[img->type]       += bitCount[BITS_COEFF_Y_MB] ;
  stats->tmp_bit_use_cbp[img->type]      += bitCount[BITS_CBP_MB];
  stats->bit_use_coeffC[img->type]       += bitCount[BITS_COEFF_UV_MB];
  stats->bit_use_delta_quant[img->type]  += bitCount[BITS_DELTA_QUANT_MB];


  if (IS_INTRA(currMB))
  {
    ++stats->intra_chroma_mode[currMB->c_ipred_mode];

    if ((currMB->cbp&15) != 0)
    {
      if (currMB->luma_transform_size_8x8_flag)
        ++stats->mode_use_transform_8x8[img->type][currMB->mb_type];
      else
        ++stats->mode_use_transform_4x4[img->type][currMB->mb_type];
    }
  }

  ++stats->mode_use[img->type][currMB->mb_type];

#ifdef MV_COMPETITION
  if (input->mv_competition > 0)
  {
    if(currMB->mb_type == 0)
    {
      if(!currMB->send_index_predictor_skip)
        ++stats->predictor_for_skip_trans[img->type];
      else
        ++stats->predictor_for_skip_use[img->type][currMB->best_predictor_for_skip];
    }
    stats->bit_use_pred_for_skip[img->type]+= bitCount[BITS_SKIP_PRED];

    if(img->type == P_SLICE)
      stats->bit_use_pred_for_mv_inter[img->type]+= bitCount[BITS_MV_PRED];
  }
#endif

  stats->bit_use_mode[img->type][currMB->mb_type]+= bitCount[BITS_INTER_MB];

  if (img->type != I_SLICE)
  {
    if (currMB->mb_type == P8x8)
    {
      for(i=0;i<4;i++)
      {
        if (currMB->b8mode[i] > 0)
          ++stats->mode_use[img->type][currMB->b8mode[i]];
        else
          ++stats->b8_mode_0_use[img->type][currMB->luma_transform_size_8x8_flag];

        if (currMB->b8mode[i]==4)
        {
          if ((currMB->luma_transform_size_8x8_flag && (currMB->cbp&15) != 0) || input->Transform8x8Mode == 2)
            ++stats->mode_use_transform_8x8[img->type][4];
          else
            ++stats->mode_use_transform_4x4[img->type][4];
        }
      }
    }
    else if (currMB->mb_type >= 0 && currMB->mb_type <=3 && ((currMB->cbp&15) != 0))
    {
      if (currMB->luma_transform_size_8x8_flag)
        ++stats->mode_use_transform_8x8[img->type][currMB->mb_type];
      else
        ++stats->mode_use_transform_4x4[img->type][currMB->mb_type];
    }
  }

  // Statistics
  if ((img->type == P_SLICE))
  {
    ++stats->quant0;
    stats->quant1 += currMB->qp;      // to find average quant for inter frames
  }
}

void set_chroma_qp(Macroblock *currMB)
{
  int i;
  for (i=0; i<2; i++)
  {
    currMB->qpc[i] = Clip3 ( -img->bitdepth_chroma_qp_scale, 51, currMB->qp + img->chroma_qp_offset[i] );
    currMB->qpc[i] = currMB->qpc[i] < 0 ? currMB->qpc[i] : QP_SCALE_CR[currMB->qpc[i]];
  }
}

/*!
************************************************************************
* \brief
*    initializes the current macroblock
*
* \param mb_addr
*    macroblock address in scan order
************************************************************************
*/
void start_macroblock(int mb_addr)
{
  int i,j,l;
  int use_bitstream_backing = (input->slice_mode == FIXED_RATE || input->slice_mode == CALLBACK);
  Macroblock *currMB = &img->mb_data[mb_addr];
  Slice *curr_slice = img->currentSlice;
  DataPartition *dataPart;
  Bitstream *currStream;
  EncodingEnvironmentPtr eep;
  int prev_mb;

#ifdef GEOM
  currMB->geo_mode = -1;
#endif


  set_MB_parameters (mb_addr);

  if(use_bitstream_backing)
  {
    if ((mb_addr&1)==0) 
    {
      // Keep the current state of the bitstreams
      if(!img->cod_counter)
        for (i=0; i<curr_slice->max_part_nr; i++)
        {
          dataPart = &(curr_slice->partArr[i]);
          currStream = dataPart->bitstream;
          currStream->stored_bits_to_go = currStream->bits_to_go;
          currStream->stored_byte_pos   = currStream->byte_pos;
          currStream->stored_byte_buf   = currStream->byte_buf;

          eep = &(dataPart->ee_cabac);
          eep->ElowS            = eep->Elow;
          eep->ErangeS          = eep->Erange;
          eep->EbufferS         = eep->Ebuffer;
          eep->Ebits_to_goS     = eep->Ebits_to_go;
          eep->Ebits_to_followS = eep->Ebits_to_follow;
          eep->EcodestrmS       = eep->Ecodestrm;
          eep->Ecodestrm_lenS   = eep->Ecodestrm_len;
          eep->CS               = eep->C;
          eep->ES               = eep->E;
        }
    }
  }

  // Save the slice number of this macroblock. When the macroblock below
  // is coded it will use this to decide if prediction for above is possible
  currMB->slice_nr = img->current_slice_nr;
  prev_mb = GetPreviousMBNr(img->current_mb_nr);
  if(img->type==I_SLICE)
  {
    Slice* currSlice = img->currentSlice;
    if (prev_mb>-1)
    {
      currMB->prev_qp = img->mb_data[prev_mb].qp;
      if (img->mb_data[prev_mb].slice_nr == img->current_slice_nr)
        currMB->prev_delta_qp = img->mb_data[prev_mb].delta_qp;
      else
        currMB->prev_delta_qp = 0;
    }
    else
    {
      currMB->prev_qp = currSlice->qp;
      currMB->prev_delta_qp = 0;
    }
    currMB->qp = img->qp = img->masterQP;
  }
  else
  {
    if (prev_mb>-1)
    {
      currMB->prev_qp       = Prev_QP;       //global variable
      currMB->prev_delta_qp = Prev_Delta_QP;  //global variable
    }
    else
    {
      currMB->prev_qp       = img->masterQP;
      currMB->prev_delta_qp = 0;
    }
  }
  currMB->delta_qp = currMB->qp - currMB->prev_qp;
  set_chroma_qp (currMB);

  // Initialize counter for MB symbols
  currMB->currSEnr=0;

  // loop filter parameter
  if (active_pps->deblocking_filter_control_present_flag)
  {
    currMB->LFDisableIdc    = img->LFDisableIdc;
    currMB->LFAlphaC0Offset = img->LFAlphaC0Offset;
    currMB->LFBetaOffset    = img->LFBetaOffset;
  }
  else
  {
    currMB->LFDisableIdc    = 0;
    currMB->LFAlphaC0Offset = 0;
    currMB->LFBetaOffset    = 0;
  }

  // If MB is next to a slice boundary, mark neighboring blocks unavailable for prediction
  if(img->type != I_SLICE) 
    CheckAvailabilityOfNeighbors_Z();
  else
    CheckAvailabilityOfNeighbors();

  CheckAvailabilityOfNeighborsCABAC();

  // Reset vectors and reference indices before doing motion search in motion_search().
  for (l=0; l<2; l++)
  {
    for (j=img->block_y; j < img->block_y + BLOCK_MULTIPLE; j++)
    {
      memset(&enc_picture->ref_idx[l][j][img->block_x], -1, BLOCK_MULTIPLE * sizeof(char));     
      memset(enc_picture->mv      [l][j][img->block_x], 0, 2 * BLOCK_MULTIPLE * sizeof(short));
      for (i=img->block_x; i < img->block_x + BLOCK_MULTIPLE; i++)
        enc_picture->ref_pic_id[l][j][i]= -1;
    }
  }

  // Reset syntax element entries in MB struct
  currMB->mb_type      = 0;
  currMB->cbp_blk      = 0;
  currMB->cbp          = 0;
  currMB->cbp_bits     = 0;
  currMB->c_ipred_mode = DC_PRED_8;

  memset (currMB->mvd, 0, BLOCK_CONTEXT * sizeof(int));  
  memset (currMB->intra_pred_modes, DC_PRED, MB_BLOCK_PARTITIONS * sizeof(char)); // changing this to char would allow us to use memset
  memset (currMB->intra_pred_modes8x8, DC_PRED, MB_BLOCK_PARTITIONS * sizeof(char));

  //initialize the whole MB as INTRA coded
  //Blocks are set to notINTRA in write_one_macroblock
  if (input->UseConstrainedIntraPred)
  {
    img->intra_block[img->current_mb_nr] = 1;
  }

  // Initialize bitcounters for this macroblock
  if(prev_mb < 0) // No slice header to account for
  {
    currMB->bitcounter[BITS_HEADER] = 0;
  }
  else if (currMB->slice_nr == img->mb_data[prev_mb].slice_nr) // current MB belongs to the
    // same slice as the last MB
  {
    currMB->bitcounter[BITS_HEADER] = 0;
  }

  currMB->bitcounter[BITS_MB_MODE       ] = 0;
  currMB->bitcounter[BITS_COEFF_Y_MB    ] = 0;
  currMB->bitcounter[BITS_INTER_MB      ] = 0;
  currMB->bitcounter[BITS_CBP_MB        ] = 0;
  currMB->bitcounter[BITS_DELTA_QUANT_MB] = 0;
  currMB->bitcounter[BITS_COEFF_UV_MB   ] = 0;

#ifdef MV_COMPETITION
  currMB->bitcounter[BITS_SKIP_PRED     ] = 0;
  currMB->bitcounter[BITS_MV_PRED       ] = 0;
#endif

  //#ifdef _FAST_FULL_ME_
  //  if(!input->FMEnable)
  //    ResetFastFullIntegerSearch ();
  //#endif
}



/*!
************************************************************************
* \brief
*    terminates processing of the current macroblock depending
*    on the chosen slice mode
************************************************************************
*/
void terminate_macroblock( Boolean *end_of_slice,      //!< returns true for last macroblock of a slice, otherwise false
                          Boolean *recode_macroblock  //!< returns true if max. slice size is exceeded an macroblock must be recoded in next slice
                          )
{
  int i;
  Slice *currSlice = img->currentSlice;
  Macroblock    *currMB    = &img->mb_data[img->current_mb_nr];
  SyntaxElement *currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  int *partMap = assignSE2partition[0];
  DataPartition *dataPart;
  Bitstream *currStream;
  int rlc_bits=0;
  EncodingEnvironmentPtr eep;
  int new_slice;
  static int skip = FALSE;


  // if previous mb in the same slice group has different slice number as the current, it's the
  // the start of new slice
  new_slice=0;
  if ( (img->current_mb_nr==0) || (GetPreviousMBNr(img->current_mb_nr)<0) )
    new_slice=1;
  else if( img->mb_data[GetPreviousMBNr(img->current_mb_nr)].slice_nr != img->current_slice_nr )
    new_slice=1;

  *recode_macroblock=FALSE;

  switch(input->slice_mode)
  {
  case NO_SLICES:
    currSlice->num_mb++;
    *recode_macroblock = FALSE;
    if ((currSlice->num_mb) == (int)img->PicSizeInMbs) // maximum number of MBs reached
      *end_of_slice = TRUE;

    // if it's end of current slice group, slice ends too
    *end_of_slice |= (img->current_mb_nr == GetLastCodedMBOfSliceGroup(0));

    break;
  case FIXED_MB:
    // For slice mode one, check if a new slice boundary follows
    currSlice->num_mb++;
    *recode_macroblock = FALSE;
    //! Check end-of-slice group condition first
    *end_of_slice = (img->current_mb_nr == GetLastCodedMBOfSliceGroup(0));
    //! Now check maximum # of MBs in slice
    *end_of_slice |= (currSlice->num_mb >= input->slice_argument);

    break;

    // For slice modes two and three, check if coding of this macroblock
    // resulted in too many bits for this slice. If so, indicate slice
    // boundary before this macroblock and code the macroblock again
  case FIXED_RATE:
    // in case of skip MBs check if there is a slice boundary
    // only for UVLC (img->cod_counter is always 0 in case of CABAC)
    if(img->cod_counter)
    {
      // write out the skip MBs to know how many bits we need for the RLC
      currSE->value1 = img->cod_counter;
      currSE->value2 = 0;
      currSE->mapping = ue_linfo;
      currSE->type = SE_MBTYPE;
      dataPart = &(currSlice->partArr[partMap[currSE->type]]);

      dataPart->writeSyntaxElement(  currSE, dataPart);
      rlc_bits=currSE->len;

      currStream = dataPart->bitstream;
      // save the bitstream as it would be if we write the skip MBs
      currStream->bits_to_go_skip  = currStream->bits_to_go;
      currStream->byte_pos_skip    = currStream->byte_pos;
      currStream->byte_buf_skip    = currStream->byte_buf;
      // restore the bitstream
      currStream->bits_to_go = currStream->stored_bits_to_go;
      currStream->byte_pos = currStream->stored_byte_pos;
      currStream->byte_buf = currStream->stored_byte_buf;
      skip = TRUE;
    }
    //! Check if the last coded macroblock fits into the size of the slice
    //! But only if this is not the first macroblock of this slice
    if (!new_slice)
    {
      if(slice_too_big(rlc_bits))
      {
        *recode_macroblock = TRUE;
        *end_of_slice = TRUE;
      }
      else if(!img->cod_counter)
        skip = FALSE;
    }
    // maximum number of MBs

    // check if current slice group is finished
    if ((*recode_macroblock == FALSE) && (img->current_mb_nr == GetLastCodedMBOfSliceGroup(0))) 
    {
      *end_of_slice = TRUE;
      if(!img->cod_counter)
        skip = FALSE;
    }

    //! (first MB OR first MB in a slice) AND bigger that maximum size of slice
    if (new_slice && slice_too_big(rlc_bits))
    {
      *end_of_slice = TRUE;
      if(!img->cod_counter)
        skip = FALSE;
    }
    if (!*recode_macroblock)
      currSlice->num_mb++;
    break;

  case  CALLBACK:
    if (img->current_mb_nr > 0 && !new_slice)
    {
      if (currSlice->slice_too_big(rlc_bits))
      {
        *recode_macroblock = TRUE;
        *end_of_slice = TRUE;
      }
    }

    if ( (*recode_macroblock == FALSE) && (img->current_mb_nr == GetLastCodedMBOfSliceGroup(0))) 
      *end_of_slice = TRUE;
    break;

  default:
    snprintf(errortext, ET_SIZE, "Slice Mode %d not supported", input->slice_mode);
    error(errortext, 600);
  }

  if(*recode_macroblock == TRUE)
  {
    // Restore everything
    for (i=0; i<currSlice->max_part_nr; i++)
    {
      dataPart = &(currSlice->partArr[i]);
      currStream = dataPart->bitstream;
      currStream->bits_to_go = currStream->stored_bits_to_go;
      currStream->byte_pos  = currStream->stored_byte_pos;
      currStream->byte_buf  = currStream->stored_byte_buf;

      eep = &(dataPart->ee_cabac);
      eep->Elow            = eep->ElowS;
      eep->Erange          = eep->ErangeS;
      eep->Ebuffer         = eep->EbufferS;
      eep->Ebits_to_go     = eep->Ebits_to_goS;
      eep->Ebits_to_follow = eep->Ebits_to_followS;
      eep->Ecodestrm       = eep->EcodestrmS;
      eep->Ecodestrm_len   = eep->Ecodestrm_lenS;
      eep->C               = eep->CS;
      eep->E               = eep->ES;       
    }
  }
}

int slice_too_big(int rlc_bits)
{
  Slice *currSlice = img->currentSlice;
  DataPartition *dataPart;
  EncodingEnvironmentPtr eep;
  int i;

  for (i=0; i<currSlice->max_part_nr; i++)
  {
    dataPart= &(currSlice->partArr[i]);
    eep = &(dataPart->ee_cabac);

    if( arienco_bits_written(eep) > (input->slice_argument*8))
      return TRUE;
  }

  return FALSE;
}

int interp_block[16][16];
extern void get_UMVLine4X8_direct(StorablePicture *ref_picture, int x_pos, int y_pos, int block[16][16], int size_x, int size_y);

static pel_t *(*get_line) (pel_t**, int, int, int, int);
/*!
************************************************************************
* \brief
*    Predict one component of a 4x4 Luma block
************************************************************************
*/
void OneComponentLumaPrediction4x4 ( int*   mpred,          //!< array of prediction values (row by row)
                                    int    pic_pix_x,      //!< absolute horizontal coordinate of 4x4 block
                                    int    pic_pix_y,      //!< absolute vertical   coordinate of 4x4 block
                                    short* mv,             //!< motion vector
                                    short  ref,            //!< reference frame 
                                    StorablePicture **list //!< reference picture list
                                    )
{
  pel_t** ref_pic = list[ref]->imgY_ups;
  int img_width =((list[ref]->size_x + 2*IMG_PAD_SIZE - 1)<<2);
  int img_height=((list[ref]->size_y + 2*IMG_PAD_SIZE - 1)<<2);
  pel_t   *ref_line;
  int j;
  int i0,j0;
  int mv_shift= img->mv_res;

  int i, dx, dy;
  unsigned int *refP;

  int mv0, mv1;
  mv0 = mv[0];
  mv1 = mv[1];

#ifdef MV_COMPETITION
  if ((skip_mode== TRUE) &&  (input->mv_competition > 0))
  {
    mv0 = mv_comp.mv_pred_skip[mv_comp.current_predictor_for_skip][0];
    mv1 = mv_comp.mv_pred_skip[mv_comp.current_predictor_for_skip][1];
  }
#endif
  if (img->mv_res)
  {
    j0      = (pic_pix_y << 3) + mv1 + IMG_PAD_SIZE_TIMES4*2;
    i0      = (pic_pix_x << 3) + mv0 + IMG_PAD_SIZE_TIMES4*2;
  }
  else
  {
    j0      = (pic_pix_y << 2) + mv1 + IMG_PAD_SIZE_TIMES4;
    i0      = (pic_pix_x << 2) + mv0 + IMG_PAD_SIZE_TIMES4;
  }


  if (img->mv_res)
  {
    if ((i0 > 0) && (i0 < (((img_width - 4*BLOCK_SIZE)<< 1) - 1))  && (j0 > 0) && (j0 < (((img_height - 4*BLOCK_SIZE)<< 1) - 1)))
      get_line = FastLine4X8;
    else
      get_line = UMVLine4X8;
  }
  else
  {
    if ((i0 > 0) && (i0 < img_width - 4*BLOCK_SIZE - 1) && (j0 > 0) && (j0 < img_height - 4*BLOCK_SIZE - 1))
      get_line = FastLine4X;
    else
      get_line = UMVLine4X;
  }


  // Calculate the interpolated values for the entire 4x4 subblock.
  dx = i0&1;
  dy = j0&1;

  if ((dx || dy) && img->mv_res) // At least i0 or j0 is 1/8-pel and img->mv_res is non-zero
  {
    get_UMVLine4X8_direct(list[ref], (i0 - (IMG_PAD_SIZE<<3)), (j0 - (IMG_PAD_SIZE<<3)), interp_block, 4, 4);
    for (j=0; j < 4; j++)
    {
      refP = &interp_block[j][0];
      for (i=0; i < 4; i++)
      {
        *mpred++ = *refP++;
      }
    }
  }
  else

  {
    for (j = j0; j < j0 + (MB_BLOCK_SIZE<<mv_shift); j+= (BLOCK_SIZE<<mv_shift))
    {
      ref_line = get_line (ref_pic, j, i0, img_height, img_width);
      *mpred++ = *ref_line;
      *mpred++ = *(ref_line + 4);
      *mpred++ = *(ref_line + 8);
      *mpred++ = *(ref_line + 12);
    }
  }
}

/*!
************************************************************************
* \brief
*    Predict one 4x4 Luma block
************************************************************************
*/
void LumaPrediction4x4 ( int   block_x,    //!< relative horizontal block coordinate of 4x4 block
                        int   block_y,    //!< relative vertical   block coordinate of 4x4 block
                        int   p_dir,      //!< prediction direction (0=list0, 1=list1, 2=bipred)
                        int   fw_mode,    //!< list0 prediction mode (1-7, 0=DIRECT if bw_mode=0)
                        int   bw_mode,    //!< list1 prediction mode (1-7, 0=DIRECT if fw_mode=0)
                        short fw_ref_idx, //!< reference frame for list0 prediction (-1: Intra4x4 pred. with fw_mode)
                        short bw_ref_idx  //!< reference frame for list1 prediction 
                        )
{
  static int fw_pred[16];
  static int bw_pred[16];

  int  i, j;
  int  block_x4  = block_x+4;
  int  block_y4  = block_y+4;
  int  pic_opix_x = img->opix_x + block_x;
  int  pic_opix_y = img->opix_y + block_y;
  int  bx        = block_x >> 2;
  int  by        = block_y >> 2;
  int* fpred     = fw_pred;
  int* bpred     = bw_pred;
  Macroblock*    currMB     = &img->mb_data[img->current_mb_nr];

  short**** mv_array = img->all_mv[by][bx];

#ifdef  USE_HP_FILTER//BROUND
  img->bipred_rounding_control = (img->nal_reference_idc!=0);
#endif

  if (currMB->bi_pred_me && fw_ref_idx == 0 && bw_ref_idx == 0 && p_dir == 2 && fw_mode==1 && bw_mode==1)
    mv_array = currMB->bi_pred_me == 1? img->bipred_mv1[by][bx] : img->bipred_mv2[by][bx];

  if ((p_dir==0)||(p_dir==2))
    OneComponentLumaPrediction4x4 (fw_pred, pic_opix_x, pic_opix_y, mv_array[LIST_0][fw_ref_idx][fw_mode], fw_ref_idx, listX[0+currMB->list_offset]);   

  if ((p_dir==1)||(p_dir==2))
    OneComponentLumaPrediction4x4 (bw_pred, pic_opix_x, pic_opix_y, mv_array[LIST_1][bw_ref_idx][bw_mode], bw_ref_idx, listX[1+currMB->list_offset]);   


  if (p_dir==2)
  {
    for   (j=block_y; j<block_y4; j++)
      for (i=block_x; i<block_x4; i++)  
#ifdef  USE_HP_FILTER//BROUND
        if(input->UseHPFilter != 0)
        {
          img->mpr[j][i] = (*fpred++ + *bpred++ + 1 - img->bipred_rounding_control) >> 1; 
        }
        else
        {
          img->mpr[j][i] = (*fpred++ + *bpred++ + 1) >> 1; 
        }
#else
        img->mpr[j][i] = (*fpred++ + *bpred++ + 1) >> 1; 
#endif
  }
  else if (p_dir==0)
  {
    for   (j=block_y; j<block_y4; j++)
      for (i=block_x; i<block_x4; i++)  
        img->mpr[j][i] = *fpred++;
  }
  else // p_dir==1
  {
    for   (j=block_y; j<block_y4; j++)
      for (i=block_x; i<block_x4; i++)  
        img->mpr[j][i] = *bpred++;
  }

}

/*!
************************************************************************
* \brief
*    Predict one 4x4 Luma block
************************************************************************
*/
void LumaPrediction4x4Bi ( int   block_x,    //!< relative horizontal block coordinate of 4x4 block
                          int   block_y,    //!< relative vertical   block coordinate of 4x4 block
                          int   p_dir,      //!< prediction direction (0=list0, 1=list1, 2=bidir)
                          int   fw_mode,    //!< list0 prediction mode (1-7, 0=DIRECT if bw_mode=0)
                          int   bw_mode,    //!< list1 prediction mode (1-7, 0=DIRECT if fw_mode=0)
                          short fw_ref_idx, //!< reference frame for list0 prediction (-1: Intra4x4 pred. with fw_mode)
                          short bw_ref_idx, //!< reference frame for list1 prediction 
                          int   list        //!< current list for prediction.
                          )
{
  static int fw_pred[16];
  static int bw_pred[16];

  int  i, j;
  int  block_x4  = block_x+4;
  int  block_y4  = block_y+4;
  int  pic_opix_x = img->opix_x + block_x;
  int  pic_opix_y = img->opix_y + block_y;
  int  bx        = block_x >> 2;
  int  by        = block_y >> 2;
  int* fpred     = fw_pred;
  int* bpred     = bw_pred;
  Macroblock*    currMB     = &img->mb_data[img->current_mb_nr];

  short   ****mv_array = list ? img->bipred_mv1[by][bx] : img->bipred_mv2[by][bx];

#ifdef  USE_HP_FILTER//BROUND
  img->bipred_rounding_control = (img->nal_reference_idc!=0);
#endif

  if ((p_dir==0)||(p_dir==2))
    OneComponentLumaPrediction4x4 (fw_pred, pic_opix_x, pic_opix_y, mv_array[LIST_0][fw_ref_idx][fw_mode], fw_ref_idx, listX[0+currMB->list_offset]);   

  if ((p_dir==1)||(p_dir==2))
    OneComponentLumaPrediction4x4 (bw_pred, pic_opix_x, pic_opix_y, mv_array[LIST_1][bw_ref_idx][bw_mode], bw_ref_idx, listX[1+currMB->list_offset]);   

  if (p_dir==2)
  {
    for   (j=block_y; j<block_y4; j++)
      for (i=block_x; i<block_x4; i++)  
#ifdef  USE_HP_FILTER//BROUND
        if(input->UseHPFilter != 0)
        {
          img->mpr[j][i] = (*fpred++ + *bpred++ + 1 - img->bipred_rounding_control) / 2; 
        }
        else
        {
          img->mpr[j][i] = (*fpred++ + *bpred++ + 1) / 2; 
        }
#else
        img->mpr[j][i] = (*fpred++ + *bpred++ + 1) / 2; 
#endif
  }
  else if (p_dir==0)
  {
    for   (j=block_y; j<block_y4; j++)
      for (i=block_x; i<block_x4; i++)  
        img->mpr[j][i] = *fpred++;
  }
  else // p_dir==1
  {
    for   (j=block_y; j<block_y4; j++)
      for (i=block_x; i<block_x4; i++)  
        img->mpr[j][i] = *bpred++;
  }

}



/*!
************************************************************************
* \brief
*    Residual Coding of an 8x8 Luma block (not for intra)
*
* \return
*    coefficient cost
************************************************************************
*/
int LumaResidualCoding8x8 (int mbi_x, 
                           int mbi_y, 
                           int mode, 
                           int mode_idx,
                           int   *cbp,        //!< Output: cbp (updated according to processed 8x8 luminance block)
                           int64 *cbp_blk,    //!< Output: block cbp (updated according to processed 8x8 luminance block)
                           int   block8x8,    //!< block number of 8x8 block
                           short p_dir,       //!< prediction direction
                           int   fw_mode,     //!< list0 prediction mode (1-7, 0=DIRECT)
                           int   bw_mode,     //!< list1 prediction mode (1-7, 0=DIRECT)
                           short fw_refframe, //!< reference picture for list0 prediction
                           short bw_refframe  //!< reference picture for list0 prediction
                           )
{
  int    block_y, block_x, pic_pix_y, pic_pix_x, i, j, nonzero = 0, cbp_blk_mask;
  int    coeff_cost = 0;
  int    mb_y       = (block8x8 >> 1) << 3;
  int    mb_x       = (block8x8 & 0x01) << 3;
  //  int    pix_y;
  int    cbp_mask   = 1 << block8x8;
  int    bxx, byy;                   // indexing curr_blk

  Macroblock* currMB = &img->mb_data[img->current_mb_nr];
  //set transform size
  int    need_8x8_transform = currMB->luma_transform_size_8x8_flag;
  static imgpel backup_mpr[100][64][64];  

  int skipped    = 0; //to allow tranform for mode=0 in P slice when img->NoResidueDirect=0 to behave like Pdirect...

  //===== loop over 4x4 blocks =====
  for (byy=0, block_y=mb_y; block_y<mb_y+8; byy+=4, block_y+=4)
  {
    pic_pix_y = img->opix_y + block_y;
    for (bxx=0, block_x=mb_x; block_x<mb_x+8; bxx+=4, block_x+=4)
    {
      pic_pix_x = img->opix_x + block_x;
      cbp_blk_mask = (block_x>>2) + block_y;

      if(isfirstQP || (mode<=0))
      {
        //===== prediction of 4x4 block =====
        LumaPrediction4x4 (block_x, block_y, p_dir, fw_mode, bw_mode, fw_refframe, bw_refframe);
        if(mode==1 || mode == 2 || mode == 3 || mode == 4 || mode == 5 || mode==6 || mode ==7)
        {
          int mb64_x = mbi_x<<4;
          int mb64_y = mbi_y<<4;

          if(mode == 4 || mode == 5 || mode==6 || mode ==7)//for submacroblock_mode_decision
            mode_idx = mode+25;

          for(j=block_y; j<block_y+4; j++)
            for(i=block_x; i<block_x+4; i++)
              backup_mpr[mode_idx][mb64_y+j][mb64_x+i] = img->mpr[j][i];
        }
      }
      else 
      {//!isfirstQP && mode>0 && mode!=4
        int mb64_x = mbi_x<<4;
        int mb64_y = mbi_y<<4;

        if(mode == 4 || mode == 5 || mode==6 || mode ==7)//for submacroblock_mode_decision
          mode_idx = mode+25;

        for(j=block_y; j<block_y+4; j++)
          for(i=block_x; i<block_x+4; i++)
            img->mpr[j][i] = backup_mpr[mode_idx][mb64_y+j][mb64_x+i];

      }

      //===== get displaced frame difference ======
      if(!need_8x8_transform)
      {
        //===== get displaced frame difference ======                
        for (j=0; j<4; j++)
          for (i=0; i<4; i++)
            img->m7[j][i] = imgY_org[pic_pix_y + j][pic_pix_x + i] - img->mpr[j+block_y][i+block_x];

        //===== DCT, Quantization, inverse Quantization, IDCT, Reconstruction =====      
        if ( (img->NoResidueDirect != 1 && !skipped  ) || ((img->qp + img->bitdepth_luma_qp_scale)==0 && img->lossless_qpprime_flag==1) )
        {
          //===== DCT, Quantization, inverse Quantization, IDCT, Reconstruction =====
#ifdef USE_INTRA_MDDT
          nonzero = dct_luma   (block_x, block_y, &coeff_cost, 0, -1);
#else
          nonzero = dct_luma   (block_x, block_y, &coeff_cost, 0);
#endif
          if (nonzero)
          {
            (*cbp_blk) |= (int64)1 << cbp_blk_mask;  // one bit for every 4x4 block
            (*cbp)     |= cbp_mask;           // one bit for the 4x4 blocks of an 8x8 block
          }
        }
      }
      else    // store the 4x4 residual to build 8x8 residual block...and later 8x8 transform will be applied on 8x8 residual block
      {
        for (j=0; j<4; j++)
          for (i=0; i<4; i++)
            img->m7[j+byy][i+bxx] = imgY_org[pic_pix_y + j][pic_pix_x+i] - img->mpr[j+block_y][i+block_x];
      }
    }
  }//end===== loop over 4x4 blocks =====


  if(need_8x8_transform)
  {
    if (img->NoResidueDirect != 1 && !skipped)
    {
#ifdef USE_INTRA_MDDT
      nonzero = dct_luma8x8   (block8x8, &coeff_cost, 0, -1);
#else
      nonzero = dct_luma8x8   (block8x8, &coeff_cost, 0);
#endif

      if (nonzero)
      {
        (*cbp_blk) |= 51 << (4*block8x8-2*(block8x8 & 0x01)); // corresponds to 110011, as if all four 4x4 blocks contain coeff, shifted to block position
        (*cbp)     |= cbp_mask;           // one bit for the 4x4 blocks of an 8x8 block
      }
    }
  }

  /*
  The purpose of the action below is to prevent that single or 'expensive' coefficients are coded.
  With 4x4 transform there is larger chance that a single coefficient in a 8x8 or 16x16 block may be nonzero.
  A single small (level=1) coefficient in a 8x8 block will cost: 3 or more bits for the coefficient,
  4 bits for EOBs for the 4x4 blocks,possibly also more bits for CBP.  Hence the total 'cost' of that single
  coefficient will typically be 10-12 bits which in a RD consideration is too much to justify the distortion improvement.
  The action below is to watch such 'single' coefficients and set the reconstructed block equal to the prediction according
  to a given criterium.  The action is taken only for inter luma blocks.

  Notice that this is a pure encoder issue and hence does not have any implication on the standard.
  coeff_cost is a parameter set in dct_luma() and accumulated for each 8x8 block.  If level=1 for a coefficient,
  coeff_cost is increased by a number depending on RUN for that coefficient.The numbers are (see also dct_luma()): 3,2,2,1,1,1,0,0,...
  when RUN equals 0,1,2,3,4,5,6, etc.
  If level >1 coeff_cost is increased by 9 (or any number above 3). The threshold is set to 3. This means for example:
  1: If there is one coefficient with (RUN,level)=(0,1) in a 8x8 block this coefficient is discarded.
  2: If there are two coefficients with (RUN,level)=(1,1) and (4,1) the coefficients are also discarded
  sum_cnt_nonz is the accumulation of coeff_cost over a whole macro block.  If sum_cnt_nonz is 5 or less for the whole MB,
  all nonzero coefficients are discarded for the MB and the reconstructed block is set equal to the prediction.
  */

  if (img->NoResidueDirect != 1 && !skipped && coeff_cost <= _LUMA_COEFF_COST_ &&
    ((img->qp + img->bitdepth_luma_qp_scale)!=0 || img->lossless_qpprime_flag==0))
  {
    coeff_cost  = 0;
    (*cbp)     &=  (63 - cbp_mask);
    (*cbp_blk) &= ~(51 << (4*block8x8-2*(block8x8 & 0x01)));

    for (j=mb_y; j<mb_y+8; j++)
      memcpy(&enc_picture->imgY[img->pix_y + j][img->pix_x + mb_x], &img->mpr[j][mb_x], 2 * BLOCK_SIZE * sizeof(imgpel));
  }

  return coeff_cost;
}


/*!
************************************************************************
* \brief
*    Set mode parameters and reference frames for an 8x8 block
************************************************************************
*/
void SetModesAndRefframe (int b8, short* p_dir, int* fw_mode, int* bw_mode, short* fw_ref, short* bw_ref)
{
  Macroblock* currMB = &img->mb_data[img->current_mb_nr];
  int         j      = 2*(b8>>1);
  int         i      = 2*(b8 & 0x01);

  *fw_mode = *bw_mode = *fw_ref = *bw_ref = -1;

  *p_dir  = currMB->b8pdir[b8];



  if (img->type!=B_SLICE)
  {
    *fw_ref = enc_picture->ref_idx[LIST_0][img->block_y+j][img->block_x+i];
    *bw_ref = 0;
    *fw_mode  = currMB->b8mode[b8];
    *bw_mode  = 0;
  }
  else
  {
    if (currMB->b8pdir[b8]==-1)
    {
      *fw_ref   = -1;
      *bw_ref   = -1;
      *fw_mode  =  0;
      *bw_mode  =  0;
    }
    else if (currMB->b8pdir[b8]==0)
    {
      *fw_ref   = enc_picture->ref_idx[LIST_0][img->block_y+j][img->block_x+i];
      *bw_ref   = 0;
      *fw_mode  = currMB->b8mode[b8];
      *bw_mode  = 0;
    }
    else if (currMB->b8pdir[b8]==1)
    {
      *fw_ref   = 0;
      *bw_ref   = enc_picture->ref_idx[LIST_1][img->block_y+j][img->block_x+i];
      *fw_mode  = 0;
      *bw_mode  = currMB->b8mode[b8];
    }
    else
    {
      *fw_ref   = enc_picture->ref_idx[LIST_0][img->block_y+j][img->block_x+i];
      *bw_ref   = enc_picture->ref_idx[LIST_1][img->block_y+j][img->block_x+i];
      *fw_mode  = currMB->b8mode[b8];
      *bw_mode  = currMB->b8mode[b8];
    }
  }
}

void ComputeResidue (imgpel **curImg, int img_m7[16][16], int mb_y, int mb_x, int opix_y, int opix_x, int width, int height)
{
  static imgpel *imgOrg, *imgPred;
  static int    *m7;
  int i, j;

  for (j = mb_y; j < mb_y + height; j++)
  {
    imgOrg = &curImg[opix_y + j][opix_x];    
    imgPred = &(img->mpr[j][mb_x]);
    m7 = &img_m7[j][mb_x]; 
    for (i = 0; i < width; i++)
    {
      *m7++ = *imgOrg++ - *imgPred++;
    }
  }
}

extern int dct_luma_16x16_new (void);
extern int dct_luma_16x8 (int blk16x8);
extern int dct_luma_8x16 (int blk8x16);

static void LumaResidualCodingBigBlock(int mbi_x, int mbi_y, int mode, int mode_idx)
{
  int block8x8;
  int fw_mode, bw_mode;
  short p_dir, refframe, bw_ref;
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];
  int nonzero; 
  int pic_pix_y, pic_pix_x;
  int byy, block_y;
  int bxx, block_x;
  int mb_x, mb_y;
  static imgpel backup_mpr[100][64][64];  

  currMB->cbp     = 0 ;
  currMB->cbp_blk = 0 ;

  assert(currMB->luma_transform_size_8x8_flag != 0);

  for (block8x8=0; block8x8<4; block8x8++)
  {    
    mb_y       = (block8x8 >> 1) << 3;
    mb_x       = (block8x8 & 0x01) << 3;

    SetModesAndRefframe (block8x8, &p_dir, &fw_mode, &bw_mode, &refframe, &bw_ref);

    if (((p_dir == 0 || p_dir == 2 )&& fw_mode < 5) || ((p_dir == 1 || p_dir == 2 ) && bw_mode < 5))
    {
      for (byy=0, block_y=mb_y; block_y<mb_y+8; byy+=4, block_y+=4)
      {
        pic_pix_y = img->opix_y + block_y;

        for (bxx=0, block_x=mb_x; block_x<mb_x+8; bxx+=4, block_x+=4)
        {
          pic_pix_x = img->opix_x + block_x;
          if(isfirstQP || (mode<=0 || mode>=4))
          {
            //===== prediction of 4x4 block =====
            LumaPrediction4x4 (block_x, block_y, p_dir, fw_mode, bw_mode, refframe, bw_ref);
            if(mode>0 && mode<4)
            {
              int i,j;
              int mb64_x = mbi_x<<4;
              int mb64_y = mbi_y<<4;
              for(j=block_y; j<block_y+4; j++)
                for(i=block_x; i<block_x+4; i++)
                  backup_mpr[mode_idx][mb64_y+j][mb64_x+i] = img->mpr[j][i];
            }
          }
          else
          {//!isfirstQP && mode>0 && mode<4
            int i,j;
            int mb64_x = mbi_x<<4;
            int mb64_y = mbi_y<<4;
            for(j=block_y; j<block_y+4; j++)
              for(i=block_x; i<block_x+4; i++)
                img->mpr[j][i] = backup_mpr[mode_idx][mb64_y+j][mb64_x+i];
          }
        }
      }
      //---img->mpr[0-15][0-15] contains 16x16 prediction
      //===== compute prediction residual ======     
      ComputeResidue (imgY_org, img->m7, mb_y, mb_x, img->opix_y, img->opix_x + mb_x, 8, 8);
    }
  }

  if( (currMB->mb_type < 2 && currMB->luma_transform_size_8x8_flag == 2) || only_16x16_transform==1  )
  {
    nonzero = dct_luma_16x16_new();
    if (nonzero)
    {
      currMB->cbp_blk = 0xffff; // all sixteen 4x4 blocks contain coeff
      currMB->cbp = 0xf;        // alll 8x8 blocks contain coeff
    }
  }
  else if(currMB->mb_type == 2 && currMB->luma_transform_size_8x8_flag == 2)    // 16x8 transform
  {
    // top block
    nonzero = dct_luma_16x8(0);
    if (nonzero)
    {
      currMB->cbp_blk |= 0x00ff; // yye: double check
      currMB->cbp |= 0x3;        // yye: double check
    }
    // bottom block
    nonzero = dct_luma_16x8(1);
    if (nonzero)
    {
      currMB->cbp_blk |= 0xff00; // yye: double check
      currMB->cbp |= 0xc;        // yye: double check
    }
  }
  else if(currMB->mb_type == 3 && currMB->luma_transform_size_8x8_flag == 2)    // 8x16 transform
  {
    // top block
    nonzero = dct_luma_8x16(0);
    if (nonzero)
    {
      currMB->cbp_blk |= 0x3333; // yye: double check
      currMB->cbp |= 0x5;        // yye: double check
    }
    // bottom block
    nonzero = dct_luma_8x16(1);
    if (nonzero)
    {
      currMB->cbp_blk |= 0xcccc; // yye: double check
      currMB->cbp |= 0xa;        // yye: double check
    }
  }

}

/*!
************************************************************************
* \brief
*    Residual Coding of a Luma macroblock (not for intra)
************************************************************************
*/
void LumaResidualCoding (int mbi_x, int mbi_y, int mode, int mode_idx)
{
  int j,block8x8;
  int fw_mode, bw_mode;
  short p_dir, refframe;
  int sum_cnt_nonz;
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];

  if(((currMB->mb_type >= 1 && currMB->mb_type <= 3) && currMB->luma_transform_size_8x8_flag == 2) ||
    (currMB->mb_type == 0 && !img->NoResidueDirect && currMB->luma_transform_size_8x8_flag == 2))
  {
    LumaResidualCodingBigBlock(mbi_x, mbi_y, mode, mode_idx);
    return;
  }

  currMB->cbp     = 0 ;
  currMB->cbp_blk = 0 ;
  sum_cnt_nonz    = 0 ;

  for (block8x8=0; block8x8<4; block8x8++)
  {
    short bw_ref;
    SetModesAndRefframe (block8x8, &p_dir, &fw_mode, &bw_mode, &refframe, &bw_ref);

    sum_cnt_nonz += LumaResidualCoding8x8 (mbi_x, mbi_y, mode, mode_idx, &(currMB->cbp), &(currMB->cbp_blk), block8x8,
      p_dir, fw_mode, bw_mode, refframe, bw_ref);
  }

  if (sum_cnt_nonz <= _LUMA_MB_COEFF_COST_ )
  {
    currMB->cbp     &= 0xfffff0 ;
    currMB->cbp_blk &= 0xff0000 ;
    for (j=0; j < MB_BLOCK_SIZE; j++)
      memcpy(&enc_picture->imgY[img->pix_y+j][img->pix_x], img->mpr[j], MB_BLOCK_SIZE * sizeof (imgpel));
  }
}

static void get_fullpel_blockUV(imgpel **  ref, int x_pres, int y_pres, int max_x, int max_y, int block[32][32], int size_x, int size_y)
{
  int i, j;

  for (j = 0; j < size_y; j++)
    for (i = 0; i < size_x; i++)
      block[j][i] = ref[max(0,min(y_pres+j,max_y))][max(0,min(x_pres+i,max_x))];
}

static void interp_blockUV_direct_X(imgpel **ref, int pres_x, int pres_y, const int *coefx, int max_x, int max_y, int block[32][32], int size_x, int size_y, int symmetric)
{
  int i, j, x;
  int result;

  const int *filtP;
  imgpel *imgP;

  // Filter tap = 8, so filter offset = (8/2)-1 = 3 and block size is 4, so (max-3-4)=(max-7)
  if ((pres_y >= 0) && (pres_y < max_y - size_y) && (pres_x >= 3) && (pres_x < max_x - 3 - size_x))
  { //Image Non-Border safe positions...so dont need Clipping
    if (!symmetric)
    {
      for (j = 0; j < size_y; j++)
        for (i = 0; i < size_x; i++)
        {
          filtP = coefx;
          imgP = &(ref[pres_y+j][pres_x+i-2]);
          for (result = 0, x = -2; x < 4; x++)
            result += (*filtP++) * (*imgP++);
          block[j][i] = result;
        }
    }
    else
    {
      for (j = 0; j < size_y; j++)
        for (i = 0; i < size_x; i++)
        {
          filtP = coefx;
          for (result = 0, x = -2; x <= 0; x++)
            result += (*filtP++) * (ref[pres_y+j][pres_x+i+x] + ref[pres_y+j][pres_x+i-x+1]);
          block[j][i] = result;
        }
    }
  }
  else
  { //Image Border positions...so needs Clipping
    if (!symmetric)
    {
      for (j = 0; j < size_y; j++)
        for (i = 0; i < size_x; i++)
        {
          for (result = 0, x = -2; x < 4; x++)
            result += ref[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i+x)] * coefx[x+2];
          block[j][i] = result;
        }
    }
    else
    {
      for (j = 0; j < size_y; j++)
        for (i = 0; i < size_x; i++)
        {
          for (result = 0, x = -2; x <= 0; x++)
            result += coefx[x+2] * (ref[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i+x)] +
            ref[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i-x+1)]);

          block[j][i] = result;
        }
    }
  }
}

static void interp_blockUV_direct_Y(imgpel **ref, int pres_x, int pres_y, const int *coefy, int max_x, int max_y, int block[32][32], int size_x,int size_y, int symmetric)
{
  int i, j, y;
  int result;

  const int *filtP;

  // Filter tap = 8, so filter offset = (8/2)-1 = 3 and block size is 4, so (max-3-4)=(max-7)
  if ((pres_y >= 3) && (pres_y < max_y - 3 - size_y) && (pres_x >= 0) && (pres_x < max_x - size_x))
  { //Image Non-Border safe positions...so dont need Clipping
    if (!symmetric)
    {
      for (j = 0; j < size_y; j++)
        for (i = 0; i < size_x; i++)
        {
          filtP = coefy;
          for (result = 0, y = -2; y < 4; y++)
            result += ref[pres_y+j+y][pres_x+i] * (*filtP++);
          block[j][i] = result;

        }
    }
    else
    {
      for (j = 0; j < size_y; j++)
        for (i = 0; i < size_x; i++)
        {
          filtP = coefy;
          for (result = 0, y = -2; y <=0; y++)
            result += (*filtP++) * (ref[pres_y+j+y][pres_x+i] +
            ref[pres_y+j-y+1][pres_x+i]);
          block[j][i] = result;
        }
    }
  }
  else
  { //Image Border positions...so needs Clipping
    if (!symmetric)
    {
      for (j = 0; j < size_y; j++)
        for (i = 0; i < size_x; i++)
        {
          filtP = coefy;
          for (result = 0, y = -2; y < 4; y++)
            result += ref[Clip3(0, max_y, pres_y+j+y)][Clip3(0, max_x, pres_x+i)] * (*filtP++);
          block[j][i] = result;
        }
    }
    else
    {
      for (j = 0; j < size_y; j++)
        for (i = 0; i < size_x; i++)
        {
          filtP = coefy;
          for (result = 0, y = -2; y <= 0; y++)
            result += (*filtP++) * (ref[Clip3(0, max_y, pres_y+j+y)][Clip3(0, max_x, pres_x+i)] +
            ref[Clip3(0, max_y, pres_y+j-y+1)][Clip3(0, max_x, pres_x+i)]);
          block[j][i] = result;
        }
    }
  }

}

void get_eighthpel_blockUV_direct(imgpel **ref, int x_pos, int y_pos, ImageParameters *img, int block[32][32], int geometry, int size_x, int size_y)
{
  int i, j;

  static const int COEF_6tap[3][6] =
  {
    {  2,  -5, 28,  9, -3, 1 },
    {  2,  -6, 20, 20, -6, 2 },
    {  1,  -3,  9, 28, -5, 2 }
  };

  static const int COEF_6tap_EIGHTH_PEL[4][6] =
  {
    {  2, -5,  60,  9, -3,  1 },
    {  4, -11, 48, 29, -9,  3 },
    {  3, -9,  29, 48, -11, 4 },
    {  1, -3,   9, 60, -5,  2 }
  };

  /*
  // Filters summing up to 256
  static const int COEF_6tap[3][6] =
  {
  { 11, -36, 226,  71, -24,  8 },
  { 10, -40, 158, 158, -40, 10 },
  { 8, -24,  71, 226, -36, 11 }
  };

  static const int COEF_6tap_EIGHTH_PEL[4][6] =
  {
  { 11, -36, 482,  71, -24,  8 },
  { 21, -76, 384, 229, -64, 18 },
  { 18, -64, 229, 384, -76, 21 },
  { 8,  -24,  71, 482, -36, 11 }
  };
  */

  int chroma_size_x = geometry?1:size_x; // If the geometry mode is ON, only 1x1 block is processed. Otherwise size_x x size_y. 
  int chroma_size_y = geometry?1:size_y; // If the geometry mode is ON, only 1x1 block is processed. Otherwise size_x x size_y. 

  int dx = (x_pos>>img->mv_res) & 7, x = 0;
  int dy = (y_pos>>img->mv_res) & 7, y = 0;
  int pres_x = x_pos >> (3+img->mv_res);
  int pres_y = y_pos >> (3+img->mv_res);

  //int dx = x_pos & 7;
  //int dy = y_pos & 7;

  //int pres_x = x_pos >> 3;
  //int pres_y = y_pos >> 3;


  int tmp[32][32+16]; // 16 is a conservative choice
  int result;

  int max_x = img->width/2 - 1;
  int max_y = img->height/2 - 1;

  int round5 = 16;
  int round6 = 32;
  int round10 = (1<<9);
  int round11 = (1<<10);
  int round12 = (1<<11);

  int filt_index, symmetric;
  const int *filtP;
  imgpel *imgP;
  int *intP;


  // choose filter depending on the sub-pel position
  if (dx == 0 && dy == 0)
  { // full-pel position, just return full-pel block
    get_fullpel_blockUV(ref, pres_x, pres_y, max_x, max_y, block, chroma_size_x, chroma_size_y);
  }
  else if (dy == 0)
  { // horizontal interpolation only
    if (dx&1)
    { // 1/8-pel position
      interp_blockUV_direct_X(ref, pres_x, pres_y, COEF_6tap_EIGHTH_PEL[dx>>1], max_x, max_y, block, chroma_size_x, chroma_size_y, 0);
      for (j = 0; j < chroma_size_y; j++)
        for (i = 0; i < chroma_size_x; i++)
          block[j][i] = Clip3(0, img->max_imgpel_value, (block[j][i] + round6) >> 6); // Filter sums up to 64 for 1/8 pel
    }
    else
    {
      filt_index = (dx>>1)-1;
      symmetric = (filt_index == 1);
      interp_blockUV_direct_X(ref, pres_x, pres_y, COEF_6tap[filt_index], max_x, max_y, block, chroma_size_x, chroma_size_y, symmetric);
      for (j = 0; j < chroma_size_y; j++)
        for (i = 0; i < chroma_size_x; i++)
          block[j][i] = Clip3(0, img->max_imgpel_value, (block[j][i] + round5) >> 5); // Filter sums up to 32 for 1/4 pel
    }
  }
  else if (dx == 0)
  { // vertical interpolation only
    if (dy&1)
    { // 1/8-pel position (vertical)
      interp_blockUV_direct_Y(ref, pres_x, pres_y, COEF_6tap_EIGHTH_PEL[dy>>1], max_x, max_y, block, chroma_size_x, chroma_size_y, 0);
      for (j = 0; j < chroma_size_y; j++)
        for (i = 0; i < chroma_size_x; i++)
          block[j][i] = Clip3(0, img->max_imgpel_value, (block[j][i] + round6) >> 6); // Filter sums up to 64 for 1/8 pel
    }
    else
    {
      filt_index = (dy>>1)-1;
      symmetric = (filt_index == 1);
      interp_blockUV_direct_Y(ref, pres_x, pres_y, COEF_6tap[filt_index], max_x, max_y, block, size_x, size_y, symmetric);
      for (j = 0; j < chroma_size_y; j++)
        for (i = 0; i < chroma_size_x; i++)
          block[j][i] = Clip3(0, img->max_imgpel_value, (block[j][i] + round5) >> 5); // Filter sums up to 32 for 1/4 pel
    }
  }
  else if (!(dx&1))
  {  // horizontal 1/4-pel
    filt_index = (dx>>1)-1;
    if ((pres_y >= 3) && (pres_y < max_y-3-size_y) && (pres_x >= 3) && (pres_x < max_x-3-size_x))
    { //Image Non-Border safe positions...so dont need Clipping
      if (filt_index != 1)
      {
        for (j = -2; j < (chroma_size_y+4); j++)
          for (i = 0; i < chroma_size_x; i++)
          {
            imgP = &(ref[pres_y+j][pres_x+i-2]);
            filtP = COEF_6tap[filt_index];
            for (tmp[i][j+2] = 0, x = -2; x < 4; x++)
              tmp[i][j+2] += (*imgP++) * (*filtP++); // indices (i,j) reversed to facilitate further filtering
          }
      }
      else
      {
        for (j = -2; j < (chroma_size_y+4); j++)
          for (i = 0; i < chroma_size_x; i++)
          {
            filtP = COEF_6tap[filt_index];
            for (tmp[i][j+2] = 0, x = -2; x <= 0; x++)
              tmp[i][j+2] += (*filtP++) * (ref[pres_y+j][pres_x+i+x] + ref[pres_y+j][pres_x+i-x+1]);
            // indices (i,j) reversed to facilitate further filtering
          }
      }
    }
    else
    {//Image Border positions...so needs Clipping
      if (filt_index != 1)
      {
        for (j = -2; j < (chroma_size_y+4); j++)
          for (i = 0; i < chroma_size_x; i++)
          {
            for (tmp[i][j+2] = 0, x = -2; x < 4; x++)
              tmp[i][j+2] += ref[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i+x )] * COEF_6tap[filt_index][x+2];
            // indices (i,j) reversed to facilitate further filtering
          }
      }
      else
      {
        for (j = -2; j < (chroma_size_y+4); j++)
          for (i = 0; i < chroma_size_x; i++)
          {
            for (tmp[i][j+2] = 0, x = -2; x <= 0; x++)
              tmp[i][j+2] += COEF_6tap[filt_index][2+x] * (ref[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i+x)] +
              ref[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i-x+1)]);
            // indices (i,j) reversed to facilitate further filtering
          }
      }
    }
    // At this point the nominal bit-depth has been increased by 5 (1/4-pel filtering, Filter sums up to 32).

    if (dy&1)
    { // 1/8-pel position (vertical)
      filt_index = (dy>>1);
      for (j = 0; j < chroma_size_y; j++)
        for (i = 0; i < chroma_size_x; i++)
        {
          intP = &(tmp[i][j]);
          filtP = COEF_6tap_EIGHTH_PEL[filt_index];
          for (result = 0, y = -2; y < 4; y++)
            result += (*filtP++) * (*intP++);
          block[j][i] = Clip3(0, img->max_imgpel_value, (result + round11) >> 11);
          // 1/8-pel filtering, Filter sums up to 512. Total nominal bit-depth increased by 17.
        }
    }
    else
    { // 1/4-pel position (vertical)
      filt_index = (dy>>1)-1;
      if (filt_index != 1)
      {
        for (j = 0; j < chroma_size_y; j++)
          for (i = 0; i < chroma_size_x; i++)
          {
            intP = &(tmp[i][j]);
            filtP = COEF_6tap[filt_index];
            for (result = 0, y = -2; y < 4; y++)
              result += (*filtP++) * (*intP++);
            block[j][i] = Clip3(0, img->max_imgpel_value, (result + round10) >> 10);
            // 1/4-pel filtering, Filter sums up to 256. Total nominal bit-depth increased by 16.
          }
      }
      else
      {
        for (j = 0; j < chroma_size_y; j++)
          for (i = 0; i < chroma_size_x; i++)
          {
            filtP = COEF_6tap[filt_index];
            for (result = 0, y = -2; y <= 0; y++)
              result += (*filtP++) * (tmp[i][j+y+2] + tmp[i][j-y+1+2]);
            block[j][i] = Clip3(0, img->max_imgpel_value, (result + round10) >> 10);
            // 1/4-pel filtering, Filter sums up to 256. Total nominal bit-depth increased by 16.
          }
      }
    }
  }
  else if (!(dy&1))
  { // vertical 1/4-pel and horizontal 1/8-pel interpolation
    if ((pres_y >= 3) && (pres_y < max_y-3-size_y) && (pres_x >= 3) && (pres_x < max_x-3-size_x))
    { //Image Non-Border safe positions...so don't need Clipping
      filt_index = (dy>>1)-1;
      if (filt_index != -1)
      {
        for (j = 0; j < chroma_size_y; j++) // y
          for (i = -2; i < chroma_size_x+4; i++) // x
          { // get vertical 1/4-pel positions
            filtP = COEF_6tap[filt_index];
            for (tmp[j][i+2] = 0, y = -2; y < 4; y++)
              tmp[j][i+2] += (*filtP++) * ref[pres_y+j+y][pres_x+i];
          }
      }
      else
      {
        for (j = 0; j < chroma_size_y; j++) // y
          for (i = -2; i < chroma_size_x+4; i++) // x
          { // get vertical 1/4-pel positions
            filtP = COEF_6tap[filt_index];
            for (tmp[j][i+2] = 0, y = -2; y <= 0; y++)
              tmp[j][i+2] += (*filtP++) * (ref[pres_y+j+y][pres_x+i] +
              ref[pres_y+j-y+1][pres_x+i]);
          }
      }
    }
    else
    { //Image Border positions...so needs Clipping
      filt_index = (dy>>1)-1;
      if (filt_index != -1)
      {
        for (j = 0; j < chroma_size_y; j++) // y
          for (i = -2; i < chroma_size_x+4; i++) // x
          { // get vertical 1/4-pel positions
            filtP = COEF_6tap[filt_index];
            for (tmp[j][i+2] = 0, y = -2; y < 4; y++)
              tmp[j][i+2] += (*filtP++) * ref[Clip3(0, max_y, pres_y+j+y)][Clip3(0, max_x, pres_x+i)];
          }
      }
      else
      {
        for (j = 0; j < chroma_size_y; j++) // y
          for (i = -2; i < chroma_size_x + 4; i++) // x
          { // get vertical 1/4-pel positions
            filtP = COEF_6tap[filt_index];
            for (tmp[j][i+2] = 0, y = -2; y <= 0; y++)
              tmp[j][i+2] += (*filtP++) * (ref[Clip3(0, max_y, pres_y+j+y)][Clip3(0, max_x, pres_x+i)] +
              ref[Clip3(0, max_y, pres_y+j-y+1)][Clip3(0, max_x, pres_x+i)]);
          }
      }
    }
    // At this point the nominal bit-depth has been increased by 8 (1/4-pel filtering, Filter sums up to 256).

    // horizontal 1/8 pel position filtering
    filt_index = (dx>>1);
    for (j = 0; j < chroma_size_y; j++)
      for (i = 0; i < chroma_size_x; i++)
      {
        intP = &(tmp[j][i]);
        filtP = COEF_6tap_EIGHTH_PEL[filt_index];
        for (result = 0, y = -2; y < 4; y++)
          result += (*filtP++) * (*intP++);
        block[j][i] = Clip3(0, img->max_imgpel_value, (result + round11) >> 11);
        // 1/8-pel filtering, Filter sums up to 64. Total nominal bit-depth increased by 11.
      }
  }       
  else
  { // 1/8-1/8 positions
    // Horizontal 1/8 pel position filtering
    filt_index = (dx>>1);
    if ((pres_y >= 3) && (pres_y < max_y-3-size_y) && (pres_x >= 3) && (pres_x < max_x-3-size_x))
    { //Image Non-Border safe positions...so dont need Clipping
      for (j = -2; j < (chroma_size_y+4); j++)
        for (i = 0; i < chroma_size_x; i++)
        {
          imgP = &(ref[pres_y+j][pres_x+i-2]);
          filtP = COEF_6tap_EIGHTH_PEL[filt_index];
          for (tmp[i][j+2] = 0, x = -2; x < 4; x++)
            tmp[i][j+2] += (*imgP++) * (*filtP++);
        }
    }
    else
    {//Image Border positions...so needs Clipping
      for (j = -2; j < (chroma_size_y+4); j++)
        for (i = 0; i < chroma_size_x; i++)
        {
          for (tmp[i][j+2] = 0, x = -2; x < 4; x++)
            tmp[i][j+2] += ref[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i+x)] * COEF_6tap_EIGHTH_PEL[filt_index][x+2];
        }
    }
    // At this point the nominal bit-depth has been increased by 6 (1/8-pel filtering, Filter sums up to 64).

    // vertical 1/8 pel position filtering
    filt_index = (dy>>1);
    for (j = 0; j < chroma_size_y; j++)
      for (i = 0; i < chroma_size_x; i++)
      {
        intP = &(tmp[i][j]);
        filtP = COEF_6tap_EIGHTH_PEL[filt_index];
        for (result = 0, y = -2; y < 4; y++)
          result += (*filtP++) * (*intP++);
        block[j][i] = Clip3(0, img->max_imgpel_value, (result + round12) >> 12);
      }

  }
}

/*!
************************************************************************
* \brief
*    Predict one component of a chroma 4x4 block
************************************************************************
*/
void OneComponentChromaPrediction4x4 (int*        mpred,      //!< array to store prediction values
                                      int         block_c_x,  //!< horizontal pixel coordinate of 4x4 block
                                      int         block_c_y,  //!< vertical   pixel coordinate of 4x4 block
                                      short****** mv,         //!< motion vector array
                                      int         list_idx,   //!< reference picture list
                                      short       ref,        //!< reference index
                                      int         blocktype,  //!< block type
                                      int         uv)         //!< chroma component
{
  int     i, j, ii, jj;
  short*  mvb;
  int mv_shift= img->mv_res;

  int     f1_x = (64/img->mb_cr_size_x) << mv_shift;
  int     f1_y = (64/img->mb_cr_size_y) << mv_shift;

  int     list_offset = img->mb_data[img->current_mb_nr].list_offset;  
  int     jjx, iix;
  int     mb_cr_y_div4 = img->mb_cr_size_y>>2;
  int     mb_cr_x_div4 = img->mb_cr_size_x>>2;
  int     jpos;

  StorablePicture **list = listX[list_idx + list_offset];
  pel_t** refimage = list[ref]->imgUV[uv];

  int blockUV[32][32];
  unsigned short m=0;

#ifdef GEOM
  int geometry=1;
#else
  int geometry=0;  // When integrating geometry, make sure that this variable gets set to 1 if the mode
#endif             //      uses geometry.

  for (j=block_c_y; j < block_c_y + BLOCK_SIZE; j++)
  {
    jjx = j/mb_cr_y_div4;
    jpos = (j + img->opix_c_y)*f1_y;//if img->mb_cr_size_y=8, f1_y = 8

    for (i=block_c_x; i < block_c_x + BLOCK_SIZE; i++)
    {
      iix = i/mb_cr_x_div4;
      mvb  = mv [jjx][iix][list_idx][ref][blocktype];

#ifdef MV_COMPETITION
      if ((skip_mode== TRUE) &&  (input->mv_competition > 0))
      {
        mvb[0] = (short) mv_comp.mv_pred_skip[mv_comp.current_predictor_for_skip][0];
        mvb[1] = (short) mv_comp.mv_pred_skip[mv_comp.current_predictor_for_skip][1];
      }
#endif

      //(ii, jj) are coordinators in 4 times upsampled luma grid
      ii   = (i + img->opix_c_x)*f1_x + mvb[0];
      jj   = jpos + mvb[1];

      if (active_sps->chroma_format_idc == 1)
        jj  += list[ref]->chroma_vector_adjustment;

      //   0  1  2  3
      //   4  5  6  7
      //   8  9 10 11
      //  12 13 14 15
      if (!geometry)
      {
        if(m==0 || m==2 || m==8 || m==10)    
        {
          get_eighthpel_blockUV_direct(refimage, ii, jj, img, blockUV, 0, 2, 2);
          mpred[m+0] = blockUV[0][0];
          mpred[m+1] = blockUV[0][1];
          mpred[m+4] = blockUV[1][0];
          mpred[m+5] = blockUV[1][1];
        }
        m++;
      }
      else
      {
        get_eighthpel_blockUV_direct(refimage, ii, jj, img, blockUV, 1, 2, 2);
        *mpred++ = blockUV[0][0];
      }

    }
  }
}


/*!
************************************************************************
* \brief
*    Predict an intra chroma 4x4 block
************************************************************************
*/
void IntraChromaPrediction4x4 (int  uv,       // <-- colour component
                               int  block_x,  // <-- relative horizontal block coordinate of 4x4 block
                               int  block_y)  // <-- relative vertical   block coordinate of 4x4 block
{
  int mode = img->mb_data[img->current_mb_nr].c_ipred_mode;
  int j;

  //===== prediction =====
  for (j=block_y; j<block_y+4; j++)
    memcpy(&img->mpr[j][block_x],&img->mprr_c[uv][mode][j][block_x], BLOCK_MULTIPLE * sizeof(imgpel));
}



/*!
************************************************************************
* \brief
*    Predict one chroma 4x4 block
************************************************************************
*/
void
ChromaPrediction4x4 (int   uv,           // <-- colour component
                     int   block_x,      // <-- relative horizontal block coordinate of 4x4 block
                     int   block_y,      // <-- relative vertical   block coordinate of 4x4 block
                     int   p_dir,        // <-- prediction direction
                     int   fw_mode,      // <-- list0  prediction mode (1-7, 0=DIRECT if bw_mode=0)
                     int   bw_mode,      // <-- list1 prediction mode (1-7, 0=DIRECT if fw_mode=0)
                     short fw_ref_idx,   // <-- reference frame for list0 prediction (if (<0) -> intra prediction)
                     short bw_ref_idx)   // <-- reference frame for list1 prediction 
{
  static int fw_pred[MB_BLOCK_SIZE];
  static int bw_pred[MB_BLOCK_SIZE];

  int  i, j;
  int  block_x4  = block_x+4;
  int  block_y4  = block_y+4;
  int* fpred     = fw_pred;
  int* bpred     = bw_pred;
  short****** mv_array = img->all_mv;

  Macroblock*    currMB     = &img->mb_data[img->current_mb_nr];

  if (currMB->bi_pred_me && fw_ref_idx == 0 && bw_ref_idx == 0 && p_dir == 2 && fw_mode==1 && bw_mode==1)
    mv_array = currMB->bi_pred_me == 1? img->bipred_mv1 : img->bipred_mv2 ;

  //===== INTRA PREDICTION =====
  if (p_dir==-1)
  {
    IntraChromaPrediction4x4 (uv, block_x, block_y);
    return;
  }

  //===== INTER PREDICTION =====
  if ((p_dir==0) || (p_dir==2))
    OneComponentChromaPrediction4x4 (fw_pred, block_x, block_y, mv_array, LIST_0, fw_ref_idx, fw_mode, uv);

  if ((p_dir==1) || (p_dir==2))
    OneComponentChromaPrediction4x4 (bw_pred, block_x, block_y, mv_array, LIST_1, bw_ref_idx, bw_mode, uv);


  if (p_dir==2)
  {
    for (j=block_y; j<block_y4; j++)
      for (i=block_x; i<block_x4; i++)  
#ifdef  USE_HP_FILTER//BROUND
        if(input->UseHPFilter != 0)
        {
          img->mpr[j][i] = (*fpred++ + *bpred++ + 1 - img->bipred_rounding_control) >> 1; 
        }
        else
        {
          img->mpr[j][i] = (*fpred++ + *bpred++ + 1) >> 1; 
        }
#else
        img->mpr[j][i] = (*fpred++ + *bpred++ + 1) >> 1; 
#endif
  }
  else if (p_dir==0)
  {
    for (j=block_y; j<block_y4; j++)
      for (i=block_x; i<block_x4; i++)  
        img->mpr[j][i] = *fpred++;
  }
  else // (p_dir==1)
  {
    for (j=block_y; j<block_y4; j++)
      for (i=block_x; i<block_x4; i++)  
        img->mpr[j][i] = *bpred++;
  }

}


/*!
************************************************************************
* \brief
*    Chroma residual coding for an macroblock
************************************************************************
*/
void ChromaResidualCoding (int mbi_x, int mbi_y, int mode, int mode_idx, int* cr_cbp)
{
  int   uv, block8, block_y, block_x, j, i;
  int   fw_mode, bw_mode;
  short p_dir, refframe, bw_ref;
  int   yuv = img->yuv_format - 1; //ADD-VG-15052004
  static imgpel backup_mpr_c[100][2][64][64];//[uv][y][x]
  int   block8x8_idx[3][4][4] =     //ADD-VG-15052004
  { { {0, 1, 0, 0}, 
  {2, 3, 0, 0},
  {0, 0, 0, 0},
  {0, 0, 0, 0}, },

  { {0, 1, 0, 0}, 
  {0, 1, 0, 0}, 
  {2, 3, 0, 0},
  {2, 3, 0, 0}  },

  { {0, 0, 1, 1},
  {0, 0, 1, 1},
  {2, 2, 3, 3},
  {2, 2, 3, 3}  }
  };

  int   skipped = 0;

  for (*cr_cbp=0, uv=0; uv<2; uv++)
  {
    //===== prediction of chrominance blocks ===d==
    block8 = 0;
    for (block_y=0; block_y < img->mb_cr_size_y; block_y+=4)
      for (block_x=0; block_x < img->mb_cr_size_x; block_x+=4)
      {
        if(isfirstQP || mode<=0 || mode>=4) //ChromaResidualCoding is not used for submacroblock_mode_decision
        {
          block8 = block8x8_idx[yuv][block_y>>2][block_x>>2];
          SetModesAndRefframe (block8, &p_dir, &fw_mode, &bw_mode, &refframe, &bw_ref);

          ChromaPrediction4x4 (uv, block_x, block_y, p_dir, fw_mode, bw_mode, refframe, bw_ref);

          if(mode>0 && mode<4)
          {
            int mb64_x = mbi_x<<3;
            int mb64_y = mbi_y<<3;
            for(j=block_y; j<block_y+4; j++)
              for(i=block_x; i<block_x+4; i++)
                backup_mpr_c[mode_idx][uv][mb64_y+j][mb64_x+i] = img->mpr[j][i];
          }
        }
        else
        {//!isfirstQP && mode>0 && mode<=P8x8
          int mb64_x = mbi_x<<3;
          int mb64_y = mbi_y<<3;
          for(j=block_y; j<block_y+4; j++)
            for(i=block_x; i<block_x+4; i++)
              img->mpr[j][i] = backup_mpr_c[mode_idx][uv][mb64_y+j][mb64_x+i];

        }
      }

      // ==== set chroma residue to zero for skip Mode in SP frames 
      if (img->NoResidueDirect)
      {
        for (j=0; j<img->mb_cr_size_y; j++)
          memcpy(&enc_picture->imgUV[uv][img->pix_c_y+j][img->pix_c_x], img->mpr[j], img->mb_cr_size_x * sizeof(imgpel));             
      }
      else
      {
        if (skipped)    //this will never happen under P_DIRECT macro ??
        {
          for (j=0; j<img->mb_cr_size_y; j++)
            memcpy(&enc_picture->imgUV[uv][img->pix_c_y+j][img->pix_c_x], img->mpr[j], img->mb_cr_size_x * sizeof(imgpel));
        }
        else
        {
          for (j=0; j<img->mb_cr_size_y; j++)
            for (i=0; i<img->mb_cr_size_x; i++)
              img->m7[j][i] = imgUV_org[uv][img->opix_c_y+j][img->opix_c_x+i] - img->mpr[j][i];
        }
      }

      //===== DCT, Quantization, inverse Quantization, IDCT, and Reconstruction =====
      if (!img->NoResidueDirect && !skipped)
      {
        *cr_cbp=dct_chroma   (uv,*cr_cbp); 
      }

  }

  //===== update currMB->cbp =====
  img->mb_data[img->current_mb_nr].cbp += ((*cr_cbp)<<4);  
}

/*!
************************************************************************
* \brief
*    Intra prediction of the chrminance layers of one macroblock
************************************************************************
*/
void IntraChromaPrediction (int *mb_up, int *mb_left, int*mb_up_left)
{

  //Macroblock *currMB = &img->mb_data[img->current_mb_nr];
  int      s, s0, s1, s2, s3, i, j;
  pel_t**  image;
  int      block_x, block_y;
  int      mb_nr = img->current_mb_nr;
  int      mb_available_up;
  int      mb_available_left[2];
  int      mb_available_up_left;
  int      ih,iv;
  int      ib,ic,iaa;
  int      uv;
  imgpel   hline[16], vline[16];
  //int      best_mode = DC_PRED_8;  //just an initilaization here, should always be overwritten
  PixelPos up;        //!< pixel position  p(0,-1)
  PixelPos left[17];  //!< pixel positions p(-1, -1..15)
  int      cr_MB_x = img->mb_cr_size_x;
  int      cr_MB_y = img->mb_cr_size_y;

  int      blk_x;
  int      blk_y;
  int      b8,b4;
  int      yuv = img->yuv_format - 1;

  static int block_pos[3][4][4]= //[yuv][b8][b4]
  {
    { {0, 1, 2, 3},{0, 0, 0, 0},{0, 0, 0, 0},{0, 0, 0, 0}},
    { {0, 1, 2, 3},{2, 3, 2, 3},{0, 0, 0, 0},{0, 0, 0, 0}},
    { {0, 1, 2, 3},{1, 1, 3, 3},{2, 3, 2, 3},{3, 3, 3, 3}}
  };


  for (i=0;i<cr_MB_y+1;i++)
  {
    getNeighbour(mb_nr, -1 , i-1 , 0, &left[i]);
  }

  getNeighbour(mb_nr, 0 , -1 , 0, &up);


  mb_available_up                             = up.available;
  mb_available_up_left                        = left[0].available;
  mb_available_left[0] = mb_available_left[1] = left[1].available;

  if(input->UseConstrainedIntraPred)
  {
    mb_available_up = up.available ? img->intra_block[up.mb_addr] : 0;
    for (i=0, mb_available_left[0]=1; i<(cr_MB_y>>1);i++)
      mb_available_left[0]  &= left[i+1].available ? img->intra_block[left[i+1].mb_addr]: 0;
    for (i=(cr_MB_y>>1), mb_available_left[1]=1; i<cr_MB_y;i++)
      mb_available_left[1] &= left[i+1].available ? img->intra_block[left[i+1].mb_addr]: 0;
    mb_available_up_left = left[0].available ? img->intra_block[left[0].mb_addr]: 0;
  }

  if (mb_up)
    *mb_up = mb_available_up;
  if (mb_left)
    *mb_left = mb_available_left[0] && mb_available_left[1];
  if (mb_up_left)
    *mb_up_left = mb_available_up_left;


  // compute all chroma intra prediction modes for both U and V
  for (uv=0; uv<2; uv++)
  {
    image = enc_picture->imgUV[uv];

    // DC prediction
    for(b8=0; b8<img->num_blk8x8_uv >> 1;b8++)
    {
      for (b4=0; b4<4; b4++)
      {
        block_y = subblk_offset_y[yuv][b8][b4];
        block_x = subblk_offset_x[yuv][b8][b4];
        blk_x = block_x;
        blk_y = block_y + 1;

        s=img->dc_pred_value_chroma;
        s0=s1=s2=s3=0;

        //===== get prediction value =====
        switch (block_pos[yuv][b8][b4])
        {
        case 0:  //===== TOP LEFT =====
          if      (mb_available_up)       for (i=blk_x;i<(blk_x+4);i++)  s0 += image[up.pos_y][up.pos_x + i];
          if      (mb_available_left[0])  for (i=blk_y;i<(blk_y+4);i++)  s2 += image[left[i].pos_y][left[i].pos_x];
          if      (mb_available_up && mb_available_left[0])  s  = (s0+s2+4) >> 3;
          else if (mb_available_up)                          s  = (s0   +2) >> 2;
          else if (mb_available_left[0])                     s  = (s2   +2) >> 2;
          break;
        case 1: //===== TOP RIGHT =====
          if      (mb_available_up)       for (i=blk_x;i<(blk_x+4);i++)  s1 += image[up.pos_y][up.pos_x + i];
          else if (mb_available_left[0])  for (i=blk_y;i<(blk_y+4);i++)  s2 += image[left[i].pos_y][left[i].pos_x];
          if      (mb_available_up)                          s  = (s1   +2) >> 2;
          else if (mb_available_left[0])                     s  = (s2   +2) >> 2;
          break;
        case 2: //===== BOTTOM LEFT =====
          if      (mb_available_left[1])  for (i=blk_y;i<(blk_y+4);i++)  s3 += image[left[i].pos_y][left[i].pos_x];
          else if (mb_available_up)       for (i=blk_x;i<(blk_x+4);i++)  s0 += image[up.pos_y][up.pos_x + i];
          if      (mb_available_left[1])                     s  = (s3   +2) >> 2;
          else if (mb_available_up)                          s  = (s0   +2) >> 2;
          break;
        case 3: //===== BOTTOM RIGHT =====
          if      (mb_available_up)       for (i=blk_x;i<(blk_x+4);i++)  s1 += image[up.pos_y][up.pos_x + i];
          if      (mb_available_left[1])  for (i=blk_y;i<(blk_y+4);i++)  s3 += image[left[i].pos_y][left[i].pos_x];
          if      (mb_available_up && mb_available_left[1])  s  = (s1+s3+4) >> 3;
          else if (mb_available_up)                          s  = (s1   +2) >> 2;
          else if (mb_available_left[1])                     s  = (s3   +2) >> 2;
          break;
        }

        //===== prediction =====
        for (j=block_y; j<block_y+4; j++)
          for (i=block_x; i<block_x+4; i++)
          {
            img->mprr_c[uv][DC_PRED_8][j][i] = s;
          }
      }
    }

    // vertical prediction
    if (mb_available_up)
    {
      memcpy(hline,&image[up.pos_y][up.pos_x], cr_MB_x * sizeof(imgpel));
      for (j=0; j<cr_MB_y; j++)
        memcpy(img->mprr_c[uv][VERT_PRED_8][j], hline, cr_MB_x * sizeof(imgpel));
    }

    // horizontal prediction 
    if (mb_available_left[0] && mb_available_left[1])
    {
      for (i=0; i<cr_MB_y; i++)
        vline[i] = image[left[i+1].pos_y][left[i+1].pos_x];
      for (i=0; i<cr_MB_x; i++)
        for (j=0; j<cr_MB_y; j++)
          img->mprr_c[uv][HOR_PRED_8][j][i] = vline[j];
    }

    // plane prediction
    if (mb_available_left[0] && mb_available_left[1] && mb_available_up && mb_available_up_left)
    {
      ih = (cr_MB_x>>1)*(hline[cr_MB_x-1] - image[left[0].pos_y][left[0].pos_x]);
      for (i=0;i<(cr_MB_x>>1)-1;i++)
        ih += (i+1)*(hline[(cr_MB_x>>1)+i] - hline[(cr_MB_x>>1)-2-i]);

      iv = (cr_MB_y>>1)*(vline[cr_MB_y-1] - image[left[0].pos_y][left[0].pos_x]);
      for (i=0;i<(cr_MB_y>>1)-1;i++)
        iv += (i+1)*(vline[(cr_MB_y>>1)+i] - vline[(cr_MB_y>>1)-2-i]);

      ib= ((cr_MB_x == 8?17:5)*ih+2*cr_MB_x)>>(cr_MB_x == 8?5:6);
      ic= ((cr_MB_y == 8?17:5)*iv+2*cr_MB_y)>>(cr_MB_y == 8?5:6);

      iaa=16*(hline[cr_MB_x-1]+vline[cr_MB_y-1]);
      for (j=0; j<cr_MB_y; j++)
        for (i=0; i<cr_MB_x; i++)
          img->mprr_c[uv][PLANE_8][j][i]=max(0,min(img->max_imgpel_value_uv,
          (iaa+(i-(cr_MB_x>>1)+1)*ib+(j-(cr_MB_y>>1)+1)*ic+16)>>5));
    }
  }
}


/*!
************************************************************************
* \brief
*    Check if all reference frames for a macroblock are zero
************************************************************************
*/
int
ZeroRef (Macroblock* currMB)
{
  int i,j;

  for (j=img->block_y; j<img->block_y + BLOCK_MULTIPLE; j++)
    for (i=img->block_x; i<img->block_x + BLOCK_MULTIPLE; i++)
    {
      if (enc_picture->ref_idx[LIST_0][j][i]!=0)
        return 0;
    }
    return 1;
}


/*!
************************************************************************
* \brief
*    Converts macroblock type to coding value
************************************************************************
*/
int MBType2Value (Macroblock* currMB)
{
  int mbtype, pdir0, pdir1;

  if (img->type!=B_SLICE)
  {
    if      (currMB->mb_type==I8MB ||currMB->mb_type==I4MB)
      return (img->type==I_SLICE ? 0 : 6);
    else if (currMB->mb_type==I16MB) return (img->type==I_SLICE ? 0 : 6) + img->i16offset;
    else if (currMB->mb_type==P8x8)
    {
      return 4;
    }
    else                             return currMB->mb_type;
  }
  else
  {
    mbtype = currMB->mb_type;
    pdir0  = currMB->b8pdir[0];
    pdir1  = currMB->b8pdir[3];

    if      (mbtype==0)       return 0;
    else if (mbtype==I4MB || mbtype==I8MB)
      return 23;
    else if (mbtype==I16MB)   return 23 + img->i16offset;
    else if (mbtype==P8x8)    return 22;
    else if (mbtype==1)       return 1;
    else if (mbtype==2)       return 2;
    else                      return 3;
  }
}


/*!
************************************************************************
* \brief
*    Writes 4x4 intra prediction modes for a macroblock
************************************************************************
*/
int writeIntra4x4Modes(void)
{
  int i;
  Macroblock    *currMB     = &img->mb_data[img->current_mb_nr];
  SyntaxElement *currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
  int           *bitCount   = currMB->bitcounter;
  Slice         *currSlice  = img->currentSlice;
  const int     *partMap    = assignSE2partition[0];
  DataPartition *dataPart   = &(currSlice->partArr[partMap[SE_INTRAPREDMODE]]);

  int rate = 0;

  currMB->IntraChromaPredModeFlag = 1;

  for(i=0;i<16;i++)
  {
    currSE->context = i;
    currSE->value1  = currMB->intra_pred_modes[i];
    currSE->value2  = 0;

#if TRACE
    if (currSE->value1 < 0 )
      snprintf(currSE->tracestring, TRACESTRING_SIZE, "Intra 4x4 mode  = predicted (context: %d)",currSE->context);
    else
      snprintf(currSE->tracestring, TRACESTRING_SIZE, "Intra 4x4 mode  = %3d (context: %d)",currSE->value1,currSE->context);
#endif

    // set symbol type and function pointers
    currSE->type = SE_INTRAPREDMODE;

    // encode and update rate
    currSE->writing = writeIntraPredMode_CABAC;
    dataPart->writeSyntaxElement (currSE, dataPart);

    bitCount[BITS_MB_MODE]+=currSE->len;
    rate += currSE->len;
    currSE++;
    currMB->currSEnr++;
  }

  return rate;
}

/*!
************************************************************************
* \brief
*    Writes 8x8 intra prediction modes for a macroblock
************************************************************************
*/
int writeIntra8x8Modes(void)
{
  int block8x8;
  Macroblock    *currMB     = &img->mb_data[img->current_mb_nr];
  SyntaxElement *currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
  int           *bitCount   = currMB->bitcounter;
  Slice         *currSlice  = img->currentSlice;
  const int     *partMap    = assignSE2partition[0];
  DataPartition *dataPart   = &(currSlice->partArr[partMap[SE_INTRAPREDMODE]]);

  int rate = 0;

  currMB->IntraChromaPredModeFlag = 1;

  for(block8x8=0;block8x8<4;block8x8++)
  {

    currSE->context = block8x8<<2;
    currSE->value1  = currMB->intra_pred_modes8x8[(block8x8<<2)];
    currSE->value2  = 0;

#if TRACE
    if (currSE->value1 < 0 )
      snprintf(currSE->tracestring, TRACESTRING_SIZE, "Intra 8x8 mode  = predicted (context: %d)",currSE->context);
    else
      snprintf(currSE->tracestring, TRACESTRING_SIZE, "Intra 8x8 mode  = %3d (context: %d)",currSE->value1,currSE->context);
#endif

    // set symbol type and function pointers
    currSE->type = SE_INTRAPREDMODE;

    // encode and update rate
    currSE->writing = writeIntraPredMode_CABAC;
    dataPart->writeSyntaxElement (currSE, dataPart);

    bitCount[BITS_MB_MODE]+=currSE->len;
    rate += currSE->len;
    currSE++;
    currMB->currSEnr++;
  }

  return rate;
}

int writeIntraModes(void)
{
  switch (img->mb_data[img->current_mb_nr].mb_type)
  {
  case I4MB:
    return writeIntra4x4Modes();
    break;
  case I8MB:
    return writeIntra8x8Modes();
    break;
  default:
    return 0;
    break;
  }
}



/*!
************************************************************************
* \brief
*    Writes chrominance coefficients
************************************************************************
*/
int writeChromaCoeff (void)
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  int*            bitCount  = currMB->bitcounter;
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[0];
  int             cbp       = currMB->cbp;
  DataPartition*  dataPart;

  int   level, run;
  int   k, uv;
  int   b8, b4;
  int*  ACLevel;
  int*  ACRun;
  int*  DCLevel;
  int*  DCRun;
  //ADD-VG-14052004
  int   chroma_dc_context[3]={CHROMA_DC, CHROMA_DC_2x4, CHROMA_DC_4x4};
  int   yuv = img->yuv_format - 1;

  //=====
  //=====   D C - C O E F F I C I E N T S
  //=====
  if (cbp > 15)  // check if any chroma bits in coded block pattern is set
  {
    for (uv=0; uv < 2; uv++)
    {
      {

        DCLevel = img->cofDC[uv+1][0];
        DCRun   = img->cofDC[uv+1][1];

        level=1;
        for (k=0; k <= img->num_cdc_coeff && level != 0; ++k)
        {
          level = currSE->value1 = DCLevel[k]; // level
          run   = currSE->value2 = DCRun  [k]; // run

          currSE->writing = writeRunLevel_CABAC;

          currSE->context     = chroma_dc_context[yuv];
          currSE->type        = (IS_INTRA(currMB) ? SE_CHR_DC_INTRA : SE_CHR_DC_INTER);
          img->is_intra_block =  IS_INTRA(currMB);
          img->is_v_block     = uv;

          // choose the appropriate data partition
          dataPart = &(currSlice->partArr[partMap[currSE->type]]);
          dataPart->writeSyntaxElement (currSE, dataPart);
          bitCount[BITS_COEFF_UV_MB] += currSE->len;
          rate                       += currSE->len;
          // proceed to next SE 
          currSE++;  
          currMB->currSEnr++;
        }
      }
    }
  }

  //=====
  //=====   A C - C O E F F I C I E N T S
  //=====
  uv=-1;   
  if (cbp >> 4 == 2) // check if chroma bits in coded block pattern = 10b
  {  
    for (b8=4; b8 < (4+img->num_blk8x8_uv); b8++)
      for (b4=0; b4 < 4; b4++)
      {
        {

          ACLevel = img->cofAC[b8][b4][0];
          ACRun   = img->cofAC[b8][b4][1];

          level=1;
          uv++;

          img->subblock_y = subblk_offset_y[yuv][b8-4][b4]>>2;
          img->subblock_x = subblk_offset_x[yuv][b8-4][b4]>>2;

          for (k=0; k < 16 && level != 0; k++)
          {
            level = currSE->value1 = ACLevel[k]; // level
            run   = currSE->value2 = ACRun  [k]; // run

            currSE->writing = writeRunLevel_CABAC;

            currSE->context     = CHROMA_AC;
            currSE->type        = (IS_INTRA(currMB) ? SE_CHR_AC_INTRA : SE_CHR_AC_INTER);
            img->is_intra_block =  IS_INTRA(currMB);
            img->is_v_block     = (uv>=(img->num_blk8x8_uv<<1));

            // choose the appropriate data partition
            dataPart = &(currSlice->partArr[partMap[currSE->type]]);
            dataPart->writeSyntaxElement (currSE, dataPart);
            bitCount[BITS_COEFF_UV_MB] += currSE->len;
            rate                       += currSE->len;
#if TRACE
            snprintf(currSE->tracestring, TRACESTRING_SIZE, "AC Chroma %2d: level =%3d run =%2d",k, level, run);
#endif

            // proceed to next SE 
            currSE++;  
            currMB->currSEnr++;
          }
        }
      }
  }

  return rate;
}




/*!
************************************************************************
* \brief
*    Converts 8x8 block type to coding value
************************************************************************
*/
int B8Mode2Value (int b8mode, int b8pdir)
{
  static const int b8start[8] = {0,0,0,0, 1, 4, 5, 10};
  static const int b8inc  [8] = {0,0,0,0, 1, 2, 2, 1};

  if (img->type!=B_SLICE)
  {
    return (b8mode-4);
  }
  else
  {
    return b8start[b8mode] + b8inc[b8mode] * b8pdir;
  }
}

/*!
************************************************************************
* \brief
*    Codes macroblock header
* \param rdopt
*    true for calls during RD-optimization
* \param coeff_rate
*    bitrate of Luma and Chroma coeff
************************************************************************
*/
int writeMBLayer (int rdopt, int *coeff_rate)
{
  int             i;
  int             mb_nr      = img->current_mb_nr;
  Macroblock*     currMB     = &img->mb_data[mb_nr];
  SyntaxElement  *currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
  int*            bitCount   = currMB->bitcounter;
  Slice*          currSlice  = img->currentSlice;
  DataPartition*  dataPart;
  const int*      partMap    = assignSE2partition[0];
  int             no_bits    = 0;
  int             mb_type;

  currMB->IntraChromaPredModeFlag = IS_INTRA(currMB);

  // choose the appropriate data partition
  dataPart = &(currSlice->partArr[partMap[SE_MBTYPE]]);

  if(img->type == I_SLICE)
  {
    //========= write mb_type (I_SLICE) =========
    currSE->value1  = MBType2Value (currMB);
    currSE->value2  = 0;
    currSE->type    = SE_MBTYPE;
    currSE->writing = writeMB_typeInfo_CABAC;
    dataPart->writeSyntaxElement( currSE, dataPart);
#if TRACE
    snprintf(currSE->tracestring, TRACESTRING_SIZE,   "mb_type (I_SLICE) (%2d,%2d) = %3d",img->mb_x, img->mb_y, currMB->mb_type);
#endif
    bitCount[BITS_MB_MODE] += currSE->len;
    no_bits                += currSE->len;
    currSE++;
    currMB->currSEnr++;

  }
  // not I_SLICE, CABAC
  else
  {

#if (!defined TEST_ONLY_Z_SCAN)
    if(MB32.write_mb32_mbtype)     //this is 1 for P8x8 mode in bigblocks
    {
      Macroblock mb32x32;
      mb32x32.mb_type = P8x8;

#ifdef GEOM
      mb32x32.geo_mode = -1;
#endif
      no_bits += write_mb_type(&mb32x32, 1);
      currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
    }
#endif


#ifdef GEOM
    if(currMB->geo_mode != -1)
      mb_type         = MBType2Value_geo (currMB);
    else
#endif
      mb_type         = MBType2Value (currMB);

    //========= write mb_skip_flag (CABAC) =========
    currSE->value1  = mb_type;
    currSE->value2  = currMB->cbp;
    currSE->type    = SE_MBTYPE;
    currSE->writing = writeMB_skip_flagInfo_CABAC;
    dataPart->writeSyntaxElement( currSE, dataPart);
#if TRACE
    snprintf(currSE->tracestring, TRACESTRING_SIZE, "mb_skip_flag");
#endif
    bitCount[BITS_MB_MODE] += currSE->len;
    no_bits                += currSE->len;
    currSE++;
    currMB->currSEnr++;

    CheckAvailabilityOfNeighborsCABAC();

    //========= write mb_type (CABAC) =========
    //mbtype!=0 for other modes and cbp!=0 is for P/B-Direct mode
    if (currMB->mb_type != 0 || currMB->cbp != 0)
    {
      currSE->value1  = mb_type;
      currSE->value2  = 0;
      currSE->type    = SE_MBTYPE;
#ifdef GEOM      
      if(currMB->geo_mode != -1)
      {
        //mbtype = 3;//use B_8x16
        currSE->dir_partition1  = best8x8pdir_geo[currMB->geo_mode][0];
        currSE->dir_partition2  = best8x8pdir_geo[currMB->geo_mode][1];
      }
      else
#endif
      {
        currSE->dir_partition1 = currMB->b8pdir[0];//this is for Partition 1
        currSE->dir_partition2 = currMB->b8pdir[3];//this is for Partition 2
      }
      currSE->writing = writeMB_typeInfo_CABAC;

      dataPart->writeSyntaxElement( currSE, dataPart);
#if TRACE
      if (img->type == B_SLICE) 
        snprintf(currSE->tracestring, TRACESTRING_SIZE, "mb_type (B_SLICE) (%2d,%2d) = %3d",img->mb_x, img->mb_y, currMB->mb_type);
      else                      
        snprintf(currSE->tracestring, TRACESTRING_SIZE, "mb_type (P_SLICE) (%2d,%2d) = %3d",img->mb_x, img->mb_y, currMB->mb_type);
#endif
      bitCount[BITS_MB_MODE] += currSE->len;
      no_bits                += currSE->len;
      currSE++;
      currMB->currSEnr++;

#ifdef GEOM
      if(currMB->mb_type == 3 || currMB->geo_mode != -1)
      {
        currSE->value1 = currMB->geo_mode;
        currSE->value2  = rho_bits[0]+THETA_BITS;
        currSE->writing = writeMB_geoInfo_CABAC;
        dataPart->writeSyntaxElement( currSE, dataPart);
#if TRACE
        snprintf(currSE->tracestring, TRACESTRING_SIZE, "mb_skip_flag");
#endif
        bitCount[BITS_MB_MODE] += currSE->len;
        no_bits                += currSE->len;
        currSE++;
        currMB->currSEnr++;
      }
#endif
    }
  }

#ifdef MV_COMPETITION
  if (input->mv_competition > 0)
  {
    if(rdopt==0)  
    {  
      if ((currMB->mb_type == 0) && (img->type==P_SLICE) && (mv_comp.nb_mode_for_skip > 1))
        no_bits  += write_predictor_index_for_skip_mode();
    }
  }
#endif


  //init NoMbPartLessThan8x8Flag
  currMB->NoMbPartLessThan8x8Flag = (IS_DIRECT(currMB) && !(active_sps->direct_8x8_inference_flag))? 0: 1;

  //===== BITS FOR 8x8 SUB-PARTITION MODES =====
  if (IS_P8x8 (currMB))
  {
    dataPart = &(currSlice->partArr[partMap[SE_MBTYPE]]);

    for (i=0; i<4; i++)
    {
      currSE->dir_partition1 = currMB->b8pdir[i];
      currSE->writing = writeB8_typeInfo_CABAC;

      currSE->value1  = B8Mode2Value (currMB->b8mode[i], currMB->b8pdir[i]);
      currSE->value2  = 0;
      currSE->type    = SE_MBTYPE;
      dataPart->writeSyntaxElement (currSE, dataPart);
      bitCount[BITS_MB_MODE]+= currSE->len;
      no_bits               += currSE->len;
      currSE++;
      currMB->currSEnr++;

      //set NoMbPartLessThan8x8Flag for P8x8 mode
      currMB->NoMbPartLessThan8x8Flag &= (currMB->b8mode[i]==0 && active_sps->direct_8x8_inference_flag) || (currMB->b8mode[i]==4);
    }

#ifdef MV_COMPETITION
    if (input->mv_competition > 0)
      pass_with_writing = !rdopt;
#endif

    no_bits += writeMotionInfo2NAL  (0);
    currSE   = &img->MB_SyntaxElements[currMB->currSEnr];
  }

  //============= Transform size flag for INTRA MBs =============
  //-------------------------------------------------------------
  //transform size flag for INTRA_4x4 and INTRA_8x8 modes
  if ((currMB->mb_type == I8MB || currMB->mb_type == I4MB) && input->Transform8x8Mode)
  {
    currSE->value1 = currMB->luma_transform_size_8x8_flag;
    currSE->type   = SE_HEADER;
    currSE->writing = writeMB_transform_size_CABAC;
    dataPart->writeSyntaxElement(currSE, dataPart);
    bitCount[BITS_MB_MODE] += currSE->len;
    no_bits                += currSE->len;
    currSE++;
    currMB->currSEnr++;
  }


  //===== BITS FOR INTRA PREDICTION MODES ====
  no_bits += writeIntraModes();
  //===== BITS FOR CHROMA INTRA PREDICTION MODE ====
  if (currMB->IntraChromaPredModeFlag && img->yuv_format != YUV400)
    no_bits += writeChromaIntraPredMode();
  else if(!rdopt) //GB CHROMA !!!!!
    currMB->c_ipred_mode = DC_PRED_8; //setting c_ipred_mode to default is not the right place here
  //resetting in rdopt.c (but where ??)
  //with cabac and bframes maybe it could crash without this default
  //since cabac needs the right neighborhood for the later MBs

  //----- motion information -----
  if (currMB->mb_type !=0 && currMB->mb_type !=P8x8)
  {
#ifdef MV_COMPETITION
    if (input->mv_competition > 0)
      pass_with_writing = !rdopt;
#endif

#ifdef GEOM
    if(currMB->geo_mode != -1)
      no_bits  += writeMotionInfo2NAL_geo  (currMB->geo_mode, 0);
    else
#endif
      no_bits  += writeMotionInfo2NAL  (0);
  }



  //mbtype!=0 for other modes and cbp!=0 is for P/B-Direct mode
  if (currMB->mb_type != 0 || currMB->cbp != 0)
  {
    no_bits    += writeCBP_Transform_size (rdopt, 0);
    no_bits    += writeDeltaQP            (rdopt, img->mb_data[img->current_mb_nr].cbp);
    *coeff_rate = writeLumaCoeff          (rdopt, 0);

    if (img->yuv_format != YUV400)
      *coeff_rate  += writeChromaCoeff ();

    no_bits  += *coeff_rate;
  }

  return no_bits;
}

void write_terminating_bit (short bit)
{
  DataPartition*          dataPart;
  const int*              partMap   = assignSE2partition[0];
  EncodingEnvironmentPtr  eep_dp;

  //--- write non-slice termination symbol if the macroblock is not the first one in its slice ---
  dataPart = &(img->currentSlice->partArr[partMap[SE_MBTYPE]]);
  dataPart->bitstream->write_flag = 1;
  eep_dp                          = &(dataPart->ee_cabac);
  biari_encode_symbol_final(eep_dp, bit); 
}


/*!
************************************************************************
* \brief
*    Write chroma intra prediction mode.
************************************************************************
*/
int writeChromaIntraPredMode(void)
{
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  int*            bitCount  = currMB->bitcounter;
  const int*      partMap   = assignSE2partition[0];
  int             rate      = 0;
  DataPartition*  dataPart;

  //===== BITS FOR CHROMA INTRA PREDICTION MODES
  currSE->writing = writeCIPredMode_CABAC;
  currSE->value1 = currMB->c_ipred_mode;
  currSE->value2 = 0;
  currSE->type = SE_INTRAPREDMODE;
  dataPart = &(currSlice->partArr[partMap[SE_INTRAPREDMODE]]);

  dataPart->writeSyntaxElement (currSE, dataPart);
  bitCount[BITS_MB_MODE] += currSE->len;
  rate                    += currSE->len;
  currSE++;
  currMB->currSEnr++;

  return rate;
}


/*!
************************************************************************
* \brief
*    Set global last_dquant according to macroblock delta qp
************************************************************************
*/

extern int last_dquant;

void set_last_dquant(void)
{
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];

  if (IS_INTERMV (currMB) || IS_INTRA (currMB) || ((img->type!=I_SLICE)  && currMB->cbp != 0))
  {
    // non-skip
    last_dquant = currMB->delta_qp;
  }
  else
  {
    // skip
    last_dquant = 0;
  }
}


/*!
************************************************************************
* \brief
*    Passes the chosen syntax elements to the NAL
************************************************************************
*/
void write_one_macroblock (int eos_bit)
{
  Macroblock* currMB   = &img->mb_data[img->current_mb_nr];
  int*        bitCount = currMB->bitcounter;
  int i;

  extern int cabac_encoding;

  //--- constrain intra prediction ---
  if(input->UseConstrainedIntraPred && (img->type==P_SLICE || img->type==B_SLICE))
  {
    if( !IS_INTRA(currMB) )
    {
      img->intra_block[img->current_mb_nr] = 0;
    }
    else
    {
      img->intra_block[img->current_mb_nr] = 1;
    }
  }

  //===== init and update number of intra macroblocks =====
  if (img->current_mb_nr==0)
    intras=0;

  if (IS_INTRA(currMB))
    intras++;

  //--- write non-slice termination symbol if the macroblock is not the first one in its slice ---
  if (img->current_mb_nr!=img->currentSlice->start_mb_nr && eos_bit)
  {
    write_terminating_bit (0);
  }

  cabac_encoding = 1;

  //--- write macroblock ---
  writeMBLayer (0, &i);  // i is temporary

  if (!((currMB->mb_type !=0 ) || ((img->type==B_SLICE) && currMB->cbp != 0) ))
  { 
    for (i=0; i < 4; i++)
      memset(img->nz_coeff [img->current_mb_nr][i], 0, (4 + img->num_blk8x8_uv) * sizeof(int));  // CAVLC
  }

  set_last_dquant();

  //--- set total bit-counter ---

#ifdef MV_COMPETITION
  if (input->mv_competition > 0)
  {
    bitCount[BITS_TOTAL_MB] = bitCount[BITS_MB_MODE]  + bitCount[BITS_COEFF_Y_MB]     
    + bitCount[BITS_INTER_MB] + bitCount[BITS_CBP_MB]  
    + bitCount[BITS_DELTA_QUANT_MB] + bitCount[BITS_COEFF_UV_MB]
    + bitCount[BITS_SKIP_PRED]
    + bitCount[BITS_MV_PRED];
  }
  else
#endif
  {
    bitCount[BITS_TOTAL_MB] = bitCount[BITS_MB_MODE]  + bitCount[BITS_COEFF_Y_MB]     
    + bitCount[BITS_INTER_MB] + bitCount[BITS_CBP_MB]  
    + bitCount[BITS_DELTA_QUANT_MB] + bitCount[BITS_COEFF_UV_MB];
  }  
  ;

  stats->bit_slice += bitCount[BITS_TOTAL_MB];

  cabac_encoding = 0;
}


/*!
************************************************************************
* \brief
*    Sets context for reference frame parameter
************************************************************************
*/
int BType2CtxRef (int btype)
{
  return (btype<4 ? 0 : 1);
}

/*!
************************************************************************
* \brief
*    Codes the reference frame
************************************************************************
*/
int writeReferenceFrame (int mode, int i, int j, int fwd_flag, int  ref, int  mb_ext_level)
{
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  int*            bitCount  = currMB->bitcounter;
  const int*      partMap   = assignSE2partition[0];
  int             rate      = 0;
  DataPartition*  dataPart;
  int             num_ref   = ( fwd_flag ? listXsize[LIST_0+currMB->list_offset]: listXsize[LIST_1+currMB->list_offset]);
  int             flag_mode = 0;

  if( num_ref == 1 )
    return 0;

  if ( num_ref == 2 )
    flag_mode = 1;

  currSE->value1 = ref;
  currSE->value2  = 0;
  currSE->type   = SE_REFFRAME;

  dataPart = &(currSlice->partArr[partMap[currSE->type]]);
  currSE->context = BType2CtxRef (mode);
  img->subblock_x = i; // position used for context determination
  img->subblock_y = j; // position used for context determination
  currSE->mb_ext_level = mb_ext_level;
  currSE->writing = writeRefFrame_CABAC;
  currSE->value2 = (fwd_flag)? LIST_0:LIST_1;
  dataPart->writeSyntaxElement (currSE, dataPart);

  bitCount[BITS_INTER_MB] += currSE->len;
  rate                    += currSE->len;
  currSE++;
  currMB->currSEnr++;

  return rate;
}

/*!
************************************************************************
* \brief
*    Writes motion vectors of an 16x16 block
************************************************************************
*/
int writeMotionVector8x8 (int  i0,
                          int  j0,
                          int  i1,
                          int  j1,
                          int  refframe,
                          int  list_idx,
                          int  mv_mode,
                          int  mb_ext_level)
{
  int            i, j, k, l, m;
  int            curr_mvd;
  DataPartition* dataPart;

  int            rate       = 0;
  int            step_h     = input->part_size[mv_mode][0] << mb_ext_level;
  int            step_v     = input->part_size[mv_mode][1] << mb_ext_level;
  Macroblock*    currMB     = &img->mb_data[img->current_mb_nr];
  SyntaxElement* currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*         currSlice  = img->currentSlice;
  int*           bitCount   = currMB->bitcounter;
  const int*     partMap    = assignSE2partition[0];
  int            refindex   = refframe;

  short******    all_mv     = img->all_mv;
  short******    pred_mv    = img->pred_mv;

  if (currMB->bi_pred_me && currMB->b8pdir[0]==2 && mv_mode == 1 && refindex == 0)
    all_mv = currMB->bi_pred_me == 1? img->bipred_mv1 : img->bipred_mv2 ;  

  for (j=j0; j<j1; j+=step_v)
  {
    for (i=i0; i<i1; i+=step_h)
    {
#ifdef ADAPTIVE_MV_RES
      int mv_shift =0; //default to 1/8 pel
      int currMvVec;
      int predMvVec;
#endif	


#ifdef MV_COMPETITION
      if((input->Transform8x8Mode)&&(pass_with_writing == 1) && (input->mv_competition > 0) &&(!input->UseRDO_Q))
      {
        short best_prediction[2];
        SetMotionVectorPredictor_Competition (best_prediction, enc_picture->ref_idx[list_idx], enc_picture->mv[list_idx], (short)refindex, list_idx, i, j,(input->blc_size[mv_mode][0]<<mb_ext_level),(input->blc_size[mv_mode][1]<<mb_ext_level),mb_ext_level);
        GetBestPredVector(best_prediction, all_mv[j][i][list_idx][refindex][mv_mode][0],all_mv[j][i][list_idx][refindex][mv_mode][1],img->lambda_mf[img->type][img->qp],refindex,list_idx,mv_mode,i*4,j*4); 
        pred_mv[j][i][list_idx][refindex][mv_mode][0]=best_prediction[0];
        pred_mv[j][i][list_idx][refindex][mv_mode][1]=best_prediction[1];
      }
#endif


#ifdef RDO_Q
      if(input->UseRDO_Q)
      {
#ifdef MV_COMPETITION
        if( (input->mv_competition > 0))    
        {
          short best_prediction[2];
          SetMotionVectorPredictor_Competition (best_prediction, enc_picture->ref_idx[list_idx], enc_picture->mv[list_idx], (short)refindex, list_idx, i, j,(input->blc_size[mv_mode][0]<<mb_ext_level),(input->blc_size[mv_mode][1]<<mb_ext_level),mb_ext_level);
          GetBestPredVector(best_prediction, all_mv[j][i][list_idx][refindex][mv_mode][0],all_mv[j][i][list_idx][refindex][mv_mode][1],img->lambda_mf[img->type][img->qp],refindex,list_idx,mv_mode,i*4,j*4); 
          pred_mv[j][i][list_idx][refindex][mv_mode][0]=best_prediction[0];
          pred_mv[j][i][list_idx][refindex][mv_mode][1]=best_prediction[1];
        }
        else
        {
#endif
          short pred_mv_stored[2];
          SetMotionVectorPredictor (pred_mv_stored, enc_picture->ref_idx[list_idx], enc_picture->mv[list_idx], (short)refindex, list_idx, i, j, (input->blc_size[mv_mode][0]<<mb_ext_level),(input->blc_size[mv_mode][1]<<mb_ext_level), mb_ext_level);
          pred_mv[j][i][list_idx][refindex][mv_mode][0]=pred_mv_stored[0];
          pred_mv[j][i][list_idx][refindex][mv_mode][1]=pred_mv_stored[1];
#ifdef MV_COMPETITION
        }  
#endif
      }
#endif   //RDO_Q

#ifdef ADAPTIVE_MV_RES
      if (img->mv_res)
      { 
        if (((all_mv[j][i][list_idx][refindex][mv_mode][0]& 1) ==0)&&((all_mv[j][i][list_idx][refindex][mv_mode][1]& 1))==0)
          mv_shift =1; //switch to 1/4 pel

        img->subblock_x = i; // position used for context determination
        img->subblock_y = j; // position used for context determination
        currSE->value1 = mv_shift;
        currSE->type   = SE_MVD; 
        currSE->mb_ext_level  = mb_ext_level;
        currSE->value2 = 0;
        currSE->writing  = writeMVRES_4_8_CABAC;
        dataPart = &(currSlice->partArr[partMap[SE_MVD]]);
        dataPart->writeSyntaxElement (currSE, dataPart);
        bitCount[BITS_INTER_MB] += currSE->len;
        rate                    += currSE->len;
        currSE++;  
        currMB->currSEnr++;
      }
      else
      {
        mv_shift=0;
      }
#endif

      for (k=0; k<2; k++) 
      {        

#ifdef ADAPTIVE_MV_RES
        if (img->mv_res && mv_shift )
        {
          int scaled_mvd ;
          currMvVec=roundInt(all_mv[j][i][list_idx][refindex][mv_mode][k],mv_shift);
          predMvVec=roundInt(pred_mv[j][i][list_idx][refindex][mv_mode][k],mv_shift);
          curr_mvd = currMvVec - predMvVec;
          scaled_mvd = curr_mvd*(1<< mv_shift);

          if(mb_ext_level)
          {
            //--- store (oversampled) mvd ---
            //only save mvd for the upper left 16x16 block for the context of the 2nd partition of 32x16 and 16x32        
            if(j==0 && i==0)
            {
              for (l=0; l < 4; l++)
                for (m=0; m < 4; m++)
                {
                  currMB->mvd[list_idx][j+l][i+m][k] = scaled_mvd; //mvd[2][BLOCK_MULTIPLE][BLOCK_MULTIPLE][2]
                }
            }
            //the mvd for the whole 32x32 is saved in mb32_mvd and will be copied to each 16x16 block's mvd in write_macroblock_cluster
            for (l=0; l < step_v; l++)
              for (m=0; m < step_h; m++)
              {
                mb32_mvd[list_idx][j+l][i+m][k] = scaled_mvd;
              }
          }
          else
          {
            //--- store (oversampled) mvd ---
            for (l=0; l < step_v; l++)
              for (m=0; m < step_h; m++)
              {
                currMB->mvd[list_idx][j+l][i+m][k] = scaled_mvd;
              }
          }
        }
        else
#endif
        {
          curr_mvd = all_mv[j][i][list_idx][refindex][mv_mode][k] - pred_mv[j][i][list_idx][refindex][mv_mode][k];
          if(mb_ext_level)
          {	        
            //--- store (oversampled) mvd ---
            //only save mvd for the upper left 16x16 block for the context of the 2nd partition of 32x16 and 16x32        
            if(j==0 && i==0)
            {
              for (l=0; l < 4; l++)
                for (m=0; m < 4; m++)
                {
                  currMB->mvd[list_idx][j+l][i+m][k] = curr_mvd; //mvd[2][BLOCK_MULTIPLE][BLOCK_MULTIPLE][2]
                }
            }
            //the mvd for the whole 32x32 is saved in mb32_mvd and will be copied to each 16x16 block's mvd in write_macroblock_cluster
            for (l=0; l < step_v; l++)
              for (m=0; m < step_h; m++)
              {
                mb32_mvd[list_idx][j+l][i+m][k] = curr_mvd;
              }
          }
          else
          {
            //--- store (oversampled) mvd ---
            for (l=0; l < step_v; l++)
              for (m=0; m < step_h; m++)
              {
                currMB->mvd[list_idx][j+l][i+m][k] = curr_mvd;
              }
          }
        }
        currSE->value1 = curr_mvd;
        currSE->value2 = 0;
        currSE->type   = SE_MVD;
        img->subblock_x = i; // position used for context determination
        img->subblock_y = j; // position used for context determination
#ifdef ADAPTIVE_MV_RES
        currSE->value2  = (2*k+list_idx)*2+mv_shift;
#else
        currSE->value2  = 2*k+list_idx; // identifies the component and the direction; only used for context determination
#endif
        currSE->mb_ext_level  = mb_ext_level;
        currSE->writing = writeMVD_CABAC;

        dataPart = &(currSlice->partArr[partMap[SE_MVD]]);
        dataPart->writeSyntaxElement (currSE, dataPart);

        bitCount[BITS_INTER_MB] += currSE->len;
        rate                    += currSE->len;
        currSE++;  
        currMB->currSEnr++;
      }
#ifdef MV_COMPETITION
      if (input->mv_competition > 0)
      { 
        int maxmode=0;
        if (img->type == P_SLICE)
          maxmode = mv_comp.nb_mode_for_mvp;
        else if (img->type == B_SLICE)
          maxmode = mv_comp.nb_mode_for_mvb;
        //----- reverse logic....0 means "send it", 1 means "dont send"
        if (send_index_mv_prediction[mv_mode][refindex][list_idx][j][i][maxmode] == 0)
          rate+=writeIndexForMotionVectorPrediction (refframe, list_idx, mv_mode, i, j);
      }
#endif
    }
  }
  return rate;
}

/*!
************************************************************************
* \brief
*    Writes motion info
************************************************************************
*/
int writeMotionInfo2NAL (int mb_ext_level)
{
  int k, j0, i0, refframe;
  int jj;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  int             no_bits   = 0;
  int   bframe          = (img->type==B_SLICE);
  int   step_h0         = ((input->blc_size[IS_P8x8(currMB) ? 4 : currMB->mb_type][0]<<mb_ext_level) >> 2);
  int   step_v0         = ((input->blc_size[IS_P8x8(currMB) ? 4 : currMB->mb_type][1]<<mb_ext_level) >> 2);

  //=== If multiple ref. frames, write reference frame for the MB ===
  if (IS_INTERMV (currMB))
  {
    for (j0=0; j0<(4<<mb_ext_level); j0+=step_v0)
    {
      jj = img->block_y+j0;
      for (i0=0; i0<(4<<mb_ext_level); i0+=step_h0)
      {
        k=(j0>>mb_ext_level)+(i0 >> (1+mb_ext_level));
        if ((currMB->b8pdir[k]==0 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)//has forward vector
        {
          no_bits += writeReferenceFrame (currMB->b8mode[k], i0, j0, 1, enc_picture->ref_idx[LIST_0][jj][img->block_x+i0], mb_ext_level);
        }
      }
    }

    for (j0=0; j0<(4<<mb_ext_level); j0+=step_v0)
    {
      jj = img->block_y+j0;
      for (i0=0; i0<(4<<mb_ext_level); i0+=step_h0)
      {
        k=(j0>>mb_ext_level)+(i0 >> (1+mb_ext_level));
        if ((currMB->b8pdir[k]==1 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)//has backward vector
        {
          no_bits += writeReferenceFrame (currMB->b8mode[k], i0, j0, 0, enc_picture->ref_idx[LIST_1][jj][img->block_x+i0], mb_ext_level);
        }
      }
    }
  }

  //===== write forward motion vectors =====
  if (IS_INTERMV (currMB))
  {
    for (j0=0; j0<(4<<mb_ext_level); j0+=step_v0)
    {
      jj = img->block_y+j0;
      for (i0=0; i0<(4<<mb_ext_level); i0+=step_h0)
      {
        k=(j0>>mb_ext_level)+(i0 >> (1+mb_ext_level));
        if ((currMB->b8pdir[k]==0 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)//has forward vector
        {
          refframe  = enc_picture->ref_idx[LIST_0][jj][img->block_x+i0];
          no_bits  += writeMotionVector8x8 (i0, j0, i0+step_h0, j0+step_v0, refframe, LIST_0, currMB->b8mode[k], mb_ext_level);
        }
      }
    }
  }


  //===== write backward motion vectors =====
  if (IS_INTERMV (currMB) && bframe)
  {
    for (j0=0; j0<(4<<mb_ext_level); j0+=step_v0)
    {
      jj = img->block_y+j0;
      for (i0=0; i0<(4<<mb_ext_level); i0+=step_h0)
      {
        k=(j0>>mb_ext_level)+(i0 >> (1+mb_ext_level));
        if ((currMB->b8pdir[k]==1 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)//has backward vector
        {
          refframe  = enc_picture->ref_idx[LIST_1][jj][img->block_x+i0];
          no_bits  += writeMotionVector8x8 (i0, j0, i0+step_h0, j0+step_v0, refframe, LIST_1, currMB->b8mode[k], mb_ext_level);
        }
      }
    }
  }
  return no_bits;
}


/*!
************************************************************************
* \brief
*    Writes Luma coeff of an 4x4 block
************************************************************************
*/
int writeLumaCoeff4x4_CABAC (int b8, int b4, int intra4x4mode)
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[0];
  int*            bitCount  = currMB->bitcounter;
  DataPartition*  dataPart;

  int   level, run;
  int   k;
  int*  ACLevel = img->cofAC[b8][b4][0];
  int*  ACRun   = img->cofAC[b8][b4][1];

  img->subblock_x = ((b8&0x1)==0) ? (((b4&0x1)==0)? 0: 1) : (((b4&0x1)==0)? 2: 3); // horiz. position for coeff_count context
  img->subblock_y = (b8<2)        ? ((b4<2)       ? 0: 1) : ((b4<2)       ? 2: 3); // vert.  position for coeff_count context

  level=1; // get inside loop
  for(k=0; k<=16 && level !=0; k++)
  {
    level = currSE->value1 = ACLevel[k]; // level
    run   = currSE->value2 = ACRun  [k]; // run

    currSE->writing = writeRunLevel_CABAC;

    currSE->context     = LUMA_4x4;
    currSE->type        = (k==0 
      ? (intra4x4mode ? SE_LUM_DC_INTRA : SE_LUM_DC_INTER) 
      : (intra4x4mode ? SE_LUM_AC_INTRA : SE_LUM_AC_INTER));
    img->is_intra_block = intra4x4mode;

    // choose the appropriate data partition
    dataPart = &(currSlice->partArr[partMap[currSE->type]]);          
    dataPart->writeSyntaxElement (currSE, dataPart);
    bitCount[BITS_COEFF_Y_MB] += currSE->len;
    rate                      += currSE->len;
#if TRACE
    snprintf(currSE->tracestring, TRACESTRING_SIZE, "Luma sng(%2d) level =%3d run =%2d", k, level,run);
#endif
    /* proceed to next SE */
    currSE++;  
    currMB->currSEnr++;

  }

  return rate;
}

/*!
************************************************************************
* \brief
*    Writes Luma coeff of an 8x8 block
************************************************************************
*/
int writeLumaCoeff8x8_CABAC (int b8, int intra_mode)
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[0];
  int*            bitCount  = currMB->bitcounter;
  DataPartition*  dataPart;

  int   level, run;
  int   k;
  int*  ACLevel = img->cofAC[b8][0][0];
  int*  ACRun   = img->cofAC[b8][0][1];

  img->subblock_x = ((b8&0x1)==0)?0:2;  // horiz. position for coeff_count context
  img->subblock_y = (b8<2)?0:2;     // vert.  position for coeff_count context


  level=1; // get inside loop
  for(k=0; k<=64 && level !=0; k++)
  {
    level = currSE->value1 = ACLevel[k]; // level
    run   = currSE->value2 = ACRun  [k]; // run

    currSE->writing = writeRunLevel_CABAC;

    currSE->context     = LUMA_8x8;
    currSE->type        = (k==0 
      ? (intra_mode ? SE_LUM_DC_INTRA : SE_LUM_DC_INTER) 
      : (intra_mode ? SE_LUM_AC_INTRA : SE_LUM_AC_INTER));
    img->is_intra_block = intra_mode;

    // choose the appropriate data partition
    dataPart = &(currSlice->partArr[partMap[img->type != B_SLICE ? currSE->type : SE_BFRAME]]);
    dataPart->writeSyntaxElement (currSE, dataPart);
    bitCount[BITS_COEFF_Y_MB] += currSE->len;
    rate                      += currSE->len;
#if TRACE
    snprintf(currSE->tracestring, TRACESTRING_SIZE, "Luma8x8 sng(%2d) level =%3d run =%2d", k, level,run);
#endif
    /* proceed to next SE */
    currSE++;  
    currMB->currSEnr++;

  }

  return rate;
}

/*!
************************************************************************
* \brief
*    Writes Luma Coeff of an 8x8 block
************************************************************************
*/
int writeLumaCoeff8x8 (int block8x8, int block_mode, int transform_size_flag)
{
  int  block4x4, rate = 0;
  int intra4x4mode = (block_mode==IBLOCK);

  if (block_mode == I8MB)
    assert(transform_size_flag == 1);


  if(!transform_size_flag) // allow here if 4x4 
  {
    for (block4x4=0; block4x4<4; block4x4++)  
      rate += writeLumaCoeff4x4_CABAC (block8x8, block4x4, intra4x4mode);
  }
  else 
    rate += writeLumaCoeff8x8_CABAC (block8x8, (block_mode == I8MB));

  return rate;
}

#ifdef USE_INTRA_MDDT
/*! 
*************************************************************************************
* \brief
*   write the luma coefficients of an INTRA16x16 macroblock
*
* \para writeLumaCoeff16x16_CABAC()
*    <paragraph>
*
* \author
*    - Yan Ye                      <yye@qualcomm.com>
*************************************************************************************
*/
int writeLumaCoeff16x16_CABAC ()
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[0];
  int*            bitCount  = currMB->bitcounter;
  DataPartition*  dataPart;

  int   level, run;
  int   k;
  int*  ACLevel = img->cofAC16[0];
  int*  ACRun   = img->cofAC16[1];

  img->subblock_x = 0;//img->pix_x;  // horiz. position for coeff_count context
  img->subblock_y = 0;//img->pix_y;  // vert.  position for coeff_count context

  level=1; // get inside loop
  for(k=0; k<=256 && level !=0; k++)
  {
    level = currSE->value1 = ACLevel[k]; // level
    run   = currSE->value2 = ACRun  [k]; // run

    currSE->writing = writeRunLevel16x16_CABAC;

    currSE->context     = LUMA_16x16;
    currSE->type        = SE_LUM_AC_INTRA;
    img->is_intra_block = 1;    // only applicable to intra blocks

    // choose the appropriate data partition
    dataPart = &(currSlice->partArr[partMap[img->type != B_SLICE ? currSE->type : SE_BFRAME]]);
    dataPart->writeSyntaxElement (currSE, dataPart);
    bitCount[BITS_COEFF_Y_MB] += currSE->len;
    rate                      += currSE->len;
#if TRACE
    snprintf(currSE->tracestring, TRACESTRING_SIZE, "Luma8x8 sng(%2d) level =%3d run =%2d", k, level,run);
#endif
    /* proceed to next SE */
    currSE++;  
    currMB->currSEnr++;

  }

  return rate;
}

int writeLumaCoeff16x16 ( )
{
  int  rate = 0;
  rate += writeLumaCoeff16x16_CABAC ();
  return rate;
}
#endif 

int writeCBP_Big(int rdopt, int cbp)
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  int*            bitCount  = currMB->bitcounter;
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[0];
  DataPartition*  dataPart;


  //=====   C B P   =====
  //---------------------
  currSE->value1 = cbp;
  currSE->writing = writeCBP_CABAC32;
  // choose the appropriate data partition
  dataPart = &(currSlice->partArr[partMap[currSE->type]]);
  dataPart->writeSyntaxElement(currSE, dataPart);

  bitCount[BITS_CBP_MB] += currSE->len;
  rate                  += currSE->len;
  // proceed to next SE
  currSE++;  
  currMB->currSEnr++;     

  return rate;
}


int writeDeltaQP(int rdopt, int cbp)  
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  int*            bitCount  = currMB->bitcounter;
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  //Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[0];
  DataPartition*  dataPart;

  if(cbp != 0 || IS_NEWINTRA (currMB))
  {
    currSE->value1 = currMB->delta_qp;
    currSE->writing = writeDquant_CABAC;

    if (IS_INTER (currMB))  currSE->type = SE_DELTA_QUANT_INTER;
    else                    currSE->type = SE_DELTA_QUANT_INTRA;

    // choose the appropriate data partition
    dataPart = &(img->currentSlice->partArr[partMap[currSE->type]]);
    dataPart->writeSyntaxElement(  currSE, dataPart);
    bitCount[BITS_DELTA_QUANT_MB] += currSE->len;
    rate                          += currSE->len;
#if TRACE
    snprintf(currSE->tracestring, TRACESTRING_SIZE, "Delta QP (%2d,%2d) = %3d",img->mb_x, img->mb_y, currMB->delta_qp);
#endif
    // proceed to next SE
    currSE++;
    currMB->currSEnr++;
  }
  return rate;
}

int writeCBP_Transform_size (int rdopt, int mb32)
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  int*            bitCount  = currMB->bitcounter;
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[0];
  int             cbp       = currMB->cbp;
  DataPartition*  dataPart;
  int             need_transform_size_flag;   //ADD-VG-24062004

  if (!IS_NEWINTRA (currMB))
  {
    //need_transform_size_flag is different from the original defination
    need_transform_size_flag = ((mb32 || (currMB->mb_type >= 0 && currMB->mb_type <= 3)||
      (IS_DIRECT(currMB) && active_sps->direct_8x8_inference_flag) ||
      (currMB->NoMbPartLessThan8x8Flag))
      && currMB->mb_type != I8MB && currMB->mb_type != I4MB
      //&& (currMB->cbp&15)  - !!!
      && input->Transform8x8Mode);

    if(need_transform_size_flag)
    {
      rate += writelumaCBPTransform_size(mb32);
    }

    if(   need_transform_size_flag 
#ifdef GEOM
      &&  ((currMB->cbp&15) == 0 || (currMB->luma_transform_size_8x8_flag==2 && (mb32 || currMB->mb_type == 1 || currMB->geo_mode!=-1 || (IS_DIRECT(currMB) && active_sps->direct_8x8_inference_flag)))))
#else
      &&  ((currMB->cbp&15) == 0 || (currMB->luma_transform_size_8x8_flag==2 && (mb32 || currMB->mb_type == 1 || (IS_DIRECT(currMB) && active_sps->direct_8x8_inference_flag)))))
#endif
    {
      rate += write_chroma_cbp();
    }
    else
    {
      //=====   C B P   =====
      //---------------------
      currSE->value1 = cbp;
      currSE->value2 = mb32;
      if (IS_OLDINTRA (currMB) || currMB->mb_type == SI4MB ||  currMB->mb_type == I8MB)
        currSE->type = SE_CBP_INTRA;
      else
        currSE->type = SE_CBP_INTER;

      currSE->writing = writeCBP_CABAC;

      // choose the appropriate data partition
      dataPart = &(currSlice->partArr[partMap[currSE->type]]);

      dataPart->writeSyntaxElement(currSE, dataPart);
      bitCount[BITS_CBP_MB] += currSE->len;
      rate                  += currSE->len;
#if TRACE
      snprintf(currSE->tracestring, TRACESTRING_SIZE, "CBP (%2d,%2d) = %3d",img->mb_x, img->mb_y, cbp);
#endif
      // proceed to next SE
      currSE++;  
      currMB->currSEnr++;
    }

  }

  return rate;
}

int writeLumaCoeff (int rdopt, int mb32)
{
  int             mb_x, mb_y, i, j, k;
  int             level, run;
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  int*            bitCount  = currMB->bitcounter;
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[0];
  int             cbp       = currMB->cbp;
  DataPartition*  dataPart;

  int   b8, b4;
  int*  DCLevel = img->cofDC[0][0];
  int*  DCRun   = img->cofDC[0][1];
  int*  ACLevel;
  int*  ACRun;

  if (!IS_NEWINTRA (currMB))
  {
    if(currMB->mb_type >= 0 && currMB->mb_type <= 3 && currMB->luma_transform_size_8x8_flag == 2)
    {
      rate += writeLumaCoefBigBlock(currMB, mb32, rdopt);
      return rate;
    }

    //=====  L U M I N A N C E   =====
    //--------------------------------
    for (i=0; i<4; i++)
      if (cbp & (1<<i))
      {
#ifdef USE_INTRA_MDDT
        if(!rdopt && currMB->mb_type == I8MB && input->UseIntraMDDT)
        {
          int pic_block_y, pic_block_x;
          int ipredmode; 
          byte (*scan)[2];

          pic_block_y = img->mb_y*4+2*(i>>1);
          pic_block_x = img->mb_x*4+2*(i &1); 
          ipredmode = img->ipredmode8x8[pic_block_y][pic_block_x];
          img->update8x8Count[ipredmode]++;

          {
            int*  ACLevel = img->cofAC[i][0][0];
            int*  ACRun   = img->cofAC[i][0][1];
            int  pos       = 0;
            int  level     = 1;
            int  k; 

            scan = img->scanOrder8x8[ipredmode];
            for (k=0; k < 64 && level != 0; k++)
            {
              level = ACLevel[k]; // level
              run   = ACRun  [k]; // run
              if(level == 0)
                break;
              pos   += run;
              img->scanStats8x8[ipredmode][pos]++;
              pos++;
            }
          }
        }
        else if(!rdopt && currMB->mb_type == I4MB && input->UseIntraMDDT)
        {
          int   b4; 

          for(b4 = 0; b4 < 4; b4++)
          {
            int*  ACLevel = img->cofAC[i][b4][0];
            int*  ACRun   = img->cofAC[i][b4][1];
            int  pos       = 0;
            int  level     = 1;
            int  k; 
            int pic_block_y, pic_block_x;
            int ipredmode; 
            byte (*scan)[2];

            pic_block_y = img->block_y+2*(i>>1)+(b4>>1);
            pic_block_x = img->block_x+2*(i &1)+(b4 &1);           
            ipredmode = img->ipredmode[pic_block_y][pic_block_x];

            scan = img->scanOrder4x4[ipredmode];
            img->update4x4Count[ipredmode]++;

            if(ACLevel[0])
            {
              for (k=0; k < 16 && level != 0; k++)
              {
                level = ACLevel[k]; // level
                run   = ACRun  [k]; // run
                if(level == 0)
                  break;
                pos   += run;
                img->scanStats4x4[ipredmode][pos]++;
                pos++;
              }
            }
          }
        }
#endif
        rate += writeLumaCoeff8x8 (i, currMB->b8mode[i], currMB->luma_transform_size_8x8_flag);
      }
#ifdef USE_INTRA_MDDT
      if(!rdopt && input->UseIntraMDDT)
      {
        int i; 
        for(i = 0; i < 9; i++)
        {
          if(img->update4x4Count[i] >= img->update4x4Thres[i])
          {
            img->update4x4[i] = 1;
            img->update4x4Count[i] = 0;
          }
          else img->update4x4[i] = 0;
          if(img->update8x8Count[i] >= img->update8x8Thres[i])
          {
            img->update8x8[i] = 1;
            img->update8x8Count[i] = 0;
          }
          else img->update8x8[i] = 0;
        }

        updateScanOrder(0);
        normalizeScanStats();
      }
#endif 
  }
  else
  {
    //=====  L U M I N A N C E   f o r   1 6 x 1 6  INTRA =====
#ifdef USE_INTRA_MDDT
    if(input->UseIntraMDDT)
    {
      if(currMB->cbp & 0xf)
        rate += writeLumaCoeff16x16 ();
    }
    else
    {
#endif
      //----------------------------------------------------
      // DC coeffs
      {
        level=1; // get inside loop
        for (k=0; k<=16 && level!=0; k++)
        {
          level = currSE->value1 = DCLevel[k]; // level
          run   = currSE->value2 = DCRun  [k]; // run

          currSE->writing = writeRunLevel_CABAC;

          currSE->context     = LUMA_16DC;
          currSE->type        = SE_LUM_DC_INTRA;   // element is of type DC
          img->is_intra_block = 1;

          // choose the appropriate data partition
          dataPart = &(currSlice->partArr[partMap[currSE->type]]);
          dataPart->writeSyntaxElement (currSE, dataPart);
          bitCount[BITS_COEFF_Y_MB] += currSE->len;
          rate                      += currSE->len;
#if TRACE
          snprintf(currSE->tracestring, TRACESTRING_SIZE, "DC luma 16x16 sng(%2d) level =%3d run =%2d", k, level, run);
#endif
          // proceed to next SE
          currSE++;
          currMB->currSEnr++;
        }
      }

      // AC coeffs
      if (cbp & 15)
      {
        for (mb_y=0; mb_y < 4; mb_y += 2)
          for (mb_x=0; mb_x < 4; mb_x += 2)
            for (j=mb_y; j < mb_y+2; j++)
              for (i=mb_x; i < mb_x+2; i++)
              {
                b8      = 2*(j >> 1) + (i >> 1);
                b4      = 2*(j & 0x01) + (i & 0x01);
                {
                  ACLevel = img->cofAC[b8][b4][0];
                  ACRun   = img->cofAC[b8][b4][1];

                  img->subblock_y = j;
                  img->subblock_x = i;

                  level=1; // get inside loop
                  for (k=0;k<16 && level !=0;k++)
                  {
                    level = currSE->value1 = ACLevel[k]; // level
                    run   = currSE->value2 = ACRun  [k]; // run

                    currSE->writing = writeRunLevel_CABAC;
                    currSE->context     = LUMA_16AC;
                    currSE->type        = SE_LUM_AC_INTRA;   // element is of type AC
                    img->is_intra_block = 1;

                    // choose the appropriate data partition
                    dataPart = &(currSlice->partArr[partMap[currSE->type]]);
                    dataPart->writeSyntaxElement (currSE, dataPart);
                    bitCount[BITS_COEFF_Y_MB] += currSE->len;
                    rate                      += currSE->len;
#if TRACE
                    snprintf(currSE->tracestring, TRACESTRING_SIZE, "AC luma 16x16 sng(%2d) level =%3d run =%2d", k, level, run);
#endif
                    // proceed to next SE
                    currSE++;
                    currMB->currSEnr++;
                  }
                }
              }
      }
#ifdef USE_INTRA_MDDT
    }
#endif
  }

  return rate;
}

/*!
************************************************************************
* \brief
*    Find best 16x16 based intra mode
*
* \par Input:
*    Image parameters, pointer to best 16x16 intra mode
*
* \par Output:
*    best 16x16 based SAD
************************************************************************/
int find_sad_16x16(int *intra_mode)
{
  int current_intra_sad_2,best_intra_sad2;
  int M1[16][16],M0[4][4][4][4],M3[4],M4[4][4];

  int i,j,k;
  int ii,jj;
  int mb_nr = img->current_mb_nr;

  PixelPos up;          //!< pixel position p(0,-1)
  PixelPos left[17];    //!< pixel positions p(-1, -1..15)

  int up_avail, left_avail, left_up_avail;

  for (i=0;i<17;i++)
  {
    getNeighbour(mb_nr, -1 ,  i-1 , 1, &left[i]);
  }

  getNeighbour(mb_nr, 0     ,  -1 , 1, &up);

  if (!(input->UseConstrainedIntraPred))
  {
    up_avail   = up.available;
    left_avail = left[1].available;
    left_up_avail = left[0].available;
  }
  else
  {
    up_avail      = up.available ? img->intra_block[up.mb_addr] : 0;
    for (i=1, left_avail=1; i<17;i++)
      left_avail  &= left[i].available ? img->intra_block[left[i].mb_addr]: 0;
    left_up_avail = left[0].available ? img->intra_block[left[0].mb_addr]: 0;
  }

  best_intra_sad2=MAX_VALUE;
  *intra_mode = DC_PRED_16;

  for (k=0;k<4;k++)
  {
    if (input->IntraDisableInterOnly == 0 || img->type != I_SLICE)
    {
      if (input->Intra16x16ParDisable && (k==VERT_PRED_16||k==HOR_PRED_16))
        continue;

      if (input->Intra16x16PlaneDisable && k==PLANE_16)
        continue;
    }
    //check if there are neighbours to predict from
    if ((k==0 && !up_avail) || (k==1 && !left_avail) || (k==3 && (!left_avail || !up_avail || !left_up_avail)))
    {
      ; // edge, do nothing
    }
    else
    {
      for (j=0;j<16;j++)
      {
        for (i=0;i<16;i++)
        {
          M1[j][i]=imgY_org[img->opix_y+j][img->opix_x+i]-img->mprr_2[k][j][i];
          M0[i & 0x03][i >> 2][j & 0x03][j >> 2]=M1[j][i];
        }
      }
      current_intra_sad_2=0;              // no SAD start handicap here
      for (jj=0;jj<4;jj++)
      {
        for (ii=0;ii<4;ii++)
        {
          for (j=0;j<4;j++)
          {
            M3[0]=M0[0][ii][j][jj]+M0[3][ii][j][jj];
            M3[1]=M0[1][ii][j][jj]+M0[2][ii][j][jj];
            M3[2]=M0[1][ii][j][jj]-M0[2][ii][j][jj];
            M3[3]=M0[0][ii][j][jj]-M0[3][ii][j][jj];

            M0[0][ii][j][jj]=M3[0]+M3[1];
            M0[2][ii][j][jj]=M3[0]-M3[1];
            M0[1][ii][j][jj]=M3[2]+M3[3];
            M0[3][ii][j][jj]=M3[3]-M3[2];
          }

          for (i=0;i<4;i++)
          {
            M3[0]=M0[i][ii][0][jj]+M0[i][ii][3][jj];
            M3[1]=M0[i][ii][1][jj]+M0[i][ii][2][jj];
            M3[2]=M0[i][ii][1][jj]-M0[i][ii][2][jj];
            M3[3]=M0[i][ii][0][jj]-M0[i][ii][3][jj];

            M0[i][ii][0][jj]=M3[0]+M3[1];
            M0[i][ii][2][jj]=M3[0]-M3[1];
            M0[i][ii][1][jj]=M3[2]+M3[3];
            M0[i][ii][3][jj]=M3[3]-M3[2];
            for (j=0;j<4;j++)
              if ((i+j)!=0)
                current_intra_sad_2 += absm(M0[i][ii][j][jj]);
          }
        }
      }

      for (j=0;j<4;j++)
        for (i=0;i<4;i++)
          M4[j][i]=M0[0][i][0][j]/4;

      // Hadamard of DC koeff
      for (j=0;j<4;j++)
      {
        M3[0]=M4[j][0]+M4[j][3];
        M3[1]=M4[j][1]+M4[j][2];
        M3[2]=M4[j][1]-M4[j][2];
        M3[3]=M4[j][0]-M4[j][3];

        M4[j][0]=M3[0]+M3[1];
        M4[j][2]=M3[0]-M3[1];
        M4[j][1]=M3[2]+M3[3];
        M4[j][3]=M3[3]-M3[2];
      }

      for (i=0;i<4;i++)
      {
        M3[0]=M4[0][i]+M4[3][i];
        M3[1]=M4[1][i]+M4[2][i];
        M3[2]=M4[1][i]-M4[2][i];
        M3[3]=M4[0][i]-M4[3][i];

        M4[0][i]=M3[0]+M3[1];
        M4[2][i]=M3[0]-M3[1];
        M4[1][i]=M3[2]+M3[3];
        M4[3][i]=M3[3]-M3[2];

        for (j=0;j<4;j++)
          current_intra_sad_2 += absm(M4[j][i]);
      }
      if(current_intra_sad_2 < best_intra_sad2)
      {
        best_intra_sad2=current_intra_sad_2;
        *intra_mode = k; // update best intra mode
      }
    }
  }
  best_intra_sad2 = best_intra_sad2/2;

  return best_intra_sad2;

}



