/* 
====================================================================================================================

The copyright in this software is being made available under the License included below. 
This software may be subject to other third party and 	contributor rights, including patent rights, and no such 
rights are granted under this license.

Copyright (c) 2010, QUALCOMM INC.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted only for 
the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
promoting such standards. The following conditions are required to be met:

* Redistributions of source code must retain the above copyright notice, this list of conditions and 
the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
the following disclaimer in the documentation and/or other materials provided with the distribution.
* The name of QUALCOMM INC.  may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

====================================================================================================================
*/
#include <stdlib.h>
#include <assert.h>
#include <memory.h>
#include "global.h"

#include "elements.h"
#include "vlc.h"
#include "image.h"
#include "mb_access.h"
#include "cabac.h"
#include "header.h"
#include <math.h>

#ifdef MV_COMPETITION
#include "mv_competition.h"
extern MV_Competition mv_comp;
extern int ******send_index_mv_prediction;        // int predModeMV2[mode][10][list][4][4][MAX_MODEPREDMV+1]; 
#endif
extern Macroblock  MB32;
extern Macroblock  MB64;//MB64X64
extern int   cbp_32[1<<(2*(MAX_MB_EXT_LEVEL-1))], cbp_32_RDCost[1<<(2*(MAX_MB_EXT_LEVEL-1))];//MB64X64

extern int   cbp16[1<<(2*MAX_MB_EXT_LEVEL)];
extern int64 cbp_blk16[1<<(2*MAX_MB_EXT_LEVEL)];
extern int   ****cofAC32[1<<(2*MAX_MB_EXT_LEVEL)];     // [8x8block][4x4block][level/run][scan_pos]
extern int   ***cofDC32[1<<(2*MAX_MB_EXT_LEVEL)];                       // [yuv][level/run][scan_pos]
extern int   luma_transform_size_8x8_flag32[1<<(2*MAX_MB_EXT_LEVEL)];
extern int    **cof32AC16x16[1<<(2*MAX_MB_EXT_LEVEL)];

extern int bestmode32;
extern int bestmode64;
//int mb64_stat[2][5];//MB64X64
extern int mb32type[1<<(2*(MAX_MB_EXT_LEVEL-1))];//MB64X64

extern int mb32_stat[2][5];
extern int mb32_stat_currentpass[2][3][5];
extern int mb64_stat[2][5];
extern int mb64_stat_currentpass[2][3][5];
extern int eos_bit_32;
extern int mb32_mvd[2][BLOCK_MULTIPLE<<MAX_MB_EXT_LEVEL][BLOCK_MULTIPLE<<MAX_MB_EXT_LEVEL][2];          //!< indices correspond to [forw,backw][block_y][block_x][x,y]
void unary_bin_encode(EncodingEnvironmentPtr eep_frame, unsigned int symbol, BiContextTypePtr ctx, int ctx_offset);
void unary_exp_golomb_mv_encode(EncodingEnvironmentPtr eep_dp, unsigned int symbol, BiContextTypePtr ctx, unsigned int max_bin);
int MBType2Value (Macroblock* currMB);
int writeIntraModes(void);
int ZeroRef (Macroblock* currMB);
int BType2CtxRef (int btype);
int writeDeltaQP(int rdopt, int cbp);
int writeCBP_Big(int rdopt, int cbp);
extern int writeCBP_Transform_size (int rdopt, int mb32);
int writeLumaCoeff (int rdopt, int mb32);

#ifdef GEOM
extern char ****geo_mask[3];//[mb_ext_level][rho][theta][y][x]
extern int curr_mb32i, curr_mbi;
extern int rho_bits[3];//[mb_ext_level]
extern int geo_stat[2][3];
extern int best8x8pdir_geo[GEO_PART_NUM][2];

int MBType2Value_geo (Macroblock* currMB);
int writeMotionInfo2NAL_geo (int geo_mode, int mb_ext_level);
void writeMB_geoInfo_CABAC (SyntaxElement *se, EncodingEnvironmentPtr eep_dp);

#endif

/*!
************************************************************************
* \brief
*    level           : 1-32x32; 2-64x64
*    CurrentMbAddr   : 16x16MB index in the raster scan order starting from 0
*    CurrentMbAddr_x : 16x16MB x-axis index
*    CurrentMbAddr_y : 16x16MB y_axis index
*    extend_x        : the horizonal next 16x16MB x-axis index in the current cluster
*    extend_y        : the vertical  next 16x16MB y-axis index in the current cluster
*    num_MB          : number of 16x16 MB in the cluster
*
************************************************************************
*/
void find_Mb_cluster(int level, int CurrentMbAddr, int *CurrentMbAddr_x, int *CurrentMbAddr_y, int *extend_x, int *extend_y, int *num_MB, int *nextMbAddr)
{
  int extend = 1<<level;
  *CurrentMbAddr_x = xPicPos[CurrentMbAddr];
  *CurrentMbAddr_y = yPicPos[CurrentMbAddr];
  *extend_x = min(*CurrentMbAddr_x+extend-1, (int)(img->PicWidthInMbs-1));
  *extend_y = min(*CurrentMbAddr_y+extend-1, (int)(img->FrameHeightInMbs-1));

  *num_MB = (*extend_x - *CurrentMbAddr_x + 1) * (*extend_y - *CurrentMbAddr_y + 1);

  if(*extend_x+1 <= (int)(img->PicWidthInMbs-1))
    *nextMbAddr = CurrentMbAddr + extend;
  else if(*extend_y+1 <= (int)(img->FrameHeightInMbs-1))
    *nextMbAddr = (*extend_y+1) * img->PicWidthInMbs;
  else
    *nextMbAddr = -1;

}

void prepare_data_for_write(int encodeMbAddr, Macroblock *MB, int mbcount, int mode)
{
  int ****i4p, ***i3p;
  Macroblock  *currMB  = &img->mb_data[encodeMbAddr];
  int **i2p;

  currMB->cbp = cbp16[mbcount];
  currMB->cbp_blk = cbp_blk16[mbcount];
  currMB->mb_type = mode;
  currMB->luma_transform_size_8x8_flag = luma_transform_size_8x8_flag32[mbcount];

  i4p=cofAC32[mbcount]; cofAC32[mbcount]=img->cofAC; img->cofAC=i4p;
  i3p=cofDC32[mbcount]; cofDC32[mbcount]=img->cofDC; img->cofDC=i3p;
  i2p=cof32AC16x16[mbcount]; cof32AC16x16[mbcount]=img->cofAC16x16; img->cofAC16x16=i2p;

  memcpy(currMB->b8mode,MB->b8mode, BLOCK_MULTIPLE * sizeof(int));
  memcpy(currMB->b8pdir,MB->b8pdir, BLOCK_MULTIPLE * sizeof(int));    
  currMB->bi_pred_me = MB->bi_pred_me;
}

void propagate_qp_info(int CurrentMbAddr, int CurrentMbAddr_x, int CurrentMbAddr_y, int extend_x, int extend_y)      
{
  int i, j, encodeMbAddr;
  int qp,prev_qp,prev_delta_qp;

  qp            = img->mb_data[CurrentMbAddr].qp;
  prev_qp       = img->mb_data[CurrentMbAddr].prev_qp;
  prev_delta_qp = img->mb_data[CurrentMbAddr].prev_delta_qp;

  for(i=CurrentMbAddr_y; i<=extend_y; i++)
  {
    for(j=CurrentMbAddr_x; j<=extend_x; j++)
    {
      encodeMbAddr = i*img->PicWidthInMbs + j;
      img->mb_data[encodeMbAddr].qp             = qp;
      img->mb_data[encodeMbAddr].prev_qp        = prev_qp;
      img->mb_data[encodeMbAddr].prev_delta_qp  = prev_delta_qp;
      set_chroma_qp(&img->mb_data[encodeMbAddr]);
      img->mb_data[encodeMbAddr].delta_qp = img->mb_data[encodeMbAddr].qp - img->mb_data[encodeMbAddr].prev_qp;
    }
  }
}

void propagate_mb_info(int mb_ext_level, Macroblock* currMB32, int CurrentMbAddr, int CurrentMbAddr_x, int CurrentMbAddr_y, int extend_x, int extend_y)      
{
  int i, j, encodeMbAddr;
#ifdef GEOM
  int k;
#endif
  //set 16x16 block type 
  for(i=CurrentMbAddr_y; i<=extend_y; i++)
  {
    for(j=CurrentMbAddr_x; j<=extend_x; j++)
    {
      encodeMbAddr = i*img->PicWidthInMbs + j;

      if(currMB32->mb_type == 0 && currMB32->cbp==0)
        img->mb_data[encodeMbAddr].skip_flag = 1;
      else
        img->mb_data[encodeMbAddr].skip_flag = 0;

      if(img->mb_data[CurrentMbAddr].mb_type == 0)
        img->mb_data[encodeMbAddr].mb_type = 0;
      else
      {
#ifndef GEOM
        if( CurrentMbAddr != encodeMbAddr)
#endif
        {
          img->mb_data[encodeMbAddr].mb_type = 1;
        }
      }
#ifdef GEOM
      for(k=0; k<4; k++)
        img->mb_data[encodeMbAddr].b8mode[k] = img->mb_data[encodeMbAddr].mb_type;

      if(currMB32->geo_mode != -1)
      {
        int m,n;
        int rho = currMB32->geo_mode>>5, theta_idx = currMB32->geo_mode & 31;  
        int start_y = (i - CurrentMbAddr_y)*16;
        int start_x = (j - CurrentMbAddr_x)*16;
        for(m=0; m<16; m++)
          for(n=0; n<16; n++)
            img->mb_data[encodeMbAddr].geo_mask[m][n] = geo_mask[mb_ext_level][rho][theta_idx][start_y+m][start_x+n];
      }
      else
      {
        img->mb_data[encodeMbAddr].geo_mask[0][0] = -1;
      }
#endif
    }//j
  }//i
}


void writeCBP_CABAC32(SyntaxElement *se, EncodingEnvironmentPtr eep_dp)
{
  //  TextureInfoContexts *ctx = img->currentSlice->tex_ctx;
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];

  int cbp = se->value1; // symbol to encode
  int curr_cbp_idx;

  curr_cbp_idx = 1;
  writeCBP_BIT_CABAC (0, cbp, cbp, currMB, curr_cbp_idx, eep_dp);
}

void writeCBP_CABAC_luma1bit(SyntaxElement *se, EncodingEnvironmentPtr eep_dp)
{
  //  TextureInfoContexts *ctx = img->currentSlice->tex_ctx;
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];

  int cbp = se->value1; // symbol to encode
  int curr_cbp_idx;

  curr_cbp_idx = 1;

  if((cbp&15)!=0)
    writeCBP_BIT_CABAC (0, 1, 1, currMB, curr_cbp_idx, eep_dp);
  else
    writeCBP_BIT_CABAC (0, 0, 0, currMB, curr_cbp_idx, eep_dp);

}

void writeMB_typeInfo_CABAC32(SyntaxElement *se, EncodingEnvironmentPtr eep_dp)
{
  int a, b;
  int act_ctx = 0;
  int act_sym;
  int bframe   = (img->type==B_SLICE);
  int mode_sym = 0;
  int mode16x16;


  MotionInfoContexts *ctx         = (img->currentSlice)->mot_ctx;
  Macroblock         *currMB      = &img->mb_data[img->current_mb_nr];
  int                curr_mb_type = se->value1;

  int dir1 = se->dir_partition1;
  int dir2 = se->dir_partition2;
  if(bframe)
    assert(curr_mb_type<=3 || curr_mb_type>=22);

  if(img->type == I_SLICE)  // INTRA-frame
  {
    if (currMB->mb_available_up == NULL)
      b = 0;
    else 
      b = ((currMB->mb_available_up->mb_type != I4MB &&  currMB->mb_available_up->mb_type != I8MB) ? 1 : 0 );

    if (currMB->mb_available_left == NULL)
      a = 0;
    else 
      a = ((currMB->mb_available_left->mb_type != I4MB &&  currMB->mb_available_left->mb_type != I8MB) ? 1 : 0 );

    act_ctx     = a + b;
    act_sym     = curr_mb_type;
    se->context = act_ctx; // store context

    if (act_sym==0) // 4x4 Intra
    {
      biari_encode_symbol(eep_dp, 0, ctx->mb_type_contexts[0] + act_ctx );
    }
    else // 16x16 Intra
    {
      biari_encode_symbol(eep_dp, 1, ctx->mb_type_contexts[0] + act_ctx );

      biari_encode_symbol_final(eep_dp, 0);

      mode_sym = act_sym-1; // Values in the range of 0...23
      act_ctx  = 4;
      act_sym  = mode_sym/12;
      biari_encode_symbol(eep_dp, (signed short) act_sym, ctx->mb_type_contexts[0] + act_ctx ); // coding of AC/no AC
      mode_sym = mode_sym % 12;
      act_sym  = mode_sym / 4; // coding of cbp: 0,1,2
      act_ctx  = 5;
      if (act_sym==0)
      {
        biari_encode_symbol(eep_dp, 0, ctx->mb_type_contexts[0] + act_ctx );
      }
      else
      {
        biari_encode_symbol(eep_dp, 1, ctx->mb_type_contexts[0] + act_ctx );
        act_ctx=6;
        biari_encode_symbol(eep_dp, (signed short) (act_sym!=1), ctx->mb_type_contexts[0] + act_ctx );
      }
      mode_sym = mode_sym & 0x03; // coding of I pred-mode: 0,1,2,3
      act_sym  = mode_sym >> 1;
      act_ctx  = 7;
      biari_encode_symbol(eep_dp, (signed short) act_sym, ctx->mb_type_contexts[0] + act_ctx );
      act_ctx  = 8;
      act_sym  = mode_sym & 0x01;
      biari_encode_symbol(eep_dp, (signed short) act_sym, ctx->mb_type_contexts[0] + act_ctx );
    }
  }
  else // INTER
  {

    if (bframe)
    {
      if (currMB->mb_available_up == NULL)
        b = 0;
      else
        b = ((currMB->mb_available_up->mb_type != 0) ? 1 : 0 );

      if (currMB->mb_available_left == NULL)
        a = 0;
      else
        a = ((currMB->mb_available_left->mb_type != 0) ? 1 : 0 );
      act_ctx = a + b;
      se->context = act_ctx; // store context
    }
    act_sym = curr_mb_type;

    if (act_sym>=(mode16x16=(bframe?24:7)))
    {
      mode_sym = act_sym-mode16x16;
      act_sym  = mode16x16; // 16x16 mode info
    }

    if (!bframe)
    {
      switch (act_sym)
      {
      case 0:       //0001    ...this will only happen when mode=0 and cbp!=0 i.e Pdirect
        //biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[1][4]);  
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[1][5]);
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[1][6]);
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[1][6]);
        break;
      case 1:       //0000    
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[1][5]);
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[1][6]);
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[1][6]);
        break;
      case 2:
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[1][5]);
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[1][7]);
        break;
      case 3:
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[1][5]);
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[1][7]);
        break;
      case 4:
      case 5:
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[1][5]);
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[1][6]);
        break;
        // case 6, 7 indicate Intra and are not valid for BigBlocks
      default:
        printf ("Unsupported MB-MODE in writeMB_typeInfo_CABAC32!\n");
        exit (1);
      }
    }
    else //===== B-FRAMES =====
    {
      if (act_sym==0)
      {
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[2][act_ctx]);
      }
      else if (act_sym==1)  //16x16           101
      {
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][act_ctx]); 
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[2][4]);
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][5]);
        //biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][6]);
        writeB_pDirInfo_CABAC (dir1 , eep_dp, &ctx->mb_type_contexts[2][6]);
      }
      else if (act_sym==2)  //16x8            110
      {
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][act_ctx]); 
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][4]);
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[2][5]);
        //biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[2][6]);
        writeB_pDirInfo_CABAC (dir1 , eep_dp, &ctx->mb_type_contexts[2][6]);
        writeB_pDirInfo_CABAC (dir2 , eep_dp, &ctx->mb_type_contexts[2][6]);
      }
      else if (act_sym==3)  //8x16            111
      {
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][act_ctx]); 
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][4]);
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][5]);
        //biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][6]);
        writeB_pDirInfo_CABAC (dir1 , eep_dp, &ctx->mb_type_contexts[2][6]);
        writeB_pDirInfo_CABAC (dir2 , eep_dp, &ctx->mb_type_contexts[2][6]);
      }
      else if (act_sym==22)  //8x8            100
      {
        biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][act_ctx]); 
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[2][4]);
        biari_encode_symbol (eep_dp, 0, &ctx->mb_type_contexts[2][5]);
        //biari_encode_symbol (eep_dp, 1, &ctx->mb_type_contexts[2][6]);
      }
      else if (act_sym>22)  //Intra modes in B slice      1101 
      {
        //Intra in BigBlocks
        //this will never happen...
      }
    }

    if(act_sym==mode16x16) // additional info for 16x16 Intra-mode
    {
      if( mode_sym==25 )
      {
        biari_encode_symbol_final(eep_dp, 1 );
        return;
      }
      biari_encode_symbol_final(eep_dp, 0 );

      act_ctx = 8;
      act_sym = mode_sym/12;
      biari_encode_symbol(eep_dp, (signed short) act_sym, ctx->mb_type_contexts[1] + act_ctx ); // coding of AC/no AC
      mode_sym = mode_sym % 12;

      act_sym = mode_sym / 4; // coding of cbp: 0,1,2
      act_ctx = 9;
      if (act_sym==0)
      {
        biari_encode_symbol(eep_dp, 0, ctx->mb_type_contexts[1] + act_ctx );
      }
      else
      {
        biari_encode_symbol(eep_dp, 1, ctx->mb_type_contexts[1] + act_ctx );
        biari_encode_symbol(eep_dp, (signed short) (act_sym!=1), ctx->mb_type_contexts[1] + act_ctx );
      }

      mode_sym = mode_sym % 4; // coding of I pred-mode: 0,1,2,3
      act_ctx  = 10;
      act_sym  = mode_sym/2;
      biari_encode_symbol(eep_dp, (signed short) act_sym, ctx->mb_type_contexts[1] + act_ctx );
      act_sym  = mode_sym%2;
      biari_encode_symbol(eep_dp, (signed short) act_sym, ctx->mb_type_contexts[1] + act_ctx );
    }
  }
}


int write_mb_type(Macroblock *MB, int mb_ext_level)
{
  int             mb_nr      = img->current_mb_nr;
  Macroblock*     currMB     = &img->mb_data[mb_nr];
  int             mb_type;  
  SyntaxElement  *currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
  int*            bitCount   = currMB->bitcounter;
  Slice*          currSlice  = img->currentSlice;
  DataPartition*  dataPart;
  const int*      partMap    = assignSE2partition[0];
  int             no_bits    = 0;

  // choose the appropriate data partition
  dataPart = &(currSlice->partArr[partMap[SE_MBTYPE]]);

  //========= write mb_skip_flag (CABAC) =========
#ifdef GEOM
  if(MB->geo_mode != -1)
    mb_type         = MBType2Value_geo (MB);
  else
#endif
    mb_type         = MBType2Value (MB); //MB64 set in write_macroblock_cluster

  currSE->value1  = mb_type;
  currSE->value2  = MB->cbp; //MB32.cbp set in write_macroblock_cluster
  currSE->type    = SE_MBTYPE;
  currSE->writing = writeMB_skip_flagInfo_CABAC;
  dataPart->writeSyntaxElement( currSE, dataPart);
#if TRACE
  snprintf(currSE->tracestring, TRACESTRING_SIZE, "mb_skip_flag");
#endif
  bitCount[BITS_MB_MODE] += currSE->len;
  no_bits                += currSE->len;
  currSE++;
  currMB->currSEnr++;


  CheckAvailabilityOfNeighborsCABAC();

  //========= write mb_type (CABAC) =========
  if (MB->mb_type != 0 || MB->cbp != 0)
  {
    currSE->value1  = mb_type;
    currSE->value2  = 0;
    currSE->type    = SE_MBTYPE;
#ifdef GEOM
    if(MB->geo_mode != -1)
    {
      //mbtype = 3;//use B_8x16
      currSE->dir_partition1  = best8x8pdir_geo[MB->geo_mode][0];
      currSE->dir_partition2  = best8x8pdir_geo[MB->geo_mode][1];
    }
    else
#endif
    {
      currSE->dir_partition1 = MB->b8pdir[0];//this is for Partition 1
      currSE->dir_partition2 = MB->b8pdir[3];//this is for Partition 2
    }

    if(mb_ext_level > 0)
      currSE->writing = writeMB_typeInfo_CABAC32;
    else
      currSE->writing = writeMB_typeInfo_CABAC;

    dataPart->writeSyntaxElement( currSE, dataPart);

    bitCount[BITS_MB_MODE] += currSE->len;
    no_bits                += currSE->len;
    currSE++;
    currMB->currSEnr++;

#ifdef GEOM
    if(MB->mb_type == 3 || MB->geo_mode != -1)
    {
      currSE->value1  = MB->geo_mode;
      currSE->value2  = rho_bits[mb_ext_level]+THETA_BITS;
      currSE->writing = writeMB_geoInfo_CABAC;
      dataPart->writeSyntaxElement( currSE, dataPart);
#if TRACE
      snprintf(currSE->tracestring, TRACESTRING_SIZE, "mb_skip_flag");
#endif
      bitCount[BITS_MB_MODE] += currSE->len;
      no_bits                += currSE->len;
      currSE++;
      currMB->currSEnr++;
    }
#endif
  }

  return no_bits;
}



extern int writeChromaCoeff (void);
/*!
************************************************************************
* \brief
*    Codes macroblock header
* \param rdopt
*    true for calls during RD-optimization
* \param coeff_rate
*    bitrate of Luma and Chroma coeff
* \param (bigger_block_type,mb16i)
*        (1,0) - 32x32, 32x16, 16x32 mode info, mv info, coeff of upper left 16x16, dequant
*        (1,1~)-                                         coeff of other      16x16
*        (0,0) - 16x16 mode, upper left 16x16 partition mode, mv, coeff, dequant
*        (0,1~)-                        16x16 partition mode, mv, coeff
************************************************************************
*/
int writeMBLayer_Big (int rdopt, int *coeff_rate, int bigger_block_type, int mb32i, int mb16i)
{
  int             i;
  int             mb_nr      = img->current_mb_nr;
  Macroblock*     currMB     = &img->mb_data[mb_nr];
  SyntaxElement  *currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
  int*            bitCount   = currMB->bitcounter;
  Slice*          currSlice  = img->currentSlice;
  DataPartition*  dataPart;
  const int*      partMap    = assignSE2partition[0];
  int             no_bits    = 0;

  currMB->IntraChromaPredModeFlag = IS_INTRA(currMB);


  // choose the appropriate data partition
  dataPart = &(currSlice->partArr[partMap[SE_MBTYPE]]);

  if(img->type == I_SLICE)
  {
    // should never come to this place
    // No big block in I slice
    assert(0);
  }
  // not I_SLICE, CABAC
  else
  {

#if (!defined TEST_ONLY_Z_SCAN) 

    if(bigger_block_type == 2)
    {
      if(mb32i==0 && mb16i==0)
      {
        no_bits += write_mb_type(&MB64, 2);
        currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
      }
    }
    else if(MB64.write_mb64_mbtype && mb16i==0)
    {
      Macroblock mb32x32;
      mb32x32.mb_type = P8x8;

#ifdef GEOM
      mb32x32.geo_mode = -1;
#endif
      no_bits += write_mb_type(&mb32x32, 2);
      currSE     = &img->MB_SyntaxElements[currMB->currSEnr];

    }
    /********************************************
    *MB type for mode 32x32, 32x16, 16x32, 16x16
    *******************************************/
    if(bigger_block_type <= 1)
    {
      if(mb16i==0)
      {
        no_bits += write_mb_type(&MB32, 1);
        currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
      }
    }
#endif

    if(bigger_block_type ==0)
    {
      no_bits += write_mb_type(currMB, 0);
      currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
    }

    CheckAvailabilityOfNeighborsCABAC();
  }

  if (input->mv_competition > 0)
  {
    if((rdopt==0) && (img->type==P_SLICE) && (mv_comp.nb_mode_for_skip > 1))  
    {  
      //make sure that this function is called only once...whether it is big blocks or 16x16
      if(bigger_block_type==2)
      {
        if(mb32i==0 && mb16i==0 && MB64.mb_type==0)
          no_bits  += write_predictor_index_for_skip_mode64();
      }
      if(bigger_block_type==1)
      { 
        //32x32 P-Skip
        if(mb16i==0 && MB32.mb_type==0)
          no_bits  += write_predictor_index_for_skip_mode32();
      }
      else if(bigger_block_type==0)
      {
        //16x16 P-Skip
        if(currMB->mb_type==0)
          no_bits  += write_predictor_index_for_skip_mode  ();
      }
    }
  }

  if (bigger_block_type == 0)
  {
    //init NoMbPartLessThan8x8Flag
    currMB->NoMbPartLessThan8x8Flag = (IS_DIRECT(currMB) && !(active_sps->direct_8x8_inference_flag))? 0: 1;
    //===== BITS FOR 8x8 SUB-PARTITION MODES =====
    if(IS_P8x8 (currMB))     //---this will happen for 16x16 having 8x8 partition
    {
      dataPart = &(currSlice->partArr[partMap[SE_MBTYPE]]);
      for (i=0; i<4; i++)
      {
        currSE->dir_partition1 = currMB->b8pdir[i];    //what value to set here-> see the input parameter in B8Mode2Value()
        currSE->writing = writeB8_typeInfo_CABAC;

        currSE->value1  = B8Mode2Value (currMB->b8mode[i], currMB->b8pdir[i]);
        currSE->value2  = 0;
        currSE->type    = SE_MBTYPE;
        dataPart->writeSyntaxElement (currSE, dataPart);

        bitCount[BITS_MB_MODE]+= currSE->len;
        no_bits               += currSE->len;
        currSE++;
        currMB->currSEnr++;

        //set NoMbPartLessThan8x8Flag for P8x8 mode
        currMB->NoMbPartLessThan8x8Flag &= (currMB->b8mode[i]==0 && active_sps->direct_8x8_inference_flag) || 
          (currMB->b8mode[i]==4);
      }

      if (input->mv_competition > 0)
        pass_with_writing = !rdopt;
      no_bits += writeMotionInfo2NAL  (bigger_block_type);
      currSE   = &img->MB_SyntaxElements[currMB->currSEnr];
    }
  }
  //============= Transform size flag for INTRA MBs =============
  //-------------------------------------------------------------
  //transform size flag for INTRA_4x4 and INTRA_8x8 modes
  if ((currMB->mb_type == I8MB || currMB->mb_type == I4MB) && input->Transform8x8Mode)
  {
    currSE->value1 = currMB->luma_transform_size_8x8_flag;
    currSE->type   = SE_HEADER;
    currSE->writing = writeMB_transform_size_CABAC;
    dataPart->writeSyntaxElement(currSE, dataPart);

    bitCount[BITS_MB_MODE] += currSE->len;
    no_bits                += currSE->len;
    currSE++;
    currMB->currSEnr++;
  }


  //===== BITS FOR INTRA PREDICTION MODES ====
  no_bits += writeIntraModes();
  //===== BITS FOR CHROMA INTRA PREDICTION MODE ====
  if (currMB->IntraChromaPredModeFlag && img->yuv_format != YUV400)
    no_bits += writeChromaIntraPredMode();
  else if(!rdopt) //GB CHROMA !!!!!
    currMB->c_ipred_mode = DC_PRED_8; //setting c_ipred_mode to default is not the right place here
  //resetting in rdopt.c (but where ??)
  //with cabac and bframes maybe it could crash without this default
  //since cabac needs the right neighborhood for the later MBs

  //----- motion information -----
  if(bigger_block_type==2)
  {
    if(mb32i==0 && mb16i==0)
    {
      if(MB64.mb_type != 0)
      {
        if (input->mv_competition > 0)
          pass_with_writing = !rdopt;

#ifdef GEOM
        if(MB64.geo_mode != -1)
          no_bits  += writeMotionInfo2NAL_geo  (MB64.geo_mode, 2);
        else
#endif
          no_bits  += writeMotionInfo2NAL  (bigger_block_type);
      }
    }
  }
  if(bigger_block_type==1)
  {
    if(mb16i==0)
    {
      if(MB32.mb_type != 0)
      {
        if (input->mv_competition > 0)
          pass_with_writing = !rdopt;
#ifdef GEOM
        if(MB32.geo_mode != -1)
          no_bits  += writeMotionInfo2NAL_geo  (MB32.geo_mode, 1);
        else
#endif
          no_bits  += writeMotionInfo2NAL  (bigger_block_type);
      }
    }
  }
  else if(bigger_block_type==0 && currMB->mb_type !=0 && currMB->mb_type !=P8x8)
  {
    if (input->mv_competition > 0)
      pass_with_writing = !rdopt;

#ifdef GEOM
    if(currMB->geo_mode != -1)
      no_bits  += writeMotionInfo2NAL_geo  (currMB->geo_mode, 0);
    else
#endif
      no_bits  += writeMotionInfo2NAL  (bigger_block_type);
  }

  if(bigger_block_type == 2)
  {
    // this function only takes care of 64x64 skip, 64x64, 64x32, 32x64....
    // so there will be only Delta_QP at 64x64 level

    if(MB64.mb_type!=0 || MB64.cbp!=0)
    {
      if(mb32i==0 && mb16i==0)    //64x64 cbp
      {
        no_bits += writeCBP_Big (rdopt, MB64.cbp);
        no_bits += writeDeltaQP (rdopt, MB64.cbp);
      }

      if(MB64.cbp!=0)           
      {
        if(mb16i == 0)            //32x32 cbp 
        {
          no_bits += writeCBP_Big  (rdopt, MB32.cbp);
        }

        if(MB32.cbp != 0)         //16x16 cbp, delta_qp, luma coeff, chroma coeff
        {
          no_bits +=    writeCBP_Transform_size (rdopt, 1);
          *coeff_rate = writeLumaCoeff          (rdopt, 1);

          if (img->yuv_format != YUV400)
            *coeff_rate  += writeChromaCoeff ();

          no_bits  += *coeff_rate;           
        }
      }
    }
  }
  else
  {
#if (!defined TEST_ONLY_Z_SCAN)    
    if(  (bigger_block_type==1 && (MB32.mb_type!=0 || MB32.cbp!=0))
      || (bigger_block_type==0 && (currMB->mb_type!=0 || currMB->cbp!=0)))
#else
    if ((currMB->mb_type!=0) || (img->type==B_SLICE && (currMB->cbp!=0)))
#endif
    {
      if(bigger_block_type==1 && mb16i==0)
      {
        no_bits += writeCBP_Big    (rdopt, MB32.cbp);
        no_bits += writeDeltaQP    (rdopt, MB32.cbp);
      }
      if( bigger_block_type==0 || (bigger_block_type==1 && MB32.cbp!=0) )
      {
        no_bits +=    writeCBP_Transform_size (rdopt, bigger_block_type);
        if(bigger_block_type==0)
          no_bits +=    writeDeltaQP            (rdopt, img->mb_data[img->current_mb_nr].cbp);
        *coeff_rate = writeLumaCoeff          (rdopt, bigger_block_type);
        if (img->yuv_format != YUV400)
          *coeff_rate  += writeChromaCoeff ();

        no_bits  += *coeff_rate;  
      }
    }
  }

  return no_bits;
}


/*!
************************************************************************
* \brief
*    Passes the chosen syntax elements to the NAL for 16x16 block inside of 64x64 block
* \param coeff_rate
*    bitrate of Luma and Chroma coeff
* \param (mb32i,mb16i)
*    mb32i: index of 32x32 block inside 64x64
*    mb16i: index of 16x16 block inside 32x32
************************************************************************
*/
int write_one_macroblock64 (int eos_bit, int mb32i, int mb16i)
{
  Macroblock* currMB   = &img->mb_data[img->current_mb_nr];
  int*        bitCount = currMB->bitcounter;
  int i;
  int no_bits;

  extern int cabac_encoding;

  //===== init and update number of intra macroblocks =====
  if (img->current_mb_nr==0)
    intras=0;

  if (IS_INTRA(currMB))
    intras++;

  //--- write non-slice termination symbol if the macroblock is not the first one in its slice ---
  if (img->current_mb_nr!=img->currentSlice->start_mb_nr && eos_bit && mb32i==0 && mb16i==0)
  {
    write_terminating_bit (0);
  }

  cabac_encoding = 1;

  //--- write macroblock ---
  no_bits = writeMBLayer_Big (0, &i, 2, mb32i, mb16i);  // i is temporary
  if (!((currMB->mb_type !=0 ) || ((img->type==B_SLICE) && currMB->cbp != 0) ))
  { 
    for (i=0; i < 4; i++)
      memset(img->nz_coeff [img->current_mb_nr][i], 0, (4 + img->num_blk8x8_uv) * sizeof(int));  // CAVLC
  }

  //--- set total bit-counter ---
  if (input->mv_competition > 0)
  {
    bitCount[BITS_TOTAL_MB] = bitCount[BITS_MB_MODE]  + bitCount[BITS_COEFF_Y_MB]     
    + bitCount[BITS_INTER_MB] + bitCount[BITS_CBP_MB]  
    + bitCount[BITS_DELTA_QUANT_MB] + bitCount[BITS_COEFF_UV_MB]
    + bitCount[BITS_SKIP_PRED]
    + bitCount[BITS_MV_PRED];
  }
  else

  {
    bitCount[BITS_TOTAL_MB] = bitCount[BITS_MB_MODE]  + bitCount[BITS_COEFF_Y_MB]     
    + bitCount[BITS_INTER_MB] + bitCount[BITS_CBP_MB]  
    + bitCount[BITS_DELTA_QUANT_MB] + bitCount[BITS_COEFF_UV_MB];
  }  


  stats->bit_use_mb_type[img->type]      += bitCount[BITS_MB_MODE];
  stats->bit_use_coeffY[img->type]       += bitCount[BITS_COEFF_Y_MB] ;
  stats->tmp_bit_use_cbp[img->type]      += bitCount[BITS_CBP_MB];
  stats->bit_use_coeffC[img->type]       += bitCount[BITS_COEFF_UV_MB];
  stats->bit_use_delta_quant[img->type]  += bitCount[BITS_DELTA_QUANT_MB];
  stats->bit_use_mode[img->type][currMB->mb_type]+= bitCount[BITS_INTER_MB];
  stats->bit_slice += bitCount[BITS_TOTAL_MB];

  cabac_encoding = 0;

  return no_bits;
}

/*!
************************************************************************
* \brief
*    Passes the chosen syntax elements to the NAL for 16x16 block inside of 32x32 block
* \param coeff_rate
*    bitrate of Luma and Chroma coeff
* \param (mb32,mbi)
*        (1,0) - 32x32, 32x16, 16x32 mode info, mv info, coeff of upper left 16x16, dequant
*        (1,1~)-                                         coeff of other      16x16
*        (0,0) - 16x16 mode, upper left 16x16 partition mode, mv, coeff, dequant
*        (0,1~)-                        16x16 partition mode, mv, coeff
************************************************************************
*/
int write_one_macroblock32 (int eos_bit, int mb32, int mbi)
{
  Macroblock* currMB   = &img->mb_data[img->current_mb_nr];
  int*        bitCount = currMB->bitcounter;
  int i;
  int no_bits;

  extern int cabac_encoding;

  //===== init and update number of intra macroblocks =====
  if (img->current_mb_nr==0)
    intras=0;

  if (IS_INTRA(currMB))
    intras++;

  //--- write non-slice termination symbol if the macroblock is not the first one in its slice ---
  if (img->current_mb_nr!=img->currentSlice->start_mb_nr && eos_bit && mbi==0 && eos_bit_32)
  {
    write_terminating_bit (0);
  }

  cabac_encoding = 1;

  //--- write macroblock ---
  no_bits = writeMBLayer_Big (0, &i, mb32, 0, mbi);  // i is temporary
  if (!((currMB->mb_type !=0 ) || ((img->type==B_SLICE) && currMB->cbp != 0) ))
  { 
    for (i=0; i < 4; i++)
      memset(img->nz_coeff [img->current_mb_nr][i], 0, (4 + img->num_blk8x8_uv) * sizeof(int));  // CAVLC
  }

  //--- set total bit-counter ---
  if (input->mv_competition > 0)
  {
    bitCount[BITS_TOTAL_MB] = bitCount[BITS_MB_MODE]  + bitCount[BITS_COEFF_Y_MB]     
    + bitCount[BITS_INTER_MB] + bitCount[BITS_CBP_MB]  
    + bitCount[BITS_DELTA_QUANT_MB] + bitCount[BITS_COEFF_UV_MB]
    + bitCount[BITS_SKIP_PRED]
    + bitCount[BITS_MV_PRED];
  }
  else
  {
    bitCount[BITS_TOTAL_MB] = bitCount[BITS_MB_MODE]  + bitCount[BITS_COEFF_Y_MB]     
    + bitCount[BITS_INTER_MB] + bitCount[BITS_CBP_MB]  
    + bitCount[BITS_DELTA_QUANT_MB] + bitCount[BITS_COEFF_UV_MB];
  }  
  stats->bit_slice += bitCount[BITS_TOTAL_MB];

  if(mb32)
  {
    stats->bit_use_mb_type[img->type]      += bitCount[BITS_MB_MODE];
    stats->bit_use_coeffY[img->type]       += bitCount[BITS_COEFF_Y_MB] ;
    stats->tmp_bit_use_cbp[img->type]      += bitCount[BITS_CBP_MB];
    stats->bit_use_coeffC[img->type]       += bitCount[BITS_COEFF_UV_MB];
    stats->bit_use_delta_quant[img->type]  += bitCount[BITS_DELTA_QUANT_MB];
    stats->bit_use_mode[img->type][currMB->mb_type]+= bitCount[BITS_INTER_MB];
  }
  cabac_encoding = 0;
  return no_bits;
}

int writeLumaCoeff16x16P_CABAC ()
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[0];
  int*            bitCount  = currMB->bitcounter;
  DataPartition*  dataPart;

  int   level, run;
  int   k;
  int*  ACLevel = img->cofAC16x16[0];
  int*  ACRun   = img->cofAC16x16[1];

  img->subblock_x = 0;//img->pix_x;  // horiz. position for coeff_count context
  img->subblock_y = 0;//img->pix_y;  // vert.  position for coeff_count context

  level=1; // get inside loop
  for(k=0; k<=256 && level !=0; k++)
  {
    level = currSE->value1 = ACLevel[k]; // level
    run   = currSE->value2 = ACRun  [k]; // run

    currSE->writing = writeRunLevel16x16P_CABAC;

    currSE->context     = LUMA_16x16P; 
    currSE->type        = SE_LUM_AC_INTER;
    img->is_intra_block = 0;    // only applicable to inter blocks

    // choose the appropriate data partition
    dataPart = &(currSlice->partArr[partMap[img->type != B_SLICE ? currSE->type : SE_BFRAME]]);
    dataPart->writeSyntaxElement (currSE, dataPart);
    bitCount[BITS_COEFF_Y_MB] += currSE->len;
    rate                      += currSE->len;

#if TRACE
    snprintf(currSE->tracestring, TRACESTRING_SIZE, "Luma8x8 sng(%2d) level =%3d run =%2d", k, level,run);
#endif
    /* proceed to next SE */
    currSE++;  
    currMB->currSEnr++;

  }

  return rate;
}

int writeLumaCoeff16x16P ()
{
  int  rate = 0;
  rate += writeLumaCoeff16x16P_CABAC ();
  return rate;
}

int writeLumaCoeff16x8_CABAC (int blk)
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[0];
  int*            bitCount  = currMB->bitcounter;
  DataPartition*  dataPart;

  int   level, run;
  int   k;
  int*  ACLevel = img->cofAC16x8[blk][0];
  int*  ACRun   = img->cofAC16x8[blk][1];

  img->subblock_x = 0;               // horiz. position for coeff_count context
  img->subblock_y = (blk == 0)?0:2;  // vert.  position for coeff_count context

  level=1; // get inside loop
  for(k=0; k<=128 && level !=0; k++)
  {
    level = currSE->value1 = ACLevel[k]; // level
    run   = currSE->value2 = ACRun  [k]; // run

    currSE->writing = writeRunLevel16x8_CABAC;

    currSE->context     = LUMA_16x8P;

    currSE->type        = SE_LUM_AC_INTER;
    img->is_intra_block = 0;    // only applicable to inter blocks

    // choose the appropriate data partition
    dataPart = &(currSlice->partArr[partMap[img->type != B_SLICE ? currSE->type : SE_BFRAME]]);
    dataPart->writeSyntaxElement (currSE, dataPart);
    bitCount[BITS_COEFF_Y_MB] += currSE->len;
    rate                      += currSE->len;
#if TRACE
    snprintf(currSE->tracestring, TRACESTRING_SIZE, "Luma8x8 sng(%2d) level =%3d run =%2d", k, level,run);
#endif
    /* proceed to next SE */
    currSE++;  
    currMB->currSEnr++;

  }

  return rate;
}

int writeLumaCoeff16x8 ( int blk16x8 )
{
  int  rate = 0;
  rate += writeLumaCoeff16x8_CABAC (blk16x8);
  return rate;
}

int writeLumaCoeff8x16_CABAC (int blk)
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[0];
  int*            bitCount  = currMB->bitcounter;
  DataPartition*  dataPart;

  int   level, run;
  int   k;
  int*  ACLevel = img->cofAC8x16[blk][0];
  int*  ACRun   = img->cofAC8x16[blk][1];

  img->subblock_x = (blk == 0)?0:2;     // horiz. position for coeff_count context
  img->subblock_y = 0;                  // vert.  position for coeff_count context

  level=1; // get inside loop
  for(k=0; k<=128 && level !=0; k++)
  {
    level = currSE->value1 = ACLevel[k]; // level
    run   = currSE->value2 = ACRun  [k]; // run

    currSE->writing = writeRunLevel8x16_CABAC;

    currSE->context     = LUMA_8x16P;

    currSE->type        = SE_LUM_AC_INTER;
    img->is_intra_block = 0;    // only applicable to inter blocks

    // choose the appropriate data partition
    dataPart = &(currSlice->partArr[partMap[img->type != B_SLICE ? currSE->type : SE_BFRAME]]);
    dataPart->writeSyntaxElement (currSE, dataPart);
    bitCount[BITS_COEFF_Y_MB] += currSE->len;
    rate                      += currSE->len;
#if TRACE
    snprintf(currSE->tracestring, TRACESTRING_SIZE, "Luma8x8 sng(%2d) level =%3d run =%2d", k, level,run);
#endif
    /* proceed to next SE */
    currSE++;  
    currMB->currSEnr++;

  }

  return rate;
}

int writeLumaCoeff8x16 ( int blk8x16 )
{
  int  rate = 0;
  rate += writeLumaCoeff8x16_CABAC (blk8x16);
  return rate;
}

int writeLumaCoefBigBlock(Macroblock *currMB, int mb32, int rdopt)
{
  int rate = 0;

  assert(currMB->mb_type >= 0 && currMB->mb_type <= 3);
  assert(currMB->luma_transform_size_8x8_flag);
#ifdef GEOM
  if((currMB->mb_type < 2 || mb32 || currMB->geo_mode!=-1)&& currMB->luma_transform_size_8x8_flag == 2)  // 16x16 partition
#else
  if((currMB->mb_type < 2 || mb32)&& currMB->luma_transform_size_8x8_flag == 2)  // 16x16 partition
#endif
  {
    if(currMB->cbp & 0xf)
      rate += writeLumaCoeff16x16P(); 
  }
  else if(currMB->mb_type == 2 && currMB->luma_transform_size_8x8_flag == 2)  // 16x8 partition
  {
    if(currMB->cbp & 0x3)
      rate += writeLumaCoeff16x8(0);
    if(currMB->cbp & 0xc)
      rate += writeLumaCoeff16x8(1);
  }
  else if(currMB->mb_type == 3 && currMB->luma_transform_size_8x8_flag == 2)  // 8x16 partition
  {
    if(currMB->cbp & 0x5)
      rate += writeLumaCoeff8x16(0);
    if(currMB->cbp & 0xa)
      rate += writeLumaCoeff8x16(1);
  }

  return rate;
}


// write 1 bit indicating luma cbp
int writelumaCBPTransform_size(int BigBlock)
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  int*            bitCount  = currMB->bitcounter;
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[0];
  int             cbp       = currMB->cbp;
  DataPartition*  dataPart;
  int             need_transform_size_flag;   //ADD-VG-24062004


  //=====   C B P   =====
  //---------------------
  currSE->value1 = cbp;

  if (IS_OLDINTRA (currMB) || currMB->mb_type == SI4MB ||  currMB->mb_type == I8MB)
    currSE->type = SE_CBP_INTRA;
  else
    currSE->type = SE_CBP_INTER;

  // choose the appropriate data partition
  dataPart = &(currSlice->partArr[partMap[currSE->type]]);

  currSE->writing = writeCBP_CABAC_luma1bit;
  dataPart->writeSyntaxElement(currSE, dataPart);
  bitCount[BITS_CBP_MB] += currSE->len;
  rate                  += currSE->len;
  // proceed to next SE
  currSE++;  
  currMB->currSEnr++;


  //============= Transform Size Flag for INTER MBs =============
  //-------------------------------------------------------------

  need_transform_size_flag = (((currMB->mb_type >= 0 && currMB->mb_type <= 3)||
    (IS_DIRECT(currMB) && active_sps->direct_8x8_inference_flag) ||
    (currMB->NoMbPartLessThan8x8Flag))
    && currMB->mb_type != I8MB && currMB->mb_type != I4MB
    && (currMB->cbp&15)
    && input->Transform8x8Mode);

  if (need_transform_size_flag)
  {
    if (input->DisableSmallTransInLargeMB && BigBlock)
      return rate;
    currSE->value1 = currMB->luma_transform_size_8x8_flag;
    currSE->type   = SE_HEADER;
    currSE->writing = writeMB_transform_size_CABAC;
    dataPart->writeSyntaxElement(currSE, dataPart);
    bitCount[BITS_MB_MODE] += currSE->len;
    rate                   += currSE->len;
    currSE++;
    currMB->currSEnr++;
  }

  return rate;
}
void writeCBP_CABAC_chroma(SyntaxElement *se, EncodingEnvironmentPtr eep_dp)
{
  TextureInfoContexts *ctx = img->currentSlice->tex_ctx;
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];

  int a, b;
  int curr_cbp_ctx;
  int cbp = se->value1; // symbol to encode
  int cbp_bit;

  if (img->yuv_format != YUV400)
  {
    // coding of chroma part
    b = 0;
    if (currMB->mb_available_up != NULL)
      b = ((currMB->mb_available_up)->cbp > 15) ? 1 : 0;

    a = 0;
    if (currMB->mb_available_left != NULL)
      a = ((currMB->mb_available_left)->cbp > 15) ? 1 : 0;

    curr_cbp_ctx = a+2*b;
    cbp_bit = (cbp > 15 ) ? 1 : 0;
    biari_encode_symbol(eep_dp, (signed short) cbp_bit, ctx->cbp_contexts[1] + curr_cbp_ctx );

    if (cbp > 15)
    {
      b = 0;
      if (currMB->mb_available_up != NULL)
      {
        if ((currMB->mb_available_up)->cbp > 15)
          b = (( ((currMB->mb_available_up)->cbp >> 4) == 2) ? 1 : 0);
      }

      a = 0;
      if (currMB->mb_available_left != NULL)
      {
        if ((currMB->mb_available_left)->cbp > 15)
          a = (( ((currMB->mb_available_left)->cbp >> 4) == 2) ? 1 : 0);
      }

      curr_cbp_ctx = a+2*b;
      cbp_bit = ((cbp>>4) == 2) ? 1 : 0;
      biari_encode_symbol(eep_dp, (signed short) cbp_bit, ctx->cbp_contexts[2] + curr_cbp_ctx );
    }
  }
}


int write_chroma_cbp()
{
  int             rate      = 0;
  Macroblock*     currMB    = &img->mb_data[img->current_mb_nr];
  int*            bitCount  = currMB->bitcounter;
  SyntaxElement*  currSE    = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*          currSlice = img->currentSlice;
  const int*      partMap   = assignSE2partition[0];
  int             cbp       = currMB->cbp;
  DataPartition*  dataPart;



  //=====   C B P   =====
  //---------------------
  currSE->value1 = cbp;

  if (IS_OLDINTRA (currMB) || currMB->mb_type == SI4MB ||  currMB->mb_type == I8MB)
    currSE->type = SE_CBP_INTRA;
  else
    currSE->type = SE_CBP_INTER;

  // choose the appropriate data partition
  dataPart = &(currSlice->partArr[partMap[currSE->type]]);
  currSE->writing = writeCBP_CABAC_chroma;        
  dataPart->writeSyntaxElement(currSE, dataPart);
  bitCount[BITS_CBP_MB] += currSE->len;
  rate                  += currSE->len;

#if TRACE
  snprintf(currSE->tracestring, TRACESTRING_SIZE, "CBP (%2d,%2d) = %3d",img->mb_x, img->mb_y, cbp);
#endif
  // proceed to next SE
  currSE++;  
  currMB->currSEnr++;

  return rate;
}

