/* 
====================================================================================================================

The copyright in this software is being made available under the License included below. 
This software may be subject to other third party and 	contributor rights, including patent rights, and no such 
rights are granted under this license.

Copyright (c) 2010, QUALCOMM INC.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted only for 
the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
promoting such standards. The following conditions are required to be met:

* Redistributions of source code must retain the above copyright notice, this list of conditions and 
the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
the following disclaimer in the documentation and/or other materials provided with the distribution.
* The name of QUALCOMM INC.  may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

====================================================================================================================
*/

#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <limits.h>
#include <float.h>
#include <memory.h>

#include "global.h"
#include "mb_access.h"
#include "image.h"
#include "mode_decision.h"
#include "mb_access.h"

#ifdef MV_COMPETITION
#include "mv_competition.h"
extern int skip_mode;
#endif

//==== MODULE PARAMETERS ====
imgpel temp_imgY[16][16]; // to temp store the Y data for 8x8 transform
imgpel temp_imgU[16][16];
imgpel temp_imgV[16][16];

const int  b8_mode_table[6]  = {0, 4, 5, 6, 7};         // DO NOT CHANGE ORDER !!!
const int  mb_mode_table[9]  = {0, 1, 2, 3, P8x8, I16MB, I4MB, I8MB}; // DO NOT CHANGE ORDER !!!

extern int curr_mbi;

#ifdef GEOM
extern int test_geo_flag;
extern int Segmentation_Selected;
int encode_one_macroblock_geo (double min_rdcost0, int mb_ext_level, int do_Re_Encode);//GEOM_RESTURCTURE
void dense_motion_field16(int mbcount, int transform8x8);
void MV_segmentation(int mb32count, int mb_ext_level);
void save_mv_for_speedup(int mb_ext_level);
#endif

/*!
*************************************************************************************
* \brief
*    Initialize Encoding parameters for Macroblock
*************************************************************************************
*/
void init_enc_mb_params(Macroblock* currMB, RD_PARAMS *enc_mb, int intra, int bslice)
{
  int mode;
  int l,k;

  //Setup list offset
  enc_mb->list_offset[LIST_0] = LIST_0 + currMB->list_offset;
  enc_mb->list_offset[LIST_1] = LIST_1 + currMB->list_offset;

  enc_mb->best_ref[LIST_0] = 0;
  enc_mb->best_ref[LIST_1] = -1;

  // Set valid modes
  enc_mb->valid[I8MB]  = input->Transform8x8Mode;
  enc_mb->valid[I4MB]  = (input->Transform8x8Mode==2) ? 0:1;
  enc_mb->valid[I16MB] = 1;

  enc_mb->valid[0]     = (!intra && input->InterSearchSkip);
  enc_mb->valid[1]     = (!intra && input->InterSearch16x16);
  enc_mb->valid[2]     = (!intra && input->InterSearch16x8);
  enc_mb->valid[3]     = (!intra && input->InterSearch8x16);
  enc_mb->valid[4]     = (!intra && input->InterSearch8x8);
  enc_mb->valid[5]     = (!intra && input->InterSearch8x4 && !(input->Transform8x8Mode==2));
  enc_mb->valid[6]     = (!intra && input->InterSearch4x8 && !(input->Transform8x8Mode==2));
  enc_mb->valid[7]     = (!intra && input->InterSearch4x4 && !(input->Transform8x8Mode==2));
  enc_mb->valid[P8x8]  = (enc_mb->valid[4] || enc_mb->valid[5] || enc_mb->valid[6] || enc_mb->valid[7]);
  enc_mb->valid[12]    = 0;//(img->type == SI_SLICE);

  if (bslice && img->nal_reference_idc)
  { 
    enc_mb->lambda_md = img->lambda_md[5][img->masterQP];
    enc_mb->lambda_me = img->lambda_me[5][img->masterQP];
    enc_mb->lambda_mf = img->lambda_mf[5][img->masterQP];
  }
  else
  {
    enc_mb->lambda_md = img->lambda_md[img->type][img->masterQP];
    enc_mb->lambda_me = img->lambda_me[img->type][img->masterQP];
    enc_mb->lambda_mf = img->lambda_mf[img->type][img->masterQP];
  }

  // Initialize bipredME decisions
  for (mode=0; mode<MAXMODE; mode++)
  {
    img->bi_pred_me[mode]=0;
  }
  for (l = LIST_0; l < BI_PRED; l++)
  {
    for(k = 0; k < listXsize[l]; k++)
    {
      listX[l][k]->chroma_vector_adjustment= 0;
    }
  }

  if(final_mb_encoding || reencoding_during_modedecision) //GEOM_RESTURCTURE
  {
    int i;
    for(i=0; i<MAXMODE; i++)
      enc_mb->valid[i] = 0;

    enc_mb->valid[final_mb_encoding_bestmode] = 1;
    //if(final_mb_encoding_bestmode==0)
    //  enc_mb->valid[2] = 1;

    if(final_mb_encoding_bestmode == P8x8)    //this will be valid for 16x16 MB
    {            
      enc_mb->valid[4]     = 1;
      enc_mb->valid[5]     = 1;
      enc_mb->valid[6]     = 1;
      enc_mb->valid[7]     = 1;

      if(bslice)
        enc_mb->valid[0]     = 1;
    }
  }
}

/*!
*************************************************************************************
* \brief
*    computation of prediction list (including biprediction) cost
*************************************************************************************
*/
void list_prediction_cost(int list, int block, int mode, RD_PARAMS enc_mb, int bmcost[5], char best_ref[2])
{
  short ref;
  int mcost;
  int cur_list = list < BI_PRED ? enc_mb.list_offset[list] : enc_mb.list_offset[LIST_0];

  //--- get cost and reference frame for forward prediction ---

  if (list < BI_PRED)
  {
    for (ref=0; ref < listXsize[cur_list]; ref++)
    {
      mcost  = REF_COST (enc_mb.lambda_mf, ref, cur_list);     
      mcost += motion_cost[mode][list][ref][block];
      if (mcost < bmcost[list])
      {
        bmcost[list]   = mcost;
        best_ref[list] = (char)ref;
      }
    }
  }
  else if (list == BI_PRED)
  {
    bmcost[list]  = (REF_COST  (enc_mb.lambda_mf, (short)best_ref[LIST_0], cur_list)
      +  REF_COST  (enc_mb.lambda_mf, (short)best_ref[LIST_1], cur_list + LIST_1));    
    bmcost[list] += BIDPartitionCost (mode, block, (short)best_ref[LIST_0], (short)best_ref[LIST_1], enc_mb.lambda_mf,0);    
  }
  else
  {
    bmcost[list]  = (REF_COST (enc_mb.lambda_mf, 0, cur_list) 
      +  REF_COST (enc_mb.lambda_mf, 0, cur_list + LIST_1));
    bmcost[list] += BPredPartitionCost(mode, block, 0, 0, enc_mb.lambda_mf, !(list&1),0);                
  }
}  

/*!
*************************************************************************************
* \brief
*    Determination of prediction list based on simple distortion computation
*************************************************************************************
*/
void determine_prediction_list(int mode, int bmcost[5], char best_ref[2], short *best_pdir, int *cost, short *bi_pred_me)
{  
  if ((!input->BiPredMotionEstimation) || (mode != 1))
  {               
    //--- get prediction direction ----
    if  (bmcost[LIST_0] <= bmcost[LIST_1] 
    && bmcost[LIST_0] <= bmcost[BI_PRED])
    {
      *best_pdir = 0;
      //      if(*cost!=INT_MAX)
      *cost += bmcost[LIST_0];
      //best_ref[LIST_1] = 0;
    }
    else if (bmcost[LIST_1] <= bmcost[LIST_0] 
    &&     bmcost[LIST_1] <= bmcost[BI_PRED])
    {
      *best_pdir = 1;
      //      if(*cost!=INT_MAX)
      *cost += bmcost[LIST_1];
      //best_ref[LIST_0] = 0;
    }
    else
    {
      *best_pdir = 2;
      //      if((bmcost[BI_PRED]==INT_MAX) | (*cost==INT_MAX))
      //        *cost=INT_MAX;
      //      else
      *cost += bmcost[BI_PRED];
      //best_ref[LIST_1] = 0;
    }
  }
  else
  {                            
    img->bi_pred_me[mode]=0;
    *bi_pred_me = 0;
    //--- get prediction direction ----
    if  (bmcost[LIST_0] <= bmcost[LIST_1] 
    && bmcost[LIST_0] <= bmcost[BI_PRED] 
    && bmcost[LIST_0] <= bmcost[BI_PRED_L0] 
    && bmcost[LIST_0] <= bmcost[BI_PRED_L1])
    {
      *best_pdir = 0;
      *cost += bmcost[LIST_0];
      //best_ref[LIST_1] = 0;
    }
    else if (bmcost[LIST_1] <= bmcost[LIST_0] 
    &&     bmcost[LIST_1] <= bmcost[BI_PRED] 
    &&     bmcost[LIST_1] <= bmcost[BI_PRED_L0] 
    &&     bmcost[LIST_1] <= bmcost[BI_PRED_L1])
    {
      *best_pdir = 1;
      *cost += bmcost[LIST_1];
      //best_ref[LIST_0] = 0;
    }
    else if (bmcost[BI_PRED] <= bmcost[LIST_0] 
    &&     bmcost[BI_PRED] <= bmcost[LIST_1] 
    &&     bmcost[BI_PRED] <= bmcost[BI_PRED_L0] 
    &&     bmcost[BI_PRED] <= bmcost[BI_PRED_L1])
    {
      *best_pdir = 2;
      *cost += bmcost[BI_PRED];
      //best_ref[LIST_1] = 0;  
    }
    else if (bmcost[BI_PRED_L0] <= bmcost[LIST_0] 
    &&     bmcost[BI_PRED_L0] <= bmcost[LIST_1] 
    &&     bmcost[BI_PRED_L0] <= bmcost[BI_PRED]
    &&     bmcost[BI_PRED_L0] <= bmcost[BI_PRED_L1])
    {
      *best_pdir = 2;
      *cost += bmcost[BI_PRED_L0];                  
      *bi_pred_me = 1;
      img->bi_pred_me[mode]=1;
      best_ref[LIST_1] = 0;
      best_ref[LIST_0] = 0;
    }
    else
    {
      *best_pdir = 2;
      *cost += bmcost[BI_PRED_L1];                                  
      *bi_pred_me = 2;
      best_ref[LIST_1] = 0;
      best_ref[LIST_0] = 0;
      img->bi_pred_me[mode]=2;
    }               
  }
}

/*!
*************************************************************************************
* \brief
*    RD decision process
*************************************************************************************
*/

void compute_mode_RD_cost(int mode, 
                          int *mode_idx,
                          Macroblock *currMB, 
                          RD_PARAMS enc_mb, 
                          double *min_rdcost, 
                          double *min_rate, 
                          int i16mode, 
                          short bslice, 
                          short *inter_skip,
                          short QPcount)
{
  //--- transform size ---           
  currMB->luma_transform_size_8x8_flag = input->Transform8x8Mode==2
    ?  (mode >= 1 && mode <= 3)
    || (mode == 0 && bslice && active_sps->direct_8x8_inference_flag)
    || ((mode == P8x8) && (enc_mb.valid[4]))
    :  0;
  //store_coding_state (cs_cm); // RD  
  SetModesAndRefframeForBlocks (mode);

  // Encode with coefficients
  img->NoResidueDirect = 0;  
  if (currMB->c_ipred_mode == DC_PRED_8 || (IS_INTRA(currMB) ))
  {
    while(1)
    {
#ifdef MV_COMPETITION
      if ((input->mv_competition > 0) && (mode == 0) && (img->type==P_SLICE))
        skip_mode = TRUE;
      else 
        skip_mode = FALSE;
#endif

      if (RDCost_for_macroblocks (enc_mb.lambda_md, mode, mode_idx, min_rdcost, min_rate, i16mode, QPcount))
      {
        store_macroblock_parameters (mode);
      }

      // Go through transform modes. 
      // Note that if currMB->cbp is 0 one could choose to skip 8x8 mode
      // although this could be due to deadzoning decisions.
      //if (input->Transform8x8Mode==1 && currMB->cbp!=0) 
      if (input->Transform8x8Mode==1)
      {
        //=========== try mb_types 1,2,3 with 8x8 transform ===========
        if ((mode >= 1 && mode <= 3) && currMB->luma_transform_size_8x8_flag == 0)
        {
          //try with 8x8 transform size
          currMB->luma_transform_size_8x8_flag = 1;
          continue;
        }
        //=========== try mb_types 1,2,3 with 16x16 transform ===========
        else if ((mode >= 1 && mode <= 3) && currMB->luma_transform_size_8x8_flag == 1)
        {
          //try with 16x16 transform size
          currMB->luma_transform_size_8x8_flag = 2;
          continue;
        }
        //=========== try P-DIRECT-MODE with 8x8 transform ===========
        else if (mode == 0 && img->type==P_SLICE && currMB->luma_transform_size_8x8_flag == 0)
        {
          //try with 8x8 transform size
          currMB->luma_transform_size_8x8_flag = 1;
          continue;
        }
        else if (mode == 0 && img->type==P_SLICE && currMB->luma_transform_size_8x8_flag == 1)
        {
          //try with 16x16 transform size
          currMB->luma_transform_size_8x8_flag = 2;
          continue;
        }
        //=========== try DIRECT-MODE with 8x8 transform ===========
        else if (mode == 0 && bslice && active_sps->direct_8x8_inference_flag && currMB->luma_transform_size_8x8_flag == 0)
        {
          //try with 8x8 transform size
          currMB->luma_transform_size_8x8_flag = 1;
          continue;
        }
        else if (mode == 0 && bslice && active_sps->direct_8x8_inference_flag && currMB->luma_transform_size_8x8_flag == 1)
        {
          //try with 16x16 transform size
          currMB->luma_transform_size_8x8_flag = 2;
          continue;
        }
        //=========== try mb_type P8x8 for mode 4 with 4x4/8x8 transform ===========
        else if ((mode == P8x8) && (enc_mb.valid[4]) && (currMB->luma_transform_size_8x8_flag == 0))
        {
          currMB->luma_transform_size_8x8_flag = 1; //check 8x8 partition for transform size 8x8
          (*mode_idx)++;    //BACKUP_PREDICTION
          continue;
        }
        else
        {
          currMB->luma_transform_size_8x8_flag = 0;
          break;
        }
      }
      else
        break;
    }

    //encode P_SKIP mode  ...above mode=0 was for P_Direct to encode with coefficients..
    if ( img->type==P_SLICE && mode == 0 && cbp && enc_mb.valid[mode] )
    {
      img->NoResidueDirect = 1; 
      currMB->luma_transform_size_8x8_flag = 0;
#ifdef MV_COMPETITION
      if ((input->mv_competition > 0) && (mode == 0) && (img->type==P_SLICE))
        skip_mode = TRUE;
      else 
        skip_mode = FALSE;
#endif

      if (RDCost_for_macroblocks (enc_mb.lambda_md, mode, mode_idx, min_rdcost, min_rate, i16mode, QPcount))
      {
        store_macroblock_parameters (mode);
      }
    }

    // Encode with no coefficients. Currently only for direct. This could be extended to all other modes as in example.
    //if (mode < P8x8 && (*inter_skip == 0) && enc_mb.valid[mode] && currMB->cbp && (currMB->cbp&15) != 15 && !input->nobskip)
    if ( bslice && mode == 0 && (*inter_skip == 0) && enc_mb.valid[mode] && cbp && !input->nobskip) 
    {
      img->NoResidueDirect = 1;
      currMB->luma_transform_size_8x8_flag = 0;
      if (RDCost_for_macroblocks (enc_mb.lambda_md, mode, mode_idx, min_rdcost, min_rate, i16mode, QPcount)) 
      {
        store_macroblock_parameters (mode);
      }
    }


  }
};

/*!
*************************************************************************************
* \brief
*    Mode Decision for an 8x8 sub-macroblock
*************************************************************************************
*/
double submacroblock_mode_decision(RD_8x8DATA *dataTr, 
                                   RD_PARAMS enc_mb, 
                                   Macroblock *currMB,
                                   short bslice, 
                                   int block, 
                                   int transform8x8, int Re_Encode)  
{
  int j0, i0, j1, i1;
  int i,j, k;
  int index;
  double min_rdcost, rdcost = 0.0;
  short best_pdir = 0;
  char best_ref[2] = {0, -1};
  int mode;
  int64 curr_cbp_blk;
  int bmcost[5] = {INT_MAX};
  int cnt_nonz = 0;
  short bi_pred_me;
  int best_cnt_nonz = 0;
  int maxindex =  (transform8x8) ? 2 : 5;
  int pix_x, pix_y;
  int block_x, block_y;
  int cost=0;

  currMB->mb_type=P8x8;
  //--- set coordinates ---
  j0 = ((block/2)<<3);    j1 = (j0>>2);
  i0 = ((block%2)<<3);    i1 = (i0>>2);

  if (transform8x8)
    currMB->luma_transform_size_8x8_flag = 1; //switch to transform size 8x8

  //--- store coding state before coding ---
  store_coding_state (cs_smb);

  //=====  LOOP OVER POSSIBLE CODING MODES FOR 8x8 SUB-PARTITION  =====
  for (min_rdcost=1e30, index=(bslice?0:1); index<maxindex; index++)
  {
    mode = b8_mode_table[index]; 
    cost = 0;
    if (enc_mb.valid[mode] && (transform8x8 == 0 || mode != 0 || (mode == 0 && active_sps->direct_8x8_inference_flag)))
    {
      curr_cbp_blk = 0;

      if (mode==0)
      {
        //--- Direct Mode ---                                      
        block_x = img->block_x+(block&1)*2;
        block_y = img->block_y+(block&2);
        best_ref[LIST_0] = direct_ref_idx[LIST_0][block_y][block_x];
        best_ref[LIST_1] = direct_ref_idx[LIST_1][block_y][block_x];
        best_pdir   = direct_pdir[block_y][block_x];
      } // if (mode==0)
      else
      {      
        if (Re_Encode)
        {   
          best_ref[LIST_0] = (char)best8x8fwref[mode][block];
          best_ref[LIST_1] = (char)best8x8bwref[mode][block];
          best_pdir        = best8x8pdir [mode][block];
        }
        else
        {
          //======= motion estimation for all reference frames ========
          //-----------------------------------------------------------
          PartitionMotionSearch (mode, block, enc_mb.lambda_mf,transform8x8, 0);

          //--- get cost and reference frame for LIST 0 prediction ---
          bmcost[LIST_0] = INT_MAX;
          list_prediction_cost(LIST_0, block, mode, enc_mb, bmcost, best_ref);

          //store LIST 0 reference index for every block
          block_x = img->block_x+(block&1)*2;
          block_y = img->block_y+(block&2);
          for (j = block_y; j< block_y + 2; j++)
          {
            for (i = block_x; i < block_x + 2; i++)
            {
              enc_picture->ref_idx   [LIST_0][j][i] = best_ref[LIST_0];
              enc_picture->ref_pic_id[LIST_0][j][i] = 
                enc_picture->ref_pic_num[enc_mb.list_offset[LIST_0]][(short)best_ref[LIST_0]];
            }
          } 

          if (bslice)
          {
            //--- get cost and reference frame for LIST 1 prediction ---
            bmcost[LIST_1] = INT_MAX;
            list_prediction_cost(LIST_1, block, mode, enc_mb, bmcost, best_ref);

            bmcost[BI_PRED] = INT_MAX;
            // Compute bipredictive cost between best list 0 and best list 1 references
            list_prediction_cost(BI_PRED, block, mode, enc_mb, bmcost, best_ref);

            //--- get prediction direction ----          
            determine_prediction_list(mode, bmcost, best_ref, &best_pdir, &cost, &bi_pred_me);

            //store backward reference index for every block
            for (j = block_y; j< block_y + 2; j++)
            {
              for (i = block_x; i < block_x + 2; i++)
              {
                enc_picture->ref_idx[LIST_0][j][i] = best_ref[LIST_0];
                enc_picture->ref_idx[LIST_1][j][i] = best_ref[LIST_1];
              }
            }
          } // if (bslice)
          else
          {
            best_pdir = 0;
            cost     = bmcost[LIST_0];
          }
          SetRefAndMotionVectors_to_ExtMBdata (block,mode, best_pdir, best_ref[LIST_0], best_ref[LIST_1],bi_pred_me,0,transform8x8);
        }   
      } // if (mode!=0)


      //--- get and check rate-distortion cost ---
      rdcost = RDCost_for_8x8blocks (&cnt_nonz, &curr_cbp_blk, enc_mb.lambda_md,
        block, mode, best_pdir, best_ref[LIST_0], best_ref[LIST_1], &min_rdcost);

      //--- set variables if best mode has changed ---
      if ( rdcost < min_rdcost )
      {
        min_rdcost                  = rdcost;
        dataTr->rdcost[block]       = rdcost;
        dataTr->part8x8mode [block] = mode;
        dataTr->part8x8pdir [block] = best_pdir;
        dataTr->part8x8fwref[block] = best_ref[LIST_0];
        dataTr->part8x8bwref[block] = best_ref[LIST_1];

        img->mb_data[img->current_mb_nr].b8mode[block] = mode;

        //--- store number of nonzero coefficients ---
        best_cnt_nonz  = cnt_nonz;

        {
          //--- store block cbp ---
          dataTr->cbp_blk8x8 &= (~(0x33 << (((block>>1)<<3)+((block%2)<<1)))); // delete bits for block
          dataTr->cbp_blk8x8 |= curr_cbp_blk;

          //--- store coefficients ---
          for (k=0; k< 4; k++)
          {
            for (j=0; j< 2; j++)
              for (i=0; i<65; i++)  
                dataTr->cofAC8x8[block][k][j][i] =  img->cofAC[block][k][j][i]; // 18->65 for ABT
          }   
          //--- store reconstruction and prediction --- 
          for (j=j0; j<j0+8; j++)    
          {
            pix_y = img->pix_y + j;
            for (i=i0; i<i0+8; i++)
            {
              pix_x = img->pix_x + i;
              dataTr->rec_mbY8x8[j][i] = enc_picture->imgY[pix_y][pix_x];
              dataTr->mpr8x8[j][i] = img->mpr[j][i];
            }     
          }
        }                  
        //--- store best 8x8 coding state ---
        if (block < 3)
          store_coding_state (cs_b8);
      } // if (rdcost <= min_rdcost)

      //--- re-set coding state as it was before coding with current mode was performed ---
      reset_coding_state (cs_smb);

    } // if ((enc_mb.valid[mode] && (transform8x8 == 0 || mode != 0 || (mode == 0 && active_sps->direct_8x8_inference_flag)))
  } // for (min_rdcost=1e30, index=(bslice?0:1); index<6; index++)  

  //----- set cbp and count of nonzero coefficients ---
  if (best_cnt_nonz)
  {
    //r    cbp8x8       |= (1 << block);
    //r    cnt_nonz_8x8 += best_cnt_nonz;
    dataTr->cbp8x8 |= (1 << block);
    dataTr->cnt_nonz_8x8 += best_cnt_nonz;
  }

  if (!transform8x8)
  {
    if (block<3)
    {
      //===== re-set reconstructed block =====
      j0   = 8*(block/2);
      i0   = 8*(block%2);
      for (j=j0; j<j0 + 2 * BLOCK_SIZE; j++)
      {
        memcpy(&enc_picture->imgY[img->pix_y + j][img->pix_x], dataTr->rec_mbY8x8[j], 2 * BLOCK_SIZE * sizeof(imgpel));
      }
    } // if (block<3)   

    //======= save motion data for 8x8 partition for transform size 4x4 ========
    StoreNewMotionVectorsBlock8x8(1, block, dataTr->part8x8mode[block], dataTr->part8x8fwref[block], dataTr->part8x8bwref[block], dataTr->part8x8pdir[block], bslice);
  }  
  else
  {  
    //======= save motion data for 8x8 partition for transform size 8x8 ========
    StoreNewMotionVectorsBlock8x8(0, block, dataTr->part8x8mode[block], dataTr->part8x8fwref[block], dataTr->part8x8bwref[block], dataTr->part8x8pdir[block], bslice);
  }  
  //===== set motion vectors and reference frames (prediction) =====
  SetRefAndMotionVectors (block, dataTr->part8x8mode[block], dataTr->part8x8pdir[block], dataTr->part8x8fwref[block], dataTr->part8x8bwref[block]);

  //===== set the coding state after current block =====
  //if (transform8x8 == 0 || block < 3)
  if (block < 3)
    reset_coding_state (cs_b8);

  return min_rdcost;
}

/*!
*************************************************************************************
* \brief
*    Mode Decision for a macroblock
*************************************************************************************
*/
double encode_one_macroblock (int Re_Encode)
{  
  int         rerun, block, mode, i, j;
  short       best_pdir;
  RD_PARAMS   enc_mb;
  double      min_rdcost=1e30;
  char        best_ref[2] = {0, -1};
  int         bmcost[5] = {INT_MAX};
  int         cost=0;
  int         min_cost = INT_MAX;
  int         intra1 = 0;
  int         best_transform_flag = 0;
  short       islice      = (img->type==I_SLICE);
  short       bslice      = (img->type==B_SLICE);
  short       pslice      = (img->type==P_SLICE);
  short       intra       = (islice || (pslice && img->mb_y==img->mb_y_upd && img->mb_y_upd!=img->mb_y_intra));
  short       runs        = 1;

  short bestQPcount=0;

  Macroblock* currMB      = &img->mb_data[img->current_mb_nr];
  // Fast Mode Decision
  short inter_skip = 0;

  isfirstQP = 1;    //its imp to set this to 1 for ME

#ifdef RDO_Q
  if(input->UseRDO_Q && active_pps->entropy_coding_mode_flag == CABAC)
  {
    estRunLevel_CABAC(LUMA_4x4);
    estRunLevel_CABAC(LUMA_8x8);
    estRunLevel_CABAC(LUMA_16AC);
#ifdef USE_INTRA_MDDT
    estRunLevel_CABAC(LUMA_16x16);
#endif
    estRunLevel_CABAC(LUMA_16x16P);
    estRunLevel_CABAC(LUMA_16x8P);
    estRunLevel_CABAC(LUMA_8x16P);
  }
#endif



  //===== Setup Macroblock encoding parameters =====
  init_enc_mb_params(currMB, &enc_mb, intra, bslice);
#ifdef GEOM//GEOM_RESTURCTURE
  if(final_mb_encoding || reencoding_during_modedecision)
  {
    if(final_mb_encoding_geo_mode != -1)
      enc_mb.valid[3] = 0;
  }
#endif  
  // Perform multiple encodings if rdopt with losses is enabled
  for (rerun=0; rerun<runs; rerun++)
  {

    // reset chroma intra predictor to default
    currMB->c_ipred_mode = DC_PRED_8;

    //=====   S T O R E   C O D I N G   S T A T E   =====
    //---------------------------------------------------
    store_coding_state (cs_cm);

    if (!intra)
    {
      //===== set direct motion vectors =====
      best_mode = 1;
      if (bslice)
      {
        Get_Direct_Motion_Vectors (0);
      }

      if (Re_Encode)
      {
        for (min_cost=INT_MAX, mode=1; mode<4; mode++)
        {
          if (enc_mb.valid[mode] && !inter_skip)
          {
            LoadRefAndMotionVectors_From_ExtMBdata(mode,best8x8pdir[mode],best8x8fwref[mode],best8x8bwref[mode],&bi_pred_me,1,1); 
          }
        }
      }
      else
      {
        //===== MOTION ESTIMATION FOR 16x16, 16x8, 8x16 BLOCKS =====      
        for (min_cost=INT_MAX, mode=1; mode<4; mode++)
        {
          bi_pred_me = 0;
          img->bi_pred_me[mode]=0;
          if (enc_mb.valid[mode] && !inter_skip)
          {
            for (cost=0, block=0; block<(mode==1?1:2); block++)
            {
              PartitionMotionSearch (mode, block, enc_mb.lambda_mf, 0, 0);

              //--- set 4x4 block indizes (for getting MV) ---
              j = (block==1 && mode==2 ? 2 : 0);
              i = (block==1 && mode==3 ? 2 : 0);

              //--- get cost and reference frame for List 0 prediction ---
              bmcost[LIST_0] = INT_MAX;
              list_prediction_cost(LIST_0, block, mode, enc_mb, bmcost, best_ref);

              if (bslice)
              {
                //--- get cost and reference frame for List 1 prediction ---
                bmcost[LIST_1] = INT_MAX;
                list_prediction_cost(LIST_1, block, mode, enc_mb, bmcost, best_ref);

                bmcost[BI_PRED] = INT_MAX;
                // Compute bipredictive cost between best list 0 and best list 1 references
                list_prediction_cost(BI_PRED, block, mode, enc_mb, bmcost, best_ref);

                // Finally, if mode 16x16, compute cost for bipredictive ME vectore
                if (input->BiPredMotionEstimation && mode == 1)
                {                
                  list_prediction_cost(BI_PRED_L0, block, mode, enc_mb, bmcost, 0);
                  list_prediction_cost(BI_PRED_L1, block, mode, enc_mb, bmcost, 0);
                }
                else
                {
                  bmcost[BI_PRED_L0] = INT_MAX;
                  bmcost[BI_PRED_L1] = INT_MAX;
                }

                // Determine prediction list based on mode cost
                determine_prediction_list(mode, bmcost, best_ref, &best_pdir, &cost, &bi_pred_me);
              }
              else // if (bslice)
              {
                best_pdir  = 0;
                cost      += bmcost[LIST_0];
              }

              //----- set reference frame and direction parameters -----
              if (mode==3)
              {
                best8x8fwref [3][block  ] = best8x8fwref [3][  block+2] = best_ref[LIST_0];
                best8x8pdir  [3][block  ] = best8x8pdir  [3][  block+2] = best_pdir;
                best8x8bwref [3][block  ] = best8x8bwref [3][  block+2] = best_ref[LIST_1];
              }
              else if (mode==2)
              {
                best8x8fwref [2][2*block] = best8x8fwref [2][2*block+1] = best_ref[LIST_0];
                best8x8pdir  [2][2*block] = best8x8pdir  [2][2*block+1] = best_pdir;
                best8x8bwref [2][2*block] = best8x8bwref [2][2*block+1] = best_ref[LIST_1];
              }
              else
              {
                best8x8fwref [1][0] = best8x8fwref [1][1] = best8x8fwref [1][2] = best8x8fwref [1][3] = best_ref[LIST_0];
                best8x8pdir  [1][0] = best8x8pdir  [1][1] = best8x8pdir  [1][2] = best8x8pdir  [1][3] = best_pdir;
                best8x8bwref [1][0] = best8x8bwref [1][1] = best8x8bwref [1][2] = best8x8bwref [1][3] = best_ref[LIST_1];
              }

              //--- set reference frames and motion vectors ---
              if (mode>1 && block==0)
                SetRefAndMotionVectors (block, mode, best_pdir, best_ref[LIST_0], best_ref[LIST_1]);            

              SetRefAndMotionVectors_to_ExtMBdata(block,mode, best_pdir, best_ref[LIST_0], best_ref[LIST_1],bi_pred_me,1,1); 
            } // for (block=0; block<(mode==1?1:2); block++)

            if ((!inter_skip) && (cost < min_cost))
            {
              best_mode = mode;
              min_cost  = cost;
              best_transform_flag = currMB->luma_transform_size_8x8_flag;
            }
          } // if (enc_mb.valid[mode])
        } // for (mode=1; mode<4; mode++)
      }
      if ((!inter_skip) && enc_mb.valid[P8x8])
      {
        giRDOpt_B8OnlyFlag = 1;

        //===== store coding state of macroblock =====
        store_coding_state (cs_mb);

        currMB->all_blk_8x8 = -1;

        if (input->Transform8x8Mode)
        { 
          if (Re_Encode)
          {
            LoadRefAndMotionVectors_From_ExtMBdata(4,best8x8pdir[4],best8x8fwref[4],best8x8bwref[4],&bi_pred_me,0,1); 
          }
          min_rdcost = compute_P8x8_RD_cost_RDOQ_Tr8x8(currMB, enc_mb, Re_Encode);
        }
        currMB->luma_transform_size_8x8_flag = 0; //switch to 4x4 transform size
        if (input->Transform8x8Mode != 2)  
        {
          if (Re_Encode)
          {
            LoadRefAndMotionVectors_From_ExtMBdata(4,best8x8pdir[4],best8x8fwref[4],best8x8bwref[4],&bi_pred_me,0,0); 
            LoadRefAndMotionVectors_From_ExtMBdata(5,best8x8pdir[5],best8x8fwref[5],best8x8bwref[5],&bi_pred_me,0,0);
            LoadRefAndMotionVectors_From_ExtMBdata(6,best8x8pdir[6],best8x8fwref[6],best8x8bwref[6],&bi_pred_me,0,0);
            LoadRefAndMotionVectors_From_ExtMBdata(7,best8x8pdir[7],best8x8fwref[7],best8x8bwref[7],&bi_pred_me,0,0);
          }
          min_rdcost = compute_P8x8_RD_cost_RDOQ_Tr4x4(currMB, enc_mb, Re_Encode);
        }

        //--- re-set coding state (as it was before 8x8 block coding) ---
        reset_coding_state (cs_mb);
        giRDOpt_B8OnlyFlag = 0;
      }

      // Find a motion vector for the Skip mode
      if(pslice)
        FindSkipModeMotionVector (0);
    }   
    else // if (!intra)
    {
      min_cost = INT_MAX;
    }

    //========= C H O O S E   B E S T   M A C R O B L O C K   M O D E =========
    //-------------------------------------------------------------------------
    min_rdcost = compute_mode_RD_cost_RDOQ(currMB, enc_mb, &bestQPcount);

    if (rerun==0)
      intra1 = IS_INTRA(currMB);
  } // for (rerun=0; rerun<runs; rerun++) 



  //=====  S E T   F I N A L   M A C R O B L O C K   P A R A M E T E R S ======
  //---------------------------------------------------------------------------  

#ifdef GEOM
  if(img->type != I_SLICE && Segmentation_Selected == 0)
    save_mv_for_speedup(0);

#ifdef GEOM//GEOM_RESTURCTURE
  if(img->type != I_SLICE && ((!Re_Encode && enc_mb.valid[3])||final_mb_encoding_geo_mode!=-1))
#else
  if(img->type != I_SLICE && enc_mb.valid[3])
#endif
  {
    if(encode_one_macroblock_geo(min_rdcost, 0, Re_Encode))//GEOM_RESTURCTURE
      return rdopt->min_rdcost;
  }
  //else
#endif
  {
    if (cbp!=0 || best_mode==I16MB)
      currMB->prev_cbp = 1;    
    else if (cbp==0)
    {
      currMB->delta_qp = 0;
      currMB->qp       = currMB->prev_qp;
      set_chroma_qp(currMB);
      img->qp          = currMB->qp;
      currMB->prev_cbp = 0;
    }    
    set_stored_macroblock_parameters (bestQPcount);
  }


  rdopt->min_rdcost = min_rdcost;

  //--- constrain intra prediction ---
  if(input->UseConstrainedIntraPred && (img->type==P_SLICE || img->type==B_SLICE))
  {
    if( !IS_INTRA(currMB) )
      img->intra_block[img->current_mb_nr] = 0;
    else
      img->intra_block[img->current_mb_nr] = 1;
  }

  return(rdopt->min_rdcost);

}

double compute_mode_RD_cost_RDOQ(Macroblock *currMB, RD_PARAMS enc_mb, short *bestQPcount)
{
  int mode_idx;
  int max_index, index=0, mode, ctr16x16;
  int i16mode=0;
  double dummy1;
  short dummy2 = 0;

  short       islice      = (img->type==I_SLICE);
  short       bslice      = (img->type==B_SLICE);
  short       pslice      = (img->type==P_SLICE);
  short       intra       = (islice || (pslice && img->mb_y==img->mb_y_upd && img->mb_y_upd!=img->mb_y_intra));

  //  Macroblock* currMB      = &img->mb_data[img->current_mb_nr];
  short   max_chroma_pred_mode;

  int deltaQP,  bestQP=currMB->qp;
  double RDOQ_min_rdcost=1e30, min_rdcost=1e30; 
  int qp_left, qp_up;
  int best_deltaQP=0;
  int deltaQPCnt; 
  int master_cr_qp=0, master_cr_cbp=0, chroma_qp=0;

  static int deltaQPTabB[] = {0, 2, 3};
  int deltaQPTabSizeB = 3;

  static int deltaQPTabP[] = {0, 1, -1, 2, -2};
  int deltaQPTabSizeP = 1;

  int deltaQPTabSize = bslice? deltaQPTabSizeB : (pslice? deltaQPTabSizeP : NumberOfQP);
  int qp_anchor;

  qp_left = (currMB->mb_available_left) ? currMB->mb_available_left->qp:img->masterQP;
  qp_up   = (currMB->mb_available_up  ) ? currMB->mb_available_up->qp  :img->masterQP;
  qp_anchor = (qp_left + qp_up + 1) >> 1;

  isfirstQP = 1;
  for (deltaQPCnt=((img->type == I_SLICE && input->UseRDO_Q)?(NumberOfQP-1):0); deltaQPCnt<deltaQPTabSize; deltaQPCnt++)
  {
    if(bslice)
      deltaQP = deltaQPTabB[deltaQPCnt];
    else
      deltaQP = deltaQPTabP[deltaQPCnt];

    currMB->qp=Clip3(0, 51, (img->masterQP+deltaQP));
    img->qp = currMB->qp;       //this is set coz img->qp is used in 8x8 DCT, KLT, 16x16 DCT, 16x8 DCT, 8x16 DCT
    //for 4x4 DCT it uses currMB->qp......inconsistency
    currMB->delta_qp = currMB->qp - currMB->prev_qp;
    set_chroma_qp(currMB);

    if(final_mb_encoding || reencoding_during_modedecision)
    {
      if(currMB->qp != finalbestQP)
        continue;
    }
    else   //skip these checking during final QP
    {
      // skip this QP if deltaQP is too large based on local stats
      if(!bslice && (deltaQP != 0) && !(currMB->qp-qp_anchor >= -2 && currMB->qp-qp_anchor <= 1) && currMB->mb_available_left && currMB->mb_available_up)
        continue; 
      if(bslice  &&                   !(currMB->qp-qp_anchor >= -1 && currMB->qp-qp_anchor <= 2) && currMB->mb_available_left && currMB->mb_available_up)
        continue;

      //prevent QP to become too high..otherwise chroma will be affected.
      if(input->successive_Bframe==0)
      {     
        static byte QP_SCALE_CR[52]=
        {
          0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,
          12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,
          28,29,29,30,31,32,32,33,34,34,35,35,36,36,37,37,
          37,38,38,38,39,39,39,39
        };

        chroma_qp = Clip3 ( -img->bitdepth_chroma_qp_scale, 51, currMB->qp + img->chroma_qp_offset[0] );
        chroma_qp = (chroma_qp < 0 ? chroma_qp : QP_SCALE_CR[chroma_qp]) + img->bitdepth_chroma_qp_scale;
        if( (!isfirstQP) && (chroma_qp > master_cr_qp) && (master_cr_cbp!=0) )
          continue;
      }
    }

    store_coding_state (cs_eachQP);
    //==========================================
    {
      int mb_available_up, mb_available_left, mb_available_up_left;
      max_index = 8;

      if (input->BiPredMotionEstimation)
        img->bi_pred_me[1] =0;  

      // precompute all new chroma intra prediction modes
      IntraChromaPrediction(&mb_available_up, &mb_available_left, &mb_available_up_left);
      max_chroma_pred_mode = PLANE_8;

      for (currMB->c_ipred_mode=DC_PRED_8; currMB->c_ipred_mode<=max_chroma_pred_mode; currMB->c_ipred_mode++)
      {
        // bypass if c_ipred_mode is not allowed
        if ( (  ((!intra || !input->IntraDisableInterOnly) && input->ChromaIntraDisable == 1 && currMB->c_ipred_mode!=DC_PRED_8) 
          || (currMB->c_ipred_mode == VERT_PRED_8 && !mb_available_up) 
          || (currMB->c_ipred_mode == HOR_PRED_8  && !mb_available_left) 
          || (currMB->c_ipred_mode == PLANE_8     && (!mb_available_left || !mb_available_up || !mb_available_up_left))))
          continue;        

        mode_idx = 0;
        //===== GET BEST MACROBLOCK MODE =====
        for (ctr16x16=0, index=0; index < max_index; index++)
        {
          mode = mb_mode_table[index];
          i16mode = 0; 
          //--- for INTER16x16 check all prediction directions ---
          if (mode==1 && bslice)
          {
            best8x8pdir[1][0] = best8x8pdir[1][1] = best8x8pdir[1][2] = best8x8pdir[1][3] = ctr16x16;

            if ( (bslice) && (input->BiPredMotionEstimation) 
              && (ctr16x16 == 2 && img->bi_pred_me[mode] < 2 && mode == 1))
              ctr16x16--;
            if (ctr16x16 < 2) 
              index--;
            ctr16x16++;
          }

          // Skip intra modes in inter slices if best inter mode is 
          // a MB partition and cbp is 0.
          if (input->SkipIntraInInterSlices && !intra && mode >= I16MB 
            && best_mode <=3 && currMB->cbp == 0)
            continue;

          if (enc_mb.valid[mode])
            compute_mode_RD_cost(mode, &mode_idx, currMB, enc_mb, &min_rdcost, &dummy1, i16mode, bslice, &dummy2, deltaQPCnt);

          if ((input->BiPredMotionEstimation) && (bslice) && ctr16x16 == 2 
            && img->bi_pred_me[mode] < 2 && mode == 1 && best8x8pdir[1][0] == 2) 
            img->bi_pred_me[mode] = img->bi_pred_me[mode] + 1;

          mode_idx++;   //BACKUP_PREDICTION

        }// for (ctr16x16=0, index=0; index<max_index; index++)
      }// for (currMB->c_ipred_mode=DC_PRED_8; currMB->c_ipred_mode<=max_chroma_pred_mode; currMB->c_ipred_mode++)

    }
    //==========================================
    reset_coding_state (cs_eachQP);

    if(!(final_mb_encoding || reencoding_during_modedecision))
    {
      if(input->successive_Bframe==0)
      {     
        //chroma_qp = currMB->qpc[0];
        if(isfirstQP) // deltaQP =0 has to be test first!
        {
          master_cr_cbp = img->mb_data[img->current_mb_nr].cbp>>4;
          master_cr_qp  = chroma_qp;
        }
      }
    }

    if (min_rdcost < RDOQ_min_rdcost)
    {
      RDOQ_min_rdcost=min_rdcost;
      bestQP=currMB->qp;
      best_deltaQP = deltaQPCnt;  //deltaQP;
      saved_best_mode = best_mode;
    }

    isfirstQP = 0;

    if(best_mode==0 && cbp==0)   //no need to check other QPs if the best mode is SKIP for current QP
      break;

  }//for (deltaQPCnt=0; deltaQPCnt<deltaQPTabSize; deltaQPCnt++)


  currMB->qp = bestQP;  //Clip3(0, 51, (img->masterQP + best_deltaQP));
  img->qp    = bestQP;
  currMB->delta_qp = currMB->qp - currMB->prev_qp;
  set_chroma_qp(currMB);

  *bestQPcount = best_deltaQP;
  return RDOQ_min_rdcost;
}

double compute_P8x8_RD_cost_RDOQ_Tr8x8  (Macroblock *currMB, RD_PARAMS enc_mb, int Re_Encode)
{
  int block=0;
  short bslice  = (img->type==B_SLICE);

  int deltaQP;
  double min_rdcost=1e30; 
  int qp_left, qp_up;
  //int best_deltaQP=0;
  int deltaQPCnt; 
  int master_cr_qp=0, master_cr_cbp=0, chroma_qp=0;

  static int deltaQPTabB[] = {0, 2, 3};
  int deltaQPTabSizeB = 3;

  static int deltaQPTabP[] = {0, 1, -1, 2, -2};
  int deltaQPTabSizeP = 1;

  int deltaQPTabSize = bslice? deltaQPTabSizeB : deltaQPTabSizeP;
  int qp_anchor;

  qp_left = (currMB->mb_available_left) ? currMB->mb_available_left->qp:img->masterQP;
  qp_up   = (currMB->mb_available_up  ) ? currMB->mb_available_up->qp  :img->masterQP;
  qp_anchor = (qp_left + qp_up + 1) >> 1;

  isfirstQP = 1;
  for (deltaQPCnt=0; deltaQPCnt<deltaQPTabSize; deltaQPCnt++)
  {
    if(bslice)
      deltaQP = deltaQPTabB[deltaQPCnt];
    else
      deltaQP = deltaQPTabP[deltaQPCnt];


    currMB->qp=Clip3(0, 51, (img->masterQP+deltaQP));
    img->qp = currMB->qp;       //this is set coz img->qp is used in 8x8 DCT, KLT, 16x16 DCT, 16x8 DCT, 8x16 DCT
    //for 4x4 DCT it uses currMB->qp......incosistency
    currMB->delta_qp = currMB->qp - currMB->prev_qp;
    set_chroma_qp(currMB);

    if(final_mb_encoding || reencoding_during_modedecision)
    {
      if(currMB->qp != finalbestQP)
        continue;
    }
    else   //skip these checking during final QP
    {
      // skip this QP if deltaQP is too large based on local stats
      if(!bslice && (deltaQP != 0) && !(currMB->qp-qp_anchor >= -2 && currMB->qp-qp_anchor <= 1) && currMB->mb_available_left && currMB->mb_available_up)
        continue; 
      if(bslice  &&                   !(currMB->qp-qp_anchor >= -1 && currMB->qp-qp_anchor <= 2) && currMB->mb_available_left && currMB->mb_available_up)
        continue;

      //prevent QP to become too high..otherwise chroma will be affected.
      if(input->successive_Bframe==0)
      {     
        static byte QP_SCALE_CR[52]=
        {
          0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,
          12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,
          28,29,29,30,31,32,32,33,34,34,35,35,36,36,37,37,
          37,38,38,38,39,39,39,39
        };

        chroma_qp = Clip3 ( -img->bitdepth_chroma_qp_scale, 51, currMB->qp + img->chroma_qp_offset[0] );
        chroma_qp = (chroma_qp < 0 ? chroma_qp : QP_SCALE_CR[chroma_qp]) + img->bitdepth_chroma_qp_scale;
        if( (!isfirstQP) && (chroma_qp > master_cr_qp) && (master_cr_cbp!=0) )
          continue;
      }
    }

    store_coding_state (cs_eachQP);
    //==========================================

    min_rdcost = 0;
    tr8x8[deltaQPCnt].cbp8x8 = tr8x8[deltaQPCnt].cbp_blk8x8 = tr8x8[deltaQPCnt].cnt_nonz_8x8 = 0;
    //===========================================================
    // Check 8x8 partition with transform size 8x8 
    //===========================================================
    //=====  LOOP OVER 8x8 SUB-PARTITIONS  (Motion Estimation & Mode Decision) =====
    for (block=0; block<4; block++)
    {
      min_rdcost += submacroblock_mode_decision(&(tr8x8[deltaQPCnt]), enc_mb, currMB, bslice, block, 1, Re_Encode);
    }


    //==========================================
    reset_coding_state (cs_eachQP);

    if(!(final_mb_encoding || reencoding_during_modedecision))
    {
      if(input->successive_Bframe==0)
      {     
        //chroma_qp = currMB->qpc[0];
        if(isfirstQP) // deltaQP =0 has to be test first!
        {
          master_cr_cbp = img->mb_data[img->current_mb_nr].cbp>>4;
          master_cr_qp  = chroma_qp;
        }
      }
    }

    //if (min_rdcost < RDOQ_min_rdcost)
    //{
    //  RDOQ_min_rdcost=min_rdcost;
    //  bestQP=currMB->qp;
    //  best_deltaQP = deltaQPCnt;//deltaQP;
    //  saved_best_mode = P8x8;
    //}

    isfirstQP = 0;

    //if(best_mode==0 && cbp==0)   //no need to check other QPs if the best mode is SKIP for current QP
    //  break;

  }//for (deltaQPCnt=0; deltaQPCnt<deltaQPTabSize; deltaQPCnt++)

  return min_rdcost;
}

double compute_P8x8_RD_cost_RDOQ_Tr4x4  (Macroblock *currMB, RD_PARAMS enc_mb, int Re_Encode)
{
  int block=0;
  short bslice  = (img->type==B_SLICE);

  int deltaQP;
  double min_rdcost=1e30; 
  int qp_left, qp_up;
  //int best_deltaQP=0;
  int deltaQPCnt; 
  int master_cr_qp=0, master_cr_cbp=0, chroma_qp=0;

  static int deltaQPTabB[] = {0, 2, 3};
  int deltaQPTabSizeB = 3;

  static int deltaQPTabP[] = {0, 1, -1, 2, -2};
  int deltaQPTabSizeP = 1;

  int deltaQPTabSize = bslice? deltaQPTabSizeB : deltaQPTabSizeP;
  int qp_anchor;

  qp_left = (currMB->mb_available_left) ? currMB->mb_available_left->qp:img->masterQP;
  qp_up   = (currMB->mb_available_up  ) ? currMB->mb_available_up->qp  :img->masterQP;
  qp_anchor = (qp_left + qp_up + 1) >> 1;

  isfirstQP = 1;
  for (deltaQPCnt=0; deltaQPCnt<deltaQPTabSize; deltaQPCnt++)
  {
    if(bslice)
      deltaQP = deltaQPTabB[deltaQPCnt];
    else
      deltaQP = deltaQPTabP[deltaQPCnt];


    currMB->qp=Clip3(0, 51, (img->masterQP+deltaQP));
    img->qp = currMB->qp;       //this is set coz img->qp is used in 8x8 DCT, KLT, 16x16 DCT, 16x8 DCT, 8x16 DCT
    //for 4x4 DCT it uses currMB->qp......incosistency
    currMB->delta_qp = currMB->qp - currMB->prev_qp;
    set_chroma_qp(currMB);

    if(final_mb_encoding || reencoding_during_modedecision)
    {
      if(currMB->qp != finalbestQP)
        continue;
    }
    else   //skip these checking during final QP
    {
      // skip this QP if deltaQP is too large based on local stats
      if(!bslice && (deltaQP != 0) && !(currMB->qp-qp_anchor >= -2 && currMB->qp-qp_anchor <= 1) && currMB->mb_available_left && currMB->mb_available_up)
        continue; 
      if(bslice  &&                   !(currMB->qp-qp_anchor >= -1 && currMB->qp-qp_anchor <= 2) && currMB->mb_available_left && currMB->mb_available_up)
        continue;

      //prevent QP to become too high..otherwise chroma will be affected.
      if(input->successive_Bframe==0)
      {     
        static byte QP_SCALE_CR[52]=
        {
          0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,
          12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,
          28,29,29,30,31,32,32,33,34,34,35,35,36,36,37,37,
          37,38,38,38,39,39,39,39
        };

        chroma_qp = Clip3 ( -img->bitdepth_chroma_qp_scale, 51, currMB->qp + img->chroma_qp_offset[0] );
        chroma_qp = (chroma_qp < 0 ? chroma_qp : QP_SCALE_CR[chroma_qp]) + img->bitdepth_chroma_qp_scale;
        if( (!isfirstQP) && (chroma_qp > master_cr_qp) && (master_cr_cbp!=0) )
          continue;
      }
    }

    store_coding_state (cs_eachQP);
    //==========================================

    min_rdcost = 0;
    tr4x4[deltaQPCnt].cbp8x8 = tr4x4[deltaQPCnt].cbp_blk8x8 = tr4x4[deltaQPCnt].cnt_nonz_8x8 = 0;
    //=================================================================
    // Check 8x8, 8x4, 4x8 and 4x4 partitions with transform size 4x4
    //=================================================================
    //=====  LOOP OVER 8x8 SUB-PARTITIONS  (Motion Estimation & Mode Decision) =====
    for (block=0; block<4; block++)
    {
      min_rdcost += submacroblock_mode_decision(&(tr4x4[deltaQPCnt]), enc_mb, currMB, bslice, block, 0, Re_Encode);
    }          

    //==========================================
    reset_coding_state (cs_eachQP);

    if(!(final_mb_encoding || reencoding_during_modedecision))
    {
      if(input->successive_Bframe==0)
      {     
        //chroma_qp = currMB->qpc[0];
        if(isfirstQP) // deltaQP =0 has to be test first!
        {
          master_cr_cbp = img->mb_data[img->current_mb_nr].cbp>>4;
          master_cr_qp  = chroma_qp;
        }
      }
    }

    //if (min_rdcost < RDOQ_min_rdcost)
    //{
    //  RDOQ_min_rdcost=min_rdcost;
    //  bestQP=currMB->qp;
    //  best_deltaQP = deltaQPCnt;//deltaQP;
    //  saved_best_mode = P8x8;
    //}

    isfirstQP = 0;

    //if(best_mode==0 && cbp==0)   //no need to check other QPs if the best mode is SKIP for current QP
    //  break;

  }//for (deltaQPCnt=0; deltaQPCnt<deltaQPTabSize; deltaQPCnt++)

  return min_rdcost;
}




