/* 
====================================================================================================================

The copyright in this software is being made available under the License included below. 
This software may be subject to other third party and 	contributor rights, including patent rights, and no such 
rights are granted under this license.

Copyright (c) 2010, QUALCOMM INC.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted only for 
the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
promoting such standards. The following conditions are required to be met:

* Redistributions of source code must retain the above copyright notice, this list of conditions and 
the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
the following disclaimer in the documentation and/or other materials provided with the distribution.
* The name of QUALCOMM INC.  may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

====================================================================================================================
*/


#include <stdlib.h>
#include <math.h>
#include <limits.h>
#include <string.h>

#include "global.h"

#include "image.h"
#include "refbuf.h"
#include "memalloc.h"
#include "mb_access.h"


#include "epzs.h"

#ifdef MV_COMPETITION
#include "mv_competition.h"
extern MV_Competition mv_comp;
#endif

#include <time.h>
#include <sys/timeb.h>

#ifdef SWITCHED_FILTERS
#include "switched_filters.h"
#endif

//! convert from H.263 QP to H.264 quant given by: quant=pow(2,QP/6)
const int QP2QUANT[40]=
{
  1, 1, 1, 1, 2, 2, 2, 2,
  3, 3, 3, 4, 4, 4, 5, 6,
  6, 7, 8, 9,10,11,13,14,
  16,18,20,23,25,29,32,36,
  40,45,51,57,64,72,81,91
};

// Vertical MV Limits (integer/halfpel/quarterpel)
// Currently only Integer Pel restrictions are used, 
// since the way values are specified 
// (i.e. mvlowbound = (levelmvlowbound + 1) and the way
// Subpel me is performed subpel will always be within range.

const int LEVELMVLIMIT[17][6] =
{
  {  -63,  63,  -128,  127,  -256,  255},
  {  -63,  63,  -128,  127,  -256,  255},
  { -127, 127,  -256,  255,  -512,  511},
  { -127, 127,  -256,  255,  -512,  511},
  { -127, 127,  -256,  255,  -512,  511},
  { -127, 127,  -256,  255,  -512,  511},
  { -255, 255,  -512,  511, -1024, 1023},
  { -255, 255,  -512,  511, -1024, 1023},
  { -255, 255,  -512,  511, -1024, 1023},
  { -511, 511, -1024, 1023, -2048, 2047},
  { -511, 511, -1024, 1023, -2048, 2047},
  { -511, 511, -1024, 1023, -2048, 2047},
  { -511, 511, -1024, 1023, -2048, 2047},
  { -511, 511, -1024, 1023, -2048, 2047},
  { -511, 511, -1024, 1023, -2048, 2047},
  { -511, 511, -1024, 1023, -2048, 2047},
  { -511, 511, -1024, 1023, -2048, 2047}

};


// These procedure pointers are used by motion_search() and one_eigthpel()
static pel_t *(*get_line_p1) (pel_t**, int, int, int, int);
static pel_t *(*get_line_p2) (pel_t**, int, int, int, int);

// Statistics, temporary
int     max_mvd;
short*  spiral_search_x;
short*  spiral_search_y;
short*  spiral_hpel_search_x;
short*  spiral_hpel_search_y;

int*    mvbits;
int*    refbits;
unsigned int*  byte_abs;
int**** motion_cost;
int     byte_abs_range;

#ifdef RDO_Q
extern int curr_mbi;
extern int curr_mb32i;
int motion_cost4[1<<(2*MAX_MB_EXT_LEVEL)][8][2][MAX_REFERENCE_PICTURES][4];
int motion_cost8[1<<(2*MAX_MB_EXT_LEVEL)][8][2][MAX_REFERENCE_PICTURES][4];
int motion32_cost4[5][8][2][MAX_REFERENCE_PICTURES][4];
int motion32_cost8[5][8][2][MAX_REFERENCE_PICTURES][4];
int ResetAvailabilityOfNeighbors32(int mb_x, int mb_y, int *reset_block_x, int *reset_block_y, int blockshape_x, int blockshape_y);
extern void CheckAvailabilityOfNeighbors32(int mb_ext_level);
#endif

#ifdef SWITCHED_FILTERS
extern short ******EPZSMotion;  //!< Array for storing Motion Vectors
#endif

void SetMotionVectorPredictor (short  pmv[2],
                               char   **refPic,
                               short  ***tmp_mv,
                               short  ref_frame,
                               int    list,
                               int    block_x,
                               int    block_y,
                               int    blockshape_x,
                               int    blockshape_y,
                               int    mb_ext_level);

extern ColocatedParams *Co_located;


/*!
************************************************************************
* \brief
*    Set motion vector predictor
************************************************************************
*/
void SetMotionVectorPredictor (short  pmv[2],
                               char   **refPic,
                               short  ***tmp_mv,
                               short  ref_frame,
                               int    list,
                               int    block_x,
                               int    block_y,
                               int    blockshape_x,
                               int    blockshape_y,
                               int    mb_ext_level)
{
  int mb_x                 = 4*block_x;
  int mb_y                 = 4*block_y;
  int mb_nr                = img->current_mb_nr;

  int mv_a, mv_b, mv_c, pred_vec=0;
  int mvPredType, rFrameL, rFrameU, rFrameUR;
  int hv;

  PixelPos block_a, block_b, block_c, block_d;

  if(mb_ext_level==0)
  {
    getLuma4x4Neighbour(mb_nr, block_x, block_y,           -1,  0, &block_a);
    getLuma4x4Neighbour(mb_nr, block_x, block_y,            0, -1, &block_b);
    getLuma4x4Neighbour(mb_nr, block_x, block_y, blockshape_x, -1, &block_c);
    getLuma4x4Neighbour(mb_nr, block_x, block_y,           -1, -1, &block_d);
  }
  else
  {
    int reset_mb_nr, reset_block_x=block_x, reset_block_y=block_y;

    reset_mb_nr = ResetAvailabilityOfNeighbors32(mb_x, mb_y, &reset_block_x, &reset_block_y, blockshape_x, blockshape_y);

    getLuma4x4Neighbour(reset_mb_nr, reset_block_x, reset_block_y,           -1,  0, &block_a);
    getLuma4x4Neighbour(reset_mb_nr, reset_block_x, reset_block_y,            0, -1, &block_b);
    getLuma4x4Neighbour(reset_mb_nr, reset_block_x, reset_block_y, blockshape_x, -1, &block_c);
    getLuma4x4Neighbour(reset_mb_nr, reset_block_x, reset_block_y,           -1, -1, &block_d);
  }

  if(mb_ext_level==0)
  {
    if (mb_y > 0)
    {
      if (mb_x < 8)  // first column of 8x8 blocks
      {
        if (mb_y==8)
        {
          if (blockshape_x == 16)      block_c.available  = 0;
        }
        else
        {
          if (mb_x+blockshape_x == 8)  block_c.available = 0;
        }
      }
      else
      {
        if (mb_x+blockshape_x == 16)   block_c.available = 0;
      }
    }
  }
  else
  {
    if (mb_y > 0)
    {
      if (mb_x < (8<<mb_ext_level))  // first column of 16x16 blocks
      {
        if (mb_y==(8<<mb_ext_level))
        {//2nd partition in 32x16 mode
          if (blockshape_x == (16<<mb_ext_level))      block_c.available  = 0;
        }
        else
        {
          error("never reach here!\n", -1);
          if (mb_x+blockshape_x == 16)  block_c.available = 0;
        }
      }
      else
      {
        error("never reach here!\n", -1);
        if (mb_x+blockshape_x == 32)   block_c.available = 0;
      }
    }
  }

  if (!block_c.available)
  {
    block_c=block_d;
  }

  mvPredType = MVPRED_MEDIAN;

  rFrameL    = block_a.available    ? refPic[block_a.pos_y][block_a.pos_x] : -1;
  rFrameU    = block_b.available    ? refPic[block_b.pos_y][block_b.pos_x] : -1;
  rFrameUR   = block_c.available    ? refPic[block_c.pos_y][block_c.pos_x] : -1;

  /* Prediction if only one of the neighbors uses the reference frame
  * we are checking
  */
  if(rFrameL == ref_frame && rFrameU != ref_frame && rFrameUR != ref_frame)       mvPredType = MVPRED_L;
  else if(rFrameL != ref_frame && rFrameU == ref_frame && rFrameUR != ref_frame)  mvPredType = MVPRED_U;
  else if(rFrameL != ref_frame && rFrameU != ref_frame && rFrameUR == ref_frame)  mvPredType = MVPRED_UR;
  // Directional predictions 
  if(blockshape_x == (8<<mb_ext_level) && blockshape_y == (16<<mb_ext_level)) //MB32X32
  {
    if(mb_x == 0)
    {
      if(rFrameL == ref_frame)
        mvPredType = MVPRED_L;
    }
    else
    {
      if( rFrameUR == ref_frame)
        mvPredType = MVPRED_UR;
    }
  }
  else if(blockshape_x == (16<<mb_ext_level) && blockshape_y == (8<<mb_ext_level)) //MB32X32
  {
    if(mb_y == 0)
    {
      if(rFrameU == ref_frame)
        mvPredType = MVPRED_U;
    }
    else
    {
      if(rFrameL == ref_frame)
        mvPredType = MVPRED_L;
    }
  }

  for (hv=0; hv < 2; hv++)
  {
    mv_a = block_a.available  ? tmp_mv[block_a.pos_y][block_a.pos_x][hv] : 0;
    mv_b = block_b.available  ? tmp_mv[block_b.pos_y][block_b.pos_x][hv] : 0;
    mv_c = block_c.available  ? tmp_mv[block_c.pos_y][block_c.pos_x][hv] : 0;


    {
      int64 xpoc = img->framepoc >> 1;
      int64 x  = enc_picture->ref_pic_num[list][ref_frame] >> 2;
      int64 L  = enc_picture->ref_pic_num[list][rFrameL  ] >> 2;
      int64 U  = enc_picture->ref_pic_num[list][rFrameU  ] >> 2;
      int64 UR = enc_picture->ref_pic_num[list][rFrameUR ] >> 2;


      if(xpoc != L)
        mv_a = (int)(mv_a * (xpoc - x) / (xpoc - L));
      if(xpoc != U)
        mv_b = (int)(mv_b * (xpoc - x) / (xpoc - U));
      if(xpoc != UR)
        mv_c = (int)(mv_c * (xpoc - x) / (xpoc - UR));

    }



    switch (mvPredType)
    {
    case MVPRED_MEDIAN:
      if(!(block_b.available || block_c.available))
      {
        pred_vec = mv_a;
      }
      else
      {
        pred_vec = mv_a+mv_b+mv_c-min(mv_a,min(mv_b,mv_c))-max(mv_a,max(mv_b,mv_c));
      }
      break;
    case MVPRED_L:
      pred_vec = mv_a;
      break;
    case MVPRED_U:
      pred_vec = mv_b;
      break;
    case MVPRED_UR:
      pred_vec = mv_c;
      break;
    default:
      break;
    }

    pmv[hv] = pred_vec;

  }
}


/*!
************************************************************************
* \brief
*    Initialize the motion search
************************************************************************
*/
void
Init_Motion_Search_Module ()
{
  int bits, i, imin, imax, k, l;

  int search_range               = input->search_range;
  int max_search_points          = max(9, (2*search_range+1)*(2*search_range+1));
  int max_ref_bits               = 1 + 2 * (int)floor(log(max(16,img->max_num_references+1)) / log(2) + 1e-10);
  int max_ref                    = (1<<((max_ref_bits>>1)+1))-1;
  int number_of_subpel_positions = (img->mv_res?8:4) * (2*search_range+3);
  int max_mv_bits                = 3 + 2 * (int)ceil (log(number_of_subpel_positions+1) / log(2) + 1e-10);
  max_mv_bits                = 31;
  max_mvd                        = (1<<( max_mv_bits >>1)   )-1;
  byte_abs_range                 = (img->max_imgpel_value > img->max_imgpel_value_uv) ? (img->max_imgpel_value + 1) * 64 : (img->max_imgpel_value_uv + 1) * 64;
  byte_abs_range                 = 1<<24;

  //=====   CREATE ARRAYS   =====
  //-----------------------------
  if ((spiral_search_x = (short*)calloc(max_search_points, sizeof(short))) == NULL)
    no_mem_exit("Init_Motion_Search_Module: spiral_search_x");
  if ((spiral_search_y = (short*)calloc(max_search_points, sizeof(short))) == NULL)
    no_mem_exit("Init_Motion_Search_Module: spiral_search_y");
  if ((spiral_hpel_search_x = (short*)calloc(max_search_points, sizeof(short))) == NULL)
    no_mem_exit("Init_Motion_Search_Module: spiral_hpel_search_x");
  if ((spiral_hpel_search_y = (short*)calloc(max_search_points, sizeof(short))) == NULL)
    no_mem_exit("Init_Motion_Search_Module: spiral_hpel_search_y");
  if ((mvbits = (int*)calloc(2*max_mvd+1, sizeof(int))) == NULL)
    no_mem_exit("Init_Motion_Search_Module: mvbits");
  if ((refbits = (int*)calloc(max_ref, sizeof(int))) == NULL)
    no_mem_exit("Init_Motion_Search_Module: refbits");
  if ((byte_abs = (unsigned int*)calloc(byte_abs_range, sizeof(unsigned int))) == NULL)
    no_mem_exit("Init_Motion_Search_Module: byte_abs");

  get_mem4Dint (&motion_cost, 8, 2, img->max_num_references, 4);

  //--- set array offsets ---
  mvbits   += max_mvd;
  byte_abs += byte_abs_range/2;


  //=====   INIT ARRAYS   =====
  //---------------------------
  //--- init array: motion vector bits ---
  mvbits[0] = 1;
  for (bits=3; bits<=max_mv_bits; bits+=2)
  {
    imax = 1    << (bits >> 1);
    imin = imax >> 1;

    for (i = imin; i < imax; i++)   mvbits[-i] = mvbits[i] = bits;
  }
  //--- init array: reference frame bits ---
  refbits[0] = 1;
  for (bits=3; bits<=max_ref_bits; bits+=2)
  {
    imax = (1   << ((bits >> 1) + 1)) - 1;
    imin = imax >> 1;

    for (i = imin; i < imax; i++)   refbits[i] = bits;
  }
  //--- init array: absolute value ---
  byte_abs[0] = 0;
  for (i=1; i<byte_abs_range/2; i++)   byte_abs[i] = byte_abs[-i] = i;
  //--- init array: search pattern ---
  spiral_search_x[0] = spiral_search_y[0] = 0;
  spiral_hpel_search_x[0] = spiral_hpel_search_y[0] = 0;
  for (k=1, l=1; l<=max(1,search_range); l++)
  {
    for (i=-l+1; i< l; i++)
    {
      spiral_search_x[k] =  i;  spiral_search_y[k] = -l;
      spiral_hpel_search_x[k] =  i<<1;  spiral_hpel_search_y[k++] = -l<<1;
      spiral_search_x[k] =  i;  spiral_search_y[k] =  l;      
      spiral_hpel_search_x[k] =  i<<1;  spiral_hpel_search_y[k++] =  l<<1;
    }
    for (i=-l;   i<=l; i++)
    {
      spiral_search_x[k] = -l;  spiral_search_y[k] =  i;
      spiral_hpel_search_x[k] = -l<<1;  spiral_hpel_search_y[k++] = i<<1;
      spiral_search_x[k] =  l;  spiral_search_y[k] =  i;      
      spiral_hpel_search_x[k] =  l<<1;  spiral_hpel_search_y[k++] = i<<1;
    }
  }


}


/*!
************************************************************************
* \brief
*    Free memory used by motion search
************************************************************************
*/
void
Clear_Motion_Search_Module ()
{
  //--- correct array offset ---
  mvbits   -= max_mvd;
  byte_abs -= byte_abs_range/2;

  //--- delete arrays ---
  free (spiral_search_x);
  free (spiral_search_y);
  free (spiral_hpel_search_x);
  free (spiral_hpel_search_y);
  free (mvbits);
  free (refbits);
  free (byte_abs);
  free_mem4Dint (motion_cost, 8, 2);

}


int
SATD (int* diff, int use_hadamard)
{
  int k, satd = 0, m[16], d[16];

  if (use_hadamard)
  {
    /*===== hadamard transform =====*/
    m[ 0] = diff[ 0] + diff[12];
    m[ 1] = diff[ 1] + diff[13];
    m[ 2] = diff[ 2] + diff[14];
    m[ 3] = diff[ 3] + diff[15];
    m[ 4] = diff[ 4] + diff[ 8];
    m[ 5] = diff[ 5] + diff[ 9];
    m[ 6] = diff[ 6] + diff[10];
    m[ 7] = diff[ 7] + diff[11];
    m[ 8] = diff[ 4] - diff[ 8];
    m[ 9] = diff[ 5] - diff[ 9];
    m[10] = diff[ 6] - diff[10];
    m[11] = diff[ 7] - diff[11];
    m[12] = diff[ 0] - diff[12];
    m[13] = diff[ 1] - diff[13];
    m[14] = diff[ 2] - diff[14];
    m[15] = diff[ 3] - diff[15];

    d[ 0] = m[ 0] + m[ 4];
    d[ 1] = m[ 1] + m[ 5];
    d[ 2] = m[ 2] + m[ 6];
    d[ 3] = m[ 3] + m[ 7];
    d[ 4] = m[ 8] + m[12];
    d[ 5] = m[ 9] + m[13];
    d[ 6] = m[10] + m[14];
    d[ 7] = m[11] + m[15];
    d[ 8] = m[ 0] - m[ 4];
    d[ 9] = m[ 1] - m[ 5];
    d[10] = m[ 2] - m[ 6];
    d[11] = m[ 3] - m[ 7];
    d[12] = m[12] - m[ 8];
    d[13] = m[13] - m[ 9];
    d[14] = m[14] - m[10];
    d[15] = m[15] - m[11];

    m[ 0] = d[ 0] + d[ 3];
    m[ 1] = d[ 1] + d[ 2];
    m[ 2] = d[ 1] - d[ 2];
    m[ 3] = d[ 0] - d[ 3];
    m[ 4] = d[ 4] + d[ 7];
    m[ 5] = d[ 5] + d[ 6];
    m[ 6] = d[ 5] - d[ 6];
    m[ 7] = d[ 4] - d[ 7];
    m[ 8] = d[ 8] + d[11];
    m[ 9] = d[ 9] + d[10];
    m[10] = d[ 9] - d[10];
    m[11] = d[ 8] - d[11];
    m[12] = d[12] + d[15];
    m[13] = d[13] + d[14];
    m[14] = d[13] - d[14];
    m[15] = d[12] - d[15];

    d[ 0] = m[ 0] + m[ 1];
    d[ 1] = m[ 0] - m[ 1];
    d[ 2] = m[ 2] + m[ 3];
    d[ 3] = m[ 3] - m[ 2];
    d[ 4] = m[ 4] + m[ 5];
    d[ 5] = m[ 4] - m[ 5];
    d[ 6] = m[ 6] + m[ 7];
    d[ 7] = m[ 7] - m[ 6];
    d[ 8] = m[ 8] + m[ 9];
    d[ 9] = m[ 8] - m[ 9];
    d[10] = m[10] + m[11];
    d[11] = m[11] - m[10];
    d[12] = m[12] + m[13];
    d[13] = m[12] - m[13];
    d[14] = m[14] + m[15];
    d[15] = m[15] - m[14];

    //===== sum up =====
    // Table lookup is faster than abs macro
    for (k=0; k<16; ++k)
    {
      //satd += absm(d[k]);
      satd += byte_abs [d [k]];
    }
    satd = ((satd+1)>>1);
  }
  else
  {
    /*===== sum up =====*/
    for (k = 0; k < 16; k++)
    {
      satd += byte_abs [diff [k]];
    }
  }

  return satd;
}

/*!
***********************************************************************
* \brief
*    Calculate SA(T)D for 8x8
***********************************************************************
*/
int
SATD8X8 (int* diff, int use_hadamard)
{
  int i, j, jj, sad=0;
  int m1[8][8], m2[8][8], m3[8][8];

  if(use_hadamard)
  {
    //horizontal
    for (j=0; j < 8; j++)
    {
      jj = j << 3;
      m2[j][0] = diff[jj  ] + diff[jj+4];
      m2[j][1] = diff[jj+1] + diff[jj+5];
      m2[j][2] = diff[jj+2] + diff[jj+6];
      m2[j][3] = diff[jj+3] + diff[jj+7];
      m2[j][4] = diff[jj  ] - diff[jj+4];
      m2[j][5] = diff[jj+1] - diff[jj+5];
      m2[j][6] = diff[jj+2] - diff[jj+6];
      m2[j][7] = diff[jj+3] - diff[jj+7];

      m1[j][0] = m2[j][0] + m2[j][2];
      m1[j][1] = m2[j][1] + m2[j][3];
      m1[j][2] = m2[j][0] - m2[j][2];
      m1[j][3] = m2[j][1] - m2[j][3];
      m1[j][4] = m2[j][4] + m2[j][6];
      m1[j][5] = m2[j][5] + m2[j][7];
      m1[j][6] = m2[j][4] - m2[j][6];
      m1[j][7] = m2[j][5] - m2[j][7];

      m2[j][0] = m1[j][0] + m1[j][1];
      m2[j][1] = m1[j][0] - m1[j][1];
      m2[j][2] = m1[j][2] + m1[j][3];
      m2[j][3] = m1[j][2] - m1[j][3];
      m2[j][4] = m1[j][4] + m1[j][5];
      m2[j][5] = m1[j][4] - m1[j][5];
      m2[j][6] = m1[j][6] + m1[j][7];
      m2[j][7] = m1[j][6] - m1[j][7];
    }

    //vertical
    for (i=0; i < 8; i++)
    {
      m3[0][i] = m2[0][i] + m2[4][i];
      m3[1][i] = m2[1][i] + m2[5][i];
      m3[2][i] = m2[2][i] + m2[6][i];
      m3[3][i] = m2[3][i] + m2[7][i];
      m3[4][i] = m2[0][i] - m2[4][i];
      m3[5][i] = m2[1][i] - m2[5][i];
      m3[6][i] = m2[2][i] - m2[6][i];
      m3[7][i] = m2[3][i] - m2[7][i];

      m1[0][i] = m3[0][i] + m3[2][i];
      m1[1][i] = m3[1][i] + m3[3][i];
      m1[2][i] = m3[0][i] - m3[2][i];
      m1[3][i] = m3[1][i] - m3[3][i];
      m1[4][i] = m3[4][i] + m3[6][i];
      m1[5][i] = m3[5][i] + m3[7][i];
      m1[6][i] = m3[4][i] - m3[6][i];
      m1[7][i] = m3[5][i] - m3[7][i];

      m2[0][i] = m1[0][i] + m1[1][i];
      m2[1][i] = m1[0][i] - m1[1][i];
      m2[2][i] = m1[2][i] + m1[3][i];
      m2[3][i] = m1[2][i] - m1[3][i];
      m2[4][i] = m1[4][i] + m1[5][i];
      m2[5][i] = m1[4][i] - m1[5][i];
      m2[6][i] = m1[6][i] + m1[7][i];
      m2[7][i] = m1[6][i] - m1[7][i];
    }
    for (j=0; j < 8; j++)
      for (i=0; i < 8; i++)
        sad += byte_abs[m2[j][i]];

    sad=((sad+2)>>2);
  }
  else
  {
    for(i=0; i<64; i++)
      sad += byte_abs [diff [i]];
  }

  return sad;
}

/*!
***********************************************************************
* \brief
*    Calculate SA(T)D for 8x8
***********************************************************************
*/
/*!
***********************************************************************
* \brief
*    Calculate SA(T)D for 8x8
***********************************************************************
*/
int
find_SATD (int c_diff[MB_PIXELS], int blocktype, int mb_ext_level)
{
  int i, sad=0;

  switch(blocktype)
  {
    //32x32
  case 1: 
    for(i=0; i<((2<<mb_ext_level)*(2<<mb_ext_level)); i++)
      sad += SATD8X8 (&c_diff[i*64],       input->hadamard);
    break;
    //32x16 16x32
  case 2:
  case 3: 
    for(i=0; i<((2<<mb_ext_level)*(1<<mb_ext_level)); i++)
      sad += SATD8X8 (&c_diff[i*64],       input->hadamard);
    break;
    //8x8
  case 4: 
    if(mb_ext_level==0)
      sad  = SATD8X8 (&c_diff[0], input->hadamard);
    else
      sad=-1;
    break;
    //8x4 4x8
  default:
    sad=-1;
    break;
  }

  return sad;
}


/*!
***********************************************************************
* \brief
*    Bipred Sub pixel block motion search
***********************************************************************
*/
int                                               //  ==> minimum motion cost after search
SubPelBlockSearchBiPred (pel_t**   orig_pic,      // <--  original pixel values for the AxB block
                         short     ref,           // <--  reference frame (0... or -1 (backward))
                         int       list,          // <--  reference picture list 
                         int       pic_pix_x,     // <--  absolute x-coordinate of regarded AxB block
                         int       pic_pix_y,     // <--  absolute y-coordinate of regarded AxB block
                         int       blocktype,     // <--  block type (1-16x16 ... 7-4x4)
                         short     pred_mv_x,     // <--  motion vector predictor (x) in sub-pel units
                         short     pred_mv_y,     // <--  motion vector predictor (y) in sub-pel units
                         short*    mv_x,          // <--> in: search center (x) / out: motion vector (x) - in pel units
                         short*    mv_y,          // <--> in: search center (y) / out: motion vector (y) - in pel units
                         short*    s_mv_x,          // <--> in: search center (x) / out: motion vector (x) - in pel units
                         short*    s_mv_y,          // <--> in: search center (y) / out: motion vector (y) - in pel units
                         int       search_pos2,   // <--  search positions for    half-pel search  (default: 9)
                         int       search_pos4,   // <--  search positions for quarter-pel search  (default: 9)
                         int       search_pos8,   // <--  search positions for eighth-pel search   (default: 9)
                         int       min_mcost,     // <--  minimum motion cost (cost for center or huge value)
                         int       lambda_factor,   // <--  lagrangian parameter for determining motion cost
                         int       mb_ext_level)
{
  int   j, i, k;
  int   c_diff[MB_PIXELS<<(2*MAX_MB_EXT_LEVEL)];
  int   diff[16], *d;  
  int   pos, best_pos, mcost, abort_search;
  int   y0, y1, y2, y3;
  int   x0;
  int   ry0, ry4, ry8, ry12, rx0;
  int   sy0, sy4, sy8, sy12, sx0;

  int   cand_mv_x, cand_mv_y;

#ifdef SWITCHED_FILTERS
  int   best_pos_sp16 = 0;
  int   min_mcost_sp16 = min_mcost;
  short mv_x_sp16 = 0;
  short mv_y_sp16 = 0;
#endif

  int   blocksize_x     = input->blc_size[blocktype][0]<<mb_ext_level;
  int   blocksize_y     = input->blc_size[blocktype][1]<<mb_ext_level;
  int   pic4_pix_x      = ((pic_pix_x + IMG_PAD_SIZE)<< 2);
  int   pic4_pix_y      = ((pic_pix_y + IMG_PAD_SIZE)<< 2);
  int   min_pos2        = (input->hadamard ? 0 : 1);
  int   max_pos2        = (input->hadamard ? max(1,search_pos2) : search_pos2);
  int   list_offset   = img->mb_data[img->current_mb_nr].list_offset; 
  short offsetSpic = 0;
  short offsetRpic = 0;

#ifdef USE_HP_FILTER//BROUND
  short offsetBi;
#else
  short offsetBi=(offsetRpic + offsetSpic + 1)>>1;
#endif
  int mv_shift= img->mv_res?1:0;
  int   test8x8transform = input->Transform8x8Mode && blocktype <= 4 && input->hadamard;
  int   cmv_x, cmv_y;
  int   smv_x = *s_mv_x + pic4_pix_x;
  int   smv_y = *s_mv_y + pic4_pix_y;

  StorablePicture *ref_picture = listX[list+list_offset][ref];
  pel_t **ref1_pic = ref_picture->imgY_ups;      
  pel_t **ref2_pic = listX[list==0? 1 +list_offset: list_offset][0  ]->imgY_ups;  
  pel_t *ref_line_p1,*ref_line_p2;
  pel_t *orig_line;
  int img_width  = ((ref_picture->size_x + 2*IMG_PAD_SIZE - 1)<<2);
  int img_height = ((ref_picture->size_y + 2*IMG_PAD_SIZE - 1)<<2);
  int max_pos_x4 = ((ref_picture->size_x - blocksize_x + 2*IMG_PAD_SIZE)<<2);
  int max_pos_y4 = ((ref_picture->size_y - blocksize_y + 2*IMG_PAD_SIZE)<<2);
#ifdef  USE_HP_FILTER//BROUND
  int   round;
  img->bipred_rounding_control = (img->nal_reference_idc!=0);
  round = 1 - img->bipred_rounding_control;
  if(input->UseHPFilter != 0)
  {
    offsetBi=offsetRpic + offsetSpic;
  }
  else
  {
    offsetBi=(offsetRpic + offsetSpic + 1)>>1;
  }
#endif

  /*********************************
  *****                       *****
  *****  HALF-PEL REFINEMENT  *****
  *****                       *****
  *********************************/

  //===== set function for getting pixel values =====
  if ((pic4_pix_x + *mv_x > 2) && (pic4_pix_x + *mv_x < max_pos_x4 - 1) &&
    (pic4_pix_y + *mv_y > 2) && (pic4_pix_y + *mv_y < max_pos_y4 - 1))
  {
    get_line_p2 = FastLine4X;
  } 
  else
  {
    get_line_p2 = UMVLine4X2;
  }

  if ((pic4_pix_x + *s_mv_x > 2) && (pic4_pix_x + *s_mv_x < max_pos_x4 - 1) &&
    (pic4_pix_y + *s_mv_y > 2) && (pic4_pix_y + *s_mv_y < max_pos_y4 - 1))
  {
    get_line_p1 = FastLine4X;
  }
  else
  {
    get_line_p1 = UMVLine4X;    
  }


  //===== loop over search positions =====
#ifdef SWITCHED_FILTERS
  for (best_pos = 0, pos = min_pos2; 
    pos < max_pos2 + (((img->filterParam == SIFO_FIRST_PASS_FPO) && (ref == 0)) ? img->noOffsets[list]: 0); 
    pos++)
  {
    if(pos < max_pos2)
    {
      cand_mv_x = *mv_x + (spiral_hpel_search_x[pos]);    // quarter-pel units
      cand_mv_y = *mv_y + (spiral_hpel_search_y[pos]);    // quarter-pel units
    }
    else
    {
      cand_mv_x = *mv_x + img->search_point_qp_x_offset[pos-max_pos2];    // quarter-pel units
      cand_mv_y = *mv_y + img->search_point_qp_y_offset[pos-max_pos2];    // quarter-pel units
    }
#else
  for (best_pos = 0, pos = min_pos2; pos < max_pos2; pos++)
  {
    cand_mv_x = *mv_x + (spiral_hpel_search_x[pos]);    // quarter-pel units
    cand_mv_y = *mv_y + (spiral_hpel_search_y[pos]);    // quarter-pel units
#endif

    //----- set motion vector cost -----

#ifdef ADAPTIVE_MV_RES
    if (input->mv_competition > 0)
      mcost = compute_mv_cost_for_each_predictor_1 (lambda_factor, mv_shift,img->mv_res, cand_mv_x, cand_mv_y, pred_mv_x, pred_mv_y, ref, list ^ 1, blocktype, pic_pix_x-img->opix_x, pic_pix_y-img->opix_y,0);
    else
      mcost = MV_COST_1(lambda_factor, mv_shift,img->mv_res, cand_mv_x, cand_mv_y, pred_mv_x, pred_mv_y,0);
#else

    if (input->mv_competition > 0)
      mcost = compute_mv_cost_for_each_predictor (lambda_factor, mv_shift, cand_mv_x, cand_mv_y, pred_mv_x, pred_mv_y, ref, list ^ 1, blocktype, pic_pix_x-img->opix_x, pic_pix_y-img->opix_y);
    else
      mcost = MV_COST (lambda_factor, mv_shift, cand_mv_x, cand_mv_y, pred_mv_x, pred_mv_y);
#endif



    if (mcost >= min_mcost) continue;

    cmv_x = cand_mv_x + pic4_pix_x;
    cmv_y = cand_mv_y + pic4_pix_y;

    //----- add up SATD -----
    for (y0=0, abort_search=0; y0<blocksize_y && !abort_search; y0+=BLOCK_SIZE)
    {
      ry0 = (y0<<2) + cmv_y;
      ry4  = ry0 + 4;
      ry8  = ry4 + 4;
      ry12 = ry8 + 4;      
      sy0 = (y0<<2) + smv_y;
      sy4  = sy0 + 4;
      sy8  = sy4 + 4;
      sy12 = sy8 + 4;
      y1 = y0 + 1;
      y2 = y1 + 1;
      y3 = y2 + 1;

      {   
#ifdef SWITCHED_FILTERS
        img->bipred_rounding_control = (img->nal_reference_idc != 0);
        round = img->currMEOffset[0] + img->currMEOffset[1] + (1 - img->bipred_rounding_control);
#endif
        for (x0=0; x0<blocksize_x; x0+=BLOCK_SIZE)
        {         
          rx0 = (x0<<2) + cmv_x;            
          sx0 = (x0<<2) + smv_x;
          d   = diff;

#ifdef  USE_HP_FILTER//BROUND
          if(input->UseHPFilter != 0)
          {
            orig_line = &orig_pic [y0][x0];    
            ref_line_p1 = get_line_p1 (ref1_pic, sy0, sx0,  img_height, img_width);
            ref_line_p2 = get_line_p2 (ref2_pic, ry0, rx0,  img_height, img_width);
            *d++ = *orig_line++  -  ((*(ref_line_p1     ) + *(ref_line_p2     ) + round) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + round) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + round) >> 1);
            *d++ = *orig_line    -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + round) >> 1);

            orig_line = &orig_pic [y1][x0];    
            ref_line_p1 = get_line_p1 (ref1_pic, sy4, sx0,  img_height, img_width);
            ref_line_p2 = get_line_p2 (ref2_pic, ry4, rx0,  img_height, img_width);
            *d++ = *orig_line++  -  ((*(ref_line_p1     ) + *(ref_line_p2     ) + round) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + round) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + round) >> 1);
            *d++ = *orig_line    -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + round) >> 1);

            orig_line = &orig_pic [y2][x0];    
            ref_line_p1 = get_line_p1 (ref1_pic, sy8, sx0,  img_height, img_width);
            ref_line_p2 = get_line_p2 (ref2_pic, ry8, rx0,  img_height, img_width);
            *d++ = *orig_line++  -  ((*(ref_line_p1     ) + *(ref_line_p2     ) + round) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + round) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + round) >> 1);
            *d++ = *orig_line    -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + round) >> 1);

            orig_line = &orig_pic [y3][x0];    
            ref_line_p1 = get_line_p1 (ref1_pic, sy12, sx0,  img_height, img_width);
            ref_line_p2 = get_line_p2 (ref2_pic, ry12, rx0,  img_height, img_width);
            *d++ = *orig_line++  -  ((*(ref_line_p1     ) + *(ref_line_p2     ) + round) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + round) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + round) >> 1);
            *d   = *orig_line    -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + round) >> 1);
          }
          else
          {
#endif
            orig_line = &orig_pic [y0][x0];    
            ref_line_p1 = get_line_p1 (ref1_pic, sy0, sx0,  img_height, img_width);
            ref_line_p2 = get_line_p2 (ref2_pic, ry0, rx0,  img_height, img_width);
            *d++ = *orig_line++  -  ((*(ref_line_p1     ) + *(ref_line_p2     ) + 1) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
            *d++ = *orig_line    -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);

            orig_line = &orig_pic [y1][x0];    
            ref_line_p1 = get_line_p1 (ref1_pic, sy4, sx0,  img_height, img_width);
            ref_line_p2 = get_line_p2 (ref2_pic, ry4, rx0,  img_height, img_width);
            *d++ = *orig_line++  -  ((*(ref_line_p1     ) + *(ref_line_p2     ) + 1) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
            *d++ = *orig_line    -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);

            orig_line = &orig_pic [y2][x0];    
            ref_line_p1 = get_line_p1 (ref1_pic, sy8, sx0,  img_height, img_width);
            ref_line_p2 = get_line_p2 (ref2_pic, ry8, rx0,  img_height, img_width);
            *d++ = *orig_line++  -  ((*(ref_line_p1     ) + *(ref_line_p2     ) + 1) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
            *d++ = *orig_line    -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);

            orig_line = &orig_pic [y3][x0];    
            ref_line_p1 = get_line_p1 (ref1_pic, sy12, sx0,  img_height, img_width);
            ref_line_p2 = get_line_p2 (ref2_pic, ry12, rx0,  img_height, img_width);
            *d++ = *orig_line++  -  ((*(ref_line_p1     ) + *(ref_line_p2     ) + 1) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
            *d   = *orig_line    -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
#ifdef  USE_HP_FILTER//BROUND
          }
#endif

          if (!test8x8transform)
          {
            if ((mcost += SATD (diff, input->hadamard)) > min_mcost)
            {
              abort_search = 1;
              break;
            }
          }
          else
          {
            int x8=x0/8, y8=y0/8;
            i = (x0&0x7) +  x8*64 + (y8 * blocksize_x/8)*64 ;
            for(k=0, j=y0; j<4 + y0; j++, k+=4)
              memcpy(&(c_diff[i + ((j&0x7)<<3)]), &diff[k], 4*sizeof(int));
          }
        }
      }
    }  

    if(test8x8transform)
      mcost += find_SATD (c_diff, blocktype, mb_ext_level);

#ifdef SWITCHED_FILTERS
    if(pos < max_pos2)
    {
      // Update cost for regular search
      if (mcost < min_mcost)
      {
        min_mcost = mcost;
        best_pos  = pos;
      }      
    }
    else
    {
      // Update results of offset search
      if(mcost < min_mcost_sp16)
      {
        min_mcost_sp16 = mcost;
        best_pos_sp16  = pos;
        mv_x_sp16 = *mv_x + img->search_point_qp_x_offset[pos - max_pos2];
        mv_y_sp16 = *mv_y + img->search_point_qp_y_offset[pos - max_pos2];
      }  
    }
#else
    if (mcost < min_mcost)
    {
      min_mcost = mcost;
      best_pos  = pos;
    }      
#endif
  }

  if (best_pos)
  {
    *mv_x += (spiral_hpel_search_x [best_pos]);
    *mv_y += (spiral_hpel_search_y [best_pos]);
  }

  test8x8transform = input->Transform8x8Mode && blocktype <= 4 && input->hadamard;

  /************************************
  *****                          *****
  *****  QUARTER-PEL REFINEMENT  *****
  *****                          *****
  ************************************/
  //===== set function for getting pixel values =====
  if ((pic4_pix_x + *mv_x > 0) && (pic4_pix_x + *mv_x < max_pos_x4) &&
    (pic4_pix_y + *mv_y > 0) && (pic4_pix_y + *mv_y < max_pos_y4))
  {
    get_line_p2 = FastLine4X;
  }
  else
  {
    get_line_p2 = UMVLine4X2;    
  }

  if ((pic4_pix_x + *s_mv_x > 0) && (pic4_pix_x + *s_mv_x < max_pos_x4) &&
    (pic4_pix_y + *s_mv_y > 0) && (pic4_pix_y + *s_mv_y < max_pos_y4))
  {
    get_line_p1 = FastLine4X;
  }
  else
  {
    get_line_p1 = UMVLine4X;    
  }


  //===== loop over search positions =====
  for (best_pos = 0, pos = 1; pos < search_pos4; pos++)
  {
    cand_mv_x = *mv_x + spiral_search_x[pos];    // quarter-pel units
    cand_mv_y = *mv_y + spiral_search_y[pos];    // quarter-pel units

    //----- set motion vector cost -----
#ifdef ADAPTIVE_MV_RES
    if (input->mv_competition > 0)
      mcost = compute_mv_cost_for_each_predictor_1 (lambda_factor, mv_shift,img->mv_res, cand_mv_x, cand_mv_y, pred_mv_x, pred_mv_y, ref, list ^ 1, blocktype, pic_pix_x-img->opix_x, pic_pix_y-img->opix_y,0);
    else
      mcost = MV_COST_1(lambda_factor, mv_shift,img->mv_res, cand_mv_x, cand_mv_y, pred_mv_x, pred_mv_y,0);
#else

    if (input->mv_competition > 0)
      mcost = compute_mv_cost_for_each_predictor (lambda_factor, mv_shift, cand_mv_x, cand_mv_y, pred_mv_x, pred_mv_y, ref, list ^ 1, blocktype, pic_pix_x-img->opix_x, pic_pix_y-img->opix_y);
    else
      mcost = MV_COST (lambda_factor, mv_shift, cand_mv_x, cand_mv_y, pred_mv_x, pred_mv_y);
#endif
    if (mcost >= min_mcost) continue;
    cmv_x = cand_mv_x + pic4_pix_x;
    cmv_y = cand_mv_y + pic4_pix_y;

    //----- add up SATD -----
    for (y0=0, abort_search=0; y0<blocksize_y && !abort_search; y0+=BLOCK_SIZE)
    {
      ry0 = (y0<<2) + cmv_y;
      ry4  = ry0 + 4;
      ry8  = ry4 + 4;
      ry12 = ry8 + 4;      
      sy0 = (y0<<2) + smv_y;
      sy4  = sy0 + 4;
      sy8  = sy4 + 4;
      sy12 = sy8 + 4;
      y1 = y0 + 1;
      y2 = y1 + 1;
      y3 = y2 + 1;

      {      
#ifdef SWITCHED_FILTERS
        img->bipred_rounding_control = (img->nal_reference_idc != 0);
        round = img->currMEOffset[0] + img->currMEOffset[1] + (1 - img->bipred_rounding_control);
#endif
        for (x0=0; x0<blocksize_x; x0+=BLOCK_SIZE)
        {
          rx0 = (x0<<2) + cmv_x;            
          sx0 = (x0<<2) + smv_x;
          d   = diff;

#ifdef  USE_HP_FILTER//BROUND
          if(input->UseHPFilter != 0)
          {
            orig_line = &orig_pic [y0][x0];    
            ref_line_p1 = get_line_p1 (ref1_pic, sy0, sx0,  img_height, img_width);
            ref_line_p2 = get_line_p2 (ref2_pic, ry0, rx0,  img_height, img_width);
            *d++ = *orig_line++  -  ((*(ref_line_p1     ) + *(ref_line_p2     ) + round) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + round) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + round) >> 1);
            *d++ = *orig_line    -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + round) >> 1);

            orig_line = &orig_pic [y1][x0];    
            ref_line_p1 = get_line_p1 (ref1_pic, sy4, sx0,  img_height, img_width);
            ref_line_p2 = get_line_p2 (ref2_pic, ry4, rx0,  img_height, img_width);
            *d++ = *orig_line++  -  ((*(ref_line_p1     ) + *(ref_line_p2     ) + round) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + round) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + round) >> 1);
            *d++ = *orig_line    -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + round) >> 1);

            orig_line = &orig_pic [y2][x0];    
            ref_line_p1 = get_line_p1 (ref1_pic, sy8, sx0,  img_height, img_width);
            ref_line_p2 = get_line_p2 (ref2_pic, ry8, rx0,  img_height, img_width);
            *d++ = *orig_line++  -  ((*(ref_line_p1     ) + *(ref_line_p2     ) + round) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + round) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + round) >> 1);
            *d++ = *orig_line    -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + round) >> 1);

            orig_line = &orig_pic [y3][x0];    
            ref_line_p1 = get_line_p1 (ref1_pic, sy12, sx0,  img_height, img_width);
            ref_line_p2 = get_line_p2 (ref2_pic, ry12, rx0,  img_height, img_width);
            *d++ = *orig_line++  -  ((*(ref_line_p1     ) + *(ref_line_p2     ) + round) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + round) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + round) >> 1);
            *d   = *orig_line    -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + round) >> 1);
          }
          else
          {
#endif
            orig_line = &orig_pic [y0][x0];    
            ref_line_p1 = get_line_p1 (ref1_pic, sy0, sx0,  img_height, img_width);
            ref_line_p2 = get_line_p2 (ref2_pic, ry0, rx0,  img_height, img_width);
            *d++ = *orig_line++  -  ((*(ref_line_p1     ) + *(ref_line_p2     ) + 1) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
            *d++ = *orig_line    -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);

            orig_line = &orig_pic [y1][x0];    
            ref_line_p1 = get_line_p1 (ref1_pic, sy4, sx0,  img_height, img_width);
            ref_line_p2 = get_line_p2 (ref2_pic, ry4, rx0,  img_height, img_width);
            *d++ = *orig_line++  -  ((*(ref_line_p1     ) + *(ref_line_p2     ) + 1) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
            *d++ = *orig_line    -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);

            orig_line = &orig_pic [y2][x0];    
            ref_line_p1 = get_line_p1 (ref1_pic, sy8, sx0,  img_height, img_width);
            ref_line_p2 = get_line_p2 (ref2_pic, ry8, rx0,  img_height, img_width);
            *d++ = *orig_line++  -  ((*(ref_line_p1     ) + *(ref_line_p2     ) + 1) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
            *d++ = *orig_line    -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);

            orig_line = &orig_pic [y3][x0];    
            ref_line_p1 = get_line_p1 (ref1_pic, sy12, sx0,  img_height, img_width);
            ref_line_p2 = get_line_p2 (ref2_pic, ry12, rx0,  img_height, img_width);
            *d++ = *orig_line++  -  ((*(ref_line_p1     ) + *(ref_line_p2     ) + 1) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
            *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
            *d   = *orig_line    -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
#ifdef  USE_HP_FILTER//BROUND
          }
#endif

          if (!test8x8transform)
          {
            if ((mcost += SATD (diff, input->hadamard)) > min_mcost)
            {
              abort_search = 1;
              break;
            }
          }
          else
          {
            int x8=x0/8, y8=y0/8;
            i = (x0&0x7) +  x8*64 + (y8 * blocksize_x/8)*64 ;
            for(k=0, j=y0; j<4 + y0; j++, k+=4)
              memcpy(&(c_diff[i + ((j&0x7)<<3)]), &diff[k], 4*sizeof(int));
          }
        }
      }
    }
    if(test8x8transform)
      mcost += find_SATD (c_diff, blocktype, mb_ext_level);

    if (mcost < min_mcost)
    {
      min_mcost = mcost;
      best_pos  = pos;
    }

  }

#ifdef SWITCHED_FILTERS
  if(best_pos || best_pos_sp16)
  {
    if(best_pos_sp16 && (min_mcost_sp16 < min_mcost))
    {
      // Replace 1/4 pixel motion vectors with the result of the offset search
      min_mcost = min_mcost_sp16;
      *mv_x = mv_x_sp16;
      *mv_y = mv_y_sp16;
    }
    else
    {
      *mv_x += spiral_search_x [best_pos];
      *mv_y += spiral_search_y [best_pos];
    }
  }
#else
  if (best_pos)
  {
    *mv_x += spiral_search_x [best_pos];
    *mv_y += spiral_search_y [best_pos];
  }
#endif

  /************************************
  *****                          *****
  *****  EIGHTH-PEL REFINEMENT  *****
  *****                          *****
  ************************************/
  if (img->mv_res)
  {
    //===== convert search center to eighth-pel units (double resolution) =====
    *mv_x <<= 1;
    *mv_y <<= 1;
    smv_x <<= 1;
    smv_y <<= 1;
    *s_mv_x <<= 1;
    *s_mv_y <<= 1;
    pic4_pix_x <<= 1;
    pic4_pix_y <<= 1;

    //===== set function for getting pixel values =====
    if ((pic4_pix_x + *mv_x > 1) && (pic4_pix_x + *mv_x < 2*max_pos_x4 - 2) &&
      (pic4_pix_y + *mv_y > 1) && (pic4_pix_y + *mv_y < 2*max_pos_y4 - 2))
      get_line_p2 = FastLine4X8;
    else
      get_line_p2 = UMVLine4X8;

    if ((pic4_pix_x + *s_mv_x > 1) && (pic4_pix_x + *s_mv_x < 2*max_pos_x4 - 2) &&
      (pic4_pix_y + *s_mv_y > 1) && (pic4_pix_y + *s_mv_y < 2*max_pos_y4 - 2))
      get_line_p1 = FastLine4X8;
    else
      get_line_p1 = UMVLine4X8;

    //===== loop over search positions =====
    for (best_pos = 0, pos = 1; pos < search_pos8; pos++)
    {
      cand_mv_x = *mv_x + spiral_search_x[pos];    // quarter-pel units
      cand_mv_y = *mv_y + spiral_search_y[pos];    // quarter-pel units

      //----- set motion vector cost -----
#ifdef ADAPTIVE_MV_RES
      if (input->mv_competition > 0)
        mcost = compute_mv_cost_for_each_predictor_1 (lambda_factor, 0,0, cand_mv_x, cand_mv_y, pred_mv_x, pred_mv_y, ref, list ^ 1, blocktype, pic_pix_x-img->opix_x, pic_pix_y-img->opix_y,1);
      else
        mcost = MV_COST_1(lambda_factor, 0,0, cand_mv_x, cand_mv_y, pred_mv_x, pred_mv_y,1);
#else

#ifdef MV_COMPETITION
      if (input->mv_competition > 0)
        mcost = compute_mv_cost_for_each_predictor (lambda_factor, 0, cand_mv_x, cand_mv_y, pred_mv_x, pred_mv_y, ref, list ^ 1, blocktype, pic_pix_x-img->opix_x, pic_pix_y-img->opix_y);
      else
#endif
        mcost = MV_COST (lambda_factor, 0, cand_mv_x, cand_mv_y, pred_mv_x, pred_mv_y);
#endif 
      if (mcost >= min_mcost) continue;
      cmv_x = cand_mv_x + pic4_pix_x;
      cmv_y = cand_mv_y + pic4_pix_y;

      //----- add up SATD -----
      for (y0=0, abort_search=0; y0<blocksize_y && !abort_search; y0+=BLOCK_SIZE)
      {
        ry0 = (y0<<2) + cmv_y;
        ry4  = ry0 + 8;
        ry8  = ry4 + 8;
        ry12 = ry8 + 8;
        sy0 = (y0<<2) + smv_y;
        sy4  = sy0 + 8;
        sy8  = sy4 + 8;
        sy12 = sy8 + 8;
        y1 = y0 + 1;
        y2 = y1 + 1;
        y3 = y2 + 1;

        for (x0=0; x0<blocksize_x; x0+=BLOCK_SIZE)
        {
          rx0 = (x0<<3) + cmv_x;
          sx0 = (x0<<3) + smv_x;
          d   = diff;

          orig_line = &orig_pic [y0][x0];
          ref_line_p1 = get_line_p1 (ref1_pic, sy0, sx0,  img_height, img_width);
          ref_line_p2 = get_line_p2 (ref2_pic, ry0, rx0,  img_height, img_width);
          *d++ = *orig_line++  -  ((*(ref_line_p1     ) + *(ref_line_p2     ) + 1) >> 1);
          *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
          *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
          *d++ = *orig_line    -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);

          orig_line = &orig_pic [y1][x0];
          ref_line_p1 = get_line_p1 (ref1_pic, sy4, sx0,  img_height, img_width);
          ref_line_p2 = get_line_p2 (ref2_pic, ry4, rx0,  img_height, img_width);
          *d++ = *orig_line++  -  ((*(ref_line_p1     ) + *(ref_line_p2     ) + 1) >> 1);
          *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
          *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
          *d++ = *orig_line    -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);

          orig_line = &orig_pic [y2][x0];
          ref_line_p1 = get_line_p1 (ref1_pic, sy8, sx0,  img_height, img_width);
          ref_line_p2 = get_line_p2 (ref2_pic, ry8, rx0,  img_height, img_width);
          *d++ = *orig_line++  -  ((*(ref_line_p1     ) + *(ref_line_p2     ) + 1) >> 1);
          *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
          *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
          *d++ = *orig_line    -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);

          orig_line = &orig_pic [y3][x0];
          ref_line_p1 = get_line_p1 (ref1_pic, sy12, sx0,  img_height, img_width);
          ref_line_p2 = get_line_p2 (ref2_pic, ry12, rx0,  img_height, img_width);
          *d++ = *orig_line++  -  ((*(ref_line_p1     ) + *(ref_line_p2     ) + 1) >> 1);
          *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
          *d++ = *orig_line++  -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);
          *d++ = *orig_line    -  ((*(ref_line_p1 += 4) + *(ref_line_p2 += 4) + 1) >> 1);

          if (!test8x8transform)
          {
            if ((mcost += SATD (diff, input->hadamard)) > min_mcost)
            {
              abort_search = 1;
              break;
            }
          }
          else
          {
            int x8=x0/8, y8=y0/8;
            i = (x0&0x7) +  x8*64 + (y8 * blocksize_x/8)*64 ;
            for(k=0, j=y0; j<4 + y0; j++, k+=4)
              memcpy(&(c_diff[i + ((j&0x7)<<3)]), &diff[k], 4*sizeof(int));
          }
        }
      }
      if(test8x8transform)
        mcost += find_SATD (c_diff, blocktype,mb_ext_level);

      if (mcost < min_mcost)
      {
        min_mcost = mcost;
        best_pos  = pos;
      }
    }
    if (best_pos)
    {
      *mv_x += spiral_search_x [best_pos];
      *mv_y += spiral_search_y [best_pos];
    }
  }
  //===== return minimum motion cost =====
  return min_mcost;
}


/*!
***********************************************************************
* \brief
*    Motion Cost for Bidirectional modes
***********************************************************************
*/
int BPredPartitionCost (int   blocktype,
                        int   block8x8,
                        short fw_ref,
                        short bw_ref,
                        int   lambda_factor,
                        int   list,
                        int mb_ext_level)
{
  static int  bx0_ext[3][5][4] = {  {{0,0,0,0}, {0,0,0,0}, {0,0,0,0}, {0,2,0,0}, {0,2,0,2}},
  {{0,0,0,0}, {0,0,0,0}, {0,0,0,0}, {0,4,0,0}, {0,4,0,4}},
  {{0,0,0,0}, {0,0,0,0}, {0,0,0,0}, {0,8,0,0}, {0,8,0,8}}
  };
  static int  by0_ext[3][5][4] = {  {{0,0,0,0}, {0,0,0,0}, {0,2,0,0}, {0,0,0,0}, {0,0,2,2}},
  {{0,0,0,0}, {0,0,0,0}, {0,4,0,0}, {0,0,0,0}, {0,0,4,4}},
  {{0,0,0,0}, {0,0,0,0}, {0,8,0,0}, {0,0,0,0}, {0,0,8,8}}
  };
  int (*bx0)[4]=bx0_ext[mb_ext_level];
  int (*by0)[4]=by0_ext[mb_ext_level];
  int   diff[64];
  int   curr_blk[MB_BLOCK_SIZE<<MAX_MB_EXT_LEVEL][MB_BLOCK_SIZE<<MAX_MB_EXT_LEVEL]; // ABT pred.error buffer

  int   bsx       = min(input->blc_size[blocktype][0],8); //scalable to MB32X32
  int   bsy       = min(input->blc_size[blocktype][1],8);

  int   pic_pix_x, pic_pix_y, block_x, block_y;
  int   v, h, mcost, i, j, k;
  int   mvd_bits  = 0;
  int   parttype  = (blocktype<4?blocktype:4);

  int   step_h0   = (input->part_size[ parttype][0])<<mb_ext_level;
  int   step_v0   = (input->part_size[ parttype][1])<<mb_ext_level;
  int   step_h    = (input->part_size[blocktype][0])<<mb_ext_level;
  int   step_v    = (input->part_size[blocktype][1])<<mb_ext_level;

  int   bxx, byy;                               // indexing curr_blk

  short   ******all_mv = list ? img->bipred_mv1 : img->bipred_mv2;
  short   ******  p_mv = img->pred_mv;



#ifdef ADAPTIVE_MV_RES
  if (img->mv_res)
  {
    for (v=by0[parttype][block8x8]; v<by0[parttype][block8x8]+step_v0; v+=step_v)
    {
      for (h=bx0[parttype][block8x8]; h<bx0[parttype][block8x8]+step_h0; h+=step_h)
      {
        if (((all_mv [v][h][LIST_0][fw_ref][blocktype][0]&1)==0) && ((all_mv [v][h][LIST_0][fw_ref][blocktype][1]&1)==0))
        {
          mvd_bits += mvbits[ roundInt(all_mv [v][h][LIST_0][fw_ref][blocktype][0],1) - roundInt(p_mv[v][h][LIST_0][fw_ref][blocktype][0],1) ];
          mvd_bits += mvbits[ roundInt(all_mv [v][h][LIST_0][fw_ref][blocktype][1],1) - roundInt(p_mv[v][h][LIST_0][fw_ref][blocktype][1],1) ];
        }
        else
        {
          mvd_bits += mvbits[ all_mv [v][h][LIST_0][fw_ref][blocktype][0] - p_mv[v][h][LIST_0][fw_ref][blocktype][0] ];
          mvd_bits += mvbits[ all_mv [v][h][LIST_0][fw_ref][blocktype][1] - p_mv[v][h][LIST_0][fw_ref][blocktype][1] ];
          mvd_bits +=1;
        }
        if (((all_mv [v][h][LIST_1][bw_ref][blocktype][0]&1)==0) && ((all_mv [v][h][LIST_1][bw_ref][blocktype][1]&1)==0))
        {
          mvd_bits += mvbits[ roundInt(all_mv [v][h][LIST_1][bw_ref][blocktype][0],1) - roundInt(p_mv[v][h][LIST_1][bw_ref][blocktype][0],1) ];
          mvd_bits += mvbits[ roundInt(all_mv [v][h][LIST_1][bw_ref][blocktype][1],1) - roundInt(p_mv[v][h][LIST_1][bw_ref][blocktype][1],1) ];
        }
        else
        {
          mvd_bits += mvbits[ all_mv [v][h][LIST_1][bw_ref][blocktype][0] - p_mv[v][h][LIST_1][bw_ref][blocktype][0] ];
          mvd_bits += mvbits[ all_mv [v][h][LIST_1][bw_ref][blocktype][1] - p_mv[v][h][LIST_1][bw_ref][blocktype][1] ];
          mvd_bits +=1;
        }
      }
    }
  }
  else
#endif
    for (v=by0[parttype][block8x8]; v<by0[parttype][block8x8]+step_v0; v+=step_v)
    {
      for (h=bx0[parttype][block8x8]; h<bx0[parttype][block8x8]+step_h0; h+=step_h)
      {
        mvd_bits += mvbits[ all_mv [v][h][LIST_0][fw_ref][blocktype][0] - p_mv[v][h][LIST_0][fw_ref][blocktype][0] ];
        mvd_bits += mvbits[ all_mv [v][h][LIST_0][fw_ref][blocktype][1] - p_mv[v][h][LIST_0][fw_ref][blocktype][1] ];

        mvd_bits += mvbits[ all_mv [v][h][LIST_1][bw_ref][blocktype][0] - p_mv[v][h][LIST_1][bw_ref][blocktype][0] ];
        mvd_bits += mvbits[ all_mv [v][h][LIST_1][bw_ref][blocktype][1] - p_mv[v][h][LIST_1][bw_ref][blocktype][1] ];
      }
    }

    mcost = WEIGHTED_COST (lambda_factor, mvd_bits);

    //----- cost of residual signal -----
    for (byy=0, v=by0[parttype][block8x8]; v<by0[parttype][block8x8]+step_v0; byy+=4, v++) //scalable to MB32X32
    {

      pic_pix_y = img->opix_y + (block_y = (v<<2));      //scalable to MB32X32

      for (bxx=0, h=bx0[parttype][block8x8]; h<bx0[parttype][block8x8]+step_h0; bxx+=4, h++)//scalable to MB32X32
      {
        pic_pix_x = img->opix_x + (block_x = (h<<2));        //scalable to MB32X32
        LumaPrediction4x4Bi (block_x, block_y, 2, blocktype, blocktype, fw_ref, bw_ref, list);

        for (k=j=0; j<4; j++)
        {
          for (  i=0; i<4; i++)
            diff[k++] = curr_blk[byy+j][bxx+i] = 
            imgY_org[pic_pix_y+j][pic_pix_x+i] - img->mpr[j+block_y][i+block_x];            
        }
        if(mb_ext_level==0)
        {
          if ((!input->Transform8x8Mode) || (blocktype>4))
          {
            mcost += SATD (diff, input->hadamard);          
          }
        }
      }
    }

    if (input->Transform8x8Mode && (blocktype<=4))  // tchen 4-29-04
    {
      for (byy=0; byy <  (input->blc_size[parttype][1]<<mb_ext_level); byy+=bsy)
        for (bxx=0; bxx< (input->blc_size[parttype][0]<<mb_ext_level); bxx+=bsx)
        {
          for (k=0, j=byy;j<byy + 8;j++, k += 8)
            memcpy(&diff[k], &(curr_blk[j][bxx]), 8 * sizeof(int));          

          mcost += SATD8X8(diff, input->hadamard);
        }
    }
    return mcost;
}

/*!
***********************************************************************
* \brief
*    Block motion search
***********************************************************************
*/

int                                         //!< minimum motion cost after search
BlockMotionSearch (short     ref,           //!< reference idx
                   int       list,          //!< reference pciture list
                   int       mb_x,          //!< x-coordinate inside macroblock   (in the unit of pel)
                   int       mb_y,          //!< y-coordinate inside macroblock   (in the unit of pel)
                   int       blocktype,     //!< block type (1-16x16 ... 7-4x4)
                   int       search_range,  //!< 1-d search range for integer-position search
                   int       lambda_factor, //!< lagrangian parameter for determining motion cost
                   int       mb_ext_level) 
{
  short     pred_mv_x, pred_mv_y, mv_x, mv_y;
  int       i, j;

  int       mvshift   = (img->mv_res ? 3 : 2);
  int       max_value = INT_MAX;
  int       min_mcost = max_value;

  int       block_x   = (mb_x>>2);   //scalable to MB32X32
  int       block_y   = (mb_y>>2);

  int       bsx       = input->blc_size[blocktype][0]<<mb_ext_level; //blc_size = part_sizes * BLOCK_SIZE 
  int       bsy       = input->blc_size[blocktype][1]<<mb_ext_level;

  int       pic_pix_x = img->opix_x + mb_x;
  int       pic_pix_y = img->opix_y + mb_y;


#ifdef WIN32
  struct _timeb tstruct1;
  struct _timeb tstruct2;
#else
  struct timeb tstruct1;
  struct timeb tstruct2;
#endif

#ifdef SWITCHED_FILTERS
  static int min_mcost_offset;
#endif

  int me_tmp_time;
  short*    pred_mv = img->pred_mv[block_y][block_x][list][ref][blocktype];
  short****** all_mv    = img->all_mv;  
  int list_offset = 0;
  int *prevSad = EPZSDistortion[list + list_offset][blocktype - 1];


  //static pel_t   orig_val [256];
  //static pel_t  *orig_pic  [16] = {orig_val,     orig_val+ 16, orig_val+ 32, orig_val+ 48,
  //  orig_val+ 64, orig_val+ 80, orig_val+ 96, orig_val+112,
  //  orig_val+128, orig_val+144, orig_val+160, orig_val+176,
  //  orig_val+192, orig_val+208, orig_val+224, orig_val+240};

  static pel_t   orig_val [(MB_BLOCK_SIZE<<MAX_MB_EXT_LEVEL)*(MB_BLOCK_SIZE<<MAX_MB_EXT_LEVEL)];
  static pel_t  *orig_pic  [MB_BLOCK_SIZE<<MAX_MB_EXT_LEVEL];

  for(i=0; i<(MB_BLOCK_SIZE<<mb_ext_level); i++)
  {
    orig_pic[i] = orig_val+(MB_BLOCK_SIZE<<mb_ext_level)*i;
  }

#ifdef WIN32
  _ftime( &tstruct1 );    // start time ms
#else
  ftime(&tstruct1);
#endif


  //==================================
  //=====   GET ORIGINAL BLOCK   =====
  //==================================
  for (j = 0; j < bsy; j++)  //scalable to MB32X32
    memcpy(orig_pic[j],&imgY_org[pic_pix_y+j][pic_pix_x], bsx *sizeof(imgpel));

  //===========================================
  //=====   GET MOTION VECTOR PREDICTOR   =====
  //===========================================

#ifdef MV_COMPETITION
  if (input->mv_competition > 0)
  {
    SetMotionVectorPredictor_Competition(pred_mv, enc_picture->ref_idx[list], enc_picture->mv[list], ref, list, block_x, block_y, bsx, bsy, mb_ext_level);
  }
  else
#endif
    SetMotionVectorPredictor             (pred_mv, enc_picture->ref_idx[list], enc_picture->mv[list], ref, list, block_x, block_y, bsx, bsy, mb_ext_level);  //scalable to MB32X32

  pred_mv_x = pred_mv[0];
  pred_mv_y = pred_mv[1];

  //==================================
  //=====   INTEGER-PEL SEARCH   =====
  //==================================

  //--- perform motion search using EPZS schemes---
  {    
    //--- set search center ---
    mv_x = pred_mv_x / (1 << mvshift);
    mv_y = pred_mv_y / (1 << mvshift);

    mv_x = Clip3(-2047 + search_range, 2047 - search_range, mv_x);
    mv_y = Clip3(LEVELMVLIMIT[img->LevelIndex][0] + search_range, LEVELMVLIMIT[img->LevelIndex][1]  - search_range, mv_y);

    min_mcost = EPZSPelBlockMotionSearch (orig_pic, ref, list, list_offset, 
      enc_picture->ref_idx, enc_picture->mv,pic_pix_x, pic_pix_y, blocktype,
      pred_mv_x, pred_mv_y, &mv_x, &mv_y, search_range, min_mcost, lambda_factor, mb_ext_level); //scalable to MB32X32

#ifdef SWITCHED_FILTERS
    if(img->filterParam == SIFO_FIRST_PASS_FPO)
    {
      if(img->currMEOffset[list] == 0)
      {
        min_mcost_offset = min_mcost;
      }
      else
      {
        if(min_mcost < min_mcost_offset)
        {
          min_mcost_offset = min_mcost;
          if(input->EPZSSpatialMem)  
          {
            EPZSMotion[list + list_offset][ref][blocktype - 1][block_y][pic_pix_x >> 2][0] = (short) mv_x;
            EPZSMotion[list + list_offset][ref][blocktype - 1][block_y][pic_pix_x >> 2][1] = (short) mv_y;
          }
        }
      }
    }
#endif  
  }

  //===== convert search center to quarter-pel units =====
  mv_x <<= 2;
  mv_y <<= 2;

  //==============================
  //=====   SUB-PEL SEARCH   =====
  //==============================
  if (!input->DisableSubpelME)
  {
    if ((ref == 0 || (ref > 0 && min_mcost < 3.5 * prevSad[pic_pix_x >> 2])))
    {
      if (input->hadamard == 1)
      {
        min_mcost = max_value;
      }

      //if (input->EPZSSubPelME)
      min_mcost =  EPZSSubPelBlockMotionSearch (orig_pic, ref, list, pic_pix_x, pic_pix_y, blocktype,
        pred_mv_x, pred_mv_y, &mv_x, &mv_y, 9, 9, 9, min_mcost, lambda_factor, mb_ext_level);            
    }
    else
    {
      if (img->mv_res)
      {
        mv_x <<= 1;
        mv_y <<= 1;
      }
    }
  }


  //===============================================
  //=====   SET MV'S AND RETURN MOTION COST   =====
  //===============================================

  if (input->mv_competition > 0)
    GetBestPredVector( pred_mv,mv_x,mv_y,lambda_factor,ref,list,blocktype,mb_x,mb_y);  

#ifdef SWITCHED_FILTERS
  if(img->filterParam == SIFO_FIRST_PASS_FPO)
  {
    if(img->currMEOffset[list] == 0)
    {
      for(j = block_y; j < block_y + (bsy>>2); j++)    
      {
        for(i = block_x; i < block_x + (bsx>>2); i++)
        {
          all_mv[j][i][list][ref][blocktype][0] = mv_x;
          all_mv[j][i][list][ref][blocktype][1] = mv_y;
        }
      }
    }
    else
    {
      if(min_mcost < img->min_mcost_offset[list])
      {
        for(j = block_y; j < block_y + (bsy>>2); j++)    
        {
          for(i = block_x; i < block_x + (bsx>>2); i++)
          {
            all_mv[j][i][list][ref][blocktype][0] = mv_x;
            all_mv[j][i][list][ref][blocktype][1] = mv_y;
          }
        }
      }
    }
  }
  else
  {
    for (j=block_y; j < block_y + (bsy>>2); j++)    //scalable to MB32X32
    {
      for (i=block_x; i < block_x + (bsx>>2); i++)
      {
        all_mv[j][i][list][ref][blocktype][0] = mv_x;
        all_mv[j][i][list][ref][blocktype][1] = mv_y;
      }
    }
  }
#else
  for(j = block_y; j < block_y + (bsy>>2); j++)    //scalable to MB32X32
  {
    for(i = block_x; i < block_x + (bsx>>2); i++)
    {
      all_mv[j][i][list][ref][blocktype][0] = mv_x;
      all_mv[j][i][list][ref][blocktype][1] = mv_y;
    }
  }
#endif

  if (img->type==B_SLICE && input->BiPredMotionEstimation!=0 && (blocktype == 1) && (ref==0))
  {

    short   ******bipred_mv = list ? img->bipred_mv1 : img->bipred_mv2;
    int     min_mcostbi = max_value;
    short   bimv_x, bimv_y, tempmv_x ,tempmv_y;
    short   pred_mv_x1, pred_mv_y1;
    short   pred_mv_x2 = 0, pred_mv_y2 = 0;
    short   iterlist=list;
    short   pred_mv_bi[2];

#ifdef SWITCHED_FILTERS
    short old_mv_x = mv_x;
    short old_mv_y = mv_y;
    short mv_x_b16, mv_y_b16, bimv_x_b16, bimv_y_b16;
    int min_mcostbi_b16 = max_value;
    int k;
    int other_list = list ^ 1;
#endif


    if (input->mv_competition > 0)
    {
      SetMotionVectorPredictor_Competition (pred_mv_bi, enc_picture->ref_idx[list ^ 1], enc_picture->mv[(list == LIST_0? LIST_1: LIST_0)], 0, (list == LIST_0? LIST_1: LIST_0), block_x, block_y, bsx, bsy, mb_ext_level);
    }
    else
      SetMotionVectorPredictor             (pred_mv_bi, enc_picture->ref_idx[list ^ 1], enc_picture->mv[(list == LIST_0? LIST_1: LIST_0)], 0, (list == LIST_0? LIST_1: LIST_0), block_x, block_y, bsx, bsy, mb_ext_level);

    mv_x = (mv_x + (1<<(mvshift-1)))>>mvshift;
    mv_y = (mv_y + (1<<(mvshift-1)))>>mvshift;


    old_mv_x = mv_x;
    old_mv_y = mv_y;

#ifdef SWITCHED_FILTERS
    img->currMEOffset[other_list] = 0;
#endif

    for (i=0;i<=input->BiPredMERefinements;i++)
    {
      if (i%2)
      {
        pred_mv_x2=pred_mv[0];
        pred_mv_y2=pred_mv[1]; 
        pred_mv_x1=pred_mv_bi[0];
        pred_mv_y1=pred_mv_bi[1]; 
        tempmv_x=bimv_x;
        tempmv_y=bimv_y;        
        bimv_x=mv_x;
        bimv_y=mv_y;
        iterlist= list ^ 1;

      }
      else
      {
        pred_mv_x1=pred_mv[0];
        pred_mv_y1=pred_mv[1]; 
        pred_mv_x2=pred_mv_bi[0];
        pred_mv_y2=pred_mv_bi[1]; 

        if (i!=0)
        {
          tempmv_x=bimv_x;
          tempmv_y=bimv_y;        
          bimv_x=mv_x;
          bimv_y=mv_y;
        }
        else
        {
          tempmv_x=mv_x;
          tempmv_y=mv_y;        
          bimv_x = (pred_mv_x2 + (1<<(mvshift-1)))>>mvshift;
          bimv_y = (pred_mv_y2 + (1<<(mvshift-1)))>>mvshift;
        }

        iterlist=list;
      }
      min_mcostbi = EPZSBiPredBlockMotionSearch (orig_pic, ref, iterlist, list_offset, 
        enc_picture->ref_idx, enc_picture->mv, 
        pic_pix_x, pic_pix_y, blocktype, 
        pred_mv_x1, pred_mv_y1, pred_mv_x2, pred_mv_y2, 
        &bimv_x, &bimv_y, &tempmv_x, &tempmv_y, 
        input->BiPredMESearchRange, min_mcostbi, lambda_factor, mb_ext_level);

      mv_x=tempmv_x;
      mv_y=tempmv_y;        
    }

    mv_x=tempmv_x << 2;
    mv_y=tempmv_y << 2;
    bimv_x = bimv_x << 2;
    bimv_y = bimv_y << 2;

    if (input->BiPredMESubPel && !input->DisableSubpelME)
    {
      if (input->hadamard)
      {
        min_mcostbi = max_value;
      }
      min_mcostbi =  SubPelBlockSearchBiPred (orig_pic, 0, iterlist, pic_pix_x, pic_pix_y, blocktype,
        pred_mv_x2, pred_mv_y2, &bimv_x, &bimv_y, &mv_x, &mv_y, 9, 9, 9,
        min_mcostbi, lambda_factor, mb_ext_level);
    }

    if (input->BiPredMESubPel==2 && !input->DisableSubpelME)
    {
      if (input->hadamard)
      {
        min_mcostbi = max_value;
      }

      if(img->mv_res)
      {
        mv_x = mv_x>>1;
        mv_y = mv_x>>1;
        bimv_x = bimv_x>>1;
        bimv_y = bimv_y>>1;
      }
      min_mcostbi =  SubPelBlockSearchBiPred (orig_pic, 0, (iterlist == LIST_0? LIST_1: LIST_0), pic_pix_x, pic_pix_y, blocktype,
        pred_mv_x, pred_mv_y, &mv_x, &mv_y, &bimv_x, &bimv_y, 9, 9, 9,
        min_mcostbi, lambda_factor, mb_ext_level);
    }

#ifdef SWITCHED_FILTERS
    if(img->filterParam == SIFO_FIRST_PASS_FPO)
    {
      // Store results of the default search
      mv_x_b16 = mv_x;
      mv_y_b16 = mv_y;
      bimv_x_b16 = bimv_x;
      bimv_y_b16 = bimv_y;
      min_mcostbi_b16 = min_mcostbi;

      // Try offsets
      if((ref == 0) && (img->noMEOffsets[other_list] > 0))
      {
        for(k = 0; k < img->noMEOffsets[other_list]; ++k)
        {
          // Restore original values
          mv_x = old_mv_x;
          mv_y = old_mv_y;
          min_mcostbi = max_value;

          img->currMEOffset[other_list] = img->offsetMETab[other_list][k];
          for (i=0; i<=input->BiPredMERefinements; i++)
          {
            if (i%2)
            {
              pred_mv_x2=pred_mv[0];
              pred_mv_y2=pred_mv[1]; 
              pred_mv_x1=pred_mv_bi[0];
              pred_mv_y1=pred_mv_bi[1]; 
              tempmv_x=bimv_x;
              tempmv_y=bimv_y;        
              bimv_x=mv_x;
              bimv_y=mv_y;
              iterlist= list ^ 1;
            }
            else
            {
              pred_mv_x1=pred_mv[0];
              pred_mv_y1=pred_mv[1]; 
              pred_mv_x2=pred_mv_bi[0];
              pred_mv_y2=pred_mv_bi[1]; 

              if (i!=0)
              {
                tempmv_x=bimv_x;
                tempmv_y=bimv_y;        
                bimv_x=mv_x;
                bimv_y=mv_y;
              }
              else
              {
                tempmv_x=mv_x;
                tempmv_y=mv_y;        

                bimv_x = (pred_mv_x2 + (1<<(mvshift-1)))>>mvshift;
                bimv_y = (pred_mv_y2 + (1<<(mvshift-1)))>>mvshift;
              }

              iterlist=list;
            }

            min_mcostbi = EPZSBiPredBlockMotionSearch (orig_pic, ref, iterlist, list_offset, 
              enc_picture->ref_idx, enc_picture->mv, 
              pic_pix_x, pic_pix_y, blocktype, 
              pred_mv_x1, pred_mv_y1, pred_mv_x2, pred_mv_y2, 
              &bimv_x, &bimv_y, &tempmv_x, &tempmv_y, 
              input->BiPredMESearchRange, min_mcostbi, lambda_factor, mb_ext_level);

            mv_x=tempmv_x;
            mv_y=tempmv_y;        
          }

          mv_x=tempmv_x << 2;
          mv_y=tempmv_y << 2;
          bimv_x = bimv_x << 2;
          bimv_y = bimv_y << 2;

          if (input->BiPredMESubPel && !input->DisableSubpelME)
          {
            if (input->hadamard)
            {
              min_mcostbi = max_value;
            }

            min_mcostbi =  SubPelBlockSearchBiPred (orig_pic, 0, iterlist, pic_pix_x, pic_pix_y, blocktype,
              pred_mv_x2, pred_mv_y2, &bimv_x, &bimv_y, &mv_x, &mv_y, 9, 9, 9,
              min_mcostbi, lambda_factor, mb_ext_level);
          }

          if (input->BiPredMESubPel==2 && !input->DisableSubpelME)
          {
            if (input->hadamard)
            {
              min_mcostbi = max_value;
            }
            if(img->mv_res)
            {
              mv_x = mv_x>>1;
              mv_y = mv_x>>1;
              bimv_x = bimv_x>>1;
              bimv_y = bimv_y>>1;
            }
            min_mcostbi =  SubPelBlockSearchBiPred (orig_pic, 0, (iterlist == LIST_0? LIST_1: LIST_0), pic_pix_x, pic_pix_y, blocktype,
              pred_mv_x, pred_mv_y, &mv_x, &mv_y, &bimv_x, &bimv_y, 9, 9, 9,
              min_mcostbi, lambda_factor, mb_ext_level);
          }

          // Check cost
          if(min_mcostbi < min_mcostbi_b16)
          {
            min_mcostbi_b16 = min_mcostbi;
            mv_x_b16 = mv_x;
            mv_y_b16 = mv_y;
            bimv_x_b16 = bimv_x;
            bimv_y_b16 = bimv_y;
          }
        }
      }

      // Returns results
      mv_x = mv_x_b16;
      mv_y = mv_y_b16;
      bimv_x = bimv_x_b16;
      bimv_y = bimv_y_b16;
      min_mcostbi = min_mcostbi_b16;
    }
#endif // SWITCHED_FILTERS


#ifdef SWITCHED_FILTERS
    if(img->filterParam == SIFO_FIRST_PASS_FPO)
    {
      if(img->currMEOffset[list] == 0)
      {
        img->min_mcost_offset_bi[list] = min_mcostbi;
        for(j=block_y; j < block_y + (bsy>>2); j++)
        {
          for(i=block_x ; i < block_x + (bsx>>2); i++)      
          {
            bipred_mv[j][i][iterlist    ][0][blocktype][0] = mv_x;
            bipred_mv[j][i][iterlist    ][0][blocktype][1] = mv_y;
            bipred_mv[j][i][iterlist ^ 1][0][blocktype][0] = bimv_x;
            bipred_mv[j][i][iterlist ^ 1][0][blocktype][1] = bimv_y;        
          }
        }
      }
      else
      {
        if(min_mcostbi < img->min_mcost_offset_bi[list])
        {
          img->min_mcost_offset_bi[list] = min_mcostbi;
          for (j=block_y; j < block_y + (bsy>>2); j++)
          {
            for (i=block_x ; i < block_x + (bsx>>2); i++)      
            {
              bipred_mv[j][i][iterlist    ][0][blocktype][0] = mv_x;
              bipred_mv[j][i][iterlist    ][0][blocktype][1] = mv_y;
              bipred_mv[j][i][iterlist ^ 1][0][blocktype][0] = bimv_x;
              bipred_mv[j][i][iterlist ^ 1][0][blocktype][1] = bimv_y;        
            }
          }
        }
      }
    }
    else
    {
      for (j=block_y; j < block_y + (bsy>>2); j++)
      {
        for (i=block_x ; i < block_x + (bsx>>2); i++)      
        {
          bipred_mv[j][i][iterlist    ][0][blocktype][0] = mv_x;
          bipred_mv[j][i][iterlist    ][0][blocktype][1] = mv_y;
          bipred_mv[j][i][iterlist ^ 1][0][blocktype][0] = bimv_x;
          bipred_mv[j][i][iterlist ^ 1][0][blocktype][1] = bimv_y;        
        }
      }
    }
#else
    for (j=block_y; j < block_y + (bsy>>2); j++) //scalable to MB32X32
    {
      for (i=block_x ; i < block_x + (bsx>>2); i++)      
      {
        bipred_mv[j][i][iterlist    ][0][blocktype][0] = mv_x;
        bipred_mv[j][i][iterlist    ][0][blocktype][1] = mv_y;
        bipred_mv[j][i][iterlist ^ 1][0][blocktype][0] = bimv_x;
        bipred_mv[j][i][iterlist ^ 1][0][blocktype][1] = bimv_y;        
      }
    }
#endif
  }


#ifdef WIN32
  _ftime(&tstruct2);   // end time ms
#else
  ftime(&tstruct2);    // end time ms
#endif

  me_tmp_time=(int)((tstruct2.time*1000+tstruct2.millitm) - (tstruct1.time*1000+tstruct1.millitm)); 
  me_tot_time += me_tmp_time;
  me_time += me_tmp_time;

  return min_mcost;
}



/*!
***********************************************************************
* \brief
*    Motion Cost for Bidirectional modes
***********************************************************************
*/
int BIDPartitionCost (int   blocktype,
                      int   block8x8,
                      short fw_ref,
                      short bw_ref,
                      int   lambda_factor, int mb_ext_level)
{
  static int  bx0_ext[3][5][4] = { {{0,0,0,0}, {0,0,0,0}, {0,0,0,0}, {0,2,0,0}, {0,2,0,2}},
  {{0,0,0,0}, {0,0,0,0}, {0,0,0,0}, {0,4,0,0}, {0,4,0,4}},
  {{0,0,0,0}, {0,0,0,0}, {0,0,0,0}, {0,8,0,0}, {0,8,0,8}}
  };
  static int  by0_ext[3][5][4] = { {{0,0,0,0}, {0,0,0,0}, {0,2,0,0}, {0,0,0,0}, {0,0,2,2}},
  {{0,0,0,0}, {0,0,0,0}, {0,4,0,0}, {0,0,0,0}, {0,0,4,4}},
  {{0,0,0,0}, {0,0,0,0}, {0,8,0,0}, {0,0,0,0}, {0,0,8,8}}
  };
  int   diff[64];
  int   curr_blk[MB_BLOCK_SIZE<<MAX_MB_EXT_LEVEL][MB_BLOCK_SIZE<<MAX_MB_EXT_LEVEL]; // ABT pred.error buffer

  int   bsx       = min(input->blc_size[blocktype][0],8); //scalable to MB32X32
  int   bsy       = min(input->blc_size[blocktype][1],8);

  int   pic_pix_x, pic_pix_y, block_x, block_y;
  int   v, h, mcost, i, j, k;
  int   mvd_bits  = 0;
  int   parttype  = (blocktype<4?blocktype:4);

  int   step_h0   = (input->part_size[ parttype][0])<<mb_ext_level;
  int   step_v0   = (input->part_size[ parttype][1])<<mb_ext_level;
  int   step_h    = (input->part_size[blocktype][0])<<mb_ext_level;
  int   step_v    = (input->part_size[blocktype][1])<<mb_ext_level;

  int   bxx, byy;                               // indexing curr_blk
  int   bx = bx0_ext[mb_ext_level][parttype][block8x8];
  int   by = by0_ext[mb_ext_level][parttype][block8x8];

  short   ******all_mv = img->all_mv;
  short   ******  p_mv = img->pred_mv;

  //----- cost for motion vector bits -----

#ifdef ADAPTIVE_MV_RES
  if (img->mv_res)
  {
    for (v=by; v<by + step_v0; v+=step_v)
    {
      for (h=bx; h<bx + step_h0; h+=step_h)
      {
        if (((all_mv [v][h][LIST_0][fw_ref][blocktype][0]&1)==0) && ((all_mv [v][h][LIST_0][fw_ref][blocktype][1]&1)==0))
        {
          mvd_bits += mvbits[ roundInt(all_mv [v][h][LIST_0][fw_ref][blocktype][0],1) - roundInt(p_mv[v][h][LIST_0][fw_ref][blocktype][0],1) ];
          mvd_bits += mvbits[ roundInt(all_mv [v][h][LIST_0][fw_ref][blocktype][1],1) - roundInt(p_mv[v][h][LIST_0][fw_ref][blocktype][1],1) ];
        }
        else
        {
          mvd_bits += mvbits[ all_mv [v][h][LIST_0][fw_ref][blocktype][0] - p_mv[v][h][LIST_0][fw_ref][blocktype][0] ];
          mvd_bits += mvbits[ all_mv [v][h][LIST_0][fw_ref][blocktype][1] - p_mv[v][h][LIST_0][fw_ref][blocktype][1] ];
          mvd_bits +=1;
        }
        if (((all_mv [v][h][LIST_1][bw_ref][blocktype][0]&1)==0) && ((all_mv [v][h][LIST_1][bw_ref][blocktype][1]&1)==0))
        {
          mvd_bits += mvbits[ roundInt(all_mv [v][h][LIST_1][bw_ref][blocktype][0],1) - roundInt(p_mv[v][h][LIST_1][bw_ref][blocktype][0],1) ];
          mvd_bits += mvbits[ roundInt(all_mv [v][h][LIST_1][bw_ref][blocktype][1],1) - roundInt(p_mv[v][h][LIST_1][bw_ref][blocktype][1],1) ];
        }
        else
        {
          mvd_bits += mvbits[ all_mv [v][h][LIST_1][bw_ref][blocktype][0] - p_mv[v][h][LIST_1][bw_ref][blocktype][0] ];
          mvd_bits += mvbits[ all_mv [v][h][LIST_1][bw_ref][blocktype][1] - p_mv[v][h][LIST_1][bw_ref][blocktype][1] ];
          mvd_bits +=1;
        }
      }
    }
  }
  else
#endif
    for (v=by; v<by + step_v0; v+=step_v)  //scalable to MB32X32
    {
      for (h=bx; h<bx + step_h0; h+=step_h)
      {
        mvd_bits += mvbits[ all_mv [v][h][LIST_0][fw_ref][blocktype][0] - p_mv[v][h][LIST_0][fw_ref][blocktype][0] ];
        mvd_bits += mvbits[ all_mv [v][h][LIST_0][fw_ref][blocktype][1] - p_mv[v][h][LIST_0][fw_ref][blocktype][1] ];

        mvd_bits += mvbits[ all_mv [v][h][LIST_1][bw_ref][blocktype][0] - p_mv[v][h][LIST_1][bw_ref][blocktype][0] ];
        mvd_bits += mvbits[ all_mv [v][h][LIST_1][bw_ref][blocktype][1] - p_mv[v][h][LIST_1][bw_ref][blocktype][1] ];
      }
    }

    mcost = WEIGHTED_COST (lambda_factor, mvd_bits);

    //----- cost of residual signal -----
    for (byy=0, v=by; v<by + step_v0; byy+=4, v++)  //scalable to MB32X32
    {
      pic_pix_y = img->opix_y + (block_y = (v<<2));  //scalable to MB32X32
      for (bxx=0, h=bx; h<bx + step_h0; bxx+=4, h++)
      {
        pic_pix_x = img->opix_x + (block_x = (h<<2));
        LumaPrediction4x4 (block_x, block_y, 2, blocktype, blocktype, fw_ref, bw_ref);
        for (k=j=0; j<4; j++)
        {
          for (  i=0; i<4; i++)
            diff[k++] = curr_blk[byy+j][bxx+i] = 
            imgY_org[pic_pix_y+j][pic_pix_x+i] - img->mpr[j+block_y][i+block_x];
        }
        if ((!input->Transform8x8Mode) || (blocktype>4))
          mcost += SATD (diff, input->hadamard);
      }
    }
    if (input->Transform8x8Mode && (blocktype<=4))  // tchen 4-29-04
    {
      for (byy=0; byy < (input->blc_size[parttype][1]<<mb_ext_level); byy+=bsy)
        for (bxx=0; bxx< (input->blc_size[parttype][0]<<mb_ext_level); bxx+=bsx)
        {        
          for (k=0, j=byy;j<byy + 8;j++, k += 8)
            memcpy(&diff[k], &(curr_blk[j][bxx]), 8 * sizeof(int));          

          mcost += SATD8X8(diff, input->hadamard);
        }
    }
    return mcost;
}


/*!
************************************************************************
* \brief
*    Find motion vector for the Skip mode
************************************************************************
*/
void FindSkipModeMotionVector (int mb_ext_level)
{
  int   bx, by;
  short ******all_mv = img->all_mv;

  short pmv[2];

  int zeroMotionAbove;
  int zeroMotionLeft;
  PixelPos mb_a, mb_b;
  int      a_mv_y = 0;
  int      b_mv_y = 0;
  int      a_ref_idx = 0;
  int      b_ref_idx = 0;
  int      a_mv_x = 0;
  int      b_mv_x = 0;


  //Macroblock *currMB = &img->mb_data[img->current_mb_nr];

  // Set all the possible values for the skipped motion vector (in the case of the skip mode) according
  // to each possible predictor

#ifdef MV_COMPETITION
  int prediction_mode_for_skip;

  for (prediction_mode_for_skip = 0; prediction_mode_for_skip<mv_comp.nb_mode_for_skip; prediction_mode_for_skip++)
  {
    if (mv_comp.predictor_for_skip[prediction_mode_for_skip] == PRED_H264_MEDIAN)  
    {
#endif
      getLuma4x4Neighbour(img->current_mb_nr,0,0,-1, 0,&mb_a);
      getLuma4x4Neighbour(img->current_mb_nr,0,0, 0,-1,&mb_b);

      if (mb_a.available)
      {
        a_mv_x    = enc_picture->mv[LIST_0][mb_a.pos_y][mb_a.pos_x][0];
        a_mv_y    = enc_picture->mv[LIST_0][mb_a.pos_y][mb_a.pos_x][1];
        a_ref_idx = enc_picture->ref_idx[LIST_0][mb_a.pos_y][mb_a.pos_x];
      }

      if (mb_b.available)
      {
        b_mv_x    = enc_picture->mv[LIST_0][mb_b.pos_y][mb_b.pos_x][0];
        b_mv_y    = enc_picture->mv[LIST_0][mb_b.pos_y][mb_b.pos_x][1];
        b_ref_idx = enc_picture->ref_idx[LIST_0][mb_b.pos_y][mb_b.pos_x];
      }

      zeroMotionLeft  = !mb_a.available ? 1 : a_ref_idx==0 && a_mv_x==0 && a_mv_y==0 ? 1 : 0;
      zeroMotionAbove = !mb_b.available ? 1 : b_ref_idx==0 && b_mv_x==0 && b_mv_y==0 ? 1 : 0;

      if (zeroMotionAbove || zeroMotionLeft)
      {
        for (by = 0;by < (4<<mb_ext_level);by++)
          for (bx = 0;bx < (4<<mb_ext_level);bx++)
          {
            memset(all_mv [by][bx][0][0][0], 0, 2* sizeof(short));
          }

#ifdef MV_COMPETITION
          if (input->mv_competition > 0)
          {
            mv_comp.mv_pred_skip[prediction_mode_for_skip][0] = 0;
            mv_comp.mv_pred_skip[prediction_mode_for_skip][1] = 0;
          }
#endif
      }
      else
      {
        SetMotionVectorPredictor (pmv, enc_picture->ref_idx[LIST_0], enc_picture->mv[LIST_0], 0, LIST_0, 0, 0, (16<<mb_ext_level), (16<<mb_ext_level), mb_ext_level);

        for (by = 0;by < (4<<mb_ext_level);by++)
          for (bx = 0;bx < (4<<mb_ext_level);bx++)
          {
            memcpy(all_mv [by][bx][0][0][0], pmv, 2* sizeof(short));
            //all_mv [by][bx][0][0][0][0] = pmv[0];
            //all_mv [by][bx][0][0][0][1] = pmv[1];
          }
#ifdef MV_COMPETITION
          if (input->mv_competition > 0)
          {
            mv_comp.mv_pred_skip[prediction_mode_for_skip][0] = pmv[0];
            mv_comp.mv_pred_skip[prediction_mode_for_skip][1] = pmv[1];
          }
#endif
      }
#ifdef MV_COMPETITION    
    }
    else   
    {
      PixelPos block_a, block_b, block_c, block_d;
      int mb_nr = img->current_mb_nr;
      int block_x = 0, block_y = 0;
      int blockshape_x = 16 << mb_ext_level; 
      int blockshape_y = 16 << mb_ext_level; 
      int hv, mv_a, mv_b, mv_c;
      int y=(int)img->current_mb_nr/(img->width/16);  // Vertical
      int x=(int)img->current_mb_nr%(img->width/16);  // Horizontal
      //      int ref_frame = 0;
      int mb_x                 = 4*block_x;
      int mb_y                 = 4*block_y;      

      if(mb_ext_level==0)
      {
        getLuma4x4Neighbour(mb_nr, block_x, block_y,           -1,  0, &block_a);
        getLuma4x4Neighbour(mb_nr, block_x, block_y,            0, -1, &block_b);
        getLuma4x4Neighbour(mb_nr, block_x, block_y, blockshape_x, -1, &block_c);
        getLuma4x4Neighbour(mb_nr, block_x, block_y,           -1, -1, &block_d);
      }
      else
      {
        int reset_mb_nr, reset_block_x=block_x, reset_block_y=block_y;

        reset_mb_nr = ResetAvailabilityOfNeighbors32(mb_x, mb_y, &reset_block_x, &reset_block_y, blockshape_x, blockshape_y);

        getLuma4x4Neighbour(reset_mb_nr, reset_block_x, reset_block_y,           -1,  0, &block_a);
        getLuma4x4Neighbour(reset_mb_nr, reset_block_x, reset_block_y,            0, -1, &block_b);
        getLuma4x4Neighbour(reset_mb_nr, reset_block_x, reset_block_y, blockshape_x, -1, &block_c);
        getLuma4x4Neighbour(reset_mb_nr, reset_block_x, reset_block_y,           -1, -1, &block_d);
      }

      if (mv_comp.predictor_for_skip[prediction_mode_for_skip] == PRED_A)
      {
        for (hv=0; hv < 2; hv++)
        {
          if (block_a.available)  
          {
            mv_a = enc_picture->mv[LIST_0][block_a.pos_y][block_a.pos_x][hv];
            pmv[hv] = mv_a;
          }
          else 
            pmv[hv] = NOT_AVAILABLE;
        }

        mv_comp.mv_pred_skip[prediction_mode_for_skip][0] = pmv[0];
        mv_comp.mv_pred_skip[prediction_mode_for_skip][1] = pmv[1];

        if (mv_comp.nb_mode_for_skip == 1)
        {
          for (by = 0;by < 4<<mb_ext_level;by++)
            for (bx = 0;bx < 4<<mb_ext_level;bx++)
              memcpy(all_mv [by][bx][0][0][0], pmv, 2* sizeof(short));
        }
      }
      else if (mv_comp.predictor_for_skip[prediction_mode_for_skip] == PRED_B)
      {
        for (hv=0; hv < 2; hv++)
        {
          if (block_b.available)  
          {
            mv_b = enc_picture->mv[LIST_0][block_b.pos_y][block_b.pos_x][hv];
            pmv[hv] = mv_b;

          }
          else 
            pmv[hv] = NOT_AVAILABLE;
        }

        mv_comp.mv_pred_skip[prediction_mode_for_skip][0] = pmv[0];
        mv_comp.mv_pred_skip[prediction_mode_for_skip][1] = pmv[1];

        if (mv_comp.nb_mode_for_skip == 1)
        {
          for (by = 0;by < 4<<mb_ext_level;by++)
            for (bx = 0;bx < 4<<mb_ext_level;bx++)
              memcpy(all_mv [by][bx][0][0][0], pmv, 2* sizeof(short));
        }
      }
      else if (mv_comp.predictor_for_skip[prediction_mode_for_skip] == PRED_C)
      {
        for (hv=0; hv < 2; hv++)
        {
          if (block_c.available)  
          {
            mv_c = enc_picture->mv[LIST_0][block_c.pos_y][block_c.pos_x][hv];
            pmv[hv] = mv_c;

          }
          else 
            pmv[hv] = NOT_AVAILABLE;
        }

        mv_comp.mv_pred_skip[prediction_mode_for_skip][0] = pmv[0];
        mv_comp.mv_pred_skip[prediction_mode_for_skip][1] = pmv[1];

        if (mv_comp.nb_mode_for_skip == 1)
        {
          for (by = 0;by < 4<<mb_ext_level;by++)
            for (bx = 0;bx < 4<<mb_ext_level;bx++)
              memcpy(all_mv [by][bx][0][0][0], pmv, 2* sizeof(short));
        }
      }
      else if (mv_comp.predictor_for_skip[prediction_mode_for_skip] == PRED_ZERO)
      {
        pmv[0] = 0;
        pmv[1] = 0;

        mv_comp.mv_pred_skip[prediction_mode_for_skip][0] = pmv[0];
        mv_comp.mv_pred_skip[prediction_mode_for_skip][1] = pmv[1];

        if (mv_comp.nb_mode_for_skip == 1)
        {
          for (by = 0;by < 4<<mb_ext_level;by++)
            for (bx = 0;bx < 4<<mb_ext_level;bx++)
              memcpy(all_mv [by][bx][0][0][0], pmv, 2* sizeof(short));
        }
      }
      else if (mv_comp.predictor_for_skip[prediction_mode_for_skip] == PRED_COLOCATED)
      {
        if (collocated_mv_available(y, x, LIST_0) == TRUE)
        {
          // The collocated is the top left MV of the macroblock, whatever the subpartition...
          // To be improved.

          pmv[0] = listX[0][0]->mv[LIST_0][y*4][x*4][0] / (listX[0][0]->ref_idx[0][y*4][x*4] + 1);
          pmv[1] = listX[0][0]->mv[LIST_0][y*4][x*4][1] / (listX[0][0]->ref_idx[0][y*4][x*4] + 1);

        }
        else
        {
          pmv[0] = NOT_AVAILABLE;
          pmv[1] = NOT_AVAILABLE;
        }

        mv_comp.mv_pred_skip[prediction_mode_for_skip][0] = pmv[0];
        mv_comp.mv_pred_skip[prediction_mode_for_skip][1] = pmv[1];

        if (mv_comp.nb_mode_for_skip == 1)
        {
          for (by = 0;by < 4<<mb_ext_level;by++)
            for (bx = 0;bx < 4<<mb_ext_level;bx++)
              memcpy(all_mv [by][bx][0][0][0], pmv, 2* sizeof(short));
        }
      }
      else if (mv_comp.predictor_for_skip[prediction_mode_for_skip] == PRED_EXTENDEDSPATIAL)
      {
        for (hv=0; hv < 2; hv++)
        {
          if ((block_a.available) && (block_b.available) && (block_c.available))  
          {
            mv_a = enc_picture->mv[LIST_0][block_a.pos_y][block_a.pos_x][hv];
            mv_b = enc_picture->mv[LIST_0][block_b.pos_y][block_b.pos_x][hv];
            mv_c = enc_picture->mv[LIST_0][block_c.pos_y][block_c.pos_x][hv];  
            pmv[hv] = mv_a+mv_b+mv_c-min(mv_a,min(mv_b,mv_c))-max(mv_a,max(mv_b,mv_c));
          }
          else if (block_a.available) 
          {
            mv_a = enc_picture->mv[LIST_0][block_a.pos_y][block_a.pos_x][hv];
            pmv[hv] = mv_a;
          }
          else if (block_b.available)  
          {
            mv_b = enc_picture->mv[LIST_0][block_b.pos_y][block_b.pos_x][hv];    
            pmv[hv] = mv_b;
          }
          else if (block_c.available)  
          {
            mv_c = enc_picture->mv[LIST_0][block_c.pos_y][block_c.pos_x][hv];
            pmv[hv] = mv_c;
          }
          else 
            pmv[hv] = 0;

          mv_comp.mv_pred_skip[prediction_mode_for_skip][0] = pmv[0];
          mv_comp.mv_pred_skip[prediction_mode_for_skip][1] = pmv[1];

          if (mv_comp.nb_mode_for_skip == 1)
          {
            for (by = 0;by < 4<<mb_ext_level;by++)
              for (bx = 0;bx < 4<<mb_ext_level;bx++)
                memcpy(all_mv [by][bx][0][0][0], pmv, 2 * sizeof(short));
          }
        }    
      }
    }
  }
#endif
}

/*!
************************************************************************
* \brief
*    Motion search for a partition
*    applicable to inter 32x32, 32x16 and 16x32 only
*    blocktype 0-32x32, 1-32x32, 2-32x16, 3-16x32
************************************************************************
*/
void
PartitionMotionSearch (int    blocktype,
                       int    block8x8, //MB32x32: block8x8=32x32, 32x16, 16x32, 16x16; MB16x16: block8x8=16x16, 16x8, 8x16, 8x8  
                       int    lambda_factor,
                       int   transform8x8,
                       int  mb_ext_level
                       )
{
  //upper left corner co-or in the unit of 4x4 pel per block 
  static int  bx0_ext[3][5][4] = { {{0,0,0,0}, {0,0,0,0}, {0,0,0,0}, {0,2,0,0}, {0,2,0,2}},
  {{0,0,0,0}, {0,0,0,0}, {0,0,0,0}, {0,4,0,0}, {0,4,0,4}},
  {{0,0,0,0}, {0,0,0,0}, {0,0,0,0}, {0,8,0,0}, {0,8,0,8}}
  };
  static int  by0_ext[3][5][4] = { {{0,0,0,0}, {0,0,0,0}, {0,2,0,0}, {0,0,0,0}, {0,0,2,2}},
  {{0,0,0,0}, {0,0,0,0}, {0,4,0,0}, {0,0,0,0}, {0,0,4,4}},
  {{0,0,0,0}, {0,0,0,0}, {0,8,0,0}, {0,0,0,0}, {0,0,8,8}}
  };

  char  **ref_array;
  short ***mv_array;
  short *all_mv;
  short ref;
  int   v, h, mcost, search_range, i, j;
  int   pic_block_x, pic_block_y;
  int   bslice    = (img->type==B_SLICE);
  int   parttype  = (blocktype<4?blocktype:4);
  int   step_h0   = (input->part_size[ parttype][0])<<mb_ext_level; // in the unit of 4x4 pel per block (at or above 8x8 level) 
  int   step_v0   = (input->part_size[ parttype][1])<<mb_ext_level; 
  int   step_h    = (input->part_size[blocktype][0])<<mb_ext_level; // all levels 
  int   step_v    = (input->part_size[blocktype][1])<<mb_ext_level;
  int   list;
  int   numlists  = bslice ? 2 : 1;
  int   list_offset = img->mb_data[img->current_mb_nr].list_offset; 
  int   block_y;
  int   *m_cost = NULL;

  int   by = by0_ext[mb_ext_level][parttype][block8x8]; //scalable to MB32X32 because by0 and bx0 have been adapted to MB32X32
  int   bx = bx0_ext[mb_ext_level][parttype][block8x8];

#ifdef SWITCHED_FILTERS
  int   mcost_temp;
#endif

  int   mb32i;

  if(mb_ext_level==1)
    mb32i = 0;
  else if(mb_ext_level==1)
    mb32i = curr_mb32i;   //0 to 3
  else
    mb32i = 4;


  //===== LOOP OVER REFERENCE FRAMES =====
  for (list=0; list<numlists;list++)
  {
    for (ref=0; ref < listXsize[list+list_offset]; ref++)
    {
      if(input->UseRDO_Q)
      {
        if (isfirstQP)    //(Motion_Selected == 0)
          m_cost = &motion_cost[blocktype][list][ref][block8x8];
        else
        {
          if(mb_ext_level==0)
          {
            if (transform8x8) motion_cost[blocktype][list][ref][block8x8] = motion_cost8[curr_mbi][blocktype][list][ref][block8x8];
            else motion_cost[blocktype][list][ref][block8x8] = motion_cost4[curr_mbi][blocktype][list][ref][block8x8];
          }
          else
          {
            if (transform8x8) motion_cost[blocktype][list][ref][block8x8] = motion32_cost8[mb32i][blocktype][list][ref][block8x8];
            else motion_cost[blocktype][list][ref][block8x8] = motion32_cost4[mb32i][blocktype][list][ref][block8x8];
          }
        }
      }
      else
        m_cost = &motion_cost[blocktype][list][ref][block8x8];

      //----- set search range ---
#ifdef _FULL_SEARCH_RANGE_
      if      (input->full_search == 2) 
        search_range = input->search_range;
      else if (input->full_search == 1) 
        search_range = input->search_range /  (min(ref,1)+1);
      else                              
        search_range = input->search_range / ((min(ref,1)+1) * min(2,blocktype));
#else
      search_range = input->search_range / ((min(ref,1)+1) * min(2,blocktype));
#endif

      //----- set arrays -----
      ref_array = enc_picture->ref_idx[list];
      mv_array  = enc_picture->mv[list];

      //----- init motion cost -----
      //motion_cost[blocktype][list][ref][block8x8] = 0;
#ifdef RDO_Q
      if (!(input->UseRDO_Q) || isfirstQP)    //(Motion_Selected == 0)
#endif
        *m_cost = 0;

      //===== LOOP OVER SUB MACRO BLOCK partitions
      //for 32x32, 32x16 and 16x32 step_h(v)0=step_h(v)
      for (v=by; v<by + step_v0; v += step_v)// by: the coordinator of 8x8 block in the unit of 4x4 in a 16x16 MB
      {
        pic_block_y = img->block_y + v;

        for (h=bx; h<bx+step_h0; h+=step_h)
        {
          all_mv = img->all_mv[v][h][list][ref][blocktype];
          pic_block_x = img->block_x + h;

          //--- motion search for block ---


          if(input->UseRDO_Q)
          {
            int       bsx       = (input->blc_size[blocktype][0])<<mb_ext_level; //blc_size = part_sizes * BLOCK_SIZE 
            int       bsy       = (input->blc_size[blocktype][1])<<mb_ext_level;

            if (isfirstQP)   //(Motion_Selected == 0)
            { 
#ifdef SWITCHED_FILTERS
              img->currMEOffset[list] = 0;
              mcost = BlockMotionSearch   (ref, list, h<<2, v<<2, blocktype, search_range, lambda_factor, mb_ext_level); //scalable to MB32X32

              if(img->filterParam == SIFO_FIRST_PASS_FPO)
              {
                img->min_mcost_offset[list] = mcost;
                if (img->noMEOffsets[list] > 0 && (ref == 0))
                {
                  for (i=0; i<img->noMEOffsets[list]; i++)
                  {
                    img->currMEOffset[list] = img->offsetMETab[list][i];
                    mcost_temp = BlockMotionSearch   (ref, list, h<<2, v<<2, blocktype, search_range, lambda_factor, mb_ext_level); //scalable to MB32X32
                    mcost = (mcost_temp < mcost)? mcost_temp: mcost;
                    img->min_mcost_offset[list] = mcost;
                  }
                }
              }
#else
              mcost = BlockMotionSearch   (ref, list, h<<2, v<<2, blocktype, search_range, lambda_factor, mb_ext_level); //scalable to MB32X32
#endif
              *m_cost += mcost;

              if (mb_ext_level==0 && blocktype>=4) //store MVs during 1st QP
              {
                if (transform8x8)     //need this saving of MVs only for blocks size 8x8, 8x4, 4x8, 4x4...
                {
                  for (j=v; j < v + (bsy>>2); j++)    
                  {
                    for (i=h; i < h + (bsx>>2); i++)
                    {
                      img->all_mv8[curr_mbi][j][i][list][ref][blocktype][0] = img->all_mv[j][i][list][ref][blocktype][0];
                      img->all_mv8[curr_mbi][j][i][list][ref][blocktype][1] = img->all_mv[j][i][list][ref][blocktype][1];

                      img->bipred_mv1_8[curr_mbi][j][i][list][ref][blocktype][0] = img->bipred_mv1[j][i][list][ref][blocktype][0];
                      img->bipred_mv1_8[curr_mbi][j][i][list][ref][blocktype][1] = img->bipred_mv1[j][i][list][ref][blocktype][1];

                      img->bipred_mv2_8[curr_mbi][j][i][list][ref][blocktype][0] = img->bipred_mv2[j][i][list][ref][blocktype][0];
                      img->bipred_mv2_8[curr_mbi][j][i][list][ref][blocktype][1] = img->bipred_mv2[j][i][list][ref][blocktype][1];
                    }
                  }
                }
                else
                {
                  for (j=v; j < v + (bsy>>2); j++)    
                  {
                    for (i=h; i < h + (bsx>>2); i++)
                    {
                      img->all_mv4[curr_mbi][j][i][list][ref][blocktype][0] = img->all_mv[j][i][list][ref][blocktype][0];
                      img->all_mv4[curr_mbi][j][i][list][ref][blocktype][1] = img->all_mv[j][i][list][ref][blocktype][1];

                      img->bipred_mv1_4[curr_mbi][j][i][list][ref][blocktype][0] = img->bipred_mv1[j][i][list][ref][blocktype][0];
                      img->bipred_mv1_4[curr_mbi][j][i][list][ref][blocktype][1] = img->bipred_mv1[j][i][list][ref][blocktype][1];

                      img->bipred_mv2_4[curr_mbi][j][i][list][ref][blocktype][0] = img->bipred_mv2[j][i][list][ref][blocktype][0];
                      img->bipred_mv2_4[curr_mbi][j][i][list][ref][blocktype][1] = img->bipred_mv2[j][i][list][ref][blocktype][1];
                    }
                  }
                }
              }
            } //end of if(isfirstQP) //(Motion_Selected == 0)
            else    //restore MVs during other QPs
            {
              if (mb_ext_level==0 && blocktype>=4) 
              {
                if (transform8x8)
                {
                  for (j=v; j < v + (bsy>>2); j++)    
                  {
                    for (i=h; i < h + (bsx>>2); i++)
                    {
                      img->all_mv[j][i][list][ref][blocktype][0] = img->all_mv8[curr_mbi][j][i][list][ref][blocktype][0];
                      img->all_mv[j][i][list][ref][blocktype][1] = img->all_mv8[curr_mbi][j][i][list][ref][blocktype][1];

                      img->bipred_mv1[j][i][list][ref][blocktype][0] = img->bipred_mv1_8[curr_mbi][j][i][list][ref][blocktype][0];
                      img->bipred_mv1[j][i][list][ref][blocktype][1] = img->bipred_mv1_8[curr_mbi][j][i][list][ref][blocktype][1];

                      img->bipred_mv2[j][i][list][ref][blocktype][0] = img->bipred_mv2_8[curr_mbi][j][i][list][ref][blocktype][0];
                      img->bipred_mv2[j][i][list][ref][blocktype][1] = img->bipred_mv2_8[curr_mbi][j][i][list][ref][blocktype][1];
                    }
                  }
                }
                else
                {
                  for (j=v; j < v + (bsy>>2); j++)    
                  {
                    for (i=h; i < h + (bsx>>2); i++)
                    {
                      img->all_mv[j][i][list][ref][blocktype][0] = img->all_mv4[curr_mbi][j][i][list][ref][blocktype][0];
                      img->all_mv[j][i][list][ref][blocktype][1] = img->all_mv4[curr_mbi][j][i][list][ref][blocktype][1];

                      img->bipred_mv1[j][i][list][ref][blocktype][0] = img->bipred_mv1_4[curr_mbi][j][i][list][ref][blocktype][0];
                      img->bipred_mv1[j][i][list][ref][blocktype][1] = img->bipred_mv1_4[curr_mbi][j][i][list][ref][blocktype][1];

                      img->bipred_mv2[j][i][list][ref][blocktype][0] = img->bipred_mv2_4[curr_mbi][j][i][list][ref][blocktype][0];
                      img->bipred_mv2[j][i][list][ref][blocktype][1] = img->bipred_mv2_4[curr_mbi][j][i][list][ref][blocktype][1];
                    }
                  }
                }
              }
            }
          }
          else
          {
#ifdef SWITCHED_FILTERS
            img->currMEOffset[list] = 0;
            mcost = BlockMotionSearch     (ref, list, h<<2, v<<2, blocktype, search_range, lambda_factor, mb_ext_level); //scalable to MB32X32
            if(img->filterParam == SIFO_FIRST_PASS_FPO)
            {
              img->min_mcost_offset[list] = mcost;
              if(img->noMEOffsets[list] > 0 && ref == 0)
              {
                for(i=0; i<img->noMEOffsets[list]; i++)
                {
                  img->currMEOffset[list] = img->offsetMETab[list][i];
                  mcost_temp = BlockMotionSearch     (ref, list, h<<2, v<<2, blocktype, search_range, lambda_factor, mb_ext_level); //scalable to MB32X32
                  mcost = (mcost_temp < mcost)? mcost_temp: mcost;
                  img->min_mcost_offset[list] = mcost;
                }
              }    
            }
#else
            mcost = BlockMotionSearch     (ref, list, h<<2, v<<2, blocktype, search_range, lambda_factor, mb_ext_level); //scalable to MB32X32
#endif

            *m_cost += mcost;
          }

          //--- set motion vectors and reference frame (for motion vector prediction) ---
          for (j=0; j<step_v; j++) //scalable to MB32X32
          {
            block_y = pic_block_y+j;  
            memset(&ref_array [block_y][pic_block_x], ref, step_h * sizeof(char));

            for (i=0; i<step_h; i++)
            {
              memcpy(mv_array  [block_y][pic_block_x+i], all_mv, 2* sizeof(short));
              //mv_array  [block_y][pic_block_x+i][0] = all_mv[0];
              //mv_array  [block_y][pic_block_x+i][1] = all_mv[1];
            }
          }
        }
      }

      if(input->UseRDO_Q)
      {
        if (isfirstQP)    // (Motion_Selected == 0)
        {
          if(mb_ext_level==0)
          {
            if (transform8x8)
              motion_cost8[curr_mbi][blocktype][list][ref][block8x8]= motion_cost[blocktype][list][ref][block8x8];
            else
              motion_cost4[curr_mbi][blocktype][list][ref][block8x8]= motion_cost[blocktype][list][ref][block8x8];
          }
          else
          {
            if (transform8x8)
              motion32_cost8[mb32i][blocktype][list][ref][block8x8]= motion_cost[blocktype][list][ref][block8x8];
            else
              motion32_cost4[mb32i][blocktype][list][ref][block8x8]= motion_cost[blocktype][list][ref][block8x8];
          }
        }
        else
        {
          if(mb_ext_level==0)
          {
            if (transform8x8)
              motion_cost[blocktype][list][ref][block8x8]= motion_cost8[curr_mbi][blocktype][list][ref][block8x8];
            else
              motion_cost[blocktype][list][ref][block8x8]= motion_cost4[curr_mbi][blocktype][list][ref][block8x8];
          }
          else
          {
            if (transform8x8)
              motion_cost[blocktype][list][ref][block8x8]= motion32_cost8[mb32i][blocktype][list][ref][block8x8];
            else
              motion_cost[blocktype][list][ref][block8x8]= motion32_cost4[mb32i][blocktype][list][ref][block8x8];
          }
        }
      }
    }
  }
}



/*!
************************************************************************
* \brief
*    Calculate Direct Motion Vectors  *****
************************************************************************
*/
void Get_Direct_Motion_Vectors (int mb_ext_level)
{

  int   block_x, block_y, pic_block_x, pic_block_y, opic_block_x, opic_block_y;
  short ****all_mvs;

  byte  **   moving_block;
  short ****   co_located_mv;
  char  ***    co_located_ref_idx;
  int64 ***    co_located_ref_id;
  char  **     ref_pic_l0 = enc_picture->ref_idx[LIST_0];
  char  **     ref_pic_l1 = enc_picture->ref_idx[LIST_1];
  //Macroblock *currMB = &img->mb_data[img->current_mb_nr];


  moving_block = Co_located->moving_block;
  co_located_mv = Co_located->mv;
  co_located_ref_idx = Co_located->ref_idx;
  co_located_ref_id = Co_located->ref_pic_id;

  if (img->direct_spatial_mv_pred_flag)  //spatial direct mode copy from decoder
  {

    short l0_refA, l0_refB, l0_refD, l0_refC;
    short l1_refA, l1_refB, l1_refD, l1_refC; 
    short l0_refX,l1_refX;
    short pmvfw[2]={0,0},pmvbw[2]={0,0};

    PixelPos mb_a, mb_b, mb_d, mb_c;              

    if(mb_ext_level)
      CheckAvailabilityOfNeighbors32(mb_ext_level);


    getLuma4x4Neighbour(img->current_mb_nr, 0, 0, -1,  0,&mb_a);
    getLuma4x4Neighbour(img->current_mb_nr, 0, 0,  0, -1,&mb_b);
    getLuma4x4Neighbour(img->current_mb_nr, 0, 0, (16<<mb_ext_level), -1,&mb_c);
    getLuma4x4Neighbour(img->current_mb_nr, 0, 0, -1, -1,&mb_d);

    l0_refA = mb_a.available ? ref_pic_l0[mb_a.pos_y][mb_a.pos_x] : -1;
    l0_refB = mb_b.available ? ref_pic_l0[mb_b.pos_y][mb_b.pos_x] : -1;
    l0_refD = mb_d.available ? ref_pic_l0[mb_d.pos_y][mb_d.pos_x] : -1;
    l0_refC = mb_c.available ? ref_pic_l0[mb_c.pos_y][mb_c.pos_x] : l0_refD;      

    l1_refA = mb_a.available ? ref_pic_l1[mb_a.pos_y][mb_a.pos_x] : -1;
    l1_refB = mb_b.available ? ref_pic_l1[mb_b.pos_y][mb_b.pos_x] : -1;
    l1_refD = mb_d.available ? ref_pic_l1[mb_d.pos_y][mb_d.pos_x] : -1;
    l1_refC = mb_c.available ? ref_pic_l1[mb_c.pos_y][mb_c.pos_x] : l1_refD;      

    l0_refX = (l0_refA >= 0 && l0_refB >= 0) ? min(l0_refA,l0_refB): max(l0_refA,l0_refB);
    l0_refX = (l0_refX >= 0 && l0_refC >= 0) ? min(l0_refX,l0_refC): max(l0_refX,l0_refC);

    l1_refX = (l1_refA >= 0 && l1_refB >= 0) ? min(l1_refA,l1_refB): max(l1_refA,l1_refB);
    l1_refX = (l1_refX >= 0 && l1_refC >= 0) ? min(l1_refX,l1_refC): max(l1_refX,l1_refC);        

    if (l0_refX >=0)
      SetMotionVectorPredictor (pmvfw, enc_picture->ref_idx[LIST_0], enc_picture->mv[LIST_0], l0_refX, LIST_0, 0, 0, (16<<mb_ext_level), (16<<mb_ext_level), mb_ext_level);

    if (l1_refX >=0)
      SetMotionVectorPredictor (pmvbw, enc_picture->ref_idx[LIST_1], enc_picture->mv[LIST_1], l1_refX, LIST_1, 0, 0, (16<<mb_ext_level), (16<<mb_ext_level), mb_ext_level);

    for (block_y=0; block_y<(4<<mb_ext_level); block_y++)
    {
      pic_block_y  = (img->pix_y  >> 2) + block_y;
      opic_block_y = (img->opix_y >> 2) + block_y;

      for (block_x=0; block_x<(4<<mb_ext_level); block_x++)
      {
        pic_block_x  = (img->pix_x  >> 2) + block_x;
        opic_block_x = (img->opix_x >> 2) + block_x;

        all_mvs = img->all_mv[block_y][block_x];

        if (l0_refX >=0)
        {
          if (!l0_refX  && !moving_block[opic_block_y][opic_block_x])
          {

            memset(all_mvs[LIST_0][0][0], 0, 2* sizeof(short));
            //all_mvs[LIST_0][0][0][0] = 0;
            //all_mvs[LIST_0][0][0][1] = 0;            
            direct_ref_idx[LIST_0][pic_block_y][pic_block_x]=0;       
          }
          else
          {
            all_mvs[LIST_0][l0_refX][0][0] = pmvfw[0];
            all_mvs[LIST_0][l0_refX][0][1] = pmvfw[1];
            direct_ref_idx[LIST_0][pic_block_y][pic_block_x]= (char)l0_refX;              
          }
        }
        else
        {
          all_mvs[LIST_0][0][0][0] = 0;
          all_mvs[LIST_0][0][0][1] = 0;
          direct_ref_idx[LIST_0][pic_block_y][pic_block_x]=-1;          
        }

        if (l1_refX >=0)
        {
          if(l1_refX==0 && !moving_block[opic_block_y][opic_block_x])
          {                  
            all_mvs[LIST_1][0][0][0] = 0;
            all_mvs[LIST_1][0][0][1] = 0;
            direct_ref_idx[LIST_1][pic_block_y][pic_block_x]= (char)l1_refX;     
          }
          else
          {
            all_mvs[LIST_1][l1_refX][0][0] = pmvbw[0];
            all_mvs[LIST_1][l1_refX][0][1] = pmvbw[1];
            direct_ref_idx[LIST_1][pic_block_y][pic_block_x]= (char)l1_refX;
          }               
        }
        else
        {      
          direct_ref_idx[LIST_1][pic_block_y][pic_block_x]=-1;

          all_mvs[LIST_1][0][0][0] = 0;
          all_mvs[LIST_1][0][0][1] = 0;
        }

        // Test Level Limits if satisfied.
        { 
          if (l0_refX < 0 && l1_refX < 0)
          {
            direct_ref_idx[LIST_0][pic_block_y][pic_block_x] = 
              direct_ref_idx[LIST_1][pic_block_y][pic_block_x] = 0;
          }
          if      (direct_ref_idx[LIST_1][pic_block_y][pic_block_x] == -1) 
            direct_pdir[pic_block_y][pic_block_x] = 0;
          else if (direct_ref_idx[LIST_0][pic_block_y][pic_block_x] == -1) 
            direct_pdir[pic_block_y][pic_block_x] = 1;
          else                                                           
            direct_pdir[pic_block_y][pic_block_x] = 2;
        }        
      }
    }
  }
}

/*!
************************************************************************
* \brief
*    control the sign of a with b
************************************************************************
*/
int sign(int a,int b)
{
  return (b >= 0) ? absm(a) : -absm(a);
}



/*!
************************************************************************
* \brief
*    based on the subpartition of 32x32 block decide the 16x16 block on which the neighborhoold will be decided 
*  \param mb_x: x-coordinate inside 32x32   (in the unit of pel )
*  \param mb_y: y-coordinate inside 32x32   (in the unit of pel )
*  \param reset_block_x
*      input x block position
*  \param reset_block_y
*      input y block position
* \return
*    16x16 block number in the raster scan order
************************************************************************
*/

int ResetAvailabilityOfNeighbors32(int mb_x, int mb_y, int *reset_block_x, int *reset_block_y, int blockshape_x, int blockshape_y)
{
  int reset_mb_nr=img->current_mb_nr;

  if(mb_x==0 && mb_y==0 && blockshape_x==64 && blockshape_y==64) // 64x64 partition
    CheckAvailabilityOfNeighbors32(2);

  if(mb_x==0 && mb_y==0 && blockshape_x==32 && blockshape_y==32) // 32x32 partition
    CheckAvailabilityOfNeighbors32(1);

  if(mb_x==0 && mb_y==0 && blockshape_x==64 && blockshape_y==32) // 1st partition 64X32
    CheckAvailabilityOfNeighbors32(2);
  if(mb_x==0 && mb_y==0 && blockshape_x==32 && blockshape_y==16) // 1st partition 32x16
    CheckAvailabilityOfNeighbors32(1);


  if(mb_x==0 && mb_y==32 && blockshape_x==64 && blockshape_y==32)// 2nd partition 64X32
  {
    int saved_mb_nr;
    saved_mb_nr = img->current_mb_nr;
    reset_mb_nr = img->current_mb_nr + img->PicWidthInMbs * 2;
    img->current_mb_nr = reset_mb_nr;
    CheckAvailabilityOfNeighbors32(0);  
    img->current_mb_nr = saved_mb_nr;
    *reset_block_y = 0;
  }
  if(mb_x==0 && mb_y==16 && blockshape_x==32 && blockshape_y==16)// 2nd partition 32x16
  {
    int saved_mb_nr;
    saved_mb_nr = img->current_mb_nr;
    reset_mb_nr = img->current_mb_nr + img->PicWidthInMbs;
    img->current_mb_nr = reset_mb_nr;
    CheckAvailabilityOfNeighbors32(0);   
    img->current_mb_nr = saved_mb_nr;
    *reset_block_y = 0;
  }

  if(mb_x==0 && mb_y==0 && blockshape_x==32 && blockshape_y==64) // 1st partition 32x64
    CheckAvailabilityOfNeighbors32(1);
  if(mb_x==0 && mb_y==0 && blockshape_x==16 && blockshape_y==32) // 1st partition 16x32
    CheckAvailabilityOfNeighbors32(0);

  if(mb_x==32 && mb_y==0 && blockshape_x==32 && blockshape_y==64)// 2nd partition 32X64
  {
    int saved_mb_nr;
    saved_mb_nr = img->current_mb_nr;
    reset_mb_nr = img->current_mb_nr + 2;
    img->current_mb_nr = reset_mb_nr;
    CheckAvailabilityOfNeighbors32(1);
    img->current_mb_nr = saved_mb_nr;
    *reset_block_x = 0; 
  }
  if(mb_x==16 && mb_y==0 && blockshape_x==16 && blockshape_y==32)// 2nd partition 16x32
  {
    int saved_mb_nr;
    saved_mb_nr = img->current_mb_nr;
    reset_mb_nr = img->current_mb_nr + 1;
    img->current_mb_nr = reset_mb_nr;
    CheckAvailabilityOfNeighbors32(0);
    img->current_mb_nr = saved_mb_nr;
    *reset_block_x = 0; 
  }
  return reset_mb_nr;
}


