/* 
====================================================================================================================

The copyright in this software is being made available under the License included below. 
This software may be subject to other third party and 	contributor rights, including patent rights, and no such 
rights are granted under this license.

Copyright (c) 2010, QUALCOMM INC.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted only for 
the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
promoting such standards. The following conditions are required to be met:

* Redistributions of source code must retain the above copyright notice, this list of conditions and 
the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
the following disclaimer in the documentation and/or other materials provided with the distribution.
* The name of QUALCOMM INC.  may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

====================================================================================================================
*/
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "global.h"
#include "mbuffer.h"
#include "mb_access.h"
#include "cabac.h"
#include "string.h"
#include "vlc.h"
#include "mb_access.h"
#include "image.h"

#ifdef MV_COMPETITION
#include "mv_competition.h"

int skip_mode;
int *******mv_predictors;          // Stores the predictor for each sub-partition and for each predictor  
// predMVXY[block_x][block_y][mode][Ref_Frame][list][prediction number][xy];
int ******send_index_mv_prediction;        // int predModeMV2[mode][10][list][4][4][MAX_MODEPREDMV+1]; 
int *****predModeMV;          // contient les modes de prediction des predicteur MV [mode][ref][list][i][j]; int predModeMV[10][18*22];
int  *ranking;            //int  ranking[MAX_VAL];
int MAX_VAL ;    

extern int *mvbits;
extern int *assignSE2partition[2];
extern void SetMotionVectorPredictor (short  pmv[2],
                                      char   **refPic,
                                      short  ***tmp_mv,
                                      short  ref_frame,
                                      int    list,
                                      int    block_x,
                                      int    block_y,
                                      int    blockshape_x,
                                      int    blockshape_y,
                                      int    mb_ext_level);

MV_Competition mv_comp;

extern Macroblock  MB32;
extern int ResetAvailabilityOfNeighbors32(int mb_x, int mb_y, int *reset_block_x, int *reset_block_y, int blockshape_x, int blockshape_y);
extern Macroblock  MB64;

/*! 
*************************************************************************************
* \brief
*    initialization of motion vector competition arrays
*
* \param 
*    
*
* \return
*    
*************************************************************************************
*/

void init_MV_Competition()
{
  int l,n,m,i;
  int  j, k=0;
  int nLength;

  nLength = strlen(input->predictors_skip)-2;

  mv_comp.predictor_for_skip = (int *) malloc(sizeof(int) * MAX_MV_PREDICTOR);
  mv_comp.predictor_for_mvp  = (int *) malloc(sizeof(int) * MAX_MV_PREDICTOR);
  mv_comp.predictor_for_mvb  = (int *) malloc(sizeof(int) * MAX_MV_PREDICTOR);

  if (input->mv_competition == 0)
  {
    mv_comp.nb_mode_for_skip = 1;
    mv_comp.nb_mode_for_mvp  = 1;
    mv_comp.nb_mode_for_mvb  = 1;

    mv_comp.predictor_for_skip[0] = PRED_H264_MEDIAN;
    mv_comp.predictor_for_mvp[0]  = PRED_H264_MEDIAN;
    mv_comp.predictor_for_mvb[0]  = PRED_H264_MEDIAN;
  }
  else if (input->mv_competition == 1)
  {
    mv_comp.nb_mode_for_skip = 2;
    mv_comp.nb_mode_for_mvp  = 2;
    mv_comp.nb_mode_for_mvb  = 2;

    for (i=1; i<8; i++)
    {
      mv_comp.predictor_for_skip[i] = 0;
      mv_comp.predictor_for_mvp[i]  = 0;
      mv_comp.predictor_for_mvb[i]  = 0;
    }

    mv_comp.predictor_for_skip[0] = PRED_EXTENDEDSPATIAL;
    mv_comp.predictor_for_mvp[0]  = PRED_H264_MEDIAN;
    mv_comp.predictor_for_mvb[0]  = PRED_H264_MEDIAN;
    mv_comp.predictor_for_skip[1] = PRED_A;
    mv_comp.predictor_for_mvp[1]  = PRED_COLOCATED;
    mv_comp.predictor_for_mvb[1]  = PRED_COLOCATED;
  }
  else
  {
    mv_comp.nb_mode_for_skip = 0;
    mv_comp.nb_mode_for_mvp  = 0;
    mv_comp.nb_mode_for_mvb  = 0;

    // Check the number of predictors + enabled predictors for skip, motion vectors for P and B frames
    for(j=1;j<nLength+1;j++)
    {
      switch (input->predictors_skip[j])
      {
      case '1':
        mv_comp.predictor_for_skip[mv_comp.nb_mode_for_skip++] = j-1;
        break;
      case '0':
        break;
      default:
        snprintf(errortext, ET_SIZE, "Invalid Predictors_skip parameter.");
        error (errortext, 400);
        break;
      }

      switch (input->predictors_mvp[j])
      {
      case '1':
        mv_comp.predictor_for_mvp[mv_comp.nb_mode_for_mvp++] = j-1;
        break;
      case '0':
        break;
      default:
        snprintf(errortext, ET_SIZE, "Invalid Predictors_mvp parameter.");
        error (errortext, 400);
        break;
      }

      switch (input->predictors_mvb[j])
      {
      case '1':
        mv_comp.predictor_for_mvb[mv_comp.nb_mode_for_mvb++] = j-1;
        break;
      case '0':
        break;
      default:
        snprintf(errortext, ET_SIZE, "Invalid Predictors_mvb parameter.");
        error (errortext, 400);
        break;
      }

    }
  }

  mv_comp.mv_pred_skip = (int**) malloc(mv_comp.nb_mode_for_skip * sizeof(int*));  
  for(j=0;j<(mv_comp.nb_mode_for_skip);j++)
  {
    mv_comp.mv_pred_skip[j] = (int*) malloc(2 * sizeof(int)) ;             
  }

  mv_predictors = (int*******)malloc((4<<MAX_MB_EXT_LEVEL)*sizeof(int******)) ;              //block_x
  for(l=0; l<(4<<MAX_MB_EXT_LEVEL); l++)
  {
    (mv_predictors)[l] = (int******)malloc((4<<MAX_MB_EXT_LEVEL)*sizeof(int*****)) ;            //block_y
    for(k=0; k<(4<<MAX_MB_EXT_LEVEL); k++)
    {
      (mv_predictors)[l][k] = (int*****)malloc(MAXMODE*sizeof(int****)) ;      //mode
      for(j=0; j<(MAXMODE); j++)
      {
        (mv_predictors)[l][k][j]=(int****)malloc((input->num_ref_frames)*sizeof(int***)); //REF FRAME
        for(n=0; n<input->num_ref_frames; n++)
        {
          (mv_predictors)[l][k][j][n] = (int***)malloc(2*sizeof(int**)) ;      //LIST_0, LIST_1
          for(m=0; m<(2); m++)
          {
            (mv_predictors)[l][k][j][n][m] = (int**)malloc(15*sizeof(int*)) ;
            for (i=0; i<15; i++)
            {
              (mv_predictors)[l][k][j][n][m][i] = (int*)malloc(2*sizeof(int));  //x,y
            }
          }
        }
      }
    }
  }

  send_index_mv_prediction = (int******)malloc(MAXMODE*sizeof(int*****));

  for (k=0; k<MAXMODE; k++)
  {
    (send_index_mv_prediction)[k] = (int*****)malloc(img->num_ref_frames*sizeof(int****)) ;   // int predModeMV2[MAXMODE][10][list][18][22][MAX_MODEPREDMV+1]; 
    for (i=0; i<img->num_ref_frames; i++)
    {
      (send_index_mv_prediction)[k][i] = (int****)malloc((2)*sizeof(int***)); 
      for (m=0; m<2; m++)
      {
        (send_index_mv_prediction)[k][i][m] = (int***)malloc((4<<MAX_MB_EXT_LEVEL)*sizeof(int**)); 
        for(l=0; l<(4<<MAX_MB_EXT_LEVEL); l++)
        {
          (send_index_mv_prediction)[k][i][m][l] = (int**)malloc((4<<MAX_MB_EXT_LEVEL)*sizeof(int*)); 
          for(j=0; j<(4<<MAX_MB_EXT_LEVEL); j++)
            (send_index_mv_prediction)[k][i][m][l][j] = (int*)malloc((max(mv_comp.nb_mode_for_mvp,mv_comp.nb_mode_for_mvb)+1)*sizeof(int));
        }
      }
    }
  }

  predModeMV= (int*****)malloc(MAXMODE*sizeof(int****)) ;
  for(j=0; j<(MAXMODE); j++)
  {
    (predModeMV)[j]= (int****)malloc(img->num_ref_frames*sizeof(int***)) ;       //int predModeMV[MAXMODE][10][list][18][22];
    for (i=0; i<img->num_ref_frames; i++)
    {
      (predModeMV)[j][i] = (int***)malloc(2*sizeof(int**));
      for (m=0; m<2; m++)
      {
        (predModeMV)[j][i][m] = (int**)malloc((4<<MAX_MB_EXT_LEVEL)*sizeof(int*));
        for (k=0; k<(4<<MAX_MB_EXT_LEVEL); k++)
          (predModeMV)[j][i][m][k] = (int*)malloc((4<<MAX_MB_EXT_LEVEL)*sizeof(int));
      }
    }
  }

  if(input->successive_Bframe > 1)
  {
    mv_previous_B_frame = (short****)malloc(2*sizeof(short***));
    for (m=0; m<2; m++)
    {
      mv_previous_B_frame[m] = (short***)malloc((img->height/4)*sizeof(short**));
      for(i=0;i<img->height/4;i+=1)
      {
        mv_previous_B_frame[m][i] = (short**)malloc((img->width/4)*sizeof(short*));
        for(j=0;j<img->width/4;j+=1)
          mv_previous_B_frame[m][i][j] = (short*)malloc(2*sizeof(short));
      }
    }

    ref_idx_previous_B_frame = (char***)malloc(2*sizeof(char**));
    for (m=0; m<2; m++)
    {
      ref_idx_previous_B_frame[m] = (char**)malloc((img->height/4)*sizeof(char*));
      for(i=0;i<img->height/4;i+=1)
      {
        ref_idx_previous_B_frame[m][i] = (char*)malloc((img->width/4)*sizeof(char));
      }
    }
  }

  ranking =(int*)malloc (MAX_VAL*sizeof(int));                       //int  ranking[MAX_VAL];
  MAX_VAL=  max(mv_comp.nb_mode_for_mvp,mv_comp.nb_mode_for_mvb)*4;
}


/*! 
*************************************************************************************
* \brief
*    free motion vector competition arrays
*
* \param 
*    
*
* \return
*    
*************************************************************************************
*/

void close_MV_Competition ()
{
  int i, j, k, l, m, n;
  for(j=0;j<(mv_comp.nb_mode_for_skip);j++)
    free(mv_comp.mv_pred_skip[j]);

  free(mv_comp.predictor_for_skip);
  free(mv_comp.predictor_for_mvp );
  free(mv_comp.predictor_for_mvb);

  for(l=0; l<(4<<MAX_MB_EXT_LEVEL); l++)
  {
    for(k=0; k<(4<<MAX_MB_EXT_LEVEL); k++)
    {
      for(j=0; j<(MAXMODE); j++)
      {
        for(n=0; n<input->num_ref_frames; n++)
        {
          for(m=0; m<(2); m++)
          {
            for (i=0; i<15; i++)
              free(mv_predictors[l][k][j][n][m][i]);

            free(mv_predictors[l][k][j][n][m]);
          }
          free(mv_predictors[l][k][j][n]);
        }
        free(mv_predictors[l][k][j]);
      }
      free(mv_predictors[l][k]);
    }
    free(mv_predictors[l]);
  }
  free(mv_predictors);




  for (k=0; k<MAXMODE; k++)
  {
    for (i=0; i<img->num_ref_frames; i++)
    {
      for (m=0; m<2; m++)
      {
        for(l=0; l<(4<<MAX_MB_EXT_LEVEL); l++)
        {
          for(j=0; j<(4<<MAX_MB_EXT_LEVEL); j++)
            free(send_index_mv_prediction[k][i][m][l][j]);
          free(send_index_mv_prediction[k][i][m][l]);
        }
        free(send_index_mv_prediction[k][i][m]);
      }
      free(send_index_mv_prediction[k][i]);
    }
    free(send_index_mv_prediction[k]);
  }
  free(send_index_mv_prediction);


  for(j=0; j<(MAXMODE); j++)
  {
    for (i=0; i<img->num_ref_frames; i++)
    {
      for (m=0; m<2; m++)
      {
        for (k=0; k<(4<<MAX_MB_EXT_LEVEL); k++)
          free(predModeMV[j][i][m][k]);
        free(predModeMV[j][i][m]);
      }
      free(predModeMV[j][i]);
    }
    free(predModeMV[j]);
  }
  free(predModeMV);

  if(input->successive_Bframe > 1)
  {
    for (m=0; m<2; m++)
    {
      for(i=0;i<img->height/4;i+=1)
      {
        for(j=0;j<img->width/4;j+=1)
        {
          free(mv_previous_B_frame[m][i][j]);
        }
        free(mv_previous_B_frame[m][i]);
      }
      free(mv_previous_B_frame[m]);
    }
    free(mv_previous_B_frame);

    for (m=0; m<2; m++)
    {
      for(i=0;i<img->height/4;i+=1)
      {
        free(ref_idx_previous_B_frame[m][i]);
      }
      free(ref_idx_previous_B_frame[m]);
    }
    free(ref_idx_previous_B_frame);
  }
  free(ranking);
}

/*! 
*************************************************************************************
* \brief
*    writes the index of the selected motion vector predictor for the skip mode
*
* \param 
*    
*
* \return int
*    
*************************************************************************************
*/

int write_predictor_index_for_skip_mode()
{
  int            rate       = 0;
  Macroblock*     currMB     = &img->mb_data[img->current_mb_nr];

  currMB->send_index_predictor_skip = send_index_for_skip_prediction(currMB->best_predictor_for_skip);
  if (currMB->send_index_predictor_skip == TRUE)
    rate += writeMotionVectorPredictorSKIP(currMB->best_predictor_for_skip);

  return rate;
}

int write_predictor_index_for_skip_mode32()
{
  int            rate       = 0;
  MB32.send_index_predictor_skip = send_index_for_skip_prediction(MB32.best_predictor_for_skip);
  if (MB32.send_index_predictor_skip == TRUE)
    rate += writeMotionVectorPredictorSKIP(MB32.best_predictor_for_skip);
  return rate;
}

int write_predictor_index_for_skip_mode64()
{
  int            rate       = 0;
  MB64.send_index_predictor_skip = send_index_for_skip_prediction(MB64.best_predictor_for_skip);
  if (MB64.send_index_predictor_skip == TRUE)
    rate += writeMotionVectorPredictorSKIP(MB64.best_predictor_for_skip);
  return rate;
}

/*! 
*************************************************************************************
* \brief
*    
*
* \param 
*    
*
* \return
*    
*************************************************************************************
*/

int writeMotionVectorPredictorSKIP (int predictor)
{
  int            rate       = 0;
  Macroblock*    currMB     = &img->mb_data[img->current_mb_nr];
  SyntaxElement* currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
  int*           bitCount   = currMB->bitcounter;
  Slice*         currSlice  = img->currentSlice;
  const int*     partMap    = assignSE2partition[0];
  DataPartition* dataPart;

  currSE->value1 = predictor;
  currSE->type = SE_MV_PREDICTOR;
  currSE->value2 = (mv_comp.nb_mode_for_skip / 2) + (mv_comp.nb_mode_for_skip % 2);     
  currSE->writing = writeMB_Motion_predictor_CABAC;
  dataPart = &(currSlice->partArr[partMap[currSE->type]]);
  dataPart->writeSyntaxElement (currSE, dataPart);

  bitCount[BITS_SKIP_PRED] += currSE->len;
  rate                     += currSE->len;
  currSE++;  
  currMB->currSEnr++;

  return rate;
}


/*! 
*************************************************************************************
* \brief
*    
*
* \param 
*    
*
* \return
*    
*************************************************************************************
*/
int writeIndexForMotionVectorPrediction (int  refframe,  int  list_idx,
                                         int  mv_mode,
                                         int i,
                                         int j)
{
  int            rate       = 0;

  Macroblock*    currMB     = &img->mb_data[img->current_mb_nr];
  int*           bitCount   = currMB->bitcounter;
  SyntaxElement* currSE     = &img->MB_SyntaxElements[currMB->currSEnr];
  Slice*         currSlice  = img->currentSlice;
  const int*     partMap    = assignSE2partition[0];
  DataPartition* dataPart;

  int maxmode = 0;

  if (img->type == I_SLICE) 
    maxmode = 1;
  else if (img->type == P_SLICE)
    maxmode = mv_comp.nb_mode_for_mvp;
  else if (img->type == B_SLICE)
    maxmode = mv_comp.nb_mode_for_mvb;

  currSE->value1 = predModeMV[mv_mode][refframe][list_idx][j][i];  
  currSE->type = SE_MV_PREDICTOR;
  currSE->value2 = (maxmode / 2) + (maxmode % 2);     

  if (maxmode > 1)
  {
    currSE->writing = writeMB_Motion_predictor_CABAC;
    dataPart = &(currSlice->partArr[partMap[currSE->type]]);
    dataPart->writeSyntaxElement (currSE, dataPart);
  }

  bitCount[BITS_MV_PRED] += currSE->len;
  rate                   += currSE->len;
  currSE++;  
  currMB->currSEnr++;

  return rate;
}

/*! 
*************************************************************************************
* \brief
*    This function is used to arithmetically encode the Motion vector 
*   predictor
*
* \param 
*    
*
* \return
*    
*************************************************************************************
*/

void writeMB_Motion_predictor_CABAC(SyntaxElement *se, EncodingEnvironmentPtr eep_dp)
{
  Macroblock          *currMB = &img->mb_data[img->current_mb_nr];
  int act_ctx = 0;
  MotionInfoContexts *ctx  = (img->currentSlice)->mot_ctx;
  int nb_symbol;
  int curr_symbol;

  if(se->value2 != 0)
  {
    nb_symbol = se->value2;
    curr_symbol = se->value1;

    while(nb_symbol!=0)
    {
      if(currMB->mb_type == 0)
        biari_encode_symbol(eep_dp, (signed short) (curr_symbol % 2), &ctx->mv_predictor_skip_contexts[act_ctx]);
      else 
        if(img->type == P_SLICE)
          biari_encode_symbol(eep_dp, (signed short) (curr_symbol % 2), &ctx->mv_predictor_mvp_contexts[act_ctx]);
        else if(img->type == B_SLICE)
          biari_encode_symbol(eep_dp, (signed short) (curr_symbol % 2), &ctx->mv_predictor_mvb_contexts[act_ctx]);

      nb_symbol--;
      curr_symbol = curr_symbol / 2;
    }
  }
}


/*! 
*************************************************************************************
* \brief
*    Returns 0 if no motion vector is available for the previous frame
* at macrobloc position (pos_j, pos_i), 1 otherwise
* because of INTRA_MODE or out of frame (checked here)Writes motion vectors predictor for SKIP
*
* \param 
*    
*
* \return
*    
*************************************************************************************
*/
short collocated_mv_available(int pos_y, int pos_x, int list)
{
  short return_val = TRUE;

  if (listX[list][0]->ref_idx[0][pos_y*4][pos_x*4] == -1)  // Collocated block in INTRA
  {
    return_val = FALSE;
  }
  return (return_val);
}

short send_index_for_skip_prediction(int best_mode_skip)
{
  short mv_x_ref, mv_y_ref;
  short i;
  short return_val = FALSE;      // Assumes it can be guessed
  short nb_of_available = 0;        // Number of modes available

  mv_x_ref = mv_comp.mv_pred_skip[best_mode_skip][0];
  mv_y_ref = mv_comp.mv_pred_skip[best_mode_skip][1];

  for (i=0; i<mv_comp.nb_mode_for_skip; i++)
  {
    //Check if among all modes enabled, one single is AVAILABLE (all others are NOT_AVAILABLE)
    if (mv_comp.mv_pred_skip[i][0] != NOT_AVAILABLE) 
      nb_of_available++;
  }

  if (nb_of_available > 1)  //Otherwise does not need to be sent: only one is available
  {
    for (i=0; i<mv_comp.nb_mode_for_skip; i++)
    {
      if ((mv_comp.mv_pred_skip[i][0] != mv_x_ref) 
        || (mv_comp.mv_pred_skip[i][1] != mv_y_ref))

        return_val = TRUE;    // One of the component for one prediction is different
      // We can return 0, the mode cannot be guessed
    }
  }

  return (return_val);
}

/*! 
*************************************************************************************
* \brief
*    Set motion vector predictor
*
* \param 
*    
*
* \return
*    
*************************************************************************************
*/

void SetMotionVectorPredictor_Competition (short  pmv[2],
                                           char   **refPic,
                                           short  ***tmp_mv,
                                           short  ref_frame,
                                           int    list,
                                           int    block_x,
                                           int    block_y,
                                           int    blockshape_x,
                                           int    blockshape_y,
                                           int    mb_ext_level)
{

  int mb_x                 = 4*block_x;
  int mb_y                 = 4*block_y;
  short  pmv2[2];
  short predictor;

  int maxmode=0;
  int modeblock;
  int prediction_mode;
  int *predictors=NULL;

  int reset_mb_nr   = img->current_mb_nr;
  int reset_block_x = block_x;
  int reset_block_y = block_y;

  if(mb_ext_level>0)
  {
    reset_mb_nr = ResetAvailabilityOfNeighbors32(mb_x, mb_y, &reset_block_x, &reset_block_y, blockshape_x, blockshape_y);
    //    block_x = reset_block_x;
    //    block_y = reset_block_y;
    //  img->current_mb_nr will give the top-left MB number of a current 32x32 block
    //  reset_mb_nr will give the actual MB number 
    // block_x and block_y are 4x4 index inside a Big Block....
    // for eg, in 32x32 MB block_x and block_y range from 0 to 7...
    //===
  }

  modeblock = determine_mode_block(blockshape_x >> mb_ext_level,blockshape_y >> mb_ext_level);

  if (img->type == I_SLICE) 
    maxmode = 1;
  else if (img->type == P_SLICE)
  {
    maxmode = mv_comp.nb_mode_for_mvp;
    predictors = mv_comp.predictor_for_mvp;
  }
  else if (img->type == B_SLICE)
  {
    maxmode = mv_comp.nb_mode_for_mvb;
    predictors = mv_comp.predictor_for_mvb;
  }

  predictor = 0;

  for (prediction_mode = 0; prediction_mode<maxmode; prediction_mode++)
  {
    if (predictors[prediction_mode] == PRED_H264_MEDIAN)  
    {
      SetMotionVectorPredictor(pmv2,refPic,tmp_mv,ref_frame,list,      block_x,      block_y,blockshape_x,blockshape_y,mb_ext_level);
      if (predictor == 0 )
      { 
        pmv[0]=pmv2[0];
        pmv[1]=pmv2[1];
      }
    }
    else if (predictors[prediction_mode] == PRED_ZERO)
    {
      pmv2[0] = 0;
      pmv2[1] = 0;
    }
    else if (predictors[prediction_mode] == PRED_A)  
    {  
      PixelPos block_a;
      getLuma4x4Neighbour(reset_mb_nr, reset_block_x, reset_block_y,           -1,  0, &block_a);
      pmv2[0] = block_a.available  ? tmp_mv[block_a.pos_y][block_a.pos_x][0] : 0;
      pmv2[1] = block_a.available  ? tmp_mv[block_a.pos_y][block_a.pos_x][1] : 0;  
    }
    else if (predictors[prediction_mode] == PRED_B)  
    {
      PixelPos  block_b;

      getLuma4x4Neighbour(reset_mb_nr, reset_block_x, reset_block_y,            0, -1, &block_b);
      pmv2[0] = block_b.available  ? tmp_mv[block_b.pos_y][block_b.pos_x][0] : 0;
      pmv2[1] = block_b.available  ? tmp_mv[block_b.pos_y][block_b.pos_x][1] : 0;
    }
    else if (predictors[prediction_mode] == PRED_C)
    {
      PixelPos block_c, block_d;

      getLuma4x4Neighbour(reset_mb_nr, reset_block_x, reset_block_y, blockshape_x, -1, &block_c);
      getLuma4x4Neighbour(reset_mb_nr, reset_block_x, reset_block_y,           -1, -1, &block_d);

      if (mb_y > 0)
      {
        if (mb_x < (8<<mb_ext_level))  // first column of 8x8 blocks
        {
          if (mb_y==(8<<mb_ext_level))
          {
            if (blockshape_x == 16<<mb_ext_level)      block_c.available  = 0;
            else                                       block_c.available &= 1;
          }
          else
          {
            if (mb_x+blockshape_x != (8<<mb_ext_level))  block_c.available &= 1;
            else                                         block_c.available  = 0;
          }
        }
        else
        {
          if (mb_x+blockshape_x != 16<<mb_ext_level)   block_c.available &= 1;
          else                                         block_c.available  = 0;
        }
      }

      if (!block_c.available)
      {
        block_c=block_d;
      }

      pmv2[0] = block_c.available  ? tmp_mv[block_c.pos_y][block_c.pos_x][0] : 0;
      pmv2[1] = block_c.available  ? tmp_mv[block_c.pos_y][block_c.pos_x][1] : 0;
    }
    else if (predictors[prediction_mode] == PRED_COLOCATED)
    {    
      int y=(int)reset_mb_nr/(img->width/16);  // Vertical
      int x=(int)reset_mb_nr%(img->width/16);  // Horizontal
      int ytop=(int)img->current_mb_nr/(img->width/16);  // Vertical
      int xtop=(int)img->current_mb_nr%(img->width/16);  // Horizontal

      //for 16x16 MB, x=xtop and y=ytop

      //---(x,y) is the 16x16MB location (even within Bigblocks)
      if (img->type == P_SLICE)
      {
        if (collocated_mv_available(y, x, LIST_0) == TRUE)
        {
          pmv2[0] = listX[0][0]->mv[LIST_0][ytop*4+block_y][xtop*4+block_x][0] * (ref_frame+1) / (listX[0][0]->ref_idx[0][ytop*4+block_y][xtop*4+block_x] + 1);
          pmv2[1] = listX[0][0]->mv[LIST_0][ytop*4+block_y][xtop*4+block_x][1] * (ref_frame+1) / (listX[0][0]->ref_idx[0][ytop*4+block_y][xtop*4+block_x] + 1);
        }
        else
        {
          pmv2[0] = 0;//NOT_AVAILABLE;
          pmv2[1] = 0;//NOT_AVAILABLE;
        }
      } 
      else if (img->type == B_SLICE)
      {
        SetMotionVectorPredictor_collocated_HB_SLICE(pmv2, ytop, xtop, list, ref_frame, block_y, block_x);
      }
    }
    mv_predictors[block_x][block_y][modeblock][ref_frame][list][prediction_mode][0] = pmv2[0];
    mv_predictors[block_x][block_y][modeblock][ref_frame][list][prediction_mode][1] = pmv2[1];

  }

  // Check if the index for the motion vector prediction needs to be sent or not
  // Stored in predModeMV2
  send_index_for_mv_prediction(block_x, block_y, modeblock, ref_frame, list, maxmode);
}

/*! 
*************************************************************************************
* \brief
*    
*
* \param 
*    
*
* \return
*    
*************************************************************************************
*/

void send_index_for_mv_prediction(int block_x, int block_y, int modeblock, short ref_frame, int list, int maxmode) 
{
  short egality_of_predictors = 0;
  short current_predictor_index;
  short nb_of_available = 0;
  short i;


  int pred0_x, pred0_y, current_pred_x, current_pred_y; 

  pred0_x = mv_predictors[block_x][block_y][modeblock][ref_frame][list][0][0];
  pred0_y = mv_predictors[block_x][block_y][modeblock][ref_frame][list][0][1];

  for (i=0; i<maxmode; i++)
  {
    //Check if among all modes enabled, one single is AVAILABLE (all others are NOT_AVAILABLE)
    if (mv_predictors[block_x][block_y][modeblock][ref_frame][list][i][0] != NOT_AVAILABLE) 
      nb_of_available++;
  }

  //  if (nb_of_available > 1)  //Otherwise does not need to be sent: only one is available
  {
    for(current_predictor_index=1;current_predictor_index<maxmode ;current_predictor_index++)
    {
      current_pred_x = mv_predictors[block_x][block_y][modeblock][ref_frame][list][current_predictor_index][0];
      current_pred_y = mv_predictors[block_x][block_y][modeblock][ref_frame][list][current_predictor_index][1];

      if ((current_pred_x == pred0_x) && (current_pred_y == pred0_y))
        egality_of_predictors++;  
    }
  }

  if(egality_of_predictors == maxmode-1)
    send_index_mv_prediction[modeblock][ref_frame][list][block_y][block_x][maxmode]=1;
  else 
    send_index_mv_prediction[modeblock][ref_frame][list][block_y][block_x][maxmode]=0;
}

/*! 
*************************************************************************************
* \brief
*    return le mode associ au blockshape x et y
*
* \param 
*    
*
* \return
*    
*************************************************************************************
*/

int determine_mode_block(int blockshape_x,int blockshape_y)
{
  short return_val=0;

  if((blockshape_x==16) && ( blockshape_y==16))
    return_val = 1;
  else if((blockshape_x==16) && ( blockshape_y==8))
    return_val = 2;
  else if((blockshape_x==8) && ( blockshape_y==16))
    return_val = 3;
  else if((blockshape_x==8) && ( blockshape_y==8))
    return_val = 4;
  else if((blockshape_x==8) && ( blockshape_y==4))
    return_val = 5;
  else if((blockshape_x==4) && ( blockshape_y==8))
    return_val =  6;
  else if((blockshape_x==4) && ( blockshape_y==4))
    return_val = 7;

  return (return_val);
}

/*! 
*************************************************************************************
* \brief
*    return  min cost for all prediction of MV 
*
* \param 
*    
*
* \return
*    
*************************************************************************************
*/

int compute_mv_cost_for_each_predictor(int f, int s, int cx, int cy, int px, int py, int ref, int list, int modeblock, int mb_x, int mb_y)  
{
  int min=MAX_VALUE;
  int i;


  int val, p;  

  int block_x = mb_x/4;
  int block_y = mb_y/4;
  int maxmode = 0;

  if (img->type == P_SLICE)
    maxmode = mv_comp.nb_mode_for_mvp;
  else if (img->type == B_SLICE)
    maxmode = mv_comp.nb_mode_for_mvb;

  for(i=0;i<maxmode;i++)
  {
    p = send_index_mv_prediction[modeblock][ref][list][block_y][block_x][maxmode];
    val = MV_COST(f,s,cx,cy,mv_predictors[block_x][block_y][modeblock][ref][list][i][0],mv_predictors[block_x][block_y][modeblock][ref][list][i][1])
      +    MVMODE_COST(f,i,p);

    if(val<min) 
      min = val;

  }
  return min ;
}


#ifdef ADAPTIVE_MV_RES
int compute_mv_cost_for_each_predictor_1(int f, int s, int d,int cx, int cy, int px, int py, int ref, int list, int modeblock, int mb_x, int mb_y,int extra_bits)  
{
  int min=MAX_VALUE;
  int i;
  int val, p;  
  int block_x = mb_x/4;
  int block_y = mb_y/4;
  int maxmode = 0;

  if (img->type == P_SLICE)
    maxmode = mv_comp.nb_mode_for_mvp;
  else if (img->type == B_SLICE)
    maxmode = mv_comp.nb_mode_for_mvb;

  for(i=0;i<maxmode;i++)
  {
    p = send_index_mv_prediction[modeblock][ref][list][block_y][block_x][maxmode];
    val = MV_COST_1(f,s,d,cx,cy,mv_predictors[block_x][block_y][modeblock][ref][list][i][0],mv_predictors[block_x][block_y][modeblock][ref][list][i][1],extra_bits)
      +    MVMODE_COST(f,i,p);

    if(val<min) 
      min = val;
  }
  return min ;
}
#endif

/*! 
*************************************************************************************
* \brief
*    
*
* \param 
*    
*
* \return
*    
*************************************************************************************
*/

int  MVMODE_COST(int f,int mode,int skip) 
{ 
  int nbbits=1;

  if(skip==1) nbbits=0;

  return nbbits*MVMODE_COST_FACTOR;

}

/*! 
*************************************************************************************
* \brief
*    get best prediction for motion vector 
*
* \param 
*    
*
* \return
*    
*************************************************************************************
*/

void GetBestPredVector (short  *pmv,short mv_x,short mv_y,int /*double*/ lambda,int ref, int list,int modeblock,int mb_x,int mb_y)
{
  //  Macroblock*    currMB     = &img->mb_data[img->current_mb_nr];
  //  int stop = 0;
  int current_predictor;

  int maxmode = 0;
  int current_cost;
  int nb_best_mode = 0;
#ifdef ADAPTIVE_MV_RES  
  int best_cost;
#else
  int best_cost = compute_mv_cost_for_each_predictor(lambda,0,mv_x,mv_y,0,0,ref,list,modeblock,mb_x,mb_y);
#endif
  int block_x = mb_x/4;
  int block_y = mb_y/4;

  if (img->type == I_SLICE) 
    maxmode = 1;
  else if (img->type == P_SLICE)
    maxmode = mv_comp.nb_mode_for_mvp;
  else if (img->type == B_SLICE)
    maxmode = mv_comp.nb_mode_for_mvb;

#ifdef ADAPTIVE_MV_RES
  if (img->mv_res)
  {
    if ((mv_x&1)==0 && (mv_y&1)==0) //1/4
      best_cost = compute_mv_cost_for_each_predictor_1(lambda,0,1,mv_x,mv_y,0,0,ref,list,modeblock,mb_x,mb_y,0);
    else
      best_cost = compute_mv_cost_for_each_predictor_1(lambda,0,0,mv_x,mv_y,0,0,ref,list,modeblock,mb_x,mb_y,1);
  }
  else
    best_cost = compute_mv_cost_for_each_predictor(lambda,0,mv_x,mv_y,0,0,ref,list,modeblock,mb_x,mb_y);
#endif

  for(current_predictor=0; current_predictor<maxmode; current_predictor++)
  {       

#ifdef ADAPTIVE_MV_RES
    if (img->mv_res)
    {
      if ((mv_x&1)==0 && (mv_y&1)==0) //1/4
      {
        current_cost = (MV_COST_1(lambda,0,1,mv_x,mv_y,mv_predictors[block_x][block_y][modeblock][ref][list][current_predictor][0],mv_predictors[block_x][block_y][modeblock][ref][list][current_predictor][1],0))+
          MVMODE_COST(lambda,current_predictor,send_index_mv_prediction[modeblock][ref][list][block_y][block_x][maxmode]);
      }
      else
        current_cost = (MV_COST_1(lambda,0,0,mv_x,mv_y,mv_predictors[block_x][block_y][modeblock][ref][list][current_predictor][0],mv_predictors[block_x][block_y][modeblock][ref][list][current_predictor][1],1))+
        MVMODE_COST(lambda,current_predictor,send_index_mv_prediction[modeblock][ref][list][block_y][block_x][maxmode]);
    }
    else
#endif
      current_cost = (MV_COST(lambda,0,mv_x,mv_y,mv_predictors[block_x][block_y][modeblock][ref][list][current_predictor][0],mv_predictors[block_x][block_y][modeblock][ref][list][current_predictor][1]))+
      MVMODE_COST(lambda,current_predictor,send_index_mv_prediction[modeblock][ref][list][block_y][block_x][maxmode]);

    if (current_cost == best_cost)
    {
      if(nb_best_mode == 0)
      {
        predModeMV[modeblock][ref][list][block_y][block_x]=current_predictor;      
        pmv[0]=(short)mv_predictors[block_x][block_y][modeblock][ref][list][current_predictor][0];         
        pmv[1]=(short)mv_predictors[block_x][block_y][modeblock][ref][list][current_predictor][1];         
      }
      send_index_mv_prediction[modeblock][ref][list][block_y][block_x][current_predictor]=1; 
      nb_best_mode++;
    }
    else 
    {
      send_index_mv_prediction[modeblock][ref][list][block_y][block_x][current_predictor]=0;
    }
  }

}


void init_mv_scale_hb()
{
  int i, j, k, l;
  int iTRb = 0 ; int iTRp = 0;
  int prescale = 0;
  int num_ref_max = 0;

  // Mv in current B frame points to LIST_0 or LIST_1
  for(i = 0 ; i < 2 ; i++)
  {
    for(k = 0 ; k <listXsize[i] ; k++)
    {
      iTRb = Clip3(-128, 127, enc_picture->poc - listX[i][k]->poc);

      // The colocated vector points to LIST_0 or LIST_1
      for(j=0;j<2;j++)
      {
        if(listX[LIST_1][0]->slice_type == B_SLICE) 
        {
          if(j==0) num_ref_max = input->B_List0_refs;
          if(j==1) num_ref_max = input->B_List1_refs;
        }
        if(listX[LIST_1][0]->slice_type == P_SLICE) 
        {
          num_ref_max = input->num_ref_frames;
        }

        for(l = 0 ; l < num_ref_max ; l++)
        {
          iTRp = (int) Clip3(-128, 127, listX[LIST_1][0]->poc - (listX[LIST_1][0]->ref_pic_num[j][l]/2));

          if(iTRp==0) img->mvscale_hb[i][j][k][l]=9999;
          else 
          {
            prescale=(16384 + absm(iTRp/2))/iTRp;
            img->mvscale_hb[i][j][k][l] = Clip3(-2048, 2047,(iTRb*prescale + 32) >> 6);
          }
        }
      }
    }
  }

}

/*! 
*************************************************************************************
* \brief
*    Set colocated motion vector for hirarchical B frames
*
* \param 
*    
*
* \return
*    
*************************************************************************************
*/
void SetMotionVectorPredictor_collocated_HB_SLICE (short  pmv[2], int y, int x, int list, int ref_frame, int block_y, int block_x)
{
  int colocated_MV_x = 0, colocated_MV_y = 0;
  int ref_frame_coloc = 0;
  pmv[0] = 0; pmv[1] = 0; 

  // B frame mv points to LIST_0
  if(list == LIST_0)
  {
    // If colocated mv points to LIST_0
    if(listX[LIST_1][0]->ref_idx[LIST_0][y*4+block_y][x*4+block_x] != -1)
    {
      colocated_MV_x = listX[LIST_1][0]->mv[LIST_0][y*4+block_y][x*4+block_x][0];
      colocated_MV_y = listX[LIST_1][0]->mv[LIST_0][y*4+block_y][x*4+block_x][1];

      ref_frame_coloc = listX[LIST_1][0]->ref_idx[LIST_0][y*4+block_y][x*4+block_x];

      // Scaling
      pmv[0] = (img->mvscale_hb[LIST_0][LIST_0][ref_frame][ref_frame_coloc] * colocated_MV_x + 128) >> 8;
      pmv[1] = (img->mvscale_hb[LIST_0][LIST_0][ref_frame][ref_frame_coloc] * colocated_MV_y + 128) >> 8;
    }
    else
    {
      // If colocated mv points to LIST_1
      if(listX[LIST_1][0]->ref_idx[LIST_1][y*4+block_y][x*4+block_x] != -1)
      {
        colocated_MV_x = listX[LIST_1][0]->mv[LIST_1][y*4+block_y][x*4+block_x][0];
        colocated_MV_y = listX[LIST_1][0]->mv[LIST_1][y*4+block_y][x*4+block_x][1];

        ref_frame_coloc = listX[LIST_1][0]->ref_idx[LIST_1][y*4+block_y][x*4+block_x];

        // Scaling
        pmv[0] = (img->mvscale_hb[LIST_0][LIST_1][ref_frame][ref_frame_coloc] * colocated_MV_x + 128) >> 8;
        pmv[1] = (img->mvscale_hb[LIST_0][LIST_1][ref_frame][ref_frame_coloc] * colocated_MV_y + 128) >> 8;
      }
      // No colocated mv is available.
      else
      {
        pmv[0] = 0;
        pmv[1] = 0;
      }
    }
  }

  // B frame mv points to LIST_1
  else
  {
    // If colocated mv points to LIST_0
    if(listX[LIST_1][0]->ref_idx[LIST_0][y*4+block_y][x*4+block_x] != -1)
    {
      colocated_MV_x = listX[LIST_1][0]->mv[LIST_0][y*4+block_y][x*4+block_x][0];
      colocated_MV_y = listX[LIST_1][0]->mv[LIST_0][y*4+block_y][x*4+block_x][1];

      ref_frame_coloc = listX[LIST_1][0]->ref_idx[LIST_0][y*4+block_y][x*4+block_x];

      // Scaling
      pmv[0] = (img->mvscale_hb[LIST_1][LIST_0][ref_frame][ref_frame_coloc] * colocated_MV_x + 128) >> 8;
      pmv[1] = (img->mvscale_hb[LIST_1][LIST_0][ref_frame][ref_frame_coloc] * colocated_MV_y + 128) >> 8;
    }

    else
    {
      // If colocated mv points to LIST_1
      if(listX[LIST_1][0]->ref_idx[LIST_1][y*4+block_y][x*4+block_x] != -1)
      {
        colocated_MV_x = listX[LIST_1][0]->mv[LIST_1][y*4+block_y][x*4+block_x][0];
        colocated_MV_y = listX[LIST_1][0]->mv[LIST_1][y*4+block_y][x*4+block_x][1];

        ref_frame_coloc = listX[LIST_1][0]->ref_idx[LIST_1][y*4+block_y][x*4+block_x];

        // Scaling
        pmv[0] = (img->mvscale_hb[LIST_1][LIST_1][ref_frame][ref_frame_coloc] * colocated_MV_x + 128) >> 8;
        pmv[1] = (img->mvscale_hb[LIST_1][LIST_1][ref_frame][ref_frame_coloc] * colocated_MV_y + 128) >> 8;
      }

      // No colocated mv is available.
      else
      {
        pmv[0] = 0;
        pmv[1] = 0;
      }
    }
  }
}

#endif
