/* 
====================================================================================================================

The copyright in this software is being made available under the License included below. 
This software may be subject to other third party and 	contributor rights, including patent rights, and no such 
rights are granted under this license.

Copyright (c) 2010, QUALCOMM INC.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted only for 
the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
promoting such standards. The following conditions are required to be met:

* Redistributions of source code must retain the above copyright notice, this list of conditions and 
the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
the following disclaimer in the documentation and/or other materials provided with the distribution.
* The name of QUALCOMM INC.  may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

====================================================================================================================
*/
#include <stdlib.h>
#include <string.h>

#include "global.h"
#include "memalloc.h"

#define MAX_ITEMS_TO_PARSE  1000

extern const int quant_coef[6][4][4];
extern const int dequant_coef[6][4][4];

extern const int quant_coef8[6][8][8];
extern const int dequant_coef8[6][8][8];


int matrix4x4_check[6] = {0, 0, 0, 0, 0, 0};
int matrix8x8_check[2] = {0, 0};

static const char MatrixType4x4[6][20] =
{
  "INTRA4X4_LUMA",
  "INTRA4X4_CHROMAU",
  "INTRA4X4_CHROMAV",
  "INTER4X4_LUMA",
  "INTER4X4_CHROMAU",
  "INTER4X4_CHROMAV"
};

static const char MatrixType8x8[2][20] =
{
  "INTRA8X8_LUMA",
  "INTER8X8_LUMA",
};

int ****LevelScale4x4Luma;
int *****LevelScale4x4Chroma;
int ****LevelScale8x8Luma;

int ****InvLevelScale4x4Luma;
int *****InvLevelScale4x4Chroma;
int ****InvLevelScale8x8Luma;

int *qp_per_matrix;
int *qp_rem_matrix;

static const short Quant_intra_default[16] =
{
  6,13,20,28,
  13,20,28,32,
  20,28,32,37,
  28,32,37,42
};

static const short Quant_inter_default[16] =
{
  10,14,20,24,
  14,20,24,27,
  20,24,27,30,
  24,27,30,34
};

static const short Quant8_intra_default[64] =
{
  6,10,13,16,18,23,25,27,
  10,11,16,18,23,25,27,29,
  13,16,18,23,25,27,29,31,
  16,18,23,25,27,29,31,33,
  18,23,25,27,29,31,33,36,
  23,25,27,29,31,33,36,38,
  25,27,29,31,33,36,38,40,
  27,29,31,33,36,38,40,42
};

static const short Quant8_inter_default[64] =
{
  9,13,15,17,19,21,22,24,
  13,13,17,19,21,22,24,25,
  15,17,19,21,22,24,25,27,
  17,19,21,22,24,25,27,28,
  19,21,22,24,25,27,28,30,
  21,22,24,25,27,28,30,32,
  22,24,25,27,28,30,32,33,
  24,25,27,28,30,32,33,35
};

/*!
***********************************************************************
* \brief
*    Allocate Q matrix arrays
***********************************************************************
*/
void allocate_QMatrix ()
{
  int bitdepth_qp_scale = 6*(input->BitDepthLuma - 8);
  int i;

  if ((qp_per_matrix = (int*)malloc((MAX_QP + 1 +  bitdepth_qp_scale)*sizeof(int))) == NULL)
    no_mem_exit("init_global_buffers: qp_per_matrix");
  if ((qp_rem_matrix = (int*)malloc((MAX_QP + 1 +  bitdepth_qp_scale)*sizeof(int))) == NULL)
    no_mem_exit("init_global_buffers: qp_per_matrix");

  for (i = 0; i < MAX_QP + bitdepth_qp_scale + 1; i++)
  {
    qp_per_matrix[i] = i / 6;
    qp_rem_matrix[i] = i % 6;
  }

  get_mem4Dint(&LevelScale4x4Luma,      2, 6, 4, 4);
  get_mem5Dint(&LevelScale4x4Chroma, 2, 2, 6, 4, 4);
  get_mem4Dint(&LevelScale8x8Luma,      2, 6, 8, 8);

  get_mem4Dint(&InvLevelScale4x4Luma,      2, 6, 4, 4);
  get_mem5Dint(&InvLevelScale4x4Chroma, 2, 2, 6, 4, 4);
  get_mem4Dint(&InvLevelScale8x8Luma,      2, 6, 8, 8);
}

/*!
***********************************************************************
* \brief
*    Free Q matrix arrays
***********************************************************************
*/
void free_QMatrix ()
{
  free(qp_rem_matrix);
  free(qp_per_matrix);

  free_mem4Dint(LevelScale4x4Luma,      2, 6);
  free_mem5Dint(LevelScale4x4Chroma, 2, 2, 6);
  free_mem4Dint(LevelScale8x8Luma,      2, 6);

  free_mem4Dint(InvLevelScale4x4Luma,      2, 6);
  free_mem5Dint(InvLevelScale4x4Chroma, 2, 2, 6);
  free_mem4Dint(InvLevelScale8x8Luma,      2, 6);

}


/*!
***********************************************************************
* \brief
*    Initialise Q matrix values.
***********************************************************************
*/
void Init_QMatrix (void)
{
  allocate_QMatrix ();
}

/*!
************************************************************************
* \brief
*    For calculating the quantisation values at frame level
*
* \par Input:
*    none
*
* \par Output:
*    none
************************************************************************
*/
void CalculateQuantParam(void)
{
  int i, j, k;

  for(k=0; k<6; k++)
    for(j=0; j<4; j++)
      for(i=0; i<4; i++)
      {
        LevelScale4x4Luma[1][k][j][i]         = quant_coef[k][j][i];
        InvLevelScale4x4Luma[1][k][j][i]      = dequant_coef[k][j][i]<<4;

        LevelScale4x4Chroma[0][1][k][j][i]    = quant_coef[k][j][i];
        InvLevelScale4x4Chroma[0][1][k][j][i] = dequant_coef[k][j][i]<<4;

        LevelScale4x4Chroma[1][1][k][j][i]    = quant_coef[k][j][i];
        InvLevelScale4x4Chroma[1][1][k][j][i] = dequant_coef[k][j][i]<<4;

        // Inter
        LevelScale4x4Luma[0][k][j][i]         = quant_coef[k][j][i];
        InvLevelScale4x4Luma[0][k][j][i]      = dequant_coef[k][j][i]<<4;

        LevelScale4x4Chroma[0][0][k][j][i]    = quant_coef[k][j][i];
        InvLevelScale4x4Chroma[0][0][k][j][i] = dequant_coef[k][j][i]<<4;

        LevelScale4x4Chroma[1][0][k][j][i]    = quant_coef[k][j][i];
        InvLevelScale4x4Chroma[1][0][k][j][i] = dequant_coef[k][j][i]<<4;
      }


}

/*!
************************************************************************
* \brief
*    Calculate the quantisation and inverse quantisation parameters
*
************************************************************************
*/
void CalculateQuant8Param()
{
  int i, j, k;

  for(k=0; k<6; k++)
    for(j=0; j<8; j++)
      for(i=0; i<8; i++)
      {
        LevelScale8x8Luma[1][k][j][i]         = quant_coef8[k][j][i];
        InvLevelScale8x8Luma[1][k][j][i]      = dequant_coef8[k][j][i]<<4;

        LevelScale8x8Luma[0][k][j][i]         = quant_coef8[k][j][i];
        InvLevelScale8x8Luma[0][k][j][i]      = dequant_coef8[k][j][i]<<4;
      }
}

static const char OffsetType4x4[15][24] = {
  "INTRA4X4_LUMA_INTRA",
  "INTRA4X4_CHROMAU_INTRA",
  "INTRA4X4_CHROMAV_INTRA",
  "INTRA4X4_LUMA_INTERP",
  "INTRA4X4_CHROMAU_INTERP",
  "INTRA4X4_CHROMAV_INTERP",
  "INTRA4X4_LUMA_INTERB",
  "INTRA4X4_CHROMAU_INTERB",
  "INTRA4X4_CHROMAV_INTERB",
  "INTER4X4_LUMA_INTERP",
  "INTER4X4_CHROMAU_INTERP",
  "INTER4X4_CHROMAV_INTERP",
  "INTER4X4_LUMA_INTERB",
  "INTER4X4_CHROMAU_INTERB",
  "INTER4X4_CHROMAV_INTERB"
};

static const char OffsetType8x8[5][24] = {
  "INTRA8X8_LUMA_INTRA",
  "INTRA8X8_LUMA_INTERP",
  "INTRA8X8_LUMA_INTERB",
  "INTER8X8_LUMA_INTERP",
  "INTER8X8_LUMA_INTERB"
};


int ****LevelOffset4x4Luma;
int *****LevelOffset4x4Chroma;
int ****LevelOffset8x8Luma;


short OffsetList4x4[15][16];
short OffsetList8x8[5][64];

void InitOffsetParam ();

const int OffsetBits = 11;

static const short Offset_intra_default_intra[16] = {
  682, 682, 682, 682,
  682, 682, 682, 682,
  682, 682, 682, 682,
  682, 682, 682, 682
};

static const short Offset_intra_default_chroma[16] = {
  682, 682, 682, 682,
  682, 682, 682, 682,
  682, 682, 682, 682,
  682, 682, 682, 682
};


static const short Offset_intra_default_inter[16] = {
  342, 342, 342, 342,
  342, 342, 342, 342,
  342, 342, 342, 342,
  342, 342, 342, 342,
};

static const short Offset_inter_default[16] = {
  342, 342, 342, 342,
  342, 342, 342, 342,
  342, 342, 342, 342,
  342, 342, 342, 342,
};

static const short Offset8_intra_default_intra[64] = {
  682, 682, 682, 682, 682, 682, 682, 682,
  682, 682, 682, 682, 682, 682, 682, 682,
  682, 682, 682, 682, 682, 682, 682, 682,
  682, 682, 682, 682, 682, 682, 682, 682,
  682, 682, 682, 682, 682, 682, 682, 682,
  682, 682, 682, 682, 682, 682, 682, 682,
  682, 682, 682, 682, 682, 682, 682, 682,
  682, 682, 682, 682, 682, 682, 682, 682
};

static const short Offset8_intra_default_inter[64] = {
  342, 342, 342, 342, 342, 342, 342, 342,
  342, 342, 342, 342, 342, 342, 342, 342,
  342, 342, 342, 342, 342, 342, 342, 342,
  342, 342, 342, 342, 342, 342, 342, 342,
  342, 342, 342, 342, 342, 342, 342, 342,
  342, 342, 342, 342, 342, 342, 342, 342,
  342, 342, 342, 342, 342, 342, 342, 342,
  342, 342, 342, 342, 342, 342, 342, 342
};

static const short Offset8_inter_default[64] = {
  342, 342, 342, 342, 342, 342, 342, 342,
  342, 342, 342, 342, 342, 342, 342, 342,
  342, 342, 342, 342, 342, 342, 342, 342,
  342, 342, 342, 342, 342, 342, 342, 342,
  342, 342, 342, 342, 342, 342, 342, 342,
  342, 342, 342, 342, 342, 342, 342, 342,
  342, 342, 342, 342, 342, 342, 342, 342,
  342, 342, 342, 342, 342, 342, 342, 342
};

/*!
***********************************************************************
* \brief
*    Allocate Q matrix arrays
***********************************************************************
*/
void allocate_QOffsets ()
{
  int max_qp_per_luma = (3 + 6*(input->BitDepthLuma) - MIN_QP)/6 + 1;
  int max_qp_per_cr = (3 + 6*(input->BitDepthChroma) - MIN_QP)/6 + 1;
  int max_qp_per = max(max_qp_per_luma,max_qp_per_cr);
  get_mem4Dint(&LevelOffset4x4Luma,      2, max_qp_per, 4, 4);
  get_mem5Dint(&LevelOffset4x4Chroma, 2, 2, max_qp_per, 4, 4);
  get_mem4Dint(&LevelOffset8x8Luma,      2, max_qp_per, 8, 8);
}

/*!
***********************************************************************
* \brief
*    Free Q matrix arrays
***********************************************************************
*/
void free_QOffsets ()
{
  int max_qp_per_luma = (3 + 6*(input->BitDepthLuma) - MIN_QP)/6 + 1;
  int max_qp_per_cr = (3 + 6*(input->BitDepthChroma) - MIN_QP)/6 + 1;
  int max_qp_per = max(max_qp_per_luma,max_qp_per_cr);
  free_mem4Dint(LevelOffset4x4Luma,      2, max_qp_per);
  free_mem5Dint(LevelOffset4x4Chroma, 2, 2, max_qp_per);
  free_mem4Dint(LevelOffset8x8Luma,      2, max_qp_per);
}

/*!
***********************************************************************
* \brief
*    Initialise Q offset matrix values.
***********************************************************************
*/
void Init_QOffsetMatrix ()
{
  allocate_QOffsets ();
  //! Now set up all offset params. This process could be reused if we wish to re-init offsets
  InitOffsetParam ();
}

/*!
************************************************************************
* \brief
*    Intit quantization offset params 
*
* \par Input:
*    none
*
* \par Output:
*    none
************************************************************************
*/
void InitOffsetParam ()
{
  int i, j, k, temp;

  for (j = 0; j < 4; j++)
  {
    for (i = 0; i < 4; i++)
    {
      temp = (i << 2) + j;

      for (k = 0; k < 3; k++)
        OffsetList4x4[k][temp] = Offset_intra_default_intra[temp];
      for (k = 3; k < 9; k++)
        OffsetList4x4[k][temp] = Offset_intra_default_inter[temp];
      for (k = 9; k < 15; k++)
        OffsetList4x4[k][temp] = Offset_inter_default[temp];
    }
  }     
  for (j = 0; j < 8; j++)
  {
    for (i = 0; i < 8; i++)
    {
      temp = (i << 3) + j;
      OffsetList8x8[0][temp]  = Offset8_intra_default_intra[temp];           
      OffsetList8x8[1][temp]  = Offset8_intra_default_inter[temp];
      OffsetList8x8[2][temp]  = Offset8_intra_default_inter[temp];
      OffsetList8x8[3][temp] = Offset8_inter_default[temp];
      OffsetList8x8[4][temp] = Offset8_inter_default[temp];
    }
  }
}


/*!
************************************************************************
* \brief
*    Calculation of the quantization offset params at the frame level
*
* \par Input:
*    none
*
* \par Output:
*    none
************************************************************************
*/
void CalculateOffsetParam ()
{
  int i, j, k, temp;
  int qp_per;
  int img_type = img->type;

  int max_qp_per_luma = (51 + img->bitdepth_luma_qp_scale - MIN_QP)/6 + 1;
  int max_qp_per_cr = (51 + img->bitdepth_chroma_qp_scale - MIN_QP)/6 + 1;

  for (k = 0; k < max(max_qp_per_luma,max_qp_per_cr); k++)
  {
    qp_per = Q_BITS + k - OffsetBits;
    for (j = 0; j < 4; j++)
    {
      for (i = 0; i < 4; i++)
      {
        temp = (i << 2) + j;

        if (img_type == I_SLICE)
        {
          LevelOffset4x4Luma[1][k][j][i] =
            (int) OffsetList4x4[0][temp] << qp_per;
          LevelOffset4x4Chroma[0][1][k][j][i] =
            (int) OffsetList4x4[1][temp] << qp_per;
          LevelOffset4x4Chroma[1][1][k][j][i] =
            (int) OffsetList4x4[2][temp] << qp_per;
        }
        else if (img_type == B_SLICE)
        {
          LevelOffset4x4Luma[1][k][j][i] =
            (int) OffsetList4x4[6][temp] << qp_per;
          LevelOffset4x4Chroma[0][1][k][j][i] =
            (int) OffsetList4x4[7][temp] << qp_per;
          LevelOffset4x4Chroma[1][1][k][j][i] =
            (int) OffsetList4x4[8][temp] << qp_per;
        }
        else
        {
          LevelOffset4x4Luma[1][k][j][i] =
            (int) OffsetList4x4[3][temp] << qp_per;
          LevelOffset4x4Chroma[0][1][k][j][i] =
            (int) OffsetList4x4[4][temp] << qp_per;
          LevelOffset4x4Chroma[1][1][k][j][i] =
            (int) OffsetList4x4[5][temp] << qp_per;
        }

        if (img_type == B_SLICE)
        {
          LevelOffset4x4Luma[0][k][j][i] =
            (int) OffsetList4x4[12][temp] << qp_per;
          LevelOffset4x4Chroma[0][0][k][j][i] =
            (int) OffsetList4x4[13][temp] << qp_per;
          LevelOffset4x4Chroma[1][0][k][j][i] =
            (int) OffsetList4x4[14][temp] << qp_per;
        }
        else
        {
          LevelOffset4x4Luma[0][k][j][i] =
            (int) OffsetList4x4[9][temp] << qp_per;
          LevelOffset4x4Chroma[0][0][k][j][i] =
            (int) OffsetList4x4[10][temp] << qp_per;
          LevelOffset4x4Chroma[1][0][k][j][i] =
            (int) OffsetList4x4[11][temp] << qp_per;
        }
      }
    }
  }
}

/*!
************************************************************************
* \brief
*    Calculate the quantisation offset parameters
*
************************************************************************
*/
void CalculateOffset8Param ()
{
  int i, j, k, temp;
  int q_bits;

  int max_qp_per_luma = (51 + img->bitdepth_luma_qp_scale - MIN_QP)/6 + 1;
  int max_qp_per_cr = (51 + img->bitdepth_chroma_qp_scale - MIN_QP)/6 + 1;

  for (k = 0; k < max(max_qp_per_luma,max_qp_per_cr); k++)
  {
    q_bits = Q_BITS_8 + k - OffsetBits;
    for (j = 0; j < 8; j++)
    {
      for (i = 0; i < 8; i++)
      {           
        temp = (i << 3) + j;

        if (img->type == I_SLICE)
          LevelOffset8x8Luma[1][k][j][i] =
          (int) OffsetList8x8[0][temp] << q_bits;
        else if (img->type == B_SLICE)
          LevelOffset8x8Luma[1][k][j][i] =
          (int) OffsetList8x8[2][temp] << q_bits;
        else
          LevelOffset8x8Luma[1][k][j][i] =
          (int) OffsetList8x8[1][temp] << q_bits;

        if (img->type == B_SLICE)
          LevelOffset8x8Luma[0][k][j][i] =
          (int) OffsetList8x8[4][temp] << q_bits;
        else
          LevelOffset8x8Luma[0][k][j][i] =
          (int) OffsetList8x8[3][temp] << q_bits;
      }
    }
  }
}
