/* 
====================================================================================================================

The copyright in this software is being made available under the License included below. 
This software may be subject to other third party and 	contributor rights, including patent rights, and no such 
rights are granted under this license.

Copyright (c) 2010, QUALCOMM INC.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted only for 
the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
promoting such standards. The following conditions are required to be met:

* Redistributions of source code must retain the above copyright notice, this list of conditions and 
the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
the following disclaimer in the documentation and/or other materials provided with the distribution.
* The name of QUALCOMM INC.  may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

====================================================================================================================
*/

#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <float.h>

#include "global.h"
#include "header.h"
#include "mb_access.h"
#include "vlc.h"
#include "image.h"
#include "cabac.h"
#include "elements.h"
#include "epzs.h"

#include <limits.h>

#ifdef MV_COMPETITION
#include "mv_competition.h"
#endif

#ifdef SWITCHED_FILTERS
#include "switched_filters.h"
#endif  // SWITCHED_FILTERS

#ifdef RDO_Q
double encode_one_macroblock_each_quant(int CurrentMbAddr);
#endif

#ifdef USE_INTRA_MDDT
#include <memory.h>
static void InitScanOrderForSlice();
extern void precompute_all_inner_product8x8();
extern void precompute_all_inner_product16x16();
#endif 

// Local declarations
static Slice *malloc_slice();
static void  free_slice(Slice *slice);
static void  init_slice(int start_mb_addr);
static void set_ref_pic_num();
extern ColocatedParams *Co_located;
extern StorablePicture **listX[6];
void poc_ref_pic_reorder(StorablePicture **list, unsigned num_ref_idx_lX_active, int *reordering_of_pic_nums_idc, int *abs_diff_pic_num_minus1, int *long_term_pic_idx, int weighted_prediction, int list_no);
void SetLagrangianMultipliers();

#ifdef RDO_Q
extern short best_mode;
extern int curr_mbi, curr_mb32i;
#endif

#ifdef GEOM
extern int Segmentation_Selected;
extern int geo_stat[2][3];
extern char ****geo_mask[3];//[mb_ext_level][rho][theta][y][x]
int cluster_counter;
void copy_partition_candidate(int rd_pass);
void set_context_geo();
#endif

int mb32_16x16_cbp[4]; //each 16x16 block cbp of 32x32 block
int mb32_mvd[2][BLOCK_MULTIPLE<<MAX_MB_EXT_LEVEL][BLOCK_MULTIPLE<<MAX_MB_EXT_LEVEL][2];          //!< indices correspond to [forw,backw][block_y][block_x][x,y]
extern int bestmode32;
extern Macroblock  MB32, MB32_16x16;
extern int   cbp16[1<<(2*MAX_MB_EXT_LEVEL)];
extern int64 cbp_blk16[1<<(2*MAX_MB_EXT_LEVEL)];
extern int   luma_transform_size_8x8_flag32[1<<(2*MAX_MB_EXT_LEVEL)];
extern int   ****cofAC32[1<<(2*MAX_MB_EXT_LEVEL)];     // [8x8block][4x4block][level/run][scan_pos]
extern int   ***cofDC32[1<<(2*MAX_MB_EXT_LEVEL)];                       // [yuv][level/run][scan_pos]
extern double encode_macroblock_cluster32(int CurrentMbAddr, int mb32count);
void start_macroblock32(int mb_addr, int mb32);
double encode_one_macroblock_Big (int level, int Re_Encode);
int write_one_macroblock32 (int eos_bit, int mb32, int mbi);
int mb32_stat[2][5];
int mb32_stat_currentpass[2][3][5];
int delta_qp_sent;
int eos_bit_32=1;
extern Macroblock  MB64;
int mb64_stat[2][5];
int mb64_stat_currentpass[2][3][5];
extern int   cbp_32[1<<(2*(MAX_MB_EXT_LEVEL-1))];
extern int bestmode64;
extern int mb32type[1<<(2*(MAX_MB_EXT_LEVEL-1))];
int write_one_macroblock64 (int eos_bit, int mb32i, int mb16i);
extern double encode_macroblock_cluster64(int CurrentMbAddr);
extern int    **cof32AC16x16[1<<(2*MAX_MB_EXT_LEVEL)];

void find_Mb_cluster(int level, int CurrentMbAddr, int *CurrentMbAddr_x, int *CurrentMbAddr_y, int *extend_x, int *extend_y, int *num_MB, int *nextMbAddr);
void propagate_qp_info(int CurrentMbAddr, int CurrentMbAddr_x, int CurrentMbAddr_y, int extend_x, int extend_y);
void prepare_data_for_write(int encodeMbAddr, Macroblock *MB, int mbcount, int mode);
void propagate_mb_info(int mb_ext_level, Macroblock* currMB32, int CurrentMbAddr, int CurrentMbAddr_x, int CurrentMbAddr_y, int extend_x, int extend_y);
int encode_one_macroblock_geo (double min_rdcost0, int mb_ext_level, int do_Re_Encode);//GEOM_RESTURCTURE

int write_macroblock_cluster32(int *NumberOfCodedMBs, int CurrentMbAddr, int mb32count)
{
  static int first_time=1;
  int no_bits=0;
  int i, j;
  //  int initialMbAddr = CurrentMbAddr;
  int encodeMbAddr;
  int CurrentMbAddr_x, CurrentMbAddr_y, extend_x, extend_y, num_MB, nextMbAddr;
#ifdef GEOM//GEOM_RESTURCTURE
  double min_rdcost=1e30;//GEOM_RESTURCTURE
#endif
  CalculateQuantParam();
  CalculateOffsetParam();

  if(input->Transform8x8Mode)
  {
    CalculateQuant8Param();
    CalculateOffset8Param();
  }

  find_Mb_cluster(1, CurrentMbAddr, &CurrentMbAddr_x, &CurrentMbAddr_y, &extend_x, &extend_y, &num_MB, &nextMbAddr);

  //prepare 32x32 block type and 32x32 block cbp
  index32x32 = mb32count;
  bestmode32 = ExtMBdata->NextMB[mb32count]->mb_type;
  MB32.mb_type = bestmode32;
  MB32.cbp=0;

  //collect statistics
  if(first_time==1)
  {
    memset(mb32_stat[0], 0, sizeof(int)*5);
    memset(mb32_stat[1], 0, sizeof(int)*5);
    first_time++;
  }
  if(num_MB==4)
  {
    mb32_stat_currentpass[img->type][img->curr_rd_pass][bestmode32<P8x8?bestmode32:4]++;
  }


  //32x32, 32x16, 16x32
  if(num_MB == 4 && bestmode32 != P8x8)
  {      
    int mbcount;
    final_mb_encoding_bestmode = bestmode32;
    start_macroblock32 (CurrentMbAddr, 1);
#ifdef GEOM
    curr_mb32i = mb32count;
#endif

    //during re-encoding 16x16 cluster, we only do best qp and best mode....
    finalbestQP = ExtMBdata->NextMB[mb32count]->qp;
#ifdef GEOM//GEOM_RESTURCTURE
    if(ExtMBdata->NextMB[mb32count]->geo_mode != -1)
      encode_one_macroblock_geo (min_rdcost, 1, do_Re_Encode);  // it's 64x64 block here
    else
#endif
      encode_one_macroblock_Big (1, do_Re_Encode);

    bestmode32=img->mb_data[img->current_mb_nr].mb_type; 
    propagate_qp_info( CurrentMbAddr, CurrentMbAddr_x,CurrentMbAddr_y,extend_x, extend_y);

#ifdef GEOM
    MB64.geo_mode = -1;
#endif
    mbcount=0;

    memset(mb32_mvd, 0, 2*(BLOCK_MULTIPLE<<MAX_MB_EXT_LEVEL)*(BLOCK_MULTIPLE<<MAX_MB_EXT_LEVEL)*2 * sizeof(int));//mb32_mvd was assigned only on predicted direction

    for(i=CurrentMbAddr_y; i<=extend_y; i++)
    {
      for(j=CurrentMbAddr_x; j<=extend_x; j++)
      {
        int b4_x,b4_y, list, k;
        encodeMbAddr = i*img->PicWidthInMbs + j;

        start_macroblock32 (encodeMbAddr, 0);//reset img->current_mb_nr and other global variables

        /******************************************
        *for the upper left 16x16 block, set b8mode
        *b8pdir for the whole 32x32 block in order
        *to write motion info in 1 shot
        *****************************************/
        prepare_data_for_write(encodeMbAddr, &MB32, mbcount, bestmode32);

        no_bits += write_one_macroblock32(1, 1, mbcount);

        if(MB32.mb_type != 0)
        {
          for(list=0; list<2; list++)            
            for(b4_y=0; b4_y<4; b4_y++)
              for(b4_x=0; b4_x<4; b4_x++)
                for(k=0; k<2; k++)
                {                    
                  img->mb_data[encodeMbAddr].mvd[list][b4_y][b4_x][k] = mb32_mvd[list][(mbcount>>1)*4+b4_y][(mbcount&1)*4+b4_x][k];
                }
        }
        mbcount ++;
      }
    }

    /*
    * propagate information to each 16x16 of 32x32 block
    */      
    propagate_mb_info(1, &MB32, CurrentMbAddr, CurrentMbAddr_x,CurrentMbAddr_y,extend_x, extend_y);
    propagate_qp_info( CurrentMbAddr, CurrentMbAddr_x,CurrentMbAddr_y,extend_x, extend_y);

    img->currentSlice->num_mb += 4;
    (*NumberOfCodedMBs) += 4;

#ifdef GEOM
    if(MB32.mb_type == 3 && MB32.geo_mode != -1)
      geo_stat[img->type][1]++;
#endif
  }//mode 32x32, 32x16, 16x32
  else
  {
    int mbcount=0;
    //int mb32_16x16_cbp_2nd[4]; //each 16x16 block cbp of 32x32 in the 2nd pass coding
#ifdef GEOM
    MB64.geo_mode = -1;
    MB32.geo_mode = -1;
#endif      
    //16x16 mode
    for(i=CurrentMbAddr_y; i<=extend_y; i++)
    {
      for(j=CurrentMbAddr_x; j<=extend_x; j++)
      {
        /****************
        * initialization
        ****************/
        encodeMbAddr = i*img->PicWidthInMbs + j;
        start_macroblock (encodeMbAddr);          
#ifdef GEOM
        curr_mbi = mb32count*4 + mbcount;
#endif
        /*********
        * coding
        *********/

        //during re-encoding 16x16 cluster, we only do best qp and best mode....
        index16x16 = mbcount;
        final_mb_encoding_bestmode = ExtMBdata->NextMB[index32x32]->NextMB[index16x16]->mb_type;
        finalbestQP = ExtMBdata->NextMB[index32x32]->NextMB[index16x16]->qp;
#ifdef GEOM//GEOM_RESTURCTURE
        final_mb_encoding_geo_mode = ExtMBdata->NextMB[index32x32]->NextMB[index16x16]->geo_mode;
#endif
        encode_one_macroblock (do_Re_Encode);

        if(num_MB == 4)
        {            
          no_bits += write_one_macroblock32(1, 0, mbcount); //32x32 block at 16x16 mode
        }
        else
          write_one_macroblock(1);

#ifdef GEOM
        if(img->mb_data[encodeMbAddr].geo_mode != -1)
        {
          int m, n;
          Macroblock *currMB = &img->mb_data[encodeMbAddr];
          int rho = currMB->geo_mode>>5, theta_idx = currMB->geo_mode & 31;  
          for(m=0; m<16; m++)
            for(n=0; n<16; n++)
              currMB->geo_mask[m][n] = geo_mask[0][rho][theta_idx][m][n];

        }
        else
        {
          img->mb_data[encodeMbAddr].geo_mask[0][0] = -1;
        }
        if(img->mb_data[encodeMbAddr].mb_type == 3 && img->mb_data[encodeMbAddr].geo_mode != -1)
          geo_stat[img->type][0]++;
#endif

        proceed2nextMacroblock();
        img->currentSlice->num_mb++;
        (*NumberOfCodedMBs)++;
        mbcount++;
        //total_cost 
      }//j
    }//i
  }

  img->currentSlice->num_mb--;//currSlice->num_mb will be updated for the last mb of this cluster in terminate_macroblock  

  return nextMbAddr;
}

int write_macroblock_cluster64(int *NumberOfCodedMBs, int CurrentMbAddr)
{
  static int first_time=1;
  int no_bits=0;
  int i, j, ii, jj;
  int encodeMbAddr;
  int CurrentMbAddr_x, CurrentMbAddr_y, extend_x, extend_y, num_MB, nextMbAddr;
#ifdef GEOM//GEOM_RESTURCTURE
  double min_rdcost=1e30;//GEOM_RESTURCTURE
#endif
  find_Mb_cluster(2, CurrentMbAddr, &CurrentMbAddr_x, &CurrentMbAddr_y, &extend_x, &extend_y, &num_MB, &nextMbAddr);

  bestmode64 = ExtMBdata->mb_type;
  MB64.mb_type = bestmode64;

  //collect statistics
  if(first_time==1)
  {
    memset(mb64_stat[0], 0, sizeof(int)*5);
    memset(mb64_stat[1], 0, sizeof(int)*5);
    first_time++;
  }
  if(num_MB==16)
  {
    mb64_stat_currentpass[img->type][img->curr_rd_pass][bestmode64<P8x8?bestmode64:4]++;
  }

  //64x64, 64x32, 32x64
  if(num_MB == 16 && bestmode64 != P8x8)
  {      
    int mb32count,mb16count;
    final_mb_encoding_bestmode = bestmode64;
    start_macroblock32 (CurrentMbAddr, 2);
    finalbestQP = ExtMBdata->qp;
#ifdef GEOM//GEOM_RESTURCTURE
    if(ExtMBdata->geo_mode != -1)
      encode_one_macroblock_geo (min_rdcost, 2, do_Re_Encode);  // it's 64x64 block here
    else
#endif
      encode_one_macroblock_Big (2 , do_Re_Encode);  // it's 64x64 block here

    propagate_qp_info( CurrentMbAddr, CurrentMbAddr_x,CurrentMbAddr_y,extend_x, extend_y);

    memset(mb32_mvd, 0, 2*(BLOCK_MULTIPLE<<MAX_MB_EXT_LEVEL)*(BLOCK_MULTIPLE<<MAX_MB_EXT_LEVEL)*2 * sizeof(int));//mb32_mvd was assigned only on predicted direction

    mb32count=0;
    for(i=CurrentMbAddr_y; i<=extend_y; i+=2)
    {
      for(j=CurrentMbAddr_x; j<=extend_x; j+=2)
      {
        MB32.cbp = cbp_32[mb32count];
        mb16count=0;
        for(ii=0; ii<2; ii++)
        {
          for(jj=0; jj<2; jj++)
          {
            int b4_x,b4_y, list, k;
            encodeMbAddr = (i+ii)*img->PicWidthInMbs + (j+jj);

            start_macroblock32 (encodeMbAddr, 0);//reset img->current_mb_nr and other global variables

            /******************************************
            *for the upper left 16x16 block, set b8mode
            *b8pdir for the whole 32x32 block in order
            *to write motion info in 1 shot
            *****************************************/               
            prepare_data_for_write(encodeMbAddr, &MB64, mb32count*4+mb16count, bestmode64);

            no_bits += write_one_macroblock64(1, mb32count, mb16count);

            if(MB64.mb_type != 0)
            {
              int mb_y = (mb32count>>1)*2 + (mb16count>>1);
              int mb_x = (mb32count&1) *2 + (mb16count&1);
              for(list=0; list<2; list++)            
                for(b4_y=0; b4_y<4; b4_y++)
                  for(b4_x=0; b4_x<4; b4_x++)
                    for(k=0; k<2; k++)
                    {                    
                      img->mb_data[encodeMbAddr].mvd[list][b4_y][b4_x][k] = mb32_mvd[list][mb_y*4+b4_y][mb_x*4+b4_x][k];
                    }
            }

            mb16count ++;
          }
        }
        mb32count++;
      }
    }
    /*
    * propagate information to each 16x16 of 32x32 block
    */      
    propagate_mb_info(2, &MB64, CurrentMbAddr, CurrentMbAddr_x,CurrentMbAddr_y,extend_x, extend_y);
    propagate_qp_info( CurrentMbAddr, CurrentMbAddr_x,CurrentMbAddr_y,extend_x, extend_y);

    img->currentSlice->num_mb += 16;
    (*NumberOfCodedMBs) += 16;

#ifdef GEOM
    if(MB64.mb_type == 3 && MB64.geo_mode != -1)
      geo_stat[img->type][2]++;
#endif

  }//mode 64x64, 64x32, 32x64 end
  else 
  {
    int mb32count=0;

    //32x32 mode
    for(i=CurrentMbAddr_y; i<=extend_y; i+=2)
    {
      for(j=CurrentMbAddr_x; j<=extend_x; j+=2)
      {
        /****************
        * initialization
        ****************/
        encodeMbAddr = i*img->PicWidthInMbs + j;
        start_macroblock (encodeMbAddr);

        //if(i+1<=extend_y && j+1<=extend_x)
        //  bestmode32 = mb32type[mb32count];
        bestmode32 = ExtMBdata->NextMB[mb32count]->mb_type;
        /*************
        * coding
        *************/
        if(num_MB==16 && mb32count == 0)
          MB64.write_mb64_mbtype = 1;

        if(num_MB==16 && mb32count != 0)
          eos_bit_32 = 0;

        write_macroblock_cluster32(NumberOfCodedMBs, encodeMbAddr, mb32count);

        if(num_MB==16 && mb32count != 0)
          eos_bit_32 = 1;

        img->currentSlice->num_mb ++; //inside write_macroblock_cluster32, currSlice->num_mb--;

        MB64.write_mb64_mbtype = 0;

        mb32count++;
        //total_cost 
      }//j
    }//i

  }


  img->currentSlice->num_mb--;//currSlice->num_mb will be updated for the last mb of this cluster in terminate_macroblock  

  return nextMbAddr;
}

int write_macroblock_cluster(int *NumberOfCodedMBs, int CurrentMbAddr)
{
  if(img->type == I_SLICE)
  {
    //start_macroblock (CurrentMbAddr);
    //finalbestQP = saveBestMode.MBsize16_qp[0];
    //encode_one_macroblock ();

    write_one_macroblock(1);

    (*NumberOfCodedMBs)++;
    proceed2nextMacroblock();

    if(CurrentMbAddr+1 == img->PicSizeInMbs)
      return -1;
    else
      return (CurrentMbAddr+1);
  }
  else
  {
    int nextMbAddr;

#ifdef GEOM_SPEEDUP5
    Motion_Selected = 0;
#endif

    nextMbAddr = write_macroblock_cluster64(NumberOfCodedMBs, CurrentMbAddr);

    return nextMbAddr;
  } 
}
/*!
************************************************************************
* \brief
*    init_ref_pic_list_reordering initializations should go here
************************************************************************
*/
void init_ref_pic_list_reordering()
{
  Slice* currSlice = img->currentSlice;

  currSlice->ref_pic_list_reordering_flag_l0 = 0;
  currSlice->ref_pic_list_reordering_flag_l1 = 0;
}


/*!
************************************************************************
*  \brief
*     This function generates the slice (and partition) header(s) 
*
*  \return number of bits used for the slice (and partition) header(s)
*
*  \par Side effects:
*      Adds slice/partition header symbols to the symbol buffer
*      increments Picture->no_slices, allocates memory for the
*      slice, sets img->currSlice
************************************************************************
*/
int start_slice()
{
  EncodingEnvironmentPtr eep;
  Slice *currSlice = img->currentSlice;
  Bitstream *currStream;
  int header_len = 0;

  currStream = (currSlice->partArr[0]).bitstream;
  currStream->write_flag = 0;
  header_len += SliceHeader (0);

  //! Initialize CABAC
  eep = &((currSlice->partArr[0]).ee_cabac);
  if (currStream->bits_to_go != 8)
    header_len+=currStream->bits_to_go;
  writeVlcByteAlign(currStream);
  arienco_start_encoding(eep, currStream->streamBuffer, &(currStream->byte_pos));
  cabac_new_slice();

  init_contexts();

  return header_len;
}



/*!
************************************************************************
* \brief
*    This function terminates a slice (but doesn't write it out), 
*    the old terminate_slice (0)
* \return
*    0 if OK,                                                         \n
*    1 in case of error
*
************************************************************************
*/
int terminate_slice(int lastslice, Picture *pic)
{
  static int MbWidthC  [4]= { 0, 8, 8,  16};
  static int MbHeightC [4]= { 0, 8, 16, 16};

  int bytes_written;
  Bitstream *currStream;
  Slice *currSlice = img->currentSlice;
  EncodingEnvironmentPtr eep;
  int i;
  int byte_pos_before_startcode_emu_prevention;
  int min_num_bytes=0;
  int stuffing_bytes=0;
  int RawMbBits;

  write_terminating_bit (1);      // only once, not for all partitions
#ifdef ADAPTIVE_LOOP_FILTER
  if(!input->UseAdaptiveLoopFilter)
  {
#endif  

    for (i=0; i<currSlice->max_part_nr; i++)
    {
      currStream = (currSlice->partArr[i]).bitstream;
      {
        eep = &((currSlice->partArr[i]).ee_cabac);
        // terminate the arithmetic code
        arienco_done_encoding(eep);
        currStream->bits_to_go = eep->Ebits_to_go;
        currStream->byte_buf = 0;
        bytes_written = currStream->byte_pos;
        img->bytes_in_picture += currStream->byte_pos;

        byte_pos_before_startcode_emu_prevention= currStream->byte_pos;
        if (lastslice && i==((currSlice->max_part_nr-1)))
        {
          RawMbBits = 256 * img->bitdepth_luma + 2 * MbWidthC[active_sps->chroma_format_idc] * MbHeightC[active_sps->chroma_format_idc] * img->bitdepth_chroma;
          min_num_bytes = ((96 * get_pic_bin_count()) - (RawMbBits * (int)img->PicSizeInMbs *3) + 1023) / 1024;
          if (min_num_bytes>img->bytes_in_picture)
          {
            stuffing_bytes = min_num_bytes - img->bytes_in_picture;
            printf ("CABAC stuffing words = %6d\n", stuffing_bytes/3);
          }
        }

        //      printf ("bytepos: %d\n", currStream->byte_pos);
        currStream->byte_pos = RBSPtoEBSP(currStream->streamBuffer, 0, currStream->byte_pos, currStream->byte_pos + stuffing_bytes);
        *(stats->em_prev_bits) += (currStream->byte_pos - byte_pos_before_startcode_emu_prevention) * 8;
      }           // CABAC
    }           // partition loop
#ifdef ADAPTIVE_LOOP_FILTER
  }
#endif

  store_contexts();
#ifdef ADAPTIVE_LOOP_FILTER
  if (input->UseAdaptiveLoopFilter)
    pic->slices[pic->no_slices-1]->model_number = img->model_number;
#endif

  if (img->type != I_SLICE || img->type != SI_SLICE)
    free_ref_pic_list_reordering_buffer (currSlice);
  return 0;   
}

/*!
************************************************************************
* \brief
*    Encodes one slice
* \par
*   returns the number of coded MBs in the SLice 
************************************************************************
*/
int encode_one_slice (int SliceGroupId, Picture *pic, int TotalCodedMBs)
{
  Boolean end_of_slice = FALSE;
  Boolean recode_macroblock;
  int len;
  int NumberOfCodedMBs = 0;
  int CurrentMbAddr;
  //  double FrameRDCost = DBL_MAX, FieldRDCost = DBL_MAX;


#ifdef RDO_Q
  int saved_CurrentMbAddr;
#endif

  img->cod_counter = 0;

#ifdef USE_INTRA_MDDT
  if(input->UseIntraMDDT)
  {
    InitScanOrderForSlice(); 
    precompute_all_inner_product8x8();
    precompute_all_inner_product16x16(); 
  }
#endif

  CurrentMbAddr = GetFirstMacroblockInSlice (SliceGroupId);
  init_slice (CurrentMbAddr);

#ifdef MV_COMPETITION
  if(input->mv_competition > 0)
  {
    enc_picture->slice_type = img->type;
    if(img->type == B_SLICE /*&& input->HierarchicalCoding!=0*/) 
      init_mv_scale_hb();
  }
#endif

  Bytes_After_Header = img->currentSlice->partArr[0].bitstream->byte_pos;
  SetLagrangianMultipliers();


#ifdef SWITCHED_FILTERS
  if(img->type != I_SLICE)
  {    
    if(input->UseHPFilter == HPF_SIFO_FPO)
    {
      if(img->filterParam == SIFO_FIRST_PASS_FPO)
      {
        if(img->type == B_SLICE)
        {
          setFirstPassSubpelOffset(LIST_0);
          setFirstPassSubpelOffset(LIST_1);
        }
        else
        {
          setFirstPassSubpelOffset(LIST_0);
        }
      }

      UnifiedOneForthPixFiltSel(img->filterParam);
      SwapFilteredFrames();
    }
  }
#endif  // SWITCHED_FILTERS

  SetCtxModelNumber ();

  len = start_slice ();

  stats->bit_slice += len;
  stats->bit_use_header[img->type] += len;

#ifdef GEOM
  cluster_counter=0;
#endif

  while (end_of_slice == FALSE) // loop over macroblocks
  {
    do_Re_Encode = noME_during_Re_Encode;
    ExtMBdata->mb_type        = -1;
    ExtMBdata->qp             = -1;
    ExtMBdata->prev_qp        = -1;
    ExtMBdata->prev_delta_qp  = -1;
    {
      int idx32,idx16;
      for (idx32=0; idx32<4;idx32++)
      {
        tmpExtMBdata= ExtMBdata->NextMB[idx32];
        tmpExtMBdata->mb_type      =-1;
        tmpExtMBdata->qp           =-1;
        tmpExtMBdata->prev_qp      =-1;
        tmpExtMBdata->prev_delta_qp=-1;
        for (idx16=0;idx16<4;idx16++)
        {
          tmpExtMBdata= ExtMBdata->NextMB[idx32]->NextMB[idx16];
          tmpExtMBdata->mb_type      =-1;
          tmpExtMBdata->qp           =-1;
          tmpExtMBdata->prev_qp      =-1;
          tmpExtMBdata->prev_delta_qp=-1;
        }
      }
    }

#ifdef GEOM
    if(img->rd_pass==0)
      Segmentation_Selected = 0;
    else
    {
      Segmentation_Selected = 1;
      copy_partition_candidate(1);
    }
#endif

    {
      recode_macroblock = FALSE;
      rdopt = &rddata_top_frame_mb;   // store data in top frame MB 

      final_mb_encoding = 0;
      reencoding_during_modedecision = 0;
      img->masterQP=img->qp;        //save master QP
      encode_one_macroblock_each_quant(CurrentMbAddr);


      final_mb_encoding = 1;
#ifdef GEOM
      Segmentation_Selected = 1;
#endif      
      reencoding_during_modedecision = 0;
      saved_CurrentMbAddr = CurrentMbAddr;
      CurrentMbAddr = write_macroblock_cluster (&NumberOfCodedMBs, CurrentMbAddr);//the return value is the next to-be-coded Mbaddr

      img->qp = img->masterQP;    //switch back the master QP....

#ifdef GEOM //arbitrary
      set_context_geo();
      if(img->rd_pass == 0)
        copy_partition_candidate(0);

      cluster_counter ++;
#endif
      terminate_macroblock (&end_of_slice, &recode_macroblock);

      if (recode_macroblock == FALSE)       // The final processing of the macroblock has been done
      {
        if (CurrentMbAddr == -1)   // end of slice
          end_of_slice = TRUE;
      }
      else
      {
        //!Go back to the previous MB to recode it
        img->current_mb_nr = GetPreviousMBNr(img->current_mb_nr);
        if(img->current_mb_nr == -1 )   // The first MB of the slice group  is too big,
          // which means it's impossible to encode picture using current slice bits restriction
        {
          snprintf (errortext, ET_SIZE, "Error encoding first MB with specified parameter, bits of current MB may be too big");
          error (errortext, 300);
        }
      }
    }
  }       //while (end_of_slice == FALSE) // loop over macroblocks


#ifdef SWITCHED_FILTERS
  if((img->type != I_SLICE) && (input->UseHPFilter == HPF_SIFO_FPO))
    SwapFilteredFrames();
#endif  // SWITCHED_FILTERS

  terminate_slice ( (NumberOfCodedMBs+TotalCodedMBs >= (int)img->PicSizeInMbs),pic );
  return NumberOfCodedMBs;
}

/*!
************************************************************************
* \brief
*    Initializes the parameters for a new slice and
*     allocates the memory for the coded slice in the Picture structure
*  \par Side effects:
*      Adds slice/partition header symbols to the symbol buffer
*      increments Picture->no_slices, allocates memory for the
*      slice, sets img->currSlice
************************************************************************
*/
static void init_slice (int start_mb_addr)
{
  int i;
  Picture *currPic = img->currentPicture;
  DataPartition *dataPart;
  Bitstream *currStream;
  Slice *currSlice;

  img->current_mb_nr = start_mb_addr;

  // Allocate new Slice in the current Picture, and set img->currentSlice
  assert (currPic != NULL);
  currPic->no_slices++;

  if (currPic->no_slices >= MAXSLICEPERPICTURE)
    error ("Too many slices per picture, increase MAXSLICEPERPICTURE in global.h.", -1);

  currPic->slices[currPic->no_slices-1] = malloc_slice();
  currSlice = currPic->slices[currPic->no_slices-1];

  img->currentSlice = currSlice;

  currSlice->picture_id = img->tr % 256;
  currSlice->qp = img->qp;
  currSlice->start_mb_nr = start_mb_addr;
  currSlice->slice_too_big = dummy_slice_too_big;

  for (i = 0; i < currSlice->max_part_nr; i++)
  {
    dataPart = &(currSlice->partArr[i]);
    dataPart->writeSyntaxElement = writeSyntaxElement_CABAC;
    currStream = dataPart->bitstream;
    currStream->bits_to_go = 8;
    currStream->byte_pos = 0;
    currStream->byte_buf = 0;
  }

  img->num_ref_idx_l0_active = active_pps->num_ref_idx_l0_active_minus1 + 1; 
  img->num_ref_idx_l1_active = active_pps->num_ref_idx_l1_active_minus1 + 1;

  // code now also considers fields. Issue whether we should account this within the appropriate input params directly
  if ((img->type == P_SLICE) && input->P_List0_refs)
  {
    img->num_ref_idx_l0_active = min(img->num_ref_idx_l0_active, input->P_List0_refs * ((img->structure !=0) + 1));
  }
  if (img->type == B_SLICE )
  {
    if (input->B_List0_refs)
    {
      img->num_ref_idx_l0_active = min(img->num_ref_idx_l0_active, input->B_List0_refs * ((img->structure !=0) + 1));
    }
    if (input->B_List1_refs)
    {
      img->num_ref_idx_l1_active = min(img->num_ref_idx_l1_active, input->B_List1_refs * ((img->structure !=0) + 1));
    }
  }
  // generate reference picture lists
  init_lists(img->type, img->structure);

  // assign list 0 size from list size
  img->num_ref_idx_l0_active = listXsize[0];
  img->num_ref_idx_l1_active = listXsize[1];

  //Perform memory management based on poc distances for HierarchicalCoding
  if (img->nal_reference_idc  && input->HierarchicalCoding && input->PocMemoryManagement && dpb.ref_frames_in_buffer==active_sps->num_ref_frames)
  {    
    poc_based_ref_management(img->frame_num);
  }
  if (input->EnableOpenGOP)
  {
    for (i = 0; i<listXsize[0]; i++)
    {    
      if (listX[0][i]->poc < img->last_valid_reference && img->ThisPOC > img->last_valid_reference)      
      {
        listXsize[0] = img->num_ref_idx_l0_active = max(1,i);
        break;
      }
    }

    for (i = 0; i<listXsize[1]; i++)
    {
      if (listX[1][i]->poc < img->last_valid_reference && img->ThisPOC > img->last_valid_reference)
      {
        listXsize[1] = img->num_ref_idx_l1_active = max(1,i);
        break;
      }
    }
  }

  init_ref_pic_list_reordering();

  //Perform reordering based on poc distances for HierarchicalCoding
  //if (img->type==P_SLICE && input->HierarchicalCoding && input->ReferenceReorder)
  if (img->type==P_SLICE && input->ReferenceReorder)
  {

    int i, num_ref;

    alloc_ref_pic_list_reordering_buffer(currSlice);

    if ((img->type != I_SLICE))
    {
      for (i=0; i<img->num_ref_idx_l0_active + 1; i++)
      {
        currSlice->reordering_of_pic_nums_idc_l0[i] = 3;
        currSlice->abs_diff_pic_num_minus1_l0[i] = 0;
        currSlice->long_term_pic_idx_l0[i] = 0;
      }

      if (img->type == B_SLICE)
      {
        for (i=0; i<img->num_ref_idx_l1_active + 1; i++)
        {
          currSlice->reordering_of_pic_nums_idc_l1[i] = 3;
          currSlice->abs_diff_pic_num_minus1_l1[i] = 0;
          currSlice->long_term_pic_idx_l1[i] = 0;
        }
      }
    }

    if ((img->type != I_SLICE))
    {
      num_ref = img->num_ref_idx_l0_active;
      poc_ref_pic_reorder(listX[LIST_0], 
        num_ref, 
        currSlice->reordering_of_pic_nums_idc_l0, 
        currSlice->abs_diff_pic_num_minus1_l0, 
        currSlice->long_term_pic_idx_l0, 0, LIST_0);

      //reference picture reordering
      reorder_ref_pic_list(listX[LIST_0], &listXsize[LIST_0], 
        img->num_ref_idx_l0_active - 1, 
        currSlice->reordering_of_pic_nums_idc_l0, 
        currSlice->abs_diff_pic_num_minus1_l0, 
        currSlice->long_term_pic_idx_l0);

      // This is not necessary since order is already poc based...  
      if (img->type == B_SLICE)
      {
        num_ref = img->num_ref_idx_l1_active;
        poc_ref_pic_reorder(listX[LIST_1], 
          num_ref, 
          currSlice->reordering_of_pic_nums_idc_l1, 
          currSlice->abs_diff_pic_num_minus1_l1, 
          currSlice->long_term_pic_idx_l1, 0, LIST_1);

        //reference picture reordering
        reorder_ref_pic_list(listX[LIST_1], &listXsize[LIST_1], 
          img->num_ref_idx_l1_active - 1, 
          currSlice->reordering_of_pic_nums_idc_l1, 
          currSlice->abs_diff_pic_num_minus1_l1, 
          currSlice->long_term_pic_idx_l1);
      }
    }
  }

  set_ref_pic_num();

  if (img->type == B_SLICE)
    compute_colocated(Co_located, listX);
  if (img->type != I_SLICE)
    EPZSSliceInit(EPZSCo_located, listX);
}


/*!
************************************************************************
* \brief
*    Allocates a slice structure along with its dependent data structures
* \return
*    Pointer to a Slice
************************************************************************
*/
static Slice *malloc_slice()
{
  int i;
  DataPartition *dataPart;
  Slice *slice;

  //  const int buffer_size = (img->width * img->height * 4); //There can be data expansion with 
  // low QP values. So, we make sure that buffer 
  // does not overflow. 4 is probably safe multiplier.
  const int buffer_size = 500 + img->FrameSizeInMbs * (128 + 256 * img->bitdepth_luma + 512 * img->bitdepth_chroma);

  if ((slice = (Slice *) calloc(1, sizeof(Slice))) == NULL) no_mem_exit ("malloc_slice: slice structure");

  // create all context models
  slice->mot_ctx = create_contexts_MotionInfo();
  slice->tex_ctx = create_contexts_TextureInfo();

  slice->max_part_nr = 1;

  //for IDR img there should be only one partition
  if(img->currentPicture->idr_flag)
    slice->max_part_nr = 1;

  assignSE2partition[0] = assignSE2partition_NoDP;
  assignSE2partition[1] = assignSE2partition_NoDP;        

  slice->num_mb = 0;          // no coded MBs so far

  if ((slice->partArr = (DataPartition *) calloc(slice->max_part_nr, sizeof(DataPartition))) == NULL) no_mem_exit ("malloc_slice: partArr");
  for (i=0; i<slice->max_part_nr; i++) // loop over all data partitions
  {
    dataPart = &(slice->partArr[i]);
    if ((dataPart->bitstream = (Bitstream *) calloc(1, sizeof(Bitstream))) == NULL) no_mem_exit ("malloc_slice: Bitstream");
    if ((dataPart->bitstream->streamBuffer = (byte *) calloc(buffer_size, sizeof(byte))) == NULL) no_mem_exit ("malloc_slice: StreamBuffer");
    // Initialize storage of bitstream parameters
  }
  return slice;
}


/*!
************************************************************************
* \brief
*    Memory frees of all Slice structures and of its dependent
*    data structures
* \par Input:
*    Image Parameters struct struct img_par *img
************************************************************************
*/
void free_slice_list(Picture *currPic)
{
  int i;

  for (i=0; i<currPic->no_slices; i++)
  {
    free_slice (currPic->slices[i]);
    currPic->slices[i]=NULL;
  }
}


/*!
************************************************************************
* \brief
*    Memory frees of the Slice structure and of its dependent
*    data structures
* \param slice:
*    Slice to be freed
************************************************************************
*/
static void free_slice(Slice *slice)
{
  int i;
  DataPartition *dataPart;

  if (slice != NULL)
  {
    for (i=0; i<slice->max_part_nr; i++) // loop over all data partitions
    {
      dataPart = &(slice->partArr[i]);
      if (dataPart != NULL)
      {
        if (dataPart->bitstream->streamBuffer != NULL)
          free(dataPart->bitstream->streamBuffer);
        if (dataPart->bitstream != NULL)
          free(dataPart->bitstream);
      }
    }
    if (slice->partArr != NULL)
      free(slice->partArr);

    delete_contexts_MotionInfo(slice->mot_ctx);
    delete_contexts_TextureInfo(slice->tex_ctx);

    free(slice);
  }
}

void set_ref_pic_num()
{
  int i;
  StorablePicture *this_ref;

  for (i=0;i<listXsize[LIST_0];i++)
  {
    this_ref = listX[LIST_0][i];
    enc_picture->ref_pic_num        [LIST_0][i] = this_ref->poc * 2; 
    enc_picture->frm_ref_pic_num    [LIST_0][i] = this_ref->frame_poc * 2; 
  }

  for (i=0;i<listXsize[LIST_1];i++)
  {
    this_ref = listX[LIST_1][i];
    enc_picture->ref_pic_num        [LIST_1][i] = this_ref->poc * 2;
    enc_picture->frm_ref_pic_num    [LIST_1][i] = this_ref->frame_poc * 2; 
  }
}

/*!
************************************************************************
* \brief
*    decide reference picture reordering, Frame only
************************************************************************
*/
void poc_ref_pic_reorder(StorablePicture **list, unsigned num_ref_idx_lX_active, int *reordering_of_pic_nums_idc, int *abs_diff_pic_num_minus1, int *long_term_pic_idx, int weighted_prediction, int list_no)
{
  unsigned i,j,k;

  int currPicNum, picNumLXPred;

  int default_order[32];
  int re_order[32];
  int tmp_reorder[32];
  int list_sign[32];
  int reorder_stop, no_reorder;
  int poc_diff[32];
  int tmp_value, diff;  

  int abs_poc_dist;
  int maxPicNum, MaxFrameNum = 1 << (log2_max_frame_num_minus4 + 4);

  maxPicNum  = MaxFrameNum;
  currPicNum = img->frame_num;

  picNumLXPred = currPicNum;

  // First assign default list order. 
  for (i=0; i<num_ref_idx_lX_active; i++)
  {
    default_order[i] = list[i]->pic_num;
  }

  // Now access all references in buffer and assign them
  // to a pottential reordering list. For each one of these 
  // references compute the poc distance compared to current
  // frame.
  for (i=0; i<dpb.ref_frames_in_buffer; i++)
  {
    re_order[i] = dpb.fs_ref[i]->frame->pic_num;

    if (dpb.fs_ref[i]->is_used==3 && (dpb.fs_ref[i]->frame->used_for_reference)&&(!dpb.fs_ref[i]->frame->is_long_term))
    {
      abs_poc_dist = abs(dpb.fs_ref[i]->frame->poc - enc_picture->poc) ;
      poc_diff[i] = abs_poc_dist;
      if (list_no == LIST_0)
      {
        list_sign[i] = (enc_picture->poc < dpb.fs_ref[i]->frame->poc) ? +1 : -1;
      }
      else
      {
        list_sign[i] = (enc_picture->poc > dpb.fs_ref[i]->frame->poc) ? +1 : -1;
      }
    }
  }


  // now sort these references based on poc (temporal) distance
  for (i=0; i< dpb.ref_frames_in_buffer-1; i++)
  {
    for (j=i+1; j< dpb.ref_frames_in_buffer; j++)
    {      
      if (poc_diff[i]>poc_diff[j] || (poc_diff[i] == poc_diff[j] && list_sign[j] > list_sign[i]))
      {

        tmp_value = poc_diff[i];
        poc_diff[i] = poc_diff[j];
        poc_diff[j] = tmp_value;
        tmp_value  = re_order[i];
        re_order[i] = re_order[j];
        re_order[j] = tmp_value ;
        tmp_value  = list_sign[i];
        list_sign[i] = list_sign[j];        
        list_sign[j] = tmp_value ;
      }
    }
  }

  // Check versus default list to see if any
  // change has happened
  no_reorder = 1;
  for (i=0; i<num_ref_idx_lX_active; i++)
  {
    if (default_order[i] != re_order[i])
    {
      no_reorder = 0;
    }
  }

  // If different, then signal reordering
  if (no_reorder==0)
  {
    for (i=0; i<num_ref_idx_lX_active; i++)
    {
      diff = re_order[i]-picNumLXPred;
      if (diff <= 0)
      {
        reordering_of_pic_nums_idc[i] = 0;
        abs_diff_pic_num_minus1[i] = abs(diff)-1;
        if (abs_diff_pic_num_minus1[i] < 0)
          abs_diff_pic_num_minus1[i] = maxPicNum -1; 
      }
      else
      {
        reordering_of_pic_nums_idc[i] = 1;
        abs_diff_pic_num_minus1[i] = abs(diff)-1;
      }
      picNumLXPred = re_order[i];

      tmp_reorder[i] = re_order[i];

      k = i;
      for (j=i; j<num_ref_idx_lX_active; j++)
      {
        if (default_order[j] != re_order[i])
        {
          ++k;
          tmp_reorder[k] = default_order[j];
        }
      }
      reorder_stop = 1;
      for(j=i+1; j<num_ref_idx_lX_active; j++)
      {
        if (tmp_reorder[j] != re_order[j])
        {
          reorder_stop = 0;
          break;
        }
      }

      if (reorder_stop==1)
      {
        ++i;
        break;
      }


      for(j=0; j<num_ref_idx_lX_active; j++)
      {
        default_order[j] = tmp_reorder[j];
      }

    }
    reordering_of_pic_nums_idc[i] = 3;

    for(j=0; j<num_ref_idx_lX_active; j++)
    {
      default_order[j] = tmp_reorder[j];
    }

    if (list_no==0)
    {
      img->currentSlice->ref_pic_list_reordering_flag_l0=1;
    }
    else
    {
      img->currentSlice->ref_pic_list_reordering_flag_l1=1;
    }
  }
}

extern int QP2QUANT[40];

void SetLagrangianMultipliers()
{
  int qp, j;
  double qp_temp;
  double lambda_scale = 1.0 - Clip3(0.0,0.5,0.05 * (double) input->jumpd);

  // RDOPT on computation of Lagrangian multipliers
  for (j = 0; j < 5; j++)
  {
    for (qp = -img->bitdepth_luma_qp_scale; qp < 52; qp++)
    {          
      qp_temp = (double)qp + img->bitdepth_luma_qp_scale - SHIFT_QP;

      if (input->UseExplicitLambdaParams) // consideration of explicit weights.
      {
        img->lambda_md[j][qp] = input->LambdaWeight[j] * pow (2, qp_temp/3.0);
        // Scale lambda due to hadamard qpel only consideration
        img->lambda_md[j][qp] = (input->hadamard == 2 ? 0.95 : 1.00) * img->lambda_md[j][qp];
        img->lambda_me[j][qp] = sqrt(img->lambda_md[j][qp]);
        img->lambda_mf[j][qp] = LAMBDA_FACTOR (img->lambda_me[j][qp]);
        if (j == B_SLICE)
        {
          img->lambda_md[5][qp] = input->LambdaWeight[5] * pow (2, qp_temp/3.0);
          img->lambda_md[5][qp] = (input->hadamard == 2 ? 0.95 : 1.00) * img->lambda_md[5][qp];
          img->lambda_me[5][qp] = sqrt(img->lambda_md[5][qp]);
          img->lambda_mf[5][qp] = LAMBDA_FACTOR (img->lambda_me[5][qp]);
        }
      }
      else
      {
#ifdef RDO_Q
        if(input->UseRDO_Q && img->type == I_SLICE && img->qp==qp)
          img->lambda_md[j][qp] = 0.57 * pow (2.0, qp_temp/3.0); 
        else
#endif
#ifdef USE_INTRA_MDDT
          if(input->UseIntraMDDT && img->type == I_SLICE && img->qp==qp)
            img->lambda_md[j][qp] = 0.57 * pow (2.0, qp_temp/3.0); 
          else
#endif
            if (input->successive_Bframe>0)
              img->lambda_md[j][qp] = 0.68 * pow (2, qp_temp/3.0) 
              * (j == B_SLICE ? Clip3(2.00,4.00,(qp_temp / 6.0)) : 1.0);
            else
              img->lambda_md[j][qp] = 0.85 * pow (2, qp_temp/3.0) 
              * ( (j == B_SLICE) ? 4.0 : 1.0);
        // Scale lambda due to hadamard qpel only consideration
        img->lambda_md[j][qp] = (input->hadamard == 2 ? 0.95 : 1.00) * img->lambda_md[j][qp];
        img->lambda_md[j][qp] = (j == B_SLICE && input->BRefPictures == 2 && img->b_frame_to_code == 0 ? 0.50 : 1.00) * img->lambda_md[j][qp];

        if (j == B_SLICE)
        {
          img->lambda_md[5][qp] = img->lambda_md[j][qp];

          if (input->HierarchicalCoding == 2)
            img->lambda_md[5][qp] *= (1.0 - min(0.4,0.2 * (double) gop_structure[img->b_frame_to_code-1].hierarchy_layer)) ;
          else
            img->lambda_md[5][qp] *= 0.80;
          img->lambda_md[5][qp] *= lambda_scale;
          img->lambda_me[5][qp] = sqrt(img->lambda_md[5][qp]);
          img->lambda_mf[5][qp] = LAMBDA_FACTOR (img->lambda_me[5][qp]);
        }
        else
          img->lambda_md[j][qp] *= lambda_scale;

        img->lambda_me[j][qp] = sqrt(img->lambda_md[j][qp]);  
        img->lambda_mf[j][qp] = LAMBDA_FACTOR (img->lambda_me[j][qp]);
      }

#ifdef  INTERNAL_BIT_DEPTH_INCREASE
      qp_temp = (double)qp + img->bitdepth_luma_qp_scale - SHIFT_QP - 6*(img->BitDepthIncrease);

      if (input->UseExplicitLambdaParams) // consideration of explicit weights.
      {
        img->lambda_md[j][qp] = input->LambdaWeight[j] * pow (2, qp_temp/3.0);
        // Scale lambda due to hadamard qpel only consideration
        img->lambda_md[j][qp] = (input->hadamard == 2 ? 0.95 : 1.00) * img->lambda_md[j][qp];
        if (j == B_SLICE)
        {
          img->lambda_md[5][qp] = input->LambdaWeight[5] * pow (2, qp_temp/3.0);
          img->lambda_md[5][qp] = (input->hadamard == 2 ? 0.95 : 1.00) * img->lambda_md[5][qp];
        }
      }
      else
      {                       
#ifdef RDO_Q
        if(input->UseRDO_Q && img->type == I_SLICE && img->qp==qp)
          img->lambda_md[j][qp] = 0.57 * pow (2.0, qp_temp/3.0); 
        else
#endif
#ifdef USE_INTRA_MDDT
          if(input->UseIntraMDDT && img->type == I_SLICE && img->qp==qp)
            img->lambda_md[j][qp] = 0.57 * pow (2.0, qp_temp/3.0); 
          else
#endif
            if (input->successive_Bframe>0)
              img->lambda_md[j][qp] = 0.68 * pow (2, qp_temp/3.0) 
              * (j == B_SLICE ? Clip3(2.00,4.00,(qp_temp / 6.0)) : 1.0);
            else
              img->lambda_md[j][qp] = 0.85 * pow (2, qp_temp/3.0) 
              * ( (j == B_SLICE) ? 4.0 : 1.0);
        // Scale lambda due to hadamard qpel only consideration
        img->lambda_md[j][qp] = (input->hadamard == 2 ? 0.95 : 1.00) * img->lambda_md[j][qp];
        img->lambda_md[j][qp] = (j == B_SLICE && input->BRefPictures == 2 && img->b_frame_to_code == 0 ? 0.50 : 1.00) * img->lambda_md[j][qp];

        if (j == B_SLICE)
        {
          img->lambda_md[5][qp] = img->lambda_md[j][qp];

          if (input->HierarchicalCoding == 2)
            img->lambda_md[5][qp] *= (1.0 - min(0.4,0.2 * (double) gop_structure[img->b_frame_to_code-1].hierarchy_layer)) ;
          else
            img->lambda_md[5][qp] *= 0.80;
          img->lambda_md[5][qp] *= lambda_scale;
        }
        else
          img->lambda_md[j][qp] *= lambda_scale;

      }
#endif
    }
  }
}

#ifdef USE_INTRA_MDDT

static const int SCANSTATS8x8[9][64] = 
{
  {85, 67, 46, 41, 32, 25, 15,  9, 44, 31, 26, 21, 15, 10,  5,  2, 27, 23, 18, 16, 10,  6,  3,  1, 18, 16, 15, 11,  8,  3,  2,  0, 13, 10, 10,  8,  6,  2,  1,  0,  7,  7,  5,  4,  3,  2,  1,  0,  4,  3,  3,  1,  1,  1,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0, },
  {79, 32, 15, 10,  7,  4,  2,  1, 56, 23, 13,  9,  6,  3,  2,  0, 38, 19, 13,  8,  4,  3,  1,  0, 34, 18, 11,  7,  4,  2,  1,  0, 36, 15,  9,  6,  4,  2,  1,  0, 26, 10,  6,  5,  3,  2,  1,  0, 17,  6,  3,  2,  2,  1,  0,  0,  6,  1,  1,  0,  0,  0,  0,  0, },
  {88, 38, 21, 12,  8,  4,  2,  1, 53, 21, 15, 10,  6,  3,  2,  1, 26, 15, 12,  9,  4,  2,  1,  0, 20, 11, 10,  6,  5,  2,  1,  0, 15, 11,  8,  6,  4,  2,  1,  0, 12,  7,  6,  5,  3,  1,  1,  0,  7,  3,  2,  3,  1,  1,  0,  0,  2,  1,  0,  1,  1,  0,  0,  0, },
  {88, 67, 41, 27, 17, 13,  5,  3, 65, 45, 33, 25, 18, 11,  6,  1, 42, 35, 31, 24, 13, 12,  4,  1, 31, 24, 27, 21, 14,  9,  5,  2, 21, 24, 18, 15, 12,  7,  5,  2, 15, 11, 10, 10,  7,  7,  2,  1,  9,  6,  4,  4,  2,  2,  1,  0,  1,  0,  0,  0,  0,  0,  0,  0, },
  {85, 69, 42, 29, 20, 12,  6,  2, 66, 53, 44, 29, 18, 10,  5,  1, 43, 46, 41, 29, 16,  7,  5,  1, 30, 30, 32, 26, 20,  9,  4,  1, 25, 21, 22, 23, 19,  9,  4,  1, 15, 14, 12, 15, 10,  7,  2,  1,  8,  6,  5,  5,  5,  3,  1,  0,  1,  1,  1,  0,  1,  0,  0,  0, },
  {89, 69, 42, 30, 22, 13,  6,  1, 63, 46, 38, 29, 22, 10,  7,  1, 38, 34, 28, 22, 15,  9,  7,  1, 25, 25, 22, 12, 11,  8,  4,  1, 21, 16, 15, 12,  7,  5,  2,  1, 13,  8,  9,  7,  3,  4,  1,  0,  4,  3,  3,  2,  2,  1,  1,  0,  1,  0,  0,  0,  0,  0,  0,  0, },
  {86, 57, 34, 20, 15,  7,  2,  0, 69, 42, 29, 19, 13,  6,  3,  0, 49, 40, 30, 19, 10,  5,  2,  0, 36, 33, 27, 18,  9,  4,  3,  1, 24, 23, 22, 13, 10,  4,  3,  0, 16, 14, 13, 12,  5,  3,  1,  0,  7,  5,  4,  4,  5,  2,  0,  0,  1,  1,  0,  1,  0,  0,  0,  0, },
  {88, 70, 51, 37, 26, 15,  7,  2, 61, 46, 40, 31, 21, 12,  7,  2, 39, 33, 27, 24, 17, 10,  4,  1, 27, 21, 23, 17, 13,  8,  4,  1, 19, 16, 16, 12, 11,  5,  3,  1, 11, 12,  9,  8,  5,  4,  2,  0,  5,  4,  4,  4,  3,  2,  1,  0,  1,  0,  0,  0,  0,  0,  0,  0, },
  {87, 57, 31, 20, 13,  7,  3,  1, 70, 44, 30, 18, 13,  6,  3,  1, 47, 40, 28, 19, 11,  5,  3,  1, 34, 32, 23, 16, 10,  5,  2,  1, 26, 26, 20, 14,  9,  3,  2,  1, 17, 15, 12, 10,  6,  2,  1,  0,  8,  8,  6,  4,  3,  1,  0,  0,  2,  1,  1,  1,  0,  0,  0,  0, },
};

static const int SCANSTATS4x4[9][16] = 
{
  {80, 66, 47, 27, 46, 35, 21,  8, 28, 23, 13,  4, 14, 11,  5,  2, },
  {75, 35, 19,  9, 66, 32, 17,  7, 60, 27, 14,  5, 42, 15,  7,  2, },
  {87, 43, 26, 12, 55, 33, 20,  8, 37, 27, 15,  6, 21, 12,  7,  2, },
  {86, 60, 36, 15, 63, 47, 31, 13, 39, 34, 24,  8, 18, 13,  9,  3, },
  {82, 54, 28,  9, 58, 53, 35,  9, 35, 41, 33, 10, 13, 15, 14,  5, },
  {85, 61, 36, 15, 56, 44, 30, 14, 33, 28, 18,  9, 15, 13,  8,  3, },
  {80, 47, 24, 11, 64, 43, 23,  9, 48, 41, 20,  9, 27, 23, 15,  4, },
  {85, 61, 39, 17, 56, 44, 31, 12, 36, 30, 19,  7, 17, 12,  8,  3, },
  {83, 44, 23,  8, 66, 41, 21,  9, 51, 40, 17,  7, 30, 24, 10,  3, },
};

static const int SCANSTATS16x16[4][256] = 
{
  {
    75,  40,  30,  26,  22,  19,  17,  14,  11,   9,   6,   4,   2,   2,   2,   1,  
      26,  16,  11,    8,   7,   6,   6,   4,   3,   2,   1,   1,   0,   0,   0,   0,
      13,   7,   5,   4,   3,   2,   2,   2,   1,   1,   0,   0,   0,   0,   0,   0,
      9,   5,   3,   2,   2,   2,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,
      7,   4,   2,   2,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      6,   3,   2,   1,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      5,   3,   2,   1,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      4,   3,   2,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      4,   3,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      3,   2,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      3,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      3,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  },
  {
    72,  24,  12,   9,   6,   5,   4,   3,   2,   2,   1,   0,   0,   0,   0,   0,
      42,  14,   7,   4,   3,   3,   3,   2,   1,   1,   0,   0,   0,   0,   0,   0,
      31,  11,   4,   2,   2,   1,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,
      25,   8,   3,   2,   1,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,
      21,   7,   3,   2,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      18,   6,   3,   1,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      16,   6,   3,   2,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      15,   6,   3,   2,   1,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,
      13,   6,   3,   2,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      11,   4,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      10,   4,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      8,   3,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      6,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      5,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      4,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
    },
    {
      79,  42,  25,  18,  13,  11,   9,   7,   5,   4,   2,   1,   1,   0,   0,   0,
        43,  25,  16,  11,   9,   8,   7,   5,   4,   3,   1,   1,   0,   0,   0,   0,
        27,  16,   9,   6,   5,   4,   4,   3,   2,   1,   1,   0,   0,   0,   0,   0,
        19,  11,   6,   4,   3,   3,   2,   2,   1,   1,   0,   0,   0,   0,   0,   0,
        15,   9,   5,   4,   3,   2,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,
        13,   8,   5,   3,   2,   2,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,
        12,   8,   5,   3,   2,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,
        10,   8,   4,   3,   2,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,
        10,   7,   4,   3,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        8,   6,   3,   2,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        7,   4,   3,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        6,   4,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        4,   3,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        3,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        3,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
    },
    {
      81,  42,  22,  14,   8,   6,   4,   3,   2,   2,   1,   0,   0,   0,   0,   0,
        43,  20,  12,   7,   5,   4,   3,   2,   2,   1,   0,   0,   0,   0,   0,   0,
        25,  12,   6,   4,   3,   2,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,
        15,   8,   4,   3,   2,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,
        11,   6,   3,   2,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        8,   5,   3,   2,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        7,   4,   2,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        6,   4,   2,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        6,   4,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        4,   3,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        4,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        3,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      },
};

/*! 
*************************************************************************************
* \brief
*   calculate the scanning order based on the input scanning stats
*
* \para calcScanOrder()
*    <paragraph>
*
* \author
*    - Yan Ye                      <yye@qualcomm.com>
*************************************************************************************
*/
static int  calcScanOrder(int *stats, int *orderX, int *orderY, int size, int width)
{
  int i, j, cOrder, cStats;
  int *stats1D = stats; 
  int order1D[256];
  int orderChanged = 0;

  for(i = 0; i < size; i++)
  {
    order1D[i] = orderX[i]+orderY[i]*width;
  }

  for (i=1; i < size; i++)
  {
    cStats = stats1D[i];
    cOrder = order1D[i];
    j = i;
    while ((j > 0) && (stats1D[j-1] < cStats) )
    {
      stats1D[j] = stats1D[j-1];
      order1D[j] = order1D[j-1];
      j = j - 1;
      orderChanged = 1;
    }
    stats1D[j] = cStats;
    order1D[j] = cOrder;
  }

  for(i = 0; i < size; i++)
  {
    orderX[i] = order1D[i]%width;
    orderY[i] = order1D[i]/width;
  }

  return orderChanged;
}

/*! 
*************************************************************************************
* \brief
*   At the beginning of each slice, initialize the scanning orders for all intra 
*   coding modes
*
* \para InitScanOrderForSlice()
*    <paragraph>
*
* \author
*    - Yan Ye                      <yye@qualcomm.com>
*************************************************************************************
*/
static void InitScanOrderForSlice()
{
  int ipredmode;
  int stats4x4[9][16];
  int stats8x8[9][64];
  int orderX4x4[16], orderY4x4[16];
  int orderX8x8[64], orderY8x8[64];
  int stats16x16[9][256];
  int orderX16x16[256], orderY16x16[256];

  for(ipredmode = 0; ipredmode < 9; ipredmode++)
  {
    int k;

    img->update4x4[ipredmode] = img->update8x8[ipredmode] = 1;
    for(k = 0; k < 16; k++)
    {
      img->scanStats4x4[ipredmode][k] = stats4x4[ipredmode][k] = SCANSTATS4x4[ipredmode][k]/2;
    }

    for(k = 0; k < 64; k++)
    {
      img->scanStats8x8[ipredmode][k] = stats8x8[ipredmode][k] = SCANSTATS8x8[ipredmode][k]/2;
    }
    img->update4x4Count[ipredmode] = img->update8x8Count[ipredmode] = 0;
    img->update4x4Thres[ipredmode] = 4;
    img->update8x8Thres[ipredmode] = 2;
  }

  for(ipredmode = 0; ipredmode < 9; ipredmode ++)
  {
    int i,dummy;

    {
      for(i = 0; i < 16; i++)
      {
        orderX4x4[i] = i%4;
        orderY4x4[i] = i/4;
      }
      dummy = calcScanOrder(stats4x4[ipredmode], orderX4x4, orderY4x4, 16, 4);
      for(i = 0; i < 16; i++)
      {
        img->scanOrder4x4[ipredmode][i][0] = orderX4x4[i];
        img->scanOrder4x4[ipredmode][i][1] = orderY4x4[i];
        img->scanStats4x4[ipredmode][i]    = stats4x4[ipredmode][i];
      }
    }

    {
      for(i = 0; i < 64; i++)
      {
        orderX8x8[i] = i%8;
        orderY8x8[i] = i/8;
      }
      dummy = calcScanOrder(stats8x8[ipredmode], orderX8x8, orderY8x8, 64, 8);    
      for(i = 0; i < 64; i++)
      {
        img->scanOrder8x8[ipredmode][i][0] = orderX8x8[i];
        img->scanOrder8x8[ipredmode][i][1] = orderY8x8[i];
        img->scanStats8x8[ipredmode][i]    = stats8x8[ipredmode][i];
      }
    }
  }

  for(ipredmode = 0; ipredmode < 4; ipredmode++)
  {
    int k, dummy;

    for(k = 0; k < 256; k++)
    {
      stats16x16[ipredmode][k] = SCANSTATS16x16[ipredmode][k]/2;
    }

    for(k = 0; k < 256; k++)
    {
      orderX16x16[k] = k%16;
      orderY16x16[k] = k/16;
    }
    dummy = calcScanOrder(stats16x16[ipredmode], orderX16x16, orderY16x16, 256, 16);
    for(k = 0;  k < 256; k++)
    {
      img->scanOrder16x16[ipredmode][k][0] = orderX16x16[k];
      img->scanOrder16x16[ipredmode][k][1] = orderY16x16[k];
    }
  }
}


/*! 
*************************************************************************************
* \brief
*   Update the scanning orders for all intra coding modes based on their coeff stats
*
* \para updateScanOrder()
*    <paragraph>
*
* \author
*    - Yan Ye                      <yye@qualcomm.com>
*************************************************************************************
*/
void updateScanOrder(int first)
{
  int ipredmode;
  int orderX[64], orderY[64];

  int orderChanged;

  for(ipredmode = 0; ipredmode < 9; ipredmode ++)
  {
    int i;

    if(img->update4x4[ipredmode])
    {
      for(i = 0; i < 16; i++)
      {
        orderX[i] = img->scanOrder4x4[ipredmode][i][0];
        orderY[i] = img->scanOrder4x4[ipredmode][i][1];
      }

      orderChanged = calcScanOrder(img->scanStats4x4[ipredmode], orderX, orderY, 16, 4);
      if(!orderChanged)
      {
        img->update4x4Thres[ipredmode] <<= 1;
      }
      else if(img->update4x4Thres[ipredmode] > 4) 
        img->update4x4Thres[ipredmode] >>= 1;
      if(orderChanged)
        for(i = 0; i < 16; i++)
        {
          img->scanOrder4x4[ipredmode][i][0] = orderX[i];
          img->scanOrder4x4[ipredmode][i][1] = orderY[i];
        }
    }
    if(img->update8x8[ipredmode])
    {
      for(i = 0; i < 64; i++)
      {
        orderX[i] = img->scanOrder8x8[ipredmode][i][0];
        orderY[i] = img->scanOrder8x8[ipredmode][i][1];
      }
      orderChanged = calcScanOrder(img->scanStats8x8[ipredmode], orderX, orderY, 64, 8);  
      if(!orderChanged)
      {
        img->update8x8Thres[ipredmode] <<= 1;
      }
      else if(img->update8x8Thres[ipredmode] > 2) 
        img->update8x8Thres[ipredmode] >>= 1;
      if(orderChanged)
        for(i = 0; i < 64; i++)
        {
          img->scanOrder8x8[ipredmode][i][0] = orderX[i];
          img->scanOrder8x8[ipredmode][i][1] = orderY[i];
        }
    }
  }
}

/*! 
*************************************************************************************
* \brief
*   Scale down the scanning stats during re-normalization
*
* \para scaleScanStats()
*    <paragraph>
*
* \author
*    - Yan Ye                      <yye@qualcomm.com>
*************************************************************************************
*/
static void scaleScanStats(int *stats, int size)
{
  int i;

  for(i = 0; i < size*size; i++)
    stats[i] >>= 1;
}

/*! 
*************************************************************************************
* \brief
*   Re-normalize the scanning stats as necessary
*
* \para normalizeScanStats()
*    <paragraph>
*
* \author
*    - Yan Ye                      <yye@qualcomm.com>
*************************************************************************************
*/
void normalizeScanStats()
{
  int ipredmode;
  int mb_type = img->mb_data[img->current_mb_nr].mb_type;

  if(mb_type == I4MB)
    for(ipredmode = 0; ipredmode < 9; ipredmode++)
    {
      if(img->scanStats4x4[ipredmode][0] >= 256)
      {
        scaleScanStats(img->scanStats4x4[ipredmode], 4);
      }
    }

  else if(mb_type == I8MB)
    for(ipredmode = 0; ipredmode < 9; ipredmode++)
    {
      if(img->scanStats8x8[ipredmode][0] >= 256)
      {
        scaleScanStats(img->scanStats8x8[ipredmode], 8);
      }
    }
}

#endif 

#ifdef RDO_Q
/*!
************************************************************************
* \brief
*    encode one macroblock for RDO_Q
*
************************************************************************
*/
double encode_one_macroblock_each_quant(int CurrentMbAddr)
{

  if(img->type == I_SLICE)
  {
    start_macroblock (CurrentMbAddr);
    encode_one_macroblock (0);
    tmpExtMBdata = ExtMBdata->NextMB[0]->NextMB[0];
    tmpExtMBdata->mb_type       = img->mb_data[CurrentMbAddr].mb_type;
    tmpExtMBdata->qp            = img->mb_data[CurrentMbAddr].qp;
    tmpExtMBdata->prev_qp       = img->mb_data[CurrentMbAddr].prev_qp;
    tmpExtMBdata->prev_delta_qp = img->mb_data[CurrentMbAddr].prev_delta_qp;
    return rdopt->min_rdcost;
  }
  else
  {
    double rd_cost;
    int prev_mb;
    Slice* currSlice = img->currentSlice;
    prev_mb = GetPreviousMBNr(CurrentMbAddr);    
    if (prev_mb>-1)
    {
      img->mb_data[CurrentMbAddr].prev_qp = img->mb_data[prev_mb].qp;
      if (img->mb_data[prev_mb].slice_nr == img->current_slice_nr)
      {
        img->mb_data[CurrentMbAddr].prev_delta_qp = img->mb_data[prev_mb].delta_qp;
      }
      else
      {
        img->mb_data[CurrentMbAddr].prev_delta_qp = 0;
      }
    }
    else
    {
      img->mb_data[CurrentMbAddr].prev_qp = currSlice->qp;
      img->mb_data[CurrentMbAddr].prev_delta_qp = 0;
    }

    img->mb_data[CurrentMbAddr].qp = img->qp = img->masterQP; 
    img->mb_data[CurrentMbAddr].delta_qp = img->mb_data[CurrentMbAddr].qp - img->mb_data[CurrentMbAddr].prev_qp;
    set_chroma_qp (&(img->mb_data[CurrentMbAddr]));
    Prev_QP      = img->mb_data[CurrentMbAddr].prev_qp;
    Prev_Delta_QP = img->mb_data[CurrentMbAddr].prev_delta_qp;

    rd_cost = encode_macroblock_cluster64(CurrentMbAddr);
    return rd_cost;

  }
}

#endif


