/* 
====================================================================================================================

The copyright in this software is being made available under the License included below. 
This software may be subject to other third party and 	contributor rights, including patent rights, and no such 
rights are granted under this license.

Copyright (c) 2010, QUALCOMM INC.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted only for 
the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
promoting such standards. The following conditions are required to be met:

* Redistributions of source code must retain the above copyright notice, this list of conditions and 
the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
the following disclaimer in the documentation and/or other materials provided with the distribution.
* The name of QUALCOMM INC.  may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

====================================================================================================================
*/
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <float.h>
#include <assert.h>
#include <memory.h>

#include "global.h"
#include "memalloc.h"
#include "malloc.h"
#include "defines.h"
#include "mbuffer.h"
#include "image.h"
#include "switched_filters.h"
#include "defines.h"

#define SQR_FILTER        37

#ifdef SWITCHED_FILTERS
static double AccErrorP[NUM_SIFO][16];                              // [Filter][Sppos]
static double SequenceAccErrorP[NUM_SIFO][16];                      // [Filter][Sppos]
static double SequenceAccErrorB[NUM_SIFO][NUM_SIFO][16][16];        // [FilterF][FilterB][SpposF][SpposB]
static int SequenceBestCombFilterB[16];                             // [Sppos]
// [YUV][Lists][Sppos]
static double SIFO_FILTER[NUM_SIFO][15][SQR_FILTER];

extern pic_parameter_set_rbsp_t *PicParSet[MAXPPS];
extern void GenerateFullPelRepresentation(pel_t ** Fourthpel, pel_t * Fullpel, int xsize, int ysize);
extern void frame_picture (Picture *frame, int method);

double STANDARD_2D_FILTER[15][SQR_FILTER] = {
  {   
    0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
      0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
      1.0/64.0  ,   -5.0/64.0  ,   52.0/64.0  ,   20.0/64.0  ,   -5.0/64.0  ,   1.0/64.0  ,
      0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
      0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
      0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0
  }, // a_pos
  {
    0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
      0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
      1.0/32.0  ,   -5.0/32.0  ,   20.0/32.0  ,   20.0/32.0  ,   -5.0/32.0  ,   1.0/32.0  ,
      0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
      0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
      0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0
    }, // b_pos
    {   0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
    0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
    1.0/64.0  ,   -5.0/64.0  ,   20.0/64.0  ,   52.0/64.0  ,   -5.0/64.0  ,   1.0/64.0  ,
    0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
    0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
    0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0
    }, // c_pos
    {
      0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
        0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
        0.0       ,    0.0       ,   52.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
        0.0       ,    0.0       ,   20.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
        0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
        0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0
    }, // d_pos
    {
      0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
        0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
        1.0/64.0  ,   -5.0/64.0  ,   40.0/64.0  ,   20.0/64.0  ,   -5.0/64.0  ,   1.0/64.0  ,
        0.0       ,    0.0       ,   20.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
        0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
        0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0
      }, // e_pos
      {
        1.0/2048.0,   -5.0/2048.0,   20.0/2048.0,   20.0/2048.0,   -5.0/2048.0,   1.0/2048.0,
          -5.0/2048.0,   25.0/2048.0, -100.0/2048.0, -100.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
          52.0/2048.0, -260.0/2048.0, 1040.0/2048.0, 1040.0/2048.0, -260.0/2048.0,  52.0/2048.0,
          20.0/2048.0, -100.0/2048.0,  400.0/2048.0,  400.0/2048.0, -100.0/2048.0,  20.0/2048.0,
          -5.0/2048.0,   25.0/2048.0, -100.0/2048.0, -100.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
          1.0/2048.0,   -5.0/2048.0,   20.0/2048.0,   20.0/2048.0,   -5.0/2048.0,   1.0/2048.0
      }, // f_pos
      {
        0.0       ,    0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,   0.0       ,
          0.0       ,    0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,   0.0       ,
          1.0/64.0  ,   -5.0/64.0  ,   20.0/64.0  ,   40.0/64.0  ,   -5.0/64.0  ,   1.0/64.0  ,
          0.0       ,    0.0       ,    0.0       ,   20.0/64.0  ,    0.0       ,   0.0       ,
          0.0       ,    0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,   0.0       ,
          0.0       ,    0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,   0.0
        }, // g_pos
        {
          0.0       ,    0.0       ,    1.0/32.0  ,    0.0       ,    0.0       ,   0.0       ,
            0.0       ,    0.0       ,   -5.0/32.0  ,    0.0       ,    0.0       ,   0.0       ,
            0.0       ,    0.0       ,   20.0/32.0  ,    0.0       ,    0.0       ,   0.0       ,
            0.0       ,    0.0       ,   20.0/32.0  ,    0.0       ,    0.0       ,   0.0       ,
            0.0       ,    0.0       ,   -5.0/32.0  ,    0.0       ,    0.0       ,   0.0       ,
            0.0       ,    0.0       ,    1.0/32.0  ,    0.0       ,    0.0       ,   0.0
        }, // h_pos
        {
          1.0/2048.0,   -5.0/2048.0,   52.0/2048.0,   20.0/2048.0,   -5.0/2048.0,   1.0/2048.0,
            -5.0/2048.0,   25.0/2048.0, -260.0/2048.0, -100.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
            20.0/2048.0, -100.0/2048.0, 1040.0/2048.0,  400.0/2048.0, -100.0/2048.0,  20.0/2048.0,
            20.0/2048.0, -100.0/2048.0, 1040.0/2048.0,  400.0/2048.0, -100.0/2048.0,  20.0/2048.0,
            -5.0/2048.0,   25.0/2048.0, -260.0/2048.0, -100.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
            1.0/2048.0,   -5.0/2048.0,   52.0/2048.0,   20.0/2048.0,   -5.0/2048.0,   1.0/2048.0
          }, // i_pos
          {
            1.0/1024.0,   -5.0/1024.0,   20.0/1024.0,   20.0/1024.0,   -5.0/1024.0,   1.0/1024.0,
              -5.0/1024.0,   25.0/1024.0, -100.0/1024.0, -100.0/1024.0,   25.0/1024.0,  -5.0/1024.0,
              20.0/1024.0, -100.0/1024.0,  400.0/1024.0,  400.0/1024.0, -100.0/1024.0,  20.0/1024.0,
              20.0/1024.0, -100.0/1024.0,  400.0/1024.0,  400.0/1024.0, -100.0/1024.0,  20.0/1024.0,
              -5.0/1024.0,   25.0/1024.0, -100.0/1024.0, -100.0/1024.0,   25.0/1024.0,  -5.0/1024.0,
              1.0/1024.0,   -5.0/1024.0,   20.0/1024.0,   20.0/1024.0,   -5.0/1024.0,   1.0/1024.0
          }, // j_pos
          {
            1.0/2048.0,   -5.0/2048.0,   20.0/2048.0,   52.0/2048.0,   -5.0/2048.0,   1.0/2048.0,
              -5.0/2048.0,   25.0/2048.0, -100.0/2048.0, -260.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
              20.0/2048.0, -100.0/2048.0,  400.0/2048.0, 1040.0/2048.0, -100.0/2048.0,  20.0/2048.0,
              20.0/2048.0, -100.0/2048.0,  400.0/2048.0, 1040.0/2048.0, -100.0/2048.0,  20.0/2048.0,
              -5.0/2048.0,   25.0/2048.0, -100.0/2048.0, -260.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
              1.0/2048.0,   -5.0/2048.0,   20.0/2048.0,   52.0/2048.0,   -5.0/2048.0,   1.0/2048.0
            }, // k_pos
            {
              0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
                0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
                0.0       ,    0.0       ,   20.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
                0.0       ,    0.0       ,   52.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
                0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
                0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0
            }, // l_pos
            {
              0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
                0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
                0.0       ,    0.0       ,   20.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
                1.0/64.0  ,   -5.0/64.0  ,   40.0/64.0  ,   20.0/64.0  ,   -5.0/64.0  ,   1.0/64.0  ,
                0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
                0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0
              }, // m_pos
              {
                1.0/2048.0,   -5.0/2048.0,   20.0/2048.0,   20.0/2048.0,   -5.0/2048.0,   1.0/2048.0,
                  -5.0/2048.0,   25.0/2048.0, -100.0/2048.0, -100.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
                  20.0/2048.0, -100.0/2048.0,  400.0/2048.0,  400.0/2048.0, -100.0/2048.0,  20.0/2048.0,
                  52.0/2048.0, -260.0/2048.0, 1040.0/2048.0, 1040.0/2048.0, -260.0/2048.0,  52.0/2048.0,
                  -5.0/2048.0,   25.0/2048.0, -100.0/2048.0, -100.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
                  1.0/2048.0,   -5.0/2048.0,   20.0/2048.0,   20.0/2048.0,   -5.0/2048.0,   1.0/2048.0
              }, // n_pos
              {
                0.0         ,    0.0         ,  0.0     ,    1.0/64.0  ,    0.0       ,   0.0       ,
                  0.0         ,    0.0         ,  0.0     ,   -5.0/64.0  ,    0.0       ,   0.0       ,
                  0.0         ,    0.0         ,  0.0     ,   20.0/64.0  ,    0.0       ,   0.0       ,
                  1.0/64.0    ,   -5.0/64.0    , 20.0/64.0,   40.0/64.0  ,   -5.0/64.0  ,   1.0/64.0  ,
                  0.0         ,    0.0         ,  0.0     ,   -5.0/64.0  ,    0.0       ,   0.0       ,
                  0.0         ,    0.0         ,  0.0     ,    1.0/64.0  ,    0.0       ,   0.0
                } // o_pos
};

//
//
//
double SYMMETRIC_1[15][SQR_FILTER] = {
  {
    0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  
      0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  
      -1.0/128.0, -2.0/128.0, 97.0/128.0, 42.0/128.0, -11.0/128.0,  3.0/128.0, 
      0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  
      0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  
      0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0
  },  // a_pos
  {
    0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  
      0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  
      3.0/128.0, -16.0/128.0, 77.0/128.0, 77.0/128.0, -16.0/128.0,  3.0/128.0, 
      0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  
      0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  
      0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0
    },  // b_pos
    {
      0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  
        0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  
        3.0/128.0, -11.0/128.0, 42.0/128.0, 97.0/128.0, -2.0/128.0, -1.0/128.0, 
        0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  
        0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  
        0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0
    },  // c_pos
    {
      0.0/128.0,  0.0/128.0,   -1.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
        0.0/128.0,  0.0/128.0,   -2.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
        0.0/128.0,  0.0/128.0,   97.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
        0.0/128.0,  0.0/128.0,   42.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
        0.0/128.0,  0.0/128.0,  -11.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
        0.0/128.0,  0.0/128.0,    3.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0
      },  // d_pos
      {
        0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  
          0.0/128.0,  -7.0/128.0,  4.0/128.0, -2.0/128.0, -3.0/128.0,  0.0/128.0,  
          0.0/128.0,   4.0/128.0, 66.0/128.0, 36.0/128.0,  1.0/128.0,  0.0/128.0,  
          0.0/128.0,  -2.0/128.0, 36.0/128.0,  7.0/128.0, -6.0/128.0,  0.0/128.0,  
          0.0/128.0,  -3.0/128.0,  1.0/128.0, -6.0/128.0,  2.0/128.0,  0.0/128.0,  
          0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0
      },  // e_pos
      {
        0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  
          0.0/128.0,  -4.0/128.0, -1.0/128.0, -1.0/128.0, -4.0/128.0,  0.0/128.0,  
          0.0/128.0,  -2.0/128.0, 56.0/128.0, 56.0/128.0, -2.0/128.0,  0.0/128.0,  
          0.0/128.0,  -7.0/128.0, 25.0/128.0, 25.0/128.0, -7.0/128.0,  0.0/128.0,  
          0.0/128.0,   1.0/128.0, -4.0/128.0, -4.0/128.0,  1.0/128.0,  0.0/128.0,  
          0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0
        },  // f_pos
        {
          0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  
            0.0/128.0,  -3.0/128.0, -2.0/128.0,  4.0/128.0, -7.0/128.0,  0.0/128.0,  
            0.0/128.0,   1.0/128.0, 36.0/128.0, 66.0/128.0,  4.0/128.0,  0.0/128.0,  
            0.0/128.0,  -6.0/128.0,  7.0/128.0, 36.0/128.0, -2.0/128.0,  0.0/128.0,  
            0.0/128.0,   2.0/128.0, -6.0/128.0,  1.0/128.0, -3.0/128.0,  0.0/128.0,  
            0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0
        },  // g_pos
        {
          0.0/128.0,  0.0/128.0,    3.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
            0.0/128.0,  0.0/128.0,  -16.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
            0.0/128.0,  0.0/128.0,   77.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
            0.0/128.0,  0.0/128.0,   77.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
            0.0/128.0,  0.0/128.0,  -16.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
            0.0/128.0,  0.0/128.0,    3.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0
          },  // h_pos
          {
            0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  
              0.0/128.0,  -4.0/128.0, -2.0/128.0, -7.0/128.0,  1.0/128.0,  0.0/128.0,  
              0.0/128.0,  -1.0/128.0, 56.0/128.0, 25.0/128.0, -4.0/128.0,  0.0/128.0,  
              0.0/128.0,  -1.0/128.0, 56.0/128.0, 25.0/128.0, -4.0/128.0,  0.0/128.0,  
              0.0/128.0,  -4.0/128.0, -2.0/128.0, -7.0/128.0,  1.0/128.0,  0.0/128.0,  
              0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0
          },  // i_pos
          {
            0.0/128.0,   0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  
              0.0/128.0,   0.0/128.0,  -6.0/128.0, -6.0/128.0,  0.0/128.0,  0.0/128.0,  
              0.0/128.0,  -6.0/128.0,  44.0/128.0, 44.0/128.0, -6.0/128.0,  0.0/128.0,  
              0.0/128.0,  -6.0/128.0,  44.0/128.0, 44.0/128.0, -6.0/128.0,  0.0/128.0,  
              0.0/128.0,   0.0/128.0,  -6.0/128.0, -6.0/128.0,  0.0/128.0,  0.0/128.0,  
              0.0/128.0,   0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0
            },  // j_pos
            {
              0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  
                0.0/128.0,   1.0/128.0, -7.0/128.0, -2.0/128.0, -4.0/128.0,  0.0/128.0,  
                0.0/128.0,  -4.0/128.0, 25.0/128.0, 56.0/128.0, -1.0/128.0,  0.0/128.0,  
                0.0/128.0,  -4.0/128.0, 25.0/128.0, 56.0/128.0, -1.0/128.0,  0.0/128.0,  
                0.0/128.0,   1.0/128.0, -7.0/128.0, -2.0/128.0, -4.0/128.0,  0.0/128.0,  
                0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0
            },  // k_pos
            {
              0.0/128.0,  0.0/128.0,    3.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
                0.0/128.0,  0.0/128.0,  -11.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
                0.0/128.0,  0.0/128.0,   42.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
                0.0/128.0,  0.0/128.0,   97.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
                0.0/128.0,  0.0/128.0,   -2.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
                0.0/128.0,  0.0/128.0,   -1.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0
              },  // l_pos
              {
                0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  
                  0.0/128.0,  -3.0/128.0,  1.0/128.0, -6.0/128.0,  2.0/128.0,  0.0/128.0,  
                  0.0/128.0,  -2.0/128.0, 36.0/128.0,  7.0/128.0, -6.0/128.0,  0.0/128.0,  
                  0.0/128.0,   4.0/128.0, 66.0/128.0, 36.0/128.0,  1.0/128.0,  0.0/128.0,  
                  0.0/128.0,  -7.0/128.0,  4.0/128.0, -2.0/128.0, -3.0/128.0,  0.0/128.0,  
                  0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0
              },  // m_pos
              {
                0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  
                  0.0/128.0,   1.0/128.0, -4.0/128.0, -4.0/128.0,  1.0/128.0,  0.0/128.0,  
                  0.0/128.0,  -7.0/128.0, 25.0/128.0, 25.0/128.0, -7.0/128.0,  0.0/128.0,  
                  0.0/128.0,  -2.0/128.0, 56.0/128.0, 56.0/128.0, -2.0/128.0,  0.0/128.0,  
                  0.0/128.0,  -4.0/128.0, -1.0/128.0, -1.0/128.0, -4.0/128.0,  0.0/128.0,  
                  0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0
                },  // n_pos
                {
                  0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  
                    0.0/128.0,   2.0/128.0, -6.0/128.0,  1.0/128.0, -3.0/128.0,  0.0/128.0,  
                    0.0/128.0,  -6.0/128.0,  7.0/128.0, 36.0/128.0, -2.0/128.0,  0.0/128.0,  
                    0.0/128.0,   1.0/128.0, 36.0/128.0, 66.0/128.0,  4.0/128.0,  0.0/128.0,  
                    0.0/128.0,  -3.0/128.0, -2.0/128.0,  4.0/128.0, -7.0/128.0,  0.0/128.0,  
                    0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0
                }  // o_pos
};


//
//
//
double SYMMETRIC_2[15][SQR_FILTER] = {
  {
    0.0/128.0,   0.0/128.0,   0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  
      0.0/128.0,   0.0/128.0,   0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  
      4.0/128.0, -14.0/128.0, 108.0/128.0, 38.0/128.0, -12.0/128.0,  4.0/128.0, 
      0.0/128.0,   0.0/128.0,   0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  
      0.0/128.0,   0.0/128.0,   0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  
      0.0/128.0,   0.0/128.0,   0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0
  },  // a_pos
  {
    0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  
      0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  
      6.0/128.0, -23.0/128.0, 81.0/128.0, 81.0/128.0, -23.0/128.0,  6.0/128.0, 
      0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  
      0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  
      0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0
    },  // b_pos
    {
      0.0/128.0,   0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  
        0.0/128.0,   0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  
        4.0/128.0, -12.0/128.0, 38.0/128.0, 108.0/128.0, -14.0/128.0, 4.0/128.0, 
        0.0/128.0,   0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  
        0.0/128.0,   0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  
        0.0/128.0,   0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0
    },  // c_pos
    {
      0.0/128.0,  0.0/128.0,    4.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
        0.0/128.0,  0.0/128.0,  -14.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
        0.0/128.0,  0.0/128.0,  108.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
        0.0/128.0,  0.0/128.0,   38.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
        0.0/128.0,  0.0/128.0,  -12.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
        0.0/128.0,  0.0/128.0,    4.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0
      },  // d_pos
      {
        0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  
          0.0/128.0,  -6.0/128.0,  2.0/128.0, -4.0/128.0, -3.0/128.0,  0.0/128.0,  
          0.0/128.0,   2.0/128.0, 71.0/128.0, 38.0/128.0,  0.0/128.0,  0.0/128.0,  
          0.0/128.0,  -4.0/128.0, 38.0/128.0,  7.0/128.0, -7.0/128.0,  0.0/128.0,  
          0.0/128.0,  -3.0/128.0,  0.0/128.0, -7.0/128.0,  4.0/128.0,  0.0/128.0,  
          0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0
      },  // e_pos
      {
        0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  
          0.0/128.0,  -3.0/128.0, -2.0/128.0, -2.0/128.0, -3.0/128.0,  0.0/128.0,  
          0.0/128.0,  -3.0/128.0, 58.0/128.0, 58.0/128.0, -3.0/128.0,  0.0/128.0,  
          0.0/128.0,  -9.0/128.0, 26.0/128.0, 26.0/128.0, -9.0/128.0,  0.0/128.0,  
          0.0/128.0,   2.0/128.0, -5.0/128.0, -5.0/128.0,  2.0/128.0,  0.0/128.0,  
          0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0
        },  // f_pos
        {
          0.0/128.0,   0.0/128.0,   0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  
            0.0/128.0,  -3.0/128.0,  -4.0/128.0,  2.0/128.0,  -6.0/128.0,  0.0/128.0,  
            0.0/128.0,   0.0/128.0,  38.0/128.0, 71.0/128.0,   2.0/128.0,  0.0/128.0,  
            0.0/128.0,  -7.0/128.0,   7.0/128.0, 38.0/128.0,  -4.0/128.0,  0.0/128.0,  
            0.0/128.0,   4.0/128.0,  -7.0/128.0,  0.0/128.0,  -3.0/128.0,  0.0/128.0,  
            0.0/128.0,   0.0/128.0,   0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0
        },  // g_pos
        {
          0.0/128.0,  0.0/128.0,    6.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
            0.0/128.0,  0.0/128.0,  -23.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
            0.0/128.0,  0.0/128.0,   81.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
            0.0/128.0,  0.0/128.0,   81.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
            0.0/128.0,  0.0/128.0,  -23.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
            0.0/128.0,  0.0/128.0,    6.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0
          },  // h_pos
          {
            0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  
              0.0/128.0,  -3.0/128.0, -3.0/128.0, -9.0/128.0,  2.0/128.0,  0.0/128.0,  
              0.0/128.0,  -2.0/128.0, 58.0/128.0, 26.0/128.0, -5.0/128.0,  0.0/128.0,  
              0.0/128.0,  -2.0/128.0, 58.0/128.0, 26.0/128.0, -5.0/128.0,  0.0/128.0,  
              0.0/128.0,  -3.0/128.0, -3.0/128.0, -9.0/128.0,  2.0/128.0,  0.0/128.0,  
              0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0
          },  // i_pos
          {
            0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  
              0.0/128.0,   1.0/128.0, -7.0/128.0, -7.0/128.0,  1.0/128.0,  0.0/128.0,  
              0.0/128.0,  -7.0/128.0, 45.0/128.0, 45.0/128.0, -7.0/128.0,  0.0/128.0,  
              0.0/128.0,  -7.0/128.0, 45.0/128.0, 45.0/128.0, -7.0/128.0,  0.0/128.0,  
              0.0/128.0,   1.0/128.0, -7.0/128.0, -7.0/128.0,  1.0/128.0,  0.0/128.0,  
              0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0
            },  // j_pos
            {
              0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  
                0.0/128.0,   2.0/128.0, -9.0/128.0, -3.0/128.0, -3.0/128.0,  0.0/128.0,  
                0.0/128.0,  -5.0/128.0, 26.0/128.0, 58.0/128.0, -2.0/128.0,  0.0/128.0,  
                0.0/128.0,  -5.0/128.0, 26.0/128.0, 58.0/128.0, -2.0/128.0,  0.0/128.0,  
                0.0/128.0,   2.0/128.0, -9.0/128.0, -3.0/128.0, -3.0/128.0,  0.0/128.0,  
                0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0
            },  // k_pos
            {
              0.0/128.0,  0.0/128.0,    4.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
                0.0/128.0,  0.0/128.0,  -12.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
                0.0/128.0,  0.0/128.0,   38.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
                0.0/128.0,  0.0/128.0,  108.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
                0.0/128.0,  0.0/128.0,  -14.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0,  
                0.0/128.0,  0.0/128.0,    4.0/128.0, 0.0/128.0,  0.0/128.0,  0.0/128.0
              },  // l_pos
              {
                0.0/128.0,   0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  
                  0.0/128.0,  -3.0/128.0,  0.0/128.0,  -7.0/128.0,  4.0/128.0,  0.0/128.0,  
                  0.0/128.0,  -4.0/128.0, 38.0/128.0,   7.0/128.0, -7.0/128.0,  0.0/128.0,  
                  0.0/128.0,   2.0/128.0, 71.0/128.0,  38.0/128.0,  0.0/128.0,  0.0/128.0,  
                  0.0/128.0,  -6.0/128.0,  2.0/128.0,  -4.0/128.0, -3.0/128.0,  0.0/128.0,  
                  0.0/128.0,   0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0
              },  // m_pos
              {
                0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  
                  0.0/128.0,   2.0/128.0, -5.0/128.0, -5.0/128.0,  2.0/128.0,  0.0/128.0,  
                  0.0/128.0,  -9.0/128.0, 26.0/128.0, 26.0/128.0, -9.0/128.0,  0.0/128.0,  
                  0.0/128.0,  -3.0/128.0, 58.0/128.0, 58.0/128.0, -3.0/128.0,  0.0/128.0,  
                  0.0/128.0,  -3.0/128.0, -2.0/128.0, -2.0/128.0, -3.0/128.0,  0.0/128.0,  
                  0.0/128.0,   0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0,  0.0/128.0
                },  // n_pos
                {
                  0.0/128.0,   0.0/128.0,   0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0,  
                    0.0/128.0,   4.0/128.0,  -7.0/128.0,  0.0/128.0,  -3.0/128.0,  0.0/128.0,  
                    0.0/128.0,  -7.0/128.0,   7.0/128.0, 38.0/128.0,  -4.0/128.0,  0.0/128.0,  
                    0.0/128.0,   0.0/128.0,  38.0/128.0, 71.0/128.0,   2.0/128.0,  0.0/128.0,  
                    0.0/128.0,  -3.0/128.0,  -4.0/128.0,  2.0/128.0,  -6.0/128.0,  0.0/128.0,  
                    0.0/128.0,   0.0/128.0,   0.0/128.0,  0.0/128.0,   0.0/128.0,  0.0/128.0
                }  // o_pos
};


//
//  Clip a value between low and high
//
int Clip3Fun(int low, int high, int val)
{
  return ((val < low)? low: (val > high)? high: val);
}


//
//  Determine noSamples
//
int get_thDC(int noMB)
{
  int thDC[4] = NO_SAMPLES;
  int mb[4]   = {   99,  396, 3600, 8100};
  int i;
  int last = 1;

  for(i = 0; i < 4; ++i)
    if(noMB <= mb[i])
      return thDC[i];
    else
      last = thDC[i];
  return last; 
}

//
//  Offset between two frames
//
double getDCdiff(int list, int ref)
{
  int i, j;
  double DCOrg = 0.0;
  double DCRef = 0.0;

  for(i = 0; i < input->img_height; ++i) //y
  {
    for(j = 0; j < input->img_width; ++j) //x
    {
      DCOrg += imgY_org[i][j];
      DCRef += listX[list][ref]->imgY[i][j];
    }
  }
  if(img->BitDepthIncrease)
  {
    DCOrg /= (1<<img->BitDepthIncrease);
    DCRef /= (1<<img->BitDepthIncrease);
  }
  return ((DCOrg - DCRef) / (double)(input->img_height * input->img_width));
}

//
//  Block-based offset between two frames
//
void blockDC(int list, int *DCmin, int *DCmax, int noSamples)
{
  int i, j, ii, jj, iiLimit, jjLimit;
  int ref_frame, noPixels;
  double DCOrg = 0.0;
  double DCRef = 0.0;
  double DCdiff, temp, err0, err1;
  int DCInt;
  int DCinterval[2][NO_DC_VAL] = {{0}};

  memset(DCinterval, 0, 2 * NO_DC_VAL * sizeof(int));
  ref_frame = 0;

  for(i = 0; i < input->img_height; i+=BL_SIZE)
  {
    for(j = 0; j < input->img_width; j+=BL_SIZE)
    {
      DCOrg = 0.0;
      DCRef = 0.0;
      noPixels = 0;

      for(ii = i; ii < i + BL_SIZE; ii++)
      {
        iiLimit = min(ii, input->img_height - 1);
        for(jj = j; jj < j + BL_SIZE; jj++)
        {
          jjLimit = min(jj, input->img_width - 1);
          DCOrg += imgY_org[iiLimit][jjLimit];
          DCRef += listX[list][ref_frame]->imgY[iiLimit][jjLimit];
          noPixels++;
        }
      }
      DCdiff = (DCOrg - DCRef) / (double)noPixels;
      DCOrg = DCOrg / (double)noPixels;
      DCRef = DCRef / (double)noPixels;

      err0 = 0.0;
      for(ii = i; ii < i + BL_SIZE; ii++)
      {
        iiLimit = min(ii, input->img_height - 1);
        for(jj = j; jj < j + BL_SIZE; jj++)
        {
          jjLimit = min(jj, input->img_width - 1);
          temp = imgY_org[iiLimit][jjLimit] - listX[list][ref_frame]->imgY[iiLimit][jjLimit];
          err0 += temp * temp;
        }
      }

      err1 = 0.0;
      for(ii = i; ii < i + BL_SIZE; ii++)
      {
        iiLimit = min(ii, input->img_height - 1);
        for(jj = j; jj < j + BL_SIZE; jj++)
        {
          jjLimit = min(jj, input->img_width - 1);
          temp = (imgY_org[iiLimit][jjLimit] - DCOrg) - (listX[list][ref_frame]->imgY[iiLimit][jjLimit] - DCRef);
          err1 += temp * temp;
        }
      }

      if(img->BitDepthIncrease)
        DCdiff /= (1<<img->BitDepthIncrease);


      DCInt = min((int)(fabs(DCdiff) + 0.5), NO_DC_VAL - 1);

      if(err1 < (err0 - err1))
      {    
        if(DCdiff < 0)
        {
          DCinterval[0][DCInt]++;
          DCInt = -DCInt;// ??? redundant 
        }
        else
        {
          DCinterval[1][DCInt]++;
        }
      }
    }
  }

  (*DCmin) = 0;
  (*DCmax) = 0;

  {
    int flag;
    for(i=0, flag=0; i<NO_DC_VAL; i++)
    {
      if(DCinterval[1][i]>=noSamples)
      {
        (*DCmax)=(*DCmax)+1;
        flag = 1;
      }
      else if(flag)
      {
        i = NO_DC_VAL;
      }
    }
    for(i=0, flag=0; i<NO_DC_VAL; i++)
    {
      if(DCinterval[0][i]>=noSamples)
      {
        (*DCmin)=(*DCmin)+1;
        flag = 1;
      }
      else if(flag)
      {
        i = NO_DC_VAL;
      }
    }
  }

  (*DCmin) = -(*DCmin);
}

//
//  Reset subpel offsets for the first pass
//
void resetFirstPassSubpelOffset(int list)
{
  int i;

  img->noOffsets[list] = img->noMEOffsets[list] = 0;
  img->currMEOffset[list] = 0;

  for(i = 0; i < 16; ++i) 
    img->offsetMETab[list][i] = 0;
}

//
//  Set subpel offsets for the first pass
//
void setFirstPassSubpelOffset(int list)
{
  int subPelPosOrder[15] = {5, 15, 13, 7, 9, 6, 11, 14, 1, 4, 3, 12, 10, 2, 8};
  int i, sign, DCmin, DCmax, thDC;
  int ref_frame, noOffsets, noOffsetsSecond, DCint, firstVal, secondVal, addOffset;
  double DCdiff[MAX_REFERENCE_PICTURES];
  int roundFact[2];
  int DCminF[2];
  int DCmaxF[2];
  int firstValF[2];
  int secondValF[2];
  int noOffsetsSecondF[2];

  //  e   o   m   g   i   f    k    n   a   d   c   l   j   b   h
  static const int search_points_x[15] = {1, -1,  1, -1,  1,  2,  -1,   2,  1,  0, -1,  0,  2,  2,  0};
  static const int search_points_y[15] = {1, -1, -1,  1,  2,  1,   2,  -1,  0,  1,  0, -1,  2,  0,  2};

  memset(DCdiff, 0, MAX_REFERENCE_PICTURES * sizeof(double));
  memset(img->offsetMETab[list], 0, 16 * sizeof(int));

  memcpy(img->search_point_qp_x_offset, search_points_x, 15 * sizeof(int));
  memcpy(img->search_point_qp_y_offset, search_points_y, 15 * sizeof(int));

  for(ref_frame = 0; ref_frame < listXsize[list]; ref_frame++)
  {    
    DCdiff[ref_frame] = getDCdiff(list, ref_frame);

    if(ref_frame==0)
    {
      sign = (DCdiff[ref_frame] >= 0)? 1: -1;
      noOffsets = (int)(fabs(DCdiff[ref_frame]) + 0.5);
      DCint = noOffsets * sign;

      if(noOffsets < 2)
      {
        DCmax = noOffsets;
        DCmin = -noOffsets;
      }
      else
      {
        thDC = get_thDC(img->FrameSizeInMbs);//determine the threshold of the number of samples  for each offset value
        blockDC(list, &DCmin, &DCmax, thDC); 
        DCmax = max(DCint, DCmax);
        DCmin = min(DCint, DCmin);
      }

      roundFact[list] = (int)ceil((double)(DCmax - DCmin) / 15.0);
      roundFact[list] = max(roundFact[list], 1);      
      DCminF[list] = (int)(fabs(DCmin) / roundFact[list] + 0.5);
      DCmaxF[list] = (int)(fabs(DCmax) / roundFact[list] + 0.5);
      DCmin = -roundFact[list] * DCminF[list];
      DCmax = roundFact[list] * DCmaxF[list];

      noOffsetsSecond = 0;
      if(noOffsets < 2)
        noOffsetsSecond = min((int)(fabs(DCdiff[ref_frame]) / 0.1 + 0.5), 15);

      if(noOffsetsSecond > 0)//number of subple using the second offset value
      {
        firstVal = 0;
        secondVal = sign;
        if(fabs(DCdiff[ref_frame]) > 0.5)
        {
          firstVal = sign; 
          secondVal = 0;
          noOffsetsSecond = 16 - noOffsetsSecond;
        }
      }
      else
      {
        firstVal = 0; 
        secondVal = 0;
      }
      addOffset = -firstVal;


      //firstVal and secondVal are based on the whole frame DCdiff, DCmin and DCmax is based on DCdiff and local DC diff
      if(((firstVal == DCmin) && (secondVal == DCmax)) || ((firstVal == DCmax) && (secondVal == DCmin)))
      {
        DCmin = 0; 
        DCmax = 0;
        DCminF[list] = 0; 
        DCmaxF[list] = 0;
      }

      for(i = 0; i < 16; i++)
        img->subpelOffset[0][list][i] = firstVal;

      img->noOffsets[list] = 0;
      for(i = DCmin; i <= DCmax; i += roundFact[list])
      {
        if((i != firstVal) && (i != secondVal))
        {
          img->subpelOffset[0][list][subPelPosOrder[img->noOffsets[list]]] = i;
          img->offsetMETab[list][img->noOffsets[list]] = i + addOffset;
          (img->noOffsets[list])++;
          if(img->noOffsets[list] == 15)
            break;
        }
      }

      img->noMEOffsets[list] = img->noOffsets[list];
      if((img->noOffsets[list] < 15) && (noOffsetsSecond > 0))
      {
        img->offsetMETab[list][img->noMEOffsets[list]] = secondVal + addOffset;
        (img->noMEOffsets[list])++;
      }

      for(i = 0; i < noOffsetsSecond; i++)
      {
        if(img->noOffsets[list] == 15)
          break;

        img->subpelOffset[0][list][subPelPosOrder[img->noOffsets[list]]] = secondVal;
        (img->noOffsets[list])++;
      }

      firstValF[list] = firstVal;
      secondValF[list] = secondVal;
      noOffsetsSecondF[list] = noOffsetsSecond;
    }
    else
    {
      sign = (DCdiff[ref_frame] >= 0)? 1: -1;

#ifdef THRESHOLD_FRAME_OFFSETS
#define TFO 2.0
      img->imgOffset[0][list][ref_frame] = (fabs(DCdiff[ref_frame]) < TFO)? 0: 1;  // ???
#else
      img->imgOffset[list][ref_frame] = min((int)(fabs(DCdiff[ref_frame]) + 0.5), 1);
#endif  // THRESHOLD_FRAME_OFFSETS

      if(sign < 0)
        img->imgOffset[0][list][ref_frame] = -img->imgOffset[0][list][ref_frame];
    }
  }

  if(img->BitDepthIncrease)
  {
    for(ref_frame = 0; ref_frame < listXsize[list]; ref_frame++)
    { 
      if(ref_frame==0)
      {
        for(i = 0; i < 16; i++)
        {
          img->subpelOffset[0][list][i] *= (1<<img->BitDepthIncrease);
          img->offsetMETab[list][i] *= (1<<img->BitDepthIncrease);
        }

      }
      else
      {
        img->imgOffset[0][list][ref_frame] *= (1<<img->BitDepthIncrease);
      }
    }
  }
}

//
//  Find a decision vector different in one position and with smallest SAD 
//
static double get_min_d(double err[NUM_SIFO][NUM_SIFO][16][16], 
                        int old_d[16], int new_d[16], double currSAD)
{
  int spp, flt, sppF, sppB;
  double minSAD = currSAD;
  int tmp_d[16];
  double tmpSAD;
  int num_sifo = NUM_SIFO;

  for(spp = 1; spp < 16; ++spp)
  {
    // Copy a fresh old_d into tmp_d        
    memcpy(tmp_d, old_d, sizeof(int) * 16);

    // Try the remaining (NUM_SIFO - 1) filters on the position spp
    for(flt = 0; flt < num_sifo - 1; ++flt)
    {
      tmp_d[spp] = (tmp_d[spp] + 1) % num_sifo;
      tmpSAD = 0;
      for(sppF = 1; sppF < 16; ++sppF)
        for(sppB = 1; sppB < 16; ++sppB)
          tmpSAD += err[tmp_d[sppF]][tmp_d[sppB]][sppF][sppB];

      if(tmpSAD < minSAD)
      {
        minSAD = tmpSAD;
        memcpy(new_d, tmp_d, sizeof(int) * 16);
      }
    }
  }
  return minSAD;
}


//
//  Compute filter combination for frame or sequence
static void ComputeFilterCombination_B_gd(double err[NUM_SIFO][NUM_SIFO][16][16], int out[16])
{
  int min_d[16] = {0};
  double oldSAD = 0.0;
  double minSAD;
  int sppF, sppB;
  int tmp_d[16] = {0};

  // Compute error for min_d = 0^n
  for(sppF = 1; sppF < 16; ++sppF)
    for(sppB = 1; sppB < 16; ++sppB)
      oldSAD += err[0][0][sppF][sppB];

  // While SAD can be improved
  while((minSAD = get_min_d(err, min_d, tmp_d, oldSAD)) < oldSAD)
  {
    oldSAD = minSAD;        
    memcpy(min_d, tmp_d, sizeof(int) * 16);
  }

  // Copy result
  memcpy(out, min_d, sizeof(int) * 16);
}


//
//  Check whether a unique filter has a RD advantage over the filter combination
//

static void ComputeBestFilter_B(double err[NUM_SIFO][NUM_SIFO][16][16], int combination[16])
{
  int bits = 0;
  int i;
  double errComb = 0.0;
  double errBest = 0.0;
  int sppF, sppB;
  int best = -1;

  // Lowest error with a single filter
  for(i = 0; i < NUM_SIFO; ++i)
  {
    double errTmp = 0.0;
    for(sppF = 1; sppF < 16; ++sppF)
      for(sppB = 1; sppB < 16; ++sppB)
        errTmp += err[i][i][sppF][sppB];
    if((best == -1) || (errTmp < errBest))
    {
      best = i;
      errBest = errTmp;
    }
  }

  // Compute bit cost of sending the filters
  for(i = 1; i < 16; ++i)
    bits += (combination[i] == 0)? 1: 2;

  // Error with filter combination
  for(sppF = 1; sppF < 16; ++sppF)
    for(sppB = 1; sppB < 16; ++sppB)
      errComb += err[combination[sppF]][combination[sppB]][sppF][sppB];


  if(errBest < (errComb + img->lambda_md[img->type][img->qp] * (double)bits))
  {
    for(i = 0; i < 16; ++i){
      combination[i] = best;
    }
    img->bestFilter = best + 1;
  }
  else
  {
    img->bestFilter = 0;
  }
}

//
//  Compute filter combination for frame and sequence
//
static void ComputeFilterCombination_B()
{
  ComputeFilterCombination_B_gd(SequenceAccErrorB, SequenceBestCombFilterB); //compute sequence filter HPF4 or HPF3 first pass
  ComputeBestFilter_B (SequenceAccErrorB, SequenceBestCombFilterB); // Compute best single sequence filter  HPF4 or HPF3 first pass
}

//
//
//
static void ResetAll(int slice)
{
  int a, b, c, d;
  static int firstP = 1;
  static int firstB = 1;

  if(slice == P_SLICE)
  {
    resetFirstPassSubpelOffset(LIST_0);

    for(a = 0; a < NUM_SIFO; ++a)
      memset(AccErrorP[a], 0, 16 * sizeof(double));

    if(firstP == 1)
    {
      firstP = 0;

      for(a = 0; a < NUM_SIFO; ++a)
        memset(SequenceAccErrorP[a], 0, 16 * sizeof(double));
    }
    else
    {
      for(a = 0; a < 16; ++a)
        for(b = 0; b < NUM_SIFO; ++b)
          SequenceAccErrorP[b][a] *= NORM_FACTOR;
    }
  }
  else if(slice == B_SLICE)
  {
    resetFirstPassSubpelOffset(LIST_0);
    resetFirstPassSubpelOffset(LIST_1);

    if(firstB == 1)
    {    
      firstB = 0;
      memset(SequenceBestCombFilterB, 0, 16 * sizeof(int));

      for(a = 0; a < NUM_SIFO; ++a)
        for(b = 0; b < NUM_SIFO; ++b)
          for(c = 0; c < 16; ++c)
            for(d = 0; d < 16; ++d)
              SequenceAccErrorB[a][b][c][d] = 0;
    }
    else
    {
      for(a = 0; a < NUM_SIFO; ++a)
        for(b = 0; b < NUM_SIFO; ++b)
          for(c = 0; c < 16; ++c)
            for(d = 0; d < 16; ++d)
              SequenceAccErrorB[a][b][c][d] *= NORM_FACTOR;
    }
  }  
}

//
//  Return the value of a pixel interpolated with filter filterNo 
//
static int GetInterpolatedPixel8TapSeperable(StorablePicture *s, int icoord, int jcoord, 
                                             int img_width, int img_height, int sub_pos, int filterNo,int i,int j)
{
#define FILTER_SIZE       8                     // Undefined at the end of the function
#define FILTER_OFFSET     (8 / 2 - 1)           // Undefined at the end of the function

  static int COEF[3][8] =
  {
    { -3, 12, -37, 229,  71, -21,  6, -1 },
    { -3, 12, -39, 158, 158, -39, 12, -3 },
    { -1,  6, -21,  71, 229, -37, 12, -3 }
  };

  int out;
  int ipVal_i,ii,jj,pos_y,pos_x;
  int filtersize=FILTER_SIZE;
  int filteroffset=FILTER_OFFSET;
  int round8 = 128;
  imgpel ** img= s->imgY;

  if (sub_pos<4)
  {
    int* p_filt;

    ipVal_i =0;
    p_filt= &COEF[sub_pos-1][0];
    for(jj = 0; jj < filtersize; ++jj)
    {
      pos_y = Clip3Fun(0, img_height - 1, i / 4);
      if((i < 0) && (pos_y > 0))
        pos_y -= 1;
      pos_x = Clip3Fun(0, img_width - 1, j / 4 - filteroffset + jj);
      if((j < 0) && (pos_x > 0))
        pos_x -=1;
      ipVal_i += *(p_filt++)* img[pos_y][pos_x];
    }
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
    out = Clip3Fun(0, (1 << input->BitDepthLuma) - 1, (int)((ipVal_i + round8)>> 8));
#else
    out = Clip3Fun(0, 255, (int)((ipVal_i + round8)>> 8));
#endif		
  }
  else if (sub_pos==4|| sub_pos==8||sub_pos==12)
  {
    int* p_filt;
    ipVal_i =0;
    p_filt= &COEF[(sub_pos>>2)-1][0];
    for(ii = 0; ii < filtersize; ++ii)
    {
      pos_y = Clip3Fun(0, img_height - 1, i / 4 - filteroffset + ii);
      if((i < 0) && (pos_y > 0))
        pos_y -= 1;
      pos_x = Clip3Fun(0, img_width - 1, j / 4);
      if((j < 0) && (pos_x > 0))
        pos_x -=1;
      ipVal_i += *(p_filt++)* img[pos_y][pos_x];
    }
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
    out = Clip3Fun(0, (1 << input->BitDepthLuma) - 1, (int)((ipVal_i + round8)>> 8));
#else
    out = Clip3Fun(0, 255, (int)((ipVal_i + round8)>> 8));
#endif
  }
  else
  {
    int* p_filt;
    int round16 = (1<<15);
    int ipVal_i_oneD[8]={0};
    int sub_pos_minus_one =(sub_pos>>2)-1;
    ipVal_i =0;
    for(jj = 0; jj < filtersize; ++jj)
    {
      pos_x = Clip3Fun(0, img_width - 1, j / 4- filteroffset + jj);
      if((j < 0) && (pos_x > 0))
        pos_x -=1;
      p_filt= &COEF[sub_pos_minus_one][0];
      for(ii = 0; ii < filtersize; ++ii)
      {
        pos_y = Clip3Fun(0, img_height - 1, i / 4 - filteroffset + ii);
        if((i < 0) && (pos_y > 0))
          pos_y -= 1;  
        ipVal_i_oneD[jj] += *(p_filt++)* img[pos_y][pos_x];
      }
    }
    p_filt= &COEF[(sub_pos&3)-1][0];
    for(jj = 0; jj < filtersize; ++jj)
    {
      ipVal_i += *(p_filt++)* ipVal_i_oneD[jj];
    }
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
    out = Clip3Fun(0, (1 << input->BitDepthLuma) - 1, (int)((ipVal_i + round16)>> 16));
#else
    out = Clip3Fun(0, 255, (int)((ipVal_i + round16)>> 16));
#endif
  }

#undef FILTER_OFFSET
#undef FILTER_SIZE
  return out;
}






static int GetInterpolatedPixel(StorablePicture *s, int icoord, int jcoord, 
                                int img_width, int img_height, int sub_pos, int filterNo)
{
  int pos_x, pos_y;
  double ipVal;
  int out;
  int i, j;
  int ii,jj;
  imgpel **imgY = s->imgY;

  assert(filterNo < NUM_SIFO);

  ipVal = 0.0;
  i = icoord - (IMG_PAD_SIZE << 2);
  j = jcoord - (IMG_PAD_SIZE << 2);

  if(sub_pos)
  {
    if((sub_pos == 7)  && (filterNo == 0))		// g
    {	
      // Strong filter position
      static const int f0[4][4] = {{  0,  5,  5,  0},
      {  5, 22, 22,  5},
      {  5, 22, 22,  5},
      {  0,  5,  5,  0}};
      int sfp = 0;
      int l, k;
      for(l = 0; l < 4; ++l)
      {
        for(k = 0; k < 4; ++k)
        {
          pos_y = Clip3Fun(0, img_height - 1, i / 4 - 1 + l);
          if((i < 0) && (pos_y > 0))
            pos_y -= 1;
          pos_x = Clip3Fun(0, img_width - 1, j / 4 - 1 + k);
          if((j < 0) && (pos_x > 0))
            pos_x -=1;
          sfp += f0[l][k] * imgY[pos_y][pos_x];
        }
      }
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
      out = Clip3Fun(0, (1 << input->BitDepthLuma) - 1, (sfp + 64) >> 7);
#else
      out = Clip3Fun(0, 255, (sfp + 64) >> 7);
#endif
    }
    else
    {
#define FILTER_SIZE       6                     // Undefined at the end of the function
#define FILTER_OFFSET     (6 / 2 - 1)           // Undefined at the end of the function

      if(filterNo==3)
      {
        return GetInterpolatedPixel8TapSeperable(s, icoord, jcoord,img_width, img_height, sub_pos, filterNo,i,j);
      }

      {
        //  x a b c     00 01 02 03
        //  d e f g     04 05 06 07
        //  h i j k     08 09 10 11
        //  l m n o     12 13 14 15
        int ii_start, ii_end, jj_start, jj_end;

        if(filterNo == 1 || filterNo == 2)
        {
          ii_start=1;
          jj_start=1;
          ii_end=FILTER_SIZE-1;   //4x4 filter
          jj_end=FILTER_SIZE-1; 
        }
        else
        {
          ii_start=0;
          jj_start=0;
          ii_end=FILTER_SIZE;
          jj_end=FILTER_SIZE;          
        }

        if(sub_pos==1 || sub_pos==2 || sub_pos==3)  //a,b,c horizontal filter
        {
          ii_start=2;
          ii_end=3;
          jj_start=0;
          jj_end=FILTER_SIZE; 
        }
        if(sub_pos==4 || sub_pos==8 || sub_pos==12) //d,h,l vertical filter
        {
          ii_start=0;
          ii_end=FILTER_SIZE;
          jj_start=2;
          jj_end=3; 
        }
        for(ii = ii_start; ii < ii_end; ++ii)
        {
          for(jj = jj_start; jj < jj_end; ++jj)
          {
            pos_y = Clip3Fun(0, img_height - 1, i / 4 - FILTER_OFFSET + ii);
            if((i < 0) && (pos_y > 0))
              pos_y -= 1;
            pos_x = Clip3Fun(0, img_width - 1, j / 4 - FILTER_OFFSET + jj);
            if((j < 0) && (pos_x > 0))
              pos_x -=1;
            ipVal += (SIFO_FILTER[filterNo][sub_pos - 1][FILTER_SIZE * ii + jj] * imgY[pos_y][pos_x]);
          }
        }
      }
      out = Clip3Fun(0, img->max_imgpel_value, (int)(ipVal + 0.5));

#undef FILTER_OFFSET
#undef FILTER_SIZE

    }
  }
  else
  {
    out = imgY[Clip3Fun(0, img_height - 1, i / 4)][Clip3Fun(0, img_width - 1, j / 4)];
  }
  return out;
}

#ifdef GEOM
int mixed4x4(int x, int y, int mb_addr)
{
  int i, j;
  int segm = img->mb_data[mb_addr].geo_mask[y][x];
  for(i=0; i<4; i++)
    for(j=0; j<4; j++)
      if(segm != img->mb_data[mb_addr].geo_mask[y+j][x+i])
        return 1;

  return 0;
}
#endif
//
//
//
static void AccumulateError_P(int x_orig, int y_orig)
{
  int x, y, xj, yi, temp, valOrg;
  int mvx = 0,  mvy = 0;
  int sub_pos = 0, mvx_sub, mvy_sub;
  int ref_frame = 0;
  int subblock = 0;
  int out4Y_width  = (img->width + 2 * IMG_PAD_SIZE) * 4 - 1;  
  int out4Y_height = (img->height + 2 * IMG_PAD_SIZE) * 4 - 1;
  imgpel ** imgY;
  int f;
#ifdef GEOM
  int mb_addr = (y_orig>>4)*img->PicWidthInMbs + (x_orig>>4);
  Boolean geo_mb = img->mb_data[mb_addr].geo_mask[0][0]!=-1;
#endif
  //(x_orig,y_orig) is the top left corner pixel of a 16x16 MB
  //(x,y)           is the top left corner pixel of corresponding 4x4 block
  for(subblock = 0; subblock < 16; ++subblock)
  {
#ifdef GEOM
    x = 4 * (subblock % 4);
    y = 4 * (subblock / 4);

    if(geo_mb)
    {
      if(mixed4x4(x, y,mb_addr))//4x4 block with pixels belonging to both geo partitions
        continue;
    }
    x += x_orig;
    y += y_orig;
#else
    x = x_orig + 4 * (subblock % 4);
    y = y_orig + 4 * (subblock / 4);
#endif
    mvx = enc_picture->mv[LIST_0][y / 4][x / 4][0];
    mvy = enc_picture->mv[LIST_0][y / 4][x / 4][1];

    if(img->mv_res)
    {
      if ((mvx&1)!=0 || (mvy&1)!=0 )
      {  
        continue;
      }
      else
      {
        mvx /=2;
        mvy /=2;
      }
    }

    mvx_sub = (mvx >= 0)? mvx % 4: (4 - abs(mvx) % 4) % 4;
    mvy_sub = (mvy >= 0)? mvy % 4: (4 - abs(mvy) % 4) % 4;
    sub_pos = mvx_sub + 4 * mvy_sub;    // pos 0..15 in a 4x4 block 

    ref_frame = enc_picture->ref_idx[LIST_0][y / 4][x / 4];


    if(ref_frame != -1)
    {
      imgY = listX[LIST_0][ref_frame]->imgY;
      for(yi = 0; yi < 4; ++yi)
      {
        for(xj = 0; xj < 4; ++xj)
        {
          int xcoord = Clip3Fun(0, out4Y_width , 4 * (x + xj) + 4 * IMG_PAD_SIZE + mvx);
          int ycoord = Clip3Fun(0, out4Y_height, 4 * (y + yi) + 4 * IMG_PAD_SIZE + mvy);

          valOrg = imgY_org[y+yi][x+xj];

          for(f = 0; f < NUM_SIFO; ++f)
          {

            temp = valOrg - GetInterpolatedPixel(listX[LIST_0][ref_frame], ycoord, xcoord, img->width, img->height, sub_pos, f);
            AccErrorP[f][sub_pos] += temp * temp;
          }
        }
      }

    }
  }
}

int offsetIBDI(int offset, int yuv)
{
  if(img->BitDepthIncrease)
  {
    if(yuv == 0)
      offset = ((offset + (1<<(img->BitDepthIncrease-1))) >> img->BitDepthIncrease) << img->BitDepthIncrease;
    else
      offset = ((offset + (1<<(img->BitDepthIncreaseChroma-1))) >> img->BitDepthIncreaseChroma) << img->BitDepthIncreaseChroma;
  }

  return offset;
}


static void ComputeBestFilter_P(double err[NUM_SIFO][16], int combination[16])
{
  int bits = 0;
  int i;
  double errComb = 0.0;
  double errBest = 0.0;
  int spp;
  int best = -1;

  // Lowest error with a single filter
  for(i = 0; i < NUM_SIFO; ++i)
  {
    double errTmp = 0.0;
    for(spp = 1; spp < 16; ++spp)
      errTmp += err[i][spp];
    if((best == -1) || (errTmp < errBest))
    {
      best = i;
      errBest = errTmp;
    }
  }


  // Compute bit cost of sending the filters
  for(i = 1; i < 16; ++i)
    bits += 2; // SIFO4Filt

  // Error with filter combination
  for(spp = 1; spp < 16; ++spp)
    errComb += err[combination[spp]][spp];

  if(errBest < (errComb + img->lambda_md[img->type][img->qp] * (double)bits))
  {
    for(i = 0; i < 16; ++i)
    {
      combination[i] = best;
    }
    img->bestFilter = best + 1;
  }
  else
  {
    img->bestFilter = 0;
  }
}

void UpdateSequenceFilters_P(void)
{
  int i, f, ind;
  double minVal;

  // Filter selection based on the sequence 
  for(i = 0; i < 16; ++i)
  {
    for(f = 0; f < NUM_SIFO; ++f)
      SequenceAccErrorP[f][i] += AccErrorP[f][i];

    img->filterSequence[i] = 0;
    minVal = SequenceAccErrorP[0][i];
    for(ind = 1; ind < NUM_SIFO; ++ind)
    {
      if(SequenceAccErrorP[ind][i] < minVal)
      {
        img->filterSequence[i] = ind;
        minVal = SequenceAccErrorP[ind][i];
      }
    }
  }

  ComputeBestFilter_P(SequenceAccErrorP, img->filterSequence);
}


//
//
//



//
//  Accumulate squared error to compute best filter combination
//
static void AccumulateError_B(int x_orig, int y_orig)
{
  int x, y, xj, yi, valOrg, i, j;
  int mvx[2],  mvy[2];
  int sub_pos[2], mvx_sub, mvy_sub;
  int ref_frame[2];
  int subblock=0;
  int out4Y_width  = (img->width + 2 * IMG_PAD_SIZE) * 4 - 1;  
  int out4Y_height = (img->height + 2 * IMG_PAD_SIZE) * 4 - 1;
  double tmp;
  imgpel ** imgY0;
  imgpel ** imgY1;
  imgpel ** imgY;
  int f;

  int num_SIFO= NUM_SIFO;
#ifdef GEOM
  int mb_addr = (y_orig>>4)*img->PicWidthInMbs + (x_orig>>4);
  Boolean geo_mb = img->mb_data[mb_addr].geo_mask[0][0]!=-1;
#endif
  for(subblock = 0; subblock < 16; ++subblock)
  {
#ifdef GEOM
    x = 4 * (subblock % 4);
    y = 4 * (subblock / 4);

    if(geo_mb)
    {
      if(mixed4x4(x, y,mb_addr))//4x4 block with pixels belonging to both geo partitions
        continue;
    }
    x += x_orig;
    y += y_orig;
#else
    x = x_orig + 4 * (subblock % 4);
    y = y_orig + 4 * (subblock / 4);
#endif
    //get motion information
    for (i = 0; i < 2; ++i)
    {
      mvx[i] = enc_picture->mv[LIST_0+i][y/4][x/4][0];
      mvy[i] = enc_picture->mv[LIST_0+i][y/4][x/4][1];

      mvx_sub = (mvx[i] >= 0)? mvx[i] % 4: (4 - abs(mvx[i]) % 4) % 4;
      mvy_sub = (mvy[i] >= 0)? mvy[i] % 4: (4 - abs(mvy[i]) % 4) % 4;

      sub_pos[i] = mvx_sub + 4 * mvy_sub;    // Pos 0..15 in a 4x4 block
      ref_frame[i] = enc_picture->ref_idx[LIST_0+i][y/4][x/4];
    }

    if ((ref_frame[0] == -1) && (ref_frame[1] == -1))
      continue;     // Intra macroblocks are not used for calculation

    if ((ref_frame[0] != -1) && (ref_frame[1] != -1)) // Bi-predicted
    {
      int tempF[NUM_SIFO], tempB[NUM_SIFO];
      int fw_ref_idx = ref_frame[0];
      int bw_ref_idx = ref_frame[1];

      imgY0 = listX[LIST_0][fw_ref_idx]->imgY;
      imgY1 = listX[LIST_1][bw_ref_idx]->imgY;

      for(yi = 0; yi < 4; ++yi)
      {
        for(xj = 0; xj < 4; ++xj)
        {
          int xcoord0 = Clip3Fun(0 , out4Y_width,  4 * (x + xj) + 4 * IMG_PAD_SIZE + mvx[0]);
          int ycoord0 = Clip3Fun(0 , out4Y_height, 4 * (y + yi) + 4 * IMG_PAD_SIZE + mvy[0]);
          int xcoord1 = Clip3Fun(0 , out4Y_width,  4 * (x + xj) + 4 * IMG_PAD_SIZE + mvx[1]);
          int ycoord1 = Clip3Fun(0 , out4Y_height, 4 * (y + yi) + 4 * IMG_PAD_SIZE + mvy[1]);

          valOrg   = imgY_org[y + yi][x + xj];
          for(f = 0; f < num_SIFO; ++f)
          {
            tempF[f] = GetInterpolatedPixel(listX[LIST_0][fw_ref_idx], ycoord0, xcoord0, img->width, img->height, sub_pos[0], f);          
            tempB[f] = GetInterpolatedPixel(listX[LIST_1][bw_ref_idx], ycoord1, xcoord1, img->width, img->height, sub_pos[1], f);
          }
          for(i = 0; i < num_SIFO; ++i)
          {
            for(j = 0; j < num_SIFO; ++j)
            {
              tmp = valOrg - ((tempF[i] + tempB[j] + 1 - img->bipred_rounding_control) >> 1); 
              SequenceAccErrorB[i][j][sub_pos[0]][sub_pos[1]] += (tmp * tmp);
            }
          }
        } // x
      } // y
    }
    else //uni-directional
    {
      int ref_index = (ref_frame[0]!= -1)? LIST_0: LIST_1;
      int ind = (ref_frame[0]!= -1)? 0: 1;

      imgY = listX[ref_index][ref_frame[ind]]->imgY;

      for(yi = 0; yi < 4; ++yi)
      {    
        for(xj = 0; xj < 4; ++xj)
        {
          int xcoordind = Clip3Fun(0 , out4Y_width,  4 * (x + xj) + 4 * IMG_PAD_SIZE + mvx[ind]);
          int ycoordind = Clip3Fun(0 , out4Y_height, 4 * (y + yi) + 4 * IMG_PAD_SIZE + mvy[ind]);

          valOrg = imgY_org[y+yi][x+xj];

          for(i = 0; i < num_SIFO; ++i)
          {
            // Unidirectional error is accumulated along the diagonal
            tmp = valOrg - GetInterpolatedPixel(listX[ref_index][ref_frame[ind]], ycoordind, xcoordind, img->width, img->height, sub_pos[ind], i);
            SequenceAccErrorB[i][i][sub_pos[ind]][sub_pos[ind]] += (tmp * tmp);
          }
        } // x
      } // y
    } // if bidirectional
  } // for(subblock = 0; subblock < 16; subblock++)
}

void SwapFilteredFrames(void)
{
  unsigned frame;

  for(frame = 0; frame < dpb.ref_frames_in_buffer; ++frame)
  {
    dpb.fs_ref[frame]->frame->p_imgY_filt = dpb.fs_ref[frame]->frame->imgY_filt;
    dpb.fs_ref[frame]->frame->imgY_filt = dpb.fs_ref[frame]->frame->imgY;
    dpb.fs_ref[frame]->frame->imgY = dpb.fs_ref[frame]->frame->p_imgY_filt;

    dpb.fs_ref[frame]->frame->p_imgY_11_filt = dpb.fs_ref[frame]->frame->imgY_11_filt;
    dpb.fs_ref[frame]->frame->imgY_11_filt = dpb.fs_ref[frame]->frame->imgY_11;
    dpb.fs_ref[frame]->frame->imgY_11 = dpb.fs_ref[frame]->frame->p_imgY_11_filt;
  }
}

static void ScanMacroblocksAndApplyFunction(void fun(int, int))
{
  int i, j;

  if((img->type == P_SLICE) || (img->type == B_SLICE))
  {
    // Loop over all macroblocks and call fun()
    for (i = 0; i < img->height; i = i + MB_BLOCK_SIZE)
      for (j = 0; j < img->width; j = j + MB_BLOCK_SIZE)
        fun(j, i);
  }
}

void UpdateSequenceFilters_B(void)
{
  int i;
  for(i = 0; i < 16; ++i)
    img->filterSequence[i] = SequenceBestCombFilterB[i];
}

int ComputeFiltersAndOffsets(void)
{
  if(img->type == I_SLICE)
    return 0;

  ResetAll(img->type);
  if(img->type == P_SLICE)
  {
    ScanMacroblocksAndApplyFunction(AccumulateError_P);       // Accumulate squared error for filter decision and offsets
    UpdateSequenceFilters_P();
  }
  else if(img->type == B_SLICE)
  {
    ScanMacroblocksAndApplyFunction(AccumulateError_B);       // Accumulate squared error for frame and sequence
    ComputeFilterCombination_B();                             // Compute filter combination (unique filter per encoded frame)
    UpdateSequenceFilters_B();
  }
  return 1; 
}

void initSIFOFilters()
{
  static int first = 1;

  if(first == 1)
  {  
    int i;
    first = 0;

    memset(img->filterSequence, 0, 16 * sizeof(int));

    for(i = 0; i < 2; ++i)
    {
      memset(img->subpelOffset[i], 0, 16 * sizeof(int));
      memset(img->imgOffset[i], 0, MAX_REFERENCE_PICTURES * sizeof(int));
    }

    // Assign filters
    for(i = 0; i < 15; ++i)
    {
      int j;
      for(j = 0; j < SQR_FILTER; ++j)
      {
        SIFO_FILTER[0][i][j] = STANDARD_2D_FILTER[i][j];
        SIFO_FILTER[1][i][j] = SYMMETRIC_1[i][j];
        SIFO_FILTER[2][i][j] = SYMMETRIC_2[i][j];
      }
    }

  }
}

static void UnifiedOneForthPixSwitchedFilters(int *filter, int *subpelOffset, int *imgOffset, int list)
{
  int i, j, offset;
  int frame;
  imgpel **out4Y;
  int img_width, img_height;
  int sub_pos, x_sub, y_sub;
  int icoord, jcoord;
  imgpel ** imgY;

  if(img->type==B_SLICE && list==1)
  {
    if(listX[0][0]->frame_poc == listX[1][0]->frame_poc) 
      return;
  }

  for(frame = 0; frame < listXsize[list]; ++frame)
  {
    img_width  = listX[list][frame]->size_x;
    img_height = listX[list][frame]->size_y;

    out4Y  = listX[list][frame]->imgY_ups;
    imgY = listX[list][frame]->imgY;

    for(i = -4 * IMG_PAD_SIZE; i < (IMG_PAD_SIZE + img_height) * 4; ++i)
    {
      for(j = -4 * IMG_PAD_SIZE; j < (IMG_PAD_SIZE + img_width) * 4; ++j)
      {
        icoord = i + (IMG_PAD_SIZE << 2);
        jcoord = j + (IMG_PAD_SIZE << 2);

        x_sub = jcoord % 4;             // x-sub-coordinate in a 4x4block
        y_sub = icoord % 4;             // y-sub-coordinate in a 4x4block

        sub_pos = x_sub + (y_sub << 2);    // pos 0..15 in a 4x4 block
        offset = (frame == 0)? subpelOffset[sub_pos]: imgOffset[frame];

        assert(filter[sub_pos] < NUM_SIFO);

        out4Y[icoord][jcoord] = (int) Clip3Fun(0, img->max_imgpel_value, offset + GetInterpolatedPixel(listX[list][frame], icoord, jcoord, img_width, img_height, sub_pos, filter[sub_pos]));
      }
    }

    offset = (frame == 0)? subpelOffset[0]: imgOffset[frame];

    for(i = 0; i < img_height; ++i)
      for(j = 0; j < img_width; ++j)
        listX[list][frame]->imgY_filt[i][j] = (int)Clip3(0, img->max_imgpel_value, listX[list][frame]->imgY[i][j] + offset);

    GenerateFullPelRepresentation(out4Y, listX[list][frame]->imgY_11_filt, img_width, img_height);  
  }
}


void UnifiedOneForthPixFiltSel(int filterParam)
{
  if(img->type == P_SLICE)
  {
    if(filterParam == SIFO_FIRST_PASS_FPO)
    {
      // SIFO_FIRST_PASS_FPO -> Sequence filters and first pass offsets
      UnifiedOneForthPixSwitchedFilters(img->filterSequence, img->subpelOffset[0][LIST_0], img->imgOffset[0][LIST_0], LIST_0);
    }
  }
  else if(img->type == B_SLICE)
  {
    if(filterParam == SIFO_FIRST_PASS_FPO)
    {
      // SIFO_FIRST_PASS_FPO -> Sequence filters and first pass offsets
      UnifiedOneForthPixSwitchedFilters(img->filterSequence, img->subpelOffset[0][LIST_0], img->imgOffset[0][LIST_0], LIST_0);
      UnifiedOneForthPixSwitchedFilters(img->filterSequence, img->subpelOffset[0][LIST_1], img->imgOffset[0][LIST_1], LIST_1);
    }
  }
  else
  {
    assert(0);    // Called on a I-slice?
  }
}


int getNonzero(int yuv, int list)
{
  int sub_pos;
  int frame;
  for(frame = 0; frame < listXsize[list]; ++frame)
  {       
    if(frame == 0)
    {     
      for(sub_pos = 0; sub_pos < 16; ++sub_pos)   
        if(img->subpelOffset[yuv][list][sub_pos] != 0)
          return 1;
    }
    else
    {
      if(img->imgOffset[yuv][list][frame] != 0)
        return 1;
    }
  }
  return 0;
}

#endif




