/* 
====================================================================================================================

The copyright in this software is being made available under the License included below. 
This software may be subject to other third party and 	contributor rights, including patent rights, and no such 
rights are granted under this license.

Copyright (c) 2010, QUALCOMM INC.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted only for 
the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
promoting such standards. The following conditions are required to be met:

* Redistributions of source code must retain the above copyright notice, this list of conditions and 
the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
the following disclaimer in the documentation and/or other materials provided with the distribution.
* The name of QUALCOMM INC.  may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

====================================================================================================================
*/
#include <math.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "global.h"
#include "mbuffer.h"
#include "elements.h"
#include "errorconcealment.h"
#include "macroblock.h"
#include "fmo.h"
#include "cabac.h"
#include "vlc.h"
#include "image.h"
#include "mb_access.h"
#include "biaridecod.h"

#include "transform8x8.h"
#ifdef MV_COMPETITION
#include "mv_competition.h"
extern MV_Competition mv_comp;                  // to each possible predictor
#endif
#include "switched_filters.h"

#if TRACE
#define TRACE_STRING(s) strncpy(currSE.tracestring, s, TRACESTRING_SIZE)
#else
#define TRACE_STRING(s) // do nothing
#endif

extern int last_dquant;
extern ColocatedParams *Co_located;

static void get_eighthpel_blockUV_direct(imgpel **  ref, int x_pos, int y_pos, ImageParameters *img, int block[32][32], int geometry, int size_x, int size_y);

#ifdef ADAPTIVE_MV_RES
short roundInt(short mv,int mvres)
{
  short roundMV=mv, sign;
  sign=mv>0? 1 : -1;
  roundMV*=sign;
  if (mvres ==1)
    roundMV=roundMV/2;
  else if (mvres ==2)
    roundMV=roundMV/4;
  roundMV*=sign;
  return(roundMV);
}

int ref_idx_mvshift(int mode, int se_value, int *mv_shift)
{
  if(mode > 4)
  {
    *mv_shift = 1;
    return se_value;
  }
  else
  {
    if(se_value <= 1)
    {
      *mv_shift = !se_value;
      return 0;
    }
    else 
    {
      *mv_shift = 1;
      return (se_value-1);
    }
  }
}

#endif

extern int no64x64, no32x32;
int cabac_debug=0;
extern int delta_qp_sent;
Macroblock  MB32;
Macroblock  MB64;

short saved_pmvfw[2], saved_pmvbw[2];
char saved_fw_rFrame,saved_bw_rFrame;
int mb32_mvd[2][BLOCK_MULTIPLE<<MAX_MB_EXT_LEVEL][BLOCK_MULTIPLE<<MAX_MB_EXT_LEVEL][2];          //!< indices correspond to [forw,backw][block_y][block_x][x,y]

void readRefFrame_CABAC32( SyntaxElement *se,
struct inp_par *inp,
struct img_par *img,
  DecodingEnvironmentPtr dep_dp);
void readMVD_CABAC32( SyntaxElement *se,
struct inp_par *inp,
struct img_par *img,
  DecodingEnvironmentPtr dep_dp);
void readCBP_CABAC32(SyntaxElement *se,
struct inp_par *inp,
struct img_par *img,
  DecodingEnvironmentPtr dep_dp);
void readCBP_CABAC_chroma(SyntaxElement *se,
struct inp_par *inp,
struct img_par *img,
  DecodingEnvironmentPtr dep_dp);
void readCBP_CABAC_luma1bit(SyntaxElement *se,
struct inp_par *inp,
struct img_par *img,
  DecodingEnvironmentPtr dep_dp);

void readRunLevel_CABAC16x16(SyntaxElement *se, struct inp_par *inp, struct img_par *img,  DecodingEnvironmentPtr dep_dp);
void read_mbtype(struct img_par *img, struct inp_par *inp, int mb_ext_level);
int Not_MB16_3(int mb_nr);
extern void CheckAvailabilityOfNeighbors_Z(void);
extern int BType2CtxRef (int btype);

#ifdef GEOM
#define _USE_MATH_DEFINES
#include <math.h>
char ****geo_mask[3];
char ****geo_4x4mask[3];//[mb_ext_level][rho][theta][y][x]
char ****geo_filter[3];//[mb_ext_level][rho][theta][y][x]
char ****geo_filterUV[3];//[mb_ext_level][rho][theta][y][x]
short ***mv_array_geo[2];
char geo_mv_predictor[4][32][32][2][6];//[0-8x8;1-16x16;2-32x32;3-64x64][rho][theta][segm][predictors]
char enc_geo_valid[GEO_PART_NUM];
int **geo_mask16_neighbor;
short geo_mv[2][2][2];
int Max_rho[3];//[mb_ext_level]
int rho_bits[3];//[mb_ext_level]
int mb32_mvd[2][BLOCK_MULTIPLE<<MAX_MB_EXT_LEVEL][BLOCK_MULTIPLE<<MAX_MB_EXT_LEVEL][2];          //!< indices correspond to [forw,backw][block_y][block_x][x,y]
int mb16_pos[][2] = {{0,0},{16,0},{0,16}, {16,16},   {32,0},{48,0},{32,16},{48,16},
{0,32},{16,32},{0,48}, {16,48}, {32,32}, {48,32}, {32,48},{48,48}};
void readMB_typeInfo_CABAC_geo( SyntaxElement *se,
struct inp_par *inp,
struct img_par *img,
  DecodingEnvironmentPtr dep_dp);
void readRefFrame_CABAC32( SyntaxElement *se,
struct inp_par *inp,
struct img_par *img,
  DecodingEnvironmentPtr dep_dp);
void readMVD_CABAC32( SyntaxElement *se,
struct inp_par *inp,
struct img_par *img,
  DecodingEnvironmentPtr dep_dp);
int get_mem3Dshort(short ****array3D, int frames, int rows, int columns);
int ResetAvailabilityOfNeighbors32(int mb_x, int mb_y, int *reset_block_x, int *reset_block_y, int blockshape_x, int blockshape_y);

typedef enum 
{
  P_A = 0,
  P_B,
  P_C,
  P_D,
  P_E,
  P_F,
  P_M //median
} PREDTYPE;

PREDTYPE geo_segm_mvp[][2]=
{ //  0          1          2          3          4          5          6          7          8          9          10          11
  {P_A,P_F}, {P_A,P_F}, {P_E,P_M}, {P_E,P_B}, {P_E,P_F}, {P_E,P_F}, {P_E,P_M}, {P_E,P_F}, {P_A,P_F}, {P_M,P_C}, {P_A, P_F}, {P_M, P_C}
};
int MB16GeoMask[GEO_PART_NUM][16]={
  {1,1,1,0,1,1,1,0,1,1,1,0,1,1,1,0},
  {1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0},
  {0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1},
  {0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1},
  {0,0,0,0,1,0,0,0,1,1,0,0,1,1,1,0},
  {1,0,0,0,1,1,0,0,1,1,1,0,1,1,1,1},
  {0,0,0,0,0,0,0,0,1,0,0,0,1,1,0,0},
  {1,1,0,0,1,1,1,0,1,1,1,1,1,1,1,1},
  {1,1,1,0,1,1,0,0,1,0,0,0,0,0,0,0},
  {1,1,1,1,1,1,1,0,1,1,0,0,1,0,0,0},
  {1,1,0,0,1,0,0,0,0,0,0,0,0,0,0,0},
  {1,1,1,1,1,1,1,1,1,1,1,0,1,1,0,0}
};
double sinLUT32[32] = {0.0000000000, 0.1950903220, 0.3826834324, 0.5555702330, 0.7071067812, 
0.8314696123, 0.9238795325, 0.9807852804, 1.0000000000, 0.9807852804, 0.9238795325, 0.8314696123,
0.7071067812, 0.5555702330, 0.3826834324, 0.1950903220, 0.0000000000, -0.1950903220, -0.3826834324,
-0.5555702330, -0.7071067812, -0.8314696123, -0.9238795325, -0.9807852804, -1.0000000000, -0.9807852804,
-0.9238795325, -0.8314696123, -0.7071067812, -0.5555702330, -0.3826834324, -0.1950903220}; //step = 11.25 deg

double cosLUT32[32] = {
  1.0000000000, 0.9807852804, 0.9238795325, 0.8314696123, 0.7071067812, 0.5555702330, 0.3826834324,
  0.1950903220, 0.0000000000, -0.1950903220, -0.3826834324, -0.5555702330, -0.7071067812, -0.8314696123, 
  -0.9238795325, -0.9807852804, -1.0000000000, -0.9807852804, -0.9238795325, -0.8314696123, -0.7071067812, 
  -0.5555702330, -0.3826834324, -0.1950903220, -0.0000000000, 0.1950903220, 0.3826834324, 0.5555702330, 0.7071067812,
  0.8314696123, 0.9238795325, 0.9807852804}; //step = 11.25 deg
  int cotLUT32[32]= {
    6689275, 41184, 19777, 12260, 8192, 5474, 3393, 1629, 0, -1629, -3393, -5474, -8192, -12260, -19777,  -41184,
    -6689275, 41184, 19777, 12260, 8192, 5474, 3393, 1629, 0, -1629, -3393, -5474, -8192, -12260, -19777,  -41184
  };
  int inv_sinLUT32[32]= {
    6689275, 41991, 21407, 14745, 11585, 9852, 8867, 8352, 8192, 8352, 8867, 9852, 11585, 14745, 21407, 41991, 
    -6689275,-41991,-21407,-14745,-11585,-9852,-8867,-8352,-8192,-8352,-8867,-9852,-11585,-14745,-21407,-41991  
  };


  int best8x8pdir_geo[2], best8x8fwref_geo[2], best8x8bwref_geo[2];

  int get_mem2Dchar(char ***array2D, int rows, int columns)
  {
    int i;

    if((*array2D      = (char**)calloc(rows,        sizeof(char*))) == NULL)
      error("get_mem2Dchar: array2D", -1);
    if(((*array2D)[0] = (char* )calloc(rows*columns,sizeof(char ))) == NULL)
      error("get_mem2Dchar: array2D", -1);

    for(i=1 ; i<rows ; i++)
      (*array2D)[i] =  (*array2D)[i-1] + columns  ;

    return rows*columns*sizeof(char);
  }

  int get_mem3Dchar(char ****array3D, int dim1, int dim2, int dim3)
  {
    int  j;

    if(((*array3D) = (char***)calloc(dim1,sizeof(char**))) == NULL)
      error("get_mem3Dchar: array3D", -1);

    for(j=0;j<dim1;j++)
      get_mem2Dchar( (*array3D)+j, dim2, dim3 ) ;

    return dim1*dim2*dim3*sizeof(char);
  }

  int get_mem4Dchar(char *****array4D, int dim1, int dim2, int dim3, int dim4)
  {
    int  j;

    if(((*array4D) = (char****)calloc(dim1,sizeof(char**))) == NULL)
      error("get_mem4Dchar: array4D", -1);

    for(j=0;j<dim1;j++)
      get_mem3Dchar( (*array4D)+j, dim2, dim3, dim4) ;

    return dim1*dim2*dim3*dim4*sizeof(char);
  }


  void free_mem2Dchar(char **array2D)
  {
    if (array2D)
    {
      if (array2D[0]) 
        free (array2D[0]);
      else error ("free_mem2Dchar: trying to free unused memory",100);

      free (array2D);

    } else
    {
      error ("free_mem2Dchar: trying to free unused memory",100);
    }
  }

  void free_mem3Dchar(char ***array3D, int frames)
  {
    int i;

    if (array3D)
    {
      for (i=0;i<frames;i++)
      { 
        free_mem2Dchar(array3D[i]);
      }
      free (array3D);
    } else
    {
      error ("free_mem3Dchar: trying to free unused memory",100);
    }
  }

  void free_mem4Dchar(char ****array4D, int idx, int frames )
  {
    int  j;

    if (array4D)
    {
      for(j=0;j<idx;j++)
        free_mem3Dchar( array4D[j], frames) ;
      free (array4D);
    } else
    {
      error ("free_mem4Dchar: trying to free unused memory",100);
    }
  }
#ifdef OBMC
  void get_mask(char **mask, char **mask4x4, char **filter, char **filterUV, int rho, int theta_idx, float theta_step, int mb_ext_level)
#else
  void get_mask(char **mask, char **mask4x4, int rho, int theta_idx, float theta_step, int mb_ext_level)
#endif
  {
#ifndef SIN_LUT
    double theta = theta_idx * theta_step;
#endif
    int i, j, x,y, segm, pred;
    //double sx, sy, rx, ry;
    //double rz;
    int sx_i,sy_i;
    int rx;
    //a  b  c  d  e  f
    static int x0[]={0, 0, 3, 0, 0, 2}; //(x0,y0) and (x1,y1) are the coordinators in a 4x4 block
    static int y0[]={3, 3, 3, 3, 1, 3};
    static int x1[]={0, 1, 3, 0, 0, 3};
    static int y1[]={2, 3, 3, 3, 0, 3};
    int coor[4];
    int blocksize = 16<<mb_ext_level;
#ifdef OBMC
    int blocksizeUV;
    int ii, jj, num[2];
    int xn[]={-1,0, 0,1};//4 connect
    int yn[]={ 0,1,-1,0};//4 connect
#endif

    //coor is used to convert the coordinators in a 4x4 block to NxN block
    if(blocksize == 8 || blocksize == 16 || blocksize==32 || blocksize==64)
    {
      coor[0] = 0; coor[1] = 3; coor[2] = blocksize-4; coor[3] = blocksize-1;
    }
    else
    {
      error ("blocisize is not correct!", -1);
    }

#ifndef SIN_LUT
    theta = theta*2*M_PI/360;//convert degree to radiant
#endif

    for(x=0;x<blocksize;x++)
    {
      for(y=0;y<blocksize;y++)
      {
        sx_i = (x*2-blocksize+1);
        sy_i = (y*2-blocksize+1);

        if (theta_idx ==0)
          rx = sx_i-rho*2;
        else if (theta_idx==16)
          rx = -sx_i-rho*2;
        else
        {
          rx = (sx_i*cotLUT32[theta_idx]+sy_i*8192-rho*2*inv_sinLUT32[theta_idx]);
          if (inv_sinLUT32[theta_idx]<0)
            rx=-rx;
        }

        //      sx = (x-blocksize/2+0.5);
        //      sy = (y-blocksize/2+0.5);
        //      
        //      // coordinators after rotation
        //#ifdef SIN_LUT
        //      rx =  sx*cosLUT32[theta_idx] + sy*sinLUT32[theta_idx] - rho;
        //      ry = -sx*sinLUT32[theta_idx] + sy*cosLUT32[theta_idx];
        //#else
        //      rx =  sx*cos(theta) + sy*sin(theta) - rho;
        //      ry = -sx*sin(theta) + sy*cos(theta);
        //#endif

        if (rx < 0)
        {
          mask[y][x]=0;
          //plot(sx, sy, '.');
        }
        else
        {
          mask[y][x]=1;
          //plot(sx, sy, 'o');
        }
      }
    }

    for(segm=0; segm<2; segm++)
    {
      for(pred=0; pred<6;pred++)
      {
        if(mask[coor[y0[pred]]] [coor[x0[pred]]] == segm && mask[coor[y1[pred]]] [coor[x1[pred]]] == segm)
        {
          geo_mv_predictor[mb_ext_level+1][rho][theta_idx][segm][pred]  = 1;        
        }
        else
        {
          geo_mv_predictor[mb_ext_level+1][rho][theta_idx][segm][pred]  = 0;        
        }
      }
    }

    for(x=0;x<blocksize;x++)
    { 
      for(y=0;y<blocksize/2;y++)
      {
        int temp = mask[y][x];
        mask[y][x] = mask[blocksize-1-y][x];
        mask[blocksize-1-y][x] = temp;
      }
    }

    //the following assignment is to solve the mismatch between Window and Linux
    if(mb_ext_level ==0)
    {
      if(rho ==0 && theta_idx == 4)
      {
        mask[1][1] = 1;
        mask[14][14] = 0;
      }
    }
    else if(mb_ext_level ==1)
    {
      if(rho ==0 && theta_idx == 4)
      {
        mask[4][4] = 0;
        mask[9][9] = 1;
        mask[22][22] = 0;
      }
      else if(rho ==0 && theta_idx == 12)
        mask[27][4] = 0;
    }
    else
    {
      if(rho ==0 && theta_idx == 4)
      {
        mask[3][3] = 1;
        mask[6][6] = 0;
        mask[8][8] = 1;
        mask[20][20] = 0;
        mask[25][25] = 1;
        mask[38][38] = 0;
        mask[55][55] = 0;
        mask[60][60] = 0;
      }
      else if(rho ==0 && theta_idx == 12)
      {
        mask[43][20] = 0;
        mask[57][6] = 0;
      }
    }

    for(x=0;x<blocksize;x+=4)
    {
      for(y=0;y<blocksize;y+=4)
      {
        segm = mask[y][x];
        mask4x4[y][x] = segm;

        for(i=x; i<x+4; i++)
        {
          for(j=y; j<y+4; j++)
          {
            if(mask[j][i] != segm)
            {
              mask4x4[y][x] = 2;
              break;
            }
          }
          if(mask4x4[y][x] == 2)
            break;
        }
      }
    }

#ifdef OBMC
    for(x=0;x<blocksize;x++)
    {
      for(y=0;y<blocksize;y++)
      {
        num[0] = 0; num[1] = 0;
#if 1
        for(i=0; i<4; i++)//4-connect
        {
          ii = x+xn[i];
          if(ii<0)          ii=0;
          if(ii>=blocksize) ii=blocksize-1;
          jj = y+yn[i];
          if(jj<0)          jj=0;
          if(jj>=blocksize) jj=blocksize-1;
          num[mask[jj][ii]]++;
        }

        if(num[0]>0 && num[1]>0)
        {
          filter[y][x] = 1+mask[y][x];
        }
        else
          filter[y][x] = 0;
#else
        for(i=-1; i<=1; i++)
        {
          for(j=-1; j<=1; j++)
          {
            ii = x+i; 
            if(ii<0)          ii=0;
            if(ii>=blocksize) ii=blocksize-1;
            jj = y+j;
            if(jj<0)          jj=0;
            if(jj>=blocksize) jj=blocksize-1;

            num[mask[jj][ii]]++;
          }
        }

        if(num[0]>1 && num[1]>1)
        {
          filter[y][x] = 1;
        }
        else
          filter[y][x] = 0;
#endif
      }
    }

    blocksizeUV = blocksize/2;
    for(x=0;x<blocksizeUV;x++)
    {
      for(y=0;y<blocksizeUV;y++)
      {
        num[0] = 0; num[1] = 0;
#if 1
        for(i=0; i<4; i++)
        {
          ii = x+xn[i];
          if(ii<0)          ii=0;
          if(ii>=blocksizeUV) ii=blocksizeUV-1;
          jj = y+yn[i];
          if(jj<0)          jj=0;
          if(jj>=blocksizeUV) jj=blocksizeUV-1;
          num[mask[jj<<1][ii<<1]]++;
        }

        if(num[0]>0 && num[1]>0)
        {
          filterUV[y][x] = 1+mask[y<<1][x<<1];
        }
        else
          filterUV[y][x] = 0;
#else
        for(i=-1; i<=1; i++)
        {
          for(j=-1; j<=1; j++)
          {
            ii = x+i;
            if(ii<0)          ii=0;
            if(ii>=blocksizeUV) ii=blocksizeUV-1;
            jj = y+j;
            if(jj<0)          jj=0;
            if(jj>=blocksizeUV) jj=blocksizeUV-1;

            num[mask[jj<<1][ii<<1]]++;
          }
        }      

        if(num[0]>1 && num[1]>1)
          filterUV[y][x] = 1;
        else
          filterUV[y][x] = 0;
#endif
      }
    }
#endif //OBMC
    //if(blocksize == 32)
    //{
    //  printf("rho = %d theta_idx = %d\n", rho, theta_idx);
    //  for(y=0;y<blocksize;y++)
    //  {
    //    for(x=0;x<blocksize;x++)
    //      printf("%d ", mask[y][x]);
    //    printf("\n");
    //  }
    //  printf("predictor\n");
    //  for(segm=0; segm<2; segm++)
    //  {
    //    for(pred=0; pred<6;pred++)
    //      printf("%d ", geo_mv_predictor[blocksize/8-1][rho][theta_idx][segm][pred]);
    //    printf("\n");
    //  }
    //}
  }

#ifdef OBMC
  void geo_partition(char ****mask, char ****mask4x4, char ****filter, char ****filterUV, int mb_ext_level)
#else
  void geo_partition(char ****mask, char ****mask4x4, int mb_ext_level)
#endif
  {

    float theta_step = THETASTEP;//in unit of degree
    int rho;// = int8(sqrt(2)*blocksize/2);
    int theta_idx;
    //int blocksize = 16<<mb_ext_level;


    switch(mb_ext_level)
    {
    case 0:
    case 1:
    case 2:
      Max_rho[mb_ext_level] = ((16<<mb_ext_level)>>1) - 1; //MAX_RHO_16;
      rho_bits[mb_ext_level] = 3+mb_ext_level;
      break;
    default:
      error("this blocksize is not supported", -1);
    }

    /*******
    case 1
    *******/
    rho=0;
    for(theta_idx=0;theta_idx<=15;theta_idx++)
    {
#ifdef OBMC
      get_mask(mask[rho][theta_idx], mask4x4[rho][theta_idx], filter[rho][theta_idx], filterUV[rho][theta_idx], rho, theta_idx, theta_step, mb_ext_level);
#else
      get_mask(mask[rho][theta_idx], mask4x4[rho][theta_idx], rho, theta_idx, theta_step, mb_ext_level);
#endif
    }


    /*******
    case 2
    *******/
    for (rho=1;rho<=Max_rho[mb_ext_level];rho++)
    {
      for(theta_idx=0;theta_idx<=(1<<THETA_BITS)-1;theta_idx++)
      {
#ifdef OBMC
        get_mask(mask[rho][theta_idx], mask4x4[rho][theta_idx], filter[rho][theta_idx], filterUV[rho][theta_idx], rho, theta_idx, theta_step, mb_ext_level);
#else
        get_mask(mask[rho][theta_idx], mask4x4[rho][theta_idx], rho, theta_idx, theta_step, mb_ext_level);
#endif
      }
    }

  }

  void init_geo()
  {
    int i;
#if GEO_SCHEME == 0 
    img->geo_scheme = 0;
#else
    img->geo_scheme = 1;
#endif

    if(img->geo_scheme != 0)
    {
      for(i=0; i<3; i++)//3 MB levels: 16, 32, 64
      {
        get_mem4Dchar(&geo_mask[i], ((16<<i)>>1), 32, 16<<i, 16<<i);
        get_mem4Dchar(&geo_4x4mask[i], ((16<<i)>>1), (1<<THETA_BITS), 16<<i, 16<<i);
#ifdef OBMC
        get_mem4Dchar(&geo_filter[i], ((16<<i)>>1), (1<<THETA_BITS), 16<<i, 16<<i);
        get_mem4Dchar(&geo_filterUV[i], ((16<<i)>>1), (1<<THETA_BITS), 8<<i, 8<<i);
        geo_partition(geo_mask[i], geo_4x4mask[i], geo_filter[i], geo_filterUV[i], i);
#else
        geo_partition(geo_mask[i], geo_4x4mask[i], i);
#endif
      }
    }


    //for(i=0; i<3; i++)//3 MB levels: 16, 32, 64
    //{
    //  int x, y, theta_idx, rho=0;
    //  for(theta_idx=0;theta_idx<=15;theta_idx++)
    //  {
    //    printf("mb_ext_level = %d rho = %d theta_idx = %d\n", i, rho, theta_idx);
    //    for(y=0;y<(16<<i);y++)
    //    {
    //      for(x=0;x<(16<<i);x++)
    //        printf("%d ", geo_mask[i][0][theta_idx][y][x]);
    //      printf("\n");
    //    }
    //  
    //  }
    //  /*******
    //   case 2
    //  *******/
    //  for (rho=1;rho<=Max_rho[i];rho++)
    //  {
    //    for(theta_idx=0;theta_idx<=(1<<THETA_BITS)-1;theta_idx++)
    //    {
    //      printf("mb_ext_level = %d rho = %d theta_idx = %d\n", i, rho, theta_idx);
    //      for(y=0;y<(16<<i);y++)
    //      {
    //        for(x=0;x<(16<<i);x++)
    //          printf("%d ", geo_mask[i][rho][theta_idx][y][x]);
    //        printf("\n");
    //      }
    //    }
    //  }

    //}

    for(i=0; i<2; i++)
      get_mem3Dshort(&(mv_array_geo[i]), img->height / BLOCK_SIZE, img->width / BLOCK_SIZE, 2);

    //get_neighborhood();
  }

  PREDTYPE find_segm_mvp(int mb_ext_level, int geo_mode, int segm, int blocksize, PixelPos *block_a, PixelPos *block_b, PixelPos *block_c, PixelPos *block_d)
  {
    int rho = geo_mode>>5, theta_idx = geo_mode & 31;
    int a,b,c,d,e,f;

    a = geo_mv_predictor[mb_ext_level+1][rho][theta_idx][segm][0] & block_a->available;
    b = geo_mv_predictor[mb_ext_level+1][rho][theta_idx][segm][1] & block_b->available;
    c = geo_mv_predictor[mb_ext_level+1][rho][theta_idx][segm][2] & block_c->available;
    d = geo_mv_predictor[mb_ext_level+1][rho][theta_idx][segm][3] & block_d->available;
    e = geo_mv_predictor[mb_ext_level+1][rho][theta_idx][segm][4] & block_a->available;
    f = geo_mv_predictor[mb_ext_level+1][rho][theta_idx][segm][5] & block_b->available;

    if(a && b && (c||d))
      return P_M;
    else if(a)
      return P_A;
    else if(b)
      return P_B;
    else if(e)
      return P_E;
    else if(f)
      return P_F;
    else if(d){
      error("no P_D available\n", -1);//return P_D;
      return -1;
    }
    else if(c)
      return P_C;
    else 
      return P_M;//error("no neighbors available \n", -1);

  }

  void SegmMVPredictor(int mb_ext_level, PREDTYPE geo_segm_pred, short pmv[2], PixelPos *block_a, PixelPos *block_b, PixelPos *block_c, PixelPos *block_d,
    short  ***tmp_mv)
  {
    int block_offset = 3<<mb_ext_level;

    //if(mb_ext_level==0)
    {
      switch(geo_segm_pred)
      {
      case P_A:
        pmv[0] = block_a->available  ? tmp_mv[block_a->pos_y][block_a->pos_x][0] : 0;
        pmv[1] = block_a->available  ? tmp_mv[block_a->pos_y][block_a->pos_x][1] : 0;
        break;
      case P_B:
        pmv[0] = block_b->available  ? tmp_mv[block_b->pos_y][block_b->pos_x][0] : 0;
        pmv[1] = block_b->available  ? tmp_mv[block_b->pos_y][block_b->pos_x][1] : 0;
        break;
      case P_C:
        pmv[0] = block_c->available  ? tmp_mv[block_c->pos_y][block_c->pos_x][0] : 0;
        pmv[1] = block_c->available  ? tmp_mv[block_c->pos_y][block_c->pos_x][1] : 0;
        break;
      case P_D:
        pmv[0] = block_d->available  ? tmp_mv[block_d->pos_y][block_d->pos_x][0] : 0;
        pmv[1] = block_d->available  ? tmp_mv[block_d->pos_y][block_d->pos_x][1] : 0;
        break;
      case P_E:
        pmv[0] = block_a->available  ? tmp_mv[block_a->pos_y+block_offset][block_a->pos_x][0] : 0;
        pmv[1] = block_a->available  ? tmp_mv[block_a->pos_y+block_offset][block_a->pos_x][1] : 0;
        break;
      case P_F:
        pmv[0] = block_b->available  ? tmp_mv[block_b->pos_y][block_b->pos_x+block_offset][0] : 0;
        pmv[1] = block_b->available  ? tmp_mv[block_b->pos_y][block_b->pos_x+block_offset][1] : 0;
        break;
      default:
        error("geo_segm_mvp[geo][segm] is out of range!", -1);
      }
    }
    /*else
    {
    error("mb_ext_level!=0", -1);
    }*/
  }

  /*!
  ************************************************************************
  * \brief
  *    Set motion vector predictor
  ************************************************************************
  */
  void SetMotionVectorPredictor_geo (int geo_mode, 
    int segm,
  struct img_par  *img,                               
    short           *pmv_x,
    short           *pmv_y,
    char            ref_frame,
    byte            list,
    char            ***refPic,
    short           ****tmp_mv,
    int             block_x,
    int             block_y,
    int             blockshape_x,
    int             blockshape_y,
    int             mb_ext_level)
  {
    int mb_x                 = BLOCK_SIZE*block_x;
    int mb_y                 = BLOCK_SIZE*block_y;
    int mb_nr                = img->current_mb_nr;

    int mv_a, mv_b, mv_c, pred_vec=0;
    int mvPredType, rFrameL, rFrameU, rFrameUR;
    int hv;


    PixelPos block_a, block_b, block_c, block_d;
    if(mb_ext_level > 0)
    {
      int reset_mb_nr, reset_block_x=block_x, reset_block_y=block_y;

      reset_mb_nr = ResetAvailabilityOfNeighbors32(mb_x, mb_y, &reset_block_x, &reset_block_y, blockshape_x, blockshape_y);

      getLuma4x4Neighbour(reset_mb_nr, reset_block_x, reset_block_y,           -1,  0, &block_a);
      getLuma4x4Neighbour(reset_mb_nr, reset_block_x, reset_block_y,            0, -1, &block_b);
      getLuma4x4Neighbour(reset_mb_nr, reset_block_x, reset_block_y, blockshape_x, -1, &block_c);
      getLuma4x4Neighbour(reset_mb_nr, reset_block_x, reset_block_y,           -1, -1, &block_d);
    }
    else
    {
      getLuma4x4Neighbour(mb_nr, block_x, block_y,           -1,  0, &block_a);
      getLuma4x4Neighbour(mb_nr, block_x, block_y,            0, -1, &block_b);
      getLuma4x4Neighbour(mb_nr, block_x, block_y, blockshape_x, -1, &block_c);
      getLuma4x4Neighbour(mb_nr, block_x, block_y,           -1, -1, &block_d);
    }

#ifdef GEOM
    //update block_a, block_b, block_c, block_d, block_available_right, block_available_below
    if(img->geo_scheme == 0)
    {
      if(geo_segm_mvp[geo_mode][segm] != P_M)
      {  int mb_ext_level = 0;
      short pmv[2];
      SegmMVPredictor(mb_ext_level, geo_segm_mvp[geo_mode][segm], pmv, &block_a, &block_b, &block_c, &block_d, tmp_mv[list]);
      *pmv_x = pmv[0];
      *pmv_y = pmv[1];
      return; 
      }
    }
    else
    {
      PREDTYPE geo_segm_pred;
      geo_segm_pred = find_segm_mvp(mb_ext_level, geo_mode, segm, blockshape_x, &block_a, &block_b, &block_c, &block_d);
      if(geo_segm_pred != P_M)
      {
        short  pmv[2];
        SegmMVPredictor(mb_ext_level, geo_segm_pred, pmv, &block_a, &block_b, &block_c, &block_d, tmp_mv[list]);
        *pmv_x  = pmv[0]; *pmv_y = pmv[1];
        return;
      }

    }
#else
    if (mb_y > 0)
    {
      if (mb_x < 8)  // first column of 8x8 blocks
      {
        if (mb_y==8)
        {
          if (blockshape_x == 16)      block_c.available  = 0;
        }
        else
        {
          if (mb_x+blockshape_x == 8)  block_c.available  = 0;
        }
      }
      else
      {
        if (mb_x+blockshape_x == 16)   block_c.available  = 0;
      }
    }
#endif
    if (!block_c.available)
    {
      block_c=block_d;
    }

    mvPredType = MVPRED_MEDIAN;

    rFrameL    = block_a.available    ? refPic[list][block_a.pos_y][block_a.pos_x] : -1;
    rFrameU    = block_b.available    ? refPic[list][block_b.pos_y][block_b.pos_x] : -1;
    rFrameUR   = block_c.available    ? refPic[list][block_c.pos_y][block_c.pos_x] : -1;



    /* Prediction if only one of the neighbors uses the reference frame
    * we are checking
    */
    if(rFrameL == ref_frame && rFrameU != ref_frame && rFrameUR != ref_frame)       mvPredType = MVPRED_L;
    else if(rFrameL != ref_frame && rFrameU == ref_frame && rFrameUR != ref_frame)  mvPredType = MVPRED_U;
    else if(rFrameL != ref_frame && rFrameU != ref_frame && rFrameUR == ref_frame)  mvPredType = MVPRED_UR;
    // Directional predictions 
    if(blockshape_x == (8<<mb_ext_level) && blockshape_y == (16<<mb_ext_level))
    {
      if(mb_x == 0)
      {
        if(rFrameL == ref_frame)
          mvPredType = MVPRED_L;
      }
      else
      {
        if( rFrameUR == ref_frame)
          mvPredType = MVPRED_UR;
      }
    }
    else if(blockshape_x == (16<<mb_ext_level) && blockshape_y == (8<<mb_ext_level))
    {
      if(mb_y == 0)
      {
        if(rFrameU == ref_frame)
          mvPredType = MVPRED_U;
      }
      else
      {
        if(rFrameL == ref_frame)
          mvPredType = MVPRED_L;
      }
    }

    for (hv=0; hv < 2; hv++)
    {
      mv_a = block_a.available  ? tmp_mv[list][block_a.pos_y][block_a.pos_x][hv] : 0;
      mv_b = block_b.available  ? tmp_mv[list][block_b.pos_y][block_b.pos_x][hv] : 0;
      mv_c = block_c.available  ? tmp_mv[list][block_c.pos_y][block_c.pos_x][hv] : 0;


      switch (mvPredType)
      {
      case MVPRED_MEDIAN:
        if(!(block_b.available || block_c.available))
          pred_vec = mv_a;
        else
          pred_vec = mv_a+mv_b+mv_c-min(mv_a,min(mv_b,mv_c))-max(mv_a,max(mv_b,mv_c));
        break;
      case MVPRED_L:
        pred_vec = mv_a;
        break;
      case MVPRED_U:
        pred_vec = mv_b;
        break;
      case MVPRED_UR:
        pred_vec = mv_c;
        break;
      default:
        break;
      }

      if (hv==0)  *pmv_x = pred_vec;
      else        *pmv_y = pred_vec;

    }
  }


  void readMotionInfoFromNAL_geo (struct img_par *img, struct inp_par *inp, int mb_ext_level)
  {
    int i,j,k;
    int step_h,step_v;
    int curr_mvd;
    Macroblock *currMB  = &img->mb_data[img->current_mb_nr];
    SyntaxElement currSE;
    Slice *currSlice    = img->currentSlice;
    DataPartition *dP;
    int *partMap        = assignSE2partition[currSlice->dp_mode];
    int bframe          = (img->type==B_SLICE);
    //int partmode        = (IS_P8x8(currMB)?4:currMB->mb_type);
#ifdef GEOM
    int segm, pdir;
    int rho = currMB->geo_mode>>5, theta_idx = currMB->geo_mode & 31;
    int step_h0         = BLOCK_STEP [1][0]<<mb_ext_level;
    int step_v0         = BLOCK_STEP [1][1]<<mb_ext_level;
#else
    int step_h0         = BLOCK_STEP [partmode][0];
    int step_v0         = BLOCK_STEP [partmode][1];
#endif
    int mv_mode, i0=0, j0=0;
    char refframe;
    short pmv[2];
    int j4, i4, ii,jj;

    int motion_vector;
    //int mv_scale = 0;

    int flag_mode;

    byte  **    moving_block;
    short ****  co_located_mv;
    char  ***   co_located_ref_idx;
    int64 ***   co_located_ref_id;
#ifdef ADAPTIVE_MV_RES
    int mv_shift=0;
#endif


    {
      moving_block = Co_located->moving_block;
      co_located_mv = Co_located->mv;
      co_located_ref_idx = Co_located->ref_idx;
      co_located_ref_id = Co_located->ref_pic_id;
    }

#ifdef GEOM
    best8x8pdir_geo[0] = currMB->b8pdir[0];
    best8x8pdir_geo[1] = currMB->b8pdir[1];
#endif


    //  If multiple ref. frames, read reference frame for the MB *********************************
    if(img->num_ref_idx_l0_active>1) 
    {
      flag_mode = ( img->num_ref_idx_l0_active == 2 ? 1 : 0);

      currSE.type = SE_REFFRAME;
      dP = &(currSlice->partArr[partMap[SE_REFFRAME]]);

      if (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag)   currSE.mapping = linfo_ue;
      else
      {
#ifdef GEOM
        if(mb_ext_level != 0)
        {
          currSE.mb_ext_level = mb_ext_level;
          currSE.reading = readRefFrame_CABAC32;
        }
        else
#endif
          currSE.reading = readRefFrame_CABAC;
      }
      for(segm=0; segm<2; segm++)
      {

        pdir = best8x8pdir_geo[segm];
        if (pdir==0 || pdir==2)                    
        {
          TRACE_STRING("ref_idx_l0");

          img->subblock_x = i0;
          img->subblock_y = j0;

          if (!IS_P8x8 (currMB) || bframe || (!bframe && !img->allrefzero))
          {
#ifdef GEOM
            currSE.context = BType2CtxRef (3);
#else
            currSE.context = BType2CtxRef (currMB->b8mode[k]);
#endif

            {
              if( (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag) && flag_mode )
              {

                currSE.len = 1;
                readSyntaxElement_FLC(&currSE, dP->bitstream);
                currSE.value1 = 1-currSE.value1;
              }
              else

              {
                currSE.value2 = LIST_0;
                dP->readSyntaxElement (&currSE,img,inp,dP);
              }
              refframe = currSE.value1;  
            }
          }
          else
          {
            refframe = 0;
          }

          for (j=j0; j<j0+step_v0;j++)
            for (i=i0; i<i0+step_h0;i++)
            {
#ifdef GEOM
              if(img->geo_scheme == 0)
              {
                if(currMB->geo_mode > -1 && MB16GeoMask[currMB->geo_mode][j*4+i]==segm)
                  dec_picture->ref_idx[LIST_0][img->block_y + j][img->block_x + i] = refframe;
              }
              else
              {                                   
                if(currMB->geo_mode > -1 && geo_mask[mb_ext_level][rho][theta_idx][j*4][i*4] == segm)
                  dec_picture->ref_idx[LIST_0][img->block_y + j][img->block_x + i] = refframe;
              }
#else
              dec_picture->ref_idx[LIST_0][img->block_y + j][img->block_x + i] = refframe;
#endif        
            }            

#ifdef GEOM
            best8x8fwref_geo[segm] = refframe;
#endif
        }      
      }
    }
    else
    {
      for (segm=0; segm<2; segm++)
      {
        pdir = best8x8pdir_geo[segm];
        if (pdir==0 || pdir==2)                    
        {
          for (j=j0; j<j0+step_v0;j++)
            for (i=i0; i<i0+step_h0;i++)
            {
#ifdef GEOM
              if(img->geo_scheme == 0)
              {
                if(currMB->geo_mode > -1 && MB16GeoMask[currMB->geo_mode][j*4+i]==segm)
                  dec_picture->ref_idx[LIST_0][img->block_y + j][img->block_x + i] = 0;
              }
              else
              {
                if(currMB->geo_mode > -1 && geo_mask[mb_ext_level][rho][theta_idx][j*4][i*4] == segm)                
                  dec_picture->ref_idx[LIST_0][img->block_y + j][img->block_x + i] = 0;
              }
#else
              dec_picture->ref_idx[LIST_0][img->block_y + j][img->block_x + i] = 0;
#endif        
            }

#ifdef GEOM
            best8x8fwref_geo[segm] = 0;
#endif      
        }

      }
    }

    //  If backward multiple ref. frames, read backward reference frame for the MB *********************************
    if(img->num_ref_idx_l1_active>1)
    {
      flag_mode = ( img->num_ref_idx_l1_active == 2 ? 1 : 0);

      currSE.type = SE_REFFRAME;
      dP = &(currSlice->partArr[partMap[SE_REFFRAME]]);
      if (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag)   currSE.mapping = linfo_ue;
      else
      {
#ifdef GEOM
        if(mb_ext_level != 0)
        {
          currSE.mb_ext_level = mb_ext_level;
          currSE.reading = readRefFrame_CABAC32;
        }
        else
#endif
          currSE.reading = readRefFrame_CABAC;
      }
      for (segm=0; segm<2; segm++)
      {
        pdir = best8x8pdir_geo[segm];
        if (pdir==1 || pdir==2)                    
        {
          TRACE_STRING("ref_idx_l1");

          img->subblock_x = i0;
          img->subblock_y = j0;

#ifdef GEOM
          currSE.context = BType2CtxRef (3);
#else
          currSE.context = BType2CtxRef (currMB->b8mode[k]);
#endif
          if( (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag) && flag_mode )
          {
            currSE.len = 1;
            readSyntaxElement_FLC(&currSE, dP->bitstream);
            currSE.value1 = 1 - currSE.value1;
          }
          else
          {
            currSE.value2 = LIST_1;
            dP->readSyntaxElement (&currSE,img,inp,dP);
          }
          refframe = currSE.value1;

          for (j=j0; j<j0+step_v0;j++)
          {
            for (i=i0; i<i0+step_h0;i++)
            {
#ifdef GEOM
              if(img->geo_scheme == 0)
              {
                if(currMB->geo_mode > -1 && MB16GeoMask[currMB->geo_mode][j*4+i]==segm)
                  dec_picture->ref_idx[LIST_1][img->block_y + j][img->block_x + i] = refframe;
              }
              else
              {
                if(currMB->geo_mode > -1 && geo_mask[mb_ext_level][rho][theta_idx][j*4][i*4] == segm)                
                  dec_picture->ref_idx[LIST_1][img->block_y + j][img->block_x + i] = refframe;
              }
#else
              dec_picture->ref_idx[LIST_1][img->block_y + j][img->block_x + i] = refframe;

#endif
            }
          }

#ifdef GEOM
          best8x8bwref_geo[segm] = refframe;
#endif
        }
      }
    }
    else
    {
      for (segm=0; segm<2; segm++)
      {
        pdir = best8x8pdir_geo[segm];
        if (pdir==1 || pdir==2)                    
        {
          for (j=j0; j<j0+step_v0;j++)
            for (i=i0; i<i0+step_h0;i++)
            {
#ifdef GEOM
              if(img->geo_scheme == 0)
              {
                if(currMB->geo_mode > -1 && MB16GeoMask[currMB->geo_mode][j*4+i]==segm)
                  dec_picture->ref_idx[LIST_1][img->block_y + j][img->block_x + i] = 0;
              }
              else
              {
                if(currMB->geo_mode > -1 && geo_mask[mb_ext_level][rho][theta_idx][j*4][i*4] == segm)  
                  dec_picture->ref_idx[LIST_1][img->block_y + j][img->block_x + i] = 0;
              }
#else
              dec_picture->ref_idx[LIST_1][img->block_y + j][img->block_x + i] = 0;
#endif        
            }


#ifdef GEOM
            best8x8bwref_geo[segm] = 0;
#endif
        }
      }
    }

    //=====  READ FORWARD MOTION VECTORS =====
    currSE.type = SE_MVD;
    dP = &(currSlice->partArr[partMap[SE_MVD]]);

    if (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag) currSE.mapping = linfo_se;
    else                     
    {
      if(mb_ext_level != 0)
      {
        currSE.mb_ext_level = mb_ext_level;
        currSE.reading = readMVD_CABAC32;
      }
      else
        currSE.reading = readMVD_CABAC;
    }

    for (segm=0; segm<2; segm++)
    {
      pdir = best8x8pdir_geo[segm];
      if (pdir==0 || pdir==2) //has forward vector
      {
#ifdef GEOM
        mv_mode = 1;
#else
        mv_mode  = currMB->b8mode[k];
#endif
        step_h   = BLOCK_STEP [mv_mode][0]<<mb_ext_level;
        step_v   = BLOCK_STEP [mv_mode][1]<<mb_ext_level;
#ifdef GEOM
        refframe = best8x8fwref_geo[segm];
#else
        refframe = dec_picture->ref_idx[LIST_0][img->block_y+j0][img->block_x+i0];
#endif
        for (j=j0; j<j0+step_v0; j+=step_v)
        {
          for (i=i0; i<i0+step_h0; i+=step_h)
          {
            j4 = img->block_y+j;
            i4 = img->block_x+i;
#ifdef ADAPTIVE_MV_RES
            if (img->mv_res)
            {
              TRACE_STRING("mv_res");
              img->subblock_x = i; // position used for context determination
              img->subblock_y = j; // position used for context determination
              if(mb_ext_level != 0)
                currSE.reading = readMV_RES_4_8_CABAC32;
              else
                currSE.reading = readMV_RES_4_8_CABAC;
              currSE.value2 = 0; // identifies the component; only used for context determination
              dP->readSyntaxElement(&currSE,img,inp,dP);
              mv_shift = currSE.value1;
              if(mb_ext_level != 0)
              {
                currSE.mb_ext_level = mb_ext_level;
                currSE.reading = readMVD_CABAC32;
              }
              else
                currSE.reading = readMVD_CABAC;
            }
            else
              mv_shift=0;

#endif            
            // first make mv-prediction
#if 0//def MV_COMPETITION       //MVC is disabled for Geometry
            if (mv_comp.mv_competition > 0)
            {
              for (k=0; k < 2; k++) 
              {
                TRACE_STRING("mvd_l0");

                img->subblock_x = i; // position used for context determination
                img->subblock_y = j; // position used for context determination
                currSE.value2 = k<<1; // identifies the component; only used for context determination
                dP->readSyntaxElement(&currSE,img,inp,dP);
                curr_mvd = currSE.value1; 

                for(ii=0;ii<step_h;ii++)
                {
                  for(jj=0;jj<step_v;jj++)
                  {
                    currMB->mvd      [LIST_0][j+jj] [i+ii] [k] = curr_mvd;
                  }
                }
              }

              SetMotionVectorPredictor_Competition (img,inp, pmv, pmv+1, refframe, LIST_0, dec_picture->ref_idx, dec_picture->mv, i, j, 4*step_h, 4*step_v);

              for (k=0; k < 2; k++) 
              {
                for(ii=0;ii<step_h;ii++)
                {
                  for(jj=0;jj<step_v;jj++)
                  {
                    dec_picture->mv  [LIST_0][j4+jj][i4+ii][k] = currMB->mvd      [LIST_0][j+jj] [i+ii] [k] + pmv[k];
                  }
                }
              }
            }

            else
#endif
            {


              SetMotionVectorPredictor_geo (currMB->geo_mode, segm, img, pmv, pmv+1, refframe, LIST_0, dec_picture->ref_idx, dec_picture->mv, i, j, 4*step_h, 4*step_v, mb_ext_level);



              for (k=0; k < 2; k++) 
              {
                TRACE_STRING("mvd_l0");

                img->subblock_x = i; // position used for context determination
                img->subblock_y = j; // position used for context determination
#ifdef ADAPTIVE_MV_RES
                currSE.value2 = (k<<2)+mv_shift; // identifies the component; only used for context determination
#else
                currSE.value2 = k<<1; // identifies the component; only used for context determination
#endif
                dP->readSyntaxElement(&currSE,img,inp,dP);
                curr_mvd = currSE.value1; 
#ifdef ADAPTIVE_MV_RES
                motion_vector=(curr_mvd+(roundInt(pmv[k],mv_shift)))<<mv_shift;           /* find motion vector */
#else
                motion_vector=curr_mvd+pmv[k];           /* find motion vector */
#endif                

#ifdef GEOM
                if(img->geo_scheme == 1)
                  geo_mv[LIST_0][segm][k] = motion_vector;
#endif

                for(ii=0;ii<step_h;ii++)
                {
                  for(jj=0;jj<step_v;jj++)
                  {
                    if(img->geo_scheme == 0)
                    {
                      if(MB16GeoMask[currMB->geo_mode][jj*step_h+ii] == segm) 
                      {
                        dec_picture->mv  [LIST_0][j4+jj][i4+ii][k] = motion_vector;
#ifdef ADAPTIVE_MV_RES
                        currMB->mvd      [LIST_0][j+jj] [i+ii] [k] = curr_mvd<<mv_shift;
#else
                        currMB->mvd      [LIST_0][j+jj] [i+ii] [k] = curr_mvd;
#endif
                      }
                    }
                    else
                    { 
                      if(geo_mask[mb_ext_level][rho][theta_idx][jj*4][ii*4] == segm) 
                      {
                        dec_picture->mv  [LIST_0][j4+jj][i4+ii][k] = motion_vector;
#ifdef GEOM
                        if(mb_ext_level == 0)
#endif
#ifdef ADAPTIVE_MV_RES
                          currMB->mvd      [LIST_0][j+jj] [i+ii][k] = curr_mvd<<mv_shift;
#else
                          currMB->mvd      [LIST_0][j+jj] [i+ii][k] = curr_mvd;
#endif
                      }

                    }
                  }
                }
#ifdef GEOM

                if(mb_ext_level != 0)
                {
                  //only save mvd for the upper left 16x16 block for the context of the 2nd partition of 32x16 and 16x32
                  if(j==0 && i==0)
                  {
                    for(ii=0;ii<4;ii++)
                    {
                      for(jj=0;jj<4;jj++)
                      {
                        if(geo_mask[mb_ext_level][rho][theta_idx][jj*4][ii*4] == segm) 
#ifdef ADAPTIVE_MV_RES
                          currMB->mvd      [LIST_0][j+jj] [i+ii] [k] = curr_mvd<<mv_shift;
#else
                          currMB->mvd      [LIST_0][j+jj] [i+ii] [k] = curr_mvd;
#endif
                      }
                    }
                  }
                  //the mvd for the whole 32x32 is saved in mb32_mvd and will be copied to each 16x16 block's mvd in write_macroblock_cluster
                  for(ii=0;ii<step_h;ii++)
                  {
                    for(jj=0;jj<step_v;jj++)
                    {
                      if(geo_mask[mb_ext_level][rho][theta_idx][jj*4][ii*4] == segm) 
#ifdef ADAPTIVE_MV_RES
                        mb32_mvd      [LIST_0][j+jj] [i+ii] [k] = curr_mvd<<mv_shift;
#else
                        mb32_mvd      [LIST_0][j+jj] [i+ii] [k] = curr_mvd;
#endif
                    }
                  }
                }
#endif

              }
            }
          }
        }
      }
    }

    //=====  READ BACKWARD MOTION VECTORS =====
    currSE.type = SE_MVD;
    dP          = &(currSlice->partArr[partMap[SE_MVD]]);

    if (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag) currSE.mapping = linfo_se;
    else
    {
      if(mb_ext_level != 0)
      {
        currSE.mb_ext_level = mb_ext_level;
        currSE.reading = readMVD_CABAC32;
      }
      else
        currSE.reading = readMVD_CABAC;
    }

    for (segm=0; segm<2; segm++)
    {
      pdir = best8x8pdir_geo[segm];
      if (pdir==1 || pdir==2) //has backward vector
      {
#ifdef GEOM
        mv_mode  = 1;
#else
        mv_mode  = currMB->b8mode[k];
#endif
        step_h   = BLOCK_STEP [mv_mode][0]<<mb_ext_level;
        step_v   = BLOCK_STEP [mv_mode][1]<<mb_ext_level;

#ifdef GEOM
        refframe = best8x8bwref_geo[segm];
#else
        refframe = dec_picture->ref_idx[LIST_1][img->block_y+j0][img->block_x+i0];
#endif        
        for (j=j0; j<j0+step_v0; j+=step_v)
        {
          for (i=i0; i<i0+step_h0; i+=step_h)
          {
            j4 = img->block_y+j;
            i4 = img->block_x+i;
#ifdef ADAPTIVE_MV_RES
            if ( img->mv_res )
            {
              TRACE_STRING("mv_res");
              img->subblock_x = i; // position used for context determination
              img->subblock_y = j; // position used for context determination
              if(mb_ext_level != 0)
                currSE.reading = readMV_RES_4_8_CABAC32;
              else
                currSE.reading = readMV_RES_4_8_CABAC;
              currSE.value2 = 0; // identifies the component; only used for context determination
              dP->readSyntaxElement(&currSE,img,inp,dP);
              mv_shift = currSE.value1;
              if(mb_ext_level != 0)
              {
                currSE.mb_ext_level = mb_ext_level;
                currSE.reading = readMVD_CABAC32;
              }
              else
                currSE.reading = readMVD_CABAC;
            }
            else
              mv_shift=0;
#endif             
            // first make mv-prediction

#if 0//def MV_COMPETITION       //MVC is disabled for Geometry
            if (mv_comp.mv_competition > 0)
            {
              for (k=0; k < 2; k++) 
              {
                TRACE_STRING("mvd_l1");

                img->subblock_x = i; // position used for context determination
                img->subblock_y = j; // position used for context determination
                currSE.value2   = (k<<1) +1; // identifies the component; only used for context determination
                dP->readSyntaxElement(&currSE,img,inp,dP);
                curr_mvd = currSE.value1; 

                for(ii=0;ii<step_h;ii++)
                  for(jj=0;jj<step_v;jj++)
                    currMB->mvd      [LIST_1][j+jj] [i+ii] [k] = curr_mvd;
              }

              SetMotionVectorPredictor_Competition (img,inp, pmv, pmv+1, refframe, LIST_1, dec_picture->ref_idx, dec_picture->mv, i, j, 4*step_h, 4*step_v);

              for (k=0; k < 2; k++) 
              {
                for(ii=0;ii<step_h;ii++)
                {
                  for(jj=0;jj<step_v;jj++)
                  {
                    dec_picture->mv  [LIST_1][j4+jj][i4+ii][k] = currMB->mvd      [LIST_1][j+jj] [i+ii] [k] + pmv[k];
                  }
                }
              }
            }
            else

#endif
            {
              SetMotionVectorPredictor_geo (currMB->geo_mode, segm, img, pmv, pmv+1, refframe, LIST_1, dec_picture->ref_idx, dec_picture->mv, i, j, 4*step_h, 4*step_v, mb_ext_level);


              for (k=0; k < 2; k++) 
              {
                TRACE_STRING("mvd_l1");

                img->subblock_x = i; // position used for context determination
                img->subblock_y = j; // position used for context determination
#ifdef ADAPTIVE_MV_RES
                currSE.value2 = (((k<<1) +1)<<1)+mv_shift; // identifies the component; only used for context determination
#else
                currSE.value2 = (k<<1) +1; // identifies the component; only used for context determination
#endif
                dP->readSyntaxElement(&currSE,img,inp,dP);

                curr_mvd = currSE.value1; 

#ifdef ADAPTIVE_MV_RES
                motion_vector=(curr_mvd+(roundInt(pmv[k],mv_shift)))<<mv_shift;           /* find motion vector */
#else
                motion_vector=curr_mvd+pmv[k];           /* find motion vector */
#endif  

#ifdef GEOM
                if(img->geo_scheme)
                  geo_mv[LIST_1][segm][k] = motion_vector;
#endif

                for(ii=0;ii<step_h;ii++)
                {
                  for(jj=0;jj<step_v;jj++)
                  {
                    if(img->geo_scheme == 0)
                    {
                      if(MB16GeoMask[currMB->geo_mode][jj*step_h+ii] == segm) 
                      {
                        dec_picture->mv  [LIST_1][j4+jj][i4+ii][k] = motion_vector;
#ifdef ADAPTIVE_MV_RES
                        currMB->mvd      [LIST_1][j+jj] [i+ii] [k] = curr_mvd<<mv_shift;
#else
                        currMB->mvd      [LIST_1][j+jj] [i+ii] [k] = curr_mvd;
#endif
                      }
                    }
                    else
                    {
                      if(geo_mask[mb_ext_level][rho][theta_idx][jj*4][ii*4] == segm) 
                      {
                        dec_picture->mv  [LIST_1][j4+jj][i4+ii][k] = motion_vector;
#ifdef GEOM
                        if(mb_ext_level == 0)
#endif
#ifdef ADAPTIVE_MV_RES
                          currMB->mvd      [LIST_1][j+jj] [i+ii][k] = curr_mvd<<mv_shift;
#else
                          currMB->mvd      [LIST_1][j+jj] [i+ii][k] = curr_mvd;
#endif
                      }
                    }
                  }
                }
#ifdef GEOM
                if(mb_ext_level != 0)
                {
                  //only save mvd for the upper left 16x16 block for the context of the 2nd partition of 32x16 and 16x32
                  if(j==0 && i==0)
                  {
                    for(ii=0;ii<4;ii++)
                    {
                      for(jj=0;jj<4;jj++)
                      {
                        if(geo_mask[mb_ext_level][rho][theta_idx][jj*4][ii*4] == segm) 
#ifdef ADAPTIVE_MV_RES
                          currMB->mvd      [LIST_1][j+jj] [i+ii] [k] = curr_mvd<<mv_shift;
#else	
                          currMB->mvd      [LIST_1][j+jj] [i+ii] [k] = curr_mvd;
#endif
                      }
                    }
                  }
                  //the mvd for the whole 32x32 is saved in mb32_mvd and will be copied to each 16x16 block's mvd in write_macroblock_cluster
                  for(ii=0;ii<step_h;ii++)
                  {
                    for(jj=0;jj<step_v;jj++)
                    {
                      if(geo_mask[mb_ext_level][rho][theta_idx][jj*4][ii*4] == segm) 
#ifdef ADAPTIVE_MV_RES
                        mb32_mvd      [LIST_1][j+jj] [i+ii] [k] = curr_mvd<<mv_shift;
#else
                        mb32_mvd      [LIST_1][j+jj] [i+ii] [k] = curr_mvd;
#endif
                    }
                  }
                }
#endif
              }
            }
          }
        }
      }

    }
    // record reference picture Ids for deblocking decisions
    for(i4=img->block_x;i4<(img->block_x+(4<<mb_ext_level));i4++)
      for(j4=img->block_y;j4<(img->block_y+(4<<mb_ext_level));j4++)
      {
        if(dec_picture->ref_idx[LIST_0][j4][i4]>=0)
          dec_picture->ref_pic_id[LIST_0][j4][i4] = dec_picture->ref_pic_num[img->current_slice_nr][LIST_0][(short)dec_picture->ref_idx[LIST_0][j4][i4]];
        else
          dec_picture->ref_pic_id[LIST_0][j4][i4] = INT64_MIN;
        if(dec_picture->ref_idx[LIST_1][j4][i4]>=0)
          dec_picture->ref_pic_id[LIST_1][j4][i4] = dec_picture->ref_pic_num[img->current_slice_nr][LIST_1][(short)dec_picture->ref_idx[LIST_1][j4][i4]];  
        else
          dec_picture->ref_pic_id[LIST_1][j4][i4] = INT64_MIN;  
      }
  }

#endif

  int readCBP_Big(struct img_par *img,struct inp_par *inp)
  {
    int mb_nr = img->current_mb_nr;
    Macroblock *currMB = &img->mb_data[mb_nr];
    int cbp;
    SyntaxElement currSE;
    Slice *currSlice = img->currentSlice;
    DataPartition *dP;
    int *partMap = assignSE2partition[currSlice->dp_mode];

    {
      //=====   C B P   =====
      //---------------------
      if (IS_OLDINTRA (currMB) || currMB->mb_type == SI4MB || currMB->mb_type == I8MB)   currSE.type = SE_CBP_INTRA;
      else                        currSE.type = SE_CBP_INTER;

      dP = &(currSlice->partArr[partMap[currSE.type]]);

      //if (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag)
      //{
      //  if (IS_OLDINTRA (currMB) || currMB->mb_type == SI4MB || currMB->mb_type == I8MB)  currSE.mapping = linfo_cbp_intra;
      //  else                        currSE.mapping = linfo_cbp_inter;
      //}
      //else
      {
        currSE.reading = readCBP_CABAC32;
      }

      TRACE_STRING("coded_block_pattern");

      if(active_pps->entropy_coding_mode_flag == CABAC)
        dP->readSyntaxElement(&currSE,img,inp,dP);
      else
        currSE.value1 = u_1 ("cbp", dP->bitstream);//tm_vlc

      currMB->cbp = cbp = currSE.value1;
      return cbp; 
    }
  }

  void readDeltaQP(struct img_par *img,struct inp_par *inp, int cbp)
  {
    int mb_nr = img->current_mb_nr;
    Macroblock *currMB = &img->mb_data[mb_nr];
    SyntaxElement currSE;
    Slice *currSlice = img->currentSlice;
    DataPartition *dP;
    int *partMap = assignSE2partition[currSlice->dp_mode];

    if (cbp !=0)
    {
      if (IS_INTER (currMB))  currSE.type = SE_DELTA_QUANT_INTER;
      else                    currSE.type = SE_DELTA_QUANT_INTRA;

      dP = &(currSlice->partArr[partMap[currSE.type]]);
      currSE.reading= readDquant_CABAC; //gabi

      TRACE_STRING("mb_qp_delta");
      dP->readSyntaxElement(&currSE,img,inp,dP);
      currMB->delta_quant = currSE.value1;
      if ((currMB->delta_quant < -(26 + img->bitdepth_luma_qp_scale/2)) || (currMB->delta_quant > (25 + img->bitdepth_luma_qp_scale/2)))
        error ("mb_qp_delta is out of range", 500);

      img->qp= ((img->qp + currMB->delta_quant + 52 + 2*img->bitdepth_luma_qp_scale)%(52+img->bitdepth_luma_qp_scale)) -
        img->bitdepth_luma_qp_scale;

      currMB->qp = img->qp;

    }
  }


  int readCBP32(struct img_par *img,struct inp_par *inp, int dquant)
  {
    int mb_nr = img->current_mb_nr;
    Macroblock *currMB = &img->mb_data[mb_nr];
    int cbp;
    SyntaxElement currSE;
    Slice *currSlice = img->currentSlice;
    DataPartition *dP;
    int *partMap = assignSE2partition[currSlice->dp_mode];
    //int coef_ctr, i0, j0, b8;
    //int ll;
    //int block_x,block_y;
    //int start_scan;
    //int run, len;
    //int levarr[16], runarr[16], numcoeff;
    //
    //int qp_const;
    //int qp_per    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)/6;
    //int qp_rem    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)%6;
    //int smb       = ((img->type==SP_SLICE) && IS_INTER (currMB)) || (img->type == SI_SLICE && currMB->mb_type == SI4MB);
    //
    //int uv;
    //int qp_const_uv[2];
    //int qp_per_uv[2];
    //int qp_rem_uv[2];
    //
    //int intra     = IS_INTRA (currMB);
    //int temp[4];
    //
    //int b4;
    //int yuv = dec_picture->chroma_format_idc-1;
    //int m5[4];
    //int m6[4];

    //int need_transform_size_flag;
    //Boolean lossless_qpprime = (Boolean)((img->qp + img->bitdepth_luma_qp_scale)==0 && img->lossless_qpprime_flag==1);

    // Residue Color Transform
    //Boolean residual_transform_dc = (Boolean)((img->residue_transform_flag==1) && (IS_OLDINTRA(currMB)||currMB->mb_type==I8MB) );
    //#ifdef USE_INTRA_MDDT
    //  int R[6] = {40, 45, 50, 57, 63, 71};
    //#endif

    //if(img->type==SP_SLICE  && currMB->mb_type!=I16MB )
    //  smb=1;
    //
    // QPI
    //init constants for every chroma qp offset
    //if (dec_picture->chroma_format_idc != YUV400)
    //{
    //  for (i=0; i<2; i++)
    //  {
    //    qp_per_uv[i] = (currMB->qpc[i] + img->bitdepth_chroma_qp_scale)/6;
    //    qp_rem_uv[i] = (currMB->qpc[i] + img->bitdepth_chroma_qp_scale)%6;
    //  }
    //}


    // read CBP if not new intra mode
    //if (!IS_NEWINTRA (currMB))
    {
      //=====   C B P   =====
      //---------------------
      if (IS_OLDINTRA (currMB) || currMB->mb_type == SI4MB || currMB->mb_type == I8MB)   currSE.type = SE_CBP_INTRA;
      else                        currSE.type = SE_CBP_INTER;

      dP = &(currSlice->partArr[partMap[currSE.type]]);

      //if (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag)
      //{
      //  if (IS_OLDINTRA (currMB) || currMB->mb_type == SI4MB || currMB->mb_type == I8MB)  currSE.mapping = linfo_cbp_intra;
      //  else                        currSE.mapping = linfo_cbp_inter;
      //}
      //else
      {
        currSE.reading = readCBP_CABAC32;
      }

      TRACE_STRING("coded_block_pattern");

      dP->readSyntaxElement(&currSE,img,inp,dP);
      currMB->cbp = cbp = currSE.value1;

      if (cbp !=0 && dquant)
      {
        if (IS_INTER (currMB))  currSE.type = SE_DELTA_QUANT_INTER;
        else                    currSE.type = SE_DELTA_QUANT_INTRA;

        dP = &(currSlice->partArr[partMap[currSE.type]]);
        currSE.reading= readDquant_CABAC; //gabi

        TRACE_STRING("mb_qp_delta");

        dP->readSyntaxElement(&currSE,img,inp,dP);
        currMB->delta_quant = currSE.value1;
        if ((currMB->delta_quant < -(26 + img->bitdepth_luma_qp_scale/2)) || (currMB->delta_quant > (25 + img->bitdepth_luma_qp_scale/2)))
          error ("mb_qp_delta is out of range", 500);

        img->qp= ((img->qp + currMB->delta_quant + 52 + 2*img->bitdepth_luma_qp_scale)%(52+img->bitdepth_luma_qp_scale)) -
          img->bitdepth_luma_qp_scale;
      }

      return cbp;   
    }
  }
  /*!
  ************************************************************************
  * \brief
  *    Checks the availability of neighboring macroblocks of
  *    the current macroblock for prediction and context determination;
  *    mb_ext_level: 2-this is 64x64 block; 1-this is 32x32 block; 0-a 16x16 block inside a 32x32 block
  ************************************************************************
  */
  void CheckAvailabilityOfNeighbors32(int mb_ext_level)
  {
    const int mb_nr = img->current_mb_nr;
    Macroblock *currMB = &img->mb_data[mb_nr];

    // mark all neighbors as unavailable
    currMB->mb_available_up   = NULL;
    currMB->mb_available_left = NULL;

    {
      currMB->mbAddrA = mb_nr - 1;
      currMB->mbAddrB = mb_nr - img->PicWidthInMbs;
      currMB->mbAddrC = mb_nr - img->PicWidthInMbs + (1<<mb_ext_level);
      currMB->mbAddrD = mb_nr - img->PicWidthInMbs - 1;

      currMB->mbAvailA = mb_is_available(currMB->mbAddrA, mb_nr) && ((mb_nr % img->PicWidthInMbs)!=0);
      currMB->mbAvailB = mb_is_available(currMB->mbAddrB, mb_nr);
      if(img->type != I_SLICE)
      {
        if(mb_ext_level)
          currMB->mbAvailC = mb_is_available(currMB->mbAddrC, mb_nr) && (((mb_nr+(1<<mb_ext_level)) % img->PicWidthInMbs)!=0);
        else
          currMB->mbAvailC = Not_MB16_3(mb_nr) && mb_is_available(currMB->mbAddrC, mb_nr) && (((mb_nr+1) % img->PicWidthInMbs)!=0);
      }
      else
        currMB->mbAvailC = mb_is_available(currMB->mbAddrC, mb_nr) && (((mb_nr+1) % img->PicWidthInMbs)!=0);

      currMB->mbAvailD = mb_is_available(currMB->mbAddrD, mb_nr) && ((mb_nr % img->PicWidthInMbs)!=0);
    }

    if (currMB->mbAvailA) currMB->mb_available_left = &(img->mb_data[currMB->mbAddrA]);
    if (currMB->mbAvailB) currMB->mb_available_up   = &(img->mb_data[currMB->mbAddrB]);
  }

  /*!
  ************************************************************************
  * \brief
  *    based on the subpartition of 32x32 block decide the 16x16 block on which the neighborhoold will be decided 
  *  \param mb_x: x-coordinate inside 32x32   (in the unit of pel )
  *  \param mb_y: y-coordinate inside 32x32   (in the unit of pel )
  *  \param reset_block_x
  *      input x block position
  *  \param reset_block_y
  *      input y block position
  * \return
  *    16x16 block number in the raster scan order
  ************************************************************************
  */

  int ResetAvailabilityOfNeighbors32(int mb_x, int mb_y, int *reset_block_x, int *reset_block_y, int blockshape_x, int blockshape_y)
  {
    int reset_mb_nr=img->current_mb_nr;

    if(mb_x==0 && mb_y==0 && blockshape_x==64 && blockshape_y==64) // 64x64 partition
      CheckAvailabilityOfNeighbors32(2);

    if(mb_x==0 && mb_y==0 && blockshape_x==32 && blockshape_y==32) // 32x32 partition
      CheckAvailabilityOfNeighbors32(1);

    if(mb_x==0 && mb_y==0 && blockshape_x==64 && blockshape_y==32) // 1st partition 64X32
      CheckAvailabilityOfNeighbors32(2);
    if(mb_x==0 && mb_y==0 && blockshape_x==32 && blockshape_y==16) // 1st partition 32x16
      CheckAvailabilityOfNeighbors32(1);


    if(mb_x==0 && mb_y==32 && blockshape_x==64 && blockshape_y==32)// 2nd partition 64X32
    {
      int saved_mb_nr;
      saved_mb_nr = img->current_mb_nr;
      reset_mb_nr = img->current_mb_nr + img->PicWidthInMbs * 2;
      img->current_mb_nr = reset_mb_nr;
      CheckAvailabilityOfNeighbors32(0);  
      img->current_mb_nr = saved_mb_nr;
      *reset_block_y = 0;
    }
    if(mb_x==0 && mb_y==16 && blockshape_x==32 && blockshape_y==16)// 2nd partition 32x16
    {
      int saved_mb_nr;
      saved_mb_nr = img->current_mb_nr;
      reset_mb_nr = img->current_mb_nr + img->PicWidthInMbs;
      img->current_mb_nr = reset_mb_nr;
      CheckAvailabilityOfNeighbors32(0);   
      img->current_mb_nr = saved_mb_nr;
      *reset_block_y = 0;
    }

    if(mb_x==0 && mb_y==0 && blockshape_x==32 && blockshape_y==64) // 1st partition 32x64
      CheckAvailabilityOfNeighbors32(1);
    if(mb_x==0 && mb_y==0 && blockshape_x==16 && blockshape_y==32) // 1st partition 16x32
      CheckAvailabilityOfNeighbors32(0);

    if(mb_x==32 && mb_y==0 && blockshape_x==32 && blockshape_y==64)// 2nd partition 32X64
    {
      int saved_mb_nr;
      saved_mb_nr = img->current_mb_nr;
      reset_mb_nr = img->current_mb_nr + 2;
      img->current_mb_nr = reset_mb_nr;
      CheckAvailabilityOfNeighbors32(1);
      img->current_mb_nr = saved_mb_nr;
      *reset_block_x = 0; 
    }
    if(mb_x==16 && mb_y==0 && blockshape_x==16 && blockshape_y==32)// 2nd partition 16x32
    {
      int saved_mb_nr;
      saved_mb_nr = img->current_mb_nr;
      reset_mb_nr = img->current_mb_nr + 1;
      img->current_mb_nr = reset_mb_nr;
      CheckAvailabilityOfNeighbors32(0);
      img->current_mb_nr = saved_mb_nr;
      *reset_block_x = 0; 
    }
    return reset_mb_nr;
  }

  /*!
  ************************************************************************
  * \brief
  *    Set motion vector predictor
  ************************************************************************
  */
  void SetMotionVectorPredictor32 (struct img_par  *img,
    short           *pmv_x,
    short           *pmv_y,
    char            ref_frame,
    byte            list,
    char            ***refPic,
    short           ****tmp_mv,
    int             block_x,
    int             block_y,
    int             blockshape_x,
    int             blockshape_y,
    int             mb_ext_level)
  {
    int mb_x                 = BLOCK_SIZE*block_x;
    int mb_y                 = BLOCK_SIZE*block_y;
    int mb_nr                = img->current_mb_nr;

    int mv_a, mv_b, mv_c, pred_vec=0;
    int mvPredType, rFrameL, rFrameU, rFrameUR;
    int hv;


    PixelPos block_a, block_b, block_c, block_d;
    {
      int reset_mb_nr, reset_block_x=block_x, reset_block_y=block_y;

      reset_mb_nr = ResetAvailabilityOfNeighbors32(mb_x, mb_y, &reset_block_x, &reset_block_y, blockshape_x, blockshape_y);

      getLuma4x4Neighbour(reset_mb_nr, reset_block_x, reset_block_y,           -1,  0, &block_a);
      getLuma4x4Neighbour(reset_mb_nr, reset_block_x, reset_block_y,            0, -1, &block_b);
      getLuma4x4Neighbour(reset_mb_nr, reset_block_x, reset_block_y, blockshape_x, -1, &block_c);
      getLuma4x4Neighbour(reset_mb_nr, reset_block_x, reset_block_y,           -1, -1, &block_d);
    }


    if (mb_y > 0)
    {
      if (mb_x < (8<<mb_ext_level))  // first column of 8x8 blocks //MB32X32
      {
        if (mb_y==(8<<mb_ext_level)) //MB32X32
        {
          if (blockshape_x == (16<<mb_ext_level))      block_c.available  = 0; //MB32X32
        }
        else
        {
          error("never reach here!\n", -1);//MB32X32
          if (mb_x+blockshape_x == 8)  block_c.available  = 0; 
        }
      }
      else
      {
        error("never reach here!\n", -1);
        if (mb_x+blockshape_x == 16)   block_c.available  = 0;
      }
    }

    if (!block_c.available)
    {
      block_c=block_d;
    }

    mvPredType = MVPRED_MEDIAN;

    rFrameL    = block_a.available    ? refPic[list][block_a.pos_y][block_a.pos_x] : -1;
    rFrameU    = block_b.available    ? refPic[list][block_b.pos_y][block_b.pos_x] : -1;
    rFrameUR   = block_c.available    ? refPic[list][block_c.pos_y][block_c.pos_x] : -1;


    /* Prediction if only one of the neighbors uses the reference frame
    * we are checking
    */
    if(rFrameL == ref_frame && rFrameU != ref_frame && rFrameUR != ref_frame)       mvPredType = MVPRED_L;
    else if(rFrameL != ref_frame && rFrameU == ref_frame && rFrameUR != ref_frame)  mvPredType = MVPRED_U;
    else if(rFrameL != ref_frame && rFrameU != ref_frame && rFrameUR == ref_frame)  mvPredType = MVPRED_UR;
    // Directional predictions 
    if(blockshape_x == (8<<mb_ext_level) && blockshape_y == (16<<mb_ext_level)) //MB32X32
    {
      if(mb_x == 0)
      {
        if(rFrameL == ref_frame)
          mvPredType = MVPRED_L;
      }
      else
      {
        if( rFrameUR == ref_frame)
          mvPredType = MVPRED_UR;
      }
    }
    else if(blockshape_x == (16<<mb_ext_level) && blockshape_y == (8<<mb_ext_level))//MB32X32
    {
      if(mb_y == 0)
      {
        if(rFrameU == ref_frame)
          mvPredType = MVPRED_U;
      }
      else
      {
        if(rFrameL == ref_frame)
          mvPredType = MVPRED_L;
      }
    }

    for (hv=0; hv < 2; hv++)
    {
      mv_a = block_a.available  ? tmp_mv[list][block_a.pos_y][block_a.pos_x][hv] : 0;
      mv_b = block_b.available  ? tmp_mv[list][block_b.pos_y][block_b.pos_x][hv] : 0;
      mv_c = block_c.available  ? tmp_mv[list][block_c.pos_y][block_c.pos_x][hv] : 0;

      {
        int64 xpoc  = img->framepoc >> 1;
        int64 x  = dec_picture->ref_pic_num[0][list][ref_frame] >> 2;
        int64 L  = dec_picture->ref_pic_num[0][list][rFrameL  ] >> 2;
        int64 U  = dec_picture->ref_pic_num[0][list][rFrameU  ] >> 2;
        int64 UR = dec_picture->ref_pic_num[0][list][rFrameUR ] >> 2;

        if(xpoc != L)
          mv_a = (int)(mv_a * (xpoc - x) / (xpoc - L));
        if(xpoc != U)
          mv_b = (int)(mv_b * (xpoc - x) / (xpoc - U));
        if(xpoc != UR)
          mv_c = (int)(mv_c * (xpoc - x) / (xpoc - UR));

      }

      switch (mvPredType)
      {
      case MVPRED_MEDIAN:
        if(!(block_b.available || block_c.available))
          pred_vec = mv_a;
        else
          pred_vec = mv_a+mv_b+mv_c-min(mv_a,min(mv_b,mv_c))-max(mv_a,max(mv_b,mv_c));
        break;
      case MVPRED_L:
        pred_vec = mv_a;
        break;
      case MVPRED_U:
        pred_vec = mv_b;
        break;
      case MVPRED_UR:
        pred_vec = mv_c;
        break;
      default:
        break;
      }

      if (hv==0)  *pmv_x = pred_vec;
      else        *pmv_y = pred_vec;

    }
  }


  /*!
  ************************************************************************
  * \brief
  *    init macroblock I and P frames
  ************************************************************************
  */
  void init_macroblock32(struct img_par *img, int mb_ext_level)
  {
    int i,j;

    for (i=0;i<(BLOCK_SIZE<<mb_ext_level);i++)
    {                           // reset vectors and pred. modes
      for(j=0;j<(BLOCK_SIZE<<mb_ext_level);j++)
      {
        dec_picture->mv[LIST_0][img->block_y+j][img->block_x+i][0]=0;
        dec_picture->mv[LIST_0][img->block_y+j][img->block_x+i][1]=0;
        dec_picture->mv[LIST_1][img->block_y+j][img->block_x+i][0]=0;
        dec_picture->mv[LIST_1][img->block_y+j][img->block_x+i][1]=0;

        img->ipredmode[img->block_x+i][img->block_y+j] = DC_PRED;
      }
    }

    for (j=0; j<(BLOCK_SIZE<<mb_ext_level); j++)
      for (i=0; i<(BLOCK_SIZE<<mb_ext_level); i++)
      {
        dec_picture->ref_idx[LIST_0][img->block_y+j][img->block_x+i] = -1;
        dec_picture->ref_idx[LIST_1][img->block_y+j][img->block_x+i] = -1;
        dec_picture->ref_pic_id[LIST_0][img->block_y+j][img->block_x+i] = INT64_MIN;
        dec_picture->ref_pic_id[LIST_1][img->block_y+j][img->block_x+i] = INT64_MIN;
      }
  }

  /*!
  ************************************************************************
  * \brief
  *    Read motion info
  ************************************************************************
  */
  void readMotionInfoFromNAL32 (struct img_par *img, struct inp_par *inp, int mb_ext_level)
  {
    int i,j,k;
    int step_h,step_v;
    int curr_mvd;
    Macroblock *currMB  = &img->mb_data[img->current_mb_nr];
    SyntaxElement currSE;
    Slice *currSlice    = img->currentSlice;
    DataPartition *dP;
    int *partMap        = assignSE2partition[currSlice->dp_mode];
    int bframe          = (img->type==B_SLICE);
    int partmode        = (IS_P8x8(currMB)?4:currMB->mb_type);

    int step_h0         = BLOCK_STEP [partmode][0]<<mb_ext_level;
    int step_v0         = BLOCK_STEP [partmode][1]<<mb_ext_level;

    int mv_mode, i0, j0;
    char refframe;
    short pmv[2];
    int j4, i4, ii,jj;

    int motion_vector;
    int mv_scale = 0;

    int flag_mode;


    byte  **    moving_block;
    short ****  co_located_mv;
    char  ***   co_located_ref_idx;
    int64 ***   co_located_ref_id;

#ifdef ADAPTIVE_MV_RES
    int mv_shift = 0;
#endif

    moving_block = Co_located->moving_block;
    co_located_mv = Co_located->mv;
    co_located_ref_idx = Co_located->ref_idx;
    co_located_ref_id = Co_located->ref_pic_id;

    if (bframe && IS_P8x8 (currMB))
    {
      if (img->direct_spatial_mv_pred_flag)
      {
        int imgblock_y= img->block_y;
        int fw_rFrameL, fw_rFrameU, fw_rFrameUL, fw_rFrameUR;
        int bw_rFrameL, bw_rFrameU, bw_rFrameUL, bw_rFrameUR;

        PixelPos mb_left, mb_up, mb_upleft, mb_upright;

        char  fw_rFrame,bw_rFrame;
        short pmvfw[2]={0,0},
          pmvbw[2]={0,0};


        getLuma4x4Neighbour(img->current_mb_nr, 0, 0, -1,  0, &mb_left);
        getLuma4x4Neighbour(img->current_mb_nr, 0, 0,  0, -1, &mb_up);
        getLuma4x4Neighbour(img->current_mb_nr, 0, 0, 16, -1, &mb_upright);
        getLuma4x4Neighbour(img->current_mb_nr, 0, 0, -1, -1, &mb_upleft);

        fw_rFrameL = mb_left.available ? dec_picture->ref_idx[LIST_0][mb_left.pos_y][mb_left.pos_x] : -1;
        fw_rFrameU = mb_up.available ? dec_picture->ref_idx[LIST_0][mb_up.pos_y][mb_up.pos_x] : -1;
        fw_rFrameUL = mb_upleft.available ? dec_picture->ref_idx[LIST_0][mb_upleft.pos_y][mb_upleft.pos_x] : -1;
        fw_rFrameUR = mb_upright.available ? dec_picture->ref_idx[LIST_0][mb_upright.pos_y][mb_upright.pos_x] : fw_rFrameUL;      

        bw_rFrameL = mb_left.available ? dec_picture->ref_idx[LIST_1][mb_left.pos_y][mb_left.pos_x] : -1;
        bw_rFrameU = mb_up.available ? dec_picture->ref_idx[LIST_1][mb_up.pos_y][mb_up.pos_x] : -1;
        bw_rFrameUL = mb_upleft.available ? dec_picture->ref_idx[LIST_1][mb_upleft.pos_y][mb_upleft.pos_x] : -1;
        bw_rFrameUR = mb_upright.available ? dec_picture->ref_idx[LIST_1][mb_upright.pos_y][mb_upright.pos_x] : bw_rFrameUL;      


        fw_rFrame = (fw_rFrameL >= 0 && fw_rFrameU >= 0) ? min(fw_rFrameL,fw_rFrameU): max(fw_rFrameL,fw_rFrameU);
        fw_rFrame = (fw_rFrame >= 0 && fw_rFrameUR >= 0) ? min(fw_rFrame,fw_rFrameUR): max(fw_rFrame,fw_rFrameUR);

        bw_rFrame = (bw_rFrameL >= 0 && bw_rFrameU >= 0) ? min(bw_rFrameL,bw_rFrameU): max(bw_rFrameL,bw_rFrameU);
        bw_rFrame = (bw_rFrame >= 0 && bw_rFrameUR >= 0) ? min(bw_rFrame,bw_rFrameUR): max(bw_rFrame,bw_rFrameUR);


        if (fw_rFrame >=0)
          SetMotionVectorPredictor (img, pmvfw, pmvfw+1, fw_rFrame, LIST_0, dec_picture->ref_idx, dec_picture->mv, 0, 0, 16, 16);

        if (bw_rFrame >=0)
          SetMotionVectorPredictor (img, pmvbw, pmvbw+1, bw_rFrame, LIST_1, dec_picture->ref_idx, dec_picture->mv, 0, 0, 16, 16);


        for (i=0;i<4;i++)
        {
          if (currMB->b8mode[i] == 0)
            for(j=2*(i/2);j<2*(i/2)+2;j++)
              for(k=2*(i%2);k<2*(i%2)+2;k++)
              {
                int j6 = imgblock_y+j;
                j4 = img->block_y+j;
                i4 = img->block_x+k;


                if (fw_rFrame >= 0)
                {

                  if  (!fw_rFrame  && ((!moving_block[j6][i4]) && (!listX[1][0]->is_long_term)))
                  {                    
                    dec_picture->mv  [LIST_0][j4][i4][0] = 0;
                    dec_picture->mv  [LIST_0][j4][i4][1] = 0;
                    dec_picture->ref_idx[LIST_0][j4][i4] = 0;                    
                  }
                  else
                  {

                    dec_picture->mv  [LIST_0][j4][i4][0] = pmvfw[0];
                    dec_picture->mv  [LIST_0][j4][i4][1] = pmvfw[1];
                    dec_picture->ref_idx[LIST_0][j4][i4] = fw_rFrame;
                  }
                }
                else
                {
                  dec_picture->mv  [LIST_0][j4][i4][0] = 0;
                  dec_picture->mv  [LIST_0][j4][i4][1] = 0;
                  dec_picture->ref_idx[LIST_0][j4][i4] = -1;
                }
                if (bw_rFrame >= 0)
                {
                  if  (bw_rFrame==0 && ((!moving_block[j6][i4])&& (!listX[1][0]->is_long_term)))
                  {
                    dec_picture->mv  [LIST_1][j4][i4][0] = 0;
                    dec_picture->mv  [LIST_1][j4][i4][1] = 0;
                    dec_picture->ref_idx[LIST_1][j4][i4] = 0;
                  }
                  else
                  {
                    dec_picture->mv  [LIST_1][j4][i4][0] = pmvbw[0];
                    dec_picture->mv  [LIST_1][j4][i4][1] = pmvbw[1];
                    dec_picture->ref_idx[LIST_1][j4][i4] = bw_rFrame;
                  }
                }
                else
                {
                  dec_picture->mv  [LIST_1][j4][i4][0] = 0;
                  dec_picture->mv  [LIST_1][j4][i4][1] = 0;
                  dec_picture->ref_idx[LIST_1][j4][i4] = -1;                               
                }

                if (fw_rFrame <0 && bw_rFrame <0)
                {
                  dec_picture->ref_idx[LIST_0][j4][i4] = 0;
                  dec_picture->ref_idx[LIST_1][j4][i4] = 0;                  
                }
              }
        }
      }
      else
      {
        for (i=0;i<4;i++)
        {
          if (currMB->b8mode[i] == 0)
          {
            for(j=2*(i/2);j<2*(i/2)+2;j++)
            {
              for(k=2*(i%2);k<2*(i%2)+2;k++)
              {

                int imgblock_y= img->block_y;
                int refList = co_located_ref_idx[LIST_0 ][imgblock_y+j][img->block_x+k]== -1 ? LIST_1 : LIST_0;
                int ref_idx = co_located_ref_idx[refList][imgblock_y + j][img->block_x + k];
                int mapped_idx=-1, iref;                             

                if (ref_idx == -1)
                {
                  dec_picture->ref_idx [LIST_0][img->block_y + j][img->block_x + k] = 0;
                  dec_picture->ref_idx [LIST_1][img->block_y + j][img->block_x + k] = 0;                
                }
                else
                {
                  for (iref=0;iref<min(img->num_ref_idx_l0_active,listXsize[LIST_0]);iref++)
                  {
#if 1
                    //int curr_mb_field = 0;

                    if(img->structure==0 )
                    {
                      // If the current MB is a frame MB and the colocated is from a field picture, 
                      // then the co_located_ref_id may have been generated from the wrong value of 
                      // frame_poc if it references it's complementary field, so test both POC values
                      if(listX[0][iref]->top_poc*2 == co_located_ref_id[refList][imgblock_y + j][img->block_x + k] 
                      || listX[0][iref]->bottom_poc*2 == co_located_ref_id[refList][imgblock_y + j][img->block_x + k])
                      {
                        mapped_idx=iref;
                        break;
                      }
                      else //! invalid index. Default to zero even though this case should not happen
                        mapped_idx=INVALIDINDEX;
                      continue;
                    }    
#endif                                        
                    if (dec_picture->ref_pic_num[img->current_slice_nr][LIST_0][iref]==co_located_ref_id[refList][imgblock_y + j][img->block_x + k])
                    {
                      mapped_idx=iref;
                      break;
                    }
                    else //! invalid index. Default to zero even though this case should not happen
                      mapped_idx=INVALIDINDEX;
                  }
                  if (INVALIDINDEX == mapped_idx)
                  {
                    error("temporal direct error\ncolocated block has ref that is unavailable",-1111);
                  }
                  dec_picture->ref_idx [LIST_0][img->block_y + j][img->block_x + k] = mapped_idx;
                  dec_picture->ref_idx [LIST_1][img->block_y + j][img->block_x + k] = 0;                
                }
              }
            }
          }
        }
      }
    } 

    //  If multiple ref. frames, read reference frame for the MB *********************************
    if(img->num_ref_idx_l0_active>1) 
    {
      flag_mode = ( img->num_ref_idx_l0_active == 2 ? 1 : 0);

      currSE.type = SE_REFFRAME;
      dP = &(currSlice->partArr[partMap[SE_REFFRAME]]);

      {
        currSE.mb_ext_level = mb_ext_level;
        currSE.reading = readRefFrame_CABAC32;
      }

      for (j0=0; j0<(4<<mb_ext_level); j0+=step_v0)
      {
        for (i0=0; i0<(4<<mb_ext_level); i0+=step_h0)
        {
          k=2*(j0>>(mb_ext_level+1))+(i0>>(mb_ext_level+1));

          if ((currMB->b8pdir[k]==0 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)
          {
            TRACE_STRING("ref_idx_l0");

            img->subblock_x = i0;
            img->subblock_y = j0;

            if (!IS_P8x8 (currMB) || bframe || (!bframe && !img->allrefzero))
            {
              currSE.context = BType2CtxRef (currMB->b8mode[k]);
              {
                if( (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag) && flag_mode )
                {
                  currSE.len = 1;
                  readSyntaxElement_FLC(&currSE, dP->bitstream);
                  currSE.value1 = 1 - currSE.value1;
                }
                else
                {
                  currSE.value2 = LIST_0;
                  dP->readSyntaxElement (&currSE,img,inp,dP);
                }
                refframe = currSE.value1;
              }
            }
            else
            {
              refframe = 0;
            }

            /*
            if (bframe && refframe>img->buf_cycle)    // img->buf_cycle should be correct for field MBs now
            {
            set_ec_flag(SE_REFFRAME);
            refframe = 1;
            }
            */

            for (j=j0; j<j0+step_v0;j++)
              for (i=i0; i<i0+step_h0;i++)
              {
                dec_picture->ref_idx[LIST_0][img->block_y + j][img->block_x + i] = refframe;
              }

          }
        }
      }
    }
    else
    {
      for (j0=0; j0<(4<<mb_ext_level); j0+=step_v0)
      {
        for (i0=0; i0<(4<<mb_ext_level); i0+=step_h0)
        {
          k=2*(j0>>(mb_ext_level+1))+(i0>>(mb_ext_level+1));
          if ((currMB->b8pdir[k]==0 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)
          {
            for (j=j0; j<j0+step_v0;j++)
              for (i=i0; i<i0+step_h0;i++)
              {
                dec_picture->ref_idx[LIST_0][img->block_y + j][img->block_x + i] = 0;
              }
          }
        }
      }
    }

    //  If backward multiple ref. frames, read backward reference frame for the MB *********************************
    if(img->num_ref_idx_l1_active>1)
    {
      flag_mode = ( img->num_ref_idx_l1_active == 2 ? 1 : 0);

      currSE.type = SE_REFFRAME;
      dP = &(currSlice->partArr[partMap[SE_REFFRAME]]);
      //if (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag)   currSE.mapping = linfo_ue;
      //else
      {
        currSE.mb_ext_level = mb_ext_level;
        currSE.reading = readRefFrame_CABAC32;
      }

      for (j0=0; j0<(4<<mb_ext_level); j0+=step_v0)
      {
        for (i0=0; i0<(4<<mb_ext_level); i0+=step_h0)
        {
          k=2*(j0>>(mb_ext_level+1))+(i0>>(mb_ext_level+1));
          if ((currMB->b8pdir[k]==1 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)
          {
            TRACE_STRING("ref_idx_l1");

            img->subblock_x = i0;
            img->subblock_y = j0;

            currSE.context = BType2CtxRef (currMB->b8mode[k]);
            if( (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag) && flag_mode )
            {
              currSE.len = 1;
              readSyntaxElement_FLC(&currSE, dP->bitstream);
              currSE.value1 = 1-currSE.value1;
            }
            else
            {
              currSE.value2 = LIST_1;
              dP->readSyntaxElement (&currSE,img,inp,dP);
            }
            refframe = currSE.value1;

            for (j=j0; j<j0+step_v0;j++)
            {
              for (i=i0; i<i0+step_h0;i++)
              {
                dec_picture->ref_idx[LIST_1][img->block_y + j][img->block_x + i] = refframe;
              }
            }
          }
        }
      }
    }
    else
    {
      for (j0=0; j0<(4<<mb_ext_level); j0+=step_v0)
      {
        for (i0=0; i0<(4<<mb_ext_level); i0+=step_h0)
        {
          k=2*(j0>>(mb_ext_level+1))+(i0>>(mb_ext_level+1));
          if ((currMB->b8pdir[k]==1 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)
          {
            for (j=j0; j<j0+step_v0;j++)
              for (i=i0; i<i0+step_h0;i++)
              {
                dec_picture->ref_idx[LIST_1][img->block_y + j][img->block_x + i] = 0;
              }
          }
        }
      }
    }

    //=====  READ FORWARD MOTION VECTORS =====
    currSE.type = SE_MVD;
    dP = &(currSlice->partArr[partMap[SE_MVD]]);

    {
      currSE.mb_ext_level = mb_ext_level;
      currSE.reading = readMVD_CABAC32;
    }


    for (j0=0; j0<(4<<mb_ext_level); j0+=step_v0)
      for (i0=0; i0<(4<<mb_ext_level); i0+=step_h0)
      {
        k=2*(j0>>(mb_ext_level+1))+(i0>>(mb_ext_level+1));
        if ((currMB->b8pdir[k]==0 || currMB->b8pdir[k]==2) && (currMB->b8mode[k] !=0))//has forward vector
        {
          mv_mode  = currMB->b8mode[k];

          step_h   = BLOCK_STEP [mv_mode][0]<<mb_ext_level;
          step_v   = BLOCK_STEP [mv_mode][1]<<mb_ext_level;

          refframe = dec_picture->ref_idx[LIST_0][img->block_y+j0][img->block_x+i0];

          for (j=j0; j<j0+step_v0; j+=step_v)
          {
            for (i=i0; i<i0+step_h0; i+=step_h)
            {
              j4 = img->block_y+j;
              i4 = img->block_x+i;
#ifdef ADAPTIVE_MV_RES
              if (img->mv_res)
              {
                TRACE_STRING("mv_res");
                img->subblock_x = i; // position used for context determination
                img->subblock_y = j; // position used for context determination
                currSE.reading = readMV_RES_4_8_CABAC32;
                currSE.value2 = 0; // identifies the component; only used for context determination
                dP->readSyntaxElement(&currSE,img,inp,dP);
                mv_shift = currSE.value1;
                currSE.reading = readMVD_CABAC32;
              }
              else
                mv_shift=0;
#endif
              // first make mv-prediction
#ifdef MV_COMPETITION
              if (mv_comp.mv_competition > 0)
              {
                int mvd[2];

                for (k=0; k < 2; k++) 
                {
                  TRACE_STRING("mvd_l0");

                  img->subblock_x = i; // position used for context determination
                  img->subblock_y = j; // position used for context determination
#ifdef ADAPTIVE_MV_RES
                  currSE.value2 = (k<<2)+mv_shift; // identifies the component; only used for context determination
#else
                  currSE.value2 = k<<1; // identifies the component; only used for context determination
#endif
                  dP->readSyntaxElement(&currSE,img,inp,dP);
                  curr_mvd = currSE.value1; 

                  mvd[k]=curr_mvd;
                }

                SetMotionVectorPredictor_Competition32 (img,inp, pmv, pmv+1, refframe, LIST_0, dec_picture->ref_idx, dec_picture->mv, i, j, 4*step_h, 4*step_v, mb_ext_level);

                for (k=0; k < 2; k++) 
                {
                  for(ii=0;ii<step_h;ii++)
                  {
                    for(jj=0;jj<step_v;jj++)
                    {
#ifdef ADAPTIVE_MV_RES
                      dec_picture->mv  [LIST_0][j4+jj][i4+ii][k] = (mvd[k] + roundInt(pmv[k],mv_shift))<<mv_shift;
                      mb32_mvd         [LIST_0][j+jj] [i+ii] [k] = mvd[k]<<mv_shift;
#else
                      dec_picture->mv        [LIST_0][j4+jj][i4+ii][k] = mvd[k] + pmv[k];
                      //the mvd for the whole 32x32 is saved in mb32_mvd and will be copied to each 16x16 block's mvd in write_macroblock_cluster
                      mb32_mvd  [LIST_0][j +jj][i +ii][k] = mvd[k];
#endif
                    }
                  }

                  //only save mvd for the upper left 16x16 block for the context of the 2nd partition of 32x16 and 16x32
                  if(j==0 && i==0)
                  {
                    for(ii=0;ii<4;ii++)
                    {
                      for(jj=0;jj<4;jj++)
                      {
#ifdef ADAPTIVE_MV_RES
                        currMB->mvd      [LIST_0][j+jj] [i+ii] [k] = mvd[k]<<mv_shift;
#else
                        currMB->mvd      [LIST_0][j+jj] [i+ii] [k] = mvd[k];
#endif
                      }
                    }
                  }
                }
              }

              else
#endif
              {

                SetMotionVectorPredictor32 (img, pmv, pmv+1, refframe, LIST_0, dec_picture->ref_idx, dec_picture->mv, i, j, 4*step_h, 4*step_v, mb_ext_level);

                for (k=0; k < 2; k++) 
                {
                  TRACE_STRING("mvd_l0");

                  img->subblock_x = i; // position used for context determination
                  img->subblock_y = j; // position used for context determination
#ifdef ADAPTIVE_MV_RES
                  currSE.value2 = (k<<2)+mv_shift; // identifies the component; only used for context determination
#else
                  currSE.value2 = k<<1; // identifies the component; only used for context determination
#endif
                  dP->readSyntaxElement(&currSE,img,inp,dP);
                  curr_mvd = currSE.value1; 
#ifdef ADAPTIVE_MV_RES
                  motion_vector=(curr_mvd+(roundInt(pmv[k],mv_shift)))<<mv_shift;           /* find motion vector */
#else
                  motion_vector=curr_mvd+pmv[k];           /* find motion vector */
#endif

                  for(ii=0;ii<step_h;ii++)
                  {
                    for(jj=0;jj<step_v;jj++)
                    {
                      dec_picture->mv  [LIST_0][j4+jj][i4+ii][k] = motion_vector;
                    }
                  }
                  //only save mvd for the upper left 16x16 block for the context of the 2nd partition of 32x16 and 16x32
                  if(j==0 && i==0)
                  {
                    for(ii=0;ii<4;ii++)
                    {
                      for(jj=0;jj<4;jj++)
                      {
#ifdef ADAPTIVE_MV_RES
                        currMB->mvd      [LIST_0][j+jj] [i+ii] [k] = curr_mvd<<mv_shift;
#else
                        currMB->mvd      [LIST_0][j+jj] [i+ii] [k] = curr_mvd;
#endif
                      }
                    }
                  }
                  //the mvd for the whole 32x32 is saved in mb32_mvd and will be copied to each 16x16 block's mvd in write_macroblock_cluster
                  for(ii=0;ii<step_h;ii++)
                  {
                    for(jj=0;jj<step_v;jj++)
                    {
#ifdef ADAPTIVE_MV_RES
                      mb32_mvd      [LIST_0][j+jj] [i+ii] [k] = curr_mvd<<mv_shift;
#else
                      mb32_mvd      [LIST_0][j+jj] [i+ii] [k] = curr_mvd;
#endif
                    }
                  }
                }
              }
            }
          }
        }
        else if (currMB->b8mode[k=2*(j0>>(mb_ext_level+1))+(i0>>(mb_ext_level+1))]==0)      
        {  
          if (!img->direct_spatial_mv_pred_flag)
          {
            int list_offset = 0;
            int imgblock_y= img->block_y;

            int refList = (co_located_ref_idx[LIST_0][imgblock_y+j0][img->block_x+i0]== -1 ? LIST_1 : LIST_0);
            int ref_idx =  co_located_ref_idx[refList][imgblock_y+j0][img->block_x+i0];          

            if (ref_idx==-1)
            {
              for (j=j0; j<j0+step_v0; j++)
                for (i=i0; i<i0+step_h0; i++)
                {            
                  dec_picture->ref_idx [LIST_1][img->block_y+j][img->block_x+i]=0;
                  dec_picture->ref_idx [LIST_0][img->block_y+j][img->block_x+i]=0; 
                  j4 = img->block_y+j;
                  i4 = img->block_x+i;            
                  for (ii=0; ii < 2; ii++) 
                  {                                    
                    dec_picture->mv [LIST_0][j4][i4][ii]=0;
                    dec_picture->mv [LIST_1][j4][i4][ii]=0;                  
                  }
                }
            }
            else 
            {        
              int mapped_idx=-1, iref;                             
              int j6;

              for (iref=0;iref<min(img->num_ref_idx_l0_active,listXsize[LIST_0]);iref++)
              {

#if 1
                //int curr_mb_field = 0;

                if(img->structure==0 )
                {
                  // If the current MB is a frame MB and the colocated is from a field picture, 
                  // then the co_located_ref_id may have been generated from the wrong value of 
                  // frame_poc if it references it's complementary field, so test both POC values
                  if(listX[0][iref]->top_poc*2 == co_located_ref_id[refList][imgblock_y + j0][img->block_x + i0] 
                  || listX[0][iref]->bottom_poc*2 == co_located_ref_id[refList][imgblock_y + j0][img->block_x + i0])
                  {
                    mapped_idx=iref;
                    break;
                  }
                  else //! invalid index. Default to zero even though this case should not happen
                    mapped_idx=INVALIDINDEX;
                  continue;
                }    
#endif                                        
                if (dec_picture->ref_pic_num[img->current_slice_nr][LIST_0][iref]==co_located_ref_id[refList][imgblock_y+j0][img->block_x+i0])
                {
                  mapped_idx=iref;
                  break;
                }
                else //! invalid index. Default to zero even though this case should not happen
                  mapped_idx=INVALIDINDEX;
              }

              if (INVALIDINDEX == mapped_idx)
              {
                error("temporal direct error\ncolocated block has ref that is unavailable",-1111);
              }


              for (j=j0; j<j0+step_v0; j++)
                for (i=i0; i<i0+step_h0; i++)
                {
                  {
                    mv_scale = img->mvscale[LIST_0][mapped_idx];

                    dec_picture->ref_idx [LIST_0][img->block_y+j][img->block_x+i] = mapped_idx;
                    dec_picture->ref_idx [LIST_1][img->block_y+j][img->block_x+i] = 0;

                    j4 = img->block_y+j;
                    j6 = imgblock_y+j;
                    i4 = img->block_x+i;

                    for (ii=0; ii < 2; ii++) 
                    {              
                      //if (iTRp==0)
                      if (mv_scale == 9999 || listX[LIST_0][mapped_idx]->is_long_term)
                        //                    if (mv_scale==9999 || Co_located->is_long_term)
                      {                      
                        dec_picture->mv  [LIST_0][j4][i4][ii]=co_located_mv[refList][j6][i4][ii];
                        dec_picture->mv  [LIST_1][j4][i4][ii]=0;
                      }
                      else
                      {
                        dec_picture->mv  [LIST_0][j4][i4][ii]=(mv_scale * co_located_mv[refList][j6][i4][ii] + 128 ) >> 8;
                        dec_picture->mv  [LIST_1][j4][i4][ii]=dec_picture->mv[LIST_0][j4][i4][ii] - co_located_mv[refList][j6][i4][ii];
                      }
                    }
                  } 
                }
            }  
          } 
        }
      }

      //=====  READ BACKWARD MOTION VECTORS =====
      currSE.type = SE_MVD;
      dP          = &(currSlice->partArr[partMap[SE_MVD]]);

      {
        currSE.mb_ext_level = mb_ext_level;
        currSE.reading = readMVD_CABAC32;
      }

      for (j0=0; j0<(4<<mb_ext_level); j0+=step_v0)
      {
        for (i0=0; i0<(4<<mb_ext_level); i0+=step_h0)
        {
          k=2*(j0>>(mb_ext_level+1))+(i0>>(mb_ext_level+1));
          if ((currMB->b8pdir[k]==1 || currMB->b8pdir[k]==2) && (currMB->b8mode[k]!=0))//has backward vector
          {
            mv_mode  = currMB->b8mode[k];
            step_h   = BLOCK_STEP [mv_mode][0]<<mb_ext_level;
            step_v   = BLOCK_STEP [mv_mode][1]<<mb_ext_level;
            refframe = dec_picture->ref_idx[LIST_1][img->block_y+j0][img->block_x+i0];

            for (j=j0; j<j0+step_v0; j+=step_v)
            {
              for (i=i0; i<i0+step_h0; i+=step_h)
              {
                j4 = img->block_y+j;
                i4 = img->block_x+i;

                // first make mv-prediction
#ifdef ADAPTIVE_MV_RES
                if (img->mv_res)
                {
                  TRACE_STRING("mv_res");
                  img->subblock_x = i; // position used for context determination
                  img->subblock_y = j; // position used for context determination
                  currSE.reading = readMV_RES_4_8_CABAC32;
                  currSE.value2 = 0; // identifies the component; only used for context determination
                  dP->readSyntaxElement(&currSE,img,inp,dP);
                  mv_shift = currSE.value1;
                  currSE.reading = readMVD_CABAC32;
                }
                else
                  mv_shift=0;
#endif            
#ifdef MV_COMPETITION
                if (mv_comp.mv_competition > 0)
                {
                  int mvd[2];
                  for (k=0; k < 2; k++) 
                  {
                    TRACE_STRING("mvd_l1");

                    img->subblock_x = i; // position used for context determination
                    img->subblock_y = j; // position used for context determination
#ifdef ADAPTIVE_MV_RES
                    currSE.value2 = (((k<<1) +1)<<1)+mv_shift; // identifies the component; only used for context determination
#else
                    currSE.value2   = (k<<1) +1; // identifies the component; only used for context determination
#endif
                    dP->readSyntaxElement(&currSE,img,inp,dP);
                    curr_mvd = currSE.value1; 

                    mvd[k] = curr_mvd;
                  }

                  SetMotionVectorPredictor_Competition32 (img,inp, pmv, pmv+1, refframe, LIST_1, dec_picture->ref_idx, dec_picture->mv, i, j, 4*step_h, 4*step_v, mb_ext_level);

                  for (k=0; k < 2; k++) 
                  {
                    for(ii=0;ii<step_h;ii++)
                    {
                      for(jj=0;jj<step_v;jj++)
                      {
#ifdef ADAPTIVE_MV_RES
                        dec_picture->mv  [LIST_1][j4+jj][i4+ii][k] = (mvd[k] + roundInt(pmv[k],mv_shift))<<mv_shift;
                        mb32_mvd         [LIST_1][j+jj] [i+ii] [k]= mvd[k]<<mv_shift;
#else
                        dec_picture->mv        [LIST_1][j4+jj][i4+ii][k] = mvd[k] + pmv[k];
                        //the mvd for the whole 32x32 is saved in mb32_mvd and will be copied to each 16x16 block's mvd in write_macroblock_cluster
                        mb32_mvd  [LIST_1][j +jj][i +ii][k] = mvd[k];
#endif
                      }
                    }

                    //only save mvd for the upper left 16x16 block for the context of the 2nd partition of 32x16 and 16x32
                    if(j==0 && i==0)
                    {
                      for(ii=0;ii<4;ii++)
                      {
                        for(jj=0;jj<4;jj++)
                        {
#ifdef ADAPTIVE_MV_RES
                          currMB->mvd      [LIST_1][j+jj] [i+ii] [k] = mvd[k]<<mv_shift;
#else
                          currMB->mvd      [LIST_1][j+jj] [i+ii] [k] = mvd[k];
#endif
                        }
                      }
                    }
                  }
                }
                else

#endif
                {
                  SetMotionVectorPredictor32 (img, pmv, pmv+1, refframe, LIST_1, dec_picture->ref_idx, dec_picture->mv, i, j, 4*step_h, 4*step_v, mb_ext_level);

                  for (k=0; k < 2; k++) 
                  {
                    TRACE_STRING("mvd_l1");

                    img->subblock_x = i; // position used for context determination
                    img->subblock_y = j; // position used for context determination
#ifdef ADAPTIVE_MV_RES
                    currSE.value2 = (((k<<1) +1)<<1)+mv_shift; // identifies the component; only used for context determination
#else
                    currSE.value2   = (k<<1) +1; // identifies the component; only used for context determination
#endif
                    dP->readSyntaxElement(&currSE,img,inp,dP);
                    curr_mvd = currSE.value1; 

#ifdef ADAPTIVE_MV_RES
                    motion_vector=(curr_mvd+(roundInt(pmv[k],mv_shift)))<<mv_shift;           /* find motion vector */
#else
                    motion_vector=curr_mvd+pmv[k];           /* find motion vector */
#endif  

                    for(ii=0;ii<step_h;ii++)
                    {
                      for(jj=0;jj<step_v;jj++)
                      {
                        dec_picture->mv  [LIST_1][j4+jj][i4+ii][k] = motion_vector;
                      }
                    }

                    //only save mvd for the upper left 16x16 block for the context of the 2nd partition of 32x16 and 16x32
                    if(j==0 && i==0)
                    {
                      for(ii=0;ii<4;ii++)
                      {
                        for(jj=0;jj<4;jj++)
                        {
#ifdef ADAPTIVE_MV_RES
                          currMB->mvd      [LIST_1][j+jj] [i+ii] [k] = curr_mvd<<mv_shift;
#else
                          currMB->mvd      [LIST_1][j+jj] [i+ii] [k] = curr_mvd;
#endif 
                        }
                      }
                    }
                    //the mvd for the whole 32x32 is saved in mb32_mvd and will be copied to each 16x16 block's mvd in write_macroblock_cluster
                    for(ii=0;ii<step_h;ii++)
                    {
                      for(jj=0;jj<step_v;jj++)
                      {
#ifdef ADAPTIVE_MV_RES
                        mb32_mvd      [LIST_1][j+jj] [i+ii] [k] = curr_mvd<<mv_shift;
#else
                        mb32_mvd      [LIST_1][j+jj] [i+ii] [k] = curr_mvd;
#endif
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
      // record reference picture Ids for deblocking decisions
      for(i4=img->block_x;i4<(img->block_x+(4<<mb_ext_level));i4++)
        for(j4=img->block_y;j4<(img->block_y+(4<<mb_ext_level));j4++)
        {
          if(dec_picture->ref_idx[LIST_0][j4][i4]>=0)
            dec_picture->ref_pic_id[LIST_0][j4][i4] = dec_picture->ref_pic_num[img->current_slice_nr][LIST_0][(short)dec_picture->ref_idx[LIST_0][j4][i4]];
          else
            dec_picture->ref_pic_id[LIST_0][j4][i4] = INT64_MIN;
          if(dec_picture->ref_idx[LIST_1][j4][i4]>=0)
            dec_picture->ref_pic_id[LIST_1][j4][i4] = dec_picture->ref_pic_num[img->current_slice_nr][LIST_1][(short)dec_picture->ref_idx[LIST_1][j4][i4]];  
          else
            dec_picture->ref_pic_id[LIST_1][j4][i4] = INT64_MIN;  
        }
  }


  /*!
  ************************************************************************
  * \brief
  *    initializes the current macroblock
  ************************************************************************
  */
  void start_macroblock(struct img_par *img,struct inp_par *inp, int CurrentMBInScanOrder)
  {
    //int i,j,k,l;
    Macroblock *currMB; 

    assert (img->current_mb_nr < img->PicSizeInMbs);

    currMB = &img->mb_data[img->current_mb_nr];
#ifdef GEOM
    currMB->geo_mode = -1;
#endif
    /* Update coordinates of the current macroblock */
    img->mb_x = PicPos[img->current_mb_nr][0];
    img->mb_y = PicPos[img->current_mb_nr][1];

    /* Define vertical positions */
    img->block_y = img->mb_y * BLOCK_SIZE;      /* luma block position */
    img->pix_y   = img->mb_y * MB_BLOCK_SIZE;   /* luma macroblock position */
    img->pix_c_y = img->mb_y * img->mb_cr_size_y; /* chroma macroblock position */

    /* Define horizontal positions */
    img->block_x = img->mb_x * BLOCK_SIZE;      /* luma block position */
    img->pix_x   = img->mb_x * MB_BLOCK_SIZE;   /* luma pixel position */

    img->pix_c_x = img->mb_x * img->mb_cr_size_x; /* chroma pixel position */

    // Save the slice number of this macroblock. When the macroblock below
    // is coded it will use this to decide if prediction for above is possible
    currMB->slice_nr = img->current_slice_nr;

    if (img->current_slice_nr >= MAX_NUM_SLICES)
    {
      error ("maximum number of supported slices exceeded, please recompile with increased value for MAX_NUM_SLICES", 200);
    }

    dec_picture->slice_id[img->mb_y][img->mb_x] = img->current_slice_nr;
    if (img->current_slice_nr > dec_picture->max_slice_id)
    {
      dec_picture->max_slice_id=img->current_slice_nr;
    }

    if(img->type != I_SLICE) 
      CheckAvailabilityOfNeighbors_Z();
    else
      CheckAvailabilityOfNeighbors(img);

    // Reset syntax element entries in MB struct
    currMB->qp          = img->qp ;
    currMB->mb_type     = 0;
    currMB->delta_quant = 0;
    currMB->cbp         = 0;
    currMB->cbp_blk     = 0;
    currMB->c_ipred_mode= DC_PRED_8; //GB

    memset(currMB->mvd, 0, 2*BLOCK_MULTIPLE*BLOCK_MULTIPLE*2 * sizeof(int));
    currMB->cbp_bits   = 0;

    // initialize img->m7 for ABT
    memset(img->m7, 0, MB_BLOCK_SIZE*MB_BLOCK_SIZE*sizeof(int));

    // store filtering parameters for this MB 
    currMB->LFDisableIdc = img->currentSlice->LFDisableIdc;
    currMB->LFAlphaC0Offset = img->currentSlice->LFAlphaC0Offset;
    currMB->LFBetaOffset = img->currentSlice->LFBetaOffset;

  }

  /*!
  ************************************************************************
  * \brief
  *    initialize macroblock position and size parameters in img
  ************************************************************************
  */
  void init_mb_coordinates(struct img_par *img,struct inp_par *inp, int CurrentMBInScanOrder)
  { 
    assert (img->current_mb_nr < img->PicSizeInMbs);

    /* Update coordinates of the current macroblock */
    img->mb_x = (img->current_mb_nr)%(img->width/MB_BLOCK_SIZE);
    img->mb_y = (img->current_mb_nr)/(img->width/MB_BLOCK_SIZE);

    /* Define vertical positions */
    img->block_y = img->mb_y * BLOCK_SIZE;      /* luma block position */
    img->pix_y   = img->mb_y * MB_BLOCK_SIZE;   /* luma macroblock position */
    img->pix_c_y = img->mb_y * img->mb_cr_size_y; /* chroma macroblock position */

    /* Define horizontal positions */
    img->block_x = img->mb_x * BLOCK_SIZE;      /* luma block position */
    img->pix_x   = img->mb_x * MB_BLOCK_SIZE;   /* luma pixel position */

    img->pix_c_x = img->mb_x * img->mb_cr_size_x; /* chroma pixel position */

    if (img->current_slice_nr >= MAX_NUM_SLICES)
    {
      error ("maximum number of supported slices exceeded, please recompile with increased value for MAX_NUM_SLICES", 200);
    }  

    // initialize img->m7 for ABT
    memset(img->m7, 0, MB_BLOCK_SIZE*MB_BLOCK_SIZE*sizeof(int));

  }



  /*!
  ************************************************************************
  * \brief
  *    set coordinates of the next macroblock
  *    check end_of_slice condition 
  ************************************************************************
  */
  int exit_macroblock(struct img_par *img,struct inp_par *inp,int eos_bit)
  {
    //! The if() statement below resembles the original code, which tested 
    //! img->current_mb_nr == img->PicSizeInMbs.  Both is, of course, nonsense
    //! In an error prone environment, one can only be sure to have a new
    //! picture by checking the tr of the next slice header!

    // printf ("exit_macroblock: FmoGetLastMBOfPicture %d, img->current_mb_nr %d\n", FmoGetLastMBOfPicture(), img->current_mb_nr);
    img->num_dec_mb++;

    if (img->num_dec_mb == img->PicSizeInMbs)
      //  if (img->current_mb_nr == FmoGetLastMBOfPicture(currSlice->structure))
    {
      //thb
      /*
      if (currSlice->next_header != EOS)
      currSlice->next_header = SOP;
      */
      //the
      //    assert (nal_startcode_follows (img, inp, eos_bit) == TRUE);
      return TRUE;
    }
    // ask for last mb in the slice  UVLC
    else
    {
      // printf ("exit_macroblock: Slice %d old MB %d, now using MB %d\n", img->current_slice_nr, img->current_mb_nr, FmoGetNextMBNr (img->current_mb_nr));

      img->current_mb_nr = FmoGetNextMBNr (img->current_mb_nr);

      if (img->current_mb_nr == -1)     // End of Slice group, MUST be end of slice
      {
        assert (nal_startcode_follows (img, inp, eos_bit) == TRUE);
        return TRUE;
      }

      if(nal_startcode_follows(img, inp, eos_bit) == FALSE) 
        return FALSE;

      if(img->type == I_SLICE  || img->type == SI_SLICE || active_pps->entropy_coding_mode_flag == CABAC)
        return TRUE;
      if(img->cod_counter<=0)
        return TRUE;
      return FALSE;
    }
  }

  /*!
  ************************************************************************
  * \brief
  *    Interpret the mb mode for P-Frames
  ************************************************************************
  */
  void interpret_mb_mode_P(struct img_par *img)
  {
    int i;
    const int ICBPTAB[6] = {0,16,32,15,31,47};
    Macroblock *currMB = &img->mb_data[img->current_mb_nr];
    int         mbmode = currMB->mb_type;

#define MODE_IS_P8x8  (mbmode==4 || mbmode==5)

    if(mbmode <4)
    {
      currMB->mb_type = mbmode;
      for (i=0;i<4;i++)
      {
        currMB->b8mode[i]   = mbmode;
        currMB->b8pdir[i]   = 0;
      }
    }
    else if(active_pps->entropy_coding_mode_flag == CABAC && (mbmode==4 || mbmode==5))
    {
      currMB->mb_type = P8x8;
      img->allrefzero = 0;
    }

    else if(mbmode==6)
    {
      currMB->mb_type = I4MB;
      for (i=0;i<4;i++)
      {
        currMB->b8mode[i] = IBLOCK;
        currMB->b8pdir[i] = -1;
      }
    }
    else if(mbmode==31)
    {
      currMB->mb_type=IPCM;

      for (i=0;i<4;i++) 
      {
        currMB->b8mode[i]=0; currMB->b8pdir[i]=-1; 
      }
      currMB->cbp= -1;
      currMB->i16mode = 0;
    }
    else
    {
      currMB->mb_type = I16MB;
      for (i=0;i<4;i++) {currMB->b8mode[i]=0; currMB->b8pdir[i]=-1; }
      currMB->cbp= ICBPTAB[(mbmode-7)>>2];
      currMB->i16mode = (mbmode-7) & 0x03;
    }
  }

  /*!
  ************************************************************************
  * \brief
  *    Interpret the mb mode for I-Frames
  ************************************************************************
  */
  void interpret_mb_mode_I(struct img_par *img)
  {
    int i;
    const int ICBPTAB[6] = {0,16,32,15,31,47};
    Macroblock *currMB   = &img->mb_data[img->current_mb_nr];
    int         mbmode   = currMB->mb_type;

    if (mbmode==0)
    {
      currMB->mb_type = I4MB;
      for (i=0;i<4;i++) {currMB->b8mode[i]=IBLOCK; currMB->b8pdir[i]=-1; }
    }
    else if(mbmode==25)
    {
      currMB->mb_type=IPCM;

      for (i=0;i<4;i++) {currMB->b8mode[i]=0; currMB->b8pdir[i]=-1; }
      currMB->cbp= -1;
      currMB->i16mode = 0;

    }
    else
    {
      currMB->mb_type = I16MB;
      for (i=0;i<4;i++) {currMB->b8mode[i]=0; currMB->b8pdir[i]=-1; }
      currMB->cbp= ICBPTAB[(mbmode-1)>>2];
      currMB->i16mode = (mbmode-1) & 0x03;
    }
  }

  /*!
  ************************************************************************
  * \brief
  *    Interpret the mb mode for B-Frames
  ************************************************************************
  */
  void interpret_mb_mode_B(struct img_par *img)
  {
    const int ICBPTAB[6] = {0,16,32,15,31,47};
    Macroblock *currMB = &img->mb_data[img->current_mb_nr];

    int i, mbmode;
    int mbtype  = currMB->mb_type;
    int *b8mode = currMB->b8mode;
    int *b8pdir = currMB->b8pdir;

    //--- set mbtype, b8type, and b8pdir ---
    if (mbtype==0)       // direct
    {
      mbmode=0;       for(i=0;i<4;i++) {b8mode[i]=0;          b8pdir[i]=2; }
    }
    else if (mbtype==23) // intra4x4
    {
      mbmode=I4MB;    for(i=0;i<4;i++) {b8mode[i]=IBLOCK;     b8pdir[i]=-1; }
    }
    else if ((mbtype>23) && (mbtype<48) ) // intra16x16
    {
      mbmode=I16MB;   for(i=0;i<4;i++) {b8mode[i]=0;          b8pdir[i]=-1; }
      currMB->cbp     = ICBPTAB[(mbtype-24)>>2];
      currMB->i16mode = (mbtype-24) & 0x03;
    }
    else if (mbtype==22) // 8x8(+split)
    {
      mbmode=P8x8;       // b8mode and pdir is transmitted in additional codewords
    }
    else if(mbtype==48)
    {
      mbmode=IPCM;
      for (i=0;i<4;i++) {currMB->b8mode[i]=0; currMB->b8pdir[i]=-1; }
      currMB->cbp= -1;
      currMB->i16mode = 0;
    }
    else if (mbtype==1)   // 16x16
    {
      mbmode=1;    for(i=0;i<4;i++)     { b8mode[i]=1; }
      b8pdir[1]=b8pdir[0]; 
      b8pdir[2]=b8pdir[0]; 
      b8pdir[3]=b8pdir[0]; 
    }
    else if (mbtype==2) // 16x8
    {
      mbmode=2;    for(i=0;i<4;i++)     { b8mode[i]=2; }
      b8pdir[1]=b8pdir[0]; 
      b8pdir[2]=b8pdir[3]; 
    }
    else          //8x16
    {
      mbmode=3;    for(i=0;i<4;i++)     { b8mode[i]=3; }
      b8pdir[1]=b8pdir[3]; 
      b8pdir[2]=b8pdir[0]; 
    }


    currMB->mb_type = mbmode;
  }

  /*!
  ************************************************************************
  * \brief
  *    Interpret the mb mode for SI-Frames
  ************************************************************************
  */
  void interpret_mb_mode_SI(struct img_par *img)
  {
    int i;
    const int ICBPTAB[6] = {0,16,32,15,31,47};
    Macroblock *currMB   = &img->mb_data[img->current_mb_nr];
    int         mbmode   = currMB->mb_type;

    if (mbmode==0)
    {
      currMB->mb_type = SI4MB;
      for (i=0;i<4;i++) {currMB->b8mode[i]=IBLOCK; currMB->b8pdir[i]=-1; }
      img->siblock[img->mb_x][img->mb_y]=1;
    }
    else if (mbmode==1)
    {
      currMB->mb_type = I4MB;
      for (i=0;i<4;i++) {currMB->b8mode[i]=IBLOCK; currMB->b8pdir[i]=-1; }
    }
    else if(mbmode==26)
    {
      currMB->mb_type=IPCM;

      for (i=0;i<4;i++) {currMB->b8mode[i]=0; currMB->b8pdir[i]=-1; }
      currMB->cbp= -1;
      currMB->i16mode = 0;

    }

    else
    {
      currMB->mb_type = I16MB;
      for (i=0;i<4;i++) {currMB->b8mode[i]=0; currMB->b8pdir[i]=-1; }
      currMB->cbp= ICBPTAB[(mbmode-1)>>2];
      currMB->i16mode = (mbmode-2) & 0x03;
    }
  }
  /*!
  ************************************************************************
  * \brief
  *    init macroblock I and P frames
  ************************************************************************
  */
  void init_macroblock(struct img_par *img)
  {
    int i,j;

    for (i=0;i<BLOCK_SIZE;i++)
    {                           // reset vectors and pred. modes
      for(j=0;j<BLOCK_SIZE;j++)
      {
        dec_picture->mv[LIST_0][img->block_y+j][img->block_x+i][0]=0;
        dec_picture->mv[LIST_0][img->block_y+j][img->block_x+i][1]=0;
        dec_picture->mv[LIST_1][img->block_y+j][img->block_x+i][0]=0;
        dec_picture->mv[LIST_1][img->block_y+j][img->block_x+i][1]=0;

        img->ipredmode[img->block_x+i][img->block_y+j] = DC_PRED;
      }
    }

    for (j=0; j<BLOCK_SIZE; j++)
      for (i=0; i<BLOCK_SIZE; i++)
      {
        dec_picture->ref_idx[LIST_0][img->block_y+j][img->block_x+i] = -1;
        dec_picture->ref_idx[LIST_1][img->block_y+j][img->block_x+i] = -1;
        dec_picture->ref_pic_id[LIST_0][img->block_y+j][img->block_x+i] = INT64_MIN;
        dec_picture->ref_pic_id[LIST_1][img->block_y+j][img->block_x+i] = INT64_MIN;
      }
  }


  /*!
  ************************************************************************
  * \brief
  *    Sets mode for 8x8 block
  ************************************************************************
  */
  void SetB8Mode (struct img_par* img, Macroblock* currMB, int value, int i)
  {
    static const int p_v2b8 [ 5] = {4, 5, 6, 7, IBLOCK};
    static const int p_v2pd [ 5] = {0, 0, 0, 0, -1};
    static const int b_v2b8 [14] = {0, 4, 4, 4, 5, 6, 5, 6, 5, 6, 7, 7, 7, IBLOCK};
    static const int b_v2pd [14] = {2, 0, 1, 2, 0, 0, 1, 1, 2, 2, 0, 1, 2, -1};

    if (img->type==B_SLICE)
    {
      currMB->b8mode[i]   = b_v2b8[value];
      currMB->b8pdir[i]   = b_v2pd[value];

    }
    else
    {
      currMB->b8mode[i]   = p_v2b8[value];
      currMB->b8pdir[i]   = p_v2pd[value];
    }

  }


  void reset_coeffs()
  {
    int i, j, iii, jjj;

    // reset luma coeffs
    for (i=0;i<BLOCK_SIZE;i++)
    { 
      for (j=0;j<BLOCK_SIZE;j++)
        for(iii=0;iii<BLOCK_SIZE;iii++)
          for(jjj=0;jjj<BLOCK_SIZE;jjj++)
            img->cof[i][j][iii][jjj]=0;
    }

    // reset chroma coeffs
    for (j=4;j<(4+img->num_blk8x8_uv);j++)
    { 
      for (i=0;i<4;i++)
        for (iii=0;iii<4;iii++)
          for (jjj=0;jjj<4;jjj++)
            img->cof[i][j][iii][jjj]=0;
    }

    // CAVLC
    if(active_pps->entropy_coding_mode_flag==UVLC)//tm_vlc
    {
      for (i=0; i < 4; i++)
        for (j=0; j < (4 + img->num_blk8x8_uv); j++)
          img->nz_coeff[img->current_mb_nr][i][j]=0;  
    }

  }

  void field_flag_inference()
  {
    Macroblock *currMB = &img->mb_data[img->current_mb_nr];

    //currMB->mb_field = 0;
  }

  void set_chroma_qp(Macroblock* currMB)
  {
    int i;
    for (i=0; i<2; i++)
    {
      currMB->qpc[i] = Clip3 ( -img->bitdepth_chroma_qp_scale, 51, currMB->qp + dec_picture->chroma_qp_offset[i] );
      currMB->qpc[i] = currMB->qpc[i] < 0 ? currMB->qpc[i] : QP_SCALE_CR[currMB->qpc[i]];
    }
  }

  /*!
  ************************************************************************
  * \brief
  *    Get the syntax elements from the NAL
  * \param (mb32,mb16count)
  *        (1,0) - 32x32, 32x16, 16x32 mode info, mv info, coeff of upper left 16x16, dequant
  *        (1,1~)-                                         coeff of other      16x16
  *        (0,0) - 16x16 mode, upper left 16x16 partition mode, mv, coeff, dequant
  *        (0,1~)-                        16x16 partition mode, mv, coeff
  ************************************************************************
  */
  int read_one_macroblock32(struct img_par *img,struct inp_par *inp, int mb16count, int mb_ext_level)
  {
    int i;
    int mb32=0;

    SyntaxElement currSE;
    Macroblock *currMB = &img->mb_data[img->current_mb_nr];

    Slice *currSlice = img->currentSlice;
    DataPartition *dP;
    int *partMap = assignSE2partition[currSlice->dp_mode];
    Macroblock *topMB = NULL;
    int  prevMbSkipped = 0;
    int  img_block_y;
    //int  check_bottom, read_bottom, read_top;

    //currMB->mb_field = 0;
    currMB->qp = img->qp ;
    for (i=0; i<2; i++)
    {
      currMB->qpc[i] = Clip3 ( -img->bitdepth_chroma_qp_scale, 51, img->qp + dec_picture->chroma_qp_offset[i] );
      currMB->qpc[i] = currMB->qpc[i] < 0 ? currMB->qpc[i] : QP_SCALE_CR[currMB->qpc[i]];
    }

    currSE.type = SE_MBTYPE;

    //  read MB mode *****************************************************************
    dP = &(currSlice->partArr[partMap[currSE.type]]);

    //if (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag)   currSE.mapping = linfo_ue;

    if(img->type == I_SLICE || img->type == SI_SLICE)
    {
      if(active_pps->entropy_coding_mode_flag  == CABAC)
        CheckAvailabilityOfNeighborsCABAC();

      //  read MB type
      TRACE_STRING("mb_type");
      currSE.reading = readMB_typeInfo_CABAC;
      dP->readSyntaxElement(&currSE,img,inp,dP);

      currMB->mb_type = currSE.value1;
      if(!dP->bitstream->ei_flag)
        currMB->ei_flag = 0;
    }
    // non I/SI-slice CABAC
    else if (active_pps->entropy_coding_mode_flag == CABAC)
    {
      CheckAvailabilityOfNeighborsCABAC();

      /********************************************
      *MB type for mode 32x32, 32x16, 16x32, 16x16
      *******************************************/
      if(mb16count==0)
      {
        read_mbtype(img, inp, 1);

        MB32.mb_type = currMB->mb_type; //untill this line, MB32.mb_type=1 could mean mode 1 of P or direct mode of B
        MB32.cbp     = currMB->cbp;
        memcpy(MB32.b8mode, currMB->b8mode, sizeof(int)*4);
        memcpy(MB32.b8pdir, currMB->b8pdir, sizeof(int)*4);
#ifdef GEOM
        MB32.geo_mode = currMB->geo_mode;
#endif
        assert((MB32.mb_type >=0 && MB32.mb_type<=3)||MB32.mb_type==P8x8);      
      }

      if( MB32.mb_type == P8x8)
        mb32=0;
      else
        mb32=1;

      // read MB skip_flag
      //if (img->MbaffFrameFlag && (img->current_mb_nr%2 == 0||prevMbSkipped))
      //  field_flag_inference();

      if(mb32==0)
      {
        read_mbtype(img, inp, 0);
      }
    }
    // VLC Non-Intra
    else
    {
      if(img->cod_counter == -1)
      {
        TRACE_STRING("mb_skip_run");
        dP->readSyntaxElement(&currSE,img,inp,dP);
        img->cod_counter = currSE.value1;
      }
      if (img->cod_counter==0)
      {

        // read MB type
        TRACE_STRING("mb_type");
        dP->readSyntaxElement(&currSE,img,inp,dP);
        if(img->type == P_SLICE || img->type == SP_SLICE)
          currSE.value1++;
        currMB->mb_type = currSE.value1;
        if(!dP->bitstream->ei_flag)
          currMB->ei_flag = 0;
        img->cod_counter--;
        currMB->skip_flag = 0;
      } 
      else
      {
        img->cod_counter--;
        currMB->mb_type = 0;
        currMB->ei_flag = 0;
        currMB->skip_flag = 1;

      }
    }

    //dec_picture->mb_field[img->current_mb_nr] = 0;

    img->siblock[img->mb_x][img->mb_y]=0;

    if ((img->type==P_SLICE ))    // inter frame
      interpret_mb_mode_P(img);
    else if (img->type==I_SLICE)                                  // intra frame
      interpret_mb_mode_I(img);
    else if ((img->type==B_SLICE))       // B frame
      interpret_mb_mode_B    (img);
    else if ((img->type==SP_SLICE))     // SP frame
      interpret_mb_mode_P(img);
    else if (img->type==SI_SLICE)     // SI frame
      interpret_mb_mode_SI(img);

    //init NoMbPartLessThan8x8Flag
    currMB->NoMbPartLessThan8x8Flag = (IS_DIRECT(currMB) && !(active_sps->direct_8x8_inference_flag))? 0: 1;

    //====== READ 8x8 SUB-PARTITION MODES (modes of 8x8 blocks) and Intra VBST block modes ======
    if (IS_P8x8 (currMB))
    {
      currSE.type    = SE_MBTYPE;
      dP = &(currSlice->partArr[partMap[SE_MBTYPE]]);

      for (i=0; i<4; i++)
      {
        if (active_pps->entropy_coding_mode_flag ==UVLC || dP->bitstream->ei_flag) currSE.mapping = linfo_ue;
        else                                                  currSE.reading = readB8_typeInfo_CABAC;

        TRACE_STRING("sub_mb_type");
        dP->readSyntaxElement (&currSE, img, inp, dP);
        SetB8Mode (img, currMB, currSE.value1, i);

        //set NoMbPartLessThan8x8Flag for P8x8 mode
        currMB->NoMbPartLessThan8x8Flag &= (currMB->b8mode[i]==0 && active_sps->direct_8x8_inference_flag) || 
          (currMB->b8mode[i]==4);
      }
      //--- init macroblock data ---
      init_macroblock       (img);
      readMotionInfoFromNAL (img, inp);
    }


    //============= Transform Size Flag for INTRA MBs =============
    //-------------------------------------------------------------
    //transform size flag for INTRA_4x4 and INTRA_8x8 modes
    if (currMB->mb_type == I4MB && img->Transform8x8Mode)
    {
      currSE.type   =  SE_HEADER;
      dP = &(currSlice->partArr[partMap[SE_HEADER]]);
      currSE.reading = readMB_transform_size_flag_CABAC;
      TRACE_STRING("transform size 8x8 flag");

      // read UVLC transform_size_8x8_flag
      //if (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag)
      //{
      //  currSE.len = 1;
      //  readSyntaxElement_FLC(&currSE, dP->bitstream);
      //} 
      //else 
      {
        dP->readSyntaxElement(&currSE,img,inp,dP);
      } 

      currMB->luma_transform_size_8x8_flag = currSE.value1;

      if (currMB->luma_transform_size_8x8_flag)
      {
        currMB->mb_type = I8MB;
        for (i=0;i<4;i++) 
        {
          currMB->b8mode[i]=I8MB;
          currMB->b8pdir[i]=-1;
        }
      }
    }
    else
    {
      currMB->luma_transform_size_8x8_flag = 0;
    }

    if(active_pps->constrained_intra_pred_flag && (img->type==P_SLICE|| img->type==B_SLICE))        // inter frame
    {
      if( !IS_INTRA(currMB) )
      {
        img->intra_block[img->current_mb_nr] = 0;
      }
    }


    dP = &(currSlice->partArr[partMap[SE_CBP_INTRA]]);
    if(IS_INTRA (currMB) && dP->bitstream->ei_flag && img->number)
    {
      currMB->mb_type = 0;
      currMB->ei_flag = 1;
      for (i=0;i<4;i++)
      {
        currMB->b8mode[i]=currMB->b8pdir[i]=0;
      }
    }
    dP = &(currSlice->partArr[partMap[currSE.type]]);

    //--- init macroblock data ---
    if (!IS_P8x8 (currMB))
    {
      if(mb32==0)
      {
        init_macroblock       (img);
      }
      else if(mb32==1 && mb16count==0)
      {
        init_macroblock32     (img, 1);
      }
    }

    if (IS_DIRECT (currMB) && img->cod_counter >= 0)
    {
      currMB->cbp = 0;

      reset_coeffs();

      if (active_pps->entropy_coding_mode_flag ==CABAC)
        img->cod_counter=-1;

      return DECODE_MB;
    }

    if (((mb32 && mb16count==0) || !mb32)
      && IS_COPY (currMB) ) //keep last macroblock
    {
      int i, j, k;
      short pmv[2];
      int zeroMotionAbove;
      int zeroMotionLeft;
      PixelPos mb_a, mb_b;
      int      a_mv_y = 0;
      int      a_ref_idx = 0;
      int      b_mv_y = 0;
      int      b_ref_idx = 0;

      getLuma4x4Neighbour(img->current_mb_nr,0,0,-1, 0,&mb_a);
      getLuma4x4Neighbour(img->current_mb_nr,0,0, 0,-1,&mb_b);

      if (mb_a.available)
      {
        a_mv_y    = dec_picture->mv[LIST_0][mb_a.pos_y][mb_a.pos_x][1];
        a_ref_idx = dec_picture->ref_idx[LIST_0][mb_a.pos_y][mb_a.pos_x];
      }

      if (mb_b.available)
      {
        b_mv_y    = dec_picture->mv[LIST_0][mb_b.pos_y][mb_b.pos_x][1];
        b_ref_idx = dec_picture->ref_idx[LIST_0][mb_b.pos_y][mb_b.pos_x];
      }

      zeroMotionLeft  = !mb_a.available ? 1 : a_ref_idx==0 && dec_picture->mv[LIST_0][mb_a.pos_y][mb_a.pos_x][0]==0 && a_mv_y==0 ? 1 : 0;
      zeroMotionAbove = !mb_b.available ? 1 : b_ref_idx==0 && dec_picture->mv[LIST_0][mb_b.pos_y][mb_b.pos_x][0]==0 && b_mv_y==0 ? 1 : 0;

      if(currMB->skip_flag)
      {
        currMB->cbp = 0;
        reset_coeffs();
      }

      img_block_y   = img->block_y;

#ifdef MV_COMPETITION
      if (mv_comp.mv_competition > 0)
      {
        short prediction_mode_for_skip;
        short predictor;

        if((mb32 && mb16count==0) || mb32==0)
        {
          for (prediction_mode_for_skip = 0; prediction_mode_for_skip<mv_comp.nb_mode_for_skip; prediction_mode_for_skip++)
          {
            if(mb32==0)
              SetMotionVectorPredictor_Skip  (pmv, pmv+1, dec_picture->ref_idx, dec_picture->mv, 0, LIST_0, 0, 0, 16, 16,prediction_mode_for_skip);
            else
              SetMotionVectorPredictor_Skip32(pmv, pmv+1, dec_picture->ref_idx, dec_picture->mv, 0, LIST_0, 0, 0, 32, 32,prediction_mode_for_skip, 1);

            mv_comp.mv_pred_skip[prediction_mode_for_skip][0] = pmv[0];
            mv_comp.mv_pred_skip[prediction_mode_for_skip][1] = pmv[1];
          }
          if ((mv_comp.nb_mode_for_skip> 1) && (read_index_for_skip_mode() == TRUE))
            predictor = readPredictorForSkip(img,inp);
          else 
            predictor = 0;

          for(i=0;i<(BLOCK_SIZE<<mb32);i++)
            for(j=0;j<(BLOCK_SIZE<<mb32);j++)
              for (k=0;k<2;k++)
                dec_picture->mv[LIST_0][img_block_y+j][img->block_x+i][k] = mv_comp.mv_pred_skip[predictor][k];
        }
      }
      else
#endif
      {
        if (zeroMotionAbove || zeroMotionLeft)
        {
          if((mb32&&mb16count==0) || !mb32)
          {
            for(i=0;i<BLOCK_SIZE<<mb32;i++)
              for(j=0;j<BLOCK_SIZE<<mb32;j++)
                for (k=0;k<2;k++)
                  dec_picture->mv[LIST_0][img->block_y+j][img->block_x+i][k] = 0;
          }
        }
        else
        {
          if((mb32&&mb16count==0) || !mb32)
          {
            if(mb32)
              SetMotionVectorPredictor32 (img, pmv, pmv+1, 0, LIST_0, dec_picture->ref_idx, dec_picture->mv, 0, 0, 16<<mb32, 16<<mb32, 1);
            else
              SetMotionVectorPredictor (img, pmv, pmv+1, 0, LIST_0, dec_picture->ref_idx, dec_picture->mv, 0, 0, 16, 16);

            for(i=0;i<BLOCK_SIZE<<mb32;i++)
              for(j=0;j<BLOCK_SIZE<<mb32;j++)
                for (k=0;k<2;k++)
                {
                  dec_picture->mv[LIST_0][img_block_y+j][img->block_x+i][k] = pmv[k];
                }
          }
        }
      }


      if((mb32 && mb16count==0) || !mb32)
      {
        for(i=0;i<BLOCK_SIZE<<mb32;i++)
          for(j=0;j<BLOCK_SIZE<<mb32;j++)
          {
            dec_picture->ref_idx[LIST_0][img_block_y+j][img->block_x+i] = 0;
            dec_picture->ref_pic_id[LIST_0][img_block_y+j][img->block_x+i] = 
              dec_picture->ref_pic_num[img->current_slice_nr][LIST_0][(short)dec_picture->ref_idx[LIST_0][img_block_y+j][img->block_x+i]];
          }        
      }

      if(currMB->skip_flag)
        return DECODE_MB;
    }
    if(currMB->mb_type!=IPCM)
    {

      // intra prediction modes for a macroblock 4x4 **********************************************
      read_ipred_modes(img,inp);

      // read inter frame vector data *********************************************************
      if(mb32 && mb16count==0)
      {
        if (MB32.mb_type != 0)
        {
#ifdef GEOM
          if(currMB->geo_mode != -1)
            readMotionInfoFromNAL_geo  (img, inp, 1);
          else
#endif
            readMotionInfoFromNAL32 (img, inp, 1);
        }
      }
      else if(mb32==0 && IS_INTERMV (currMB) && (!IS_P8x8(currMB)))
      {
#ifdef GEOM
        if(currMB->geo_mode != -1)
          readMotionInfoFromNAL_geo  (img, inp, 0);
        else
#endif     
          readMotionInfoFromNAL (img, inp);
      }
      // read CBP and Coeffs  ***************************************************************

      if(mb32 && mb16count==0)
      {      
        //MB32.cbp=readCBP32(img,inp,!delta_qp_sent);//with dquant
        MB32.cbp = readCBP_Big(img,inp);
        readDeltaQP(img,inp,MB32.cbp);
      }
      if(!mb32 || (mb32 && MB32.cbp !=0 ) )
        readCBPandCoeffsFromNAL (img,inp, mb32,0,mb16count);
    }
    else
    {
      //read pcm_alignment_zero_bit and pcm_byte[i] 

      // here dP is assigned with the same dP as SE_MBTYPE, because IPCM syntax is in the 
      // same category as MBTYPE
      dP = &(currSlice->partArr[partMap[SE_MBTYPE]]);
      readIPCMcoeffsFromNAL(img,inp,dP);
    }

    return DECODE_MB;
  }

  int mbtype_mapping[]={3,0,1,2};
  void re_mapping(int *mbtype_value, int mb_ext_level)
  {
    if(mb_ext_level>0)
      *mbtype_value = mbtype_mapping[*mbtype_value];
  }
#ifdef GEOM
  void readMB_typeInfo_CAVLC_geo( SyntaxElement *se,
  struct inp_par *inp,
  struct img_par *img,
    DataPartition *dP)
  {
    Macroblock *currMB = &img->mb_data[img->current_mb_nr];

    if(u_1 ("mb_geo_flag", dP->bitstream))
    {
      int i, geo_mode_bits, geo_mode=0;

      if(img->geo_scheme == 0)
        geo_mode_bits = 4;
      else
        geo_mode_bits = se->value2;


      for(i=0; i<geo_mode_bits; i++)
      {
        if(u_1 ("mb_geo_mode", dP->bitstream))
          geo_mode |= 1<<i;
      }
      currMB->geo_mode = geo_mode;
    }
  }
#endif


  void read_mbtype(struct img_par *img, struct inp_par *inp, int mb_ext_level)//MB64X64
  {
    Slice *currSlice = img->currentSlice;
    SyntaxElement currSE;
    Macroblock *currMB = &img->mb_data[img->current_mb_nr];
    int *partMap = assignSE2partition[currSlice->dp_mode];
    DataPartition *dP;
    currSE.type = SE_MBTYPE;

    //  read MB mode *****************************************************************
    dP = &(currSlice->partArr[partMap[currSE.type]]);

    // read skip flag
    //CheckAvailabilityOfNeighborsCABAC();
    TRACE_STRING("mb_skip_flag");
    currSE.reading = readMB_skip_flagInfo_CABAC;
    dP->readSyntaxElement(&currSE,img,inp,dP);

    currMB->mb_type   = currSE.value1;
    currMB->skip_flag = !(currSE.value1);
    currMB->cbp = currSE.value2;

    if(!dP->bitstream->ei_flag)
      currMB->ei_flag = 0;

    if ((img->type==B_SLICE) && currSE.value1==0 && currSE.value2==0)
      img->cod_counter=0;

    //CheckAvailabilityOfNeighborsCABAC();

    // read MB type
    if (currMB->mb_type != 0 )
    {
      if(mb_ext_level != 0)
        currSE.reading = readMB_typeInfo_CABAC32;
      else
        currSE.reading = readMB_typeInfo_CABAC;

      TRACE_STRING("mb_type");
      dP->readSyntaxElement(&currSE,img,inp,dP);
      currMB->mb_type = currSE.value1;
      currMB->b8pdir[0] = currSE.dir_partition1;
      currMB->b8pdir[3] = currSE.dir_partition2;

      if(!dP->bitstream->ei_flag)
        currMB->ei_flag = 0;
    }

    if ((img->type==P_SLICE ))    // inter frame
      interpret_mb_mode_P(img);
    else if (img->type==I_SLICE)                                  // intra frame
      interpret_mb_mode_I(img);
    else if ((img->type==B_SLICE))       // B frame
      interpret_mb_mode_B    (img);
    else if ((img->type==SP_SLICE))     // SP frame
      interpret_mb_mode_P(img);
    else if (img->type==SI_SLICE)     // SI frame
      interpret_mb_mode_SI(img);
#ifdef GEOM

    if(img->type != I_SLICE && currMB->mb_type == 3)
    {
      currSE.value2 = rho_bits[mb_ext_level]+THETA_BITS;
      currSE.reading = readMB_typeInfo_CABAC_geo;
      TRACE_STRING("mb_type");
      dP->readSyntaxElement(&currSE,img,inp,dP);
      if(!dP->bitstream->ei_flag)
        currMB->ei_flag = 0;
    }
#endif
  }

  //mb16count: counter of 16x16 (0~15 if mb_ext_level ==2; 0~3 if mb_ext_level==1)
  //mb_ext_level: 1 - this is a 16x16 of 32x32
  //              2 - this is a 16x16 of 64x64              
  int read_one_macroblock64(struct img_par *img,struct inp_par *inp, int mb32count, int mb16count, int mb_ext_level)
  {
    int i;
    int mb64=0, mb32=0; // mb64:0- 32x32 mode, 1-the others; mb32:0-16x16 mode, 1-the others

    SyntaxElement currSE;
    Macroblock *currMB = &img->mb_data[img->current_mb_nr];

    Slice *currSlice = img->currentSlice;
    DataPartition *dP;
    int *partMap = assignSE2partition[currSlice->dp_mode];
    Macroblock *topMB = NULL;
    int  prevMbSkipped = 0;
    int  img_block_y;

    //currMB->mb_field = 0;

    currMB->qp = img->qp ;
    for (i=0; i<2; i++)
    {
      currMB->qpc[i] = Clip3 ( -img->bitdepth_chroma_qp_scale, 51, img->qp + dec_picture->chroma_qp_offset[i] );
      currMB->qpc[i] = currMB->qpc[i] < 0 ? currMB->qpc[i] : QP_SCALE_CR[currMB->qpc[i]];
    }

    currSE.type = SE_MBTYPE;

    //  read MB mode *****************************************************************
    dP = &(currSlice->partArr[partMap[currSE.type]]);

    //if (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag)   currSE.mapping = linfo_ue;

    if(img->type == I_SLICE || img->type == SI_SLICE)
    {
      if(active_pps->entropy_coding_mode_flag  == CABAC)
        CheckAvailabilityOfNeighborsCABAC();

      //  read MB type
      TRACE_STRING("mb_type");
      currSE.reading = readMB_typeInfo_CABAC;
      dP->readSyntaxElement(&currSE,img,inp,dP);

      currMB->mb_type = currSE.value1;
      if(!dP->bitstream->ei_flag)
        currMB->ei_flag = 0;
    }
    // non I/SI-slice CABAC
    else //if (active_pps->entropy_coding_mode_flag == CABAC)
    {
      if (active_pps->entropy_coding_mode_flag == CABAC)
        CheckAvailabilityOfNeighborsCABAC();

      if(mb_ext_level == 2)
      {
        /********************************************
        *MB type for mode 64x64, 64x32, 32x64, 32x32
        *******************************************/
        if(mb32count==0 && mb16count==0)
        {
          read_mbtype(img, inp, 2);

          MB64.mb_type = currMB->mb_type; //untill this line, MB32.mb_type=1 could mean mode 1 of P or direct mode of B
          MB64.cbp     = currMB->cbp;
          memcpy(MB64.b8mode, currMB->b8mode, sizeof(int)*4);
          memcpy(MB64.b8pdir, currMB->b8pdir, sizeof(int)*4);
#ifdef GEOM
          MB64.geo_mode = currMB->geo_mode;
#endif
          assert((MB64.mb_type >=0 && MB64.mb_type<=3)||MB64.mb_type==P8x8);      
        }

        if( MB64.mb_type == P8x8)
          mb64=0;
        else
          mb64=1;
      }

      /********************************************
      *MB type for mode 32x32, 32x16, 16x32, 16x16
      *******************************************/
      if((mb_ext_level==2 && mb64==0) || (mb_ext_level==1))
      {
        if((mb16count&3)==0)//mb16count: counter of 16x16 (0~15 if mb_ext_level ==2; 0~3 if mb_ext_level==1)
        {
          read_mbtype(img, inp, 1);

          MB32.mb_type = currMB->mb_type; //untill this line, MB32.mb_type=1 could mean mode 1 of P or direct mode of B
          MB32.cbp     = currMB->cbp;
          memcpy(MB32.b8mode, currMB->b8mode, sizeof(int)*4);
          memcpy(MB32.b8pdir, currMB->b8pdir, sizeof(int)*4);
#ifdef GEOM
          MB32.geo_mode = currMB->geo_mode;
#endif
          assert((MB32.mb_type >=0 && MB32.mb_type<=3)||MB32.mb_type==P8x8);        
        }
        if( MB32.mb_type == P8x8)
          mb32=0;
        else
          mb32=1;
      }

      if((mb_ext_level==2 && mb64==0 && mb32==0) || (mb_ext_level==1 && mb32==0))
      {
        read_mbtype(img, inp, 0);
      }
    }
    img->siblock[img->mb_x][img->mb_y]=0;    

    //init NoMbPartLessThan8x8Flag
    currMB->NoMbPartLessThan8x8Flag = (IS_DIRECT(currMB) && !(active_sps->direct_8x8_inference_flag))? 0: 1;

    //====== READ 8x8 SUB-PARTITION MODES (modes of 8x8 blocks) and Intra VBST block modes ======
    if (IS_P8x8 (currMB))
    {
      currSE.type    = SE_MBTYPE;
      dP = &(currSlice->partArr[partMap[SE_MBTYPE]]);
      for (i=0; i<4; i++)
      {
        if (active_pps->entropy_coding_mode_flag ==UVLC || dP->bitstream->ei_flag) currSE.mapping = linfo_ue;
        else                                                  currSE.reading = readB8_typeInfo_CABAC;

        TRACE_STRING("sub_mb_type");
        dP->readSyntaxElement (&currSE, img, inp, dP);
        SetB8Mode (img, currMB, currSE.value1, i);

        //set NoMbPartLessThan8x8Flag for P8x8 mode
        currMB->NoMbPartLessThan8x8Flag &= (currMB->b8mode[i]==0 && active_sps->direct_8x8_inference_flag) || 
          (currMB->b8mode[i]==4);
      }
      //--- init macroblock data ---
      init_macroblock       (img);
      readMotionInfoFromNAL (img, inp);
    }


    //============= Transform Size Flag for INTRA MBs =============
    //-------------------------------------------------------------
    //transform size flag for INTRA_4x4 and INTRA_8x8 modes
    if (currMB->mb_type == I4MB && img->Transform8x8Mode)
    {
      currSE.type   =  SE_HEADER;
      dP = &(currSlice->partArr[partMap[SE_HEADER]]);
      currSE.reading = readMB_transform_size_flag_CABAC;
      TRACE_STRING("transform size 8x8 flag");

      // read UVLC transform_size_8x8_flag
      //if (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag)
      //{
      //  currSE.len = 1;
      //  readSyntaxElement_FLC(&currSE, dP->bitstream);
      //} 
      //else 
      {
        dP->readSyntaxElement(&currSE,img,inp,dP);
      } 

      currMB->luma_transform_size_8x8_flag = currSE.value1;

      if (currMB->luma_transform_size_8x8_flag)
      {
        currMB->mb_type = I8MB;
        for (i=0;i<4;i++) 
        {
          currMB->b8mode[i]=I8MB;
          currMB->b8pdir[i]=-1;
        }
      }
    }
    else
    {
      currMB->luma_transform_size_8x8_flag = 0;
    }

    if(active_pps->constrained_intra_pred_flag && (img->type==P_SLICE|| img->type==B_SLICE))        // inter frame
    {
      if( !IS_INTRA(currMB) )
      {
        img->intra_block[img->current_mb_nr] = 0;
      }
    }

    //! TO for error concealment
    //! If we have an INTRA Macroblock and we lost the partition
    //! which contains the intra coefficients Copy MB would be better 
    //! than just a gray block.
    //! Seems to be a bit at the wrong place to do this right here, but for this case 
    //! up to now there is no other way.
    dP = &(currSlice->partArr[partMap[SE_CBP_INTRA]]);
    if(IS_INTRA (currMB) && dP->bitstream->ei_flag && img->number)
    {
      currMB->mb_type = 0;
      currMB->ei_flag = 1;
      for (i=0;i<4;i++)
      {
        currMB->b8mode[i]=currMB->b8pdir[i]=0;
      }
    }
    dP = &(currSlice->partArr[partMap[currSE.type]]);
    //! End TO


    //--- init macroblock data ---
    if (!IS_P8x8 (currMB))
    {
      if((mb_ext_level==2 && mb64==0 && mb32==0) || (mb_ext_level==1 && mb32==0))
      {
        init_macroblock       (img);
      }
      else if((mb_ext_level==2 && mb64==0 && mb32==1 && (mb16count&3)==0) || (mb_ext_level==1 && mb32==1 && (mb16count&3)==0))
      {
        init_macroblock32     (img, 1);
      }
      else if(mb_ext_level==2 && mb64==1 && mb32count == 0 && mb16count==0)
      {
        init_macroblock32     (img, 2);
      }
    }

    if (IS_DIRECT (currMB) && img->cod_counter >= 0)
    {
      currMB->cbp = 0;
      reset_coeffs();//reset_coeffs has been called in image.c 

      if (active_pps->entropy_coding_mode_flag ==CABAC)
        img->cod_counter=-1;

      return DECODE_MB;
    }

    if (((mb_ext_level==1 && ((mb32 && mb16count==0) || !mb32)) ||
      (mb_ext_level==2 && ((mb64 && mb32count==0 && mb16count==0) || (!mb64 && mb32 && (mb16count&3)==0) || (!mb64 &&!mb32))))
      && IS_COPY (currMB) )//keep last macroblock
    {
      int i, j, k;
      short pmv[2];
      int zeroMotionAbove;
      int zeroMotionLeft;
      PixelPos mb_a, mb_b;
      int      a_mv_y = 0;
      int      a_ref_idx = 0;
      int      b_mv_y = 0;
      int      b_ref_idx = 0;

      getLuma4x4Neighbour(img->current_mb_nr,0,0,-1, 0,&mb_a);
      getLuma4x4Neighbour(img->current_mb_nr,0,0, 0,-1,&mb_b);

      if (mb_a.available)
      {
        a_mv_y    = dec_picture->mv[LIST_0][mb_a.pos_y][mb_a.pos_x][1];
        a_ref_idx = dec_picture->ref_idx[LIST_0][mb_a.pos_y][mb_a.pos_x];
      }

      if (mb_b.available)
      {
        b_mv_y    = dec_picture->mv[LIST_0][mb_b.pos_y][mb_b.pos_x][1];
        b_ref_idx = dec_picture->ref_idx[LIST_0][mb_b.pos_y][mb_b.pos_x];
      }

      zeroMotionLeft  = !mb_a.available ? 1 : a_ref_idx==0 && dec_picture->mv[LIST_0][mb_a.pos_y][mb_a.pos_x][0]==0 && a_mv_y==0 ? 1 : 0;
      zeroMotionAbove = !mb_b.available ? 1 : b_ref_idx==0 && dec_picture->mv[LIST_0][mb_b.pos_y][mb_b.pos_x][0]==0 && b_mv_y==0 ? 1 : 0;

      if(currMB->skip_flag)
      {
        currMB->cbp = 0;
        reset_coeffs();
      }

      img_block_y   = img->block_y;

#ifdef MV_COMPETITION
      if (mv_comp.mv_competition > 0)
      {
        short prediction_mode_for_skip;
        short predictor;

        if(((mb64 && mb32count==0 && mb16count==0) || (!mb64 && mb32 && (mb16count&3)==0) || (!mb64 && !mb32)))
        {

          for (prediction_mode_for_skip = 0; prediction_mode_for_skip<mv_comp.nb_mode_for_skip; prediction_mode_for_skip++)
          {
            //      SetMotionVectorPredictor_Skip  (pmv, pmv+1, dec_picture->ref_idx, dec_picture->mv, 0, LIST_0, 0, 0, 16, 16,prediction_mode_for_skip);
            if(mb32 || mb64)
              SetMotionVectorPredictor_Skip32(pmv, pmv+1, dec_picture->ref_idx, dec_picture->mv, 0, LIST_0, 0, 0, 16<<(mb64*2+mb32), 16<<(mb64*2+mb32), prediction_mode_for_skip, (mb64*2+mb32));
            else
              SetMotionVectorPredictor_Skip  (pmv, pmv+1, dec_picture->ref_idx, dec_picture->mv, 0, LIST_0, 0, 0, 16, 16,prediction_mode_for_skip);

            mv_comp.mv_pred_skip[prediction_mode_for_skip][0] = pmv[0];
            mv_comp.mv_pred_skip[prediction_mode_for_skip][1] = pmv[1];
          }

          if ((mv_comp.nb_mode_for_skip> 1) && (read_index_for_skip_mode() == TRUE))
            predictor = readPredictorForSkip(img,inp);
          else 
            predictor = 0;

          if(mb32 || mb64)
          {
            for(i=0;i<(BLOCK_SIZE<<(mb64*2+mb32));i++)
              for(j=0;j<(BLOCK_SIZE<<(mb64*2+mb32));j++)
                for (k=0;k<2;k++)
                  dec_picture->mv[LIST_0][img_block_y+j][img->block_x+i][k] = mv_comp.mv_pred_skip[predictor][k];
          }
          else
          {
            for(i=0;i<BLOCK_SIZE;i++)
              for(j=0;j<BLOCK_SIZE;j++)
                for (k=0;k<2;k++)
                  dec_picture->mv[LIST_0][img_block_y+j][img->block_x+i][k] = mv_comp.mv_pred_skip[predictor][k];
          }

        }
      }
      else
#endif
      {
        if (zeroMotionAbove || zeroMotionLeft)
        {
          for(i=0;i<(BLOCK_SIZE<<(mb64*2+mb32));i++)
            for(j=0;j<(BLOCK_SIZE<<(mb64*2+mb32));j++)
              for (k=0;k<2;k++)
                dec_picture->mv[LIST_0][img->block_y+j][img->block_x+i][k] = 0;
        }
        else
        {
          if(mb32 || mb64)
          {
            SetMotionVectorPredictor32 (img, pmv, pmv+1, 0, LIST_0, dec_picture->ref_idx, dec_picture->mv, 0, 0, 16<<(mb64*2+mb32), 16<<(mb64*2+mb32), (mb64*2+mb32));
            for(i=0;i<(BLOCK_SIZE<<(mb64*2+mb32));i++)
              for(j=0;j<(BLOCK_SIZE<<(mb64*2+mb32));j++)
                for (k=0;k<2;k++)
                {
                  dec_picture->mv[LIST_0][img_block_y+j][img->block_x+i][k] = pmv[k];
                }
          }
          else
          {
            SetMotionVectorPredictor (img, pmv, pmv+1, 0, LIST_0, dec_picture->ref_idx, dec_picture->mv, 0, 0, 16, 16);

            for(i=0;i<BLOCK_SIZE;i++)
              for(j=0;j<BLOCK_SIZE;j++)
                for (k=0;k<2;k++)
                {
                  dec_picture->mv[LIST_0][img_block_y+j][img->block_x+i][k] = pmv[k];
                }
          }
        }


      }


      for(i=0;i<(BLOCK_SIZE<<(mb64*2+mb32));i++)
        for(j=0;j<(BLOCK_SIZE<<(mb64*2+mb32));j++)
        {
          dec_picture->ref_idx[LIST_0][img_block_y+j][img->block_x+i] = 0;
          dec_picture->ref_pic_id[LIST_0][img_block_y+j][img->block_x+i] = 
            dec_picture->ref_pic_num[img->current_slice_nr][LIST_0][(short)dec_picture->ref_idx[LIST_0][img_block_y+j][img->block_x+i]];
        }           

        if(currMB->skip_flag)
          return DECODE_MB;
    }
    if(currMB->mb_type!=IPCM)
    {

      // intra prediction modes for a macroblock 4x4 **********************************************
      read_ipred_modes(img,inp);

      // read inter frame vector data *********************************************************
      if(mb64 && mb32count==0 && mb16count==0)
      {
        if (MB64.mb_type != 0)
        {
#ifdef GEOM
          if(currMB->geo_mode != -1)
            readMotionInfoFromNAL_geo  (img, inp, 2);
          else
#endif
            readMotionInfoFromNAL32 (img, inp, 2);
        }
      }
      else if((mb_ext_level==2 && !mb64 && mb32 && (mb16count&3)==0) || (mb_ext_level==1 && mb32 && (mb16count&3)==0))
      {
        if (MB32.mb_type != 0)
        {
#ifdef GEOM
          if(currMB->geo_mode != -1)
            readMotionInfoFromNAL_geo  (img, inp, 1);
          else
#endif
            readMotionInfoFromNAL32 (img, inp, 1);
        }
      }
      else if(((mb_ext_level==2 && mb64==0 && mb32==0)||(mb_ext_level==1 && mb32==0)) && IS_INTERMV (currMB) && (!IS_P8x8(currMB)))
      {
#ifdef GEOM
        if(currMB->geo_mode != -1)
          readMotionInfoFromNAL_geo  (img, inp, 0);
        else
#endif
          readMotionInfoFromNAL (img, inp);
      }

      // read CBP and Coeffs  ***************************************************************
      if(mb64 && mb32count==0 && mb16count==0)
      {
        //MB64.cbp=readCBP32(img,inp,1);//with dquant
        MB64.cbp = readCBP_Big(img,inp);
        readDeltaQP(img,inp,MB64.cbp);
      }

      if(mb_ext_level==2 && mb64 && MB64.cbp !=0 && (mb16count&3)==0)
      {      
        MB32.cbp = readCBP_Big(img,inp);
      }

      if((mb_ext_level==2 && !mb64 && mb32 && (mb16count&3)==0) || (mb_ext_level==1 && mb32 && (mb16count&3)==0) )
      {      
        MB32.cbp = readCBP_Big(img,inp);
        readDeltaQP(img,inp,MB32.cbp);
      }

      if(mb_ext_level == 2)
      {
        if( (mb64 && MB64.cbp!=0 && MB32.cbp !=0)
          ||(!mb64 && mb32 && MB32.cbp !=0 )
          ||(!mb64 && !mb32) )
        {
          readCBPandCoeffsFromNAL (img,inp, (mb64|mb32), mb64, 0);
        }
      }
      else
      {
        if(!mb32 || (mb32 && MB32.cbp !=0 ) )
          readCBPandCoeffsFromNAL (img,inp, mb32,0,mb16count);

      }
    }
    else
    {
      //read pcm_alignment_zero_bit and pcm_byte[i] 

      // here dP is assigned with the same dP as SE_MBTYPE, because IPCM syntax is in the 
      // same category as MBTYPE
      dP = &(currSlice->partArr[partMap[SE_MBTYPE]]);
      readIPCMcoeffsFromNAL(img,inp,dP);
    }

    return DECODE_MB;
  }


  /*!
  ************************************************************************
  * \brief
  *    Get the syntax elements from the NAL
  ************************************************************************
  */
  int read_one_macroblock(struct img_par *img,struct inp_par *inp)
  {
    int i;

    SyntaxElement currSE;
    Macroblock *currMB = &img->mb_data[img->current_mb_nr];

    Slice *currSlice = img->currentSlice;
    DataPartition *dP;
    int *partMap = assignSE2partition[currSlice->dp_mode];
    Macroblock *topMB = NULL;
    int  prevMbSkipped = 0;
    int  img_block_y;

    currMB->qp = img->qp ;
    for (i=0; i<2; i++)
    {
      currMB->qpc[i] = Clip3 ( -img->bitdepth_chroma_qp_scale, 51, img->qp + dec_picture->chroma_qp_offset[i] );
      currMB->qpc[i] = currMB->qpc[i] < 0 ? currMB->qpc[i] : QP_SCALE_CR[currMB->qpc[i]];
    }

    currSE.type = SE_MBTYPE;

    //  read MB mode *****************************************************************
    dP = &(currSlice->partArr[partMap[currSE.type]]);

    //if (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag)   currSE.mapping = linfo_ue;

    if(img->type == I_SLICE || img->type == SI_SLICE)
    {
      if(active_pps->entropy_coding_mode_flag  == CABAC)
        CheckAvailabilityOfNeighborsCABAC();

      //  read MB type
      TRACE_STRING("mb_type");
      currSE.reading = readMB_typeInfo_CABAC;
      dP->readSyntaxElement(&currSE,img,inp,dP);

      currMB->mb_type = currSE.value1;
      if(!dP->bitstream->ei_flag)
        currMB->ei_flag = 0;
    }
    // non I/SI-slice CABAC
    else if (active_pps->entropy_coding_mode_flag == CABAC)
    {
      CheckAvailabilityOfNeighborsCABAC();
      TRACE_STRING("mb_skip_flag");
      currSE.reading = readMB_skip_flagInfo_CABAC;
      dP->readSyntaxElement(&currSE,img,inp,dP);

      currMB->mb_type   = currSE.value1;      //---is non-skip then mbtype=1..so for Direct also mbtype=1 here...
      currMB->skip_flag = !(currSE.value1);
      currMB->cbp = currSE.value2;

      if(!dP->bitstream->ei_flag)
        currMB->ei_flag = 0;

      if ((img->type==B_SLICE) && currSE.value1==0 && currSE.value2==0)
        img->cod_counter=0;

      //CheckAvailabilityOfNeighborsCABAC();

      // read MB type
      if (currMB->mb_type != 0 )
      {
        currSE.reading = readMB_typeInfo_CABAC;

        TRACE_STRING("mb_type");
        dP->readSyntaxElement(&currSE,img,inp,dP);
        currMB->mb_type = currSE.value1;
        currMB->b8pdir[0] = currSE.dir_partition1;
        currMB->b8pdir[3] = currSE.dir_partition2;

        if(!dP->bitstream->ei_flag)
          currMB->ei_flag = 0;
      }
    }

    img->siblock[img->mb_x][img->mb_y]=0;

    if ((img->type==P_SLICE ))    // inter frame
      interpret_mb_mode_P(img);
    else if (img->type==I_SLICE)                                  // intra frame
      interpret_mb_mode_I(img);
    else if ((img->type==B_SLICE))       // B frame
      interpret_mb_mode_B    (img);
    else if ((img->type==SP_SLICE))     // SP frame
      interpret_mb_mode_P(img);
    else if (img->type==SI_SLICE)     // SI frame
      interpret_mb_mode_SI(img);


#ifdef GEOM
    if(img->type != I_SLICE && currMB->mb_type == 3)
    {
      currSE.value2 = rho_bits[0]+THETA_BITS;
      currSE.reading = readMB_typeInfo_CABAC_geo;
      TRACE_STRING("mb_type");
      dP->readSyntaxElement(&currSE,img,inp,dP);      

      if(!dP->bitstream->ei_flag)
        currMB->ei_flag = 0;
    }
#endif


    //init NoMbPartLessThan8x8Flag
    currMB->NoMbPartLessThan8x8Flag = (IS_DIRECT(currMB) && !(active_sps->direct_8x8_inference_flag))? 0: 1;

    //====== READ 8x8 SUB-PARTITION MODES (modes of 8x8 blocks) and Intra VBST block modes ======
    if (IS_P8x8 (currMB))
    {
      currSE.type    = SE_MBTYPE;
      dP = &(currSlice->partArr[partMap[SE_MBTYPE]]);

      for (i=0; i<4; i++)
      {
        if (active_pps->entropy_coding_mode_flag ==UVLC || dP->bitstream->ei_flag)
          currSE.mapping = linfo_ue;
        else
          currSE.reading = readB8_typeInfo_CABAC;

        TRACE_STRING("sub_mb_type");
        dP->readSyntaxElement (&currSE, img, inp, dP);

        SetB8Mode (img, currMB, currSE.value1, i);

        //set NoMbPartLessThan8x8Flag for P8x8 mode
        currMB->NoMbPartLessThan8x8Flag &= (currMB->b8mode[i]==0 && active_sps->direct_8x8_inference_flag) || 
          (currMB->b8mode[i]==4);
      }
      //--- init macroblock data ---
      init_macroblock       (img);
      readMotionInfoFromNAL (img, inp);
    }


    //============= Transform Size Flag for INTRA MBs =============
    //-------------------------------------------------------------
    //transform size flag for INTRA_4x4 and INTRA_8x8 modes
    if (currMB->mb_type == I4MB && img->Transform8x8Mode)
    {
      currSE.type   =  SE_HEADER;
      dP = &(currSlice->partArr[partMap[SE_HEADER]]);
      currSE.reading = readMB_transform_size_flag_CABAC;
      TRACE_STRING("transform size 8x8 flag");

      // read UVLC transform_size_8x8_flag
      //if (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag)
      //{
      //  currSE.len = 1;
      //  readSyntaxElement_FLC(&currSE, dP->bitstream);
      //} 
      //else 
      {
        dP->readSyntaxElement(&currSE,img,inp,dP);
      } 

      currMB->luma_transform_size_8x8_flag = currSE.value1;

      if (currMB->luma_transform_size_8x8_flag)
      {
        currMB->mb_type = I8MB;
        for (i=0;i<4;i++) {currMB->b8mode[i]=I8MB; currMB->b8pdir[i]=-1; }
      }
    }
    else
    {
      currMB->luma_transform_size_8x8_flag = 0;
    }

    if(active_pps->constrained_intra_pred_flag && (img->type==P_SLICE|| img->type==B_SLICE))        // inter frame
    {
      if( !IS_INTRA(currMB) )
      {
        img->intra_block[img->current_mb_nr] = 0;
      }
    }

    //! TO for error concealment
    //! If we have an INTRA Macroblock and we lost the partition
    //! which contains the intra coefficients Copy MB would be better 
    //! than just a gray block.
    //! Seems to be a bit at the wrong place to do this right here, but for this case 
    //! up to now there is no other way.
    dP = &(currSlice->partArr[partMap[SE_CBP_INTRA]]);
    if(IS_INTRA (currMB) && dP->bitstream->ei_flag && img->number)
    {
      currMB->mb_type = 0;
      currMB->ei_flag = 1;
      for (i=0;i<4;i++) {currMB->b8mode[i]=currMB->b8pdir[i]=0; }
    }
    dP = &(currSlice->partArr[partMap[currSE.type]]);
    //! End TO


    //--- init macroblock data ---
    if (!IS_P8x8 (currMB))
      init_macroblock       (img);

    if (active_pps->entropy_coding_mode_flag ==CABAC)
    {
      if (IS_DIRECT (currMB) && img->cod_counter >= 0)
      {
        currMB->cbp = 0;
        reset_coeffs();

        img->cod_counter=-1;

        return DECODE_MB;
      }
    }
    else
    {
      if (IS_DIRECT (currMB) && currMB->skip_flag == 1)
      {
        currMB->cbp = 0;
        reset_coeffs();
        return DECODE_MB;
      }
    }

    //---PSKIP and PDIRECT

    if (IS_COPY (currMB)) //keep last macroblock
    {
      int i, j, k;
      short pmv[2];
      int zeroMotionAbove;
      int zeroMotionLeft;
      PixelPos mb_a, mb_b;
      int      a_mv_y = 0;
      int      a_ref_idx = 0;
      int      b_mv_y = 0;
      int      b_ref_idx = 0;

      getLuma4x4Neighbour(img->current_mb_nr,0,0,-1, 0,&mb_a);
      getLuma4x4Neighbour(img->current_mb_nr,0,0, 0,-1,&mb_b);

      if (mb_a.available)
      {
        a_mv_y    = dec_picture->mv[LIST_0][mb_a.pos_y][mb_a.pos_x][1];
        a_ref_idx = dec_picture->ref_idx[LIST_0][mb_a.pos_y][mb_a.pos_x];
      }

      if (mb_b.available)
      {
        b_mv_y    = dec_picture->mv[LIST_0][mb_b.pos_y][mb_b.pos_x][1];
        b_ref_idx = dec_picture->ref_idx[LIST_0][mb_b.pos_y][mb_b.pos_x];
      }

      zeroMotionLeft  = !mb_a.available ? 1 : a_ref_idx==0 && dec_picture->mv[LIST_0][mb_a.pos_y][mb_a.pos_x][0]==0 && a_mv_y==0 ? 1 : 0;
      zeroMotionAbove = !mb_b.available ? 1 : b_ref_idx==0 && dec_picture->mv[LIST_0][mb_b.pos_y][mb_b.pos_x][0]==0 && b_mv_y==0 ? 1 : 0;

      if(currMB->skip_flag)
      {
        currMB->cbp = 0;
        reset_coeffs();
      }
      img_block_y   = img->block_y;

#ifdef MV_COMPETITION

      if (mv_comp.mv_competition > 0)

      {
        short prediction_mode_for_skip;
        short predictor;

        for (prediction_mode_for_skip = 0; prediction_mode_for_skip<mv_comp.nb_mode_for_skip; prediction_mode_for_skip++)
        {
          SetMotionVectorPredictor_Skip(pmv, pmv+1, dec_picture->ref_idx, dec_picture->mv, 0, LIST_0, 0, 0, 16, 16,prediction_mode_for_skip);
          mv_comp.mv_pred_skip[prediction_mode_for_skip][0] = pmv[0];
          mv_comp.mv_pred_skip[prediction_mode_for_skip][1] = pmv[1];
        }

        if ((mv_comp.nb_mode_for_skip> 1) && (read_index_for_skip_mode() == TRUE))
          predictor = readPredictorForSkip(img,inp);
        else 
          predictor = 0;

        for(i=0;i<BLOCK_SIZE;i++)
          for(j=0;j<BLOCK_SIZE;j++)
            for (k=0;k<2;k++)
              dec_picture->mv[LIST_0][img_block_y+j][img->block_x+i][k] = mv_comp.mv_pred_skip[predictor][k];
      }
      else
#endif

      {


        if (zeroMotionAbove || zeroMotionLeft)
        {
          for(i=0;i<BLOCK_SIZE;i++)
            for(j=0;j<BLOCK_SIZE;j++)
              for (k=0;k<2;k++)
                dec_picture->mv[LIST_0][img->block_y+j][img->block_x+i][k] = 0;
        }
        else
        {
          SetMotionVectorPredictor (img, pmv, pmv+1, 0, LIST_0, dec_picture->ref_idx, dec_picture->mv, 0, 0, 16, 16);

          for(i=0;i<BLOCK_SIZE;i++)
            for(j=0;j<BLOCK_SIZE;j++)
              for (k=0;k<2;k++)
              {
                dec_picture->mv[LIST_0][img_block_y+j][img->block_x+i][k] = pmv[k];
              }
        }


      }

      for(i=0;i<BLOCK_SIZE;i++)
        for(j=0;j<BLOCK_SIZE;j++)
        {
          dec_picture->ref_idx[LIST_0][img_block_y+j][img->block_x+i] = 0;
          dec_picture->ref_pic_id[LIST_0][img_block_y+j][img->block_x+i] = 
            dec_picture->ref_pic_num[img->current_slice_nr][LIST_0][(short)dec_picture->ref_idx[LIST_0][img_block_y+j][img->block_x+i]];
        }

        if(currMB->skip_flag)
          return DECODE_MB;
    }

    if(currMB->mb_type!=IPCM)
    {

      // intra prediction modes for a macroblock 4x4 **********************************************
      read_ipred_modes(img,inp);

      // read inter frame vector data *********************************************************
      if (IS_INTERMV (currMB) && (!IS_P8x8(currMB)))
      {
#ifdef GEOM
        if(currMB->geo_mode != -1)
          readMotionInfoFromNAL_geo  (img, inp, 0);
        else
#endif
          readMotionInfoFromNAL (img, inp);
      }
      // read CBP and Coeffs  ***************************************************************
      readCBPandCoeffsFromNAL (img,inp,0,0,0);
    }
    else
    {
      //read pcm_alignment_zero_bit and pcm_byte[i] 

      // here dP is assigned with the same dP as SE_MBTYPE, because IPCM syntax is in the 
      // same category as MBTYPE
      dP = &(currSlice->partArr[partMap[SE_MBTYPE]]);
      readIPCMcoeffsFromNAL(img,inp,dP);
    }

    return DECODE_MB;
  }



  void init_decoding_engine_IPCM(struct img_par *img)
  {
    Slice *currSlice = img->currentSlice;
    Bitstream *currStream;
    int ByteStartPosition;
    int PartitionNumber;
    int i;

    if(currSlice->dp_mode==PAR_DP_1)
      PartitionNumber=1;
    else if(currSlice->dp_mode==PAR_DP_3)
      PartitionNumber=3;
    else
    {
      printf("Partition Mode is not supported\n");
      exit(1);
    }

    for(i=0;i<PartitionNumber;i++)
    {
      currStream = currSlice->partArr[i].bitstream;
      ByteStartPosition = currStream->read_len;


      arideco_start_decoding (&currSlice->partArr[i].de_cabac, currStream->streamBuffer, ByteStartPosition, &currStream->read_len, img->type);
    }
  }



  void readIPCMcoeffsFromNAL(struct img_par *img, struct inp_par *inp, struct datapartition *dP)
  {
    SyntaxElement currSE;
    int i,j;

    //For CABAC, we don't need to read bits to let stream byte aligned
    //  because we have variable for integer bytes position
    if(active_pps->entropy_coding_mode_flag  == CABAC)
    {
      //read luma and chroma IPCM coefficients
      currSE.len=8;
      TRACE_STRING("pcm_byte luma");

      for(i=0;i<MB_BLOCK_SIZE;i++)
      {
        for(j=0;j<MB_BLOCK_SIZE;j++)
        {
          readIPCMBytes_CABAC(&currSE, dP->bitstream);
          img->cof[i/4][j/4][i%4][j%4]=currSE.value1;
        }
      } 
      if (dec_picture->chroma_format_idc != YUV400)
      {
        TRACE_STRING("pcm_byte chroma");
        for(i=0;i<img->mb_cr_size_y;i++)
        {
          for(j=0;j<img->mb_cr_size_x;j++)
          {
            readIPCMBytes_CABAC(&currSE, dP->bitstream);
            img->cof[i/4][j/4+4][i%4][j%4]=currSE.value1;
          }
        } 
        for(i=0;i<img->mb_cr_size_y;i++)
        {
          for(j=0;j<img->mb_cr_size_x;j++)
          {
            readIPCMBytes_CABAC(&currSE, dP->bitstream);
            img->cof[i/4+2][j/4+4][i%4][j%4]=currSE.value1;
          }
        } 
      }
      //If the decoded MB is IPCM MB, decoding engine is initialized

      // here the decoding engine is directly initialized without checking End of Slice
      // The reason is that, whether current MB is the last MB in slice or not, there is
      // at least one 'end of slice' syntax after this MB. So when fetching bytes in this 
      // initialisation process, we can guarantee there is bits available in bitstream. 

      init_decoding_engine_IPCM(img);
    }
    else
    { 
      //read bits to let stream byte aligned

      if((dP->bitstream->frame_bitoffset)%8!=0)
      {
        TRACE_STRING("pcm_alignment_zero_bit");
        currSE.len=8-(dP->bitstream->frame_bitoffset)%8;
        readSyntaxElement_FLC(&currSE, dP->bitstream);
      }

      //read luma and chroma IPCM coefficients
      currSE.len=img->bitdepth_luma;
      TRACE_STRING("pcm_byte luma");

      for(i=0;i<MB_BLOCK_SIZE;i++)
      {
        for(j=0;j<MB_BLOCK_SIZE;j++)
        {
          readSyntaxElement_FLC(&currSE, dP->bitstream);
          img->cof[i/4][j/4][i%4][j%4]=currSE.value1;
        }
      } 
      currSE.len=img->bitdepth_chroma;
      if (dec_picture->chroma_format_idc != YUV400)
      {
        TRACE_STRING("pcm_byte chroma");
        for(i=0;i<img->mb_cr_size_y;i++)
        {
          for(j=0;j<img->mb_cr_size_x;j++)
          {
            readSyntaxElement_FLC(&currSE, dP->bitstream);
            img->cof[i/4][j/4+4][i%4][j%4]=currSE.value1;
          }
        } 
        for(i=0;i<img->mb_cr_size_y;i++)
        {
          for(j=0;j<img->mb_cr_size_x;j++)
          {
            readSyntaxElement_FLC(&currSE, dP->bitstream);
            img->cof[i/4+2][j/4+4][i%4][j%4]=currSE.value1;
          }
        }
      }
    }
  }



  void read_ipred_modes(struct img_par *img,struct inp_par *inp)
  {
    int b8,i,j,bi,bj,bx,by,dec;
    SyntaxElement currSE;
    Slice *currSlice;
    DataPartition *dP;
    int *partMap;
    Macroblock *currMB;
    int ts, ls;
    int mostProbableIntraPredMode;
    int upIntraPredMode;
    int leftIntraPredMode;
    int IntraChromaPredModeFlag;
    int bs_x, bs_y;
    int ii,jj;

    PixelPos left_block;
    PixelPos top_block;

    currMB = &img->mb_data[img->current_mb_nr];

    IntraChromaPredModeFlag = IS_INTRA(currMB);

    currSlice = img->currentSlice;
    partMap = assignSE2partition[currSlice->dp_mode];

    currSE.type = SE_INTRAPREDMODE;

    TRACE_STRING("intra4x4_pred_mode");
    dP = &(currSlice->partArr[partMap[currSE.type]]);

    //if (!(active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag)) 
    currSE.reading = readIntraPredMode_CABAC;

    for(b8=0;b8<4;b8++)  //loop 8x8 blocks
    {
      if((currMB->b8mode[b8]==IBLOCK )||(currMB->b8mode[b8]==I8MB))
      {
        bs_x = bs_y = (currMB->b8mode[b8] == I8MB)?8:4;

        IntraChromaPredModeFlag = 1;

        ii=(bs_x>>2);
        jj=(bs_y>>2);

        for(j=0;j<2;j+=jj)  //loop subblocks
          for(i=0;i<2;i+=ii)
          {
            //get from stream
            //if (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag)
            //  readSyntaxElement_Intra4x4PredictionMode(&currSE,img,inp,dP);
            //else 
            {
              currSE.context=(b8<<2)+(j<<1)+i;
              dP->readSyntaxElement(&currSE,img,inp,dP);
            }

            bx = ((b8&1)<<1) + i;
            by = (b8&2)      + j;

            getLuma4x4Neighbour(img->current_mb_nr, bx, by, -1,  0, &left_block);
            getLuma4x4Neighbour(img->current_mb_nr, bx, by,  0, -1, &top_block);

            //get from array and decode
            bi = img->block_x + bx;
            bj = img->block_y + by;

            if (active_pps->constrained_intra_pred_flag)
            {
              left_block.available = left_block.available ? img->intra_block[left_block.mb_addr] : 0;
              top_block.available  = top_block.available  ? img->intra_block[top_block.mb_addr]  : 0;
            }

            // !!not sure if the follwing is still correct...
            ts=ls=0;   // Check to see if the neighboring block is SI
            if (IS_OLDINTRA(currMB) && img->type == SI_SLICE)           // need support for MBINTLC1
            {
              if (left_block.available)
                if (img->siblock [left_block.pos_x][left_block.pos_y])
                  ls=1;

              if (top_block.available)
                if (img->siblock [top_block.pos_x][top_block.pos_y])
                  ts=1;
            }

            upIntraPredMode            = (top_block.available  &&(ts == 0)) ? img->ipredmode[top_block.pos_x ][top_block.pos_y ] : -1;
            leftIntraPredMode          = (left_block.available &&(ls == 0)) ? img->ipredmode[left_block.pos_x][left_block.pos_y] : -1;

            mostProbableIntraPredMode  = (upIntraPredMode < 0 || leftIntraPredMode < 0) ? DC_PRED : upIntraPredMode < leftIntraPredMode ? upIntraPredMode : leftIntraPredMode;

            dec = (currSE.value1 == -1) ? mostProbableIntraPredMode : currSE.value1 + (currSE.value1 >= mostProbableIntraPredMode);

            //set
            for(jj=0;jj<(bs_y>>2);jj++)   //loop 4x4s in the subblock for 8x8 prediction setting
              for(ii=0;ii<(bs_x>>2);ii++)
                img->ipredmode[bi+ii][bj+jj]=dec;
          }
      }
    }

    if (IntraChromaPredModeFlag && dec_picture->chroma_format_idc != YUV400)
    {
      currSE.type = SE_INTRAPREDMODE;
      TRACE_STRING("intra_chroma_pred_mode");
      dP = &(currSlice->partArr[partMap[currSE.type]]);

      /*    if (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag) currSE.mapping = linfo_ue;
      else */                                                   currSE.reading = readCIPredMode_CABAC;

      dP->readSyntaxElement(&currSE,img,inp,dP);
      currMB->c_ipred_mode = currSE.value1;

      if (currMB->c_ipred_mode < DC_PRED_8 || currMB->c_ipred_mode > PLANE_8)
      {
        error("illegal chroma intra pred mode!\n", 600);
      }
    }
  }

  /*!
  ************************************************************************
  * \brief
  *    Set motion vector predictor
  ************************************************************************
  */
  void SetMotionVectorPredictor (struct img_par  *img,
    short           *pmv_x,
    short           *pmv_y,
    char            ref_frame,
    byte            list,
    char            ***refPic,
    short           ****tmp_mv,
    int             block_x,
    int             block_y,
    int             blockshape_x,
    int             blockshape_y)
  {
    int mb_x                 = BLOCK_SIZE*block_x;
    int mb_y                 = BLOCK_SIZE*block_y;
    int mb_nr                = img->current_mb_nr;

    int mv_a, mv_b, mv_c, pred_vec=0;
    int mvPredType, rFrameL, rFrameU, rFrameUR;
    int hv;


    PixelPos block_a, block_b, block_c, block_d;

    getLuma4x4Neighbour(mb_nr, block_x, block_y,           -1,  0, &block_a);
    getLuma4x4Neighbour(mb_nr, block_x, block_y,            0, -1, &block_b);
    getLuma4x4Neighbour(mb_nr, block_x, block_y, blockshape_x, -1, &block_c);
    getLuma4x4Neighbour(mb_nr, block_x, block_y,           -1, -1, &block_d);

    if (mb_y > 0)
    {
      if (mb_x < 8)  // first column of 8x8 blocks
      {
        if (mb_y==8)
        {
          if (blockshape_x == 16)      block_c.available  = 0;
        }
        else
        {
          if (mb_x+blockshape_x == 8)  block_c.available  = 0;
        }
      }
      else
      {
        if (mb_x+blockshape_x == 16)   block_c.available  = 0;
      }
    }

    if (!block_c.available)
    {
      block_c=block_d;
    }

    mvPredType = MVPRED_MEDIAN;

    rFrameL    = block_a.available    ? refPic[list][block_a.pos_y][block_a.pos_x] : -1;
    rFrameU    = block_b.available    ? refPic[list][block_b.pos_y][block_b.pos_x] : -1;
    rFrameUR   = block_c.available    ? refPic[list][block_c.pos_y][block_c.pos_x] : -1;

    /* Prediction if only one of the neighbors uses the reference frame
    * we are checking
    */
    if(rFrameL == ref_frame && rFrameU != ref_frame && rFrameUR != ref_frame)       mvPredType = MVPRED_L;
    else if(rFrameL != ref_frame && rFrameU == ref_frame && rFrameUR != ref_frame)  mvPredType = MVPRED_U;
    else if(rFrameL != ref_frame && rFrameU != ref_frame && rFrameUR == ref_frame)  mvPredType = MVPRED_UR;
    // Directional predictions 
    if(blockshape_x == 8 && blockshape_y == 16)
    {
      if(mb_x == 0)
      {
        if(rFrameL == ref_frame)
          mvPredType = MVPRED_L;
      }
      else
      {
        if( rFrameUR == ref_frame)
          mvPredType = MVPRED_UR;
      }
    }
    else if(blockshape_x == 16 && blockshape_y == 8)
    {
      if(mb_y == 0)
      {
        if(rFrameU == ref_frame)
          mvPredType = MVPRED_U;
      }
      else
      {
        if(rFrameL == ref_frame)
          mvPredType = MVPRED_L;
      }
    }

    for (hv=0; hv < 2; hv++)
    {
      mv_a = block_a.available  ? tmp_mv[list][block_a.pos_y][block_a.pos_x][hv] : 0;
      mv_b = block_b.available  ? tmp_mv[list][block_b.pos_y][block_b.pos_x][hv] : 0;
      mv_c = block_c.available  ? tmp_mv[list][block_c.pos_y][block_c.pos_x][hv] : 0;

      {
        int64 xpoc  = img->framepoc >> 1;
        int64 x  = dec_picture->ref_pic_num[0][list][ref_frame] >> 2;
        int64 L  = dec_picture->ref_pic_num[0][list][rFrameL  ] >> 2;
        int64 U  = dec_picture->ref_pic_num[0][list][rFrameU  ] >> 2;
        int64 UR = dec_picture->ref_pic_num[0][list][rFrameUR ] >> 2;

        if(xpoc != L)
          mv_a = (int)(mv_a * (xpoc - x) / (xpoc - L));
        if(xpoc != U)
          mv_b = (int)(mv_b * (xpoc - x) / (xpoc - U));
        if(xpoc != UR)
          mv_c = (int)(mv_c * (xpoc - x) / (xpoc - UR));
      }

      switch (mvPredType)
      {
      case MVPRED_MEDIAN:
        if(!(block_b.available || block_c.available))
          pred_vec = mv_a;
        else
          pred_vec = mv_a+mv_b+mv_c-min(mv_a,min(mv_b,mv_c))-max(mv_a,max(mv_b,mv_c));
        break;
      case MVPRED_L:
        pred_vec = mv_a;
        break;
      case MVPRED_U:
        pred_vec = mv_b;
        break;
      case MVPRED_UR:
        pred_vec = mv_c;
        break;
      default:
        break;
      }

      if (hv==0)  *pmv_x = pred_vec;
      else        *pmv_y = pred_vec;

    }
  }


  /*!
  ************************************************************************
  * \brief
  *    Set context for reference frames
  ************************************************************************
  */
  int
    BType2CtxRef (int btype)
  {
    if (btype<4)  return 0;
    else          return 1;
  }

  /*!
  ************************************************************************
  * \brief
  *    Read motion info
  ************************************************************************
  */
  void readMotionInfoFromNAL (struct img_par *img, struct inp_par *inp)
  {
    int i,j,k;
    int step_h,step_v;
    int curr_mvd;
    Macroblock *currMB  = &img->mb_data[img->current_mb_nr];
    SyntaxElement currSE;
    Slice *currSlice    = img->currentSlice;
    DataPartition *dP;
    int *partMap        = assignSE2partition[currSlice->dp_mode];
    int bframe          = (img->type==B_SLICE);
    int partmode        = (IS_P8x8(currMB)?4:currMB->mb_type);
    int step_h0         = BLOCK_STEP [partmode][0];
    int step_v0         = BLOCK_STEP [partmode][1];

    int mv_mode, i0, j0;
    char refframe;
    short pmv[2];
    int j4, i4, ii,jj;

    int motion_vector;
    int mv_scale = 0;

    int flag_mode;


    byte  **    moving_block;
    short ****  co_located_mv;
    char  ***   co_located_ref_idx;
    int64 ***   co_located_ref_id;

#ifdef ADAPTIVE_MV_RES
    int mv_shift=0;
#endif

    moving_block = Co_located->moving_block;
    co_located_mv = Co_located->mv;
    co_located_ref_idx = Co_located->ref_idx;
    co_located_ref_id = Co_located->ref_pic_id;

    if (bframe && IS_P8x8 (currMB))
    {
      if (img->direct_spatial_mv_pred_flag)
      {
        int imgblock_y= img->block_y;
        int fw_rFrameL, fw_rFrameU, fw_rFrameUL, fw_rFrameUR;
        int bw_rFrameL, bw_rFrameU, bw_rFrameUL, bw_rFrameUR;

        PixelPos mb_left, mb_up, mb_upleft, mb_upright;

        char  fw_rFrame,bw_rFrame;
        short pmvfw[2]={0,0},
          pmvbw[2]={0,0};


        getLuma4x4Neighbour(img->current_mb_nr, 0, 0, -1,  0, &mb_left);
        getLuma4x4Neighbour(img->current_mb_nr, 0, 0,  0, -1, &mb_up);
        getLuma4x4Neighbour(img->current_mb_nr, 0, 0, 16, -1, &mb_upright);
        getLuma4x4Neighbour(img->current_mb_nr, 0, 0, -1, -1, &mb_upleft);


        fw_rFrameL = mb_left.available ? dec_picture->ref_idx[LIST_0][mb_left.pos_y][mb_left.pos_x] : -1;
        fw_rFrameU = mb_up.available ? dec_picture->ref_idx[LIST_0][mb_up.pos_y][mb_up.pos_x] : -1;
        fw_rFrameUL = mb_upleft.available ? dec_picture->ref_idx[LIST_0][mb_upleft.pos_y][mb_upleft.pos_x] : -1;
        fw_rFrameUR = mb_upright.available ? dec_picture->ref_idx[LIST_0][mb_upright.pos_y][mb_upright.pos_x] : fw_rFrameUL;      

        bw_rFrameL = mb_left.available ? dec_picture->ref_idx[LIST_1][mb_left.pos_y][mb_left.pos_x] : -1;
        bw_rFrameU = mb_up.available ? dec_picture->ref_idx[LIST_1][mb_up.pos_y][mb_up.pos_x] : -1;
        bw_rFrameUL = mb_upleft.available ? dec_picture->ref_idx[LIST_1][mb_upleft.pos_y][mb_upleft.pos_x] : -1;
        bw_rFrameUR = mb_upright.available ? dec_picture->ref_idx[LIST_1][mb_upright.pos_y][mb_upright.pos_x] : bw_rFrameUL;      


        fw_rFrame = (fw_rFrameL >= 0 && fw_rFrameU >= 0) ? min(fw_rFrameL,fw_rFrameU): max(fw_rFrameL,fw_rFrameU);
        fw_rFrame = (fw_rFrame >= 0 && fw_rFrameUR >= 0) ? min(fw_rFrame,fw_rFrameUR): max(fw_rFrame,fw_rFrameUR);

        bw_rFrame = (bw_rFrameL >= 0 && bw_rFrameU >= 0) ? min(bw_rFrameL,bw_rFrameU): max(bw_rFrameL,bw_rFrameU);
        bw_rFrame = (bw_rFrame >= 0 && bw_rFrameUR >= 0) ? min(bw_rFrame,bw_rFrameUR): max(bw_rFrame,bw_rFrameUR);


        if (fw_rFrame >=0)
          SetMotionVectorPredictor (img, pmvfw, pmvfw+1, fw_rFrame, LIST_0, dec_picture->ref_idx, dec_picture->mv, 0, 0, 16, 16);

        if (bw_rFrame >=0)
          SetMotionVectorPredictor (img, pmvbw, pmvbw+1, bw_rFrame, LIST_1, dec_picture->ref_idx, dec_picture->mv, 0, 0, 16, 16);


        for (i=0;i<4;i++)
        {
          if (currMB->b8mode[i] == 0)
            for(j=2*(i/2);j<2*(i/2)+2;j++)
              for(k=2*(i%2);k<2*(i%2)+2;k++)
              {
                int j6 = imgblock_y+j;
                j4 = img->block_y+j;
                i4 = img->block_x+k;


                if (fw_rFrame >= 0)
                {

                  if  (!fw_rFrame  && ((!moving_block[j6][i4]) && (!listX[1][0]->is_long_term)))
                  {                    
                    dec_picture->mv  [LIST_0][j4][i4][0] = 0;
                    dec_picture->mv  [LIST_0][j4][i4][1] = 0;
                    dec_picture->ref_idx[LIST_0][j4][i4] = 0;                    
                  }
                  else
                  {

                    dec_picture->mv  [LIST_0][j4][i4][0] = pmvfw[0];
                    dec_picture->mv  [LIST_0][j4][i4][1] = pmvfw[1];
                    dec_picture->ref_idx[LIST_0][j4][i4] = fw_rFrame;
                  }
                }
                else
                {
                  dec_picture->mv  [LIST_0][j4][i4][0] = 0;
                  dec_picture->mv  [LIST_0][j4][i4][1] = 0;
                  dec_picture->ref_idx[LIST_0][j4][i4] = -1;
                }
                if (bw_rFrame >= 0)
                {
                  if  (bw_rFrame==0 && ((!moving_block[j6][i4])&& (!listX[1][0]->is_long_term)))
                  {
                    dec_picture->mv  [LIST_1][j4][i4][0] = 0;
                    dec_picture->mv  [LIST_1][j4][i4][1] = 0;
                    dec_picture->ref_idx[LIST_1][j4][i4] = 0;
                  }
                  else
                  {
                    dec_picture->mv  [LIST_1][j4][i4][0] = pmvbw[0];
                    dec_picture->mv  [LIST_1][j4][i4][1] = pmvbw[1];
                    dec_picture->ref_idx[LIST_1][j4][i4] = bw_rFrame;
                  }
                }
                else
                {
                  dec_picture->mv  [LIST_1][j4][i4][0] = 0;
                  dec_picture->mv  [LIST_1][j4][i4][1] = 0;
                  dec_picture->ref_idx[LIST_1][j4][i4] = -1;                               
                }

                if (fw_rFrame <0 && bw_rFrame <0)
                {
                  dec_picture->ref_idx[LIST_0][j4][i4] = 0;
                  dec_picture->ref_idx[LIST_1][j4][i4] = 0;                  
                }
              }
        }
      }
      else
      {
        for (i=0;i<4;i++)
        {
          if (currMB->b8mode[i] == 0)
          {
            for(j=2*(i/2);j<2*(i/2)+2;j++)
            {
              for(k=2*(i%2);k<2*(i%2)+2;k++)
              {

                int imgblock_y= img->block_y;
                int refList = co_located_ref_idx[LIST_0 ][imgblock_y+j][img->block_x+k]== -1 ? LIST_1 : LIST_0;
                int ref_idx = co_located_ref_idx[refList][imgblock_y + j][img->block_x + k];
                int mapped_idx=-1, iref;                             

                if (ref_idx == -1)
                {
                  dec_picture->ref_idx [LIST_0][img->block_y + j][img->block_x + k] = 0;
                  dec_picture->ref_idx [LIST_1][img->block_y + j][img->block_x + k] = 0;                
                }
                else
                {
                  for (iref=0;iref<min(img->num_ref_idx_l0_active,listXsize[LIST_0]);iref++)
                  {
#if 1
                    //int curr_mb_field = 0;

                    if(img->structure==0 )
                    {
                      // If the current MB is a frame MB and the colocated is from a field picture, 
                      // then the co_located_ref_id may have been generated from the wrong value of 
                      // frame_poc if it references it's complementary field, so test both POC values
                      if(listX[0][iref]->top_poc*2 == co_located_ref_id[refList][imgblock_y + j][img->block_x + k] 
                      || listX[0][iref]->bottom_poc*2 == co_located_ref_id[refList][imgblock_y + j][img->block_x + k])
                      {
                        mapped_idx=iref;
                        break;
                      }
                      else //! invalid index. Default to zero even though this case should not happen
                        mapped_idx=INVALIDINDEX;
                      continue;
                    }    
#endif                                        
                    if (dec_picture->ref_pic_num[img->current_slice_nr][LIST_0][iref]==co_located_ref_id[refList][imgblock_y + j][img->block_x + k])
                    {
                      mapped_idx=iref;
                      break;
                    }
                    else //! invalid index. Default to zero even though this case should not happen
                      mapped_idx=INVALIDINDEX;
                  }
                  if (INVALIDINDEX == mapped_idx)
                  {
                    error("temporal direct error\ncolocated block has ref that is unavailable",-1111);
                  }
                  dec_picture->ref_idx [LIST_0][img->block_y + j][img->block_x + k] = mapped_idx;
                  dec_picture->ref_idx [LIST_1][img->block_y + j][img->block_x + k] = 0;                
                }
              }
            }
          }
        }
      }
    } 

    //  If multiple ref. frames, read reference frame for the MB *********************************
    if(img->num_ref_idx_l0_active>1) 
    {
      flag_mode = ( img->num_ref_idx_l0_active == 2 ? 1 : 0);

      currSE.type = SE_REFFRAME;
      dP = &(currSlice->partArr[partMap[SE_REFFRAME]]);
      currSE.reading = readRefFrame_CABAC;

      for (j0=0; j0<4; j0+=step_v0)
      {
        for (i0=0; i0<4; i0+=step_h0)
        {
          k=2*(j0/2)+(i0/2); // k is reused as loop index below. ????

          if ((currMB->b8pdir[k]==0 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)
          {
            TRACE_STRING("ref_idx_l0");

            img->subblock_x = i0;
            img->subblock_y = j0;

            if (!IS_P8x8 (currMB) || bframe || (!bframe && !img->allrefzero))
            {
              currSE.context = BType2CtxRef (currMB->b8mode[k]);
              {
                if( (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag) && flag_mode )
                {
                  currSE.len = 1;
                  readSyntaxElement_FLC(&currSE, dP->bitstream);
                  currSE.value1 = 1 - currSE.value1;
                }
                else
                {
                  currSE.value2 = LIST_0;
                  dP->readSyntaxElement (&currSE,img,inp,dP);
                }
                refframe = currSE.value1;
              }
            }
            else
            {
              refframe = 0;
            }

            for (j=j0; j<j0+step_v0;j++)
              for (i=i0; i<i0+step_h0;i++)
              {
                dec_picture->ref_idx[LIST_0][img->block_y + j][img->block_x + i] = refframe;
              }
          }
        }
      }
    }
    else
    {
      for (j0=0; j0<4; j0+=step_v0)
      {
        for (i0=0; i0<4; i0+=step_h0)
        {
          k=2*(j0/2)+(i0/2);
          if ((currMB->b8pdir[k]==0 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)
          {
            for (j=j0; j<j0+step_v0;j++)
              for (i=i0; i<i0+step_h0;i++)
              {
                dec_picture->ref_idx[LIST_0][img->block_y + j][img->block_x + i] = 0;
              }
          }
        }
      }
    }

    //  If backward multiple ref. frames, read backward reference frame for the MB *********************************
    if(img->num_ref_idx_l1_active>1)
    {
      flag_mode = ( img->num_ref_idx_l1_active == 2 ? 1 : 0);

      currSE.type = SE_REFFRAME;
      dP = &(currSlice->partArr[partMap[SE_REFFRAME]]);
      currSE.reading = readRefFrame_CABAC;

      for (j0=0; j0<4; j0+=step_v0)
      {
        for (i0=0; i0<4; i0+=step_h0)
        {
          k=2*(j0/2)+(i0/2);
          if ((currMB->b8pdir[k]==1 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)
          {
            TRACE_STRING("ref_idx_l1");

            img->subblock_x = i0;
            img->subblock_y = j0;

            currSE.context = BType2CtxRef (currMB->b8mode[k]);
            if( (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag) && flag_mode )
            {
              currSE.len = 1;
              readSyntaxElement_FLC(&currSE, dP->bitstream);
              currSE.value1 = 1-currSE.value1;
            }
            else
            {
              currSE.value2 = LIST_1;
              dP->readSyntaxElement (&currSE,img,inp,dP);
            }
            refframe = currSE.value1;

            for (j=j0; j<j0+step_v0;j++)
            {
              for (i=i0; i<i0+step_h0;i++)
              {
                dec_picture->ref_idx[LIST_1][img->block_y + j][img->block_x + i] = refframe;
              }
            }
          }
        }
      }
    }
    else
    {
      for (j0=0; j0<4; j0+=step_v0)
      {
        for (i0=0; i0<4; i0+=step_h0)
        {
          k=2*(j0/2)+(i0/2);
          if ((currMB->b8pdir[k]==1 || currMB->b8pdir[k]==2) && currMB->b8mode[k]!=0)
          {
            for (j=j0; j<j0+step_v0;j++)
              for (i=i0; i<i0+step_h0;i++)
              {
                dec_picture->ref_idx[LIST_1][img->block_y + j][img->block_x + i] = 0;
              }
          }
        }
      }
    }

    //=====  READ FORWARD MOTION VECTORS =====
    currSE.type = SE_MVD;
    dP = &(currSlice->partArr[partMap[SE_MVD]]);
    currSE.reading = readMVD_CABAC;

    for (j0=0; j0<4; j0+=step_v0)
      for (i0=0; i0<4; i0+=step_h0)
      {
        k=2*(j0/2)+(i0/2);

        if ((currMB->b8pdir[k]==0 || currMB->b8pdir[k]==2) && (currMB->b8mode[k] !=0))//has forward vector
        {
          mv_mode  = currMB->b8mode[k];
          step_h   = BLOCK_STEP [mv_mode][0];
          step_v   = BLOCK_STEP [mv_mode][1];

          refframe = dec_picture->ref_idx[LIST_0][img->block_y+j0][img->block_x+i0];

          for (j=j0; j<j0+step_v0; j+=step_v)
          {
            for (i=i0; i<i0+step_h0; i+=step_h)
            {
              j4 = img->block_y+j;
              i4 = img->block_x+i;
#ifdef ADAPTIVE_MV_RES
              if(img->mv_res)
              {
                img->subblock_x = i; // position used for context determination
                img->subblock_y = j; // position used for context determination
                TRACE_STRING("mv_res");
                currSE.reading = readMV_RES_4_8_CABAC;
                currSE.value2 = 0; // identifies the component; only used for context determination
                dP->readSyntaxElement(&currSE,img,inp,dP);
                mv_shift = currSE.value1;
                currSE.reading = readMVD_CABAC;
              }
              else
                mv_shift =0;
#endif
              // first make mv-prediction
#ifdef MV_COMPETITION
              if (mv_comp.mv_competition > 0)
              {
                for (k=0; k < 2; k++) 
                {
                  TRACE_STRING("mvd_l0");

                  img->subblock_x = i; // position used for context determination
                  img->subblock_y = j; // position used for context determination
#ifdef ADAPTIVE_MV_RES
                  currSE.value2 = (k<<2)+mv_shift; // identifies the component; only used for context determination
#else
                  currSE.value2 = k<<1; // identifies the component; only used for context determination
#endif
                  dP->readSyntaxElement(&currSE,img,inp,dP);
                  curr_mvd = currSE.value1; 

                  for(ii=0;ii<step_h;ii++)
                  {
                    for(jj=0;jj<step_v;jj++)
                    {
                      currMB->mvd      [LIST_0][j+jj] [i+ii] [k] = curr_mvd;
                    }
                  }
                }

                SetMotionVectorPredictor_Competition (img,inp, pmv, pmv+1, refframe, LIST_0, dec_picture->ref_idx, dec_picture->mv, i, j, 4*step_h, 4*step_v);

                for (k=0; k < 2; k++) 
                {
                  for(ii=0;ii<step_h;ii++)
                  {
                    for(jj=0;jj<step_v;jj++)
                    {
#ifdef ADAPTIVE_MV_RES
                      dec_picture->mv  [LIST_0][j4+jj][i4+ii][k] = (currMB->mvd      [LIST_0][j+jj] [i+ii] [k] + roundInt(pmv[k],mv_shift))<<mv_shift;
                      currMB->mvd      [LIST_0][j+jj] [i+ii] [k] =  currMB->mvd      [LIST_0][j+jj] [i+ii] [k]<<mv_shift;
#else
                      dec_picture->mv  [LIST_0][j4+jj][i4+ii][k] = currMB->mvd      [LIST_0][j+jj] [i+ii] [k] + pmv[k];
#endif
                    }
                  }
                }
              }

              else
#endif
              {


                SetMotionVectorPredictor (img, pmv, pmv+1, refframe, LIST_0, dec_picture->ref_idx, dec_picture->mv, i, j, 4*step_h, 4*step_v);



                for (k=0; k < 2; k++) 
                {
                  TRACE_STRING("mvd_l0");

                  img->subblock_x = i; // position used for context determination
                  img->subblock_y = j; // position used for context determination
#ifdef ADAPTIVE_MV_RES
                  currSE.value2 = (k<<2)+mv_shift; // identifies the component; only used for context determination
#else
                  currSE.value2 = k<<1; // identifies the component; only used for context determination
#endif
                  dP->readSyntaxElement(&currSE,img,inp,dP);
                  curr_mvd = currSE.value1; 
#ifdef ADAPTIVE_MV_RES                
                  motion_vector=(curr_mvd+roundInt(pmv[k],mv_shift))<<mv_shift;      /* find motion vector */
#else
                  motion_vector=curr_mvd+pmv[k];           /* find motion vector */
#endif

                  for(ii=0;ii<step_h;ii++)
                  {
                    for(jj=0;jj<step_v;jj++)
                    {
                      dec_picture->mv  [LIST_0][j4+jj][i4+ii][k] = motion_vector;
#ifdef ADAPTIVE_MV_RES
                      currMB->mvd      [LIST_0][j+jj] [i+ii] [k] = curr_mvd<<mv_shift;
#else
                      currMB->mvd      [LIST_0][j+jj] [i+ii] [k] = curr_mvd;
#endif
                    }
                  }
                }
              }
            }
          }
        }
        else if (currMB->b8mode[k=2*(j0/2)+(i0/2)]==0)      
        {  
          if (!img->direct_spatial_mv_pred_flag)
          {
            int imgblock_y= img->block_y;

            int refList = (co_located_ref_idx[LIST_0][imgblock_y+j0][img->block_x+i0]== -1 ? LIST_1 : LIST_0);
            int ref_idx =  co_located_ref_idx[refList][imgblock_y+j0][img->block_x+i0];          

            if (ref_idx==-1)
            {
              for (j=j0; j<j0+step_v0; j++)
                for (i=i0; i<i0+step_h0; i++)
                {            
                  dec_picture->ref_idx [LIST_1][img->block_y+j][img->block_x+i]=0;
                  dec_picture->ref_idx [LIST_0][img->block_y+j][img->block_x+i]=0; 
                  j4 = img->block_y+j;
                  i4 = img->block_x+i;            
                  for (ii=0; ii < 2; ii++) 
                  {                                    
                    dec_picture->mv [LIST_0][j4][i4][ii]=0;
                    dec_picture->mv [LIST_1][j4][i4][ii]=0;                  
                  }
                }
            }
            else 
            {        
              int mapped_idx=-1, iref;                             
              int j6;

              for (iref=0;iref<min(img->num_ref_idx_l0_active,listXsize[LIST_0]);iref++)
              {

#if 1
                //int curr_mb_field = 0;

                if(img->structure==0 )
                {
                  // If the current MB is a frame MB and the colocated is from a field picture, 
                  // then the co_located_ref_id may have been generated from the wrong value of 
                  // frame_poc if it references it's complementary field, so test both POC values
                  if(listX[0][iref]->top_poc*2 == co_located_ref_id[refList][imgblock_y + j0][img->block_x + i0] 
                  || listX[0][iref]->bottom_poc*2 == co_located_ref_id[refList][imgblock_y + j0][img->block_x + i0])
                  {
                    mapped_idx=iref;
                    break;
                  }
                  else //! invalid index. Default to zero even though this case should not happen
                    mapped_idx=INVALIDINDEX;
                  continue;
                }    
#endif                                        
                if (dec_picture->ref_pic_num[img->current_slice_nr][LIST_0][iref]==co_located_ref_id[refList][imgblock_y+j0][img->block_x+i0])
                {
                  mapped_idx=iref;
                  break;
                }
                else //! invalid index. Default to zero even though this case should not happen
                  mapped_idx=INVALIDINDEX;
              }

              if (INVALIDINDEX == mapped_idx)
              {
                error("temporal direct error\ncolocated block has ref that is unavailable",-1111);
              }


              for (j=j0; j<j0+step_v0; j++)
                for (i=i0; i<i0+step_h0; i++)
                {
                  {
                    mv_scale = img->mvscale[LIST_0][mapped_idx];

                    dec_picture->ref_idx [LIST_0][img->block_y+j][img->block_x+i] = mapped_idx;
                    dec_picture->ref_idx [LIST_1][img->block_y+j][img->block_x+i] = 0;

                    j4 = img->block_y+j;
                    j6 = imgblock_y+j;
                    i4 = img->block_x+i;

                    for (ii=0; ii < 2; ii++) 
                    {              
                      //if (iTRp==0)
                      if (mv_scale == 9999 || listX[LIST_0][mapped_idx]->is_long_term)
                        //                    if (mv_scale==9999 || Co_located->is_long_term)
                      {                      
                        dec_picture->mv  [LIST_0][j4][i4][ii]=co_located_mv[refList][j6][i4][ii];
                        dec_picture->mv  [LIST_1][j4][i4][ii]=0;
                      }
                      else
                      {
                        dec_picture->mv  [LIST_0][j4][i4][ii]=(mv_scale * co_located_mv[refList][j6][i4][ii] + 128 ) >> 8;
                        dec_picture->mv  [LIST_1][j4][i4][ii]=dec_picture->mv[LIST_0][j4][i4][ii] - co_located_mv[refList][j6][i4][ii];
                      }
                    }
                  } 
                }
            }  
          } 
        }
      }

      //=====  READ BACKWARD MOTION VECTORS =====
      currSE.type = SE_MVD;
      dP          = &(currSlice->partArr[partMap[SE_MVD]]);
      /*  
      if (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag) currSE.mapping = linfo_se;
      else*/                                                    currSE.reading = readMVD_CABAC;

      for (j0=0; j0<4; j0+=step_v0)
      {
        for (i0=0; i0<4; i0+=step_h0)
        {
          k=2*(j0/2)+(i0/2);
          if ((currMB->b8pdir[k]==1 || currMB->b8pdir[k]==2) && (currMB->b8mode[k]!=0))//has backward vector
          {
            mv_mode  = currMB->b8mode[k];
            step_h   = BLOCK_STEP [mv_mode][0];
            step_v   = BLOCK_STEP [mv_mode][1];

            refframe = dec_picture->ref_idx[LIST_1][img->block_y+j0][img->block_x+i0];

            for (j=j0; j<j0+step_v0; j+=step_v)
            {
              for (i=i0; i<i0+step_h0; i+=step_h)
              {
                j4 = img->block_y+j;
                i4 = img->block_x+i;

                // first make mv-prediction
#ifdef ADAPTIVE_MV_RES
                if (img->mv_res)
                {
                  img->subblock_x = i; // position used for context determination
                  img->subblock_y = j; // position used for context determination
                  TRACE_STRING("mv_res");
                  currSE.reading = readMV_RES_4_8_CABAC;
                  currSE.value2 = 0; // identifies the component; only used for context determination
                  dP->readSyntaxElement(&currSE,img,inp,dP);
                  mv_shift = currSE.value1;
                  currSE.reading = readMVD_CABAC;
                }
                else
                  mv_shift =0;
#endif            
#ifdef MV_COMPETITION
                if (mv_comp.mv_competition > 0)
                {
                  for (k=0; k < 2; k++) 
                  {
                    TRACE_STRING("mvd_l1");

                    img->subblock_x = i; // position used for context determination
                    img->subblock_y = j; // position used for context determination
#ifdef ADAPTIVE_MV_RES
                    currSE.value2 = (((k<<1)+1)<<1)+mv_shift; // identifies the component; only used for context determination
#else
                    currSE.value2   = (k<<1) +1; // identifies the component; only used for context determination
#endif
                    dP->readSyntaxElement(&currSE,img,inp,dP);
                    curr_mvd = currSE.value1; 

                    for(ii=0;ii<step_h;ii++)
                      for(jj=0;jj<step_v;jj++)
                        currMB->mvd      [LIST_1][j+jj] [i+ii] [k] = curr_mvd;
                  }

                  SetMotionVectorPredictor_Competition (img,inp, pmv, pmv+1, refframe, LIST_1, dec_picture->ref_idx, dec_picture->mv, i, j, 4*step_h, 4*step_v);

                  for (k=0; k < 2; k++) 
                  {
                    for(ii=0;ii<step_h;ii++)
                    {
                      for(jj=0;jj<step_v;jj++)
                      {
#ifdef ADAPTIVE_MV_RES
                        dec_picture->mv  [LIST_1][j4+jj][i4+ii][k] = (currMB->mvd      [LIST_1][j+jj] [i+ii] [k] + roundInt(pmv[k],mv_shift))<<mv_shift;
                        currMB->mvd      [LIST_1][j+jj] [i+ii] [k] =  currMB->mvd      [LIST_1][j+jj] [i+ii] [k]<<mv_shift;
#else
                        dec_picture->mv  [LIST_1][j4+jj][i4+ii][k] = currMB->mvd      [LIST_1][j+jj] [i+ii] [k] + pmv[k];
#endif

                      }
                    }
                  }
                }
                else

#endif
                {
                  SetMotionVectorPredictor (img, pmv, pmv+1, refframe, LIST_1, dec_picture->ref_idx, dec_picture->mv, i, j, 4*step_h, 4*step_v);


                  for (k=0; k < 2; k++) 
                  {
                    TRACE_STRING("mvd_l1");

                    img->subblock_x = i; // position used for context determination
                    img->subblock_y = j; // position used for context determination
#ifdef ADAPTIVE_MV_RES
                    currSE.value2 = (((k<<1)+1)<<1)+mv_shift; // identifies the component; only used for context determination
#else
                    currSE.value2   = (k<<1) +1; // identifies the component; only used for context determination
#endif
                    dP->readSyntaxElement(&currSE,img,inp,dP);
                    curr_mvd = currSE.value1; 

#ifdef ADAPTIVE_MV_RES                
                    motion_vector=(curr_mvd+roundInt(pmv[k],mv_shift))<<mv_shift;      /* find motion vector */
#else
                    motion_vector=curr_mvd+pmv[k];           /* find motion vector */
#endif

                    for(ii=0;ii<step_h;ii++)
                    {
                      for(jj=0;jj<step_v;jj++)
                      {
                        dec_picture->mv  [LIST_1][j4+jj][i4+ii][k] = motion_vector;
#ifdef ADAPTIVE_MV_RES
                        currMB->mvd      [LIST_1][j+jj] [i+ii] [k] = curr_mvd<<mv_shift;
#else
                        currMB->mvd      [LIST_1][j+jj] [i+ii] [k] = curr_mvd;
#endif
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
      // record reference picture Ids for deblocking decisions

      for(i4=img->block_x;i4<(img->block_x+4);i4++)
        for(j4=img->block_y;j4<(img->block_y+4);j4++)
        {
          if(dec_picture->ref_idx[LIST_0][j4][i4]>=0)
            dec_picture->ref_pic_id[LIST_0][j4][i4] = dec_picture->ref_pic_num[img->current_slice_nr][LIST_0][(short)dec_picture->ref_idx[LIST_0][j4][i4]];
          else
            dec_picture->ref_pic_id[LIST_0][j4][i4] = INT64_MIN;
          if(dec_picture->ref_idx[LIST_1][j4][i4]>=0)
            dec_picture->ref_pic_id[LIST_1][j4][i4] = dec_picture->ref_pic_num[img->current_slice_nr][LIST_1][(short)dec_picture->ref_idx[LIST_1][j4][i4]];  
          else
            dec_picture->ref_pic_id[LIST_1][j4][i4] = INT64_MIN;  
        }
  }



  /*!
  ************************************************************************
  * \brief
  *    Get the Prediction from the Neighboring Blocks for Number of Nonzero Coefficients 
  *    
  *    Luma Blocks
  ************************************************************************
  */
  int predict_nnz(struct img_par *img, int i,int j)
  {
    PixelPos pix;

    int pred_nnz = 0;
    int cnt      = 0;
    int mb_nr    = img->current_mb_nr;

    // left block
    getLuma4x4Neighbour(mb_nr, i, j, -1, 0, &pix);

    if (pix.available && active_pps->constrained_intra_pred_flag && (img->currentSlice->dp_mode==PAR_DP_3))
    {
      pix.available &= img->intra_block[pix.mb_addr];
      cnt--;
    }

    if (pix.available)
    {
      pred_nnz = img->nz_coeff [pix.mb_addr ][pix.x][pix.y];
      cnt++;
    }

    // top block
    getLuma4x4Neighbour(mb_nr, i, j, 0, -1, &pix);

    if (pix.available && active_pps->constrained_intra_pred_flag && (img->currentSlice->dp_mode==PAR_DP_3))
    {
      pix.available &= img->intra_block[pix.mb_addr];
      cnt--;
    }

    if (pix.available)
    {
      pred_nnz += img->nz_coeff [pix.mb_addr ][pix.x][pix.y];
      cnt++;
    }

    if (cnt==2)
    {
      pred_nnz++;
      pred_nnz>>=1; 
    }

    return pred_nnz;
  }


  /*!
  ************************************************************************
  * \brief
  *    Get the Prediction from the Neighboring Blocks for Number of Nonzero Coefficients 
  *    
  *    Chroma Blocks   
  ************************************************************************
  */
  int predict_nnz_chroma(struct img_par *img, int i,int j)
  {
    PixelPos pix;

    int pred_nnz = 0;
    int cnt      =0;
    int mb_nr    = img->current_mb_nr;
    int j_off_tab [12] = {0,0,0,0,4,4,4,4,8,8,8,8};
    int j_off = j_off_tab[j];

    if (dec_picture->chroma_format_idc != YUV444)
    {
      //YUV420 and YUV422
      // left block
      getChroma4x4Neighbour(mb_nr, i%2, j-4, -1, 0, &pix);

      if (pix.available && active_pps->constrained_intra_pred_flag && (img->currentSlice->dp_mode==PAR_DP_3))
      {
        pix.available &= img->intra_block[pix.mb_addr];
        cnt--;
      }

      if (pix.available)
      {
        pred_nnz = img->nz_coeff [pix.mb_addr ][2 * (i/2) + pix.x][4 + pix.y];
        cnt++;
      }

      // top block
      getChroma4x4Neighbour(mb_nr, i%2, j-4, 0, -1, &pix);

      if (pix.available && active_pps->constrained_intra_pred_flag && (img->currentSlice->dp_mode==PAR_DP_3))
      {
        pix.available &= img->intra_block[pix.mb_addr];
        cnt--;
      }

      if (pix.available)
      {
        pred_nnz += img->nz_coeff [pix.mb_addr ][2 * (i/2) + pix.x][4 + pix.y];
        cnt++;
      }
    }
    else
    {
      //YUV444
      // left block
      getChroma4x4Neighbour(mb_nr, i, j-j_off, -1, 0, &pix);

      if (pix.available && active_pps->constrained_intra_pred_flag && (img->currentSlice->dp_mode==PAR_DP_3))
      {
        pix.available &= img->intra_block[pix.mb_addr];
        cnt--;
      }

      if (pix.available)
      {
        pred_nnz = img->nz_coeff [pix.mb_addr ][pix.x][j_off + pix.y];
        cnt++;
      }

      // top block
      getChroma4x4Neighbour(mb_nr, i, j-j_off, 0, -1, &pix);

      if (pix.available && active_pps->constrained_intra_pred_flag && (img->currentSlice->dp_mode==PAR_DP_3))
      {
        pix.available &= img->intra_block[pix.mb_addr];
        cnt--;
      }

      if (pix.available)
      {
        pred_nnz += img->nz_coeff [pix.mb_addr ][pix.x][j_off + pix.y];
        cnt++;
      }    
    }

    if (cnt==2)
    {
      pred_nnz++;
      pred_nnz>>=1;
    }

    return pred_nnz;
  }


  /*!
  ************************************************************************
  * \brief
  *    Calculate the quantisation and inverse quantisation parameters
  *
  ************************************************************************
  */
  void CalculateQuant8Param()
  {
    int i, j, k, temp;

    for(k=0; k<6; k++)
      for(j=0; j<8; j++)
      {
        for(i=0; i<8; i++)
        {
          temp = (i<<3)+j;
          InvLevelScale8x8Luma_Intra[k][j][i] = dequant_coef8[k][j][i]*qmatrix[6][temp];
          InvLevelScale8x8Luma_Inter[k][j][i] = dequant_coef8[k][j][i]*qmatrix[7][temp];
        }
      }
  }


  const byte SNGL_SCAN16x16[256][2] = 
  {
    { 0, 0}, { 1, 0}, { 0, 1}, { 0, 2}, { 1, 1}, { 2, 0}, { 3, 0}, { 2, 1}, { 1, 2}, { 0, 3}, { 0, 4}, { 1, 3}, { 2, 2}, { 3, 1}, { 4, 0}, { 5, 0},
    { 4, 1}, { 3, 2}, { 2, 3}, { 1, 4}, { 0, 5}, { 0, 6}, { 1, 5}, { 2, 4}, { 3, 3}, { 4, 2}, { 5, 1}, { 6, 0}, { 7, 0}, { 6, 1}, { 5, 2}, { 4, 3},
    { 3, 4}, { 2, 5}, { 1, 6}, { 0, 7}, { 0, 8}, { 1, 7}, { 2, 6}, { 3, 5}, { 4, 4}, { 5, 3}, { 6, 2}, { 7, 1}, { 8, 0}, { 9, 0}, { 8, 1}, { 7, 2}, 
    { 6, 3}, { 5, 4}, { 4, 5}, { 3, 6}, { 2, 7}, { 1, 8}, { 0, 9}, { 0,10}, { 1, 9}, { 2, 8}, { 3, 7}, { 4, 6}, { 5, 5}, { 6, 4}, { 7, 3}, { 8, 2},
    { 9, 1}, {10, 0}, {11, 0}, {10, 1}, { 9, 2}, { 8, 3}, { 7, 4}, { 6, 5}, { 5, 6}, { 4, 7}, { 3, 8}, { 2, 9}, { 1,10}, { 0,11}, { 0,12}, { 1,11},
    { 2,10}, { 3, 9}, { 4, 8}, { 5, 7}, { 6, 6}, { 7, 5}, { 8, 4}, { 9, 3}, {10, 2}, {11, 1}, {12, 0}, {13, 0}, {12, 1}, {11, 2}, {10, 3}, { 9, 4},
    { 8, 5}, { 7, 6}, { 6, 7}, { 5, 8}, { 4, 9}, { 3,10}, { 2,11}, { 1,12}, { 0,13}, { 0,14}, { 1,13}, { 2,12}, { 3,11}, { 4,10}, { 5, 9}, { 6, 8},
    { 7, 7}, { 8, 6}, { 9, 5}, {10, 4}, {11, 3}, {12, 2}, {13, 1}, {14, 0}, {15, 0}, {14, 1}, {13, 2}, {12, 3}, {11, 4}, {10, 5}, { 9, 6}, { 8, 7},
    { 7, 8}, { 6, 9}, { 5,10}, { 4,11}, { 3,12}, { 2,13}, { 1,14}, { 0,15}, { 1,15}, { 2,14}, { 3,13}, { 4,12}, { 5,11}, { 6,10}, { 7, 9}, { 8, 8},
    { 9, 7}, {10, 6}, {11, 5}, {12, 4}, {13, 3}, {14, 2}, {15, 1}, {15, 2}, {14, 3}, {13, 4}, {12, 5}, {11, 6}, {10, 7}, { 9, 8}, { 8, 9}, { 7,10},
    { 6,11}, { 5,12}, { 4,13}, { 3,14}, { 2,15}, { 3,15}, { 4,14}, { 5,13}, { 6,12}, { 7,11}, { 8,10}, { 9, 9}, {10, 8}, {11, 7}, {12, 6}, {13, 5},
    {14, 4}, {15, 3}, {15, 4}, {14, 5}, {13, 6}, {12, 7}, {11, 8}, {10, 9}, { 9,10}, { 8,11}, { 7,12}, { 6,13}, { 5,14}, { 4,15}, { 5,15}, { 6,14},
    { 7,13}, { 8,12}, { 9,11}, {10,10}, {11, 9}, {12, 8}, {13, 7}, {14, 6}, {15, 5}, {15, 6}, {14, 7}, {13, 8}, {12, 9}, {11,10}, {10,11}, { 9,12},
    { 8,13}, { 7,14}, { 6,15}, { 7,15}, { 8,14}, { 9,13}, {10,12}, {11,11}, {12,10}, {13, 9}, {14, 8}, {15, 7}, {15, 8}, {14, 9}, {13,10}, {12,11},
    {11,12}, {10,13}, { 9,14}, { 8,15}, { 9,15}, {10,14}, {11,13}, {12,12}, {13,11}, {14,10}, {15, 9}, {15,10}, {14,11}, {13,12}, {12,13}, {11,14},
    {10,15}, {11,15}, {12,14}, {13,13}, {14,12}, {15,11}, {15,12}, {14,13}, {13,14}, {12,15}, {13,15}, {14,14}, {15,13}, {15,14}, {14,15}, {15,15}
  };

  const byte SNGL_SCAN16x8[128][2] = 
  {
    { 0, 0}, { 0, 1}, { 1, 0}, { 2, 0}, { 1, 1}, { 0, 2}, { 0, 3}, { 1, 2}, { 2, 1}, { 3, 0}, { 4, 0}, { 3, 1}, { 2, 2}, { 1, 3}, { 0, 4}, { 0, 5}, 
    { 1, 4}, { 2, 3}, { 3, 2}, { 4, 1}, { 5, 0}, { 6, 0}, { 5, 1}, { 4, 2}, { 3, 3}, { 2, 4}, { 1, 5}, { 0, 6}, { 0, 7}, { 1, 6}, { 2, 5}, { 3, 4}, 
    { 4, 3}, { 5, 2}, { 6, 1}, { 7, 0}, { 8, 0}, { 7, 1}, { 6, 2}, { 5, 3}, { 4, 4}, { 3, 5}, { 2, 6}, { 1, 7}, { 2, 7}, { 3, 6}, { 4, 5}, { 5, 4}, 
    { 6, 3}, { 7, 2}, { 8, 1}, { 9, 0}, {10, 0}, { 9, 1}, { 8, 2}, { 7, 3}, { 6, 4}, { 5, 5}, { 4, 6}, { 3, 7}, { 4, 7}, { 5, 6}, { 6, 5}, { 7, 4}, 
    { 8, 3}, { 9, 2}, {10, 1}, {11, 0}, {12, 0}, {11, 1}, {10, 2}, { 9, 3}, { 8, 4}, { 7, 5}, { 6, 6}, { 5, 7}, { 6, 7}, { 7, 6}, { 8, 5}, { 9, 4}, 
    {10, 3}, {11, 2}, {12, 1}, {13, 0}, {14, 0}, {13, 1}, {12, 2}, {11, 3}, {10, 4}, { 9, 5}, { 8, 6}, { 7, 7}, { 8, 7}, { 9, 6}, {10, 5}, {11, 4}, 
    {12, 3}, {13, 2}, {14, 1}, {15, 0}, {15, 1}, {14, 2}, {13, 3}, {12, 4}, {11, 5}, {10, 6}, { 9, 7}, {10, 7}, {11, 6}, {12, 5}, {13, 4}, {14, 3},
    {15, 2}, {15, 3}, {14, 4}, {13, 5}, {12, 6}, {11, 7}, {12, 7}, {13, 6}, {14, 5}, {15, 4}, {15, 5}, {14, 6}, {13, 7}, {14, 7}, {15, 6}, {15, 7}
  };

  static const int dequant_coef_16x16[6][16][16] =
  {
    {
      {160, 334, 264, 248, 168, 248, 373, 236, 160, 236, 373, 248, 168, 248, 264, 334},
      {334, 695, 550, 516, 350, 516, 777, 492, 334, 492, 777, 516, 350, 516, 550, 695},
      {264, 550, 435, 408, 277, 408, 615, 389, 264, 389, 615, 408, 277, 408, 435, 550},
      {248, 516, 408, 384, 260, 384, 577, 365, 248, 365, 577, 384, 260, 384, 408, 516},
      {168, 350, 277, 260, 177, 260, 392, 248, 168, 248, 392, 260, 177, 260, 277, 350},
      {248, 516, 408, 384, 260, 384, 577, 365, 248, 365, 577, 384, 260, 384, 408, 516},
      {373, 777, 615, 577, 392, 577, 869, 550, 373, 550, 869, 577, 392, 577, 615, 777},
      {236, 492, 389, 365, 248, 365, 550, 348, 236, 348, 550, 365, 248, 365, 389, 492},
      {160, 334, 264, 248, 168, 248, 373, 236, 160, 236, 373, 248, 168, 248, 264, 334},
      {236, 492, 389, 365, 248, 365, 550, 348, 236, 348, 550, 365, 248, 365, 389, 492},
      {373, 777, 615, 577, 392, 577, 869, 550, 373, 550, 869, 577, 392, 577, 615, 777},
      {248, 516, 408, 384, 260, 384, 577, 365, 248, 365, 577, 384, 260, 384, 408, 516},
      {168, 350, 277, 260, 177, 260, 392, 248, 168, 248, 392, 260, 177, 260, 277, 350},
      {248, 516, 408, 384, 260, 384, 577, 365, 248, 365, 577, 384, 260, 384, 408, 516},
      {264, 550, 435, 408, 277, 408, 615, 389, 264, 389, 615, 408, 277, 408, 435, 550},
      {334, 695, 550, 516, 350, 516, 777, 492, 334, 492, 777, 516, 350, 516, 550, 695},
    },
    {
      {180, 374, 296, 278, 189, 278, 419, 265, 180, 265, 419, 278, 189, 278, 296, 374},
      {374, 780, 617, 580, 393, 580, 873, 552, 374, 552, 873, 580, 393, 580, 617, 780},
      {296, 617, 488, 458, 311, 458, 690, 436, 296, 436, 690, 458, 311, 458, 488, 617},
      {278, 580, 458, 431, 292, 431, 648, 410, 278, 410, 648, 431, 292, 431, 458, 580},
      {189, 393, 311, 292, 198, 292, 440, 278, 189, 278, 440, 292, 198, 292, 311, 393},
      {278, 580, 458, 431, 292, 431, 648, 410, 278, 410, 648, 431, 292, 431, 458, 580},
      {419, 873, 690, 648, 440, 648, 976, 617, 419, 617, 976, 648, 440, 648, 690, 873},
      {265, 552, 436, 410, 278, 410, 617, 390, 265, 390, 617, 410, 278, 410, 436, 552},
      {180, 374, 296, 278, 189, 278, 419, 265, 180, 265, 419, 278, 189, 278, 296, 374},
      {265, 552, 436, 410, 278, 410, 617, 390, 265, 390, 617, 410, 278, 410, 436, 552},
      {419, 873, 690, 648, 440, 648, 976, 617, 419, 617, 976, 648, 440, 648, 690, 873},
      {278, 580, 458, 431, 292, 431, 648, 410, 278, 410, 648, 431, 292, 431, 458, 580},
      {189, 393, 311, 292, 198, 292, 440, 278, 189, 278, 440, 292, 198, 292, 311, 393},
      {278, 580, 458, 431, 292, 431, 648, 410, 278, 410, 648, 431, 292, 431, 458, 580},
      {296, 617, 488, 458, 311, 458, 690, 436, 296, 436, 690, 458, 311, 458, 488, 617},
      {374, 780, 617, 580, 393, 580, 873, 552, 374, 552, 873, 580, 393, 580, 617, 780},
      },
      {
        {202, 420, 332, 312, 212, 312, 470, 297, 202, 297, 470, 312, 212, 312, 332, 420},
        {420, 876, 693, 651, 441, 651, 979, 619, 420, 619, 979, 651, 441, 651, 693, 876},
        {332, 693, 548, 514, 349, 514, 774, 490, 332, 490, 774, 514, 349, 514, 548, 693},
        {312, 651, 514, 483, 328, 483, 728, 460, 312, 460, 728, 483, 328, 483, 514, 651},
        {212, 441, 349, 328, 222, 328, 494, 312, 212, 312, 494, 328, 222, 328, 349, 441},
        {312, 651, 514, 483, 328, 483, 728, 460, 312, 460, 728, 483, 328, 483, 514, 651},
        {470, 979, 774, 728, 494, 728, 1095, 693, 470, 693, 1095, 728, 494, 728, 774, 979},
        {297, 619, 490, 460, 312, 460, 693, 438, 297, 438, 693, 460, 312, 460, 490, 619},
        {202, 420, 332, 312, 212, 312, 470, 297, 202, 297, 470, 312, 212, 312, 332, 420},
        {297, 619, 490, 460, 312, 460, 693, 438, 297, 438, 693, 460, 312, 460, 490, 619},
        {470, 979, 774, 728, 494, 728, 1095, 693, 470, 693, 1095, 728, 494, 728, 774, 979},
        {312, 651, 514, 483, 328, 483, 728, 460, 312, 460, 728, 483, 328, 483, 514, 651},
        {212, 441, 349, 328, 222, 328, 494, 312, 212, 312, 494, 328, 222, 328, 349, 441},
        {312, 651, 514, 483, 328, 483, 728, 460, 312, 460, 728, 483, 328, 483, 514, 651},
        {332, 693, 548, 514, 349, 514, 774, 490, 332, 490, 774, 514, 349, 514, 548, 693},
        {420, 876, 693, 651, 441, 651, 979, 619, 420, 619, 979, 651, 441, 651, 693, 876},
      },
      {
        {226, 472, 373, 350, 238, 350, 527, 334, 226, 334, 527, 350, 238, 350, 373, 472},
        {472, 983, 777, 730, 496, 730, 1099, 695, 472, 695, 1099, 730, 496, 730, 777, 983},
        {373, 777, 615, 577, 392, 577, 869, 550, 373, 550, 869, 577, 392, 577, 615, 777},
        {350, 730, 577, 543, 368, 543, 817, 516, 350, 516, 817, 543, 368, 543, 577, 730},
        {238, 496, 392, 368, 250, 368, 554, 350, 238, 350, 554, 368, 250, 368, 392, 496},
        {350, 730, 577, 543, 368, 543, 817, 516, 350, 516, 817, 543, 368, 543, 577, 730},
        {527, 1099, 869, 817, 554, 817, 1229, 777, 527, 777, 1229, 817, 554, 817, 869, 1099},
        {334, 695, 550, 516, 350, 516, 777, 492, 334, 492, 777, 516, 350, 516, 550, 695},
        {226, 472, 373, 350, 238, 350, 527, 334, 226, 334, 527, 350, 238, 350, 373, 472},
        {334, 695, 550, 516, 350, 516, 777, 492, 334, 492, 777, 516, 350, 516, 550, 695},
        {527, 1099, 869, 817, 554, 817, 1229, 777, 527, 777, 1229, 817, 554, 817, 869, 1099},
        {350, 730, 577, 543, 368, 543, 817, 516, 350, 516, 817, 543, 368, 543, 577, 730},
        {238, 496, 392, 368, 250, 368, 554, 350, 238, 350, 554, 368, 250, 368, 392, 496},
        {350, 730, 577, 543, 368, 543, 817, 516, 350, 516, 817, 543, 368, 543, 577, 730},
        {373, 777, 615, 577, 392, 577, 869, 550, 373, 550, 869, 577, 392, 577, 615, 777},
        {472, 983, 777, 730, 496, 730, 1099, 695, 472, 695, 1099, 730, 496, 730, 777, 983},
        },
        {
          {254, 529, 419, 393, 267, 393, 592, 374, 254, 374, 592, 393, 267, 393, 419, 529},
          {529, 1104, 873, 820, 556, 820, 1234, 780, 529, 780, 1234, 820, 556, 820, 873, 1104},
          {419, 873, 690, 648, 440, 648, 976, 617, 419, 617, 976, 648, 440, 648, 690, 873},
          {393, 820, 648, 609, 413, 609, 917, 580, 393, 580, 917, 609, 413, 609, 648, 820},
          {267, 556, 440, 413, 280, 413, 622, 393, 267, 393, 622, 413, 280, 413, 440, 556},
          {393, 820, 648, 609, 413, 609, 917, 580, 393, 580, 917, 609, 413, 609, 648, 820},
          {592, 1234, 976, 917, 622, 917, 1380, 873, 592, 873, 1380, 917, 622, 917, 976, 1234},
          {374, 780, 617, 580, 393, 580, 873, 552, 374, 552, 873, 580, 393, 580, 617, 780},
          {254, 529, 419, 393, 267, 393, 592, 374, 254, 374, 592, 393, 267, 393, 419, 529},
          {374, 780, 617, 580, 393, 580, 873, 552, 374, 552, 873, 580, 393, 580, 617, 780},
          {592, 1234, 976, 917, 622, 917, 1380, 873, 592, 873, 1380, 917, 622, 917, 976, 1234},
          {393, 820, 648, 609, 413, 609, 917, 580, 393, 580, 917, 609, 413, 609, 648, 820},
          {267, 556, 440, 413, 280, 413, 622, 393, 267, 393, 622, 413, 280, 413, 440, 556},
          {393, 820, 648, 609, 413, 609, 917, 580, 393, 580, 917, 609, 413, 609, 648, 820},
          {419, 873, 690, 648, 440, 648, 976, 617, 419, 617, 976, 648, 440, 648, 690, 873},
          {529, 1104, 873, 820, 556, 820, 1234, 780, 529, 780, 1234, 820, 556, 820, 873, 1104},
        },
        {
          {285, 594, 470, 441, 299, 441, 664, 420, 285, 420, 664, 441, 299, 441, 470, 594},
          {594, 1239, 979, 920, 624, 920, 1385, 876, 594, 876, 1385, 920, 624, 920, 979, 1239},
          {470, 979, 774, 728, 494, 728, 1095, 693, 470, 693, 1095, 728, 494, 728, 774, 979},
          {441, 920, 728, 684, 464, 684, 1029, 651, 441, 651, 1029, 684, 464, 684, 728, 920},
          {299, 624, 494, 464, 315, 464, 698, 441, 299, 441, 698, 464, 315, 464, 494, 624},
          {441, 920, 728, 684, 464, 684, 1029, 651, 441, 651, 1029, 684, 464, 684, 728, 920},
          {664, 1385, 1095, 1029, 698, 1029, 1549, 979, 664, 979, 1549, 1029, 698, 1029, 1095, 1385},
          {420, 876, 693, 651, 441, 651, 979, 619, 420, 619, 979, 651, 441, 651, 693, 876},
          {285, 594, 470, 441, 299, 441, 664, 420, 285, 420, 664, 441, 299, 441, 470, 594},
          {420, 876, 693, 651, 441, 651, 979, 619, 420, 619, 979, 651, 441, 651, 693, 876},
          {664, 1385, 1095, 1029, 698, 1029, 1549, 979, 664, 979, 1549, 1029, 698, 1029, 1095, 1385},
          {441, 920, 728, 684, 464, 684, 1029, 651, 441, 651, 1029, 684, 464, 684, 728, 920},
          {299, 624, 494, 464, 315, 464, 698, 441, 299, 441, 698, 464, 315, 464, 494, 624},
          {441, 920, 728, 684, 464, 684, 1029, 651, 441, 651, 1029, 684, 464, 684, 728, 920},
          {470, 979, 774, 728, 494, 728, 1095, 693, 470, 693, 1095, 728, 494, 728, 774, 979},
          {594, 1239, 979, 920, 624, 920, 1385, 876, 594, 876, 1385, 920, 624, 920, 979, 1239},
          }
  };

  static const int dequant_coef_16x8[6][8][16] =
  {
    {
      {226, 472, 373, 350, 238, 350, 527, 334, 226, 334, 527, 350, 238, 350, 373, 472},
      {373, 777, 615, 577, 392, 577, 869, 550, 373, 550, 869, 577, 392, 577, 615, 777},
      {238, 496, 392, 368, 250, 368, 554, 350, 238, 350, 554, 368, 250, 368, 392, 496},
      {527, 1099, 869, 817, 554, 817, 1229, 777, 527, 777, 1229, 817, 554, 817, 869, 1099},
      {226, 472, 373, 350, 238, 350, 527, 334, 226, 334, 527, 350, 238, 350, 373, 472},
      {527, 1099, 869, 817, 554, 817, 1229, 777, 527, 777, 1229, 817, 554, 817, 869, 1099},
      {238, 496, 392, 368, 250, 368, 554, 350, 238, 350, 554, 368, 250, 368, 392, 496},
      {373, 777, 615, 577, 392, 577, 869, 550, 373, 550, 869, 577, 392, 577, 615, 777},
    },
    {
      {254, 529, 419, 393, 267, 393, 592, 374, 254, 374, 592, 393, 267, 393, 419, 529},
      {419, 873, 690, 648, 440, 648, 976, 617, 419, 617, 976, 648, 440, 648, 690, 873},
      {267, 556, 440, 413, 280, 413, 622, 393, 267, 393, 622, 413, 280, 413, 440, 556},
      {592, 1234, 976, 917, 622, 917, 1380, 873, 592, 873, 1380, 917, 622, 917, 976, 1234},
      {254, 529, 419, 393, 267, 393, 592, 374, 254, 374, 592, 393, 267, 393, 419, 529},
      {592, 1234, 976, 917, 622, 917, 1380, 873, 592, 873, 1380, 917, 622, 917, 976, 1234},
      {267, 556, 440, 413, 280, 413, 622, 393, 267, 393, 622, 413, 280, 413, 440, 556},
      {419, 873, 690, 648, 440, 648, 976, 617, 419, 617, 976, 648, 440, 648, 690, 873},
      },
      {
        {285, 594, 470, 441, 299, 441, 664, 420, 285, 420, 664, 441, 299, 441, 470, 594},
        {470, 979, 774, 728, 494, 728, 1095, 693, 470, 693, 1095, 728, 494, 728, 774, 979},
        {299, 624, 494, 464, 315, 464, 698, 441, 299, 441, 698, 464, 315, 464, 494, 624},
        {664, 1385, 1095, 1029, 698, 1029, 1549, 979, 664, 979, 1549, 1029, 698, 1029, 1095, 1385},
        {285, 594, 470, 441, 299, 441, 664, 420, 285, 420, 664, 441, 299, 441, 470, 594},
        {664, 1385, 1095, 1029, 698, 1029, 1549, 979, 664, 979, 1549, 1029, 698, 1029, 1095, 1385},
        {299, 624, 494, 464, 315, 464, 698, 441, 299, 441, 698, 464, 315, 464, 494, 624},
        {470, 979, 774, 728, 494, 728, 1095, 693, 470, 693, 1095, 728, 494, 728, 774, 979},
      },
      {
        {320, 667, 527, 496, 336, 496, 746, 472, 320, 472, 746, 496, 336, 496, 527, 667},
        {527, 1099, 869, 817, 554, 817, 1229, 777, 527, 777, 1229, 817, 554, 817, 869, 1099},
        {336, 701, 554, 521, 353, 521, 783, 496, 336, 496, 783, 521, 353, 521, 554, 701},
        {746, 1555, 1229, 1155, 783, 1155, 1738, 1099, 746, 1099, 1738, 1155, 783, 1155, 1229, 1555},
        {320, 667, 527, 496, 336, 496, 746, 472, 320, 472, 746, 496, 336, 496, 527, 667},
        {746, 1555, 1229, 1155, 783, 1155, 1738, 1099, 746, 1099, 1738, 1155, 783, 1155, 1229, 1555},
        {336, 701, 554, 521, 353, 521, 783, 496, 336, 496, 783, 521, 353, 521, 554, 701},
        {527, 1099, 869, 817, 554, 817, 1229, 777, 527, 777, 1229, 817, 554, 817, 869, 1099},
        },
        {
          {359, 749, 592, 556, 377, 556, 837, 529, 359, 529, 837, 556, 377, 556, 592, 749},
          {592, 1234, 976, 917, 622, 917, 1380, 873, 592, 873, 1380, 917, 622, 917, 976, 1234},
          {377, 787, 622, 584, 396, 584, 879, 556, 377, 556, 879, 584, 396, 584, 622, 787},
          {837, 1745, 1380, 1296, 879, 1296, 1951, 1234, 837, 1234, 1951, 1296, 879, 1296, 1380, 1745},
          {359, 749, 592, 556, 377, 556, 837, 529, 359, 529, 837, 556, 377, 556, 592, 749},
          {837, 1745, 1380, 1296, 879, 1296, 1951, 1234, 837, 1234, 1951, 1296, 879, 1296, 1380, 1745},
          {377, 787, 622, 584, 396, 584, 879, 556, 377, 556, 879, 584, 396, 584, 622, 787},
          {592, 1234, 976, 917, 622, 917, 1380, 873, 592, 873, 1380, 917, 622, 917, 976, 1234},
        },
        {
          {403, 840, 664, 624, 424, 624, 940, 594, 403, 594, 940, 624, 424, 624, 664, 840},
          {664, 1385, 1095, 1029, 698, 1029, 1549, 979, 664, 979, 1549, 1029, 698, 1029, 1095, 1385},
          {424, 883, 698, 656, 445, 656, 987, 624, 424, 624, 987, 656, 445, 656, 698, 883},
          {940, 1959, 1549, 1455, 987, 1455, 2190, 1385, 940, 1385, 2190, 1455, 987, 1455, 1549, 1959},
          {403, 840, 664, 624, 424, 624, 940, 594, 403, 594, 940, 624, 424, 624, 664, 840},
          {940, 1959, 1549, 1455, 987, 1455, 2190, 1385, 940, 1385, 2190, 1455, 987, 1455, 1549, 1959},
          {424, 883, 698, 656, 445, 656, 987, 624, 424, 624, 987, 656, 445, 656, 698, 883},
          {664, 1385, 1095, 1029, 698, 1029, 1549, 979, 664, 979, 1549, 1029, 698, 1029, 1095, 1385},
          }
  };

  static const int dequant_coef_8x16[6][16][8] =
  {
    {
      {226, 373, 238, 527, 226, 527, 238, 373},
      {472, 777, 496, 1099, 472, 1099, 496, 777},
      {373, 615, 392, 869, 373, 869, 392, 615},
      {350, 577, 368, 817, 350, 817, 368, 577},
      {238, 392, 250, 554, 238, 554, 250, 392},
      {350, 577, 368, 817, 350, 817, 368, 577},
      {527, 869, 554, 1229, 527, 1229, 554, 869},
      {334, 550, 350, 777, 334, 777, 350, 550},
      {226, 373, 238, 527, 226, 527, 238, 373},
      {334, 550, 350, 777, 334, 777, 350, 550},
      {527, 869, 554, 1229, 527, 1229, 554, 869},
      {350, 577, 368, 817, 350, 817, 368, 577},
      {238, 392, 250, 554, 238, 554, 250, 392},
      {350, 577, 368, 817, 350, 817, 368, 577},
      {373, 615, 392, 869, 373, 869, 392, 615},
      {472, 777, 496, 1099, 472, 1099, 496, 777},
    },
    {
      {254, 419, 267, 592, 254, 592, 267, 419},
      {529, 873, 556, 1234, 529, 1234, 556, 873},
      {419, 690, 440, 976, 419, 976, 440, 690},
      {393, 648, 413, 917, 393, 917, 413, 648},
      {267, 440, 280, 622, 267, 622, 280, 440},
      {393, 648, 413, 917, 393, 917, 413, 648},
      {592, 976, 622, 1380, 592, 1380, 622, 976},
      {374, 617, 393, 873, 374, 873, 393, 617},
      {254, 419, 267, 592, 254, 592, 267, 419},
      {374, 617, 393, 873, 374, 873, 393, 617},
      {592, 976, 622, 1380, 592, 1380, 622, 976},
      {393, 648, 413, 917, 393, 917, 413, 648},
      {267, 440, 280, 622, 267, 622, 280, 440},
      {393, 648, 413, 917, 393, 917, 413, 648},
      {419, 690, 440, 976, 419, 976, 440, 690},
      {529, 873, 556, 1234, 529, 1234, 556, 873},
      },
      {
        {285, 470, 299, 664, 285, 664, 299, 470},
        {594, 979, 624, 1385, 594, 1385, 624, 979},
        {470, 774, 494, 1095, 470, 1095, 494, 774},
        {441, 728, 464, 1029, 441, 1029, 464, 728},
        {299, 494, 315, 698, 299, 698, 315, 494},
        {441, 728, 464, 1029, 441, 1029, 464, 728},
        {664, 1095, 698, 1549, 664, 1549, 698, 1095},
        {420, 693, 441, 979, 420, 979, 441, 693},
        {285, 470, 299, 664, 285, 664, 299, 470},
        {420, 693, 441, 979, 420, 979, 441, 693},
        {664, 1095, 698, 1549, 664, 1549, 698, 1095},
        {441, 728, 464, 1029, 441, 1029, 464, 728},
        {299, 494, 315, 698, 299, 698, 315, 494},
        {441, 728, 464, 1029, 441, 1029, 464, 728},
        {470, 774, 494, 1095, 470, 1095, 494, 774},
        {594, 979, 624, 1385, 594, 1385, 624, 979},
      },
      {
        {320, 527, 336, 746, 320, 746, 336, 527},
        {667, 1099, 701, 1555, 667, 1555, 701, 1099},
        {527, 869, 554, 1229, 527, 1229, 554, 869},
        {496, 817, 521, 1155, 496, 1155, 521, 817},
        {336, 554, 353, 783, 336, 783, 353, 554},
        {496, 817, 521, 1155, 496, 1155, 521, 817},
        {746, 1229, 783, 1738, 746, 1738, 783, 1229},
        {472, 777, 496, 1099, 472, 1099, 496, 777},
        {320, 527, 336, 746, 320, 746, 336, 527},
        {472, 777, 496, 1099, 472, 1099, 496, 777},
        {746, 1229, 783, 1738, 746, 1738, 783, 1229},
        {496, 817, 521, 1155, 496, 1155, 521, 817},
        {336, 554, 353, 783, 336, 783, 353, 554},
        {496, 817, 521, 1155, 496, 1155, 521, 817},
        {527, 869, 554, 1229, 527, 1229, 554, 869},
        {667, 1099, 701, 1555, 667, 1555, 701, 1099},
        },
        {
          {359, 592, 377, 837, 359, 837, 377, 592},
          {749, 1234, 787, 1745, 749, 1745, 787, 1234},
          {592, 976, 622, 1380, 592, 1380, 622, 976},
          {556, 917, 584, 1296, 556, 1296, 584, 917},
          {377, 622, 396, 879, 377, 879, 396, 622},
          {556, 917, 584, 1296, 556, 1296, 584, 917},
          {837, 1380, 879, 1951, 837, 1951, 879, 1380},
          {529, 873, 556, 1234, 529, 1234, 556, 873},
          {359, 592, 377, 837, 359, 837, 377, 592},
          {529, 873, 556, 1234, 529, 1234, 556, 873},
          {837, 1380, 879, 1951, 837, 1951, 879, 1380},
          {556, 917, 584, 1296, 556, 1296, 584, 917},
          {377, 622, 396, 879, 377, 879, 396, 622},
          {556, 917, 584, 1296, 556, 1296, 584, 917},
          {592, 976, 622, 1380, 592, 1380, 622, 976},
          {749, 1234, 787, 1745, 749, 1745, 787, 1234},
        },
        {
          {403, 664, 424, 940, 403, 940, 424, 664},
          {840, 1385, 883, 1959, 840, 1959, 883, 1385},
          {664, 1095, 698, 1549, 664, 1549, 698, 1095},
          {624, 1029, 656, 1455, 624, 1455, 656, 1029},
          {424, 698, 445, 987, 424, 987, 445, 698},
          {624, 1029, 656, 1455, 624, 1455, 656, 1029},
          {940, 1549, 987, 2190, 940, 2190, 987, 1549},
          {594, 979, 624, 1385, 594, 1385, 624, 979},
          {403, 664, 424, 940, 403, 940, 424, 664},
          {594, 979, 624, 1385, 594, 1385, 624, 979},
          {940, 1549, 987, 2190, 940, 2190, 987, 1549},
          {624, 1029, 656, 1455, 624, 1455, 656, 1029},
          {424, 698, 445, 987, 424, 987, 445, 698},
          {624, 1029, 656, 1455, 624, 1455, 656, 1029},
          {664, 1095, 698, 1549, 664, 1549, 698, 1095},
          {840, 1385, 883, 1959, 840, 1959, 883, 1385},
          }
  };


  static void readLumaCoeff16x16_CABAC (struct img_par *img,struct inp_par *inp, int mb32)
  {
    int mb_nr = img->current_mb_nr;
    Macroblock *currMB = &img->mb_data[mb_nr];
    int cbp = currMB->cbp;
    SyntaxElement currSE;
    Slice *currSlice = img->currentSlice;
    DataPartition *dP;
    int *partMap = assignSE2partition[currSlice->dp_mode];
    int coef_ctr;// i0, j0;
    int start_scan;
    int any_coeff;
    int numCoeff = MB_BLOCK_SIZE*MB_BLOCK_SIZE+1;

    int level;
    int run, len;

    int qp_per    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)/6;
    int qp_rem    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)%6;

    int i, j, k;

    for(i = 0; i < MB_BLOCK_SIZE; i++)
      memset(img->cof16x16P[i], 0, sizeof(int)*MB_BLOCK_SIZE);

    if (cbp & 0xf)  // are there any coeff in current block at all
    {
      start_scan = 0; // take all coeffs
      coef_ctr = start_scan-1;
      level    = 1;

      for(k=start_scan;(k < numCoeff) && (level != 0);k++)
      {
        //============ read =============
        /*
        * make distinction between INTRA and INTER coded
        * luminance coefficients
        */
        currSE.context      = LUMA_16x16P;
        currSE.type         = (k==0 ? SE_LUM_DC_INTER : SE_LUM_AC_INTER);

        dP = &(currSlice->partArr[partMap[currSE.type]]);
        currSE.reading = mb32 ? readRunLevel_CABAC16x16 : readRunLevel_CABAC;
        dP->readSyntaxElement(&currSE,img,inp,dP);

        level = currSE.value1;
        run   = currSE.value2;
        len   = currSE.len;

        //============ decode =============
        if (level != 0)    /* leave if len=1 */
        {
          any_coeff=1;
          coef_ctr += run+1;

          i=SNGL_SCAN16x16[coef_ctr][0];
          j=SNGL_SCAN16x16[coef_ctr][1];

          currMB->cbp_blk |= 0xffffffff;
          img->cof16x16P[j][i] = level * dequant_coef_16x16[qp_rem][j][i] * (1<<qp_per); 

          //if(i!=0 || j != 0) img->cof16x16P[j][i]=0;
        }
      }
    }
  }


  static void readLumaCoeff16x8_CABAC (struct img_par *img,struct inp_par *inp, int blk)
  {
    int mb_nr = img->current_mb_nr;
    Macroblock *currMB = &img->mb_data[mb_nr];
    int cbp = currMB->cbp;
    SyntaxElement currSE;
    Slice *currSlice = img->currentSlice;
    DataPartition *dP;
    int *partMap = assignSE2partition[currSlice->dp_mode];
    int coef_ctr;// i0, j0;
    int start_scan;
    int any_coeff;
    int numCoeff = MB_BLOCK_SIZE*MB_BLOCK_SIZE/2+1;

    int level;
    int run, len;

    int qp_per    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)/6;
    int qp_rem    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)%6;

    int i, j, k;

    for(i = 0; i < MB_BLOCK_SIZE; i++)
      memset(img->cof16x8[blk][i], 0, sizeof(int)*MB_BLOCK_SIZE);

    img->subblock_x = 0;               // horiz. position for coeff_count context
    img->subblock_y = (blk == 0)?0:2;  // vert.  position for coeff_count context

    if (cbp & (0x3 << blk*2))  // are there any coeff in current block at all
    {
      start_scan = 0; // take all coeffs
      coef_ctr = start_scan-1;
      level    = 1;

#ifdef VLC_CBP_BIT
      set_CBP_block_bit(currMB, LUMA_16x8P, 0, (blk == 0)?0:2, 1);
#endif //VLC_CBP_BIT
      for(k=start_scan;(k < numCoeff) && (level != 0);k++)
      {
        //============ read =============
        /*
        * make distinction between INTRA and INTER coded
        * luminance coefficients
        */
        currSE.context      = LUMA_16x8P;
        currSE.type         = (k==0 ? SE_LUM_DC_INTER : SE_LUM_AC_INTER);

        dP = &(currSlice->partArr[partMap[currSE.type]]);
        currSE.reading = readRunLevel_CABAC;
        dP->readSyntaxElement(&currSE,img,inp,dP);

        level = currSE.value1;
        run   = currSE.value2;
        len   = currSE.len;


        //============ decode =============
        if (level != 0)    /* leave if len=1 */
        {
          any_coeff=1;
          coef_ctr += run+1;

          i=SNGL_SCAN16x8[coef_ctr][0];
          j=SNGL_SCAN16x8[coef_ctr][1];

          currMB->cbp_blk |= (0xff<<blk*8);

          img->cof16x8[blk][j][i] = level * dequant_coef_16x8[qp_rem][j][i] * (1<<qp_per); 
        }
      }
    }
  }


  static void readLumaCoeff8x16_CABAC (struct img_par *img,struct inp_par *inp, int blk)
  {
    int mb_nr = img->current_mb_nr;
    Macroblock *currMB = &img->mb_data[mb_nr];
    int cbp = currMB->cbp;
    SyntaxElement currSE;
    Slice *currSlice = img->currentSlice;
    DataPartition *dP;
    int *partMap = assignSE2partition[currSlice->dp_mode];
    int coef_ctr;// i0, j0;
    int start_scan;
    int any_coeff;
    int numCoeff = MB_BLOCK_SIZE*MB_BLOCK_SIZE/2+1;

    int level;
    int run, len;

    int qp_per    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)/6;
    int qp_rem    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)%6;

    int i, j, k;

    for(i = 0; i < MB_BLOCK_SIZE; i++)
      memset(img->cof8x16[blk][i], 0, sizeof(int)*MB_BLOCK_SIZE);

    img->subblock_x = (blk == 0)?0:2;     // horiz. position for coeff_count context
    img->subblock_y = 0;                  // vert.  position for coeff_count context

    if (cbp & (0x5<<blk))  // are there any coeff in current block at all
    {
      start_scan = 0; // take all coeffs
      coef_ctr = start_scan-1;
      level    = 1;

#ifdef VLC_CBP_BIT
      set_CBP_block_bit(currMB, LUMA_8x16P, (blk == 0)?0:2, 0, 1);
#endif //VLC_CBP_BIT
      for(k=start_scan;(k < numCoeff) && (level != 0);k++)
      {
        //============ read =============
        /*
        * make distinction between INTRA and INTER coded
        * luminance coefficients
        */
        currSE.context      = LUMA_8x16P;
        currSE.type         = (k==0 ? SE_LUM_DC_INTER : SE_LUM_AC_INTER);

        dP = &(currSlice->partArr[partMap[currSE.type]]);
        currSE.reading = readRunLevel_CABAC;
        dP->readSyntaxElement(&currSE,img,inp,dP);

        level = currSE.value1;
        run   = currSE.value2;
        len   = currSE.len;

        //============ decode =============
        if (level != 0)    /* leave if len=1 */
        {
          any_coeff=1;
          coef_ctr += run+1;

          j=SNGL_SCAN16x8[coef_ctr][0];
          i=SNGL_SCAN16x8[coef_ctr][1];

          currMB->cbp_blk |= (0x3333 << blk*2);

          img->cof8x16[blk][j][i] = level * dequant_coef_8x16[qp_rem][j][i] * (1<<qp_per); 
        }
      }
    }
  }


  void readLumaCoeffBigBlock_CABAC (struct img_par *img,struct inp_par *inp, int mb32)
  {
    int mb_nr = img->current_mb_nr;
    Macroblock *currMB = &img->mb_data[mb_nr];
    int i;

#ifdef GEOM
    if(currMB->mb_type == P16x16 || currMB->mb_type == 0 || mb32 || currMB->geo_mode!=-1)
#else
    if(currMB->mb_type == P16x16 || currMB->mb_type == 0 || mb32)
#endif
    {
      if(currMB->cbp & 0xf)
        readLumaCoeff16x16_CABAC(img, inp, mb32);
      else
      {
        // set coefficients to all zero
        for(i = 0; i < MB_BLOCK_SIZE; i++)
          memset(img->cof16x16P[i], 0, sizeof(int)*MB_BLOCK_SIZE);
      }
    }
    else if(currMB->mb_type == P16x8)
    {
      readLumaCoeff16x8_CABAC(img, inp, 0);
      readLumaCoeff16x8_CABAC(img, inp, 1);
    }
    else //if(currMB->mb_type == P8x16)
    {
      readLumaCoeff8x16_CABAC(img, inp, 0);
      readLumaCoeff8x16_CABAC(img, inp, 1);
    }
  }




  /*!
  ************************************************************************
  * \brief
  *    Get coefficients (run/level) of one 8x8 block  
  *    from the NAL (CABAC Mode)
  ************************************************************************
  */
  void readLumaCoeff8x8_CABAC (struct img_par *img,struct inp_par *inp, int b8)
  {
    int i,j,k;
    int level;
    int mb_nr = img->current_mb_nr;
    Macroblock *currMB = &img->mb_data[mb_nr];
    int cbp = currMB->cbp;
    SyntaxElement currSE;
    Slice *currSlice = img->currentSlice;
    DataPartition *dP;
    int *partMap = assignSE2partition[currSlice->dp_mode];
    int coef_ctr;// i0, j0;
    int start_scan;
    int boff_x, boff_y;
    int any_coeff;
    int dq_lshift = 0, dq_rshift = 0, dq_round = 0;

    int run, len;

    int qp_per    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)/6;
    int qp_rem    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)%6;
    Boolean lossless_qpprime = ((img->qp + img->bitdepth_luma_qp_scale)==0 && img->lossless_qpprime_flag==1);
#ifdef USE_INTRA_MDDT
    int R[6] = {40, 45, 50, 57, 63, 71};
    int img_block_x = (img->mb_x)*4 + 2*(b8%2);
    int img_block_y = (img->mb_y)*4 + 2*(b8/2);
    int ipmode = img->ipredmode[img_block_x][img_block_y];
#endif

    if (qp_per < 6)
    {
      dq_rshift = 6 - qp_per;
      dq_round  = 1<<(5-qp_per);
    }
    else
      dq_lshift = qp_per - 6;

    img->is_intra_block = IS_INTRA(currMB);

    if (cbp & (1<<b8))  // are there any coeff in current block at all
    {
      // === set offset in current macroblock ===
      boff_x = (b8%2)<<3;
      boff_y = (b8/2)<<3;

      img->subblock_x = boff_x>>2; // position for coeff_count ctx
      img->subblock_y = boff_y>>2; // position for coeff_count ctx

      start_scan = 0; // take all coeffs
      coef_ctr = start_scan-1;
      level    = 1;

      for(k=start_scan;(k < 65) && (level != 0);k++)
      {
        //============ read =============
        /*
        * make distinction between INTRA and INTER coded
        * luminance coefficients
        */
        currSE.context      = LUMA_8x8;
        currSE.type         = (IS_INTRA(currMB) ?
          (k==0 ? SE_LUM_DC_INTRA : SE_LUM_AC_INTRA) :
          (k==0 ? SE_LUM_DC_INTER : SE_LUM_AC_INTER));

#if TRACE
        sprintf(currSE.tracestring, "Luma8x8 sng ");
#endif
        dP = &(currSlice->partArr[partMap[currSE.type]]);
        currSE.reading = readRunLevel_CABAC;

        dP->readSyntaxElement(&currSE,img,inp,dP);
        level = currSE.value1;
        run   = currSE.value2;
        len   = currSE.len;


        //============ decode =============
        if (level != 0)    /* leave if len=1 */
        {
          any_coeff=1;
          coef_ctr += run+1;

          i=SNGL_SCAN8x8[coef_ctr][0];
          j=SNGL_SCAN8x8[coef_ctr][1];

          currMB->cbp_blk |= 51 << (4*b8-2*(b8%2)); // corresponds to 110011, as if all four 4x4 blocks contain coeff, shifted to block position

          if(lossless_qpprime)
          {
            img->m7[boff_x + i][boff_y + j] = level;
          }
          else if (qp_per>=6)
          {
            if(img->is_intra_block == 1)
#ifdef USE_INTRA_MDDT
            {
              if(img->UseIntraMDDT)
              {
                i=img->scanOrder8x8[ipmode][coef_ctr][0];
                j=img->scanOrder8x8[ipmode][coef_ctr][1];
                if(ipmode == 2)
                  img->m7[boff_x + i][boff_y + j] = level*InvLevelScale8x8Luma_Intra[qp_rem][i][j]<<dq_lshift; // dequantization
                else
                  img->m7[boff_x + i][boff_y + j] = level*(R[qp_rem]<<qp_per); // dequantization
                img->scanStats8x8[ipmode][coef_ctr]++;
              }
              else
                img->m7[boff_x + i][boff_y + j] = level*InvLevelScale8x8Luma_Intra[qp_rem][i][j]<<dq_lshift; // dequantization
            }
#else
              img->m7[boff_x + i][boff_y + j] = level*InvLevelScale8x8Luma_Intra[qp_rem][i][j]<<dq_lshift; // dequantization
#endif
            else
              img->m7[boff_x + i][boff_y + j] = level*InvLevelScale8x8Luma_Inter[qp_rem][i][j]<<dq_lshift; // dequantization
          }
          else
          {
            if(img->is_intra_block == 1)
#ifdef USE_INTRA_MDDT
            {
              if(img->UseIntraMDDT)
              {
                i=img->scanOrder8x8[ipmode][coef_ctr][0];
                j=img->scanOrder8x8[ipmode][coef_ctr][1];
                if(ipmode == 2)
                  img->m7[boff_x + i][boff_y + j] = (level*InvLevelScale8x8Luma_Intra[qp_rem][i][j]+dq_round)>>dq_rshift; // dequantization
                else
                  img->m7[boff_x + i][boff_y + j] = level*(R[qp_rem]<<qp_per); // dequantization
                img->scanStats8x8[ipmode][coef_ctr]++;
              }
              else
                img->m7[boff_x + i][boff_y + j] = (level*InvLevelScale8x8Luma_Intra[qp_rem][i][j]+dq_round)>>dq_rshift; // dequantization
            }
#else
              img->m7[boff_x + i][boff_y + j] = (level*InvLevelScale8x8Luma_Intra[qp_rem][i][j]+dq_round)>>dq_rshift; // dequantization
#endif
            else
              img->m7[boff_x + i][boff_y + j] = (level*InvLevelScale8x8Luma_Inter[qp_rem][i][j]+dq_round)>>dq_rshift; // dequantization
          }
        }
      }
    }


  }





#ifdef GEOM_ADAPTIVE_TRANSFORM     
  int derived_transform_size(int mb64, int geo_mode)
  {
    int x, y, x0, y0, mb_x, mb_y, CurrentMbAddr_x, CurrentMbAddr_y;
    int mb_nr = img->current_mb_nr;
    int rho = geo_mode>>5, theta_idx = geo_mode & 31, segm;
    char **mask4x4;

    mask4x4 = geo_4x4mask[mb64+1][rho][theta_idx];

    CurrentMbAddr_x = mb_nr % img->PicWidthInMbs;
    CurrentMbAddr_y = mb_nr / img->PicWidthInMbs;


    if(mb64)
    {
      mb_x = CurrentMbAddr_x%4;
      mb_y = CurrentMbAddr_y%4;
    }
    else
    {
      mb_x = CurrentMbAddr_x%2;
      mb_y = CurrentMbAddr_y%2;
    }

    x0=mb_x<<2;
    y0=mb_y<<2;

    segm = mask4x4[y0][x0];
    if(segm != 2)
    {
      for (y=y0; y<y0+4; y++)
      {      
        for (x=x0; x<x0+4; x++)
        {  
          if(mask4x4[y][x] != segm)
            return 0; //4x4 transform
        }
      }
      return 2;//16x16 transform
    }
    else
      return 0;//4x4 transform
  }
#endif

  void readlumaCBPTransform_size(struct inp_par *inp, int BigBlock, int mb64)
  {
    int mb_nr = img->current_mb_nr;
    Macroblock *currMB = &img->mb_data[mb_nr];
    int cbp;
    SyntaxElement currSE;
    Slice *currSlice = img->currentSlice;
    DataPartition *dP;
    int *partMap = assignSE2partition[currSlice->dp_mode];
    int smb       = ((img->type==SP_SLICE) && IS_INTER (currMB)) || (img->type == SI_SLICE && currMB->mb_type == SI4MB);

    int intra     = IS_INTRA (currMB);
    int yuv = dec_picture->chroma_format_idc-1;

    int need_transform_size_flag;


    if(img->type==SP_SLICE  && currMB->mb_type!=I16MB )
      smb=1;

    //=====   C B P   =====
    //---------------------
    if (IS_OLDINTRA (currMB) || currMB->mb_type == SI4MB || currMB->mb_type == I8MB)   currSE.type = SE_CBP_INTRA;
    else                        currSE.type = SE_CBP_INTER;

    dP = &(currSlice->partArr[partMap[currSE.type]]);
    {
      currSE.reading = readCBP_CABAC_luma1bit;
      TRACE_STRING("coded_block_pattern");

      dP->readSyntaxElement(&currSE,img,inp,dP);
      currMB->cbp = cbp = currSE.value1;
    }

    //============= Transform size flag for INTER MBs =============
    //-------------------------------------------------------------
    need_transform_size_flag = (((currMB->mb_type >= 0 && currMB->mb_type <= 3)||
      (IS_DIRECT(currMB) && active_sps->direct_8x8_inference_flag) ||
      (currMB->NoMbPartLessThan8x8Flag))
      && currMB->mb_type != I8MB && currMB->mb_type != I4MB
      && (currMB->cbp&15)
      && img->Transform8x8Mode);


    if (need_transform_size_flag)
    {
      if (img->DisableSmallTransInLargeMB && BigBlock)
      {
#ifdef GEOM_ADAPTIVE_TRANSFORM
        if(mb64 && MB64.geo_mode>-1)
          currMB->luma_transform_size_8x8_flag = derived_transform_size(1, MB64.geo_mode);
        else if(!mb64 && MB32.geo_mode>-1)
          currMB->luma_transform_size_8x8_flag = derived_transform_size(0, MB32.geo_mode);
        else
#endif
          currMB->luma_transform_size_8x8_flag = 2;
        return;
      }

      currSE.type   =  SE_HEADER;
      dP = &(currSlice->partArr[partMap[SE_HEADER]]);
      currSE.reading = readMB_transform_size_flag_CABAC;
      TRACE_STRING("transform size 8x8 flag");
      {
        dP->readSyntaxElement(&currSE,img,inp,dP);
        currMB->luma_transform_size_8x8_flag = currSE.value1;
      }
    }                 
  }

  int read_chroma_cbp(struct inp_par *inp)
  {
    int mb_nr = img->current_mb_nr;
    Macroblock *currMB = &img->mb_data[mb_nr];
    int cbp;
    SyntaxElement currSE;
    Slice *currSlice = img->currentSlice;
    DataPartition *dP;
    int *partMap = assignSE2partition[currSlice->dp_mode];
    int smb       = ((img->type==SP_SLICE) && IS_INTER (currMB)) || (img->type == SI_SLICE && currMB->mb_type == SI4MB);

    int intra     = IS_INTRA (currMB);
    int yuv = dec_picture->chroma_format_idc-1;

    if(img->type==SP_SLICE  && currMB->mb_type!=I16MB )
      smb=1;

    //=====   C B P   =====
    //---------------------
    if (IS_OLDINTRA (currMB) || currMB->mb_type == SI4MB || currMB->mb_type == I8MB)   currSE.type = SE_CBP_INTRA;
    else                        currSE.type = SE_CBP_INTER;

    dP = &(currSlice->partArr[partMap[currSE.type]]);
    currSE.reading = readCBP_CABAC_chroma;
    TRACE_STRING("coded_block_pattern");

    dP->readSyntaxElement(&currSE,img,inp,dP);
    cbp = currSE.value1;

    return cbp;
  }




  /*!
  ************************************************************************
  * \brief
  *    Get coded block pattern and coefficients (run/level)
  *    from the NAL
  ************************************************************************
  */
  void readCBPandCoeffsFromNAL(struct img_par *img,struct inp_par *inp, int mb32, int mb64, int mbcount)
  {
    int i,j,k;
    int level;
    int mb_nr = img->current_mb_nr;
    int m2,jg2;// i1,j1;
    Macroblock *currMB = &img->mb_data[mb_nr];
    int cbp;
    SyntaxElement currSE;
    Slice *currSlice = img->currentSlice;
    DataPartition *dP;
    int *partMap = assignSE2partition[currSlice->dp_mode];
    int iii,jjj;
    int coef_ctr, i0, j0, b8;
    int ll;
    int block_x,block_y;
    int start_scan;
    int run, len;

    int qp_const;
    int qp_per    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)/6;
    int qp_rem    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)%6;
    int smb       = ((img->type==SP_SLICE) && IS_INTER (currMB)) || (img->type == SI_SLICE && currMB->mb_type == SI4MB);

    int uv;
    int qp_const_uv[2];
    int qp_per_uv[2];
    int qp_rem_uv[2];

    int intra     = IS_INTRA (currMB);
    int temp[4];

    int b4;
    int yuv = dec_picture->chroma_format_idc-1;
    int m5[4];
    int m6[4];

    int need_transform_size_flag;
    Boolean lossless_qpprime = ((img->qp + img->bitdepth_luma_qp_scale)==0 && img->lossless_qpprime_flag==1);

    // Residue Color Transform
    Boolean residual_transform_dc = ((img->residue_transform_flag==1) && (IS_OLDINTRA(currMB)||currMB->mb_type==I8MB) );
#ifdef USE_INTRA_MDDT
    int R[6] = {40, 45, 50, 57, 63, 71};
#endif

    if(img->type==SP_SLICE  && currMB->mb_type!=I16MB )
      smb=1;

    // QPI
    //init constants for every chroma qp offset
    if (dec_picture->chroma_format_idc != YUV400)
    {
      for (i=0; i<2; i++)
      {
        qp_per_uv[i] = (currMB->qpc[i] + img->bitdepth_chroma_qp_scale)/6;
        qp_rem_uv[i] = (currMB->qpc[i] + img->bitdepth_chroma_qp_scale)%6;
      }
    }

    // read CBP if not new intra mode
    if (!IS_NEWINTRA (currMB))
    {
      int luma_cbp;
      //need_transform_size_flag is different from the original defination

      need_transform_size_flag = ((mb32 || (currMB->mb_type >= 0 && currMB->mb_type <= 3)||
        (IS_DIRECT(currMB) && active_sps->direct_8x8_inference_flag) ||
        (currMB->NoMbPartLessThan8x8Flag))
        && currMB->mb_type != I8MB && currMB->mb_type != I4MB
        //&& (currMB->cbp&15) - !!!
        && img->Transform8x8Mode);

      if(need_transform_size_flag)
      {
        readlumaCBPTransform_size(inp, mb32, mb64);
        luma_cbp = currMB->cbp;
      }

      if(need_transform_size_flag &&
        (luma_cbp==0 || (currMB->luma_transform_size_8x8_flag == 2
#ifdef GEOM
        && (mb32 || currMB->mb_type == 1 || currMB->geo_mode != -1 || (IS_DIRECT(currMB) && active_sps->direct_8x8_inference_flag)))))
#else
        && (mb32 || currMB->mb_type == 1 || (IS_DIRECT(currMB) && active_sps->direct_8x8_inference_flag)))))
#endif
      {
        int cr_cbp;
        cr_cbp = read_chroma_cbp(inp);
        currMB->cbp = cbp = (luma_cbp == 1?15:0) + cr_cbp;
        if (!cbp)
        {
          last_dquant=0;
        }
      }
      else
      {
        //=====   C B P   =====
        //---------------------
        if (IS_OLDINTRA (currMB) || currMB->mb_type == SI4MB || currMB->mb_type == I8MB)   currSE.type = SE_CBP_INTRA;
        else                        currSE.type = SE_CBP_INTER;

        dP = &(currSlice->partArr[partMap[currSE.type]]);
        currSE.value2 = mb32;
        currSE.reading = readCBP_CABAC;

        TRACE_STRING("coded_block_pattern");

        dP->readSyntaxElement(&currSE,img,inp,dP);
        currMB->cbp = cbp = currSE.value1;
      }

      //=====   DQUANT   =====
      //----------------------
      // Delta quant only if nonzero coeffs
      delta_qp_sent=0;
      if ( ((!mb32 && !delta_qp_sent)||img->type==I_SLICE) && cbp !=0 )
      {
        if (IS_INTER (currMB))  currSE.type = SE_DELTA_QUANT_INTER;
        else                    currSE.type = SE_DELTA_QUANT_INTRA;

        dP = &(currSlice->partArr[partMap[currSE.type]]);
        currSE.reading= readDquant_CABAC; //gabi

        TRACE_STRING("mb_qp_delta");

        dP->readSyntaxElement(&currSE,img,inp,dP);
        currMB->delta_quant = currSE.value1;

        if ((currMB->delta_quant < -(26 + img->bitdepth_luma_qp_scale/2)) || (currMB->delta_quant > (25 + img->bitdepth_luma_qp_scale/2)))
          error ("mb_qp_delta is out of range", 500);

        img->qp= ((img->qp + currMB->delta_quant + 52 + 2*img->bitdepth_luma_qp_scale)%(52+img->bitdepth_luma_qp_scale)) -
          img->bitdepth_luma_qp_scale;
      }
    }
    else
    {
      cbp = currMB->cbp;
    }

    for (i=0;i<BLOCK_SIZE;i++)
      for (j=0;j<BLOCK_SIZE;j++)
        for(iii=0;iii<BLOCK_SIZE;iii++)
          for(jjj=0;jjj<BLOCK_SIZE;jjj++)
            img->cof[i][j][iii][jjj]=0;// reset luma coeffs


    if (IS_NEWINTRA (currMB)) // read DC coeffs for new intra modes
    {
      delta_qp_sent=0;
      assert(mb32==0); 
      if ( (mb32 && !mbcount) || (!mb32 && !delta_qp_sent) || img->type==I_SLICE ) 
      {
        currSE.type = SE_DELTA_QUANT_INTRA;
        dP = &(currSlice->partArr[partMap[currSE.type]]);
        currSE.reading= readDquant_CABAC;
        dP->readSyntaxElement(&currSE,img,inp,dP);
        currMB->delta_quant = currSE.value1;

        if ((currMB->delta_quant < -(26 + img->bitdepth_luma_qp_scale/2)) || (currMB->delta_quant > (25 + img->bitdepth_luma_qp_scale/2)))
          error ("mb_qp_delta is out of range", 500);

        img->qp= ((img->qp + currMB->delta_quant + 52 + 2*img->bitdepth_luma_qp_scale)%(52+img->bitdepth_luma_qp_scale)) -
          img->bitdepth_luma_qp_scale;
      }
      for (i=0;i<BLOCK_SIZE;i++)
        for (j=0;j<BLOCK_SIZE;j++)
          img->ipredmode[img->block_x+i][img->block_y+j]=DC_PRED;

      {  // CABAC
#ifdef USE_INTRA_MDDT
        if(img->UseIntraMDDT)
        {
          for(k = 0; k < 16; k++)
            memset(img->cof16x16[k], 0, sizeof(int)*16);
          if(currMB->cbp & 0xf)
          {
            currSE.context      = LUMA_16x16;
            currSE.type         = SE_LUM_AC_INTRA;
            dP = &(currSlice->partArr[partMap[currSE.type]]);
            img->is_intra_block = 1;

            currSE.reading = readRunLevel16x16_CABAC;

            qp_per    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)/6;
            qp_rem    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)%6;

            coef_ctr=-1;
            level = 1;                            // just to get inside the loop
            for(k=0;(k<257) && (level!=0);k++)
            {
#if TRACE
              snprintf(currSE.tracestring, TRACESTRING_SIZE, "luma 16x16 ");
#endif
              dP->readSyntaxElement(&currSE,img,inp,dP);
              level = currSE.value1;
              run   = currSE.value2;
              len   = currSE.len;

              if (level != 0)                     // leave if len=1
              {
                coef_ctr=coef_ctr+run+1;
                i0=img->scanOrder16x16[currMB->i16mode][coef_ctr][0];
                j0=img->scanOrder16x16[currMB->i16mode][coef_ctr][1];
                img->cof16x16[j0][i0]=level*(R[qp_rem]<<qp_per);// add new intra DC coeff
              }
            }
          }
        }
        else
        {
#endif

          currSE.type = SE_LUM_DC_INTRA;
          dP = &(currSlice->partArr[partMap[currSE.type]]);

          currSE.context      = LUMA_16DC;
          currSE.type         = SE_LUM_DC_INTRA;
          img->is_intra_block = 1;

          //if (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag)
          //{
          //  currSE.mapping = linfo_levrun_inter;
          //}
          //else
          //{
          currSE.reading = readRunLevel_CABAC;
          //}



          coef_ctr=-1;
          level = 1;                            // just to get inside the loop
          for(k=0;(k<17) && (level!=0);k++)
          {
#if TRACE
            snprintf(currSE.tracestring, TRACESTRING_SIZE, "DC luma 16x16 ");
#endif
            dP->readSyntaxElement(&currSE,img,inp,dP);
            level = currSE.value1;
            run   = currSE.value2;
            len   = currSE.len;

            if (level != 0)                     // leave if len=1
            {
              coef_ctr=coef_ctr+run+1;

              i0=SNGL_SCAN[coef_ctr][0];
              j0=SNGL_SCAN[coef_ctr][1];

              img->cof[i0][j0][0][0]=level;// add new intra DC coeff
            }
          }
#ifdef USE_INTRA_MDDT
        }
#endif
      }
#ifdef USE_INTRA_MDDT
      if(active_pps->entropy_coding_mode_flag == UVLC || !img->UseIntraMDDT)
#endif
        if(!lossless_qpprime)
          itrans_2(img);// transform new intra DC
    }

    currMB->qp = img->qp;

    set_chroma_qp(currMB);

    qp_per    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)/6;
    qp_rem    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)%6;
    qp_const  = 1<<(3-qp_per);

    //init constants for every chroma qp offset
    if (dec_picture->chroma_format_idc != YUV400)
    {
      for(i=0; i < 2; i++)
      {
        qp_per_uv[i] = (currMB->qpc[i] + img->bitdepth_chroma_qp_scale)/6;
        qp_rem_uv[i] = (currMB->qpc[i] + img->bitdepth_chroma_qp_scale)%6;
      }
    }


    if(currMB->luma_transform_size_8x8_flag == 2)
    {
      assert(active_pps->entropy_coding_mode_flag == CABAC);
      readLumaCoeffBigBlock_CABAC(img, inp, mb32);
    }
    else

#ifdef USE_INTRA_MDDT
      if(!IS_NEWINTRA(currMB) || active_pps->entropy_coding_mode_flag == UVLC || !img->UseIntraMDDT)
#endif
        // luma coefficients
        for (block_y=0; block_y < 4; block_y += 2) /* all modes */
        {
          for (block_x=0; block_x < 4; block_x += 2)
          {

            b8 = 2*(block_y/2) + block_x/2;
#ifdef USE_INTRA_MDDT
            if (img->UseIntraMDDT && (cbp & (1<<b8)))
            {
              if(currMB->mb_type == I8MB)
              {
                int ipmode8x8 = img->ipredmode[(img->mb_x)*4 + block_x][(img->mb_y)*4 + block_y];
                img->update8x8Count[ipmode8x8]++;
              }
              else if(currMB->mb_type == I4MB)
              {
                for (j=block_y; j < block_y+2; j++)
                {
                  for (i=block_x; i < block_x+2; i++)
                  {
                    int ipmode4x4 = img->ipredmode[(img->mb_x)*4 + i][(img->mb_y)*4 + j];
                    img->update4x4Count[ipmode4x4]++;
                  }
                }
              }
            }
#endif

            { 
              if(currMB->luma_transform_size_8x8_flag)
                readLumaCoeff8x8_CABAC(img, inp, b8); //======= 8x8 trannsform size & CABAC ========
              else
              {
                //======= Other Modes & CABAC ========
                //------------------------------------
                for (j=block_y; j < block_y+2; j++)
                {
                  for (i=block_x; i < block_x+2; i++)
                  {
#ifdef USE_INTRA_MDDT
                    int img_block_x = (img->mb_x)*4 + i; 
                    int img_block_y = (img->mb_y)*4 + j; 
                    int ipmode = img->ipredmode[img_block_x][img_block_y];
#endif 
                    if (IS_NEWINTRA (currMB))   start_scan = 1; // skip DC coeff
                    else                        start_scan = 0; // take all coeffs

                    img->subblock_x = i; // position for coeff_count ctx
                    img->subblock_y = j; // position for coeff_count ctx

                    if (cbp & (1<<b8))  // are there any coeff in current block at all
                    {
                      coef_ctr = start_scan-1;
                      level    = 1;

                      for(k=start_scan;(k<17) && (level!=0);k++)
                      {
                        /*
                        * make distinction between INTRA and INTER coded
                        * luminance coefficients
                        */
                        currSE.context      = (IS_NEWINTRA(currMB) ? LUMA_16AC : LUMA_4x4);
                        currSE.type         = (IS_INTRA(currMB) ?
                          (k==0 ? SE_LUM_DC_INTRA : SE_LUM_AC_INTRA) :
                          (k==0 ? SE_LUM_DC_INTER : SE_LUM_AC_INTER));
                        img->is_intra_block = IS_INTRA(currMB);

                        dP = &(currSlice->partArr[partMap[currSE.type]]);
                        currSE.reading = readRunLevel_CABAC;
                        dP->readSyntaxElement(&currSE,img,inp,dP);

                        level = currSE.value1;
                        run   = currSE.value2;
                        len   = currSE.len;

                        if (level != 0)    /* leave if len=1 */
                        {
                          coef_ctr             += run+1;

                          i0=SNGL_SCAN[coef_ctr][0];
                          j0=SNGL_SCAN[coef_ctr][1];

                          currMB->cbp_blk      |= (int64)1 << ((j<<2) + i) ;

                          if(lossless_qpprime)
                          {
                            img->cof[i][j][i0][j0]= level;
                          }
                          else if(qp_per<4)
                          {
#ifdef USE_INTRA_MDDT
                            if(currMB->mb_type == I4MB && img->UseIntraMDDT)
                            {
                              i0 = img->scanOrder4x4[ipmode][coef_ctr][0];
                              j0 = img->scanOrder4x4[ipmode][coef_ctr][1];
                              if(ipmode == 2)
                                img->cof[i][j][i0][j0]= (level*InvLevelScale4x4Luma_Intra[qp_rem][i0][j0]+qp_const)>>(4-qp_per);
                              else 
                                img->cof[i][j][i0][j0]= level*(R[qp_rem]<<qp_per);
                              img->scanStats4x4[ipmode][coef_ctr]++;
                            }
                            else 
#endif
                              if(intra == 1)
                                img->cof[i][j][i0][j0]= (level*InvLevelScale4x4Luma_Intra[qp_rem][i0][j0]+qp_const)>>(4-qp_per);
                              else
                                img->cof[i][j][i0][j0]= (level*InvLevelScale4x4Luma_Inter[qp_rem][i0][j0]+qp_const)>>(4-qp_per);
                          }
                          else
                          {
#ifdef USE_INTRA_MDDT
                            if( currMB->mb_type == I4MB && img->UseIntraMDDT)
                            {
                              i0 = img->scanOrder4x4[ipmode][coef_ctr][0];
                              j0 = img->scanOrder4x4[ipmode][coef_ctr][1];
                              if(ipmode == 2)
                                img->cof[i][j][i0][j0]= (level*InvLevelScale4x4Luma_Intra[qp_rem][i0][j0])<<(qp_per-4);
                              else
                                img->cof[i][j][i0][j0]= level*(R[qp_rem]<<qp_per);
                              img->scanStats4x4[ipmode][coef_ctr]++;
                            }
                            else 
#endif
                              if(intra == 1)
                                img->cof[i][j][i0][j0]= (level*InvLevelScale4x4Luma_Intra[qp_rem][i0][j0])<<(qp_per-4);
                              else
                                img->cof[i][j][i0][j0]= (level*InvLevelScale4x4Luma_Inter[qp_rem][i0][j0])<<(qp_per-4);
                          }

                        }
                      }
                    }
                  }
                }
              }
            } 
          }
        }

        if (dec_picture->chroma_format_idc != YUV400)
        {
          for (j=4;j<(4+img->num_blk8x8_uv);j++) // reset all chroma coeffs before read
            for (i=0;i<4;i++)
              for (iii=0;iii<4;iii++)
                for (jjj=0;jjj<4;jjj++)
                  img->cof[i][j][iii][jjj]=0;

          m2 =img->mb_x*2;
          jg2=img->mb_y*2;

          qp_const_uv[0] = 1<<(3-qp_per_uv[0]);
          qp_const_uv[1] = 1<<(3-qp_per_uv[1]);


          //========================== CHROMA DC ============================
          //-----------------------------------------------------------------
          // chroma DC coeff
          if(cbp>15)
          {
#ifdef VLC_CBP_BIT
            int cbpBits;
            if (active_pps->entropy_coding_mode_flag == UVLC)
            {
              dP = &(currSlice->partArr[partMap[SE_CHR_DC_INTRA]]);
              readCbpBlockBit(CHROMA_DC, 0, 0, dP, &cbpBits);
            }
#endif // VLC_CBP_BIT
            for (ll=0;ll<3;ll+=2)
            {
              uv = ll>>1;

              if (dec_picture->chroma_format_idc == YUV420)
              {
                //===================== CHROMA DC YUV420 ======================
                for (i=0;i<4;i++)
                  img->cofu[i]=0;

                {
                  coef_ctr=-1;
                  level=1;
                  for(k=0;(k<(img->num_cdc_coeff+1))&&(level!=0);k++)
                  {
                    currSE.context      = CHROMA_DC;
                    currSE.type         = (IS_INTRA(currMB) ? SE_CHR_DC_INTRA : SE_CHR_DC_INTER);
                    img->is_intra_block =  IS_INTRA(currMB);
                    img->is_v_block     = ll;

                    dP = &(currSlice->partArr[partMap[currSE.type]]);
                    currSE.reading = readRunLevel_CABAC;
                    dP->readSyntaxElement(&currSE,img,inp,dP);
                    level = currSE.value1;
                    run = currSE.value2;
                    len = currSE.len;
                    if (level != 0)
                    {
                      currMB->cbp_blk |= 0xf0000 << (ll<<1) ;
                      coef_ctr=coef_ctr+run+1;
                      // Bug: img->cofu has only 4 entries, hence coef_ctr MUST be <4 (which is
                      // caught by the assert().  If it is bigger than 4, it starts patching the
                      // img->predmode pointer, which leads to bugs later on.
                      //
                      // This assert() should be left in the code, because it captures a very likely
                      // bug early when testing in error prone environments (or when testing NAL
                      // functionality).
                      assert (coef_ctr < img->num_cdc_coeff);
                      img->cofu[coef_ctr]=level;
                    }
                  }
                }

                if (smb // check to see if MB type is SPred or SIntra4x4 
                  || lossless_qpprime)
                {
                  img->cof[0+ll][4][0][0]=img->cofu[0];   img->cof[1+ll][4][0][0]=img->cofu[1];
                  img->cof[0+ll][5][0][0]=img->cofu[2];   img->cof[1+ll][5][0][0]=img->cofu[3];
                }
                else
                {
                  temp[0]=(img->cofu[0]+img->cofu[1]+img->cofu[2]+img->cofu[3]);
                  temp[1]=(img->cofu[0]-img->cofu[1]+img->cofu[2]-img->cofu[3]);
                  temp[2]=(img->cofu[0]+img->cofu[1]-img->cofu[2]-img->cofu[3]);
                  temp[3]=(img->cofu[0]-img->cofu[1]-img->cofu[2]+img->cofu[3]);

                  for (i=0;i<img->num_cdc_coeff;i++)
                  {
                    if(qp_per_uv[uv]<5)
                    {
                      if(intra == 1)
                        temp[i]=(temp[i]*InvLevelScale4x4Chroma_Intra[uv][qp_rem_uv[uv]][0][0])>>(5-qp_per_uv[uv]);
                      else
                        temp[i]=(temp[i]*InvLevelScale4x4Chroma_Inter[uv][qp_rem_uv[uv]][0][0])>>(5-qp_per_uv[uv]);
                    }
                    else
                    {
                      if(intra == 1)
                        temp[i]=(temp[i]*InvLevelScale4x4Chroma_Intra[uv][qp_rem_uv[uv]][0][0])<<(qp_per_uv[uv]-5);
                      else
                        temp[i]=(temp[i]*InvLevelScale4x4Chroma_Inter[uv][qp_rem_uv[uv]][0][0])<<(qp_per_uv[uv]-5);
                    }
                  }
                  img->cof[0+ll][4][0][0]=temp[0];
                  img->cof[1+ll][4][0][0]=temp[1];
                  img->cof[0+ll][5][0][0]=temp[2];
                  img->cof[1+ll][5][0][0]=temp[3];
                }
              }
              else if (dec_picture->chroma_format_idc == YUV422)
              {
                int i,j,j1;
                int uv_idx = ll;
                int m3[2][4] = {{0,0,0,0},{0,0,0,0}};
                int m4[2][4] = {{0,0,0,0},{0,0,0,0}};
                int qp_per_uv_dc = (currMB->qpc[uv] + 3 + img->bitdepth_chroma_qp_scale)/6;       //for YUV422 only
                int qp_rem_uv_dc = (currMB->qpc[uv] + 3 + img->bitdepth_chroma_qp_scale)%6;       //for YUV422 only

                //===================== CHROMA DC YUV422 ======================
                //if (active_pps->entropy_coding_mode_flag == UVLC)
                //{
                //  readCoeff4x4_CAVLC(img, inp, CHROMA_DC, 0, 0,
                //    levarr, runarr, &numcoeff);
                //  coef_ctr=-1;
                //  level=1;
                //  for(k = 0; k < numcoeff; k++)
                //  {
                //    if (levarr[k] != 0)
                //    {
                //      currMB->cbp_blk |= ((int64)0xff0000) << (ll<<2);
                //      coef_ctr=coef_ctr+runarr[k]+1;
                //      i0=SCAN_YUV422[coef_ctr][0];
                //      j0=SCAN_YUV422[coef_ctr][1];
                //      
                //      m3[i0][j0]=levarr[k];
                //    }
                //  }
                //}
                //else
                {
                  coef_ctr=-1;
                  level=1;
                  for(k=0;(k<9)&&(level!=0);k++)
                  {
                    currSE.context      = CHROMA_DC_2x4;
                    currSE.type         = (IS_INTRA(currMB) ? SE_CHR_DC_INTRA : SE_CHR_DC_INTER);
                    img->is_intra_block =  IS_INTRA(currMB);
                    img->is_v_block     = ll;

#if TRACE
                    snprintf(currSE.tracestring, TRACESTRING_SIZE, "2x4 DC Chroma ");
#endif
                    dP = &(currSlice->partArr[partMap[currSE.type]]);

                    //if (active_pps->entropy_coding_mode_flag == UVLC || dP->bitstream->ei_flag)
                    //  currSE.mapping = linfo_levrun_c2x2;
                    //else
                    currSE.reading = readRunLevel_CABAC;

                    dP->readSyntaxElement(&currSE,img,inp,dP);
                    level = currSE.value1;
                    run = currSE.value2;
                    len = currSE.len;
                    if (level != 0)
                    {
                      currMB->cbp_blk |= ((int64)0xff0000) << (ll<<2) ;
                      coef_ctr=coef_ctr+run+1;
                      assert (coef_ctr < img->num_cdc_coeff);
                      i0=SCAN_YUV422[coef_ctr][0];
                      j0=SCAN_YUV422[coef_ctr][1];

                      m3[i0][j0]=level;
                    }
                  }
                }
                // inverse CHROMA DC YUV422 transform
                // horizontal
                if(!lossless_qpprime)
                {
                  m4[0][0] = m3[0][0] + m3[1][0];
                  m4[0][1] = m3[0][1] + m3[1][1];
                  m4[0][2] = m3[0][2] + m3[1][2];
                  m4[0][3] = m3[0][3] + m3[1][3];

                  m4[1][0] = m3[0][0] - m3[1][0];
                  m4[1][1] = m3[0][1] - m3[1][1];
                  m4[1][2] = m3[0][2] - m3[1][2];
                  m4[1][3] = m3[0][3] - m3[1][3];
                }
                else
                {
                  for(i=0;i<2;i++)
                    for(j=0;j<4;j++)
                      img->cof[i+uv_idx][j+4][0][0]=m3[i][j];
                }

                // vertical
                for (i=0;i<2 && !lossless_qpprime;i++)
                {
                  for (j=0; j < 4;j++)    //TODO: remove m5 with m4
                    m5[j]=m4[i][j]; 

                  m6[0]=m5[0]+m5[2];
                  m6[1]=m5[0]-m5[2];
                  m6[2]=m5[1]-m5[3];
                  m6[3]=m5[1]+m5[3];

                  for (j=0;j<2;j++)
                  {
                    j1=3-j;
                    if(qp_per_uv_dc<4)
                    {
                      if(intra == 1)
                      {
                        img->cof[i+uv_idx][j +4][0][0]=((((m6[j]+m6[j1])*InvLevelScale4x4Chroma_Intra[uv][qp_rem_uv_dc][0][0]+(1<<(3-qp_per_uv_dc)))>>(4-qp_per_uv_dc))+2)>>2;
                        img->cof[i+uv_idx][j1+4][0][0]=((((m6[j]-m6[j1])*InvLevelScale4x4Chroma_Intra[uv][qp_rem_uv_dc][0][0]+(1<<(3-qp_per_uv_dc)))>>(4-qp_per_uv_dc))+2)>>2;
                      }
                      else
                      {
                        img->cof[i+uv_idx][j +4][0][0]=((((m6[j]+m6[j1])*InvLevelScale4x4Chroma_Inter[uv][qp_rem_uv_dc][0][0]+(1<<(3-qp_per_uv_dc)))>>(4-qp_per_uv_dc))+2)>>2;
                        img->cof[i+uv_idx][j1+4][0][0]=((((m6[j]-m6[j1])*InvLevelScale4x4Chroma_Inter[uv][qp_rem_uv_dc][0][0]+(1<<(3-qp_per_uv_dc)))>>(4-qp_per_uv_dc))+2)>>2;
                      }
                    }
                    else
                    {
                      if(intra == 1)
                      {
                        img->cof[i+uv_idx][j +4][0][0]=((((m6[j]+m6[j1])*InvLevelScale4x4Chroma_Intra[uv][qp_rem_uv_dc][0][0])<<(qp_per_uv_dc-4))+2)>>2;
                        img->cof[i+uv_idx][j1+4][0][0]=((((m6[j]-m6[j1])*InvLevelScale4x4Chroma_Intra[uv][qp_rem_uv_dc][0][0])<<(qp_per_uv_dc-4))+2)>>2;
                      }
                      else
                      {
                        img->cof[i+uv_idx][j +4][0][0]=((((m6[j]+m6[j1])*InvLevelScale4x4Chroma_Inter[uv][qp_rem_uv_dc][0][0])<<(qp_per_uv_dc-4))+2)>>2;
                        img->cof[i+uv_idx][j1+4][0][0]=((((m6[j]-m6[j1])*InvLevelScale4x4Chroma_Inter[uv][qp_rem_uv_dc][0][0])<<(qp_per_uv_dc-4))+2)>>2;
                      }
                    }
                  }//for (j=0;j<2;j++)
                }//for (i=0;i<2;i++)
              }//else if (dec_picture->chroma_format_idc == YUV422)
              else
              {
                //===================== CHROMA DC YUV444 ======================
                int i,j,i1,j1;
                int uv_idx = 4 + (ll<<1);

                {
                  coef_ctr=-1;
                  level=1;
                  for(k=0;(k<17)&&(level!=0);k++)
                  {
                    currSE.context      = CHROMA_DC_4x4;
                    currSE.type         = (IS_INTRA(currMB) ? SE_CHR_DC_INTRA : SE_CHR_DC_INTER);
                    img->is_intra_block =  IS_INTRA(currMB);
                    img->is_v_block     = ll;

                    dP = &(currSlice->partArr[partMap[currSE.type]]);
                    currSE.reading = readRunLevel_CABAC;

                    dP->readSyntaxElement(&currSE,img,inp,dP);
                    level = currSE.value1;
                    run = currSE.value2;
                    len = currSE.len;
                    if (level != 0)
                    {
                      currMB->cbp_blk |= ((int64)0xffff0000) << (ll<<3) ;
                      coef_ctr=coef_ctr+run+1;
                      assert (coef_ctr < img->num_cdc_coeff);
                      i0=SNGL_SCAN[coef_ctr][0];
                      j0=SNGL_SCAN[coef_ctr][1];

                      img->cof[i0][j0+uv_idx][0][0]=level;
                    }
                  }
                }

                // Residue Color Transform
                if(!residual_transform_dc)
                {
                  // inverse CHROMA DC YUV444 transform
                  // horizontal
                  for (j=uv_idx; (j < 4+uv_idx) && !lossless_qpprime;j++)
                  {
                    for (i=0;i<4;i++)
                      m5[i]=img->cof[i][j][0][0];

                    m6[0]=m5[0]+m5[2];
                    m6[1]=m5[0]-m5[2];
                    m6[2]=m5[1]-m5[3];
                    m6[3]=m5[1]+m5[3];

                    for (i=0;i<2;i++)
                    {
                      i1=3-i;
                      img->cof[i][j][0][0]= m6[i]+m6[i1];
                      img->cof[i1][j][0][0]=m6[i]-m6[i1];
                    }
                  }

                  // vertical
                  for (i=0;i<4 && !lossless_qpprime;i++)
                  {
                    for (j=0; j < 4;j++)
                      m5[j]=img->cof[i][j+uv_idx][0][0]; 

                    m6[0]=m5[0]+m5[2];
                    m6[1]=m5[0]-m5[2];
                    m6[2]=m5[1]-m5[3];
                    m6[3]=m5[1]+m5[3];

                    for (j=0;j<2;j++)
                    {
                      j1=3-j;
                      if(qp_per_uv[uv]<4)
                      {
                        if(intra == 1)
                        {
                          img->cof[i][j +uv_idx][0][0]=((((m6[j]+m6[j1])*InvLevelScale4x4Chroma_Intra[uv][qp_rem_uv[uv]][0][0]+(1<<(3-qp_per_uv[uv])))>>(4-qp_per_uv[uv]))+2)>>2;
                          img->cof[i][j1+uv_idx][0][0]=((((m6[j]-m6[j1])*InvLevelScale4x4Chroma_Intra[uv][qp_rem_uv[uv]][0][0]+(1<<(3-qp_per_uv[uv])))>>(4-qp_per_uv[uv]))+2)>>2;
                        } 
                        else 
                        {
                          img->cof[i][j +uv_idx][0][0]=((((m6[j]+m6[j1])*InvLevelScale4x4Chroma_Inter[uv][qp_rem_uv[uv]][0][0]+(1<<(3-qp_per_uv[uv])))>>(4-qp_per_uv[uv]))+2)>>2;
                          img->cof[i][j1+uv_idx][0][0]=((((m6[j]-m6[j1])*InvLevelScale4x4Chroma_Inter[uv][qp_rem_uv[uv]][0][0]+(1<<(3-qp_per_uv[uv])))>>(4-qp_per_uv[uv]))+2)>>2;
                        }
                      }
                      else
                      {
                        if(intra == 1)
                        {
                          img->cof[i][j +uv_idx][0][0]=((((m6[j]+m6[j1])*InvLevelScale4x4Chroma_Intra[uv][qp_rem_uv[uv]][0][0])<<(qp_per_uv[uv]-4))+2)>>2;
                          img->cof[i][j1+uv_idx][0][0]=((((m6[j]-m6[j1])*InvLevelScale4x4Chroma_Intra[uv][qp_rem_uv[uv]][0][0])<<(qp_per_uv[uv]-4))+2)>>2;
                        } 
                        else 
                        {
                          img->cof[i][j +uv_idx][0][0]=((((m6[j]+m6[j1])*InvLevelScale4x4Chroma_Inter[uv][qp_rem_uv[uv]][0][0])<<(qp_per_uv[uv]-4))+2)>>2;
                          img->cof[i][j1+uv_idx][0][0]=((((m6[j]-m6[j1])*InvLevelScale4x4Chroma_Inter[uv][qp_rem_uv[uv]][0][0])<<(qp_per_uv[uv]-4))+2)>>2;
                        }
                      }
                    } 
                  }//for (i=0;i<4;i++)
                } 
                else
                { //residual_transform_dc

                  for (i=0;i<4  && !lossless_qpprime ;i++)
                    for (j=0; j < 4;j++)
                    {
                      if(qp_per_uv[uv]<4)
                      {
                        if(intra == 1)
                        {
                          img->cof[i][j +uv_idx][0][0]=((((img->cof[i][j +uv_idx][0][0])*InvLevelScale4x4Chroma_Intra[uv][qp_rem_uv[uv]][0][0]+(1<<(3-qp_per_uv[uv])))>>(4-qp_per_uv[uv])));
                        } 
                        else 
                        {
                          img->cof[i][j +uv_idx][0][0]=((((img->cof[i][j +uv_idx][0][0])*InvLevelScale4x4Chroma_Inter[uv][qp_rem_uv[uv]][0][0]+(1<<(3-qp_per_uv[uv])))>>(4-qp_per_uv[uv])));
                        }
                      } 
                      else 
                      {
                        if(intra == 1) 
                        {
                          img->cof[i][j +uv_idx][0][0]=((((img->cof[i][j +uv_idx][0][0])*InvLevelScale4x4Chroma_Intra[uv][qp_rem_uv[uv]][0][0])<<(qp_per_uv[uv]-4)));
                        } 
                        else 
                        {
                          img->cof[i][j +uv_idx][0][0]=((((img->cof[i][j +uv_idx][0][0])*InvLevelScale4x4Chroma_Inter[uv][qp_rem_uv[uv]][0][0])<<(qp_per_uv[uv]-4)));
                        }
                      }
                    }
                } //residual_transform_dc
              }//else (dec_picture->chroma_format_idc == YUV444)
            }//for (ll=0;ll<3;ll+=2)
          }

          // chroma AC coeff, all zero fram start_scan
          if(active_pps->entropy_coding_mode_flag==UVLC)//tm)vlc
          {
            if (cbp<=31)
              for (j=4; j < (4 + img->num_blk8x8_uv); j++)
                for (i=0; i < 4; i++)
                  img->nz_coeff [img->current_mb_nr ][i][j]=0;
          }   

          //========================== CHROMA AC ============================
          //-----------------------------------------------------------------
          // chroma AC coeff, all zero fram start_scan
          if (cbp>31)
          {
            for (b8=0; b8 < img->num_blk8x8_uv; b8++)
            {
              for (b4=0; b4 < 4; b4++)
              {
                i = cofuv_blk_x[yuv][b8][b4];
                j = cofuv_blk_y[yuv][b8][b4];

                img->is_v_block = uv = (b8 > ((img->num_blk8x8_uv>>1) - 1 ));

                {
                  coef_ctr=0;
                  level=1;

                  img->subblock_y = subblk_offset_y[yuv][b8][b4]>>2;
                  img->subblock_x = subblk_offset_x[yuv][b8][b4]>>2;

                  currSE.context      = CHROMA_AC;
                  currSE.type         = (IS_INTRA(currMB) ? SE_CHR_AC_INTRA : SE_CHR_AC_INTER);
                  img->is_intra_block =  IS_INTRA(currMB);

                  for(k=0;(k<16)&&(level!=0);k++)
                  {
                    dP = &(currSlice->partArr[partMap[currSE.type]]);
                    currSE.reading = readRunLevel_CABAC;
                    dP->readSyntaxElement(&currSE,img,inp,dP);
                    level = currSE.value1;
                    run = currSE.value2;
                    len = currSE.len;

                    if (level != 0)
                    {
                      currMB->cbp_blk |= ((int64)1) << cbp_blk_chroma[b8][b4];
                      coef_ctr=coef_ctr+run+1;

                      i0=SNGL_SCAN[coef_ctr][0];
                      j0=SNGL_SCAN[coef_ctr][1];

                      if(lossless_qpprime)
                      {
                        img->cof[i][j][i0][j0]=level;
                      }
                      else if(qp_per_uv[uv]<4)
                      {
                        if(intra == 1)
                          img->cof[i][j][i0][j0]=(level*InvLevelScale4x4Chroma_Intra[img->is_v_block][qp_rem_uv[uv]][i0][j0]+qp_const_uv[uv])>>(4-qp_per_uv[uv]);
                        else
                          img->cof[i][j][i0][j0]=(level*InvLevelScale4x4Chroma_Inter[img->is_v_block][qp_rem_uv[uv]][i0][j0]+qp_const_uv[uv])>>(4-qp_per_uv[uv]);
                      }
                      else
                      {
                        if(intra == 1)
                          img->cof[i][j][i0][j0]=(level*InvLevelScale4x4Chroma_Intra[img->is_v_block][qp_rem_uv[uv]][i0][j0])<<(qp_per_uv[uv]-4);
                        else
                          img->cof[i][j][i0][j0]=(level*InvLevelScale4x4Chroma_Inter[img->is_v_block][qp_rem_uv[uv]][i0][j0])<<(qp_per_uv[uv]-4);
                      }
                    }
                  } //for(k=0;(k<16)&&(level!=0);k++)
                } //else / if (active_pps->entropy_coding_mode_flag == UVLC)
              } //for (b4=0; b4 < 4; b4++)
            } //for (b8=0; b8 < img->num_blk8x8_uv; b8++)
          } //if (cbp>31)
        } //if (dec_picture->chroma_format_idc != YUV400)
#ifdef USE_INTRA_MDDT
        if(img->UseIntraMDDT)
        {
          int i;
          for(i = 0; i < 9; i++)
          {
            if(img->update4x4Count[i] >= img->update4x4Thres[i])
            {
              img->update4x4[i] = 1;
              img->update4x4Count[i] = 0;
            }
            else img->update4x4[i] = 0;
            if(img->update8x8Count[i] >= img->update8x8Thres[i])
            {
              img->update8x8[i] = 1;
              img->update8x8Count[i] = 0;
            }
            else img->update8x8[i] = 0;
          }
          updateScanOrder(0);
          normalizeScanStats();
        }
#endif
  }

  void decode_ipcm_mb(struct img_par *img)
  {
    int i,j;

    Macroblock *currMb = &img->mb_data[img->current_mb_nr];    

    //Copy coefficients to decoded picture buffer
    //IPCM coefficients are stored in img->cof which is set in function readIPCMcoeffsFromNAL()

    for(i=0;i<16;i++)
      for(j=0;j<16;j++)
        dec_picture->imgY[img->pix_y+i][img->pix_x+j]=img->cof[i/4][j/4][i%4][j%4];

    if (dec_picture->chroma_format_idc != YUV400)
    {
      for(i=0;i<img->mb_cr_size_y;i++)
        for(j=0;j<img->mb_cr_size_x;j++)
          dec_picture->imgUV[0][img->pix_c_y+i][img->pix_c_x+j]=img->cof[i/4][j/4+4][i%4][j%4];   //TODO-VG

      for(i=0;i<img->mb_cr_size_y;i++)
        for(j=0;j<img->mb_cr_size_x;j++)
          dec_picture->imgUV[1][img->pix_c_y+i][img->pix_c_x+j]=img->cof[i/4+2][j/4+4][i%4][j%4]; //TODO-VG
    }



    // for deblocking filter
    currMb->qp=0;
    set_chroma_qp(currMb);

    // for CAVLC: Set the nz_coeff to 16. 
    // These parameters are to be used in CAVLC decoding of neighbour blocks
    for(i=0;i<4;i++)
      for (j=0;j<(4 + img->num_blk8x8_uv);j++)
        img->nz_coeff[img->current_mb_nr][i][j]=16;


    // for CABAC decoding of MB skip flag 
    currMb->skip_flag = 0;

    //for deblocking filter CABAC
    currMb->cbp_blk=0xFFFF;

    //For CABAC decoding of Dquant
    last_dquant=0;
  }


  static int decode_one_component_i_slice(Macroblock *currMB, StorablePicture *dec_picture,int mb_ext_level, int mb_count)
  {
    switch(currMB->mb_type)
    {
    case I4MB:
      mb_pred_intra4x4(currMB);
      break;
    case I8MB:
      mb_pred_intra8x8(currMB);
      break;
    case I16MB:
      mb_pred_intra16x16(currMB);
      break;
    }
    intrapred_chroma(img, 0);
    perform_transform_chroma(currMB, dec_picture, 0);
    intrapred_chroma(img, 1);
    perform_transform_chroma(currMB, dec_picture, 1);

    return 1;
  }

  static int decode_one_component(Macroblock *currMB, StorablePicture *dec_picture,int mb_ext_level, int mb_count)
  {
    if(mb_ext_level==0)
    {
      switch(currMB->mb_type)
      {
      case 0:
      case 1:
        //Y
        perform_mc_luma(currMB, dec_picture, mb_ext_level, mb_count, currMB->b8pdir[0], 0, 0, MB_BLOCK_SIZE, MB_BLOCK_SIZE);
        perform_transform_luma(currMB, dec_picture, mb_ext_level);
        //U
        perform_mc_chroma_MxN(currMB, 0, 0, 8, 8, 0);
        perform_transform_chroma(currMB, dec_picture, 0);
        //V
        perform_mc_chroma_MxN(currMB, 0, 0, 8, 8, 1);
        perform_transform_chroma(currMB, dec_picture, 1);
        break;
      case 2: //16x8
        //Y
        perform_mc_luma(currMB, dec_picture, mb_ext_level, mb_count, currMB->b8pdir[0], 0, 0, MB_BLOCK_SIZE, 8);
        perform_mc_luma(currMB, dec_picture, mb_ext_level, mb_count, currMB->b8pdir[2], 0, 2, MB_BLOCK_SIZE, 8);
        perform_transform_luma(currMB, dec_picture, mb_ext_level);

        perform_mc_chroma_MxN(currMB, 0, 0, 8, 4, 0);
        perform_mc_chroma_MxN(currMB, 0, 1, 8, 4, 0);
        perform_transform_chroma(currMB, dec_picture, 0);

        perform_mc_chroma_MxN(currMB, 0, 0, 8, 4, 1);
        perform_mc_chroma_MxN(currMB, 0, 1, 8, 4, 1);
        perform_transform_chroma(currMB, dec_picture, 1);
        break;
      case 3: //8x16
        //Y
        perform_mc_luma(currMB, dec_picture, mb_ext_level, mb_count, currMB->b8pdir[0], 0, 0, 8, MB_BLOCK_SIZE);
        perform_mc_luma(currMB, dec_picture, mb_ext_level, mb_count, currMB->b8pdir[1], 2, 0, 8, MB_BLOCK_SIZE);
        perform_transform_luma(currMB, dec_picture, mb_ext_level);

        perform_mc_chroma_MxN(currMB, 0, 0, 4, 8, 0);
        perform_mc_chroma_MxN(currMB, 1, 0, 4, 8, 0);
        perform_transform_chroma(currMB, dec_picture, 0);

        perform_mc_chroma_MxN(currMB, 0, 0, 4, 8, 1);
        perform_mc_chroma_MxN(currMB, 1, 0, 4, 8, 1);
        perform_transform_chroma(currMB, dec_picture, 1);
        break;
      }
    }
    else
    {
      //Y
      perform_mc_luma(currMB, dec_picture, mb_ext_level, mb_count, currMB->b8pdir[0], 0, 0, MB_BLOCK_SIZE, MB_BLOCK_SIZE);
      perform_transform_luma(currMB, dec_picture, mb_ext_level);

      //U
      perform_mc_chroma_MxN(currMB, 0, 0, 8, 8, 0);
      perform_transform_chroma(currMB, dec_picture, 0);

      //V
      perform_mc_chroma_MxN(currMB, 0, 0, 8, 8, 1);
      perform_transform_chroma(currMB, dec_picture, 1);
    }
    return 1;
  }


  void mb_pred_intra4x4(Macroblock *currMB)
  {
    int i=0, j=0,k, j4=0,i4=0;  
    int ioff,joff;
    int block8x8;   // needed for ABT
    const byte decode_block_scan[16] = {0,1,4,5,2,3,6,7,8,9,12,13,10,11,14,15};

    for (block8x8 = 0; block8x8 < 4; block8x8++)
    {
      for (k = block8x8 * 4; k < block8x8 * 4 + 4; k ++)
      {
        i =  (decode_block_scan[k] & 3);
        j = ((decode_block_scan[k] >> 2) & 3);

        ioff = (i << 2);
        joff = (j << 2);
        i4   = img->block_x + i;
        j4   = img->block_y + j;

        intrapred(img,ioff,joff,i4,j4);
        if(img->UseIntraMDDT)
        {
          int ipmode = img->ipredmode[i4][j4];
          if(ipmode == 2)
            itrans(img, ioff, joff, i, j, 0);
          else
            itransklt_sep(img, ioff, joff, i, j, ipmode);
        }
        else
          itrans   (img,ioff,joff,i,j, 0);      // use DCT transform and make 4x4 block m7 from prediction block mpr

        move_data_to_dec_picture_transpose(BLOCK_SIZE, BLOCK_SIZE, i4*BLOCK_SIZE, j4*BLOCK_SIZE, 0, 0);
      }
    }
  }


  void mb_pred_intra16x16(Macroblock *currMB)
  {
    intrapred_luma_16x16(img, currMB->i16mode);
    itransklt16x16_sep_fast(img, currMB->i16mode);

    move_data_to_dec_picture_transpose(MB_BLOCK_SIZE, MB_BLOCK_SIZE, img->pix_x, img->pix_y, 0, 0);
  }

  void mb_pred_intra8x8(Macroblock *currMB)
  {
    int block8x8;   // needed for ABT

    for (block8x8 = 0; block8x8 < 4; block8x8++)
    {
      //=========== 8x8 BLOCK TYPE ============
      int ioff = (block8x8 & 0x01) << 3;
      int joff = (block8x8 >> 1  ) << 3;

      intrapred8x8(img, block8x8);
      if(img->UseIntraMDDT)
      {
        int img_block_x = (img->mb_x)*4 + 2*(block8x8%2);
        int img_block_y = (img->mb_y)*4 + 2*(block8x8/2);
        int ipmode = img->ipredmode[img_block_x][img_block_y];

        if(ipmode == 2)
          itrans8x8(img, ioff, joff);
        else
          itrans8x8klt_sep_fast(img, ioff, joff, ipmode);
      }
      else
        itrans8x8(img,ioff,joff); 

      //cannot move this out of loop as 8x8 prediction needs updated dec_picture...
      move_data_to_dec_picture_transpose(8, 8, img->pix_x + ioff, img->pix_y + joff, ioff, joff);
    }
  }

  void perform_mc_luma(Macroblock *currMB, StorablePicture *dec_picture, int mb_ext_level, int mb_count, int pred_dir, int i, int j, int block_size_x, int block_size_y)
  {
    int tmp_block[16][16];
    int tmp_blockbw[16][16];
    int ii=0,jj=0,i1=0,j1=0,j4=0,i4=0;
    int vec1_x=0,vec1_y=0,vec2_x=0,vec2_y=0;
    int ioff,joff;

    int bw_pred=0, fw_pred=0;
    int mv_mul;

    short ref_idx, fw_refframe=-1, bw_refframe=-1;
    int mv_mode; // = currMB->ref_frame;
    short fw_ref_idx=-1, bw_ref_idx=-1;

    short  *** mv_array, ***fw_mv_array, ***bw_mv_array;
    StorablePicture **list;
    char fw_rFrame=-1,bw_rFrame=-1;

#ifdef EIGHTH_PEL
    mv_mul = 4 << img->mv_res;
#else
    mv_mul=4;
#endif

    //following arrays are stored as img->mpr[x][y] and tmp_block[x][y]..1st x and then y--------- 

    mv_mode  = currMB->b8mode[(i>>1)+j]; //in the following mv_mode only differentiate skip, intra and the others
    //pred_dir = currMB->b8pdir[0];
    assert (pred_dir<=2);
    i4=img->block_x + i;
    j4=img->block_y + j;
    ioff = (i << 2);
    joff = (j << 2);  

    if (pred_dir != 2)
    {
      //===== FORWARD/BACKWARD PREDICTION =====
      fw_refframe = ref_idx  = dec_picture->ref_idx[LIST_0 + pred_dir][j4][i4];
      mv_array = dec_picture->mv[LIST_0 + pred_dir];
      list     = listX[0+pred_dir];
      vec1_x = i4*4*mv_mul + mv_array[j4][i4][0];
      vec1_y = j4*4*mv_mul + mv_array[j4][i4][1];

      getblock2DFilt(ref_idx, list, i4<<2, j4<<2, mv_array[j4][i4][0], mv_array[j4][i4][1],img->width, img->height, tmp_block, block_size_x,block_size_y);

      if (img->apply_weights)
      {
        for(ii=0;ii<block_size_x;ii++)
          for(jj=0;jj<block_size_y;jj++)  
            img->mpr[ii+ioff][jj+joff] = Clip1(((img->wp_weight[pred_dir][ref_idx][0] *  tmp_block[ii][jj]+ img->wp_round_luma) >>img->luma_log2_weight_denom)  + img->wp_offset[pred_dir][ref_idx][0] );
      }
      else
      {
        for(ii=0;ii<block_size_x;ii++)
          for(jj=0;jj<block_size_y;jj++)  
            img->mpr[ii+ioff][jj+joff] = tmp_block[ii][jj];
      }
    }
    else //bi-directional
    {
      assert(mv_mode != 0);
      //===== BI-DIRECTIONAL PREDICTION =====
      fw_mv_array = dec_picture->mv[LIST_0];
      bw_mv_array = dec_picture->mv[LIST_1];

      fw_refframe = dec_picture->ref_idx[LIST_0][j4][i4];
      bw_refframe = dec_picture->ref_idx[LIST_1][j4][i4];

      fw_ref_idx = fw_refframe;
      bw_ref_idx = bw_refframe;

      vec1_x = i4*4*mv_mul + fw_mv_array[j4][i4][0];
      vec2_x = i4*4*mv_mul + bw_mv_array[j4][i4][0];
      vec1_y = j4*4*mv_mul + fw_mv_array[j4][i4][1];
      vec2_y = j4*4*mv_mul + bw_mv_array[j4][i4][1];

      getblock2DFilt(fw_refframe, listX[0], i4<<2, j4<<2, fw_mv_array[j4][i4][0], fw_mv_array[j4][i4][1], img->width, img->height, tmp_block  , block_size_x,block_size_y);
      getblock2DFilt(bw_refframe, listX[1], i4<<2, j4<<2, bw_mv_array[j4][i4][0], bw_mv_array[j4][i4][1], img->width, img->height, tmp_blockbw, block_size_x,block_size_y);

      if(img->apply_weights)
      {
        int alpha_fw, alpha_bw;

        alpha_fw = img->wbp_weight[0][fw_ref_idx][bw_ref_idx][0];
        alpha_bw = img->wbp_weight[1][fw_ref_idx][bw_ref_idx][0];

        if(img->use_high_precision_flag != 0)
        {
          for(ii=0;ii<block_size_x;ii++)
            for(jj=0;jj<block_size_y;jj++)   
              img->mpr[ii+ioff][jj+joff] = (int)Clip1((alpha_fw * tmp_block[ii][jj] + alpha_bw * tmp_blockbw[ii][jj]  
            + ((img->wp_offset[0][fw_ref_idx][0] + img->wp_offset[1][bw_ref_idx][0] ) << img->luma_log2_weight_denom)
              + (1<<img->luma_log2_weight_denom)*(1- img->bipred_rounding_control)) >> (img->luma_log2_weight_denom+1)) ;
        }
        else
        {
          for(ii=0;ii<block_size_x;ii++)
            for(jj=0;jj<block_size_y;jj++)   
              img->mpr[ii+ioff][jj+joff] = (int)Clip1(((alpha_fw * tmp_block[ii][jj] + alpha_bw * tmp_blockbw[ii][jj]  
            + (1<<img->luma_log2_weight_denom)) >> (img->luma_log2_weight_denom+1)) 
              + ((img->wp_offset[0][fw_ref_idx][0] + img->wp_offset[1][bw_ref_idx][0] + 1) >>1));
        }
      }
      else
      {
        if(img->use_high_precision_flag != 0)
        {
          int round = 1 - img->bipred_rounding_control;
          for(ii=0;ii<block_size_x;ii++)
            for(jj=0;jj<block_size_y;jj++)     
              img->mpr[ii+ioff][jj+joff] = (tmp_block[ii][jj]+tmp_blockbw[ii][jj]+round)>>1;
        }
        else
        {
          for(ii=0;ii<block_size_x;ii++)
            for(jj=0;jj<block_size_y;jj++)    
              img->mpr[ii+ioff][jj+joff] = (tmp_block[ii][jj]+tmp_blockbw[ii][jj]+1)>>1;
        }
      }
    }       
  }

  void move_data_to_dec_picture_transpose(int size_x,int size_y, int dec_pic_ioff, int dec_pic_joff, int m7_ioff, int m7_joff)
  {
    imgpel *imgDec;
    imgpel *m7_pointer;
    int jj;
    for(jj=0;jj<size_y;jj++)
    {
      imgDec = &dec_picture->imgY[jj+dec_pic_joff][dec_pic_ioff];
      m7_pointer = &img->m7_transpose[jj+m7_joff][m7_ioff];
      memcpy(imgDec,m7_pointer,sizeof(imgpel)*size_x);
    }
  }
  void move_data_to_dec_picture_chroma_transpose(int size_x,int size_y, int dec_pic_ioff, int dec_pic_joff, int m7_ioff, int m7_joff, int uv)
  {
    imgpel *imgDec;
    int jj;

    for(jj=0;jj<size_y;jj++)
    {
      imgDec = &dec_picture->imgUV[uv][jj+dec_pic_joff][dec_pic_ioff];
      memcpy(imgDec,&img->m7_transpose[jj+m7_joff][m7_ioff],sizeof(imgpel)*size_x);
    }
  }

  void move_data_to_dec_picture(int size_x,int size_y, int dec_pic_ioff, int dec_pic_joff, int m7_ioff, int m7_joff)
  {
    imgpel *imgDec;
    int ii,jj;
    for(jj=0;jj<size_y;jj++)
    {
      imgDec = &dec_picture->imgY[jj+dec_pic_joff][dec_pic_ioff];
      for(ii=0;ii<size_x;ii++)
      {
        *imgDec++ = img->m7[ii+m7_ioff][jj+m7_joff];
      }
    }
  }
  void move_data_to_dec_picture_chroma(int size_x,int size_y, int dec_pic_ioff, int dec_pic_joff, int m7_ioff, int m7_joff, int uv)
  {
    imgpel *imgDec;
    int ii,jj;
    for(jj=0;jj<size_y;jj++)
    {
      imgDec = &dec_picture->imgUV[uv][jj+dec_pic_joff][dec_pic_ioff];
      for(ii=0;ii<size_x;ii++)
      {
        *imgDec++ = img->m7[ii+m7_ioff][jj+m7_joff];
      }
    }
  }

  void perform_transform_luma(Macroblock *currMB, StorablePicture *dec_picture, int mb_ext_level)
  {
    int block8x8,k,i,j,ioff,joff,i4,j4;
    const byte decode_block_scan[16] = {0,1,4,5,2,3,6,7,8,9,12,13,10,11,14,15};

    //================================================================
    for (block8x8=0; block8x8<4; block8x8++)
    {
      for (k = block8x8*4; k < block8x8*4+4; k ++)
      {
        if(currMB->luma_transform_size_8x8_flag == 0)
        { 
          i = (decode_block_scan[k] & 3);
          j = ((decode_block_scan[k] >> 2) & 3);
          i4=img->block_x + i;
          j4=img->block_y + j;
          ioff=i<<2;
          joff=j<<2;
          itrans   (img,ioff,joff,i,j, 0);      // use DCT transform and make 4x4 block m7 from prediction block mpr

          move_data_to_dec_picture_transpose(BLOCK_SIZE, BLOCK_SIZE, i4*BLOCK_SIZE, j4*BLOCK_SIZE, 0, 0);
        }
      }     //4x4 loop
      if(currMB->luma_transform_size_8x8_flag == 1)       
      {
        ioff = 8*(block8x8%2);
        joff = 8*(block8x8/2);
        itrans8x8(img,ioff,joff);      // use DCT transform and make 8x8 block m7 from prediction block mpr
      }
    }   //8x8 loop

    if(currMB->luma_transform_size_8x8_flag == 2) 
    {
#ifdef GEOM
      if(currMB->mb_type == 0 || currMB->mb_type == P16x16 || currMB->geo_mode != -1 || (mb_ext_level == 1 && MB32.mb_type != P8x8) || (mb_ext_level == 2 && MB64.mb_type != P8x8)) //MB64X64
#else
      if(currMB->mb_type == 0 || currMB->mb_type == P16x16 || (mb_ext_level == 1 && MB32.mb_type != P8x8) || (mb_ext_level == 2 && MB64.mb_type != P8x8)) //MB64X64
#endif
      {
        itrans16x16(img);
      }
      else if(currMB->mb_type == P16x8)
      {
        itrans16x8(img, 0);
        itrans16x8(img, 1);
      }
      else // currMB->mb_type == P8x16
      {
        assert(currMB->mb_type == P8x16);
        itrans8x16(img, 0);
        itrans8x16(img, 1);
      }
    }

    if(currMB->luma_transform_size_8x8_flag > 0)  
    {
      move_data_to_dec_picture_transpose(MB_BLOCK_SIZE, MB_BLOCK_SIZE, img->pix_x, img->pix_y, 0, 0);
    }
  }



  void perform_transform_chroma(Macroblock *currMB, StorablePicture *dec_picture, int uv)
  {
    int b8,b4,ioff,joff,i4,j4;
    int yuv = dec_picture->chroma_format_idc - 1;
    for (b8=0;b8<(img->num_blk8x8_uv/2);b8++)
    {
      for(b4=0;b4<4;b4++)
      {
        joff = subblk_offset_y[yuv][b8][b4];
        ioff = subblk_offset_x[yuv][b8][b4];
        j4=img->pix_c_y+joff;
        i4=img->pix_c_x+ioff;
        itrans(img,ioff,joff, cofuv_blk_x[yuv][b8+uv][b4], cofuv_blk_y[yuv][b8+uv][b4], 1);
      }
    }
    move_data_to_dec_picture_chroma_transpose(8,8,img->pix_c_x,img->pix_c_y,0,0,uv);
  }




  void prepare_B_DIRECT(Macroblock *currMB, int mb_ext_level, int mb_count, char *fw_rFrame,char *bw_rFrame, short *pmvfw, short *pmvbw)
  {
    if(img->type==B_SLICE &&( (mb_ext_level == 1 && MB32.mb_type == 0) || (mb_ext_level == 2 && MB64.mb_type == 0) ) )
    {
      if((mb_count==0 && mb_ext_level == 2) || ((mb_count&3)==0 && mb_ext_level == 1))
      {
        char fw_rFrameL, fw_rFrameU, fw_rFrameUL, fw_rFrameUR;
        char bw_rFrameL, bw_rFrameU, bw_rFrameUL, bw_rFrameUR;    

        PixelPos mb_left, mb_up, mb_upleft, mb_upright;              

        CheckAvailabilityOfNeighbors32(mb_ext_level);


        getLuma4x4Neighbour(img->current_mb_nr,0,0,-1, 0,&mb_left);
        getLuma4x4Neighbour(img->current_mb_nr,0,0, 0,-1,&mb_up);
        getLuma4x4Neighbour(img->current_mb_nr,0,0, 16<<mb_ext_level, -1,&mb_upright);
        getLuma4x4Neighbour(img->current_mb_nr,0,0, -1,-1,&mb_upleft);

        {
          fw_rFrameL = mb_left.available ? dec_picture->ref_idx[LIST_0][mb_left.pos_y][mb_left.pos_x] : -1;
          fw_rFrameU = mb_up.available ? dec_picture->ref_idx[LIST_0][mb_up.pos_y][mb_up.pos_x] : -1;
          fw_rFrameUL = mb_upleft.available ? dec_picture->ref_idx[LIST_0][mb_upleft.pos_y][mb_upleft.pos_x] : -1;
          fw_rFrameUR = mb_upright.available ? dec_picture->ref_idx[LIST_0][mb_upright.pos_y][mb_upright.pos_x] : fw_rFrameUL;      

          bw_rFrameL = mb_left.available ? dec_picture->ref_idx[LIST_1][mb_left.pos_y][mb_left.pos_x] : -1;
          bw_rFrameU = mb_up.available ? dec_picture->ref_idx[LIST_1][mb_up.pos_y][mb_up.pos_x] : -1;
          bw_rFrameUL = mb_upleft.available ? dec_picture->ref_idx[LIST_1][mb_upleft.pos_y][mb_upleft.pos_x] : -1;
          bw_rFrameUR = mb_upright.available ? dec_picture->ref_idx[LIST_1][mb_upright.pos_y][mb_upright.pos_x] : bw_rFrameUL;      
        }

        *fw_rFrame = (fw_rFrameL >= 0 && fw_rFrameU >= 0) ? min(fw_rFrameL,fw_rFrameU): max(fw_rFrameL,fw_rFrameU);
        *fw_rFrame = (*fw_rFrame >= 0 && fw_rFrameUR >= 0) ? min(*fw_rFrame,fw_rFrameUR): max(*fw_rFrame,fw_rFrameUR);

        *bw_rFrame = (bw_rFrameL >= 0 && bw_rFrameU >= 0) ? min(bw_rFrameL,bw_rFrameU): max(bw_rFrameL,bw_rFrameU);
        *bw_rFrame = (*bw_rFrame >= 0 && bw_rFrameUR >= 0) ? min(*bw_rFrame,bw_rFrameUR): max(*bw_rFrame,bw_rFrameUR);        

        if (*fw_rFrame >=0)
          SetMotionVectorPredictor32 (img, pmvfw, pmvfw+1, *fw_rFrame, LIST_0, dec_picture->ref_idx, dec_picture->mv, 0, 0, 16<<mb_ext_level, 16<<mb_ext_level, mb_ext_level);
        if (*bw_rFrame >=0)
          SetMotionVectorPredictor32 (img, pmvbw, pmvbw+1, *bw_rFrame, LIST_1, dec_picture->ref_idx, dec_picture->mv, 0, 0, 16<<mb_ext_level, 16<<mb_ext_level, mb_ext_level);

        memcpy(saved_pmvfw, pmvfw, sizeof(short)*2);
        memcpy(saved_pmvbw, pmvbw, sizeof(short)*2);
        saved_fw_rFrame = *fw_rFrame;
        saved_bw_rFrame = *bw_rFrame;
      }
      else
      {
        memcpy(pmvfw, saved_pmvfw, sizeof(short)*2);
        memcpy(pmvbw, saved_pmvbw, sizeof(short)*2);
        *fw_rFrame = saved_fw_rFrame;
        *bw_rFrame = saved_bw_rFrame;
      }
    }
    else
    {
      if (img->type==B_SLICE && img->direct_spatial_mv_pred_flag && (IS_DIRECT (currMB) || 
        (IS_P8x8(currMB) && !(currMB->b8mode[0] && currMB->b8mode[1] && currMB->b8mode[2] && currMB->b8mode[3]))))
      {
        char fw_rFrameL, fw_rFrameU, fw_rFrameUL, fw_rFrameUR;
        char bw_rFrameL, bw_rFrameU, bw_rFrameUL, bw_rFrameUR;    

        PixelPos mb_left, mb_up, mb_upleft, mb_upright;              

        getLuma4x4Neighbour(img->current_mb_nr,0,0,-1, 0,&mb_left);
        getLuma4x4Neighbour(img->current_mb_nr,0,0, 0,-1,&mb_up);
        getLuma4x4Neighbour(img->current_mb_nr,0,0,16, -1,&mb_upright);
        getLuma4x4Neighbour(img->current_mb_nr,0,0, -1,-1,&mb_upleft);

        {
          fw_rFrameL = mb_left.available ? dec_picture->ref_idx[LIST_0][mb_left.pos_y][mb_left.pos_x] : -1;
          fw_rFrameU = mb_up.available ? dec_picture->ref_idx[LIST_0][mb_up.pos_y][mb_up.pos_x] : -1;
          fw_rFrameUL = mb_upleft.available ? dec_picture->ref_idx[LIST_0][mb_upleft.pos_y][mb_upleft.pos_x] : -1;
          fw_rFrameUR = mb_upright.available ? dec_picture->ref_idx[LIST_0][mb_upright.pos_y][mb_upright.pos_x] : fw_rFrameUL;      

          bw_rFrameL = mb_left.available ? dec_picture->ref_idx[LIST_1][mb_left.pos_y][mb_left.pos_x] : -1;
          bw_rFrameU = mb_up.available ? dec_picture->ref_idx[LIST_1][mb_up.pos_y][mb_up.pos_x] : -1;
          bw_rFrameUL = mb_upleft.available ? dec_picture->ref_idx[LIST_1][mb_upleft.pos_y][mb_upleft.pos_x] : -1;
          bw_rFrameUR = mb_upright.available ? dec_picture->ref_idx[LIST_1][mb_upright.pos_y][mb_upright.pos_x] : bw_rFrameUL;      
        }

        *fw_rFrame = (fw_rFrameL >= 0 && fw_rFrameU >= 0) ? min(fw_rFrameL,fw_rFrameU): max(fw_rFrameL,fw_rFrameU);
        *fw_rFrame = (*fw_rFrame >= 0 && fw_rFrameUR >= 0) ? min(*fw_rFrame,fw_rFrameUR): max(*fw_rFrame,fw_rFrameUR);

        *bw_rFrame = (bw_rFrameL >= 0 && bw_rFrameU >= 0) ? min(bw_rFrameL,bw_rFrameU): max(bw_rFrameL,bw_rFrameU);
        *bw_rFrame = (*bw_rFrame >= 0 && bw_rFrameUR >= 0) ? min(*bw_rFrame,bw_rFrameUR): max(*bw_rFrame,bw_rFrameUR);        

        if (*fw_rFrame >=0)
          SetMotionVectorPredictor (img, pmvfw, (pmvfw+1), *fw_rFrame, LIST_0, dec_picture->ref_idx, dec_picture->mv, 0, 0, 16, 16);
        if (*bw_rFrame >=0)
          SetMotionVectorPredictor (img, pmvbw, pmvbw+1, *bw_rFrame, LIST_1, dec_picture->ref_idx, dec_picture->mv, 0, 0, 16, 16);
      }
    }
  }

  /*!
  ************************************************************************
  * \brief
  *    decode one macroblock
  ************************************************************************
  */
  int decode_one_macroblock(struct img_par *img,struct inp_par *inp, int mb_ext_level, int mb_count)
  {
    int tmp_block[16][16];
    int tmp_blockbw[16][16];
    int i=0,j=0,k,l,ii=0,jj=0,i1=0,j1=0,j4=0,i4=0;
    int /*uv, */hv;
    int vec1_x=0,vec1_y=0,vec2_x=0,vec2_y=0;
    int ioff,joff;
    int block8x8;   // needed for ABT

    int bw_pred=0, fw_pred=0/*, pred, ifx*/;
    //int ii0,jj0,ii1,jj1,if1,jf1,if0,jf0;
    int mv_mul/*, f1_x, f1_y, f2_x, f2_y, f3, f4*/;

    const byte decode_block_scan[16] = {0,1,4,5,2,3,6,7,8,9,12,13,10,11,14,15};

    Macroblock *currMB   = &img->mb_data[img->current_mb_nr];
    short ref_idx, fw_refframe=-1, bw_refframe=-1;
    int mv_mode, pred_dir/*, intra_prediction*/; // = currMB->ref_frame;
    short fw_ref_idx=-1, bw_ref_idx=-1;

    short  *** mv_array, ***fw_mv_array, ***bw_mv_array;

    int mv_scale;

    //int mb_nr     = img->current_mb_nr;
    //int smb       = ((img->type==SP_SLICE) && IS_INTER (currMB)) || (img->type == SI_SLICE && currMB->mb_type == SI4MB);
    int max_y_cr;

    StorablePicture **list;

    //int jf;

    char fw_rFrame=-1,bw_rFrame=-1;

    short pmvfw[2]={0,0},
      pmvbw[2]={0,0};              

    int direct_pdir=-1;

    //int curr_mb_field = 0;

    byte  **    moving_block;
    short ****  co_located_mv;
    char  ***   co_located_ref_idx;
    int64 ***   co_located_ref_id;

    int need_4x4_transform = (!currMB->luma_transform_size_8x8_flag);

    //int b8, b4;
    //int uv_shift;
    int yuv = dec_picture->chroma_format_idc - 1;

    // Residue Color Transform
    int residue_transform_flag = img->residue_transform_flag;
    //int residue_R, residue_G, residue_B, temp;
#ifdef GEOM
    int rho = currMB->geo_mode>>5, theta_idx = currMB->geo_mode & 31;
    int mask4x4[4][4]/*, mpr[4][4]*/;
    int segm, start_x, start_y;
#ifdef OBMC
    int mpr2[2][4][4]/*, mprUV2[2]*/;
    char **smfilter;
    if(currMB->geo_mode>-1)
      smfilter=geo_filter[mb_ext_level][rho][theta_idx];
#endif
    switch(mb_ext_level)
    {
    case 0:
      start_x=start_y=0;
      break;
    case 1:
      start_x = ((mb_count&3)&1)*16;
      start_y = ((mb_count&3)>>1)*16;
      break;
    case 2:
      start_x = mb16_pos[mb_count][0];
      start_y = mb16_pos[mb_count][1];
      break;
    default:
      error("mb_ext_level is out of range\n", -1);
    }

#endif

    if(IS_INTRA(currMB))
    {
      decode_one_component_i_slice(currMB,dec_picture,mb_ext_level, mb_count);
      return 0;
    }

#ifdef GEOM
    if(currMB->geo_mode==-1)
#endif
      if(!(IS_DIRECT(currMB)))
      {
        if(currMB->mb_type<4)
        {
          decode_one_component(currMB,dec_picture,mb_ext_level, mb_count);
          return 0;
        }
      }

      moving_block = Co_located->moving_block;
      co_located_mv = Co_located->mv;
      co_located_ref_idx = Co_located->ref_idx;
      co_located_ref_id = Co_located->ref_pic_id;
      max_y_cr = dec_picture->size_y_cr-1;

      for (l=0;l<(2);l++)
      {
        for(k = 0; k < listXsize[l]; k++)
        {
          listX[l][k]->chroma_vector_adjustment= 0;
          //if(img->structure == TOP_FIELD && img->structure != listX[l][k]->structure)
          //  listX[l][k]->chroma_vector_adjustment = -2;
          //if(img->structure == BOTTOM_FIELD && img->structure != listX[l][k]->structure)
          //  listX[l][k]->chroma_vector_adjustment = 2;
        }
      }


#ifdef EIGHTH_PEL
      mv_mul = 4 << img->mv_res;
#else
      mv_mul=4;
#endif

      // luma decoding **************************************************

      prepare_B_DIRECT(currMB,mb_ext_level,mb_count,&fw_rFrame,&bw_rFrame,pmvfw,pmvbw);

      for (block8x8=0; block8x8<4; block8x8++)
      {
        if (currMB->b8mode[block8x8] == I8MB)
        {
          //=========== 8x8 BLOCK TYPE ============
          ioff = 8*(block8x8%2);
          joff = 8*(block8x8/2);

          //PREDICTION
          if(!residue_transform_flag)
            intrapred8x8(img, block8x8);

#ifdef USE_INTRA_MDDT
          if(img->UseIntraMDDT)
          {
            int img_block_x = (img->mb_x)*4 + 2*(block8x8%2);
            int img_block_y = (img->mb_y)*4 + 2*(block8x8/2);
            int ipmode = img->ipredmode[img_block_x][img_block_y];

            if(ipmode == 2)
              itrans8x8(img, ioff, joff);
            else
              itrans8x8klt_sep_fast(img, ioff, joff, ipmode);
          }
          else
            itrans8x8(img,ioff,joff); 
#else
          itrans8x8(img,ioff,joff);      // use DCT transform and make 8x8 block m7 from prediction block mpr
#endif

          move_data_to_dec_picture_transpose(8, 8, img->pix_x + ioff, img->pix_y + joff, ioff, joff);
          continue;
        }

        for (k = block8x8*4; k < block8x8*4+4; k ++)
        {
#if defined GEOM && GEO_SCHEME == 1
          int num_segm, segm0=0, segm1=0;
#endif
          i = (decode_block_scan[k] & 3);
          j = ((decode_block_scan[k] >> 2) & 3);

          ioff=i*4;
          i4=img->block_x+i;

          joff=j*4;
          j4=img->block_y+j;

          mv_mode  = currMB->b8mode[2*(j/2)+(i/2)]; //in the following mv_mode only differentiate skip, intra and the others



#if defined GEOM && GEO_SCHEME == 1

          num_segm = 1;
          if(currMB->geo_mode > -1)
          {            
            segm1 = geo_mask[mb_ext_level][rho][theta_idx][start_y+j*4][start_x+i*4];
            for(jj=0; jj<4; jj++)
              for(ii=0; ii<4; ii++)
              {               
                mask4x4[jj][ii] = geo_mask[mb_ext_level][rho][theta_idx][start_y+j*4+jj][start_x+i*4+ii];
#ifndef OBMC
                if(mask4x4[jj][ii] != segm1)
#endif
                  num_segm = 2;
              }
              if(num_segm == 1)
                segm0=segm1;
              else
              {
                segm0=0; 
                segm1=1;
              }
          }

          for(segm=segm0; segm<=segm1; segm++)
          {
#endif

#ifdef GEOM
            if(currMB->geo_mode > -1)
            {
              if(img->geo_scheme == 0)
                pred_dir = best8x8pdir_geo[MB16GeoMask[currMB->geo_mode][decode_block_scan[k]]]; 
              else
                pred_dir = best8x8pdir_geo[segm]; 
            }
            else
#endif
              pred_dir = currMB->b8pdir[2*(j/2)+(i/2)];

            assert (pred_dir<=2);

            // PREDICTION
            if (mv_mode==IBLOCK)
            {
              // Residue Color Transform
              if(!residue_transform_flag)
              {
                //===== INTRA PREDICTION =====
                if (intrapred(img,ioff,joff,i4,j4)==SEARCH_SYNC)  /* make 4x4 prediction block mpr from given prediction img->mb_mode */
                  return SEARCH_SYNC;                   /* bit error */
              }
            }
            else if (!IS_NEWINTRA (currMB))
            {
              if (pred_dir != 2)
              {
                //===== FORWARD/BACKWARD PREDICTION =====
#if defined GEOM && GEO_SCHEME == 1          
                if(currMB->geo_mode > -1)
                {
                  fw_refframe = ref_idx  = pred_dir==0 ? best8x8fwref_geo[segm] : best8x8bwref_geo[segm];
                  mv_array_geo[pred_dir][j4][i4][0] = geo_mv[LIST_0 + pred_dir][segm][0];
                  mv_array_geo[pred_dir][j4][i4][1] = geo_mv[LIST_0 + pred_dir][segm][1];
                  mv_array = mv_array_geo[pred_dir];
                }
                else
#endif
                {
                  fw_refframe = ref_idx  = dec_picture->ref_idx[LIST_0 + pred_dir][j4][i4];
                  mv_array = dec_picture->mv[LIST_0 + pred_dir];
                }

                list     = listX[pred_dir];
                vec1_x = i4*4*mv_mul + mv_array[j4][i4][0];

                //if (!curr_mb_field)
                //{
                vec1_y = j4*4*mv_mul + mv_array[j4][i4][1];
                //}
                //else
                //{
                //  if (mb_nr%2 == 0) 
                //    vec1_y = (img->block_y * 2 + joff) * mv_mul + mv_array[j4][i4][1];
                //  else
                //    vec1_y = ((img->block_y-4) * 2 + joff)* mv_mul + mv_array[j4][i4][1];
                //}

                getblock2DFilt(ref_idx, list, i4*4, j4*4, mv_array[j4][i4][0], mv_array[j4][i4][1],
                  img->width, img->height, tmp_block, BLOCK_SIZE, BLOCK_SIZE);

                if (img->apply_weights)
                {

                  for(ii=0;ii<BLOCK_SIZE;ii++)
                    for(jj=0;jj<BLOCK_SIZE;jj++)  
                      img->mpr[ii+ioff][jj+joff] = Clip1(((img->wp_weight[pred_dir][ref_idx][0] *  tmp_block[ii][jj]+ img->wp_round_luma) >>img->luma_log2_weight_denom)  + img->wp_offset[pred_dir][ref_idx][0] );
                }
                else
                {
                  for(ii=0;ii<BLOCK_SIZE;ii++)
                    for(jj=0;jj<BLOCK_SIZE;jj++)  
                      img->mpr[ii+ioff][jj+joff] = tmp_block[ii][jj];
                }
#if defined GEOM && GEO_SCHEME == 1
                if(num_segm == 2)
                {
                  for(jj=0; jj<4; jj++)
                    for(ii=0; ii<4; ii++)
#ifdef OBMC
                      mpr2[segm][jj][ii] = img->mpr[ii+ioff][jj+joff];
#else
                      if(mask4x4[jj][ii] == segm)
                        mpr[jj][ii] = img->mpr[ii+ioff][jj+joff];
#endif
                }         
#endif

              }
              else //bi-directional
              {
                if (mv_mode != 0)
                {


                  //
#if defined GEOM && GEO_SCHEME == 1

                  if(currMB->geo_mode > -1)
                  {
                    mv_array_geo[0][j4][i4][0] = geo_mv[LIST_0][segm][0];
                    mv_array_geo[0][j4][i4][1] = geo_mv[LIST_0][segm][1];
                    fw_mv_array = mv_array_geo[0];

                    mv_array_geo[1][j4][i4][0] = geo_mv[LIST_1][segm][0];
                    mv_array_geo[1][j4][i4][1] = geo_mv[LIST_1][segm][1];
                    bw_mv_array = mv_array_geo[1];

                    fw_refframe = best8x8fwref_geo[segm];
                    bw_refframe = best8x8bwref_geo[segm];
                  }
                  else
#endif
                  {
                    //===== BI-DIRECTIONAL PREDICTION =====
                    fw_mv_array = dec_picture->mv[LIST_0];
                    bw_mv_array = dec_picture->mv[LIST_1];

                    fw_refframe = dec_picture->ref_idx[LIST_0][j4][i4];
                    bw_refframe = dec_picture->ref_idx[LIST_1][j4][i4];
                  }
                  fw_ref_idx = fw_refframe;
                  bw_ref_idx = bw_refframe;
                }
                else
                {
                  //===== DIRECT PREDICTION =====
                  fw_mv_array = dec_picture->mv[LIST_0];
                  bw_mv_array = dec_picture->mv[LIST_1];
                  bw_refframe = 0;

                  if (img->direct_spatial_mv_pred_flag )
                  {
                    int imgblock_y= img->block_y;
                    int j6=imgblock_y+j;


                    if (fw_rFrame >=0)
                    {
                      if (!fw_rFrame  && ((!moving_block[j6][i4]) && (!listX[1][0]->is_long_term)))
                      {
                        dec_picture->mv  [LIST_0][j4][i4][0]= 0;
                        dec_picture->mv  [LIST_0][j4][i4][1]= 0;
                        dec_picture->ref_idx[LIST_0][j4][i4] = 0;
                      }
                      else
                      {
                        dec_picture->mv  [LIST_0][j4][i4][0]= pmvfw[0];
                        dec_picture->mv  [LIST_0][j4][i4][1]= pmvfw[1];
                        dec_picture->ref_idx[LIST_0][j4][i4] = fw_rFrame;
                      }
                    }
                    else
                    {
                      dec_picture->ref_idx[LIST_0][j4][i4] = -1;                
                      dec_picture->mv  [LIST_0][j4][i4][0]= 0;
                      dec_picture->mv  [LIST_0][j4][i4][1]= 0;
                    }

                    if (bw_rFrame >=0)
                    {
                      if  (bw_rFrame==0 && ((!moving_block[j6][i4]) && (!listX[1][0]->is_long_term)))
                      {                  

                        dec_picture->mv  [LIST_1][j4][i4][0]= 0;
                        dec_picture->mv  [LIST_1][j4][i4][1]= 0;
                        dec_picture->ref_idx[LIST_1][j4][i4] = bw_rFrame;

                      }
                      else
                      {
                        dec_picture->mv  [LIST_1][j4][i4][0]= pmvbw[0];
                        dec_picture->mv  [LIST_1][j4][i4][1]= pmvbw[1];

                        dec_picture->ref_idx[LIST_1][j4][i4] = bw_rFrame;
                      }               
                    }
                    else
                    {                  
                      dec_picture->mv  [LIST_1][j4][i4][0]=0;
                      dec_picture->mv  [LIST_1][j4][i4][1]=0;
                      dec_picture->ref_idx[LIST_1][j4][i4] = -1;

                    }

                    if (fw_rFrame < 0 && bw_rFrame < 0)
                    {
                      dec_picture->ref_idx[LIST_0][j4][i4] = 0;
                      dec_picture->ref_idx[LIST_1][j4][i4] = 0;
                    }

                    fw_refframe = (dec_picture->ref_idx[LIST_0][j4][i4]!=-1) ? dec_picture->ref_idx[LIST_0][j4][i4]:0;
                    bw_refframe = (dec_picture->ref_idx[LIST_1][j4][i4]!=-1) ? dec_picture->ref_idx[LIST_1][j4][i4]:0;

                    fw_ref_idx = fw_refframe;
                    bw_ref_idx = bw_refframe;

                    if      (dec_picture->ref_idx[LIST_1][j4][i4]==-1) direct_pdir = 0;
                    else if (dec_picture->ref_idx[LIST_0][j4][i4]==-1) direct_pdir = 1;
                    else                                               direct_pdir = 2;

                  }
                  else // Temporal Mode
                  {

                    int imgblock_y=  img->block_y;
                    int j6= imgblock_y + j;

                    int refList = (co_located_ref_idx[LIST_0][j6][i4]== -1 ? LIST_1 : LIST_0);
                    int ref_idx =  co_located_ref_idx[refList][j6][i4];


                    if(ref_idx==-1) // co-located is intra mode
                    {
                      for(hv=0; hv<2; hv++)
                      {
                        dec_picture->mv  [LIST_0][j4][i4][hv]=0;
                        dec_picture->mv  [LIST_1][j4][i4][hv]=0;                    
                      }

                      dec_picture->ref_idx[LIST_0][j4][i4] = 0;
                      dec_picture->ref_idx[LIST_1][j4][i4] = 0;

                      fw_refframe = 0;
                      fw_ref_idx = 0;
                    }
                    else // co-located skip or inter mode
                    {
                      int mapped_idx=0;
                      int iref;          

                      {
                        for (iref=0;iref<min(img->num_ref_idx_l0_active,listXsize[LIST_0]);iref++)
                        {
#if 1
                          //if(img->structure==0 && curr_mb_field==0)
                          {
                            // If the current MB is a frame MB and the colocated is from a field picture, 
                            // then the co_located_ref_id may have been generated from the wrong value of 
                            // frame_poc if it references it's complementary field, so test both POC values
                            if(listX[0][iref]->top_poc*2 == co_located_ref_id[refList][j6][i4] || listX[0][iref]->bottom_poc*2 == co_located_ref_id[refList][j6][i4])
                            {
                              mapped_idx=iref;
                              break;
                            }
                            else //! invalid index. Default to zero even though this case should not happen
                              mapped_idx=INVALIDINDEX;
                            continue;
                          }    
#endif                                        
                          if (dec_picture->ref_pic_num[img->current_slice_nr][LIST_0][iref]==co_located_ref_id[refList][j6][i4])
                          {
                            mapped_idx=iref;
                            break;
                          }
                          else //! invalid index. Default to zero even though this case should not happen
                          {                        
                            mapped_idx=INVALIDINDEX;
                          }
                        }
                        if (INVALIDINDEX == mapped_idx)
                        {
                          error("temporal direct error\ncolocated block has ref that is unavailable",-1111);
                        }
                      }

                      fw_ref_idx = mapped_idx;
                      mv_scale = img->mvscale[LIST_0][mapped_idx];

                      //! In such case, an array is needed for each different reference.
                      if (mv_scale == 9999 || listX[LIST_0][mapped_idx]->is_long_term)
                      {
                        dec_picture->mv  [LIST_0][j4][i4][0]=co_located_mv[refList][j6][i4][0];
                        dec_picture->mv  [LIST_0][j4][i4][1]=co_located_mv[refList][j6][i4][1];

                        dec_picture->mv  [LIST_1][j4][i4][0]=0;
                        dec_picture->mv  [LIST_1][j4][i4][1]=0;
                      }
                      else
                      {
                        dec_picture->mv  [LIST_0][j4][i4][0]=(mv_scale * co_located_mv[refList][j6][i4][0] + 128 ) >> 8;
                        dec_picture->mv  [LIST_0][j4][i4][1]=(mv_scale * co_located_mv[refList][j6][i4][1] + 128 ) >> 8;

                        dec_picture->mv  [LIST_1][j4][i4][0]=dec_picture->mv  [LIST_0][j4][i4][0] - co_located_mv[refList][j6][i4][0] ;
                        dec_picture->mv  [LIST_1][j4][i4][1]=dec_picture->mv  [LIST_0][j4][i4][1] - co_located_mv[refList][j6][i4][1] ;
                      }

                      fw_refframe = dec_picture->ref_idx[LIST_0][j4][i4] = mapped_idx; //listX[1][0]->ref_idx[refList][j4][i4];
                      bw_refframe = dec_picture->ref_idx[LIST_1][j4][i4] = 0;

                      fw_ref_idx = fw_refframe;
                      bw_ref_idx = bw_refframe;
                    }
                  }
                  // store reference picture ID determined by direct mode
                  dec_picture->ref_pic_id[LIST_0][j4][i4] = dec_picture->ref_pic_num[img->current_slice_nr][LIST_0][(short)dec_picture->ref_idx[LIST_0][j4][i4]];
                  dec_picture->ref_pic_id[LIST_1][j4][i4] = dec_picture->ref_pic_num[img->current_slice_nr][LIST_1][(short)dec_picture->ref_idx[LIST_1][j4][i4]];  
                }

                if (mv_mode==0 && img->direct_spatial_mv_pred_flag )
                {
                  if (dec_picture->ref_idx[LIST_0][j4][i4] >= 0)
                  {
                    vec1_x = i4*4*mv_mul + fw_mv_array[j4][i4][0];
                    vec1_y = j4*4*mv_mul + fw_mv_array[j4][i4][1];
                    getblock2DFilt(fw_refframe, listX[0], i4*4, j4*4, fw_mv_array[j4][i4][0], fw_mv_array[j4][i4][1],
                      img->width, img->height, tmp_block, BLOCK_SIZE, BLOCK_SIZE);
                  }
                  if (dec_picture->ref_idx[LIST_1][j4][i4] >= 0)
                  {
                    vec2_x = i4*4*mv_mul + bw_mv_array[j4][i4][0];
                    vec2_y = j4*4*mv_mul + bw_mv_array[j4][i4][1];
                    getblock2DFilt(bw_refframe, listX[1], i4*4, j4*4, bw_mv_array[j4][i4][0], bw_mv_array[j4][i4][1],
                      img->width, img->height, tmp_blockbw, BLOCK_SIZE, BLOCK_SIZE);
                  }              
                }
                else
                {
                  vec1_x = i4*4*mv_mul + fw_mv_array[j4][i4][0];
                  vec2_x = i4*4*mv_mul + bw_mv_array[j4][i4][0];
                  vec1_y = j4*4*mv_mul + fw_mv_array[j4][i4][1];
                  vec2_y = j4*4*mv_mul + bw_mv_array[j4][i4][1];
                  getblock2DFilt(fw_refframe, listX[0], i4*4, j4*4, fw_mv_array[j4][i4][0], fw_mv_array[j4][i4][1], img->width, img->height, tmp_block, BLOCK_SIZE, BLOCK_SIZE);
                  getblock2DFilt(bw_refframe, listX[1], i4*4, j4*4, bw_mv_array[j4][i4][0], bw_mv_array[j4][i4][1], img->width, img->height, tmp_blockbw, BLOCK_SIZE, BLOCK_SIZE);
                }

                if (mv_mode==0 && img->direct_spatial_mv_pred_flag && direct_pdir==0)
                {
                  if (img->apply_weights)
                  {
                    for(ii=0;ii<BLOCK_SIZE;ii++)
                      for(jj=0;jj<BLOCK_SIZE;jj++)  
                        img->mpr[ii+ioff][jj+joff] = Clip1(((tmp_block[ii][jj] * img->wp_weight[0][fw_ref_idx][0]  
                      + img->wp_round_luma)>>img->luma_log2_weight_denom) 
                        + img->wp_offset[0][fw_ref_idx][0]);
                  }
                  else
                  {
                    for(ii=0;ii<BLOCK_SIZE;ii++)
                      for(jj=0;jj<BLOCK_SIZE;jj++)  
                        img->mpr[ii+ioff][jj+joff] = tmp_block[ii][jj];
                  }
                }
                else if (mv_mode==0 && img->direct_spatial_mv_pred_flag && direct_pdir==1)
                {              
                  if (img->apply_weights)
                  {
                    for(ii=0;ii<BLOCK_SIZE;ii++)
                      for(jj=0;jj<BLOCK_SIZE;jj++)  
                        img->mpr[ii+ioff][jj+joff] = Clip1(((tmp_blockbw[ii][jj] * img->wp_weight[1][bw_ref_idx][0] 
                      + img->wp_round_luma)>>img->luma_log2_weight_denom) 
                        + img->wp_offset[1][bw_ref_idx][0]);
                  }
                  else
                  {
                    for(ii=0;ii<BLOCK_SIZE;ii++)
                      for(jj=0;jj<BLOCK_SIZE;jj++)  
                        img->mpr[ii+ioff][jj+joff] = tmp_blockbw[ii][jj];
                  }
                }
                else if(img->apply_weights)
                {
                  int alpha_fw, alpha_bw;

                  if (mv_mode==0 && img->direct_spatial_mv_pred_flag==0 )bw_ref_idx=0;    //temporal direct 

                  alpha_fw = img->wbp_weight[0][fw_ref_idx][bw_ref_idx][0];
                  alpha_bw = img->wbp_weight[1][fw_ref_idx][bw_ref_idx][0];

                  if(img->use_high_precision_flag != 0)
                  {
                    for(ii=0;ii<BLOCK_SIZE;ii++)
                      for(jj=0;jj<BLOCK_SIZE;jj++)  
                        img->mpr[ii+ioff][jj+joff] = (int)Clip1((alpha_fw * tmp_block[ii][jj] + alpha_bw * tmp_blockbw[ii][jj]  
                      + ((img->wp_offset[0][fw_ref_idx][0] + img->wp_offset[1][bw_ref_idx][0] ) << img->luma_log2_weight_denom)
                        + (1<<img->luma_log2_weight_denom)*(1- img->bipred_rounding_control)) >> (img->luma_log2_weight_denom+1)) ;
                  }
                  else
                  {
                    for(ii=0;ii<BLOCK_SIZE;ii++)
                      for(jj=0;jj<BLOCK_SIZE;jj++)  
                        img->mpr[ii+ioff][jj+joff] = (int)Clip1(((alpha_fw * tmp_block[ii][jj] + alpha_bw * tmp_blockbw[ii][jj]  
                      + (1<<img->luma_log2_weight_denom)) >> (img->luma_log2_weight_denom+1)) 
                        + ((img->wp_offset[0][fw_ref_idx][0] + img->wp_offset[1][bw_ref_idx][0] + 1) >>1));
                  }
                }
                else
                {
                  if(img->use_high_precision_flag != 0)
                  {
                    int round = 1 - img->bipred_rounding_control;
                    for(ii=0;ii<BLOCK_SIZE;ii++)
                      for(jj=0;jj<BLOCK_SIZE;jj++)  
                        img->mpr[ii+ioff][jj+joff] = (tmp_block[ii][jj]+tmp_blockbw[ii][jj]+round)/2;
                  }
                  else
                  {
                    for(ii=0;ii<BLOCK_SIZE;ii++)
                      for(jj=0;jj<BLOCK_SIZE;jj++)  
                        img->mpr[ii+ioff][jj+joff] = (tmp_block[ii][jj]+tmp_blockbw[ii][jj]+1)/2;
                  }
                }
#if defined GEOM && GEO_SCHEME == 1
                if(num_segm == 2)
                {
                  for(jj=0; jj<4; jj++)
                    for(ii=0; ii<4; ii++)
#ifdef OBMC
                      mpr2[segm][jj][ii] = img->mpr[ii+ioff][jj+joff];
#else
                      if(mask4x4[jj][ii] == segm)
                        mpr[jj][ii] = img->mpr[ii+ioff][jj+joff];
#endif
                }
#endif

              }       

            }

#if defined GEOM && GEO_SCHEME == 1
          }
          if(num_segm == 2)
          {
            for(jj=0; jj<4; jj++)
              for(ii=0; ii<4; ii++)
              {
#ifdef OBMC
                if(smfilter[start_y+joff+jj][start_x+ioff+ii])//OBMC
                {
                  int diff = abs(mpr2[0][jj][ii] - mpr2[1][jj][ii]);

                  if(diff < 45)
                  {
                    if(mask4x4[jj][ii] == 0)
                      img->mpr[ii+ioff][jj+joff] = (int)((mpr2[0][jj][ii]*2 + mpr2[1][jj][ii]  +1.5)/3);
                    else
                      img->mpr[ii+ioff][jj+joff] = (int)((mpr2[0][jj][ii]   + mpr2[1][jj][ii]*2+1.5)/3);
                  }
                  else
                  {
                    if(mask4x4[jj][ii] == 0)
                      img->mpr[ii+ioff][jj+joff] = (mpr2[0][jj][ii]*3 + mpr2[1][jj][ii]  +2)>>2;
                    else
                      img->mpr[ii+ioff][jj+joff] = (mpr2[0][jj][ii]   + mpr2[1][jj][ii]*3+2)>>2;
                  }
                }
                else
                {
                  segm = mask4x4[jj][ii];
                  img->mpr[ii+ioff][jj+joff] = mpr2[segm][jj][ii];
                }
#else
                img->mpr[ii+ioff][jj+joff] = mpr[jj][ii];
#endif       
              }
          }

#endif


          // =============== 4x4 itrans ================
          // -------------------------------------------
          //if (0)
          //{
          //  if(!IS_NEWINTRA (currMB))// modif ES added
          //    itrans_sp(img,ioff,joff,i,j);
          //  else //modif ES added
          //    itrans(img,ioff,joff,i,j,0); // modif ES added
          //}
          //else
          //{
          if(need_4x4_transform)
          {
            if(img->type==SP_SLICE && currMB->mb_type != I16MB){ // ES added
              itrans_sp   (img,ioff,joff,i,j);//ES added
            }
            else{
#ifdef USE_INTRA_MDDT
              if(currMB->mb_type == I4MB && img->UseIntraMDDT)
              {
                int ipmode = img->ipredmode[i4][j4];
                if(ipmode == 2)
                  itrans(img, ioff, joff, i, j, 0);
                else
                  itransklt_sep(img, ioff, joff, i, j, ipmode);
              }
              else
#endif
                itrans   (img,ioff,joff,i,j, 0);      // use DCT transform and make 4x4 block m7 from prediction block mpr
            }
          }
          //}
          if(need_4x4_transform)
          {
            move_data_to_dec_picture_transpose(BLOCK_SIZE, BLOCK_SIZE, i4*BLOCK_SIZE, j4*BLOCK_SIZE, 0, 0);
          }// if(need_4x4_transform)
        }

        if(currMB->luma_transform_size_8x8_flag == 2)
        {
          // perform itrans outside of b8 loop
          continue;
        }
        else
          if(!need_4x4_transform)       
          {
            // =============== 8x8 itrans ================
            // -------------------------------------------
            ioff = 8*(block8x8%2);
            joff = 8*(block8x8/2);

            itrans8x8(img,ioff,joff);      // use DCT transform and make 8x8 block m7 from prediction block mpr
          }
      }

      if(currMB->luma_transform_size_8x8_flag == 2) 
      {
#ifdef GEOM
        if(currMB->mb_type == 0 || currMB->mb_type == P16x16 || currMB->geo_mode != -1 || (mb_ext_level == 1 && MB32.mb_type != P8x8) || (mb_ext_level == 2 && MB64.mb_type != P8x8)) //MB64X64
#else
        if(currMB->mb_type == 0 || currMB->mb_type == P16x16 || (mb_ext_level == 1 && MB32.mb_type != P8x8) || (mb_ext_level == 2 && MB64.mb_type != P8x8)) //MB64X64
#endif
        {
          itrans16x16(img);
        }
        else if(currMB->mb_type == P16x8)
        {
          itrans16x8(img, 0);
          itrans16x8(img, 1);
        }
        else // currMB->mb_type == P8x16
        {
          assert(currMB->mb_type == P8x16);
          itrans8x16(img, 0);
          itrans8x16(img, 1);
        }
      }

      if(currMB->luma_transform_size_8x8_flag > 0 && !IS_INTRA(currMB))
      {
        move_data_to_dec_picture_transpose(MB_BLOCK_SIZE, MB_BLOCK_SIZE, img->pix_x, img->pix_y, 0, 0);
      }

      {
        //2x2 chroma processing
        decode_one_macroblock_chroma(currMB, mb_ext_level, mb_count);
      }

      return 0;
  }


  void decode_one_macroblock_chroma(Macroblock * currMB,int mb_ext_level, int mb_count)
  {
    int i=0,j=0,k,l,ii=0,jj=0,i1=0,j1=0,j4=0,i4=0;
    int uv;
    int ioff,joff;

    int bw_pred=0, fw_pred=0, pred, ifx;
    int ii0,jj0,ii1,jj1,if1,jf1,if0,jf0;
    int f1_x, f1_y, f2_x, f2_y, f3, f4;

    short ref_idx, fw_refframe=-1, bw_refframe=-1;
    int mv_mode, pred_dir, intra_prediction; // = currMB->ref_frame;
    short fw_ref_idx=-1, bw_ref_idx=-1;

    short  *** mv_array, ***fw_mv_array, ***bw_mv_array;

    int smb       = 0;
    int max_y_cr;

    StorablePicture **list;

    int jf;

    char fw_rFrame=-1,bw_rFrame=-1;
    int direct_pdir=-1;

    //int curr_mb_field = 0;

    int b8, b4;
    int uv_shift;
    int yuv = dec_picture->chroma_format_idc - 1;

#ifdef GEOM
    int rho = currMB->geo_mode>>5, theta_idx = currMB->geo_mode & 31;
    //int mask4x4[4][4], mpr[4][4];
    int segm, start_x, start_y;
#ifdef OBMC
    int /*mpr2[2][4][4],*/ mprUV2[2];
    char **smfilter;
    if(currMB->geo_mode>-1)
      smfilter=geo_filter[mb_ext_level][rho][theta_idx];
#endif
    switch(mb_ext_level)
    {
    case 0:
      start_x=start_y=0;
      break;
    case 1:
      start_x = ((mb_count&3)&1)*16;
      start_y = ((mb_count&3)>>1)*16;
      break;
    case 2:
      start_x = mb16_pos[mb_count][0];
      start_y = mb16_pos[mb_count][1];
      break;
    default:
      error("mb_ext_level is out of range\n", -1);
    }

#endif

    max_y_cr = dec_picture->size_y_cr-1;

    for (l=0;l<(2);l++)
    {
      for(k = 0; k < listXsize[l]; k++)
      {
        listX[l][k]->chroma_vector_adjustment= 0;
      }
    }

    if (dec_picture->chroma_format_idc != YUV400)
    {
#ifdef OBMC
      int curr_segm, segmi, segm0=0, segm1=0;
#endif
      int blockUV[32][32];
      int blockUV_fw[32][32];
      int blockUV_bw[32][32];
      int blockUV_2x2[32][32];

      int geometry = 0; // When geometry mode is ON, this should be set to 1.
#ifdef GEOM
      if(currMB->geo_mode > -1)
        geometry = 1; // When geometry mode is ON, this should be set to 1.
#endif
      // chroma decoding *******************************************************
      //f1_x = (64/img->mb_cr_size_x) << img->mv_res;
      //f1_y = (64/img->mb_cr_size_y) << img->mv_res;
      f1_x = (8) << img->mv_res;
      f1_y = (8) << img->mv_res;
      f2_x=f1_x-1;
      f2_y=f1_y-1;

      f3=f1_x*f1_y;
      f4=f3>>1;

      for(uv=0;uv<2;uv++)
      {
        uv_shift = uv*(img->num_blk8x8_uv/2);
        intra_prediction = IS_INTRA (currMB);

        if (intra_prediction)
        {
          intrapred_chroma(img, uv);
        }

        for (b8=0;b8<(img->num_blk8x8_uv/2);b8++)
        {
          for(b4=0;b4<4;b4++)
          {
            joff = subblk_offset_y[yuv][b8][b4];
            ioff = subblk_offset_x[yuv][b8][b4];
            j4=img->pix_c_y+joff;
            i4=img->pix_c_x+ioff;

            mv_mode  = currMB->b8mode[block8x8_idx[yuv][b8][b4]];
#ifdef GEOM                  
            if (!intra_prediction)
            {
              //--- FORWARD/BACKWARD PREDICTION ---
              for(jj=0;jj<4;jj++)
              {
                //jf=(j4+jj)/(img->mb_cr_size_y/4);     // jf  = Subblock_y-coordinate
                jf=(j4+jj)>>1;     // jf  = Subblock_y-coordinate

                for(ii=0;ii<4;ii++)
                {
#ifdef OBMC                
                  if(currMB->geo_mode > -1)
                  {
                    curr_segm = geo_mask[mb_ext_level][rho][theta_idx][start_y + ((joff+jj)<<1)][start_x + ((ioff+ii)<<1)];

                    if(geo_filterUV[mb_ext_level][rho][theta_idx][(start_y>>1) + (joff+jj)][(start_x>>1) + (ioff+ii)] > 0)
                      segm1=1;
                    else
                      segm1=0;
                  }
                  else
                  {
                    segm1=0;
                  }

                  for(segmi=segm0; segmi<=segm1; segmi++)
                  {
#endif
                    //ifx=(i4+ii)/(img->mb_cr_size_x/4);  // ifx = Subblock_x-coordinate
                    ifx=(i4+ii)>>1;  // ifx = Subblock_x-coordinate

                    if(currMB->geo_mode > -1)
                    {
                      if(img->geo_scheme == 0)
                        pred_dir = best8x8pdir_geo[MB16GeoMask[currMB->geo_mode][(((joff+jj)>>1)<<2)+((ioff+ii)>>1)]]; 
                      else
                      {
#ifdef OBMC
                        if(segm1==1)
                          segm = segmi;
                        else
                          segm = curr_segm;
#else               
                        segm = geo_mask[mb_ext_level][rho][theta_idx][start_y + ((joff+jj)<<1)][start_x + ((ioff+ii)<<1)];
#endif
                        pred_dir = best8x8pdir_geo[segm];                     
                      }

                    }
                    else
#endif
                      pred_dir = currMB->b8pdir[block8x8_idx[yuv][b8][b4]];
                    assert (pred_dir<=2);


#ifndef GEOM                  
                    if (!intra_prediction)
                    {
#endif
                      if (pred_dir != 2)
                      {
                        //--- FORWARD/BACKWARD PREDICTION ---
#if defined GEOM && GEO_SCHEME == 1
                        if(currMB->geo_mode > -1)
                        {
                          mv_array_geo[pred_dir][jf][ifx][0] = geo_mv[LIST_0 + pred_dir][segm][0];
                          mv_array_geo[pred_dir][jf][ifx][1] = geo_mv[LIST_0 + pred_dir][segm][1];
                          mv_array = mv_array_geo[pred_dir];
                        }
                        else
#endif
                        {
                          mv_array = dec_picture->mv[LIST_0 + pred_dir];
                        }

                        list = listX[pred_dir];
#ifndef GEOM                  
                        for(jj=0;jj<4;jj++)
                        {
                          jf=(j4+jj)/(img->mb_cr_size_y/4);     // jf  = Subblock_y-coordinate
                          for(ii=0;ii<4;ii++)
                          {
                            ifx=(i4+ii)/(img->mb_cr_size_x/4);  // ifx = Subblock_x-coordinate
#endif
#if defined GEOM && GEO_SCHEME == 1
                            if(currMB->geo_mode > -1)
                            {
                              fw_refframe = ref_idx = pred_dir==0 ? best8x8fwref_geo[segm] : best8x8bwref_geo[segm];
                            }
                            else
#endif
                            {
                              fw_refframe = ref_idx = dec_picture->ref_idx[LIST_0+pred_dir][jf][ifx];
                            }

                            i1=(i4+ii)*f1_x+mv_array[jf][ifx][0];
                            j1=(j4+jj)*f1_y+mv_array[jf][ifx][1];

                            if (active_sps->chroma_format_idc == 1)
                              j1 += list[ref_idx]->chroma_vector_adjustment;

                            ii0=max (0, min (i1/f1_x,   img->width_cr-1));
                            jj0=max (0, min (j1/f1_y,   max_y_cr));
                            ii1=max (0, min ((i1+f2_x)/f1_x, img->width_cr-1));
                            jj1=max (0, min ((j1+f2_y)/f1_y, max_y_cr));

                            if1=(i1 & f2_x);
                            jf1=(j1 & f2_y);
                            if0=f1_x-if1;
                            jf0=f1_y-jf1;

                            if (!geometry)
                            {
                              if (!(ii&0x01) && !(jj&0x01)) // ii=0,2 and jj=0,2
                              {
                                get_eighthpel_blockUV_direct(list[ref_idx]->imgUV[uv], i1, j1, img, blockUV_2x2, 0, 2, 2);
                                blockUV[jj][ii] = blockUV_2x2[0][0];
                                blockUV[jj][ii+1] = blockUV_2x2[0][1];
                                blockUV[jj+1][ii] = blockUV_2x2[1][0];
                                blockUV[jj+1][ii+1] = blockUV_2x2[1][1];
                              }
                            }

                            if (img->apply_weights)
                            {
                              if (!geometry)
                              {
                                pred = blockUV[jj][ii];
                              }
                              else
                              {
                                get_eighthpel_blockUV_direct(list[ref_idx]->imgUV[uv], i1, j1, img, blockUV, 1, 2, 2);
                                pred = blockUV[0][0];
                              }

                              img->mpr[ii+ioff][jj+joff] = Clip1_Chr(((img->wp_weight[pred_dir][ref_idx][uv+1] * pred  + img->wp_round_chroma)>>img->chroma_log2_weight_denom) + img->wp_offset[pred_dir][ref_idx][uv+1]);
                            }
                            else
                            {
                              if (!geometry)
                              {
                                img->mpr[ii+ioff][jj+joff] = blockUV[jj][ii];
                              }
                              else
                              {
                                get_eighthpel_blockUV_direct(list[ref_idx]->imgUV[uv], i1, j1, img, blockUV, 1, 2, 2);
                                img->mpr[ii+ioff][jj+joff] = blockUV[0][0];
                              }
#ifndef GEOM                  
                            }
                          }
#endif
                        }
                      }
                      else
                      {
#if defined GEOM && GEO_SCHEME == 1
                        if(currMB->geo_mode > -1)
                        {
                          mv_array_geo[0][jf][ifx][0] = geo_mv[LIST_0][segm][0];
                          mv_array_geo[0][jf][ifx][1] = geo_mv[LIST_0][segm][1];
                          fw_mv_array = mv_array_geo[0];

                          mv_array_geo[1][jf][ifx][0] = geo_mv[LIST_1][segm][0];
                          mv_array_geo[1][jf][ifx][1] = geo_mv[LIST_1][segm][1];
                          bw_mv_array = mv_array_geo[1];
                        }
                        else
#endif
                        {
                          fw_mv_array = dec_picture->mv[LIST_0];
                          bw_mv_array = dec_picture->mv[LIST_1];
                        }
#ifndef GEOM                  
                        for(jj=0;jj<4;jj++)
                        {
                          jf=(j4+jj)/(img->mb_cr_size_y/4);     // jf  = Subblock_y-coordinate
                          for(ii=0;ii<4;ii++)
                          {
                            ifx=(i4+ii)/(img->mb_cr_size_x/4);  // ifx = Subblock_x-coordinate
#endif
                            direct_pdir = 2;

                            if (mv_mode == 0 && img->direct_spatial_mv_pred_flag)
                            {
                              //===== DIRECT PREDICTION =====
                              //char ref0=dec_picture->ref_idx[LIST_0][2*(jf/2)][(ifx/2)*2];
                              //char ref1=dec_picture->ref_idx[LIST_1][2*(jf/2)][(ifx/2)*2];
                              char ref0=dec_picture->ref_idx[LIST_0][(jf>>1)<<1][(ifx>>1)<<1];
                              char ref1=dec_picture->ref_idx[LIST_1][(jf>>1)<<1][(ifx>>1)<<1];

                              if (ref0!=-1)
                              {
                                fw_refframe = ref0;
                                fw_ref_idx = fw_refframe;
                              }
                              if (ref1!=-1)
                              {
                                bw_refframe = ref1;
                                bw_ref_idx = bw_refframe;
                              }

                              if      (ref1==-1) direct_pdir = 0;
                              else if (ref0==-1) direct_pdir = 1;

                              if (direct_pdir == 0 || direct_pdir == 2)
                              {
                                i1=(i4+ii)*f1_x+fw_mv_array[jf][ifx][0];
                                j1=(j4+jj)*f1_y+fw_mv_array[jf][ifx][1];

                                if (active_sps->chroma_format_idc == 1)
                                  j1 += listX[0][fw_refframe]->chroma_vector_adjustment;

                                ii0=max (0, min (i1/f1_x,   img->width_cr-1));
                                jj0=max (0, min (j1/f1_y,   max_y_cr));
                                ii1=max (0, min ((i1+f2_x)/f1_x, img->width_cr-1));
                                jj1=max (0, min ((j1+f2_y)/f1_y, max_y_cr));


                                if1=(i1 & f2_x);
                                jf1=(j1 & f2_y);
                                if0=f1_x-if1;
                                jf0=f1_y-jf1;

                                if (!geometry)
                                {
                                  if (!(ii&0x01) && !(jj&0x01)) // ii=0,2 and jj=0,2
                                  {
                                    get_eighthpel_blockUV_direct(listX[0][fw_refframe]->imgUV[uv], i1, j1, img, blockUV_2x2, 0, 2, 2);
                                    blockUV_fw[jj][ii] = blockUV_2x2[0][0];
                                    blockUV_fw[jj][ii+1] = blockUV_2x2[0][1];
                                    blockUV_fw[jj+1][ii] = blockUV_2x2[1][0];
                                    blockUV_fw[jj+1][ii+1] = blockUV_2x2[1][1];
                                  }
                                }


                                if (!geometry)
                                {
                                  fw_pred = blockUV_fw[jj][ii];
                                }
                                else
                                {
                                  get_eighthpel_blockUV_direct(listX[0][fw_refframe]->imgUV[uv], i1, j1, img, blockUV, 1, 2, 2);
                                  fw_pred = blockUV[0][0];
                                }
                              }
                              if (direct_pdir == 1 || direct_pdir == 2)
                              {
                                i1=(i4+ii)*f1_x+bw_mv_array[jf][ifx][0];
                                j1=(j4+jj)*f1_y+bw_mv_array[jf][ifx][1];

                                if (active_sps->chroma_format_idc == 1)
                                  j1 += listX[1][bw_refframe]->chroma_vector_adjustment;

                                ii0=max (0, min (i1/f1_x,   img->width_cr-1));
                                jj0=max (0, min (j1/f1_y,   max_y_cr));
                                ii1=max (0, min ((i1+f2_x)/f1_x, img->width_cr-1));
                                jj1=max (0, min ((j1+f2_y)/f1_y, max_y_cr));

                                if1=(i1 & f2_x);
                                jf1=(j1 & f2_y);
                                if0=f1_x-if1;
                                jf0=f1_y-jf1;

                                if (!geometry)
                                {
                                  if (!(ii&0x01) && !(jj&0x01)) // ii=0,2 and jj=0,2
                                  {
                                    get_eighthpel_blockUV_direct(listX[1][bw_refframe]->imgUV[uv], i1, j1, img, blockUV_2x2, 0, 2, 2);
                                    blockUV_bw[jj][ii] = blockUV_2x2[0][0];
                                    blockUV_bw[jj][ii+1] = blockUV_2x2[0][1];
                                    blockUV_bw[jj+1][ii] = blockUV_2x2[1][0];
                                    blockUV_bw[jj+1][ii+1] = blockUV_2x2[1][1];
                                  }
                                }

                                if (!geometry)
                                {
                                  bw_pred = blockUV_bw[jj][ii];
                                }
                                else
                                {
                                  get_eighthpel_blockUV_direct(listX[1][bw_refframe]->imgUV[uv], i1, j1, img, blockUV, 1, 2, 2);
                                  bw_pred = blockUV[0][0];
                                }
                              }

                            }
                            else
                            {
                              //===== BI-DIRECTIONAL PREDICTION =====
#if defined GEOM && GEO_SCHEME == 1
                              if(currMB->geo_mode > -1)
                              {
                                fw_refframe = ref_idx = best8x8fwref_geo[segm];
                                bw_refframe = ref_idx = best8x8bwref_geo[segm];
                              }
                              else
#endif
                              {
                                fw_refframe = dec_picture->ref_idx[LIST_0][jf][ifx];
                                bw_refframe = dec_picture->ref_idx[LIST_1][jf][ifx];
                              } 
                              fw_ref_idx = fw_refframe;
                              bw_ref_idx = bw_refframe;

                              i1=(i4+ii)*f1_x+fw_mv_array[jf][ifx][0];
                              j1=(j4+jj)*f1_y+fw_mv_array[jf][ifx][1];

                              if (active_sps->chroma_format_idc == 1)
                                j1 += listX[0][fw_refframe]->chroma_vector_adjustment;

                              ii0=max (0, min (i1/f1_x,   img->width_cr-1));
                              jj0=max (0, min (j1/f1_y,   max_y_cr));
                              ii1=max (0, min ((i1+f2_x)/f1_x, img->width_cr-1));
                              jj1=max (0, min ((j1+f2_y)/f1_y, max_y_cr));

                              if1=(i1 & f2_x);
                              jf1=(j1 & f2_y);
                              if0=f1_x-if1;
                              jf0=f1_y-jf1;

                              if (!geometry)
                              {
                                if (!(ii&0x01) && !(jj&0x01)) // ii=0,2 and jj=0,2
                                {
                                  get_eighthpel_blockUV_direct(listX[0][fw_refframe]->imgUV[uv], i1, j1, img, blockUV_2x2, 0, 2, 2);
                                  blockUV_fw[jj][ii] = blockUV_2x2[0][0];
                                  blockUV_fw[jj][ii+1] = blockUV_2x2[0][1];
                                  blockUV_fw[jj+1][ii] = blockUV_2x2[1][0];
                                  blockUV_fw[jj+1][ii+1] = blockUV_2x2[1][1];
                                }
                              }

                              if (!geometry)
                              {
                                fw_pred = blockUV_fw[jj][ii];
                              }
                              else
                              {
                                get_eighthpel_blockUV_direct(listX[0][fw_refframe]->imgUV[uv], i1, j1, img, blockUV, 1, 2, 2);
                                fw_pred = blockUV[0][0];
                              }
                              i1=(i4+ii)*f1_x+bw_mv_array[jf][ifx][0];
                              j1=(j4+jj)*f1_y+bw_mv_array[jf][ifx][1];

                              if (active_sps->chroma_format_idc == 1)
                                j1 += listX[1][bw_refframe]->chroma_vector_adjustment;

                              ii0=max (0, min (i1/f1_x,   img->width_cr-1));
                              jj0=max (0, min (j1/f1_y,   max_y_cr));
                              ii1=max (0, min ((i1+f2_x)/f1_x, img->width_cr-1));
                              jj1=max (0, min ((j1+f2_y)/f1_y, max_y_cr));

                              if1=(i1 & f2_x);
                              jf1=(j1 & f2_y);
                              if0=f1_x-if1;
                              jf0=f1_y-jf1;


                              if (!geometry)
                              {
                                if (!(ii&0x01) && !(jj&0x01)) // ii=0,2 and jj=0,2
                                {
                                  get_eighthpel_blockUV_direct(listX[1][bw_refframe]->imgUV[uv], i1, j1, img, blockUV_2x2, 0, 2, 2);
                                  blockUV_bw[jj][ii] = blockUV_2x2[0][0];
                                  blockUV_bw[jj][ii+1] = blockUV_2x2[0][1];
                                  blockUV_bw[jj+1][ii] = blockUV_2x2[1][0];
                                  blockUV_bw[jj+1][ii+1] = blockUV_2x2[1][1];
                                }
                              }

                              if (!geometry)
                              {
                                bw_pred = blockUV_bw[jj][ii];
                              }
                              else
                              {
                                get_eighthpel_blockUV_direct(listX[1][bw_refframe]->imgUV[uv], i1, j1, img, blockUV, 1, 2, 2);
                                bw_pred = blockUV[0][0];
                              }
                            }

                            if (img->apply_weights)
                            {
                              if (img->direct_spatial_mv_pred_flag && direct_pdir==1)
                              {
                                img->mpr[ii+ioff][jj+joff]= Clip1_Chr(((img->wp_weight[1][bw_ref_idx][uv+1] * bw_pred  + img->wp_round_chroma)>>img->chroma_log2_weight_denom) + img->wp_offset[1][bw_ref_idx][uv+1]);   // Replaced with integer only operations
                              }
                              else if (img->direct_spatial_mv_pred_flag && direct_pdir==0)
                              {
                                img->mpr[ii+ioff][jj+joff]=Clip1_Chr(((img->wp_weight[0][fw_ref_idx][uv+1] * fw_pred + img->wp_round_chroma)>>img->chroma_log2_weight_denom) + img->wp_offset[0][fw_ref_idx][uv+1]);   // Replaced with integer only operations
                              }
                              else
                              {


                                int alpha_fw = img->wbp_weight[0][fw_ref_idx][bw_ref_idx][uv+1];
                                int alpha_bw = img->wbp_weight[1][fw_ref_idx][bw_ref_idx][uv+1];

                                if(img->use_high_precision_flag != 0)
                                {
                                  img->mpr[ii+ioff][jj+joff]= Clip1_Chr(((alpha_fw * fw_pred + alpha_bw * bw_pred  + (1<<img->chroma_log2_weight_denom) - img->bipred_rounding_control) >> (img->chroma_log2_weight_denom + 1))+ ((img->wp_offset[0][fw_ref_idx][uv+1] + img->wp_offset[1][bw_ref_idx][uv+1] + 1)>>1) );
                                }
                                else
                                {
                                  img->mpr[ii+ioff][jj+joff]= Clip1_Chr(((alpha_fw * fw_pred + alpha_bw * bw_pred  + (1<<img->chroma_log2_weight_denom)) >> (img->chroma_log2_weight_denom + 1))+ ((img->wp_offset[0][fw_ref_idx][uv+1] + img->wp_offset[1][bw_ref_idx][uv+1] + 1)>>1) );
                                }
                              }
                            }
                            else
                            {
                              if (img->direct_spatial_mv_pred_flag && direct_pdir==1)
                              {
                                img->mpr[ii+ioff][jj+joff]=bw_pred;
                              }
                              else if (img->direct_spatial_mv_pred_flag && direct_pdir==0)
                              {
                                img->mpr[ii+ioff][jj+joff]=fw_pred;
                              } 
                              else
                              {
                                if(img->use_high_precision_flag != 0)
                                {
                                  img->mpr[ii+ioff][jj+joff]=(fw_pred + bw_pred + 1 - img->bipred_rounding_control)/2;
                                }
                                else
                                {
                                  img->mpr[ii+ioff][jj+joff]=(fw_pred + bw_pred + 1 )/2;
                                }
                              }
                            }
                          }
#ifdef OBMC
                          if(segm1==1)
                            mprUV2[segmi] = img->mpr[ii+ioff][jj+joff];
                        }//segmi
                        if(segm1==1)
                        {
                          if(curr_segm == 0)
                            img->mpr[ii+ioff][jj+joff] = (mprUV2[0]*2 + mprUV2[1])/3;
                          else
                            img->mpr[ii+ioff][jj+joff] = (mprUV2[0]   + mprUV2[1]*2)/3;
                        }
#endif
                      }//ii
                    }//jj

                  } //if (!intra_prediction)

                  if (!smb)
                  {
                    itrans(img,ioff,joff, cofuv_blk_x[yuv][b8+uv_shift][b4], cofuv_blk_y[yuv][b8+uv_shift][b4], 1);
                  }
                }
              }
              move_data_to_dec_picture_chroma_transpose(8,8,img->pix_c_x,img->pix_c_y,0,0,uv);
            }  
          }
        }

        void perform_mc_chroma_MxN(Macroblock *currMB, int x, int y, int size_x, int size_y, int uv)
        {
          int i=0,j=0,ii=0,jj=0,i1=0,j1=0,j4=0,i4=0;
          int ioff,joff;

          int bw_pred=0, fw_pred=0, pred;
          int f1_x, f1_y, f2_x, f2_y, f3, f4;

          short ref_idx, fw_refframe=-1, bw_refframe=-1;
          int mv_mode, pred_dir;
          short fw_ref_idx=-1, bw_ref_idx=-1;

          short  *** mv_array, ***fw_mv_array, ***bw_mv_array;

          StorablePicture **list;

          int jfx,ifx;
          int yuv = dec_picture->chroma_format_idc - 1;

          int b8=0;
          int b4=x+(y<<1);

          if (dec_picture->chroma_format_idc != YUV400)
          {
            int blockUV[32][32];
            int blockUV_fw[32][32];
            int blockUV_bw[32][32];
            int geometry = 0; //When geometry mode is ON, this should be set to 1.
            // chroma decoding *******************************************************
            f1_x = (8) << img->mv_res;
            f1_y = (8) << img->mv_res;
            f2_x=f1_x-1;
            f2_y=f1_y-1;

            f3=f1_x*f1_y;
            f4=f3>>1;

            joff = subblk_offset_y[yuv][b8][b4];
            ioff = subblk_offset_x[yuv][b8][b4];
            j4=img->pix_c_y+joff;
            i4=img->pix_c_x+ioff;

            mv_mode  = currMB->b8mode[block8x8_idx[yuv][b8][b4]];
            pred_dir = currMB->b8pdir[block8x8_idx[yuv][b8][b4]];
            assert (pred_dir<=2);
            jfx=(j4)>>1;  // jfx = Subblock_y-coordinate
            ifx=(i4)>>1;  // ifx = Subblock_x-coordinate

            if (pred_dir != 2)
            {
              //--- FORWARD/BACKWARD PREDICTION ---
              mv_array = dec_picture->mv[LIST_0 + pred_dir];
              list = listX[0+pred_dir];
              fw_refframe = ref_idx = dec_picture->ref_idx[LIST_0+pred_dir][jfx][ifx];

              i1=(i4)*f1_x+mv_array[jfx][ifx][0];
              j1=(j4)*f1_y+mv_array[jfx][ifx][1];

              get_eighthpel_blockUV_direct(list[ref_idx]->imgUV[uv], i1, j1, img, blockUV, 0, size_x, size_y);
              for(ii=0;ii<size_x;ii++)
                for(jj=0;jj<size_y;jj++)
                {
                  pred = blockUV[jj][ii];
                  if (img->apply_weights)
                    img->mpr[ii+ioff][jj+joff] = Clip1_Chr(((img->wp_weight[pred_dir][ref_idx][uv+1] * pred  + img->wp_round_chroma)>>img->chroma_log2_weight_denom) + img->wp_offset[pred_dir][ref_idx][uv+1]);
                  else
                    img->mpr[ii+ioff][jj+joff] = pred;
                }
            }
            else
            {
              fw_mv_array = dec_picture->mv[LIST_0];
              bw_mv_array = dec_picture->mv[LIST_1];
              //===== BI-DIRECTIONAL PREDICTION =====
              fw_refframe = dec_picture->ref_idx[LIST_0][jfx][ifx];
              bw_refframe = dec_picture->ref_idx[LIST_1][jfx][ifx];
              fw_ref_idx = fw_refframe;
              bw_ref_idx = bw_refframe;

              i1=(i4)*f1_x+fw_mv_array[jfx][ifx][0];
              j1=(j4)*f1_y+fw_mv_array[jfx][ifx][1];
              get_eighthpel_blockUV_direct(listX[0][fw_refframe]->imgUV[uv], i1, j1, img, blockUV_fw, 0, size_x, size_y);

              i1=(i4)*f1_x+bw_mv_array[jfx][ifx][0];
              j1=(j4)*f1_y+bw_mv_array[jfx][ifx][1];
              get_eighthpel_blockUV_direct(listX[1][bw_refframe]->imgUV[uv], i1, j1, img, blockUV_bw, 0, size_x, size_y);

              for(ii=0;ii<size_x;ii++)
                for(jj=0;jj<size_y;jj++)
                {
                  fw_pred = blockUV_fw[jj][ii];
                  bw_pred = blockUV_bw[jj][ii];
                  if (img->apply_weights)
                  {
                    int alpha_fw = img->wbp_weight[0][fw_ref_idx][bw_ref_idx][uv+1];
                    int alpha_bw = img->wbp_weight[1][fw_ref_idx][bw_ref_idx][uv+1];

                    if(img->use_high_precision_flag != 0)
                      img->mpr[ii+ioff][jj+joff]= Clip1_Chr(((alpha_fw * fw_pred + alpha_bw * bw_pred  + (1<<img->chroma_log2_weight_denom) - img->bipred_rounding_control) >> (img->chroma_log2_weight_denom + 1))+ ((img->wp_offset[0][fw_ref_idx][uv+1] + img->wp_offset[1][bw_ref_idx][uv+1] + 1)>>1) );
                    else
                      img->mpr[ii+ioff][jj+joff]= Clip1_Chr(((alpha_fw * fw_pred + alpha_bw * bw_pred  + (1<<img->chroma_log2_weight_denom)) >> (img->chroma_log2_weight_denom + 1))+ ((img->wp_offset[0][fw_ref_idx][uv+1] + img->wp_offset[1][bw_ref_idx][uv+1] + 1)>>1) );
                  }
                  else
                  {
                    if(img->use_high_precision_flag != 0)
                      img->mpr[ii+ioff][jj+joff]=(fw_pred + bw_pred + 1 - img->bipred_rounding_control)>>1;
                    else
                      img->mpr[ii+ioff][jj+joff]=(fw_pred + bw_pred + 1 )>>1;
                  }
                }
            }
          }
        }

        static void get_fullpel_blockUV(imgpel **  ref, int x_pres, int y_pres, int max_x, int max_y, int block[32][32], int size_x, int size_y)
        {
          int i, j;

          for (j = 0; j < size_y; j++)
            for (i = 0; i < size_x; i++)
              block[j][i] = ref[max(0,min(y_pres+j,max_y))][max(0,min(x_pres+i,max_x))];
        }

        static void interp_blockUV_direct_X(imgpel **ref, int pres_x, int pres_y, const int *coefx, int max_x, int max_y, int block[32][32], int size_x, int size_y, int symmetric)
        {
          int i, j, x;
          int result;

          const int *filtP;
          imgpel *imgP;

          // Filter tap = 8, so filter offset = (8/2)-1 = 3 and block size is 4, so (max-3-4)=(max-7)
          if ((pres_y >= 0) && (pres_y < max_y - size_y) && (pres_x >= 3) && (pres_x < max_x - 3 - size_x))
          { //Image Non-Border safe positions...so dont need Clipping
            if (!symmetric)
            {
              for (j = 0; j < size_y; j++)
                for (i = 0; i < size_x; i++)
                {
                  filtP = coefx;
                  imgP = &(ref[pres_y+j][pres_x+i-2]);
                  for (result = 0, x = -2; x < 4; x++)
                    result += (*filtP++) * (*imgP++);
                  block[j][i] = result;
                }
            }
            else
            {
              for (j = 0; j < size_y; j++)
                for (i = 0; i < size_x; i++)
                {
                  filtP = coefx;
                  for (result = 0, x = -2; x <= 0; x++)
                    result += (*filtP++) * (ref[pres_y+j][pres_x+i+x] + ref[pres_y+j][pres_x+i-x+1]);
                  block[j][i] = result;
                }
            }
          }
          else
          { //Image Border positions...so needs Clipping
            if (!symmetric)
            {
              for (j = 0; j < size_y; j++)
                for (i = 0; i < size_x; i++)
                {
                  for (result = 0, x = -2; x < 4; x++)
                    result += ref[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i+x)] * coefx[x+2];
                  block[j][i] = result;
                }
            }
            else
            {
              for (j = 0; j < size_y; j++)
                for (i = 0; i < size_x; i++)
                {
                  for (result = 0, x = -2; x <= 0; x++)
                    result += coefx[x+2] * (ref[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i+x)] +
                    ref[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i-x+1)]);

                  block[j][i] = result;
                }
            }
          }
        }

        static void interp_blockUV_direct_Y(imgpel **ref, int pres_x, int pres_y, const int *coefy, int max_x, int max_y, int block[32][32], int size_x,int size_y, int symmetric)
        {
          int i, j, y;
          int result;

          const int *filtP;

          // Filter tap = 8, so filter offset = (8/2)-1 = 3 and block size is 4, so (max-3-4)=(max-7)
          if ((pres_y >= 3) && (pres_y < max_y - 3 - size_y) && (pres_x >= 0) && (pres_x < max_x - size_x))
          { //Image Non-Border safe positions...so dont need Clipping
            if (!symmetric)
            {
              for (j = 0; j < size_y; j++)
                for (i = 0; i < size_x; i++)
                {
                  filtP = coefy;
                  for (result = 0, y = -2; y < 4; y++)
                    result += ref[pres_y+j+y][pres_x+i] * (*filtP++);
                  block[j][i] = result;

                }
            }
            else
            {
              for (j = 0; j < size_y; j++)
                for (i = 0; i < size_x; i++)
                {
                  filtP = coefy;
                  for (result = 0, y = -2; y <=0; y++)
                    result += (*filtP++) * (ref[pres_y+j+y][pres_x+i] +
                    ref[pres_y+j-y+1][pres_x+i]);
                  block[j][i] = result;
                }
            }
          }
          else
          { //Image Border positions...so needs Clipping
            if (!symmetric)
            {
              for (j = 0; j < size_y; j++)
                for (i = 0; i < size_x; i++)
                {
                  filtP = coefy;
                  for (result = 0, y = -2; y < 4; y++)
                    result += ref[Clip3(0, max_y, pres_y+j+y)][Clip3(0, max_x, pres_x+i)] * (*filtP++);
                  block[j][i] = result;
                }
            }
            else
            {
              for (j = 0; j < size_y; j++)
                for (i = 0; i < size_x; i++)
                {
                  filtP = coefy;
                  for (result = 0, y = -2; y <= 0; y++)
                    result += (*filtP++) * (ref[Clip3(0, max_y, pres_y+j+y)][Clip3(0, max_x, pres_x+i)] +
                    ref[Clip3(0, max_y, pres_y+j-y+1)][Clip3(0, max_x, pres_x+i)]);
                  block[j][i] = result;
                }
            }
          }

        }

        static void get_eighthpel_blockUV_direct(imgpel **ref, int x_pos, int y_pos, ImageParameters *img, int block[32][32], int geometry, int size_x, int size_y)
        {
          int i, j;

          static const int COEF_6tap[3][6] =
          {
            {  2,  -5, 28,  9, -3, 1 },
            {  2,  -6, 20, 20, -6, 2 },
            {  1,  -3,  9, 28, -5, 2 }
          };

          static const int COEF_6tap_EIGHTH_PEL[4][6] =
          {
            {  2, -5,  60,  9, -3,  1 },
            {  4, -11, 48, 29, -9,  3 },
            {  3, -9,  29, 48, -11, 4 },
            {  1, -3,   9, 60, -5,  2 }
          };

          /*
          // Filters summing up to 256
          static const int COEF_6tap[3][6] =
          {
          { 11, -36, 226,  71, -24,  8 },
          { 10, -40, 158, 158, -40, 10 },
          { 8, -24,  71, 226, -36, 11 }
          };

          static const int COEF_6tap_EIGHTH_PEL[4][6] =
          {
          { 11, -36, 482,  71, -24,  8 },
          { 21, -76, 384, 229, -64, 18 },
          { 18, -64, 229, 384, -76, 21 },
          { 8,  -24,  71, 482, -36, 11 }
          };
          */

          int chroma_size_x = geometry?1:size_x; // If the geometry mode is ON, only 1x1 block is processed. Otherwise size_x x size_y. 
          int chroma_size_y = geometry?1:size_y; // If the geometry mode is ON, only 1x1 block is processed. Otherwise size_x x size_y. 

#ifdef EIGHTH_PEL
          int dx = (x_pos>>img->mv_res) & 7, x = 0;
          int dy = (y_pos>>img->mv_res) & 7, y = 0;
          int pres_x = x_pos >> (3+img->mv_res);
          int pres_y = y_pos >> (3+img->mv_res);
#else
          int dx = x_pos & 7, x = 0;
          int dy = y_pos & 7, y = 0;
          int pres_x = x_pos >> 3;
          int pres_y = y_pos >> 3;
#endif

          //int dx = x_pos & 7;
          //int dy = y_pos & 7;

          //int pres_x = x_pos >> 3;
          //int pres_y = y_pos >> 3;


          int tmp[32][32+16]; // 16 is a conservative choice
          int result;

          int max_x = img->width/2 - 1;
          int max_y = img->height/2 - 1;

          int round5 = 16;
          int round6 = 32;
          int round10 = (1<<9);
          int round11 = (1<<10);
          int round12 = (1<<11);

          int filt_index, symmetric;
          const int *filtP;
          imgpel *imgP;
          int *intP;

          // choose filter depending on the sub-pel position
          if (dx == 0 && dy == 0)
          { // full-pel position, just return full-pel block
            get_fullpel_blockUV(ref, pres_x, pres_y, max_x, max_y, block, chroma_size_x, chroma_size_y);
          }
          else if (dy == 0)
          { // horizontal interpolation only
            if (dx&1)
            { // 1/8-pel position
              interp_blockUV_direct_X(ref, pres_x, pres_y, COEF_6tap_EIGHTH_PEL[dx>>1], max_x, max_y, block, chroma_size_x, chroma_size_y, 0);
              for (j = 0; j < chroma_size_y; j++)
                for (i = 0; i < chroma_size_x; i++)
                  block[j][i] = Clip3(0, img->max_imgpel_value, (block[j][i] + round6) >> 6); // Filter sums up to 64 for 1/8 pel
            }
            else
            {
              filt_index = (dx>>1)-1;
              symmetric = (filt_index == 1);
              interp_blockUV_direct_X(ref, pres_x, pres_y, COEF_6tap[filt_index], max_x, max_y, block, chroma_size_x, chroma_size_y, symmetric);
              for (j = 0; j < chroma_size_y; j++)
                for (i = 0; i < chroma_size_x; i++)
                  block[j][i] = Clip3(0, img->max_imgpel_value, (block[j][i] + round5) >> 5); // Filter sums up to 32 for 1/4 pel
            }
          }
          else if (dx == 0)
          { // vertical interpolation only
            if (dy&1)
            { // 1/8-pel position (vertical)
              interp_blockUV_direct_Y(ref, pres_x, pres_y, COEF_6tap_EIGHTH_PEL[dy>>1], max_x, max_y, block, chroma_size_x, chroma_size_y, 0);
              for (j = 0; j < chroma_size_y; j++)
                for (i = 0; i < chroma_size_x; i++)
                  block[j][i] = Clip3(0, img->max_imgpel_value, (block[j][i] + round6) >> 6); // Filter sums up to 64 for 1/8 pel
            }
            else
            {
              filt_index = (dy>>1)-1;
              symmetric = (filt_index == 1);
              interp_blockUV_direct_Y(ref, pres_x, pres_y, COEF_6tap[filt_index], max_x, max_y, block, size_x, size_y, symmetric);
              for (j = 0; j < chroma_size_y; j++)
                for (i = 0; i < chroma_size_x; i++)
                  block[j][i] = Clip3(0, img->max_imgpel_value, (block[j][i] + round5) >> 5); // Filter sums up to 32 for 1/4 pel
            }
          }
          else if (!(dx&1))
          {  // horizontal 1/4-pel
            filt_index = (dx>>1)-1;
            if ((pres_y >= 3) && (pres_y < max_y-3-size_y) && (pres_x >= 3) && (pres_x < max_x-3-size_x))
            { //Image Non-Border safe positions...so dont need Clipping
              if (filt_index != 1)
              {
                for (j = -2; j < (chroma_size_y+4); j++)
                  for (i = 0; i < chroma_size_x; i++)
                  {
                    imgP = &(ref[pres_y+j][pres_x+i-2]);
                    filtP = COEF_6tap[filt_index];
                    for (tmp[i][j+2] = 0, x = -2; x < 4; x++)
                      tmp[i][j+2] += (*imgP++) * (*filtP++); // indices (i,j) reversed to facilitate further filtering
                  }
              }
              else
              {
                for (j = -2; j < (chroma_size_y+4); j++)
                  for (i = 0; i < chroma_size_x; i++)
                  {
                    filtP = COEF_6tap[filt_index];
                    for (tmp[i][j+2] = 0, x = -2; x <= 0; x++)
                      tmp[i][j+2] += (*filtP++) * (ref[pres_y+j][pres_x+i+x] + ref[pres_y+j][pres_x+i-x+1]);
                    // indices (i,j) reversed to facilitate further filtering
                  }
              }
            }
            else
            {//Image Border positions...so needs Clipping
              if (filt_index != 1)
              {
                for (j = -2; j < (chroma_size_y+4); j++)
                  for (i = 0; i < chroma_size_x; i++)
                  {
                    for (tmp[i][j+2] = 0, x = -2; x < 4; x++)
                      tmp[i][j+2] += ref[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i+x )] * COEF_6tap[filt_index][x+2];
                    // indices (i,j) reversed to facilitate further filtering
                  }
              }
              else
              {
                for (j = -2; j < (chroma_size_y+4); j++)
                  for (i = 0; i < chroma_size_x; i++)
                  {
                    for (tmp[i][j+2] = 0, x = -2; x <= 0; x++)
                      tmp[i][j+2] += COEF_6tap[filt_index][2+x] * (ref[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i+x)] +
                      ref[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i-x+1)]);
                    // indices (i,j) reversed to facilitate further filtering
                  }
              }
            }
            // At this point the nominal bit-depth has been increased by 5 (1/4-pel filtering, Filter sums up to 32).

            if (dy&1)
            { // 1/8-pel position (vertical)
              filt_index = (dy>>1);
              for (j = 0; j < chroma_size_y; j++)
                for (i = 0; i < chroma_size_x; i++)
                {
                  intP = &(tmp[i][j]);
                  filtP = COEF_6tap_EIGHTH_PEL[filt_index];
                  for (result = 0, y = -2; y < 4; y++)
                    result += (*filtP++) * (*intP++);
                  block[j][i] = Clip3(0, img->max_imgpel_value, (result + round11) >> 11);
                  // 1/8-pel filtering, Filter sums up to 512. Total nominal bit-depth increased by 17.
                }
            }
            else
            { // 1/4-pel position (vertical)
              filt_index = (dy>>1)-1;
              if (filt_index != 1)
              {
                for (j = 0; j < chroma_size_y; j++)
                  for (i = 0; i < chroma_size_x; i++)
                  {
                    intP = &(tmp[i][j]);
                    filtP = COEF_6tap[filt_index];
                    for (result = 0, y = -2; y < 4; y++)
                      result += (*filtP++) * (*intP++);
                    block[j][i] = Clip3(0, img->max_imgpel_value, (result + round10) >> 10);
                    // 1/4-pel filtering, Filter sums up to 256. Total nominal bit-depth increased by 16.
                  }
              }
              else
              {
                for (j = 0; j < chroma_size_y; j++)
                  for (i = 0; i < chroma_size_x; i++)
                  {
                    filtP = COEF_6tap[filt_index];
                    for (result = 0, y = -2; y <= 0; y++)
                      result += (*filtP++) * (tmp[i][j+y+2] + tmp[i][j-y+1+2]);
                    block[j][i] = Clip3(0, img->max_imgpel_value, (result + round10) >> 10);
                    // 1/4-pel filtering, Filter sums up to 256. Total nominal bit-depth increased by 16.
                  }
              }
            }
          }
          else if (!(dy&1))
          { // vertical 1/4-pel and horizontal 1/8-pel interpolation
            if ((pres_y >= 3) && (pres_y < max_y-3-size_y) && (pres_x >= 3) && (pres_x < max_x-3-size_x))
            { //Image Non-Border safe positions...so don't need Clipping
              filt_index = (dy>>1)-1;
              if (filt_index != -1)
              {
                for (j = 0; j < chroma_size_y; j++) // y
                  for (i = -2; i < chroma_size_x+4; i++) // x
                  { // get vertical 1/4-pel positions
                    filtP = COEF_6tap[filt_index];
                    for (tmp[j][i+2] = 0, y = -2; y < 4; y++)
                      tmp[j][i+2] += (*filtP++) * ref[pres_y+j+y][pres_x+i];
                  }
              }
              else
              {
                for (j = 0; j < chroma_size_y; j++) // y
                  for (i = -2; i < chroma_size_x+4; i++) // x
                  { // get vertical 1/4-pel positions
                    filtP = COEF_6tap[filt_index];
                    for (tmp[j][i+2] = 0, y = -2; y <= 0; y++)
                      tmp[j][i+2] += (*filtP++) * (ref[pres_y+j+y][pres_x+i] +
                      ref[pres_y+j-y+1][pres_x+i]);
                  }
              }
            }
            else
            { //Image Border positions...so needs Clipping
              filt_index = (dy>>1)-1;
              if (filt_index != -1)
              {
                for (j = 0; j < chroma_size_y; j++) // y
                  for (i = -2; i < chroma_size_x+4; i++) // x
                  { // get vertical 1/4-pel positions
                    filtP = COEF_6tap[filt_index];
                    for (tmp[j][i+2] = 0, y = -2; y < 4; y++)
                      tmp[j][i+2] += (*filtP++) * ref[Clip3(0, max_y, pres_y+j+y)][Clip3(0, max_x, pres_x+i)];
                  }
              }
              else
              {
                for (j = 0; j < chroma_size_y; j++) // y
                  for (i = -2; i < chroma_size_x + 4; i++) // x
                  { // get vertical 1/4-pel positions
                    filtP = COEF_6tap[filt_index];
                    for (tmp[j][i+2] = 0, y = -2; y <= 0; y++)
                      tmp[j][i+2] += (*filtP++) * (ref[Clip3(0, max_y, pres_y+j+y)][Clip3(0, max_x, pres_x+i)] +
                      ref[Clip3(0, max_y, pres_y+j-y+1)][Clip3(0, max_x, pres_x+i)]);
                  }
              }
            }
            // At this point the nominal bit-depth has been increased by 8 (1/4-pel filtering, Filter sums up to 256).

            // horizontal 1/8 pel position filtering
            filt_index = (dx>>1);
            for (j = 0; j < chroma_size_y; j++)
              for (i = 0; i < chroma_size_x; i++)
              {
                intP = &(tmp[j][i]);
                filtP = COEF_6tap_EIGHTH_PEL[filt_index];
                for (result = 0, y = -2; y < 4; y++)
                  result += (*filtP++) * (*intP++);
                block[j][i] = Clip3(0, img->max_imgpel_value, (result + round11) >> 11);
                // 1/8-pel filtering, Filter sums up to 64. Total nominal bit-depth increased by 11.
              }
          }       
          else
          { // 1/8-1/8 positions
            // Horizontal 1/8 pel position filtering
            filt_index = (dx>>1);
            if ((pres_y >= 3) && (pres_y < max_y-3-size_y) && (pres_x >= 3) && (pres_x < max_x-3-size_x))
            { //Image Non-Border safe positions...so dont need Clipping
              for (j = -2; j < (chroma_size_y+4); j++)
                for (i = 0; i < chroma_size_x; i++)
                {
                  imgP = &(ref[pres_y+j][pres_x+i-2]);
                  filtP = COEF_6tap_EIGHTH_PEL[filt_index];
                  for (tmp[i][j+2] = 0, x = -2; x < 4; x++)
                    tmp[i][j+2] += (*imgP++) * (*filtP++);
                }
            }
            else
            {//Image Border positions...so needs Clipping
              for (j = -2; j < (chroma_size_y+4); j++)
                for (i = 0; i < chroma_size_x; i++)
                {
                  for (tmp[i][j+2] = 0, x = -2; x < 4; x++)
                    tmp[i][j+2] += ref[Clip3(0, max_y, pres_y+j)][Clip3(0, max_x, pres_x+i+x)] * COEF_6tap_EIGHTH_PEL[filt_index][x+2];
                }
            }
            // At this point the nominal bit-depth has been increased by 6 (1/8-pel filtering, Filter sums up to 64).

            // vertical 1/8 pel position filtering
            filt_index = (dy>>1);
            for (j = 0; j < chroma_size_y; j++)
              for (i = 0; i < chroma_size_x; i++)
              {
                intP = &(tmp[i][j]);
                filtP = COEF_6tap_EIGHTH_PEL[filt_index];
                for (result = 0, y = -2; y < 4; y++)
                  result += (*filtP++) * (*intP++);
                block[j][i] = Clip3(0, img->max_imgpel_value, (result + round12) >> 12);
              }

          }
        }



