/* 
====================================================================================================================

The copyright in this software is being made available under the License included below. 
This software may be subject to other third party and 	contributor rights, including patent rights, and no such 
rights are granted under this license.

Copyright (c) 2010, QUALCOMM INC.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted only for 
the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
promoting such standards. The following conditions are required to be met:

* Redistributions of source code must retain the above copyright notice, this list of conditions and 
the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
the following disclaimer in the documentation and/or other materials provided with the distribution.
* The name of QUALCOMM INC.  may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

====================================================================================================================
*/

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <string.h>
#include <math.h>

#include "global.h"

#include "image.h"
#include "mb_access.h"
//#include "block.h"
#include "vlc.h"

#include "memalloc.h"




//! single scan pattern
const byte SNGL_SCAN[16][2] =
{
  {0,0},{1,0},{0,1},{0,2},
  {1,1},{2,0},{3,0},{2,1},
  {1,2},{0,3},{1,3},{2,2},
  {3,1},{3,2},{2,3},{3,3}
};

//! array used to find expencive coefficients
const byte COEFF_COST[2][16] =
{
  {3,2,2,1,1,1,0,0,0,0,0,0,0,0,0,0},
  {9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9}
};

//! single scan pattern
const byte SNGL_SCAN8x8[64][2] = {
  {0,0}, {1,0}, {0,1}, {0,2}, {1,1}, {2,0}, {3,0}, {2,1}, 
  {1,2}, {0,3}, {0,4}, {1,3}, {2,2}, {3,1}, {4,0}, {5,0},
  {4,1}, {3,2}, {2,3}, {1,4}, {0,5}, {0,6}, {1,5}, {2,4},
  {3,3}, {4,2}, {5,1}, {6,0}, {7,0}, {6,1}, {5,2}, {4,3},
  {3,4}, {2,5}, {1,6}, {0,7}, {1,7}, {2,6}, {3,5}, {4,4},
  {5,3}, {6,2}, {7,1}, {7,2}, {6,3}, {5,4}, {4,5}, {3,6},
  {2,7}, {3,7}, {4,6}, {5,5}, {6,4}, {7,3}, {7,4}, {6,5},
  {5,6}, {4,7}, {5,7}, {6,6}, {7,5}, {7,6}, {6,7}, {7,7}
};

//! array used to find expensive coefficients
const byte COEFF_COST8x8[2][64] =
{
  {3,3,3,3,2,2,2,2,2,2,2,2,1,1,1,1,
  1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
  {9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,
  9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,
  9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,
  9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9}
};

//! bit cost for coefficients
const byte COEFF_BIT_COST[3][16][16]=
{
  { // 2x2 scan 
    { 3, 5, 7, 9, 9,11,11,11,11,13,13,13,13,13,13,13},
    { 5, 7, 9, 9,11,11,11,11,13,13,13,13,13,13,13,13},
    { 7, 9, 9,11,11,11,11,13,13,13,13,13,13,13,13,15},
    { 7, 9, 9,11,11,11,11,13,13,13,13,13,13,13,13,15},
    { 7, 7, 9, 9, 9, 9,11,11,11,11,11,11,11,11,13,13},
    { 7, 7, 9, 9, 9, 9,11,11,11,11,11,11,11,11,13,13},
    { 7, 7, 9, 9, 9, 9,11,11,11,11,11,11,11,11,13,13},
    { 7, 7, 9, 9, 9, 9,11,11,11,11,11,11,11,11,13,13},
    { 7, 7, 9, 9, 9, 9,11,11,11,11,11,11,11,11,13,13},
    { 7, 7, 9, 9, 9, 9,11,11,11,11,11,11,11,11,13,13},
    { 7, 7, 9, 9, 9, 9,11,11,11,11,11,11,11,11,13,13},
    { 7, 7, 9, 9, 9, 9,11,11,11,11,11,11,11,11,13,13},
    { 7, 7, 9, 9, 9, 9,11,11,11,11,11,11,11,11,13,13},
    { 7, 7, 9, 9, 9, 9,11,11,11,11,11,11,11,11,13,13},
    { 7, 7, 9, 9, 9, 9,11,11,11,11,11,11,11,11,13,13},
    { 7, 7, 9, 9, 9, 9,11,11,11,11,11,11,11,11,13,13},
  },
  {  // double scan
    { 3, 5, 7, 7, 7, 9, 9, 9, 9,11,11,13,13,13,13,15},
    { 5, 9, 9,11,11,13,13,13,13,15,15,15,15,15,15,15},
    { 7,11,11,13,13,13,13,15,15,15,15,15,15,15,15,17},
    { 9,11,11,13,13,13,13,15,15,15,15,15,15,15,15,17},
    { 9,11,11,13,13,13,13,15,15,15,15,15,15,15,15,17},
    {11,11,13,13,13,13,15,15,15,15,15,15,15,15,17,17},
    {11,11,13,13,13,13,15,15,15,15,15,15,15,15,17,17},
    {11,11,13,13,13,13,15,15,15,15,15,15,15,15,17,17},
    {11,11,13,13,13,13,15,15,15,15,15,15,15,15,17,17},
    {11,11,13,13,13,13,15,15,15,15,15,15,15,15,17,17},
    {11,11,13,13,13,13,15,15,15,15,15,15,15,15,17,17},
    {11,11,13,13,13,13,15,15,15,15,15,15,15,15,17,17},
    {11,11,13,13,13,13,15,15,15,15,15,15,15,15,17,17},
    {11,11,13,13,13,13,15,15,15,15,15,15,15,15,17,17},
    },
    {    // single scan
      { 3, 7, 9, 9,11,13,13,15,15,15,15,17,17,17,17,17},
      { 5, 9,11,13,13,15,15,15,15,17,17,17,17,17,17,17},
      { 5, 9,11,13,13,15,15,15,15,17,17,17,17,17,17,17},
      { 7,11,13,13,15,15,15,15,17,17,17,17,17,17,17,17},
      { 7,11,13,13,15,15,15,15,17,17,17,17,17,17,17,17},
      { 7,11,13,13,15,15,15,15,17,17,17,17,17,17,17,17},
      { 9,11,13,13,15,15,15,15,17,17,17,17,17,17,17,17},
      { 9,11,13,13,15,15,15,15,17,17,17,17,17,17,17,17},
      { 9,11,13,13,15,15,15,15,17,17,17,17,17,17,17,17},
      { 9,11,13,13,15,15,15,15,17,17,17,17,17,17,17,17},
      {11,13,13,15,15,15,15,17,17,17,17,17,17,17,17,19},
      {11,13,13,15,15,15,15,17,17,17,17,17,17,17,17,19},
      {11,13,13,15,15,15,15,17,17,17,17,17,17,17,17,19},
      {11,13,13,15,15,15,15,17,17,17,17,17,17,17,17,19},
      {11,13,13,15,15,15,15,17,17,17,17,17,17,17,17,19},
      {11,13,13,15,15,15,15,17,17,17,17,17,17,17,17,19},
    },
};

//! single scan pattern
const byte SCAN_YUV422  [8][2] =
{
  {0,0},{0,1},
  {1,0},{0,2},
  {0,3},{1,1},
  {1,2},{1,3}
};

//! look up tables for FRExt-chroma support
const unsigned char hor_offset[4][4][4] =
{{{0, 0, 0, 0}, 
{0, 0, 0, 0}, 
{0, 0, 0, 0}, 
{0, 0, 0, 0}},

{{0, 4, 0, 4}, 
{0, 0, 0, 0}, 
{0, 0, 0, 0}, 
{0, 0, 0, 0}},

{{0, 4, 0, 4}, 
{0, 4, 0, 4}, 
{0, 0, 0, 0}, 
{0, 0, 0, 0}},

{{0, 4, 0, 4}, 
{8,12, 8,12},
{0, 4, 0, 4},
{8,12, 8,12}}};

const unsigned char ver_offset[4][4][4] =
{ {{0, 0, 0, 0}, 
{0, 0, 0, 0}, 
{0, 0, 0, 0}, 
{0, 0, 0, 0}},             

{{0, 0, 4, 4}, 
{0, 0, 0, 0}, 
{0, 0, 0, 0}, 
{0, 0, 0, 0}},

{{0, 0, 4, 4}, 
{8, 8,12,12}, 
{0, 0, 0, 0}, 
{0, 0, 0, 0}},

{{0, 0, 4, 4},
{0, 0, 4, 4},
{8, 8,12,12},
{8, 8,12,12}}};

static unsigned char cbp_blk_chroma[8][4] =
{ {16, 17, 18, 19},
{20, 21, 22, 23},
{24, 25, 26, 27},
{28, 29, 30, 31},
{32, 33, 34, 35},
{36, 37, 38, 39},
{40, 41, 42, 43},
{44, 45, 46, 47} };


const int quant_coef[6][4][4] = {
  {{13107, 8066,13107, 8066},{ 8066, 5243, 8066, 5243},{13107, 8066,13107, 8066},{ 8066, 5243, 8066, 5243}},
  {{11916, 7490,11916, 7490},{ 7490, 4660, 7490, 4660},{11916, 7490,11916, 7490},{ 7490, 4660, 7490, 4660}},
  {{10082, 6554,10082, 6554},{ 6554, 4194, 6554, 4194},{10082, 6554,10082, 6554},{ 6554, 4194, 6554, 4194}},
  {{ 9362, 5825, 9362, 5825},{ 5825, 3647, 5825, 3647},{ 9362, 5825, 9362, 5825},{ 5825, 3647, 5825, 3647}},
  {{ 8192, 5243, 8192, 5243},{ 5243, 3355, 5243, 3355},{ 8192, 5243, 8192, 5243},{ 5243, 3355, 5243, 3355}},
  {{ 7282, 4559, 7282, 4559},{ 4559, 2893, 4559, 2893},{ 7282, 4559, 7282, 4559},{ 4559, 2893, 4559, 2893}}
};

const int dequant_coef[6][4][4] = {
  {{10, 13, 10, 13},{ 13, 16, 13, 16},{10, 13, 10, 13},{ 13, 16, 13, 16}},
  {{11, 14, 11, 14},{ 14, 18, 14, 18},{11, 14, 11, 14},{ 14, 18, 14, 18}},
  {{13, 16, 13, 16},{ 16, 20, 16, 20},{13, 16, 13, 16},{ 16, 20, 16, 20}},
  {{14, 18, 14, 18},{ 18, 23, 18, 23},{14, 18, 14, 18},{ 18, 23, 18, 23}},
  {{16, 20, 16, 20},{ 20, 25, 20, 25},{16, 20, 16, 20},{ 20, 25, 20, 25}},
  {{18, 23, 18, 23},{ 23, 29, 23, 29},{18, 23, 18, 23},{ 23, 29, 23, 29}}
};
static const int A[4][4] = {
  { 16, 20, 16, 20},
  { 20, 25, 20, 25},
  { 16, 20, 16, 20},
  { 20, 25, 20, 25}
};


const byte QP_SCALE_CR[52]=
{
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,
  12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,
  28,29,29,30,31,32,32,33,34,34,35,35,36,36,37,37,
  37,38,38,38,39,39,39,39
};

#ifdef RDO_Q
const int estErr4x4[6][4][4]=
{
  {
    {25600, 27040, 25600, 27040}, 
    {27040, 25600, 27040, 25600}, 
    {25600, 27040, 25600, 27040}, 
    {27040, 25600, 27040, 25600} 
  },
  {
    {30976, 31360, 30976, 31360}, 
    {31360, 32400, 31360, 32400}, 
    {30976, 31360, 30976, 31360}, 
    {31360, 32400, 31360, 32400} 
  },
  {
    {43264, 40960, 43264, 40960}, 
    {40960, 40000, 40960, 40000}, 
    {43264, 40960, 43264, 40960}, 
    {40960, 40000, 40960, 40000} 
  },
  {
    {50176, 51840, 50176, 51840}, 
    {51840, 52900, 51840, 52900}, 
    {50176, 51840, 50176, 51840}, 
    {51840, 52900, 51840, 52900} 
  },
  {
    {65536, 64000, 65536, 64000}, 
    {64000, 62500, 64000, 62500}, 
    {65536, 64000, 65536, 64000}, 
    {64000, 62500, 64000, 62500} 
  },
  {
    {82944, 84640, 82944, 84640}, 
    {84640, 84100, 84640, 84100}, 
    {82944, 84640, 82944, 84640}, 
    {84640, 84100, 84640, 84100} 
  }
};

const int estErr8x8[6][8][8]={
  {
    {6553600, 6677056, 6400000, 6677056, 6553600, 6677056, 6400000, 6677056}, 
    {6677056, 6765201, 6658560, 6765201, 6677056, 6765201, 6658560, 6765201}, 
    {6400000, 6658560, 6553600, 6658560, 6400000, 6658560, 6553600, 6658560}, 
    {6677056, 6765201, 6658560, 6765201, 6677056, 6765201, 6658560, 6765201}, 
    {6553600, 6677056, 6400000, 6677056, 6553600, 6677056, 6400000, 6677056}, 
    {6677056, 6765201, 6658560, 6765201, 6677056, 6765201, 6658560, 6765201}, 
    {6400000, 6658560, 6553600, 6658560, 6400000, 6658560, 6553600, 6658560}, 
    {6677056, 6765201, 6658560, 6765201, 6677056, 6765201, 6658560, 6765201} 
  },
  {
    {7929856, 8156736, 8028160, 8156736, 7929856, 8156736, 8028160, 8156736}, 
    {8156736, 7537770, 7814560, 7537770, 8156736, 7537770, 7814560, 7537770}, 
    {8028160, 7814560, 7840000, 7814560, 8028160, 7814560, 7840000, 7814560}, 
    {8156736, 7537770, 7814560, 7537770, 8156736, 7537770, 7814560, 7537770}, 
    {7929856, 8156736, 8028160, 8156736, 7929856, 8156736, 8028160, 8156736}, 
    {8156736, 7537770, 7814560, 7537770, 8156736, 7537770, 7814560, 7537770}, 
    {8028160, 7814560, 7840000, 7814560, 8028160, 7814560, 7840000, 7814560}, 
    {8156736, 7537770, 7814560, 7537770, 8156736, 7537770, 7814560, 7537770} 
  },
  {
    {11075584, 10653696, 11151360, 10653696, 11075584, 10653696, 11151360, 10653696}, 
    {10653696, 11045652, 11109160, 11045652, 10653696, 11045652, 11109160, 11045652}, 
    {11151360, 11109160, 11289600, 11109160, 11151360, 11109160, 11289600, 11109160}, 
    {10653696, 11045652, 11109160, 11045652, 10653696, 11045652, 11109160, 11045652}, 
    {11075584, 10653696, 11151360, 10653696, 11075584, 10653696, 11151360, 10653696}, 
    {10653696, 11045652, 11109160, 11045652, 10653696, 11045652, 11109160, 11045652}, 
    {11151360, 11109160, 11289600, 11109160, 11151360, 11109160, 11289600, 11109160}, 
    {10653696, 11045652, 11109160, 11045652, 10653696, 11045652, 11109160, 11045652} 
  },
  {
    {12845056, 12503296, 12544000, 12503296, 12845056, 12503296, 12544000, 12503296}, 
    {12503296, 13050156, 12588840, 13050156, 12503296, 13050156, 12588840, 13050156}, 
    {12544000, 12588840, 12960000, 12588840, 12544000, 12588840, 12960000, 12588840}, 
    {12503296, 13050156, 12588840, 13050156, 12503296, 13050156, 12588840, 13050156}, 
    {12845056, 12503296, 12544000, 12503296, 12845056, 12503296, 12544000, 12503296}, 
    {12503296, 13050156, 12588840, 13050156, 12503296, 13050156, 12588840, 13050156}, 
    {12544000, 12588840, 12960000, 12588840, 12544000, 12588840, 12960000, 12588840}, 
    {12503296, 13050156, 12588840, 13050156, 12503296, 13050156, 12588840, 13050156} 
  },
  {
    {16777216, 16646400, 16384000, 16646400, 16777216, 16646400, 16384000, 16646400}, 
    {16646400, 16370116, 16692640, 16370116, 16646400, 16370116, 16692640, 16370116}, 
    {16384000, 16692640, 16646400, 16692640, 16384000, 16692640, 16646400, 16692640}, 
    {16646400, 16370116, 16692640, 16370116, 16646400, 16370116, 16692640, 16370116}, 
    {16777216, 16646400, 16384000, 16646400, 16777216, 16646400, 16384000, 16646400}, 
    {16646400, 16370116, 16692640, 16370116, 16646400, 16370116, 16692640, 16370116}, 
    {16384000, 16692640, 16646400, 16692640, 16384000, 16692640, 16646400, 16692640}, 
    {16646400, 16370116, 16692640, 16370116, 16646400, 16370116, 16692640, 16370116} 
  },
  {
    {21233664, 21381376, 21667840, 21381376, 21233664, 21381376, 21667840, 21381376}, 
    {21381376, 21381376, 21374440, 21381376, 21381376, 21381376, 21374440, 21381376}, 
    {21667840, 21374440, 21529600, 21374440, 21667840, 21374440, 21529600, 21374440}, 
    {21381376, 21381376, 21374440, 21381376, 21381376, 21381376, 21374440, 21381376}, 
    {21233664, 21381376, 21667840, 21381376, 21233664, 21381376, 21667840, 21381376}, 
    {21381376, 21381376, 21374440, 21381376, 21381376, 21381376, 21374440, 21381376}, 
    {21667840, 21374440, 21529600, 21374440, 21667840, 21374440, 21529600, 21374440}, 
    {21381376, 21381376, 21374440, 21381376, 21381376, 21381376, 21374440, 21381376} 
  }
};

#endif


const int quant_coef8[6][8][8] = 
{
  { 
    {13107, 12222,  16777,  12222,  13107,  12222,  16777,  12222},
    {12222, 11428,  15481,  11428,  12222,  11428,  15481,  11428},
    {16777, 15481,  20972,  15481,  16777,  15481,  20972,  15481},
    {12222, 11428,  15481,  11428,  12222,  11428,  15481,  11428},
    {13107, 12222,  16777,  12222,  13107,  12222,  16777,  12222},
    {12222, 11428,  15481,  11428,  12222,  11428,  15481,  11428},
    {16777, 15481,  20972,  15481,  16777,  15481,  20972,  15481},
    {12222, 11428,  15481,  11428,  12222,  11428,  15481,  11428}
  },
  {
    {11916, 11058,  14980,  11058,  11916,  11058,  14980,  11058},
    {11058, 10826,  14290,  10826,  11058,  10826,  14290,  10826},
    {14980, 14290,  19174,  14290,  14980,  14290,  19174,  14290},
    {11058, 10826,  14290,  10826,  11058,  10826,  14290,  10826},
    {11916, 11058,  14980,  11058,  11916,  11058,  14980,  11058},
    {11058, 10826,  14290,  10826,  11058,  10826,  14290,  10826},
    {14980, 14290,  19174,  14290,  14980,  14290,  19174,  14290},
    {11058, 10826,  14290,  10826,  11058,  10826,  14290,  10826}
  },
  {
    {10082, 9675,   12710,  9675,   10082,  9675, 12710,  9675},
    {9675,  8943,   11985,  8943,   9675,   8943, 11985,  8943},
    {12710, 11985,  15978,  11985,  12710,  11985,  15978,  11985},
    {9675,  8943,   11985,  8943,   9675,   8943, 11985,  8943},
    {10082, 9675,   12710,  9675,   10082,  9675, 12710,  9675},
    {9675,  8943,   11985,  8943,   9675, 8943, 11985,  8943},
    {12710, 11985,  15978,  11985,  12710,  11985,  15978,  11985},
    {9675,  8943,   11985,  8943,   9675, 8943, 11985,  8943}
  },
  {
    {9362,  8931, 11984,  8931, 9362, 8931, 11984,  8931},
    {8931,  8228, 11259,  8228, 8931, 8228, 11259,  8228},
    {11984, 11259,  14913,  11259,  11984,  11259,  14913,  11259},
    {8931,  8228, 11259,  8228, 8931, 8228, 11259,  8228},
    {9362,  8931, 11984,  8931, 9362, 8931, 11984,  8931},
    {8931,  8228, 11259,  8228, 8931, 8228, 11259,  8228},
    {11984, 11259,  14913,  11259,  11984,  11259,  14913,  11259},
    {8931,  8228, 11259,  8228, 8931, 8228, 11259,  8228}
  },
  {
    {8192,  7740, 10486,  7740, 8192, 7740, 10486,  7740},
    {7740,  7346, 9777, 7346, 7740, 7346, 9777, 7346},
    {10486, 9777, 13159,  9777, 10486,  9777, 13159,  9777},
    {7740,  7346, 9777, 7346, 7740, 7346, 9777, 7346},
    {8192,  7740, 10486,  7740, 8192, 7740, 10486,  7740},
    {7740,  7346, 9777, 7346, 7740, 7346, 9777, 7346},
    {10486, 9777, 13159,  9777, 10486,  9777, 13159,  9777},
    {7740,  7346, 9777, 7346, 7740, 7346, 9777, 7346}
  },
  {
    {7282,  6830, 9118, 6830, 7282, 6830, 9118, 6830},
    {6830,  6428, 8640, 6428, 6830, 6428, 8640, 6428},
    {9118,  8640, 11570,  8640, 9118, 8640, 11570,  8640},
    {6830,  6428, 8640, 6428, 6830, 6428, 8640, 6428},
    {7282,  6830, 9118, 6830, 7282, 6830, 9118, 6830},
    {6830,  6428, 8640, 6428, 6830, 6428, 8640, 6428},
    {9118,  8640, 11570,  8640, 9118, 8640, 11570,  8640},
    {6830,  6428, 8640, 6428, 6830, 6428, 8640, 6428}
  }
};

const int dequant_coef8[6][8][8] = 
{
  {
    {20,  19, 25, 19, 20, 19, 25, 19},
    {19,  18, 24, 18, 19, 18, 24, 18},
    {25,  24, 32, 24, 25, 24, 32, 24},
    {19,  18, 24, 18, 19, 18, 24, 18},
    {20,  19, 25, 19, 20, 19, 25, 19},
    {19,  18, 24, 18, 19, 18, 24, 18},
    {25,  24, 32, 24, 25, 24, 32, 24},
    {19,  18, 24, 18, 19, 18, 24, 18}
  },
  {
    {22,  21, 28, 21, 22, 21, 28, 21},
    {21,  19, 26, 19, 21, 19, 26, 19},
    {28,  26, 35, 26, 28, 26, 35, 26},
    {21,  19, 26, 19, 21, 19, 26, 19},
    {22,  21, 28, 21, 22, 21, 28, 21},
    {21,  19, 26, 19, 21, 19, 26, 19},
    {28,  26, 35, 26, 28, 26, 35, 26},
    {21,  19, 26, 19, 21, 19, 26, 19}
  },
  {
    {26,  24, 33, 24, 26, 24, 33, 24},
    {24,  23, 31, 23, 24, 23, 31, 23},
    {33,  31, 42, 31, 33, 31, 42, 31},
    {24,  23, 31, 23, 24, 23, 31, 23},
    {26,  24, 33, 24, 26, 24, 33, 24},
    {24,  23, 31, 23, 24, 23, 31, 23},
    {33,  31, 42, 31, 33, 31, 42, 31},
    {24,  23, 31, 23, 24, 23, 31, 23}
  },
  {
    {28,  26, 35, 26, 28, 26, 35, 26},
    {26,  25, 33, 25, 26, 25, 33, 25},
    {35,  33, 45, 33, 35, 33, 45, 33},
    {26,  25, 33, 25, 26, 25, 33, 25},
    {28,  26, 35, 26, 28, 26, 35, 26},
    {26,  25, 33, 25, 26, 25, 33, 25},
    {35,  33, 45, 33, 35, 33, 45, 33},
    {26,  25, 33, 25, 26, 25, 33, 25}
  },
  {
    {32,  30, 40, 30, 32, 30, 40, 30},
    {30,  28, 38, 28, 30, 28, 38, 28},
    {40,  38, 51, 38, 40, 38, 51, 38},
    {30,  28, 38, 28, 30, 28, 38, 28},
    {32,  30, 40, 30, 32, 30, 40, 30},
    {30,  28, 38, 28, 30, 28, 38, 28},
    {40,  38, 51, 38, 40, 38, 51, 38},
    {30,  28, 38, 28, 30, 28, 38, 28}
  },
  {
    {36,  34, 46, 34, 36, 34, 46, 34},
    {34,  32, 43, 32, 34, 32, 43, 32},
    {46,  43, 58, 43, 46, 43, 58, 43},
    {34,  32, 43, 32, 34, 32, 43, 32},
    {36,  34, 46, 34, 36, 34, 46, 34},
    {34,  32, 43, 32, 34, 32, 43, 32},
    {46,  43, 58, 43, 46, 43, 58, 43},
    {34,  32, 43, 32, 34, 32, 43, 32}
  }

};




void forward4x4(int block[16][16], int tblock[16][16], int pos_y, int pos_x)
{
  int i, ii;  
  int tmp[16];
  int *pTmp = tmp, *pblock;
  int p0,p1,p2,p3;
  int t0,t1,t2,t3;

  // Horizontal
  for (i=pos_y; i < pos_y + BLOCK_SIZE; i++)
  {
    pblock = &block[i][pos_x];
    p0 = *(pblock++);
    p1 = *(pblock++);
    p2 = *(pblock++);
    p3 = *(pblock  );

    t0 = p0 + p3;
    t1 = p1 + p2;
    t2 = p1 - p2;
    t3 = p0 - p3;

    *(pTmp++) =  t0 + t1;
    *(pTmp++) = (t3 << 1) + t2;
    *(pTmp++) =  t0 - t1;    
    *(pTmp++) =  t3 - (t2 << 1);
  }

  // Vertical 
  for (i=0; i < BLOCK_SIZE; i++)
  {
    pTmp = tmp + i;
    p0 = *pTmp;
    p1 = *(pTmp += BLOCK_SIZE);
    p2 = *(pTmp += BLOCK_SIZE);
    p3 = *(pTmp += BLOCK_SIZE);

    t0 = p0 + p3;
    t1 = p1 + p2;
    t2 = p1 - p2;
    t3 = p0 - p3;

    ii = pos_x + i;
    tblock[pos_y    ][ii] = t0 +  t1;
    tblock[pos_y + 1][ii] = t2 + (t3 << 1);
    tblock[pos_y + 2][ii] = t0 -  t1;
    tblock[pos_y + 3][ii] = t3 - (t2 << 1);
  }
}

void inverse4x4(int tblock[16][16], int block[16][16], int pos_y, int pos_x)
{
  int i, ii;  
  int tmp[16];
  int *pTmp = tmp, *pblock;
  int p0,p1,p2,p3;
  int t0,t1,t2,t3;

  // Horizontal
  for (i = pos_y; i < pos_y + BLOCK_SIZE; i++)
  {
    pblock = &tblock[i][pos_x];
    t0 = *(pblock++);
    t1 = *(pblock++);
    t2 = *(pblock++);
    t3 = *(pblock  );

    p0 =  t0 + t2;
    p1 =  t0 - t2;
    p2 = (t1 >> 1) - t3;
    p3 =  t1 + (t3 >> 1);

    *(pTmp++) = p0 + p3;
    *(pTmp++) = p1 + p2;
    *(pTmp++) = p1 - p2;
    *(pTmp++) = p0 - p3;
  }

  //  Vertical 
  for (i = 0; i < BLOCK_SIZE; i++)
  {
    pTmp = tmp + i;
    t0 = *pTmp;
    t1 = *(pTmp += BLOCK_SIZE);
    t2 = *(pTmp += BLOCK_SIZE);
    t3 = *(pTmp += BLOCK_SIZE);

    p0 = t0 + t2;
    p1 = t0 - t2;
    p2 =(t1 >> 1) - t3;
    p3 = t1 + (t3 >> 1);

    ii = i + pos_x;
    block[pos_y    ][ii] = p0 + p3;
    block[pos_y + 1][ii] = p1 + p2;
    block[pos_y + 2][ii] = p1 - p2;
    block[pos_y + 3][ii] = p0 - p3;
  }
}

//following functions perform 8 additions, 8 assignments. Should be a bit faster
void hadamard2x2(int block[16][16], int tblock[4])
{
  int p0,p1,p2,p3;

  p0 = block[0][0] + block[0][4];
  p1 = block[0][0] - block[0][4];
  p2 = block[4][0] + block[4][4];
  p3 = block[4][0] - block[4][4];

  tblock[0] = (p0 + p2);
  tblock[1] = (p1 + p3);
  tblock[2] = (p0 - p2);
  tblock[3] = (p1 - p3);
}

void ihadamard2x2(int tblock[4], int block[4])
{
  int t0,t1,t2,t3;

  t0 = tblock[0] + tblock[1];
  t1 = tblock[0] - tblock[1];
  t2 = tblock[2] + tblock[3];
  t3 = tblock[2] - tblock[3];

  block[0] = (t0 + t2);
  block[1] = (t1 + t3);
  block[2] = (t0 - t2);
  block[3] = (t1 - t3);
}

void forward8x8(int block[16][16], int tblock[16][16], int pos_y, int pos_x)
{
  int i, ii;  
  int tmp[64];
  int *pTmp = tmp, *pblock;
  int a0, a1, a2, a3;
  int p0, p1, p2, p3, p4, p5 ,p6, p7;
  int b0, b1, b2, b3, b4, b5, b6, b7;

  // Horizontal
  for (i=pos_y; i < pos_y + BLOCK_SIZE_8x8; i++)
  {
    pblock = &block[i][pos_x];
    p0 = *(pblock++);
    p1 = *(pblock++);
    p2 = *(pblock++);
    p3 = *(pblock++);
    p4 = *(pblock++);
    p5 = *(pblock++);
    p6 = *(pblock++);
    p7 = *(pblock  );

    a0 = p0 + p7;
    a1 = p1 + p6;
    a2 = p2 + p5;
    a3 = p3 + p4;

    b0 = a0 + a3;
    b1 = a1 + a2;
    b2 = a0 - a3;
    b3 = a1 - a2;

    a0 = p0 - p7;
    a1 = p1 - p6;
    a2 = p2 - p5;
    a3 = p3 - p4;

    b4 = a1 + a2 + ((a0 >> 1) + a0);
    b5 = a0 - a3 - ((a2 >> 1) + a2);
    b6 = a0 + a3 - ((a1 >> 1) + a1);
    b7 = a1 - a2 + ((a3 >> 1) + a3);

    *(pTmp++) =  b0 + b1;
    *(pTmp++) =  b4 + (b7 >> 2);
    *(pTmp++) =  b2 + (b3 >> 1);
    *(pTmp++) =  b5 + (b6 >> 2);
    *(pTmp++) =  b0 - b1;
    *(pTmp++) =  b6 - (b5 >> 2);
    *(pTmp++) = (b2 >> 1) - b3;                 
    *(pTmp++) = (b4 >> 2) - b7;
  }

  // Vertical 
  for (i=0; i < BLOCK_SIZE_8x8; i++)
  {
    pTmp = tmp + i;
    p0 = *pTmp;
    p1 = *(pTmp += BLOCK_SIZE_8x8);
    p2 = *(pTmp += BLOCK_SIZE_8x8);
    p3 = *(pTmp += BLOCK_SIZE_8x8);
    p4 = *(pTmp += BLOCK_SIZE_8x8);
    p5 = *(pTmp += BLOCK_SIZE_8x8);
    p6 = *(pTmp += BLOCK_SIZE_8x8);
    p7 = *(pTmp += BLOCK_SIZE_8x8);

    a0 = p0 + p7;
    a1 = p1 + p6;
    a2 = p2 + p5;
    a3 = p3 + p4;

    b0 = a0 + a3;
    b1 = a1 + a2;
    b2 = a0 - a3;
    b3 = a1 - a2;

    a0 = p0 - p7;
    a1 = p1 - p6;
    a2 = p2 - p5;
    a3 = p3 - p4;

    b4 = a1 + a2 + ((a0 >> 1) + a0);
    b5 = a0 - a3 - ((a2 >> 1) + a2);
    b6 = a0 + a3 - ((a1 >> 1) + a1);
    b7 = a1 - a2 + ((a3 >> 1) + a3);

    ii = pos_x + i;
    tblock[pos_y    ][ii] =  b0 + b1;
    tblock[pos_y + 1][ii] =  b4 + (b7 >> 2);
    tblock[pos_y + 2][ii] =  b2 + (b3 >> 1);
    tblock[pos_y + 3][ii] =  b5 + (b6 >> 2);
    tblock[pos_y + 4][ii] =  b0 - b1;
    tblock[pos_y + 5][ii] =  b6 - (b5 >> 2);
    tblock[pos_y + 6][ii] = (b2 >> 1) - b3;
    tblock[pos_y + 7][ii] = (b4 >> 2) - b7;
  }
}

void inverse8x8(int tblock[16][16], int block[16][16], int pos_y, int pos_x)
{
  int i, ii;
  int tmp[64];
  int *pTmp = tmp, *pblock;
  int a0, a1, a2, a3;
  int p0, p1, p2, p3, p4, p5 ,p6, p7;  
  int b0, b1, b2, b3, b4, b5, b6, b7;

  // Horizontal  
  for (i=pos_y; i < pos_y + BLOCK_SIZE_8x8; i++)
  {
    pblock = &tblock[i][pos_x];
    p0 = *(pblock++);
    p1 = *(pblock++);
    p2 = *(pblock++);
    p3 = *(pblock++);
    p4 = *(pblock++);
    p5 = *(pblock++);
    p6 = *(pblock++);
    p7 = *(pblock  );

    a0 = p0 + p4;
    a1 = p0 - p4;
    a2 = p6 - (p2 >> 1);
    a3 = p2 + (p6 >> 1);

    b0 =  a0 + a3;
    b2 =  a1 - a2;
    b4 =  a1 + a2;
    b6 =  a0 - a3;

    a0 = -p3 + p5 - p7 - (p7 >> 1);    
    a1 =  p1 + p7 - p3 - (p3 >> 1);    
    a2 = -p1 + p7 + p5 + (p5 >> 1);    
    a3 =  p3 + p5 + p1 + (p1 >> 1);


    b1 =  a0 + (a3>>2);    
    b3 =  a1 + (a2>>2);    
    b5 =  a2 - (a1>>2);
    b7 =  a3 - (a0>>2);                

    *(pTmp++) = b0 + b7;
    *(pTmp++) = b2 - b5;
    *(pTmp++) = b4 + b3;
    *(pTmp++) = b6 + b1;
    *(pTmp++) = b6 - b1;
    *(pTmp++) = b4 - b3;
    *(pTmp++) = b2 + b5;
    *(pTmp++) = b0 - b7;
  }

  //  Vertical 
  for (i=0; i < BLOCK_SIZE_8x8; i++)
  {
    pTmp = tmp + i;
    p0 = *pTmp;
    p1 = *(pTmp += BLOCK_SIZE_8x8);
    p2 = *(pTmp += BLOCK_SIZE_8x8);
    p3 = *(pTmp += BLOCK_SIZE_8x8);
    p4 = *(pTmp += BLOCK_SIZE_8x8);
    p5 = *(pTmp += BLOCK_SIZE_8x8);
    p6 = *(pTmp += BLOCK_SIZE_8x8);
    p7 = *(pTmp += BLOCK_SIZE_8x8);

    a0 =  p0 + p4;
    a1 =  p0 - p4;
    a2 =  p6 - (p2>>1);
    a3 =  p2 + (p6>>1);

    b0 = a0 + a3;
    b2 = a1 - a2;
    b4 = a1 + a2;
    b6 = a0 - a3;

    a0 = -p3 + p5 - p7 - (p7 >> 1);
    a1 =  p1 + p7 - p3 - (p3 >> 1);
    a2 = -p1 + p7 + p5 + (p5 >> 1);
    a3 =  p3 + p5 + p1 + (p1 >> 1);


    b1 =  a0 + (a3 >> 2);
    b7 =  a3 - (a0 >> 2);
    b3 =  a1 + (a2 >> 2);
    b5 =  a2 - (a1 >> 2);

    ii = i + pos_x;
    block[pos_y    ][ii] = b0 + b7;
    block[pos_y + 1][ii] = b2 - b5;
    block[pos_y + 2][ii] = b4 + b3;
    block[pos_y + 3][ii] = b6 + b1;
    block[pos_y + 4][ii] = b6 - b1;
    block[pos_y + 5][ii] = b4 - b3;
    block[pos_y + 6][ii] = b2 + b5;
    block[pos_y + 7][ii] = b0 - b7;
  }
}

/*!
************************************************************************
* \brief
*    For new intra pred routines
*
* \par Input:
*    Image par, 16x16 based intra mode
*
* \par Output:
*    none
************************************************************************
*/
int dct_luma_16x16(int new_intra_mode)
{
  //int qp_const;
  int i,j;
  int ii,jj;
  int jdiv, jmod;
  int M1[16][16];
  int M4[4][4];
  int M5[4],M6[4];
  int M0[4][4][4][4];
  int run,scan_pos,coeff_ctr,level;
  int qp_per,qp_rem,q_bits;
  int ac_coef = 0;

  Macroblock *currMB = &img->mb_data[img->current_mb_nr];

  int   b8, b4;
  int*  DCLevel = img->cofDC[0][0];
  int*  DCRun   = img->cofDC[0][1];
  int*  ACLevel;
  int*  ACRun;
  int **levelscale,**leveloffset;
  int **invlevelscale;
  Boolean lossless_qpprime = ((currMB->qp + img->bitdepth_luma_qp_scale)==0 && img->lossless_qpprime_flag==1);
  const byte (*pos_scan)[2] = SNGL_SCAN;
#ifdef RDO_Q
  levelDataStruct levelData[16];
  double  lambda_md=0;
  double normFact = pow(2, (2 * DQ_BITS + 19));
  double err;
  int lowerInt, levelTrellis[16], k, kStart, kStop, noCoeff, estBits;
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
  normFact = normFact*(1<<(2*img->BitDepthIncrease));
#endif
#endif

  // Note that we could just use currMB->qp here
  qp_per    = qp_per_matrix[(currMB->qp + img->bitdepth_luma_qp_scale - MIN_QP)];
  qp_rem    = qp_rem_matrix[(currMB->qp + img->bitdepth_luma_qp_scale - MIN_QP)];

  q_bits    = Q_BITS+qp_per;

#ifdef RDO_Q
  if(input->UseRDO_Q)
  {
    if ((img->type==B_SLICE) && img->nal_reference_idc)
    {
      lambda_md = img->lambda_md[5][img->masterQP];  
    }
    else
    {
      lambda_md = img->lambda_md[img->type][img->masterQP];
    }
  }
#endif

  levelscale    = LevelScale4x4Luma[1][qp_rem];
  leveloffset   = LevelOffset4x4Luma[1][qp_per];
  invlevelscale = InvLevelScale4x4Luma[1][qp_rem];

  for (j=0;j<16;j++)
  {
    jdiv = j >> 2;
    jmod = j & 0x03;
    jj = img->opix_y+j;
    for (i=0;i<16;i++)
    {
      M1[j][i]=imgY_org[jj][img->opix_x+i]-img->mprr_2[new_intra_mode][j][i];
      M0[jdiv][i >> 2][jmod][i & 0x03]=M1[j][i];
    }
  }

  for (jj=0;jj<4 && !lossless_qpprime;jj++)
  {
    for (ii=0;ii<4;ii++)
    {
      for (j=0;j<4;j++)
      {
        M5[0] = M0[jj][ii][j][0] + M0[jj][ii][j][3];
        M5[1] = M0[jj][ii][j][1] + M0[jj][ii][j][2];
        M5[2] = M0[jj][ii][j][1] - M0[jj][ii][j][2];
        M5[3] = M0[jj][ii][j][0] - M0[jj][ii][j][3];

        M4[j][0] = M5[0]   + M5[1];
        M4[j][2] = M5[0]   - M5[1];
        M4[j][1] = M5[3]*2 + M5[2];
        M4[j][3] = M5[3]   - M5[2]*2;
      }
      // vertical
      for (i=0;i<4;i++)
      {
        M5[0] = M4[0][i] + M4[3][i];
        M5[1] = M4[1][i] + M4[2][i];
        M5[2] = M4[1][i] - M4[2][i];
        M5[3] = M4[0][i] - M4[3][i];

        M0[jj][ii][0][i] = M5[0]   + M5[1];
        M0[jj][ii][2][i] = M5[0]   - M5[1];
        M0[jj][ii][1][i] = M5[3]*2 + M5[2];
        M0[jj][ii][3][i] = M5[3]   - M5[2]*2;
      }
    }
  }

  // pick out DC coeff

  for (j=0;j<4;j++)
  {
    for (i=0;i<4;i++)
      M4[j][i]= M0[j][i][0][0];
  }

  if (!lossless_qpprime)
  {
    for (j=0;j<4;j++)
    {
      M5[0] = M4[j][0]+M4[j][3];
      M5[1] = M4[j][1]+M4[j][2];
      M5[2] = M4[j][1]-M4[j][2];
      M5[3] = M4[j][0]-M4[j][3];

      M4[j][0] = M5[0]+M5[1];
      M4[j][2] = M5[0]-M5[1];
      M4[j][1] = M5[3]+M5[2];
      M4[j][3] = M5[3]-M5[2];
    }

    // vertical

    for (i=0;i<4;i++)
    {    
      M5[0] = M4[0][i]+M4[3][i];
      M5[1] = M4[1][i]+M4[2][i];
      M5[2] = M4[1][i]-M4[2][i];
      M5[3] = M4[0][i]-M4[3][i];

      M4[0][i]=(M5[0]+M5[1])>>1;
      M4[2][i]=(M5[0]-M5[1])>>1;
      M4[1][i]=(M5[3]+M5[2])>>1;
      M4[3][i]=(M5[3]-M5[2])>>1;
    }
  }
  // quant

  run=-1;
  scan_pos=0;

  for (coeff_ctr=0;coeff_ctr<16;coeff_ctr++)
  {
    i=pos_scan[coeff_ctr][0];
    j=pos_scan[coeff_ctr][1];

    run++;

    if(lossless_qpprime)
      level= absm(M4[j][i]);
    else
      level= (absm(M4[j][i]) * levelscale[0][0] + (leveloffset[0][0]<<1)) >> (q_bits+1);

    if (level != 0)
    {
      DCLevel[scan_pos] = sign(level,M4[j][i]);
      DCRun  [scan_pos] = run;
      ++scan_pos;
      run=-1;
    }
    if(!lossless_qpprime)
      M4[j][i]=sign(level,M4[j][i]);
  }
  DCLevel[scan_pos]=0;

  // invers DC transform
  for (j=0;j<4 && !lossless_qpprime;j++)
  {
    M6[0]=M4[j][0]+M4[j][2];
    M6[1]=M4[j][0]-M4[j][2];
    M6[2]=M4[j][1]-M4[j][3];
    M6[3]=M4[j][1]+M4[j][3];

    M4[j][0] = M6[0]+M6[3];
    M4[j][1] = M6[1]+M6[2];
    M4[j][2] = M6[1]-M6[2];
    M4[j][3] = M6[0]-M6[3];
  }

  for (i=0;i<4 && !lossless_qpprime;i++)
  {

    M6[0]=M4[0][i]+M4[2][i];
    M6[1]=M4[0][i]-M4[2][i];
    M6[2]=M4[1][i]-M4[3][i];
    M6[3]=M4[1][i]+M4[3][i];

    if(qp_per<6)
    {
      M0[0][i][0][0] = ((M6[0]+M6[3])*invlevelscale[0][0]+(1<<(5-qp_per)))>>(6-qp_per);
      M0[1][i][0][0] = ((M6[1]+M6[2])*invlevelscale[0][0]+(1<<(5-qp_per)))>>(6-qp_per);
      M0[2][i][0][0] = ((M6[1]-M6[2])*invlevelscale[0][0]+(1<<(5-qp_per)))>>(6-qp_per);
      M0[3][i][0][0] = ((M6[0]-M6[3])*invlevelscale[0][0]+(1<<(5-qp_per)))>>(6-qp_per);
    }
    else
    {
      M0[0][i][0][0] = ((M6[0]+M6[3])*invlevelscale[0][0])<<(qp_per-6);
      M0[1][i][0][0] = ((M6[1]+M6[2])*invlevelscale[0][0])<<(qp_per-6);
      M0[2][i][0][0] = ((M6[1]-M6[2])*invlevelscale[0][0])<<(qp_per-6);
      M0[3][i][0][0] = ((M6[0]-M6[3])*invlevelscale[0][0])<<(qp_per-6);
    }   
  }

  // AC inverse trans/quant for MB
  for (jj=0;jj<4;jj++)
  {
    for (ii=0;ii<4;ii++)
    {
      for (j=0;j<4;j++)
      {
        memcpy(M4[j],M0[jj][ii][j], BLOCK_SIZE * sizeof(int));
      }

      run      = -1;
      scan_pos =  0;
      b8       = 2*(jj >> 1) + (ii >> 1);
      b4       = 2*(jj & 0x01) + (ii & 0x01);
      ACLevel  = img->cofAC [b8][b4][0];
      ACRun    = img->cofAC [b8][b4][1];

#ifdef RDO_Q
      if(input->UseRDO_Q && active_pps->entropy_coding_mode_flag == CABAC)
      {
        kStart=0; kStop=0; noCoeff=0;
        for (coeff_ctr=0;coeff_ctr < 15;coeff_ctr++)
        {
          i=pos_scan[coeff_ctr+1][0]; // scan is shifted due to DC
          j=pos_scan[coeff_ctr+1][1]; // scan is shifted due to DC

          levelData[coeff_ctr].levelDouble=absm(M4[j][i]*levelscale[i][j]);
          level = (int)(levelData[coeff_ctr].levelDouble >> q_bits);
          lowerInt=(((int)levelData[coeff_ctr].levelDouble-(level<<q_bits))<(1<<(q_bits-1)))? 1 : 0;

          levelData[coeff_ctr].level[0]=0;
          if (level==0 && lowerInt==1){
            levelData[coeff_ctr].noLevels=1;
          }
          else if (level==0 && lowerInt==0){
            levelData[coeff_ctr].level[1] = level+1;
            levelData[coeff_ctr].noLevels=2;
            kStop=coeff_ctr;
            noCoeff++;
          }
          else if (level>0 && lowerInt==1){
            levelData[coeff_ctr].level[1] = level;
            levelData[coeff_ctr].noLevels=2;
            kStop=coeff_ctr;
            noCoeff++;
          }
          else{
            levelData[coeff_ctr].level[1] = level;
            levelData[coeff_ctr].level[2] = level+1;
            levelData[coeff_ctr].noLevels=3;
            kStop=coeff_ctr;
            kStart=coeff_ctr;
            noCoeff++;
          }

          for (k=0; k<levelData[coeff_ctr].noLevels; k++){
            err=(double)(levelData[coeff_ctr].level[k]<<q_bits)-(double)levelData[coeff_ctr].levelDouble;
            levelData[coeff_ctr].errLevel[k]=(err*err*(double)estErr4x4[qp_rem][i][j])/normFact; 
          }
        }

        estBits=est_write_and_store_CBP_block_bit(currMB, LUMA_16AC);
        est_writeRunLevel_CABAC(levelData, levelTrellis, LUMA_16AC, lambda_md, kStart, kStop, noCoeff, estBits);
      }
#endif

      for (coeff_ctr=1;coeff_ctr<16;coeff_ctr++) // set in AC coeff
      {

        i=pos_scan[coeff_ctr][0];
        j=pos_scan[coeff_ctr][1];

        run++;
#ifdef RDO_Q
        if(input->UseRDO_Q)
        {
          level=levelTrellis[coeff_ctr-1];
        }
        else
        {
          if(lossless_qpprime)
            level= absm( M4[j][i]);
          else          
            level= ( absm( M4[j][i]) * levelscale[i][j] + leveloffset[i][j]) >> q_bits;
        }
#else
        if(lossless_qpprime)
          level= absm( M4[j][i]);
        else          
          level= ( absm( M4[j][i]) * levelscale[i][j] + leveloffset[i][j]) >> q_bits;
#endif

        if (level != 0)
        {
          ac_coef = 15;
          ACLevel[scan_pos] = sign(level,M4[j][i]);
          ACRun  [scan_pos] = run;
          ++scan_pos;
          run=-1;
        }

        if(!lossless_qpprime)
        {
          level=sign(level, M4[j][i]);
          if(qp_per<4)
            M4[j][i]=(level*invlevelscale[i][j]+(1<<(3-qp_per)))>>(4-qp_per);
          else
            M4[j][i]=(level*invlevelscale[i][j])<<(qp_per-4);
        }
      }
      ACLevel[scan_pos] = 0;


      // IDCT horizontal
      for (j=0;j<4 && !lossless_qpprime;j++)
      {
        M6[0] = M4[j][0]     +  M4[j][2];
        M6[1] = M4[j][0]     -  M4[j][2];
        M6[2] =(M4[j][1]>>1) -  M4[j][3];
        M6[3] = M4[j][1]     + (M4[j][3]>>1);

        M4[j][0] = M6[0] + M6[3];
        M4[j][1] = M6[1] + M6[2];
        M4[j][2] = M6[1] - M6[2];
        M4[j][3] = M6[0] - M6[3];
      }

      // vert
      for (i=0;i<4 && !lossless_qpprime;i++)
      {
        M6[0]= M4[0][i]     +  M4[2][i];
        M6[1]= M4[0][i]     -  M4[2][i];
        M6[2]=(M4[1][i]>>1) -  M4[3][i];
        M6[3]= M4[1][i]     + (M4[3][i]>>1);

        M0[jj][ii][0][i] = M6[0] + M6[3];
        M0[jj][ii][1][i] = M6[1] + M6[2];
        M0[jj][ii][2][i] = M6[1] - M6[2];
        M0[jj][ii][3][i] = M6[0] - M6[3];
      }
    }
  }

  for (jj=0;jj<BLOCK_MULTIPLE; jj++)
    for (ii=0;ii<BLOCK_MULTIPLE; ii++)
      for (j=0;j<BLOCK_SIZE;j++)
      {
        memcpy(&M1[jj*BLOCK_SIZE + j][ii*BLOCK_SIZE], M0[jj][ii][j], BLOCK_SIZE * sizeof(int));
      }



      if(lossless_qpprime)
      {
        for (j=0;j<16;j++)
        {
          jj = img->pix_y+j;
          for (i=0;i<16;i++)
          {
            enc_picture->imgY[jj][img->pix_x+i]=(imgpel)(M1[j][i]+img->mprr_2[new_intra_mode][j][i]);
          }
        }
      }
      else
      {
        for (j=0;j<16;j++)
        {
          jj = img->pix_y+j;
          for (i=0;i<16;i++)
          {
            enc_picture->imgY[jj][img->pix_x+i]=(imgpel)clip1a((M1[j][i]+((long)img->mprr_2[new_intra_mode][j][i]<<DQ_BITS)+DQ_ROUND)>>DQ_BITS);
          }
        }
      }


      return ac_coef;
}


/*!
************************************************************************
* \brief
*    The routine performs transform,quantization,inverse transform, adds the diff.
*    to the prediction and writes the result to the decoded luma frame. Includes the
*    RD constrained quantization also.
*
* \par Input:
*    block_x,block_y: Block position inside a macro block (0,4,8,12).
*
* \par Output_
*    nonzero: 0 if no levels are nonzero.  1 if there are nonzero levels.             \n
*    coeff_cost: Counter for nonzero coefficients, used to discard expensive levels.
************************************************************************
*/
#ifdef USE_INTRA_MDDT
int dct_luma(int block_x,int block_y,int *coeff_cost, int intra, int ipmode)
#else
int dct_luma(int block_x,int block_y,int *coeff_cost, int intra)
#endif
{
  int sign(int a,int b);

  int i,j,ilev,coeff_ctr;
  int ii,jj;
  int m4[16][16];

  //int qp_const;
  int level,scan_pos,run;
  int nonzero;
  int qp_per,qp_rem,q_bits;

  int   pos_x   = block_x >> BLOCK_SHIFT;
  int   pos_y   = block_y >> BLOCK_SHIFT;
  int   b8      = 2*(pos_y >> 1) + (pos_x >> 1);
  int   b4      = 2*(pos_y & 0x01) + (pos_x & 0x01);
  int*  ACLevel = img->cofAC[b8][b4][0];
  int*  ACRun   = img->cofAC[b8][b4][1];
  int   pix_y;


  Macroblock *currMB = &img->mb_data[img->current_mb_nr];

  Boolean lossless_qpprime = ((currMB->qp + img->bitdepth_luma_qp_scale)==0 && img->lossless_qpprime_flag==1);
  int **levelscale,**leveloffset;
  int **invlevelscale;
#ifdef USE_INTRA_MDDT
  byte pos_scan[16][2];
#else
  const byte (*pos_scan)[2] = SNGL_SCAN;
#endif

#ifdef RDO_Q
  levelDataStruct levelData[16];
  double  lambda_md=0, normFact=pow(2,(2*DQ_BITS+19));
  double err;
  int lowerInt, levelTrellis[16], kStart=0, kStop=0, noCoeff, estBits;
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
  normFact = normFact*(1<<(2*img->BitDepthIncrease));
#endif
#endif
#ifdef USE_INTRA_MDDT
  if(input->UseIntraMDDT && intra)
  {
    for(i = 0; i < 16; i++)
    {
      pos_scan[i][0] = img->scanOrder4x4[ipmode][i][0];
      pos_scan[i][1] = img->scanOrder4x4[ipmode][i][1];
    }
  }
  else
  {
    for(i = 0; i < 16; i++)
    {
      pos_scan[i][0] = SNGL_SCAN[i][0];
      pos_scan[i][1] = SNGL_SCAN[i][1];
    }
  }
#endif

  qp_per    = (currMB->qp + img->bitdepth_luma_qp_scale - MIN_QP)/6; 
  qp_rem    = (currMB->qp + img->bitdepth_luma_qp_scale - MIN_QP)%6; 
  q_bits    = Q_BITS+qp_per;

#ifdef RDO_Q
  if(input->UseRDO_Q)
  {
    if ((img->type==B_SLICE) && img->nal_reference_idc)
    {
      lambda_md = img->lambda_md[5][img->masterQP];  
    }
    else
    {
      lambda_md = img->lambda_md[img->type][img->masterQP];
    }
  }
#endif

  levelscale    = LevelScale4x4Luma[intra][qp_rem];
  leveloffset   = LevelOffset4x4Luma[intra][qp_per];
  invlevelscale = InvLevelScale4x4Luma[intra][qp_rem];

  //  Horizontal transform
  if (!lossless_qpprime)
  {
    forward4x4(img->m7, m4, 0, 0);
  }

  // Quant

  nonzero=FALSE;

  run=-1;
  scan_pos=0;
#ifdef RDO_Q
  if(input->UseRDO_Q && active_pps->entropy_coding_mode_flag == CABAC)
  {
    noCoeff=0;
    for (coeff_ctr=0;coeff_ctr < 16;coeff_ctr++)
    {
      i=pos_scan[coeff_ctr][0];
      j=pos_scan[coeff_ctr][1];

      levelData[coeff_ctr].levelDouble = absm(m4[j][i] * levelscale[i][j]);
      level = (int)(levelData[coeff_ctr].levelDouble >> q_bits);
      lowerInt=(((int)levelData[coeff_ctr].levelDouble - (level << q_bits)) < (1 <<( q_bits - 1)))? 1: 0;

      levelData[coeff_ctr].level[0]=0;
      if (level==0 && lowerInt==1){
        levelData[coeff_ctr].noLevels=1;
      }
      else if (level==0 && lowerInt==0){
        levelData[coeff_ctr].level[1] = level+1;
        levelData[coeff_ctr].noLevels=2;
        kStop=coeff_ctr;
        noCoeff++;
      }
      else if (level>0 && lowerInt==1){
        levelData[coeff_ctr].level[1] = level;
        levelData[coeff_ctr].noLevels=2;
        kStop=coeff_ctr;
        noCoeff++;
      }
      else{
        levelData[coeff_ctr].level[1] = level;
        levelData[coeff_ctr].level[2] = level+1;
        levelData[coeff_ctr].noLevels=3;
        kStop=coeff_ctr;
        kStart=coeff_ctr;
        noCoeff++;
      }

      for (ii=0; ii<levelData[coeff_ctr].noLevels; ii++){
        err=(double)(levelData[coeff_ctr].level[ii]<<q_bits)-(double)levelData[coeff_ctr].levelDouble;
        levelData[coeff_ctr].errLevel[ii]=(err*err*(double)estErr4x4[qp_rem][i][j])/normFact; 
      }
    }

    estBits=est_write_and_store_CBP_block_bit(currMB, LUMA_4x4);
    est_writeRunLevel_CABAC(levelData, levelTrellis, LUMA_4x4, lambda_md, kStart, kStop, noCoeff, estBits);
  }
#endif

  for (coeff_ctr=0;coeff_ctr < 16;coeff_ctr++)
  {

    i=pos_scan[coeff_ctr][0];
    j=pos_scan[coeff_ctr][1];

    run++;
    ilev=0;

#ifdef RDO_Q
    if(input->UseRDO_Q)    
      level = levelTrellis[coeff_ctr];
    else
    {
      if(lossless_qpprime)
        level = absm (img->m7[j][i]);
      else
        level = (absm (m4[j][i]) * levelscale[i][j] + leveloffset[i][j]) >> q_bits;
    }
#else
    if(lossless_qpprime)
      level = absm (img->m7[j][i]);
    else
      level = (absm (m4[j][i]) * levelscale[i][j] + leveloffset[i][j]) >> q_bits;
#endif

    if (level != 0)
    {
      nonzero=TRUE;

      *coeff_cost += (level > 1 || lossless_qpprime) ? MAX_VALUE : COEFF_COST[input->disthres][run];

      if(lossless_qpprime)
        ACLevel[scan_pos] = sign(level,img->m7[j][i]);
      else
        ACLevel[scan_pos] = sign(level,m4[j][i]);

      ACRun  [scan_pos] = run;
      ++scan_pos;
      run=-1;                     // reset zero level counter

      level=sign(level, m4[j][i]);

      if(lossless_qpprime)
      {
        ilev=level;
      }
      else
      {
        ilev=((((level*invlevelscale[i][j])<< qp_per) + 8 ) >> 4);
      }
    }
    if(!lossless_qpprime)
      m4[j][i]=ilev;
  }

  ACLevel[scan_pos] = 0;  

  //     IDCT.
  //     horizontal

  inverse4x4(m4, img->m7, 0, 0);

  for( j=0; j<BLOCK_SIZE; j++)
  {
    jj = block_y+j;    
    pix_y = img->pix_y + jj;
    for( i=0; i<BLOCK_SIZE; i++)
    {
      ii = block_x+i;
      img->m7[j][i] = clip1a((img->m7[j][i]+((long)img->mpr[jj][ii] << DQ_BITS)+DQ_ROUND)>>DQ_BITS);
      enc_picture->imgY[pix_y][img->pix_x + ii]=img->m7[j][i];
    }
  }

  return nonzero;
}


/*!
************************************************************************
* \brief
*    The routine performs transform,quantization,inverse transform, adds the diff.
*    to the prediction and writes the result to the decoded luma frame. Includes the
*    RD constrained quantization also.
*
* \par Input:
*    b8: Block position inside a macro block (0,1,2,3).
*
* \par Output:
*    nonzero: 0 if no levels are nonzero.  1 if there are nonzero levels.  
*    coeff_cost: Counter for nonzero coefficients, used to discard expensive levels.
************************************************************************
*/

#define MC(coeff) ((coeff)&3)

#ifdef USE_INTRA_MDDT
int dct_luma8x8(int b8,int *coeff_cost, int intra, int ipmode)
#else
int dct_luma8x8(int b8,int *coeff_cost, int intra)
#endif
{
  int sign(int a,int b);

  int i,j,ilev,coeff_ctr;
  int level,scan_pos,run;
  int nonzero;
  int qp_per,qp_rem,q_bits;
  int dq_lshift = 0, dq_rshift = 0, dq_round = 0;

  int block_x = 8*(b8 & 0x01);
  int block_y = 8*(b8 >> 1);
  int*  ACLevel = img->cofAC[b8][0][0];
  int*  ACRun   = img->cofAC[b8][0][1];

  int scan_poss[4],runs[4];
  int pix_x, pix_y, ipix_y;
  int **levelscale,**leveloffset;
  int **invlevelscale;
  int MCcoeff;
#ifdef RDO_Q
  levelDataStruct levelData[64];
  double  lambda_md=0, normFact=pow(2,(2*Q_BITS_8+9));
  double err;
  int lowerInt, levelTrellis[64], ii, kStart=0, kStop=0, noCoeff;
#endif

  Boolean lossless_qpprime = ((img->qp + img->bitdepth_luma_qp_scale)==0 && img->lossless_qpprime_flag==1);

  qp_per    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)/6;
  qp_rem    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)%6;
  q_bits    = Q_BITS_8+qp_per;

#ifdef RDO_Q
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
  normFact = normFact*(1<<(2*img->BitDepthIncrease));
#endif
  if(input->UseRDO_Q)
  {
    if ((img->type==B_SLICE) && img->nal_reference_idc)
    {
      lambda_md = img->lambda_md[5][img->masterQP];  
    }
    else
    {
      lambda_md = img->lambda_md[img->type][img->masterQP];
    }
  }
#endif

  levelscale    = LevelScale8x8Luma[intra][qp_rem];
  leveloffset   = LevelOffset8x8Luma[intra][qp_per];
  invlevelscale = InvLevelScale8x8Luma[intra][qp_rem];

  if (qp_per < 6)
  {
    dq_rshift = 6 - qp_per;
    dq_round  = 1<<(5-qp_per);
  }
  else
    dq_lshift = qp_per - 6;


  // horizontal transform
  if (!lossless_qpprime) 
  {
    forward8x8(img->m7, img->m7, 0, 0);
  }


#ifdef RDO_Q
  if(input->UseRDO_Q && active_pps->entropy_coding_mode_flag == CABAC)
  {
    noCoeff=0;
    for (coeff_ctr=0;coeff_ctr < 64;coeff_ctr++)
    {

#ifdef USE_INTRA_MDDT
      if(input->UseIntraMDDT && intra)
      {
        i=img->scanOrder8x8[ipmode][coeff_ctr][0];
        j=img->scanOrder8x8[ipmode][coeff_ctr][1];
      }
      else
      {
        i=SNGL_SCAN8x8[coeff_ctr][0];
        j=SNGL_SCAN8x8[coeff_ctr][1];
      }
#else
      i=SNGL_SCAN8x8[coeff_ctr][0];
      j=SNGL_SCAN8x8[coeff_ctr][1];
#endif       

      levelData[coeff_ctr].levelDouble = absm(img->m7[j][i] * levelscale[i][j]);
      level = (int)(levelData[coeff_ctr].levelDouble >> q_bits);
      lowerInt = (((int)levelData[coeff_ctr].levelDouble - (level << q_bits)) < (1 << (q_bits - 1)))? 1: 0;

      levelData[coeff_ctr].level[0]=0;
      if (level==0 && lowerInt==1){
        levelData[coeff_ctr].noLevels=1;
      }
      else if (level==0 && lowerInt==0)
      {
        levelData[coeff_ctr].level[1] = level+1;
        levelData[coeff_ctr].noLevels=2;
        kStop=coeff_ctr;
        noCoeff++;
      }
      else if (level>0 && lowerInt==1)
      {
        levelData[coeff_ctr].level[1] = level;
        levelData[coeff_ctr].noLevels=2;
        kStop=coeff_ctr;
        noCoeff++;
      }
      else
      {
        levelData[coeff_ctr].level[1] = level;
        levelData[coeff_ctr].level[2] = level+1;
        levelData[coeff_ctr].noLevels=3;
        kStop=coeff_ctr;
        kStart=coeff_ctr;
        noCoeff++;
      }

      for (ii=0; ii<levelData[coeff_ctr].noLevels; ii++)
      {
        err=(double)(levelData[coeff_ctr].level[ii]<<q_bits)-(double)levelData[coeff_ctr].levelDouble;
        levelData[coeff_ctr].errLevel[ii]=(err*err*(double)estErr8x8[qp_rem][i][j])/normFact; 
      }
    }
    est_writeRunLevel_CABAC(levelData, levelTrellis, LUMA_8x8, lambda_md, kStart, kStop, noCoeff, 0);
  }
#endif

  // Quant  
  nonzero=FALSE;

  run=-1;
  scan_pos=0;

  runs[0]=runs[1]=runs[2]=runs[3]=-1;
  scan_poss[0]=scan_poss[1]=scan_poss[2]=scan_poss[3]=0;

  for (coeff_ctr=0;coeff_ctr < 64;coeff_ctr++)
  {

#ifdef USE_INTRA_MDDT
    if(input->UseIntraMDDT && intra)
    {
      i=img->scanOrder8x8[ipmode][coeff_ctr][0];
      j=img->scanOrder8x8[ipmode][coeff_ctr][1];
    }
    else 
    {
      i=SNGL_SCAN8x8[coeff_ctr][0];
      j=SNGL_SCAN8x8[coeff_ctr][1];
    }
#else
    i=SNGL_SCAN8x8[coeff_ctr][0];
    j=SNGL_SCAN8x8[coeff_ctr][1];
#endif
    MCcoeff = MC(coeff_ctr);
    run++;
    ilev=0;

    runs[MCcoeff]++;

#ifdef RDO_Q
    if(input->UseRDO_Q)
      level = levelTrellis[coeff_ctr];
    else
    {
      if(lossless_qpprime)
        level = absm (img->m7[j][i]);
      else 
        level = (absm (img->m7[j][i]) * levelscale[i][j] + leveloffset[i][j]) >> q_bits;
    }
#else
    if(lossless_qpprime)
      level = absm (img->m7[j][i]);
    else 
      level = (absm (img->m7[j][i]) * levelscale[i][j] + leveloffset[i][j]) >> q_bits;
#endif

    if (level != 0)
    {
      nonzero=TRUE;

      *coeff_cost += (level > 1 || lossless_qpprime) ? MAX_VALUE :COEFF_COST8x8[input->disthres][run];
      ACLevel[scan_pos] = sign(level,img->m7[j][i]);
      ACRun  [scan_pos] = run;
      ++scan_pos;
      run=-1;                     // reset zero level counter

      level = sign(level, img->m7[j][i]);
      if(lossless_qpprime)
      {
        ilev = level;
      }
      else 
      {
        if (qp_per>=6)
          ilev = level*invlevelscale[i][j]<<dq_lshift; // dequantization
        else
          ilev = (level*invlevelscale[i][j] + dq_round)>>dq_rshift; // dequantization
      }
    }
    if(!lossless_qpprime)
      img->m7[j][i] = ilev;
  }

  ACLevel[scan_pos] = 0;

  //    Inverse Transform
  // horizontal inverse transform
  if (!lossless_qpprime)
  {
    inverse8x8(img->m7,img->m7, 0, 0);
  }

  for( j=0; j<2*BLOCK_SIZE; j++)
  {
    pix_y = block_y+j;    
    ipix_y = img->pix_y + pix_y;
    for( i=0; i<2*BLOCK_SIZE; i++)
    {
      pix_x = block_x+i;
      if(lossless_qpprime)
        img->m7[j][i] = img->m7[j][i]+img->mpr[pix_y][block_x+i];
      else
        img->m7[j][i] = clip1a((img->m7[j][i]+((long)img->mpr[pix_y][pix_x] << DQ_BITS_8)+DQ_ROUND_8)>>DQ_BITS_8);
      enc_picture->imgY[ipix_y][img->pix_x + pix_x]=img->m7[j][i];
    }
  }

  //  Decoded block moved to frame memory
  return nonzero;
}


/*!
************************************************************************
* \brief
*    Transform,quantization,inverse transform for chroma.
*    The main reason why this is done in a separate routine is the
*    additional 2x2 transform of DC-coeffs. This routine is called
*    ones for each of the chroma components.
*
* \par Input:
*    uv    : Make difference between the U and V chroma component  \n
*    cr_cbp: chroma coded block pattern
*
* \par Output:
*    cr_cbp: Updated chroma coded block pattern.
************************************************************************
*/
int dct_chroma(int uv,int cr_cbp)
{
  int i,j,ilev,n2,n1,coeff_ctr,level ,scan_pos,run;
  int m1[BLOCK_SIZE],m5[BLOCK_SIZE],m6[BLOCK_SIZE];
  int coeff_cost;
  int cr_cbp_tmp;
  int DCcoded=0 ;
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];

  int qp_per,qp_rem,q_bits;

  int   b4;
  int*  DCLevel = img->cofDC[uv+1][0];
  int*  DCRun   = img->cofDC[uv+1][1];
  int*  ACLevel;
  int*  ACRun;
  int   intra = IS_INTRA (currMB);
  int   uv_scale = uv*(img->num_blk8x8_uv >> 1);

  //FRExt
  int64 cbpblk_pattern[4]={0, 0xf0000, 0xff0000, 0xffff0000};
  int yuv = img->yuv_format;
  int b8;
  int m3[4][4];
  int m4[4][4];
  int qp_per_dc = 0;
  int qp_rem_dc = 0;
  int q_bits_422 = 0;  
  int ***levelscale, ***leveloffset;
  int ***invlevelscale;
  short pix_c_x, pix_c_y;
  const byte (*pos_scan)[2] = SNGL_SCAN;

  Boolean lossless_qpprime = ((currMB->qp + img->bitdepth_luma_qp_scale)==0 && img->lossless_qpprime_flag==1);
#ifdef RDO_Q
  levelDataStruct levelData[16];
  double  lambda_md=0;
  double normFact = pow(2, (2 * DQ_BITS + 19));
  double err;
  int lowerInt, levelTrellis[16], k, kStart, kStop, noCoeff, estBits;
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
  normFact = normFact*(1<<(2*img->BitDepthIncrease));
#endif
#endif

  qp_per    = qp_per_matrix[(currMB->qpc[uv] + img->bitdepth_chroma_qp_scale)];
  qp_rem    = qp_rem_matrix[(currMB->qpc[uv] + img->bitdepth_chroma_qp_scale)];

  q_bits    = Q_BITS+qp_per;

#ifdef RDO_Q
  //if (input->successive_Bframe>0)
  if(input->UseRDO_Q)
  {
    if ((img->type==B_SLICE) && img->nal_reference_idc)
    {
      lambda_md = img->lambda_md[5][img->masterQP];  
    }
    else
    {
      if (input->successive_Bframe==0 && img->type != B_SLICE)
      {
        int masterQP_C;
        masterQP_C = Clip3 ( -img->bitdepth_chroma_qp_scale, 51, img->masterQP + img->chroma_qp_offset[uv] );
        masterQP_C = (masterQP_C < 0 ? masterQP_C : QP_SCALE_CR[masterQP_C]) + img->bitdepth_chroma_qp_scale;

        if(masterQP_C < img->masterQP)
        {
          lambda_md = img->lambda_md[img->type][masterQP_C];
        }
        else
        {
          lambda_md = img->lambda_md[img->type][img->masterQP];
        }
      }
      else
      {
        lambda_md = img->lambda_md[img->type][img->masterQP];
      }
    }
  }
#endif


  levelscale = LevelScale4x4Chroma[uv][intra];
  leveloffset = LevelOffset4x4Chroma[uv][intra];
  invlevelscale = InvLevelScale4x4Chroma[uv][intra];

  if (img->yuv_format == YUV422)
  {
    //for YUV422 only
    qp_per_dc = (currMB->qpc[uv] + 3 + img->bitdepth_chroma_qp_scale)/6;
    qp_rem_dc = (currMB->qpc[uv] + 3 + img->bitdepth_chroma_qp_scale)%6;

    q_bits_422 = Q_BITS+qp_per_dc;  
  }


  //============= dct transform ===============  
  for (n2=0; n2 < img->mb_cr_size_y; n2 += BLOCK_SIZE)
  {
    for (n1=0; n1 < img->mb_cr_size_x; n1 += BLOCK_SIZE)
    {
      forward4x4(img->m7, img->m7, n2, n1);
    }
  }

  if (yuv == YUV420)
  {
    //================== CHROMA DC YUV420 ===================
    //     2X2 transform of DC coeffs.
    if(lossless_qpprime)
    {
      m1[0]=img->m7[0][0];
      m1[1]=img->m7[0][4];
      m1[2]=img->m7[4][0];
      m1[3]=img->m7[4][4];
    }
    else 
    {
      hadamard2x2(img->m7, m1);
    }

    //     Quant of chroma 2X2 coeffs.
    run=-1;
    scan_pos=0;

    for (coeff_ctr=0; coeff_ctr < 4; coeff_ctr++)
    {
      run++;
      ilev=0;

      if(lossless_qpprime)
        level =absm(m1[coeff_ctr]);
      else 
        level =(absm(m1[coeff_ctr]) * levelscale[qp_rem][0][0] + (leveloffset[qp_per][0][0]<<1)) >> (q_bits+1);

      if (level  != 0)
      {
        currMB->cbp_blk |= 0xf0000 << (uv << 2) ;    // if one of the 2x2-DC levels is != 0 set the
        cr_cbp=max(1,cr_cbp);                     // coded-bit all 4 4x4 blocks (bit 16-19 or 20-23)
        DCcoded = 1 ;
        DCLevel[scan_pos] = sign(level ,m1[coeff_ctr]);
        DCRun  [scan_pos] = run;
        scan_pos++;
        run=-1;

        ilev=sign(level, m1[coeff_ctr]);
      }
      if(!lossless_qpprime)
        m1[coeff_ctr]=ilev;
    }
    DCLevel[scan_pos] = 0;

    //  Inverse transform of 2x2 DC levels
    if(!lossless_qpprime)
    {
      ihadamard2x2(m1, m5);

      if(qp_per<5)
      {
        for(i=0; i<4; i++)
          m1[i]=(m5[i] * invlevelscale[qp_rem][0][0])>>(5-qp_per);
      }
      else
      {
        for(i=0; i<4; i++)
          m1[i]=(m5[i] * invlevelscale[qp_rem][0][0])<<(qp_per-5);
      }

      img->m7[0][0] = m1[0];
      img->m7[0][4] = m1[1];
      img->m7[4][0] = m1[2];
      img->m7[4][4] = m1[3];
    }
  }
  else if(yuv == YUV422)
  {
    //================== CHROMA DC YUV422 ===================
    //transform DC coeff
    //horizontal

    //pick out DC coeff
    for (j=0; j < img->mb_cr_size_y; j+=BLOCK_SIZE)
    {
      for (i=0; i < img->mb_cr_size_x; i+=BLOCK_SIZE)
        m3[i>>2][j>>2]= img->m7[j][i];
    } 
    //horizontal
    if(!lossless_qpprime)
    {
      m4[0][0] = m3[0][0] + m3[1][0];
      m4[0][1] = m3[0][1] + m3[1][1];
      m4[0][2] = m3[0][2] + m3[1][2];
      m4[0][3] = m3[0][3] + m3[1][3];

      m4[1][0] = m3[0][0] - m3[1][0];
      m4[1][1] = m3[0][1] - m3[1][1];
      m4[1][2] = m3[0][2] - m3[1][2];
      m4[1][3] = m3[0][3] - m3[1][3];

      // vertical
      for (i=0;i<2;i++)
      {
        m5[0] = m4[i][0] + m4[i][3];
        m5[1] = m4[i][1] + m4[i][2];
        m5[2] = m4[i][1] - m4[i][2];
        m5[3] = m4[i][0] - m4[i][3];

        m4[i][0] = (m5[0] + m5[1]);
        m4[i][2] = (m5[0] - m5[1]);
        m4[i][1] = (m5[3] + m5[2]);
        m4[i][3] = (m5[3] - m5[2]);
      }
    }

    run=-1;
    scan_pos=0;

    //quant of chroma DC-coeffs
    for (coeff_ctr=0;coeff_ctr<8;coeff_ctr++)
    {
      i=SCAN_YUV422[coeff_ctr][0];
      j=SCAN_YUV422[coeff_ctr][1];

      run++;

      if(lossless_qpprime)
      {
        level = absm(m3[i][j]);
        m4[i][j]=m3[i][j];
      }
      else 
        level =(absm(m4[i][j]) * levelscale[qp_rem_dc][0][0] + (leveloffset[qp_per_dc][0][0]*2)) >> (q_bits_422+1);

      if (level != 0)
      {
        //YUV422
        currMB->cbp_blk |= 0xff0000 << (uv << 3) ;   // if one of the DC levels is != 0 set the
        cr_cbp=max(1,cr_cbp);                           // coded-bit all 4 4x4 blocks (bit 16-31 or 32-47) //YUV444
        DCcoded = 1 ;

        DCLevel[scan_pos] = sign(level,m4[i][j]);
        DCRun  [scan_pos] = run;
        ++scan_pos;
        run=-1;
      }
      if(!lossless_qpprime)
        m3[i][j]=sign(level,m4[i][j]);
    }
    DCLevel[scan_pos]=0;

    //inverse DC transform
    //horizontal
    if(!lossless_qpprime)
    {
      m4[0][0] = m3[0][0] + m3[1][0];
      m4[0][1] = m3[0][1] + m3[1][1];
      m4[0][2] = m3[0][2] + m3[1][2];
      m4[0][3] = m3[0][3] + m3[1][3];

      m4[1][0] = m3[0][0] - m3[1][0];
      m4[1][1] = m3[0][1] - m3[1][1];
      m4[1][2] = m3[0][2] - m3[1][2];
      m4[1][3] = m3[0][3] - m3[1][3];      

      // vertical
      for (i=0;i<2;i++)
      {       
        m6[0]=m4[i][0]+m4[i][2];
        m6[1]=m4[i][0]-m4[i][2];
        m6[2]=m4[i][1]-m4[i][3];
        m6[3]=m4[i][1]+m4[i][3];

        if(qp_per_dc<4)
        {          
          img->m7[0 ][i*4]=((((m6[0]+m6[3])*invlevelscale[qp_rem_dc][0][0]+(1<<(3-qp_per_dc)))>>(4-qp_per_dc))+2)>>2;
          img->m7[4 ][i*4]=((((m6[1]+m6[2])*invlevelscale[qp_rem_dc][0][0]+(1<<(3-qp_per_dc)))>>(4-qp_per_dc))+2)>>2;
          img->m7[8 ][i*4]=((((m6[1]-m6[2])*invlevelscale[qp_rem_dc][0][0]+(1<<(3-qp_per_dc)))>>(4-qp_per_dc))+2)>>2;
          img->m7[12][i*4]=((((m6[0]-m6[3])*invlevelscale[qp_rem_dc][0][0]+(1<<(3-qp_per_dc)))>>(4-qp_per_dc))+2)>>2;
        }
        else
        {
          img->m7[0 ][i*4]=((((m6[0]+m6[3])*invlevelscale[qp_rem_dc][0][0])<<(qp_per_dc-4))+2)>>2;
          img->m7[4 ][i*4]=((((m6[1]+m6[2])*invlevelscale[qp_rem_dc][0][0])<<(qp_per_dc-4))+2)>>2;
          img->m7[8 ][i*4]=((((m6[1]-m6[2])*invlevelscale[qp_rem_dc][0][0])<<(qp_per_dc-4))+2)>>2;
          img->m7[12][i*4]=((((m6[0]-m6[3])*invlevelscale[qp_rem_dc][0][0])<<(qp_per_dc-4))+2)>>2;
        }
      }//for (i=0;i<2;i++)    
    }
  }
  else if(yuv == YUV444)
  {
    //================== CHROMA DC YUV444 ===================
    //transform DC coeff
    //pick out DC coeff
    for (j=0; j < img->mb_cr_size_y; j+=BLOCK_SIZE)
    {
      for (i=0; i < img->mb_cr_size_x; i+=BLOCK_SIZE)
        m4[i>>2][j>>2]= img->m7[j][i];
    }

    //horizontal
    for (j=0;j<4 && !lossless_qpprime;j++)
    {
      m5[0] = m4[0][j] + m4[3][j];
      m5[1] = m4[1][j] + m4[2][j];
      m5[2] = m4[1][j] - m4[2][j];
      m5[3] = m4[0][j] - m4[3][j];

      m4[0][j]=m5[0]+m5[1];
      m4[2][j]=m5[0]-m5[1];
      m4[1][j]=m5[3]+m5[2];
      m4[3][j]=m5[3]-m5[2];
    }
    // vertical
    for (i=0;i<4 && !lossless_qpprime;i++)
    {
      m5[0] = m4[i][0] + m4[i][3];
      m5[1] = m4[i][1] + m4[i][2];
      m5[2] = m4[i][1] - m4[i][2];
      m5[3] = m4[i][0] - m4[i][3];

      m4[i][0]=(m5[0]+m5[1])>>1;
      m4[i][2]=(m5[0]-m5[1])>>1;
      m4[i][1]=(m5[3]+m5[2])>>1;
      m4[i][3]=(m5[3]-m5[2])>>1;
    }

    run=-1;
    scan_pos=0;

    //quant of chroma DC-coeffs
    for (coeff_ctr=0;coeff_ctr<16;coeff_ctr++)
    {
      i=SNGL_SCAN[coeff_ctr][0];
      j=SNGL_SCAN[coeff_ctr][1];

      run++;

      if(lossless_qpprime)
        level = absm(m4[i][j]);
      else 
        level =(absm(m4[i][j]) * levelscale[qp_rem][0][0] + (leveloffset[qp_per][0][0]*2)) >> (q_bits+1);

      if (level != 0)
      {
        //YUV444
        currMB->cbp_blk |= ((int64)0xffff0000) << (uv << 4) ;   // if one of the DC levels is != 0 set the
        cr_cbp=max(1,cr_cbp);                           // coded-bit all 4 4x4 blocks (bit 16-31 or 32-47) //YUV444
        DCcoded = 1 ;

        DCLevel[scan_pos] = sign(level,m4[i][j]);
        DCRun  [scan_pos] = run;
        ++scan_pos;
        run=-1;
      }
      if(!lossless_qpprime)
        m4[i][j]=sign(level,m4[i][j]);
    }
    DCLevel[scan_pos]=0;

    // inverse DC transform
    //horizontal
    for (j=0;j<4 && !lossless_qpprime;j++)
    {     
      m6[0] = m4[0][j] + m4[2][j];
      m6[1] = m4[0][j] - m4[2][j];
      m6[2] = m4[1][j] - m4[3][j];
      m6[3] = m4[1][j] + m4[3][j];

      m4[0][j] = m6[0] + m6[3];
      m4[1][j] = m6[1] + m6[2];
      m4[2][j] = m6[1] - m6[2];
      m4[3][j] = m6[0] - m6[3];
    }

    //vertical
    for (i=0;i<4 && !lossless_qpprime;i++)
    {
      m6[0]=m4[i][0]+m4[i][2];
      m6[1]=m4[i][0]-m4[i][2];
      m6[2]=m4[i][1]-m4[i][3];
      m6[3]=m4[i][1]+m4[i][3];

      if(qp_per<4)
      {
        img->m7[0 ][i*4] = ((((m6[0] + m6[3])*invlevelscale[qp_rem][0][0]+(1<<(3-qp_per)))>>(4-qp_per))+2)>>2;
        img->m7[4 ][i*4] = ((((m6[1] + m6[2])*invlevelscale[qp_rem][0][0]+(1<<(3-qp_per)))>>(4-qp_per))+2)>>2;
        img->m7[8 ][i*4] = ((((m6[1] - m6[2])*invlevelscale[qp_rem][0][0]+(1<<(3-qp_per)))>>(4-qp_per))+2)>>2;
        img->m7[12][i*4] = ((((m6[0] - m6[3])*invlevelscale[qp_rem][0][0]+(1<<(3-qp_per)))>>(4-qp_per))+2)>>2;
      }
      else
      {
        img->m7[0 ][i*4] = ((((m6[0]+m6[3])*invlevelscale[qp_rem][0][0])<<(qp_per-4))+2)>>2;
        img->m7[4 ][i*4] = ((((m6[1]+m6[2])*invlevelscale[qp_rem][0][0])<<(qp_per-4))+2)>>2;
        img->m7[8 ][i*4] = ((((m6[1]-m6[2])*invlevelscale[qp_rem][0][0])<<(qp_per-4))+2)>>2;
        img->m7[12][i*4] = ((((m6[0]-m6[3])*invlevelscale[qp_rem][0][0])<<(qp_per-4))+2)>>2;
      }
    }
  }

  //     Quant of chroma AC-coeffs.
  coeff_cost=0;
  cr_cbp_tmp=0;

  for (b8=0; b8 < (img->num_blk8x8_uv >> 1); b8++)
  {
    for (b4=0; b4 < 4; b4++)
    {
      n1 = hor_offset[yuv][b8][b4];
      n2 = ver_offset[yuv][b8][b4];
      ACLevel = img->cofAC[4+b8+uv_scale][b4][0];
      ACRun   = img->cofAC[4+b8+uv_scale][b4][1];
      run=-1;
      scan_pos=0;

#ifdef RDO_Q
      if(input->UseRDO_Q && active_pps->entropy_coding_mode_flag == CABAC)
      {
        kStart=0; kStop=0; noCoeff=0;
        for (coeff_ctr=0;coeff_ctr < 15;coeff_ctr++)
        {
          i=pos_scan[coeff_ctr+1][0]; // scan is shifted due to DC
          j=pos_scan[coeff_ctr+1][1]; // scan is shifted due to DC

          levelData[coeff_ctr].levelDouble=absm(img->m7[n2+j][n1+i]*levelscale[qp_rem][i][j]);
          level = (int)(levelData[coeff_ctr].levelDouble >> q_bits);
          lowerInt=(((int)levelData[coeff_ctr].levelDouble-(level<<q_bits))<(1<<(q_bits-1)))? 1 : 0;

          levelData[coeff_ctr].level[0]=0;
          if (level==0 && lowerInt==1){
            levelData[coeff_ctr].noLevels=1;
          }
          else if (level==0 && lowerInt==0){
            levelData[coeff_ctr].level[1] = level+1;
            levelData[coeff_ctr].noLevels=2;
            kStop=coeff_ctr;
            noCoeff++;
          }
          else if (level>0 && lowerInt==1){
            levelData[coeff_ctr].level[1] = level;
            levelData[coeff_ctr].noLevels=2;
            kStop=coeff_ctr;
            noCoeff++;
          }
          else{
            levelData[coeff_ctr].level[1] = level;
            levelData[coeff_ctr].level[2] = level+1;
            levelData[coeff_ctr].noLevels=3;
            kStop=coeff_ctr;
            kStart=coeff_ctr;
            noCoeff++;
          }

          for (k=0; k<levelData[coeff_ctr].noLevels; k++){
            err=(double)(levelData[coeff_ctr].level[k]<<q_bits)-(double)levelData[coeff_ctr].levelDouble;
            levelData[coeff_ctr].errLevel[k]=(err*err*(double)estErr4x4[qp_rem][i][j])/normFact; 
          }
        }

        estBits=est_write_and_store_CBP_block_bit(currMB, LUMA_16AC);
        est_writeRunLevel_CABAC(levelData, levelTrellis, LUMA_16AC, lambda_md, kStart, kStop, noCoeff, estBits);
      }
#endif


      for (coeff_ctr=1; coeff_ctr < 16; coeff_ctr++)// start change rd_quant
      {

        i=pos_scan[coeff_ctr][0];
        j=pos_scan[coeff_ctr][1];

        ++run;
        ilev=0;

#ifdef RDO_Q
        if(input->UseRDO_Q)
        {
          level=levelTrellis[coeff_ctr-1];
        }
        else
#endif
        {
          if(lossless_qpprime)
            level = absm(img->m7[n2+j][n1+i]);
          else 
            level=(absm(img->m7[n2+j][n1+i])*levelscale[qp_rem][i][j]+leveloffset[qp_per][i][j])>>q_bits;
        }

        if (level  != 0)
        {
          currMB->cbp_blk |= ((int64)1) << cbp_blk_chroma[b8 + uv_scale][b4];
          if (level > 1 || lossless_qpprime)
            coeff_cost += MAX_VALUE;                // set high cost, shall not be discarded
          else
            coeff_cost += COEFF_COST[input->disthres][run];

          cr_cbp_tmp=2;
          ACLevel[scan_pos] = sign(level,img->m7[n2+j][n1+i]);
          ACRun  [scan_pos] = run;
          ++scan_pos;
          run=-1;

          level=sign(level, img->m7[n2+j][n1+i]);
          if(lossless_qpprime)
          {
            ilev = level;
          }
          else if(qp_per<4)
          {
            ilev=(level*invlevelscale[qp_rem][i][j]+(1<<(3-qp_per)))>>(4-qp_per);
          }
          else
          {
            ilev=(level*invlevelscale[qp_rem][i][j])<<(qp_per-4);
          }
        }
        if(!lossless_qpprime)
          img->m7[n2+j][n1+i]=ilev;
      }
      ACLevel[scan_pos] = 0;
    }
  }

  // * reset chroma coeffs
  if(coeff_cost < _CHROMA_COEFF_COST_ && !lossless_qpprime)
  {
    cr_cbp_tmp = 0 ;

    for (b8=0; b8 < (img->num_blk8x8_uv >> 1); b8++)
    {
      for (b4=0; b4 < 4; b4++)
      {
        n1 = hor_offset[yuv][b8][b4];
        n2 = ver_offset[yuv][b8][b4];
        ACLevel = img->cofAC[4+b8+uv_scale][b4][0];
        ACRun   = img->cofAC[4+b8+uv_scale][b4][1];
        if( DCcoded == 0) 
          currMB->cbp_blk &= ~((int64)cbpblk_pattern[yuv] << (uv << (1+yuv)));  // if no chroma DC's: then reset coded-bits of this chroma subblock

        ACLevel[0] = 0;
        for (coeff_ctr=1; coeff_ctr < 16; coeff_ctr++)// ac coeff
        {
          i=pos_scan[coeff_ctr][0];
          j=pos_scan[coeff_ctr][1];

          img->m7[n2+j][n1+i]=0;
          ACLevel[coeff_ctr] = 0;
        }
      }
    }
  }

  if(cr_cbp_tmp==2)   
    cr_cbp = 2;

  //     IDCT.
  //     Horizontal.

  for (n2=0; n2 < img->mb_cr_size_y && !lossless_qpprime; n2 += BLOCK_SIZE)
    for (n1=0; n1 < img->mb_cr_size_x; n1 += BLOCK_SIZE)
      inverse4x4(img->m7, img->m7, n2, n1);

  {
    //  Decoded block moved to memory
    imgpel *imgOrg;
    for (j=0; j < img->mb_cr_size_y; j++)
    {
      pix_c_y = img->pix_c_y+j;
      imgOrg = &enc_picture->imgUV[uv][pix_c_y][img->pix_c_x];
      for (i=0; i < img->mb_cr_size_x; i++)
      {
        pix_c_x = img->pix_c_x+i;
        img->m7[j][i] = Clip1((img->m7[j][i]+((long)img->mpr[j][i]<< DQ_BITS)+DQ_ROUND)>>DQ_BITS);
        *imgOrg++ = img->m7[j][i];
      }
    }
  }

  return cr_cbp;
}




#ifdef USE_INTRA_MDDT

const int KLTCol[9][4][4]=
{
  { // 0
    { -42,  -61,  -73,  -74},
    {  74,   65,  -16,  -81},
    { -80,   37,   73,  -57},
    { -53,   84,  -74,   33},
  },
  { // 1
    { -35,  -62,  -79,  -71},
    {  82,   65,  -24,  -70},
    { -78,   46,   59,  -68},
    { -49,   78,  -78,   42},
    },
    { // 2
      { -47,  -63,  -72,  -71},
      {  80,   57,  -25,  -78},
      { -75,   50,   67,  -62},
      { -46,   82,  -78,   37},
    },
    { // 3
      { -30,  -60,  -79,  -75},
      {  72,   73,  -10,  -76},
      { -84,   29,   69,  -61},
      { -56,   82,  -73,   34},
      },
      { // 4
        { -30,  -61,  -80,  -74},
        {  71,   73,  -11,  -77},
        { -85,   28,   68,  -62},
        { -57,   81,  -73,   35},
      },
      { // 5
        { -29,  -57,  -79,  -78},
        {  72,   74,   -7,  -75},
        { -83,   26,   71,  -61},
        { -58,   83,  -71,   32},
        },
        { // 6
          { -32,  -61,  -79,  -74},
          {  70,   73,  -14,  -77},
          { -85,   31,   67,  -61},
          { -56,   79,  -74,   38},
        },
        { // 7
          { -34,  -61,  -77,  -75},
          {  71,   71,  -10,  -79},
          {  83,  -28,  -72,   59},
          {  57,  -83,   72,  -32},
          },
          { // 8
            { -45,  -61,  -71,  -74},
            {  91,   51,  -31,  -68},
            { -68,   67,   54,  -66},
            { -38,   74,  -86,   45},
          },
};

const int KLTRow[9][4][4]=
{
  { // 0
    { -41,   86,  -74,  -42},
    { -62,   56,   56,   79},
    { -75,  -26,   58,  -82},
    { -72,  -71,  -66,   41},
  },
  { // 1
    { -37,   75,   79,   56},
    { -61,   67,  -33,  -84},
    { -74,  -13,  -74,   73},
    { -76,  -78,   60,  -30},
    },
    { // 2
      { -44,   84,  -73,  -45},
      { -63,   56,   51,   82},
      { -73,  -25,   65,  -79},
      { -72,  -75,  -65,   36},
    },
    { // 3
      { -47,   87,  -71,  -39},
      { -65,   50,   64,   75},
      { -73,  -34,   52,  -85},
      { -68,  -71,  -68,   46},
      },
      { // 4
        { -31,   72,  -81,  -60},
        { -60,   73,   29,   81},
        { -78,  -13,   70,  -72},
        { -75,  -75,  -63,   34},
      },
      { // 5
        { -34,   76,  -82,  -53},
        { -61,   70,   39,   79},
        { -78,  -18,   65,  -76},
        { -74,  -74,  -63,   39},
        },
        { // 6
          { -31,   73,   83,   57},
          { -60,   71,  -28,  -83},
          { -77,   -9,  -71,   73},
          { -76,  -77,   60,  -31},
        },
        { // 7
          { -46,   86,  -72,  -40},
          { -65,   52,   62,   75},
          { -73,  -34,   53,  -84},
          { -68,  -72,  -68,   46},
          },
          { // 8
            { -34,   75,   81,   55},
            { -60,   69,  -33,  -83},
            { -76,  -12,  -72,   73},
            { -76,  -77,   61,  -32},
          },
};

/*! 
*************************************************************************************
* \brief
*   Perform mode-dependent 4x4 directional transform 
*
* \para klt_luma_sep()
*    <paragraph>
*
* \author
*    - Yan Ye                      <yye@qualcomm.com>
*************************************************************************************
*/
int klt_luma_sep(int block_x,int block_y,int *coeff_cost, int ipmode)
{
  int sign(int a,int b);

  int i,j,k,coeff_ctr;
  int scan_pos,run;
  int nonzero;
  int qp_per,qp_rem, q_bits;

  int   pos_x   = block_x/BLOCK_SIZE;
  int   pos_y   = block_y/BLOCK_SIZE;
  int   b8      = 2*(pos_y/2) + (pos_x/2);
  int   b4      = 2*(pos_y%2) + (pos_x%2);
  int*  ACLevel = img->cofAC[b8][b4][0];
  int*  ACRun   = img->cofAC[b8][b4][1];
  byte (*pos_scan)[2] = img->scanOrder4x4[ipmode];

  int coeff [BLOCK_SIZE][BLOCK_SIZE], ilev[BLOCK_SIZE][BLOCK_SIZE], 
    temp2D[BLOCK_SIZE][BLOCK_SIZE], y2D [BLOCK_SIZE][BLOCK_SIZE];

  int m=4, n=10, KLTprec=14;

  int64 qpOffset;
  int64 roundOffset;
  int64 level;

  int R[6] = {40, 45, 50, 57, 63, 71};
  int Q[6] = {410,   364,   328,   287,   260,   231};

  long shift;
  int shiftBy2;

  Macroblock *currMB = &img->mb_data[img->current_mb_nr];

#ifdef RDO_Q
  levelDataStruct levelData[16];
  double lambda_md = 0.0;
  double err;
  int lowerInt;
  int levelRDOQ[16];
  int kStart = 0;
  int kStop = 0;
  int noCoeff;
  int estBits;

  int levelscale[4][4];
  int invlevelscale[4][4];
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
  double normFact = pow(2, 2 * (KLTprec + m + n + 8) - 15)*(1<<(2*img->BitDepthIncrease));
#else
  double normFact = pow(2, 2 * (KLTprec + m + n + 8) - 15);
#endif
#endif

  pos_scan = img->scanOrder4x4[ipmode];

  qp_per    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)/6; 
  qp_rem    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)%6; 
  q_bits    = Q_BITS + qp_per;

  qpOffset=n+qp_per+KLTprec-2+4;
  roundOffset=((int64)1<<(qpOffset))/3;

  for(j=0;j<4;j++)
    for(i=0;i<4;i++)
    {
      levelscale[i][j]    = Q[qp_rem]<<4;
      invlevelscale[i][j] = R[qp_rem]<<4;
    }

#ifdef RDO_Q
    if(input->UseRDO_Q)
    {
      if ((img->type == B_SLICE) && img->nal_reference_idc)
      {
        lambda_md = img->lambda_md[5][img->masterQP];  
      }
      else
      {
        lambda_md = img->lambda_md[img->type][img->masterQP];
      }
    }
#endif

    // column transform first
    for (i=0;i<(BLOCK_SIZE);i++){
      for (j=0;j<(BLOCK_SIZE);j++){
        temp2D[i][j] = 0;
        for(k = 0; k < BLOCK_SIZE; k++)
          temp2D[i][j]+=KLTCol[ipmode][i][k]*(long)(img->m7[k][j]);
      }
    }

    // row transform second
    for (i=0;i<(BLOCK_SIZE);i++){
      for (j=0;j<(BLOCK_SIZE);j++){
        coeff[i][j] = 0;
        for(k = 0; k < BLOCK_SIZE; k++)
          coeff[i][j]+=temp2D[i][k]*KLTRow[ipmode][k][j];
      }
    }

#ifdef RDO_Q
    if(input->UseRDO_Q)
    {
      noCoeff = 0;
      for (coeff_ctr = 0; coeff_ctr < 16; coeff_ctr++)
      {
        i = pos_scan[coeff_ctr][0];
        j = pos_scan[coeff_ctr][1];

        levelData[coeff_ctr].levelDouble = (int64)absm(coeff[j][i]) * (int64)(levelscale[i][j]);
        level = (int)(levelData[coeff_ctr].levelDouble >> qpOffset);
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
        lowerInt = ((levelData[coeff_ctr].levelDouble - ((int64)level << qpOffset)) < ((int64)1 << (qpOffset - 1)))? 1: 0;
#else
        lowerInt = ((levelData[coeff_ctr].levelDouble - (level << qpOffset)) < ((int64)1 << (qpOffset - 1)))? 1: 0;
#endif

        levelData[coeff_ctr].level[0] = 0;
        if (level == 0 && lowerInt == 1)
        {
          levelData[coeff_ctr].noLevels = 1;
        }
        else if (level == 0 && lowerInt == 0)
        {
          levelData[coeff_ctr].level[1] = level + 1;
          levelData[coeff_ctr].noLevels = 2;
          kStop = coeff_ctr;
          noCoeff++;
        }
        else if (level > 0 && lowerInt == 1)
        {
          levelData[coeff_ctr].level[1] = level;
          levelData[coeff_ctr].noLevels = 2;
          kStop = coeff_ctr;
          noCoeff++;
        }
        else
        {
          levelData[coeff_ctr].level[1] = level;
          levelData[coeff_ctr].level[2] = level + 1;
          levelData[coeff_ctr].noLevels = 3;
          kStop = coeff_ctr;
          kStart = coeff_ctr;
          noCoeff++;
        }

        for (k = 0; k < levelData[coeff_ctr].noLevels; k++)
        {      
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
          err = (double)((int64)levelData[coeff_ctr].level[k] << qpOffset) - (double) levelData[coeff_ctr].levelDouble;
#else
          err = (double)(levelData[coeff_ctr].level[k] << qpOffset) - (double) levelData[coeff_ctr].levelDouble;
#endif
          err = err * (double)invlevelscale[i][j];
          levelData[coeff_ctr].errLevel[k] = (err * err) / normFact; 
        }
      }

      estBits = est_write_and_store_CBP_block_bit(currMB, LUMA_4x4);
      est_writeRunLevel_CABAC(levelData, levelRDOQ, LUMA_4x4, lambda_md, kStart, kStop, noCoeff, estBits);
    }
#endif

    // quantization and zig-zag scanning 
    nonzero=FALSE;

    run=-1;
    scan_pos=0;

    for (coeff_ctr=0;coeff_ctr < 16;coeff_ctr++)
    {
      i=pos_scan[coeff_ctr][0];
      j=pos_scan[coeff_ctr][1];

      run++;

      if(input->UseRDO_Q)
      {
        level = levelRDOQ[coeff_ctr];
      }
      else
      {
        level = (absm((int64)coeff[j][i])*(int64)levelscale[i][j]+roundOffset);
        level = level >> (qpOffset);
      }

      if (level != 0)
      {
        nonzero=TRUE;

        *coeff_cost += (level > 1) ? MAX_VALUE : COEFF_COST[input->disthres][run];

        ACLevel[scan_pos] = sign((int)level, coeff[j][i]);
        ACRun  [scan_pos] = run;
        ++scan_pos;
        run=-1;                     // reset zero level counter

        ilev[j][i] = sign((int)level, coeff[j][i]) * ((invlevelscale[i][j] << qp_per) >> 4); 
      }
      else 
        ilev[j][i]=0;
    }
    ACLevel[scan_pos] = 0;

    // inverse transform, column first
    for (i=0;i<(BLOCK_SIZE);i++){
      for (j=0;j<(BLOCK_SIZE);j++){
        temp2D[i][j] = 0;
        for(k = 0; k < BLOCK_SIZE; k++)
          temp2D[i][j]+=KLTCol[ipmode][k][i]*(ilev[k][j]);
      }
    }

    // inverse transform, row second 
    for (i=0;i<(BLOCK_SIZE);i++)
    {
      for (j=0;j<(BLOCK_SIZE);j++)
      {
        y2D[i][j] = 0;
        for(k = 0; k < BLOCK_SIZE; k++)
          y2D[i][j]+=temp2D[i][k]*KLTRow[ipmode][j][k];
      }
    }

    // store the encoded image
    shift=KLTprec+m+2; shiftBy2=1<<(shift-1);
    for (j=0; j < BLOCK_SIZE; j++)
      for (i=0; i < BLOCK_SIZE; i++)
      {
        short temp;

        temp=(short)((y2D[j][i]+((int64)img->mpr[j+block_y][i+block_x]<<shift)+shiftBy2)>>shift);
        temp=min(img->max_imgpel_value,max(0,temp));
        enc_picture->imgY[img->pix_y+block_y+j][img->pix_x+block_x+i]=temp;
      }

      return nonzero;
}


const int KLTRow8[9][8][8] = 
{ // 0 
  {
    { -22,   57,   63,  -59,   53,  -41,   23,   13},
    { -30,   65,   41,    7,  -46,   59,  -55,  -29},
    { -38,   51,  -16,   62,  -37,  -17,   66,   47},
    { -45,   30,  -58,   31,   46,  -46,  -36,  -60},
    { -51,    2,  -57,  -44,   31,   55,  -18,   64},
    { -55,  -27,  -16,  -59,  -51,    1,   54,  -58},
    { -56,  -48,   32,    4,  -37,  -60,  -57,   41},
    { -51,  -48,   51,   51,   55,   45,   27,  -17},
  },
  { // 1
    { -18,  -38,   50,  -57,   57,  -56,   45,  -24},
    { -30,  -58,   57,  -27,  -13,   45,  -63,   44},
    { -38,  -57,   14,   49,  -58,   17,   45,  -59},
    { -44,  -42,  -42,   55,   22,  -58,   -3,   64},
    { -50,  -15,  -65,  -20,   53,   35,  -40,  -59},
    { -54,   19,  -37,  -63,  -37,   27,   60,   45},
    { -57,   49,   17,  -15,  -50,  -66,  -52,  -28},
    { -56,   60,   53,   49,   49,   37,   21,   10},
    },
    { // 2
      { -24,  -55,   65,  -51,   47,  -49,   31,  -20},
      { -34,  -62,   45,    1,  -33,   60,  -55,   41},
      { -41,  -50,  -19,   62,  -48,   -6,   46,  -60},
      { -47,  -30,  -60,   30,   45,  -47,  -16,   65},
      { -50,    0,  -53,  -50,   40,   51,  -28,  -59},
      { -53,   29,  -14,  -58,  -54,    7,   63,   47},
      { -54,   50,   26,    1,  -37,  -62,  -66,  -26},
      { -50,   53,   50,   55,   54,   40,   31,    8},
    },
    { // 3
      { -31,   61,   64,  -55,   48,  -37,   25,   12},
      { -39,   63,   30,   13,  -49,   60,  -54,  -30},
      { -45,   44,  -29,   63,  -35,  -24,   61,   47},
      { -49,   17,  -62,   19,   56,  -39,  -37,  -59},
      { -51,  -12,  -50,  -52,   25,   60,  -10,   64},
      { -51,  -37,   -7,  -55,  -58,   -9,   51,  -58},
      { -49,  -50,   36,   10,  -27,  -58,  -64,   43},
      { -44,  -49,   52,   55,   51,   48,   33,  -18},
      },
      { // 4
        { -11,  -25,   42,  -56,   61,  -61,   49,  -30},
        { -22,  -49,   61,  -46,    4,   40,  -61,   48},
        { -34,  -61,   34,   30,  -62,   25,   40,  -58},
        { -44,  -53,  -24,   59,   10,  -60,    1,   61},
        { -52,  -25,  -63,   -4,   57,   33,  -40,  -56},
        { -58,   15,  -45,  -62,  -30,   26,   59,   44},
        { -60,   50,   16,  -17,  -48,  -61,  -54,  -29},
        { -55,   59,   55,   51,   47,   37,   25,   12},
      },
      { // 5
        { -12,   28,   44,  -60,   63,  -61,   42,   20},
        { -24,   53,   60,  -40,   -7,   51,  -62,  -36},
        { -36,   62,   27,   37,  -59,   11,   54,   51},
        { -46,   49,  -30,   56,   20,  -57,  -18,  -62},
        { -52,   19,  -64,  -11,   53,   41,  -27,   63},
        { -57,  -19,  -42,  -62,  -35,   17,   55,  -53},
        { -59,  -50,   20,  -14,  -46,  -58,  -55,   36},
        { -54,  -57,   56,   50,   48,   38,   27,  -15},
        },
        { // 6
          { -10,  -27,   45,  -61,   54,  -63,   49,  -25},
          { -22,  -50,   60,  -43,    0,   42,  -65,   45},
          { -34,  -61,   31,   34,  -59,   28,   44,  -57},
          { -44,  -52,  -24,   57,   10,  -62,   -2,   63},
          { -52,  -25,  -61,   -7,   60,   26,  -39,  -59},
          { -57,   15,  -45,  -59,  -27,   32,   59,   47},
          { -60,   49,   13,  -18,  -56,  -58,  -50,  -29},
          { -55,   59,   57,   50,   50,   32,   21,   11},
        },
        { // 7
          { -30,   60,   64,  -56,   50,  -38,   25,   12},
          { -38,   63,   32,   11,  -48,   60,  -55,  -30},
          { -44,   46,  -26,   63,  -35,  -22,   61,   48},
          { -48,   19,  -62,   22,   53,  -41,  -35,  -60},
          { -51,  -10,  -52,  -51,   27,   58,  -13,   64},
          { -52,  -35,   -9,  -56,  -57,   -7,   52,  -57},
          { -51,  -50,   36,    9,  -29,  -59,  -62,   42},
          { -45,  -49,   51,   54,   52,   48,   32,  -18},
          },
          { // 8
            { -13,  -33,   47,  -61,   58,  -57,   45,  -25},
            { -25,  -55,   60,  -34,   -8,   47,  -61,   45},
            { -36,  -61,   22,   44,  -60,   13,   42,  -58},
            { -46,  -47,  -36,   54,   24,  -58,   -7,   63},
            { -52,  -17,  -64,  -18,   53,   40,  -34,  -58},
            { -57,   20,  -36,  -60,  -41,   21,   60,   46},
            { -58,   49,   20,  -14,  -44,  -62,  -58,  -29},
            { -54,   57,   55,   51,   49,   37,   26,   10},
          },
};

const int KLTCol8[9][8][8] = 
{
  {  // 0
    { -22,  -32,  -40,  -44,  -49,  -53,  -55,  -55},
    { -37,  -52,  -56,  -43,  -15,   19,   51,   64},
    {  43,   54,   24,  -34,  -70,  -46,   14,   53},
    { -60,  -31,   39,   61,   -5,  -59,  -23,   50},
    {  64,   -9,  -58,   10,   53,  -31,  -52,   48},
    { -58,   54,   16,  -62,   39,   22,  -56,   30},
    {  35,  -61,   50,  -12,  -36,   62,  -55,   23},
    { -21,   45,  -59,   61,  -57,   49,  -32,   13},
  },
  {  // 1
    { -26,  -37,  -42,  -43,  -47,  -54,  -55,  -49},
    {  66,   66,   37,   16,    2,  -19,  -49,  -57},
    { -53,  -38,   32,   79,   48,  -20,  -42,  -18},
    { -30,   28,   63,    9,  -67,  -61,   11,   48},
    {  68,  -26,  -43,   29,   16,  -56,  -23,   67},
    { -43,   57,   -9,  -51,   53,   -1,  -58,   48},
    {  29,  -57,   60,  -29,  -23,   55,  -59,   30},
    { -17,   34,  -51,   60,  -62,   55,  -40,   18},
    },
    { // 2
      { -29,  -38,  -43,  -46,  -49,  -51,  -51,  -49},
      {  52,   61,   50,   26,   -4,  -32,  -53,  -55},
      {  55,   43,   -9,  -58,  -64,  -22,   33,   50},
      { -57,   -6,   59,   40,  -37,  -61,   -3,   53},
      {  58,  -34,  -50,   40,   37,  -50,  -36,   51},
      { -42,   55,   -5,  -54,   52,    5,  -62,   45},
      {  33,  -62,   57,  -24,  -25,   55,  -56,   27},
      {  17,  -36,   52,  -59,   62,  -55,   39,  -16},
    },
    { // 3
      { -13,  -27,  -37,  -44,  -52,  -57,  -58,  -53},
      { -27,  -52,  -62,  -50,  -17,   21,   51,   56},
      {  43,   61,   27,  -35,  -67,  -37,   21,   50},
      { -56,  -39,   36,   56,  -15,  -62,  -12,   55},
      {  67,   -3,  -59,   18,   50,  -37,  -44,   49},
      { -60,   46,   13,  -58,   40,   18,  -61,   39},
      {  44,  -61,   47,  -10,  -34,   59,  -56,   26},
      { -24,   43,  -57,   62,  -59,   49,  -32,   12},
      },
      { // 4
        { -11,  -22,  -34,  -44,  -53,  -59,  -60,  -53},
        {  22,   47,   62,   54,   25,  -15,  -51,  -59},
        {  39,   62,   38,  -23,  -64,  -43,   16,   54},
        { -58,  -47,   29,   59,   -5,  -58,  -16,   52},
        {  68,    1,  -59,   14,   52,  -33,  -46,   47},
        { -62,   46,   14,  -57,   37,   22,  -60,   38},
        {  45,  -63,   49,  -12,  -32,   58,  -54,   25},
        {  21,  -40,   55,  -63,   61,  -50,   32,  -12},
      },
      { // 5
        { -11,  -22,  -32,  -42,  -51,  -58,  -61,  -57},
        { -24,  -45,  -60,  -55,  -31,    9,   49,   62},
        {  42,   62,   40,  -17,  -62,  -47,   11,   52},
        { -59,  -45,   28,   61,   -1,  -58,  -21,   50},
        {  63,    2,  -59,    7,   57,  -28,  -51,   48},
        { -62,   43,   24,  -61,   31,   27,  -59,   34},
        {  46,  -64,   45,   -5,  -38,   60,  -51,   23},
        { -24,   45,  -58,   61,  -57,   48,  -31,   12},
        },
        { // 6
          { -13,  -26,  -37,  -45,  -52,  -56,  -59,  -54},
          {  27,   53,   62,   48,   18,  -19,  -51,  -57},
          {  40,   58,   29,  -30,  -65,  -43,   17,   57},
          { -57,  -45,   34,   60,   -9,  -60,  -15,   50},
          {  65,   -3,  -62,   19,   53,  -36,  -43,   45},
          { -66,   51,    7,  -54,   43,   12,  -55,   37},
          {  42,  -61,   54,  -25,  -19,   54,  -59,   29},
          {  17,  -34,   50,  -60,   63,  -56,   39,  -15},
        },
        { // 7
          { -14,  -27,  -37,  -44,  -52,  -56,  -58,  -55},
          { -29,  -52,  -61,  -48,  -20,   18,   51,   60},
          {  43,   59,   29,  -30,  -67,  -42,   17,   53},
          { -55,  -39,   32,   60,   -6,  -62,  -20,   54},
          {  66,   -1,  -61,   10,   54,  -30,  -48,   46},
          { -62,   47,   17,  -60,   37,   22,  -59,   35},
          {  43,  -61,   47,  -10,  -35,   60,  -55,   25},
          { -23,   44,  -57,   61,  -58,   49,  -31,   12},
          },
          { // 8
            { -29,  -37,  -42,  -46,  -49,  -51,  -52,  -50},
            {  67,   68,   44,   13,  -15,  -33,  -44,  -45},
            {  66,   21,  -43,  -67,  -43,    1,   37,   47},
            { -54,   23,   61,    1,  -62,  -46,   15,   54},
            {  45,  -50,  -22,   59,   12,  -64,  -25,   55},
            { -34,   58,  -28,  -35,   61,  -15,  -59,   49},
            {  24,  -53,   63,  -41,   -5,   49,  -63,   31},
            {  10,  -27,   44,  -57,   64,  -62,   46,  -19},
          },
};

static int KLTCol8inProd[9][8], KLTRow8inProd[9][8];

void compute_inner_product8x8(int matrix[8][8], int isLeft, int *inner_prod)
{
  int i, j;

  if(isLeft)
    for(i = 0; i < 8; i++)
    {
      inner_prod[i] = 0;
      for(j = 0; j < 4; j++)
      {
        inner_prod[i] += matrix[i][2*j+1]*matrix[i][2*j];
      }
    }
  else
    for(i = 0; i < 8; i++)
    {
      inner_prod[i] = 0;
      for(j = 0; j < 4; j++)
      {
        inner_prod[i] += matrix[2*j+1][i]*matrix[2*j][i];
      }
    }
}

/*! 
*************************************************************************************
* \brief
*   Precompute inner products for all 8x8 transform matrices
*
* \para precompute_all_inner_product8x8()
*    <paragraph>
*
* \author
*    - Yan Ye                      <yye@qualcomm.com>
*************************************************************************************
*/
void precompute_all_inner_product8x8()
{
  int i, j, k;
  int temp2D[8][8];

  for(i = 0; i < 9; i++)
  {
    if(i == 2) // handled by DCT 
      continue;

    for(j = 0; j < 8; j++)
      for(k = 0; k < 8; k++)
        temp2D[j][k] = KLTCol8[i][j][k];
    compute_inner_product8x8(temp2D, 1, KLTCol8inProd[i]);

    for(j = 0; j < 8; j++)
      for(k = 0; k < 8; k++)
        temp2D[j][k] = KLTRow8[i][j][k];
    compute_inner_product8x8(temp2D, 0, KLTRow8inProd[i]);
  }
}

/*! 
*************************************************************************************
* \brief
*   Perform fast mode-dependent 8x8 directional transform 
*
* \para klt_luma8x8_sep_fast()
*    <paragraph>
*
* \author
*    - Yan Ye                      <yye@qualcomm.com>
*************************************************************************************
*/

int klt_luma8x8_sep_fast(int b8,int *coeff_cost, int ipmode)
{
  int sign(int a,int b);

  int i,j,k,coeff_ctr;
  int scan_pos,run;
  int nonzero;
  int qp_per,qp_rem, q_bits;

  int block_x = 8*(b8 & 0x01);
  int block_y = 8*(b8 >> 1);
  int*  ACLevel = img->cofAC[b8][0][0];
  int*  ACRun   = img->cofAC[b8][0][1];

  int coeff [2*BLOCK_SIZE][2*BLOCK_SIZE], ilev [2*BLOCK_SIZE][2*BLOCK_SIZE], 
    temp2D[2*BLOCK_SIZE][2*BLOCK_SIZE], y2D  [2*BLOCK_SIZE][2*BLOCK_SIZE]; 
  int inprod[BLOCK_SIZE*2];
  int m=4, n=10, KLTprec=14;
  int64 qpOffset;
  int64 roundOffset;
  int64 level;

  int R[6] = {40, 45, 50, 57, 63, 71};
  int Q[6] = {410,   364,   328,   287,   260,   231};
  long shift, shiftBy2;
  int MCcoeff;
  int runs[4], scan_poss[4]; 

#ifdef RDO_Q
  levelDataStruct levelData[64];
  double lambda_md = 0.0;
  double err;
  int lowerInt;
  int levelRDOQ[64];
  int kStart = 0;
  int kStop = 0;
  int noCoeff;

  int levelscale[8][8];
  int invlevelscale[8][8];
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
  double normFact = pow(2, 2 * (KLTprec + m + n + 8) - 15)*(1<<(2*img->BitDepthIncrease));
#else
  double normFact = pow(2, 2 * (KLTprec + m + n + 8) - 15);
#endif
#endif


  qp_per    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)/6; 
  qp_rem    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)%6; 
  q_bits    = Q_BITS + qp_per;

  qpOffset=n+qp_per+KLTprec+2;
  roundOffset=((int64)1<<(qpOffset))/3;

  for(j=0;j<8;j++)
    for(i=0;i<8;i++)
    {
      levelscale[i][j]    = Q[qp_rem]<<4;
      invlevelscale[i][j] = R[qp_rem]<<4;
    }

#ifdef RDO_Q
    if(input->UseRDO_Q)
    {
      if ((img->type == B_SLICE) && img->nal_reference_idc)
      {
        lambda_md = img->lambda_md[5][img->masterQP];  
      }
      else
      {
        lambda_md = img->lambda_md[img->type][img->masterQP];
      }
    }
#endif

    for (j=0;j<(2*BLOCK_SIZE);j++)
      for(k = 0; k < 2*BLOCK_SIZE; k++)
        temp2D[j][k] = img->m7[j][k];

    compute_inner_product8x8(temp2D, 0, inprod);

    // column transform first
    for (i=0;i<(2*BLOCK_SIZE);i++)
    {
      for (j=0;j<(2*BLOCK_SIZE);j++)
      {
        temp2D[i][j] = 0;
        for(k = 0; k < BLOCK_SIZE; k++)
          temp2D[i][j]+=
          (KLTCol8[ipmode][i][2*k]+img->m7[2*k+1][j])*(KLTCol8[ipmode][i][2*k+1]+img->m7[2*k][j]);
        temp2D[i][j] = temp2D[i][j] - KLTCol8inProd[ipmode][i] - inprod[j];
      }
    }

    compute_inner_product8x8(temp2D, 1, inprod);

    // row transform second
    for (i=0;i<(2*BLOCK_SIZE);i++)
    {
      for (j=0;j<(2*BLOCK_SIZE);j++)
      {
        coeff[i][j] = 0;
        for(k = 0; k < BLOCK_SIZE; k++)
          coeff[i][j]+=
          (temp2D[i][2*k+1]+KLTRow8[ipmode][2*k][j])*(temp2D[i][2*k]+KLTRow8[ipmode][2*k+1][j]);
        coeff[i][j] = coeff[i][j] - inprod[i] - KLTRow8inProd[ipmode][j];
      }
    }

#ifdef RDO_Q
    if(input->UseRDO_Q)
    {
      noCoeff = 0;
      for (coeff_ctr = 0; coeff_ctr < 64; coeff_ctr++)
      {
        i=img->scanOrder8x8[ipmode][coeff_ctr][0];
        j=img->scanOrder8x8[ipmode][coeff_ctr][1];

        levelData[coeff_ctr].levelDouble = (int64)absm(coeff[j][i]) * (int64)(levelscale[i][j]);
        level = (int)(levelData[coeff_ctr].levelDouble >> qpOffset);
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
        lowerInt = ((levelData[coeff_ctr].levelDouble - ((int64)level << qpOffset)) < ((int64) 1 << (qpOffset - 1)))? 1: 0;
#else
        lowerInt = ((levelData[coeff_ctr].levelDouble - (level << qpOffset)) < ((int64) 1 << (qpOffset - 1)))? 1: 0;
#endif
        levelData[coeff_ctr].level[0] = 0;

        if (level == 0 && lowerInt == 1)
        {
          levelData[coeff_ctr].noLevels = 1;
        }
        else if (level == 0 && lowerInt == 0)
        {
          levelData[coeff_ctr].level[1] = level + 1;
          levelData[coeff_ctr].noLevels = 2;
          kStop = coeff_ctr;
          noCoeff++;
        }
        else if (level > 0 && lowerInt == 1)
        {
          levelData[coeff_ctr].level[1] = level;
          levelData[coeff_ctr].noLevels = 2;
          kStop = coeff_ctr;
          noCoeff++;
        }
        else
        {
          levelData[coeff_ctr].level[1] = level;
          levelData[coeff_ctr].level[2] = level + 1;
          levelData[coeff_ctr].noLevels = 3;
          kStop = coeff_ctr;
          kStart = coeff_ctr;
          noCoeff++;
        }

        for (k = 0; k < levelData[coeff_ctr].noLevels; k++)
        {      
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
          err = (double)((int64)levelData[coeff_ctr].level[k] << qpOffset) - (double) levelData[coeff_ctr].levelDouble;
#else
          err = (double)(levelData[coeff_ctr].level[k] << qpOffset) - (double) levelData[coeff_ctr].levelDouble;
#endif
          err = err * (double)invlevelscale[i][j];
          levelData[coeff_ctr].errLevel[k] = (err * err) / normFact; 
        }
      }

      est_writeRunLevel_CABAC(levelData, levelRDOQ, LUMA_8x8, lambda_md, kStart, kStop, noCoeff, 0);
    }
#endif

    nonzero=FALSE;

    run=-1;
    scan_pos=0;

    runs[0]=runs[1]=runs[2]=runs[3]=-1;
    scan_poss[0]=scan_poss[1]=scan_poss[2]=scan_poss[3]=0;

    for (coeff_ctr=0;coeff_ctr < 64;coeff_ctr++)
    {   
      i=img->scanOrder8x8[ipmode][coeff_ctr][0];
      j=img->scanOrder8x8[ipmode][coeff_ctr][1];

      run++;
      ilev[j][i]=0;

      MCcoeff = (coeff_ctr&3);
      runs[MCcoeff]++;

#ifdef RDO_Q
      if(input->UseRDO_Q)
      {
        level = levelRDOQ[coeff_ctr];
      }
      else
#endif
      {
        level = (absm((int64)coeff[j][i])*(int64)levelscale[i][j]+roundOffset);
        level = level >> (qpOffset);
      }

      if (level != 0)
      {
        nonzero=TRUE;

        if (level > 1)
          *coeff_cost += MAX_VALUE;                // set high cost, shall not be discarded
        else
          *coeff_cost += COEFF_COST8x8[input->disthres][run];

        ACLevel[scan_pos] = sign((int)level, coeff[j][i]);
        ACRun  [scan_pos] = run;
        ++scan_pos;
        run=-1;                     // reset zero level counter

        ilev[j][i] = sign((int)level, coeff[j][i]) * ((invlevelscale[i][j] << qp_per) >> 4); 
      }
      else
        ilev[j][i] = 0;
    }

    ACLevel[scan_pos] = 0;

    // inverse transform, column first
    for (i=0;i<(2*BLOCK_SIZE);i++)
    {
      for (j=0;j<(2*BLOCK_SIZE);j++)
      {
        temp2D[i][j] = 0;
        for(k = 0; k < 2*BLOCK_SIZE; k++)
          temp2D[i][j]+=KLTCol8[ipmode][k][i]*(ilev[k][j]);
      }
    }

    // inverse transform, row second 
    for (i=0;i<(2*BLOCK_SIZE);i++)
    {
      for (j=0;j<(2*BLOCK_SIZE);j++)
      {
        y2D[i][j] = 0;
        for(k = 0; k < 2*BLOCK_SIZE; k++)
          y2D[i][j]+=temp2D[i][k]*KLTRow8[ipmode][j][k];
      }
    }

    shift=KLTprec+m+2; shiftBy2=1<<(shift-1);
    for (j=0;j<2*BLOCK_SIZE;j++)
    {
      for (i=0;i<2*BLOCK_SIZE;i++)
      {
        int temp;

#ifdef  INTERNAL_BIT_DEPTH_INCREASE
        temp=(int)((y2D[j][i]+((int64)img->mpr[j+block_y][i+block_x]<<shift)+shiftBy2)>>shift);
#else
        temp=(y2D[j][i]+(img->mpr[j+block_y][i+block_x]<<shift)+shiftBy2)>>shift;
#endif
        temp=min(img->max_imgpel_value,max(0,temp));
        enc_picture->imgY[img->pix_y+block_y+j][img->pix_x+block_x+i]= (imgpel) temp;
      }
    }

    return nonzero;
}



const int KLTCol16[4][16][16]=
{
  {
    { -26,  -27,  -28,  -30,  -31,  -32,  -32,  -33,  -33,  -34,  -34,  -34,  -34,  -34,  -34,  -34},
    {  39,   41,   40,   38,   32,   27,   21,   16,   -3,   -9,  -17,  -24,  -34,  -39,  -44,  -46},
    {  36,   36,   29,   20,    0,  -12,  -24,  -32,  -49,  -45,  -36,  -22,    6,   24,   40,   49},
    { -40,  -36,  -22,   -3,   32,   46,   51,   45,  -21,  -33,  -37,  -33,  -13,    3,   19,   29},
    { -21,  -14,    3,   18,   29,   19,    0,  -19,  -52,  -32,    6,   43,   60,   33,  -17,  -58},
    {  50,   26,  -18,  -53,  -51,  -13,   34,   57,  -12,  -22,  -16,    2,   28,   21,   -4,  -27},
    { -28,   -6,   24,   32,   -9,  -27,  -16,   15,   56,    6,  -48,  -49,   26,   48,   11,  -39},
    {  49,   -2,  -52,  -36,   39,   46,   -7,  -52,   22,   20,  -12,  -31,    0,   29,   14,  -23},
    { -27,   14,   30,  -10,  -30,   12,   27,  -13,  -45,   33,   44,  -32,  -46,   36,   42,  -38},
    {  41,  -37,  -39,   41,   34,  -46,  -30,   49,  -25,    2,   26,   -3,  -31,   13,   26,  -20},
    { -19,   25,    8,  -27,   18,    6,  -20,    6,   40,  -53,  -12,   57,  -49,    1,   51,  -31},
    {  35,  -51,   -7,   52,  -55,   10,   51,  -40,   22,  -20,  -12,   27,  -16,   -4,   18,  -10},
    { -11,   24,  -21,    6,   16,  -31,   28,  -10,  -24,   55,  -53,   20,   22,  -52,   52,  -21},
    { -25,   55,  -53,   23,   18,  -48,   52,  -25,   13,  -22,   16,   -2,  -17,   33,  -30,   12},
    {  14,  -32,   39,  -30,   26,  -29,   20,   -6,  -14,   35,  -47,   40,  -40,   48,  -37,   14},
    {  15,  -33,   43,  -39,   41,  -49,   41,  -22,   18,  -32,   37,  -30,   26,  -28,   20,   -8},
  },
  {
    { -27,  -30,  -32,  -33,  -33,  -34,  -34,  -34,  -34,  -33,  -33,  -33,  -32,  -31,  -30,  -28},
    {  47,   46,   41,   35,   28,   20,   11,    4,   -8,  -14,  -22,  -29,  -35,  -39,  -42,  -42},
    {  53,   42,   22,    4,  -15,  -28,  -39,  -45,  -37,  -29,  -15,   -3,   15,   27,   38,   42},
    { -43,  -22,    5,   24,   37,   37,   30,   17,  -37,  -45,  -43,  -34,  -13,    9,   32,   44},
    {  54,   11,  -34,  -53,  -33,    0,   34,   51,   22,    2,  -20,  -35,  -30,   -9,   17,   32},
    { -29,    2,   27,   28,   -1,  -22,  -30,  -20,   51,   42,    7,  -31,  -56,  -31,   16,   51},
    {  33,  -16,  -46,  -22,   51,   46,   -4,  -49,  -24,   12,   35,   25,  -31,  -32,   -1,   27},
    { -24,   22,   32,   -7,  -38,  -15,   24,   31,  -49,  -19,   39,   51,  -20,  -48,  -14,   34},
    {  35,  -42,  -28,   48,   20,  -39,  -25,   34,   20,  -27,  -23,   27,   28,  -34,  -30,   36},
    { -26,   37,   14,  -46,    2,   34,    0,  -28,   41,  -11,  -43,   11,   49,  -26,  -42,   38},
    {  20,  -36,    6,   35,  -50,   15,   46,  -36,  -27,   39,    9,  -41,   35,   -3,  -37,   26},
    { -15,   28,  -10,  -22,   46,  -24,  -33,   41,  -45,   39,   25,  -51,   28,    9,  -35,   21},
    {   7,  -18,   18,   -7,   -9,   20,  -18,    5,   22,  -45,   36,   -4,  -38,   69,  -64,   27},
    { -23,   59,  -65,   35,    9,  -42,   51,  -27,   10,   -9,    0,    7,  -17,   25,  -22,    8},
    { -11,   29,  -39,   33,  -35,   45,  -34,   11,   17,  -43,   52,  -37,   28,  -29,   21,   -8},
    {   7,  -20,   30,  -30,   37,  -48,   44,  -25,   26,  -47,   49,  -36,   25,  -22,   15,   -6},
    },
    {
      { -25,  -28,  -31,  -32,  -33,  -34,  -35,  -36,  -35,  -34,  -34,  -33,  -32,  -31,  -29,  -27},
      {  49,   47,   42,   36,   27,   19,   10,    3,   -8,  -14,  -22,  -28,  -34,  -38,  -41,  -40},
      {  50,   41,   24,    6,  -15,  -28,  -38,  -43,  -37,  -28,  -15,   -2,   16,   30,   41,   46},
      { -38,  -23,    0,   20,   38,   40,   34,   20,  -34,  -45,  -45,  -37,  -12,   10,   31,   41},
      { -54,  -16,   32,   56,   37,    2,  -33,  -50,  -24,   -5,   17,   32,   30,   11,  -15,  -32},
      {  31,    2,  -27,  -32,    0,   24,   31,   17,  -52,  -41,   -4,   36,   53,   27,  -17,  -49},
      { -30,   12,   42,   20,  -45,  -43,    3,   47,   27,  -13,  -41,  -27,   34,   38,    1,  -35},
      {  33,  -20,  -41,    1,   43,   16,  -26,  -31,   47,   19,  -33,  -45,   16,   45,   13,  -35},
      { -34,   39,   31,  -45,  -25,   42,   28,  -42,  -19,   31,   20,  -32,  -22,   33,   24,  -30},
      {  25,  -32,  -13,   40,   -2,  -34,    4,   31,  -48,   15,   47,  -18,  -47,   29,   38,  -36},
      { -26,   42,   -1,  -41,   51,  -11,  -46,   36,   17,  -34,    2,   33,  -38,    7,   35,  -25},
      {  15,  -28,    7,   23,  -40,   20,   29,  -36,   44,  -41,  -20,   53,  -38,   -3,   41,  -25},
      { -22,   52,  -53,   24,   17,  -49,   52,  -23,  -11,   31,  -33,   15,    8,  -24,   26,  -11},
      {  13,  -33,   37,  -22,    3,   13,  -22,   16,  -26,   50,  -42,   11,   27,  -56,   55,  -24},
      {  13,  -33,   44,  -37,   38,  -46,   36,  -14,   -7,   24,  -34,   30,  -32,   41,  -34,   14},
      {   7,  -19,   27,  -27,   33,  -42,   39,  -23,   25,  -44,   49,  -39,   35,  -36,   27,  -11},
    },
    {
      { -21,  -24,  -27,  -30,  -32,  -34,  -35,  -36,  -36,  -36,  -35,  -35,  -34,  -33,  -31,  -29},
      {  42,   43,   42,   38,   32,   25,   16,    9,   -1,   -9,  -18,  -26,  -33,  -39,  -43,  -44},
      {  45,   42,   31,   17,   -4,  -18,  -31,  -39,  -44,  -37,  -23,   -8,   11,   26,   40,   47},
      { -40,  -31,  -10,   12,   36,   44,   42,   31,  -28,  -40,  -43,  -36,  -16,    5,   25,   36},
      {  47,   23,  -17,  -46,  -46,  -15,   23,   48,   34,   12,  -16,  -37,  -36,  -14,   16,   37},
      { -35,  -11,   22,   39,   16,  -14,  -35,  -33,   46,   39,    8,  -27,  -50,  -28,   15,   47},
      {  30,    0,  -34,  -33,   30,   38,    7,  -31,  -39,    5,   46,   40,  -38,  -44,   -4,   35},
      { -39,    9,   47,   21,  -45,  -34,   18,   45,  -39,  -19,   27,   41,  -13,  -37,  -12,   26},
      {  34,  -28,  -35,   30,   33,  -32,  -32,   33,   31,  -32,  -31,   34,   30,  -34,  -30,   33},
      { -32,   34,   27,  -41,  -17,   39,   13,  -39,   39,  -14,  -39,   17,   42,  -27,  -36,   33},
      {  30,  -39,   -9,   42,  -40,    6,   39,  -27,  -28,   39,    8,  -40,   41,   -8,  -40,   30},
      { -24,   35,    0,  -35,   46,  -16,  -37,   38,  -38,   37,   17,  -46,   34,    1,  -36,   23},
      {  19,  -42,   39,  -15,  -20,   47,  -47,   19,   17,  -43,   43,  -19,  -13,   35,  -37,   16},
      { -19,   44,  -45,   22,    7,  -29,   36,  -20,   22,  -41,   34,   -9,  -24,   49,  -47,   20},
      { -15,   36,  -43,   32,  -30,   37,  -28,    9,   10,  -31,   41,  -33,   35,  -47,   38,  -14},
      {  11,  -28,   36,  -33,   37,  -47,   41,  -22,   21,  -39,   43,  -34,   30,  -32,   24,   -8},
      },
};

const int KLTRow16[4][16][16]=
{
  {
    { -23,   41,   40,  -47,  -44,   32,  -41,   37,  -38,   22,  -37,    6,  -28,    2,  -14,   -6},
    { -27,   44,   39,  -34,  -26,    8,    6,  -17,   35,  -26,   50,   -9,   66,   -5,   31,   14},
    { -29,   43,   29,   -5,   11,  -25,   46,  -44,   31,   -9,    6,    1,  -74,    5,  -33,  -17},
    { -31,   40,   16,   23,   41,  -38,   33,   -4,  -35,   29,  -61,   11,   52,    1,   12,    5},
    { -31,   34,    0,   41,   45,  -11,  -34,   44,  -30,   -3,   61,  -18,  -33,  -11,   26,   23},
    { -33,   24,  -20,   43,   15,   30,  -44,   11,   43,  -24,  -13,    8,   30,   18,  -54,  -50},
    { -34,   13,  -39,   32,  -24,   47,   -4,  -32,   30,    4,  -39,   13,  -25,  -17,   49,   52},
    { -36,    5,  -47,   12,  -42,   25,   39,  -24,  -54,   37,   40,  -32,   13,   13,  -24,  -22},
    { -35,   -2,  -45,  -18,  -31,  -41,   24,   42,  -12,  -56,   -2,   56,   -5,  -20,   18,  -24},
    { -35,  -11,  -36,  -35,   -6,  -50,  -15,   24,   34,   16,  -11,  -42,    4,   36,  -34,   58},
    { -35,  -21,  -18,  -42,   27,  -18,  -40,  -30,   16,   41,   -2,  -19,   -5,  -40,   40,  -60},
    { -34,  -30,    0,  -36,   45,   24,  -21,  -44,  -35,  -14,   19,   61,    4,   37,  -18,   29},
    { -34,  -36,   16,  -18,   33,   44,   36,   21,  -11,  -47,  -23,  -53,    2,  -49,  -19,    9},
    { -33,  -40,   30,    9,    5,   23,   37,   45,   29,   26,    3,    7,   -9,   67,   44,  -26},
    { -31,  -41,   40,   32,  -26,  -15,    0,   10,   19,   46,   22,   41,    9,  -58,  -39,   21},
    { -27,  -37,   39,   40,  -41,  -38,  -36,  -38,  -30,  -45,  -17,  -32,   -4,   24,   16,   -8},
  },
  {
    { -23,  -34,   37,   37,  -38,   41,  -38,   45,  -38,   23,  -32,   32,   -6,    1,   34,   13},
    { -26,  -39,   40,   35,  -29,   22,   -7,   -5,   23,  -21,   37,  -40,   10,   -2,  -72,  -28},
    { -28,  -41,   34,   19,    1,  -18,   34,  -49,   39,  -18,   15,  -10,   -7,    5,   74,   29},
    { -30,  -39,   23,   -4,   32,  -46,   37,  -22,  -21,   23,  -48,   51,    3,  -14,  -48,   -9},
    { -31,  -35,    5,  -31,   49,  -31,  -13,   44,  -40,   16,   26,  -55,   -2,   30,   23,  -30},
    { -33,  -28,  -15,  -45,   29,   12,  -38,   30,   31,  -35,   14,   20,    5,  -45,  -16,   62},
    { -34,  -20,  -31,  -45,   -7,   43,  -16,  -25,   37,   -9,  -20,   34,   -4,   45,   12,  -63},
    { -35,  -12,  -40,  -29,  -35,   36,   24,  -40,  -35,   46,  -12,  -50,    0,  -31,   -6,   28},
    { -35,   -2,  -47,   11,  -45,  -24,   42,   24,  -34,  -31,   51,   41,    3,   37,   -8,   17},
    { -35,    8,  -42,   33,  -23,  -42,    0,   35,   32,  -11,  -34,  -20,   -5,  -60,   21,  -44},
    { -35,   18,  -29,   45,   15,  -23,  -42,  -12,   30,   30,  -28,  -18,   10,   62,  -22,   45},
    { -35,   28,  -12,   41,   44,   16,  -34,  -44,  -29,   15,   54,   30,  -24,  -35,    9,  -23},
    { -34,   36,   10,   19,   41,   44,   34,    5,  -27,  -57,  -35,  -17,   56,    5,    7,   -1},
    { -34,   42,   27,   -7,   13,   27,   47,   38,   28,   19,   -2,   -4,  -82,   10,  -14,   10},
    { -33,   44,   39,  -29,  -21,  -12,    4,   14,   28,   55,   31,   18,   70,   -9,   10,   -8},
    { -30,   42,   41,  -38,  -40,  -40,  -43,  -32,  -32,  -46,  -19,  -11,  -28,    4,   -3,    2},
    },
    {
      { -23,   43,   45,  -42,  -43,   36,  -38,  -37,   38,  -28,   28,  -15,  -23,   21,   -2,   -4},
      { -28,   46,   41,  -30,  -22,    8,    6,   18,  -36,   35,  -38,   23,   54,  -49,    5,   10},
      { -31,   43,   28,   -3,   15,  -28,   43,   43,  -31,   11,   -4,   -3,  -60,   56,   -5,  -12},
      { -33,   38,   11,   24,   43,  -40,   29,    3,   37,  -41,   48,  -26,   37,  -37,   -7,    1},
      { -34,   30,   -8,   41,   43,   -8,  -35,  -44,   27,    9,  -50,   45,  -15,   14,   32,   27},
      { -35,   20,  -26,   42,   11,   30,  -41,  -15,  -40,   28,   12,  -27,    8,   -1,  -55,  -56},
      { -36,    9,  -39,   30,  -26,   44,   -2,   32,  -29,   -9,   33,  -25,   -5,   -2,   55,   57},
      { -36,    1,  -44,   10,  -45,   23,   36,   31,   43,  -34,  -24,   53,    1,   -1,  -35,  -24},
      { -36,   -6,  -41,  -24,  -35,  -39,   28,  -37,   25,   43,  -24,  -56,    5,    3,   33,  -23},
      { -35,  -14,  -31,  -40,   -8,  -47,  -13,  -30,  -35,   -3,   29,   29,  -11,   -2,  -52,   56},
      { -34,  -23,  -15,  -45,   26,  -17,  -41,   24,  -23,  -37,   11,   27,   12,    4,   54,  -58},
      { -33,  -32,    3,  -36,   45,   25,  -24,   46,   33,    2,  -43,  -48,  -17,  -16,  -31,   29},
      { -32,  -37,   19,  -14,   34,   45,   38,  -15,   20,   51,   43,   29,   35,   41,    4,    3},
      { -30,  -40,   32,   12,    5,   24,   41,  -45,  -29,  -22,   -7,    2,  -55,  -60,   11,  -16},
      { -27,  -41,   41,   34,  -25,  -15,   -1,  -12,  -24,  -48,  -36,  -25,   49,   52,  -10,   13},
      { -24,  -37,   41,   41,  -39,  -41,  -40,   37,   31,   44,   29,   16,  -20,  -21,    4,   -5},
    },
    {
      { -19,   37,   41,  -42,   45,  -39,   35,   39,   38,  -25,   33,  -25,   -8,   32,   -4,   -6},
      { -23,   42,   42,  -34,   28,  -13,    1,  -10,  -29,   27,  -39,   32,   18,  -73,   10,   15},
      { -27,   42,   34,  -12,   -8,   24,  -36,  -46,  -36,   16,  -15,    5,  -16,   78,  -10,  -17},
      { -30,   40,   21,   13,  -39,   42,  -35,  -17,   28,  -31,   51,  -41,    6,  -49,   -3,    3},
      { -32,   34,    4,   36,  -47,   17,   26,   45,   37,  -10,  -37,   52,   -1,   22,   25,   31},
      { -34,   26,  -16,   44,  -20,  -22,   38,   27,  -36,   33,    1,  -23,    3,  -14,  -45,  -64},
      { -36,   17,  -33,   39,   17,  -44,    9,  -26,  -36,    6,   24,  -31,   -3,   11,   45,   65},
      { -37,    8,  -41,   22,   41,  -31,  -28,  -42,   40,  -42,   -6,   50,   -1,   -7,  -28,  -28},
      { -37,    1,  -45,  -17,   38,   35,  -35,   32,   31,   38,  -41,  -47,    7,    3,   34,  -18},
      { -36,   -8,  -37,  -36,   16,   46,    5,   32,  -33,   -1,   32,   27,  -14,    1,  -61,   47},
      { -36,  -19,  -22,  -44,  -20,   20,   43,  -18,  -28,  -33,   23,   21,   17,    0,   63,  -48},
      { -35,  -29,   -4,  -39,  -43,  -20,   36,  -45,   30,   -2,  -50,  -40,  -24,   -5,  -33,   23},
      { -34,  -35,   13,  -20,  -37,  -45,  -39,   10,   23,   54,   39,   24,   51,   14,   -2,    3},
      { -32,  -41,   29,    6,   -9,  -26,  -45,   39,  -26,  -24,   -6,    3,  -79,  -17,   20,  -13},
      { -30,  -43,   40,   30,   23,   13,   -3,   13,  -26,  -51,  -31,  -23,   70,   13,  -17,   10},
      { -27,  -41,   43,   40,   40,   41,   40,  -30,   30,   47,   25,   15,  -28,   -5,    7,   -4},
      },
};


static int KLTCol16inProd[4][16], KLTRow16inProd[4][16];
/*! 
*************************************************************************************
* \brief
*   Compute inner product for the input 16x16 matrix 
*
* \para compute_inner_product16x16()
*    <paragraph>
*
* \author
*    - Yan Ye                      <yye@qualcomm.com>
*************************************************************************************
*/
void compute_inner_product16x16(int matrix[16][16], int isLeft, int *inner_prod)
{
  int i, j;

  if(isLeft)
    for(i = 0; i < 16; i++)
    {
      inner_prod[i] = 0;
      for(j = 0; j < 8; j++)
      {
        inner_prod[i] += matrix[i][2*j+1]*matrix[i][2*j];
      }
    }
  else
    for(i = 0; i < 16; i++)
    {
      inner_prod[i] = 0;
      for(j = 0; j < 8; j++)
      {
        inner_prod[i] += matrix[2*j+1][i]*matrix[2*j][i];
      }
    }
}

/*! 
*************************************************************************************
* \brief
*   Precompute inner products for all 16x16 transform matrices
*
* \para precompute_all_inner_product16x16()
*    <paragraph>
*
* \author
*    - Yan Ye                      <yye@qualcomm.com>
*************************************************************************************
*/
void precompute_all_inner_product16x16()
{
  int i, j, k;
  int temp2D[16][16];

  for(i = 0; i < 4; i++)
  {
    for(j = 0; j < 16; j++)
      for(k = 0; k < 16; k++)
        temp2D[j][k] = KLTCol16[i][j][k];
    compute_inner_product16x16(temp2D, 1, KLTCol16inProd[i]);

    for(j = 0; j < 16; j++)
      for(k = 0; k < 16; k++)
        temp2D[j][k] = KLTRow16[i][j][k];
    compute_inner_product16x16(temp2D, 0, KLTRow16inProd[i]);
  }
}

/*! 
*************************************************************************************
* \brief
*   Perform fast mode-dependent 16x16 directional transform 
*
* \para klt_luma16x16_sep_fast()
*    <paragraph>
*
* \author
*    - Yan Ye                      <yye@qualcomm.com>
*************************************************************************************
*/

int klt_luma16x16_sep_fast(int ipmode)
{
  int sign(int a,int b);

  int i,j,k,coeff_ctr;
  int scan_pos,run;
  int nonzero;
  int qp_per,qp_rem, q_bits;

  int*  ACLevel = img->cofAC16[0];
  int*  ACRun   = img->cofAC16[1];
  byte (*pos_scan)[2] = img->scanOrder16x16[ipmode];

  int coeff [MB_BLOCK_SIZE][MB_BLOCK_SIZE], ilev[MB_BLOCK_SIZE][MB_BLOCK_SIZE], 
    temp2D[MB_BLOCK_SIZE][MB_BLOCK_SIZE], y2D [MB_BLOCK_SIZE][MB_BLOCK_SIZE];
  int M1[MB_BLOCK_SIZE][MB_BLOCK_SIZE];
  int inprod[MB_BLOCK_SIZE];

  int m=4, n=10, KLTprec=14;

  int64 qpOffset;
  int64 roundOffset;
  int64 level;

  int R[6] = {40, 45, 50, 57, 63, 71};
  int Q[6] = {410,   364,   328,   287,   260,   231};
  long shift;
  int shiftBy2;

#ifdef RDO_Q
  levelDataStruct levelData[256];
  double lambda_md = 0.0;
  double err;
  int lowerInt;
  int levelRDOQ[256];
  int kStart = 0;
  int kStop = 0;
  int noCoeff;

  int levelscale[16][16];
  int invlevelscale[16][16];
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
  double normFact = pow(2, 2 * (KLTprec + m + n + 8) - 15)*(1<<(2*img->BitDepthIncrease));
#else
  double normFact = pow(2, 2 * (KLTprec + m + n + 8) - 15);
#endif
#endif

  pos_scan = img->scanOrder16x16[ipmode];

  qp_per    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)/6; 
  qp_rem    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)%6; 
  q_bits    = Q_BITS + qp_per;

  qpOffset=n+qp_per+KLTprec+2;
  roundOffset=((int64)1<<(qpOffset))/3;

  for(j=0;j<16;j++)
    for(i=0;i<16;i++)
    {
      levelscale[i][j]    = Q[qp_rem]<<4;
      invlevelscale[i][j] = R[qp_rem]<<4;
    }

#ifdef RDO_Q
    if(input->UseRDO_Q)
    {
      if ((img->type == B_SLICE) && img->nal_reference_idc)
      {
        lambda_md = img->lambda_md[5][img->masterQP];  
      }
      else
      {
        lambda_md = img->lambda_md[img->type][img->masterQP];
      }
    }
#endif

    for (j=0;j<16;j++)
    {
      for (i=0;i<16;i++)
      {
        M1[j][i]=imgY_org[img->opix_y+j][img->opix_x+i]-img->mprr_2[ipmode][j][i];
      }
    }

    compute_inner_product16x16(M1, 0, inprod);

    // column transform first
    for (i=0;i<(MB_BLOCK_SIZE);i++)
    {
      for (j=0;j<(MB_BLOCK_SIZE);j++)
      {
        temp2D[i][j] = 0;
        for(k = 0; k < MB_BLOCK_SIZE/2; k++)
          temp2D[i][j]+=
          (KLTCol16[ipmode][i][2*k]+M1[2*k+1][j])*(KLTCol16[ipmode][i][2*k+1]+M1[2*k][j]);
        temp2D[i][j] = temp2D[i][j] - KLTCol16inProd[ipmode][i] - inprod[j];
      }
    }

    compute_inner_product16x16(temp2D, 1, inprod);

    // row transform second
    for (i=0;i<(MB_BLOCK_SIZE);i++)
    {
      for (j=0;j<(MB_BLOCK_SIZE);j++)
      {
        coeff[i][j] = 0;
        for(k = 0; k < MB_BLOCK_SIZE/2; k++)
          coeff[i][j]+=
          (temp2D[i][2*k+1]+KLTRow16[ipmode][2*k][j])*(temp2D[i][2*k]+KLTRow16[ipmode][2*k+1][j]);
        coeff[i][j] = coeff[i][j] - inprod[i] - KLTRow16inProd[ipmode][j];
      }
    }

#ifdef RDO_Q
    if(input->UseRDO_Q)
    {
      noCoeff = 0;
      for (coeff_ctr = 0; coeff_ctr < 256; coeff_ctr++)
      {
        i = pos_scan[coeff_ctr][0];
        j = pos_scan[coeff_ctr][1];

        levelData[coeff_ctr].levelDouble = (int64)absm(coeff[j][i]) * (int64)(levelscale[i][j]);
        level = (int)(levelData[coeff_ctr].levelDouble >> qpOffset);
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
        lowerInt = ((levelData[coeff_ctr].levelDouble - ((int64)level << qpOffset)) < ((int64) 1 << (qpOffset - 1)))? 1: 0;
#else
        lowerInt = ((levelData[coeff_ctr].levelDouble - (level << qpOffset)) < ((int64) 1 << (qpOffset - 1)))? 1: 0;
#endif

        levelData[coeff_ctr].level[0] = 0;
        if (level == 0 && lowerInt == 1)
        {
          levelData[coeff_ctr].noLevels = 1;
        }
        else if (level == 0 && lowerInt == 0)
        {
          levelData[coeff_ctr].level[1] = level + 1;
          levelData[coeff_ctr].noLevels = 2;
          kStop = coeff_ctr;
          noCoeff++;
        }
        else if (level > 0 && lowerInt == 1)
        {
          levelData[coeff_ctr].level[1] = level;
          levelData[coeff_ctr].noLevels = 2;
          kStop = coeff_ctr;
          noCoeff++;
        }
        else
        {
          levelData[coeff_ctr].level[1] = level;
          levelData[coeff_ctr].level[2] = level + 1;
          levelData[coeff_ctr].noLevels = 3;
          kStop = coeff_ctr;
          kStart = coeff_ctr;
          noCoeff++;
        }

        for (k = 0; k < levelData[coeff_ctr].noLevels; k++)
        {      
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
          err = (double)((int64)levelData[coeff_ctr].level[k] << qpOffset) - (double) levelData[coeff_ctr].levelDouble;
#else
          err = (double)(levelData[coeff_ctr].level[k] << qpOffset) - (double) levelData[coeff_ctr].levelDouble;
#endif
          err = err * (double)invlevelscale[i][j];
          levelData[coeff_ctr].errLevel[k] = (err * err) / normFact; 
        }
      }

      est_writeRunLevel_CABAC(levelData, levelRDOQ, LUMA_16x16, lambda_md, kStart, kStop, noCoeff, 0);
    }
#endif

    // quantization and zig-zag scanning 
    nonzero=FALSE;

    run=-1;
    scan_pos=0;

    for (coeff_ctr=0;coeff_ctr < 256;coeff_ctr++)
    {
      i=pos_scan[coeff_ctr][0];
      j=pos_scan[coeff_ctr][1];

      run++;

#ifdef RDO_Q
      if(input->UseRDO_Q)
      {
        level = levelRDOQ[coeff_ctr];
      }
      else
#endif
      {
        level = (absm((int64)coeff[j][i])*(int64)levelscale[i][j]+roundOffset);
        level = level >> (qpOffset);
      }

      if (level != 0)
      {
        nonzero=TRUE;

        ACLevel[scan_pos] = sign((int)level, coeff[j][i]);
        ACRun  [scan_pos] = run;
        ++scan_pos;
        run=-1;                     // reset zero level counter

        ilev[j][i] = sign((int)level, coeff[j][i]) * ((invlevelscale[i][j] << qp_per) >> 4); 
      }
      else 
        ilev[j][i]=0;
    }
    ACLevel[scan_pos] = 0;

    // inverse transform, column first
    for (i=0;i<(MB_BLOCK_SIZE);i++)
    {
      for (j=0;j<(MB_BLOCK_SIZE);j++)
      {
        temp2D[i][j] = 0;
        for(k = 0; k < MB_BLOCK_SIZE; k++)
          temp2D[i][j]+=KLTCol16[ipmode][k][i]*(ilev[k][j]);
      }
    }

    // inverse transform, row second 
    for (i=0;i<(MB_BLOCK_SIZE);i++)
    {
      for (j=0;j<(MB_BLOCK_SIZE);j++)
      {
        y2D[i][j] = 0;
        for(k = 0; k < MB_BLOCK_SIZE; k++)
          y2D[i][j]+=temp2D[i][k]*KLTRow16[ipmode][j][k];
      }
    }

    // store the encoded image
    shift=KLTprec+m+2; shiftBy2=1<<(shift-1);
    for (j=0; j < MB_BLOCK_SIZE; j++)
      for (i=0; i < MB_BLOCK_SIZE; i++)
      {
        short temp;

#ifdef  INTERNAL_BIT_DEPTH_INCREASE
        temp=(short)((y2D[j][i]+((int64)img->mprr_2[ipmode][j][i]<<shift)+shiftBy2)>>shift);
#else
        temp=(short)((y2D[j][i]+((int)img->mprr_2[ipmode][j][i]<<shift)+shiftBy2)>>shift);
#endif
        temp=min(img->max_imgpel_value,max(0,temp));
        enc_picture->imgY[img->pix_y+j][img->pix_x+i]=temp;
      }

      return nonzero;
}


#endif

#define M_PI       3.14159265358979323846

// The butterfly parameters used to create these matrices for 16x16 transform
static const int DQ_BITS_16 = 12;
static const int Q_BITS_16  = 15;

static const int DCT16_2D_fw_UP_SHIFT = 8;
static const int DCT16_2D_bw_UP_SHIFT = 0;

// Orthonormal

// 4x4 transform
// static const int A4_II=5, B4_II=12;
// static const int shift4_II=3;
static const int A4_II=2, B4_II=5;
static const int shift4_II=2;

// static double norm4_II = sqrt((double) (A4_II*A4_II + B4_II*B4_II)) / (1 << shift4_II);

// Even part of 8x8 transform
static const int A8_II = 2;
static const int B8_II = 5;

// static const double norm8_II_AB = norm4_II;

// Odd part of 8x8 transform
static const int G8_II=19, D8_II=4, E8_II=16, Z8_II=11;
static const int shift8_II=5;

// static const double norm8_II_GDEZ = sqrt((double) (G8_II*G8_II + D8_II*D8_II)) / (1 << shift8_II);

// Givens rotation matrix butterfly
static const int B8_IV=6, D8_IV=11, F8_IV=21, H8_IV=27, G8_IV=34, E8_IV=38, C8_IV=42, A8_IV=43;
static const int shift8_IV=6;

// static const double norm8_IV = sqrt((double) (A8_IV*A8_IV + B8_IV*B8_IV)) / (1 << shift8_IV);

static const int quant_coef_16x16[6][16][16] =
{ 
  {
    {3277, 6822, 5394, 5069, 3444, 5069, 7636, 4827, 3277, 4827, 7636, 5069, 3444, 5069, 5394, 6822},
    {6822, 14248, 11252, 10588, 7183, 10588, 15930, 10063, 6822, 10063, 15930, 10588, 7183, 10588, 11252, 14248},
    {5394, 11252, 8892, 8369, 5673, 8369, 12579, 7955, 5394, 7955, 12579, 8369, 5673, 8369, 8892, 11252},
    {5069, 10588, 8369, 7849, 5335, 7849, 11835, 7484, 5069, 7484, 11835, 7849, 5335, 7849, 8369, 10588},
    {3444, 7183, 5673, 5335, 3607, 5335, 8017, 5069, 3444, 5069, 8017, 5335, 3607, 5335, 5673, 7183},
    {5069, 10588, 8369, 7849, 5335, 7849, 11835, 7484, 5069, 7484, 11835, 7849, 5335, 7849, 8369, 10588},
    {7636, 15930, 12579, 11835, 8017, 11835, 17804, 11252, 7636, 11252, 17804, 11835, 8017, 11835, 12579, 15930},
    {4827, 10063, 7955, 7484, 5069, 7484, 11252, 7114, 4827, 7114, 11252, 7484, 5069, 7484, 7955, 10063},
    {3277, 6822, 5394, 5069, 3444, 5069, 7636, 4827, 3277, 4827, 7636, 5069, 3444, 5069, 5394, 6822},
    {4827, 10063, 7955, 7484, 5069, 7484, 11252, 7114, 4827, 7114, 11252, 7484, 5069, 7484, 7955, 10063},
    {7636, 15930, 12579, 11835, 8017, 11835, 17804, 11252, 7636, 11252, 17804, 11835, 8017, 11835, 12579, 15930},
    {5069, 10588, 8369, 7849, 5335, 7849, 11835, 7484, 5069, 7484, 11835, 7849, 5335, 7849, 8369, 10588},
    {3444, 7183, 5673, 5335, 3607, 5335, 8017, 5069, 3444, 5069, 8017, 5335, 3607, 5335, 5673, 7183},
    {5069, 10588, 8369, 7849, 5335, 7849, 11835, 7484, 5069, 7484, 11835, 7849, 5335, 7849, 8369, 10588},
    {5394, 11252, 8892, 8369, 5673, 8369, 12579, 7955, 5394, 7955, 12579, 8369, 5673, 8369, 8892, 11252},
    {6822, 14248, 11252, 10588, 7183, 10588, 15930, 10063, 6822, 10063, 15930, 10588, 7183, 10588, 11252, 14248},
  },
  {
    {2913, 6092, 4811, 4522, 3061, 4522, 6797, 4299, 2913, 4299, 6797, 4522, 3061, 4522, 4811, 6092},
    {6092, 12695, 10030, 9419, 6397, 9419, 14178, 8969, 6092, 8969, 14178, 9419, 6397, 9419, 10030, 12695},
    {4811, 10030, 7926, 7455, 5053, 7455, 11212, 7097, 4811, 7097, 11212, 7455, 5053, 7455, 7926, 10030},
    {4522, 9419, 7455, 6993, 4751, 6993, 10539, 6662, 4522, 6662, 10539, 6993, 4751, 6993, 7455, 9419},
    {3061, 6397, 5053, 4751, 3224, 4751, 7143, 4522, 3061, 4522, 7143, 4751, 3224, 4751, 5053, 6397},
    {4522, 9419, 7455, 6993, 4751, 6993, 10539, 6662, 4522, 6662, 10539, 6993, 4751, 6993, 7455, 9419},
    {6797, 14178, 11212, 10539, 7143, 10539, 15852, 10030, 6797, 10030, 15852, 10539, 7143, 10539, 11212, 14178},
    {4299, 8969, 7097, 6662, 4522, 6662, 10030, 6347, 4299, 6347, 10030, 6662, 4522, 6662, 7097, 8969},
    {2913, 6092, 4811, 4522, 3061, 4522, 6797, 4299, 2913, 4299, 6797, 4522, 3061, 4522, 4811, 6092},
    {4299, 8969, 7097, 6662, 4522, 6662, 10030, 6347, 4299, 6347, 10030, 6662, 4522, 6662, 7097, 8969},
    {6797, 14178, 11212, 10539, 7143, 10539, 15852, 10030, 6797, 10030, 15852, 10539, 7143, 10539, 11212, 14178},
    {4522, 9419, 7455, 6993, 4751, 6993, 10539, 6662, 4522, 6662, 10539, 6993, 4751, 6993, 7455, 9419},
    {3061, 6397, 5053, 4751, 3224, 4751, 7143, 4522, 3061, 4522, 7143, 4751, 3224, 4751, 5053, 6397},
    {4522, 9419, 7455, 6993, 4751, 6993, 10539, 6662, 4522, 6662, 10539, 6993, 4751, 6993, 7455, 9419},
    {4811, 10030, 7926, 7455, 5053, 7455, 11212, 7097, 4811, 7097, 11212, 7455, 5053, 7455, 7926, 10030},
    {6092, 12695, 10030, 9419, 6397, 9419, 14178, 8969, 6092, 8969, 14178, 9419, 6397, 9419, 10030, 12695},
    },
    {
      {2595, 5425, 4289, 4029, 2729, 4029, 6060, 3836, 2595, 3836, 6060, 4029, 2729, 4029, 4289, 5425},
      {5425, 11304, 8930, 8392, 5701, 8392, 12643, 7998, 5425, 7998, 12643, 8392, 5701, 8392, 8930, 11304},
      {4289, 8930, 7058, 6643, 4503, 6643, 9995, 6315, 4289, 6315, 9995, 6643, 4503, 6643, 7058, 8930},
      {4029, 8392, 6643, 6241, 4229, 6241, 9381, 5938, 4029, 5938, 9381, 6241, 4229, 6241, 6643, 8392},
      {2729, 5701, 4503, 4229, 2876, 4229, 6362, 4029, 2729, 4029, 6362, 4229, 2876, 4229, 4503, 5701},
      {4029, 8392, 6643, 6241, 4229, 6241, 9381, 5938, 4029, 5938, 9381, 6241, 4229, 6241, 6643, 8392},
      {6060, 12643, 9995, 9381, 6362, 9381, 14130, 8930, 6060, 8930, 14130, 9381, 6362, 9381, 9995, 12643},
      {3836, 7998, 6315, 5938, 4029, 5938, 8930, 5652, 3836, 5652, 8930, 5938, 4029, 5938, 6315, 7998},
      {2595, 5425, 4289, 4029, 2729, 4029, 6060, 3836, 2595, 3836, 6060, 4029, 2729, 4029, 4289, 5425},
      {3836, 7998, 6315, 5938, 4029, 5938, 8930, 5652, 3836, 5652, 8930, 5938, 4029, 5938, 6315, 7998},
      {6060, 12643, 9995, 9381, 6362, 9381, 14130, 8930, 6060, 8930, 14130, 9381, 6362, 9381, 9995, 12643},
      {4029, 8392, 6643, 6241, 4229, 6241, 9381, 5938, 4029, 5938, 9381, 6241, 4229, 6241, 6643, 8392},
      {2729, 5701, 4503, 4229, 2876, 4229, 6362, 4029, 2729, 4029, 6362, 4229, 2876, 4229, 4503, 5701},
      {4029, 8392, 6643, 6241, 4229, 6241, 9381, 5938, 4029, 5938, 9381, 6241, 4229, 6241, 6643, 8392},
      {4289, 8930, 7058, 6643, 4503, 6643, 9995, 6315, 4289, 6315, 9995, 6643, 4503, 6643, 7058, 8930},
      {5425, 11304, 8930, 8392, 5701, 8392, 12643, 7998, 5425, 7998, 12643, 8392, 5701, 8392, 8930, 11304},
    },
    {
      {2320, 4827, 3818, 3592, 2431, 3592, 5404, 3411, 2320, 3411, 5404, 3592, 2431, 3592, 3818, 4827},
      {4827, 10073, 7965, 7484, 5069, 7484, 11263, 7124, 4827, 7124, 11263, 7484, 5069, 7484, 7965, 10073},
      {3818, 7965, 6289, 5918, 4009, 5918, 8902, 5626, 3818, 5626, 8902, 5918, 4009, 5918, 6289, 7965},
      {3592, 7484, 5918, 5551, 3769, 5551, 8359, 5294, 3592, 5294, 8359, 5551, 3769, 5551, 5918, 7484},
      {2431, 5069, 4009, 3769, 2553, 3769, 5673, 3592, 2431, 3592, 5673, 3769, 2553, 3769, 4009, 5069},
      {3592, 7484, 5918, 5551, 3769, 5551, 8359, 5294, 3592, 5294, 8359, 5551, 3769, 5551, 5918, 7484},
      {5404, 11263, 8902, 8359, 5673, 8359, 12589, 7965, 5404, 7965, 12589, 8359, 5673, 8359, 8902, 11263},
      {3411, 7124, 5626, 5294, 3592, 5294, 7965, 5032, 3411, 5032, 7965, 5294, 3592, 5294, 5626, 7124},
      {2320, 4827, 3818, 3592, 2431, 3592, 5404, 3411, 2320, 3411, 5404, 3592, 2431, 3592, 3818, 4827},
      {3411, 7124, 5626, 5294, 3592, 5294, 7965, 5032, 3411, 5032, 7965, 5294, 3592, 5294, 5626, 7124},
      {5404, 11263, 8902, 8359, 5673, 8359, 12589, 7965, 5404, 7965, 12589, 8359, 5673, 8359, 8902, 11263},
      {3592, 7484, 5918, 5551, 3769, 5551, 8359, 5294, 3592, 5294, 8359, 5551, 3769, 5551, 5918, 7484},
      {2431, 5069, 4009, 3769, 2553, 3769, 5673, 3592, 2431, 3592, 5673, 3769, 2553, 3769, 4009, 5069},
      {3592, 7484, 5918, 5551, 3769, 5551, 8359, 5294, 3592, 5294, 8359, 5551, 3769, 5551, 5918, 7484},
      {3818, 7965, 6289, 5918, 4009, 5918, 8902, 5626, 3818, 5626, 8902, 5918, 4009, 5918, 6289, 7965},
      {4827, 10073, 7965, 7484, 5069, 7484, 11263, 7124, 4827, 7124, 11263, 7484, 5069, 7484, 7965, 10073},
      },
      {
        {2064, 4307, 3399, 3199, 2167, 3199, 4811, 3046, 2064, 3046, 4811, 3199, 2167, 3199, 3399, 4307},
        {4307, 8969, 7089, 6662, 4522, 6662, 10030, 6347, 4307, 6347, 10030, 6662, 4522, 6662, 7089, 8969},
        {3399, 7089, 5606, 5269, 3571, 5269, 7926, 5015, 3399, 5015, 7926, 5269, 3571, 5269, 5606, 7089},
        {3199, 6662, 5269, 4949, 3359, 4949, 7447, 4710, 3199, 4710, 7447, 4949, 3359, 4949, 5269, 6662},
        {2167, 4522, 3571, 3359, 2280, 3359, 5053, 3199, 2167, 3199, 5053, 3359, 2280, 3359, 3571, 4522},
        {3199, 6662, 5269, 4949, 3359, 4949, 7447, 4710, 3199, 4710, 7447, 4949, 3359, 4949, 5269, 6662},
        {4811, 10030, 7926, 7447, 5053, 7447, 11212, 7089, 4811, 7089, 11212, 7447, 5053, 7447, 7926, 10030},
        {3046, 6347, 5015, 4710, 3199, 4710, 7089, 4485, 3046, 4485, 7089, 4710, 3199, 4710, 5015, 6347},
        {2064, 4307, 3399, 3199, 2167, 3199, 4811, 3046, 2064, 3046, 4811, 3199, 2167, 3199, 3399, 4307},
        {3046, 6347, 5015, 4710, 3199, 4710, 7089, 4485, 3046, 4485, 7089, 4710, 3199, 4710, 5015, 6347},
        {4811, 10030, 7926, 7447, 5053, 7447, 11212, 7089, 4811, 7089, 11212, 7447, 5053, 7447, 7926, 10030},
        {3199, 6662, 5269, 4949, 3359, 4949, 7447, 4710, 3199, 4710, 7447, 4949, 3359, 4949, 5269, 6662},
        {2167, 4522, 3571, 3359, 2280, 3359, 5053, 3199, 2167, 3199, 5053, 3359, 2280, 3359, 3571, 4522},
        {3199, 6662, 5269, 4949, 3359, 4949, 7447, 4710, 3199, 4710, 7447, 4949, 3359, 4949, 5269, 6662},
        {3399, 7089, 5606, 5269, 3571, 5269, 7926, 5015, 3399, 5015, 7926, 5269, 3571, 5269, 5606, 7089},
        {4307, 8969, 7089, 6662, 4522, 6662, 10030, 6347, 4307, 6347, 10030, 6662, 4522, 6662, 7089, 8969},
      },
      {
        {1840, 3836, 3030, 2851, 1935, 2851, 4289, 2712, 1840, 2712, 4289, 2851, 1935, 2851, 3030, 3836},
        {3836, 7992, 6322, 5938, 4029, 5938, 8937, 5652, 3836, 5652, 8937, 5938, 4029, 5938, 6322, 7992},
        {3030, 6322, 4997, 4690, 3181, 4690, 7065, 4465, 3030, 4465, 7065, 4690, 3181, 4690, 4997, 6322},
        {2851, 5938, 4690, 4407, 2990, 4407, 6637, 4196, 2851, 4196, 6637, 4407, 2990, 4407, 4690, 5938},
        {1935, 4029, 3181, 2990, 2027, 2990, 4503, 2851, 1935, 2851, 4503, 2990, 2027, 2990, 3181, 4029},
        {2851, 5938, 4690, 4407, 2990, 4407, 6637, 4196, 2851, 4196, 6637, 4407, 2990, 4407, 4690, 5938},
        {4289, 8937, 7065, 6637, 4503, 6637, 9988, 6322, 4289, 6322, 9988, 6637, 4503, 6637, 7065, 8937},
        {2712, 5652, 4465, 4196, 2851, 4196, 6322, 3999, 2712, 3999, 6322, 4196, 2851, 4196, 4465, 5652},
        {1840, 3836, 3030, 2851, 1935, 2851, 4289, 2712, 1840, 2712, 4289, 2851, 1935, 2851, 3030, 3836},
        {2712, 5652, 4465, 4196, 2851, 4196, 6322, 3999, 2712, 3999, 6322, 4196, 2851, 4196, 4465, 5652},
        {4289, 8937, 7065, 6637, 4503, 6637, 9988, 6322, 4289, 6322, 9988, 6637, 4503, 6637, 7065, 8937},
        {2851, 5938, 4690, 4407, 2990, 4407, 6637, 4196, 2851, 4196, 6637, 4407, 2990, 4407, 4690, 5938},
        {1935, 4029, 3181, 2990, 2027, 2990, 4503, 2851, 1935, 2851, 4503, 2990, 2027, 2990, 3181, 4029},
        {2851, 5938, 4690, 4407, 2990, 4407, 6637, 4196, 2851, 4196, 6637, 4407, 2990, 4407, 4690, 5938},
        {3030, 6322, 4997, 4690, 3181, 4690, 7065, 4465, 3030, 4465, 7065, 4690, 3181, 4690, 4997, 6322},
        {3836, 7992, 6322, 5938, 4029, 5938, 8937, 5652, 3836, 5652, 8937, 5938, 4029, 5938, 6322, 7992},
        }
};

static const int dequant_coef_16x16[6][16][16] =
{
  {
    {160, 334, 264, 248, 168, 248, 373, 236, 160, 236, 373, 248, 168, 248, 264, 334},
    {334, 695, 550, 516, 350, 516, 777, 492, 334, 492, 777, 516, 350, 516, 550, 695},
    {264, 550, 435, 408, 277, 408, 615, 389, 264, 389, 615, 408, 277, 408, 435, 550},
    {248, 516, 408, 384, 260, 384, 577, 365, 248, 365, 577, 384, 260, 384, 408, 516},
    {168, 350, 277, 260, 177, 260, 392, 248, 168, 248, 392, 260, 177, 260, 277, 350},
    {248, 516, 408, 384, 260, 384, 577, 365, 248, 365, 577, 384, 260, 384, 408, 516},
    {373, 777, 615, 577, 392, 577, 869, 550, 373, 550, 869, 577, 392, 577, 615, 777},
    {236, 492, 389, 365, 248, 365, 550, 348, 236, 348, 550, 365, 248, 365, 389, 492},
    {160, 334, 264, 248, 168, 248, 373, 236, 160, 236, 373, 248, 168, 248, 264, 334},
    {236, 492, 389, 365, 248, 365, 550, 348, 236, 348, 550, 365, 248, 365, 389, 492},
    {373, 777, 615, 577, 392, 577, 869, 550, 373, 550, 869, 577, 392, 577, 615, 777},
    {248, 516, 408, 384, 260, 384, 577, 365, 248, 365, 577, 384, 260, 384, 408, 516},
    {168, 350, 277, 260, 177, 260, 392, 248, 168, 248, 392, 260, 177, 260, 277, 350},
    {248, 516, 408, 384, 260, 384, 577, 365, 248, 365, 577, 384, 260, 384, 408, 516},
    {264, 550, 435, 408, 277, 408, 615, 389, 264, 389, 615, 408, 277, 408, 435, 550},
    {334, 695, 550, 516, 350, 516, 777, 492, 334, 492, 777, 516, 350, 516, 550, 695},
  },
  {
    {180, 374, 296, 278, 189, 278, 419, 265, 180, 265, 419, 278, 189, 278, 296, 374},
    {374, 780, 617, 580, 393, 580, 873, 552, 374, 552, 873, 580, 393, 580, 617, 780},
    {296, 617, 488, 458, 311, 458, 690, 436, 296, 436, 690, 458, 311, 458, 488, 617},
    {278, 580, 458, 431, 292, 431, 648, 410, 278, 410, 648, 431, 292, 431, 458, 580},
    {189, 393, 311, 292, 198, 292, 440, 278, 189, 278, 440, 292, 198, 292, 311, 393},
    {278, 580, 458, 431, 292, 431, 648, 410, 278, 410, 648, 431, 292, 431, 458, 580},
    {419, 873, 690, 648, 440, 648, 976, 617, 419, 617, 976, 648, 440, 648, 690, 873},
    {265, 552, 436, 410, 278, 410, 617, 390, 265, 390, 617, 410, 278, 410, 436, 552},
    {180, 374, 296, 278, 189, 278, 419, 265, 180, 265, 419, 278, 189, 278, 296, 374},
    {265, 552, 436, 410, 278, 410, 617, 390, 265, 390, 617, 410, 278, 410, 436, 552},
    {419, 873, 690, 648, 440, 648, 976, 617, 419, 617, 976, 648, 440, 648, 690, 873},
    {278, 580, 458, 431, 292, 431, 648, 410, 278, 410, 648, 431, 292, 431, 458, 580},
    {189, 393, 311, 292, 198, 292, 440, 278, 189, 278, 440, 292, 198, 292, 311, 393},
    {278, 580, 458, 431, 292, 431, 648, 410, 278, 410, 648, 431, 292, 431, 458, 580},
    {296, 617, 488, 458, 311, 458, 690, 436, 296, 436, 690, 458, 311, 458, 488, 617},
    {374, 780, 617, 580, 393, 580, 873, 552, 374, 552, 873, 580, 393, 580, 617, 780},
    },
    {
      {202, 420, 332, 312, 212, 312, 470, 297, 202, 297, 470, 312, 212, 312, 332, 420},
      {420, 876, 693, 651, 441, 651, 979, 619, 420, 619, 979, 651, 441, 651, 693, 876},
      {332, 693, 548, 514, 349, 514, 774, 490, 332, 490, 774, 514, 349, 514, 548, 693},
      {312, 651, 514, 483, 328, 483, 728, 460, 312, 460, 728, 483, 328, 483, 514, 651},
      {212, 441, 349, 328, 222, 328, 494, 312, 212, 312, 494, 328, 222, 328, 349, 441},
      {312, 651, 514, 483, 328, 483, 728, 460, 312, 460, 728, 483, 328, 483, 514, 651},
      {470, 979, 774, 728, 494, 728, 1095, 693, 470, 693, 1095, 728, 494, 728, 774, 979},
      {297, 619, 490, 460, 312, 460, 693, 438, 297, 438, 693, 460, 312, 460, 490, 619},
      {202, 420, 332, 312, 212, 312, 470, 297, 202, 297, 470, 312, 212, 312, 332, 420},
      {297, 619, 490, 460, 312, 460, 693, 438, 297, 438, 693, 460, 312, 460, 490, 619},
      {470, 979, 774, 728, 494, 728, 1095, 693, 470, 693, 1095, 728, 494, 728, 774, 979},
      {312, 651, 514, 483, 328, 483, 728, 460, 312, 460, 728, 483, 328, 483, 514, 651},
      {212, 441, 349, 328, 222, 328, 494, 312, 212, 312, 494, 328, 222, 328, 349, 441},
      {312, 651, 514, 483, 328, 483, 728, 460, 312, 460, 728, 483, 328, 483, 514, 651},
      {332, 693, 548, 514, 349, 514, 774, 490, 332, 490, 774, 514, 349, 514, 548, 693},
      {420, 876, 693, 651, 441, 651, 979, 619, 420, 619, 979, 651, 441, 651, 693, 876},
    },
    {
      {226, 472, 373, 350, 238, 350, 527, 334, 226, 334, 527, 350, 238, 350, 373, 472},
      {472, 983, 777, 730, 496, 730, 1099, 695, 472, 695, 1099, 730, 496, 730, 777, 983},
      {373, 777, 615, 577, 392, 577, 869, 550, 373, 550, 869, 577, 392, 577, 615, 777},
      {350, 730, 577, 543, 368, 543, 817, 516, 350, 516, 817, 543, 368, 543, 577, 730},
      {238, 496, 392, 368, 250, 368, 554, 350, 238, 350, 554, 368, 250, 368, 392, 496},
      {350, 730, 577, 543, 368, 543, 817, 516, 350, 516, 817, 543, 368, 543, 577, 730},
      {527, 1099, 869, 817, 554, 817, 1229, 777, 527, 777, 1229, 817, 554, 817, 869, 1099},
      {334, 695, 550, 516, 350, 516, 777, 492, 334, 492, 777, 516, 350, 516, 550, 695},
      {226, 472, 373, 350, 238, 350, 527, 334, 226, 334, 527, 350, 238, 350, 373, 472},
      {334, 695, 550, 516, 350, 516, 777, 492, 334, 492, 777, 516, 350, 516, 550, 695},
      {527, 1099, 869, 817, 554, 817, 1229, 777, 527, 777, 1229, 817, 554, 817, 869, 1099},
      {350, 730, 577, 543, 368, 543, 817, 516, 350, 516, 817, 543, 368, 543, 577, 730},
      {238, 496, 392, 368, 250, 368, 554, 350, 238, 350, 554, 368, 250, 368, 392, 496},
      {350, 730, 577, 543, 368, 543, 817, 516, 350, 516, 817, 543, 368, 543, 577, 730},
      {373, 777, 615, 577, 392, 577, 869, 550, 373, 550, 869, 577, 392, 577, 615, 777},
      {472, 983, 777, 730, 496, 730, 1099, 695, 472, 695, 1099, 730, 496, 730, 777, 983},
      },
      {
        {254, 529, 419, 393, 267, 393, 592, 374, 254, 374, 592, 393, 267, 393, 419, 529},
        {529, 1104, 873, 820, 556, 820, 1234, 780, 529, 780, 1234, 820, 556, 820, 873, 1104},
        {419, 873, 690, 648, 440, 648, 976, 617, 419, 617, 976, 648, 440, 648, 690, 873},
        {393, 820, 648, 609, 413, 609, 917, 580, 393, 580, 917, 609, 413, 609, 648, 820},
        {267, 556, 440, 413, 280, 413, 622, 393, 267, 393, 622, 413, 280, 413, 440, 556},
        {393, 820, 648, 609, 413, 609, 917, 580, 393, 580, 917, 609, 413, 609, 648, 820},
        {592, 1234, 976, 917, 622, 917, 1380, 873, 592, 873, 1380, 917, 622, 917, 976, 1234},
        {374, 780, 617, 580, 393, 580, 873, 552, 374, 552, 873, 580, 393, 580, 617, 780},
        {254, 529, 419, 393, 267, 393, 592, 374, 254, 374, 592, 393, 267, 393, 419, 529},
        {374, 780, 617, 580, 393, 580, 873, 552, 374, 552, 873, 580, 393, 580, 617, 780},
        {592, 1234, 976, 917, 622, 917, 1380, 873, 592, 873, 1380, 917, 622, 917, 976, 1234},
        {393, 820, 648, 609, 413, 609, 917, 580, 393, 580, 917, 609, 413, 609, 648, 820},
        {267, 556, 440, 413, 280, 413, 622, 393, 267, 393, 622, 413, 280, 413, 440, 556},
        {393, 820, 648, 609, 413, 609, 917, 580, 393, 580, 917, 609, 413, 609, 648, 820},
        {419, 873, 690, 648, 440, 648, 976, 617, 419, 617, 976, 648, 440, 648, 690, 873},
        {529, 1104, 873, 820, 556, 820, 1234, 780, 529, 780, 1234, 820, 556, 820, 873, 1104},
      },
      {
        {285, 594, 470, 441, 299, 441, 664, 420, 285, 420, 664, 441, 299, 441, 470, 594},
        {594, 1239, 979, 920, 624, 920, 1385, 876, 594, 876, 1385, 920, 624, 920, 979, 1239},
        {470, 979, 774, 728, 494, 728, 1095, 693, 470, 693, 1095, 728, 494, 728, 774, 979},
        {441, 920, 728, 684, 464, 684, 1029, 651, 441, 651, 1029, 684, 464, 684, 728, 920},
        {299, 624, 494, 464, 315, 464, 698, 441, 299, 441, 698, 464, 315, 464, 494, 624},
        {441, 920, 728, 684, 464, 684, 1029, 651, 441, 651, 1029, 684, 464, 684, 728, 920},
        {664, 1385, 1095, 1029, 698, 1029, 1549, 979, 664, 979, 1549, 1029, 698, 1029, 1095, 1385},
        {420, 876, 693, 651, 441, 651, 979, 619, 420, 619, 979, 651, 441, 651, 693, 876},
        {285, 594, 470, 441, 299, 441, 664, 420, 285, 420, 664, 441, 299, 441, 470, 594},
        {420, 876, 693, 651, 441, 651, 979, 619, 420, 619, 979, 651, 441, 651, 693, 876},
        {664, 1385, 1095, 1029, 698, 1029, 1549, 979, 664, 979, 1549, 1029, 698, 1029, 1095, 1385},
        {441, 920, 728, 684, 464, 684, 1029, 651, 441, 651, 1029, 684, 464, 684, 728, 920},
        {299, 624, 494, 464, 315, 464, 698, 441, 299, 441, 698, 464, 315, 464, 494, 624},
        {441, 920, 728, 684, 464, 684, 1029, 651, 441, 651, 1029, 684, 464, 684, 728, 920},
        {470, 979, 774, 728, 494, 728, 1095, 693, 470, 693, 1095, 728, 494, 728, 774, 979},
        {594, 1239, 979, 920, 624, 920, 1385, 876, 594, 876, 1385, 920, 624, 920, 979, 1239},
        }
};

static const int quant_coef_16x8[6][8][16] =
{
  {
    {4640, 9655, 7636, 7183, 4862, 7183, 10809, 6822, 4640, 6822, 10809, 7183, 4862, 7183, 7636, 9655},
    {7636, 15930, 12579, 11835, 8017, 11835, 17804, 11252, 7636, 11252, 17804, 11835, 8017, 11835, 12579, 15930},
    {4862, 10138, 8017, 7539, 5107, 7539, 11346, 7183, 4862, 7183, 11346, 7539, 5107, 7539, 8017, 10138},
    {10809, 22525, 17804, 16717, 11346, 16717, 25178, 15930, 10809, 15930, 25178, 16717, 11346, 16717, 17804, 22525},
    {4640, 9655, 7636, 7183, 4862, 7183, 10809, 6822, 4640, 6822, 10809, 7183, 4862, 7183, 7636, 9655},
    {10809, 22525, 17804, 16717, 11346, 16717, 25178, 15930, 10809, 15930, 25178, 16717, 11346, 16717, 17804, 22525},
    {4862, 10138, 8017, 7539, 5107, 7539, 11346, 7183, 4862, 7183, 11346, 7539, 5107, 7539, 8017, 10138},
    {7636, 15930, 12579, 11835, 8017, 11835, 17804, 11252, 7636, 11252, 17804, 11835, 8017, 11835, 12579, 15930},
  },
  {
    {4128, 8614, 6797, 6397, 4334, 6397, 9622, 6092, 4128, 6092, 9622, 6397, 4334, 6397, 6797, 8614},
    {6797, 14178, 11212, 10539, 7143, 10539, 15852, 10030, 6797, 10030, 15852, 10539, 7143, 10539, 11212, 14178},
    {4334, 9044, 7143, 6717, 4560, 6717, 10105, 6397, 4334, 6397, 10105, 6717, 4560, 6717, 7143, 9044},
    {9622, 20061, 15852, 14894, 10105, 14894, 22423, 14178, 9622, 14178, 22423, 14894, 10105, 14894, 15852, 20061},
    {4128, 8614, 6797, 6397, 4334, 6397, 9622, 6092, 4128, 6092, 9622, 6397, 4334, 6397, 6797, 8614},
    {9622, 20061, 15852, 14894, 10105, 14894, 22423, 14178, 9622, 14178, 22423, 14894, 10105, 14894, 15852, 20061},
    {4334, 9044, 7143, 6717, 4560, 6717, 10105, 6397, 4334, 6397, 10105, 6717, 4560, 6717, 7143, 9044},
    {6797, 14178, 11212, 10539, 7143, 10539, 15852, 10030, 6797, 10030, 15852, 10539, 7143, 10539, 11212, 14178},
    },
    {
      {3679, 7672, 6060, 5701, 3870, 5701, 8579, 5425, 3679, 5425, 8579, 5701, 3870, 5701, 6060, 7672},
      {6060, 12643, 9995, 9381, 6362, 9381, 14130, 8930, 6060, 8930, 14130, 9381, 6362, 9381, 9995, 12643},
      {3870, 8058, 6362, 5979, 4053, 5979, 9005, 5701, 3870, 5701, 9005, 5979, 4053, 5979, 6362, 8058},
      {8579, 17874, 14130, 13273, 9005, 13273, 19977, 12643, 8579, 12643, 19977, 13273, 9005, 13273, 14130, 17874},
      {3679, 7672, 6060, 5701, 3870, 5701, 8579, 5425, 3679, 5425, 8579, 5701, 3870, 5701, 6060, 7672},
      {8579, 17874, 14130, 13273, 9005, 13273, 19977, 12643, 8579, 12643, 19977, 13273, 9005, 13273, 14130, 17874},
      {3870, 8058, 6362, 5979, 4053, 5979, 9005, 5701, 3870, 5701, 9005, 5979, 4053, 5979, 6362, 8058},
      {6060, 12643, 9995, 9381, 6362, 9381, 14130, 8930, 6060, 8930, 14130, 9381, 6362, 9381, 9995, 12643},
    },
    {
      {3277, 6832, 5404, 5069, 3444, 5069, 7636, 4827, 3277, 4827, 7636, 5069, 3444, 5069, 5404, 6832},
      {5404, 11263, 8902, 8359, 5673, 8359, 12589, 7965, 5404, 7965, 12589, 8359, 5673, 8359, 8902, 11263},
      {3444, 7173, 5673, 5325, 3617, 5325, 8027, 5069, 3444, 5069, 8027, 5325, 3617, 5325, 5673, 7173},
      {7636, 15920, 12589, 11825, 8027, 11825, 17804, 11263, 7636, 11263, 17804, 11825, 8027, 11825, 12589, 15920},
      {3277, 6832, 5404, 5069, 3444, 5069, 7636, 4827, 3277, 4827, 7636, 5069, 3444, 5069, 5404, 6832},
      {7636, 15920, 12589, 11825, 8027, 11825, 17804, 11263, 7636, 11263, 17804, 11825, 8027, 11825, 12589, 15920},
      {3444, 7173, 5673, 5325, 3617, 5325, 8027, 5069, 3444, 5069, 8027, 5325, 3617, 5325, 5673, 7173},
      {5404, 11263, 8902, 8359, 5673, 8359, 12589, 7965, 5404, 7965, 12589, 8359, 5673, 8359, 8902, 11263},
      },
      {
        {2921, 6084, 4811, 4522, 3069, 4522, 6806, 4307, 2921, 4307, 6806, 4522, 3069, 4522, 4811, 6084},
        {4811, 10030, 7926, 7447, 5053, 7447, 11212, 7089, 4811, 7089, 11212, 7447, 5053, 7447, 7926, 10030},
        {3069, 6389, 5053, 4751, 3224, 4751, 7151, 4522, 3069, 4522, 7151, 4751, 3224, 4751, 5053, 6389},
        {6806, 14186, 11212, 10539, 7151, 10539, 15861, 10030, 6806, 10030, 15861, 10539, 7151, 10539, 11212, 14186},
        {2921, 6084, 4811, 4522, 3069, 4522, 6806, 4307, 2921, 4307, 6806, 4522, 3069, 4522, 4811, 6084},
        {6806, 14186, 11212, 10539, 7151, 10539, 15861, 10030, 6806, 10030, 15861, 10539, 7151, 10539, 11212, 14186},
        {3069, 6389, 5053, 4751, 3224, 4751, 7151, 4522, 3069, 4522, 7151, 4751, 3224, 4751, 5053, 6389},
        {4811, 10030, 7926, 7447, 5053, 7447, 11212, 7089, 4811, 7089, 11212, 7447, 5053, 7447, 7926, 10030},
      },
      {
        {2602, 5425, 4289, 4029, 2729, 4029, 6060, 3836, 2602, 3836, 6060, 4029, 2729, 4029, 4289, 5425},
        {4289, 8937, 7065, 6637, 4503, 6637, 9988, 6322, 4289, 6322, 9988, 6637, 4503, 6637, 7065, 8937},
        {2729, 5695, 4503, 4229, 2869, 4229, 6368, 4029, 2729, 4029, 6368, 4229, 2869, 4229, 4503, 5695},
        {6060, 12637, 9988, 9387, 6368, 9387, 14130, 8937, 6060, 8937, 14130, 9387, 6368, 9387, 9988, 12637},
        {2602, 5425, 4289, 4029, 2729, 4029, 6060, 3836, 2602, 3836, 6060, 4029, 2729, 4029, 4289, 5425},
        {6060, 12637, 9988, 9387, 6368, 9387, 14130, 8937, 6060, 8937, 14130, 9387, 6368, 9387, 9988, 12637},
        {2729, 5695, 4503, 4229, 2869, 4229, 6368, 4029, 2729, 4029, 6368, 4229, 2869, 4229, 4503, 5695},
        {4289, 8937, 7065, 6637, 4503, 6637, 9988, 6322, 4289, 6322, 9988, 6637, 4503, 6637, 7065, 8937},
        }
};

static const int dequant_coef_16x8[6][8][16] =
{
  {
    {226, 472, 373, 350, 238, 350, 527, 334, 226, 334, 527, 350, 238, 350, 373, 472},
    {373, 777, 615, 577, 392, 577, 869, 550, 373, 550, 869, 577, 392, 577, 615, 777},
    {238, 496, 392, 368, 250, 368, 554, 350, 238, 350, 554, 368, 250, 368, 392, 496},
    {527, 1099, 869, 817, 554, 817, 1229, 777, 527, 777, 1229, 817, 554, 817, 869, 1099},
    {226, 472, 373, 350, 238, 350, 527, 334, 226, 334, 527, 350, 238, 350, 373, 472},
    {527, 1099, 869, 817, 554, 817, 1229, 777, 527, 777, 1229, 817, 554, 817, 869, 1099},
    {238, 496, 392, 368, 250, 368, 554, 350, 238, 350, 554, 368, 250, 368, 392, 496},
    {373, 777, 615, 577, 392, 577, 869, 550, 373, 550, 869, 577, 392, 577, 615, 777},
  },
  {
    {254, 529, 419, 393, 267, 393, 592, 374, 254, 374, 592, 393, 267, 393, 419, 529},
    {419, 873, 690, 648, 440, 648, 976, 617, 419, 617, 976, 648, 440, 648, 690, 873},
    {267, 556, 440, 413, 280, 413, 622, 393, 267, 393, 622, 413, 280, 413, 440, 556},
    {592, 1234, 976, 917, 622, 917, 1380, 873, 592, 873, 1380, 917, 622, 917, 976, 1234},
    {254, 529, 419, 393, 267, 393, 592, 374, 254, 374, 592, 393, 267, 393, 419, 529},
    {592, 1234, 976, 917, 622, 917, 1380, 873, 592, 873, 1380, 917, 622, 917, 976, 1234},
    {267, 556, 440, 413, 280, 413, 622, 393, 267, 393, 622, 413, 280, 413, 440, 556},
    {419, 873, 690, 648, 440, 648, 976, 617, 419, 617, 976, 648, 440, 648, 690, 873},
    },
    {
      {285, 594, 470, 441, 299, 441, 664, 420, 285, 420, 664, 441, 299, 441, 470, 594},
      {470, 979, 774, 728, 494, 728, 1095, 693, 470, 693, 1095, 728, 494, 728, 774, 979},
      {299, 624, 494, 464, 315, 464, 698, 441, 299, 441, 698, 464, 315, 464, 494, 624},
      {664, 1385, 1095, 1029, 698, 1029, 1549, 979, 664, 979, 1549, 1029, 698, 1029, 1095, 1385},
      {285, 594, 470, 441, 299, 441, 664, 420, 285, 420, 664, 441, 299, 441, 470, 594},
      {664, 1385, 1095, 1029, 698, 1029, 1549, 979, 664, 979, 1549, 1029, 698, 1029, 1095, 1385},
      {299, 624, 494, 464, 315, 464, 698, 441, 299, 441, 698, 464, 315, 464, 494, 624},
      {470, 979, 774, 728, 494, 728, 1095, 693, 470, 693, 1095, 728, 494, 728, 774, 979},
    },
    {
      {320, 667, 527, 496, 336, 496, 746, 472, 320, 472, 746, 496, 336, 496, 527, 667},
      {527, 1099, 869, 817, 554, 817, 1229, 777, 527, 777, 1229, 817, 554, 817, 869, 1099},
      {336, 701, 554, 521, 353, 521, 783, 496, 336, 496, 783, 521, 353, 521, 554, 701},
      {746, 1555, 1229, 1155, 783, 1155, 1738, 1099, 746, 1099, 1738, 1155, 783, 1155, 1229, 1555},
      {320, 667, 527, 496, 336, 496, 746, 472, 320, 472, 746, 496, 336, 496, 527, 667},
      {746, 1555, 1229, 1155, 783, 1155, 1738, 1099, 746, 1099, 1738, 1155, 783, 1155, 1229, 1555},
      {336, 701, 554, 521, 353, 521, 783, 496, 336, 496, 783, 521, 353, 521, 554, 701},
      {527, 1099, 869, 817, 554, 817, 1229, 777, 527, 777, 1229, 817, 554, 817, 869, 1099},
      },
      {
        {359, 749, 592, 556, 377, 556, 837, 529, 359, 529, 837, 556, 377, 556, 592, 749},
        {592, 1234, 976, 917, 622, 917, 1380, 873, 592, 873, 1380, 917, 622, 917, 976, 1234},
        {377, 787, 622, 584, 396, 584, 879, 556, 377, 556, 879, 584, 396, 584, 622, 787},
        {837, 1745, 1380, 1296, 879, 1296, 1951, 1234, 837, 1234, 1951, 1296, 879, 1296, 1380, 1745},
        {359, 749, 592, 556, 377, 556, 837, 529, 359, 529, 837, 556, 377, 556, 592, 749},
        {837, 1745, 1380, 1296, 879, 1296, 1951, 1234, 837, 1234, 1951, 1296, 879, 1296, 1380, 1745},
        {377, 787, 622, 584, 396, 584, 879, 556, 377, 556, 879, 584, 396, 584, 622, 787},
        {592, 1234, 976, 917, 622, 917, 1380, 873, 592, 873, 1380, 917, 622, 917, 976, 1234},
      },
      {
        {403, 840, 664, 624, 424, 624, 940, 594, 403, 594, 940, 624, 424, 624, 664, 840},
        {664, 1385, 1095, 1029, 698, 1029, 1549, 979, 664, 979, 1549, 1029, 698, 1029, 1095, 1385},
        {424, 883, 698, 656, 445, 656, 987, 624, 424, 624, 987, 656, 445, 656, 698, 883},
        {940, 1959, 1549, 1455, 987, 1455, 2190, 1385, 940, 1385, 2190, 1455, 987, 1455, 1549, 1959},
        {403, 840, 664, 624, 424, 624, 940, 594, 403, 594, 940, 624, 424, 624, 664, 840},
        {940, 1959, 1549, 1455, 987, 1455, 2190, 1385, 940, 1385, 2190, 1455, 987, 1455, 1549, 1959},
        {424, 883, 698, 656, 445, 656, 987, 624, 424, 624, 987, 656, 445, 656, 698, 883},
        {664, 1385, 1095, 1029, 698, 1029, 1549, 979, 664, 979, 1549, 1029, 698, 1029, 1095, 1385},
        }
};

static const int quant_coef_8x16[6][16][8] =
{
  {
    {4640, 7636, 4862, 10809, 4640, 10809, 4862, 7636},
    {9655, 15930, 10138, 22525, 9655, 22525, 10138, 15930},
    {7636, 12579, 8017, 17804, 7636, 17804, 8017, 12579},
    {7183, 11835, 7539, 16717, 7183, 16717, 7539, 11835},
    {4862, 8017, 5107, 11346, 4862, 11346, 5107, 8017},
    {7183, 11835, 7539, 16717, 7183, 16717, 7539, 11835},
    {10809, 17804, 11346, 25178, 10809, 25178, 11346, 17804},
    {6822, 11252, 7183, 15930, 6822, 15930, 7183, 11252},
    {4640, 7636, 4862, 10809, 4640, 10809, 4862, 7636},
    {6822, 11252, 7183, 15930, 6822, 15930, 7183, 11252},
    {10809, 17804, 11346, 25178, 10809, 25178, 11346, 17804},
    {7183, 11835, 7539, 16717, 7183, 16717, 7539, 11835},
    {4862, 8017, 5107, 11346, 4862, 11346, 5107, 8017},
    {7183, 11835, 7539, 16717, 7183, 16717, 7539, 11835},
    {7636, 12579, 8017, 17804, 7636, 17804, 8017, 12579},
    {9655, 15930, 10138, 22525, 9655, 22525, 10138, 15930},
  },
  {
    {4128, 6797, 4334, 9622, 4128, 9622, 4334, 6797},
    {8614, 14178, 9044, 20061, 8614, 20061, 9044, 14178},
    {6797, 11212, 7143, 15852, 6797, 15852, 7143, 11212},
    {6397, 10539, 6717, 14894, 6397, 14894, 6717, 10539},
    {4334, 7143, 4560, 10105, 4334, 10105, 4560, 7143},
    {6397, 10539, 6717, 14894, 6397, 14894, 6717, 10539},
    {9622, 15852, 10105, 22423, 9622, 22423, 10105, 15852},
    {6092, 10030, 6397, 14178, 6092, 14178, 6397, 10030},
    {4128, 6797, 4334, 9622, 4128, 9622, 4334, 6797},
    {6092, 10030, 6397, 14178, 6092, 14178, 6397, 10030},
    {9622, 15852, 10105, 22423, 9622, 22423, 10105, 15852},
    {6397, 10539, 6717, 14894, 6397, 14894, 6717, 10539},
    {4334, 7143, 4560, 10105, 4334, 10105, 4560, 7143},
    {6397, 10539, 6717, 14894, 6397, 14894, 6717, 10539},
    {6797, 11212, 7143, 15852, 6797, 15852, 7143, 11212},
    {8614, 14178, 9044, 20061, 8614, 20061, 9044, 14178},
    },
    {
      {3679, 6060, 3870, 8579, 3679, 8579, 3870, 6060},
      {7672, 12643, 8058, 17874, 7672, 17874, 8058, 12643},
      {6060, 9995, 6362, 14130, 6060, 14130, 6362, 9995},
      {5701, 9381, 5979, 13273, 5701, 13273, 5979, 9381},
      {3870, 6362, 4053, 9005, 3870, 9005, 4053, 6362},
      {5701, 9381, 5979, 13273, 5701, 13273, 5979, 9381},
      {8579, 14130, 9005, 19977, 8579, 19977, 9005, 14130},
      {5425, 8930, 5701, 12643, 5425, 12643, 5701, 8930},
      {3679, 6060, 3870, 8579, 3679, 8579, 3870, 6060},
      {5425, 8930, 5701, 12643, 5425, 12643, 5701, 8930},
      {8579, 14130, 9005, 19977, 8579, 19977, 9005, 14130},
      {5701, 9381, 5979, 13273, 5701, 13273, 5979, 9381},
      {3870, 6362, 4053, 9005, 3870, 9005, 4053, 6362},
      {5701, 9381, 5979, 13273, 5701, 13273, 5979, 9381},
      {6060, 9995, 6362, 14130, 6060, 14130, 6362, 9995},
      {7672, 12643, 8058, 17874, 7672, 17874, 8058, 12643},
    },
    {
      {3277, 5404, 3444, 7636, 3277, 7636, 3444, 5404},
      {6832, 11263, 7173, 15920, 6832, 15920, 7173, 11263},
      {5404, 8902, 5673, 12589, 5404, 12589, 5673, 8902},
      {5069, 8359, 5325, 11825, 5069, 11825, 5325, 8359},
      {3444, 5673, 3617, 8027, 3444, 8027, 3617, 5673},
      {5069, 8359, 5325, 11825, 5069, 11825, 5325, 8359},
      {7636, 12589, 8027, 17804, 7636, 17804, 8027, 12589},
      {4827, 7965, 5069, 11263, 4827, 11263, 5069, 7965},
      {3277, 5404, 3444, 7636, 3277, 7636, 3444, 5404},
      {4827, 7965, 5069, 11263, 4827, 11263, 5069, 7965},
      {7636, 12589, 8027, 17804, 7636, 17804, 8027, 12589},
      {5069, 8359, 5325, 11825, 5069, 11825, 5325, 8359},
      {3444, 5673, 3617, 8027, 3444, 8027, 3617, 5673},
      {5069, 8359, 5325, 11825, 5069, 11825, 5325, 8359},
      {5404, 8902, 5673, 12589, 5404, 12589, 5673, 8902},
      {6832, 11263, 7173, 15920, 6832, 15920, 7173, 11263},
      },
      {
        {2921, 4811, 3069, 6806, 2921, 6806, 3069, 4811},
        {6084, 10030, 6389, 14186, 6084, 14186, 6389, 10030},
        {4811, 7926, 5053, 11212, 4811, 11212, 5053, 7926},
        {4522, 7447, 4751, 10539, 4522, 10539, 4751, 7447},
        {3069, 5053, 3224, 7151, 3069, 7151, 3224, 5053},
        {4522, 7447, 4751, 10539, 4522, 10539, 4751, 7447},
        {6806, 11212, 7151, 15861, 6806, 15861, 7151, 11212},
        {4307, 7089, 4522, 10030, 4307, 10030, 4522, 7089},
        {2921, 4811, 3069, 6806, 2921, 6806, 3069, 4811},
        {4307, 7089, 4522, 10030, 4307, 10030, 4522, 7089},
        {6806, 11212, 7151, 15861, 6806, 15861, 7151, 11212},
        {4522, 7447, 4751, 10539, 4522, 10539, 4751, 7447},
        {3069, 5053, 3224, 7151, 3069, 7151, 3224, 5053},
        {4522, 7447, 4751, 10539, 4522, 10539, 4751, 7447},
        {4811, 7926, 5053, 11212, 4811, 11212, 5053, 7926},
        {6084, 10030, 6389, 14186, 6084, 14186, 6389, 10030},
      },
      {
        {2602, 4289, 2729, 6060, 2602, 6060, 2729, 4289},
        {5425, 8937, 5695, 12637, 5425, 12637, 5695, 8937},
        {4289, 7065, 4503, 9988, 4289, 9988, 4503, 7065},
        {4029, 6637, 4229, 9387, 4029, 9387, 4229, 6637},
        {2729, 4503, 2869, 6368, 2729, 6368, 2869, 4503},
        {4029, 6637, 4229, 9387, 4029, 9387, 4229, 6637},
        {6060, 9988, 6368, 14130, 6060, 14130, 6368, 9988},
        {3836, 6322, 4029, 8937, 3836, 8937, 4029, 6322},
        {2602, 4289, 2729, 6060, 2602, 6060, 2729, 4289},
        {3836, 6322, 4029, 8937, 3836, 8937, 4029, 6322},
        {6060, 9988, 6368, 14130, 6060, 14130, 6368, 9988},
        {4029, 6637, 4229, 9387, 4029, 9387, 4229, 6637},
        {2729, 4503, 2869, 6368, 2729, 6368, 2869, 4503},
        {4029, 6637, 4229, 9387, 4029, 9387, 4229, 6637},
        {4289, 7065, 4503, 9988, 4289, 9988, 4503, 7065},
        {5425, 8937, 5695, 12637, 5425, 12637, 5695, 8937},
        }
};

static const int dequant_coef_8x16[6][16][8] =
{
  {
    {226, 373, 238, 527, 226, 527, 238, 373},
    {472, 777, 496, 1099, 472, 1099, 496, 777},
    {373, 615, 392, 869, 373, 869, 392, 615},
    {350, 577, 368, 817, 350, 817, 368, 577},
    {238, 392, 250, 554, 238, 554, 250, 392},
    {350, 577, 368, 817, 350, 817, 368, 577},
    {527, 869, 554, 1229, 527, 1229, 554, 869},
    {334, 550, 350, 777, 334, 777, 350, 550},
    {226, 373, 238, 527, 226, 527, 238, 373},
    {334, 550, 350, 777, 334, 777, 350, 550},
    {527, 869, 554, 1229, 527, 1229, 554, 869},
    {350, 577, 368, 817, 350, 817, 368, 577},
    {238, 392, 250, 554, 238, 554, 250, 392},
    {350, 577, 368, 817, 350, 817, 368, 577},
    {373, 615, 392, 869, 373, 869, 392, 615},
    {472, 777, 496, 1099, 472, 1099, 496, 777},
  },
  {
    {254, 419, 267, 592, 254, 592, 267, 419},
    {529, 873, 556, 1234, 529, 1234, 556, 873},
    {419, 690, 440, 976, 419, 976, 440, 690},
    {393, 648, 413, 917, 393, 917, 413, 648},
    {267, 440, 280, 622, 267, 622, 280, 440},
    {393, 648, 413, 917, 393, 917, 413, 648},
    {592, 976, 622, 1380, 592, 1380, 622, 976},
    {374, 617, 393, 873, 374, 873, 393, 617},
    {254, 419, 267, 592, 254, 592, 267, 419},
    {374, 617, 393, 873, 374, 873, 393, 617},
    {592, 976, 622, 1380, 592, 1380, 622, 976},
    {393, 648, 413, 917, 393, 917, 413, 648},
    {267, 440, 280, 622, 267, 622, 280, 440},
    {393, 648, 413, 917, 393, 917, 413, 648},
    {419, 690, 440, 976, 419, 976, 440, 690},
    {529, 873, 556, 1234, 529, 1234, 556, 873},
    },
    {
      {285, 470, 299, 664, 285, 664, 299, 470},
      {594, 979, 624, 1385, 594, 1385, 624, 979},
      {470, 774, 494, 1095, 470, 1095, 494, 774},
      {441, 728, 464, 1029, 441, 1029, 464, 728},
      {299, 494, 315, 698, 299, 698, 315, 494},
      {441, 728, 464, 1029, 441, 1029, 464, 728},
      {664, 1095, 698, 1549, 664, 1549, 698, 1095},
      {420, 693, 441, 979, 420, 979, 441, 693},
      {285, 470, 299, 664, 285, 664, 299, 470},
      {420, 693, 441, 979, 420, 979, 441, 693},
      {664, 1095, 698, 1549, 664, 1549, 698, 1095},
      {441, 728, 464, 1029, 441, 1029, 464, 728},
      {299, 494, 315, 698, 299, 698, 315, 494},
      {441, 728, 464, 1029, 441, 1029, 464, 728},
      {470, 774, 494, 1095, 470, 1095, 494, 774},
      {594, 979, 624, 1385, 594, 1385, 624, 979},
    },
    {
      {320, 527, 336, 746, 320, 746, 336, 527},
      {667, 1099, 701, 1555, 667, 1555, 701, 1099},
      {527, 869, 554, 1229, 527, 1229, 554, 869},
      {496, 817, 521, 1155, 496, 1155, 521, 817},
      {336, 554, 353, 783, 336, 783, 353, 554},
      {496, 817, 521, 1155, 496, 1155, 521, 817},
      {746, 1229, 783, 1738, 746, 1738, 783, 1229},
      {472, 777, 496, 1099, 472, 1099, 496, 777},
      {320, 527, 336, 746, 320, 746, 336, 527},
      {472, 777, 496, 1099, 472, 1099, 496, 777},
      {746, 1229, 783, 1738, 746, 1738, 783, 1229},
      {496, 817, 521, 1155, 496, 1155, 521, 817},
      {336, 554, 353, 783, 336, 783, 353, 554},
      {496, 817, 521, 1155, 496, 1155, 521, 817},
      {527, 869, 554, 1229, 527, 1229, 554, 869},
      {667, 1099, 701, 1555, 667, 1555, 701, 1099},
      },
      {
        {359, 592, 377, 837, 359, 837, 377, 592},
        {749, 1234, 787, 1745, 749, 1745, 787, 1234},
        {592, 976, 622, 1380, 592, 1380, 622, 976},
        {556, 917, 584, 1296, 556, 1296, 584, 917},
        {377, 622, 396, 879, 377, 879, 396, 622},
        {556, 917, 584, 1296, 556, 1296, 584, 917},
        {837, 1380, 879, 1951, 837, 1951, 879, 1380},
        {529, 873, 556, 1234, 529, 1234, 556, 873},
        {359, 592, 377, 837, 359, 837, 377, 592},
        {529, 873, 556, 1234, 529, 1234, 556, 873},
        {837, 1380, 879, 1951, 837, 1951, 879, 1380},
        {556, 917, 584, 1296, 556, 1296, 584, 917},
        {377, 622, 396, 879, 377, 879, 396, 622},
        {556, 917, 584, 1296, 556, 1296, 584, 917},
        {592, 976, 622, 1380, 592, 1380, 622, 976},
        {749, 1234, 787, 1745, 749, 1745, 787, 1234},
      },
      {
        {403, 664, 424, 940, 403, 940, 424, 664},
        {840, 1385, 883, 1959, 840, 1959, 883, 1385},
        {664, 1095, 698, 1549, 664, 1549, 698, 1095},
        {624, 1029, 656, 1455, 624, 1455, 656, 1029},
        {424, 698, 445, 987, 424, 987, 445, 698},
        {624, 1029, 656, 1455, 624, 1455, 656, 1029},
        {940, 1549, 987, 2190, 940, 2190, 987, 1549},
        {594, 979, 624, 1385, 594, 1385, 624, 979},
        {403, 664, 424, 940, 403, 940, 424, 664},
        {594, 979, 624, 1385, 594, 1385, 624, 979},
        {940, 1549, 987, 2190, 940, 2190, 987, 1549},
        {624, 1029, 656, 1455, 624, 1455, 656, 1029},
        {424, 698, 445, 987, 424, 987, 445, 698},
        {624, 1029, 656, 1455, 624, 1455, 656, 1029},
        {664, 1095, 698, 1549, 664, 1549, 698, 1095},
        {840, 1385, 883, 1959, 840, 1959, 883, 1385},
        }
};


#ifdef RDO_Q
static const int estErr16x16[6][16][16]=
{
  {
    {6553600, 6539552, 6521472, 6603264, 6547968, 6554624, 6558272, 6565184, 6553600, 6559168, 6558272, 6594816, 6547968, 6546816, 6531968, 6575648},
    {6539552, 6533224, 6545040, 6565432, 6523104, 6526040, 6554080, 6567576, 6576640, 6557752, 6554080, 6545552, 6523104, 6538984, 6559424, 6555248},
    {6521472, 6535520, 6548784, 6558128, 6531904, 6530848, 6569104, 6564624, 6553088, 6568544, 6569104, 6558432, 6531904, 6526440, 6568640, 6558624},
    {6603264, 6517326, 6526284, 6574484, 6550168, 6567514, 6537468, 6540438, 6561216, 6532130, 6537468, 6576230, 6550168, 6562908, 6539436, 6535828},
    {6547968, 6493400, 6513888, 6588080, 6572736, 6512096, 6563280, 6562832, 6547968, 6558032, 6563280, 6547728, 6572736, 6530352, 6563808, 6531216},
    {6554624, 6521282, 6539180, 6568244, 6547864, 6557910, 6535556, 6536394, 6581312, 6532602, 6535556, 6564302, 6547864, 6558888, 6546524, 6539508},
    {6558272, 6523492, 6544136, 6578232, 6557568, 6524084, 6542024, 6562812, 6561280, 6559824, 6542024, 6550632, 6557568, 6532020, 6575504, 6541664},
    {6565184, 6538720, 6512480, 6580464, 6571264, 6522016, 6572864, 6553360, 6562432, 6552352, 6572864, 6547152, 6571264, 6526432, 6560192, 6561152},
    {6553600, 6539552, 6521472, 6603264, 6547968, 6554624, 6558272, 6565184, 6553600, 6559168, 6558272, 6594816, 6547968, 6546816, 6531968, 6575648},
    {6559168, 6538720, 6512480, 6580464, 6571264, 6522016, 6572864, 6553360, 6562432, 6552352, 6572864, 6547152, 6571264, 6526432, 6560192, 6561152},
    {6558272, 6523492, 6544136, 6578232, 6557568, 6524084, 6542024, 6562812, 6561280, 6559824, 6542024, 6550632, 6557568, 6532020, 6575504, 6541664},
    {6594816, 6521262, 6538908, 6568196, 6547832, 6557874, 6535532, 6536374, 6581312, 6532538, 6535532, 6564270, 6547832, 6558828, 6546420, 6539080},
    {6547968, 6493400, 6513888, 6588080, 6572736, 6512096, 6563280, 6562832, 6547968, 6558032, 6563280, 6547728, 6572736, 6530352, 6563808, 6531216},
    {6546816, 6517306, 6528276, 6574156, 6550136, 6567478, 6537444, 6540418, 6561216, 6532066, 6537444, 6576198, 6550136, 6562568, 6540076, 6535860},
    {6531968, 6535520, 6548784, 6558128, 6531904, 6530848, 6569104, 6564624, 6553088, 6568544, 6569104, 6558432, 6531904, 6526440, 6568640, 6558624},
    {6575648, 6533224, 6545040, 6565432, 6523104, 6526040, 6554080, 6567576, 6576640, 6557752, 6554080, 6545552, 6523104, 6538984, 6559424, 6555248},
  },
  {
    {8294400, 8203776, 8204672, 8314880, 8260096, 8194880, 8270528, 8265472, 8294400, 8258560, 8270528, 8239680, 8260096, 8236352, 8216448, 8244224},
    {8203776, 8230648, 8239296, 8292504, 8233312, 8247848, 8265280, 8258896, 8245312, 8251072, 8265280, 8261392, 8233312, 8271064, 8271536, 8247264},
    {8204672, 8221680, 8238656, 8290216, 8239840, 8202136, 8257424, 8229544, 8240128, 8242064, 8257424, 8245192, 8239840, 8228056, 8241088, 8246848},
    {8314880, 8242242, 8208708, 8298168, 8260816, 8218544, 8257464, 8269266, 8260128, 8256028, 8257464, 8284394, 8260816, 8246350, 8235748, 8262792},
    {8260096, 8203984, 8191328, 8265592, 8200192, 8224344, 8282384, 8264088, 8302464, 8256528, 8282384, 8278720, 8200192, 8201216, 8245232, 8244440},
    {8194880, 8242126, 8214028, 8296420, 8256768, 8227136, 8252000, 8261154, 8239008, 8256028, 8252000, 8282950, 8256768, 8254870, 8233604, 8263904},
    {8270528, 8240740, 8244064, 8281040, 8269376, 8228188, 8262784, 8253128, 8273920, 8251304, 8262784, 8254596, 8269376, 8246824, 8263056, 8260300},
    {8265472, 8222224, 8191456, 8288944, 8238592, 8240640, 8250432, 8251792, 8262272, 8240912, 8250432, 8268624, 8238592, 8242288, 8216864, 8255392},
    {8294400, 8203776, 8204672, 8314880, 8260096, 8194880, 8270528, 8265472, 8294400, 8258560, 8270528, 8239680, 8260096, 8236352, 8216448, 8244224},
    {8258560, 8222224, 8191456, 8288944, 8238592, 8240640, 8250432, 8251792, 8262272, 8240912, 8250432, 8268624, 8238592, 8242288, 8216864, 8255392},
    {8270528, 8240740, 8244064, 8281040, 8269376, 8228188, 8262784, 8253128, 8273920, 8251304, 8262784, 8254596, 8269376, 8246824, 8263056, 8260300},
    {8239680, 8242114, 8214036, 8296432, 8256784, 8227104, 8246928, 8261130, 8239008, 8256004, 8246928, 8282906, 8256784, 8254838, 8233524, 8263784},
    {8260096, 8203984, 8191328, 8265592, 8200192, 8224344, 8282384, 8264088, 8302464, 8256528, 8282384, 8278720, 8200192, 8201216, 8245232, 8244440},
    {8236352, 8242230, 8208716, 8298180, 8260832, 8218512, 8256976, 8269242, 8260128, 8256004, 8256976, 8284350, 8260832, 8246318, 8235668, 8263804},
    {8216448, 8221680, 8238656, 8290216, 8239840, 8202136, 8257424, 8229544, 8240128, 8242064, 8257424, 8245192, 8239840, 8228056, 8241088, 8246848},
    {8244224, 8230648, 8239296, 8292504, 8233312, 8247848, 8265280, 8258896, 8245312, 8251072, 8265280, 8261392, 8233312, 8271064, 8271536, 8247264},
    },
    {
      {10445824, 10357824, 10373376, 10458848, 10427008, 10340768, 10433728, 10387968, 10445824, 10380288, 10433728, 10391136, 10427008, 10387744, 10417344, 10403200},
      {10357824, 10385432, 10393488, 10432040, 10365856, 10373280, 10392960, 10389920, 10404480, 10386376, 10392960, 10420080, 10365856, 10371464, 10421200, 10400704},
      {10373376, 10396752, 10407264, 10411184, 10364928, 10354512, 10411136, 10405200, 10418624, 10418080, 10411136, 10402160, 10364928, 10364600, 10422832, 10416648},
      {10458848, 10379644, 10373816, 10417408, 10405456, 10335894, 10435648, 10409940, 10405920, 10406242, 10435648, 10368660, 10405456, 10364384, 10399060, 10407714},
      {10427008, 10348112, 10344464, 10454072, 10336512, 10395768, 10429792, 10383280, 10427008, 10385336, 10429792, 10444192, 10336512, 10387640, 10413856, 10392672},
      {10340768, 10383852, 10373664, 10420672, 10414480, 10332770, 10432176, 10414112, 10374112, 10410290, 10432176, 10374632, 10414480, 10360444, 10393908, 10408062},
      {10433728, 10376740, 10371040, 10446976, 10414208, 10407184, 10406880, 10412400, 10422432, 10417128, 10406880, 10418432, 10414208, 10407112, 10404752, 10398732},
      {10387968, 10355328, 10377952, 10441040, 10390016, 10375344, 10412288, 10406000, 10384384, 10403776, 10412288, 10410368, 10390016, 10378384, 10409920, 10392544},
      {10445824, 10357824, 10373376, 10458848, 10427008, 10340768, 10433728, 10387968, 10445824, 10380288, 10433728, 10391136, 10427008, 10387744, 10417344, 10403200},
      {10380288, 10355328, 10377952, 10441040, 10390016, 10375344, 10412288, 10406000, 10384384, 10403776, 10412288, 10410368, 10390016, 10378384, 10409920, 10392544},
      {10433728, 10376740, 10371040, 10446976, 10414208, 10407184, 10406880, 10412400, 10422432, 10417128, 10406880, 10418432, 10414208, 10407112, 10404752, 10398732},
      {10391136, 10383788, 10373576, 10420688, 10416736, 10336078, 10432192, 10414084, 10374112, 10410174, 10432192, 10374596, 10416736, 10362900, 10393924, 10408034},
      {10427008, 10348112, 10344464, 10454072, 10336512, 10395768, 10429792, 10383280, 10427008, 10385336, 10429792, 10444192, 10336512, 10387640, 10413856, 10392672},
      {10387744, 10379580, 10373728, 10417424, 10405248, 10336186, 10435664, 10409912, 10405920, 10406202, 10435664, 10368624, 10405248, 10364572, 10399076, 10407686},
      {10417344, 10396752, 10407264, 10411184, 10364928, 10354512, 10411136, 10405200, 10418624, 10418080, 10411136, 10402160, 10364928, 10364600, 10422832, 10416648},
      {10403200, 10385432, 10393488, 10432040, 10365856, 10373280, 10392960, 10389920, 10404480, 10386376, 10392960, 10420080, 10365856, 10371464, 10421200, 10400704},
    },
    {
      {13075456, 13077888, 13047936, 13144768, 13103360, 13006400, 13072320, 13157312, 13075456, 13148608, 13072320, 13090816, 13103360, 13028160, 13137536, 13128896},
      {13077888, 13072320, 13067280, 13122120, 13131808, 13052528, 13085408, 13105352, 13130304, 13106144, 13085408, 13100696, 13131808, 13067240, 13100272, 13097416},
      {13047936, 13046984, 13088272, 13156464, 13115136, 13048168, 13084048, 13125624, 13122560, 13119648, 13084048, 13101264, 13115136, 13064040, 13151008, 13083328},
      {13144768, 13052806, 13040296, 13156004, 13105536, 13080634, 13113632, 13086818, 13074528, 13082916, 13113632, 13123222, 13105536, 13075148, 13093188, 13083366},
      {13103360, 13097112, 13063264, 13175888, 13109440, 13066584, 13118432, 13063248, 13141440, 13052088, 13118432, 13121968, 13109440, 13108672, 13082288, 13137288},
      {13006400, 13056002, 13043864, 13170708, 13103040, 13070290, 13126400, 13081550, 13074528, 13080516, 13126400, 13125278, 13103040, 13075900, 13083804, 13081582},
      {13072320, 13080268, 13044056, 13129708, 13077424, 13112860, 13097632, 13097052, 13076608, 13084520, 13097632, 13141340, 13077424, 13100932, 13096848, 13097628},
      {13157312, 13066448, 13090080, 13130864, 13046208, 13052080, 13108160, 13135152, 13153280, 13115504, 13108160, 13091104, 13046208, 13077968, 13118848, 13110496},
      {13075456, 13077888, 13047936, 13144768, 13103360, 13006400, 13072320, 13157312, 13075456, 13148608, 13072320, 13090816, 13103360, 13028160, 13137536, 13128896},
      {13148608, 13066448, 13090080, 13130864, 13046208, 13052080, 13108160, 13135152, 13153280, 13115504, 13108160, 13091104, 13046208, 13077968, 13118848, 13110496},
      {13072320, 13080268, 13044056, 13129708, 13077424, 13112860, 13097632, 13097052, 13076608, 13084520, 13097632, 13141340, 13077424, 13100932, 13096848, 13097628},
      {13090816, 13055514, 13043992, 13170708, 13103040, 13072726, 13126816, 13081530, 13074464, 13080492, 13126816, 13127730, 13103040, 13075852, 13083636, 13081114},
      {13103360, 13097112, 13063264, 13175888, 13109440, 13066584, 13118432, 13063248, 13141440, 13052088, 13118432, 13121968, 13109440, 13108672, 13082288, 13137288},
      {13028160, 13052778, 13041400, 13156004, 13105536, 13080802, 13114712, 13086798, 13074464, 13082892, 13114712, 13123406, 13105536, 13075100, 13092044, 13083358},
      {13137536, 13046984, 13088272, 13156464, 13115136, 13048168, 13084048, 13125624, 13122560, 13119648, 13084048, 13101264, 13115136, 13064040, 13151008, 13083328},
      {13128896, 13072320, 13067280, 13122120, 13131808, 13052528, 13085408, 13105352, 13130304, 13106144, 13085408, 13100696, 13131808, 13067240, 13100272, 13097416},
      },
      {
        {16516096, 16452800, 16464000, 16564128, 16457984, 16410752, 16525120, 16495104, 16516096, 16485376, 16525120, 16518944, 16457984, 16457472, 16564608, 16509952},
        {16452800, 16485392, 16506256, 16549168, 16512032, 16493912, 16523648, 16503968, 16511552, 16497392, 16523648, 16524992, 16512032, 16510448, 16542768, 16516936},
        {16464000, 16481480, 16488128, 16562080, 16538752, 16456376, 16525568, 16506256, 16547840, 16502608, 16525568, 16509192, 16538752, 16493648, 16526112, 16520600},
        {16564128, 16468990, 16462540, 16583456, 16485504, 16489680, 16535036, 16530182, 16512608, 16518426, 16535036, 16534942, 16485504, 16465296, 16473804, 16504718},
        {16457984, 16447480, 16506256, 16556232, 16483616, 16446960, 16532384, 16488072, 16517760, 16472744, 16532384, 16525208, 16483616, 16459608, 16516736, 16493616},
        {16410752, 16469254, 16459236, 16579052, 16482880, 16485228, 16517444, 16538562, 16499328, 16530506, 16517444, 16529146, 16482880, 16473084, 16476780, 16501762},
        {16525120, 16479300, 16501056, 16569000, 16512768, 16492716, 16519784, 16516016, 16515648, 16523652, 16519784, 16547720, 16512768, 16496060, 16503176, 16505212},
        {16495104, 16461296, 16478592, 16585008, 16466624, 16495696, 16530560, 16517792, 16490624, 16502144, 16530560, 16522784, 16466624, 16542128, 16543072, 16494528},
        {16516096, 16452800, 16464000, 16564128, 16457984, 16410752, 16525120, 16495104, 16516096, 16485376, 16525120, 16518944, 16457984, 16457472, 16564608, 16509952},
        {16485376, 16461296, 16478592, 16585008, 16466624, 16495696, 16530560, 16517792, 16490624, 16502144, 16530560, 16522784, 16466624, 16542128, 16543072, 16494528},
        {16525120, 16479300, 16501056, 16569000, 16512768, 16492716, 16519784, 16516016, 16515648, 16523652, 16519784, 16547720, 16512768, 16496060, 16503176, 16505212},
        {16518944, 16469306, 16458868, 16579300, 16482880, 16485176, 16517388, 16538358, 16499232, 16530726, 16517388, 16529118, 16482880, 16472760, 16476700, 16501662},
        {16457984, 16447480, 16506256, 16556232, 16483616, 16446960, 16532384, 16488072, 16517760, 16472744, 16532384, 16525208, 16483616, 16459608, 16516736, 16493616},
        {16457472, 16469426, 16462476, 16581036, 16485504, 16489628, 16534980, 16530486, 16512512, 16518026, 16534980, 16534914, 16485504, 16467624, 16473724, 16503946},
        {16564608, 16481480, 16488128, 16562080, 16538752, 16456376, 16525568, 16506256, 16547840, 16502608, 16525568, 16509192, 16538752, 16493648, 16526112, 16520600},
        {16509952, 16485392, 16506256, 16549168, 16512032, 16493912, 16523648, 16503968, 16511552, 16497392, 16523648, 16524992, 16512032, 16510448, 16542768, 16516936},
      },
      {
        {20793600, 20710912, 20780928, 20866752, 20650240, 20664288, 20807808, 20813952, 20793600, 20803072, 20807808, 20783168, 20650240, 20726240, 20843072, 20775040},
        {20710912, 20776264, 20751808, 20825704, 20796224, 20773472, 20805136, 20809104, 20776832, 20794296, 20805136, 20812504, 20796224, 20756160, 20806048, 20805560},
        {20780928, 20753480, 20742080, 20893952, 20828416, 20814368, 20813760, 20824800, 20844864, 20834256, 20813760, 20836864, 20828416, 20814224, 20809504, 20797464},
        {20866752, 20761770, 20806304, 20884710, 20859472, 20798528, 20791660, 20818944, 20784448, 20813764, 20791660, 20845912, 20859472, 20819740, 20818452, 20796428},
        {20650240, 20725072, 20780256, 20917720, 20797440, 20761776, 20809632, 20779328, 20717184, 20777344, 20809632, 20831432, 20797440, 20834944, 20838496, 20788096},
        {20664288, 20754622, 20812200, 20886398, 20814640, 20807964, 20787812, 20820632, 20773312, 20814368, 20787812, 20849064, 20814640, 20824040, 20822316, 20788296},
        {20807808, 20776780, 20785872, 20866696, 20773248, 20773064, 20811240, 20796400, 20797184, 20797116, 20811240, 20806532, 20773248, 20802984, 20819680, 20796804},
        {20813952, 20770864, 20786976, 20864080, 20731712, 20746560, 20785920, 20779840, 20808960, 20772752, 20785920, 20840160, 20731712, 20742928, 20842400, 20801408},
        {20793600, 20710912, 20780928, 20866752, 20650240, 20664288, 20807808, 20813952, 20793600, 20803072, 20807808, 20783168, 20650240, 20726240, 20843072, 20775040},
        {20803072, 20770864, 20786976, 20864080, 20731712, 20746560, 20785920, 20779840, 20808960, 20772752, 20785920, 20840160, 20731712, 20742928, 20842400, 20801408},
        {20807808, 20776780, 20785872, 20866696, 20773248, 20773064, 20811240, 20796400, 20797184, 20797116, 20811240, 20806532, 20773248, 20802984, 20819680, 20796804},
        {20783168, 20754626, 20810472, 20886354, 20814640, 20807960, 20786100, 20820616, 20773120, 20814340, 20786100, 20849032, 20814640, 20824028, 20820556, 20788276},
        {20650240, 20725072, 20780256, 20917720, 20797440, 20761776, 20809632, 20779328, 20717184, 20777344, 20809632, 20831432, 20797440, 20834944, 20838496, 20788096},
        {20726240, 20761774, 20806416, 20884586, 20859472, 20798524, 20791492, 20818928, 20784256, 20813736, 20791492, 20845976, 20859472, 20819728, 20818532, 20796408},
        {20843072, 20753480, 20742080, 20893952, 20828416, 20814368, 20813760, 20824800, 20844864, 20834256, 20813760, 20836864, 20828416, 20814224, 20809504, 20797464},
        {20775040, 20776264, 20751808, 20825704, 20796224, 20773472, 20805136, 20809104, 20776832, 20794296, 20805136, 20812504, 20796224, 20756160, 20806048, 20805560},
        }
};

static const int estErr16x8[6][8][16]=
{
  {
    {6537728, 6561808, 6561216, 6541312, 6570720, 6536592, 6544592, 6575992, 6537728, 6576576, 6544592, 6537072, 6570720, 6541312, 6561216, 6561808},
    {6561216, 6544245, 6561328, 6548068, 6557328, 6543605, 6553920, 6559996, 6561216, 6558773, 6553920, 6543133, 6557328, 6547397, 6561328, 6545679},
    {6570720, 6570592, 6557328, 6555360, 6574704, 6550124, 6548172, 6535978, 6570720, 6536164, 6548172, 6550124, 6574704, 6555360, 6557328, 6570592},
    {6544592, 6544326, 6553920, 6558908, 6548172, 6560268, 6549248, 6543349, 6544592, 6543333, 6549248, 6559577, 6548172, 6558785, 6553920, 6546167},
    {6537728, 6561808, 6561216, 6541312, 6570720, 6536592, 6544592, 6575992, 6537728, 6576576, 6544592, 6537072, 6570720, 6541312, 6561216, 6561808},
    {6544592, 6544326, 6553920, 6558908, 6548172, 6560268, 6549248, 6543349, 6544592, 6543333, 6549248, 6559577, 6548172, 6558785, 6553920, 6546167},
    {6570720, 6570592, 6557328, 6555360, 6574704, 6550124, 6548172, 6535978, 6570720, 6536164, 6548172, 6550124, 6574704, 6555360, 6557328, 6570592},
    {6561216, 6544245, 6561328, 6548068, 6557328, 6543605, 6553920, 6559996, 6561216, 6558773, 6553920, 6543133, 6557328, 6547397, 6561328, 6545679},
  },
  {
    {8258048, 8236432, 8262544, 8242152, 8280224, 8248040, 8257824, 8232896, 8258048, 8232896, 8257824, 8248040, 8280224, 8242152, 8273856, 8236432},
    {8262544, 8264686, 8254738, 8254148, 8265696, 8252967, 8269356, 8258978, 8269248, 8259172, 8269356, 8252637, 8265696, 8254148, 8254738, 8266321},
    {8280224, 8249646, 8265696, 8251128, 8241808, 8250936, 8269416, 8243320, 8280224, 8244450, 8269416, 8250936, 8241808, 8251128, 8265696, 8249646},
    {8257824, 8252619, 8269356, 8270265, 8269416, 8264114, 8265214, 8264187, 8257824, 8263493, 8265214, 8265482, 8269416, 8268447, 8269356, 8253448},
    {8258048, 8236432, 8262544, 8242152, 8280224, 8248040, 8257824, 8232896, 8258048, 8232896, 8257824, 8248040, 8280224, 8242152, 8273856, 8236432},
    {8257824, 8252619, 8269356, 8270265, 8269416, 8264114, 8265214, 8264187, 8257824, 8263493, 8265214, 8265482, 8269416, 8268447, 8269356, 8253448},
    {8280224, 8249646, 8265696, 8251128, 8241808, 8250936, 8269416, 8243320, 8280224, 8244450, 8269416, 8250936, 8241808, 8251128, 8265696, 8249646},
    {8273856, 8264686, 8254738, 8254148, 8265696, 8252967, 8269356, 8258978, 8269248, 8259172, 8269356, 8252637, 8265696, 8254148, 8254738, 8266321},
    },
    {
      {10396800, 10405312, 10418624, 10388648, 10382496, 10393440, 10378656, 10391600, 10396800, 10391600, 10378656, 10393440, 10382496, 10390936, 10418624, 10405312},
      {10418624, 10393404, 10389936, 10417543, 10421492, 10415965, 10410708, 10413473, 10418624, 10413473, 10408920, 10415965, 10421492, 10418877, 10389936, 10395779},
      {10382496, 10392304, 10421492, 10418376, 10429536, 10417804, 10398984, 10381676, 10382496, 10381752, 10398984, 10417804, 10429536, 10418376, 10421492, 10392304},
      {10378656, 10397057, 10405152, 10410201, 10398984, 10405626, 10404694, 10393826, 10378656, 10393046, 10404428, 10407612, 10398984, 10409335, 10408052, 10399267},
      {10396800, 10405312, 10418624, 10388648, 10382496, 10393440, 10378656, 10391600, 10396800, 10391600, 10378656, 10393440, 10382496, 10390936, 10418624, 10405312},
      {10378656, 10397057, 10405152, 10410201, 10398984, 10405626, 10404694, 10393826, 10378656, 10393046, 10404428, 10407612, 10398984, 10409335, 10408052, 10399267},
      {10382496, 10392304, 10421492, 10418376, 10429536, 10417804, 10398984, 10381676, 10382496, 10381752, 10398984, 10417804, 10429536, 10418376, 10421492, 10392304},
      {10418624, 10393404, 10389936, 10417543, 10421492, 10415965, 10410708, 10413473, 10418624, 10413473, 10408920, 10415965, 10421492, 10418877, 10389936, 10395779},
    },
    {
      {13107200, 13099736, 13100768, 13127840, 13095936, 13137384, 13106016, 13123616, 13107200, 13123616, 13106016, 13137384, 13095936, 13127840, 13100768, 13099736},
      {13100768, 13097979, 13113980, 13120604, 13111280, 13119554, 13108290, 13098317, 13100768, 13096179, 13106864, 13119193, 13111280, 13120347, 13114438, 13099257},
      {13095936, 13114646, 13111280, 13123698, 13124160, 13125930, 13081800, 13131744, 13095936, 13131744, 13079108, 13125472, 13124160, 13123124, 13111280, 13116190},
      {13106016, 13105269, 13104094, 13113994, 13081800, 13106568, 13111772, 13099860, 13106016, 13099860, 13111772, 13109054, 13081800, 13111508, 13106168, 13106499},
      {13107200, 13099736, 13100768, 13127840, 13095936, 13137384, 13106016, 13123616, 13107200, 13123616, 13106016, 13137384, 13095936, 13127840, 13100768, 13099736},
      {13106016, 13105269, 13104094, 13113994, 13081800, 13106568, 13111772, 13099860, 13106016, 13099860, 13111772, 13109054, 13081800, 13111508, 13106168, 13106499},
      {13095936, 13114646, 13111280, 13123698, 13124160, 13125930, 13081800, 13131744, 13095936, 13131744, 13079108, 13125472, 13124160, 13123124, 13111280, 13116190},
      {13100768, 13097979, 13113980, 13120604, 13111280, 13119554, 13108290, 13098317, 13100768, 13096179, 13106864, 13119193, 13111280, 13120347, 13114438, 13099257},
      },
      {
        {16496768, 16526776, 16515648, 16505136, 16471904, 16503336, 16515856, 16484888, 16496768, 16487136, 16515856, 16501008, 16471904, 16505136, 16515648, 16526776},
        {16515648, 16510829, 16532462, 16528117, 16521956, 16530171, 16511402, 16529568, 16515648, 16529729, 16511402, 16530348, 16521956, 16528117, 16532462, 16512533},
        {16471904, 16535722, 16521956, 16503942, 16495200, 16509440, 16488080, 16506906, 16471904, 16506906, 16488080, 16509540, 16495200, 16504576, 16521956, 16535722},
        {16515856, 16507822, 16511402, 16504325, 16488080, 16504494, 16506552, 16511197, 16515856, 16511332, 16506552, 16504494, 16488080, 16504325, 16511402, 16507822},
        {16496768, 16526776, 16515648, 16505136, 16471904, 16503336, 16515856, 16484888, 16496768, 16487136, 16515856, 16501008, 16471904, 16505136, 16515648, 16526776},
        {16515856, 16507822, 16511402, 16504325, 16488080, 16504494, 16506552, 16511197, 16515856, 16511332, 16506552, 16504494, 16488080, 16504325, 16511402, 16507822},
        {16471904, 16535722, 16521956, 16503942, 16495200, 16509440, 16488080, 16506906, 16471904, 16506906, 16488080, 16509540, 16495200, 16504576, 16521956, 16535722},
        {16515648, 16510829, 16532462, 16528117, 16521956, 16530171, 16511402, 16529568, 16515648, 16529729, 16511402, 16530348, 16521956, 16528117, 16532462, 16512533},
      },
      {
        {20788352, 20769328, 20757312, 20785152, 20854016, 20790920, 20812784, 20792496, 20788352, 20792496, 20812784, 20790920, 20854016, 20785152, 20757312, 20769328},
        {20757312, 20799857, 20806386, 20813768, 20801280, 20810249, 20813910, 20786937, 20757312, 20785982, 20813910, 20813432, 20801280, 20813868, 20809504, 20799857},
        {20854016, 20809670, 20801280, 20812720, 20797248, 20822060, 20817152, 20777424, 20854016, 20777424, 20817152, 20822060, 20797248, 20812720, 20801280, 20809670},
        {20812784, 20807671, 20813910, 20800207, 20811292, 20806752, 20796264, 20797000, 20812784, 20796486, 20796264, 20806752, 20811292, 20799424, 20816524, 20810797},
        {20788352, 20769328, 20757312, 20785152, 20854016, 20790920, 20812784, 20792496, 20788352, 20792496, 20812784, 20790920, 20854016, 20785152, 20757312, 20769328},
        {20812784, 20807671, 20813910, 20800207, 20811292, 20806752, 20796264, 20797000, 20812784, 20796486, 20796264, 20806752, 20811292, 20799424, 20816524, 20810797},
        {20854016, 20809670, 20801280, 20812720, 20797248, 20822060, 20817152, 20777424, 20854016, 20777424, 20817152, 20822060, 20797248, 20812720, 20801280, 20809670},
        {20757312, 20799857, 20806386, 20813768, 20801280, 20810249, 20813910, 20786937, 20757312, 20785982, 20813910, 20813432, 20801280, 20813868, 20809504, 20799857},
        }
};

static const int estErr8x16[6][16][8]=
{
  {
    {6537728, 6561216, 6570720, 6544592, 6537728, 6544592, 6570720, 6561216},
    {6561808, 6545161, 6570592, 6546270, 6561808, 6546270, 6570592, 6546413},
    {6561216, 6561328, 6557328, 6553920, 6561216, 6553920, 6557328, 6561328},
    {6541312, 6546531, 6555360, 6559680, 6539092, 6558866, 6555360, 6548174},
    {6570720, 6557328, 6574704, 6548172, 6570720, 6548172, 6574704, 6557328},
    {6536592, 6542570, 6550124, 6562225, 6536832, 6560810, 6550124, 6543378},
    {6544592, 6553920, 6548172, 6549248, 6544592, 6549248, 6548172, 6553920},
    {6575992, 6558044, 6535978, 6544863, 6576284, 6543536, 6535978, 6558044},
    {6537728, 6561216, 6570720, 6544592, 6537728, 6544592, 6570720, 6561216},
    {6576576, 6558044, 6535978, 6544863, 6576284, 6543536, 6535978, 6558044},
    {6544592, 6553920, 6548172, 6549248, 6544592, 6549248, 6548172, 6553920},
    {6537072, 6542916, 6550124, 6561531, 6536832, 6561022, 6550124, 6543161},
    {6570720, 6557328, 6574704, 6548172, 6570720, 6548172, 6574704, 6557328},
    {6541312, 6545857, 6555360, 6559703, 6539092, 6558584, 6555360, 6547459},
    {6561216, 6561328, 6557328, 6553920, 6561216, 6553920, 6557328, 6561328},
    {6561808, 6545161, 6570592, 6546270, 6561808, 6546270, 6570592, 6546413},
  },
  {
    {8258048, 8262544, 8280224, 8257824, 8258048, 8257824, 8280224, 8273856},
    {8236432, 8264375, 8249646, 8253024, 8236432, 8253024, 8249646, 8264912},
    {8262544, 8254738, 8265696, 8269356, 8269248, 8269356, 8265696, 8254738},
    {8242152, 8253281, 8251362, 8272309, 8245296, 8270265, 8251362, 8253281},
    {8280224, 8265696, 8241808, 8269416, 8280224, 8269416, 8241808, 8265696},
    {8248040, 8252543, 8252364, 8266382, 8248396, 8266052, 8252364, 8252543},
    {8257824, 8269356, 8269416, 8265214, 8257824, 8265214, 8269416, 8269356},
    {8232896, 8257417, 8243922, 8263464, 8235748, 8261776, 8243922, 8259615},
    {8258048, 8262544, 8280224, 8257824, 8258048, 8257824, 8280224, 8273856},
    {8232896, 8257417, 8243922, 8263464, 8235748, 8261776, 8243922, 8259615},
    {8257824, 8269356, 8269416, 8265214, 8257824, 8265214, 8269416, 8269356},
    {8248040, 8252543, 8252364, 8266382, 8248396, 8266217, 8252364, 8252543},
    {8280224, 8265696, 8241808, 8269416, 8280224, 8269416, 8241808, 8265696},
    {8242152, 8253281, 8251362, 8272309, 8245296, 8270265, 8251362, 8253281},
    {8273856, 8254738, 8265696, 8269356, 8269248, 8269356, 8265696, 8254738},
    {8236432, 8264375, 8249646, 8253024, 8236432, 8253024, 8249646, 8264912},
    },
    {
      {10396800, 10418624, 10382496, 10378656, 10396800, 10378656, 10382496, 10418624},
      {10405312, 10393705, 10392304, 10400801, 10405312, 10400176, 10392304, 10394841},
      {10418624, 10389936, 10421492, 10410708, 10418624, 10408920, 10421492, 10389936},
      {10388648, 10419016, 10418376, 10411477, 10389792, 10409073, 10418376, 10419016},
      {10382496, 10421492, 10429536, 10398984, 10382496, 10398984, 10429536, 10421492},
      {10393440, 10416104, 10417804, 10410965, 10393440, 10406552, 10417804, 10416104},
      {10378656, 10405152, 10398984, 10404694, 10378656, 10404428, 10398984, 10408052},
      {10391600, 10411119, 10379340, 10393891, 10391600, 10393502, 10379340, 10411718},
      {10396800, 10418624, 10382496, 10378656, 10396800, 10378656, 10382496, 10418624},
      {10391600, 10411119, 10379340, 10393891, 10391600, 10393502, 10379340, 10411718},
      {10378656, 10405152, 10398984, 10404694, 10378656, 10404428, 10398984, 10408052},
      {10393440, 10416104, 10417804, 10410507, 10393440, 10406047, 10417804, 10416104},
      {10382496, 10421492, 10429536, 10398984, 10382496, 10398984, 10429536, 10421492},
      {10390936, 10419016, 10418376, 10411506, 10389792, 10409182, 10418376, 10419016},
      {10418624, 10389936, 10421492, 10410708, 10418624, 10408920, 10421492, 10389936},
      {10405312, 10393705, 10392304, 10400801, 10405312, 10400176, 10392304, 10394841},
    },
    {
      {13107200, 13100768, 13095936, 13106016, 13107200, 13106016, 13095936, 13100768},
      {13099736, 13097834, 13120122, 13108461, 13100236, 13107559, 13120122, 13098935},
      {13100768, 13113980, 13111280, 13108290, 13100768, 13106864, 13111280, 13114438},
      {13127840, 13119844, 13124056, 13112751, 13127840, 13111737, 13124056, 13120870},
      {13095936, 13111280, 13124160, 13081800, 13095936, 13079108, 13124160, 13111280},
      {13137384, 13117503, 13129060, 13110535, 13137384, 13110076, 13129060, 13119444},
      {13106016, 13104094, 13081800, 13111772, 13106016, 13111772, 13081800, 13106168},
      {13123616, 13093474, 13131744, 13099860, 13123616, 13099860, 13131744, 13094530},
      {13107200, 13100768, 13095936, 13106016, 13107200, 13106016, 13095936, 13100768},
      {13123616, 13093474, 13131744, 13099860, 13123616, 13099860, 13131744, 13094530},
      {13106016, 13104094, 13081800, 13111772, 13106016, 13111772, 13081800, 13106168},
      {13137384, 13119112, 13129060, 13111589, 13137384, 13110435, 13129060, 13120256},
      {13095936, 13111280, 13124160, 13081800, 13095936, 13079108, 13124160, 13111280},
      {13127840, 13120058, 13124056, 13113994, 13127840, 13112980, 13124056, 13120737},
      {13100768, 13113980, 13111280, 13108290, 13100768, 13106864, 13111280, 13114438},
      {13099736, 13097834, 13120122, 13108461, 13100236, 13107559, 13120122, 13098935},
      },
      {
        {16496768, 16515648, 16471904, 16515856, 16496768, 16515856, 16471904, 16515648},
        {16526776, 16511134, 16538304, 16509850, 16530708, 16509850, 16538304, 16511134},
        {16515648, 16532462, 16521956, 16511402, 16515648, 16511402, 16521956, 16532462},
        {16505136, 16526111, 16503016, 16504325, 16505136, 16504325, 16503016, 16527669},
        {16471904, 16521956, 16495200, 16488080, 16471904, 16488080, 16495200, 16521956},
        {16503336, 16527690, 16509440, 16504494, 16502172, 16504494, 16509440, 16531929},
        {16515856, 16511402, 16488080, 16506552, 16515856, 16506552, 16488080, 16511402},
        {16484888, 16527938, 16506906, 16511197, 16486012, 16511197, 16506906, 16528572},
        {16496768, 16515648, 16471904, 16515856, 16496768, 16515856, 16471904, 16515648},
        {16487136, 16527938, 16506906, 16511197, 16486012, 16511197, 16506906, 16528572},
        {16515856, 16511402, 16488080, 16506552, 16515856, 16506552, 16488080, 16511402},
        {16501008, 16528665, 16509440, 16504494, 16502172, 16504494, 16509440, 16531495},
        {16471904, 16521956, 16495200, 16488080, 16471904, 16488080, 16495200, 16521956},
        {16505136, 16525663, 16503016, 16504325, 16505136, 16504325, 16503016, 16527669},
        {16515648, 16532462, 16521956, 16511402, 16515648, 16511402, 16521956, 16532462},
        {16526776, 16511134, 16538304, 16509850, 16530708, 16509850, 16538304, 16511134},
      },
      {
        {20788352, 20757312, 20854016, 20812784, 20788352, 20812784, 20854016, 20757312},
        {20769328, 20799857, 20810084, 20810797, 20769328, 20809892, 20810084, 20800624},
        {20757312, 20806386, 20801280, 20813910, 20757312, 20813910, 20801280, 20809504},
        {20785152, 20809200, 20812720, 20798718, 20785152, 20797612, 20812720, 20810176},
        {20854016, 20801280, 20797248, 20817152, 20854016, 20817152, 20797248, 20801280},
        {20790920, 20809862, 20822060, 20807535, 20790920, 20807535, 20822060, 20811064},
        {20812784, 20813910, 20811292, 20796264, 20812784, 20796264, 20811292, 20816524},
        {20792496, 20782936, 20777424, 20795339, 20792496, 20793881, 20777424, 20788202},
        {20788352, 20757312, 20854016, 20812784, 20788352, 20812784, 20854016, 20757312},
        {20792496, 20782936, 20777424, 20795339, 20792496, 20793881, 20777424, 20788202},
        {20812784, 20813910, 20811292, 20796264, 20812784, 20796264, 20811292, 20816524},
        {20790920, 20810249, 20822060, 20808534, 20790920, 20808534, 20822060, 20811451},
        {20854016, 20801280, 20797248, 20817152, 20854016, 20817152, 20797248, 20801280},
        {20785152, 20804976, 20812720, 20800827, 20785152, 20799397, 20812720, 20809369},
        {20757312, 20806386, 20801280, 20813910, 20757312, 20813910, 20801280, 20809504},
        {20769328, 20799857, 20810084, 20810797, 20769328, 20809892, 20810084, 20800624},
        }
};

#endif // RDO_Q

static void dct4_II_fw_LLM_int(int *x)
{
  int n=4, n1=2;
  int i;
  int *y;

  if ((y = (int* ) calloc(n, sizeof(int))) == NULL)
    no_mem_exit("get_mem1Dint: y");

  // First butterfly
  for (i=0; i < n1; i++) {
    y[i] = x[i] + x[n-1-i];
    y[n-1-i] = x[i] - x[n-1-i];
  }

  // Process even part of the transform

  x[0] = y[0] + y[1];
  x[2] = y[0] - y[1];

  // Process even part of the transform

  x[1] = (A4_II * y[2] + B4_II * y[3]) >> shift4_II;
  x[3] = (A4_II * y[3] - B4_II * y[2]) >> shift4_II;

  // Free memory
  free(y);
}

static void dct4_II_bw_LLM_int(int *x)
{
  int n=4;
  int *y;

  if ((y = (int* ) calloc(n, sizeof(int))) == NULL)
    no_mem_exit("get_mem1Dint: y");

  // Process even part of the transform 

  y[0] = x[0] + x[2];
  y[1] = x[0] - x[2];

  // Process odd part of the tranform
  y[2] = (A4_II * x[1] - B4_II * x[3]) >> shift4_II;
  y[3] = (B4_II * x[1] + A4_II * x[3]) >> shift4_II;

  // Big butterfly
  x[0] = y[0] + y[3];
  x[1] = y[1] + y[2];
  x[2] = y[1] - y[2];
  x[3] = y[0] - y[3];

  // Free memory
  free(y);
}

static void dct8_II_fw_LLM_int(int *x)
{
  int n=8, n1=4, n2=2;
  int i, temp;
  int *y;

  if ((y = (int* ) calloc(n, sizeof(int))) == NULL)
    no_mem_exit("get_mem1Dint: y");

  // First butterfly
  for (i=0; i < n1; i++) {
    y[i] = x[i] + x[n-1-i];
    y[n-1-i] = x[i] - x[n-1-i];
  }

  // Process even part of the transform

  // 1st buttefly for even part
  for (i=0; i < n2; i++)
  {
    x[i] = y[i] + y[n1-1-i];
    x[n1-1-i] = y[i] - y[n1-1-i];
  }

  // 2nd butterfly for even part
  y[0] = x[0] + x[1];
  y[1] = x[0] - x[1];

  y[2] = (A8_II * x[2] + B8_II * x[3]) >> shift4_II;
  y[3] = (A8_II * x[3] - B8_II * x[2]) >> shift4_II;

  // Process odd part of the transform

  // First stage for odd part of the transform
  x[4] = (E8_II * y[4] + Z8_II * y[7]) >> shift8_II;
  x[7] = (E8_II * y[7] - Z8_II * y[4]) >> shift8_II;

  x[5] = (G8_II * y[5] + D8_II * y[6]) >> shift8_II;
  x[6] = (G8_II * y[6] - D8_II * y[5]) >> shift8_II;

  // Second stage for odd part of the transform
  y[4] = x[4] + x[6];
  y[5] = -x[5] + x[7];
  y[6] = -x[6] + x[4];
  y[7] = x[5] + x[7];

  // Third stage for odd part of the transform
  temp = y[7] - y[4];
  y[7] = y[7] + y[4];
  y[4] = temp;

  // Permutation
  x[0] = y[0];
  x[1] = y[7];
  x[2] = y[2];
  x[3] = y[5];
  x[4] = y[1];
  x[5] = y[6];
  x[6] = y[3];
  x[7] = y[4];

  // Free memory
  free(y);
}

static void dct8_IV_fw_LLM_int(int *x)
{
  int n=8, n1=4;
  int i;
  int *y;

  if ((y = (int* ) calloc(n, sizeof(int))) == NULL)
    no_mem_exit("get_mem1Dint: y");

  // Givens rotation

  y[0] = (A8_IV * x[0] + B8_IV * x[7]) >> shift8_IV;
  y[1] = (C8_IV * x[1] + D8_IV * x[6]) >> shift8_IV;
  y[2] = (E8_IV * x[2] + F8_IV * x[5]) >> shift8_IV;
  y[3] = (G8_IV * x[3] + H8_IV * x[4]) >> shift8_IV;

  y[4] = (-H8_IV * x[3] + G8_IV * x[4]) >> shift8_IV;
  y[5] = (-F8_IV * x[2] + E8_IV * x[5]) >> shift8_IV;
  y[6] = (-D8_IV * x[1] + C8_IV * x[6]) >> shift8_IV;
  y[7] = (-B8_IV * x[0] + A8_IV * x[7]) >> shift8_IV;

  // Sign change for odd term lower half
  y[5] = -y[5];
  y[7] = -y[7];

  // Calculate 2 4-point type II DCTs

  dct4_II_fw_LLM_int(y);
  dct4_II_fw_LLM_int(y+4);

  // Order reversal followed by sign alteration for the lower half

  for (i=0; i < n1; i++)
  {
    x[i] = y[i];
  }
  x[4] = y[7];
  x[5] = -y[6];
  x[6] = y[5];
  x[7] = -y[4];

  //

  y[0] = x[0];
  y[1] = x[1] + x[4];
  y[2] = x[2] + x[5];
  y[3] = x[3] + x[6];
  y[4] = x[1] - x[4];
  y[5] = x[2] - x[5];
  y[6] = x[3] - x[6];
  y[7] = -x[7];

  // Transpose of permutation matrix

  x[0] = y[0];
  x[1] = y[4];
  x[2] = y[1];
  x[3] = y[5];
  x[4] = y[2];
  x[5] = y[6];
  x[6] = y[3];
  x[7] = y[7];

  // Free memory
  free(y);
}

void dct8_II_bw_LLM_int(int *x)
{
  // Assumes that the input vector has already been pre-scaled by
  // appropriate factors.

  int n=8, n1=4;
  int i, temp;
  int *y;

  if ((y = (int* ) calloc(n, sizeof(int))) == NULL)
    no_mem_exit("get_mem1Dint: y");

  // Input reordering
  y[0] = x[0];
  y[1] = x[4];
  y[2] = x[2];
  y[3] = x[6];
  y[4] = x[7];
  y[5] = x[3];
  y[6] = x[5];
  y[7] = x[1];

  // Process even part of the transform (top 2)

  // stage 1
  x[0] = y[0] + y[1];
  x[1] = y[0] - y[1];

  x[2] = (A8_II * y[2] - B8_II * y[3]) >> shift4_II;
  x[3] = (B8_II * y[2] + A8_II * y[3]) >> shift4_II;

  // stage 2
  y[0] = x[0] + x[3];
  y[1] = x[1] + x[2];
  y[2] = x[1] - x[2];
  y[3] = x[0] - x[3];

  // Process odd part of the transform (bottom 4)

  // stage 1
  temp = -y[4] + y[7];
  y[7] =  y[4] + y[7];
  y[4] =  temp;

  // stage 2
  x[4] =  y[4] + y[6];
  x[5] = -y[5] + y[7];
  x[6] =  y[4] - y[6];
  x[7] =  y[5] + y[7];

  // stage 3
  y[4] = (E8_II * x[4] - Z8_II * x[7]) >> shift8_II;
  y[5] = (G8_II * x[5] - D8_II * x[6]) >> shift8_II;
  y[6] = (D8_II * x[5] + G8_II * x[6]) >> shift8_II;
  y[7] = (Z8_II * x[4] + E8_II * x[7]) >> shift8_II;

  // Big butterfly

  for (i=0; i < n1; i++) {
    x[i] = y[i] + y[n-1-i];
    x[n-1-i] = y[i] - y[n-1-i];
  }

  // Free memory
  free(y);
}

static void dct8_IV_bw_LLM_int(int *x)
{
  int n=8, n1=4;
  int i;
  int *y;

  if ((y = (int* ) calloc(n, sizeof(int))) == NULL)
    no_mem_exit("get_mem1Dint: y");

  // permutation
  y[0] = x[0];
  y[1] = x[2];
  y[2] = x[4];
  y[3] = x[6];
  y[4] = x[1];
  y[5] = x[3];
  y[6] = x[5];
  y[7] = x[7];

  //
  x[0] = y[0];
  x[1] = y[1] + y[4];
  x[4] = y[1] - y[4];
  x[2] = y[2] + y[5];
  x[5] = y[2] - y[5];
  x[3] = y[3] + y[6];
  x[6] = y[3] - y[6];
  x[7] = -y[7];

  // Order reversal followed by sign alteration for the lower half
  for (i=0; i < n1; i++)
  {
    y[i] = x[i];
  }
  y[4] = -x[7];
  y[5] =  x[6];
  y[6] = -x[5];
  y[7] =  x[4];

  // 2 Inverse 4-point DCTs of type 2 
  dct4_II_bw_LLM_int(y);
  dct4_II_bw_LLM_int(y+4);

  // Sign change in the lower half
  y[5] = -y[5];
  y[7] = -y[7];

  // Inverse of Givens rotation. 

  x[0] = (A8_IV * y[0] - B8_IV * y[7]) >> shift8_IV;
  x[1] = (C8_IV * y[1] - D8_IV * y[6]) >> shift8_IV;
  x[2] = (E8_IV * y[2] - F8_IV * y[5]) >> shift8_IV;
  x[3] = (G8_IV * y[3] - H8_IV * y[4]) >> shift8_IV;

  x[4] = (H8_IV * y[3] + G8_IV * y[4]) >> shift8_IV;
  x[5] = (F8_IV * y[2] + E8_IV * y[5]) >> shift8_IV;
  x[6] = (D8_IV * y[1] + C8_IV * y[6]) >> shift8_IV;
  x[7] = (B8_IV * y[0] + A8_IV * y[7]) >> shift8_IV;

  free(y);
}

/*
* This implements 16-point forward DCT in integer precision (Type II).
* This is the C_II_16 matrix. So the 0th coefficient should be further
* scaled by (1/SQRT(2)). Also each term should be scaled by (SQRT(1/8))
* to obtain a orthonormal decomposition.
*
* It decomposes the 16 point into type II and type IV. So the normalization
* factors for the butterflies have also to be accounted for.
*/

static void dct16_II_fw_LLM_int(int *x)
{
  int n=16, n1=8;
  int i;
  int *y;

  if ((y = (int* ) calloc(n, sizeof(int))) == NULL)
    no_mem_exit("get_mem1Dint: y");

  // First butterfly
  for (i=0; i < n1; i++) {
    y[i] = x[i] + x[n-1-i];
    y[i+n1] = x[i] - x[n-1-i];
  }

  // Calculate Type II and Type IV DCTs for upper and lower half, respectively.

  // dct8_II implementation is the same as C_8^II except that the 0th term for dct8_II
  // is scaled by (1/SQRT(2)). So to get to C_8^II, the 0th term of dct8_II should be
  // scaled up by SQRT(2).
  dct8_II_fw_LLM_int(y);
  dct8_IV_fw_LLM_int(y+8);

  // Permutation
  for (i=0; i < n1; i++) {
    x[2*i] = y[i];
    x[2*i+1] = y[i+n1];
  }

  free(y);
}

/*
* This implements 16-point inverse DCT (Type II) in int.
* 
*/

static void dct16_II_bw_LLM_int(int *x)
{
  int n=16, n1=8;
  int i;
  int *y;

  if ((y = (int* ) calloc(n, sizeof(int))) == NULL)
    no_mem_exit("get_mem1Dint: y");

  // Permutation
  for (i=0; i < n1; i++) {
    y[i] = x[2*i];
    y[i+n1] = x[2*i+1];
  }

  // Calculate Type II and Type IV inverse DCTs for upper and lower half, respectively.

  dct8_II_bw_LLM_int(y);
  dct8_IV_bw_LLM_int(y+8);

  // Big butterfly

  for (i=0; i < n1; i++) {
    x[i] = y[i] + y[i+n1];
    x[n-1-i] = y[i] - y[i+n1];
  }

  // Free memory
  free(y);
}

/*
* This implements 16x16 2D DCT (Type II) in int.
* 
*/

static void dct_2D_16x16_II_fw_int(int (*x)[16], int **y)
{
  int i, j, m, n;

  int num_rows = 16;
  int num_cols = 16;

  int *tmpI;

  if ((tmpI = (int* ) calloc(num_rows, sizeof(int))) == NULL)
    no_mem_exit("get_mem1Dint: tmpI");

  // Pre-shift left by DCT16_2D_UP_SHIFT
  for (m=0; m < num_rows; m++)
  {
    for (n=0; n < num_cols; n++)
    {
      x[m][n] = x[m][n] << DCT16_2D_fw_UP_SHIFT;
    }
  }

  // Rows first
  for (i=0; i < num_rows; i++)
  {
    memcpy(y[i], x[i], num_cols*sizeof(int));
    dct16_II_fw_LLM_int(y[i]);
  }

  // Columns next
  for (j=0; j < num_cols; j++)
  {
    for (i=0; i < num_rows; i++)
    {
      tmpI[i] = y[i][j];
    }
    dct16_II_fw_LLM_int(tmpI);
    for (i=0; i < num_rows; i++)
    {
      y[i][j] = tmpI[i];
    }
  }

  // Post-shift right by DCT16_2D_UP_SHIFT
  for (m=0; m < num_rows; m++)
  {
    for (n=0; n < num_cols; n++)
    {
      y[m][n] = (y[m][n] + (1 << (DCT16_2D_fw_UP_SHIFT-1))) >> DCT16_2D_fw_UP_SHIFT;
    }
  }

  free(tmpI);
}

/*
* This implements 16x16 inverse 2D DCT (Type II) in int.
* 
*/

static void dct_2D_16x16_II_bw_int(int **x, int (*y)[16])
{
  int i, j;

  int num_rows = 16;
  int num_cols = 16;

  int *tmpI;
  if ((tmpI = (int* ) calloc(num_rows, sizeof(int))) == NULL)
    no_mem_exit("get_mem1Dint: tmpI");

  // Rows first
  for (i=0; i < num_rows; i++)
  {
    memcpy(y[i], x[i], num_cols*sizeof(int));
    dct16_II_bw_LLM_int(y[i]);
  }

  // Columns next
  for (j=0; j < num_cols; j++)
  {
    for (i=0; i < num_rows; i++)
    {
      tmpI[i] = y[i][j];
    }
    dct16_II_bw_LLM_int(tmpI);
    for (i=0; i < num_rows; i++)
    {
      y[i][j] = tmpI[i];
    }
  }

  free(tmpI);
}

/*
* This implements 16x8 (width x height) 2D DCT (Type II) in int.
* 
*/

void dct_2D_16x8_II_fw_int(int (*x)[16], int **y)
{
  int i, j, m, n;

  int num_rows = 8;
  int num_cols = 16;

  int *tmpI;
  if ((tmpI = (int* ) calloc(num_rows, sizeof(int))) == NULL)
    no_mem_exit("get_mem1Dint: tmpI");

  // Pre-shift left by DCT16_2D_UP_SHIFT
  for (m=0; m < num_rows; m++)
  {
    for (n=0; n < num_cols; n++)
    {
      x[m][n] = x[m][n] << DCT16_2D_fw_UP_SHIFT;
    }
  }

  // Rows first
  for (i=0; i < num_rows; i++)
  {
    memcpy(y[i], x[i], num_cols*sizeof(int));
    dct16_II_fw_LLM_int(y[i]);
  }

  // Columns next
  for (j=0; j < num_cols; j++)
  {
    for (i=0; i < num_rows; i++)
    {
      tmpI[i] = y[i][j];
    }
    dct8_II_fw_LLM_int(tmpI);
    for (i=0; i < num_rows; i++)
    {
      y[i][j] = tmpI[i];
    }
  }

  // Post-shift right by DCT16_2D_UP_SHIFT
  for (m=0; m < num_rows; m++)
  {
    for (n=0; n < num_cols; n++)
    {
      y[m][n] = (y[m][n] + (1 << (DCT16_2D_fw_UP_SHIFT-1))) >> DCT16_2D_fw_UP_SHIFT;
    }
  }

  free(tmpI);
}

/*
* This implements 16x8 (width x height) inverse 2D DCT (Type II) in int.
* 
*/

void dct_2D_16x8_II_bw_int(int **x, int (*y)[16])
{
  int i, j;

  int num_rows = 8;
  int num_cols = 16;

  int *tmpI;
  if ((tmpI = (int* ) calloc(num_rows, sizeof(int))) == NULL)
    no_mem_exit("get_mem1Dint: tmpI");

  // Rows first
  for (i=0; i < num_rows; i++)
  {
    memcpy(y[i], x[i], num_cols*sizeof(int));
    dct16_II_bw_LLM_int(y[i]);
  }

  // Columns next
  for (j=0; j < num_cols; j++)
  {
    for (i=0; i < num_rows; i++)
    {
      tmpI[i] = y[i][j];
    }
    dct8_II_bw_LLM_int(tmpI);
    for (i=0; i < num_rows; i++)
    {
      y[i][j] = tmpI[i];
    }
  }

  free(tmpI);
}

/*
* This implements 8x16 (width x height) 2D DCT (Type II) in int.
* 
*/

void dct_2D_8x16_II_fw_int(int (*x)[16], int **y)
{
  int i, j, m, n;

  int num_rows = 16;
  int num_cols = 8;

  int *tmpI;
  if ((tmpI = (int* ) calloc(num_rows, sizeof(int))) == NULL)
    no_mem_exit("get_mem1Dint: tmpI");

  // Pre-shift left by DCT16_2D_UP_SHIFT
  for (m=0; m < num_rows; m++)
  {
    for (n=0; n < num_cols; n++)
    {
      x[m][n] = x[m][n] << DCT16_2D_fw_UP_SHIFT;
    }
  }

  // Rows first
  for (i=0; i < num_rows; i++)
  {
    memcpy(y[i], x[i], num_cols*sizeof(int));
    dct8_II_fw_LLM_int(y[i]);
  }

  // Columns next
  for (j=0; j < num_cols; j++)
  {
    for (i=0; i < num_rows; i++)
    {
      tmpI[i] = y[i][j];
    }
    dct16_II_fw_LLM_int(tmpI);
    for (i=0; i < num_rows; i++)
    {
      y[i][j] = tmpI[i];
    }
  }

  // Post-shift right by DCT16_2D_UP_SHIFT
  for (m=0; m < num_rows; m++)
  {
    for (n=0; n < num_cols; n++)
    {
      y[m][n] = (y[m][n] + (1 << (DCT16_2D_fw_UP_SHIFT-1))) >> DCT16_2D_fw_UP_SHIFT;
    }
  }

  free(tmpI);
}

/*
* This implements 8x16 (width x height) 2D inverse DCT (Type II) in int.
* 
*/

void dct_2D_8x16_II_bw_int(int **x, int (*y)[16])
{
  int i, j;

  int num_rows = 16;
  int num_cols = 8;

  int *tmpI;
  if ((tmpI = (int* ) calloc(num_rows, sizeof(int))) == NULL)
    no_mem_exit("get_mem1Dint: tmpI");

  // Rows first
  for (i=0; i < num_rows; i++)
  {
    memcpy(y[i], x[i], num_cols*sizeof(int));
    dct8_II_bw_LLM_int(y[i]);
  }

  // Columns next
  for (j=0; j < num_cols; j++)
  {
    for (i=0; i < num_rows; i++)
    {
      tmpI[i] = y[i][j];
    }
    dct16_II_bw_LLM_int(tmpI);
    for (i=0; i < num_rows; i++)
    {
      y[i][j] = tmpI[i];
    }
  }

  free(tmpI);
}



const byte SNGL_SCAN16x16[256][2] = 
{
  { 0, 0}, { 1, 0}, { 0, 1}, { 0, 2}, { 1, 1}, { 2, 0}, { 3, 0}, { 2, 1}, { 1, 2}, { 0, 3}, { 0, 4}, { 1, 3}, { 2, 2}, { 3, 1}, { 4, 0}, { 5, 0},
  { 4, 1}, { 3, 2}, { 2, 3}, { 1, 4}, { 0, 5}, { 0, 6}, { 1, 5}, { 2, 4}, { 3, 3}, { 4, 2}, { 5, 1}, { 6, 0}, { 7, 0}, { 6, 1}, { 5, 2}, { 4, 3},
  { 3, 4}, { 2, 5}, { 1, 6}, { 0, 7}, { 0, 8}, { 1, 7}, { 2, 6}, { 3, 5}, { 4, 4}, { 5, 3}, { 6, 2}, { 7, 1}, { 8, 0}, { 9, 0}, { 8, 1}, { 7, 2}, 
  { 6, 3}, { 5, 4}, { 4, 5}, { 3, 6}, { 2, 7}, { 1, 8}, { 0, 9}, { 0,10}, { 1, 9}, { 2, 8}, { 3, 7}, { 4, 6}, { 5, 5}, { 6, 4}, { 7, 3}, { 8, 2},
  { 9, 1}, {10, 0}, {11, 0}, {10, 1}, { 9, 2}, { 8, 3}, { 7, 4}, { 6, 5}, { 5, 6}, { 4, 7}, { 3, 8}, { 2, 9}, { 1,10}, { 0,11}, { 0,12}, { 1,11},
  { 2,10}, { 3, 9}, { 4, 8}, { 5, 7}, { 6, 6}, { 7, 5}, { 8, 4}, { 9, 3}, {10, 2}, {11, 1}, {12, 0}, {13, 0}, {12, 1}, {11, 2}, {10, 3}, { 9, 4},
  { 8, 5}, { 7, 6}, { 6, 7}, { 5, 8}, { 4, 9}, { 3,10}, { 2,11}, { 1,12}, { 0,13}, { 0,14}, { 1,13}, { 2,12}, { 3,11}, { 4,10}, { 5, 9}, { 6, 8},
  { 7, 7}, { 8, 6}, { 9, 5}, {10, 4}, {11, 3}, {12, 2}, {13, 1}, {14, 0}, {15, 0}, {14, 1}, {13, 2}, {12, 3}, {11, 4}, {10, 5}, { 9, 6}, { 8, 7},
  { 7, 8}, { 6, 9}, { 5,10}, { 4,11}, { 3,12}, { 2,13}, { 1,14}, { 0,15}, { 1,15}, { 2,14}, { 3,13}, { 4,12}, { 5,11}, { 6,10}, { 7, 9}, { 8, 8},
  { 9, 7}, {10, 6}, {11, 5}, {12, 4}, {13, 3}, {14, 2}, {15, 1}, {15, 2}, {14, 3}, {13, 4}, {12, 5}, {11, 6}, {10, 7}, { 9, 8}, { 8, 9}, { 7,10},
  { 6,11}, { 5,12}, { 4,13}, { 3,14}, { 2,15}, { 3,15}, { 4,14}, { 5,13}, { 6,12}, { 7,11}, { 8,10}, { 9, 9}, {10, 8}, {11, 7}, {12, 6}, {13, 5},
  {14, 4}, {15, 3}, {15, 4}, {14, 5}, {13, 6}, {12, 7}, {11, 8}, {10, 9}, { 9,10}, { 8,11}, { 7,12}, { 6,13}, { 5,14}, { 4,15}, { 5,15}, { 6,14},
  { 7,13}, { 8,12}, { 9,11}, {10,10}, {11, 9}, {12, 8}, {13, 7}, {14, 6}, {15, 5}, {15, 6}, {14, 7}, {13, 8}, {12, 9}, {11,10}, {10,11}, { 9,12},
  { 8,13}, { 7,14}, { 6,15}, { 7,15}, { 8,14}, { 9,13}, {10,12}, {11,11}, {12,10}, {13, 9}, {14, 8}, {15, 7}, {15, 8}, {14, 9}, {13,10}, {12,11},
  {11,12}, {10,13}, { 9,14}, { 8,15}, { 9,15}, {10,14}, {11,13}, {12,12}, {13,11}, {14,10}, {15, 9}, {15,10}, {14,11}, {13,12}, {12,13}, {11,14},
  {10,15}, {11,15}, {12,14}, {13,13}, {14,12}, {15,11}, {15,12}, {14,13}, {13,14}, {12,15}, {13,15}, {14,14}, {15,13}, {15,14}, {14,15}, {15,15}
};


int dct_luma_16x16_new (void)
{
  int nonzero = FALSE;  
  int i,j,coeff_ctr;
  int scan_pos,run;
  int*  ACLevel = img->cofAC16x16[0];
  int*  ACRun   = img->cofAC16x16[1];

  imgpel **img_enc      = enc_picture->imgY;
  int    (*curr_res)[MB_BLOCK_SIZE] = img->m7; 

  const byte (*pos_scan)[2] = SNGL_SCAN16x16;
  int qp_per    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)/6; 
  int qp_rem    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)%6; 

  int qpShift, qpRound;
  int level;
  int **M1;

  int k;
  levelDataStruct levelData[256];
  double lambda_md = 0.0;
  double err;
  int lowerInt;
  int levelRDOQ[256];
  int kStart = 0;
  int kStop = 0;
  int noCoeff;

#ifdef INTERNAL_BIT_DEPTH_INCREASE
  double normFact = pow(2, 2*(Q_BITS_16 + DQ_BITS_16)-15) * (1<<(2*img->BitDepthIncrease));
#else
  double normFact = pow(2, 2*(Q_BITS_16 + DQ_BITS_16)-15); 
#endif

  get_mem2Dint(&M1,16,16);

  //  Forward 16x16 transform
  dct_2D_16x16_II_fw_int(curr_res, M1);

  // Quantization process
  qpShift = qp_per + Q_BITS;
  qpRound = (1<<(qpShift))/6; // Only inter block currently get encoded with this mode


  // quantization and zig-zag scanning 
  if(input->UseRDO_Q)
  {
    if ((img->type == B_SLICE) && img->nal_reference_idc)
    {
      lambda_md = img->lambda_md[5][img->masterQP];  
    }
    else
    {
      lambda_md = img->lambda_md[img->type][img->masterQP];
    }
    noCoeff = 0;
    for (coeff_ctr = 0; coeff_ctr < 256; coeff_ctr++)
    {
      i = pos_scan[coeff_ctr][0];
      j = pos_scan[coeff_ctr][1];

      levelData[coeff_ctr].levelDouble = absm(M1[j][i]) * quant_coef_16x16[qp_rem][j][i];
      level = (int)(levelData[coeff_ctr].levelDouble >> qpShift);
      lowerInt=(((int)levelData[coeff_ctr].levelDouble - (level << qpShift)) < (1 << (qpShift-1)))? 1 : 0;

      levelData[coeff_ctr].level[0] = 0;
      if (level == 0 && lowerInt == 1)
      {
        levelData[coeff_ctr].noLevels = 1;
      }
      else if (level == 0 && lowerInt == 0)
      {
        levelData[coeff_ctr].level[1] = level + 1;
        levelData[coeff_ctr].noLevels = 2;
        kStop = coeff_ctr;
        noCoeff++;
      }
      else if (level > 0 && lowerInt == 1)
      {
        levelData[coeff_ctr].level[1] = level;
        levelData[coeff_ctr].noLevels = 2;
        kStop = coeff_ctr;
        noCoeff++;
      }
      else
      {
        levelData[coeff_ctr].level[1] = level;
        levelData[coeff_ctr].level[2] = level + 1;
        levelData[coeff_ctr].noLevels = 3;
        kStop = coeff_ctr;
        kStart = coeff_ctr;
        noCoeff++;
      }

      for (k = 0; k < levelData[coeff_ctr].noLevels; k++)
      {      
        err = (double)(levelData[coeff_ctr].level[k] << qpShift) - (double) levelData[coeff_ctr].levelDouble;
        levelData[coeff_ctr].errLevel[k] = (err * err * (double) estErr16x16[qp_rem][i][j]) / normFact;
      }
    }

    est_writeRunLevel_CABAC(levelData, levelRDOQ, LUMA_16x16P, lambda_md, kStart, kStop, noCoeff, 0);
  }

  nonzero=FALSE;
  run=-1;
  scan_pos=0;

  for (coeff_ctr=0;coeff_ctr < 256;coeff_ctr++)
  {
    i=pos_scan[coeff_ctr][0];
    j=pos_scan[coeff_ctr][1];

    run++;

    if(input->UseRDO_Q)
    {
      level = levelRDOQ[coeff_ctr];
    }
    else
    {
      level = (absm(M1[j][i]) * quant_coef_16x16[qp_rem][j][i]) + qpRound;
      level = level >>(qpShift);
    }

    if (level != 0)
    {
      nonzero=TRUE;

      ACLevel[scan_pos] = sign(level, M1[j][i]);
      ACRun  [scan_pos] = run;
      ++scan_pos;
      run=-1;                     // reset zero level counter

      M1[j][i] = sign(level, M1[j][i]) * dequant_coef_16x16[qp_rem][j][i] * (1<<qp_per); 
    }
    else 
      M1[j][i]=0;
  }
  ACLevel[scan_pos] = 0;

  if (nonzero)
  {
    int shift, shiftBy2;
    // Inverse 16x16 transform
    dct_2D_16x16_II_bw_int(M1, curr_res);

    // store the encoded image
    shift=DQ_BITS_16; 
    shiftBy2=1<<(shift-1);
    for (j=0; j < MB_BLOCK_SIZE; j++)
    {
      for (i=0; i < MB_BLOCK_SIZE; i++)
      {
        short temp;
        temp=(short)((curr_res[j][i]+((int)img->mpr[j][i]<<shift)+shiftBy2)>>shift);
        temp=min(img->max_imgpel_value,max(0,temp));
        img_enc[img->pix_y+j][img->pix_x+i]=temp;

        //printf("%d(%d) ", img->mpr[j][i], curr_res[j][i]);

      }
      //printf("\n");
    }
  }
  else // if (nonzero) => No transformed residual. Just use prediction.
  {
    for (j=0; j < MB_BLOCK_SIZE; j++)
    {
      memcpy(&(img_enc[img->pix_y + j][img->pix_x]),&(img->mpr[j][0]), MB_BLOCK_SIZE * sizeof(imgpel));
    }
  }

  free_mem2Dint(M1);

  return nonzero;
}

const byte SNGL_SCAN16x8[128][2] = 
{
  { 0, 0}, { 0, 1}, { 1, 0}, { 2, 0}, { 1, 1}, { 0, 2}, { 0, 3}, { 1, 2}, { 2, 1}, { 3, 0}, { 4, 0}, { 3, 1}, { 2, 2}, { 1, 3}, { 0, 4}, { 0, 5}, 
  { 1, 4}, { 2, 3}, { 3, 2}, { 4, 1}, { 5, 0}, { 6, 0}, { 5, 1}, { 4, 2}, { 3, 3}, { 2, 4}, { 1, 5}, { 0, 6}, { 0, 7}, { 1, 6}, { 2, 5}, { 3, 4}, 
  { 4, 3}, { 5, 2}, { 6, 1}, { 7, 0}, { 8, 0}, { 7, 1}, { 6, 2}, { 5, 3}, { 4, 4}, { 3, 5}, { 2, 6}, { 1, 7}, { 2, 7}, { 3, 6}, { 4, 5}, { 5, 4}, 
  { 6, 3}, { 7, 2}, { 8, 1}, { 9, 0}, {10, 0}, { 9, 1}, { 8, 2}, { 7, 3}, { 6, 4}, { 5, 5}, { 4, 6}, { 3, 7}, { 4, 7}, { 5, 6}, { 6, 5}, { 7, 4}, 
  { 8, 3}, { 9, 2}, {10, 1}, {11, 0}, {12, 0}, {11, 1}, {10, 2}, { 9, 3}, { 8, 4}, { 7, 5}, { 6, 6}, { 5, 7}, { 6, 7}, { 7, 6}, { 8, 5}, { 9, 4}, 
  {10, 3}, {11, 2}, {12, 1}, {13, 0}, {14, 0}, {13, 1}, {12, 2}, {11, 3}, {10, 4}, { 9, 5}, { 8, 6}, { 7, 7}, { 8, 7}, { 9, 6}, {10, 5}, {11, 4}, 
  {12, 3}, {13, 2}, {14, 1}, {15, 0}, {15, 1}, {14, 2}, {13, 3}, {12, 4}, {11, 5}, {10, 6}, { 9, 7}, {10, 7}, {11, 6}, {12, 5}, {13, 4}, {14, 3},
  {15, 2}, {15, 3}, {14, 4}, {13, 5}, {12, 6}, {11, 7}, {12, 7}, {13, 6}, {14, 5}, {15, 4}, {15, 5}, {14, 6}, {13, 7}, {14, 7}, {15, 6}, {15, 7}
};


int dct_luma_16x8 (int blk16x8)
{
  int nonzero = FALSE;  
  int i,j,coeff_ctr;
  int scan_pos,run;

  int*  ACLevel = img->cofAC16x8[blk16x8][0];
  int*  ACRun   = img->cofAC16x8[blk16x8][1];

  imgpel **img_enc      = enc_picture->imgY;
  int    (*curr_res)[MB_BLOCK_SIZE] = &img->m7 [blk16x8*MB_BLOCK_SIZE/2]; 

  const byte (*pos_scan)[2] = SNGL_SCAN16x8;
  int qp_per    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)/6; 
  int qp_rem    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)%6; 

  int qpShift, qpRound;
  int level;
  int **M1;

  int k;
  levelDataStruct levelData[256];
  double lambda_md = 0.0;
  double err;
  int lowerInt;
  int levelRDOQ[256];
  int kStart = 0;
  int kStop = 0;
  int noCoeff;

#ifdef INTERNAL_BIT_DEPTH_INCREASE
  double normFact = pow(2, 2*(Q_BITS_16 + DQ_BITS_16)-15) * (1<<(2*img->BitDepthIncrease));
#else
  double normFact = pow(2, 2*(Q_BITS_16 + DQ_BITS_16)-15); 
#endif

  get_mem2Dint(&M1,16,16);

  //  Forward 16x16 transform
  dct_2D_16x8_II_fw_int(curr_res, M1);

  // Quantization process
  qpShift = qp_per + Q_BITS;
  qpRound = (1<<(qpShift))/6; // Only inter block currently get encoded with this mode

  // quantization and zig-zag scanning 
  if(input->UseRDO_Q)
  {
    if ((img->type == B_SLICE) && img->nal_reference_idc)
    {
      lambda_md = img->lambda_md[5][img->masterQP];  
    }
    else
    {
      lambda_md = img->lambda_md[img->type][img->masterQP];
    }

    noCoeff = 0;
    for (coeff_ctr = 0; coeff_ctr < 128; coeff_ctr++)
    {
      i = pos_scan[coeff_ctr][0];
      j = pos_scan[coeff_ctr][1];

      levelData[coeff_ctr].levelDouble = absm(M1[j][i]) * quant_coef_16x8[qp_rem][j][i];
      level = (int) (levelData[coeff_ctr].levelDouble >> qpShift);
      lowerInt = ((levelData[coeff_ctr].levelDouble - (level << qpShift)) < ((int64)1 << (qpShift - 1)))? 1: 0;

      levelData[coeff_ctr].level[0] = 0;
      if (level == 0 && lowerInt == 1)
      {
        levelData[coeff_ctr].noLevels = 1;
      }
      else if (level == 0 && lowerInt == 0)
      {
        levelData[coeff_ctr].level[1] = level + 1;
        levelData[coeff_ctr].noLevels = 2;
        kStop = coeff_ctr;
        noCoeff++;
      }
      else if (level > 0 && lowerInt == 1)
      {
        levelData[coeff_ctr].level[1] = level;
        levelData[coeff_ctr].noLevels = 2;
        kStop = coeff_ctr;
        noCoeff++;
      }
      else
      {
        levelData[coeff_ctr].level[1] = level;
        levelData[coeff_ctr].level[2] = level + 1;
        levelData[coeff_ctr].noLevels = 3;
        kStop = coeff_ctr;
        kStart = coeff_ctr;
        noCoeff++;
      }

      for (k = 0; k < levelData[coeff_ctr].noLevels; k++)
      {      
        err = (double)(levelData[coeff_ctr].level[k] << qpShift) - (double) levelData[coeff_ctr].levelDouble;
        levelData[coeff_ctr].errLevel[k] = (err * err * (double) estErr16x8[qp_rem][j][i]) / normFact; 
      }
    }

    est_writeRunLevel_CABAC(levelData, levelRDOQ, LUMA_16x8P, lambda_md, kStart, kStop, noCoeff, 0);
  }

  nonzero=FALSE;
  run=-1;
  scan_pos=0;

  for (coeff_ctr=0;coeff_ctr < 128;coeff_ctr++)
  {
    i=pos_scan[coeff_ctr][0];
    j=pos_scan[coeff_ctr][1];

    run++;

    if(input->UseRDO_Q)
    {
      level = levelRDOQ[coeff_ctr];
    }
    else
    {
      level = (abs(M1[j][i]) * quant_coef_16x8[qp_rem][j][i]) + qpRound;
      level = level >>(qpShift);
    }

    if (level != 0)
    {
      nonzero=TRUE;

      ACLevel[scan_pos] = sign(level, M1[j][i]);
      ACRun  [scan_pos] = run;
      ++scan_pos;
      run=-1;                     // reset zero level counter

      M1[j][i] = sign(level, M1[j][i]) * dequant_coef_16x8[qp_rem][j][i] * (1 << qp_per); 
    }
    else 
      M1[j][i]=0;
  }
  ACLevel[scan_pos] = 0;

  if (nonzero)
  {
    int shift, shiftBy2;

    // Inverse 16x8 transform
    dct_2D_16x8_II_bw_int(M1, curr_res);

    // store the encoded image
    shift = DQ_BITS_16; 
    shiftBy2 = 1<<(shift-1);
    for (j=0; j < MB_BLOCK_SIZE/2; j++)
      for (i=0; i < MB_BLOCK_SIZE; i++)
      {
        short temp;

        temp=(short)((curr_res[j][i]+((int)img->mpr[blk16x8*MB_BLOCK_SIZE/2+j][i]<<shift)+shiftBy2)>>shift);
        temp=min(img->max_imgpel_value,max(0,temp));
        img_enc[img->pix_y+blk16x8*MB_BLOCK_SIZE/2+j][img->pix_x+i]=temp;
      }
  }
  else // if (nonzero) => No transformed residual. Just use prediction.
  {
    for (j=0; j < MB_BLOCK_SIZE/2; j++)
    {
      memcpy(&(img_enc[img->pix_y + blk16x8*MB_BLOCK_SIZE/2 + j][img->pix_x]),&(img->mpr[blk16x8*MB_BLOCK_SIZE/2+j][0]), MB_BLOCK_SIZE * sizeof(imgpel));
    }
  }

  free_mem2Dint(M1);

  return nonzero;
}

int dct_luma_8x16 (int blk8x16)
{
  int nonzero = FALSE;  
  int i,j,coeff_ctr;
  int scan_pos,run;

  int*  ACLevel = img->cofAC8x16[blk8x16][0];
  int*  ACRun   = img->cofAC8x16[blk8x16][1];

  imgpel **img_enc      = enc_picture->imgY;
  int    res[MB_BLOCK_SIZE][MB_BLOCK_SIZE];
  int    (*curr_res)[MB_BLOCK_SIZE] = res;

  const byte (*pos_scan)[2] = SNGL_SCAN16x8;
  int qp_per    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)/6; //CHA-VG01
  int qp_rem    = (img->qp + img->bitdepth_luma_qp_scale - MIN_QP)%6; //CHA-VG01

  int qpShift, qpRound;
  int shift, shiftBy2;
  int level;
  int **M1;

  int k;
  levelDataStruct levelData[256];
  double lambda_md = 0.0;
  double err;
  int lowerInt;
  int levelRDOQ[256];
  int kStart = 0;
  int kStop = 0;
  int noCoeff;

#ifdef INTERNAL_BIT_DEPTH_INCREASE
  double normFact = pow(2, 2*(Q_BITS_16 + DQ_BITS_16)-15) * (1<<(2*img->BitDepthIncrease));
#else
  double normFact = pow(2, 2*(Q_BITS_16 + DQ_BITS_16)-15); 
#endif

  //  Forward 8x16 transform
  for(j = 0; j < MB_BLOCK_SIZE; j++)
    for(i = 0; i < MB_BLOCK_SIZE/2; i++)
      curr_res[j][i] = img->m7[j][i+blk8x16*MB_BLOCK_SIZE/2];

  get_mem2Dint(&M1,16,16);

  //  Forward 16x16 transform
  dct_2D_8x16_II_fw_int(curr_res, M1);

  // Quantization process
  qpShift = qp_per + Q_BITS;
  qpRound = (1<<(qpShift))/6;

  // quantization and zig-zag scanning 
  if(input->UseRDO_Q)
  {
    if ((img->type == B_SLICE) && img->nal_reference_idc)
    {
      lambda_md = img->lambda_md[5][img->masterQP];  
    }
    else
    {
      lambda_md = img->lambda_md[img->type][img->masterQP];
    }

    noCoeff = 0;
    for (coeff_ctr = 0; coeff_ctr < 128; coeff_ctr++)
    {
      j = pos_scan[coeff_ctr][0];
      i = pos_scan[coeff_ctr][1];

      levelData[coeff_ctr].levelDouble = absm(M1[j][i]) * quant_coef_8x16[qp_rem][j][i];
      level = (int)(levelData[coeff_ctr].levelDouble >> qpShift);
      lowerInt = ((levelData[coeff_ctr].levelDouble - (level << qpShift)) < ((int64)1 << (qpShift - 1)))? 1: 0;

      levelData[coeff_ctr].level[0] = 0;
      if (level == 0 && lowerInt == 1)
      {
        levelData[coeff_ctr].noLevels = 1;
      }
      else if (level == 0 && lowerInt == 0)
      {
        levelData[coeff_ctr].level[1] = level + 1;
        levelData[coeff_ctr].noLevels = 2;
        kStop = coeff_ctr;
        noCoeff++;
      }
      else if (level > 0 && lowerInt == 1)
      {
        levelData[coeff_ctr].level[1] = level;
        levelData[coeff_ctr].noLevels = 2;
        kStop = coeff_ctr;
        noCoeff++;
      }
      else
      {
        levelData[coeff_ctr].level[1] = level;
        levelData[coeff_ctr].level[2] = level + 1;
        levelData[coeff_ctr].noLevels = 3;
        kStop = coeff_ctr;
        kStart = coeff_ctr;
        noCoeff++;
      }

      for (k = 0; k < levelData[coeff_ctr].noLevels; k++)
      {      
        err = (double)(levelData[coeff_ctr].level[k] << qpShift) - (double) levelData[coeff_ctr].levelDouble;
        levelData[coeff_ctr].errLevel[k] = (err * err * (double) estErr16x16[qp_rem][j][i]) / normFact; 
      }
    }

    est_writeRunLevel_CABAC(levelData, levelRDOQ, LUMA_8x16P, lambda_md, kStart, kStop, noCoeff, 0);
  }

  nonzero=FALSE;

  run=-1;
  scan_pos=0;

  for (coeff_ctr=0;coeff_ctr < 128;coeff_ctr++)
  {
    j=pos_scan[coeff_ctr][0];
    i=pos_scan[coeff_ctr][1];

    run++;

    if(input->UseRDO_Q)
    {
      level = levelRDOQ[coeff_ctr];
    }
    else
    {
      level = (abs(M1[j][i]) * quant_coef_8x16[qp_rem][j][i]) + qpRound;
      level = level >>(qpShift);
    }

    if (level != 0)
    {
      nonzero=TRUE;

      ACLevel[scan_pos] = sign(level, M1[j][i]);
      ACRun  [scan_pos] = run;
      ++scan_pos;
      run=-1;                     // reset zero level counter

      M1[j][i] = sign(level, M1[j][i]) * dequant_coef_8x16[qp_rem][j][i] * (1 << qp_per); 
    }
    else 
      M1[j][i]=0;
  }
  ACLevel[scan_pos] = 0;

  if (nonzero)
  {
    // Inverse 8x16 transform

    dct_2D_8x16_II_bw_int(M1, curr_res);

    // store the encoded image
    shift = DQ_BITS_16; 
    shiftBy2 = 1<<(shift-1);
    for (j=0; j < MB_BLOCK_SIZE; j++)
      for (i=0; i < MB_BLOCK_SIZE/2; i++)
      {
        short temp;

        temp=(short)((curr_res[j][i]+((int)img->mpr[j][i+blk8x16*MB_BLOCK_SIZE/2]<<shift)+shiftBy2)>>shift);
        temp=min(img->max_imgpel_value,max(0,temp));
        img_enc[img->pix_y+j][img->pix_x+blk8x16*MB_BLOCK_SIZE/2+i]=temp;
      }
  }
  else // if (nonzero) => No transformed residual. Just use prediction.
  {
    for (j=0; j < MB_BLOCK_SIZE; j++)
    {
      memcpy(&(img_enc[img->pix_y + j][img->pix_x+ blk8x16*MB_BLOCK_SIZE/2]),&(img->mpr[j][blk8x16*MB_BLOCK_SIZE/2]), MB_BLOCK_SIZE/2 * sizeof(imgpel));
    }
  }

  free_mem2Dint(M1);

  return nonzero;
}


